#**********************************************************************
#
#	$Id: Mrules.mk,v 1.1 2024/10/20 01:45:28 warme Exp $
#
#	File:	Mrules.mk
#	Rev:	a-2
#	Date:	10/19/2024
#
#	Copyright (c) 2012, 2024 by David M. Warme.  All rights
#	reserved.  This work is licensed under a Creative Commons
#	Attribution-NonCommercial-ShareAlike 4.0 International License.
#
#**********************************************************************
#
#	Global "make" include file for CALIB.  This file is included
#	in all other Makefiles in the build heirarchy, and defines
#	the various rules that we use globally.
#
#**********************************************************************
#
#	Modification Log:
#
#	a-1:	10/10/2012	warme
#		: Created.
#	a-2:	10/19/2024	warme
#		: Cleanup for initial release.
#
#**********************************************************************

# List of all our phoney targets:

.PHONY: \
	all \
	all_prelude \
	all_subdirs \
	all_postlude \
	clean \
	clean_prelude \
	clean_subdirs \
	clean_postlude \
	dist \
	dist_prelude \
	dist_normal \
	dist_subdirs \
	dist_postlude \
	distclean \
	distclean_subdirs \
	distclean_postlude \
	doc \
	doc_prelude \
	doc_subdirs \
	install \
	install_prelude \
	install_subdirs

# Delete all default suffixes.  Add back the few that we will use.

.SUFFIXES:
.SUFFIXES: .cc .c .o

#
# Rules used to re-configure the program if any of the configuration
# files change.
#

configure_deps :=
ifneq ($(MAKECMDGOALS),clean)
  ifneq ($(MAKECMDGOALS),distclean)
	configure_deps += $(top)/aclocal.m4
	configure_deps += $(top)/configure.in
  endif
endif

$(top)/configure: $(configure_deps)
	cd $(top) && autoconf

$(top)/Minit.mk: \
		$(top)/calib_config.in \
		$(top)/Minit.mk.in \
		$(top)/config.status
	cd $(top) && ./config.status

# Note: the one timestamp file is updated after *all* of the
# config.h's are rebuilt.  If any of the config.h.in's are newer than
# the one timestamp file, this means we need to rerun the config.status.

$(top)/include/calib/stamp-config.h: \
		$(top)/include/calib/config.h.in \
		$(top)/config.status
	cd $(top) && ./config.status

$(top)/config.status: $(top)/configure
	cd $(top) && ./config.status --recheck

#
# Here is how to bring all of the generated include files up to date.
#

generated_includes: \
	$(top)/include/calib/stamp-config.h \
	$(generated_h_files)

#
# Include file search dir arguments for the C/C++ compilers/preprocessors.
#

all_include_search_dirs := \
	. \
	$(top)/include \
	$(include_search_dirs)

all_include_search_dirs := $(strip $(all_include_search_dirs))

dash_I_args := $(addprefix -I,$(all_include_search_dirs))

#
# Rules for updating dependency files.
#

%.dep: %.cc
	@echo Finding dependencies for $<
	-@rm -f $(srcdir)/$@
	$(verbose)$(SHELL) -ec '$(CXX) -MM $(dash_I_args) -std=c++11 $< \
		| sed '\''s/\($*\.o\)[ :]*/$*\.o $@ : /g'\'' > $@'

%.dep: %.c
	@echo Finding dependencies for $<
	-@rm -f $(srcdir)/$@
	$(verbose)$(SHELL) -ec '$(CC) -MM $(dash_I_args) $< \
		| sed '\''s/\($*\.o\)[ :]*/$*\.o $@ : /g'\'' > $@'

#
# Define all of the dependency files (and their dependencies!!!).
#

dependency_files := \
	$(cxx_source_files:.cc=.dep) \
	$(extra_cxx_source_files:.cc=.dep) \
	$(c_source_files:.c=.dep) \
	$(cxx_test_drivers:.cc=.dep) \
	$(c_test_drivers:.c=.dep) \
	$(generated_cxx_files:.cc=.dep) \
	$(generated_c_files:.c=.dep) \
	$(extra_dep_files)

#
# We must rebuild the dependencies if Makefile changes, since it
# could change the include search path or other variables that affect
# the dependency files.
#

$(dependency_files): Makefile

#
# We should also rebuild the dependency files if either Minit.mk of
# Mrules.mk change.
#

$(dependency_files): \
	$(top)/Minit.mk \
	$(top)/Mrules.mk

#
# Now include all of the dependency files.
#
ifneq ($(MAKECMDGOALS),clean)
  ifneq ($(MAKECMDGOALS),distclean)
    ifneq ($(MAKECMDGOALS),disttree)
      sinclude $(dependency_files)
    endif
  endif
endif


ifdef VERBOSE
 verbose=
else
 verbose=@
endif

# Now generate a list of all the object files we are working with.

object_files := \
	$(cxx_source_files:.cc=.o) \
	$(c_source_files:.c=.o) \
	$(generated_cxx_files:.cc=.o) \
	$(generated_c_files:.c=.o) \
	$(extra_object_files)

#
# Rules to compile a C++ (.cc) file to create the object file.
#

cxx_and_c_compile_targets := \
	$(cxx_source_files:.cc=.o) \
	$(extra_cxx_source_files:.cc=.o) \
	$(c_source_files:.c=.o) \
	$(cxx_test_drivers:.cc=.o) \
	$(c_test_drivers:.c=.o) \
	$(generated_cxx_files:.cc=.o) \
	$(generated_c_files:.c=.o)

# Warning flags for C++ compiling.
#  -Wno-class-memaccess avoids message about clearing non-trivial type
#	with memset
#  -Wno-parenthese avoids message about unneccesary parentheses
CXX_WFLAGS = \
	-Wall 			\
	-Woverloaded-virtual 	\
	-Wno-ctor-dtor-privacy 	\
	-Wnon-virtual-dtor	\
	-Wno-sign-compare	\
	-Wno-class-memaccess	\
	-Wno-parentheses

# Treat all warnings as errors.  This is good for developers, but not
# for released versions because it causes builds to fail (for no good
# reason) on other compilers and platforms.
CXX_WFLAGS += $(WERROR_FLAG)

# A place to add other warning flags
ifdef EXTRA_WARN
CXX_WFLAGS += 
endif


OPTFLAGS = -O3

MOST_CXX_FLAGS = $(OPTFLAGS) $(CXX_WFLAGS) $(dash_I_args) $(CPPFLAGS) -std=c++11

# Rule for compiling a C++ file.

%.o: %.cc
	@echo Compiling $<
	$(verbose)$(CXX) -c $(MOST_CXX_FLAGS) $< -o $@

#
# Rules to compile a C (.c) file to create the object file.
#

# Warning flags for C compiling.
C_WFLAGS = \
	-Wall \
	$(WERROR_FLAG)

MOST_C_FLAGS = $(OPTFLAGS) $(C_WFLAGS) $(dash_I_args) $(CPPFLAGS)

# Rule for compiling a C file.

%.o: %.c
	@echo Compiling $<
	$(verbose)$(CC) -c $(MOST_C_FLAGS) $< -o $@

#
# Rule for creating a library from a set of (object) files.
#

%.a:
	@echo Generating library $@
	-$(verbose)rm -f $@
	$(verbose)$(AR) cr $@ $^
	$(verbose)$(RANLIB) $@

#
# Now for the various standard make targets.
# (Note that very old versions of sh/bash do not permit
#  	for i in list; do stmts; done
# where list is empty, so we use ifneq/endif to guard against this.)
#
# The "all" target.
#

subdirs := $(strip $(subdirs))

all: generated_includes all_prelude all_subdirs all_postlude $(goals)

all_prelude:

all_subdirs:
ifneq ($(subdirs),)
	@$(foreach dir,$(subdirs),$(MAKE) -C $(dir) all || exit $$?;)
endif

all_postlude:

#
# Rules for making things clean.
#

# By default, we delete ALL of the object files.  The user can selectively
# retain normal, debug or profile objects (e.g. RETAIN_DEBUG=yes).

lib_goals := $(filter %.a, $(goals))
obj_goals := $(filter %.o, $(goals))

clean_lib_files :=
clean_objects :=

clean_lib_files += $(lib_goals)

clean_objects += \
	$(c_source_files:.c=.o) \
	$(cxx_source_files:.cc=.o) \
	$(extra_c_source_files:.c=.o) \
	$(extra_cxx_source_files:.cc=.o) \
	$(c_test_drivers:.c=.o) \
	$(cxx_test_drivers:.cc=.o) \
	$(generated_c_files:.c=.o) \
	$(generated_cxx_files:.cc=.o) \
	$(extra_object_files)

# Only clean postscript files if we can rebuild them, AND if the
# user specifically says CLEAN_PS=yes.

clean_ps_files :=
ifdef CLEAN_PS
  ifeq ($(HAVE_LATEX),yes)
	# Only delete those we have the ability to regenerate!
	ifneq ($(DVIPS),)
		clean_ps_files += $(latex_src_files:.tex=.ps)
	endif
	ifneq ($(PS2PDF),)
		clean_ps_files += $(latex_src_files:.tex=.pdf)
	endif
  endif
endif

all_files_to_clean := \
	$(clean_lib_files) \
	$(clean_objects) \
	$(generated_c_files) \
	$(generated_cxx_files) \
	$(generated_h_files) \
	$(dependency_files) \
	$(clean_ps_files) \
	$(latex_src_files:.tex=.aux) \
	$(latex_src_files:.tex=.dvi) \
	$(latex_src_files:.tex=.log) \
	$(extra_clean_files)

all_files_to_clean := $(strip $(all_files_to_clean))

extra_clean_subdirs := $(strip $(extra_clean_subdirs))

define cleanFileActions
	@echo Removing files
	$(verbose)-rm -f $(all_files_to_clean)
endef

ifneq ($(extra_clean_subdirs),)
define cleanDirActions
	@Removing subdirs
	$(verbose)-rm -rf $(extra_clean_subdirs)
endef
endif

clean: clean_prelude clean_subdirs clean_postlude

clean_prelude:
	$(cleanFileActions)
	$(cleanDirActions)

clean_subdirs:
ifneq ($(subdirs),)
	@$(foreach dir,$(subdirs),$(MAKE) -C $(dir) clean || exit $$?; )
endif

clean_postlude:

#
# Rules for cleaning things back to distribution level.  Note that this
# may require re-running the configuration script.
#

distclean_files := $(strip $(distclean_files))
distclean_subdirs := $(strip $(distclean_subdirs))

ifneq ($(distclean_files),)
define distcleanFileActions
	@echo Removing files
	$(verbose)-rm -f $(distclean_files)
endef
endif

ifneq ($(distclean_subdirs),)
define distcleanDirActions
	@echo Removing directories
	$(verbose) -rm -rf $(distclean_subdirs)
endef
endif

define distcleanActions
	@-true
	$(distcleanFileActions)
	$(distcleanDirActions)
	$(cleanFileActions)
	$(cleanDirActions)
endef

# We do this "bottom-up" so that the top-level Minit.mk file is deleted
# at the end.
distclean: distclean_subdirs distclean_postlude

distclean_subdirs:
ifneq ($(subdirs),)
	@$(foreach dir,$(subdirs),$(MAKE) -C $(dir) distclean || exit $$?;)
endif

distclean_postlude:
	$(distcleanActions)

#
# Rules for making a distribution hierarchy.
#

files_to_distribute := \
	Makefile \
	$(cxx_source_files) \
	$(extra_cxx_source_files) \
	$(c_source_files) \
	$(header_files) \
	$(cxx_test_drivers) \
	$(c_test_drivers) \
	$(latex_src_files) \
	$(latex_src_files:.tex=.ps) \
	$(latex_src_files:.tex=.pdf) \
	$(extra_dist_files)

files_to_distribute := $(sort $(strip $(files_to_distribute)))

scripts_to_distribute := $(strip $(filter %.sh,$(files_to_distribute)))

toppath := $(shell (cd $(top); pwd))
relpath := $(shell pwd | sed -e 's@^$(toppath)@@' -e 's@^$$@.@' -e 's@^/@@')

distdir := $(top)/$(dist_subdir)/$(relpath)

distdir := $(patsubst %/.,%,$(distdir))

dist:	dist_prelude dist_normal dist_subdirs dist_postlude

# The local Makefile can add commands for this target...
dist_prelude:

dist_normal: $(files_to_distribute)
	@echo Copying files...
	$(verbose)mkdir $(distdir)
	$(verbose)cp $(files_to_distribute) $(distdir)
	$(verbose)for distfile in $(files_to_distribute); \
	do \
		touch -r $$distfile $(distdir)/$$distfile; \
	done
	$(verbose)(cd $(distdir); chmod 644 $(files_to_distribute))
ifneq ($(scripts_to_distribute),)
	$(verbose)(cd $(distdir); chmod 755 $(scripts_to_distribute))
endif

dist_subdirs:
ifneq ($(subdirs),)
	@$(foreach dir,$(sort $(subdirs)),$(MAKE) -C $(dir) dist || exit $$? ;)
endif

# The local Makefile can add commands for this target...
dist_postlude:

#
# Rules for installing everything.
#

install: install_prelude install_subdirs

# Each local Makefile provides its own commands for doing this...
install_prelude:

install_subdirs:
ifneq ($(subdirs),)
	@$(foreach dir,$(subdirs),$(MAKE) -C $(dir) install || exit $$?; )
endif

#
# Rules for making documentation.  This always causes the postscript
# files to be made.  If we don't have latex/dvips/dvipdf, then we are
# careful to never remove the postscript and/or PDF files.
#
# In general, we must run latex twice in order to get all the forward
# references correct.
#

.PRECIOUS: $(latex_src_files:.tex=.ps) $(latex_src_files:.tex=.pdf)

doc : doc_prelude doc_subdirs

doc_prelude: $(latex_src_files:.tex=.ps) $(latex_src_files:.tex=.pdf)

doc_subdirs:
ifneq ($(subdirs),)
	@$(foreach dir,$(subdirs),$(MAKE) -C $(dir) doc || exit $$?; )
endif

ifeq ($(HAVE_LATEX),yes)

.SECONDARY: $(latex_src_files:.tex=.dvi)

%.dvi : %.tex
	$(LATEX) $*
	$(LATEX) $*

ifneq ($(DVIPS),)
%.ps : %.dvi
	if $(DVIPS) -f -t letter <$< >__$@; then \
	  mv -f __$@ $@; else rm -f __$@; fi
endif

ifneq ($(PS2PDF),)
%.pdf : %.ps
	$(PS2PDF) $<
endif

endif
