/**********************************************************************

	$Id: combdist.c,v 1.1 2024/10/19 23:48:02 warme Exp $

	File:	combdist.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2015, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Compute centroid distances for TSP 3-toothed comb
	inequalities.

***********************************************************************

	Modification Log:

	a-1:	08/30/2015	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/calib.h"
#include "calib/cputime.h"
#include "calib/fatal.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "hilbert10.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Some macros that avoid confusing certain editors.
 */

#define	LBRACK_STRING	"["
#define	RBRACK_STRING	"]"
#define	LPAREN_STRING	"("
#define	RPAREN_STRING	")"


/*
 * Global information.
 */

struct ginfo {
	int 		nverts;		/* Number of vertex types */
	int 		nedges;		/* Number of edge types */
	const char **	nvarnames;	/* "n" variable names */
	const char **	xvarnames;	/* "x" variables names */
	int *		v_to_e;		/* Xlate pair of vertices to edge */
	int *		edge_v1;	/* First vertex of each edge */
	int *		edge_v2;	/* Second vertex of each edge */
	struct calib_Zxyz_dom *
			nring;		/* Ring of "n" vars */
	struct calib_Zxyz_dom *
			nxring;		/* Ring of both the "n" and "x" vars */
	struct calib_rat_dom *
			nrat;		/* Rational functions of "n" vars */
	struct calib_rat_dom *
			nxrat;		/* Rational functions of "n" and */
					/* "x" vars */
	int *		nx2n_varmap;	/* Varmap for converting nxring */
					/* to nring */
	int		numrx;		/* Number of remaining "x" vars */
	const char **	rxvarnames;	/* Remaining "x" variable names */
	int *		rx_to_x;	/* Map remaining x var to orig x */
	int *		x_to_rx;	/* Map orig x var to remaining */
					/* x var, or -1 if eliminated */
};

/*
 * An object to hold a complete solution.
 *
 * There is a large factor `pden' in the denominator of both `dist2'
 * and each component of `point'.  We remove this factor and store it
 * separately for compactness (especially when printing these!).
 */

struct solution {
	const struct calib_rat_dom *	d2_dom;	/* Domain of dist2 */
	struct calib_rat_obj *		dist2;	/* Distance squared */
	const struct calib_rat_dom *	pt_dom;	/* Domain of point */
	struct calib_subst_list *	point;	/* Closest point */
	struct calib_Zxyz_obj *		pden;	/* Common denominator */
};

/*
 * An object to hold pre-specified pivot sequences for both
 * the `hull' system and the `gradient' system.  Either or
 * both of these pointers can be NULL.
 */

struct PPS {
	const int *	hull_ps;	/* Pivot sequence for hull system */
	const int *	gradient_ps;	/* Pivot sequence for gradient system */
};

/*
 * A `state' object for remembering how far down we are within
 * a given pre-specified pivot sequence.
 */

struct PPSState {
	const int *	piv_seq;	/* Pivot sequence */
	int		index;		/* Index into piv_seq[] */
};

/*
 * A system of equations.
 */

struct sys {
	int			nrows;	/* Number of rows */
	int			ncols;	/* Number of columns */
	const char **		vars;	/* Var name for each column */
	struct calib_Zxyz_obj **
				A;	/* Matrix */
	struct calib_Zxyz_obj **
				b;	/* Right hand side vector */
	struct calib_Zxyz_dom *	ring;	/* Coefficient ring */
	int *			lbound;	/* Lower bounds on coefficient vars */
	int *			rflag;	/* j=rflag[i]: -1 or pivot at (i,j) */
	int *			cflag;	/* i=cflag[j]: -1 or pivot at (i,j) */
	char *			name;	/* Name of system */
};

/*
 * An object to contain a description of the "size" of a polynomial.
 * We use this to measure the "quality" of an otherwise valid pivot
 * element.  We want to choose pivot elements that are the "simplest"
 * possible sub-expressions, and this object allows us to compare the
 * complexity of two prospective pivot candidates.
 */

struct psize {
	int	nterms;		/* Number of terms */
	int	maxdeg;		/* Maximum degree of any one variable */
	int	totdeg;		/* Total degree of polynomial */
	int	nbits;		/* Max coefficient length */
	int	sign;		/* Sign of leading coefficient */
};

/*
 * An object to record a potential pivot candidate, together with
 * the "complexity" of that pivot element.  We sort these from
 * simplest to most complicated, and then test them in this order until
 * we find one that appears unable to become zero under any valid
 * assignment of its variables.  It is the "can be zero" check that is
 * expensive, so this way we pay this cost only on the simplest
 * candidates possible.
 */

struct pcand {
	int		prow;	/* Pivot row */
	int		pcol;	/* Pivot column */
	struct psize	psize;	/* "Size" of the pivot */
	struct pcand *	next;	/* Next pivot in list */
};

/*
 * Local Functions
 */

static void		add_int_into (struct calib_Zxyz_obj * poly, int val);
static void		cancel_pden_from_dist2 (struct calib_rat_obj *	dist2,
						struct calib_Zxyz_obj *	pden);
static calib_bool	check_solution (const struct solution *	rsp,
					const struct solution *	gsp,
					const calib_bool *	v_not_exist);
static void		clear_ginfo (struct ginfo * gip);
static void		clear_solution (struct solution * sp);
static void		clear_system (struct sys * sys);
static int		compare_poly_sizes (const struct psize *	p1,
					    const struct psize *	p2);
static void		compute_dist2_and_closest_point (
					struct solution *	sp,
					const calib_bool *	v_not_exist,
					const struct PPS *	piv_seqs,
					struct ginfo *		gip);
static struct calib_subst_list *
			compute_gradient_substitutions (struct ginfo *	gip,
							struct sys *	gsys);
static struct calib_subst_list *
			compute_hull_substitutions (
						struct ginfo *	gip,
						struct sys *	hsys,
						const int *	col2edge);
static struct calib_subst_list *
			compute_point (struct ginfo *			gip,
				       struct calib_subst_list *	hslist,
				       struct calib_subst_list *	gslist);
static int *		compute_var_max_degrees (
					struct calib_Zxyz_obj *	p,
					int *			totaldeg);
static const char **	copy_varlist (const char ** varlist, int nvars);
static void		create_dist2_expr (
					struct ginfo *		gip,
					struct calib_rat_obj *	dist2,
					const calib_bool *	v_not_exist);
static void		create_gradient_system (
					struct ginfo *			gip,
					struct sys *			gsys,
					const struct calib_rat_obj *	dist2);
static int *		create_hull_system (struct ginfo *	gip,
					    struct sys *	hsys,
					    const calib_bool *	v_not_exist);
static void		create_xvar_map (struct ginfo *		gip,
					 const struct sys *	hsys,
					 const int *		col2edge);
static void		debug_print_system (const struct sys * sys);
static void		debug_print_system_complexity (const struct sys * sys);
static void		free_varlist (const char ** varlist, int nvars);
static void		gauss_choose_pivot (struct sys *	sys,
					    int *		p,
					    int *		q,
					    struct PPSState *	state_ptr);
static void		gauss_eliminate (struct sys *	sys,
					 const int *	piv_seq);
static void		gauss_elim_row (struct sys *	sys,
					int		p,
					int		q,
					int		i,
					int		giter);
static struct calib_Zxyz_obj *
			get_point_denom (struct calib_subst_list * point);
static void		get_poly_size (struct psize *		psize,
				       struct calib_Zxyz_obj *	p);
static void		init_vars (struct ginfo * gip);
static calib_bool	is_constant_poly (const struct calib_Zxyz_obj * poly);
static calib_bool	is_one_poly (const struct calib_Zxyz_obj * poly);
static void		make_all_vars_ring (struct ginfo * gip);
static struct calib_Zxyz_obj *
			make_comb_lhs_term (struct calib_Zxyz_dom *	ring,
					     int			coeff,
					     int			u,
					     int			v);
static struct calib_Zxyz_obj *
			make_comb_rhs (struct calib_Zxyz_dom *	ring,
				       const calib_bool *	v_not_exist);
static struct calib_Zxyz_obj *
			make_int (struct calib_Zxyz_dom * ring, int val);
static struct calib_Zxyz_obj *
			make_var (struct calib_Zxyz_dom * ring, int var);
static struct calib_Zxyz_obj *
			make_2x_num_edges (struct calib_Zxyz_dom *	ring,
					   int				u,
					   int				v);
static void		mul_by_int (struct calib_Zxyz_obj * p, int val);
static void		partial_derivative (
					struct calib_Zxyz_obj *		result,
					const struct calib_Zxyz_obj *	poly,
					int				var);
static void		perform_gradient_substitutions (
					struct calib_rat_obj *		result,
					const struct calib_rat_obj *	ratp,
					struct calib_subst_list *	slist,
					struct ginfo *			gip);
static void		perform_hull_substitutions (
					struct calib_rat_obj *		result,
					const struct calib_rat_obj *	ratp,
					struct calib_subst_list *	slist,
					struct ginfo *			gip);
static calib_bool	pivot_can_be_zero (
				const struct calib_Zxyz_obj *	p,
				int *				lbound);
static calib_bool	pivots_in_order (const struct pcand *	p1,
					 const struct pcand *	p2);
static void		print_factored (struct calib_Zxyz_obj * poly);
static void		print_factored_maxima (struct calib_Zxyz_obj * poly);
static void		print_maxima_set_matrix_elements (
				const char *			name,
				struct calib_Zxyz_obj **	A,
				int				nrows,
				int				ncols);
static void		print_named_subst_factored (
					const char *			vardef,
					const struct calib_subst_list *	slist,
					const char *			numf,
					const char *			denf);
static void		print_poly_factored_nnl (
					const struct calib_Zxyz_obj *	poly,
					const char *			fact);
static void		rat_print_factored_nnl (
					const struct calib_rat_obj *	rp,
					const char *			numf,
					const char *			denf);
static void		print_solution (struct solution * sp);
static void		print_substitutions (
					const struct calib_subst_list * slist);
static void		print_sys_to_maxima (const struct sys * sys);
static void		print_vars (const struct calib_Zxyz_obj * poly);
static void		print_unfactored (struct calib_Zxyz_obj * poly);
static calib_bool	solution_polys_match (
					const struct calib_Zxyz_obj *	rpoly,
					const struct calib_Zxyz_obj *	gpoly,
					const calib_bool *	v_not_exist);
static struct pcand *	sort_pivot_candidates (struct pcand * plist);

/*
 * Pre-determined Pivot Sequences for the (n6 EQ 0) AND (n7 > 0) case.
 *
 * The default pivot choices made by the algorithm result in the situation
 * where the final pivot for the `gradient' system (where there is only
 * one remaining unpivoted position) winds up being a polynomial whose value
 * can be zero for valid combinations of n0-n5, and n7.
 *
 * To remedy this situation we need to tweak only the `hull' pivot sequence.
 */

const int		n6_zero_n7_positive_hull_pivots [] = {
	1, 10,		/* Eliminate x14. */
	-1,
};

const struct PPS	n6_zero_n7_positive_pivots = {
	&n6_zero_n7_positive_hull_pivots [0],
	NULL,
};

/*
 * The main program.
 */

	int
main (

int		argc,
char **		argv
)
{
calib_bool		do_print;
calib_bool		v_not_exist [8];
struct calib_Zxyz_dom *	nring;
struct ginfo		ginfo;
struct solution		gen_sol, sol1, sol2, sol3;

	do_print = TRUE;

	memset (&ginfo, 0, sizeof (ginfo));

	memset (&gen_sol, 0, sizeof (gen_sol));
	memset (&sol1, 0, sizeof (sol1));
	memset (&sol2, 0, sizeof (sol2));
	memset (&sol3, 0, sizeof (sol3));

	init_vars (&ginfo);

	nring = calib_make_Zxyz_dom (ginfo.nverts, ginfo.nvarnames);
	ginfo.nring = nring;

	make_all_vars_ring (&ginfo);

	/* Make rational function domains. */
	ginfo.nrat	= calib_make_rat_dom (nring);
	ginfo.nxrat	= calib_make_rat_dom (ginfo.nxring);

	memset (v_not_exist, FALSE, sizeof (v_not_exist));

	/* Solve the general problem first. */
	v_not_exist [6] = FALSE;
	v_not_exist [7] = FALSE;
	compute_dist2_and_closest_point (&gen_sol,
					 v_not_exist,
					 NULL,
					 &ginfo);

	if (do_print) {
		printf ("\nGeneral solution:\n");
		print_solution (&gen_sol);
	}

	/* Solve the n6=0 and n7>0 problem.				*/
	/* NOTE: the pivot sequence chosen automatically for this	*/
	/*	 sub-problem fails: the final pivot for the gradient	*/
	/*	 system (for which there is only one unpivoted row and	*/
	/*	 column remaining, and therefore no choice) winds up	*/
	/*	 being a polynomial that can be zero for valid values	*/
	/*	 of the parameters.  Tweaking the first pivot of the	*/
	/*	 hull system (with a pre-specified pivot sequence)	*/
	/*	 eliminates this issue.					*/
	v_not_exist [6] = TRUE;
	v_not_exist [7] = FALSE;
	compute_dist2_and_closest_point (&sol1,
					 v_not_exist,
					 &n6_zero_n7_positive_pivots,
					 &ginfo);

	check_solution (&sol1, &gen_sol, v_not_exist);

	if (do_print) {
		printf ("\nSolution for n6=0 and n7>0:\n");
		print_solution (&sol1);
	}
	clear_solution (&sol1);

	/* Solve the n6>0 and n7=0 problem. */
	v_not_exist [6] = FALSE;
	v_not_exist [7] = TRUE;
	compute_dist2_and_closest_point (&sol2,
					 v_not_exist,
					 NULL,
					 &ginfo);

	check_solution (&sol2, &gen_sol, v_not_exist);

	if (do_print) {
		printf ("\nSolution for n6>0 and n7=0:\n");
		print_solution (&sol2);
	}
	clear_solution (&sol2);

	/* Solve the n6=0 and n7=0 problem. */
	v_not_exist [6] = TRUE;
	v_not_exist [7] = TRUE;
	compute_dist2_and_closest_point (&sol3,
					 v_not_exist,
					 NULL,
					 &ginfo);

	check_solution (&sol3, &gen_sol, v_not_exist);

	if (do_print) {
		printf ("\nSolution for n6=0 and n7=0:\n");
		print_solution (&sol3);
	}
	clear_solution (&sol3);

	clear_solution (&gen_sol);

	clear_ginfo (&ginfo);

	calib_shutdown ();

	exit (0);
}

/*
 * Compute the solution, consisting of:
 *
 *	1. The point Y in the affine hull of the 3-toothed comb facet that
 *	   is closest to the centroid C of the polytope.
 *
 *	2. The distance-squared between Y and C.
 *
 * These are obtained as exact, symbolic rational functions of 8 symbolic
 * parameters n0 through n7, which are the cardinalities of the particular
 * vertex subsets within the 3-toothed comb.
 *
 * Two of these parameters (n6 and n7) are allowed to be zero (i.e., the
 * corresponding vertex subsets are empty).  Unfortunately, when either
 * or both of n6 and/or n7 are zero, the "degree 2" constraint associated
 * with that subset (in the general formulation) transforms from
 *
 *		\sum_{e \in \delta(v)} x_e = 2
 *
 * to
 *
 *		0 = 2
 *
 * because none of these edges actually exist when vertex v does not
 * actually exist.
 *
 * Our caller passes in an array of flags that indicates which of the
 * vertex classes do not actually exist.  We set up and solve only
 * that portion of the problem that actually exists.  The results of
 * said computation will not be tainted by any 0 = 2 assumption, and
 * should be absolutely correct.  The objective is to determine
 * whether or not the general formulae also give the correct solutions
 * when n6=0 and/or n7=0.  So either we vindicate the general
 * solution, or we get correct expressions for the various cases.
 */

	static
	void
compute_dist2_and_closest_point (

struct solution *	sp,		/* IN/OUT: solution */
const calib_bool *	v_not_exist,	/* IN: flags indicating which */
					/*     vertices do not exist */
const struct PPS *	piv_seqs,	/* IN: optional Pre-determined */
					/*     Pivot Sequences */
struct ginfo *		gip		/* IN/OUT: global info */
)
{
int *				col2edge;
calib_cpu_time_t		t0, t1;
const int *			hull_ps;
const int *			gradient_ps;
struct calib_subst_list *	hslist;
struct calib_subst_list *	gslist;
struct calib_subst_list *	point;
struct calib_Zxyz_dom *		nring;
struct calib_rat_dom *		nrat;
struct calib_rat_dom *		nxrat;
struct calib_Zxyz_obj *		pden;
struct calib_rat_obj *		rp;
struct calib_rat_obj		dist2_expr, affine_dist2, dist2;
struct sys			hsys;
struct sys			gsys;
char				tbuf [32];

	CALIB_FATAL_ERROR_IF ((sp	   EQ NULL) OR
			      (v_not_exist EQ NULL) OR
			      (gip	   EQ NULL));

	t0 = calib_get_cpu_time ();

	hull_ps		= NULL;
	gradient_ps	= NULL;
	if (piv_seqs NE NULL) {
		hull_ps		= piv_seqs -> hull_ps;
		gradient_ps	= piv_seqs -> gradient_ps;
	}

	nring	= gip -> nring;
	nrat	= gip -> nrat;
	nxrat	= gip -> nxrat;

	nxrat -> init (nxrat, &dist2_expr);
	nxrat -> init (nxrat, &affine_dist2);
	nrat  -> init (nrat,  &dist2);

	create_dist2_expr (gip, &dist2_expr, v_not_exist);

	col2edge = create_hull_system (gip, &hsys, v_not_exist);

#if 0
	debug_print_system (&hsys);
#endif

	gauss_eliminate (&hsys, hull_ps);

	hslist = compute_hull_substitutions (gip, &hsys, col2edge);

	perform_hull_substitutions (&affine_dist2, &dist2_expr, hslist, gip);

	create_xvar_map (gip, &hsys, col2edge);

	clear_system (&hsys);

	create_gradient_system (gip, &gsys, &affine_dist2);

#if 0
	debug_print_system_complexity (&gsys);
#endif

	gauss_eliminate (&gsys, gradient_ps);

	gslist = compute_gradient_substitutions (gip, &gsys);

	clear_system (&gsys);

	perform_gradient_substitutions (&dist2, &affine_dist2, gslist, gip);

	point = compute_point (gip, hslist, gslist);

	pden = get_point_denom (point);

	cancel_pden_from_dist2 (&dist2, pden);

	/* Copy `dist2' into a dynamically-allocated rat object. */
	rp = CALIB_NEW (struct calib_rat_obj);
	nrat -> init (nrat, rp);
	nring -> set (&(rp -> numer), &dist2.numer);
	nring -> set (&(rp -> denom), &dist2.denom);

	/* Pass solution information back to caller. */
	sp -> d2_dom	= nrat;
	sp -> dist2	= rp;
	sp -> pt_dom	= nxrat;
	sp -> point	= point;
	sp -> pden	= pden;

	nrat  -> clear (&dist2);
	nxrat -> clear (&affine_dist2);
	nxrat -> clear (&dist2_expr);

	nxrat -> free_slist (gslist);
	nxrat -> free_slist (hslist);

	t1 = calib_get_cpu_time ();
	calib_convert_cpu_time (t1 - t0, tbuf);
	printf ("CPU Time:\t%s\n", tbuf);
}

/*
 * Print out the solution in a form that maxima can read.
 * It defines `d' to be the common denominator factor, and then defines
 * `dist2' and `point' using `d' as a short-hand for this factor.
 */

	static
	void
print_solution (

struct solution *	sp		/* IN: solution to print */
)
{
const struct calib_Zxyz_dom *	ring;

	CALIB_FATAL_ERROR_IF (sp EQ NULL);
	CALIB_FATAL_ERROR_IF (sp -> pden EQ NULL);
	CALIB_FATAL_ERROR_IF (sp -> dist2 EQ NULL);
	CALIB_FATAL_ERROR_IF (sp -> point EQ NULL);

	ring = sp -> pden -> dom;

	printf ("\n"
		"d : ");
	ring -> print_maxima_nnl (sp -> pden);
	printf ("$\n\n");

	printf ("\n"
		"dist2 : ");
	rat_print_factored_nnl (sp -> dist2, NULL, "d");
	printf ("$\n\n");

	print_named_subst_factored ("point", sp -> point, NULL, "d");
}

/*
 * Check a restricted solution to see whether or not it matches the
 * general solution evaluated with n6=0 and/or n7=0, as appropriate.
 */

	static
	calib_bool
check_solution (

const struct solution *	rsp,	/* IN: restricted solution to check */
const struct solution *	gsp,	/* IN: general solution to compare against */
const calib_bool *	v_not_exist
				/* IN: which vertices do not exist */
)
{
int				i, j, rn, gn;
calib_bool			correct;
const char *			varname;
struct calib_rsubst *		rsubst;
struct calib_rsubst *		gsubst;

	CALIB_FATAL_ERROR_IF ((rsp	   EQ NULL) OR
			      (gsp	   EQ NULL) OR
			      (v_not_exist EQ NULL));

	correct = TRUE;

	if (NOT solution_polys_match (rsp -> pden, gsp -> pden, v_not_exist)) {
		printf ("\tDenominator does not match!\n");
		correct = FALSE;
	}

	if (NOT solution_polys_match (&(rsp -> dist2 -> numer),
				      &(gsp -> dist2 -> numer),
				      v_not_exist)) {
		printf ("\tDist2 numerator does not match!\n");
		correct = FALSE;
	}

	/* In general, gn > rn here.  Only compare those members of	*/
	/* gsp -> point that actually appear in rsp -> point.		*/

	rn	= rsp -> point -> num_subst;
	rsubst	= rsp -> point -> subst;
	gn	= gsp -> point -> num_subst;
	gsubst	= gsp -> point -> subst;
	for (i = 0; i < rn; i++) {
		varname = rsubst [i].varname;
		for (j = 0; j < gn; j++) {
			if (strcmp (gsubst [j].varname, varname) EQ 0) break;
		}
		CALIB_FATAL_ERROR_IF (j >= gn);
		if (NOT solution_polys_match (&(rsubst [i].val.numer),
					      &(gsubst [j].val.numer),
					      v_not_exist)) {
			printf ("\tPoint mismatch for %s numerator!\n",
				varname);
			correct = FALSE;
		}
		if (NOT solution_polys_match (&(rsubst [i].val.denom),
					      &(gsubst [j].val.denom),
					      v_not_exist)) {
			printf ("\tPoint mismatch for %s denominator!\n",
				varname);
			correct = FALSE;
		}
	}

	if (correct) {
		printf ("\nSolutions match.\n");
	}
	else {
		printf ("\n"
			"********************************\n"
			" SOLUTIONS DO NOT MATCH!\n"
			"********************************\n");
	}
	return (correct);
}

/*
 * Compare a polynomial from the restricted solution against the
 * (suitably restricted) polynomial from the general solution.
 * Returns TRUE if they match, FALSE otherwise.
 */

	static
	calib_bool
solution_polys_match (

const struct calib_Zxyz_obj *
			rpoly,		/* IN: poly from restricted solution */
const struct calib_Zxyz_obj *
			gpoly,		/* IN: poly from general solution */
const calib_bool *	v_not_exist	/* IN: which vertices do not exist */
)
{
int				nvars;
const struct calib_Zxyz_dom *	ring;
const calib_bool *		evflags;
mpz_ptr				zeros;
struct calib_Zxyz_obj		GEVAL;

	CALIB_FATAL_ERROR_IF ((rpoly	   EQ NULL) OR
			      (gpoly	   EQ NULL) OR
			      (v_not_exist EQ NULL));

	ring = rpoly -> dom;
	CALIB_FATAL_ERROR_IF (ring NE gpoly -> dom);

	nvars = ring -> nvars;

	zeros = calib_new_Z_vector (nvars);
	ring -> init (ring, &GEVAL);

	/* Evaluate (at zero) variables for which the corresponding	*/
	/* class of vertices does not exist (cardinality zero).		*/
	evflags = v_not_exist;

	ring -> eval_var_subset (&GEVAL, gpoly, evflags, zeros);

	/* Subtract these two polynomials.  They match if the result	*/
	/* is zero.							*/
	ring -> sub (&GEVAL, &GEVAL, rpoly);

	CALIB_FATAL_ERROR_IF (GEVAL.nterms < 0);

	return (GEVAL.nterms EQ 0);
}

/*
 * Clear out the given solution object.
 */

	static
	void
clear_solution (

struct solution *	sp		/* IN/OUT: solution object to clear */
)
{
struct calib_Zxyz_obj *		pden;
struct calib_rat_obj *		dist2;
const struct calib_Zxyz_dom *	ring;
const struct calib_rat_dom *	d2_dom;
const struct calib_rat_dom *	pt_dom;

	CALIB_FATAL_ERROR_IF (sp EQ NULL);

	pden	= sp -> pden;
	dist2	= sp -> dist2;
	d2_dom	= sp -> d2_dom;
	pt_dom	= sp -> pt_dom;

	if (pden NE NULL) {
		ring = pden -> dom;
		ring -> free (pden);
	}
	if (dist2 NE NULL) {
		d2_dom -> clear (dist2);
		free (dist2);
	}
	pt_dom -> free_slist (sp -> point);

	memset (sp, 0, sizeof (*sp));
}

/*
 * Initialize the variables used to represent:
 *
 *	- "n" vars: the number of vertices in each vertex class.
 *	- "x" vars: the weight of each member of an edge class.
 */

	static
	void
init_vars (

struct ginfo *		gip		/* IN/OUT: global info */
)
{
int		u, v, e;
int		nv, ne;
const char **	nvnames;
const char **	xvnames;
int *		v2e;
int *		ev1;
int *		ev2;
char		buf [8];

	nv = 8;
	ne = nv * (nv + 1) / 2;

	nvnames = CALIB_NEWA (nv, const char *);

	for (v = 0; v < nv; v++) {
		buf [0] = 'n';
		buf [1] = '0' + v;
		buf [2] = '\0';
		nvnames [v] = strdup (buf);
	}

	ev1 = CALIB_NEWA (ne, int);
	ev2 = CALIB_NEWA (ne, int);
	v2e = CALIB_NEWA (nv * nv, int);
	xvnames = CALIB_NEWA (ne, const char *);

	e = 0;
	for (u = 0; u < nv; u++) {
		for (v = u; v < nv; v++) {
			ev1 [e] = u;
			ev2 [e] = v;
			v2e [u * nv + v] = e;
			v2e [v * nv + u] = e;
			sprintf (buf, "x%d%d", u, v);
			xvnames [e] = strdup (buf);
			++e;
		}
	}

	gip -> nverts	= nv;
	gip -> nedges	= ne;
	gip -> nvarnames	= nvnames;
	gip -> xvarnames	= xvnames;
	gip -> v_to_e		= v2e;
	gip -> edge_v1		= ev1;
	gip -> edge_v2		= ev2;
}

/*
 * Make a ring for polynomials over all the variables, both "n" and "x".
 */

	static
	void
make_all_vars_ring (

struct ginfo *	gip		/* IN/OUT: global info */
)
{
int		i, j;
int		nverts, nedges, nvars;
const char **	nvarnames;
const char **	xvarnames;
const char **	varlist;
int *		cv_varmap;

	nverts		= gip -> nverts;
	nedges		= gip -> nedges;
	nvarnames	= gip -> nvarnames;
	xvarnames	= gip -> xvarnames;

	/* Make a new ring containing all of the variables. */
	nvars	= nverts + nedges;

	varlist = CALIB_NEWA (nvars, const char *);
	j = 0;
	for (i = 0; i < nverts; i++) {
		varlist [j++] = nvarnames [i];
	}
	for (i = 0; i < nedges; i++) {
		varlist [j++] = xvarnames [i];
	}
	CALIB_FATAL_ERROR_IF (j NE nvars);

	gip -> nxring = calib_make_Zxyz_dom (nvars, varlist);

	free (varlist);

	/* Make a variable list for converting nxring to nring. */
	cv_varmap = CALIB_NEWA (nvars, int);
	for (i = 0; i < nverts; i++) {
		cv_varmap [i] = i;
	}
	for (; i < nvars; i++) {
		cv_varmap [i] = -1;
	}
	gip -> nx2n_varmap = cv_varmap;
}

/*
 * Create an expression that represents the distance squared from
 * a point (described by the xvars) and the centroid.
 * This is a rational function having all variables in the numerator
 * and only the "n" variables in the denominator.  We use the same
 * ring (of all variables) for both polynomials, however, so that
 * they are computationally compatible.
 *
 * This expression has the following form:
 *
 *	\sum_{e=(i,j)} NumEdges(i,j) (x_e - C_e)**2
 *
 * The expression has the form numerator / (2 * (n-1)**2) where the
 * 1/2 factor comes from the NumEdges(i,i) terms and the 1/(n-1)**2
 * factor comes from C_e = 2 / (n-1), where n is the sum of all
 * the n vars (n0 + ... + n7).
 */

	static
	void
create_dist2_expr (

struct ginfo *		gip,		/* IN: global info */
struct calib_rat_obj *	dist2,		/* IN/OUT: distance squared expression */
const calib_bool *	v_not_exist	/* IN: which vertices do not exist */
)
{
int			i, j, e;
int			nverts, nedges;
int *			edge_v1;
int *			edge_v2;
struct calib_Zxyz_dom *	nxring;
struct calib_rat_dom *	nxrat;
struct calib_Zxyz_obj *	nvarp;
struct calib_Zxyz_obj *	xvarp;
struct calib_Zxyz_obj *	nep;
struct calib_Zxyz_obj	nMinusOne, TERM;

	nverts	= gip -> nverts;
	nedges	= gip -> nedges;
	edge_v1	= gip -> edge_v1;
	edge_v2	= gip -> edge_v2;
	nxring	= gip -> nxring;
	nxrat	= gip -> nxrat;

	nxring -> init (nxring, &nMinusOne);
	nxring -> init (nxring, &TERM);

	/* Create a polynomial representing the value "n-1",	*/
	/* where n = n0 + ... + n7.				*/
	for (i = 0; i < nverts; i++) {
		if (v_not_exist [i]) continue;
		nvarp = make_var (nxring, i);
		nxring -> add (&nMinusOne, &nMinusOne, nvarp);
		nxring -> free (nvarp);
	}
	add_int_into (&nMinusOne, -1);

	nxring -> set (&(dist2 -> denom), &nMinusOne);
	nxring -> mul (&(dist2 -> denom),
		       &(dist2 -> denom),
		       &(dist2 -> denom));
	mul_by_int  (&(dist2 -> denom), 2);

	/* Now add all necessary terms into the numerator. */
	for (e = 0; e < nedges; e++) {
		i = edge_v1 [e];
		j = edge_v2 [e];

		if ((v_not_exist [i]) OR (v_not_exist [j])) {
			/* At least one of the vertices does not exist.	*/
			/* This edge, therefore, does not exist.	*/
			continue;
		}

		xvarp = make_var (nxring, nverts + e);
		nxring -> set (&TERM, xvarp);
		nxring -> free (xvarp);

		nxring -> mul (&TERM, &TERM, &nMinusOne);

		add_int_into (&TERM, -2);

		nxring -> mul (&TERM, &TERM, &TERM);

		nep = make_2x_num_edges (nxring, i, j);
		nxring -> mul (&TERM, &TERM, nep);
		nxring -> free (nep);

		nxring -> add (&(dist2 -> numer), &(dist2 -> numer), &TERM);
	}

	nxrat -> canonicalize (dist2);

	nxring -> clear (&TERM);
	nxring -> clear (&nMinusOne);
}

/*
 * Create the system of equations that define the affine hull
 * of the 3-toothed comb facet.
 *
 * Return value is a `col2edge[]' map, which caller must free.
 */

	static
	int *
create_hull_system (

struct ginfo *	gip,		/* IN: global info */
struct sys *	hsys,		/* IN/OUT: hull system */
const calib_bool *
		v_not_exist	/* IN: which vertices do not exist */
)
{
int				i, j, k, nverts, nedges;
int				nrows, ncols, nvars, ndeg2;
int *				v_to_e;
int *				edge_v1;
int *				edge_v2;
int *				col2edge;
int *				edge2col;
int *				lbound;
const char **			vars;
struct calib_Zxyz_dom *		ring;
struct calib_Zxyz_obj **	A;
struct calib_Zxyz_obj **	b;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		p;

	memset (hsys, 0, sizeof (*hsys));

	nverts	= gip -> nverts;
	nedges	= gip -> nedges;
	v_to_e	= gip -> v_to_e;
	edge_v1	= gip -> edge_v1;
	edge_v2	= gip -> edge_v2;
	ring	= gip -> nring;
	nvars	= ring -> nvars;

	/* Count the number of degree-2 equations.  We only have such	*/
	/* equations for vertex classes that exist.			*/
	ndeg2 = 0;
	for (i = 0; i < nverts; i++) {
		if (v_not_exist [i]) {
			/* No vertex, no degree-2 equation! */
		}
		else {
			++ndeg2;
		}
	}

	/* Determine which edges actually exist.  Our system has one	*/
	/* column per vertex that exists.  Fill in an array of variable	*/
	/* indices that each column represents, as well as a map from	*/
	/* variable index to column number (or -1).			*/
	col2edge = CALIB_NEWA (nedges, int);
	edge2col = CALIB_NEWA (nedges, int);
	for (i = 0; i < nedges; i++) {
		edge2col [i] = -1;
	}

	ncols = 0;
	for (i = 0; i < nedges; i++) {
		j = edge_v1 [i];
		k = edge_v2 [i];

		if ((v_not_exist [j]) OR (v_not_exist [k])) {
			/* This edge does not exist. */
		}
		else {
			col2edge [ncols] = i;
			edge2col [i] = ncols;
			++ncols;
		}
	}

#define EDGE(u,v)					\
	({ int e = v_to_e [(u)*nverts + (v)];		\
	   int col = edge2col [e];			\
	   CALIB_FATAL_ERROR_IF (col < 0);		\
	   col;						\
	})

	nrows	= ndeg2 + 1;

	A	= CALIB_NEWA (nrows * ncols,	struct calib_Zxyz_obj *);
	b	= CALIB_NEWA (nrows,		struct calib_Zxyz_obj *);
	memset (A, 0, nrows * ncols * sizeof (A [0]));
	memset (b, 0, nrows	    * sizeof (b [0]));

	/* The first equations are "degree 2" equations, one for each	*/
	/* class of vertex type.  These have the following form:	*/
	/*								*/
	/*	(n_i - 1) x_i \sum_{j \ne i} n_j x_{i,j} = 2		*/
	/*								*/
	/* WARNING: we leave the zero elements as NULL pointers, not	*/
	/* polynomial objects that represent the value zero!		*/

	rowp = A;
	k = 0;
	for (i = 0; i < nverts; i++) {
		if (v_not_exist [i]) continue;

		for (j = 0; j < nverts; j++) {
			if (v_not_exist [j]) continue;
			p = make_var (ring, j);
			if (j EQ i) {
				add_int_into (p, -1);
			}
			rowp [EDGE (i, j)] = p;
		}
		b [k] = make_int (ring, 2);

		rowp += ncols;
		++k;
	}

	/* The last equation is the 3-toothed-comb equation.	Once	*/
	/* again, this equation is multiplied by 2 in order to		*/
	/* eliminate the factor of 1/2 that would normally arise from	*/
	/* the NumEdges(i,i) terms.					*/
	/*								*/
	/*	0----1				teeth = 0+1, 2+3, 4+5	*/
	/*	|				handle = 6		*/
	/*	2----3		7		outside = 7		*/
	/*	|				tooth tines = 1, 3, 5	*/
	/*	4----5				tooth bases = 0, 2, 4	*/
	/*	|							*/
	/*	6							*/
	/*								*/
	/* The comb inequality is:					*/
	/*								*/
	/*	x(H) + \sum_i x(Ti) <= |H| + \sum_i |Ti| - (3*t+1)/2	*/
	/*								*/
	/* where t = 3 and 1 <= i <= t.  This means that vertex sets	*/
	/* 0, 2 and 4 count twice on both the left and right-hand sides	*/
	/* (because they are members of both H and Ti).			*/

#define COMB_TERM(coeff, u, v)						\
    do {								\
	if ((v_not_exist [u]) OR (v_not_exist [v])) {			\
		/* Edge does not exist.  No term. */			\
	}								\
	else {								\
		rowp [EDGE (u,v)] = make_comb_lhs_term (ring, coeff, u, v); \
	}								\
    } while (FALSE)

	COMB_TERM (2, 0, 0);
	COMB_TERM (1, 0, 1);
	COMB_TERM (1, 0, 2);
	COMB_TERM (1, 0, 4);
	COMB_TERM (1, 0, 6);

	COMB_TERM (1, 1, 1);

	COMB_TERM (2, 2, 2);
	COMB_TERM (1, 2, 3);
	COMB_TERM (1, 2, 4);
	COMB_TERM (1, 2, 6);

	COMB_TERM (1, 3, 3);

	COMB_TERM (2, 4, 4);
	COMB_TERM (1, 4, 5);
	COMB_TERM (1, 4, 6);

	COMB_TERM (1, 5, 5);

	COMB_TERM (1, 6, 6);

	b [k] = make_comb_rhs (ring, v_not_exist);

	CALIB_FATAL_ERROR_IF (nvars NE 8);

	vars = CALIB_NEWA (ncols, const char *);
	for (i = 0; i < ncols; i++) {
		j = col2edge [i];
		vars [i] = strdup (gip -> xvarnames [j]);
	}

	/* We set the lower bounds correctly for *all* "n" variables,	*/
	/* even those that may not exist.				*/
	lbound = CALIB_NEWA (nvars, int);
	for (j = 0; j < nvars; j++) {
		lbound [j] = 1;
	}
	lbound [6] = 0;
	lbound [7] = 0;

	free (edge2col);

	hsys -> nrows	= nrows;
	hsys -> ncols	= ncols;
	hsys -> vars	= vars;
	hsys -> A	= A;
	hsys -> b	= b;
	hsys -> ring	= ring;
	hsys -> lbound	= lbound;
	hsys -> name	= "hull";

#undef	COMB_TERM
#undef	EDGE

	return (col2edge);
}

/*
 * Make a polynomial term that is coeff*2*NumEdges(u,v).
 */

	static
	struct calib_Zxyz_obj *
make_comb_lhs_term (

struct calib_Zxyz_dom *	ring,	/* IN: polynomial ring to use */
int			coeff,	/* IN: integer multiplier */
int			u,	/* IN: first vertex */
int			v	/* IN: second vertex */
)
{
struct calib_Zxyz_obj *	result;

	result = make_2x_num_edges (ring, u, v);

	mul_by_int (result, coeff);

	return (result);
}

/*
 * Make a polynomial that represents the right-hand-side of the
 * 3-toothed comb inequality.
 */

	static
	struct calib_Zxyz_obj *
make_comb_rhs (

struct calib_Zxyz_dom *	ring,		/* IN: polynomial ring to use */
const calib_bool *	v_not_exist	/* IN: which vertices do not exist */
)
{
int			i;
struct calib_Zxyz_obj *	result;
struct calib_Zxyz_obj *	p;

	result = CALIB_NEW (struct calib_Zxyz_obj);
	ring -> init (ring, result);

	for (i = 0; i <= 6; i++) {
		p = make_var (ring, i);
		if (v_not_exist [i]) continue;
		/* The base of each tooth gets counted twice. */
		if ((i EQ 0) OR (i EQ 2) OR (i EQ 4)) {
			mul_by_int (p, 2);
		}
		ring -> add (result, result, p);
		ring -> free (p);
	}

	add_int_into (result, -5);

	mul_by_int (result, 2);

	return (result);
}

/*
 * Create a polynomial that gives TWO TIMES the number of edges having
 * one vertex of class u and the other vertex of class b.
 *
 * We return 2X the number of edges in order to avoid the factor of 1/2
 * that would otherwise occur in the (u EQ v) case.
 *
 * The result is:
 *
 *		n_u * (n_u - 1)		if (u EQ v)
 *		2 * n_u * n_v		if (u NE v)
 */

	static
	struct calib_Zxyz_obj *
make_2x_num_edges (

struct calib_Zxyz_dom *	ring,	/* IN: polynomial ring to use */
int			u,	/* IN: first vertex class */
int			v	/* IN: second vertex class */
)
{
struct calib_Zxyz_obj	*	p1;
struct calib_Zxyz_obj	*	p2;
struct calib_Zxyz_obj	*	p3;
struct calib_Zxyz_obj *		result;
struct calib_Zxyz_obj		tmp1;

	result = CALIB_NEW (struct calib_Zxyz_obj);
	ring -> init (ring, result);

	p1 = make_var (ring, u);
	ring -> init (ring, &tmp1);

	if (u EQ v) {

		p2 = make_int (ring, 1);
		ring -> init (ring, &tmp1);
		ring -> sub (&tmp1, p1, p2);
		ring -> mul (result, p1, &tmp1);

		ring -> free (p2);
	}
	else {
		p2 = make_var (ring, v);
		p3 = make_int (ring, 2);

		ring -> mul (&tmp1, p1, p2);
		ring -> mul (result, p3, &tmp1);

		ring -> free (p3);
		ring -> free (p2);
	}

	ring -> clear (&tmp1);
	ring -> free (p1);

	return (result);
}

/*
 * Perform Gaussian elimination on the given system.
 */

	static
	void
gauss_eliminate (

struct sys *	sys,		/* IN/OUT: system to perform Gaussian */
				/*	   elimination upon */
const int *	piv_seq		/* IN: pre-specified pivot sequence (or NULL) */
)
{
int				i, j, p, q;
int				nrows, ncols;
int				iter;
int *				rflag;
int *				cflag;
struct calib_Zxyz_obj **	A;
struct PPSState *		state_ptr;
struct PPSState			state;

	CALIB_FATAL_ERROR_IF (sys EQ NULL);

	/* Manage pre-specified pivot sequence, if we were given one. */
	state_ptr = NULL;
	if (piv_seq NE NULL) {
		/* Need state object to keep track of where we are in	*/
		/* the pivot sequence.					*/
		state.piv_seq	= piv_seq;
		state.index	= 0;
		state_ptr = &state;
	}

	nrows	= sys -> nrows;
	ncols	= sys -> ncols;
	A	= sys -> A;

	printf ("Entering gauss_eliminate for `%s' system (%d x %d)\n",
		sys -> name, nrows, ncols);

	CALIB_FATAL_ERROR_IF (nrows > ncols);

	rflag	= CALIB_NEWA (nrows, int);
	for (i = 0; i < nrows; i++) {
		rflag [i] = -1;
	}
	sys -> rflag	= rflag;

	cflag	= CALIB_NEWA (ncols, int);
	for (j = 0; j < ncols; j++) {
		cflag [j] = -1;
	}
	sys -> cflag	= cflag;

	for (iter = 0; iter < nrows; iter++) {
		/* Choose a pivot element A[p,q]. */
		gauss_choose_pivot (sys, &p, &q, state_ptr);

		/* Verify we are not repeating a pivot row or column. */
		CALIB_FATAL_ERROR_IF (rflag [p] >= 0);
		CALIB_FATAL_ERROR_IF (cflag [q] >= 0);

		/* Record pivot taken. */
		rflag [p] = q;
		cflag [q] = p;

		printf (" Pivot %d at (%d,%d):\n", iter + 1, p, q);
		print_factored_maxima (A [p * ncols + q]);

		/* Eliminate down pivot column q. */
		for (i = 0; i < nrows; i++) {
			if (i EQ p) continue;
			gauss_elim_row (sys, p, q, i, iter);
		}
	}

	printf ("Leaving gauss_eliminate for `%s' system (%d x %d)\n",
		sys -> name, nrows, ncols);
}

/*
 * Choose a pivot element for Gaussian elimination.  This is complex
 * because the matrix elements are symbolic polynomials, and we must
 * attempt to filter out pivot candidates whose value can be zero
 * for some valid combination of the symbolic coefficients.
 * These coefficients are n0 through n7.  We can assume:
 * n0, n1, n2, n3, n4, n5 >= 1, and n6, n7 >= 0.
 */

	static
	void
gauss_choose_pivot (

struct sys *		sys,		/* IN: system to choose pivot for */
int *			p,		/* IN/OUT: selected pivot row */
int *			q,		/* IN/OUT: selected pivot column */
struct PPSState *	state_ptr	/* IN/OUT: pre-specified pivot */
					/*	   sequence state object */
					/*	   (or NULL if none) */
)
{
int				i, j, nrows, ncols;
int *				lbound;
int *				rflag;
int *				cflag;
struct calib_Zxyz_obj **	A;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		poly;
struct pcand *			pc;
struct pcand *			plist;
struct pcand **			phookp;

	nrows	= sys -> nrows;
	ncols	= sys -> ncols;
	A	= sys -> A;
	lbound	= sys -> lbound;
	rflag	= sys -> rflag;
	cflag	= sys -> cflag;

	if (state_ptr NE NULL) {
		/* We have been given a pre-deteremined pivot sequence.	*/
		/* Fetch the next (p,q) pivot position from that	*/
		/* sequence.  If the sequence is depleted, fall through	*/
		/* to the automatic pivot selection code.		*/
		i = state_ptr -> index;
		*p = state_ptr -> piv_seq [i++];
		if (*p >= 0) {
			*q = state_ptr -> piv_seq [i++];
			CALIB_FATAL_ERROR_IF (*q < 0);

			state_ptr -> index = i;

			/* Even the pre-determined pivots must be	*/
			/* checked against our validity criteria.	*/
			poly = A [*p * ncols + *q];

			CALIB_FATAL_ERROR_IF (poly EQ NULL);
			if (pivot_can_be_zero (poly, lbound)) {
				printf ("\n"
					"********************************\n"
					" WARNING: PRE-DETERMINED PIVOT"
					" AT (%d,%d) CAN BE ZERO!!!\n"
					"********************************\n",
					*p, *q);
			}
			return;
		}
	}

	plist = NULL;
	phookp = &plist;

	rowp = A;
	for (i = 0; i < nrows; i++, rowp += ncols) {
		if (rflag [i] >= 0) {
			/* Already pivoted in this row.  Skip. */
			continue;
		}
		for (j = 0; j < ncols; j++) {
			if (cflag [j] >= 0) {
				/* Already pivoted in this column.  Skip. */
				continue;
			}
			poly = rowp [j];
			if ((poly EQ NULL) OR (poly -> nterms <= 0)) {
				/* Element is zero. */
				continue;
			}
			/* We have a potential pivot candidate.  Record	*/
			/* it together with its "size" information.	*/
			pc = CALIB_NEW (struct pcand);
			memset (pc, 0, sizeof (*pc));
			pc -> prow	= i;
			pc -> pcol	= j;
			get_poly_size (&(pc -> psize), poly);
			pc -> next	= NULL;

			*phookp = pc;
			phookp = &(pc -> next);
		}
	}

	CALIB_FATAL_ERROR_IF (plist EQ NULL);

	/* Sort the pivot candidates from "simplest" to "most complex". */
	plist = sort_pivot_candidates (plist);

	for (pc = plist; pc NE NULL; pc = pc -> next) {
		i = pc -> prow;
		j = pc -> pcol;
		poly = A [i * ncols + j];
		CALIB_FATAL_ERROR_IF (poly EQ NULL);

		if (NOT pivot_can_be_zero (poly, lbound)) {
			break;
		}
	}

	/* No suitable pivot could be found. */
	CALIB_FATAL_ERROR_IF (pc EQ NULL);

	*p = pc -> prow;
	*q = pc -> pcol;

	while (plist NE NULL) {
		pc = plist;
		plist = pc -> next;
		pc -> next = NULL;
		free (pc);
	}
}

/*
 * Compute the "size" information for the given polynomial.
 */

	static
	void
get_poly_size (

struct psize *		psize,	/* IN/OUT: pivot size info */
struct calib_Zxyz_obj *	p	/* IN: polynomial to get size info for */
)
{
int		i, k, nvars, nterms;
int		totaldeg, hipow, nbits, sign;
int *		maxdeg;

	nterms	= p -> nterms;
	nvars	= p -> dom -> nvars;

	maxdeg = compute_var_max_degrees (p, &totaldeg);

	hipow = 0;
	for (i = 0; i < nvars; i++) {
		k = maxdeg [i];
		if (k > hipow) {
			hipow = k;
		}
	}

	nbits = 0;
	for (i = 0; i < nterms; i++) {
		k = mpz_sizeinbase (&(p -> coeff [i]), 2);
		if (k > nbits) {
			nbits = k;
		}
	}
	sign = (nterms > 0) ? mpz_sgn (p -> coeff) : 0;

	memset (psize, 0, sizeof (*psize));

	psize -> nterms		= p -> nterms;
	psize -> maxdeg		= hipow;
	psize -> totdeg		= totaldeg;
	psize -> nbits		= nbits;
	psize -> sign		= sign;

	free (maxdeg);
}

/*
 * Compute the maximum degree of each variable,
 * as well as the total degree of the given polynomial.
 */

	static
	int *
compute_var_max_degrees (

struct calib_Zxyz_obj *
		p,		/* IN: polynomial to compute degrees for */
int *		totaldeg	/* IN/OUT: total degree of polynomial */
)
{
int		i, j, k, nvars, nterms, totdeg, maxtotdeg;
int *		maxdeg;
int *		pow;

	CALIB_FATAL_ERROR_IF (p EQ NULL);

	nterms	= p -> nterms;
	nvars	= p -> dom -> nvars;

	maxdeg = CALIB_NEWA (nvars, int);
	memset (maxdeg, 0, nvars * sizeof (maxdeg [0]));

	maxtotdeg = 0;

	pow = p -> pow;
	for (i = 0; i < nterms; i++, pow += nvars) {
		totdeg = 0;
		for (j = 0; j < nvars; j++) {
			k = pow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			if (k > maxdeg [j]) {
				maxdeg [j] = k;
			}
			totdeg += k;
		}
		if (totdeg > maxtotdeg) {
			maxtotdeg = totdeg;
		}
	}

	if (totaldeg NE NULL) {
		*totaldeg = maxtotdeg;
	}

	return (maxdeg);
}

/*
 * Sort the given list of pivot candidates in order from
 * "simplest" to "most complex".
 */

	static
	struct pcand *
sort_pivot_candidates (

struct pcand *	plist		/* IN: list of pivot candidates to sort */
)
{
int		i, j, k, n, ncand;
struct pcand *	p1;
struct pcand *	p2;
struct pcand **	array;
struct pcand *	root;
struct pcand **	hookp;

	CALIB_FATAL_ERROR_IF (plist EQ NULL);
	ncand = 0;
	for (p1 = plist; p1 NE NULL; p1 = p1 -> next) {
		++ncand;
	}

	array = CALIB_NEWA (ncand, struct pcand *);
	i = 0;
	for (p1 = plist; p1 NE NULL; p1 = p1 -> next) {
		array [i] = p1;
		++i;
	}
	CALIB_FATAL_ERROR_IF (i NE ncand);

	/* Construct the heap via sift-downs in O(n) time. */
	n = ncand;
	for (k = n / 2; k >= 0; k--) {
		j = k;
		for (;;) {
			i = 2 * j + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the left subchild preceeds the right. */
				p1 = array [i];
				p2 = array [i + 1];
				if (pivots_in_order (p1, p2)) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			p1 = array [j];
			p2 = array [i];
			if (pivots_in_order (p2, p1)) {
				/* Child node preceeds parent, sift-down */
				/* is done.				 */
				break;
			}
			/* Sift down and continue. */
			array [j] = p2;
			array [i] = p1;
			j = i;
		}
	}

	/* Now do the actual sorting.  Exchange first/last and sift down. */
	while (n > 1) {
		/* Item at array[0] is greatest among all items remaining */
		/* in the heap (i.e., must come last in the sorted order). */
		/* Swap first and last items, and decrease heap size by one. */
		--n;
		p1 = array [0];
		array [0] = array [n];
		array [n] = p1;

		/* Now restore the heap property by sifting array[0] down. */
		j = 0;
		for (;;) {
			i = 2 * j + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the left subchild preceeds the right. */
				p1 = array [i];
				p2 = array [i + 1];
				if (pivots_in_order (p1, p2)) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			p1 = array [j];
			p2 = array [i];
			if (pivots_in_order (p2, p1)) {
				/* Child node preceeds parent, sift-down */
				/* is done.				 */
				break;
			}
			/* Sift down and continue. */
			array [j] = p2;
			array [i] = p1;
			j = i;
		}
	}

	/* Thread the nodes back together into a list in final order. */
	root = NULL;
	hookp = &root;
	for (i = 0; i < ncand; i++) {
		p1 = array [i];
		p1 -> next = NULL;
		*hookp = p1;
		hookp = &(p1 -> next);
	}

	free (array);

	return (root);
}

/*
 * Return TRUE if-and-only-if pivot `p1' is "simpler" that pivot `p2',
 * and should be considered first.
 */

	static
	calib_bool
pivots_in_order (

const struct pcand *	p1,	/* IN: First pivot candidate */
const struct pcand *	p2	/* IN: Second pivot candidate */
)
{
int		cmp;

	cmp = compare_poly_sizes (&(p1 -> psize), &(p2 -> psize));

	if (cmp < 0) return (TRUE);
	if (cmp > 0) return (FALSE);

	if (p1 -> prow < p2 -> prow) return (TRUE);
	if (p1 -> prow > p2 -> prow) return (FALSE);

	if (p1 -> pcol < p2 -> pcol) return (TRUE);
	if (p1 -> pcol > p2 -> pcol) return (FALSE);

	/* Should not have two copies of the same pivot! */
	CALIB_FATAL_ERROR;

	return (FALSE);
}

/*
 * Compare the two given polynomial "size" objects, and return
 * -1, 0, or +1, as if doing *p1 - *p2.
 */

	static
	int
compare_poly_sizes (

const struct psize *	p1,	/* IN: size info for first pivot */
const struct psize *	p2	/* IN: size info for second pivot */
)
{
	CALIB_FATAL_ERROR_IF ((p1 EQ NULL) OR (p2 EQ NULL));

	if (p1 -> nterms < p2 -> nterms) return (-1);
	if (p1 -> nterms > p2 -> nterms) return (+1);

	if (p1 -> maxdeg < p2 -> maxdeg) return (-1);
	if (p1 -> maxdeg > p2 -> maxdeg) return (+1);

	if (p1 -> totdeg < p2 -> totdeg) return (-1);
	if (p1 -> totdeg > p2 -> totdeg) return (+1);

	if (p1 -> nbits < p2 -> nbits) return (-1);
	if (p1 -> nbits > p2 -> nbits) return (+1);

	if (p1 -> sign > p2 -> sign) return (-1);
	if (p1 -> sign > p2 -> sign) return (+1);

	return (0);
}

/*
 * Eliminate A[i,q] using pivot element A[p,q].
 * Peforms the entire row operation and reduces all common
 * factors that may arise across row i.
 */

	static
	void
gauss_elim_row (

struct sys *		sys,		/* IN/OUT: system to eliminate in */
int			p,		/* IN: index of pivot row */
int			q,		/* IN: index of pivot column */
int			i,		/* IN: row to modify via elimination */
int			giter		/* IN: pivot iteration we are executing */
)
{
int				j, k;
int				ncols;
struct calib_Zxyz_obj **	A;
struct calib_Zxyz_obj **	b;
struct calib_Zxyz_obj *		elem;
struct calib_Zxyz_obj **	rowi;
struct calib_Zxyz_obj **	pivrow;
struct calib_Zxyz_obj *		piv;
struct calib_Zxyz_obj *		ppp;
struct calib_Zxyz_obj *		pip;
struct calib_Zxyz_obj **	pptrs;
struct calib_Zxyz_dom *		ring;
struct calib_Zxyz_obj		red_piv, red_elem, red_gcd, PROD1, PROD2;

	ncols	= sys -> ncols;
	A	= sys -> A;
	b	= sys -> b;
	ring	= sys -> ring;

	elem	= A [i * ncols + q];
	if (elem EQ NULL) {
		/* The element we wish to eliminate is already	*/
		/* NULL (implicitly zero).  We are done!	*/
		return;
	}
	if (elem -> nterms <= 0) {
		/* The element we wish to eliminate is present,	*/
		/* and already has zero value.  Free it up and	*/
		/* NULL it out.					*/

		CALIB_FATAL_ERROR_IF (elem -> nterms < 0);

		A [i * ncols + q] = NULL;
		ring -> free (elem);
		return;
	}

	/* Need to do actual elimination across row i. */

	rowi	= &A [i * ncols];
	pivrow	= &A [p * ncols];
	piv	= pivrow [q];

	CALIB_FATAL_ERROR_IF (piv EQ NULL);

	/* This is important -- "elem" and "piv" can have common	*/
	/* factors, and these must be removed before doing the		*/
	/* elimination.  Even though we only have two polynomials, we	*/
	/* go ahead and use the "gcd_n" operation because it also	*/
	/* provides us the cofactors.					*/

	ring -> init (ring, &red_piv);
	ring -> init (ring, &red_elem);
	ring -> init (ring, &PROD1);
	ring -> init (ring, &PROD2);
	ring -> init (ring, &red_gcd);

	{ const struct calib_Zxyz_obj * poly_ptrs [2];
	  struct calib_Zxyz_obj * cf_ptrs [2];
	  poly_ptrs [0] = piv;
	  cf_ptrs   [0] = &red_piv;
	  poly_ptrs [1] = elem;
	  cf_ptrs   [1] = &red_elem;
	  ring -> gcd_n (&red_gcd, &cf_ptrs [0], 2, &poly_ptrs [0]);
	}

	/* Create a list of the columns to reduce when we're done. */
	pptrs = CALIB_NEWA (ncols + 1, struct calib_Zxyz_obj *);
	k = 0;

	/* Perform the elimination. */
	for (j = 0; j < ncols; j++) {
		ppp = pivrow [j];
		pip = rowi [j];
		if (j EQ q) {
			/* Skip doing the math -- the result is zero. */
			if (pip NE NULL) {
				ring -> free (pip);
				rowi [j] = NULL;
			}
			continue;
		}
		if (ppp EQ NULL) {
			if (pip EQ NULL) {
				/* Both elements are NULL.  Skip! */
				continue;
			}
			PROD1.nterms = 0;
		}
		else {
			ring -> mul (&PROD1, &red_elem, ppp);
		}
		if (pip EQ NULL) {
			PROD2.nterms = 0;
		}
		else {
			ring -> mul (&PROD2, &red_piv, pip);
		}
		ring -> sub (&PROD2, &PROD2, &PROD1);
		if (PROD2.nterms EQ 0) {
			/* New element is zero. */
			if (pip NE NULL) {
				ring -> free (pip);
				rowi [j] = NULL;
			}
		}
		else {
			if (pip NE NULL) {
				ring -> set (pip, &PROD2);
			}
			else {
				pip = ring -> dup (&PROD2);
				rowi [j] = pip;
			}
			pptrs [k++] = pip;
		}
	}

	/* Eliminate right-hand-sides. */
	ppp = b [p];
	pip = b [i];
	if (ppp EQ NULL) {
		PROD1.nterms = 0;
	}
	else {
		ring -> mul (&PROD1, &red_elem, ppp);
	}
	if (pip EQ NULL) {
		PROD2.nterms = 0;
	}
	else {
		ring -> mul (&PROD2, &red_piv, pip);
	}
	ring -> sub (&PROD2, &PROD2, &PROD1);
	if (PROD2.nterms EQ 0) {
		/* New element is zero. */
		if (pip NE NULL) {
			ring -> free (pip);
			b [i] = NULL;
		}
	}
	else {
		if (pip NE NULL) {
			ring -> set (pip, &PROD2);
		}
		else {
			pip = ring -> dup (&PROD2);
			b [i] = pip;
		}
		pptrs [k++] = pip;
	}

	if (k >= 2) {
		/* Eliminate common factors across entire row. */
		ring -> gcd_n (&red_gcd,
			       &pptrs [0],
			       k,
			       (const struct calib_Zxyz_obj **) &pptrs [0]);
	}

	free (pptrs);
	ring -> clear (&red_gcd);
	ring -> clear (&PROD2);
	ring -> clear (&PROD1);
	ring -> clear (&red_elem);
	ring -> clear (&red_piv);
}

/*
 * Compute the set of variable substitutions implied by the
 * solution of the set of affine hull equations.
 */

	static
	struct calib_subst_list *
compute_hull_substitutions (

struct ginfo *		gip,		/* IN/OUT: global info */
struct sys *		hsys,		/* IN: affine hull system */
const int *		col2edge	/* IN: edge number for each column */
)
{
int				i, j, q;
int				nrows, ncols, nverts;
int *				rflag;
int *				cflag;
struct calib_Zxyz_dom *		nxring;
struct calib_rat_dom *		nxrat;
struct calib_subst_list *	slist;
struct calib_rsubst *		array;
struct calib_rsubst *		rsp;
struct calib_Zxyz_obj **	A;
struct calib_Zxyz_obj **	b;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		numer;
struct calib_Zxyz_obj *		denom;
struct calib_Zxyz_obj *		piv;
struct calib_Zxyz_obj *		coeff;
struct calib_Zxyz_obj *		xvarp;
struct calib_Zxyz_obj		TERM;

	printf ("Entering compute_hull_substitutions\n");

	nverts	= gip -> nverts;
	nxring	= gip -> nxring;
	nxrat	= gip -> nxrat;

	nrows	= hsys -> nrows;
	ncols	= hsys -> ncols;
	A	= hsys -> A;
	b	= hsys -> b;
	rflag	= hsys -> rflag;
	cflag	= hsys -> cflag;

	array = CALIB_NEWA (nrows, struct calib_rsubst);
	memset (array, 0, nrows * sizeof (array [0]));

	nxring -> init (nxring, &TERM);

	rowp = A;
	for (i = 0; i < nrows; i++, rowp += ncols) {
		rsp = &array [i];

		q = rflag [i];
		CALIB_FATAL_ERROR_IF ((q < 0) OR (q >= ncols));

		piv = rowp [q];
		CALIB_FATAL_ERROR_IF (piv EQ NULL);

		rsp -> var = nverts + col2edge [q];
		nxrat -> init (nxrat, &(rsp -> val));

		numer = &(rsp -> val.numer);
		denom = &(rsp -> val.denom);

		/* Convert right-hand side from nring to nxring. */
		coeff = b [i];
		if (coeff NE NULL) {
			nxring -> convert_with_varmap (numer,
						       coeff,
						       gip -> nx2n_varmap);
		}

		for (j = 0; j < ncols; j++) {
			coeff = rowp [j];
			if (cflag [j] >= 0) {
				/* This column has been eliminated. */
				if (j EQ q) {
					CALIB_FATAL_ERROR_IF (coeff NE piv);
				}
				else {
					CALIB_FATAL_ERROR_IF (coeff NE NULL);
				}
				continue;
			}
			if (coeff EQ NULL) continue;

			/* Convert coefficient from nring to nxring. */
			nxring -> convert_with_varmap (&TERM,
						       coeff,
						       gip -> nx2n_varmap);

			/* Multiply by the "x" var for this column. */
			xvarp = make_var (nxring, nverts + col2edge [j]);
			nxring -> mul (&TERM, &TERM, xvarp);
			nxring -> free (xvarp);

			/* Subtract from right-hand-side. */
			nxring -> sub (numer, numer, &TERM);
		}

		/* Convert the pivot from nring to nxring. */
		nxring -> convert_with_varmap (denom, piv, gip -> nx2n_varmap);

		nxrat -> canonicalize (&(rsp -> val));
	}

	nxring -> clear (&TERM);

	slist = CALIB_NEW (struct calib_subst_list);
	slist -> num_subst	= nrows;
	slist -> subst		= array;

	printf ("Leaving compute_hull_substitutions\n");

	return (slist);
}

/*
 * Apply the substitutions obtained from the hull system of equations
 * to the distance-squared expression in order to eliminate these
 * x variables.
 *
 * There is no change of ring here -- everything stays in the nxring,
 * because many x vars still remain.  Fortunately, the denominators
 * wind up being not very complex.
 */

	static
	void
perform_hull_substitutions (

struct calib_rat_obj *	result,	/* IN/OUT: result			*/
const struct calib_rat_obj *
			ratp,	/* IN: rational function to apply	*/
				/*	   substitutions to		*/
struct calib_subst_list *
			slist,	/* IN: list of substitutions to apply	*/
struct ginfo *		gip	/* IN/OUT: global info			*/
)
{
int				i, k, n, nvars, nverts;
int *				varmap;
const struct calib_Zxyz_dom *	nxring;
const struct calib_rat_dom *	nxrat;
struct calib_rsubst *		rsp;

	CALIB_FATAL_ERROR_IF ((gip   EQ NULL) OR
			      (ratp  EQ NULL) OR
			      (slist EQ NULL));

	printf ("Entering perform_hull_substitutions (%d substitutions)\n",
		slist -> num_subst);

	nverts	= gip -> nverts;
	nxring	= gip -> nxring;
	nxrat	= gip -> nxrat;
	nvars	= nxring -> nvars;

	/* Start with a varmap that is a no-op. */
	varmap = CALIB_NEWA (nvars, int);
	for (i = 0; i < nvars; i++) {
		varmap [i] = i;
	}

	/* Remove from the varmap all vars that are being	*/
	/* substituted away.					*/
	n	= slist -> num_subst;
	rsp	= slist -> subst;
	for (i = 0; i < n; i++, rsp++) {
		k = rsp -> var;
		CALIB_FATAL_ERROR_IF ((k < nverts) OR (k >= nvars));
		CALIB_FATAL_ERROR_IF (varmap [k] < 0);
		varmap [k] = -1;
	}

	/* Perform substitution with respect to these domains. */
	nxrat -> substitute_with_varmap (result, ratp, varmap, slist);

	free (varmap);

	printf ("Leaving perform_hull_substitutions\n");
}

/*
 * Create a map from the space of all x variables down to the smaller space
 * of x variables that have not been eliminated, and also a reverse map.
 *
 * The hull system tells us which x vars were eliminated.
 */

	static
	void
create_xvar_map (

struct ginfo *		gip,		/* IN: global info */
const struct sys *	hsys,		/* IN: hull system */
const int *		col2edge	/* IN: maps column ==> edge */
)
{
int		i, j, k, e;
int		nedges;
int		nrows, ncols;
int		numrx;
int *		cflag;
const char **	rxvarnames;
int *		x_to_rx;
int *		rx_to_x;

	nedges	= gip -> nedges;

	nrows	= hsys -> nrows;
	ncols	= hsys -> ncols;
	cflag	= hsys -> cflag;

	/* Compute number of remaining x vars.				*/
	/* (Note: ncols = number of edges that exist, nrows = number	*/
	/* of edges that get eliminated by solving the hull system.)	*/
	numrx = ncols - nrows;

	rxvarnames = CALIB_NEWA (numrx, const char *);

	x_to_rx = CALIB_NEWA (nedges, int);
	rx_to_x = CALIB_NEWA (numrx, int);

	for (i = 0; i < nedges; i++) {
		x_to_rx [i] = -1;
	}

	j = 0;
	for (i = 0; i < ncols; i++) {
		k = cflag [i];
		if (k < 0) {
			/* Column i did *NOT* get eliminated. */
			e = col2edge [i];
			rxvarnames [j] = strdup (gip -> xvarnames [e]);
			x_to_rx [e] = j;
			rx_to_x [j] = e;
			++j;
		}
	}
	CALIB_FATAL_ERROR_IF (j NE numrx);

	gip -> numrx		= numrx;
	gip -> rxvarnames	= rxvarnames;
	gip -> rx_to_x		= rx_to_x;
	gip -> x_to_rx		= x_to_rx;
}

/*
 * We compute the gradient of dist2 (with respect to each of the
 * remaining xvars).  We want the gradient to be zero, which gives
 * rise to a system of equations that we set up here.
 */

	static
	void
create_gradient_system (

struct ginfo *		gip,	/* IN: global info */
struct sys *		gsys,	/* IN/OUT: gradient system */
const struct calib_rat_obj *
			dist2	/* IN: dist2 expression */
)
{
int				i, j, k, m, n, q;
int				nverts, numrx, var;
int				xvar, xvcount, konst_count;
struct calib_Zxyz_dom *		nring;
struct calib_Zxyz_dom *		nxring;
int *				rx_to_x;
int *				x_to_rx;
struct calib_Zxyz_obj **	A;
struct calib_Zxyz_obj **	b;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		rxcoeff;
struct calib_Zxyz_obj *		dst;
int *				rxvarcount;
int *				dpow;
int *				lbound;
struct calib_Zxyz_obj **	poly_ptrs;
struct calib_Zxyz_obj		DERIV;

	nverts	= gip -> nverts;
	nring	= gip -> nring;
	nxring	= gip -> nxring;
	numrx	= gip -> numrx;
	rx_to_x	= gip -> rx_to_x;
	x_to_rx	= gip -> x_to_rx;

	k = numrx * numrx;
	A = CALIB_NEWA (k, struct calib_Zxyz_obj *);
	b = CALIB_NEWA (numrx, struct calib_Zxyz_obj *);
	memset (A, 0, k * sizeof (A [0]));
	memset (b, 0, numrx * sizeof (b [0]));

	nxring -> init (nxring, &DERIV);

	rxvarcount = CALIB_NEWA (numrx, int);
	poly_ptrs = CALIB_NEWA (numrx + 1, struct calib_Zxyz_obj *);

	rxcoeff = CALIB_NEWA (numrx + 1, struct calib_Zxyz_obj);
	for (i = 0; i <= numrx; i++) {
		nring -> init (nring, &rxcoeff [i]);
	}

	rowp = A;
	for (i = 0; i < numrx; i++, rowp += numrx) {
		var = nverts + rx_to_x [i];
		partial_derivative (&DERIV, &(dist2 -> numer), var);

		/* Scan the terms of the dervative.  Each term should	*/
		/* have at most one "x" variable in it.  If present,	*/
		/* this "x" var should have degree exactly 1.  Count	*/
		/* the number of terms for each such x variable, and	*/
		/* count those "constant" terms that have no "x" vars.	*/

		memset (rxvarcount, 0, numrx * sizeof (int));
		konst_count = 0;

		n = nxring -> nvars;
		dpow = DERIV.pow;
		for (j = 0; j < DERIV.nterms; j++, dpow += n) {
			xvcount = 0;
			xvar = -1;
			for (k = nverts; k < n; k++) {
				m = dpow [k];
				if (m <= 0) {
					CALIB_FATAL_ERROR_IF (m < 0);
					continue;
				}
				CALIB_FATAL_ERROR_IF (m NE 1);
				++xvcount;
				xvar = k;
			}
			if (xvcount > 0) {
				CALIB_FATAL_ERROR_IF (xvcount NE 1);
				q = x_to_rx [xvar - nverts];
				CALIB_FATAL_ERROR_IF (q < 0);
				++(rxvarcount [q]);
			}
			else {
				/* No x vars, so it is a constant. */
				++konst_count;
			}
		}

		/* Empty each bucket and allocate sufficient terms	*/
		/* for each.						*/
		for (j = 0; j < numrx; j++) {
			rxcoeff [j].nterms = 0;
			nring -> alloc (&rxcoeff [j], rxvarcount [j]);
		}
		rxcoeff [numrx].nterms = 0;
		nring -> alloc (&rxcoeff [numrx], konst_count);

		/* Scan the terms again, this time appending each to	*/
		/* the proper destination bucket.			*/
		dpow = DERIV.pow;
		for (j = 0; j < DERIV.nterms; j++, dpow += n) {
			xvcount = 0;
			xvar = -1;
			for (k = nverts; k < n; k++) {
				m = dpow [k];
				if (m <= 0) {
					CALIB_FATAL_ERROR_IF (m < 0);
					continue;
				}
				CALIB_FATAL_ERROR_IF (m NE 1);
				++xvcount;
				xvar = k;
			}
			if (xvcount > 0) {
				CALIB_FATAL_ERROR_IF (xvcount NE 1);
				q = x_to_rx [xvar - nverts];
				CALIB_FATAL_ERROR_IF (q < 0);
				dst = &rxcoeff [q];
			}
			else {
				/* No x vars, so it is a constant. */
				dst = &rxcoeff [numrx];
			}

			/* Append the term to the destination bucket.	*/
			/* Note that we discard all of the powers for	*/
			/* the x vars while copying.			*/
			q = dst -> nterms;
			memcpy (&(dst -> pow [q * nverts]),
				&dpow [0],
				nverts * sizeof (int));
			mpz_set (&(dst -> coeff [q]), &DERIV.coeff [j]);
			dst -> nterms = q + 1;
		}

		/* Verify that the correct number of terms appeared in	*/
		/* each output bucket.					*/
		for (j = 0; j < numrx; j++) {
			q = rxcoeff [j].nterms;
			CALIB_FATAL_ERROR_IF (q > rxcoeff [j].size);
			CALIB_FATAL_ERROR_IF (q NE rxvarcount [j]);
		}
		q = rxcoeff [numrx].nterms;
		CALIB_FATAL_ERROR_IF (q > rxcoeff [numrx].size);
		CALIB_FATAL_ERROR_IF (q NE konst_count);

		/* Distribute the coefficients into this row. */
		k = 0;
		for (j = 0; j < numrx; j++) {
			if (rxcoeff [j].nterms <= 0) continue;
			dst = nring -> dup (&rxcoeff [j]);
			rowp [j] = dst;
			poly_ptrs [k++] = dst;
		}
		if (rxcoeff [numrx].nterms > 0) {
			/* Negate constants and move to RHS. */
			mul_by_int (&rxcoeff [numrx], -1);
			dst = nring -> dup (&rxcoeff [numrx]);
			b [i] = dst;
			poly_ptrs [k++] = dst;
		}

		/* Cancel any common factors across the row. */
		nring -> gcd_n (&rxcoeff [0],
				&poly_ptrs [0],
				k,
				(const struct calib_Zxyz_obj **) &poly_ptrs [0]);
	}

	for (i = 0; i <= numrx; i++) {
		nring -> clear (&rxcoeff [i]);
	}
	free (rxcoeff);
	free (poly_ptrs);
	free (rxvarcount);

	nxring -> clear (&DERIV);

	lbound = CALIB_NEWA (nring -> nvars, int);
	for (j = 0; j < nring -> nvars; j++) {
		lbound [j] = 1;
	}
	lbound [6] = 0;
	lbound [7] = 0;

	memset (gsys, 0, sizeof (*gsys));

	gsys -> nrows	= numrx;
	gsys -> ncols	= numrx;
	gsys -> vars	= copy_varlist (gip -> rxvarnames, numrx);
	gsys -> A	= A;
	gsys -> b	= b;
	gsys -> ring	= nring;
	gsys -> lbound	= lbound;
	gsys -> name	= "gradient";
}

/*
 * Compute the partial derivative of the given polynomial with respect
 * to the given variable.
 */

	static
	void
partial_derivative (

struct calib_Zxyz_obj *	result,	/* IN/OUT: resulting derivative */
const struct calib_Zxyz_obj *
			poly,	/* IN: polynomial to take derivative of */
int			var	/* IN: var to differentiate with respect to */
)
{
int				i, j, k;
int				nterms, nvars;
const struct calib_Zxyz_dom *	ring;
int *				pow;
mpz_ptr				coeff;
int *				ipow;
int *				tpow;
struct calib_Zxyz_obj		DERIV;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (poly EQ NULL));

	ring = poly -> dom;
	CALIB_FATAL_ERROR_IF (ring NE result -> dom);

	nterms	= poly -> nterms;
	pow	= poly -> pow;
	coeff	= poly -> coeff;
	nvars	= ring -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	/* Count the number of terms that will appear in the derivative. */
	k = 0;
	ipow = pow;
	for (i = 0; i < nterms; i++, ipow += nvars) {
		if (ipow [var] > 0) {
			++k;
		}
	}

	/* Set up temporary to hold the derivative, and pre-allocate	*/
	/* all of the terms that will be needed.			*/
	ring -> init (ring, &DERIV);
	ring -> alloc (&DERIV, k);

	tpow = DERIV.pow;

	ipow = pow;
	j = 0;
	for (i = 0; i < nterms; i++, ipow += nvars) {
		k = ipow [var];
		if (k <= 0) {
			CALIB_FATAL_ERROR_IF (k < 0);
			continue;
		}
		memcpy (tpow, ipow, nvars * sizeof (int));
		tpow [var] = k - 1;
		mpz_mul_ui (&DERIV.coeff [j], &coeff [i], k);
		tpow += nvars;
		++j;
	}
	DERIV.nterms = j;
	CALIB_FATAL_ERROR_IF (j NE DERIV.size);

	ring -> set (result, &DERIV);

	ring -> clear (&DERIV);
}

/*
 * Clear out the given system of equations, freeing up all of the
 * memory.  Does *not* free the given pointer.
 */

	static
	void
clear_system (

struct sys *	sys		/* IN: system of equations to clear */
)
{
int				i, j, nrows, ncols;
struct calib_Zxyz_dom *		ring;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		p;

	CALIB_FATAL_ERROR_IF (sys EQ NULL);

	nrows	= sys -> nrows;
	ncols	= sys -> ncols;
	ring	= sys -> ring;

	if (sys -> rflag NE NULL) {
		free (sys -> rflag);
	}
	if (sys -> cflag NE NULL) {
		free (sys -> cflag);
	}
	if (sys -> lbound NE NULL) {
		free (sys -> lbound);
	}
	if (sys -> vars NE NULL) {
		free_varlist (sys -> vars, ncols);
	}
	rowp = sys -> A;
	if (rowp NE NULL) {
		for (i = 0; i < nrows; i++, rowp += ncols) {
			for (j = 0; j < ncols; j++) {
				p = rowp [j];
				if (p NE NULL) {
					ring -> free (p);
				}
			}
		}
		free (sys -> A);
	}
	rowp = sys -> b;
	if (rowp NE NULL) {
		for (i = 0; i < nrows; i++) {
			p = rowp [i];
			if (p NE NULL) {
				ring -> free (p);
			}
		}
		free (rowp);
	}

	memset (sys, 0, sizeof (*sys));
}

/*
 * Free up the memory held by the given ginfo object.
 */

	static
	void
clear_ginfo (

struct ginfo *		gip	/* IN/OUT: global info */
)
{
	CALIB_FATAL_ERROR_IF (gip EQ NULL);

	CALIB_FREE (gip -> x_to_rx);
	CALIB_FREE (gip -> rx_to_x);
	if (gip -> rxvarnames NE NULL) {
		free_varlist (gip -> rxvarnames, gip -> numrx);
	}
	calib_free_rat_dom (gip -> nxrat);
	calib_free_rat_dom (gip -> nrat);
	if (gip -> nxring NE NULL) {
		calib_free_Zxyz_dom (gip -> nxring);
	}
	if (gip -> nring NE NULL) {
		calib_free_Zxyz_dom (gip -> nring);
	}
	CALIB_FREE (gip -> edge_v2);
	CALIB_FREE (gip -> edge_v1);
	CALIB_FREE (gip -> v_to_e);
	if (gip -> xvarnames NE NULL) {
		free_varlist (gip -> xvarnames, gip -> nedges);
	}
	if (gip -> nvarnames NE NULL) {
		free_varlist (gip -> nvarnames, gip -> nverts);
	}

	memset (gip, 0, sizeof (*gip));
}

/*
 * Return TRUE if there is a valid assignment of integer values to the
 * variables of the given polynomial that causes it to be zero.
 * (An assignment is valid if it respects the given lower bounds.)
 *
 * Note: This is a variant of Hilbert's 10th problem, and is therefore
 * impossible to solve!  Our approach tries pretty hard to filter out
 * most cases, but there is a chance that this algorithm returns FALSE
 * when it should instead return TRUE.
 *
 * The polynomials should be in variables n0, n1, n2, n3, n4, n5, n6, n7
 * which must all have integer values.  n0...n5 >= 1, and n6,n7 >= 0.
 */

	static
	calib_bool
pivot_can_be_zero (

const struct calib_Zxyz_obj *
			p,	/* IN: polynomial to test */
int *			lbound	/* IN: variable lower bounds */
)
{
calib_bool	can_be_zero;

	/* Use our batch of heuristics. */
	can_be_zero = hilbert_10_heuristic (p, lbound);

	return (can_be_zero);
}

/*
 * Compute the set of variable substitutions implied by the
 * solution of the set of gradient equations.
 */

	static
	struct calib_subst_list *
compute_gradient_substitutions (

struct ginfo *		gip,		/* IN/OUT: global info */
struct sys *		gsys		/* IN: gradient system */
)
{
int				i, q;
int				nrows, ncols, nverts;
int *				rflag;
int *				rx_to_x;
struct calib_Zxyz_dom *		nxring;
struct calib_rat_dom *		nxrat;
struct calib_subst_list *	slist;
struct calib_rsubst *		array;
struct calib_rsubst *		rsp;
struct calib_Zxyz_obj **	A;
struct calib_Zxyz_obj **	b;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		numer;
struct calib_Zxyz_obj *		denom;
struct calib_Zxyz_obj *		piv;
struct calib_Zxyz_obj *		coeff;

	printf ("Entering compute_gradient_substitutions\n");

	nverts	= gip -> nverts;
	nxring	= gip -> nxring;
	nxrat	= gip -> nxrat;
	rx_to_x	= gip -> rx_to_x;

	nrows	= gsys -> nrows;
	ncols	= gsys -> ncols;
	A	= gsys -> A;
	b	= gsys -> b;
	rflag	= gsys -> rflag;

	array = CALIB_NEWA (nrows, struct calib_rsubst);
	memset (array, 0, nrows * sizeof (array [0]));

	rowp = A;
	for (i = 0; i < nrows; i++, rowp += ncols) {
		rsp = &array [i];

		q = rflag [i];
		CALIB_FATAL_ERROR_IF ((q < 0) OR (q >= ncols));

		piv = rowp [q];
		CALIB_FATAL_ERROR_IF (piv EQ NULL);

		rsp -> var = nverts + rx_to_x [q];
		nxrat -> init (nxrat, &(rsp -> val));

		numer = &(rsp -> val.numer);
		denom = &(rsp -> val.denom);

		/* Convert right-hand side from nring to nxring. */
		coeff = b [i];
		if (coeff NE NULL) {
			nxring -> convert_with_varmap (numer,
						       coeff,
						       gip -> nx2n_varmap);
		}

		/* Convert the pivot from nring to nxring. */
		nxring -> convert_with_varmap (denom, piv, gip -> nx2n_varmap);

		nxrat -> canonicalize (&(rsp -> val));
	}

	slist = CALIB_NEW (struct calib_subst_list);
	slist -> num_subst	= nrows;
	slist -> subst		= array;

	printf ("Leaving compute_gradient_substitutions\n");

	return (slist);
}

/*
 * Apply the substitutions obtained from the gradient system of equations
 * to the affine distance-squared in order to eliminate the remaining
 * x variables.
 *
 * The ring changes from nxring down to nring.
 */

	static
	void
perform_gradient_substitutions (

struct calib_rat_obj *	result,	/* IN/OUT: result			*/
const struct calib_rat_obj *
			ratp,	/* IN: rational function to apply	*/
				/*	   substitutions to		*/
struct calib_subst_list *
			slist,	/* IN: list of substitutions to apply	*/
struct ginfo *		gip	/* IN/OUT: global info			*/
)
{
int				i, nvars, nverts;
int *				varmap;
const struct calib_Zxyz_dom *	nxring;
const struct calib_rat_dom *	nxrat;

	CALIB_FATAL_ERROR_IF ((gip   EQ NULL) OR
			      (ratp  EQ NULL) OR
			      (slist EQ NULL));

	printf ("Entering perform_gradient_substitutions (%d substitutions)\n",
		slist -> num_subst);

	nverts	= gip -> nverts;
	nxring	= gip -> nxring;
	nxrat	= gip -> nxrat;
	nvars	= nxring -> nvars;

	/* Start with a varmap that eliminates everything. */
	varmap = CALIB_NEWA (nvars, int);
	for (i = 0; i < nvars; i++) {
		varmap [i] = -1;
	}

	/* Map the "n" variables to themselves. */
	for (i = 0; i < nverts; i++) {
		varmap [i] = i;
	}

	/* Perform substitution with respect to these domains. */
	nxrat -> substitute_with_varmap (result, ratp, varmap, slist);

	free (varmap);

	printf ("Leaving perform_gradient_substitutions\n");
}

/*
 * Compute the "closest point" -- i.e., rational expressions for each
 * coordinate (edge class) of the point in the affine hull of our
 * comb facet that is closest to the centroid.
 *
 * This is obtained by combining the hull substitutions with the
 * gradient substitutions, and converting these down to rational
 * expressions over the "nring" ("n" variables only -- all "x"
 * variables eliminated).
 *
 * The result is a sorted sequence of "named" rational expressions
 * (one per "x" variable), stored as a subst_list object.
 */

	static
	struct calib_subst_list *
compute_point (

struct ginfo *			gip,	/* IN/OUT: global info */
struct calib_subst_list *	hslist,	/* IN: hull system substitutions */
struct calib_subst_list *	gslist	/* IN: gradient system substitutions */
)
{
int				i, j, k, n;
int				nverts, nedges, nvars1, nvars2, nsubst;
calib_bool *			mark;
int *				varmap;
const struct calib_Zxyz_dom *	nring;
const struct calib_Zxyz_dom *	nxring;
const struct calib_rat_dom *	nrat;
const struct calib_rat_dom *	nxrat;
struct calib_subst_list *	point;
struct calib_subst_list *	slp;
struct calib_rsubst **		map;
struct calib_rsubst *		array;
struct calib_rsubst *		rsp;
struct calib_rsubst *		dst;
struct calib_subst_list *	lists [2];

	CALIB_FATAL_ERROR_IF ((gip    EQ NULL) OR
			      (hslist EQ NULL) OR
			      (gslist EQ NULL));

	nverts	= gip -> nverts;
	nedges	= gip -> nedges;
	nring	= gip -> nring;
	nxring	= gip -> nxring;
	nrat	= gip -> nrat;
	nxrat	= gip -> nxrat;

	nvars1	= nxring -> nvars;
	nvars2	= nring -> nvars;

	CALIB_FATAL_ERROR_IF (nvars1 NE nverts + nedges);
	CALIB_FATAL_ERROR_IF (nvars2 NE nverts);

	/* Verify that there is at most one substitution (among both	*/
	/* the hslist and gslist for each "x" variable), and that none	*/
	/* of the "n" variables is substituted.				*/
	lists [0] = hslist;
	lists [1] = gslist;
	mark = CALIB_NEWA (nvars1, calib_bool);
	memset (mark, FALSE, nvars1 * sizeof (mark [0]));
	for (i = 0; i < 2; i++) {
		slp	= lists [i];
		n	= slp -> num_subst;
		rsp	= slp -> subst;
		for (j = 0; j < n; j++, rsp++) {
			k = rsp -> var;
			CALIB_FATAL_ERROR_IF ((k < 0) OR (k >= nvars1));
			CALIB_FATAL_ERROR_IF (mark [k]);
			mark [k] = TRUE;
		}
	}
	for (i = 0; i < nverts; i++) {
		/* Verify that no "n" var is substituted. */
		CALIB_FATAL_ERROR_IF (mark [i]);
	}
	/* Count how many "x" vars got marked.				*/
	/* NOTE: some "x" variables may not exist within the problem	*/
	/*	 as we set it up (i.e., the n6=0 and/or n7=0 cases).	*/
	nsubst = 0;
	for (; i < nvars1; i++) {
		if (mark [i]) {
			++nsubst;
		}
	}

	/* Create the array of "named rational values.", one for each	*/
	/* "x" variable that has a substitutions.  The rational		*/
	/* expression values, however, reside in nring (which has no	*/
	/* "x" variables).  This is why	we need to use rsp->varname	*/
	/* instead of rsp->var.  Also create a map from variable index	*/
	/* to the partituclar array entry.				*/
	map = CALIB_NEWA (nvars1, struct calib_rsubst *);
	memset (map, 0, nvars1 * sizeof (map [0]));

	array = NULL;
	if (nsubst > 0) {
		array = CALIB_NEWA (nsubst, struct calib_rsubst);
		memset (array, 0, nsubst * sizeof (array [0]));
	}
	rsp = array;
	for (i = nverts; i < nvars1; i++) {
		if (NOT mark [i]) continue;

		rsp -> var	= -1;
		rsp -> varname	= strdup (nxring -> vars [i]);
		nrat -> init (nrat, &(rsp -> val));

		map [i] = rsp;

		++rsp;
	}
	free (mark);

	/* Create a varmap for mapping the nxring down to the	*/
	/* nring (for expressions living in the nxring that no	*/
	/* longer contain any of the "x" variables).		*/
	varmap = CALIB_NEWA (nvars1, int);
	for (i = 0; i < nverts; i++) {
		varmap [i] = i;
	}
	for (; i < nvars1; i++) {
		varmap [i] = -1;
	}

	/* The hull substitutions still have "x" vars lurking.	*/
	/* Eliminate all of these by applying the gradient	*/
	/* substitutions to each, and map the result down to	*/
	/* a rational expression over the nring.		*/
	n	= hslist -> num_subst;
	rsp	= hslist -> subst;
	for (i = 0; i < n; i++, rsp++) {
		k = rsp -> var;
		CALIB_FATAL_ERROR_IF ((k < 0) OR (k > nvars1));
		dst = map [k];
		CALIB_FATAL_ERROR_IF (dst EQ NULL);

		nxrat -> substitute_with_varmap (&(dst -> val),
						 &(rsp -> val),
						 varmap,
						 gslist);
	}

	/* The gradient substitutions already have all "x"	*/
	/* variables eliminated, but still reside in nxring	*/
	/* form.  Map to nring form.				*/
	n	= gslist -> num_subst;
	rsp	= gslist -> subst;
	for (i = 0; i < n; i++, rsp++) {
		k = rsp -> var;
		CALIB_FATAL_ERROR_IF ((k < 0) OR (k > nvars1));
		dst = map [k];
		CALIB_FATAL_ERROR_IF (dst EQ NULL);

		nxring -> convert_with_varmap (&(dst -> val.numer),
					       &(rsp -> val.numer),
					       varmap);
		nxring -> convert_with_varmap (&(dst -> val.denom),
					       &(rsp -> val.denom),
					       varmap);
	}

	free (varmap);
	free (map);

	point = CALIB_NEW (struct calib_subst_list);
	memset (point, 0, sizeof (*point));

	point -> num_subst	= nsubst;
	point -> subst		= array;

	return (point);
}

/*
 * Verify that each component of the "closest point" has the same
 * denominator.  Cancel this denominator to 1 throughout, and
 * return a dynamically allocated copy of the common denominator.
 *
 * NOTE: this polynomial also appears as a factor in the denominator
 *	 of the final "dist2" expression.
 */

	static
	struct calib_Zxyz_obj *
get_point_denom (

struct calib_subst_list *	point	/* IN: the closest point */
)
{
int				i, n;
const struct calib_Zxyz_dom *	ring;
struct calib_rsubst *		rsp;
struct calib_Zxyz_obj *		denom;
struct calib_Zxyz_obj		TEMP;

	CALIB_FATAL_ERROR_IF (point EQ NULL);

	n	= point -> num_subst;
	rsp	= point -> subst;

	CALIB_FATAL_ERROR_IF (n <= 0);

	ring	= rsp -> val.denom.dom;
	ring -> init (ring, &TEMP);

	denom = ring -> dup (&(rsp -> val.denom));

	for (i = 0; i < n; i++, rsp++) {
		ring -> sub (&TEMP, &(rsp -> val.denom), denom);
		CALIB_FATAL_ERROR_IF (TEMP.nterms NE 0);
		ring -> clear (&(rsp -> val.denom));
		ring -> init (ring, &(rsp -> val.denom));
		ring -> set_si (&(rsp -> val.denom), 1);
	}

	ring -> clear (&TEMP);

	return (denom);
}

/*
 * Cancel the given factor from the denominator of "dist2".
 */

	static
	void
cancel_pden_from_dist2 (

struct calib_rat_obj *	dist2,	/* IN: distance squared expression */
struct calib_Zxyz_obj *	pden	/* IN: denominator from closest point */
)
{
int				npoly;
const struct calib_Zxyz_dom *	ring;
const struct calib_Zxyz_obj *	poly_ptrs [2];
struct calib_Zxyz_obj *		cf_ptrs [2];
struct calib_Zxyz_obj		GCD, CF2;

	CALIB_FATAL_ERROR_IF ((dist2 EQ NULL) OR (pden EQ NULL));

	ring = dist2 -> numer.dom;
	CALIB_FATAL_ERROR_IF (ring NE dist2 -> denom.dom);
	CALIB_FATAL_ERROR_IF (ring NE pden -> dom);

	ring -> init (ring, &GCD);
	ring -> init (ring, &CF2);

	npoly = 2;
	poly_ptrs [0]	= &(dist2 -> denom);
	poly_ptrs [1]	= pden;
	cf_ptrs [0]	= &(dist2 -> denom);
	cf_ptrs [1]	= &CF2;

	ring -> gcd_n (&GCD, &cf_ptrs [0], npoly, &poly_ptrs [0]);

	CALIB_FATAL_ERROR_IF (GCD.nterms NE pden -> nterms);

	ring -> clear (&CF2);
	ring -> clear (&GCD);
}

/*
 * Debug print the given system of equations.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
debug_print_system (

const struct sys *	sys	/* IN: system to print */
)
{
int				i, j, nrows, ncols;
const char **			vars;
struct calib_Zxyz_obj **	A;
struct calib_Zxyz_obj **	b;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		p;

	nrows	= sys -> nrows;
	ncols	= sys -> ncols;
	vars	= sys -> vars;
	A	= sys -> A;
	b	= sys -> b;

	rowp = A;
	for (i = 0; i < nrows; i++) {
		for (j = 0; j < ncols; j++) {
			p = rowp [j];
			if (p EQ NULL) continue;

			printf ("A[%d,%s]:\n", i, vars [j]);
			print_factored_maxima (p);
		}

		p = b [i];
		if (p NE NULL) {
			printf ("b[%d] =\n", i);
			print_factored_maxima (p);
		}
		else {
			printf ("b[%d] = 0\n", i);
		}

		rowp += ncols;
	}
}

#pragma GCC diagnostic pop

/*
 * Debug print the "complexity" of the given system of equations.
 * This is just the number of terms in each coefficient polynomial.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
debug_print_system_complexity (

const struct sys *	sys	/* IN: system to print */
)
{
int				i, j, nrows, ncols;
int				nterms, maxA, maxb;
int				Acols, bcols;
struct calib_Zxyz_obj **	A;
struct calib_Zxyz_obj **	b;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		p;
char				buf [32];

	nrows	= sys -> nrows;
	ncols	= sys -> ncols;
	A	= sys -> A;
	b	= sys -> b;

	/* Pass 1: find largest number of terms, both in the A	*/
	/* matrix, and in the b vector.				*/
	maxA = 0;
	maxb = 0;
	rowp = A;
	for (i = 0; i < nrows; i++, rowp += ncols) {
		for (j = 0; j < ncols; j++) {
			p = rowp [j];
			if (p EQ NULL) continue;

			if (p -> nterms > maxA) {
				maxA = p -> nterms;
			}
		}

		p = b [i];
		if (p EQ NULL) continue;
		if (p -> nterms > maxb) {
			maxb = p -> nterms;
		}
	}

	/* Translate into number of columns. */
	sprintf (buf, "%d", maxA);
	Acols = 1 + strlen (buf);

	sprintf (buf, "%d", maxb);
	bcols = 1 + strlen (buf);

	/* Pass 2: print the term counts. */
	rowp = A;
	for (i = 0; i < nrows; i++, rowp += ncols) {
		for (j = 0; j < ncols; j++) {
			p = rowp [j];
			nterms = (p NE NULL) ? p -> nterms : 0;
			printf ("%*d", Acols, nterms);
		}

		p = b [i];
		nterms = (p NE NULL) ? p -> nterms : 0;
		printf ("%*d\n", bcols, nterms);
	}
}

#pragma GCC diagnostic pop

/*
 * Return TRUE if-and-only-if the given polynomial is a constant.
 */

	static
	calib_bool
is_constant_poly (

const struct calib_Zxyz_obj *	poly		/* IN: polynomial to test */
)
{
int			i, k, nterms, nvars;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	nterms = poly -> nterms;

	if (nterms <= 0) {
		CALIB_FATAL_ERROR_IF (nterms < 0);
		/* Zero is a constant. */
		return (TRUE);
	}

	if (nterms > 1) {
		/* Cannot have multiple terms without being	*/
		/* some non-constant polynomial.		*/
		return (FALSE);
	}

	/* Exactly one term.  Need to see if constant. */

	nvars = poly -> dom -> nvars;
	for (i = 0; i < nvars; i++) {
		k = poly -> pow [i];
		if (k > 0) {
			/* At least 1 var to power >= 1. */
			return (FALSE);
		}
		CALIB_FATAL_ERROR_IF (k < 0);
	}

	/* Polynomial is a non-zero constant. */
	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is the
 * constant 1.
 */

	static
	calib_bool
is_one_poly (

const struct calib_Zxyz_obj *	poly		/* IN: polynomial to test */
)
{
int			i, k, nterms, nvars;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	nterms = poly -> nterms;

	if (nterms NE 1) {
		CALIB_FATAL_ERROR_IF (nterms < 0);
		/* Cant' be 1 unless it has exactly 1 term. */
		return (FALSE);
	}

	/* Exactly one term.  Check coefficient value. */
	if (mpz_cmp_ui (&(poly -> coeff [0]), 1) NE 0) {
		return (FALSE);
	}

	nvars = poly -> dom -> nvars;
	for (i = 0; i < nvars; i++) {
		k = poly -> pow [i];
		if (k > 0) {
			/* At least 1 var to power >= 1. */
			return (FALSE);
		}
		CALIB_FATAL_ERROR_IF (k < 0);
	}

	/* Polynomial is exactly 1. */
	return (TRUE);
}

/*
 * Print the given polynomial in factored form.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
print_factored (

struct calib_Zxyz_obj *	poly	/* IN: polynomial to factor and print */
)
{
const struct calib_Zxyz_dom *	ring;
struct calib_Zxyz_factor *	factors;
struct calib_genrep *		gp1;

	ring = poly -> dom;

	factors = ring -> factor (poly);

	gp1 = ring -> factors_to_genrep (factors);

	calib_genrep_print (gp1);
	calib_genrep_free (gp1);

	ring -> free_factors (factors);
}

#pragma GCC diagnostic pop

/*
 * Print the given polynomial in factored form, but
 * use "maxima" input syntax -- don't try to "pretty print" it.
 */

	static
	void
print_factored_maxima (

struct calib_Zxyz_obj *	poly	/* IN: polynomial to factor and print */
)
{
	print_poly_factored_nnl (poly, NULL);
	printf ("\n");
}

/*
 * Print the given polynomial longhand, without factoring.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
print_unfactored (

struct calib_Zxyz_obj *	poly	/* IN: polynomial to print */
)
{
const struct calib_Zxyz_dom *	ring;
struct calib_genrep *		gp1;

	ring = poly -> dom;

	gp1 = ring -> to_genrep (poly);

	calib_genrep_print (gp1);

	calib_genrep_free (gp1);
}

#pragma GCC diagnostic pop

/*
 * Print the set of all variables referenced by the given polynomial.
 * This can be any subset of the ring variables.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
print_vars (

const struct calib_Zxyz_obj *	poly	/* IN: polynomial to print vars of */
)
{
int		i, j, k, nterms, nvars;
calib_bool *	ref;
int *		pow;
char **		vars;
const char *	sep;

	if (poly EQ NULL) {
		printf ("NULL pointer given!\n");
		return;
	}

	nterms	= poly -> nterms;
	pow	= poly -> pow;
	nvars	= poly -> dom -> nvars;
	vars	= poly -> dom -> vars;

	ref = CALIB_NEWA (nvars, calib_bool);
	memset (ref, FALSE, nvars * sizeof (ref [0]));

	for (i = 0; i < nterms; i++, pow += nvars) {
		for (j = 0; j < nvars; j++) {
			k = pow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			if (k > 0) {
				ref [j] = TRUE;
			}
		}
	}

	sep = "";
	for (i = 0; i < nvars; i++) {
		if (NOT ref [i]) continue;
		printf ("%s%s", sep, vars [i]);
		sep = " ";
	}
	printf ("\n");

	free (ref);
}

#pragma GCC diagnostic pop

/*
 * Print a system of equations in a syntax that maxima can read.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
print_sys_to_maxima (

const struct sys *	sys		/* IN: system to print */
)
{
int		nrows, ncols;

	if (sys EQ NULL) {
		printf ("Ooops! You called me with NULL!\n");
		return;
	}

	nrows	= sys -> nrows;
	ncols	= sys -> ncols;

	printf ("block " LPAREN_STRING
		LBRACK_STRING "m,n,A,b" RBRACK_STRING ",\n"
		"  m : %d,\n"
		"  n : %d,\n"
		"  A : zeromatrix " LPAREN_STRING "m,n" RPAREN_STRING ",\n"
		"  b : zeromatrix " LPAREN_STRING "m,1" RPAREN_STRING ",\n",
		nrows, ncols);
	print_maxima_set_matrix_elements ("A", sys -> A, nrows, ncols);
	print_maxima_set_matrix_elements ("b", sys -> b, nrows, 1);

	printf ("  " LBRACK_STRING "A,b" RBRACK_STRING RPAREN_STRING "$\n");
}

#pragma GCC diagnostic pop

/*
 * Generate a sequence of maxima assignment statements to set the
 * individual non-zero elements of the matrix.
 */

	static
	void
print_maxima_set_matrix_elements (

const char *		name,		/* IN: Matrix variable name */
struct calib_Zxyz_obj **
			A,		/* IN: Matrix of polynomial values */
int			nrows,		/* IN: Number of rows */
int			ncols		/* IN: Number of columns */
)
{
int				i, j;
struct calib_Zxyz_obj **	rowp;
struct calib_Zxyz_obj *		p;
const struct calib_Zxyz_dom *	ring;

	rowp = A;
	for (i = 0; i < nrows; i++, rowp += ncols) {
		for (j = 0; j < ncols; j++) {
			p = rowp [j];
			if (p EQ NULL) continue;
			if (p -> nterms <= 0) continue;

			printf ("  %s" LBRACK_STRING "%d,%d" RBRACK_STRING
				" : ",
				name, i+1, j+1);
			ring = p -> dom;
			ring -> print_maxima_nnl (p);
			printf (",\n");
		}
	}
}

/*
 * Print the given substitution list in a format that can be
 * read directly by maxima.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
print_substitutions (

const struct calib_subst_list *	slist	/* IN: list of subsitutions to print */
)
{
int				i, n;
struct calib_rsubst *		rsp;
const struct calib_Zxyz_dom *	ring;

	if (slist EQ NULL) {
		printf ("Ooops -- gave me a NULL pointer!\n");
		return;
	}

	printf (LBRACK_STRING "\n");

	n	= slist -> num_subst;
	rsp	= slist -> subst;
	for (i = 0; i < n; i++, rsp++) {
		if (i > 0) {
			printf (", ");
		}
		ring = rsp -> val.numer.dom;
		printf ("%s = " LPAREN_STRING, ring -> vars [rsp -> var]);
		ring -> print_maxima (&(rsp -> val.numer));
		printf (RPAREN_STRING " / " LPAREN_STRING);
		ring -> print_maxima (&(rsp -> val.denom));
		printf (RPAREN_STRING);
	}

	printf (RBRACK_STRING "\n");
}

#pragma GCC diagnostic pop

/*
 * Print a maxima assignment statement defining the
 * given set of substitutions, which are printed in
 * *factored* form.
 *
 * The caller may optionally specify an additional (textual) factor
 * separately for both the numerator and denominator.  If non-NULL
 * this string will be added as an additional trailing factor.
 */

	static
	void
print_named_subst_factored (

const char *			vardef,	/* IN: var being defined */
const struct calib_subst_list *	slist,	/* IN: substitutions */
const char *			numf,	/* IN: extra factor for numerator */
const char *			denf	/* IN: extra factor for denominator */
)
{
int				i, k, n;
struct calib_rsubst *		rsp;
const struct calib_Zxyz_dom *	ring;
const char *			lhs_name;

	CALIB_FATAL_ERROR_IF ((vardef EQ NULL) OR (slist  EQ NULL));

	printf ("%s : " LBRACK_STRING, vardef);

	n = slist -> num_subst;
	rsp	= slist -> subst;
	for (i = 0; i < n; i++, rsp++) {
		k = rsp -> var;
		if (k EQ -1) {
			/* Name doesn't have to be a var in the val's ring. */
			lhs_name = rsp -> varname;
		}
		else {
			/* Var index must be in same ring as value! */
			ring = rsp -> val.numer.dom;
			CALIB_FATAL_ERROR_IF ((k < 0) OR (k >= ring -> nvars));
			lhs_name = ring -> vars [k];
		}
		CALIB_FATAL_ERROR_IF (lhs_name EQ NULL);

		printf (" %s = ", lhs_name);
		rat_print_factored_nnl (&(rsp -> val), numf, denf);
		if (i < n - 1) {
			printf (",\n");
		}
	}

	printf (RBRACK_STRING "$\n\n");
}

/*
 * Print a rational expression value in *factored* form, with no newline.
 *
 * The caller may optionally specify an additional (textual) factor
 * separately for both the numerator and denominator.  If non-NULL,
 * this string will be added as an additional trailing factor.
 */

	static
	void
rat_print_factored_nnl (

const struct calib_rat_obj *
			rp,	/* IN: rational expression to print */
const char *		numf,	/* IN: extra factor for numerator */
const char *		denf	/* IN: extra factor for denominator */
)
{
const struct calib_Zxyz_dom *	ring;

	CALIB_FATAL_ERROR_IF (rp EQ NULL);

	ring = rp -> numer.dom;
	CALIB_FATAL_ERROR_IF (ring NE rp -> denom.dom);

	printf (LPAREN_STRING);
	print_poly_factored_nnl (&(rp -> numer), numf);
	printf (RPAREN_STRING " / " LPAREN_STRING);
	print_poly_factored_nnl (&(rp -> denom), denf);
	printf (RPAREN_STRING);
}

/*
 * Print the given polynomial in *factored* form, with no newline.
 *
 * The caller may optionally specify an additional (textual) factor.
 * If non-NULL, this string will be added as an additional trailing
 * factor.
 */

	static
	void
print_poly_factored_nnl (

const struct calib_Zxyz_obj *
			poly,	/* IN: polynomial to print */
const char *		fact	/* IN: optional extra factor string */
)
{
int				k;
struct calib_Zxyz_factor *	factors;
struct calib_Zxyz_factor *	fp;
const struct calib_Zxyz_dom *	ring;
const char *			sep;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	ring = poly -> dom;

	if (is_constant_poly (poly)) {
		/* Handle constant polynomials specially. */
		if (is_one_poly (poly)) {
			if (fact EQ NULL) {
				printf ("1");
			}
			else {
				printf ("%s", fact);
			}
		}
		else {
			ring -> print_maxima_nnl (poly);
			if (fact NE NULL) {
				printf ("*%s", fact);
			}
		}
		return;
	}

	/* A non-constant polynomial. */

	factors = ring -> factor (poly);

	sep = "";
	for (fp = factors; fp NE NULL; fp = fp -> next) {
		printf ("%s", sep);
		if (fp -> factor -> nterms <= 1) {
			/* There are not multiple terms added together,	*/
			/* so there is no need for parens.		*/
			ring -> print_maxima_nnl (fp -> factor);
		}
		else {
			printf (LPAREN_STRING);
			ring -> print_maxima_nnl (fp -> factor);
			printf (RPAREN_STRING);
		}
		k = fp -> multiplicity;
		CALIB_FATAL_ERROR_IF (k <= 0);
		if (k > 1) {
			printf ("^%d", k);
		}
		sep = " * ";
	}
	if (fact NE NULL) {
		printf ("%s%s", sep, fact);
	}

	ring -> free_factors (factors);
}

/*
 * Multiply the given polynomial by the given integer constant.
 * This side-effects the given polynomial.
 */

	static
	void
mul_by_int (

struct calib_Zxyz_obj *	poly,	/* IN: polynomial to multiply by integer */
int			val	/* IN: integer value to multiply by */
)
{
mpz_t		IVAL;

	if (val NE 1) {
		mpz_init_set_si (IVAL, val);
		poly -> dom -> mul_z (poly, poly, IVAL);
		mpz_clear (IVAL);
	}
}

/*
 * Add the given integer value into the given polynomial.
 * This side-effects the given polynomial.
 */

	static
	void
add_int_into (

struct calib_Zxyz_obj *	poly,	/* IN/OUT: polynomial to add into */
int			val	/* IN: integer value to add */
)
{
const struct calib_Zxyz_dom *	ring;
struct calib_Zxyz_obj		VALUE;

	ring = poly -> dom;

	ring -> init_si (ring, &VALUE, val);

	ring -> add (poly, poly, &VALUE);

	ring -> clear (&VALUE);
}

/*
 * Create a polynomial object representing the given constant integer value.
 */

	static
	struct calib_Zxyz_obj *
make_int (

struct calib_Zxyz_dom *	ring,	/* IN: polynomial ring to use */
int			val	/* IN: integer value to represent */
)
{
struct calib_Zxyz_obj *	p;

	p = CALIB_NEW (struct calib_Zxyz_obj);
	ring -> init_si (ring, p, val);

	return (p);
}

/*
 * Create a polynomial object representing the given variable number.
 */

	static
	struct calib_Zxyz_obj *
make_var (

struct calib_Zxyz_dom *	ring,	/* IN: polynomial ring to use */
int			var	/* IN: index of variable to yield */
)
{
int			nvars;
int *			pow;
struct calib_Zxyz_obj *	p;

	CALIB_FATAL_ERROR_IF (ring EQ NULL);

	nvars = ring -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	pow = CALIB_NEWA (nvars, int);
	memset (pow, 0, nvars * sizeof (pow [0]));
	pow [var] = 1;

	p = CALIB_NEW (struct calib_Zxyz_obj);

	p -> nterms	= 1;
	p -> size	= 1;
	p -> dom	= ring;
	p -> pow	= pow;
	p -> coeff	= calib_new_Z_vector (1);
	mpz_set_si (p -> coeff, 1);

	return (p);
}

/*
 * Make a copy of the given variable list.
 */

	static
	const char **
copy_varlist (

const char **		varlist,	/* IN: list of variable names */
int			nvars		/* IN: number of variables */
)
{
int			i;
const char **		list;

	list = CALIB_NEWA (nvars, const char *);
	for (i = 0; i < nvars; i++) {
		list [i] = strdup (varlist [i]);
	}

	return (list);
}

/*
 * Free up the given variable list.
 */

	static
	void
free_varlist (

const char **		varlist,	/* IN: variable list to free */
int			nvars		/* IN: number of variables */
)
{
int		i;

	for (i = 0; i < nvars; i++) {
		free ((void *) (varlist [i]));
	}
	free ((void *) varlist);
}
