/**********************************************************************

	$Id: hilbert10.c,v 1.1 2024/10/19 23:48:03 warme Exp $

	File:	hilbert10.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2015, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Heuristics for "solving" Hilbert's tenth problem.

***********************************************************************

	Modification Log:

	a-1:	08/30/2015	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "hilbert10.h"

#include "calib/fatal.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/types.h"
#include "calib/Zxyz.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
 * Data for recursive enumeration of polynomial evaluation points.
 */

struct ptest {
	int	nvars;		/* Number of local vars to enumerate */
	int	nvals;		/* Row size of power table */
	int	limit;		/* Numver of values to try for each var */
	int *	ival;		/* Buffer to hold local var values */
	int *	vlist;		/* i = vlist [j] means local var j	*/
				/* corresponds to var i from orig polynomial */
	int *	vmap;		/* Let j = vmap [i].  If j < 0, then orig */
				/* var i from polynomial shouldn't be */
				/* used, and has no corresponding local var. */
				/* Otherwise, j is the local var index. */
	mpz_srcptr powtable;	/* Precomputed table of powers */
	struct calib_Zxyz_obj *
		p;		/* Polynomial to test */
	int *	lbound;		/* Variable lower bounds */
};

/*
 * Local Functions
 */

static int *		compute_var_max_degrees (
					struct calib_Zxyz_obj *	p,
					int *		   	totaldeg);
static calib_bool	irreducible_poly_can_be_zero (
					struct calib_Zxyz_obj *	p,
					int *			lbound);
static calib_bool	is_bounded_linear_Diophantine_solvable (
					struct calib_Zxyz_obj *	p,
					int *			lbound);
static calib_bool	is_hard_linear_Diophantine_solvable (
						mpz_srcptr	a,
						mpz_srcptr	b,
						int		n);
static calib_bool	is_mixed_linear_Diophantine_solvable (
						mpz_srcptr	a,
						mpz_srcptr	b,
						int		n);
static calib_bool	is_nonneg_linear_Diophantine_solvable (
						mpz_srcptr	a,
						mpz_srcptr	b,
						int		n);
static calib_bool	poly_test_recurse (int var, struct ptest * ptest);

/*
 * Return TRUE if there is a valid assignment of integer values to the
 * variables of the given polynomial that causes it to be zero.
 * (An assignment is valid if it respects the given lower bounds.)
 *
 * Note: This is a variant of Hilbert's 10th problem, and is therefore
 * impossible to solve!  Our approach tries pretty hard to filter out
 * most cases, but there is a chance that this algorithm returns FALSE
 * when it should instead return TRUE.
 */

	calib_bool
hilbert_10_heuristic (

const struct calib_Zxyz_obj *
		p,		/* IN: polynomial to test */
int *		lbound		/* IN: variable lower bounds */
)
{
int				i, j;
int				nterms, nvars;
calib_bool			in_all_terms;
calib_bool			result;
int *				tpow;
struct calib_Zxyz_factor *	factors;
struct calib_Zxyz_factor *	fp;
const struct calib_Zxyz_dom *	ring;

	if (p EQ NULL) {
		/* We interpret a NULL polynomial to be implictly zero. */
		return (TRUE);
	}

	ring = p -> dom;
	nvars = ring -> nvars;

	CALIB_FATAL_ERROR_IF (nvars NE 8);

	nterms = p -> nterms;
	if (nterms <= 0) {
		CALIB_FATAL_ERROR_IF (nterms < 0);

		/* Polynomial is identically zero! */
		return (TRUE);
	}
	if (nterms EQ 1) {
		CALIB_FATAL_ERROR_IF (mpz_sgn (&(p -> coeff [0])) EQ 0);
		for (i = 0; i < nvars; i++) {
			if ((p -> pow [i] > 0) AND
			    (lbound [i] <= 0)) {
				/* Var i can be zero, which forces the	*/
				/* polynomial to be zero.		*/
				return (TRUE);
			}
		}
		return (FALSE);
	}

	/* Two or more terms.  Check each variable that is allowed to	*/
	/* be zero and see if that variable is present in every term.	*/
	/* If so, then setting that variable to zero forces the entire	*/
	/* polynomial to be zero.					*/
	for (i = 0; i < nvars; i++) {
		if (lbound [i] > 0) continue;
		/* Var i is allowed to be zero. */
		in_all_terms = TRUE;
		tpow = p -> pow;
		for (j = 0; j < nterms; j++, tpow += nvars) {
			if (tpow [i] <= 0) {
				/* Var i is not present in this term. */
				in_all_terms = FALSE;
				break;
			}
		}
		if (in_all_terms) {
			/* Setting var i to be zero forces the given	*/
			/* polynomial to be zero.			*/
			return (TRUE);
		}
	}

	/* Time to expend more effort.  Factor it! */
	factors = ring -> factor (p);

	result = FALSE;
	for (fp = factors; fp NE NULL; fp = fp -> next) {
		if (irreducible_poly_can_be_zero (fp -> factor, lbound)) {
			result = TRUE;
			break;
		}
	}

	ring -> free_factors (factors);

	return (result);
}

/*
 * We are given an *irreducible* polynomial, together with a set of
 * lower bounds on its variables.
 *
 * Return TRUE if there is a valid assignment of integer values to the
 * variables of this polynomial that cause it to be zero.
 *
 * This is still a variant of Hilbert's 10th problem, and therefore
 * impossible to solve in general.  This algorithm gives the correct
 * answer for an important set of special cases (i.e., univariate and
 * linear), and then performs exhaustive evaluation up to a small
 * limit for each variable present.
 */

	static
	calib_bool
irreducible_poly_can_be_zero (

struct calib_Zxyz_obj *	p,	/* IN: polynomial to test */
int *			lbound	/* IN: lower bounds on variables */
)
{
int				i, j, k;
int				nterms, nvars, nvars2;
int				var, degree, totaldeg;
int				hipow, maxlb, limit, nvals;
calib_bool			result;
int *				maxdeg;
int *				vlist;
mpz_ptr				powtable;
int *				vmap;
int *				ival;
mpz_srcptr			lc, tc;
const struct calib_Zxyz_dom *	ring;
struct ptest			ptest;

	CALIB_FATAL_ERROR_IF (p EQ NULL);

	ring = p -> dom;
	nvars = ring -> nvars;

	CALIB_FATAL_ERROR_IF (nvars NE 8);

	nterms = p -> nterms;

	/* Identically zero polynomials should have been weeded	*/
	/* out by our caller.					*/
	CALIB_FATAL_ERROR_IF (nterms <= 0);

	if (nterms EQ 1) {
		CALIB_FATAL_ERROR_IF (mpz_sgn (&(p -> coeff [0])) EQ 0);
		for (i = 0; i < nvars; i++) {
			if ((p -> pow [i] > 0) AND (lbound [i] <= 0)) {
				/* Setting var i to zero forces the	*/
				/* polynomial to be zero.		*/
				return (TRUE);
			}
		}
		return (FALSE);
	}

	/* Compute max degree in each variable. */
	maxdeg = compute_var_max_degrees (p, &totaldeg);

	/* See if the polynomial is univariate.  If so, determine which	*/
	/* variable it is, and the degree of the polynomial.		*/
	k = 0;
	var = -1;
	for (i = 0; i < nvars; i++) {
		if (maxdeg [i] > 0) {
			++k;
			var = i;
		}
	}
	if ((k EQ 1) AND (var >= 0)) {
		/* We have a univariate polynomial! */
		degree = maxdeg [var];
		CALIB_FATAL_ERROR_IF (degree NE totaldeg);
		free (maxdeg);

		if (degree > 1) {
			/* Polynomial is irreducible, and so cannot	*/
			/* have any integer root!			*/
			return (FALSE);
		}

		CALIB_FATAL_ERROR_IF (degree < 1);
		CALIB_FATAL_ERROR_IF (nterms NE 2);

		/* Terms are *highest* degree first. */
		lc = &(p -> coeff [0]);
		tc = &(p -> coeff [1]);
		if (mpz_cmp_ui (lc, 1) EQ 0) {
			if (mpz_cmp_si (tc, - lbound [var]) <= 0) {
				/* We have, e.g., (var - k) EQ 0, where	*/
				/* k >= lower bound for var.		*/
				return (TRUE);
			}
		}
		else if (mpz_cmp_si (lc, -1) EQ 0) {
			if (mpz_cmp_si (tc, lbound [var]) >= 0) {
				/* We have, e.g., (-var + k) EQ 0,	*/
				/* where k >= lower bound for var.	*/
				return (TRUE);
			}
		}
		return (FALSE);
	}

	/* See if polynomial is at most linear in each variable, and	*/
	/* count how many variables are involved if so.			*/
	k = 0;
	for (i = 0; i < nvars; i++) {
		j = maxdeg [i];
		if (j <= 0) continue;
		if (j > 1) {
			k = -1;
			break;
		}
		++k;
	}

	if ((k > 0) AND (totaldeg <= 1)) {
		CALIB_FATAL_ERROR_IF (totaldeg < 1);

		/* We have a polynomial linear in k variables. */
		CALIB_FATAL_ERROR_IF (k < 2);

		/* This is a linear Diophantine equation with		*/
		/* lower-bounded variables.				*/
		result = is_bounded_linear_Diophantine_solvable (p, lbound);

		free (maxdeg);

		return (result);
	}

	/* It is time to do some exhaustive enumeration.  Make a list	*/
	/* of all the vars that are present.				*/
	k = 0;
	for (i = 0; i < nvars; i++) {
		if (maxdeg [i] > 0) {
			++k;
		}
	}
	CALIB_FATAL_ERROR_IF (k < 2);
	nvars2 = k;
	vlist = CALIB_NEWA (nvars2, int);
	k = 0;
	hipow = 0;
	for (i = 0; i < nvars; i++) {
		j = maxdeg [i];
		if (j > 0) {
			vlist [k] = i;
			++k;
		}
		if (j > hipow) {
			hipow = j;
		}
	}
	CALIB_FATAL_ERROR_IF (k NE nvars2);

	/* Set the enumeration limit for how many values we are	*/
	/* going to test for each variable.  Note that the cost	*/
	/* is *exponential*:					*/
	/*							*/
	/*	cost = (limit**nvars2) * poly_eval_cost,	*/
	/*							*/
	/* where poly_eval_cost is nterms*(nvars+1) multiplies	*/
	/* and nterms additions (in mpz arithmetic).		*/

	switch (nvars) {
	case 2:
	case 3:
	case 4:		limit = 20;	break;
	case 5:		limit = 10;	break;
	case 6:		limit = 7;	break;
	case 7:
	case 8:		limit = 6;	break;
	default:
		CALIB_FATAL_ERROR;
		limit = 0;
		break;
	}

	/* The range of values we need powers of gets increased by the	*/
	/* lower bounds involved, so find the largest lower bound.	*/
	maxlb = lbound [vlist [0]];
	for (i = 1; i < nvars2; i++) {
		k = lbound [vlist [i]];
		if (k > maxlb) {
			maxlb = k;
		}
	}

	/* Create an array that maps original variable indices	*/
	/* to either -1 or our local variable index.		*/
	vmap = CALIB_NEWA (nvars, int);
	for (i = 0; i < nvars; i++) {
		vmap [i] = -1;
	}
	for (i = 0; i < nvars2; i++) {
		vmap [vlist [i]] = i;
	}

	/* Precompute a table of the necessary powers of each	*/
	/* value we can plug into the polynomial.		*/
	nvals = maxlb + limit + 1;
	k = (hipow + 1) * nvals;
	powtable = calib_new_Z_vector (k);
	for (i = 0; i < nvals; i++) {
		for (j = 0; j <= hipow; j++) {
			mpz_ui_pow_ui (&powtable [j * nvals + i], i, j);
		}
	}
	ival = CALIB_NEWA (nvars2, int);
	ptest.nvars	= nvars2;
	ptest.nvals	= nvals;
	ptest.limit	= limit;
	ptest.ival	= ival;
	ptest.vlist	= vlist;
	ptest.vmap	= vmap;
	ptest.powtable	= powtable;
	ptest.p		= p;
	ptest.lbound	= lbound;

	result = poly_test_recurse (0, &ptest);

	free (ival);
	free (vmap);
	calib_free_Z_vector (powtable, k);
	free (vlist);
	free (maxdeg);

	return (result);
}

/*
 * Trying to find a valid assignment of the integer variables that
 * make the given polynomial become zero.  This is the recursive
 * portion of that enumeration effort.
 *
 * Returns TRUE is a zero is found, FALSE otherwise.
 */

	static
	calib_bool
poly_test_recurse (

int		var,		/* IN: Variable to choose at this level */
struct ptest *	ptest		/* IN: Data used in the recursion */
)
{
int			i, j, k, m;
int			nvars, nvals, limit;
int			nterms, npv, val;
calib_bool		result;
int *			ppow;
int *			ival;
int *			vmap;
mpz_srcptr		powtable;
struct calib_Zxyz_obj * p;
int *			lbound;
mpz_t			SUM, PROD;

	nvars	= ptest -> nvars;
	ival	= ptest -> ival;
	if (var >= nvars) {
		/* Recursive base case.					*/
		/* We have a complete assignment of vars.  Time to	*/
		/* evaluate the polynomial.				*/
		nvals		= ptest -> nvals;
		vmap		= ptest -> vmap;
		powtable	= ptest -> powtable;
		p		= ptest -> p;
		lbound		= ptest -> lbound;

		mpz_init (SUM);
		mpz_init (PROD);
	
		nterms	= p -> nterms;
		npv	= p -> dom -> nvars;
		ppow	= p -> pow;
		for (i = 0; i < nterms; i++, ppow += npv) {
			mpz_set (PROD, &(p -> coeff [i]));
			for (j = 0; j < npv; j++) {
				k = ppow [j];
				m = vmap [j];
				if (m < 0) {
					CALIB_FATAL_ERROR_IF (k NE 0);
				}
				else {
					val = lbound [j] + ival [m];
					mpz_mul (PROD,
						 PROD,
						 &powtable [k * nvals + val]);
				}
			}
			mpz_add (SUM, SUM, PROD);
		}

		result = (mpz_sgn (SUM) EQ 0);

		mpz_clear (PROD);
		mpz_clear (SUM);

		return (result);
	}

	limit = ptest -> limit;

	for (i = 0; i < limit; i++) {
		ival [var] = i;
		result = poly_test_recurse (var + 1, ptest);
		if (result) {
			return (TRUE);
		}
	}

	return (FALSE);
}

/*
 * Return TRUE if-and-only-if the given linear Diophantine equation
 * has a solution with variables that satisfy given lower bounds.
 * Returns FALSE if no such solution exists.
 *
 * This decision problem is NP-complete in general, so there can be
 * difficult "TRUE" instances that we classify as "FALSE".
 */

	static
	calib_bool
is_bounded_linear_Diophantine_solvable (

struct calib_Zxyz_obj *	p,		/* IN: linear polynomial to test */
int *			lbound		/* IN: variable lower bounds */
)
{
int		i, j, k, n, nterms, nvars, var, tdeg;
calib_bool	result;
int *		pow;
calib_bool *	flags;
mpz_ptr		a, b;
mpz_t		TMP;

	nterms	= p -> nterms;
	nvars	= p -> dom -> nvars;

	/* Extract the coefficients and constant term. */
	a = calib_new_Z_vector (nvars + 1);
	b = a + nvars;
	flags = CALIB_NEWA (nvars + 1, calib_bool);
	memset (flags, FALSE, (nvars + 1) * sizeof (flags [0]));

	pow = p -> pow;
	for (i = 0; i < nterms; i++, pow += nvars) {
		var = -1;
		tdeg = 0;
		for (j = 0; j < nvars; j++) {
			k = pow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			if (k > 0) {
				CALIB_FATAL_ERROR_IF (k > 1);
				var = j;
				++tdeg;
			}
		}
		if (tdeg EQ 0) {
			CALIB_FATAL_ERROR_IF (flags [nvars]);
			flags [nvars] = TRUE;

			/* Negate the constant term so that it becomes	*/
			/* the right-hand-side of an equation.		*/
			mpz_neg (b, &(p -> coeff [i]));
		}
		else {
			CALIB_FATAL_ERROR_IF (tdeg NE 1);
			CALIB_FATAL_ERROR_IF (flags [var]);
			flags [var] = TRUE;
			mpz_set (&a [var], &(p -> coeff [i]));
		}
	}
	free (flags);

	/* Process the lower bounds, transforming each variable.  The	*/
	/* original var satisfies x[i] >= lbound[i].  We transform	*/
	/* x[i] = y[i] + lbound[i], where y[i] >= 0.			*/
	mpz_init (TMP);
	for (i = 0; i < nvars; i++) {
		mpz_mul_si (TMP, &a [i], lbound [i]);
		mpz_sub (b, b, TMP);
	}
	mpz_clear (TMP);

	if (mpz_sgn (b) EQ 0) {
		/* Setting all x vars to their lower bounds yields zero! */
		calib_free_Z_vector (a, nvars + 1);
		return (TRUE);
	}

	/* Eliminate zero coefficients. */
	n = 0;
	for (i = 0; i < nvars; i++) {
		if (mpz_sgn (&a [i]) NE 0) {
			mpz_set (&a [n], &a [i]);
			++n;
		}
	}

	result = is_nonneg_linear_Diophantine_solvable (a, b, n);

	calib_free_Z_vector (a, nvars + 1);

	return (result);
}

/*
 * Return TRUE if-and-only-if the given linear Diophantine equation
 * has a solution over the non-negative integer variables.
 * Returns FALSE if no such solution exists.
 *
 * This decision problem is NP-complete in general, so there can be
 * difficult "TRUE" instances that we classify as "FALSE".
 */

	static
	calib_bool
is_nonneg_linear_Diophantine_solvable (

mpz_srcptr		a,		/* IN: left-hand-side coefficients */
mpz_srcptr		b,		/* IN: right-hand-side constant */
int			n		/* IN: number of LHS coefficients */
)
{
int		i, j1, j2, k;
int		bsign, s0, s1;
calib_bool	swapped, result;
int *		order;
mpz_ptr		sorted_a, tmp;

	CALIB_FATAL_ERROR_IF (n < 2);

	bsign = mpz_sgn (b);
	if (bsign EQ 0) {
		return (TRUE);
	}

	/* Sort the coefficients into increasing order.  Bubble sort is	*/
	/* completely acceptable here, since n <= 8.			*/
	order = CALIB_NEWA (n, int);
	for (i = 0; i < n; i++) {
		order [i] = i;
	}
	do {
		swapped = FALSE;
		for (i = 1; i < n; i++) {
			j1 = order [i-1];
			j2 = order [i];
			if (mpz_cmp (&a [j1], &a [j2]) > 0) {
				order [i-1] = j2;
				order [i]   = j1;
				swapped = TRUE;
			}
		}
	} while (swapped);

	/* Eliminate duplicates. */
	k = 1;
	for (i = 1; i < n; i++) {
		j1 = order [k - 1];
		j2 = order [i];
		if (mpz_cmp (&a [j1], &a [j2]) NE 0) {
			order [k] = j2;
			++k;
		}
	}

	/* Make a sorted copy of the coefficients. */
	sorted_a = calib_new_Z_vector (k + 1);
	for (i = 0; i < k; i++) {
		mpz_set (&sorted_a [i], &a [order [i]]);
	}
	tmp = &sorted_a [k];

#if 0
	printf ("is_nonneg_linear_Diophantine_solvable:");
	for (i = 0; i < k; i++) {
		printf (" ");
		mpz_out_str (stdout, 10, &sorted_a [i]);
	}
	printf (" = ");
	mpz_out_str (stdout, 10, b);
	printf ("\n");
#endif

	s0 = mpz_sgn (&sorted_a [0]);
	s1 = mpz_sgn (&sorted_a [k-1]);
	if (s0 > 0) {
		/* All coefficients are positive! */
		if (bsign < 0) {
			/* Cannot add positives to get a negative. */
			result = FALSE;
		}
		else {
			result = is_hard_linear_Diophantine_solvable (
							sorted_a, b, k);
		}
	}
	else if (s1 < 0) {
		/* All coefficients are negative! */
		if (bsign > 0) {
			/* Cannot add negatives to get a positive. */
			result = FALSE;
		}
		else {
			/* Negate and reverse coefficients. */
			for (i = 0; i < k; i++) {
				mpz_neg (&sorted_a [i], &sorted_a [i]);
			}
			j1 = 0;
			j2 = k-1;
			while (j1 < j2) {
				mpz_set (tmp, &sorted_a [j1]);
				mpz_set (&sorted_a [j1], &sorted_a [j2]);
				mpz_set (&sorted_a [j2], tmp);
				++j1;
				--j2;
			}
			mpz_neg (tmp, b);
			result = is_hard_linear_Diophantine_solvable (
							sorted_a, tmp, k);
		}
	}
	else {
		/* Some coefficients are negative, some positive. */
		result = is_mixed_linear_Diophantine_solvable (
							sorted_a, b, k);
	}

	calib_free_Z_vector (sorted_a, k + 1);
	free (order);

	return (result);
}

/*
 * Return TRUE if-and-only-if the given linear Diophantine equation
 * has a solution over the non-negative integer variables.
 * Returns FALSE if no such solution exists.
 *
 * The coefficients are already sorted from smallest to largest,
 * and we know that there is a mix of negative and positive values.
 * This allows some interesting reductions to be applied.
 */

	static
	calib_bool
is_mixed_linear_Diophantine_solvable (

mpz_srcptr		a,		/* IN: left-hand-side coefficients */
mpz_srcptr		b,		/* IN: right-hand-side constant */
int			n		/* IN: number of LHS coefficients */
)
{
int		i, bsign;
calib_bool	result;
mpz_t		g;

	CALIB_FATAL_ERROR_IF (n < 0);

	bsign = mpz_sgn (b);
	if (bsign EQ 0) {
		return (TRUE);
	}

	for (i = 0; i < n; i++) {
		if ((mpz_sgn (&a [i]) EQ bsign) AND
		    (mpz_divisible_p (b, &a [i]))) {
			return (TRUE);
		}
	}

	if (n EQ 2) {
		mpz_init (g);
		mpz_gcd (g, &a [0], &b [0]);
		result = mpz_divisible_p (b, g);
		mpz_clear (g);
		return (result);
	}

	/* FIXME -- write this! */
	CALIB_FATAL_ERROR;

	return (FALSE);
}

/*
 * Return TRUE if-and-only-if the given linear Diophantine equation
 * has a solution over the non-negative integer variables.
 * Returns FALSE if no such solution exists.
 *
 * The coefficients are already sorted from smallest to largest,
 * all are strictly positive, and the right-hand-side coefficient is
 * strictly positive.  This is a classic NP-complete decision problem,
 * so we may occasionally give some "FALSE" answers that should
 * actually be "TRUE".
 */

	static
	calib_bool
is_hard_linear_Diophantine_solvable (

mpz_srcptr		a,		/* IN: left-hand-side coefficients */
mpz_srcptr		b,		/* IN: right-hand-side constant */
int			n		/* IN: number of LHS coefficients */
)
{
	/* FIXME -- write this! */
	CALIB_FATAL_ERROR;

	return (FALSE);
}

/*
 * Compute the maximum degree of each variable,
 * as well as the total degree of the given polynomial.
 */

	static
	int *
compute_var_max_degrees (

struct calib_Zxyz_obj *
		p,		/* IN: polynomial to compute degrees for */
int *		totaldeg	/* IN/OUT: total degree of polynomial */
)
{
int		i, j, k, nvars, nterms, totdeg, maxtotdeg;
int *		maxdeg;
int *		pow;

	CALIB_FATAL_ERROR_IF (p EQ NULL);

	nterms	= p -> nterms;
	nvars	= p -> dom -> nvars;

	maxdeg = CALIB_NEWA (nvars, int);
	memset (maxdeg, 0, nvars * sizeof (maxdeg [0]));

	maxtotdeg = 0;

	pow = p -> pow;
	for (i = 0; i < nterms; i++, pow += nvars) {
		totdeg = 0;
		for (j = 0; j < nvars; j++) {
			k = pow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			if (k > maxdeg [j]) {
				maxdeg [j] = k;
			}
			totdeg += k;
		}
		if (totdeg > maxtotdeg) {
			maxtotdeg = totdeg;
		}
	}

	if (totaldeg NE NULL) {
		*totaldeg = maxtotdeg;
	}

	return (maxdeg);
}
