/**********************************************************************

	$Id: ratint.c,v 1.1 2024/10/19 23:51:10 warme Exp $

	File:	ratint.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2015, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Integration of rational functions.

***********************************************************************

	Modification Log:

	a-1:	09/21/2024	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/calib.h"
#include "calib/cputime.h"
#include "calib/fatal.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Local Types
 */

/*
 * An object to hold a list of genreps.
 */

struct grlist {
	struct calib_genrep_list *	root;
	struct calib_genrep_list **	hookp;
};

/*
 * Global data used by ratint() and its callees.
 */

struct RatInt {
	const struct calib_rat_dom *
			dom;		/* Rational function domain */
	const struct calib_rat_obj *
			integrand;	/* Original integrand */
	int		var;		/* Index of integration variable */
	char *		varname;	/* Name of integration variable */
	struct grlist	poly;		/* Integral of polynomial part */
	struct grlist	rats;		/* Rational function (genrep form) */
	struct grlist	atans;		/* Arctangents */
	struct grlist	logs;		/* Log terms */
	struct grlist	integrals;	/* Unfinished integrals */
	struct calib_rat_obj
			rat;		/* Rational function */
};

/*
 * An object to represent a linked list of rational roots.  These store
 * the absolute value of each rational root Ci in strictly increasing value,
 * together with separate flags indicating the presence of +Ci and -Ci.
 * This allows us to generate logs of ratios in Trager's method.
 */

struct qroot {
	mpq_t		root;		/* Absolute value Ci of rational */
					/* root */
	calib_bool	pflag;		/* If true, +Ci is a root */
	calib_bool	mflag;		/* If true, -Ci is a root */
	struct qroot *	next;		/* Next node in linked list */
};

/*
 * An object to represent a linked list of partial-fraction integrands
 * having the following form:
 *
 *		Q * (R / S^k)
 *
 * where R, S \in Z[x] are primitive, S is square-free, deg(R) < k*deg(S),
 * and Q is a rational number.
 */

struct pfrac {
	mpq_t			Q;	/* Rational multiplier */
	struct calib_Zx_obj	R;	/* Numerator polynomial */
	struct calib_Zx_obj	S;	/* Denominator polynomial */
	int			k;	/* Denominator exponent */
	struct pfrac *		next;	/* Next partial fraction in list */
};

/*
 * An object that represents sqrt(Q) * R, where Q is GMP rational and
 * R is a rational function.  We use this to represent the derivative
 * of our solution, which may contain constants of the form sqrt(k),
 * where k is a GMP integer.
 */

struct sqrt_rat {
	calib_bool		err;	/* Some error occurred? */
	mpq_t			Q;	/* A factor of sqrt(Q) */
	struct calib_rat_obj	R;	/* Rational function */
};

/*
 * An object to keep track of how well we did on a suite of problems.
 */

struct score {
	int	syntax;		/* Number of syntax errors */
	int	problems;	/* Number of problems attempted */
	int	incorrect;	/* Number of incorrect solutions */
	int	unfinished;	/* Number of "unfinished" solutions */
	int	deriv_errs;	/* Number of errors during differentiation */
};

/*
 * Local Functions
 */

static struct calib_genrep *
		assemble_result (struct RatInt * rip);
static void	check_solution (struct score *			score,
				const struct calib_rat_obj *	op,
				const struct calib_genrep *	gp);
static void	clear_global_data (struct RatInt * rip);
static struct calib_Zx_obj *
		convert_to_Zx (const struct calib_Zxyz_obj * op, int var);
static void	decode_args (int argc, char ** argv);
static void	differentiate (struct sqrt_rat *		rop,
			       const struct calib_genrep *	op);
static calib_bool
		emit_arc_tangent (struct RatInt *		rip,
				  mpq_srcptr			Q,
				  const struct calib_Zx_obj *	R,
				  const struct calib_Zx_obj *	S);
static void	emit_rational_func_term (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	R,
				const struct calib_Zx_obj *	S,
				int				k);
static void	find_arc_tangent (struct RatInt *		rip,
				  mpq_srcptr			Q,
				  const struct calib_Zx_obj *	R,
				  const struct calib_Zx_obj *	S);
static void	free_rational_roots (struct qroot * list);
static calib_bool
		genrep_is_constant (const struct calib_genrep *	gp);
static struct calib_genrep *
		get_integrand (struct calib_genrep * gp);
static void	grlist_append (struct grlist *		list,
			       struct calib_genrep *	expr);
static void	grlist_init (struct grlist * list);
static void	grlist_take (struct grlist *	list1,
			     struct grlist *	list2);
static void	handle_quadratic_denominator (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	R,
				const struct calib_Zx_obj *	S);
static calib_bool
		has_function (const struct calib_genrep *	gp,
			      calib_si_t			fidx);
static void	hermite_reduction (struct RatInt *		rip,
				   const struct pfrac *		pfp);
static void	init_global_data (struct RatInt *		rip,
				  const struct calib_rat_obj *	expr,
				  int				var);
static void	integrate_polynomial_part (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	zpoly);
static void	integrate_primitive_rational_part (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	R,
				const struct calib_Zx_obj *	S);
static calib_bool
		is_rational_function (const struct calib_rat_dom *	dom,
				      struct calib_genrep *		gp);
static struct calib_genrep *
		make_integral (struct calib_genrep *	expr,
			       const char *		varname);
static struct calib_genrep *
		make_rational_func_term (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	R,
				const struct calib_Zx_obj *	S,
				int				k);
static struct calib_genrep *
		make_sqrt (mpz_srcptr A, mpz_srcptr B);
static struct pfrac *
		partial_fraction_decomposition (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	R,
				int				nfact,
				struct calib_Zx_factor *	flist);
static struct pfrac *
		pfrac_create (mpq_srcptr			Q,
				  const struct calib_Zx_obj *	R,
				  const struct calib_Zx_obj *	S,
				  int				k);
static void	pfrac_free (struct pfrac * list);
static void	primitize (mpq_ptr Q, struct calib_Zx_obj * poly);
static void	process_log_term (struct RatInt *		rip,
				  mpq_srcptr			C,
				  const struct calib_Zx_obj *	Vnum,
				  const struct calib_Zx_obj *	Vden);
static void	process_trager_residuals (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_rat_obj *	rat);
static struct calib_genrep *
		ratint (const struct calib_rat_obj * op, int var);
static struct qroot *
		rational_root_list (const struct calib_Zx_obj *	poly);
static void	reduce_sqrt (mpz_ptr A, mpz_ptr B, mpz_srcptr D);
static struct calib_Zx_factor *
		remove_constant_factors (
				mpz_ptr				zval,
				struct calib_Zx_factor *	sqfactors);
static void	report_score (const struct score *	score);
static void	sqrt_rat_add (struct sqrt_rat *		rop,
			      const struct sqrt_rat *	op1,
			      const struct sqrt_rat *	op2);
static void	sqrt_rat_clear (struct sqrt_rat *	op);
static void	sqrt_rat_div (struct sqrt_rat *		rop,
			      const struct sqrt_rat *	op1,
			      const struct sqrt_rat *	op2);
static void	sqrt_rat_init (const struct calib_rat_dom *	dom,
			       struct sqrt_rat *		rop);
static void	sqrt_rat_ipow (struct sqrt_rat *		rop,
			       const struct sqrt_rat *		op1,
			       int				op2);
static void	sqrt_rat_mul (struct sqrt_rat *		rop,
			      const struct sqrt_rat *	op1,
			      const struct sqrt_rat *	op2);
static void	sqrt_rat_set (struct sqrt_rat *		rop,
			      const struct sqrt_rat *	op);
static void	sqrt_rat_set_genrep (struct sqrt_rat *			rop,
				     const struct calib_genrep *	op);
static void	sqrt_rat_set_q (struct sqrt_rat *	rop,
				mpq_srcptr		op);
static void	sqrt_rat_set_si (struct sqrt_rat *	rop,
				 calib_si_t		op);
static void	sqrt_rat_set_z (struct sqrt_rat *	rop,
				mpz_srcptr		op);
static void	sqrt_rat_square (struct sqrt_rat *		rop,
				 const struct sqrt_rat *	op);
static struct pfrac *
		square_free_partial_fraction_decomp (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	R,
				const struct calib_Zx_obj *	S);
static void	subtract_out_arc_tangent (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				struct calib_Zx_obj *		R,
				const struct calib_Zx_obj *	S);
static void	trager_split (struct RatInt *		rip,
			      mpq_srcptr		Q,
			      struct calib_Zx_obj *	V,
			      mpz_srcptr		N,
			      mpz_srcptr		D,
			      struct calib_rat_obj *	RAT);
static void	tragers_method (struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	R,
				const struct calib_Zx_obj *	S);
static void	unfinished_rat_integral (
				struct RatInt *			rip,
				mpq_srcptr			Q,
				const struct calib_Zx_obj *	R,
				const struct calib_Zx_obj *	S,
				int				k);
static void	usage (void);

/*
 * Local Data
 */

static calib_bool	flag_factor_log;
static calib_bool	flag_factor_poly;
static calib_bool	flag_factor_rat;
static calib_bool	flag_trace;
static char *		me;

/*
 * The main program for 'ratint'.
 */

	int
main (

int		argc,
char **		argv
)
{
int				status;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_genrep_varlist	varlist;
struct score			score;
struct calib_rat_obj		RAT;

	me = argv [0];

	decode_args (argc, argv);

	memset (&score, 0, sizeof (score));

	/* Read-eval-print loop. */
	for (;;) {
		gp1	= NULL;
		gp2	= NULL;
		gp3	= NULL;
		memset (&varlist, 0, sizeof (varlist));
		Zxyz	= NULL;
		dom	= NULL;

		calib_prompt ("(ratint) ");
		gp1 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			++score.syntax;
			continue;
		}
		if (gp1 EQ NULL) break;

		do {		/* Used only for "break". */

			calib_genrep_get_varlist (&varlist, gp1);

			if (varlist.nvars NE 1) {
				fprintf (stderr,
					 "Expression must contain exactly"
					 " one variable\n");
				++score.syntax;
				break;
			}

			Zxyz	= calib_make_Zxyz_dom (varlist.nvars,
						       varlist.vlist);
			dom	= calib_make_rat_dom (Zxyz);

			/* Note: because there is just one variable,	*/
			/* we know that the variable is "correct" in	*/
			/* the 2nd arg to integrate(expr, var).		*/

			gp2 = get_integrand (gp1);
			if (gp2 EQ NULL) {
				fprintf (stderr,
					 "Must be integrate (expr, var);\n");
				++score.syntax;
				break;
			}

			/* Diagnose when integrand is not a rational func. */
			if (NOT is_rational_function (dom, gp2)) {
				fprintf (stderr,
					 "Integrand must be a"
					 " rational function.\n");
				++score.syntax;
				break;
			}

			dom -> init (dom, &RAT);
			dom -> set_genrep (&RAT, gp2);

			++score.problems;

			/* Do the integration. */
			gp3 = ratint (&RAT, 0);

			calib_genrep_prettyprint (gp1);
			printf (" =\n");
			calib_genrep_prettyprint (gp3);
			printf ("\n");

			check_solution (&score, &RAT, gp3);

			dom -> clear (&RAT);

		} while (FALSE);

		calib_free_rat_dom (dom);
		calib_free_Zxyz_dom (Zxyz);
		calib_genrep_clear_varlist (&varlist);
		calib_genrep_free (gp3);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
	}

	report_score (&score);

	calib_shutdown ();

	exit (0);
}

/*
 * Decode the command line arguments.
 */

	static
	void
decode_args (

int		argc,		/* IN: Number of arguments */
char **		argv		/* IN: Argument strings */
)
{
int		i;
char		c;
const char *	ap;

	flag_factor_log		= FALSE;
	flag_factor_rat		= FALSE;
	flag_trace		= FALSE;

	for (i = 1; i < argc; i++) {
		ap = argv [i];
		c = *ap++;
		if (c NE '-') {
			usage ();
		}
		for (;;) {
			c = *ap++;
			if (c EQ '\0') break;
			switch (c) {
			case 'l':
				flag_factor_log = TRUE;
				break;

			case 'p':
				flag_factor_poly = TRUE;
				break;

			case 'r':
				flag_factor_rat = TRUE;
				break;

			case 't':
				flag_trace = TRUE;
				break;

			default:
				usage ();
				break;
			}
		}
	}
}

/*
 * Display command usage and exit(1).
 */

	static
	void
usage (void)

{
	fprintf (stderr,
		 "Usage: %s [-[al]]\n"
		 "\t-l\tFactor arguments of generated log() terms.\n"
		 "\t-p\tFactor polynomial part of solution.\n"
		 "\t-r\tFactor generated rational function term.\n"
		 "\t-t\tEnable some tracing.\n",
		 me);

	exit (1);
}

/*
 * Report the score to the user.
 */

	static
	void
report_score (

const struct score *	score	/* IN: Score to report */
)
{
	CALIB_FATAL_ERROR_IF (score EQ NULL);

	if (score -> problems <= 1) {
		/* We only report any scores when two or more problems	*/
		/* have been solved (e.g., when doing a "batch" of	*/
		/* problems).						*/
		return;
	}

	printf ("\n");
	if (score -> syntax > 0) {
		printf ("There were %d syntax errors.\n", score -> syntax);
	}
	printf ("Total problems attempted:\t%d\n", score -> problems);
	printf ("Incorrect solutions:\t\t%d\n", score -> incorrect);
	printf ("Unfinished solutions:\t\t%d\n", score -> unfinished);
	if (score -> deriv_errs > 0) {
		printf ("Errors while calculating derivatives: %d\n",
			score -> deriv_errs);
	}
}

/*
 * Returns "expr" when given a genrep of the following form:
 *
 *	integrate (expr, var)
 *
 * Returns NULL if the genrep does not have this exact form.
 */

	static
	struct calib_genrep *
get_integrand (

struct calib_genrep *	gp	/* IN: Genrep from which to get integrand */
)
{
calib_si_t			idx;
struct calib_genrep *		f;
struct calib_genrep *		expr;
struct calib_genrep *		var;
struct calib_genrep_list *	args;

	CALIB_FATAL_ERROR_IF (gp EQ NULL);

	do {		/* Used only for "break". */
		if (gp -> op NE CALIB_GENREP_OP_FUNC) break;

		f	= gp -> u.func.func;
		args	= gp -> u.func.args;

		if (f -> op NE CALIB_GENREP_OP_SI) break;

		idx	= f -> u.siop;
		if (idx NE CALIB_GENREP_FUNC_INTEGRATE) break;

		if (args EQ NULL) break;
		expr	= args -> operand;
		args	= args -> next;

		if (args EQ NULL) break;
		var	= args -> operand;
		args	= args -> next;

		if (var -> op NE CALIB_GENREP_OP_VAR) break;
		if (args NE NULL) break;

		/* Expression has the required form.  Return a *copy*	*/
		/* of the integrand.					*/
		expr = calib_genrep_dup (expr);

		return (expr);

	} while (FALSE);

	return (NULL);
}

/*
 * Return TRUE if-and-only-if the given genrep can be loaded into
 * a rational function.
 */

	static
	calib_bool
is_rational_function (

const struct calib_rat_dom *	dom,	/* IN: Rational function domain */
struct calib_genrep *		gp	/* IN: Genrep to test */
)
{
int				idx;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_genrep_list *	lp;
struct calib_genrep *		f;
const char *			fname;

	CALIB_FATAL_ERROR_IF ((dom EQ NULL) OR (gp EQ NULL));

	switch (gp -> op) {
	case CALIB_GENREP_OP_ADD:
	case CALIB_GENREP_OP_MUL:
		for (lp = gp -> u.list; lp NE NULL; lp = lp -> next) {
			if (NOT is_rational_function (dom, lp -> operand)) {
				return (FALSE);
			}
		}
		return (TRUE);

	case CALIB_GENREP_OP_IPOW:
		if (is_rational_function (dom, gp -> u.ipow.base)) {
			return (TRUE);
		}
		return (FALSE);

	case CALIB_GENREP_OP_VAR:
		Zxyz = dom -> ring;
		idx = Zxyz -> lookup_var (Zxyz, gp -> u.var);
		if (idx < 0) {
			return (FALSE);
		}
		return (TRUE);

	case CALIB_GENREP_OP_SI:
	case CALIB_GENREP_OP_Z:
	case CALIB_GENREP_OP_Q:
		return (TRUE);

	case CALIB_GENREP_OP_FUNC:
		f = gp -> u.func.func;
		if (f -> op EQ CALIB_GENREP_OP_VAR) {
			fname = f -> u.var;
		}
		else if (f -> op EQ CALIB_GENREP_OP_SI) {
			fname = calib_genrep_builtin_func_index_to_name (f -> u.siop);
			CALIB_FATAL_ERROR_IF (fname EQ NULL);
		}
		else {
			CALIB_FATAL_ERROR;
		}
		fprintf (stderr,
			 "Rational functions cannot contain functions"
			 " like %s()\n",
			 fname);
		return (FALSE);

	default:
		CALIB_FATAL_ERROR;
		return (FALSE);
	}
}

/*
 * Compute the indefinite integral of the given rational function,
 * returning it as a genrep.
 */

	static
	struct calib_genrep *
ratint (

const struct calib_rat_obj *	op,	/* IN: Rational func to integrate */
int				var	/* IN: Which var to integrate */
)
{
const struct calib_Zx_dom *	Zx;
struct calib_genrep *		gp;
struct calib_Zx_obj *		R;
struct calib_Zx_obj *		S;
struct RatInt *			rip;
mpq_t				Q;
mpz_t				DVAL;
struct calib_Zx_obj		QUOT, REM;
struct RatInt			global_data;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (var < 0));

	if (op -> numer.nterms <= 0) {
		/* Integral of zero is zero. */
		gp = calib_genrep_si (0);
		return (gp);
	}

	/* Division by zero? */
	CALIB_FATAL_ERROR_IF (op -> denom.nterms <= 0);

	rip	= &global_data;
	init_global_data (rip, op, var);

	/* Convert to Zx. */
	Zx = calib_get_Zx_dom ();

	R = convert_to_Zx (&(op -> numer), var);
	S = convert_to_Zx (&(op -> denom), var);

	/* Remove integer contents into Q. */
	mpq_init (Q);
	Zx -> prim_part (mpq_numref (Q), R, R);
	Zx -> prim_part (mpq_denref (Q), S, S);
	mpq_canonicalize (Q);		/* Might need to flip signs. */

	Zx -> init (&QUOT);
	Zx -> init (&REM);
	mpz_init (DVAL);

	/* Split off any polynomial part via division:			*/
	/*								*/
	/*		DVAL * R = QUOT * S + REM			*/
	/* ==>		R 	 = (QUOT / DVAL) * S + (REM / DVAL)	*/
	/* ==>		R/S	 = (QUOT / DVAL) + (REM / DVAL) / S	*/
	/* ==>		R/S	 = (1/DVAL) * (QUOT + REM / S)		*/
	/* ==>		Q*R/S	 = (Q/DVAL) * (QUOT + REM / S).		*/

	Zx -> div (&QUOT, &REM, DVAL, R, S);

	calib_mpq_div_z (Q, Q, DVAL);

	integrate_polynomial_part (rip, Q, &QUOT);

	primitize (Q, &REM);

	integrate_primitive_rational_part (rip, Q, &REM, S);

	mpz_clear (DVAL);
	Zx -> clear (&REM);
	Zx -> clear (&QUOT);
	Zx -> free (S);
	Zx -> free (R);
	mpq_clear (Q);

	gp = assemble_result (rip);

	clear_global_data (rip);

	return (gp);
}

/*
 * Initialze the given global data (RatInt) object.
 */

	static
	void
init_global_data (

struct RatInt *		rip,	/* IN/OUT: RatInt object to initialize */
const struct calib_rat_obj *
			expr,	/* IN: Original integrand */
int			var	/* IN: Index of var to integrate */
)
{
int				nvars;
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR (expr EQ NULL));

	dom	= expr -> dom;
	Zxyz	= dom -> ring;
	nvars	= Zxyz -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	memset (rip, 0, sizeof (*rip));

	rip -> dom		= dom;
	rip -> integrand	= expr;
	rip -> var		= var;
	rip -> varname		= strdup (Zxyz -> vars [var]);

	grlist_init (&(rip -> poly));
	grlist_init (&(rip -> rats));
	grlist_init (&(rip -> atans));
	grlist_init (&(rip -> logs));
	grlist_init (&(rip -> integrals));
	dom -> init (dom, &(rip -> rat));
}

/*
 * Clear out the given global data.
 */

	static
	void
clear_global_data (

struct RatInt *		rip	/* IN/OUT: RatInt object to clear */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF (rip EQ NULL);

	dom	= rip -> dom;

	CALIB_FREE (rip -> varname);

	dom -> clear (&(rip -> rat));

	memset (rip, 0, sizeof (*rip));
}

/*
 * Integrate the polynomial portion.
 */

	static
	void
integrate_polynomial_part (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	zpoly	/* IN: Zx polynomial */
)
{
int				deg;
const struct calib_Qx_dom *	Qx;
struct calib_Qx_factor *	flist;
struct calib_genrep *		gp;
struct calib_Qx_obj		QPOLY;

	CALIB_FATAL_ERROR_IF ((rip   EQ NULL) OR
			      (Q     EQ NULL) OR
			      (zpoly EQ NULL));

	deg = zpoly -> degree;
	if (deg < 0) {
		/* Polynomial part is zero. */
		return;
	}

	Qx = calib_get_Qx_dom ();

	Qx -> init (&QPOLY);
	Qx -> set_Zx (&QPOLY, zpoly);

	CALIB_FATAL_ERROR_IF (QPOLY.degree NE deg);

	Qx -> mul_q (&QPOLY, &QPOLY, Q);

	Qx -> integral (&QPOLY, &QPOLY);

	if (flag_factor_poly) {
		flist = Qx -> factor (&QPOLY);
		gp = Qx -> factors_to_genrep (flist, rip -> varname);
		Qx -> free_factors (flist);
	}
	else {
		gp = Qx -> to_genrep (&QPOLY, rip -> varname);
	}

	grlist_append (&(rip -> poly), gp);

	Qx -> clear (&QPOLY);
}

/*
 * Integrate the reduced rational portion.  The integrand has the form:
 *
 *			Q * (R / S)
 *
 * where R, S \in Z[x], and Q is a rational number.
 * We know that R and S are primitive, and we should have deg(R) < deg(S).
 */

	static
	void
integrate_primitive_rational_part (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S	/* IN: Denominator polynomial */
)
{
int			deg1, deg2;
struct pfrac *		sqfp;
struct pfrac *		pfp;

	CALIB_FATAL_ERROR_IF ((rip   EQ NULL) OR
			      (Q     EQ NULL) OR
			      (R     EQ NULL) OR
			      (S     EQ NULL));

	deg1	= R -> degree;
	deg2	= S -> degree;

	CALIB_FATAL_ERROR_IF (deg1 >= deg2);

	if (deg1 < 0) {
		/* Rational part is zero. */
		return;
	}

	sqfp = square_free_partial_fraction_decomp (rip, Q, R, S);

	for (pfp = sqfp; pfp NE NULL; pfp = pfp -> next) {
		hermite_reduction (rip, pfp);
	}

	pfrac_free (sqfp);
}

/*
 * We are given F = Q*(R/S).  Perform a partial fraction decomposition
 * (based upon a square-free factorization of the denominator) having
 * the following form:
 *
 *	F = \sum_{i=1}^m	Q_i * R_i / (S_i)^i
 */

	static
	struct pfrac *
square_free_partial_fraction_decomp (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S	/* IN: Denominator polynomial */
)
{
int				k;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_factor *	sqfactors;
struct calib_Zx_factor *	fp;
struct calib_Zx_obj *		poly;
struct pfrac *			root;
struct pfrac **			hookp;
struct pfrac *			pfp;
mpz_t				CFACT, DVAL;
mpq_t				QFACT, QF2;
struct calib_Zx_obj		NUMER, DENOM, G, A, B, SK, QUOT, REM;

	Zx = calib_get_Zx_dom ();

	Zx -> init (&NUMER);
	Zx -> init (&DENOM);
	Zx -> init (&G);
	Zx -> init (&A);
	Zx -> init (&B);
	Zx -> init (&SK);
	Zx -> init (&QUOT);
	Zx -> init (&REM);
	mpz_init (CFACT);
	mpz_init (DVAL);
	mpq_init (QFACT);
	mpq_init (QF2);

	Zx -> set (&NUMER, R);
	Zx -> set (&DENOM, S);

	/* Compute square-free factorization of denominator S. */
	sqfactors = Zx -> factor_square_free (S);

	/* Process and remove constant factors (e.g., -1). */
	sqfactors = remove_constant_factors (CFACT, sqfactors);

	calib_mpq_div_z (QFACT, Q, CFACT);

	root	= NULL;
	hookp	= &root;
	for (fp = sqfactors; fp NE NULL; fp = fp -> next) {
		poly	= fp -> factor;
		k	= fp -> multiplicity;
		Zx -> ipow (&SK, poly, k);
		Zx -> div (&QUOT, &REM, DVAL, &DENOM, &SK);
		/* Division must be exact. */
		CALIB_FATAL_ERROR_IF (REM.degree >= 0);
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (DVAL, 1) NE 0);

		/* Compute A*QUOT + B*SK = 1. */
		Zx -> extgcd (&G, &A, &B, &QUOT, &SK);
		CALIB_FATAL_ERROR_IF (G.degree NE 0);
		CALIB_FATAL_ERROR_IF (mpz_sgn (&G.coeff [0]) EQ 0);

		/* Fold G into denominator of QFACT.	*/
		calib_mpq_div_z (QFACT, QFACT, &G.coeff [0]);

		Zx -> mul (&A, &A, &NUMER);
		Zx -> div (NULL, &A, DVAL, &A, &SK);
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (DVAL, 1) NE 0);

		Zx -> mul (&B, &B, &NUMER);
		Zx -> div (NULL, &B, DVAL, &B, &QUOT);
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (DVAL, 1) NE 0);

		/* We now have:					*/
		/*						*/
		/*	A*QUOT + B*SK = NUMER			*/
		/* ==>	A*QUOT/DENOM + B*SK/DENOM = NUMER/DENOM	*/
		/* ==>	A/SK + B/QUOT = NUMER/DENOM		*/
		/*						*/
		/* Emit a partial fraction for QFACT*A/poly^k	*/
		/* (but move integer content from A into	*/
		/* rational multiplier first).			*/
		Zx -> prim_part (DVAL, &A, &A);
		calib_mpq_mul_z (QF2, QFACT, DVAL);

		pfp = pfrac_create (QF2, &A, poly, k);
		*hookp = pfp;
		hookp = &(pfp -> next);

		/* Now move integer content from B into QFACT.	*/
		Zx -> prim_part (DVAL, &B, &B);
		calib_mpq_mul_z (QFACT, QFACT, DVAL);

		Zx -> set (&NUMER, &B);
		Zx -> set (&DENOM, &QUOT);
	}

	Zx -> free_factors (sqfactors);

	mpq_clear (QF2);
	mpq_clear (QFACT);
	mpz_clear (DVAL);
	Zx -> clear (&REM);
	Zx -> clear (&QUOT);
	Zx -> clear (&SK);
	Zx -> clear (&B);
	Zx -> clear (&A);
	Zx -> clear (&G);
	Zx -> clear (&DENOM);
	Zx -> clear (&NUMER);

	return (root);
}

/*
 * Remove constant factors from the given list of square-free factors,
 * returning their product in zval.
 */

	static
	struct calib_Zx_factor *
remove_constant_factors (

mpz_ptr			zval,		/* IN/OUT: Constant multiplier */
struct calib_Zx_factor *
			sqfactors	/* IN: List of square-free factors */
)
{
int				k;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_factor *	root;
struct calib_Zx_factor *	fp;
struct calib_Zx_factor **	hookp;
struct calib_Zx_obj *		poly;
mpz_t				CFACT, TEMP;

	Zx = calib_get_Zx_dom ();

	mpz_init (CFACT);
	mpz_init (TEMP);

	mpz_set_ui (CFACT, 1);

	root	= sqfactors;
	hookp	= &root;
	for (;;) {
		fp = *hookp;
		if (fp EQ NULL) break;

		poly	= fp -> factor;
		k	= fp -> multiplicity;

		CALIB_FATAL_ERROR_IF (k <= 0);

		if (poly -> degree <= 0) {
			/* We have a constant factor. */
			CALIB_FATAL_ERROR_IF (poly -> degree < 0);

			mpz_pow_ui (TEMP, &(poly -> coeff [0]), k);
			mpz_mul (CFACT, CFACT, TEMP);

			/* Remove this factor from the list. */
			Zx -> free (fp -> factor);
			*hookp = fp -> next;
			fp -> next = NULL;
			free (fp);
		}
		else {
			/* Non-constant factor.  Keep it. */
			hookp = &(fp -> next);
		}
	}

	mpz_set (zval, CFACT);

	mpz_clear (TEMP);
	mpz_clear (CFACT);

	return (root);
}

/*
 * Use Hermite's reduction to integrate a partial fraction of the form:
 *
 *		Q * (R / S^k)
 *
 * where R,S \in Z[x], primitive and relatively prime, Q is a rational
 * number, and k >= 1.
 */

	static
	void
hermite_reduction (

struct RatInt *			rip,	/* IN/OUT: Global data */
const struct pfrac *		pfp	/* IN: Square-free partial fraction */
)
{
int				k;
const struct calib_Zx_dom *	Zx;
mpz_t				DVAL, ZCONT, Za, Zb, Zc, Zd, Zg, ZD, ZEprime;
mpq_t				Q, BCQ, QD, QE, QEprime, TMP;
struct calib_Zx_obj		R, S, Sprime, G, B, C, D, E, Eprime;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR (pfp EQ NULL));

	if (flag_trace) {
		struct calib_genrep *	gp;
		printf ("Hermite reduction:\n");
		gp = make_rational_func_term (rip,
					      pfp -> Q,
					      &(pfp -> R),
					      &(pfp -> S),
					      pfp -> k);
		calib_genrep_prettyprint (gp);
		calib_genrep_free (gp);
	}

	k	= pfp -> k;

	if (k <= 1) {
		CALIB_FATAL_ERROR_IF (k < 1);

		/* Use Trager's method to integrate the final	*/
		/* "simple" rational function.			*/
		tragers_method (rip, pfp -> Q, &(pfp -> R), &(pfp -> S));
		return;
	}

	Zx = calib_get_Zx_dom ();

	mpz_init (DVAL);
	mpz_init (ZCONT);
	mpz_init (Za);
	mpz_init (Zb);
	mpz_init (Zc);
	mpz_init (Zd);
	mpz_init (Zg);
	mpz_init (ZD);
	mpz_init (ZEprime);
	mpq_init (Q);
	mpq_init (BCQ);
	mpq_init (QD);
	mpq_init (QE);
	mpq_init (QEprime);
	mpq_init (TMP);
	Zx -> init (&R);
	Zx -> init (&S);
	Zx -> init (&Sprime);
	Zx -> init (&G);
	Zx -> init (&B);
	Zx -> init (&C);
	Zx -> init (&D);
	Zx -> init (&E);
	Zx -> init (&Eprime);

	mpq_set (Q, pfp -> Q);
	Zx -> set (&R, &(pfp -> R));
	Zx -> set (&S, &(pfp -> S));

	/* We will need one or more iterations of Hermite's reduction.	*/
	/* Perform necessary initialization.				*/
	Zx -> derivative (&Sprime, &S);
	Zx -> extgcd (&G, &B, &C, &S, &Sprime);
	CALIB_FATAL_ERROR_IF (G.degree NE 0);
	mpq_set_ui (BCQ, 1, 1);
	calib_mpq_div_z (BCQ, BCQ, &G.coeff [0]);

	while (k > 1) {
		/* Apply Hermite's reduction. */
		Zx -> mul (&D, &B, &R);
		mpq_set (QD, BCQ);
		primitize (QD, &D);

		Zx -> mul (&E, &C, &R);
		mpq_set (QE, BCQ);
		primitize (QE, &E);

		/* Form rational function -E/((k-1)*S^(k-1)).	*/
		/* Note: we need to reduce E modulo S here.	*/
		mpq_set_ui (TMP, 1, k-1);
		mpq_mul (TMP, TMP, QE);
		mpq_mul (TMP, TMP, Q);
		mpq_neg (TMP, TMP);

		emit_rational_func_term (rip, TMP, &E, &S, k - 1);

		/* Compute reduced sub-problem. */
		Zx -> derivative (&Eprime, &E);
		mpq_set (QEprime, QE);
		primitize (QEprime, &Eprime);
		mpz_set_ui (mpq_numref (TMP), 1);
		mpz_set_ui (mpq_denref (TMP), k - 1);
		mpq_mul (QEprime, QEprime, TMP);

		mpz_set (Za, mpq_numref (QD));
		mpz_set (Zb, mpq_denref (QD));
		mpz_set (Zc, mpq_numref (QEprime));
		mpz_set (Zd, mpq_denref (QEprime));
		mpz_gcd (Zg, Zb, Zd);
		mpz_divexact (ZD, Zd, Zg);
		mpz_mul (ZD, ZD, Za);
		mpz_divexact (ZEprime, Zb, Zg);
		mpz_mul (ZEprime, ZEprime, Zc);
		Zx -> mul_z (&D, &D, ZD);
		Zx -> mul_z (&Eprime, &Eprime, ZEprime);
		Zx -> add (&D, &D, &Eprime);
		mpz_divexact (Zb, Zb, Zg);
		mpz_mul (Zb, Zb, Zd);
		calib_mpq_div_z (Q, Q, Zb);
		primitize (Q, &D);

		Zx -> set (&R, &D);
		--k;
	}

	CALIB_FATAL_ERROR_IF (k < 1);

	/* Reduce numerator modulo denominator so that deg(num)<deg(den). */
	Zx -> div (NULL, &R, DVAL, &R, &S);
	primitize (Q, &R);

	if (R.degree >= 0) {
		/* Now use Trager's method to integrate the	*/
		/* final "simple" rational function.		*/
		tragers_method (rip, Q, &R, &S);
	}
	else {
		/* No logarithmic part. */
	}

	Zx -> clear (&Eprime);
	Zx -> clear (&E);
	Zx -> clear (&D);
	Zx -> clear (&C);
	Zx -> clear (&B);
	Zx -> clear (&G);
	Zx -> clear (&Sprime);
	Zx -> clear (&S);
	Zx -> clear (&R);
	mpq_clear (TMP);
	mpq_clear (QEprime);
	mpq_clear (QE);
	mpq_clear (QD);
	mpq_clear (BCQ);
	mpq_clear (Q);
	mpz_clear (ZEprime);
	mpz_clear (ZD);
	mpz_clear (Zg);
	mpz_clear (Zd);
	mpz_clear (Zc);
	mpz_clear (Zb);
	mpz_clear (Za);
	mpz_clear (ZCONT);
	mpz_clear (DVAL);
}

/*
 * Emit a rational function term having the following form:
 *
 *		Q * (R / S^k)
 */

	static
	void
emit_rational_func_term (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S,	/* IN: Denominator polynomial */
int				k	/* IN: Denominator exponent */
)
{
int				sign;
const struct calib_rat_dom *	dom;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
mpq_t				QABS;
struct calib_rat_obj		RAT;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR
			      (Q   EQ NULL) OR
			      (R   EQ NULL) OR
			      (S   EQ NULL) OR
			      (k   <= 0));

	dom = rip -> dom;

	sign = mpq_sgn (Q);

	mpq_init (QABS);
	mpq_abs (QABS, Q);

	gp1 = make_rational_func_term (rip, QABS, R, S, k);

	mpq_clear (QABS);

	if (sign < 0) {
		gp2 = calib_genrep_neg (gp1);
		calib_genrep_free (gp1);
		gp1 = gp2;
	}

	dom -> init (dom, &RAT);
	dom -> set_genrep (&RAT, gp1);
	dom -> add (&(rip -> rat), &(rip -> rat), &RAT);
	calib_genrep_free (gp1);
	dom -> clear (&RAT);
}

/*
 * Use Trager's method to integrate a "simple" rational function having the
 * form:
 *
 *		Q * (R / S)
 *
 * where R,S \in Z[x], primitive and relatively prime, S is square-free,
 * and Q is a rational number.
 */

	static
	void
tragers_method (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S	/* IN: Denominator polynomial */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
const struct calib_Zx_dom *	Zx;
const struct calib_Zxyz_dom *	xydom;
struct qroot *			root_list;
struct qroot *			qrp;
mpz_ptr				N, D;
const char *			vars [2];
mpq_t				Ci, CVAL;
struct calib_Zx_obj		Sprime, RES, Vplus, Vminus;
struct calib_Zxyz_obj		H, Rxy, Sxy, Sprimexy, Yxy, RESxy;
struct calib_rat_obj		RAT;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR
			      (Q   EQ NULL) OR
			      (R   EQ NULL) OR
			      (S   EQ NULL));

	dom	= rip -> dom;
	Zxyz	= dom -> ring;
	Zx	= calib_get_Zx_dom ();

	/* Make a Zxyz domain with variables x and y. */
	vars [0] = "x";
	vars [1] = "y";
	xydom	 = calib_make_Zxyz_dom (2, vars);

	mpq_init (Ci);
	mpq_init (CVAL);
	Zx -> init (&Sprime);
	Zx -> init (&RES);
	Zx -> init (&Vplus);
	Zx -> init (&Vminus);
	xydom -> init (xydom, &H);
	xydom -> init (xydom, &Rxy);
	xydom -> init (xydom, &Sxy);
	xydom -> init (xydom, &Sprimexy);
	xydom -> init (xydom, &Yxy);
	xydom -> init (xydom, &RESxy);
	dom -> init (dom, &RAT);

	/* Construct polynomial H(x,y) = R(x) - y*S'(x).	*/
	Zx -> derivative (&Sprime, S);
	xydom -> copy_from_Zx (&Rxy, R, 0);
	xydom -> copy_from_Zx (&Sxy, S, 0);
	xydom -> copy_from_Zx (&Sprimexy, &Sprime, 0);
	xydom -> set_var_power (&Yxy, 1, 1);
	xydom -> mul (&H, &Yxy, &Sprimexy);
	xydom -> sub (&H, &Rxy, &H);

	/* Now compute resultant_x (H(x,y), S(x)).	*/
	xydom -> resultant (&RESxy, &H, &Sxy, 0);

	xydom -> copy_into_Zx (&RES, &RESxy, 1);

	/* Obtain all rational roots Ci of this resultant.  This list	*/
	/* contains strictly positive Ci values in increasing order,	*/
	/* together with separate flags indicating whether +Ci and/or	*/
	/* -Ci are roots.  This allows us to easily generate ratios of	*/
	/* logs, when appropriate.					*/
	root_list = rational_root_list (&RES);

	/* Form local copy of R/S integrand in "rat" form.  We deflate	*/
	/* this rational func for each log term that Trager's method	*/
	/* reveals (by subtracting off the derivative of said log	*/
	/* term).  Once we have processed all of the rational roots of	*/
	/* this resultant, we are left with either zero, or a rational	*/
	/* integrand that cannot be resolved via Trager's method.	*/
	Zxyz -> copy_from_Zx (&RAT.numer, R, 0);
	Zxyz -> copy_from_Zx (&RAT.denom, S, 0);

	for (qrp = root_list; qrp NE NULL; qrp = qrp -> next) {
		mpq_set (Ci, qrp -> root);

		CALIB_FATAL_ERROR_IF (mpq_sgn (Ci) <= 0);

		N = mpq_numref (Ci);
		D = mpq_denref (Ci);

		Zx -> set_si (&Vplus, 1);
		if (qrp -> pflag) {
			/* Process positive rational root +N/D. */
			trager_split (rip, Q, &Vplus, N, D, &RAT);
		}

		Zx -> set_si (&Vminus, 1);
		if (qrp -> mflag) {
			/* Process negative rational root -N/D. */
			mpz_neg (N, N);
			trager_split (rip, Q, &Vminus, N, D, &RAT);
			mpz_abs (N, N);
		}

		mpq_mul (CVAL, Q, Ci);

		if (qrp -> pflag) {
			if (qrp -> mflag) {
				/* We have a ratio of logs. */
				process_log_term (rip, CVAL, &Vplus, &Vminus);
			}
			else {
				/* Only positive Ci. */
				process_log_term (rip, CVAL, &Vplus, NULL);
			}
		}
		else {
			/* Only negative Ci. */
			mpq_neg (CVAL, CVAL);
			process_log_term (rip, CVAL, &Vminus, NULL);
		}
	}
	free_rational_roots (root_list);

	if (RAT.numer.nterms > 0) {
		/* We have a rational integrand remaining that Trager's	*/
		/* method did not solve.				*/
		process_trager_residuals (rip, Q, &RAT);
	}

	dom -> clear (&RAT);
	xydom -> clear (&RESxy);
	xydom -> clear (&Yxy);
	xydom -> clear (&Sprimexy);
	xydom -> clear (&Sxy);
	xydom -> clear (&Rxy);
	xydom -> clear (&H);
	Zx -> clear (&Vminus);
	Zx -> clear (&Vplus);
	Zx -> clear (&RES);
	Zx -> clear (&Sprime);
	mpq_clear (CVAL);
	mpq_clear (Ci);

	calib_free_Zxyz_dom (xydom);
}

/*
 * Given a polynomial in Z[x], return a list of all its rational roots.
 * The list contains strictly positive rational Ci in strictly increasing
 * order, together with separate flags indicating whether +Ci and/or -Ci
 * are roots.
 */

	static
	struct qroot *
rational_root_list (

const struct calib_Zx_obj *	poly	/* IN: Polynomial for which to get */
					/*     rational roots */
)
{
int				cmp, sign;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_factor *	flist;
struct calib_Zx_factor *	fp;
struct calib_Zx_obj *		factor;
struct qroot *			root_list;
struct qroot **			hookp;
struct qroot *			qrp;
mpq_t				ROOT, Ci;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	mpq_init (ROOT);
	mpq_init (Ci);

	Zx = calib_get_Zx_dom ();

	/* Factor poly.  Note: we only need factors of degree one --	*/
	/* we should implement a "factor_degree_limit" method in Z[x]	*/
	/* that efficiently supports this.  FIXME!			*/

	flist = Zx -> factor (poly);

	root_list = NULL;
	for (fp = flist; fp NE NULL; fp = fp -> next) {
		factor = fp -> factor;
		if (factor -> degree NE 1) {
			/* Not a rational root.  Skip. */
			continue;
		}

		calib_mpq_set_z_z (ROOT,
				   &(factor -> coeff [0]),
				   &(factor -> coeff [1]));
		mpq_neg (ROOT, ROOT);

		mpq_abs (Ci, ROOT);

		/* Find (or add) Ci to the root_list. */
		hookp = &root_list;
		for (;;) {
			qrp = *hookp;
			if (qrp EQ NULL) break;
			cmp = mpq_cmp (qrp -> root, Ci);
			if (cmp EQ 0) {
				/* Found existing node with Ci. */
				break;
			}
			if (cmp > 0) {
				/* Must insert new node here. */
				qrp = NULL;
				break;
			}
			hookp = &(qrp -> next);
		}
		if (qrp EQ NULL) {
			/* Must add new entry for Ci at *hookp. */
			qrp = CALIB_NEW (struct qroot);
			mpq_init (qrp -> root);
			mpq_set  (qrp -> root, Ci);
			qrp -> pflag	= FALSE;
			qrp -> mflag	= FALSE;
			qrp -> next	= *hookp;
			*hookp = qrp;
		}
		sign = mpq_sgn (ROOT);
		if (sign > 0) {
			CALIB_FATAL_ERROR_IF (qrp -> pflag);
			qrp -> pflag = TRUE;
		}
		else if (sign < 0) {
			CALIB_FATAL_ERROR_IF (qrp -> mflag);
			qrp -> mflag = TRUE;
		}
		else {
			CALIB_FATAL_ERROR;
		}
	}

	Zx -> free_factors (flist);

	mpq_clear (Ci);
	mpq_clear (ROOT);

	return (root_list);
}

/*
 * Free the given list of rational roots.
 */

	static
	void
free_rational_roots (

struct qroot *		list	/* IN: Linked list of roots to free */
)
{
struct qroot *		tmp;

	while (list NE NULL) {
		tmp		= list -> next;
		list -> next	= NULL;

		mpq_clear (list -> root);

		memset (list, 0, sizeof (*list));
		free (list);

		list = tmp;
	}
}

/*
 * Split off the V polynomial associated with the given rational root
 * Ci = N/D, reducing the rational function accordingly.
 */

	static
	void
trager_split (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational multiplier */
struct calib_Zx_obj *		V,	/* IN/OUT: V polynomial */
mpz_srcptr			N,	/* IN: Numerator of Ci */
mpz_srcptr			D,	/* IN: Denominator of Ci */
struct calib_rat_obj *		RAT	/* IN: Rational function to deflate */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_obj		R, S, Sprime, J, TMP;
struct calib_rat_obj		RAT2;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR
			      (V   EQ NULL) OR
			      (N   EQ NULL) OR
			      (D   EQ NULL) OR
			      (RAT EQ NULL));

	dom	= rip -> dom;
	Zxyz	= dom -> ring;
	Zx	= calib_get_Zx_dom ();

	Zx -> init (&R);
	Zx -> init (&S);
	Zx -> init (&Sprime);
	Zx -> init (&J);
	Zx -> init (&TMP);
	dom -> init (dom, &RAT2);

	/* Let R/S be the remaining rational integrand. */
	Zxyz -> copy_into_Zx (&R, &(RAT -> numer), rip -> var);
	Zxyz -> copy_into_Zx (&S, &(RAT -> denom), rip -> var);

	/* Compute polynomial J(x) = D*R(x) - N*S'(x). */
	Zx -> derivative (&Sprime, &S);
	Zx -> mul_z (&J, &R, D);
	Zx -> mul_z (&TMP, &Sprime, N);
	Zx -> sub (&J, &J, &TMP);

	/* Compute TMP = gcd (J, S). */
	Zx -> gcd (&TMP, &J, &S);

	/* GCD doesn't always give us a primitive polynomial.  (Integer	*/
	/* content comes outside the log as an additive integration	*/
	/* constant.)							*/
	Zx -> prim_part (NULL, V, &TMP);

	/* "Deflate" this term from our current rational integrand by	*/
	/* subtracting off Ci*V'/V (where Ci = N/D).			*/
	Zx -> derivative (&TMP, V);
	Zxyz -> copy_from_Zx (&RAT2.numer, &TMP, rip -> var);
	Zxyz -> copy_from_Zx (&RAT2.denom, V,    rip -> var);
	Zxyz -> mul_z (&RAT2.numer, &RAT2.numer, N);
	Zxyz -> mul_z (&RAT2.denom, &RAT2.denom, D);
	dom -> canonicalize (&RAT2);

	dom -> sub (RAT, RAT, &RAT2);

	dom -> clear (&RAT2);
	Zx -> clear (&TMP);
	Zx -> clear (&J);
	Zx -> clear (&Sprime);
	Zx -> clear (&S);
	Zx -> clear (&R);
}

/*
 * Process a term of the form C*log(Vnum/Vden).  Note that Vden is allowed
 * to be NULL, in which case it is assumed to be 1.
 */

	static
	void
process_log_term (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			C,	/* IN: Rational constant */
const struct calib_Zx_obj *	Vnum,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	Vden	/* IN: Denominator polynomial */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_Zxyz_factor *	flist;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
struct calib_rat_obj		RAT;

	CALIB_FATAL_ERROR_IF ((rip  EQ NULL) OR
			      (C    EQ NULL) OR
			      (Vnum EQ NULL));

	dom	= rip -> dom;
	Zxyz	= dom -> ring;

	/* Form rational function Vnum/Vden. */

	dom -> init (dom, &RAT);
	Zxyz -> copy_from_Zx (&RAT.numer, Vnum, rip -> var);
	if (Vden NE NULL) {
		Zxyz -> copy_from_Zx (&RAT.denom, Vden, rip -> var);
	}

	gp1 = calib_genrep_q (C);

	if (flag_factor_log) {
		flist = dom -> factor (&RAT);
		gp2 = Zxyz -> factors_to_genrep (flist);
		Zxyz -> free_factors (flist);
	}
	else {
		gp2 = dom -> to_genrep (&RAT);
	}
	dom -> clear (&RAT);

	gp2 = calib_genrep_log (gp2);

	gp3 = calib_genrep_mul2 (gp1, gp2);
	calib_genrep_free (gp2);
	calib_genrep_free (gp1);

	grlist_append (&(rip -> logs), gp3);
}

/*
 * Process a residual rational function that Trager's method was unable
 * to solve.  We first factor the denominator.  If there are two or more
 * factors, then partial-fraction decomposition is a good next step.
 * (Without introducing algebraic constants C, none of these partial
 * fractions will have the form C*V'/V needed to get new log terms, but
 * we can potentially discover atan() terms, and/or split the problem
 * into smaller / simpler pieces.)
 */

	static
	void
process_trager_residuals (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_rat_obj *	rat	/* IN: Rational function */
)
{
int				nfactors, deg;
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_factor *	flist;
struct calib_Zx_factor *	fp;
struct calib_Zx_factor **	hookp;
struct calib_Zx_obj *		poly;
struct pfrac *			pfracs;
struct pfrac *			pfp;
mpq_t				QVAL, QR, QS;
struct calib_Zx_obj		R, S;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR (Q EQ NULL) OR (rat EQ NULL));

	dom	= rat -> dom;
	Zxyz	= dom -> ring;
	Zx	= calib_get_Zx_dom ();

	mpq_init (QVAL);
	mpq_init (QR);
	mpq_init (QS);
	Zx -> init (&R);
	Zx -> init (&S);

	mpq_set (QVAL, Q);
	mpq_set_ui (QR, 1, 1);
	mpq_set_ui (QS, 1, 1);

	Zxyz -> copy_into_Zx (&R, &(rat -> numer), 0);
	Zxyz -> copy_into_Zx (&S, &(rat -> denom), 0);
	primitize (QR, &R);
	primitize (QS, &S);
	mpq_mul (QVAL, QVAL, QR);
	mpq_div (QVAL, QVAL, QS);

	/* It is possible that a single term of the numerator forms an	*/
	/* arc-tangent with the denominator.  Subtract out such a term	*/
	/* if present.							*/
	subtract_out_arc_tangent (rip, QVAL, &R, &S);

	if (R.degree < 0) {
		/* Nothing left to do. */
	}
	else {
		/* Factor the denominator S.  If we have two or more	*/
		/* factors, then we will do partial-fraction		*/
		/* decomposition.					*/

		flist = Zx -> factor (&S);

		/* Count the factors of degree >= 1. */
		nfactors = 0;
		hookp = &flist;
		for (;;) {
			fp = *hookp;
			if (fp EQ NULL) break;

			/* S should already be square-free. */
			CALIB_FATAL_ERROR_IF (fp -> multiplicity NE 1);

			poly = fp -> factor;
			deg = poly -> degree;
			if (deg > 0) {
				/* We have a factor of positive degree. */
				++nfactors;
				hookp = &(fp -> next);
			}
			else {
				CALIB_FATAL_ERROR_IF (deg < 0);
				/* We made S be primitive above.  Only	*/
				/* valid case here is poly EQ -1.	*/
				CALIB_FATAL_ERROR_IF (mpz_cmp_si (&(poly -> coeff [0]),
								  -1) NE 0);

				/* Negate QVAL and remove this factor. */
				mpq_neg (QVAL, QVAL);

				*hookp = fp -> next;
				fp -> next = NULL;
				Zx -> free_factors (fp);
			}
		}

		if (nfactors >= 2) {
			pfracs = partial_fraction_decomposition (rip,
								 QVAL,
								 &R,
								 nfactors,
								 flist);

			for (pfp = pfracs; pfp NE NULL; pfp = pfp -> next) {
				CALIB_FATAL_ERROR_IF (pfp -> k NE 1);

				find_arc_tangent (rip,
						  pfp -> Q,
						  &(pfp -> R),
						  &(pfp -> S));
			}

			pfrac_free (pfracs);
		}
		else {
			find_arc_tangent (rip, QVAL, &R, &S);
		}

		Zx -> free_factors (flist);
	}

	Zx -> clear (&S);
	Zx -> clear (&R);
	mpq_clear (QS);
	mpq_clear (QR);
	mpq_clear (QVAL);
}

/*
 * Perform partial-fraction decomposition on the given Q*R/S, where
 * S is given as a list of N factors.
 */

	static
	struct pfrac *
partial_fraction_decomposition (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
int				nfact,	/* IN: Number of denominator factors */
struct calib_Zx_factor *	flist	/* IN: List of denominator factors */
)
{
int				i;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_factor *	fp;
struct calib_Zx_obj **		array;
struct calib_Zx_obj *		poly;
struct pfrac *			root;
struct pfrac **			hookp;
struct pfrac *			pfp;
mpz_srcptr			gval;
mpz_t				DVAL;
mpq_t				QVAL, QA, QB;
struct calib_Zx_obj		NUMER, DENOM, S1, REM, G, A, B;

	CALIB_FATAL_ERROR_IF ((rip   EQ NULL) OR
			      (Q     EQ NULL) OR
			      (R     EQ NULL) OR
			      (flist EQ NULL) OR
			      (nfact  < 2));

	/* Make an array of the denominator factors. */
	array	= CALIB_NEWA (nfact, struct calib_Zx_obj *);
	i = 0;
	for (fp = flist; fp NE NULL; fp = fp -> next) {
		CALIB_FATAL_ERROR_IF (fp -> multiplicity NE 1);
		array [i++] = fp -> factor;
	}
	CALIB_FATAL_ERROR_IF (i NE nfact);

	Zx	= calib_get_Zx_dom ();

	mpz_init (DVAL);
	mpq_init (QVAL);
	mpq_init (QA);
	mpq_init (QB);
	Zx -> init (&NUMER);
	Zx -> init (&DENOM);
	Zx -> init (&S1);
	Zx -> init (&REM);
	Zx -> init (&G);
	Zx -> init (&A);
	Zx -> init (&B);

	/* Get local copy of numerator and denominator polynomials.	*/
	/* (We need to multiply out the denominator factors.)		*/
	mpq_set (QVAL, Q);
	Zx -> set (&NUMER, R);

	Zx -> set_si (&DENOM, 1);
	for (i = 0; i < nfact; i++) {
		Zx -> mul (&DENOM, &DENOM, array [i]);
	}

	/* Perform partial fraction decomposition. */
	root = NULL;
	hookp = &root;
	for (i = nfact - 1; i > 0; i--) {
		poly = array [i];

		/* Get product of other factors via division. */
		Zx -> div (&S1, &REM, DVAL, &DENOM, poly);

		/* Division must be exact. */
		CALIB_FATAL_ERROR_IF (REM.degree >= 0);
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (DVAL, 1) NE 0);

		/* Solve A*S1 + B*poly = 1. */
		Zx -> extgcd (&G, &A, &B, &S1, poly);
		CALIB_FATAL_ERROR_IF (G.degree NE 0);
		gval = &G.coeff [0];

		mpz_set_ui (mpq_numref (QA), 1);
		mpz_set    (mpq_denref (QA), gval);
		mpz_set_ui (mpq_numref (QB), 1);
		mpz_set    (mpq_denref (QB), gval);

		Zx -> mul (&A, &A, &NUMER);
		Zx -> div (NULL, &A, DVAL, &A, poly);
		calib_mpq_div_z (QA, QA, DVAL);
		primitize (QA, &A);

		Zx -> mul (&B, &B, &NUMER);
		Zx -> div (NULL, &B, DVAL, &B, &S1);
		calib_mpq_div_z (QB, QB, DVAL);
		primitize (QB, &B);

		/* Generate partial function for A/poly. */
		mpq_mul (QA, QA, QVAL);
		pfp = pfrac_create (QA, &A, poly, 1);

		*hookp = pfp;
		hookp = &(pfp -> next);

		/* Reset QVAL, NUMER, DENOM to be just for remaining	*/
		/* factors.						*/
		mpq_mul (QB, QB, QVAL);
		mpq_set (QVAL, QB);
		Zx -> set (&NUMER, &B);
		Zx -> set (&DENOM, &S1);
	}

	/* Generate final partial fraction. */
	pfp = pfrac_create (QVAL, &NUMER, &DENOM, 1);
	*hookp = pfp;

	Zx -> clear (&B);
	Zx -> clear (&A);
	Zx -> clear (&G);
	Zx -> clear (&REM);
	Zx -> clear (&S1);
	Zx -> clear (&DENOM);
	Zx -> clear (&NUMER);
	mpq_clear (QB);
	mpq_clear (QA);
	mpq_clear (QVAL);
	mpz_clear (DVAL);

	free (array);

	return (root);
}

/*
 * Handle the case where a single term of R forms an arc-tangent with S.
 * If this happens, subtract this term out from R and emit said arc-tangent.
 */

	static
	void
subtract_out_arc_tangent (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
struct calib_Zx_obj *		R,	/* IN/OUT: Numerator polynomial */
const struct calib_Zx_obj *	S	/* IN: Denominator polynomial */
)
{
int				k;
const struct calib_Zx_dom *	Zx;
mpz_srcptr			zp;
mpq_t				QVAL;
struct calib_Zx_obj		R2, R3;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR
			      (Q   EQ NULL) OR
			      (R   EQ NULL) OR
			      (S   EQ NULL));

	k = S -> degree;

	/* Linear denominators should have become log() terms by now. */
	CALIB_FATAL_ERROR_IF (k < 2);

	if ((k & 1) NE 0) {
		/* Odd degree. */
		return;
	}

	Zx = calib_get_Zx_dom ();

	if (k EQ 2) {
		/* Denominator is quadratic, but irreducible.  This	*/
		/* affords some very special cases.			*/
		handle_quadratic_denominator (rip, Q, R, S);

		/* Nothing remains for our caller to consider. */
		Zx -> set_si (R, 0);
		return;
	}

	k = k / 2;

	if (R -> degree < k - 1) {
		/* Required numerator coefficient does not exist. */
		return;
	}

	zp = &(R -> coeff [k - 1]);
	if (mpz_sgn (zp) EQ 0) {
		/* Required numerator coefficient does not exist. */
		return;
	}

	/* An atan() is potentially available.  Try generating it now. */
	Zx -> init (&R2);
	Zx -> init (&R3);
	mpq_init (QVAL);

	/* Generate term we're considering to subtract from R. */
	Zx -> set_var_power (&R2, k - 1);
	mpz_set (&R2.coeff [k - 1], zp);

	/* Now get primitive term for arc-tangent consideration. */
	Zx -> set (&R3, &R2);

	mpq_set (QVAL, Q);
	primitize (QVAL, &R3);

	if (emit_arc_tangent (rip, QVAL, &R3, S)) {
		/* Remove this term from R. */
		Zx -> sub (R, R, &R2);
	}

	mpq_clear (QVAL);
	Zx -> clear (&R3);
	Zx -> clear (&R2);
}

/*
 * We have R, S both irreducible and primitive, deg(R) < deg(S), and further
 * that S is quadratic:
 *
 *		S	=	a*x^2 + b*x + c
 *
 * We perform a special trick here that always completely solves the
 * negative discriminant case (a log(), an atan(), or both -- but never
 * any unfinished integrals).
 *
 * We do not currently introduce sqrt() with log() terms, which would be
 * necessary in the case of positive discriminant, so these do become
 * unfinished integrals for now.
 */

	static
	void
handle_quadratic_denominator (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S	/* IN: Denominator polynomial */
)
{
int				sign;
calib_bool			ok;
const struct calib_Zx_dom *	Zx;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
mpz_srcptr			a, b, c;
mpz_t				D, TEMP, DVAL;
mpq_t				F, G;
struct calib_Zx_obj		Sprime, QUOT, REM, R2;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR
			      (Q   EQ NULL) OR
			      (R   EQ NULL) OR
			      (S   EQ NULL));

	CALIB_FATAL_ERROR_IF (S -> degree NE 2);
	CALIB_FATAL_ERROR_IF ((R -> degree < 0) OR (R -> degree > 1));

	a	= &(S -> coeff [2]);
	b	= &(S -> coeff [1]);
	c	= &(S -> coeff [0]);

	Zx	= calib_get_Zx_dom ();

	mpz_init (D);
	mpz_init (TEMP);
	mpz_init (DVAL);
	mpq_init (F);
	mpq_init (G);
	Zx -> init (&Sprime);
	Zx -> init (&QUOT);
	Zx -> init (&REM);
	Zx -> init (&R2);

	/* Compute the discriminant. */
	mpz_mul (D, b, b);
	mpz_mul_ui (TEMP, a, 4);
	mpz_mul (TEMP, TEMP, c);
	mpz_sub (D, D, TEMP);

	sign = mpz_sgn (D);

	/* Denominator is irreducible and cannot have repeated roots. */
	CALIB_FATAL_ERROR_IF (sign EQ 0);

	/* Compute rational constants F and G such that R decomposes as	*/
	/*								*/
	/*		R = F*S' + G					*/

	Zx -> derivative (&Sprime, S);
	Zx -> div (&QUOT, &REM, DVAL, R, &Sprime);

	if (QUOT.degree >= 0) {
		CALIB_FATAL_ERROR_IF (QUOT.degree > 0);

		calib_mpq_set_z_z (F, &QUOT.coeff [0], DVAL);

		/* We integrate F*S'/S, giving rise to (Q*F) * log(S). */
		mpq_mul (F, Q, F);
		gp1 = calib_genrep_q (F);
		gp2 = Zx -> to_genrep (S, rip -> varname);
		gp2 = calib_genrep_log (gp2);
		gp3 = calib_genrep_mul2 (gp1, gp2);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		gp1 = gp3;
		grlist_append (&(rip -> logs), gp1);
	}

	if (REM.degree >= 0) {
		CALIB_FATAL_ERROR_IF (REM.degree > 0);

		calib_mpq_set_z_z (G, &REM.coeff [0], DVAL);

		mpq_mul (G, Q, G);
		Zx -> set_si (&R2, 1);

		if (sign > 0) {
			/* We could decompose this into two log()	*/
			/* terms, but not without introducing sqrt().	*/
			unfinished_rat_integral (rip, G, &R2, S, 1);
		}
		else {
			ok = emit_arc_tangent (rip, G, &R2, S);

			/* This should always generate an atan(). */
			CALIB_FATAL_ERROR_IF (NOT ok);
		}
	}

	Zx -> clear (&R2);
	Zx -> clear (&REM);
	Zx -> clear (&QUOT);
	Zx -> clear (&Sprime);
	mpq_clear (G);
	mpq_clear (F);
	mpz_clear (DVAL);
	mpz_clear (TEMP);
	mpz_clear (D);
}

/*
 * Consider whether the given Q*R/S integrates as an arctangent.
 * If not, we punt it as an unfinished integral.
 */

	static
	void
find_arc_tangent (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S	/* IN: Denominator polynomial */
)
{
	if (NOT emit_arc_tangent (rip, Q, R, S)) {
		unfinished_rat_integral (rip, Q, R, S, 1);
	}
}

/*
 * Return TRUE if-and-only-if the integrand has the form required for an
 * arc-tangent that we herein emit.  Returns FALSE otherwise.
 */

	static
	calib_bool
emit_arc_tangent (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S	/* IN: Denominator polynomial */
)
{
int				i, k, deg, num_nz;
calib_bool			result;
const struct calib_Zx_dom *	Zx;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
mpz_srcptr			lc, tc, a, b, c;
mpq_t				C1, QAP;
mpz_t				D, TEMP, A, B;
struct calib_Zx_obj		ARGPOLY;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR
			      (Q   EQ NULL) OR
			      (R   EQ NULL) OR
			      (S   EQ NULL));

	deg = S -> degree;
	if (deg < 2) {
		return (FALSE);
	}

	if ((deg & 0x01) NE 0) {
		/* Odd degree.  Cannot be atan(). */
		return (FALSE);
	}

	k	= deg / 2;

	lc	= &(S -> coeff [deg]);
	tc	= &(S -> coeff [0]);

	/* Cannot have constant term of 0:			*/
	/*	1/x would have generated a log() term.		*/
	/*	1(f(x) * x) would have split via pfrac.		*/
	CALIB_FATAL_ERROR_IF (mpz_sgn (tc) EQ 0);

	/* Cannot have lead coefficient of zero. */
	CALIB_FATAL_ERROR_IF (mpz_sgn (lc) EQ 0);

	/* Count the number of non-zero coefficients. */
	num_nz = 0;
	for (i = 0; i <= deg; i++) {
		if (mpz_sgn (&(S -> coeff [i])) NE 0) {
			++num_nz;
		}
	}
	CALIB_FATAL_ERROR_IF (num_nz < 2);

	if (num_nz > 3) {
		return (FALSE);
	}

	/* Have either 2 or 3 non-zero coefficients. */

	a	= lc;
	b	= &(S -> coeff [k]);
	c	= tc;

	if ((mpz_sgn (b) EQ 0) AND (num_nz NE 2)) {
		/* Coefficients are not x^2k, x^k, x^0. */
		return (FALSE);
	}

	/* Now examine the numerator, which must have just one	*/
	/* non-zero coefficient of degree k-1.			*/
	if (R -> degree NE k - 1) {
		return (FALSE);
	}
	for (i = k - 2; i >= 0; i--) {
		if (mpz_sgn (&(R -> coeff [i])) NE 0) {
			return (FALSE);
		}
	}

	/* This looks very promising.  Compute the discriminant. */
	mpz_init (D);
	mpz_init (TEMP);

	mpz_mul (D, b, b);
	mpz_mul_ui (TEMP, a, 4);
	mpz_mul (TEMP, TEMP, c);
	mpz_sub (D, D, TEMP);

	result = FALSE;
	if (mpz_sgn (D) < 0) {
		/* We have an arc-tangent! */

		mpz_neg (D, D);

		Zx = calib_get_Zx_dom ();
		Zx -> init (&ARGPOLY);
		mpz_init (A);
		mpz_init (B);
		mpq_init (C1);
		mpq_init (QAP);

		/* The integral is now:					*/
		/*							*/
		/*	2*atan((2*a*x^k + b) / sqrt(D)) / (k*sqrt(D))	*/
		/*							*/
		/* But of course we should simplify these sqrt() as	*/
		/* much as we can, getting rid of them completely when	*/
		/* possible.  Reduce sqrt(D) = A*sqrt(B), where		*/
		/* (hopefully) B is square-free.  This gives:		*/
		/*							*/
		/*		  2*a*x^k + b				*/
		/*	2 * atan (-----------)				*/
		/*		   A*sqrt(B)				*/
		/*      ----------------------				*/
		/*	     k*A*sqrt(B)				*/

		reduce_sqrt (A, B, D);

		Zx -> set_var_power (&ARGPOLY, k);
		mpz_mul_ui (TEMP, a, 2);
		Zx -> mul_z (&ARGPOLY, &ARGPOLY, TEMP);
		mpz_set (&ARGPOLY.coeff [0], b);
		mpz_set_ui (mpq_numref (QAP), 1);
		mpz_set    (mpq_denref (QAP), A);
		primitize (QAP, &ARGPOLY);

		gp1 = Zx -> to_genrep (&ARGPOLY, rip -> varname);

		gp2 = calib_genrep_z (mpq_numref (QAP));
		gp3 = calib_genrep_mul2 (gp2, gp1);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		gp1 = gp3;

		gp2 = make_sqrt (mpq_denref (QAP), B);
		gp3 = calib_genrep_div2 (gp1, gp2);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		gp1 = gp3;

		gp1 = calib_genrep_atan (gp1);

		mpq_set_ui (C1, 2, k);
		mpq_mul (C1, Q, C1);
		calib_mpq_div_z (C1, C1, A);

		gp2 = calib_genrep_z (mpq_numref (C1));
		gp3 = calib_genrep_mul2 (gp2, gp1);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		gp1 = gp3;

		gp2 = make_sqrt (mpq_denref (C1), B);
		gp3 = calib_genrep_div2 (gp1, gp2);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		gp1 = gp3;

		grlist_append (&(rip -> atans), gp1);

		mpq_clear (QAP);
		mpq_clear (C1);
		mpz_clear (B);
		mpz_clear (A);
		Zx -> clear (&ARGPOLY);

		result = TRUE;
	}

	mpz_clear (TEMP);
	mpz_clear (D);

	return (result);
}

/*
 * For given D, calculate A and B such that sqrt(D) = A*sqrt(B), where
 * B is (hopefully) square-free.  This requires factoring a general
 * integer, so we are limited in what we can do efficiently.
 */

	static
	void
reduce_sqrt (

mpz_ptr		A,		/* IN/OUT: Result A */
mpz_ptr		B,		/* IN/OUT: Result B */
mpz_srcptr	D		/* IN/OUT: Given integer D */
)
{
int		i;
mp_bitcnt_t	pow;
mpz_t		VAL, TEMP, A2, B2;

	CALIB_FATAL_ERROR_IF ((A EQ NULL) OR (B EQ NULL) OR (D EQ NULL));

	CALIB_FATAL_ERROR_IF (mpz_sgn (D) < 0);

	if (mpz_perfect_square_p (D)) {
		mpz_sqrt (A, D);
		mpz_set_ui (B, 1);
		return;
	}

	mpz_init (VAL);
	mpz_init (TEMP);
	mpz_init (A2);
	mpz_init (B2);

	mpz_set (VAL, D);

	mpz_set_ui (A2, 1);
	mpz_set_ui (B2, 1);

	/* Remove all factors of 2. */
	mpz_set_ui (TEMP, 2);
	pow = mpz_remove (VAL, VAL, TEMP);
	if ((pow & 1) NE 0) {
		mpz_mul_ui (B2, B2, 2);
	}
	mpz_ui_pow_ui (A2, 2, pow / 2);

	if (mpz_cmp_ui (VAL, 1) > 0) {
		for (i = 3; i < 10000; i += 2) {
			mpz_set_ui (TEMP, i);
			pow = mpz_remove (VAL, VAL, TEMP);
			if (pow EQ 0) continue;

			if ((pow & 1) NE 0) {
				mpz_mul_ui (B2, B2, i);
			}
			mpz_ui_pow_ui (TEMP, i, pow / 2);
			mpz_mul (A2, A2, TEMP);

			if (mpz_cmp_ui (VAL, 1) EQ 0) break;
		}
		/* Any residuals must remain inside the sqrt(). */
		mpz_mul (B2, B2, VAL);
	}

	mpz_set (A, A2);
	mpz_set (B, B2);

	mpz_clear (B2);
	mpz_clear (A2);
	mpz_clear (TEMP);
	mpz_clear (VAL);
}

/*
 * Make the appropriate genrep for A*sqrt(B).
 */

	static
	struct calib_genrep *
make_sqrt (

mpz_srcptr		A,		/* IN: Integer A */
mpz_srcptr		B		/* IN: Integer B */
)
{
struct calib_genrep *	gp1;
struct calib_genrep *	gp2;
struct calib_genrep *	gp3;

	CALIB_FATAL_ERROR_IF ((A EQ NULL) OR (B EQ NULL));

	gp1 = calib_genrep_z (A);
	if (mpz_cmp_ui (B, 1) EQ 0) {
		return (gp1);
	}

	gp2 = calib_genrep_z (B);
	gp2 = calib_genrep_sqrt (gp2);

	gp3 = calib_genrep_mul2 (gp1, gp2);
	calib_genrep_free (gp2);
	calib_genrep_free (gp1);
	gp1 = gp3;

	return (gp1);
}

/*
 * Construct a pfrac object for Q*R/S^k.
 */

	static
	struct pfrac *
pfrac_create (

mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S,	/* IN: Denominator polynomial */
int				k	/* IN: Denominator exponent */
)
{
struct pfrac *			p;
const struct calib_Zx_dom *	Zx;

	Zx = calib_get_Zx_dom ();

	p = CALIB_NEW (struct pfrac);

	memset (p, 0, sizeof (*p));

	mpq_init (p -> Q);
	Zx -> init (&(p -> R));
	Zx -> init (&(p -> S));

	mpq_set (p -> Q, Q);
	Zx -> set (&(p -> R), R);
	Zx -> set (&(p -> S), S);
	p -> k = k;

	return (p);
}

/*
 * Free up the given list of pfrac objects.
 */

	static
	void
pfrac_free (

struct pfrac *		list	/* IN: List of sqf_pfrac objects to free */
)
{
const struct calib_Zx_dom *	Zx;
struct pfrac *			p;
struct pfrac *			tmp;

	Zx = calib_get_Zx_dom ();

	p = list;
	while (p NE NULL) {
		tmp = p -> next;
		p -> next = NULL;

		mpq_clear (p -> Q);
		Zx -> clear (&(p -> R));
		Zx -> clear (&(p -> S));

		memset (p, 0, sizeof (*p));
		free (p);

		p = tmp;
	}
}

/*
 * Make a genrep representing
 *
 *		integrate (Q * (R / S^k), x)
 *
 * where R, S in Z[x] and primitive; Q is GMP rational and k >= 1.
 */

	static
	void
unfinished_rat_integral (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S,	/* IN: Denominator polynomial */
int				k	/* IN: Denominator exponent */
)
{
int				sign;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
mpq_t				QABS;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR
			      (Q   EQ NULL) OR
			      (R   EQ NULL) OR
			      (S   EQ NULL) OR
			      (k   <= 0));

	sign = mpq_sgn (Q);

	mpq_init (QABS);
	mpq_abs (QABS, Q);

	gp1 = make_rational_func_term (rip, QABS, R, S, k);

	mpq_clear (QABS);

	gp1 = make_integral (gp1, rip -> varname);

	if (sign < 0) {
		gp2 = calib_genrep_neg (gp1);
		calib_genrep_free (gp1);
		gp1 = gp2;
	}

	grlist_append (&(rip -> integrals), gp1);
}

/*
 * Make a genrep having the following form:
 *
 *	Q * (R / S^k)
 *
 * where R, S in Z[x] and primitive; Q is GMP rational; and k >= 1.
 */

	static
	struct calib_genrep *
make_rational_func_term (

struct RatInt *			rip,	/* IN/OUT: Global data */
mpq_srcptr			Q,	/* IN: Rational factor */
const struct calib_Zx_obj *	R,	/* IN: Numerator polynomial */
const struct calib_Zx_obj *	S,	/* IN: Denominator polynomial */
int				k	/* IN: Denominator exponent */
)
{
int				sign;
const struct calib_Zx_dom *	Zx;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
struct calib_genrep *		gp4;
mpq_t				QFACTOR;

	CALIB_FATAL_ERROR_IF ((rip EQ NULL) OR
			      (Q   EQ NULL) OR
			      (R   EQ NULL) OR
			      (S   EQ NULL) OR
			      (k   <= 0));

	sign	= mpq_sgn (Q);

	mpq_init (QFACTOR);
	mpq_abs (QFACTOR, Q);

	Zx = calib_get_Zx_dom ();

	gp1 = Zx -> to_genrep (R, rip -> varname);
	gp2 = Zx -> to_genrep (S, rip -> varname);

	/* Apply negative exponent to gp2. */
	gp3 = calib_genrep_ipow (gp2, - k);
	calib_genrep_free (gp2);
	gp2 = gp3;

	if (mpz_cmp_ui (mpq_numref (QFACTOR), 1) NE 0) {
		gp3 = calib_genrep_z (mpq_numref (QFACTOR));
		gp4 = calib_genrep_mul2 (gp3, gp1);
		calib_genrep_free (gp3);
		calib_genrep_free (gp1);
		gp1 = gp4;
	}
	if (mpz_cmp_ui (mpq_denref (QFACTOR), 1) NE 0) {
		gp3 = calib_genrep_z (mpq_denref (QFACTOR));
		gp4 = calib_genrep_ipow (gp3, -1);
		calib_genrep_free (gp3);
		gp3 = gp4;
		gp4 = calib_genrep_mul2 (gp3, gp2);
		calib_genrep_free (gp3);
		calib_genrep_free (gp2);
		gp2 = gp4;
	}

	gp3 = calib_genrep_mul2 (gp1, gp2);
	calib_genrep_free (gp1);
	calib_genrep_free (gp2);
	gp1 = gp3;

	if (sign < 0) {
		gp2 = calib_genrep_neg (gp1);
		calib_genrep_free (gp1);
		gp1 = gp2;
	}

	mpq_clear (QFACTOR);

	return (gp1);
}

/*
 * Make an integrate() genrep.
 */

	static
	struct calib_genrep *
make_integral (

struct calib_genrep *	expr,		/* IN: The integrand */
const char *		varname		/* IN: Name of integration var */
)
{
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_genrep *		f;

	gp = calib_genrep_var (varname);
	lp = calib_genrep_new_list (gp, NULL);

	lp = calib_genrep_new_list (expr, lp);

	f = calib_genrep_new ();
	f -> op		= CALIB_GENREP_OP_SI;
	f -> u.siop	= CALIB_GENREP_FUNC_INTEGRATE;

	gp = calib_genrep_new ();
	gp -> op		= CALIB_GENREP_OP_FUNC;
	gp -> u.func.func	= f;
	gp -> u.func.args	= lp;

	return (gp);
}

/*
 * Assemble the final result.
 */

	static
	struct calib_genrep *
assemble_result (

struct RatInt *		rip	/* IN/OUT: global data */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_obj *		R;
struct calib_Zx_obj *		S;
struct calib_Zxyz_factor *	flist;
struct grlist *			list;
struct calib_genrep *		gp;
struct calib_genrep_list *	lp;
struct grlist			terms;
mpz_t				DVAL;

	CALIB_FATAL_ERROR_IF (rip EQ NULL);

	dom	= rip -> dom;
	Zxyz	= dom -> ring;
	Zx	= calib_get_Zx_dom ();

	if (rip -> rat.numer.nterms > 0) {
		/* We have a rational function term.  Reduce its	*/
		/* numerator modulo its denominator (so that		*/
		/* degree(numerator) < degree (denominator)).  Failure	*/
		/* to do this would introduce additional, incorrect	*/
		/* polynomial terms into the solution.			*/
		R = convert_to_Zx (&(rip -> rat.numer), rip -> var);
		S = convert_to_Zx (&(rip -> rat.denom), rip -> var);
		mpz_init (DVAL);
		Zx -> div (NULL, R, DVAL, R, S);
		Zxyz -> copy_from_Zx (&(rip -> rat.numer), R, rip -> var);
		Zxyz -> mul_z (&(rip -> rat.denom), &(rip -> rat.denom), DVAL);
		dom -> canonicalize (&(rip -> rat));
		mpz_clear (DVAL);
		Zx -> free (S);
		Zx -> free (R);

		/* Convert updated rat to genrep form, placing this in	*/
		/* its proper position of the final form.		*/
		if (flag_factor_rat) {
			flist = dom -> factor (&(rip -> rat));
			gp = Zxyz -> factors_to_genrep (flist);
			Zxyz -> free_factors (flist);
		}
		else {
			gp = dom -> to_genrep (&(rip -> rat));
		}

		grlist_append (&(rip -> rats), gp);
	}

	/* Assemble all the terms of the final solution. */
	list	= &terms;
	grlist_init (list);

	grlist_take (list, &(rip -> poly));
	grlist_take (list, &(rip -> rats));
	grlist_take (list, &(rip -> atans));
	grlist_take (list, &(rip -> logs));
	grlist_take (list, &(rip -> integrals));

	lp = list -> root;
	grlist_init (list);

	if (lp EQ NULL) {
		gp = calib_genrep_si (0);
	}
	else if (lp -> next EQ NULL) {
		gp = lp -> operand;
		lp -> operand = NULL;
		lp -> operand = NULL;
		free (lp);
	}
	else {
		gp = calib_genrep_new ();
		gp -> op	= CALIB_GENREP_OP_ADD;
		gp -> u.list	= lp;
	}

	return (gp);
}

/*
 * Check the solution by differentiating it and comparing with the
 * original integrand.
 */

	static
	void
check_solution (

struct score *			score,	/* IN/OUT: Solution statistics */
const struct calib_rat_obj *	op,	/* IN: Original rational integrand */
const struct calib_genrep *	gp	/* IN: Integral solution */
)
{
calib_bool			correct;
const struct calib_rat_dom *	dom;
struct sqrt_rat			DERIV;

	dom	= op -> dom;

	sqrt_rat_init (dom, &DERIV);

	differentiate (&DERIV, gp);
	
	correct = FALSE;

	if (DERIV.err) {
		++(score -> deriv_errs);
	}
	else {
		if (mpq_cmp_ui (DERIV.Q, 1, 1) EQ 0) {
			/* All sqrt() have vanished / canceled. */
			dom -> sub (&DERIV.R, &DERIV.R, op);

			if (DERIV.R.numer.nterms EQ 0) {
				correct = TRUE;
			}
		}
	}
	if (NOT correct) {
		printf ("\n"
			"\t*** WARNING: SOLUTION NOT CORRECT ***\n"
			"\n");
		++(score -> incorrect);
	}

	if (has_function (gp, CALIB_GENREP_FUNC_INTEGRATE)) {
		++(score -> unfinished);
	}

	sqrt_rat_clear (&DERIV);
}

/*
 * Set rop to be the derivative of gp.
 */

	static
	void
differentiate (

struct sqrt_rat *		rop,	/* IN/OUT: Result */
const struct calib_genrep *	gp	/* IN: Expression to differentiate */
)
{
int					i, k, n;
const struct calib_rat_dom *		dom;
const struct calib_genrep_list *	lp;
const struct calib_genrep **		garray;
struct sqrt_rat *			rarray;
const struct calib_genrep *		gp1;
const struct calib_genrep *		base;
const struct calib_genrep *		fgp;
const struct calib_genrep_list *	args;
const struct calib_genrep *		arg1;
struct sqrt_rat				TEMP1, TEMP2, KONST;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (gp EQ NULL));

	dom = rop -> R.dom;

	switch (gp -> op) {
	case CALIB_GENREP_OP_ADD:
		sqrt_rat_init (dom, &TEMP1);
		sqrt_rat_set_si (rop, 0);
		for (lp = gp -> u.list; lp NE NULL; lp = lp -> next) {
			differentiate (&TEMP1, lp -> operand);
			sqrt_rat_add (rop, rop, &TEMP1);
		}
		sqrt_rat_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_MUL:
		sqrt_rat_init (dom, &TEMP1);
		sqrt_rat_init (dom, &TEMP2);
		sqrt_rat_init (dom, &KONST);

		/* Create an array of the *non-constant* factors, while	*/
		/* multiplying all of the constant factors into KONST.	*/
		/* Doing this allows us to never attempt calling	*/
		/* sqrt_rat_set_genrep() on any log(), atan() or	*/
		/* integrate() term (sqrt_rat has no way to represent	*/
		/* these).  But it is OK to differentiate() these	*/
		/* functions (as long as constant multipliers have	*/
		/* been stripped away so that they don't participate	*/
		/* in the product-rule).				*/
		n = 0;
		for (lp = gp -> u.list; lp NE NULL; lp = lp -> next) {
			if (NOT genrep_is_constant (lp -> operand)) {
				++n;
			}
		}
		garray = CALIB_NEWA (n, const struct calib_genrep *);
		rarray = CALIB_NEWA (n, struct sqrt_rat);

		sqrt_rat_set_si (&KONST, 1);
		i = 0;
		for (lp = gp -> u.list; lp NE NULL; lp = lp -> next) {
			gp1 = lp -> operand;
			if (genrep_is_constant (gp1)) {
				sqrt_rat_set_genrep (&TEMP1, gp1);
				sqrt_rat_mul (&KONST, &KONST, &TEMP1);
			}
			else {
				garray [i] = gp1;
				sqrt_rat_init (dom, &rarray [i]);
				sqrt_rat_set_genrep (&rarray [i], garray [i]);
				++i;
			}
		}
		CALIB_FATAL_ERROR_IF (i NE n);

		sqrt_rat_set_si (rop, 0);
		for (i = 0; i < n; i++) {
			sqrt_rat_set_si (&TEMP1, 1);
			for (k = 0; k < n; k++) {
				if (k EQ i) {
					differentiate (&TEMP2, garray [k]);
					sqrt_rat_mul (&TEMP1, &TEMP1, &TEMP2);
				}
				else {
					sqrt_rat_mul (&TEMP1, &TEMP1, &rarray [k]);
				}
			}
			sqrt_rat_add (rop, rop, &TEMP1);
		}
		sqrt_rat_mul (rop, rop, &KONST);

		for (i = 0; i < n; i++) {
			sqrt_rat_clear (&rarray [i]);
		}
		free (rarray);
		free (garray);

		sqrt_rat_clear (&KONST);
		sqrt_rat_clear (&TEMP2);
		sqrt_rat_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_IPOW:
		base	= gp -> u.ipow.base;
		k	= gp -> u.ipow.power;

		CALIB_FATAL_ERROR_IF (k EQ 0);

		sqrt_rat_init (dom, &TEMP1);
		sqrt_rat_init (dom, &TEMP2);

		sqrt_rat_set_genrep (&TEMP1, base);
		differentiate (&TEMP2, base);

		sqrt_rat_ipow (rop, &TEMP1, k - 1);
		sqrt_rat_mul (rop, rop, &TEMP2);
		sqrt_rat_set_si (&TEMP2, k);
		sqrt_rat_mul (rop, rop, &TEMP2);

		sqrt_rat_clear (&TEMP2);
		sqrt_rat_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_VAR:
		/* There is only one variable. */
		sqrt_rat_set_si (rop, 1);
		break;

	case CALIB_GENREP_OP_SI:
	case CALIB_GENREP_OP_Z:
	case CALIB_GENREP_OP_Q:
		sqrt_rat_set_si (rop, 0);
		break;

	case CALIB_GENREP_OP_FUNC:
		fgp	= gp -> u.func.func;
		args	= gp -> u.func.args;

		CALIB_FATAL_ERROR_IF ((fgp EQ NULL) OR (args EQ NULL));

		arg1	= args -> operand;

		CALIB_FATAL_ERROR_IF (fgp -> op NE CALIB_GENREP_OP_SI);

		switch (fgp -> u.siop) {
		case CALIB_GENREP_FUNC_LOG:
			sqrt_rat_init (dom, &TEMP1);
			sqrt_rat_init (dom, &TEMP2);

			sqrt_rat_set_genrep (&TEMP1, arg1);
			differentiate (&TEMP2, arg1);

			sqrt_rat_div (rop, &TEMP2, &TEMP1);

			sqrt_rat_clear (&TEMP2);
			sqrt_rat_clear (&TEMP1);
			break;

		case CALIB_GENREP_FUNC_SQRT:
			/* Verify that arg is an integer constant. */
			CALIB_FATAL_ERROR_IF (arg1 -> op NE CALIB_GENREP_OP_Z);
			sqrt_rat_set_si (rop, 0);
			break;

		case CALIB_GENREP_FUNC_ATAN:
			sqrt_rat_init (dom, &TEMP1);
			sqrt_rat_init (dom, &TEMP2);

			sqrt_rat_set_genrep (&TEMP2, arg1);
			sqrt_rat_square (&TEMP2, &TEMP2);
			sqrt_rat_set_si (&TEMP1, 1);
			sqrt_rat_add (&TEMP2, &TEMP2, &TEMP1);

			differentiate (&TEMP1, arg1);

			sqrt_rat_div (rop, &TEMP1, &TEMP2);

			sqrt_rat_clear (&TEMP2);
			sqrt_rat_clear (&TEMP1);
			break;

		case CALIB_GENREP_FUNC_INTEGRATE:
			sqrt_rat_set_genrep (rop, arg1);
			break;
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Return TRUE if-and-only-if the given genrep is constant (in the sense
 * that the value can be stored in a sqrt_rat object -- for example,
 * atan(3) does not qualify, but shouldn't happen either).
 */

	static
	calib_bool
genrep_is_constant (

const struct calib_genrep *	gp	/* IN: Genrep to test */
)
{
const struct calib_genrep_list *	lp;
const struct calib_genrep *		fgp;
const struct calib_genrep_list *	args;
const struct calib_genrep *		arg1;

	CALIB_FATAL_ERROR_IF (gp EQ NULL);

	switch (gp -> op) {
	case CALIB_GENREP_OP_ADD:
	case CALIB_GENREP_OP_MUL:
		for (lp = gp -> u.list; lp NE NULL; lp = lp -> next) {
			if (NOT genrep_is_constant (lp -> operand)) {
				return (FALSE);
			}
		}
		return (TRUE);

	case CALIB_GENREP_OP_IPOW:
		if (genrep_is_constant (gp -> u.ipow.base)) {
			return (TRUE);
		}
		return (FALSE);

	case CALIB_GENREP_OP_VAR:
		return (FALSE);

	case CALIB_GENREP_OP_SI:
	case CALIB_GENREP_OP_Z:
	case CALIB_GENREP_OP_Q:
		return (TRUE);

	case CALIB_GENREP_OP_FUNC:
		fgp	= gp -> u.func.func;
		args	= gp -> u.func.args;

		CALIB_FATAL_ERROR_IF ((fgp EQ NULL) OR (args EQ NULL));

		arg1	= args -> operand;

		CALIB_FATAL_ERROR_IF (fgp -> op NE CALIB_GENREP_OP_SI);

		switch (fgp -> u.siop) {
		case CALIB_GENREP_FUNC_SQRT:
			CALIB_FATAL_ERROR_IF (arg1 -> op NE CALIB_GENREP_OP_Z);
			return (TRUE);

		case CALIB_GENREP_FUNC_LOG:
		case CALIB_GENREP_FUNC_ATAN:
		case CALIB_GENREP_FUNC_INTEGRATE:
			CALIB_FATAL_ERROR_IF (genrep_is_constant (arg1));
			return (FALSE);

		default:
			CALIB_FATAL_ERROR;
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	return (FALSE);
}

/*
 * Initialize the given sqrt_rat object.
 */

	static
	void
sqrt_rat_init (

const struct calib_rat_dom *	dom,	/* IN: Rational domain */
struct sqrt_rat *		rop	/* IN/OUT: Object to initialize */
)
{
	rop -> err = FALSE;
	mpq_init (rop -> Q);
	dom -> init (dom, &(rop -> R));
}

/*
 * Clear the given sqrt_rat object.
 */

	static
	void
sqrt_rat_clear (

struct sqrt_rat *	op	/* IN: Object to clear */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	dom = op -> R.dom;

	dom -> clear (&(op -> R));
	mpq_clear (op -> Q);
	op -> err	= FALSE;

	memset (op, 0, sizeof (*op));
}

/*
 * Set rop to op in "sqrt_rat" domain.
 */

	static
	void
sqrt_rat_set (

struct sqrt_rat *	rop,		/* IN/OUT: Result */
const struct sqrt_rat *	op		/* IN: Source sqrt_rat operand */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {
		rop -> err = op -> err;

		mpq_set (rop -> Q, op -> Q);

		dom = rop -> R.dom;
		dom -> set (&(rop -> R), &(op -> R));
	}
}

/*
 * Set rop to op (an integer) in "sqrt_rat" domain.
 */

	static
	void
sqrt_rat_set_si (

struct sqrt_rat *	rop,		/* IN/OUT: Result */
calib_si_t		op		/* IN: Integer operand */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	mpq_set_ui (rop -> Q, 1, 1);

	dom = rop -> R.dom;

	dom -> set_si (&(rop -> R), op);
}

/*
 * Set rop to op (a GMP integer) in "sqrt_rat" domain.
 */

	static
	void
sqrt_rat_set_z (

struct sqrt_rat *	rop,		/* IN/OUT: Result */
mpz_srcptr		op		/* IN: Integer operand */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	mpq_set_ui (rop -> Q, 1, 1);

	dom = rop -> R.dom;

	dom -> set_z (&(rop -> R), op);
}

/*
 * Set rop to op (a GMP rational) in "sqrt_rat" domain.
 */

	static
	void
sqrt_rat_set_q (

struct sqrt_rat *	rop,		/* IN/OUT: Result */
mpq_srcptr		op		/* IN: Integer operand */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	mpq_set_ui (rop -> Q, 1, 1);

	dom = rop -> R.dom;

	dom -> set_q (&(rop -> R), op);
}

/*
 * Set rop to op1 + op2 in "sqrt_rat" domain.
 */

	static
	void
sqrt_rat_add (

struct sqrt_rat *		rop,	/* IN/OUT: Result */
const struct sqrt_rat *		op1,	/* IN: Operand 1 */
const struct sqrt_rat *		op2	/* IN: Operand 2 */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	if ((op1 -> err) OR (op2 -> err)) {
		rop -> err = TRUE;
	}

	if (op1 -> R.numer.nterms <= 0) {
		sqrt_rat_set (rop, op2);
	}
	else if (op2 -> R.numer.nterms <= 0) {
		sqrt_rat_set (rop, op1);
	}
	else {
		if (mpq_cmp (op1 -> Q, op2 -> Q) NE 0) {
			/* sqrt() factors do not match. */
			rop -> err = TRUE;
		}
		mpq_set (rop -> Q, op1 -> Q);
		dom = rop -> R.dom;
		dom -> add (&(rop -> R), &(op1 -> R), &(op2 -> R));
	}
}

/*
 * Set rop to op1 * op2 in "sqrt_rat" domain.
 */

	static
	void
sqrt_rat_mul (

struct sqrt_rat *		rop,	/* IN/OUT: Result */
const struct sqrt_rat *		op1,	/* IN: Operand 1 */
const struct sqrt_rat *		op2	/* IN: Operand 2 */
)
{
const struct calib_rat_dom *	dom;
mpz_ptr				zp;
mpq_t				QVAL, TEMP;
struct calib_rat_obj		RAT;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	if ((op1 -> err) OR (op2 -> err)) {
		rop -> err = TRUE;
	}

	dom = rop -> R.dom;

	mpq_init (QVAL);
	mpq_init (TEMP);
	dom -> init (dom, &RAT);

	mpq_mul (QVAL, op1 -> Q, op2 -> Q);
	dom -> mul (&RAT, &(op1 -> R), &(op2 -> R));

	zp = mpq_numref (QVAL);
	if ((mpz_cmp_ui (zp, 1) NE 0) AND mpz_perfect_square_p (zp)) {
		mpq_set_ui (TEMP, 1, 1);
		mpz_sqrt (mpq_numref (TEMP), zp);

		mpz_set_ui (zp, 1);

		dom -> mul_q (&RAT, &RAT, TEMP);
	}

	zp = mpq_denref (QVAL);
	if ((mpz_cmp_ui (zp, 1) NE 0) AND mpz_perfect_square_p (zp)) {
		mpq_set_ui (TEMP, 1, 1);
		mpz_sqrt (mpq_denref (TEMP), zp);

		mpz_set_ui (zp, 1);

		dom -> mul_q (&RAT, &RAT, TEMP);
	}

	mpq_set (rop -> Q, QVAL);
	dom -> set (&(rop -> R), &RAT);

	dom -> clear (&RAT);
	mpq_clear (TEMP);
	mpq_clear (QVAL);
}

/*
 * Set rop to op1 / op2 in "sqrt_rat" domain.
 */

	static
	void
sqrt_rat_div (

struct sqrt_rat *		rop,	/* IN/OUT: Result */
const struct sqrt_rat *		op1,	/* IN: Operand 1 */
const struct sqrt_rat *		op2	/* IN: Operand 2 */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	if ((op1 -> err) OR (op2 -> err)) {
		rop -> err = TRUE;
	}

	if (mpq_sgn (op2 -> Q) EQ 0) {
		rop -> err = TRUE;
		mpq_set_ui (rop -> Q, 1, 1);
	}
	else {
		mpq_div (rop -> Q, op1 -> Q, op2 -> Q);
	}

	dom = rop -> R.dom;

	if (op2 -> R.numer.nterms <= 0) {
		rop -> err = TRUE;
		sqrt_rat_set_si (rop, 0);
	}
	else {
		dom -> div (&(rop -> R), &(op1 -> R), &(op2 -> R));
	}
}

/*
 * Set rop to op^2 in "sqrt_rat" domain.
 * This eliminates sqrt() factors.
 */

	static
	void
sqrt_rat_square (

struct sqrt_rat *		rop,	/* IN/OUT: Result */
const struct sqrt_rat *		op	/* IN: Operand */
)
{
const struct calib_rat_dom *	dom;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (op -> err) {
		rop -> err = TRUE;
	}

	dom = rop -> R.dom;

	dom -> mul (&(rop -> R), &(op -> R), &(op -> R));
	dom -> mul_q (&(rop -> R), &(rop -> R), op -> Q);

	mpq_set_ui (rop -> Q, 1, 1);
}

/*
 * Set rop to op1 ^ op2 in "sqrt_rat" domain.
 */

	static
	void
sqrt_rat_ipow (

struct sqrt_rat *		rop,	/* IN/OUT: Result */
const struct sqrt_rat *		op1,	/* IN: Operand 1 */
int				op2	/* IN: Operand 2 */
)
{
int				k;
const struct calib_rat_dom *	dom;
mpq_t				QVAL1, QVAL2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL));

	if (op1 -> err) {
		rop -> err = TRUE;
	}

	k = op2;

	mpq_init (QVAL1);
	mpq_init (QVAL2);

	mpq_set (QVAL1, op1 -> Q);
	if (k < 0) {
		mpq_inv (QVAL1, QVAL1);
		k = - k;
	}

	mpq_set_ui (QVAL2, 1, 1);
	if ((k & 1) NE 0) {
		/* Odd power. */
		mpq_set (QVAL2, QVAL1);
	}
	k /= 2;
	mpz_pow_ui (mpq_numref (QVAL1), mpq_numref (QVAL1), k);
	mpz_pow_ui (mpq_denref (QVAL1), mpq_denref (QVAL1), k);

	dom = rop -> R.dom;

	dom -> ipow (&(rop -> R), &(op1 -> R), op2);
	dom -> mul_q (&(rop -> R), &(rop -> R), QVAL1);

	mpq_set (rop -> Q, QVAL2);

	mpq_clear (QVAL2);
	mpq_clear (QVAL1);
}

/*
 * Convert given expression "op" into sqrt_rat form, storing the
 * result into "rop".
 */

	static
	void
sqrt_rat_set_genrep (

struct sqrt_rat *		rop,	/* IN/OUT: Result */
const struct calib_genrep *	op	/* IN: Genrep expression */
)
{
int					k;
const struct calib_rat_dom *		dom;
const struct calib_genrep_list *	lp;
const struct calib_genrep *		base;
const struct calib_genrep *		fgp;
const struct calib_genrep_list *	args;
const struct calib_genrep *		arg1;
struct sqrt_rat				TEMP1;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> R.dom;

	switch (op -> op) {
	case CALIB_GENREP_OP_ADD:
		sqrt_rat_init (dom, &TEMP1);

		sqrt_rat_set_si (rop, 0);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			sqrt_rat_set_genrep (&TEMP1, lp -> operand);
			sqrt_rat_add (rop, rop, &TEMP1);
		}

		sqrt_rat_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_MUL:
		sqrt_rat_init (dom, &TEMP1);

		sqrt_rat_set_si (rop, 1);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			sqrt_rat_set_genrep (&TEMP1, lp -> operand);
			sqrt_rat_mul (rop, rop, &TEMP1);
		}

		sqrt_rat_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_IPOW:
		base	= op -> u.ipow.base;
		k	= op -> u.ipow.power;

		sqrt_rat_init (dom, &TEMP1);

		sqrt_rat_set_genrep (&TEMP1, base);
		sqrt_rat_ipow (rop, &TEMP1, k);

		sqrt_rat_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_VAR:
		sqrt_rat_set_si (rop, 0);
		dom -> set_var_power (&(rop -> R), 0, 1);
		break;

	case CALIB_GENREP_OP_SI:
		sqrt_rat_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		sqrt_rat_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		sqrt_rat_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_FUNC:
		fgp	= op -> u.func.func;
		args	= op -> u.func.args;

		CALIB_FATAL_ERROR_IF ((fgp EQ NULL) OR (args EQ NULL));

		arg1	= args -> operand;

		CALIB_FATAL_ERROR_IF (fgp -> op NE CALIB_GENREP_OP_SI);

		switch (fgp -> u.siop) {
		case CALIB_GENREP_FUNC_LOG:
		case CALIB_GENREP_FUNC_ATAN:
		case CALIB_GENREP_FUNC_INTEGRATE:
			/* These should all get multiplied by zero	*/
			/* during differentiation, which will cause	*/
			/* the error flag to disappear.			*/
			sqrt_rat_set_si (rop, 1);
			rop -> err = TRUE;
			break;

		case CALIB_GENREP_FUNC_SQRT:
			CALIB_FATAL_ERROR_IF (arg1 -> op NE CALIB_GENREP_OP_Z);
			sqrt_rat_set_si (rop, 1);
			mpq_set_z (rop -> Q, arg1 -> u.zop);
			break;

		default:
			CALIB_FATAL_ERROR;
			break;
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Convert the given Zxyz polynomial (in given var) to Zx form.
 */

	static
	struct calib_Zx_obj *
convert_to_Zx (

const struct calib_Zxyz_obj *	op,	/* IN: Zxyz polynomial to convert */
int				var	/* IN: Polynomial variable */
)
{
struct calib_Zx_obj *		poly;
const struct calib_Zxyz_dom *	Zxyz;
const struct calib_Zx_dom *	Zx;

	CALIB_FATAL_ERROR_IF ((op EQ 0) OR (var < 0));

	Zxyz	= op -> dom;
	Zx	= calib_get_Zx_dom ();

	poly = CALIB_NEW (struct calib_Zx_obj);

	Zx -> init (poly);
	Zxyz -> copy_into_Zx (poly, op, var);
	
	return (poly);
}

/*
 * Make the given polynomial be primitive, moving any integer content
 * into the given rational number Q (via multiplication into Q's current
 * value).
 */

	static
	void
primitize (

mpq_ptr			Q,		/* IN/OUT: Rational multiplier */
struct calib_Zx_obj *	poly		/* IN/OUT: Polynomial to make */
					/*	   primitive */
)
{
const struct calib_Zx_dom *	Zx;
mpz_t				ZCONT;

	mpz_init (ZCONT);

	Zx = calib_get_Zx_dom ();

	Zx -> prim_part (ZCONT, poly, poly);

	calib_mpq_mul_z (Q, Q, ZCONT);

	mpz_clear (ZCONT);
}

/*
 * Return TRUE if-and-only-if the given genrep contains at least one
 * occurrence of the given builtin function index.
 */

	static
	calib_bool
has_function (

const struct calib_genrep *	gp,	/* IN: Genrep to test */
calib_si_t			fidx	/* IN: Index of builtin function */
)
{
calib_bool				result;
const struct calib_genrep_list *	lp;
const struct calib_genrep *		fp;
const struct calib_genrep_list *	args;

	CALIB_FATAL_ERROR_IF ((gp EQ NULL) OR (fidx < 0));

	result = FALSE;

	switch (gp -> op) {
	case CALIB_GENREP_OP_ADD:
	case CALIB_GENREP_OP_MUL:
		for (lp = gp -> u.list; lp NE NULL; lp = lp -> next) {
			result = has_function (lp -> operand, fidx);
			if (result) break;
		}
		break;

	case CALIB_GENREP_OP_IPOW:
		result = has_function (gp -> u.ipow.base, fidx);
		break;

	case CALIB_GENREP_OP_VAR:
	case CALIB_GENREP_OP_SI:
	case CALIB_GENREP_OP_Z:
	case CALIB_GENREP_OP_Q:
		break;

	case CALIB_GENREP_OP_FUNC:
		/* Check the function. */
		fp	= gp -> u.func.func;
		args	= gp -> u.func.args;
		if (fp -> op EQ CALIB_GENREP_OP_SI) {
			if (fp -> u.siop EQ fidx) {
				result = TRUE;
				break;
			}
		}
		else {
			CALIB_FATAL_ERROR_IF (fp -> op NE CALIB_GENREP_OP_VAR);
		}
		/* Check all of the arguments. */
		for (lp = args; lp NE NULL; lp = lp -> next) {
			result = has_function (lp -> operand, fidx);
			if (result) break;
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	return (result);
}

/*
 * Initialize the given grlist to empty.
 */

	static
	void
grlist_init (

struct grlist *		list	/* IN/OUT: grlist to initialize */
)
{
	list -> root	= NULL;
	list -> hookp	= &(list -> root);
}

/*
 * Append the given expression onto the given grlist.
 */

	static
	void
grlist_append (

struct grlist *		list,	/* IN/OUT: grlist to append onto */
struct calib_genrep *	expr	/* IN: Expression to append */
)
{
struct calib_genrep_list *	lp;

	CALIB_FATAL_ERROR_IF ((list EQ NULL) OR (expr EQ NULL));

	lp	= calib_genrep_new_list (expr, NULL);

	*(list -> hookp)	= lp;
	list -> hookp		= &(lp -> next);
}

/*
 * Destructively *take* the items from list2 and append them onto list1.
 */

	static
	void
grlist_take (

struct grlist *		list1,		/* IN/OUT: List to append to */
struct grlist *		list2		/* IN/OUT: List to take from */
)
{
struct calib_genrep_list *	root;

	CALIB_FATAL_ERROR_IF ((list1 EQ NULL) OR
			      (list2 EQ NULL) OR
			      (list1 EQ list2));

	root	= list2 -> root;
	if (root NE NULL) {
		*(list1 -> hookp)	= root;
		list1 -> hookp		= list2 -> hookp;

		list2 -> root	= NULL;
		list2 -> hookp	= &(list2 -> root);
	}
}
