/**********************************************************************

	$Id: GFpk.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	GFpk.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Galois field GF(p^k), where p is a prime number and k > 1
	is an integer.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef GFPK_H
#define GFPK_H

#include "calib/types.h"
#include "calib/Zp.h"

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_genrep;
struct calib_GFpk_dom;
struct calib_GFpk_obj;
struct calib_Random;
struct calib_Za_dom;
struct calib_Za_obj;
struct calib_Zp_dom;
struct calib_Zpx_obj;

/*
 * An instance of a value in GF(p^k).
 */

struct calib_GFpk_obj {
	const struct calib_GFpk_dom *
		dom;		/* GF(p^k) domain containing this value */
	mpz_ptr	coeff;		/* Coefficients.  This is an array of	*/
				/* k integers in Z_p.			*/
};

/*
 * A Galois field GF(p^k).  This is equivalent to
 * Zp/f(x), where f(x) is a monic irreducible polynomial of
 * degree k in Zp[x], which must be specified.
 */

struct calib_GFpk_dom {
	const struct calib_Zp_dom *
			cf;	/* Coefficient field */
	int		k;	/* Degree of modulus polynomial */
	mpz_srcptr	coeff;	/* Coefficients of modulus polynomial */

	void		(*init) (const struct calib_GFpk_dom *	f,
				 struct calib_GFpk_obj *	x);
	void		(*clear) (struct calib_GFpk_obj *	x);
	void		(*set) (struct calib_GFpk_obj *		rop,
				const struct calib_GFpk_obj *	op);
	void		(*set_si) (struct calib_GFpk_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_GFpk_obj *	rop,
				  mpz_srcptr			op);
	void		(*set_q) (struct calib_GFpk_obj *	rop,
				  mpq_srcptr			op);
	void		(*set_var_power)
				(struct calib_GFpk_obj *	rop,
				 int				power);
	void		(*add) (struct calib_GFpk_obj *		rop,
				const struct calib_GFpk_obj *	op1,
				const struct calib_GFpk_obj *	op2);
	void		(*sub) (struct calib_GFpk_obj *		rop,
				const struct calib_GFpk_obj *	op1,
				const struct calib_GFpk_obj *	op2);
	void		(*neg) (struct calib_GFpk_obj *		rop,
				const struct calib_GFpk_obj *	op);
	void		(*mul) (struct calib_GFpk_obj *		rop,
				const struct calib_GFpk_obj *	op1,
				const struct calib_GFpk_obj *	op2);
	void		(*mul_z) (struct calib_GFpk_obj *	rop,
				  const struct calib_GFpk_obj *	op1,
				  mpz_srcptr			op2);
	void		(*mul_a) (struct calib_GFpk_obj *	rop,
				  const struct calib_GFpk_obj *	op);
	void		(*ipow) (struct calib_GFpk_obj *	rop,
				 const struct calib_GFpk_obj *	op,
				 int				power);
	void		(*inv) (struct calib_GFpk_obj *		rop,
				const struct calib_GFpk_obj *	op);
	void		(*pth_root) (struct calib_GFpk_obj *		rop,
				     const struct calib_GFpk_obj *	op);
	void		(*cvZa) (struct calib_GFpk_obj *	rop,
				 const struct calib_Za_obj *	op);
	int		(*degree) (const struct calib_GFpk_obj *	op);
	calib_bool	(*zerop) (const struct calib_GFpk_obj *	op);
	calib_bool	(*onep) (const struct calib_GFpk_obj *	op);
	void		(*set_random) (struct calib_GFpk_obj *	rop,
				       struct calib_Random *	randp);
	void		(*set_genrep) (struct calib_GFpk_obj *		rop,
				       const struct calib_genrep *	op,
				       const char *			var);
	struct calib_genrep *
			(*to_genrep) (const struct calib_GFpk_obj *	op,
				      const char *			var);
};

/*
 * Construct the field GF(p^k) = Zp/f(x), where f(x) is the given
 * monic, irreducible polynomial in Zp[x].
 */

extern void	calib_free_GFpk_dom (struct calib_GFpk_dom * cf);
extern struct calib_GFpk_dom *
		calib_make_GFpk_dom (
				const struct calib_Zp_dom *	cf,
				const struct calib_Zpx_obj *	gpoly);
extern struct calib_GFpk_dom *
		calib_make_GFpk_dom_z (
				const struct calib_Zp_dom *	cf,
				int				k,
				mpz_srcptr			gcoeff);

#ifdef __cplusplus
 }
#endif

#endif
