/**********************************************************************

	$Id: GFpkx.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	GFpkx.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring of polynomials K[x] over the field K = GF(q),
	where GF(q) is a Galois field, q = p^k, and p is a prime.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef GFPKX_H
#define GFPKX_H

#include "calib/Zp.h"

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_GFpk_dom;
struct calib_GFpk_obj;
struct calib_GFpkx_dom;
struct calib_GFpkx_factor;
struct calib_GFpkx_obj;
struct calib_Random;
struct calib_Zax_dom;
struct calib_Zax_obj;

/*
 * An instance of a polynomial in K[x], where K = GF(p^k).
 */

struct calib_GFpkx_obj {
	int	degree;		/* Degree of polynomial			*/
	int	size;		/* Size of coeff buffer (degree < size)	*/
	const struct calib_GFpkx_dom *
		dom;		/* Domain of polynomial			*/
	struct calib_GFpk_obj *
		coeff;		/* Coefficients of polynomial.		*/
};

/*
 * The following structure represents the factorization of a polynomial
 * over GF(p^k)[x] into monic, irreducible factors.  It forms a linked list
 * of factors, including the multiplicity of each factor.
 *
 * The factor() function:
 * - Produces only monic factors (of degree 1 or greater).  The caller
 *   must manually fetch the lead-coefficient to obtain the constant
 *   factor.
 * - Dynamically allocates the factors and (linked list thereof).  It is
 *   the responsibility of the caller to free these when done using the
 *   provided free_factors() function.
 */

struct calib_GFpkx_factor {
	int				multiplicity;
	struct calib_GFpkx_obj *	factor;
	struct calib_GFpkx_factor *	next;
};

/* A ring of polynomials over a field GF(p^k). */

struct calib_GFpkx_dom {
	const struct calib_GFpk_dom *
			cf;	/* Coefficient field */
	void		(*init) (const struct calib_GFpkx_dom *	K_of_x,
				 struct calib_GFpkx_obj *	x);
	void		(*init_degree)
				(const struct calib_GFpkx_dom *	K_of_x,
				 struct calib_GFpkx_obj *	x,
				 int				degree);
	void		(*alloc)
				(struct calib_GFpkx_obj *	rop,
				 int				degree);
	void		(*clear) (struct calib_GFpkx_obj *	x);
	void		(*set) (struct calib_GFpkx_obj *	rop,
				const struct calib_GFpkx_obj *	op);
	void		(*set_si) (struct calib_GFpkx_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_GFpkx_obj *	rop,
				  mpz_srcptr			op);
	void		(*set_q) (struct calib_GFpkx_obj *	rop,
				  mpq_srcptr			op);
	void		(*set_var_power) (
				struct calib_GFpkx_obj *	rop,
				int				power);
	void		(*add) (struct calib_GFpkx_obj *	rop,
				const struct calib_GFpkx_obj *	op1,
				const struct calib_GFpkx_obj *	op2);
	void		(*sub) (struct calib_GFpkx_obj *	rop,
				const struct calib_GFpkx_obj *	op1,
				const struct calib_GFpkx_obj *	op2);
	void		(*neg) (struct calib_GFpkx_obj *	rop,
				const struct calib_GFpkx_obj *	op);
	void		(*mul) (struct calib_GFpkx_obj *	rop,
				const struct calib_GFpkx_obj *	op1,
				const struct calib_GFpkx_obj *	op2);
	void		(*mul_z) (struct calib_GFpkx_obj *		rop,
				  const struct calib_GFpkx_obj *	op1,
				  mpz_srcptr				op2);
	void		(*ipow) (struct calib_GFpkx_obj *	rop,
				 const struct calib_GFpkx_obj *	op,
				 int				power);
	struct calib_GFpkx_obj *
			(*dup) (const struct calib_GFpkx_obj *	op);
	void		(*free) (struct calib_GFpkx_obj * poly);
	void		(*eval) (struct calib_GFpk_obj *	rop,
				 const struct calib_GFpkx_obj *	poly,
				 const struct calib_GFpk_obj *	value);
	void		(*div) (struct calib_GFpkx_obj *	quotient,
				struct calib_GFpkx_obj *	remainder,
				const struct calib_GFpkx_obj *	a,
				const struct calib_GFpkx_obj *	b);
	void		(*gcd) (struct calib_GFpkx_obj *	gcd,
				const struct calib_GFpkx_obj *	a,
				const struct calib_GFpkx_obj *	b);
	void		(*extgcd) (struct calib_GFpkx_obj *		gcd,
				   struct calib_GFpkx_obj *		xa,
				   struct calib_GFpkx_obj *		xb,
				   const struct calib_GFpkx_obj *	a,
				   const struct calib_GFpkx_obj *	b);
	struct calib_GFpkx_obj *
			(*cvZax) (const struct calib_GFpkx_dom *	K_of_x,
				  const struct calib_Zax_obj *		op);
	struct calib_GFpkx_factor *
			(*factor) (const struct calib_GFpkx_obj *	poly);
	void		(*free_factors) (struct calib_GFpkx_factor *	factors);
	void		(*set_random)
				(struct calib_GFpkx_obj *	rop,
				 int				d,
				 struct calib_Random *		randp);
	calib_bool	(*zerop) (const struct calib_GFpkx_obj *	op);
	calib_bool	(*onep) (const struct calib_GFpkx_obj *		op);
	void		(*set_genrep)
				(struct calib_GFpkx_obj *	rop,
				 const struct calib_genrep *	op,
				 const char *			xvar,
				 const char *			avar);
	struct calib_genrep *
			(*to_genrep)
				(const struct calib_GFpkx_obj *	op,
				 const char *			xvar,
				 const char *			avar);
	struct calib_genrep *
			(*factors_to_genrep) (
				const struct calib_GFpkx_factor *	factors,
				const char *				xvar,
				const char *				avar);
};

/*
 * Construct the ring of polynomials over field K, where K is
 * the given field GF(p^k).
 */

extern void	calib_free_GFpkx_dom (struct calib_GFpkx_dom * K_of_x);
extern struct calib_GFpkx_dom *
		calib_make_GFpkx_dom (const struct calib_GFpk_dom * K);

#ifdef __cplusplus
 }
#endif

#endif
