/**********************************************************************

	$Id: Qa.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	Qa.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The algebraic field Q(a), where a is an algebraic extension
	defined to be the root of a given irreducible polynomial
	g(a) \in Z[a].

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_QA_H
#define CALIB_QA_H

#include "calib/types.h"
#include <gmp.h>

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_genrep;
struct calib_Qa_dom;
struct calib_Qa_obj;
struct calib_Za_obj;
struct calib_Zx_obj;

/*
 * Which multiplicative inverse algorithm to use.
 */

enum calib_Qa_inv_method {
	CALIB_QA_INV_METHOD_MODULAR,
	CALIB_QA_INV_METHOD_PRS
};

/*
 * The "settings" object for the Qa domain.
 */

struct calib_Qa_settings {
	/* Multiplicative inverse method. */
	enum calib_Qa_inv_method	inv_method;
};

/*
 * Newly created Qa domains default to these settings.
 */

extern struct calib_Qa_settings		calib_Qa_default_settings;

/*
 * An instance of a value in Q(a).  We represent this as the product
 * of a rational multiplier times a primitive Z[x] polynomial whose leading
 * coefficient (if any) is strictly positive.
 */

struct calib_Qa_obj {
	mpq_t	qfact;		/* Rational multiplier			*/
	const struct calib_Qa_dom *
		dom;		/* Q(a) domain containing this value	*/
	int	degree;		/* Degree of this value in a */
	mpz_ptr	coeff;		/* Coefficients.  This is always an	*/
				/* array of k integers so that		*/
				/* reallocation is never required.	*/
};

/*
 * The algebraic field Q(a), where a is an algebraic extension
 * defined to be the root of a given irreducible polynomial g(a) \in Z[a].
 */

struct calib_Qa_dom {
	struct calib_Zx_obj *
			gpoly;	/* Generator polynomial */

	struct calib_Za_dom *
			Za_dom;	/* Corresponding Z(a) domain (or NULL) */

	const struct calib_Qa_dom *
			ai_dom;	/* Qa domain for algebraic integer	*/
				/* (or NULL)				*/
	struct calib_Qa_settings
			settings;	/* Algorithmic settings */
	void		(*init) (const struct calib_Qa_dom *	f,
				 struct calib_Qa_obj *		x);
	void		(*clear) (struct calib_Qa_obj *		x);
	void		(*set) (struct calib_Qa_obj *		rop,
				const struct calib_Qa_obj *	op);
	void		(*set_si) (struct calib_Qa_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_Qa_obj *	rop,
				  mpz_srcptr		op);
	void		(*set_q) (struct calib_Qa_obj *	rop,
				  mpq_srcptr		op);
	void		(*set_var_power)
				(struct calib_Qa_obj *	rop,
				 int			power);
	void		(*set_Za_q)
				(struct calib_Qa_obj *		rop,
				 const struct calib_Za_obj *	op1,
				 mpq_srcptr			op2);
	void		(*set_Zx)
				(struct calib_Qa_obj *		rop,
				 const struct calib_Zx_obj *	op);
	void		(*add) (struct calib_Qa_obj *		rop,
				const struct calib_Qa_obj *	op1,
				const struct calib_Qa_obj *	op2);
	void		(*sub) (struct calib_Qa_obj *		rop,
				const struct calib_Qa_obj *	op1,
				const struct calib_Qa_obj *	op2);
	void		(*neg) (struct calib_Qa_obj *		rop,
				const struct calib_Qa_obj *	op);
	void		(*mul) (struct calib_Qa_obj *		rop,
				const struct calib_Qa_obj *	op1,
				const struct calib_Qa_obj *	op2);
	void		(*mul_si)
				(struct calib_Qa_obj *		rop,
				 const struct calib_Qa_obj *	op1,
				 calib_si_t			op2);
	void		(*mul_z)
				(struct calib_Qa_obj *		rop,
				 const struct calib_Qa_obj *	op1,
				 mpz_srcptr			op2);
	void		(*mul_q)
				(struct calib_Qa_obj *		rop,
				 const struct calib_Qa_obj *	op1,
				 mpq_srcptr			op2);
	void		(*ipow) (struct calib_Qa_obj *		rop,
				 const struct calib_Qa_obj *	op,
				 int				power);
	void		(*inv) (struct calib_Qa_obj *		rop,
				const struct calib_Qa_obj *	op);
	calib_bool	(*is_algint)
				(const struct calib_Qa_dom *	dom);
	const struct calib_Qa_dom *
			(*algint_dom)
				(const struct calib_Qa_dom *	dom);
	void		(*to_algint)
				(struct calib_Qa_obj *		rop,
				 const struct calib_Qa_obj *	op);
	void		(*from_algint)
				(struct calib_Qa_obj *		rop,
				 const struct calib_Qa_obj *	op);
	int		(*degree) (const struct calib_Qa_obj *	op);
	calib_bool	(*zerop) (const struct calib_Qa_obj *	op);
	calib_bool	(*onep) (const struct calib_Qa_obj *	op);
	void		(*set_genrep) (struct calib_Qa_obj *		rop,
				       const struct calib_genrep *	op,
				       const char *			var);
	struct calib_genrep *
			(*to_genrep) (const struct calib_Qa_obj*	op,
				      const char *			var);
	size_t		(*bit_size)
				(const struct calib_Qa_obj *	op);
	mpq_ptr		(*get_coeffs) (const struct calib_Qa_obj *	op);
	void		(*set_coeffs) (struct calib_Qa_obj *	rop,
				       int			degree,
				       mpq_srcptr		coeffs);
	const struct calib_Za_dom *
			(*get_Za_dom) (const struct calib_Qa_dom *	f);
};

/*
 * Construct the field Q(a) = Q[a]/g(a), where g(a) is the given
 * irreducible polynomial in Z[a].
 */

extern void	calib_free_Qa_dom (struct calib_Qa_dom * cf);
extern struct calib_Qa_dom *
		calib_make_Qa_dom (const struct calib_Zx_obj * gpoly);

#ifdef __cplusplus
 }
#endif

#endif
