/**********************************************************************

	$Id: Qax.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	Qax.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring of polynomials K[x] over the algebraic
	extension field K = Q(a).

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_QAX_H
#define CALIB_QAX_H

#include "calib/types.h"
#include <gmp.h>

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_genrep;
struct calib_Qa_dom;
struct calib_Qa_obj;
struct calib_Qax_dom;
struct calib_Qax_factor;
struct calib_Qax_obj;
struct calib_Qax_settings;
struct calib_Zx_obj;

/*
 * Which factorization algorithm to use (for square-free polynomials).
 */

enum calib_Qax_factor_method {
	CALIB_QAX_FACTOR_METHOD_WEINBERGER_ROTHSCHILD,
	CALIB_QAX_FACTOR_METHOD_NORM
};

/*
 * The "settings" object for the Qax domain.
 */

struct calib_Qax_settings {
	/* Factorization method for square-free Qax polynomials. */
	enum calib_Qax_factor_method	factor_method;

	/* Print initial factors during Weinberger-Rothschild? */
	calib_bool			WR_print_initial_factors;

	/* Print lifted factors during Weinberger-Rothschild? */
	calib_bool			WR_print_lifted_factors;

	/* Print trial factor combinations during Weinberger-Rothschild? */
	calib_bool			WR_print_trial_factor_combinations;
};

/*
 * Newly created Qax domains default to these settings.
 */

extern struct calib_Qax_settings	calib_Qax_default_settings;

/*
 * An instance of a polynomial in K[x], where K = Q(a).
 */

struct calib_Qax_obj {
	int	degree;		/* Degree of polynomial			*/
	int	size;		/* Size of coeff buffer			*/
	const struct calib_Qax_dom *
		dom;		/* Qax domain of polynomial		*/
	struct calib_Qa_obj *
		coeff;		/* Coefficients of polynomial.		*/
};

/*
 * The following structure represents the factorization of a polynomial
 * over Q(a)[x] into irreducibles.  It forms a linked list of factors,
 * including the multiplicity of each factor.
 *
 * The factor() function:
 * - Produces an optional degree 0 factor, followed by monic factors of
 *   degree >= 1.
 * - Dynamically allocates the factors and (linked list thereof).  It is
 *   the responsibility of the caller to free these when done using the
 *   provided free_factors() function.
 */

struct calib_Qax_factor {
	int				multiplicity;
	struct calib_Qax_obj *		factor;
	struct calib_Qax_factor *	next;
};

/* A ring of polynomials over a field Q(a). */

struct calib_Qax_dom {
	const struct calib_Qa_dom *
			cf;		/* Coefficient field */
	struct calib_Zax_dom *
			Zax_dom; 	/* Corresponding Z(a)[x] domain */
	struct calib_Qax_dom *
			ai_dom;		/* Corresponding algebraic integer */
					/* domain (or NULL) */
	struct calib_Qax_settings
			settings;	/* Algorithmic settings */
	void		(*init) (const struct calib_Qax_dom *	K_of_x,
				 struct calib_Qax_obj *		x);
	void		(*init_degree) (const struct calib_Qax_dom *	K_of_x,
					struct calib_Qax_obj *		x,
					int				degree);
	void		(*alloc) (struct calib_Qax_obj *	rop,
				  int				degree);
	void		(*clear) (struct calib_Qax_obj *	x);
	void		(*set) (struct calib_Qax_obj *		rop,
				const struct calib_Qax_obj *	op);
	void		(*set_si) (struct calib_Qax_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_Qax_obj *	rop,
				  mpz_srcptr			op);
	void		(*set_q) (struct calib_Qax_obj *	rop,
				  mpq_srcptr			op);
	void		(*set_Qa) (struct calib_Qax_obj *	rop,
				   const struct calib_Qa_obj *	op);
	void		(*set_var_power)
				(struct calib_Qax_obj *		rop,
				 int				power);
	void		(*set_Zx) (struct calib_Qax_obj *	rop,
				   const struct calib_Zx_obj *	op);
	void		(*add) (struct calib_Qax_obj *		rop,
				const struct calib_Qax_obj *	op1,
				const struct calib_Qax_obj *	op2);
	void		(*sub) (struct calib_Qax_obj *		rop,
				const struct calib_Qax_obj *	op1,
				const struct calib_Qax_obj *	op2);
	void		(*neg) (struct calib_Qax_obj *		rop,
				const struct calib_Qax_obj *	op);
	void		(*mul) (struct calib_Qax_obj *		rop,
				const struct calib_Qax_obj *	op1,
				const struct calib_Qax_obj *	op2);
	void		(*ipow) (struct calib_Qax_obj *		rop,
				 const struct calib_Qax_obj *	op,
				 int				power);
	struct calib_Qax_obj *
			(*dup) (const struct calib_Qax_obj *	op);
	void		(*free) (struct calib_Qax_obj * poly);
	void		(*eval) (struct calib_Qa_obj *		rop,
				 const struct calib_Qax_obj *	poly,
				 const struct calib_Qa_obj *	value);
	void		(*eval_poly) (struct calib_Qax_obj *		result,
				      const struct calib_Qax_obj *	poly,
				      const struct calib_Qax_obj *	value);
	void		(*div) (struct calib_Qax_obj *		quotient,
				struct calib_Qax_obj *		remainder,
				const struct calib_Qax_obj *	a,
				const struct calib_Qax_obj *	b);
	void		(*gcd) (struct calib_Qax_obj *		gcd,
				const struct calib_Qax_obj *	a,
				const struct calib_Qax_obj *	b);
	void		(*extgcd) (struct calib_Qax_obj *	gcd,
				   struct calib_Qax_obj *	xa,
				   struct calib_Qax_obj *	xb,
				   const struct calib_Qax_obj *	a,
				   const struct calib_Qax_obj *	b);
	struct calib_Qax_factor *
			(*factor) (const struct calib_Qax_obj *	poly);
	struct calib_Qax_factor *
			(*factor_square_free)
				(const struct calib_Qax_obj *	sqfpoly);
	struct calib_Qax_factor *
			(*finish_sqf)
				(const struct calib_Qax_factor * sqfactors);
	void		(*free_factors)
				(struct calib_Qax_factor * factors);
	const struct calib_Qax_dom *
			(*algint_dom)
				(const struct calib_Qax_dom *	dom);
	void		(*to_algint)
				(struct calib_Qax_obj *		rop,
				 const struct calib_Qax_obj *	op);
	void		(*from_algint)
				(struct calib_Qax_obj *		rop,
				 const struct calib_Qax_obj *	op);
	calib_bool	(*zerop) (const struct calib_Qax_obj *	op);
	calib_bool	(*onep) (const struct calib_Qax_obj *	op);
	void		(*set_genrep)
				(struct calib_Qax_obj *		rop,
				 const struct calib_genrep *	op,
				 const char *			xvar,
				 const char *			avar);
	struct calib_genrep *
			(*to_genrep) (const struct calib_Qax_obj *	op,
				      const char *			xvar,
				      const char *			avar);
	struct calib_genrep *
			(*factors_to_genrep) (
				const struct calib_Qax_factor *	factors,
				const char *			xvar,
				const char *			avar);
	size_t		(*bit_size)
				(const struct calib_Qax_obj *	op);
	size_t		(*factors_bit_size)
				(const struct calib_Qax_factor *	factors);
	const struct calib_Zax_dom *
			(*get_Zax_dom) (const struct calib_Qax_dom *	K_of_x);
};

/*
 * Construct the ring of polynomials over field K, where K is
 * the given field Q(a).
 */

extern void	calib_free_Qax_dom (struct calib_Qax_dom * K_of_x);
extern struct calib_Qax_dom *
		calib_make_Qax_dom (const struct calib_Qa_dom * K);

#ifdef __cplusplus
 }
#endif

#endif
