/**********************************************************************

	$Id: Qx.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	Qx.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The ring Q[x] of polynomials over the rationals Q.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef QX_H
#define QX_H

#include "calib/types.h"
#include <gmp.h>

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_genrep;
struct calib_Qx_dom;
struct calib_Qx_factor;
struct calib_Qx_obj;
struct calib_Zx_obj;


/*
 * An instance of a polynomial in Q[x].  We represent these as the product
 * of a rational factor with a primitive Z[x] polynomial whose leading
 * coefficient (if any) is strictly positive.
 */

struct calib_Qx_obj {
	mpq_t	qfact;		/* Rational multiplier			 */
	int	degree;		/* Degree of polynomial			 */
	int	size;		/* Size of coeff[] array (degree < size) */
	mpz_ptr	coeff;		/* Coefficients of polynomial.		 */
};

/*
 * The following structure represents the factorization of a polynomial
 * over Q[x] into monic, irreducible factors.  It forms a linked list
 * of factors, including the multiplicity of each factor.
 *
 * The factor() function:
 * - Produces monic factors of degree 1 or greater.  If the given polynomial
 *   is not monic, then the first factor will be the necessary constant
 *   (a non-monic polynomial of degree 0 and multiplicity 1).
 * - Dynamically allocates the factors and (linked list thereof).  It is
 *   the responsibility of the caller to free these when done using the
 *   provided free_factors() function.
 */

struct calib_Qx_factor {
	int				multiplicity;
	struct calib_Qx_obj *		factor;
	struct calib_Qx_factor *	next;
};

/* The ring of polynomials over Q. */

struct calib_Qx_dom {
	void		(*init) (struct calib_Qx_obj *	x);
	void		(*init_degree)
				(struct calib_Qx_obj *	x,
				 int			degree);
	void		(*alloc) (struct calib_Qx_obj *	rop,
				  int			degree);
	void		(*clear) (struct calib_Qx_obj *	x);
	void		(*set) (struct calib_Qx_obj *		rop,
				const struct calib_Qx_obj *	op);
	void		(*set_si) (struct calib_Qx_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_Qx_obj *		rop,
				  mpz_srcptr			op);
	void		(*set_q) (struct calib_Qx_obj *		rop,
				  mpq_srcptr			op);
	void		(*set_var_power) (struct calib_Qx_obj *	rop,
					  int			power);
	void		(*set_Zx)
				(struct calib_Qx_obj *		rop,
				 const struct calib_Zx_obj *	op);
	void		(*add) (struct calib_Qx_obj *		rop,
				const struct calib_Qx_obj *	op1,
				const struct calib_Qx_obj *	op2);
	void		(*sub) (struct calib_Qx_obj *		rop,
				const struct calib_Qx_obj *	op1,
				const struct calib_Qx_obj *	op2);
	void		(*neg) (struct calib_Qx_obj *		rop,
				const struct calib_Qx_obj *	op);
	void		(*mul) (struct calib_Qx_obj *		rop,
				const struct calib_Qx_obj *	op1,
				const struct calib_Qx_obj *	op2);
	void		(*mul_si) (struct calib_Qx_obj *	rop,
				   const struct calib_Qx_obj *	op1,
				   calib_si_t			op2);
	void		(*mul_z) (struct calib_Qx_obj *		rop,
				  const struct calib_Qx_obj *	op1,
				  mpz_srcptr			op2);
	void		(*mul_q) (struct calib_Qx_obj *		rop,
				  const struct calib_Qx_obj *	op1,
				  mpq_srcptr			op2);
	void		(*ipow) (struct calib_Qx_obj *		rop,
				 const struct calib_Qx_obj *	op,
				 int				power);
	struct calib_Qx_obj *
			(*dup) (const struct calib_Qx_obj * op);
	void		(*free) (struct calib_Qx_obj * poly);
	void		(*eval) (mpq_ptr			rop,
				 const struct calib_Qx_obj *	poly,
				 mpq_srcptr			value);
	void		(*div) (struct calib_Qx_obj *		quotient,
				struct calib_Qx_obj *		remainder,
				const struct calib_Qx_obj *	a,
				const struct calib_Qx_obj *	b);
	void		(*gcd) (struct calib_Qx_obj *		gcd,
				const struct calib_Qx_obj *	a,
				const struct calib_Qx_obj *	b);
	void		(*extgcd) (struct calib_Qx_obj *	gcd,
				   struct calib_Qx_obj *	xa,
				   struct calib_Qx_obj *	xb,
				   const struct calib_Qx_obj *	a,
				   const struct calib_Qx_obj *	b);
	struct calib_Qx_factor *
			(*factor) (const struct calib_Qx_obj *	poly);
	void		(*free_factors) (struct calib_Qx_factor * factors);
	void		(*derivative) (struct calib_Qx_obj *		rop,
				       const struct calib_Qx_obj *	op);
	void		(*integral) (struct calib_Qx_obj *		rop,
				     const struct calib_Qx_obj *	op);
	calib_bool	(*zerop) (const struct calib_Qx_obj *	op);
	calib_bool	(*onep) (const struct calib_Qx_obj *	op);
	void		(*set_genrep) (struct calib_Qx_obj *		rop,
				       const struct calib_genrep *	op,
				       const char *			var);
	struct calib_genrep *
			(*to_genrep) (const struct calib_Qx_obj *	op,
				      const char *			var);
	struct calib_genrep *
			(*factors_to_genrep) (
				const struct calib_Qx_factor *	factors,
				const char *			var);
	mpq_ptr		(*get_coeffs) (const struct calib_Qx_obj *	op);
	void		(*set_coeffs) (struct calib_Qx_obj *	rop,
				       int			degree,
				       mpq_srcptr		coeffs);
};

/*
 * Return the ring of polynomials Q[x].
 */

const struct calib_Qx_dom *	calib_get_Qx_dom (void);

#ifdef __cplusplus
 }
#endif

#endif
