/**********************************************************************

	$Id: Za.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	Za.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The ring Z(a), where a is an algebraic extension defined
	to be the root of a given monic, irreducible polynomial
	g(a) \in Z[a].  It is a quotient ring obtained as Z[a] mod g(a).

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_ZA_H
#define CALIB_ZA_H

#include "calib/types.h"
#include <gmp.h>

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_genrep;
struct calib_Za_dom;
struct calib_Za_obj;

/*
 * An instance of a value in Z(a).
 */

struct calib_Za_obj {
	const struct calib_Za_dom *
		dom;		/* Z(a) domain containing this value	*/
	mpz_ptr	coeff;		/* Coefficients.  This is an array of	*/
				/* k integers.				*/
};

/*
 * The ring Z(a), where a is an algebraic extension defined to be
 * the root of a given monic, irreducible polynomial g(a) \in Z[a].
 * It is a quotient ring obtained as Z[a] mod g(a).
 */

struct calib_Za_dom {
	struct calib_Zx_obj *
			gpoly;	/* Generator polynomial */

	void		(*init) (const struct calib_Za_dom *	rp,
				 struct calib_Za_obj *		x);
	void		(*clear) (struct calib_Za_obj *		x);
	void		(*set) (struct calib_Za_obj *		rop,
				const struct calib_Za_obj *	op);
	void		(*set_si) (struct calib_Za_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_Za_obj *	rop,
				  mpz_srcptr		op);
	void		(*set_q) (struct calib_Za_obj *	rop,
				  mpq_srcptr		op);
	void		(*set_var_power)
				(struct calib_Za_obj *	rop,
				 int			power);
	void		(*add) (struct calib_Za_obj *		rop,
				const struct calib_Za_obj *	op1,
				const struct calib_Za_obj *	op2);
	void		(*sub) (struct calib_Za_obj *		rop,
				const struct calib_Za_obj *	op1,
				const struct calib_Za_obj *	op2);
	void		(*neg) (struct calib_Za_obj *		rop,
				const struct calib_Za_obj *	op);
	void		(*mul) (struct calib_Za_obj *		rop,
				const struct calib_Za_obj *	op1,
				const struct calib_Za_obj *	op2);
	void		(*mul_z) (struct calib_Za_obj *		rop,
				  const struct calib_Za_obj *	op1,
				  mpz_srcptr			op2);
	void		(*mul_a) (struct calib_Za_obj *		rop,
				  const struct calib_Za_obj *	op);
	void		(*ipow) (struct calib_Za_obj *		rop,
				 const struct calib_Za_obj *	op,
				 int				power);
	void		(*pinv) (struct calib_Za_obj *		rop,
				 mpz_ptr			d,
				 const struct calib_Za_obj *	op);
	void		(*div_z_exact)
				(struct calib_Za_obj *		rop,
				 const struct calib_Za_obj *	op1,
				 mpz_srcptr			op2);
	void		(*prim_part) (
				mpz_ptr				content,
				struct calib_Za_obj *		ppart,
				const struct calib_Za_obj *	op);
	void		(*cvZa) (struct calib_Za_obj *		rop,
				 const struct calib_Za_obj *	op);
	int		(*degree) (const struct calib_Za_obj *	op);
	calib_bool	(*zerop) (const struct calib_Za_obj *	op);
	calib_bool	(*onep) (const struct calib_Za_obj *	op);
	void		(*set_genrep) (struct calib_Za_obj *		rop,
				       const struct calib_genrep *	op,
				       const char *			var);
	struct calib_genrep *
			(*to_genrep) (const struct calib_Za_obj *	op,
				      const char *			var);
};

/*
 * Construct the ring Z(a) = Z[a] mod g(a), where g(a) is a given
 * monic, irreducible polynomial in Z[x].
 */

extern void	calib_free_Za_dom (struct calib_Za_dom * rp);
extern struct calib_Za_dom *
		calib_make_Za_dom (const struct calib_Zx_obj * gpoly);

#ifdef __cplusplus
 }
#endif

#endif
