/**********************************************************************

	$Id: Zax.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	Zax.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring of polynomials R[x] over the ring R = Z(a).

***********************************************************************

	Modification Log:

	a-1:	07/04/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_ZAX_H
#define CALIB_ZAX_H

#include "calib/types.h"
#include <gmp.h>

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_genrep;
struct calib_GFpkx_dom;
struct calib_GFpkx_obj;
struct calib_Qax_dom;
struct calib_Qax_obj;
struct calib_Za_dom;
struct calib_Za_obj;
struct calib_Zax_dom;
struct calib_Zax_obj;

/*
 * An instance of a polynomial in K[x], where K = Z(a).
 */

struct calib_Zax_obj {
	int			degree;	/* Degree of polynomial		*/
	int			size;	/* Size of coeff buffer		*/
	const struct calib_Zax_dom *
				dom;	/* Domain of polynomial		*/
	struct calib_Za_obj *	coeff;	/* Polynomial coefficients	*/
};

/*
 * A ring of polynomials over a ring Z(a).
 *
 * This is a skeletal implementation, designed only to support
 * the Q(a)[x] factorization algorithm.
 */

struct calib_Zax_dom {
	const struct calib_Za_dom *
			cr;	/* Coefficient ring */
	void		(*init) (const struct calib_Zax_dom *	R_of_x,
				 struct calib_Zax_obj *		x);
	void		(*init_degree)
				 (const struct calib_Zax_dom *	R_of_x,
				  struct calib_Zax_obj *	x,
				  int				degree);
	void		(*init_si) (
				const struct calib_Zax_dom *	R_of_x,
				struct calib_Zax_obj *		x,
				calib_si_t			op);
	void		(*alloc) (struct calib_Zax_obj *	rop,
				  int				degree);
	void		(*clear) (struct calib_Zax_obj *	x);
	void		(*set) (struct calib_Zax_obj *		rop,
				const struct calib_Zax_obj *	op);
	void		(*set_si) (struct calib_Zax_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_Zax_obj *	rop,
				  mpz_srcptr			op);
	void		(*set_q) (struct calib_Zax_obj *	rop,
				  mpq_srcptr			op);
	void		(*set_var_power) (struct calib_Zax_obj *	rop,
					  int				power);
	void		(*add) (struct calib_Zax_obj *		rop,
				const struct calib_Zax_obj *	op1,
				const struct calib_Zax_obj *	op2);
	void		(*sub) (struct calib_Zax_obj *		rop,
				const struct calib_Zax_obj *	op1,
				const struct calib_Zax_obj *	op2);
	void		(*neg) (struct calib_Zax_obj *		rop,
				const struct calib_Zax_obj *	op);
	void		(*mul) (struct calib_Zax_obj *		rop,
				const struct calib_Zax_obj *	op1,
				const struct calib_Zax_obj *	op2);
	void		(*mul_z) (
				struct calib_Zax_obj *		rop,
				const struct calib_Zax_obj *	op1,
				mpz_srcptr			op2);
	void		(*ipow) (struct calib_Zax_obj *		rop,
				 const struct calib_Zax_obj *	op,
				 int				power);
	struct calib_Zax_obj *
			(*dup) (const struct calib_Zax_obj *	op);
	void		(*free) (struct calib_Zax_obj * poly);
	void		(*eval) (struct calib_Za_obj *		rop,
				 const struct calib_Zax_obj *	op,
				 const struct calib_Za_obj *	value);
	void		(*div) (struct calib_Zax_obj *		quotient,
				struct calib_Zax_obj *		remainder,
				mpz_ptr				d,
				const struct calib_Zax_obj *	a,
				const struct calib_Zax_obj *	b);
	void		(*div_z_exact) (
				struct calib_Zax_obj *		rop,
				const struct calib_Zax_obj *	op1,
				mpz_srcptr			op2);
	void		(*extgcd) (struct calib_Zax_obj *	gcd,
				   struct calib_Zax_obj *	xa,
				   struct calib_Zax_obj *	xb,
				   const struct calib_Zax_obj *	a,
				   const struct calib_Zax_obj *	b);
	struct calib_Zax_obj *
			(*cvZax) (const struct calib_Zax_dom *	R_of_x,
				  const struct calib_Zax_obj *	poly);
	struct calib_Zax_obj *
			(*cvGFpkx) (const struct calib_Zax_dom *	R_of_x,
				    const struct calib_GFpkx_obj *	gfpoly);
	struct calib_Zax_obj *
			(*cvQax) (const struct calib_Zax_dom *	R_of_x,
				  const struct calib_Qax_obj *	qpoly,
				  mpz_srcptr			d);
	calib_bool	(*zerop) (const struct calib_Zax_obj *	op);
	calib_bool	(*onep) (const struct calib_Zax_obj *	op);
	void		(*set_genrep) (struct calib_Zax_obj *		rop,
				       const struct calib_genrep *	op,
				       const char *			xvar,
				       const char *			avar);
	struct calib_genrep *
			(*to_genrep) (const struct calib_Zax_obj *	op,
				      const char *			xvar,
				      const char *			avar);
};

/*
 * Construct the ring of polynomials over ring R, where R is
 * the given ring Z(a).
 */

extern void	calib_free_Zax_dom (struct calib_Zax_dom * R_of_x);
extern struct calib_Zax_dom *
		calib_make_Zax_dom (const struct calib_Za_dom * R);

#ifdef __cplusplus
 }
#endif

#endif
