/**********************************************************************

	$Id: Zpx.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	Zpx.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring Zp[x] of polynomials over the field Zp, the integers
	modulo a prime p.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_ZPX_H
#define CALIB_ZPX_H

#include "calib/Zp.h"

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_Qa_obj;
struct calib_Zpx_dom;
struct calib_Zpx_factor;
struct calib_Zpx_obj;
struct calib_Zx_obj;

/*
 * Which factorization algorithm to use (for square-free polynomials).
 */

enum calib_Zpx_factor_method {
	CALIB_ZPX_FACTOR_METHOD_BERLEKAMP,
	CALIB_ZPX_FACTOR_METHOD_DDF,
};

/*
 * The "settings" object for the Zpx domain.
 */

struct calib_Zpx_settings {
	/* Factorization method for square-free Zpx polynomials. */
	enum calib_Zpx_factor_method	factor_method;
};

/*
 * Newly created Zpx domains default to these settings.
 */

extern struct calib_Zpx_settings	calib_Zpx_default_settings;

/*
 * An instance of a polynomial in Zp[x].
 */

struct calib_Zpx_obj {
	int	degree;		/* Degree of polynomial			 */
	int	size;		/* Size of coeff[] array (degree < size) */
	const struct calib_Zpx_dom *
		dom;		/* Polynomial domain.			 */
	mpz_ptr	coeff;		/* Coefficients of polynomial.		 */
};

/*
 * The following structure represents the factorization of a polynomial
 * over Zp[x] into monic, irreducible factors.  It forms a linked list
 * of factors, including the multiplicity of each factor.
 *
 * The factor() function:
 * - Produces monic factors of degree 1 or greater.  If the given polynomial
 *   is not monic, then the first factor will be the necessary constant
 *   (a non-monic polynomial of degree 0 and multiplicity 1).
 * - Dynamically allocates the factors and (linked list thereof).  It is
 *   the responsibility of the caller to free these when done using the
 *   provided free_factors() function.
 */

struct calib_Zpx_factor {
	int				multiplicity;
	struct calib_Zpx_obj *		factor;
	struct calib_Zpx_factor *	next;
};

/* A ring of polynomials over a field Zp. */

struct calib_Zpx_dom {
	const struct calib_Zp_dom *
			cf;	/* Coefficient field */
	struct calib_Zpx_settings
			settings;	/* Algorithmic settings */
	void		(*init) (const struct calib_Zpx_dom *	K_of_x,
				 struct calib_Zpx_obj *		x);
	void		(*init_degree)
				(const struct calib_Zpx_dom *	K_of_x,
				 struct calib_Zpx_obj *		x,
				 int				degree);
	void		(*alloc) (struct calib_Zpx_obj *	rop,
				  int				degree);
	void		(*clear) (struct calib_Zpx_obj *	x);
	void		(*set) (struct calib_Zpx_obj *		rop,
				const struct calib_Zpx_obj *	op);
	void		(*set_si) (struct calib_Zpx_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_Zpx_obj *	rop,
				  mpz_srcptr			op);
	void		(*set_q) (struct calib_Zpx_obj *	rop,
				  mpq_srcptr			op);
	void		(*set_var_power)
				(struct calib_Zpx_obj *		rop,
				 int				power);
	void		(*set_Zx) (struct calib_Zpx_obj *	rop,
				   const struct calib_Zx_obj *	op);
	void		(*set_Qa) (struct calib_Zpx_obj *	rop,
				   const struct calib_Qa_obj *	op);
	void		(*add) (struct calib_Zpx_obj *		rop,
				const struct calib_Zpx_obj *	op1,
				const struct calib_Zpx_obj *	op2);
	void		(*sub) (struct calib_Zpx_obj *		rop,
				const struct calib_Zpx_obj *	op1,
				const struct calib_Zpx_obj *	op2);
	void		(*neg) (struct calib_Zpx_obj *		rop,
				const struct calib_Zpx_obj *	op);
	void		(*mul) (struct calib_Zpx_obj *		rop,
				const struct calib_Zpx_obj *	op1,
				const struct calib_Zpx_obj *	op2);
	void		(*mul_z) (struct calib_Zpx_obj *	rop,
				  const struct calib_Zpx_obj *	op1,
				  mpz_srcptr			op2);
	void		(*ipow) (struct calib_Zpx_obj *		rop,
				 const struct calib_Zpx_obj *	op,
				 int				power);
	struct calib_Zpx_obj *
			(*dup) (const struct calib_Zpx_obj *	op);
	void		(*free) (struct calib_Zpx_obj * poly);
	void		(*monicize) (struct calib_Zpx_obj *	poly);
	void		(*eval) (mpz_ptr			rop,
				 const struct calib_Zpx_obj *	poly,
				 mpz_srcptr			value);
	void		(*div) (struct calib_Zpx_obj *		quotient,
				struct calib_Zpx_obj *		remainder,
				const struct calib_Zpx_obj *	a,
				const struct calib_Zpx_obj *	b);
	void		(*gcd) (struct calib_Zpx_obj *		gcd,
				const struct calib_Zpx_obj *	a,
				const struct calib_Zpx_obj *	b);
	void		(*extgcd) (struct calib_Zpx_obj *	gcd,
				   struct calib_Zpx_obj *	xa,
				   struct calib_Zpx_obj *	xb,
				   const struct calib_Zpx_obj *	a,
				   const struct calib_Zpx_obj *	b);
	struct calib_Zpx_factor *
			(*factor) (const struct calib_Zpx_obj *	poly);
	struct calib_Zpx_factor *
			(*factor_square_free) (
				const struct calib_Zpx_obj *	poly);
	struct calib_Zpx_factor *
			(*finish_sqf) (
				const struct calib_Zpx_factor *	sqfactors);
	void		(*free_factors)
				(struct calib_Zpx_factor *	factors);
	void		(*resultant) (mpz_ptr				result,
				      const struct calib_Zpx_obj *	a,
				      const struct calib_Zpx_obj *	b);
	void		(*derivative)
				(struct calib_Zpx_obj *		rop,
				 const struct calib_Zpx_obj *	op);
	calib_bool	(*zerop) (const struct calib_Zpx_obj *	op);
	calib_bool	(*onep) (const struct calib_Zpx_obj *	op);
	void		(*set_genrep) (struct calib_Zpx_obj *		rop,
				       const struct calib_genrep *	op,
				       const char *			var);
	struct calib_genrep *
			(*to_genrep) (const struct calib_Zpx_obj *	op,
				      const char *			var);
	struct calib_genrep *
			(*factors_to_genrep)
				(const struct calib_Zpx_factor *	factors,
				 const char *				var);
	void		(*print_maxima) (const struct calib_Zpx_obj *	op);
};

/*
 * Construct the ring of polynomials Zp[x] for the given Zp.
 */

extern void	calib_free_Zpx_dom (struct calib_Zpx_dom *	K_of_x);
extern struct calib_Zpx_dom *
		calib_make_Zpx_dom (const struct calib_Zp_dom *	cf);

#ifdef __cplusplus
 }
#endif

#endif
