/**********************************************************************

	$Id: Zx.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	Zx.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring Z[x] of polynomials over the ring Z, the integers.

***********************************************************************

	Modification Log:

	a-1:	05/04/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_ZX_H
#define CALIB_ZX_H

#include "calib/types.h"
#include <gmp.h>

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_genrep;
struct calib_Zpx_dom;
struct calib_Zpx_obj;
struct calib_Zx_dom;
struct calib_Zx_factor;
struct calib_Zx_obj;
struct calib_Zx_settings;

/*
 * Which factorization algorithm to use (for square-free polynomials).
 */

enum calib_Zx_factor_method {
	CALIB_ZX_FACTOR_METHOD_ZASSENHAUS,
	CALIB_ZX_FACTOR_METHOD_VAN_HOEIJ,
};

/*
 * Which method to use for lifting modular factors.
 */

enum calib_Zx_lift_method {
	CALIB_ZX_LIFT_METHOD_LINEAR,
	CALIB_ZX_LIFT_METHOD_QUADRATIC,
};

/*
 * Which method to use for gcd (gcd, a, b).  (The gcd_n() function always
 * uses modular.)
 */

enum calib_Zx_gcd_method {
	CALIB_ZX_GCD_METHOD_MODULAR,
	CALIB_ZX_GCD_METHOD_PRS
};

/*
 * The "settings" object for the Zx domain.
 */

struct calib_Zx_settings {
	/* Factorization method for square-free Zx polynomials. */
	enum calib_Zx_factor_method	factor_method;

	/* Method for lifting modular factors. */
	enum calib_Zx_lift_method	lift_method;

	/* Method to use for two-operand gcd(). */
	enum calib_Zx_gcd_method	gcd_method;

	/* If number of remaining modular factors does not exceed this,	*/
	/* then use exhaustive enumeration instead of Van Hoeij.	*/
	int				VH_max_enumerate;

	/* Enable trace output from Z[x] factorization algorithm.	*/
	/* Zero is none.  Higher values give more output.		*/
	int				factor_trace_level;

	/* Van Hoeij uses resultant-based trace root bound? */
	calib_bool			VH_use_resultant_trace_root_bound;

	/* Validate lifted modular factors? */
	calib_bool			validate_lifted_modular_factors;
};

/*
 * All ssettings for Zx domain are global.
 */

extern struct calib_Zx_settings		calib_Zx_settings;

/*
 * An instance of a polynomial in Z[x].
 */

struct calib_Zx_obj {
	int	degree;		/* Degree of polynomial			 */
	int	size;		/* Size of coeff[] array (degree < size) */
	mpz_ptr	coeff;		/* Coefficients of polynomial.		 */
};

/*
 * The following structure represents the factorization of a polynomial
 * over Z[x] into irreducible factors.  It forms a linked list
 * of factors, including the multiplicity of each factor.
 *
 * The factor() function:
 * - Produces irreducible factors of degree 1 or greater.  The one exception
 *   is that the first factor can be a constant factor (of degree 0 and
 *   multiplicity 1).
 * - Dynamically allocates the factors and (linked list thereof).  It is
 *   the responsibility of the caller to free these when done using the
 *   provided free_factors() function.
 */

struct calib_Zx_factor {
	int				multiplicity;
	struct calib_Zx_obj *		factor;
	struct calib_Zx_factor *	next;
};

/* The ring of polynomials over Z. */

struct calib_Zx_dom {
	void		(*init) (struct calib_Zx_obj *		x);
	void		(*init_si) (struct calib_Zx_obj *	x,
				    calib_si_t			op);
	void		(*alloc) (struct calib_Zx_obj *		rop,
				  int				degree);
	void		(*clear) (struct calib_Zx_obj *		x);
	void		(*set) (struct calib_Zx_obj *		rop,
				const struct calib_Zx_obj *	op);
	void		(*set_si) (struct calib_Zx_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_Zx_obj *		rop,
				  mpz_srcptr			op);
	void		(*set_var_power) (struct calib_Zx_obj *	rop,
					  int			power);
	void		(*add) (struct calib_Zx_obj *		rop,
				const struct calib_Zx_obj *	op1,
				const struct calib_Zx_obj *	op2);
	void		(*sub) (struct calib_Zx_obj *		rop,
				const struct calib_Zx_obj *	op1,
				const struct calib_Zx_obj *	op2);
	void		(*neg) (struct calib_Zx_obj *		rop,
				const struct calib_Zx_obj *	op);
	void		(*mul) (struct calib_Zx_obj *		rop,
				const struct calib_Zx_obj *	op1,
				const struct calib_Zx_obj *	op2);
	void		(*mul_z) (struct calib_Zx_obj *		rop,
				  const struct calib_Zx_obj *	op1,
				  mpz_srcptr			op2);
	void		(*ipow) (struct calib_Zx_obj *		rop,
				 const struct calib_Zx_obj *	op,
				 int				power);
	struct calib_Zx_obj *
			(*dup) (const struct calib_Zx_obj * op);
	void		(*free) (struct calib_Zx_obj * poly);
	void		(*eval) (mpz_ptr			rop,
				 const struct calib_Zx_obj *	poly,
				 mpz_srcptr			value);
	void		(*div) (struct calib_Zx_obj *		quotient,
				struct calib_Zx_obj *		remainder,
				mpz_ptr				d,
				const struct calib_Zx_obj *	a,
				const struct calib_Zx_obj *	b);
	calib_bool	(*exactly_divides) (
				struct calib_Zx_obj *		quotient,
				const struct calib_Zx_obj *	a,
				const struct calib_Zx_obj *	b);
	void		(*div_z_exact) (struct calib_Zx_obj *		rop,
					const struct calib_Zx_obj *	op1,
					mpz_srcptr			op2);
	void		(*gcd) (struct calib_Zx_obj *			gcd,
				const struct calib_Zx_obj *		a,
				const struct calib_Zx_obj *		b);
	void		(*gcd_n) (struct calib_Zx_obj *			gcd,
				  struct calib_Zx_obj **		cofact,
				  int					npoly,
				  const struct calib_Zx_obj * const *	poly_ptrs);
	void		(*extgcd) (struct calib_Zx_obj *	gcd,
				   struct calib_Zx_obj *	xa,
				   struct calib_Zx_obj *	xb,
				   const struct calib_Zx_obj *	a,
				   const struct calib_Zx_obj *	b);
	void		(*prim_part) (mpz_ptr				content,
				      struct calib_Zx_obj *		ppart,
				      const struct calib_Zx_obj *	op);
	void		(*resultant) (mpz_ptr				result,
				      const struct calib_Zx_obj *	a,
				      const struct calib_Zx_obj *	b);
	void		(*discriminant) (mpz_ptr			result,
					 const struct calib_Zx_obj *	op);
	void		(*derivative)
				(struct calib_Zx_obj *		rop,
				 const struct calib_Zx_obj *	op);
	struct calib_Zx_obj *
			(*cvZpx) (const struct calib_Zpx_obj *	poly);
	struct calib_Zx_factor *
			(*factor) (const struct calib_Zx_obj *	poly);
	struct calib_Zx_factor *
			(*factor_square_free) (const struct calib_Zx_obj * poly);
	struct calib_Zx_factor *
			(*finish_sqf) (const struct calib_Zx_factor * sqfactors);
	void		(*free_factors) (struct calib_Zx_factor *	factors);
	calib_bool	(*zerop) (const struct calib_Zx_obj *	op);
	calib_bool	(*onep) (const struct calib_Zx_obj *	op);
	void		(*set_genrep) (struct calib_Zx_obj *		rop,
				       const struct calib_genrep *	op,
				       const char *			var);
	struct calib_genrep *
			(*to_genrep) (const struct calib_Zx_obj *	op,
				      const char *			var);
	struct calib_genrep *
			(*factors_to_genrep)
				(const struct calib_Zx_factor *	factors,
				 const char *			var);
	void		(*print_maxima) (const struct calib_Zx_obj *	op);
};

/*
 * Return the ring of polynomials Z[x].
 */

const struct calib_Zx_dom *	calib_get_Zx_dom (void);

#ifdef __cplusplus
 }
#endif

#endif
