/**********************************************************************

	$Id: Zxyz.h,v 1.1 2024/10/20 00:09:10 warme Exp $

	File:	Zxyz.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The ring Z[x,y,z] of multi-variate polynomials over the
	ring Z, of integers.

***********************************************************************

	Modification Log:

	a-1:	08/05/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_ZXYZ_H
#define CALIB_ZXYZ_H

#include "calib/types.h"
#include <gmp.h>

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_genrep;
struct calib_Qax_dom;
struct calib_Qax_obj;
struct calib_Zp_dom;
struct calib_Zpx_dom;
struct calib_Zpx_obj;
struct calib_Zx_obj;
struct calib_Zxyz_dom;
struct calib_Zxyz_factor;
struct calib_Zxyz_obj;


/*
 * An instance of a polynomial in Z[x,y,z].  (Any number of variables
 * is allowed, and they can have arbitrary names, not just x, y and z.)
 */

struct calib_Zxyz_obj {
	int			nterms;	/* Number of terms in polynomial */
	int			size;	/* Size of pow[] and coeff[] arrays */
	const struct calib_Zxyz_dom *
				dom;	/* Domain of polynomial */
	int *			pow;	/* Powers of each var, each term */
	mpz_ptr			coeff;	/* Coefficient for each term	*/
};

/*
 * The following structure represents the factorization of a polynomial
 * over Z[x,y,z] into irreducible factors.  It forms a linked list
 * of factors, including the multiplicity of each factor.
 *
 * The factor() function:
 * - Produces irreducible factors of degree 1 or greater.
 * - Dynamically allocates the factors and (linked list thereof).  It is
 *   the responsibility of the caller to free these when done using the
 *   provided free_factors() function.
 */

struct calib_Zxyz_factor {
	int				multiplicity;
	struct calib_Zxyz_obj *		factor;
	struct calib_Zxyz_factor *	next;
};

/* The ring of multi-variate polynomials over Z. */

struct calib_Zxyz_dom {
	int		nvars;		/* Number of variables */
	char **		vars;		/* Name of each variable */

	void		(*init) (const struct calib_Zxyz_dom *	dom,
				 struct calib_Zxyz_obj *	x);
	void		(*init_si) (const struct calib_Zxyz_dom *	dom,
				    struct calib_Zxyz_obj *		x,
				    calib_si_t				op);
	void		(*alloc) (struct calib_Zxyz_obj *	rop,
				  int				nterms);
	void		(*clear) (struct calib_Zxyz_obj *	x);
	void		(*set) (struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op);
	void		(*set_si) (struct calib_Zxyz_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_Zxyz_obj *	rop,
				  mpz_srcptr			op);
	void		(*set_var_power)
				 (struct calib_Zxyz_obj *	rop,
				  int				var,
				  int				power);
	void		(*add) (struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op1,
				const struct calib_Zxyz_obj *	op2);
	void		(*add_n) (struct calib_Zxyz_obj *		rop,
				  int					npoly,
				  const struct calib_Zxyz_obj **	poly_ptrs);
	void		(*sub) (struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op1,
				const struct calib_Zxyz_obj *	op2);
	void		(*neg) (struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op);
	void		(*mul) (struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op1,
				const struct calib_Zxyz_obj *	op2);
	void		(*mul_z) (struct calib_Zxyz_obj *	rop,
				  const struct calib_Zxyz_obj *	op1,
				  mpz_srcptr			op2);
	void		(*ipow) (struct calib_Zxyz_obj *	rop,
				 const struct calib_Zxyz_obj *	op,
				 int				power);
	struct calib_Zxyz_obj *
			(*dup) (const struct calib_Zxyz_obj * op);
	void		(*free) (struct calib_Zxyz_obj * poly);
	void		(*eval) (mpz_ptr			rop,
				 const struct calib_Zxyz_obj *	poly,
				 mpz_srcptr			values);
	void		(*eval_var_subset) (
				struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	poly,
				const calib_bool *		evflags,
				mpz_srcptr			values);
	void		(*div) (struct calib_Zxyz_obj *		quotient,
				struct calib_Zxyz_obj *		remainder,
				struct calib_Zxyz_obj *		d,
				const struct calib_Zxyz_obj *	a,
				const struct calib_Zxyz_obj *	b,
				int				var);
	void		(*div_z_exact) (struct calib_Zxyz_obj *		rop,
					const struct calib_Zxyz_obj *	op1,
					mpz_srcptr			op2);
	int		(*div_remove) (struct calib_Zxyz_obj *		rop,
				       const struct calib_Zxyz_obj *	op1,
				       const struct calib_Zxyz_obj *	op2);
	void		(*gcd) (struct calib_Zxyz_obj *		gcd,
				const struct calib_Zxyz_obj *	a,
				const struct calib_Zxyz_obj *	b);
	void		(*gcd_n) (struct calib_Zxyz_obj *		gcd,
				  struct calib_Zxyz_obj **		cofact,
				  int					npoly,
				  const struct calib_Zxyz_obj **	array);
	void		(*extgcd) (struct calib_Zxyz_obj *		gcd,
				   struct calib_Zxyz_obj *		xa,
				   struct calib_Zxyz_obj *		xb,
				   mpz_ptr				d,
				   const struct calib_Zxyz_obj *	a,
				   const struct calib_Zxyz_obj *	b);
	void		(*z_content) (mpz_ptr				zcont,
				      const struct calib_Zxyz_obj *	op);
	void		(*strip_z_content)
				    (mpz_ptr				zcont,
				     struct calib_Zxyz_obj *		poly);
	void		(*prim_part) (
				struct calib_Zxyz_obj *		content,
				struct calib_Zxyz_obj *		ppart,
				const struct calib_Zxyz_obj *	op,
				int				var);
	void		(*resultant) (struct calib_Zxyz_obj *		result,
				      const struct calib_Zxyz_obj *	a,
				      const struct calib_Zxyz_obj *	b,
				      int				var);
	struct calib_Zxyz_factor *
			(*resultant_old) (const struct calib_Zxyz_obj *	a,
					  const struct calib_Zxyz_obj *	b,
					  int				var);
	void		(*discriminant) (struct calib_Zxyz_obj *	rop,
					 const struct calib_Zxyz_obj *	op,
					 int				var);
	void		(*cvZpxyz) (struct calib_Zxyz_obj *		rop,
				    const struct calib_Zp_dom *		Zp,
				    const struct calib_Zxyz_obj *	op);
	struct calib_Zxyz_factor *
			(*factor) (const struct calib_Zxyz_obj * poly);
	struct calib_Zxyz_factor *
			(*sqf_factor) (const struct calib_Zxyz_obj * poly);
	void		(*free_factors) (struct calib_Zxyz_factor * factors);
	struct calib_Zxyz_obj *
			(*map_to_subring) (
				struct calib_Zxyz_obj *		result,
				const struct calib_Zxyz_obj *	poly,
				const struct calib_Zxyz_dom *	subring,
				const int *			varmap);
	void		(*copy_into_superring) (
				struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op);
	void		(*convert_with_varmap) (
				struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op,
				const int *			varmap);
	void		(*copy_from_Zx) (
				struct calib_Zxyz_obj *		rop,
				const struct calib_Zx_obj *	op,
				int				var);
	void		(*copy_into_Zx) (struct calib_Zx_obj *		rop,
					 const struct calib_Zxyz_obj *	op,
					 int				var);
	void		(*copy_from_Zpx) (struct calib_Zxyz_obj *	rop,
					 const struct calib_Zpx_obj *	op,
					 int				var);
	void		(*copy_into_Zpx) (struct calib_Zpx_obj *	rop,
					  const struct calib_Zxyz_obj *	op,
					  int				var);
	void		(*copy_from_Qax) (struct calib_Zxyz_obj * 	rop,
					  mpz_ptr		 	denom,
					  const struct calib_Qax_obj *	op,
					  int			 	xvar,
					  int			 	avar);
	void		(*copy_into_Qax) (struct calib_Qax_obj *  	rop,
					  const struct calib_Zxyz_obj *	op,
					  int			  	xvar,
					  int			  	avar);
	struct calib_Zxyz_dom *
			(*add_vars) (const struct calib_Zxyz_dom *	dom,
				     int				nvars,
				     const char * const *		newvars);
	calib_bool	(*zerop) (const struct calib_Zxyz_obj *	op);
	calib_bool	(*onep) (const struct calib_Zxyz_obj *	op);
	void		(*set_genrep) (struct calib_Zxyz_obj *		rop,
				       const struct calib_genrep *	op);
	struct calib_genrep *
			(*to_genrep) (const struct calib_Zxyz_obj *	op);
	struct calib_genrep *
			(*factors_to_genrep)
				(const struct calib_Zxyz_factor * factors);
	void		(*print_maxima) (const struct calib_Zxyz_obj * op);
	void		(*print_maxima_nnl) (const struct calib_Zxyz_obj * op);
	int		(*lookup_var) (const struct calib_Zxyz_dom *	dom,
				       const char *			var);
};

/*
 * Construct the ring of polynomials Z[x,y,z], for a given list of vars.
 */

extern void	calib_free_Zxyz_dom (const struct calib_Zxyz_dom * dom);
extern struct calib_Zxyz_dom *
		calib_make_Zxyz_dom (int			nvars,
				     const char * const *	vars);

#ifdef __cplusplus
 }
#endif

#endif
