/**********************************************************************

	$Id: genrep.h,v 1.1 2024/10/20 00:09:11 warme Exp $

	File:	genrep.h
	Rev:	a-3
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	General representation for expressions.

***********************************************************************

	Modification Log:

	a-1:	04/23/2012	warme
		: Created.
	a-2:	10/11/2019	warme
		: Add new pretty print functions.
	a-3:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_GENREP_H
#define CALIB_GENREP_H

#ifdef __cplusplus
 extern "C" {
#endif

#include "calib/logic.h"
#include "calib/types.h"
#include <gmp.h>
#include <stddef.h>
#include <stdio.h>

struct calib_genrep;
struct calib_genrep_list;
struct calib_genrep_varlist;

/*
 * The general representation for expressions.  Each operator is an
 * n-ary function.  Arithmetic operators include +, * and ^, but we
 * only allow "int" powers of things for now.
 *
 * This reduced set of operators requires a-b to be expressed as
 * a + (-1*b), and a/b as a * (b ^ (-1)).
 */

struct calib_genrep {
	char	op;	/* Operator. */
	union {
		struct calib_genrep_list *	list;	/* List of operand subexprs */
		char *				var;	/* Variable */
		calib_si_t			siop;	/* Signed integer */
		mpz_ptr				zop;	/* Integer operand */
		mpq_ptr				qop;	/* Rational operand */
		struct {
			struct calib_genrep *	base;
			int			power;
		} ipow;
		struct {
			struct calib_genrep *		func;
			struct calib_genrep_list *	args;
		} func;
	} u;
};

struct calib_genrep_list {
	struct calib_genrep *		operand;	/* Current operand */
	struct calib_genrep_list *	next;		/* List of other operands */
};

/*
 * Structure used to represent a list of variable names.
 */

struct calib_genrep_varlist {
	int			nvars;	/* Number of vars in the list */
	const char * const *	vlist;	/* List of variable names.  Note */
					/* that vlist[nvars] EQ NULL. */
};

/*
 * The suite of operators.
 */

#define CALIB_GENREP_OP_ADD	0
#define CALIB_GENREP_OP_MUL	1
#define CALIB_GENREP_OP_IPOW	2
#define	CALIB_GENREP_OP_VAR	3
#define	CALIB_GENREP_OP_SI	4
#define CALIB_GENREP_OP_Z	5
#define CALIB_GENREP_OP_Q	6
#define	CALIB_GENREP_OP_FUNC	7

/*
 * Useful macros.
 */

#define CALIB_GENREP_VARP(gp)	((gp) -> op EQ CALIB_GENREP_OP_VAR)
#define CALIB_GENREP_INTP(gp)	((gp) -> op EQ CALIB_GENREP_OP_Z)
#define CALIB_GENREP_RATP(gp)	((gp) -> op EQ CALIB_GENREP_OP_Q)
#define CALIB_GENREP_SIP(gp)	((gp) -> op EQ CALIB_GENREP_OP_SI)
#define CALIB_GENREP_NUMBERP(gp) \
				(((gp) -> op EQ CALIB_GENREP_OP_SI) OR \
				 ((gp) -> op EQ CALIB_GENREP_OP_Z) OR \
				 ((gp) -> op EQ CALIB_GENREP_OP_Q))
#define CALIB_GENREP_ZEROP(gp)	((CALIB_GENREP_SIP (gp) AND \
				  ((gp) -> u.siop EQ 0)) OR \
				 (CALIB_GENREP_INTP (gp) AND \
				  (mpz_sgn ((gp) -> u.zop) EQ 0)) OR \
				 (CALIB_GENREP_RATP (gp) AND \
				  (mpz_sgn (mpq_numref ((gp) -> u.qop)) EQ 0)))
#define CALIB_GENREP_ONEP(gp)	((CALIB_GENREP_SIP (gp) AND \
				  ((gp) -> u.siop EQ 1)) OR \
				 (CALIB_GENREP_INTP (gp) AND \
				  (mpz_cmp_ui ((gp) -> u.zop, 1) EQ 0)) OR \
				 (CALIB_GENREP_RATP (gp) AND \
				  (mpq_cmp_ui ((gp) -> u.qop, 1, 1) EQ 0)))

/*
 * Function signatures.
 */

extern struct calib_genrep *
		calib_genrep_add2 (const struct calib_genrep *	op1,
				   const struct calib_genrep *	op2);
extern const char *
		calib_genrep_builtin_func_index_to_name (calib_si_t	fidx);
extern calib_si_t
		calib_genrep_builtin_func_name_to_index (const char * fname);
extern void	calib_genrep_clear_varlist (
				struct calib_genrep_varlist *	list);
extern size_t	calib_genrep_convert_abs_Z_to_decimal_string (
				mpz_srcptr	zval,
				char **		str_out);
extern struct calib_genrep *
		calib_genrep_div2 (const struct calib_genrep *	op1,
				   const struct calib_genrep *	op2);
extern struct calib_genrep *
		calib_genrep_dup (const struct calib_genrep * op);
extern struct calib_genrep_list *
		 calib_genrep_dup_list (const struct calib_genrep_list * list);
extern struct calib_genrep *
		calib_genrep_fread (FILE * fp, int * status);
extern void	calib_genrep_free (struct calib_genrep *	op);
extern void	calib_genrep_free_list (struct calib_genrep_list * list);
extern void	calib_genrep_fprint (FILE *				fp,
				     const struct calib_genrep *	op);
extern void	calib_genrep_fwprint (FILE *				fp,
				      const struct calib_genrep *	op,
				      int				width);
extern struct calib_genrep *
		calib_genrep_func (struct calib_genrep *	f,
				   struct calib_genrep_list *	args);
extern struct calib_genrep *
		calib_genrep_func_si_1 (calib_si_t		fidx,
					struct calib_genrep *	op);
extern struct calib_genrep *
		calib_genrep_func_si_2 (calib_si_t		fidx,
					struct calib_genrep *	arg1,
					struct calib_genrep *	arg2);
extern struct calib_genrep *
		calib_genrep_func_str (const char *			fname,
				       struct calib_genrep_list *	args);
extern void	calib_genrep_get_varlist (
				struct calib_genrep_varlist *	list,
				const struct calib_genrep *	op);
extern struct calib_genrep *
		calib_genrep_ipow (const struct calib_genrep * op1, int op2);
extern struct calib_genrep *
		calib_genrep_mul2 (const struct calib_genrep *	op1,
				   const struct calib_genrep *	op2);
extern struct calib_genrep *
		calib_genrep_neg (struct calib_genrep * op);
extern struct calib_genrep *
		calib_genrep_new ();
extern struct calib_genrep_list *
		 calib_genrep_new_list (struct calib_genrep *	  	op,
					struct calib_genrep_list *	next);
extern struct calib_genrep *
		calib_genrep_poly_term (struct calib_genrep *	coeff,
					const char *		var,
					int			pow);
extern void	calib_genrep_prettyprint (const struct calib_genrep * op);
extern void	calib_genrep_prettyprint_file (
					FILE *				fp,
					const struct calib_genrep *	op);
extern void	calib_genrep_prettyprint_file_width (
					FILE *				fp,
					const struct calib_genrep *	op,
					int				width);
extern void	calib_genrep_print_maxima (
					FILE *				fp,
					const struct calib_genrep *	op,
					int				width);
extern void	calib_genrep_prettyprint_width (
					const struct calib_genrep *	op,
					int				width);
extern void	calib_genrep_print (const struct calib_genrep * op);
extern struct calib_genrep *
		calib_genrep_q (mpq_srcptr	op);
extern struct calib_genrep *
		calib_genrep_read (int * status);
extern struct calib_genrep *
		calib_genrep_si (calib_si_t op);
extern struct calib_genrep *
		calib_genrep_sub2 (struct calib_genrep *	op1,
				   struct calib_genrep *	op2);
extern struct calib_genrep *
		calib_genrep_var (const char * var);
extern void	calib_genrep_wprint (const struct calib_genrep *	op,
				     int				width);
extern struct calib_genrep *
		calib_genrep_z (mpz_srcptr	op);

/*
 * Builtin function numbers.
 */

#define	CALIB_GENREP_BUILTIN_FUNC1_LIST(proc) \
	proc (exp,	EXP) \
	proc (log,	LOG) \
	proc (abs,	ABS) \
	proc (sqrt,	SQRT) \
	proc (sin,	SIN) \
	proc (cos,	COS) \
	proc (tan,	TAN) \
	proc (cot,	COT) \
	proc (sec,	SEC) \
	proc (csc,	CSC) \
	proc (asin,	ASIN) \
	proc (acos,	ACOS) \
	proc (atan,	ATAN) \
	proc (acot,	ACOT) \
	proc (asec,	ASEC) \
	proc (acsc,	ACSC) \
	proc (sinh,	SINH) \
	proc (cosh,	COSH) \
	proc (tanh,	TANH) \
	proc (coth,	COTH) \
	proc (sech,	SECH) \
	proc (csch,	CSCH) \
	proc (asinh,	ASINH) \
	proc (acosh,	ACOSH) \
	proc (atanh,	ATANH) \
	proc (acoth,	ACOTH) \
	proc (asech,	ASECH) \
	proc (acsch,	ACSCh) \
	/* end of list */

#define	CALIB_GENREP_BUILTIN_FUNC2_LIST(proc) \
	proc (atan2,		ATAN2) \
	proc (integrate,	INTEGRATE) \
	/* end of list */

/*
 * Define builtin function numbers
 */

#define	_CALIB_BUILTIN_DEFINE_ENUM_MEMBER(func, FUNC) \
	CALIB_GENREP_FUNC_ ## FUNC,

enum calib_genrep_builtin_funcs {
	CALIB_GENREP_BUILTIN_FUNC1_LIST (_CALIB_BUILTIN_DEFINE_ENUM_MEMBER)
	CALIB_GENREP_BUILTIN_FUNC2_LIST (_CALIB_BUILTIN_DEFINE_ENUM_MEMBER)
	CALIB_GENREP_NUM_BUILTIN_FUNCTIONS
};

/*
 * Functions to generate standard / builtin functions.
 */

#define	_CALIB_DECLARE_FUNC1(func, FUNC) \
extern struct calib_genrep * calib_genrep_ ## func (struct calib_genrep * x);

CALIB_GENREP_BUILTIN_FUNC1_LIST (_CALIB_DECLARE_FUNC1)

#define	_CALIB_DECLARE_FUNC2(func, FUNC) \
extern struct calib_genrep * calib_genrep_ ## func (struct calib_genrep * x1, \
						    struct calib_genrep * x2);

CALIB_GENREP_BUILTIN_FUNC2_LIST (_CALIB_DECLARE_FUNC2)

#ifdef __cplusplus
 }
#endif

#endif
