/**********************************************************************

	$Id: rat.h,v 1.1 2024/10/20 00:09:11 warme Exp $

	File:	rat.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Rational functions: the quotient field P/P,
	where P=Z[x,y,z], the ring of multi-variate polynomials
	over the integers.

***********************************************************************

	Modification Log:

	a-1:	10/01/2015	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_RAT_H
#define CALIB_RAT_H

#include "calib/types.h"
#include "calib/Zxyz.h"
#include <gmp.h>

#ifdef __cplusplus
 extern "C" {
#endif

struct calib_rat_dom;
struct calib_rat_obj;
struct calib_rsubst;
struct calib_subst_list;

/*
 * A rational function: polynomial / polynomial, where both the numerator
 * and the denominator are members of the same Zxyz ring.
 */

struct calib_rat_obj {
	const struct calib_rat_dom *
				dom;	/* Rational function domain */
	struct calib_Zxyz_obj	numer;	/* Numerator polynomial */
	struct calib_Zxyz_obj	denom;	/* Denominator polynomial */
};

/*
 * An object to represent a set of variable substitutions,
 * each variable being replaced with a given rational expression.
 */

struct calib_subst_list {
	int			num_subst;	/* Number of substitutions */
	struct calib_rsubst *	subst;		/* Array of substitutions */
};

/*
 * An object to represent a rational expression to substitute
 * for a given variable.
 * In order to accomodate variables that are not part of the polynomial
 * domain of `val', the `varname' field can be the name of an arbitrary
 * variable.  Convention:
 *
 *	((var >=  0) AND (varname EQ NULL)) OR
 *	((var EQ -1) AND (varname NE NULL))
 */

struct calib_rsubst {
	int			var;	/* Variable number to be substituted */
					/* (or -1) */
	const char *		varname;
					/* Variable name (or NULL) */
	struct calib_rat_obj	val;	/* Rational value to replace */
					/* var with */
};

/*
 * The field of rational functions R / R, where R = Z[x,y,z]
 * (general multi-variate polynomials with integer coefficients).
 */

struct calib_rat_dom {
	const struct calib_Zxyz_dom *
			ring;	/* Ring/domain for numer, denom */
	void		(*init) (const struct calib_rat_dom *	dom,
				 struct calib_rat_obj *		x);
	void		(*clear) (struct calib_rat_obj *	x);
	void		(*set) (struct calib_rat_obj *		rop,
				const struct calib_rat_obj *	op);
	void		(*set_si) (struct calib_rat_obj *	rop,
				   calib_si_t			op);
	void		(*set_z) (struct calib_rat_obj *	rop,
				  mpz_srcptr			op);
	void		(*set_q) (struct calib_rat_obj *	rop,
				  mpq_srcptr			op);
	void		(*set_var_power)
				 (struct calib_rat_obj *	rop,
				  int				var,
				  int				power);
	void		(*add) (struct calib_rat_obj *		rop,
				const struct calib_rat_obj *	op1,
				const struct calib_rat_obj *	op2);
	void		(*sub) (struct calib_rat_obj *		rop,
				const struct calib_rat_obj *	op1,
				const struct calib_rat_obj *	op2);
	void		(*neg) (struct calib_rat_obj *		rop,
				const struct calib_rat_obj *	op);
	void		(*mul) (struct calib_rat_obj *		rop,
				const struct calib_rat_obj *	op1,
				const struct calib_rat_obj *	op2);
	void		(*mul_z) (struct calib_rat_obj *	rop,
				  const struct calib_rat_obj *	op1,
				  mpz_srcptr			op2);
	void		(*mul_q) (struct calib_rat_obj *	rop,
				  const struct calib_rat_obj *	op1,
				  mpq_srcptr			op2);
	void		(*ipow) (struct calib_rat_obj *	rop,
				 const struct calib_rat_obj *	op,
				 int				power);
	void		(*inv) (struct calib_rat_obj *		rop,
				const struct calib_rat_obj *	op);
	void		(*div) (struct calib_rat_obj *		rop,
				const struct calib_rat_obj *	op1,
				const struct calib_rat_obj *	op2);
	void		(*canonicalize) (struct calib_rat_obj *	op);
	struct calib_Zxyz_factor *
			(*factor) (const struct calib_rat_obj *	op);
	void		(*substitute_with_varmap)
				(struct calib_rat_obj *		rop,
				 const struct calib_rat_obj *	op,
				 const int *			varmap,
				 struct calib_subst_list *	slist);
	void		(*free_slist) (struct calib_subst_list * slist);
	calib_bool	(*zerop) (const struct calib_rat_obj *	op);
	calib_bool	(*onep) (const struct calib_rat_obj *	op);
	void		(*set_genrep) (struct calib_rat_obj *		rop,
				       const struct calib_genrep *	op);
	struct calib_genrep *
			(*to_genrep) (const struct calib_rat_obj *	op);
	void		(*print) (const struct calib_rat_obj *	op);
	void		(*print_nnl) (const struct calib_rat_obj * op);
};

extern struct calib_rat_dom *
		calib_make_rat_dom (const struct calib_Zxyz_dom * ring);
extern void	calib_free_rat_dom (const struct calib_rat_dom * dom);

#ifdef __cplusplus
 }
#endif

#endif
