/**********************************************************************

	$Id: GFpk.c,v 1.1 2024/10/20 00:36:25 warme Exp $

	File:	GFpk.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Galois field GF(p^k), where p is a prime number and k > 1
	is an integer.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/GFpk.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/random.h"
#include "calib/Za.h"
#include "calib/Zpx.h"
#include "calib/Zx.h"
#include <gmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Local Functions
 */

static void	GFpk_add (struct calib_GFpk_obj *	rop,
			  const struct calib_GFpk_obj *	op1,
			  const struct calib_GFpk_obj *	op2);
static void	GFpk_clear (struct calib_GFpk_obj *	x);
static void	GFpk_cvZa (struct calib_GFpk_obj *		rop,
			   const struct calib_Za_obj *		op);
static int	GFpk_degree (const struct calib_GFpk_obj *	op);
static void	GFpk_init (const struct calib_GFpk_dom *	f,
			   struct calib_GFpk_obj *		x);
static void	GFpk_inv (struct calib_GFpk_obj *	rop,
			  const struct calib_GFpk_obj *	op);
static void	GFpk_ipow (struct calib_GFpk_obj *		rop,
			   const struct calib_GFpk_obj *	op,
			   int					power);
static void	GFpk_mul (struct calib_GFpk_obj *	rop,
			  const struct calib_GFpk_obj *	op1,
			  const struct calib_GFpk_obj *	op2);
static void	GFpk_mul_a (struct calib_GFpk_obj *		rop,
			    const struct calib_GFpk_obj *	op);
static void	GFpk_mul_z (struct calib_GFpk_obj *		rop,
			    const struct calib_GFpk_obj *	op1,
			    mpz_srcptr				op2);
static void	GFpk_neg (struct calib_GFpk_obj *	rop,
			  const struct calib_GFpk_obj *	op);
static calib_bool
		GFpk_onep (const struct calib_GFpk_obj *	op);
static void	GFpk_pth_root (struct calib_GFpk_obj *		rop,
			       const struct calib_GFpk_obj *	op);
static void	GFpk_set (struct calib_GFpk_obj *	rop,
			  const struct calib_GFpk_obj *	op);
static void	GFpk_set_genrep (struct calib_GFpk_obj *	rop,
				 const struct calib_genrep *	op,
				 const char *			var);
static void	GFpk_set_q (struct calib_GFpk_obj *	rop,
			    mpq_srcptr			op);
static void	GFpk_set_random (struct calib_GFpk_obj *	rop,
				 struct calib_Random *		randp);
static void	GFpk_set_si (struct calib_GFpk_obj *	rop,
			     calib_si_t			op);
static void	GFpk_set_var_power (struct calib_GFpk_obj *	rop,
				    int				power);
static void	GFpk_set_z (struct calib_GFpk_obj *	rop,
			    mpz_srcptr			op);
static void	GFpk_sub (struct calib_GFpk_obj *	rop,
			  const struct calib_GFpk_obj *	op1,
			  const struct calib_GFpk_obj *	op2);
static struct calib_genrep *
		GFpk_to_genrep (const struct calib_GFpk_obj *	op,
				const char *			var);
static calib_bool
		GFpk_zerop (const struct calib_GFpk_obj *	op);

/*
 * Initialize the given x to hold a value from the given GF(p^k) field.
 */

	static
	void
GFpk_init (

const struct calib_GFpk_dom *	f,	/* IN: description of field */
struct calib_GFpk_obj *		x	/* IN/OUT: value to initialize */
)
{
	CALIB_FATAL_ERROR_IF ((f EQ NULL) OR (x EQ NULL));

	x -> dom	= f;
	x -> coeff	= calib_new_Z_vector (f -> k);
}

/*
 * Clear the given GF(p^k) value object.
 */

	static
	void
GFpk_clear (

struct calib_GFpk_obj *		x	/* IN/OUT: value to clear */
)
{
const struct calib_GFpk_dom *	f;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	f = x -> dom;

	calib_free_Z_vector (x -> coeff, f -> k);

	x -> dom	= NULL;
	x -> coeff	= NULL;
}

/*
 * Set the destination GF(p^k) value from the source GF(p^k) value.
 */

	static
	void
GFpk_set (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
const struct calib_GFpk_obj *	op		/* IN: GFpk value to set */
)
{
const struct calib_GFpk_dom *	f;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {

		f = rop -> dom;
		CALIB_FATAL_ERROR_IF (f NE op -> dom);

		calib_copy_Z_vector (rop -> coeff, op -> coeff, f -> k);
	}
}

/*
 * Convert the given integer into the corresponding member of GF(p^k).
 */

	static
	void
GFpk_set_si (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
calib_si_t			op		/* IN: integer to convert */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_ptr				coeff;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	f	= rop -> dom;
	coeff	= rop -> coeff;

	k  = f -> k;
	cf = f -> cf;

	cf -> set_si (cf, &coeff [0], op);

	for (i = 1; i < k; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
}

/*
 * Convert the given GMP integer into the corresponding member of GF(p^k).
 */

	static
	void
GFpk_set_z (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
mpz_srcptr			op		/* IN: integer to convert */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_ptr				coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f	= rop -> dom;
	coeff	= rop -> coeff;

	k  = f -> k;
	cf = f -> cf;

	cf -> set_z (cf, &coeff [0], op);

	for (i = 1; i < k; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
}

/*
 * Convert the given GMP rational into the corresponding member of GF(p^k).
 */

	static
	void
GFpk_set_q (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
mpq_srcptr			op		/* IN: rational to convert */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_ptr				coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f	= rop -> dom;
	coeff	= rop -> coeff;

	k  = f -> k;
	cf = f -> cf;

	cf -> set_q (cf, &coeff [0], op);

	for (i = 1; i < k; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
}

/*
 * Set the GF(p^k) value to be a**power (where "a" is the GFpk domain's
 * polynomial variable).
 */

	static
	void
GFpk_set_var_power (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
int				power		/* IN: power to set */
)
{
int				k;
const struct calib_GFpk_dom *	f;		/* IN: description of field */
struct calib_GFpk_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (power < 0));

	f	= rop -> dom;

	k  = f -> k;

	if (power < k) {
		/* Simple case: one term. */
		GFpk_set_si (rop, 0);
		mpz_set_ui (&(rop -> coeff [power]), 1);
	}
	else {
		/* Higher power must be reduced modulo the generator poly. */
		GFpk_init (f, &TEMP);
		GFpk_set_var_power (&TEMP, 1);
		GFpk_ipow (rop, &TEMP, power);
		GFpk_clear (&TEMP);
	}
}

/*
 * Addition in GF(p^k).
 */

	static
	void
GFpk_add (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
const struct calib_GFpk_obj *	op1,		/* IN: first operand */
const struct calib_GFpk_obj *	op2		/* IN: second operand */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_srcptr			coeff1, coeff2;
mpz_ptr				res;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE f) OR (op2 -> dom NE f));

	cf	= f -> cf;
	k	= f -> k;

	coeff1	= op1 -> coeff;
	coeff2	= op2 -> coeff;

	res = calib_new_Z_vector (k);

	for (i = 0; i < k; i++) {
		cf -> add (cf, &res [i], &coeff1 [i], &coeff2 [i]);
	}

	calib_copy_Z_vector (rop -> coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Subtraction in GF(p^k).
 */

	static
	void
GFpk_sub (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
const struct calib_GFpk_obj *	op1,		/* IN: first operand */
const struct calib_GFpk_obj *	op2		/* IN: second operand */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_srcptr			coeff1, coeff2;
mpz_ptr				res;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE f) OR (op2 -> dom NE f));

	cf	= f -> cf;
	k	= f -> k;

	coeff1	= op1 -> coeff;
	coeff2	= op2 -> coeff;

	res = calib_new_Z_vector (k);

	for (i = 0; i < k; i++) {
		cf -> sub (cf, &res [i], &coeff1 [i], &coeff2 [i]);
	}

	calib_copy_Z_vector (rop -> coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Negation in GF(p^k).
 */

	static
	void
GFpk_neg (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
const struct calib_GFpk_obj *	op		/* IN: operand */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_srcptr			coeff;
mpz_ptr				res;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE f);

	cf	= f -> cf;
	k	= f -> k;

	if (rop EQ op) {
		/* Negation in place. */
		res = rop -> coeff;
		for (i = 0; i < k; i++) {
			cf -> neg (cf, &res [i], &res [i]);
		}
	}
	else {
		res = calib_new_Z_vector (k);

		coeff = op -> coeff;
		for (i = 0; i < k; i++) {
			cf -> neg (cf, &res [i], &coeff [i]);
		}

		calib_copy_Z_vector (rop -> coeff, res, k);
		calib_free_Z_vector (res, k);
	}
}

/*
 * Multiplication in GF(p^k).
 */

	static
	void
GFpk_mul (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
const struct calib_GFpk_obj *	op1,		/* IN: first operand */
const struct calib_GFpk_obj *	op2		/* IN: second operand */
)
{
int				i, j, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_srcptr			gpoly;
mpz_srcptr			coeff1, coeff2;
mpz_srcptr			bj, lc;
mpz_ptr				res;
mpz_t				FACTOR, TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE f) OR (op2 -> dom NE f));

	cf	= f -> cf;
	k	= f -> k;
	gpoly	= &(f -> coeff [0]);

	coeff1	= op1 -> coeff;
	coeff2	= op2 -> coeff;

	res = calib_new_Z_vector (k);

	bj = &coeff2 [k - 1];
	if (mpz_sgn (bj) EQ 0) {
		/* res is currently already completely zero. */
	}
	else if (mpz_cmp_ui (bj, 1) EQ 0) {
		calib_copy_Z_vector (res, coeff1, k);
	}
	else {
		for (i = 0; i < k; i++) {
			cf -> mul (cf, &res [i], &coeff1 [i], bj);
		}
	}

	mpz_init (FACTOR);
	mpz_init (TEMP);

	for (j = k - 2; j >= 0; j--) {
		/* Multiply by a^1. */
		lc = &res [k - 1];
		if (mpz_sgn (lc) EQ 0) {
			/* No correction necessary, just shift. */
			for (i = k - 1; i > 0; i--) {
				mpz_set (&res [i], &res [i - 1]);
			}
			mpz_set_ui (&res [0], 0);
		}
		else if (mpz_cmp_ui (lc, 1) EQ 0) {
			/* Subtract generator polynomial while shifting. */
			for (i = k - 1; i > 0; i--) {
				cf -> sub (cf, &res [i], &res [i - 1], &gpoly [i]);
			}
			cf -> neg (cf, &res [0], &gpoly [0]);
		}
		else {
			/* Subtract lc*gpoly while shifting. */
			cf -> neg (cf, FACTOR, lc);
			for (i = k - 1; i > 0; i--) {
				cf -> mul (cf, TEMP, FACTOR, &gpoly [i]);
				cf -> add (cf, &res [i], &res [i -1], TEMP);
			}
			cf -> mul (cf, &res [0], FACTOR, &gpoly [0]);
		}

		/* Get next bj and do res += bj*coeff1. */
		bj = &coeff2 [j];
		if (mpz_sgn (bj) EQ 0) {
			/* Nothing to add. */
		}
		else if (mpz_cmp_ui (bj, 1) EQ 0) {
			for (i = 0; i < k; i++) {
				cf -> add (cf, &res [i], &res [i], &coeff1 [i]);
			}
		}
		else {
			for (i = 0; i < k; i++) {
				cf -> mul (cf, TEMP, bj, &coeff1 [i]);
				cf -> add (cf, &res [i], &res [i], TEMP);
			}
		}
	}

	mpz_clear (TEMP);
	mpz_clear (FACTOR);

	calib_copy_Z_vector (rop -> coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Multiply the given element of GF(p^k) by the given integer z.
 */

	static
	void
GFpk_mul_z (

struct calib_GFpk_obj *		rop,	/* IN/OUT: result */
const struct calib_GFpk_obj *	op1,	/* IN: first operand (in GF(p^k)) */
mpz_srcptr			op2	/* IN: second operand (in Z) */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_srcptr			coeff;
mpz_ptr				res;
mpz_t				ZMODP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE f);

	cf	= f -> cf;
	k	= f -> k;

	res = calib_new_Z_vector (k);

	mpz_init (ZMODP);
	cf -> set_z (cf, ZMODP, op2);

	if (mpz_sgn (ZMODP) NE 0) {
		coeff = op1 -> coeff;
		for (i = 0; i < k; i++) {
			cf -> mul (cf, &res [i], &coeff [i], ZMODP);
		}
	}

	mpz_clear (ZMODP);

	calib_copy_Z_vector (rop -> coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Multiply the given element of GF(p^k) by 'a' (the generator polynomial
 * variable).
 */

	static
	void
GFpk_mul_a (

struct calib_GFpk_obj *		rop,	/* IN/OUT: result */
const struct calib_GFpk_obj *	op	/* IN: element to multiply by 'x' */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_srcptr			gpoly;
mpz_srcptr			coeff;
mpz_srcptr			lc;
mpz_ptr				res;
mpz_t				FACTOR, TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f = rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE f);

	cf	= f -> cf;
	k	= f -> k;
	gpoly	= &(f -> coeff [0]);

	res = calib_new_Z_vector (k);

	coeff	= op -> coeff;

	lc = &coeff [k - 1];
	if (mpz_sgn (lc) EQ 0) {
		/* No correction necessary, just shift. */
		for (i = k - 1; i > 0; i--) {
			mpz_set (&res [i], &coeff [i - 1]);
		}
		/* res [0] is already zero. */
	}
	else if (mpz_cmp_ui (lc, 1) EQ 1) {
		/* Subtract generator polynomial while shifting. */
		for (i = k - 1; i > 0; i--) {
			cf -> sub (cf, &res [i], &coeff [i - 1], &gpoly [i]);
		}
		cf -> neg (cf, &res [0], &gpoly [0]);
	}
	else {
		/* Subtract lc*gpoly while shifting. */
		mpz_init (FACTOR);
		mpz_init (TEMP);
		cf -> neg (cf, FACTOR, lc);
		for (i = k - 1; i > 0; i--) {
			cf -> mul (cf, TEMP, FACTOR, &gpoly [i]);
			cf -> add (cf, &res [i], &coeff [i - 1], TEMP);
		}
		cf -> mul (cf, &res [0], FACTOR, &gpoly [0]);
		mpz_clear (TEMP);
		mpz_clear (FACTOR);
	}
	calib_copy_Z_vector (rop -> coeff, res, k);

	calib_free_Z_vector (res, k);
}

/*
 * Take the given GF(p^k) value to the given integer power.
 */

	static
	void
GFpk_ipow (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
const struct calib_GFpk_obj *	op,		/* IN: value to exponentiate */
int				power		/* IN: integer power */
)
{
const struct calib_GFpk_dom *	f;
struct calib_GFpk_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f = rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE f);

	GFpk_init (f, &RES);
	GFpk_init (f, &PREG);

	GFpk_set_si (&RES, 1);
	if (power >= 0) {
		GFpk_set (&PREG, op);
	}
	else {
		GFpk_inv (&PREG, op);
		power = - power;
	}

	while (power > 0) {
		if ((power & 1) NE 0) {
			GFpk_mul (&RES, &RES, &PREG);
		}
		GFpk_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}

	GFpk_set (rop, &RES);

	GFpk_clear (&PREG);
	GFpk_clear (&RES);
}

/*
 * Multiplicative inverse in GF(p^k).
 */

	static
	void
GFpk_inv (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
const struct calib_GFpk_obj *	op		/* IN: operand to invert */
)
{
int				i, j, k;
int				adeg, bdeg, rdeg;
mpz_ptr				a, b, q, r, m1, m2, m3, n1, n2, n3, tmp;
mpz_ptr				zp;
mpz_ptr				alc, blc, rlc;
mpz_ptr				coeff;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_srcptr			gpoly;
mpz_t				FACTOR, TEMP;
struct calib_GFpk_obj		Q, M1, M2, M3, N1, N2, N3, TMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f = rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE f);

	cf	= f -> cf;
	k	= f -> k;
	gpoly	= &(f -> coeff [0]);

	/* Allocate local buffers. */
	zp	= calib_new_Z_vector (11 * k + 2);
	a = zp;		zp += (k + 1);
	b = zp;		zp += k;
	q = zp;		zp += (k + 1);
	r = zp;		zp += k;
	m1 = zp;	zp += k;
	m2 = zp;	zp += k;
	m3 = zp;	zp += k;
	n1 = zp;	zp += k;
	n2 = zp;	zp += k;
	n3 = zp;	zp += k;
	tmp = zp;	zp += k;

	Q.dom		= f;
	M1.dom		= f;
	M2.dom		= f;
	M3.dom		= f;
	N1.dom		= f;
	N2.dom		= f;
	N3.dom		= f;
	TMP.dom		= f;
	Q.coeff		= q;
	M1.coeff	= m1;
	M2.coeff	= m2;
	M3.coeff	= m3;
	N1.coeff	= n1;
	N2.coeff	= n2;
	N3.coeff	= n3;
	TMP.coeff	= tmp;

	mpz_init (FACTOR);
	mpz_init (TEMP);

	calib_copy_Z_vector (a, gpoly, k + 1);
	calib_copy_Z_vector (b, op -> coeff, k);

	adeg = k;
	bdeg = k - 1;
	while ((bdeg >= 0) AND (mpz_sgn (&b [bdeg]) EQ 0)) {
		--bdeg;
	}

	/* a = m1*gpoly + n1*op */
	mpz_set_ui (&m1 [0], 1);	mpz_set_ui (&n1 [0], 0);
	/* b = m2*gpoly + n2*op */
	mpz_set_ui (&m2 [0], 0);	mpz_set_ui (&n2 [0], 1);

	while (bdeg >= 0) {
		/* q,r = a divrem b.  Polynomial division in Zp[x]. */
		calib_copy_Z_vector (r, a, adeg + 1);
		for (i = 0; i <= k; i++) {
			mpz_set_ui (&q [i], 0);
		}
		rdeg = adeg;
		while (rdeg >= bdeg) {
			rlc = &r [rdeg];
			blc = &b [bdeg];
			j = rdeg - bdeg;
			/* factor = rlc/blc. */
			cf -> inv (cf, FACTOR, blc);
			cf -> mul (cf, FACTOR, FACTOR, rlc);
			mpz_set (&q [j], FACTOR);
			for (i = bdeg; i >= 0; i--) {
				cf -> mul (cf, TEMP, FACTOR, &b [i]);
				cf -> sub (cf, &r [i + j], &r [i + j], TEMP);
			}
			CALIB_FATAL_ERROR_IF (mpz_sgn (&r [rdeg]) NE 0);
			while ((rdeg >= 0) AND (mpz_sgn (&r [rdeg]) EQ 0)) {
				--rdeg;
			}
		}

		/* m3 = m1 - q*m2 */
		GFpk_mul (&TMP, &Q, &M2);
		GFpk_sub (&M3, &M1, &TMP);

		/* n3 = n1 - q*n2 */
		GFpk_mul (&TMP, &Q, &N2);
		GFpk_sub (&N3, &N1, &TMP);

		for (i = 0; i < k; i++) {
			mpz_set (&a [i], &b [i]);
			mpz_set (&b [i], &r [i]);
			mpz_set (&m1 [i], &m2 [i]);
			mpz_set (&m2 [i], &m3 [i]);
			mpz_set (&n1 [i], &n2 [i]);
			mpz_set (&n2 [i], &n3 [i]);
		}
		adeg = bdeg;
		bdeg = rdeg;
	}

	CALIB_FATAL_ERROR_IF (adeg NE 0);

	/* We have a = m1*gpoly + n1*op. */
	alc = &a [0];
	cf -> inv (cf, FACTOR, alc);
	coeff = rop -> coeff;
	for (i = 0; i < k; i++) {
		cf -> mul (cf, &coeff [i], FACTOR, &n1 [i]);
	}

	mpz_clear (TEMP);
	mpz_clear (FACTOR);

	calib_free_Z_vector (a, 11 * k + 2);
}

/*
 * Compute y, the p-th root of the given element op \in GF(p^k),
 * so that y^p = op.  This can be calculated as y = op^M, where M = p^(k-1).
 */

	static
	void
GFpk_pth_root (

struct calib_GFpk_obj *		rop,	/* IN/OUT: result */
const struct calib_GFpk_obj *	op	/* IN: element to take pth-root of */
)
{
int				k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
struct calib_GFpk_obj		A, B;
mpz_t				M;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f = rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE f);

	cf = f -> cf;
	k  = f -> k;

	mpz_init (M);
	GFpk_init (f, &A);
	GFpk_init (f, &B);

	/* Compute M = p^(k-1). */
	mpz_pow_ui (M, cf -> p, k - 1);

	GFpk_set_si (&A, 1);
	GFpk_set (&B, op);

	/* Exponentiation. */
	while (mpz_sgn (M) > 0) {
		if (mpz_odd_p (M)) {
			GFpk_mul (&A, &A, &B);
		}
		GFpk_mul (&B, &B, &B);
		mpz_tdiv_q_2exp (M, M, 1);
	}
	GFpk_set (rop, &A);

	GFpk_clear (&B);
	GFpk_clear (&A);
	mpz_clear (M);
}

/*
 * Convert the given element of Z(a) into the corresponding element of GF(p^k).
 * It is intended that the generator polynomial defining Z(a) be different
 * from the generator polynomial defining GF(p^k).
 */

	static
	void
GFpk_cvZa (

struct calib_GFpk_obj *		rop,	/* IN/OUT: result */
const struct calib_Za_obj *	op	/* IN: source element of Z(a) to */
					/*     convert */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	Zp;
const struct calib_Za_dom *	Za;
struct calib_GFpk_obj		RES;
mpz_srcptr			coeff;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f	= rop -> dom;
	Zp	= f -> cf;

	Za	= op -> dom;
	coeff	= op -> coeff;
	k	= Za -> gpoly -> degree;

	GFpk_init (f, &RES);

	mpz_init (TEMP);
	for (i = k - 1; i >= 0; i--) {
		GFpk_mul_a (&RES, &RES);

		Zp -> set_z (Zp, TEMP, &coeff [i]);
		Zp -> add (Zp, &RES.coeff [0], &RES.coeff [0], TEMP);
	}
	mpz_clear (TEMP);

	GFpk_set (rop, &RES);

	GFpk_clear (&RES);
}

/*
 * Return the degree of the given element of GF(p^k), i.e., the highest
 * non-zero power of the corresponding polynomial in the quotient
 * field Zp[x]/g(x).  The degree of the zero element is -1.
 */

	static
	int
GFpk_degree (

const struct calib_GFpk_obj *	op	/* IN: field element to get degree */
					/*     of */
)
{
int				deg;
const struct calib_GFpk_dom *	f;
mpz_srcptr			coeff;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	f	= op -> dom;
	coeff	= op -> coeff;

	deg = f -> k - 1;
	while ((deg >= 0) AND (mpz_sgn (&coeff [deg]) EQ 0)) {
		--deg;
	}
	return (deg);
}

/*
 * Is the given element of GF(p^k) zero?
 */

	static
	calib_bool
GFpk_zerop (

const struct calib_GFpk_obj *	op	/* IN: field element to test */
)
{
int		deg;

	deg = GFpk_degree (op);
	if (deg < 0) {
		return (TRUE);
	}
	return (FALSE);
}


/*
 * Is the given element of GF(p^k) one?
 */

	static
	calib_bool
GFpk_onep (

const struct calib_GFpk_obj *	op	/* IN: field element to test */
)
{
int	deg;

	deg = GFpk_degree (op);

	if ((deg EQ 0) AND (mpz_cmp_ui (&(op -> coeff [0]), 1) EQ 0)) {
		return (TRUE);
	}
	return (FALSE);
}

/*
 * Set rop to a random GF(p^k) value using the given random number generator.
 */

	static
	void
GFpk_set_random (

struct calib_GFpk_obj *	rop,	/* IN/OUT: randomized GF(p^k) value */
struct calib_Random *	randp	/* IN/OUT: random number generator to use */
)
{
int				i, k;
const struct calib_GFpk_dom *	cf;
const struct calib_Zp_dom *	Zp;
mpz_ptr				coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (randp EQ NULL));

	cf	= rop -> dom;
	coeff	= rop -> coeff;
	k	= cf -> k;
	Zp	= cf -> cf;

	for (i = 0; i < k; i++) {
		Zp -> set_random (Zp, &coeff [i], randp);
	}
}

/*
 * Convert the given general expression into a member of GF(p^k) = Zp[x]/f(x).
 * All of the variables in the given expression *MUST* match the given
 * variable name.
 */

	static
	void
GFpk_set_genrep (

struct calib_GFpk_obj *		rop,	/* IN/OUT: result */
const struct calib_genrep *	op,	/* IN: expression to convert */
const char *			var	/* IN: var for polynomial */
)
{
int				i, invert;
unsigned int			pow;
const struct calib_GFpk_dom *	f;
struct calib_genrep_list *	lp;
struct calib_GFpk_obj		TEMP1, TEMP2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (var EQ NULL));


	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		GFpk_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		GFpk_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		GFpk_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		CALIB_FATAL_ERROR_IF (strcmp (op -> u.var, var) NE 0);
		GFpk_set_var_power (rop, 1);
		break;

	case CALIB_GENREP_OP_ADD:
		f	= rop -> dom;
		GFpk_init (f, &TEMP1);
		GFpk_init (f, &TEMP2);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			GFpk_set_genrep (&TEMP2, lp -> operand, var);
			GFpk_add (&TEMP1, &TEMP1, &TEMP2);
		}
		GFpk_set (rop, &TEMP1);
		GFpk_clear (&TEMP2);
		GFpk_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_MUL:
		f	= rop -> dom;
		GFpk_init (f, &TEMP1);
		GFpk_init (f, &TEMP2);
		GFpk_set_si (&TEMP1, 1);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			GFpk_set_genrep (&TEMP2, lp -> operand, var);
			GFpk_mul (&TEMP1, &TEMP1, &TEMP2);
		}
		GFpk_set (rop, &TEMP1);
		GFpk_clear (&TEMP2);
		GFpk_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;
		if (i < 0) {
			pow = - i;
			invert = TRUE;
		}
		else {
			pow = i;
			invert = FALSE;
		}

		f	= rop -> dom;
		GFpk_init (f, &TEMP1);
		GFpk_init (f, &TEMP2);
		GFpk_set_si (&TEMP1, 1);

		if (op -> op EQ CALIB_GENREP_OP_VAR) {
			/* Optimize important special case. */
			GFpk_set_var_power (&TEMP1, pow);
		}
		else {
			GFpk_set_genrep (&TEMP2, op -> u.ipow.base, var);

			while (pow > 0) {
				if ((pow & 1) NE 0) {
					GFpk_mul (&TEMP1, &TEMP1, &TEMP2);
				}
				GFpk_mul (&TEMP2, &TEMP2, &TEMP2);
				pow >>= 1;
			}
		}
		if (invert) {
			GFpk_inv (&TEMP1, &TEMP1);
		}
		GFpk_set (rop, &TEMP1);
		GFpk_clear (&TEMP2);
		GFpk_clear (&TEMP1);
		break;

	default:
		CALIB_FATAL_ERROR;
	}
}

/*
 * Convert the given element of GF(p^k) = Zp[x]/f(x) into
 * general representation.  Use the variable name given.
 */

	static
	struct calib_genrep *
GFpk_to_genrep (

const struct calib_GFpk_obj *	op,	/* IN: element to convert */
const char *			var	/* IN: variable name to use */
)
{
int				i, k;
const struct calib_GFpk_dom *	f;
const struct calib_Zp_dom *	cf;
mpz_srcptr			coeff;
mpz_srcptr			value;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (var EQ NULL));

	f	= op -> dom;
	coeff	= op -> coeff;

	cf	= f -> cf;
	k	= f -> k;

	root = NULL;
	hookp = &root;

	for (i = k - 1; i >= 0; i--) {
		value = &coeff [i];
		if (mpz_sgn (value) EQ 0) continue;
		vgp = cf -> to_genrep (cf, value);
		gp1 = calib_genrep_poly_term (vgp, var, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	return (gp1);
}

/*
 * Prototype GFpk domain object.
 */

static const struct calib_GFpk_dom	GFpk_dom = {
	.init			= GFpk_init,
	.clear			= GFpk_clear,
	.set			= GFpk_set,
	.set_si			= GFpk_set_si,
	.set_z			= GFpk_set_z,
	.set_q			= GFpk_set_q,
	.set_var_power		= GFpk_set_var_power,
	.add			= GFpk_add,
	.sub			= GFpk_sub,
	.neg			= GFpk_neg,
	.mul			= GFpk_mul,
	.mul_z			= GFpk_mul_z,
	.mul_a			= GFpk_mul_a,
	.ipow			= GFpk_ipow,
	.inv			= GFpk_inv,
	.pth_root		= GFpk_pth_root,
	.cvZa			= GFpk_cvZa,
	.zerop			= GFpk_zerop,
	.onep			= GFpk_onep,
	.set_random		= GFpk_set_random,
	.set_genrep		= GFpk_set_genrep,
	.to_genrep		= GFpk_to_genrep,
};

/*
 * Construct the field GF(p^k) = Zp/g(x), where g(x) is the given
 * monic, irreducible generator polynomial in Zp[x].
 */

	struct calib_GFpk_dom *
calib_make_GFpk_dom (

const struct calib_Zp_dom *	cf,	/* IN: coefficient field Zp */
const struct calib_Zpx_obj *	gpoly	/* IN: generator polynomial */
)
{
	return (calib_make_GFpk_dom_z (cf, gpoly -> degree, gpoly -> coeff));
}


/*
 * Construct the field GF(p^k) = Zp/g(x), where g(x) is the given
 * monic, irreducible generator polynomial in Zp[x].
 */

	struct calib_GFpk_dom *
calib_make_GFpk_dom_z (

const struct calib_Zp_dom *	cf,	/* IN: coefficient field Zp */
int				k,	/* IN: degree of g(x) */
mpz_srcptr			gcoeff	/* IN: coefficients of g(x) */
)
{
int			i, n;
struct calib_GFpk_dom *	f;
mpz_ptr			buf;
mpz_srcptr		zp;
mpz_srcptr		lc;
mpz_t			INV;

	/* Verify that coefficients are in range 0..p-1. */
	zp  = gcoeff;
	for (i = 0; i <= k; i++, zp++) {
		CALIB_FATAL_ERROR_IF (mpz_sgn (zp) < 0);
		CALIB_FATAL_ERROR_IF (mpz_cmp (zp, cf -> p) >= 0);
	}

	f = CALIB_NEW (struct calib_GFpk_dom);

	/* Initialize "most" things. */
	*f = GFpk_dom;

	n = k + 1;
	buf = calib_new_Z_vector (n);
	calib_copy_Z_vector (buf, gcoeff, n);

	f -> cf		= cf;
	f -> k		= k;
	f -> coeff	= buf;

	lc = &buf [k];
	if (mpz_cmp_ui (lc, 1) NE 0) {
		/* Not monic.  Force it to be monic. */
		printf ("Warning: generator polynomial is not monic.\n");
		mpz_init (INV);
		cf -> inv (cf, INV, lc);
		for (i = 0; i < k; i++) {
			cf -> mul (cf, &buf [i], &buf [i], INV);
		}
		mpz_set_ui (&buf [k], 1);
	}

	return (f);
}

/*
 * Free up the given GF(p^k) field description.
 */

	void
calib_free_GFpk_dom (

struct calib_GFpk_dom *	f	/* IN: field description to free */
)
{
int		k;

	if (f NE NULL) {
		k = f -> k;
		calib_free_Z_vector ((mpz_ptr) (f -> coeff), k + 1);
		memset (f, 0, sizeof (*f));
		free (f);
	}
}
