/**********************************************************************

	$Id: GFpkx.c,v 1.1 2024/10/20 00:36:25 warme Exp $

	File:	GFpkx.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring of polynomials K[x] over the field K = GF(q),
	where GF(q) is a Galois field, q = p^k, and p is a prime.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/GFpkx.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/GFpk.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/random.h"
#include "calib/Za.h"
#include "calib/Zax.h"
#include "calib/Zx.h"
#include <gmp.h>
#include <stdlib.h>
#include <string.h>


static void	GFpkx_add (struct calib_GFpkx_obj *		rop,
			   const struct calib_GFpkx_obj *	op1,
			   const struct calib_GFpkx_obj *	op2);
static void	GFpkx_alloc (struct calib_GFpkx_obj *		rop,
			     int				degree);
static void	GFpkx_clear (struct calib_GFpkx_obj *		x);
static void	GFpkx_clear_array (struct calib_GFpkx_obj *	parray,
				   int				npoly);
static int	GFpkx_count_multiplicity (
				struct calib_GFpkx_obj *	poly,
				const struct calib_GFpkx_obj *	factor);
static struct calib_GFpkx_factor *
		GFpkx_compute_square_free_factorization (
				struct calib_GFpkx_obj *	poly);
static void	GFpkx_copy_GFpk_vector (
				struct calib_GFpk_obj *		dst,
				const struct calib_GFpk_obj *	src,
				int				n);
static struct calib_GFpkx_obj *
		GFpkx_cvZax (const struct calib_GFpkx_dom *	K_of_x,
			     const struct calib_Zax_obj *	op);
static struct calib_GFpkx_factor *
		GFpkx_ddf (const struct calib_GFpkx_obj *	sqfpoly,
			   struct calib_Random *		randp);
static void	GFpkx_derivative (struct calib_GFpkx_obj *		deriv,
				  const struct calib_GFpkx_obj *	poly);
static void	GFpkx_div (struct calib_GFpkx_obj *		quotient,
			   struct calib_GFpkx_obj *		remainder,
			   const struct calib_GFpkx_obj *	op1,
			   const struct calib_GFpkx_obj *	op2);
static struct calib_GFpkx_obj *
		GFpkx_dup (const struct calib_GFpkx_obj *	op);
static void	GFpkx_eval (struct calib_GFpk_obj *		rop,
			    const struct calib_GFpkx_obj *	op,
			    const struct calib_GFpk_obj *	value);
static void	GFpkx_extgcd (struct calib_GFpkx_obj *		gcd,
			      struct calib_GFpkx_obj *		xa,
			      struct calib_GFpkx_obj *		xb,
			      const struct calib_GFpkx_obj *	a,
			      const struct calib_GFpkx_obj *	b);
static struct calib_GFpkx_factor *
		GFpkx_factor (const struct calib_GFpkx_obj *	poly);
static struct calib_genrep *
		GFpkx_factors_to_genrep (
			const struct calib_GFpkx_factor *	factors,
			const char *				xvar,
			const char *				avar);
static void	GFpkx_free (struct calib_GFpkx_obj * poly);
static void	GFpkx_free_GFpk_vector (struct calib_GFpk_obj *	array,
				        int			n);
static void	GFpkx_free_factors (struct calib_GFpkx_factor *	factors);
static void	GFpkx_gcd (struct calib_GFpkx_obj *		gcd,
			   const struct calib_GFpkx_obj *	a,
			   const struct calib_GFpkx_obj *	b);
static void	GFpkx_init (const struct calib_GFpkx_dom *	K_of_x,
			    struct calib_GFpkx_obj *		x);
static void	GFpkx_init_array_degree (
				const struct calib_GFpkx_dom *	K_of_x,
				struct calib_GFpkx_obj *	parray,
				int				npoly,
				int				degree);
static void	GFpkx_init_degree (const struct calib_GFpkx_dom *	K_of_x,
				   struct calib_GFpkx_obj *		poly,
				   int					degree);
static void	GFpkx_ipow (struct calib_GFpkx_obj *		rop,
			    const struct calib_GFpkx_obj *	op,
			    int					power);
static void	GFpkx_monicize (struct calib_GFpkx_obj *	poly);
static void	GFpkx_mul (struct calib_GFpkx_obj *		rop,
			   const struct calib_GFpkx_obj *	op1,
			   const struct calib_GFpkx_obj *	op2);
static void	GFpkx_mul_z (struct calib_GFpkx_obj *		rop,
			     const struct calib_GFpkx_obj *	op1,
			     mpz_srcptr				op2);
static void	GFpkx_neg (struct calib_GFpkx_obj *		rop,
			   const struct calib_GFpkx_obj *	op);
static struct calib_GFpk_obj *
		GFpkx_new_GFpk_vector (const struct calib_GFpk_dom *	cf,
				       int				n);
static calib_bool
		GFpkx_onep (const struct calib_GFpkx_obj *	op);
static void	GFpkx_poly_mul_mod_poly (
				struct calib_GFpkx_obj *	rop,
				const struct calib_GFpkx_obj *	a,
				const struct calib_GFpkx_obj *	b,
				const struct calib_GFpkx_obj *	modpoly);
static void	GFpkx_poly_to_int_pow_mod_poly (
				struct calib_GFpkx_obj *	rop,
				const struct calib_GFpkx_obj *	poly,
				mpz_srcptr			exponent,
				const struct calib_GFpkx_obj *	modpoly);
static void	GFpkx_set_random (struct calib_GFpkx_obj *		poly,
				   int					d,
				   struct calib_Random *		randp);
static void	GFpkx_realloc_copy_poly (struct calib_GFpkx_obj *	poly,
					 const struct calib_GFpk_obj *	coeff,
					 int				degree);
static void	GFpkx_set (struct calib_GFpkx_obj *		rop,
			   const struct calib_GFpkx_obj *	op);
static void	GFpkx_set_genrep (struct calib_GFpkx_obj *		rop,
				  const struct calib_genrep *		op,
				  const char *				xvar,
				  const char *				avar);
static void	GFpkx_set_q (struct calib_GFpkx_obj *		rop,
			     mpq_srcptr				op);
static void	GFpkx_set_si (struct calib_GFpkx_obj *		rop,
			      calib_si_t			op);
static void	GFpkx_set_var_power (struct calib_GFpkx_obj *	rop,
				     int			power);
static void	GFpkx_set_z (struct calib_GFpkx_obj *		rop,
			     mpz_srcptr				op);
static struct calib_GFpkx_factor **
		GFpkx_split_factors (struct calib_GFpkx_factor **	hookp,
				     const struct calib_GFpkx_obj *	poly,
				     int				d,
				     struct calib_Random *		randp);
static void	GFpkx_sub (struct calib_GFpkx_obj *		rop,
			   const struct calib_GFpkx_obj *	op1,
			   const struct calib_GFpkx_obj *	op2);
static struct calib_genrep *
		GFpkx_to_genrep (const struct calib_GFpkx_obj *	op,
				 const char *			xvar,
				 const char *			avar);
static calib_bool
		GFpkx_zerop (const struct calib_GFpkx_obj *	op);

/* Standard "max" of two integers. */

#define _MAX(A,B)	(((A) > (B)) ? (A) : (B))

/*
 * Macros to allocate/zero a local polynomial of the given degree.
 */

#define ALLOCATE_LOCAL_POLY(K_OF_X, PNAME, DEGREE) \
	struct calib_GFpkx_obj	PNAME; \
	GFpkx_init_degree ((K_OF_X), &PNAME, (DEGREE))

#define FREE_LOCAL_POLY(PNAME) \
	GFpkx_clear (&PNAME)

/*
 * Macros to allocate/zero an array of local polynomials, each having
 * the given degree.
 */

#define ALLOCATE_LOCAL_POLY_ARRAY(K_OF_X, PNAME, NPOLY, DEGREE) \
	struct calib_GFpkx_obj	PNAME [NPOLY]; \
	GFpkx_init_array_degree ((K_OF_X), &PNAME [0], (NPOLY), (DEGREE))

#define FREE_LOCAL_POLY_ARRAY(PNAME, NPOLY) \
	GFpkx_clear_array(&PNAME [0], (NPOLY))

/*
 * Initialize the given polynomial.  Its value after initialization is zero.
 */

	static
	inline
	void
GFpkx_init (

const struct calib_GFpkx_dom *	K_of_x,	/* IN: polynomial domain */
struct calib_GFpkx_obj *	x	/* IN/OUT: polynomial to initialize */
)
{
	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR (x EQ NULL));

	x -> degree	= -1;
	x -> size	= 0;
	x -> dom	= K_of_x;
	x -> coeff	= NULL;
}

/*
 * Initialize the given polynomial to zero, but provide it with a
 * coefficient buffer sufficient to hold a polynomial of the given degree.
 */

	static
	inline
	void
GFpkx_init_degree (

const struct calib_GFpkx_dom *	K_of_x,	/* IN: domain of polynomial */
struct calib_GFpkx_obj *	x,	/* IN: polynomial to initialize */
int				degree	/* IN: maximum degree */
)
{
int				size;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		coeff;

	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR
			      (x      EQ NULL) OR
			      (degree  < -1));

	cf = K_of_x -> cf;

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	coeff = GFpkx_new_GFpk_vector (cf, size);

	x -> degree	= -1;
	x -> size	= size;
	x -> dom	= K_of_x;
	x -> coeff	= coeff;
}

/*
 * Initialize the given array of polynomials to zero, but provide each
 * with a coefficient buffer sufficient to hold a polynomial of the
 * given degree.
 *
 * Although it is tempting to allocate all of the coefficients for all of
 * the polynomials into one big long vector, we cannot do so because the
 * coefficient vector for each polynomial must be separately free-able,
 * and re-allocatable.
 */

	static
	inline
	void
GFpkx_init_array_degree (

const struct calib_GFpkx_dom *
		K_of_x,		/* IN: domain of polynomial */
struct calib_GFpkx_obj *
		parray,		/* IN/OUT: array of polynomials */
int		npoly,		/* IN: number of polynomials in array */
int		degree		/* IN: max degree of each polynomial */
)
{
int				i;
struct calib_GFpkx_obj *	pp;

	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR (degree < -1));

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (parray EQ NULL);

	pp = &parray [0];
	for (i = 0; i < npoly; i++, pp++) {
		GFpkx_init_degree (K_of_x, pp, degree);
	}
}

/*
 * Force the given polynomial to have buffer space sufficient to hold
 * a polynomial of at least the given degree.  This does not change
 * the value currently residing in this polynomial.
 */

	static
	void
GFpkx_alloc (

struct calib_GFpkx_obj *	rop,	/* IN/OUT: poly to alloc buffer for */
int				degree	/* IN: degree desired */
)
{
int				i, size, dsize;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		p1;
struct calib_GFpk_obj *		p2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (degree < -1));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;

	size	= degree + 1;
	dsize	= rop -> size;

	if (dsize < size) {
		p1 = rop -> coeff;
		p2 = GFpkx_new_GFpk_vector (cf, size);
		for (i = 0; i < dsize; i++) {
			cf -> set (&p2 [i], &p1 [i]);
		}
		GFpkx_free_GFpk_vector (p1, dsize);
		rop -> coeff = p1;
		rop -> size  = size;
	}
}

/*
 * Clear out the given polynomial.
 */

	static
	inline
	void
GFpkx_clear (

struct calib_GFpkx_obj *	x	/* IN/OUT: polynomial to clear */
)
{
int			size;
struct calib_GFpk_obj *	coeff;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	coeff = x -> coeff;
	if (coeff NE NULL) {
		size = x -> size;
		CALIB_FATAL_ERROR_IF (size <= 0);
		GFpkx_free_GFpk_vector (coeff, size);
	}
	
	x -> degree	= -1;
	x -> size	= 0;
	x -> dom	= NULL;
	x -> coeff	= NULL;
}

/*
 * Clear out the given array of polynomials.
 */

	static
	inline
	void
GFpkx_clear_array (

struct calib_GFpkx_obj * parray,	/* IN: array of polynomials to clear */
int			npoly		/* IN: number of polynomials in parray */
)
{
int		i;

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (parray EQ NULL);

	for (i = 0; i < npoly; i++) {
		GFpkx_clear (&parray [i]);
	}
}

/*
 * Set the given destination polynomial from the given coefficient
 * array.  Reallocate the destination polynomial's coefficient array,
 * if necessary.
 */

	static
	inline
	void
GFpkx_realloc_copy_poly (

struct calib_GFpkx_obj *	poly,	/* IN: destination polynomial */
const struct calib_GFpk_obj *	coeff,	/* IN: source coefficient array */
int				degree	/* IN: degree to set */
)
{
int				i, cur_size, new_size;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		newcf;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (degree < -1));

	poly -> degree	= degree;

	if (degree < 0) {
		/* We allow (coeff EQ NULL) here (but do not require). */
		return;
	}

	CALIB_FATAL_ERROR_IF (coeff EQ NULL);

	new_size = degree + 1;
	if (new_size > 0) {
		cf = poly -> dom -> cf;
		cur_size = poly -> size;
		if (new_size > cur_size) {
			GFpkx_free_GFpk_vector (poly -> coeff, cur_size);
			newcf = GFpkx_new_GFpk_vector (cf, new_size);
			poly -> coeff	= newcf;
			poly -> size	= new_size;
		}
		for (i = 0; i < new_size; i++) {
			cf -> set (&(poly -> coeff [i]), &coeff [i]);
		}
	}
}

/*
 * Allocate a vector of GFpk coefficients.
 */

	static
	struct calib_GFpk_obj *
GFpkx_new_GFpk_vector (

const struct calib_GFpk_dom *	cf,	/* IN: Coefficient field */
int				n	/* IN: array size */
)
{
int			i;
struct calib_GFpk_obj *	array;

	array = NULL;
	if (n > 0) {
		array = CALIB_NEWA (n, struct calib_GFpk_obj);
		for (i = 0; i < n; i++) {
			cf -> init (cf, &array [i]);
		}
	}

	return (array);
}

/*
 * Free up a vector of GFpk coefficients.
 */

	static
	void
GFpkx_free_GFpk_vector (

struct calib_GFpk_obj *	array,		/* IN: array to free */
int			n		/* IN: array elements */
)
{
int				i;
const struct calib_GFpk_dom *	cf;

	if (array NE NULL) {
		for (i = 0; i < n; i++) {
			cf = array [i].dom;
			cf -> clear (&array [i]);
		}
		free (array);
	}
}

/*
 * Copy a vector of GFpk objects.
 */

	static
	void
GFpkx_copy_GFpk_vector (

struct calib_GFpk_obj *		dst,	/* IN/OUT: destination vector */
const struct calib_GFpk_obj *	src,	/* IN: source vector */
int				n	/* IN: num elements to copy */
)
{
int				i;
const struct calib_GFpk_dom *	cf;

	if  ((dst NE src) AND (n > 0)) {
		cf = dst -> dom;
		for (i = 0; i < n; i++) {
			cf -> set (&dst [i], &src [i]);
		}
	}
}

/*
 * Set the given polynomial to the given constant integer value.
 */

	static
	void
GFpkx_set (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: destination */
const struct calib_GFpkx_obj *	op		/* IN: source */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {
		GFpkx_realloc_copy_poly (rop, op -> coeff, op -> degree);
	}
}

/*
 * Set the given polynomial to the given constant integer value.
 */

	static
	void
GFpkx_set_si (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: destination */
calib_si_t			op		/* IN: source */
)
{
const struct calib_GFpk_dom *	cf;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	if (op EQ 0) {
		rop -> degree = -1;
	}
	else {
		GFpkx_alloc (rop, 0);
		cf = rop -> dom -> cf;
		cf -> set_si (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given polynomial to the given GMP integer value.
 */

	static
	void
GFpkx_set_z (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: destination */
mpz_srcptr			op		/* IN: source */
)
{
const struct calib_GFpk_dom *	cf;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpz_sgn (op) EQ 0) {
		rop -> degree = -1;
	}
	else {
		GFpkx_alloc (rop, 0);
		cf = rop -> dom -> cf;
		cf -> set_z (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given polynomial to the given GMP rational value.
 */

	static
	void
GFpkx_set_q (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: destination */
mpq_srcptr			op		/* IN: source */
)
{
const struct calib_GFpk_dom *	cf;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpq_sgn (op) EQ 0) {
		rop -> degree = -1;
	}
	else {
		GFpkx_alloc (rop, 0);
		cf = rop -> dom -> cf;
		cf -> set_q (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given GF(p^k)[x] polynomial to x**power.
 */

	static
	void
GFpkx_set_var_power (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: result */
int				power		/* IN: power to set */
)
{
int				i;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (power < 0));

	K_of_x	= rop -> dom;
	coeff	= rop -> coeff;
	cf	= K_of_x -> cf;

	GFpkx_alloc (rop, power);

	for (i = 0; i < power; i++) {
		cf -> set_si (&coeff [i], 0);
	}
	cf -> set_si (&coeff [power], 1);
	rop -> degree = power;
}

/*
 * Addition of two polynomials over GF(p^k).
 */

	static
	void
GFpkx_add (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: result */
const struct calib_GFpkx_obj *	op1,		/* IN: first operand */
const struct calib_GFpkx_obj *	op2		/* IN: second operand */
)
{
int				i, rdeg, deg1, deg2, rsize;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		res;
struct calib_GFpk_obj *		rp;
const struct calib_GFpk_obj *	cp1;
const struct calib_GFpk_obj *	cp2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE K_of_x);
	CALIB_FATAL_ERROR_IF (op2 -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = GFpkx_new_GFpk_vector (cf, rsize);

	cp1 = &(op1 -> coeff [0]);
	cp2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, rp++, cp1++, cp2++) {
		if (i <= deg1) {
			if (i <= deg2) {
				cf -> add (rp, cp1, cp2);
			}
			else {
				cf -> set (rp, cp1);
			}
		}
		else {
			cf -> set (rp, cp2);
		}
	}
	while ((rdeg >= 0) AND (cf -> zerop (&res [rdeg]))) {
		--rdeg;
	}

	GFpkx_realloc_copy_poly (rop, res, rdeg);
	GFpkx_free_GFpk_vector (res, rsize);
}

/*
 * Subtraction of two polynomials over GF(p^k).
 */

	static
	void
GFpkx_sub (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: result */
const struct calib_GFpkx_obj *	op1,		/* IN: first operand */
const struct calib_GFpkx_obj *	op2		/* IN: second operand */
)
{
int				i, rdeg, deg1, deg2, rsize;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		res;
struct calib_GFpk_obj *		rp;
const struct calib_GFpk_obj *	cp1;
const struct calib_GFpk_obj *	cp2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE K_of_x);
	CALIB_FATAL_ERROR_IF (op2 -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = GFpkx_new_GFpk_vector (cf, rsize);

	cp1 = &(op1 -> coeff [0]);
	cp2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, rp++, cp1++, cp2++) {
		if (i <= deg1) {
			if (i <= deg2) {
				cf -> sub (rp, cp1, cp2);
			}
			else {
				cf -> set (rp, cp1);
			}
		}
		else {
			cf -> neg (rp, cp2);
		}
	}
	while ((rdeg >= 0) AND (cf -> zerop (&res [rdeg]))) {
		--rdeg;
	}
	GFpkx_realloc_copy_poly (rop, res, rdeg);
	GFpkx_free_GFpk_vector (res, rsize);
}

/*
 * Negate the given polynomial over GF(p^k).
 */

	static
	void
GFpkx_neg (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: result */
const struct calib_GFpkx_obj *	op		/* IN: operand to negate */
)
{
int				i, deg;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		dstp;
const struct calib_GFpk_obj *	srcp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	if (rop EQ op) {
		deg	= rop -> degree;
		dstp	= rop -> coeff;
		for (i = 0; i <= deg; i++, dstp++) {
			cf -> neg (dstp, dstp);
		}
	}
	else {
		deg	= op -> degree;
		srcp	= op -> coeff;
		GFpkx_alloc (rop, deg);
		dstp	= rop -> coeff;
		for (i = 0; i <= deg; i++, dstp++, srcp++) {
			cf -> neg (dstp, srcp);
		}
		rop -> degree = deg;
	}
}

/*
 * Multiplcation of two polynomials over GF(p^k).
 */

	static
	void
GFpkx_mul (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: result */
const struct calib_GFpkx_obj *	op1,		/* IN: first operand */
const struct calib_GFpkx_obj *	op2		/* IN: second operand */
)
{
int				i, j, rdeg, deg1, deg2, rsize;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		res;
struct calib_GFpk_obj *		rp1;
struct calib_GFpk_obj *		rp2;
const struct calib_GFpk_obj *	cp1;
const struct calib_GFpk_obj *	cp2;
struct calib_GFpk_obj		TMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE K_of_x);
	CALIB_FATAL_ERROR_IF (op2 -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	if ((deg1 < 0) OR (deg2 < 0)) {
		/* Multipying by zero. */
		rop -> degree = -1;
		return;
	}

	rdeg	= deg1 + deg2;

	rsize = rdeg + 1;
	res = GFpkx_new_GFpk_vector (cf, rsize);
	cf -> init (cf, &TMP);

	cp1 = &(op1 -> coeff [0]);
	rp1 = &res [0];
	for (i = 0; i <= deg1; i++, rp1++, cp1++) {
		cp2 = &(op2 -> coeff [0]);
		rp2 = rp1;
		for (j = 0; j <= deg2; j++, rp2++, cp2++) {
			cf -> mul (&TMP, cp1, cp2);
			cf -> add (rp2, rp2, &TMP);
		}
	}
	while ((rdeg >= 0) AND (cf -> zerop (&res [rdeg]))) {
		--rdeg;
	}

	GFpkx_realloc_copy_poly (rop, res, rdeg);
	cf -> clear (&TMP);
	GFpkx_free_GFpk_vector (res, rsize);
}

/*
 * Multiplcation of a polynomial in GF(p^k) by an integer z.
 */

	static
	void
GFpkx_mul_z (

struct calib_GFpkx_obj *	rop,		/* IN/OUT: result */
const struct calib_GFpkx_obj *	op1,		/* IN: first operand */
mpz_srcptr			op2		/* IN: second operand */
)
{
int				i, adeg, rdeg, rsize;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		res;
struct calib_GFpk_obj *		dstp;
const struct calib_GFpk_obj *	srcp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	adeg	= op1 -> degree;

	if ((adeg < 0) OR (mpz_sgn (op2) EQ 0)) {
		/* Multipying by zero. */
		rop -> degree = -1;
		return;
	}

	rdeg = adeg;
	rsize = rdeg + 1;
	res = GFpkx_new_GFpk_vector (cf, rsize);
	srcp = &(op1 -> coeff [0]);
	dstp = res;

	for (i = 0; i <= adeg; i++, srcp++, dstp++) {
		cf -> mul_z (dstp, srcp, op2);
	}
	while ((rdeg >= 0) AND (cf -> zerop (&res [rdeg]))) {
		--rdeg;
	}

	GFpkx_realloc_copy_poly (rop, res, rdeg);
	GFpkx_free_GFpk_vector (res, rsize);
}

/*
 * Take the given polynomial to the given integer power.
 */

	static
	void
GFpkx_ipow (

struct calib_GFpkx_obj *	rop,	/* IN/OUT: result polynomial */
const struct calib_GFpkx_obj *	op,	/* IN: polynomial to exponentiate */
int				power	/* IN: integer power */
)
{
const struct calib_GFpkx_dom *	K_of_x;
struct calib_GFpkx_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (power < 0));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	GFpkx_init (K_of_x, &RES);
	GFpkx_init (K_of_x, &PREG);

	GFpkx_set_si (&RES, 1);
	GFpkx_set (&PREG, op);

	while (power > 0) {
		if ((power & 1) NE 0) {
			GFpkx_mul (&RES, &RES, &PREG);
		}
		GFpkx_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	GFpkx_set (rop, &RES);
	GFpkx_clear (&PREG);
	GFpkx_clear (&RES);
}

/*
 * Create a dynamically allocated copy of the given polynomial.
 */

	static
	struct calib_GFpkx_obj *
GFpkx_dup (

const struct calib_GFpkx_obj *	op	/* IN: polynomial to duplicate */
)
{
int				deg, size;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpkx_obj *	dst;
struct calib_GFpk_obj *		cp;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg	= op -> degree;
	K_of_x	= op -> dom;
	cf	= K_of_x -> cf;

	size	= deg + 1;

	dst	= CALIB_NEW (struct calib_GFpkx_obj);
	cp	= NULL;
	if (size > 0) {
		cp = GFpkx_new_GFpk_vector (cf, size);
		GFpkx_copy_GFpk_vector (cp, op -> coeff, size);
	}
	dst -> degree	= deg;
	dst -> size	= size;
	dst -> dom	= K_of_x;
	dst -> coeff	= cp;

	return (dst);
}

/*
 * Free the given polynomial, which must have been dynamically allocated.
 */

	static
	void
GFpkx_free (

struct calib_GFpkx_obj *	poly	/* IN: polynomial to free */
)
{
	if (poly NE NULL) {
		GFpkx_clear (poly);
		free (poly);
	}
}

/*
 * Evaluate the given poly(value).
 */

	static
	void
GFpkx_eval (

struct calib_GFpk_obj *		rop,		/* IN/OUT: result */
const struct calib_GFpkx_obj *	op,		/* IN: poly to evaluate */
const struct calib_GFpk_obj *	value		/* IN: value to use */
)
{
int				i;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj		SUM;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (value EQ NULL));

	K_of_x	= op -> dom;
	cf	= K_of_x -> cf;

	CALIB_FATAL_ERROR_IF ((rop -> dom NE cf) OR (value -> dom NE cf));

	cf -> init (cf, &SUM);

	for (i = op -> degree; i >= 0; i--) {
		cf -> mul (&SUM, &SUM, value);
		cf -> add (&SUM, &SUM, &(op -> coeff [i]));
	}
	cf -> set (rop, &SUM);

	cf -> clear (&SUM);
}

/*
 * The division algorith (with remainder):
 * Given polynomials a, b \in GF(p^k)[x], find polynomials q, r \in GF(p^k)[x]
 * such that:
 *
 *	a = q*b + r	AND	degree (r) < degree (b)
 *
 * This is easy and clean because we have polynomials over a field,
 * which supports division.
 */

	static
	void
GFpkx_div (

struct calib_GFpkx_obj *	quotient,	/* IN/OUT: quotient */
struct calib_GFpkx_obj *	remainder,	/* IN/OUT: remainder */
const struct calib_GFpkx_obj *	op1,		/* IN: dividend */
const struct calib_GFpkx_obj *	op2		/* IN: divisor */
)
{
int				i, j, adeg, bdeg, qdeg, rdeg;
int				size, total_size;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		quot;
struct calib_GFpk_obj *		rem;
struct calib_GFpk_obj *		qp;
struct calib_GFpk_obj *		rp;
struct calib_GFpk_obj *		zp;
const struct calib_GFpk_obj *	ap;
const struct calib_GFpk_obj *	bp;
struct calib_GFpk_obj		BLC_INV, TMP;

	CALIB_FATAL_ERROR_IF ((op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= op1 -> dom;

	CALIB_FATAL_ERROR_IF (op2 -> dom NE K_of_x);
	CALIB_FATAL_ERROR_IF ((quotient NE NULL) AND
			      (quotient -> dom NE K_of_x));
	CALIB_FATAL_ERROR_IF ((remainder NE NULL) AND
			      (remainder -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	adeg	= op1 -> degree;
	bdeg	= op2 -> degree;

	/* Division by zero is not permitted. */
	CALIB_FATAL_ERROR_IF (bdeg < 0);

	if (adeg < 0) {
		if (quotient NE NULL) {
			quotient -> degree	= -1;
		}
		if (remainder NE NULL) {
			remainder -> degree	= -1;
		}
		return;
	}

	/* Local coefficient buffers. */
	size = adeg + 1;
	total_size = 2 * size;
	zp = GFpkx_new_GFpk_vector (cf, total_size);
	quot	= zp;		zp += size;
	rem	= zp;		zp += size;

	cf -> init (cf, &BLC_INV);
	cf -> init (cf, &TMP);

	ap = &(op1 -> coeff [0]);
	bp = &(op2 -> coeff [0]);

	if (bdeg <= 0) {
		/* Division by a constant.  Do not allow division by zero. */
		CALIB_FATAL_ERROR_IF ((bdeg < 0) OR (cf -> zerop (bp)));

		cf -> inv (&BLC_INV, bp);
		rp = &rem [0];
		for (i = 0; i <= adeg; i++, ap++, rp++) {
			cf -> mul (rp, ap, &BLC_INV);
		}
		if (quotient NE NULL) {
			GFpkx_realloc_copy_poly (quotient, rem, adeg);
		}
		if (remainder NE NULL) {
			remainder -> degree = -1;
		}

		GFpkx_free_GFpk_vector (quot, total_size);

		return;
	}

	qdeg = adeg - bdeg;
	if (qdeg < -1) {
		qdeg = -1;
	}

	rdeg = adeg;
	GFpkx_copy_GFpk_vector (&rem [0], ap, rdeg + 1);

	/* Compute multiplicative inverse of lc(b). */
	cf -> inv (&BLC_INV, &(op2 -> coeff [bdeg]));

	while (rdeg >= bdeg) {
		j = rdeg - bdeg;
		qp = &quot [j];
		cf -> mul (qp, &rem [rdeg], &BLC_INV);
		rp = &rem [rdeg - 1];
		bp = &(op2 -> coeff [bdeg - 1]);
		for (i = bdeg - 1; i >= 0; i--, rp--, bp--) {
			cf -> mul (&TMP, qp, bp);
			cf -> sub (rp, rp, &TMP);
		}
		--rdeg;
		rp = &rem [rdeg];
		while ((rdeg >= 0) AND (cf -> zerop (rp))) {
			--rdeg;
			--rp;
		}
	}

	if (quotient NE NULL) {
		GFpkx_realloc_copy_poly (quotient, quot, qdeg);
	}

	if (remainder NE NULL) {
		GFpkx_realloc_copy_poly (remainder, rem, rdeg);
	}

	cf -> clear (&TMP);
	cf -> clear (&BLC_INV);
	GFpkx_free_GFpk_vector (quot, total_size);
}

/*
 * Compute the Greatest Common Divisor of two polynomials over GF(p^k).
 * The result is always monic unless a = b = 0.
 *
 * We use the conventional Euclidean algorithm.
 */

	static
	void
GFpkx_gcd (

struct calib_GFpkx_obj *	gcd,		/* IN: GCD result */
const struct calib_GFpkx_obj *	a,		/* IN: first operand */
const struct calib_GFpkx_obj *	b		/* IN: second operand */
)
{
int				i, adeg, bdeg, maxdeg;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpkx_obj *	ap;
struct calib_GFpkx_obj *	bp;
struct calib_GFpkx_obj *	rp;
struct calib_GFpkx_obj *	qp;
struct calib_GFpkx_obj *	temp_ptr;
struct calib_GFpk_obj *		lca_inv;
struct calib_GFpk_obj *		cp;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	K_of_x	= gcd -> dom;

	CALIB_FATAL_ERROR_IF (a -> dom NE K_of_x);
	CALIB_FATAL_ERROR_IF (b -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	adeg = a -> degree;
	bdeg = b -> degree;
	maxdeg = _MAX (adeg, bdeg);

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 4, maxdeg);

	temp_ptr = &TMP [0];
	ap = temp_ptr++;
	bp = temp_ptr++;
	rp = temp_ptr++;
	qp = temp_ptr++;

	GFpkx_set (ap, a);
	GFpkx_set (bp, b);

	while (bp -> degree >= 0) {
		GFpkx_div (qp, rp, ap, bp);
		temp_ptr = ap;
		ap = bp;
		bp = rp;
		rp = temp_ptr;
	}

	adeg = ap -> degree;
	if (adeg >= 0) {
		/* Monicize result. */
		lca_inv = &(qp -> coeff [0]);
		cf -> inv (lca_inv, &(ap -> coeff [adeg]));
		cp = &(ap -> coeff [0]);
		for (i = 0; i <= adeg; i++, cp++) {
			cf -> mul (cp, cp, lca_inv);
		}
	}

	GFpkx_set (gcd, ap);

	FREE_LOCAL_POLY_ARRAY (TMP, 4);
}

/*
 * The Extended Euclidean Algorithm for polynomials in K[x],
 * where K = GF(p^k).  Given polynomials a, b \in K[x],
 * compute polynomials g, xa, xb \in K[x] such that
 *
 * 1.	a is a multiple of g,
 * 2.	b is a multiple of g,
 * 3.	xa * a + xb * b = g.
 *
 * These three conditions imply that g = gcd(a,b).  The polynomial g
 * is monic (unless a = b = 0, in which case g = 0).
 */

	static
	void
GFpkx_extgcd (

struct calib_GFpkx_obj *	gcd,		/* IN/OUT: GCD result */
struct calib_GFpkx_obj *	xa,		/* IN/OUT: a multiplier */
struct calib_GFpkx_obj *	xb,		/* IN/OUT: b multiplier */
const struct calib_GFpkx_obj *	a,		/* IN: operand a */
const struct calib_GFpkx_obj *	b		/* IN: operand b */
)
{
int				i, adeg, bdeg, maxdeg, deg;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpkx_obj *	ap;
struct calib_GFpkx_obj *	bp;
struct calib_GFpkx_obj *	rp;
struct calib_GFpkx_obj *	qp;
struct calib_GFpkx_obj *	m1;
struct calib_GFpkx_obj *	m2;
struct calib_GFpkx_obj *	m3;
struct calib_GFpkx_obj *	n1;
struct calib_GFpkx_obj *	n2;
struct calib_GFpkx_obj *	n3;
struct calib_GFpkx_obj *	TEMP;
struct calib_GFpkx_obj *	temp_ptr;
struct calib_GFpk_obj *		lca_inv;
struct calib_GFpk_obj *		cp;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR
			      (xa  EQ NULL) OR
			      (xb  EQ NULL) OR
			      (a   EQ NULL) OR
			      (b   EQ NULL));

	K_of_x	= gcd -> dom;

	CALIB_FATAL_ERROR_IF ((xa -> dom NE K_of_x) OR
			      (xb -> dom NE K_of_x) OR
			      (a  -> dom NE K_of_x) OR
			      (b  -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	adeg = a -> degree;
	bdeg = b -> degree;
	maxdeg = _MAX (adeg, bdeg);

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 11, maxdeg);

	temp_ptr = &TMP [0];
	ap = temp_ptr++;
	bp = temp_ptr++;
	rp = temp_ptr++;
	qp = temp_ptr++;
	m1 = temp_ptr++;
	m2 = temp_ptr++;
	m3 = temp_ptr++;
	n1 = temp_ptr++;
	n2 = temp_ptr++;
	n3 = temp_ptr++;
	TEMP = temp_ptr++;

	GFpkx_set (ap, a);
	GFpkx_set (bp, b);

	GFpkx_set_si (m1, 1);		/* ap = m1*a + n1*b */
	GFpkx_set_si (n1, 0);
	GFpkx_set_si (m2, 0);		/* bp = m2*a + n2*b */
	GFpkx_set_si (n2, 1);

	while (bp -> degree >= 0) {
		GFpkx_div (qp, rp, ap, bp);

		/* r = ap - q*bp			*/
		/*   = (m1*a + n1*b) - q*(m2*a + n2*b)	*/
		/*   = (m1 - q*m2)*a + (n1 - q*n2)*b	*/
		/*   = m3*a	     + m3*b,		*/
		/* so we compute:			*/
		/*	m3 = m1 - q*m2			*/
		/*	n3 = n1 - q*n2			*/
		GFpkx_mul (TEMP, qp, m2);
		GFpkx_sub (m3, m1, TEMP);
		GFpkx_mul (TEMP, qp, n2);
		GFpkx_sub (n3, n1, TEMP);

		temp_ptr = ap;
		ap = bp;
		bp = rp;
		rp = temp_ptr;

		temp_ptr = m1;
		m1 = m2;
		m2 = m3;
		m3 = temp_ptr;

		temp_ptr = n1;
		n1 = n2;
		n2 = n3;
		n3 = temp_ptr;
	}

	/* We have ap = M1*a + N1*b. */

	adeg = ap -> degree;
	if (adeg >= 0) {
		/* Monicize result. */
		lca_inv = &(qp -> coeff [0]);
		cf -> inv (lca_inv, &(ap -> coeff [adeg]));

		cp = &(ap -> coeff [0]);
		for (i = 0; i <= adeg; i++, cp++) {
			cf -> mul (cp, cp, lca_inv);
		}

		cp = &(m1 -> coeff [0]);
		deg = m1 -> degree;
		for (i = 0; i <= deg; i++, cp++) {
			cf -> mul (cp, cp, lca_inv);
		}

		cp = &(n1 -> coeff [0]);
		deg = n1 -> degree;
		for (i = 0; i <= deg; i++, cp++) {
			cf -> mul (cp, cp, lca_inv);
		}
	}

	GFpkx_set (gcd, ap);
	GFpkx_set (xa, m1);
	GFpkx_set (xb, n1);

	FREE_LOCAL_POLY_ARRAY (TMP, 11);
}

/*
 * Convert the given polynomial in Z(a)[x] into GF(p^k)[x] form.
 */

	static
	struct calib_GFpkx_obj *
GFpkx_cvZax (

const struct calib_GFpkx_dom *	K_of_x,	/* IN: dst polynomial ring */
const struct calib_Zax_obj *	op	/* IN: polynomial to convert */
)
{
int				i, j, k, deg, size;
const struct calib_Zx_obj *	gpoly;
const struct calib_Zax_dom *	Zax;
const struct calib_GFpk_dom *	gfpk;
const struct calib_Zp_dom *	Zp;
struct calib_GFpkx_obj *	new_poly;
mpz_srcptr			szp;
mpz_ptr				dzp;
struct calib_GFpk_obj *		coeff;
mpz_srcptr			p;
mpz_t				cval;

	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR (op EQ NULL));

	deg	= op -> degree;
	Zax	= op -> dom;

	/* Verify that the generator polynomial is identical in	*/
	/* both domains (equivalent mod p).			*/
	gfpk	= K_of_x -> cf;
	Zp	= gfpk -> cf;
	k	= gfpk -> k;
	gpoly	= Zax -> cr -> gpoly;
	CALIB_FATAL_ERROR_IF (k NE gpoly -> degree);
	p = Zp -> p;
	mpz_init (cval);
	for (i = 0; i <= k; i++) {
		mpz_sub (cval,
			 &(gpoly -> coeff [i]),
			 &(gfpk -> coeff [i]));
		mpz_mod (cval, cval, p);
		CALIB_FATAL_ERROR_IF (mpz_sgn (cval) NE 0);
	}
	mpz_clear (cval);

	size = deg + 1;
	coeff = NULL;
	if (size > 0) {
		coeff = GFpkx_new_GFpk_vector (gfpk, size);
	}

	for (i = 0; i <= deg; i++) {
		dzp = &coeff [i].coeff [0];
		szp = &(op -> coeff [i].coeff [0]);
		for (j = 0; j < k; j++, dzp++, szp++) {
			Zp -> set_z (Zp, dzp, szp);
		}
	}

	new_poly = CALIB_NEW (struct calib_GFpkx_obj);
	new_poly -> degree	= deg;
	new_poly -> size	= size;
	new_poly -> dom		= K_of_x;
	new_poly -> coeff	= coeff;

	return (new_poly);
}

/*
 * Factor the given polynomial in GF(p^k)[x] completely into its monic,
 * irreducible factors over GF(p^k)[x].  The correct multiplicity of each
 * factor is also determined.
 */

	static
	struct calib_GFpkx_factor *
GFpkx_factor (

const struct calib_GFpkx_obj *	poly	/* IN: polynomial to factor */
)
{
int				deg, multiplicity;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpkx_factor *	sqfactors;
struct calib_GFpkx_factor *	list;
struct calib_GFpkx_factor **	hookp;
struct calib_GFpkx_factor *	next_list;
struct calib_GFpkx_factor *	factors;
struct calib_GFpkx_obj *	sqfpoly;
struct calib_GFpkx_obj *	irred_poly;
struct calib_Random		random;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;

	deg = poly -> degree;

	if (deg <= 0) {
		if ((poly -> degree EQ 0) AND
		    (cf -> onep (&(poly -> coeff [0])))) {
			/* NULL factors ===> 1. */
			return (NULL);
		}
		/* Polynomial is constant. */
		list = CALIB_NEW (struct calib_GFpkx_factor);
		list -> multiplicity	= 1;
		list -> factor		= GFpkx_dup (poly);
		list -> next		= NULL;
		return (list);
	}

	calib_random_init (&random);

	/* Make a local copy of the polynomial and monicize it.	*/
	ALLOCATE_LOCAL_POLY (K_of_x, LPOLY, deg);
	GFpkx_set (&LPOLY, poly);
	GFpkx_monicize (&LPOLY);

	sqfactors = GFpkx_compute_square_free_factorization (&LPOLY);

	list = NULL;
	hookp = &list;

	while (sqfactors NE NULL) {
		/* Remove first square-free factor from the list. */
		multiplicity	= sqfactors -> multiplicity;
		sqfpoly		= sqfactors -> factor;
		next_list	= sqfactors -> next;
		sqfactors -> next = NULL;
		free (sqfactors);
		sqfactors = next_list;

		CALIB_FATAL_ERROR_IF ((sqfpoly EQ NULL) OR
				      (sqfpoly -> degree <= 0));

		/* Factor into irreducibles using distinct-degree	*/
		/* factorization, with randomized splitting of factors	*/
		/* within each degree.					*/
		factors = GFpkx_ddf (sqfpoly, &random);
		while (factors NE NULL) {
			irred_poly  = factors -> factor;
			CALIB_FATAL_ERROR_IF ((irred_poly EQ NULL) OR
					      (irred_poly -> degree <= 0));
			/* Set multiplicity and append to output list. */
			factors -> multiplicity = multiplicity;
			next_list = factors -> next;
			factors -> next = NULL;
			*hookp = factors;
			hookp = &(factors -> next);
			factors = next_list;
		}
		GFpkx_free (sqfpoly);
	}

	FREE_LOCAL_POLY (LPOLY);

	return (list);
}

/*
 * Factor the given polynomial into factors that are both square-free and
 * pairwise relatively prime, correctly determining the multiplicity of
 * each such factor.  This results in one non-trivial factor for each
 * distinct multiplicity present -- that factor being the product of all
 * factors having that that multiplicity.
 *
 * NOTE: the given polynomial is assumed to be monic, and may be
 *	 modified by this routine.
 */

	static
	struct calib_GFpkx_factor *
GFpkx_compute_square_free_factorization (

struct calib_GFpkx_obj *	poly	/* IN: polynomial to factor */
)
{
int				i, deg;
int				multiplicity, multiplicity_multiplier;
unsigned long			p;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpkx_factor *	list;
struct calib_GFpkx_factor *	fp;
struct calib_GFpkx_factor **	hookp;
struct calib_GFpkx_obj *	pp;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;

	list = NULL;
	hookp = &list;

	/* Given polynomial must be monic, and degree >= 1. */
	deg = poly -> degree;
	CALIB_FATAL_ERROR_IF (deg <= 0);
	CALIB_FATAL_ERROR_IF (NOT (cf -> onep (&(poly -> coeff [deg]))));

	/* Assume poly has the following factorization:			*/
	/*								*/
	/*	poly(x) = f1(x) * (f2(x))^2 * ... * (fp(x))^p * ...	*/
	/*								*/
	/* and let g = gcd(poly, poly').  Then g is like poly,		*/
	/* except that one copy of each factor "WHOSE MULTIPLICITY IS	*/
	/* *NOT* A MULTPLE OF p" is removed.				*/

	ALLOCATE_LOCAL_POLY (K_of_x, f, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, fprime, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, a, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, b, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, c, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, d, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, rem, deg - 1);

	/* Yun's algorithm with a GF(p^k)[x] "double twist", initialize. */
	multiplicity_multiplier = 1;

	GFpkx_set (&f, poly);

	GFpkx_derivative (&fprime, &f);

	while (f.degree > 0) {
		/* We have a new f to factor. */
		while (fprime.degree < 0) {
			/* Derivative is 0.  This implies that poly has	*/
			/* the form:					*/
			/*	poly(x) = g(x^p) = (h(x))^p.		*/
			/* If we were working in Zp[x], then we would	*/
			/* have g(x) = h(x), but it is not this easy	*/
			/* because we are instead working in GF(q)[x],	*/
			/* where q = p^k.  Suppose g(x) = \sum g_i x^i	*/
			/* and h(x) = \sum h_i x^i.  We need		*/
			/* g_i = (h_i)^p, or (g_i)^(1/p) = h_i, so we	*/
			/* need to be able to compute p-th roots of	*/
			/* elements of GF(p^k).				*/
			deg = f.degree;
			/* Note: we now know that p is sufficiently	*/
			/* small to divide a positive "int" degree.	*/
			p = mpz_get_ui (cf -> cf -> p);
			CALIB_FATAL_ERROR_IF ((deg % p) NE 0);
			deg /= p;
			for (i = 0; i <= deg; i++) {
				cf -> pth_root (&f.coeff [i],
						&f.coeff [i * p]);
			}
			f.degree = deg;
			multiplicity_multiplier *= p;

			GFpkx_derivative (&fprime, &f);
		}
		GFpkx_gcd (&a, &f, &fprime);

		GFpkx_div (&b, &rem, &f, &a);
		CALIB_FATAL_ERROR_IF (rem.degree >= 0);

		GFpkx_div (&c, &rem, &fprime, &a);
		CALIB_FATAL_ERROR_IF (rem.degree >= 0);

		GFpkx_derivative (&d, &b);
		GFpkx_sub (&d, &c, &d);

		for (;;) {
			/* Standard Yun's algorithm, iteration. */
			GFpkx_gcd (&a, &b, &d);

			if (a.degree > 0) {
				/* We have a non-trivial factor that is	*/
				/* also square-free.  Determine its	*/
				/* multiplicity and store it.		*/
				pp = GFpkx_dup (&a);
				GFpkx_monicize (pp);

				multiplicity = GFpkx_count_multiplicity (&f, pp);

				CALIB_FATAL_ERROR_IF (multiplicity <= 0);

				fp = CALIB_NEW (struct calib_GFpkx_factor);
				fp -> multiplicity	= multiplicity * multiplicity_multiplier;
				fp -> factor		= pp;
				fp -> next		= NULL;
				*hookp = fp;
				hookp = &(fp -> next);
			}
			GFpkx_div (&b, &rem, &b, &a);
			CALIB_FATAL_ERROR_IF (rem.degree >= 0);

			GFpkx_div (&c, &rem, &d, &a);
			CALIB_FATAL_ERROR_IF (rem.degree >= 0);

			GFpkx_derivative (&d, &b);
			GFpkx_sub (&d, &c, &d);

			if (b.degree <= 0) break;
		}

		/* We had better have b EQ 1. */
		CALIB_FATAL_ERROR_IF ((b.degree NE 0) OR
				      (NOT (cf -> onep (&b.coeff [0]))));

		if ((f.degree EQ 0) AND
		    (cf -> onep (&f.coeff [0]))) {
			/* All factors have been divided out of f.	*/
			/* We are done.					*/
			break;
		}

		/* Must have f' = 0, demonstrating that f has the form:	*/
		/*	f(x) = g(x^p) = (h(x))^p			*/
		GFpkx_derivative (&fprime, &f);
		CALIB_FATAL_ERROR_IF (fprime.degree >= 0);
	}

	FREE_LOCAL_POLY (rem);
	FREE_LOCAL_POLY (d);
	FREE_LOCAL_POLY (c);
	FREE_LOCAL_POLY (b);
	FREE_LOCAL_POLY (a);
	FREE_LOCAL_POLY (fprime);
	FREE_LOCAL_POLY (f);

	return (list);
}

/*
 * Compute the derivative of the given polynomial in GF(p^k)[x].
 */

	static
	void
GFpkx_derivative (

struct calib_GFpkx_obj *	deriv,	/* IN/OUT: derivative polynomial */
const struct calib_GFpkx_obj *	poly	/* IN: polynomial to differentiate */
)
{
int				i, deg, rsize;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
const struct calib_Zp_dom *	zpf;
struct calib_GFpk_obj *		result;
const struct calib_GFpk_obj *	pcoeff;
mpz_t				CBUF;

	CALIB_FATAL_ERROR_IF ((deriv EQ NULL) OR (poly EQ NULL));

	K_of_x	= poly -> dom;

	CALIB_FATAL_ERROR_IF (deriv -> dom NE K_of_x);

	cf	= K_of_x -> cf;
	zpf	= cf -> cf;

	deg	= poly -> degree;
	pcoeff	= poly -> coeff;

	if (deg <= 0) {
		deriv -> degree = -1;
		return;
	}

	rsize = deg;
	result = GFpkx_new_GFpk_vector (cf, rsize);
	mpz_init (CBUF);

	for (i = 1; i <= deg; i++) {
		zpf -> set_si (zpf, CBUF, i);
		cf -> mul_z (&result [(i - 1)],
			     &pcoeff [i],
			     CBUF);
	}

	do {
		--deg;
	} while ((deg >= 0) AND (cf -> zerop (&result [deg])));

	GFpkx_realloc_copy_poly (deriv, result, deg);

	mpz_clear (CBUF);
	GFpkx_free_GFpk_vector (result, rsize);
}

/*
 * Determine the multiplicity of the given factor of the given polynomial
 * by counting how many times it divides out evenly.
 */

	static
	int
GFpkx_count_multiplicity (

struct calib_GFpkx_obj *	poly,	/* IN: polynomial to divide */
const struct calib_GFpkx_obj *	factor	/* IN: factor to divide out */
)
{
int				pdeg, fdeg, multiplicity;
const struct calib_GFpkx_dom *	K_of_x;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (factor EQ NULL));

	K_of_x	= poly -> dom;

	CALIB_FATAL_ERROR_IF (factor -> dom NE K_of_x);

	pdeg	= poly -> degree;
	fdeg	= factor -> degree;

	ALLOCATE_LOCAL_POLY (K_of_x, quot, pdeg);
	ALLOCATE_LOCAL_POLY (K_of_x, rem,  fdeg);

	multiplicity = 0;
	for (;;) {
		GFpkx_div (&quot, &rem, poly, factor);
		if (rem.degree >= 0) {
			/* Remainder is not zero.  We are done. */
			break;
		}
		++multiplicity;
		GFpkx_set (poly, &quot);
	}

	FREE_LOCAL_POLY (rem);
	FREE_LOCAL_POLY (quot);

	return (multiplicity);
}

/*
 * Given a square-free polynomial over GF(p^k)[x], factor it completely into
 * irreducibles using the Distinct-Degree-Factorization method.  This
 * method exploits the fact that when q = p^k, the polynomial x^{q^d} - x
 * is the product of all irreducible polynomials whose degree divides d.
 *
 * To factor f(x), simply compute p(x) = ([x^s] mod f(x)), where s = q^d.
 * Then gcd(p(x) - x, f(x)) gives the product of all factors whose degree
 * k divides d.  If we do this for d = 1, 2, ... then we recover the
 * irreducible factors in increasing degree.
 */

	static
	struct calib_GFpkx_factor *
GFpkx_ddf (

const struct calib_GFpkx_obj *	sqfpoly,/* IN: square-free polynomial to */
					/*     factor */
struct calib_Random *		randp	/* IN/OUT: random number generator */
)
{
int				d, k, deg, fdeg;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpkx_factor *	list;
struct calib_GFpkx_factor **	hookp;
struct calib_GFpkx_factor *	fp;
struct calib_GFpkx_obj *	lpoly;
struct calib_GFpkx_obj *	reg1;
struct calib_GFpkx_obj *	reg2;
struct calib_GFpkx_obj *	f;
struct calib_GFpkx_obj *	temp_ptr;
mpz_t				Q;

	CALIB_FATAL_ERROR_IF (sqfpoly EQ NULL);

	K_of_x	= sqfpoly -> dom;
	cf	= K_of_x -> cf;
	k	= cf -> k;

	mpz_init_set (Q, cf -> cf -> p);
	mpz_pow_ui (Q, Q, k);

	list = NULL;
	hookp = &list;

	deg = sqfpoly -> degree;

	/* Assure that sqfpoly is monic. */
	CALIB_FATAL_ERROR_IF ((deg <= 0) OR
			      (NOT cf -> onep (&(sqfpoly -> coeff [deg]))));

	/* Allocate local polynomials. */
	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 4, deg);
	temp_ptr = &TMP [0];
	lpoly	= temp_ptr++;
	reg1	= temp_ptr++;
	reg2	= temp_ptr++;
	f	= temp_ptr++;

	/* Make a local copy of sqfpoly that we can modify. */
	GFpkx_set (lpoly, sqfpoly);

	ALLOCATE_LOCAL_POLY (K_of_x, xpoly, 1);
	GFpkx_set_var_power (&xpoly, 1);

	/* reg1 is initially x**(q**d), where d=0 (i.e., x**1). */
	GFpkx_set (reg1, &xpoly);

	for (d = 1; ; d++) {
		if (2 * d > lpoly -> degree) {
			/* Residual lpoly is irreducible. */
			break;
		}

		/* reg1 = (reg1**q) mod lpoly, so that	*/
		/* reg1 = x**(q**d) mod lpoly.		*/
		GFpkx_poly_to_int_pow_mod_poly (reg1,
						reg1,
						Q,
						lpoly);
		GFpkx_sub (reg2, reg1, &xpoly);
		GFpkx_gcd (f, reg2, lpoly);

		fdeg = f -> degree;
		if (fdeg > 0) {
			CALIB_FATAL_ERROR_IF (fdeg > lpoly -> degree);
			GFpkx_div (lpoly, NULL, lpoly, f);
			/* Further reduce reg1 by new lpoly. */
			GFpkx_div (NULL, reg1, reg1, lpoly);
			hookp = GFpkx_split_factors (hookp, f, d, randp);
		}
	}

	mpz_clear (Q);

	if (lpoly -> degree > 0) {
		/* Append final irreducible factor (highest degree). */
		f = GFpkx_dup (lpoly);
		GFpkx_monicize (f);

		fp = CALIB_NEW (struct calib_GFpkx_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= f;
		fp -> next		= NULL;
		*hookp = fp;
		hookp = &(fp -> next);
	}

	FREE_LOCAL_POLY (xpoly);
	FREE_LOCAL_POLY_ARRAY (TMP, 4);

	return (list);
}

/*
 * Given a polynomial f(x) \in GF(p^k)[x] whose irreducible factors all have
 * degree d, factor f(x) into these irreducible factors.  This is trivial
 * if degree(f) is 0 or d.  Otherwise we use a randomized algorithm to
 * split f(x) into irreducible factors.
 */

	static
	struct calib_GFpkx_factor **
GFpkx_split_factors (

struct calib_GFpkx_factor **	hookp,	/* IN: where to append next factor */
const struct calib_GFpkx_obj *	poly,	/* IN: polynomial f(x) to factor */
int				d,	/* IN: degree of irreducible factors */
struct calib_Random *		randp	/* IN: random number generator */
)
{
int				k, deg, fdeg, rdeg;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpkx_factor *	fp;
struct calib_GFpkx_obj *	pp;
mpz_t				R;

	CALIB_FATAL_ERROR_IF ((hookp EQ NULL) OR
			      (poly EQ NULL) OR
			      (d <= 0) OR
			      (randp EQ NULL));

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;
	k	= cf -> k;

	deg = poly -> degree;

	/* Degree must be a multiple of d! */
	CALIB_FATAL_ERROR_IF ((deg < 0) OR ((deg % d) NE 0));
	if (deg EQ 0) {
		/* Polynomial is constant -- nothing to split. */
		return (hookp);
	}
	if (deg EQ d) {
		/* We have a single irreducible factor! */
		pp = GFpkx_dup (poly);
		GFpkx_monicize (pp);

		fp = CALIB_NEW (struct calib_GFpkx_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= pp;
		fp -> next		= NULL;
		*hookp = fp;
		hookp = &(fp -> next);
		return (hookp);
	}

	/* f(x) has degree j*d, where j >= 2.			*/
	/* We require that p is an odd prime.  Let:		*/
	/*							*/
	/*	q = p**k,					*/
	/*	s = q**d,					*/
	/*	G(x)  = x**s - x				*/
	/*	r = (s-1)/2,					*/
	/*	H1(x) = x**r - 1				*/
	/*	H2(x) = x**r + 1				*/
	/*							*/
	/* Let R(x) be a random polynomial \in GF(p^k)[x] of	*/
	/* degree d.						*/
	/*							*/
	/* Since G(x) is the product of all monic, irreducible	*/
	/* polynomials whose degree divides d, then G(R(x))	*/
	/* has these same factors (and perhaps more).  It turns	*/
	/* out that H1(R(x)) has about half of these factors,	*/
	/* so that gcd(f(x), H1(R(x))) is a non-trivial factor	*/
	/* of f(x) with probability at least 1/2.		*/

	mpz_init (R);
	mpz_set (R, cf -> cf -> p);
	mpz_pow_ui (R, R, k);
	mpz_pow_ui (R, R, d);
	mpz_sub_ui (R, R, 1);
	mpz_tdiv_q_2exp (R, R, 1);	/* R <<= 1 */

	ALLOCATE_LOCAL_POLY (K_of_x, xpoly, 1);
	GFpkx_set_var_power (&xpoly, 1);

	ALLOCATE_LOCAL_POLY (K_of_x, rpoly, d);
	ALLOCATE_LOCAL_POLY (K_of_x, f,     deg);
	ALLOCATE_LOCAL_POLY (K_of_x, reg1,  deg);
	ALLOCATE_LOCAL_POLY (K_of_x, reg2,  deg);

	/* Make a local, modifiable copy of f(x). */
	GFpkx_set (&f, poly);

	fdeg = f.degree;
	while (fdeg > d) {
		GFpkx_set_random (&rpoly, d, randp);
		GFpkx_poly_to_int_pow_mod_poly (&reg1, &rpoly, R, &f);
		cf -> sub (&reg1.coeff [0],
			   &reg1.coeff [0],
			   &xpoly.coeff [1]);	/* This is 1 as GF(p^k). */
		if ((reg1.degree EQ 0) AND
		    (cf -> zerop (&reg1.coeff [0]))) {
			/* H1(R(x)) == 0 (mod f(x)), so this is an	*/
			/* unlucky (and very unusual) draw.		*/
			continue;
		}
		GFpkx_gcd (&reg2, &reg1, &f);
		rdeg = reg2.degree;
		if ((rdeg <= 0) OR (rdeg >= fdeg)) {
			/* We failed to split.  Try again. */
			continue;
		}

		/* We have a non-trivial split.  Recurse to either	*/
		/* save off the degree d factor, or split further.	*/
		hookp = GFpkx_split_factors (hookp, &reg2, d, randp);
		GFpkx_div (&f, NULL, &f, &reg2);
		fdeg = f.degree;
	}

	/* Should have one irredicible left.  Save it off. */
	CALIB_FATAL_ERROR_IF (fdeg NE d);
	hookp = GFpkx_split_factors (hookp, &f, d, randp);

	FREE_LOCAL_POLY (reg2);
	FREE_LOCAL_POLY (reg1);
	FREE_LOCAL_POLY (f);
	FREE_LOCAL_POLY (rpoly);
	FREE_LOCAL_POLY (xpoly);

	mpz_clear (R);

	return (hookp);
}

/*
 * Compute ((P(x)^K) mod M(x)), P(x), M(x) \in GF(p^k)[x], K is a possibly
 * very large non-negative integer.
 */

	static
	void
GFpkx_poly_to_int_pow_mod_poly (

struct calib_GFpkx_obj *	rop,	  /* IN/OUT: result */
const struct calib_GFpkx_obj *	poly,	  /* IN: the polynomial P(x) */
mpz_srcptr			exponent, /* IN: the power K */
const struct calib_GFpkx_obj *	modpoly	  /* IN: the modulus polynomial M(x) */
)
{
int				mdeg;
const struct calib_GFpkx_dom *	K_of_x;
struct calib_GFpkx_obj *	reg1;
struct calib_GFpkx_obj *	reg2;
mpz_t				POWER;

	CALIB_FATAL_ERROR_IF ((rop	EQ NULL) OR
			      (poly	EQ NULL) OR
			      (exponent	EQ NULL) OR
			      (modpoly	EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((poly    -> dom NE K_of_x) OR
			      (modpoly -> dom NE K_of_x));

	mdeg = modpoly -> degree;

	mpz_init_set (POWER, exponent);

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 2, mdeg - 1);
	reg1	= &TMP [0];		/* Successive squarings of poly. */
	reg2	= &TMP [1];		/* Result. */

	GFpkx_div (NULL, reg1, poly, modpoly);
	GFpkx_set_si (reg2, 1);

	while (mpz_sgn (POWER) > 0) {
		if (mpz_odd_p (POWER)) {
			GFpkx_poly_mul_mod_poly (reg2, reg2, reg1, modpoly);
		}
		GFpkx_poly_mul_mod_poly (reg1, reg1, reg1, modpoly);

		/* POWER >>= 1. */
		mpz_tdiv_q_2exp (POWER, POWER, 1);
	}

	GFpkx_set (rop, reg2);

	FREE_LOCAL_POLY_ARRAY (TMP, 2);
	mpz_clear (POWER);
}

/*
 * Compute ((A(x) * B(x)) mod M(x)), where A(x), B(x), M(x) \in GF(p^k)[x].
 */

	static
	void
GFpkx_poly_mul_mod_poly (

struct calib_GFpkx_obj *	rop,	/* IN/OUT: result */
const struct calib_GFpkx_obj *	a,	/* IN: the polynomial A(x) */
const struct calib_GFpkx_obj *	b,	/* IN: the polynomial B(x) */
const struct calib_GFpkx_obj *	modpoly	/* IN: the modulus polynomial M(x) */
)
{
int				i, j, adeg, mdeg, maxdeg;
int				r1deg, r2deg;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpkx_obj *	ap;
struct calib_GFpkx_obj *	bp;
struct calib_GFpkx_obj *	reg1;
struct calib_GFpkx_obj *	reg2;
struct calib_GFpkx_obj *	temp_ptr;
struct calib_GFpk_obj *		apcoeff;
struct calib_GFpk_obj *		bpcoeff;
struct calib_GFpk_obj *		r1coeff;
struct calib_GFpk_obj *		r2coeff;
struct calib_GFpk_obj *		modpcoeff;
struct calib_GFpk_obj		factor, shift_temp, bi;

	CALIB_FATAL_ERROR_IF ((rop	EQ NULL) OR
			      (a	EQ NULL) OR
			      (b	EQ NULL) OR
			      (modpoly	EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((a       -> dom NE K_of_x) OR
			      (b       -> dom NE K_of_x) OR
			      (modpoly -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	mdeg = modpoly -> degree;
	maxdeg = mdeg - 1;		/* max degree of something mod M(X). */

	/* Verify that modpoly is monic. */
	CALIB_FATAL_ERROR_IF ((mdeg < 1) OR
			      (NOT (cf -> onep (&(modpoly -> coeff [mdeg])))));

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 4, mdeg - 1);
	temp_ptr = &TMP [0];
	ap	= temp_ptr++;		/* A(x) mod M(x) */
	bp	= temp_ptr++;		/* B(x) mod M(x) */
	reg1	= temp_ptr++;		/* Result. */
	reg2	= temp_ptr++;		/* bi*A(x) */

	cf -> init (cf, &factor);
	cf -> init (cf, &shift_temp);
	cf -> init (cf, &bi);

	GFpkx_div (NULL, ap, a, modpoly);
	GFpkx_div (NULL, bp, b, modpoly);

	adeg = ap -> degree;

	reg1 -> degree	  = -1;

	apcoeff		= ap -> coeff;
	bpcoeff		= bp -> coeff;
	r1coeff		= reg1 -> coeff;
	r2coeff		= reg2 -> coeff;
	modpcoeff	= modpoly -> coeff;

	for (i = bp -> degree; i >= 0; i--) {
		/* Compute reg1 = ((reg1 * x**1) mod M(x)).	*/
		/* This just shifts the coefficients left,	*/
		/* mod-ing out those that fall off the end.	*/
		r1deg = reg1 -> degree;
		if (r1deg < maxdeg) {
			/* Simple shift suffices. */
			++r1deg;
			for (j = r1deg - 1; j >= 0; j--) {
				cf -> set (&r1coeff [j + 1], &r1coeff [j]);
			}
			cf -> set_si (&r1coeff [0], 0);
		}
		else {
			/* Need to mod out while shifting. */
			cf -> set (&factor, &r1coeff [r1deg]);
			for (j = r1deg; j > 0; j--) {
				cf -> mul (&shift_temp,
					   &factor,
					   &modpcoeff [j]);
				cf -> sub (&r1coeff [j],
					   &r1coeff [j - 1],
					   &shift_temp);
			}
			cf -> mul (&shift_temp,
				   &factor,
				   &modpcoeff [0]);
			cf -> neg (&r1coeff [0],
				   &shift_temp);
		}
		while ((r1deg >= 0) AND
		       (cf -> zerop (&r1coeff [r1deg]))) {
			--r1deg;
		}
		reg1 -> degree = r1deg;

		cf -> set (&bi, &bpcoeff [i]);
		for (j = 0; j <= adeg; j++) {
			cf -> mul (&r2coeff [j],
				   &bi,
				   &apcoeff [j]);
		}
		r2deg = adeg;
		while ((r2deg >= 0) AND
		       (cf -> zerop (&r2coeff [r2deg]))) {
			--r2deg;
		}
		reg2 -> degree = r2deg;
		GFpkx_add (reg1, reg1, reg2);
	}

	GFpkx_set (rop, reg1);

	cf -> clear (&bi);
	cf -> clear (&shift_temp);
	cf -> clear (&factor);
	FREE_LOCAL_POLY_ARRAY (TMP, 4);
}

/*
 * Modify the given polynomial to be monic.
 */

	static
	void
GFpkx_monicize (

struct calib_GFpkx_obj *	poly	/* IN/OUT: polynomial to monicize */
)
{
int				i, deg;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		lc;
struct calib_GFpk_obj *		coeff;
struct calib_GFpk_obj		INV;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	deg = poly -> degree;
	if (deg >= 0) {
		K_of_x	= poly -> dom;
		cf	= K_of_x -> cf;
		coeff = poly -> coeff;
		lc = &coeff [deg];
		if (NOT (cf -> onep (lc))) {
			cf -> init (cf, &INV);
			cf -> inv (&INV, lc);
			for (i = 0; i <= deg; i++) {
				cf -> mul (&coeff [i], &coeff [i], &INV);
			}
			cf -> clear (&INV);
		}
	}
}

/*
 * Free the given factorization, as produced by GFpkx_factor().
 */

	static
	void
GFpkx_free_factors (

struct calib_GFpkx_factor *	factors
)
{
struct calib_GFpkx_factor *	ftemp;

	while (factors NE NULL) {
		ftemp = factors -> next;
		GFpkx_free (factors -> factor);
		memset (factors, 0, sizeof (*factors));
		free (factors);
		factors = ftemp;
	}
}

/*
 * Choose a random polynomial \in GF(p^k)[x] of degree d.
 */

	static
	void
GFpkx_set_random (

struct calib_GFpkx_obj *	rop,	/* IN/OUT: random polynomial */
int				d,	/* IN: degree desired */
struct calib_Random *		randp	/* IN: random number generator */
)
{
int				i, rsize;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_GFpk_obj *		result;
struct calib_GFpk_obj *		lc;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (d <= 0) OR (randp EQ NULL));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;

	rsize = d + 1;
	result = GFpkx_new_GFpk_vector (cf, rsize);

	mpz_init (TEMP);

	for (i = 0; i < d; i++) {
		cf -> set_random (&result [i], randp);
	}

	/* Lead coefficient must not be zero.  Rare, but redraw if it	*/
	/* happens.							*/
	lc = &result [d];
	do {
		cf -> set_random (lc, randp);
	} while (cf -> zerop (lc));

	mpz_clear (TEMP);

	GFpkx_realloc_copy_poly (rop, result, d);
	GFpkx_free_GFpk_vector (result, rsize);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically zero.
 */

	static
	calib_bool
GFpkx_zerop (

const struct calib_GFpkx_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg >= 0) {
		return (FALSE);
	}

	CALIB_FATAL_ERROR_IF (deg < -1);

	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically one.
 */

	static
	calib_bool
GFpkx_onep (

const struct calib_GFpkx_obj *	op	/* IN: Polynomial to test */
)
{
int				deg;
const struct calib_GFpk_dom *	cf;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg NE 0) {
		return (FALSE);
	}

	cf = op -> dom -> cf;

	if (cf -> onep (&(op -> coeff [0])) EQ 0) {
		return (TRUE);
	}

	return (FALSE);
}

/*
 * Convert the given general expression into a member of GF(p^k)[x].
 * All of the variables in the given expression *MUST* match either the
 * given xvar, or avar.  (xvar is the "main" variable of the polynomial,
 * whereas avar is the variable for the coefficient field.)
 */

	static
	void
GFpkx_set_genrep (

struct calib_GFpkx_obj *	rop,	/* IN/OUT: result polynomial */
const struct calib_genrep *	op,	/* IN: expression to convert */
const char *			xvar,	/* IN: main var of polynomial */
const char *			avar	/* IN: var for coefficient field */
)
{
int				i;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_GFpkx_obj		POLY;
struct calib_GFpkx_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop    EQ NULL) OR
			      (op     EQ NULL) OR
			      (xvar   EQ NULL) OR
			      (avar   EQ NULL));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		GFpkx_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		GFpkx_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		GFpkx_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		if (strcmp (op -> u.var, xvar) EQ 0) {
			GFpkx_set_var_power (rop, 1);
		}
		else if (strcmp (op -> u.var, avar) EQ 0) {
			GFpkx_alloc (rop, 0);
			cf -> set_var_power (&(rop -> coeff [0]), 1);
		}
		else {
			CALIB_FATAL_ERROR;
		}
		break;

	case CALIB_GENREP_OP_ADD:
		GFpkx_init (K_of_x, &POLY);
		GFpkx_init (K_of_x, &TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			GFpkx_set_genrep (&TEMP,
					  lp -> operand,
					  xvar,
					  avar);
			GFpkx_add (&POLY, &POLY, &TEMP);
		}
		GFpkx_set (rop, &POLY);

		GFpkx_clear (&TEMP);
		GFpkx_clear (&POLY);
		break;

	case CALIB_GENREP_OP_MUL:
		GFpkx_init (K_of_x, &POLY);
		GFpkx_init (K_of_x, &TEMP);
		GFpkx_set_si (&POLY, 1);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			GFpkx_set_genrep (&TEMP,
					  lp -> operand,
					  xvar,
					  avar);
			GFpkx_mul (&POLY, &POLY, &TEMP);
		}
		GFpkx_set (rop, &POLY);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;
		CALIB_FATAL_ERROR_IF (i < 0);

		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);

		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			/* Optimize important special cases. */
			if (strcmp (op -> u.var, xvar) EQ 0) {
				GFpkx_set_var_power (rop, i);
			}
			else if (strcmp (op -> u.var, avar) EQ 0) {
				GFpkx_alloc (rop, 0);
				cf -> set_var_power (&(rop -> coeff [0]), i);
			}
			else {
				CALIB_FATAL_ERROR;
			}
		}
		else {
			GFpkx_init (K_of_x, &TEMP);
			GFpkx_set_genrep (&TEMP, gp, xvar, avar);
			GFpkx_ipow (rop, &TEMP, i);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Convert the given element of K[x] (where K = GF(p^k)) into
 * general representation.  The generated expression uses xvar for
 * the "main" polynomial variable, whereas avar is used for the
 * variable in the coefficient field K.
 */

	static
	struct calib_genrep *
GFpkx_to_genrep (

const struct calib_GFpkx_obj *	op,	/* IN: polynomial to convert */
const char *			xvar,	/* IN: main var of polynomial */
const char *			avar	/* IN: var for coefficient field */
)
{
int				i, degree;
const struct calib_GFpkx_dom *	K_of_x;
const struct calib_GFpk_dom *	cf;
const struct calib_GFpk_obj *	coeff;
const struct calib_GFpk_obj *	valp;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op     EQ NULL) OR
			      (xvar   EQ NULL) OR
			      (avar   EQ NULL));

	K_of_x	= op -> dom;
	cf	= K_of_x -> cf;

	degree	= op -> degree;
	coeff	= op -> coeff;

	root = NULL;
	hookp = &root;

	for (i = degree; i >= 0; i--) {
		valp = &coeff [i];
		if (cf -> zerop (valp)) continue;
		vgp = cf -> to_genrep (valp, avar);
		gp1 = calib_genrep_poly_term (vgp, xvar, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	return (gp1);
}

/*
 * Convert the given list of factors into general representation.
 * Use the variables given: xvar for the main variable, avar for
 * coefficient field.
 */

	static
	struct calib_genrep *
GFpkx_factors_to_genrep (

const struct calib_GFpkx_factor *
				factors,/* IN: list of factors */
const char *			xvar,	/* IN: polynomial main var */
const char *			avar	/* IN: var for coefficient field */
)
{
int					pow;
const struct calib_GFpkx_factor *	fp;
struct calib_genrep_list *		root;
struct calib_genrep_list **		hookp;
struct calib_genrep_list *		lp;
struct calib_genrep *			gp1;
struct calib_genrep *			gp2;

	root = NULL;
	hookp = &root;

	for (fp = factors; fp NE NULL; fp = fp -> next) {
		gp1 = GFpkx_to_genrep (fp -> factor, xvar, avar);
		pow = fp -> multiplicity;
		CALIB_FATAL_ERROR_IF (pow <= 0);
		if (pow > 1) {
			gp2 = CALIB_NEW (struct calib_genrep);
			gp2 -> op		= CALIB_GENREP_OP_IPOW;
			gp2 -> u.ipow.base	= gp1;
			gp2 -> u.ipow.power	= pow;
			gp1 = gp2;
		}
		lp = calib_genrep_new_list (gp1, NULL);
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (1);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_MUL;
		gp1 -> u.list	= root;
	}
	return (gp1);
}

/*
 * Prototype GFpkx domain.
 */

static const struct calib_GFpkx_dom	GFpkx_dom = {
	.init			= GFpkx_init,
	.init_degree		= GFpkx_init_degree,
	.alloc			= GFpkx_alloc,
	.clear			= GFpkx_clear,
	.set			= GFpkx_set,
	.set_si			= GFpkx_set_si,
	.set_z			= GFpkx_set_z,
	.set_q			= GFpkx_set_q,
	.set_var_power		= GFpkx_set_var_power,
	.add			= GFpkx_add,
	.sub			= GFpkx_sub,
	.neg			= GFpkx_neg,
	.mul			= GFpkx_mul,
	.mul_z			= GFpkx_mul_z,
	.ipow			= GFpkx_ipow,
	.dup			= GFpkx_dup,
	.free			= GFpkx_free,
	.eval			= GFpkx_eval,
	.div			= GFpkx_div,
	.gcd			= GFpkx_gcd,
	.extgcd			= GFpkx_extgcd,
	.cvZax			= GFpkx_cvZax,
	.factor			= GFpkx_factor,
	.free_factors		= GFpkx_free_factors,
	.zerop			= GFpkx_zerop,
	.onep			= GFpkx_onep,
	.set_random		= GFpkx_set_random,
	.set_genrep		= GFpkx_set_genrep,
	.to_genrep		= GFpkx_to_genrep,
	.factors_to_genrep	= GFpkx_factors_to_genrep,
};

/*
 * Construct the ring of polynomials over field K, where K is
 * the given field GF(p^k).
 */

	struct calib_GFpkx_dom *
calib_make_GFpkx_dom (

const struct calib_GFpk_dom *	K	/* IN: coefficient field */
)
{
struct calib_GFpkx_dom *	K_of_x;

	K_of_x = CALIB_NEW (struct calib_GFpkx_dom);

	/* Initialize "most" things. */
	*K_of_x = GFpkx_dom;

	K_of_x -> cf		= K;

	return (K_of_x);
}

/*
 * Free up the given GF(p^k)[x] ring description.
 */

	void
calib_free_GFpkx_dom (

struct calib_GFpkx_dom *	K_of_x	/* IN: GF(p^k)[x] ring description */
)
{
	if (K_of_x NE NULL) {
		/* No subordinate data to free. */
		memset (K_of_x, 0, sizeof (*K_of_x));
		free (K_of_x);
	}
}
