/**********************************************************************

	$Id: Qa.c,v 1.1 2024/10/20 00:36:25 warme Exp $

	File:	Qa.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The algebraic field Q(a), where a is an algebraic extension
	defined to be the root of a given irreducible polynomial
	g(a) \in Z[a].

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Qa.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/Qx.h"
#include "calib/Za.h"
#include "calib/Zpx.h"
#include "calib/Zx.h"
#include <gmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
 * Default settings object for newly-created Qa domains.
 *
 * API permits these to be changed, which affects future creation of
 * new Qa domains.
 */

struct calib_Qa_settings	calib_Qa_default_settings = {

	/* Default to PRS inverse method.  (Modular isn't ready	*/
	/* for actual use yet.)					*/
	.inv_method		= CALIB_QA_INV_METHOD_PRS,
};

/*
 * Local Functions
 */

static void	Qa_add (struct calib_Qa_obj *		rop,
			const struct calib_Qa_obj *	op1,
			const struct calib_Qa_obj *	op2);
static size_t	Qa_bit_size (const struct calib_Qa_obj *	op);
static void	Qa_canonicalize (struct calib_Qa_obj *	op);
static void	Qa_clear (struct calib_Qa_obj *		x);
static void	Qa_common_denominator (mpz_ptr				lcm,
				       mpz_ptr				mul1,
				       mpz_ptr				mul2,
				       const struct calib_Qa_obj *	op1,
				       const struct calib_Qa_obj *	op2);
static void	Qa_convert_Q_vector_to_Z (mpz_ptr	lcm,
					  mpz_ptr	zarray,
					  mpq_srcptr	qarray,
					  int		n);
static void	Qa_cv_modular_to_rational (mpq_ptr		qval,
					   mpz_srcptr		zval,
					   mpz_ptr		dlcm,
					   mpz_srcptr		modulus);
static int	Qa_degree (const struct calib_Qa_obj *	op);
static void	Qa_from_algint (struct calib_Qa_obj *		rop,
				const struct calib_Qa_obj *	op);
static mpq_ptr	Qa_get_coeffs (const struct calib_Qa_obj *	op);
static const struct calib_Za_dom *
		Qa_get_Za_dom (const struct calib_Qa_dom * f);
static void	Qa_init (const struct calib_Qa_dom *	f,
			 struct calib_Qa_obj *		x);
static void	Qa_inv (struct calib_Qa_obj *		rop,
			const struct calib_Qa_obj *	op);
static void	Qa_inv_modular (struct calib_Qa_obj *		rop,
				const struct calib_Qa_obj *	op);
static calib_bool
		Qa_inv_prime_acceptable (calib_int64u			p,
					 const struct calib_Zpx_dom *	Zpx,
					 const struct calib_Qa_obj *	op,
					 const struct calib_Zx_obj *	gpoly);
static void	Qa_inv_prs (struct calib_Qa_obj *	rop,
			    const struct calib_Qa_obj *	op);
static void	Qa_ipow (struct calib_Qa_obj *		rop,
			 const struct calib_Qa_obj *	op,
			 int				power);
static calib_bool
		Qa_is_algint (const struct calib_Qa_dom *	dom);
static void	Qa_mul (struct calib_Qa_obj *		rop,
			const struct calib_Qa_obj *	op1,
			const struct calib_Qa_obj *	op2);
static void	Qa_mul_q (struct calib_Qa_obj *	rop,
			  const struct calib_Qa_obj *	op1,
			  mpq_srcptr			op2);
static void	Qa_mul_si (struct calib_Qa_obj *	rop,
			   const struct calib_Qa_obj *	op1,
			   calib_si_t			op2);
static void	Qa_mul_z (struct calib_Qa_obj *	rop,
			  const struct calib_Qa_obj *	op1,
			  mpz_srcptr			op2);
static void	Qa_neg (struct calib_Qa_obj *		rop,
			const struct calib_Qa_obj *	op);
static calib_bool
		Qa_onep (const struct calib_Qa_obj *	op);
static calib_int64u
		Qa_previous_prime (calib_int64u p);
static void	Qa_set (struct calib_Qa_obj *		rop,
			const struct calib_Qa_obj *	op);
static void	Qa_set_coeffs (struct calib_Qa_obj *	rop,
			       int			degree,
			       mpq_srcptr		coeffs);
static void	Qa_set_genrep (struct calib_Qa_obj *		rop,
			       const struct calib_genrep *	op,
			       const char *			var);
static void	Qa_set_q (struct calib_Qa_obj *		rop,
			  mpq_srcptr			op);
static void	Qa_set_si (struct calib_Qa_obj *	rop,
			   calib_si_t			src);
static void	Qa_set_var_power (struct calib_Qa_obj *	rop,
				  int			power);
static void	Qa_set_z (struct calib_Qa_obj *	rop,
			   mpz_srcptr		src);
static void	Qa_set_Za_q (struct calib_Qa_obj *		rop,
			     const struct calib_Za_obj *	op1,
			     mpq_srcptr				op2);
static void	Qa_set_Zx (struct calib_Qa_obj *	rop,
			   const struct calib_Zx_obj *	op);
static void	Qa_sub (struct calib_Qa_obj *		rop,
			const struct calib_Qa_obj *	op1,
			const struct calib_Qa_obj *	op2);
static void	Qa_to_algint (struct calib_Qa_obj *		rop,
			      const struct calib_Qa_obj *	op);
static struct calib_genrep *
		Qa_to_genrep (const struct calib_Qa_obj *	op,
			      const char *			var);
static calib_bool
		Qa_zerop (const struct calib_Qa_obj *	op);

/* Standard "max" of two integers. */

#define _MAX(A,B)	(((A) > (B)) ? (A) : (B))

/*
 * Initialize the given x to hold a value from the given Q(a) field.
 */

	static
	void
Qa_init (

const struct calib_Qa_dom *	f,	/* IN: description of field */
struct calib_Qa_obj *		x	/* IN/OUT: value to initialize */
)
{
	CALIB_FATAL_ERROR_IF ((f EQ NULL) OR (x EQ NULL));

	mpq_init (x -> qfact);
	mpq_set_ui (x -> qfact, 1, 1);

	x -> dom	= f;
	x -> degree	= -1;
	x -> coeff	= calib_new_Z_vector (f -> gpoly -> degree);
}

/*
 * Clear the given Q(a) value object.
 */

	static
	void
Qa_clear (

struct calib_Qa_obj *		x	/* IN/OUT: value to clear */
)
{
const struct calib_Qa_dom *	f;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	mpq_clear (x -> qfact);

	f	= x -> dom;

	calib_free_Z_vector (x -> coeff, f -> gpoly -> degree);

	memset (x, 0, sizeof (*x));
}

/*
 * Set the destination Q(a) value from the source Q(a) value.
 */

	static
	void
Qa_set (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_Qa_obj *	op		/* IN: Qa value to set */
)
{
int				k;
const struct calib_Qa_dom *	f;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {
		f = rop -> dom;
		CALIB_FATAL_ERROR_IF (op -> dom NE f);

		mpq_set (rop -> qfact, op -> qfact);

		k = op -> degree;

		CALIB_FATAL_ERROR_IF (k >= f -> gpoly -> degree);

		rop -> degree	= k;
		calib_copy_Z_vector (rop -> coeff, op -> coeff, k + 1);
	}
}

/*
 * Set the destination Q(a) value from the source signed integer.
 */

	static
	void
Qa_set_si (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
calib_si_t			op	/* IN: integer to set */
)
{
	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	if (op EQ 0) {
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
	}
	else {
		mpq_set_si (rop -> qfact, op, 1);
		mpz_set_ui (&(rop -> coeff [0]), 1);
		rop -> degree	= 0;
	}
}

/*
 * Set the destination Q(a) value from the source GMP integer.
 */

	static
	void
Qa_set_z (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
mpz_srcptr			op	/* IN: GMP integer to convert */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpz_sgn (op) EQ 0) {
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
	}
	else {
		mpq_set_z (rop -> qfact, op);
		mpz_set_ui (&(rop -> coeff [0]), 1);
		rop -> degree	= 0;
	}
}

/*
 * Set the destination Q(a) value from the source GMP rational.
 */

	static
	void
Qa_set_q (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
mpq_srcptr			op	/* IN: GMP rational to set */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpq_sgn (op) EQ 0) {
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
	}
	else {
		mpq_set (rop -> qfact, op);
		mpz_set_ui (&(rop -> coeff [0]), 1);
		rop -> degree	= 0;
	}
}

/*
 * Set the given Q(a) value to a**power.
 */

	static
	void
Qa_set_var_power (

struct calib_Qa_obj *	rop,		/* IN/OUT: result */
int			power		/* IN: power to use */
)
{
int				i, k;
const struct calib_Qa_dom *	f;
struct calib_Qa_obj		TEMP;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	f	= rop -> dom;
	k	= f -> gpoly -> degree;

	CALIB_FATAL_ERROR_IF (k <= 1);

	if (power < 0) {
		Qa_init (f, &TEMP);
		Qa_set_var_power (&TEMP, 1);
		Qa_inv (&TEMP, &TEMP);
		Qa_ipow (&TEMP, &TEMP, -k);
		Qa_set (rop, &TEMP);
		Qa_clear (&TEMP);
	}
	else if (power < k) {
		/* Simple case. */
		mpq_set_ui (rop -> qfact, 1, 1);
		mpz_set_ui (&(rop -> coeff [power]), 1);
		for (i = power - 1; i >= 0; i--) {
			mpz_set_ui (&(rop -> coeff [i]), 0);
		}
		rop -> degree = power;
	}
	else {
		/* Must reduce result modulo generator polynomial. */
		Qa_init (f, &TEMP);
		Qa_set_var_power (&TEMP, 1);
		Qa_ipow (rop, &TEMP, power);
		Qa_clear (&TEMP);
	}
}

/*
 * Set rop = op1 * op2, where rop is a Qa value, op1 is a Za value,
 * and op2 is a GMP rational.  The generator polynomials for rop and
 * op1 must have the same degree (so that this conversion can be done
 * one coefficient at a time).
 */

	static
	void
Qa_set_Za_q (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
const struct calib_Za_obj *	op1,	/* IN: first operand */
mpq_srcptr			op2	/* IN: second operand */
)
{
int		k1, k2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	k1	= rop -> dom -> gpoly -> degree;
	k2	= op1 -> dom -> gpoly -> degree;

	CALIB_FATAL_ERROR_IF (k1 NE k2);
	CALIB_FATAL_ERROR_IF (k1 < 2);

	mpq_set (rop -> qfact, op2);
	rop -> degree = k1 - 1;
	calib_copy_Z_vector (rop -> coeff, op1 -> coeff, k1);

	Qa_canonicalize (rop);
}

/*
 * Set the given Q(a) object from the given Z[x] polynomial (the x
 * variables become a variables).
 */

	static
	void
Qa_set_Zx (

struct calib_Qa_obj *		rop,	/* IN/OUT: Resulting Q(a) value */
const struct calib_Zx_obj *	op	/* IN: Z[x] polynomial */
)
{
int				k1, k2;
const struct calib_Qa_dom *	dom;
const struct calib_Zx_obj *	gpoly;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;
	gpoly	= dom -> gpoly;
	k1	= gpoly -> degree;
	k2	= op -> degree;

	/* This routine does not yet support reduction modulo gpoly. */
	CALIB_FATAL_ERROR_IF (k2 >= k1);

	mpq_set_ui (rop -> qfact, 1, 1);
	rop -> degree	= k2;
	calib_copy_Z_vector (rop -> coeff, op -> coeff, k2 + 1);

	Qa_canonicalize (rop);
}

/*
 * Force the given Q(a) object into canonical form:
 *
 *	- Proper degree
 *	- Z[x] polynomial primitive
 *	- Leading coefficient (if any) is strictly positive
 */

	static
	void
Qa_canonicalize (

struct calib_Qa_obj *	op	/* IN/OUT: Q(a) object to canonicalize */
)
{
int		i, deg, sign;
mpz_srcptr	lc;
mpz_ptr		coeff;
mpz_t		GCD;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg	= op -> degree;
	coeff	= op -> coeff;

	while ((deg >= 0) AND (mpz_sgn (&coeff [deg]) EQ 0)) {
		--deg;
	}

	if (deg < 0) {
		CALIB_FATAL_ERROR_IF (deg < -1);
		mpq_set_ui (op -> qfact, 1, 1);
		op -> degree = -1;
	}
	else {
		op -> degree = deg;
		mpz_init (GCD);
		lc = &coeff [deg];
		mpz_abs (GCD, lc);
		for (i = deg - 1; i >= 0; i--) {
			mpz_gcd (GCD, GCD, &coeff [i]);
			if (mpz_cmp_ui (GCD, 1) EQ 0) break;
		}

		sign = mpz_sgn (lc);
		if (sign <= 0) {
			CALIB_FATAL_ERROR_IF (sign EQ 0);
			mpz_neg (GCD, GCD);
		}

		if (mpz_cmp_ui (GCD, 1) NE 0) {
			calib_mpq_mul_z (op -> qfact, op -> qfact, GCD);
			for (i = 0; i <= deg; i++) {
				mpz_divexact (&coeff [i], &coeff [i], GCD);
			}
		}

		mpz_clear (GCD);
	}
}

/*
 * Addition in Q(a).
 */

	static
	void
Qa_add (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
const struct calib_Qa_obj *	op1,	/* IN: first operand */
const struct calib_Qa_obj *	op2	/* IN: second operand */
)
{
int				i, k, rdeg, deg1, deg2, rsize, size1, size2;
mpz_ptr				rp, cbuf1, cbuf2, rcoeff;
const struct calib_Qa_dom *	f;
mpq_t				QFACT;
mpz_t				LCM, MUL1, MUL2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE f) OR (op2 -> dom NE f));

	k	= f -> gpoly -> degree;
	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		CALIB_FATAL_ERROR_IF (rdeg < -1);
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	CALIB_FATAL_ERROR_IF (rdeg >= k);

	rsize	= rdeg + 1;
	size1	= deg1 + 1;
	size2	= deg2 + 1;

	/* We make both buffers be full size, so that we don't have to	*/
	/* worry about one buffer being shorter than the other.		*/
	rp	= calib_new_Z_vector (2 * rsize);
	cbuf1	= rp;		rp += rsize;
	cbuf2	= rp;

	mpq_init (QFACT);
	mpz_init (LCM);
	mpz_init (MUL1);
	mpz_init (MUL2);

	Qa_common_denominator (LCM, MUL1, MUL2, op1, op2);

	mpq_set_ui (QFACT, 1, 1);
	calib_mpq_div_z (rop -> qfact, QFACT, LCM);

	calib_copy_Z_vector (cbuf1, op1 -> coeff, size1);
	calib_copy_Z_vector (cbuf2, op2 -> coeff, size2);

	for (i = 0; i < size1; i++) {
		mpz_mul (&cbuf1 [i], &cbuf1 [i], MUL1);
	}

	for (i = 0; i < size2; i++) {
		mpz_mul (&cbuf2 [i], &cbuf2 [i], MUL2);
	}

	rcoeff	= rop -> coeff;
	for (i = 0; i < rsize; i++) {
		mpz_add (&rcoeff [i], &cbuf1 [i], &cbuf2 [i]);
	}
	rop -> degree = rdeg;
	
	Qa_canonicalize (rop);

	calib_free_Z_vector (cbuf1, 2 * rsize);

	mpz_clear (MUL2);
	mpz_clear (MUL1);
	mpz_clear (LCM);
	mpq_clear (QFACT);
}

/*
 * Subtraction in Q(a).
 */

	static
	void
Qa_sub (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_Qa_obj *	op1,		/* IN: first operand */
const struct calib_Qa_obj *	op2		/* IN: second operand */
)
{
int				i, k, rdeg, deg1, deg2, rsize, size1, size2;
mpz_ptr				rp, cbuf1, cbuf2, rcoeff;
const struct calib_Qa_dom *	f;
mpq_t				QFACT;
mpz_t				LCM, MUL1, MUL2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE f) OR (op2 -> dom NE f));

	k	= f -> gpoly -> degree;
	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		CALIB_FATAL_ERROR_IF (rdeg < -1);
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	CALIB_FATAL_ERROR_IF (rdeg >= k);

	rsize	= rdeg + 1;
	size1	= deg1 + 1;
	size2	= deg2 + 1;

	/* We make both buffers be full size, so that we don't have to	*/
	/* worry about one buffer being shorter than the other.		*/
	rp	= calib_new_Z_vector (2 * rsize);
	cbuf1	= rp;		rp += rsize;
	cbuf2	= rp;

	mpq_init (QFACT);
	mpz_init (LCM);
	mpz_init (MUL1);
	mpz_init (MUL2);

	Qa_common_denominator (LCM, MUL1, MUL2, op1, op2);


	mpq_set_ui (QFACT, 1, 1);
	calib_mpq_div_z (rop -> qfact, QFACT, LCM);

	calib_copy_Z_vector (cbuf1, op1 -> coeff, size1);
	calib_copy_Z_vector (cbuf2, op2 -> coeff, size2);

	for (i = 0; i < size1; i++) {
		mpz_mul (&cbuf1 [i], &cbuf1 [i], MUL1);
	}

	for (i = 0; i < size2; i++) {
		mpz_mul (&cbuf2 [i], &cbuf2 [i], MUL2);
	}

	rcoeff	= rop -> coeff;
	for (i = 0; i < rsize; i++) {
		mpz_sub (&rcoeff [i], &cbuf1 [i], &cbuf2 [i]);
	}
	rop -> degree = rdeg;
	
	Qa_canonicalize (rop);

	calib_free_Z_vector (cbuf1, 2 * rsize);

	mpz_clear (MUL2);
	mpz_clear (MUL1);
	mpz_clear (LCM);
	mpq_clear (QFACT);
}

/*
 * Compute the common denominator for Q(a) polynomials op1 and op2,
 * together with the multipliers mul1 and mul2 needed to conform them.
 */

	static
	void
Qa_common_denominator (

mpz_ptr				lcm,	/* IN/OUT: LCM of denominators */
mpz_ptr				mul1,	/* IN/OUT: Multiplier for op1 */
mpz_ptr				mul2,	/* IN/OUT: Multiplier for op2 */
const struct calib_Qa_obj *	op1,	/* IN: First operand polynomial */
const struct calib_Qa_obj *	op2	/* IN: Second operand polynomial */
)
{
mpz_srcptr	n1, n2, d1, d2;
mpz_t		G, TEMP;

	CALIB_FATAL_ERROR_IF ((lcm  EQ NULL) OR
			      (mul1 EQ NULL) OR
			      (mul2 EQ NULL) OR
			      (op1 EQ NULL) OR
			      (op2 EQ NULL));

	n1	= mpq_numref (op1 -> qfact);
	n2	= mpq_numref (op2 -> qfact);

	d1	= mpq_denref (op1 -> qfact);
	d2	= mpq_denref (op2 -> qfact);

	mpz_init (G);
	mpz_init (TEMP);

	mpz_gcd (G, d1, d2);

	mpz_set (TEMP, d1);
	mpz_divexact (TEMP, TEMP, G);
	mpz_mul (lcm, TEMP, d2);

	mpz_divexact (TEMP, d2, G);
	mpz_mul (mul1, n1, TEMP);

	mpz_divexact (TEMP, d1, G);
	mpz_mul (mul2, n2, TEMP);

	mpz_clear (TEMP);
	mpz_clear (G);
}

/*
 * Negation in Q(a).
 */

	static
	void
Qa_neg (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_Qa_obj *	op		/* IN: operand */
)
{
int				k, deg;
const struct calib_Qa_dom *	f;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE f);

	k	= f -> gpoly -> degree;

	if (rop EQ op) {
		/* Negation in place. */
		deg	= rop -> degree;
		if (deg < 0) {
			CALIB_FATAL_ERROR_IF (deg < -1);
			return;
		}
		CALIB_FATAL_ERROR_IF (deg >= k);
		mpq_neg (rop -> qfact, rop -> qfact);
	}
	else {
		deg	= op -> degree;

		CALIB_FATAL_ERROR_IF ((deg < -1) OR (deg >= k));

		mpq_neg (rop -> qfact, op -> qfact);
		calib_copy_Z_vector (rop -> coeff, op -> coeff, deg + 1);
		rop -> degree = deg;
	}
}

/*
 * Multiplication in Q(a).
 */

	static
	void
Qa_mul (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
const struct calib_Qa_obj *	op1,	/* IN: first operand */
const struct calib_Qa_obj *	op2	/* IN: second operand */
)
{
int				k, deg1, deg2, rdeg;
const struct calib_Qa_dom *	f;
const struct calib_Zx_dom *	Zx;
const struct calib_Zx_obj *	zpoly;
mpq_t				QFACT;
mpz_t				DVAL;
struct calib_Zx_obj		RES, OP1, OP2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE f) OR (op2 -> dom NE f));

	zpoly	= f -> gpoly;
	k	= zpoly -> degree;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;

	Zx = calib_get_Zx_dom ();

	Zx -> init (&RES);
	mpq_init (QFACT);
	mpz_init (DVAL);

	mpq_mul (QFACT, op1 -> qfact, op2 -> qfact);

	/* Assemble "fake" Zx versions of op1 and op2 (used only in a	*/
	/* "read-only" fashion).					*/
	OP1.degree	= deg1;
	OP1.size	= deg1 + 1;
	OP1.coeff	= op1 -> coeff;

	OP2.degree	= deg2;
	OP2.size	= deg2 + 1;
	OP2.coeff	= op2 -> coeff;

	/* Multiply two polynomials in Z[x].  This is an O(n^2)		*/
	/* operation, but does not involve any integer GCD's, so the	*/
	/* arithmetic is quite fast.					*/
	Zx -> mul (&RES, &OP1, &OP2);

	/* Use pseudo-division in Z[x] to compute remainder when	*/
	/* dividing by zpoly (ignoring quotient).  Once again, this 	*/
	/* does not involve reducing any rational numbers to lowest	*/
	/* terms (integer GCD's), and is therefore reasonably fast.	*/
	Zx -> div (NULL, &RES, DVAL, &RES, zpoly);

	/* Fold 1/DVAL into rational multiplier. */
	calib_mpq_div_z (rop -> qfact, QFACT, DVAL);

	rdeg	= RES.degree;
	CALIB_FATAL_ERROR_IF (rdeg >= k);

	calib_copy_Z_vector (rop -> coeff, RES.coeff, rdeg + 1);
	rop -> degree = rdeg;

	Qa_canonicalize (rop);

	mpz_clear (DVAL);
	mpq_clear (QFACT);
	Zx -> clear (&RES);
}

/*
 * Multiply the given Q(a) value by the given signed integer.
 */

	static
	void
Qa_mul_si (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
const struct calib_Qa_obj *	op1,	/* IN: first operand */
calib_si_t			op2	/* IN: second operand */
)
{
int				deg;
const struct calib_Qa_dom *	f;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE f);

	deg	= op1 -> degree;

	if ((op2 EQ 0) OR (deg < 0)) {
		Qa_set_si (rop, 0);
		return;
	}

	Qa_set (rop, op1);
	calib_mpq_mul_si (rop -> qfact, op1 -> qfact, op2);
}

/*
 * Multiply the given Q(a) value by the given GMP integer.
 */

	static
	void
Qa_mul_z (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
const struct calib_Qa_obj *	op1,	/* IN: first operand */
mpz_srcptr			op2	/* IN: second operand */
)
{
int				deg;
const struct calib_Qa_dom *	f;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE f);

	deg	= op1 -> degree;

	if ((mpz_sgn (op2) EQ 0) OR (deg < 0)) {
		Qa_set_si (rop, 0);
		return;
	}

	Qa_set (rop, op1);
	calib_mpq_mul_z (rop -> qfact, op1 -> qfact, op2);
}

/*
 * Multiply the given Q(a) value by the given GMP rational.
 */

	static
	void
Qa_mul_q (

struct calib_Qa_obj *		rop,	/* IN/OUT: result */
const struct calib_Qa_obj *	op1,	/* IN: first operand */
mpq_srcptr			op2	/* IN: second operand */
)
{
int				deg;
const struct calib_Qa_dom *	f;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE f);

	deg	= op1 -> degree;

	if ((mpq_sgn (op2) EQ 0) OR (deg < 0)) {
		Qa_set_si (rop, 0);
		return;
	}

	Qa_set (rop, op1);
	mpq_mul (rop -> qfact, op1 -> qfact, op2);
}

/*
 * Take the given Q(a) value to the given integer power.
 */

	static
	void
Qa_ipow (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_Qa_obj *	op,		/* IN: value to exponentiate */
int				power		/* IN: integer power */
)
{
const struct calib_Qa_dom *	f;
struct calib_Qa_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE f);

	Qa_init (f, &RES);
	Qa_init (f, &PREG);
	Qa_set_si (&RES, 1);

	if (power >= 0) {
		Qa_set (&PREG, op);
	}
	else {
		Qa_inv (&PREG, op);
		power = - power;
	}

	while (power > 0) {
		if ((power & 1) NE 0) {
			Qa_mul (&RES, &RES, &PREG);
		}
		Qa_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	Qa_set (rop, &RES);

	Qa_clear (&PREG);
	Qa_clear (&RES);
}

/*
 * Multiplicative inverse in Q(a).
 */

	static
	void
Qa_inv (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_Qa_obj *	op		/* IN: operand to invert */
)
{
int				deg;
const struct calib_Qa_dom *	dom;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	deg	= op -> degree;

	/* Cannot divide by zero! */
	CALIB_FATAL_ERROR_IF (deg < 0);

	if (deg EQ 0) {
		Qa_set_si (rop, 1);
		mpq_inv (rop -> qfact, op -> qfact);
		return;
	}

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	switch (dom -> settings.inv_method) {
	case CALIB_QA_INV_METHOD_MODULAR:
		Qa_inv_modular (rop, op);
		break;

	case CALIB_QA_INV_METHOD_PRS:
		Qa_inv_prs (rop, op);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Multiplicative inverse in Q(a) using modular / Chinese Remainder Theorem
 * method.
 *
 * This code is not yet ready for "prime" time (so to speak):
 *
 * - We need upper bounds on the magnitude of coefficients in the solution
 *   so that we know when to stop increasing the modulus.  (Note: any such
 *   bound is likely to be *very* pessimistic, so a practical algorithm
 *   should perform rational reconstruction and test the resulting solution
 *   more often -- but not too often.)
 *
 * - We need to get the DLCM technique working correctly so that rational
 *   reconstruction does not take so long.
 *
 * - The biggest problem, however, is that we do not know what the solution
 *   looks like (up front), especially the coefficient denominators.  All
 *   it takes is for *one* of our primes to evenly divide *one* of these
 *   denominators -- making it impossible to reconstruct the correct
 *   rational solution from the modular solution.  The fact that we have to
 *   use so many different primes increases the probability of this
 *   happening.
 *
 * In light of this, it would probably be better to explore a p-adic
 * approach, which uses just a single prime.  If we have lifted the solution
 * to a sufficiently high power p^k and it doesn work, we can try again with
 * a different prime.
 */

#define	DEBUG_QA_INV_MODULAR	0

	static
	void
Qa_inv_modular (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_Qa_obj *	op		/* IN: operand to invert */
)
{
int				i, k, deg, deg2, poll_freq, poll_iter;
calib_int64u			p;
const struct calib_Qa_dom *	dom;
struct calib_Zp_dom *		Zp;
struct calib_Zpx_dom *		Zpx;
const struct calib_Zx_obj *	gpoly;
mpz_ptr				zvec;
mpq_ptr				qvec;
mpz_t				MODULUS, NEW_MODULUS, DLCM;
mpz_t				ZGCD, INV1, INV2, TEMP1, TEMP2;
mpq_t				FRAC;
struct calib_Zpx_obj		OP, GPOLY, G, A, B;
struct calib_Qa_obj		INVERSE, PROD;

#if DEBUG_QA_INV_MODULAR
struct calib_Qa_obj		TRUE_INVERSE;
struct calib_Zpx_obj		ZPX_INV, ZPX_TEMP;
#endif
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	gpoly	= dom -> gpoly;
	deg	= op -> degree;
	k	= gpoly -> degree;

	zvec	= calib_new_Z_vector (k);
	qvec	= calib_new_Q_vector (k);

	/* Make a (Zp, Zpx) domain pair. */
	Zp	= calib_make_Zp_dom_ui (3);
	Zpx	= calib_make_Zpx_dom (Zp);

	Zpx -> init (Zpx, &OP);
	Zpx -> init (Zpx, &GPOLY);
	Zpx -> init (Zpx, &G);
	Zpx -> init (Zpx, &A);
	Zpx -> init (Zpx, &B);
	dom -> init (dom, &INVERSE);
	dom -> init (dom, &PROD);

#if DEBUG_QA_INV_MODULAR
	/* This takes a lot of time, but gives the correct answer	*/
	/* against which we can compare					*/
	dom -> init (dom, &TRUE_INVERSE);
	Zpx -> init (Zpx, &ZPX_INV);
	Zpx -> init (Zpx, &ZPX_TEMP);

	Qa_inv_prs (&TRUE_INVERSE, op);
#endif

	Zpx -> alloc (&OP, deg);

	mpz_init (MODULUS);
	mpz_init (NEW_MODULUS);
	mpz_init (DLCM);
	mpz_init (ZGCD);
	mpz_init (INV1);
	mpz_init (INV2);
	mpz_init (TEMP1);
	mpz_init (TEMP2);
	mpq_init (FRAC);

	/* Initial solution is correct modulo 1. */
	mpz_set_ui (MODULUS, 1);

#if 1
	poll_freq = 1200;
#elif 0
	/* (For debugging quintic.dat.) */
	poll_freq = 1074;
#else
	poll_freq = 100;
#endif

	/* Start with largest 64-bit prime. */
	p	= 18446744073709551557lu;
	for (poll_iter = 0;; p = Qa_previous_prime (p)) {

		mpz_set_ui (Zp -> p, p);
		if (NOT Qa_inv_prime_acceptable (p, Zpx, op, gpoly)) {
			/* This prime unacceptable.  Get a new one. */
			continue;
		}

		/* Convert op and gpoly into Zp[x] form. */
		Zpx -> set_Qa (&OP, op);
		Zpx -> set_Zx (&GPOLY, gpoly);

		/* Solve A*op + B*gpoly = 1 (mod p). */
		Zpx -> extgcd (&G, &A, &B, &OP, &GPOLY);

		if (G.degree NE 0) {
			CALIB_FATAL_ERROR_IF (G.degree < 0);

			/* Apparently, op and gpoly have some modular	*/
			/* factor in common (mod p).  Cannot use this	*/
			/* modulus.					*/
			continue;
		}

		/* GCD had better be 1 (monic). */
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (&G.coeff [0], 1) NE 0);

#if DEBUG_QA_INV_MODULAR
		Zpx -> set_Qa (&ZPX_INV, &TRUE_INVERSE);
		Zpx -> sub (&ZPX_TEMP, &ZPX_INV, &A);
		if (ZPX_TEMP.degree >= 0) {
			printf ("Modular solution does not match.\n");
		}
#endif

		/* Use Chinese Remainder Theorem to combine inverse A	*/
		/* that from previous moduli.				*/
		mpz_gcdext (ZGCD, INV1, INV2, MODULUS, Zp -> p);

		/* The GCD must be 1! */
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (ZGCD, 1) NE 0);

		deg2 = A.degree;
		CALIB_FATAL_ERROR_IF (deg2 >= k);

		mpz_mul (NEW_MODULUS, MODULUS, Zp -> p);

		for (i = 0; i <= A.degree; i++) {
			mpz_mul (TEMP1, INV1, MODULUS);
			mpz_mul (TEMP2, INV2, Zp -> p);
			mpz_mul (TEMP1, TEMP1, &A.coeff [i]);
			mpz_mul (TEMP2, TEMP2, &zvec [i]);
			mpz_add (TEMP1, TEMP1, TEMP2);
			mpz_fdiv_r (&zvec [i], TEMP1, NEW_MODULUS);
		}
		mpz_set (MODULUS, NEW_MODULUS);

		++poll_iter;
		if (poll_iter >= poll_freq) {
			/* Do rational reconstruction. */
			mpz_set_ui (DLCM, 1);
			for (i = 0; i < k; i++) {
#if 0
				/* Disable DLCM trick for now. */
				mpz_set_ui (DLCM, 1);
#endif
				Qa_cv_modular_to_rational (&qvec [i],
							   &zvec [i],
							   DLCM,
							   MODULUS);
			}

			/* Convert qvec into Q(a) value. */
			Qa_set_coeffs (&INVERSE, k - 1, qvec);

			/* See if op * INVERSE = 1. */
			Qa_mul (&PROD, op, &INVERSE);

			if (Qa_onep (&PROD)) {
				Qa_set (rop, &INVERSE);
				break;
			}
			poll_iter = 0;
		}
	}

	mpq_clear (FRAC);
	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
	mpz_clear (INV2);
	mpz_clear (INV1);
	mpz_clear (ZGCD);
	mpz_clear (DLCM);
	mpz_clear (NEW_MODULUS);
	mpz_clear (MODULUS);

#if DEBUG_QA_INV_MODULAR
	dom -> clear (&TRUE_INVERSE);
	Zpx -> clear (&ZPX_INV);
	Zpx -> clear (&ZPX_TEMP);
#endif

	dom -> clear (&PROD);
	dom -> clear (&INVERSE);
	Zpx -> clear (&B);
	Zpx -> clear (&A);
	Zpx -> clear (&G);
	Zpx -> clear (&GPOLY);
	Zpx -> clear (&OP);
	calib_free_Zpx_dom (Zpx);
	calib_free_Zp_dom (Zp);
}

/*
 * Return true if-and-only-if the given prime is acceptable for use with
 * the modular multiplicative inverse algorithm.
 */

	static
	calib_bool
Qa_inv_prime_acceptable (

calib_int64u			p,		/* IN: Prime p to test */
const struct calib_Zpx_dom *	Zpx,		/* IN: Zp[x] domain */
const struct calib_Qa_obj *	op,		/* IN: Q(a) value to invert */
const struct calib_Zx_obj *	gpoly		/* IN: Generator poly for */
						/*     Q(a) domain */
)
{
calib_bool			is_sqf;
mpz_srcptr			d;
struct calib_Zpx_factor *	factors;
struct calib_Zpx_factor *	fp;
struct calib_Zpx_obj		POLY;

	CALIB_FATAL_ERROR_IF ((Zpx   EQ NULL) OR
			      (op    EQ NULL) OR
			      (gpoly EQ NULL));

	d = mpq_denref (op -> qfact);

	if (mpz_divisible_ui_p (d, p)) {
		/* This prime evenly divides one or more denominators	*/
		/* of the Q(a) value being inverted.  This would cause	*/
		/* division by zero and is therefore unacceptable.	*/
		return (FALSE);
	}

	/* See if gpoly is square-free modulo p. */
	Zpx -> init (Zpx, &POLY);

	Zpx -> set_Zx (&POLY, gpoly);
	factors = Zpx -> factor_square_free (&POLY);

	is_sqf = TRUE;
	for (fp = factors; fp NE NULL; fp = fp -> next) {
		if (fp -> multiplicity NE 1) {
			is_sqf = FALSE;
			break;
		}
	}
	Zpx -> free_factors (factors);

	Zpx -> clear (&POLY);

	if (NOT is_sqf) {
		/* The generator polynomial isn't square-free modulo p.	*/
		/* Can't use this modulus.				*/
		return (FALSE);
	}

	/* Prime p is acceptable! */
	return (TRUE);
}

/*
 * Given an odd integer, return the next-lowest prime number.
 */

	static
	calib_int64u
Qa_previous_prime (

calib_int64u	p		/* IN: Given integer p */
)
{
int		i, reps;
mpz_t		NUM;

	/* Must have an odd number. */
	CALIB_FATAL_ERROR_IF ((p & 1) EQ 0);

	mpz_init (NUM);

	reps = 30;

	for (;;) {
		CALIB_FATAL_ERROR_IF (p <= 3);

		p -= 2;

		mpz_set_ui (NUM, p);

		i = mpz_probab_prime_p (NUM, reps);
		if (i NE 0) break;
	}

	mpz_clear (NUM);

	return (p);
}

/*
 * Convert the given zval (a remainder modulo the given modulus) int its
 * equivalent rational number p/q.  The recipe works as follows.
 * Let HMOD = floor (modulus / 2).
 * We need to obtain the following modular equivalence:
 *
 *	q * zval == p (mod modulus)
 *
 * where ((p^2 < HMOD) AND (q^2 < HMOD)).  We accomplish this by
 * executing the Extended Euclidean Algorithm starting with A=modulus and
 * B=zval, and stopping just as soon as B^2 < HMOD.
 *
 * We use the "DLCM" trick to accelerate this process.  DLCM is the
 * least-common-multple of all previously converted denominators.  We
 * pre-multiply each entering zval by DLCM (mod modulus), giving us a
 * zval much more likely to satisfy the B^2 < HMOD stopping criteria.
 */

	static
	void
Qa_cv_modular_to_rational (

mpq_ptr			qval,		/* IN/OUT: Rational value p/q */
mpz_srcptr		zval,		/* IN: Integer residue to convert */
mpz_ptr			dlcm,		/* IN/OUT: DLCM value */
mpz_srcptr		modulus		/* IN: Modulus value */
)
{
int		sign;
mpz_t		A, B, Q, R, M1, M2, M3, N1, N2, N3, TEMP, HMOD;

	CALIB_FATAL_ERROR_IF ((qval    EQ NULL) OR
			      (zval    EQ NULL) OR
			      (dlcm    EQ NULL) OR
			      (modulus EQ NULL));

	sign = mpz_sgn (zval);
	if (sign EQ 0) {
		mpq_set_ui (qval, 0, 1);
		return;
	}

	mpz_init (A);
	mpz_init (B);
	mpz_init (Q);
	mpz_init (R);
	mpz_init (M1);
	mpz_init (M2);
	mpz_init (M3);
	mpz_init (N1);
	mpz_init (N2);
	mpz_init (N3);
	mpz_init (TEMP);
	mpz_init (HMOD);

	mpz_fdiv_q_2exp (HMOD, modulus, 1);

	mpz_set (A, modulus);
	mpz_set (B, zval);
	sign = +1;
	if (mpz_cmp (B, HMOD) > 0) {
		mpz_sub (B, modulus, B);
		sign = -1;
	}

	mpz_mul (B, B, dlcm);
	mpz_fdiv_r (B, B, modulus);

	mpz_set_ui (M1, 1);	mpz_set_ui (N1, 0);	/* A = 1*M + 0*zval */
	mpz_set_ui (M2, 0);	mpz_set_ui (N2, 1);	/* B = 0*M + 1*zval */

	while (mpz_sgn (B) > 0) {
		/* Invariant: B = M2*modulus + N2*|zval|. */
		mpz_mul (TEMP, B, B);
		if (mpz_cmp (TEMP, HMOD) < 0) {
			break;
		}

		mpz_fdiv_qr (Q, R, A, B);	/* R = A - Q*B */
		/* M3 = M1 - Q*M2;	*/
		mpz_mul (TEMP, Q, M2);
		mpz_sub (M3, M1, TEMP);
		/* N3 = N1 - Q*N2;	*/
		mpz_mul (TEMP, Q, N2);
		mpz_sub (N3, N1, TEMP);
		mpz_set (A, B);
		mpz_set (B, R);
		mpz_set (M1, M2);
		mpz_set (M2, M3);
		mpz_set (N1, N2);
		mpz_set (N2, N3);
	}

	/* We have B = M2*modulus + N2*|zval|, and B^2 < HMOD.  We	*/
	/* should also have N2^2 < HMOD, but we don't really care	*/
	/* too much or check this.  The important thing is that		*/
	/* |zval| == B/N2 (mod modulus).				*/

	mpz_set (mpq_numref (qval), B);
	mpz_mul (mpq_denref (qval), N2, dlcm);
	mpq_canonicalize (qval);
	if (sign < 0) {
		mpq_neg (qval, qval);
	}

	/* Update the dlcm value. */
	mpz_gcd (TEMP, dlcm, mpq_denref (qval));
	mpz_divexact (TEMP, mpq_denref (qval), TEMP);
	mpz_mul (TEMP, dlcm, TEMP);
	mpz_fdiv_r (dlcm, TEMP, modulus);

	mpz_clear (HMOD);
	mpz_clear (TEMP);
	mpz_clear (N3);
	mpz_clear (N2);
	mpz_clear (N1);
	mpz_clear (M3);
	mpz_clear (M2);
	mpz_clear (M1);
	mpz_clear (R);
	mpz_clear (Q);
	mpz_clear (B);
	mpz_clear (A);
}

/*
 * Multiplicative inverse in Q(a) using polynomial remainder sequence.
 */

	static
	void
Qa_inv_prs (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_Qa_obj *	op		/* IN: operand to invert */
)
{
int				k, deg, rdeg;
const struct calib_Qa_dom *	f;
const struct calib_Zx_dom *	Zx;
const struct calib_Zx_obj *	zpoly;
mpz_srcptr			denom;
struct calib_Zx_obj		AX, BX, GCD, OP;
mpq_t				QFACT;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	f	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE f);

	zpoly	= f -> gpoly;
	k	= zpoly -> degree;

	CALIB_FATAL_ERROR_IF (k < 2);

	deg	= op -> degree;

	CALIB_FATAL_ERROR_IF ((deg < 0) OR (deg >= k));

	Zx = calib_get_Zx_dom ();

	mpq_init (QFACT);

	Zx -> init (&AX);
	Zx -> init (&BX);
	Zx -> init (&GCD);

	/* Assemble a "fake" Zx version of op. */
	OP.degree	= deg;
	OP.size		= deg + 1;
	OP.coeff	= op -> coeff;

	/* Now use the Extended Euclidean Algorithm \in Z[x].  This	*/
	/* uses only purely integer arithmetic, does not require	*/
	/* reducing rational numbers to lowest form, and is therefore	*/
	/* reasonably fast.						*/
	Zx -> extgcd (&GCD, &AX, &BX, &OP, zpoly);

	/* This GCD *must* be a non-zero constant, strictly positive. */
	CALIB_FATAL_ERROR_IF (GCD.degree NE 0);

	denom = &GCD.coeff [0];
	CALIB_FATAL_ERROR_IF (mpz_sgn (denom) <= 0);

	/* Combine QFACT = (1 / op -> qfact) / denom. */
	mpq_inv (QFACT, op -> qfact);	
	calib_mpq_div_z (rop -> qfact, QFACT, denom);

	rdeg	= AX.degree;
	CALIB_FATAL_ERROR_IF ((rdeg < 0) OR (rdeg >= k));

	calib_copy_Z_vector (rop -> coeff, AX.coeff, rdeg + 1);
	rop -> degree = rdeg;

	Qa_canonicalize (rop);

	Zx -> clear (&GCD);
	Zx -> clear (&BX);
	Zx -> clear (&AX);

	mpq_clear (QFACT);
}

/*
 * Return TRUE if-and-only-if the given Qa domain is an algebraic integer
 * (i.e., the generator polynomial is monic).
 */

	static
	calib_bool
Qa_is_algint (

const struct calib_Qa_dom *	dom	/* IN: Qa domain to test */
)
{
int				deg;
const struct calib_Zx_obj *	gpoly;
mpz_srcptr			lc;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	gpoly	= dom -> gpoly;
	deg	= gpoly -> degree;
	lc	= &(gpoly -> coeff [deg]);

	CALIB_FATAL_ERROR_IF (deg < 2);

	if (mpz_cmp_ui (lc, 1) EQ 0) {
		/* gpoly is monic, so it defines an algebraic integer. */
		return (TRUE);
	}
	return (FALSE);
}

/*
 * For given Qa domain dom, return a Qa domain defining the corresponding
 * algebraic integer.  If dom's generator polynomial is monic, this will
 * always return dom.  Otherwise, we return a new Qa domain representing
 * the corresponding algebraic integer.
 * DO NOT FREE THIS DOMAIN, SINCE IT IS OWNED BY dom!
 */

	static
	const struct calib_Qa_dom *
Qa_algint_dom (

const struct calib_Qa_dom *	dom	/* IN: Qa domain for which to get */
					/*     algebraic integer domain */
)
{
int				i, deg;
const struct calib_Zx_dom *	Zx;
const struct calib_Zx_obj *	gpoly;
const struct calib_Qa_dom *	ai_dom;
struct calib_Qa_dom *		writable_dom;
mpz_srcptr			lc;
mpz_ptr				coeff;
mpz_t				D, DPOW;
struct calib_Zx_obj		GPOLY;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	if (dom -> is_algint (dom)) {
		/* Generator polynomial is monic.  This domain	*/
		/* already defines the algebraic integer.	*/
		return (dom);
	}

	/* See if we already have it. */
	ai_dom	= dom -> ai_dom;
	if (ai_dom NE NULL) {
		return (ai_dom);
	}

	/* Compute the algebraic integer polynomial.  Let g(a) be of	*/
	/* degree k, and d = lc(g(a)).  This is a^(k-1) * g(a/d).	*/
	gpoly	= dom -> gpoly;
	deg	= gpoly -> degree;
	lc	= &(gpoly -> coeff [deg]);

	CALIB_FATAL_ERROR_IF (deg < 2);

	Zx	= calib_get_Zx_dom ();

	mpz_init (D);
	mpz_init (DPOW);
	Zx -> init (&GPOLY);

	mpz_set (D, lc);
	mpz_set (DPOW, lc);
	Zx -> set (&GPOLY, gpoly);

	coeff	= GPOLY.coeff;

	mpz_set_ui (&coeff [deg], 1);

	/* deg-1 coefficient is unchanged. */

	for (i = deg - 2;; i--) {
		mpz_mul (&coeff [i], &coeff [i], DPOW);

		if (i <= 0) break;

		mpz_mul (DPOW, DPOW, D);
	}

	ai_dom = calib_make_Qa_dom (&GPOLY);

	/* Cache this polynomial for later use. */
	writable_dom	= (struct calib_Qa_dom *) dom;

	writable_dom -> ai_dom	= ai_dom;

	Zx -> clear (&GPOLY);
	mpz_clear (DPOW);
	mpz_clear (D);

	return (ai_dom);
}

/*
 * This operation translates a Q(a) value from domain B to domain A.
 * Two cases are permitted:
 *
 *	1. A = B, or
 *	2. A is the algebraic integer corresponding to B.
 *
 * The result (intepreted with respect to domain A) should have exactly the
 * same value as the source value (interpreted with respect to domain B).
 */

	static
	void
Qa_to_algint (

struct calib_Qa_obj *		rop,	/* IN/OUT: Resulting Q(a) value */
const struct calib_Qa_obj *	op	/* IN: Source Q(a) value */
)
{
int				i, k, deg;
const struct calib_Qa_dom *	A;
const struct calib_Qa_dom *	B;
const struct calib_Zx_obj *	gpoly;
mpq_ptr				scoeff, rcoeff, srcp, dstp;
mpz_t				D, DPOW, LCM;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	A	= rop -> dom;
	B	= op  -> dom;

	if (A EQ B) {
		/* No change of domain. */
		Qa_set (rop, op);
		return;
	}

	/* Verify that A is the algebraic integer of B. */
	CALIB_FATAL_ERROR_IF (A NE B -> ai_dom);

	gpoly	= B -> gpoly;
	k	= gpoly -> degree;

	CALIB_FATAL_ERROR_IF (k < 2);

	deg	= op -> degree;

	if (deg < 0) {
		CALIB_FATAL_ERROR_IF (deg < -1);

		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	CALIB_FATAL_ERROR_IF (deg >= k);

	/* Proceed to convert.  Let A = Q(alpha), B = Q(beta), g(x) the	*/
	/* generator polynomial of B, d = lc(g(x)), and f(beta) be the	*/
	/* polynomial representing 'op'.  The relationship is that	*/
	/* alpha/d = beta, so the result is just polynomial f(alpha/d).	*/

	mpz_init (D);
	mpz_init (DPOW);
	mpz_init (LCM);

	mpz_set (D, &(gpoly -> coeff [k]));
	mpz_set (DPOW, D);

	scoeff	= Qa_get_coeffs (op);

	rcoeff	= calib_new_Q_vector (deg + 1);

	srcp	= scoeff;
	dstp	= rcoeff;

	mpq_set (dstp, srcp);

	++srcp;
	++dstp;

	for (i = 1; i <= deg; i++, dstp++, srcp++) {
		calib_mpq_div_z (dstp, srcp, DPOW);
		mpz_mul (DPOW, DPOW, D);
	}

	Qa_convert_Q_vector_to_Z (LCM, rop -> coeff, rcoeff, deg + 1);
	rop -> degree = deg;

	mpq_set_ui (rop -> qfact, 1, 1);
	calib_mpq_div_z (rop -> qfact, rop -> qfact, LCM);

	Qa_canonicalize (rop);

	calib_free_Q_vector (rcoeff, deg + 1);

	mpz_clear (LCM);
	mpz_clear (DPOW);
	mpz_clear (D);
}

/*
 * This operation translates a Q(a) value from domain B to domain A.
 * Two cases are permitted:
 *
 *	1. A = B, or
 *	2. B is the algebraic integer corresponding to A.
 *
 * The result (intepreted with respect to domain A) should have exactly the
 * same value as the source value (interpreted with respect to domain B).
 */

	static
	void
Qa_from_algint (

struct calib_Qa_obj *		rop,	/* IN/OUT: Resulting Q(a) value */
const struct calib_Qa_obj *	op	/* IN: Source Q(a) value */
)
{
int				i, k, deg;
const struct calib_Qa_dom *	A;
const struct calib_Qa_dom *	B;
const struct calib_Zx_obj *	gpoly;
mpq_ptr				scoeff, rcoeff, srcp, dstp;
mpz_t				D, DPOW, LCM;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	A	= rop -> dom;
	B	= op  -> dom;

	if (A EQ B) {
		/* No change of domain. */
		Qa_set (rop, op);
		return;
	}

	/* Verify that B is the algebraic integer of A. */
	CALIB_FATAL_ERROR_IF (B NE A -> ai_dom);

	gpoly	= A -> gpoly;
	k	= gpoly -> degree;

	CALIB_FATAL_ERROR_IF (k < 2);

	deg	= op -> degree;

	if (deg < 0) {
		CALIB_FATAL_ERROR_IF (deg < -1);

		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	CALIB_FATAL_ERROR_IF (deg >= k);

	/* Proceed to convert.  Let A = Q(alpha), B = Q(beta), g(x) the	*/
	/* generator polynomial of A, d = lc(g(x)), and f(beta) be the	*/
	/* polynomial representing 'op'.  The relationship is that	*/
	/* alpha = d*beta, so the result is just polynomial f(d*alpha).	*/

	mpz_init (D);
	mpz_init (DPOW);
	mpz_init (LCM);

	mpz_set (D, &(gpoly -> coeff [k]));
	mpz_set (DPOW, D);

	scoeff	= Qa_get_coeffs (op);

	rcoeff	= calib_new_Q_vector (deg + 1);
	
	srcp	= scoeff;
	dstp	= rcoeff;

	mpq_set (dstp, srcp);

	++srcp;
	++dstp;

	for (i = 1; i <= deg; i++, dstp++, srcp++) {
		calib_mpq_mul_z (dstp, srcp, DPOW);
		mpz_mul (DPOW, DPOW, D);
	}

	Qa_convert_Q_vector_to_Z (LCM, rop -> coeff, rcoeff, deg + 1);
	rop -> degree = deg;

	mpq_set_ui (rop -> qfact, 1, 1);
	calib_mpq_div_z (rop -> qfact, rop -> qfact, LCM);

	Qa_canonicalize (rop);

	calib_free_Q_vector (rcoeff, deg + 1);

	mpz_clear (LCM);
	mpz_clear (DPOW);
	mpz_clear (D);
}

/*
 * Convert the given Q vector into the corresponding LCM (of denominators)
 * and Z vector.
 */

	static
	void
Qa_convert_Q_vector_to_Z (

mpz_ptr		lcm,		/* IN/OUT: Common denominator */
mpz_ptr		zarray,		/* IN/OUT: Vector of Z values */
mpq_srcptr	qarray,		/* IN: Vector of Q values */
int		n		/* IN: Length of vectors */
)
{
int		i;
mpz_srcptr	zp;
mpz_t		G, LCM, TEMP;

	CALIB_FATAL_ERROR_IF ((lcm    EQ NULL) OR
			      (zarray EQ NULL) OR
			      (qarray EQ NULL));

	mpz_init (G);
	mpz_init (LCM);
	mpz_init (TEMP);

	/* Compute least-common-multiple of the denominators. */
	mpz_set_ui (LCM, 1);

	for (i = 0; i < n; i++) {
		zp = mpq_denref (&qarray [i]);
		mpz_gcd (G, LCM, zp);
		mpz_divexact (TEMP, zp, G);
		mpz_mul (LCM, LCM, TEMP);
	}

	mpz_set (lcm, LCM);

	if (mpz_cmp_ui (LCM, 1) EQ 0) {
		/* All denominators are 1. */
		for (i = 0; i < n; i++) {
			mpz_set (&zarray [i], mpq_numref (&qarray [i]));
		}
	}
	else {
		for (i = 0; i < n; i++) {
			mpz_set (TEMP, LCM);
			mpz_divexact (TEMP, TEMP, mpq_denref (&qarray [i]));
			mpz_mul (&zarray [i], TEMP, mpq_numref (&qarray [i]));
		}
	}

	mpz_clear (TEMP);
	mpz_clear (LCM);
	mpz_clear (G);
}

/*
 * Return the degree of the given element of Q(a), i.e., the highest
 * non-zero power of the corresponding polynomial in the quotient
 * field Z[a]/g(a).  The degree of the zero element is -1.
 */

	static
	int
Qa_degree (

const struct calib_Qa_obj *	op	/* IN: field element to get degree */
					/*     of */
)
{
	CALIB_FATAL_ERROR_IF (op EQ NULL);

	return (op -> degree);
}

/*
 * Is the given element of Q(a) zero?
 */

	static
	calib_bool
Qa_zerop (

const struct calib_Qa_obj *	op	/* IN: field element to test */
)
{
int		deg;

	deg = Qa_degree (op);
	if (deg < 0) {
		return (TRUE);
	}
	return (FALSE);
}


/*
 * Is the given element of Q(a) one?
 */

	static
	calib_bool
Qa_onep (

const struct calib_Qa_obj *	op	/* IN: field element to test */
)
{
int				deg;
mpz_srcptr			lc;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg	= Qa_degree (op);
	if (deg EQ 0) {
		lc = &(op -> coeff [0]);
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (lc, 1) NE 0);
		if (mpq_cmp_ui (op -> qfact, 1, 1) EQ 0) {
			return (TRUE);
		}
	}
	return (FALSE);
}

/*
 * Convert the given general expression into a member of Q(a) = Z[a]/g(a).
 * All of the variables in the given expression *MUST* match the given
 * variable name.
 */

	static
	void
Qa_set_genrep (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_genrep *	op,		/* IN: expression to convert */
const char *			var		/* IN: var for polynomial */
)
{
const struct calib_Qa_dom *	f;
unsigned int			pow;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_Qa_obj		TEMP1, TEMP2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (var EQ NULL));

	f	= rop -> dom;

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Qa_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Qa_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		Qa_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		CALIB_FATAL_ERROR_IF (strcmp (op -> u.var, var) NE 0);
		Qa_set_var_power (rop, 1);
		break;

	case CALIB_GENREP_OP_ADD:
		Qa_init (f, &TEMP1);
		Qa_init (f, &TEMP2);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Qa_set_genrep (&TEMP2, lp -> operand, var);
			Qa_add (&TEMP1, &TEMP1, &TEMP2);
		}
		Qa_set (rop, &TEMP1);
		Qa_clear (&TEMP2);
		Qa_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_MUL:
		Qa_init (f, &TEMP1);
		Qa_init (f, &TEMP2);
		Qa_set_si (&TEMP1, 1);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Qa_set_genrep (&TEMP2, lp -> operand, var);
			Qa_mul (&TEMP1, &TEMP1, &TEMP2);
		}
		Qa_set (rop, &TEMP1);
		Qa_clear (&TEMP2);
		Qa_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_IPOW:
		pow = op -> u.ipow.power;

		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);

		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			/* Optimize important special case. */
			CALIB_FATAL_ERROR_IF (strcmp (gp -> u.var, var) NE 0);
			Qa_set_var_power (rop, pow);
		}
		else {
			Qa_init (f, &TEMP2);
			Qa_set_genrep (&TEMP2, gp, var);
			Qa_ipow (rop, &TEMP2, pow);
			Qa_clear (&TEMP2);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
	}
}

/*
 * Convert the given element of Q(a) = Z[a]/g(a) into
 * general representation.  Use the variable name given.
 */

	static
	struct calib_genrep *
Qa_to_genrep (

const struct calib_Qa_obj *	op,	/* IN: element to convert */
const char *			var	/* IN: variable name to use */
)
{
int				i, k, deg;
const struct calib_Qa_dom *	f;
mpq_ptr				coeff;
mpq_srcptr			value;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (var EQ NULL));

	deg	= op -> degree;
	f	= op -> dom;
	k	= f -> gpoly -> degree;

	CALIB_FATAL_ERROR_IF (k < 2);
	CALIB_FATAL_ERROR_IF ((deg < -1) OR (deg >= k));

	/* Get array of fully rational coefficients. */
	coeff	= Qa_get_coeffs (op);

	root = NULL;
	hookp = &root;

	for (i = deg; i >= 0; i--) {
		value = &coeff [i];
		if (mpq_sgn (value) EQ 0) continue;
		vgp = calib_genrep_q (value);
		gp1 = calib_genrep_poly_term (vgp, var, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	calib_free_Q_vector (coeff, deg + 1);

	return (gp1);
}

/*
 * Compute the bit-size of the given Q(a) value.
 */

	static
	size_t
Qa_bit_size (

const struct calib_Qa_obj *	op	/* IN: Q(a) value to get size of */
)
{
int		deg;
mpq_ptr		coeffs;
size_t		nbits;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg	= op -> degree;
	coeffs	= Qa_get_coeffs (op);

	nbits = calib_mpq_bit_size_vector (coeffs, deg + 1);

	calib_free_Q_vector (coeffs, deg + 1);

	return (nbits);
}

/*
 * Return the Z(a) version of this Q(a) domain.
 *
 * NOTE: Do *NOT* free this domain, because it is owned by the Q(a) domain!
 */

	static
	const struct calib_Za_dom *
Qa_get_Za_dom (

const struct calib_Qa_dom *	f	/* IN: Given Q(a) domain */
)
{
struct calib_Za_dom *		dom;
const struct calib_Zx_dom *	Zx;
struct calib_Qa_dom *		writable_f;
struct calib_Zx_obj *		zpoly;

	CALIB_FATAL_ERROR_IF (f EQ NULL);

	dom = f -> Za_dom;
	if (dom EQ NULL) {
		/* Dynamically create the corresponding Z(a) domain. */
		Zx = calib_get_Zx_dom ();
		zpoly = Zx -> dup (f -> gpoly);

		dom = calib_make_Za_dom (zpoly);

		/* We must force write access to f -> Za_dom here! */
		writable_f = (struct calib_Qa_dom *) f;
		writable_f -> Za_dom = dom;
	}

	return (dom);
}

/*
 * Return a dynamically allocated array of the rational coefficients
 * of this Q(a) value.  (The caller must free it.)
 */

	static
	mpq_ptr
Qa_get_coeffs (

const struct calib_Qa_obj *	op	/* IN: Q(a) value for which to	*/
					/*     get the coefficients */
)
{
int		i, deg;
mpz_srcptr	scoeff;
mpq_srcptr	qfact;
mpq_ptr		dcoeff;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	qfact	= op -> qfact;
	deg	= op -> degree;
	scoeff	= op -> coeff;

	CALIB_FATAL_ERROR_IF (deg < -1);

	dcoeff	= calib_new_Q_vector (deg + 1);

	for (i = 0; i <= deg; i++) {
		calib_mpq_mul_z (&dcoeff [i], qfact, &scoeff [i]);
	}

	return (dcoeff);
}

/*
 * Set Q(a) value rop from the given vector of rational coefficients,
 * having specified degree.  (The coeffs array has degree+1 elements.)
 */

	static
	void
Qa_set_coeffs (

struct calib_Qa_obj *		rop,	/* IN/OUT: Resulting Q(a) value */
int				degree,	/* IN: Degree of given coefficients */
mpq_srcptr			coeffs	/* IN: Rational coefficients */
)
{
int				i, k, deg2;
const struct calib_Qa_dom *	dom;
const struct calib_Zx_dom *	Zx;
const struct calib_Zx_obj *	gpoly;
mpz_ptr				zvec;
mpq_t				QFACT;
mpz_t				G, LCM, TEMP, DVAL;
struct calib_Zx_obj		ZXPOLY;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	dom	= rop -> dom;
	gpoly	= dom -> gpoly;
	k	= gpoly -> degree;

	while ((degree >= 0) AND (mpq_sgn (&coeffs [degree]) EQ 0)) {
		--degree;
	}

	if (degree < 0) {
		CALIB_FATAL_ERROR_IF (degree < -1);
		Qa_set_si (rop, 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (coeffs EQ NULL);

	mpq_init (QFACT);
	mpz_init (G);
	mpz_init (LCM);
	mpz_init (TEMP);
	mpz_init (DVAL);

	/* Compute common denominator. */
	mpz_set_ui (LCM, 1);
	for (i = 0; i <= degree; i++) {
		mpz_gcd (G, LCM, mpq_denref (&coeffs [i]));
		mpz_divexact (TEMP, mpq_denref (&coeffs [i]), G);
		mpz_abs (TEMP, TEMP);
		mpz_mul (LCM, LCM, TEMP);
	}

	mpz_set_ui (mpq_numref (QFACT), 1);
	mpz_set    (mpq_denref (QFACT), LCM);

	deg2	= degree;

	/* Clear denominators. */
	zvec	= calib_new_Z_vector (deg2 + 1);
	for (i = 0; i <= degree; i++) {
		mpz_divexact (TEMP, LCM, mpq_denref (&coeffs [i]));
		mpz_mul (&zvec [i], TEMP, mpq_numref (&coeffs [i]));
	}

	/* Remove integer content. */
	mpz_abs (G, &zvec [0]);
	for (i = 1; i <= degree; i++) {
		mpz_gcd (G, G, &zvec [i]);
		if (mpz_cmp_ui (G, 1) EQ 0) break;
	}
	if (mpz_cmp_ui (G, 1) NE 0) {
		if (mpz_sgn (&zvec [degree]) < 0) {
			mpz_neg (G, G);
		}
		mpz_set (mpq_numref (QFACT), G);
		for (i = 0; i <= degree; i++) {
			mpz_divexact (&zvec [i], &zvec [i], G);
		}
	}
	mpq_canonicalize (QFACT);

	if (degree >= k) {
		ZXPOLY.degree	= degree;
		ZXPOLY.size	= degree + 1;
		ZXPOLY.coeff	= zvec;
		Zx = calib_get_Zx_dom ();
		Zx -> div (NULL, &ZXPOLY, DVAL, &ZXPOLY, gpoly);

		degree = ZXPOLY.degree;
		CALIB_FATAL_ERROR_IF (degree >= k);

		mpz_mul (mpq_denref (QFACT), mpq_denref (QFACT), DVAL);
		mpq_canonicalize (QFACT);

		/* Remove any new integer content. */
		mpz_abs (G, &zvec [0]);
		for (i = 1; i <= degree; i++) {
			mpz_gcd (G, G, &zvec [i]);
			if (mpz_cmp_ui (G, 1) EQ 0) break;
		}
		if (mpz_cmp_ui (G, 1) NE 0) {
			if (mpz_sgn (&zvec [degree]) < 0) {
				mpz_neg (G, G);
			}
			mpz_mul (mpq_numref (QFACT), mpq_numref (QFACT), G);
			mpq_canonicalize (QFACT);
			for (i = 0; i <= degree; i++) {
				mpz_divexact (&zvec [i], &zvec [i], G);
			}
		}
	}

	mpq_set (rop -> qfact, QFACT);
	calib_copy_Z_vector (rop -> coeff, zvec, degree + 1);
	rop -> degree	= degree;

	calib_free_Z_vector (zvec, deg2 + 1);

	mpz_clear (DVAL);
	mpz_clear (TEMP);
	mpz_clear (LCM);
	mpz_clear (G);
	mpq_clear (QFACT);
}

/*
 * Prototype Qa domain object.
 */

static const struct calib_Qa_dom	Qa_dom = {
	.init		= Qa_init,
	.clear		= Qa_clear,
	.set		= Qa_set,
	.set_si		= Qa_set_si,
	.set_z		= Qa_set_z,
	.set_q		= Qa_set_q,
	.set_var_power	= Qa_set_var_power,
	.set_Za_q	= Qa_set_Za_q,
	.set_Zx		= Qa_set_Zx,
	.add		= Qa_add,
	.sub		= Qa_sub,
	.neg		= Qa_neg,
	.mul		= Qa_mul,
	.mul_si		= Qa_mul_si,
	.mul_z		= Qa_mul_z,
	.mul_q		= Qa_mul_q,
	.ipow		= Qa_ipow,
	.inv		= Qa_inv,
	.is_algint	= Qa_is_algint,
	.algint_dom	= Qa_algint_dom,
	.to_algint	= Qa_to_algint,
	.from_algint	= Qa_from_algint,
	.degree		= Qa_degree,
	.zerop		= Qa_zerop,
	.onep		= Qa_onep,
	.set_genrep	= Qa_set_genrep,
	.to_genrep	= Qa_to_genrep,
	.bit_size	= Qa_bit_size,
	.get_coeffs	= Qa_get_coeffs,
	.get_Za_dom	= Qa_get_Za_dom,
};

/*
 * Construct the field Q(a) = Z[a]/g(a), where g(a) is the given
 * irreducible generator polynomial in Z[x].
 */

	struct calib_Qa_dom *
calib_make_Qa_dom (

const struct calib_Zx_obj *	gpoly		/* IN: generator polynomial */
)
{
int				k, sign;
mpz_srcptr			lc;
struct calib_Qa_dom *		f;
const struct calib_Zx_dom *	Zx;

	/* Verify that the given generator polynomial has degree >= 2.	*/
	/* If the degree is less than 2, then remainders have degree at	*/
	/* most 0 and there is no representation for the var "a"!	*/
	k	= gpoly -> degree;
	CALIB_FATAL_ERROR_IF (k < 2);

	Zx = calib_get_Zx_dom ();

	f = CALIB_NEW (struct calib_Qa_dom);

	*f = Qa_dom;

	/* Make our local copy of gpoly. */
	f -> gpoly		= Zx -> dup (gpoly);

	lc = &(f -> gpoly -> coeff [k]);
	sign = mpz_sgn (lc);
	if (sign <= 0) {
		CALIB_FATAL_ERROR_IF (sign EQ 0);
		Zx -> neg (f -> gpoly, f -> gpoly);
	}

	f -> settings	= calib_Qa_default_settings;

	return (f);
}

/*
 * Free up the given Q(a) field description.
 */

	void
calib_free_Qa_dom (

struct calib_Qa_dom *	f	/* IN: field description to free */
)
{
const struct calib_Zx_dom *	Zx;

	if (f NE NULL) {
		Zx = calib_get_Zx_dom ();
		Zx -> free (f -> gpoly);
		calib_free_Za_dom (f -> Za_dom);
		memset (f, 0, sizeof (*f));
		free (f);
	}
}
