/**********************************************************************

	$Id: Qx.c,v 1.1 2024/10/20 00:36:26 warme Exp $

	File:	Qx.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The ring Q[x] of polynomials over the rationals Q.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Qx.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/random.h"
#include "calib/Zx.h"
#include <gmp.h>
#include <stdlib.h>
#include <string.h>


static void	Qx_add (struct calib_Qx_obj *		rop,
			const struct calib_Qx_obj *	op1,
			const struct calib_Qx_obj *	op2);
static void	Qx_alloc (struct calib_Qx_obj *	rop,
			  int			degree);
static void	Qx_clear (struct calib_Qx_obj *	x);
static void	Qx_clear_array (struct calib_Qx_obj *	parray,
				int			npoly);
static void	Qx_common_denominator (mpz_ptr				lcm,
				       mpz_ptr				mul1,
				       mpz_ptr				mul2,
				       const struct calib_Qx_obj *	op1,
				       const struct calib_Qx_obj *	op2);
static void	Qx_convert_Q_vector_to_Z (mpz_ptr	lcm,
					  mpz_ptr	zarray,
					  mpq_srcptr	qarray,
					  int		n);
static void	Qx_derivative (struct calib_Qx_obj *		rop,
			       const struct calib_Qx_obj *	op);
static void	Qx_div (struct calib_Qx_obj *		quotient,
			struct calib_Qx_obj *		remainder,
			const struct calib_Qx_obj *	a,
			const struct calib_Qx_obj *	b);
static struct calib_Qx_obj *
		Qx_dup (const struct calib_Qx_obj *	op);
static void	Qx_eval (mpq_ptr			rop,
			 const struct calib_Qx_obj *	poly,
			 mpq_srcptr			value);
static void	Qx_extgcd (struct calib_Qx_obj *	gcd,
			   struct calib_Qx_obj *	xa,
			   struct calib_Qx_obj *	xb,
			   const struct calib_Qx_obj *	a,
			   const struct calib_Qx_obj *	b);
static struct calib_Qx_factor *
		Qx_factor (const struct calib_Qx_obj *	poly);
static struct calib_genrep *
		Qx_factors_to_genrep (
				const struct calib_Qx_factor *	factors,
				const char *			var);
static void	Qx_free (struct calib_Qx_obj *	poly);
static void	Qx_free_factors (struct calib_Qx_factor *	factors);
static void	Qx_gcd (struct calib_Qx_obj *		gcd,
			const struct calib_Qx_obj *	a,
			const struct calib_Qx_obj *	b);
static mpq_ptr	Qx_get_coeffs (const struct calib_Qx_obj *	op);
static void	Qx_init (struct calib_Qx_obj *	x);
static void	Qx_init_array_degree (struct calib_Qx_obj *	parray,
				      int			npoly,
				      int			degree);
static void	Qx_init_degree (struct calib_Qx_obj *	x,
				int			degree);
static void	Qx_integral (struct calib_Qx_obj *		rop,
			     const struct calib_Qx_obj *	op);
static void	Qx_ipow (struct calib_Qx_obj *		rop,
			 const struct calib_Qx_obj *	op,
			 int				power);
static void	Qx_mul (struct calib_Qx_obj *		rop,
			const struct calib_Qx_obj *	op1,
			const struct calib_Qx_obj *	op2);
static void	Qx_mul_q (struct calib_Qx_obj *		rop,
			  const struct calib_Qx_obj *	op1,
			  mpq_srcptr			op2);
static void	Qx_mul_si (struct calib_Qx_obj *	rop,
			   const struct calib_Qx_obj *	op1,
			   calib_si_t			op2);
static void	Qx_mul_z (struct calib_Qx_obj *		rop,
			  const struct calib_Qx_obj *	op1,
			  mpz_srcptr			op2);
static void	Qx_neg (struct calib_Qx_obj *		rop,
			const struct calib_Qx_obj *	op);
static calib_bool
		Qx_onep (const struct calib_Qx_obj *	op);
static void	Qx_realloc_copy_poly (struct calib_Qx_obj *	poly,
				      mpq_srcptr		qfact,
				      mpz_srcptr		coeff,
				      int			degree);
static void	Qx_set (struct calib_Qx_obj *		rop,
			const struct calib_Qx_obj *	op);
static void	Qx_set_coeffs (struct calib_Qx_obj *	rop,
			       int			degree,
			       mpq_srcptr		coeffs);
static void	Qx_set_genrep (struct calib_Qx_obj *		rop,
			       const struct calib_genrep *	op,
			       const char *			var);
static void	Qx_set_q (struct calib_Qx_obj *		rop,
			  mpq_srcptr			op);
static void	Qx_set_si (struct calib_Qx_obj *	rop,
			   calib_si_t			op);
static void	Qx_set_var_power (struct calib_Qx_obj *	rop,
				  int			power);
static void	Qx_set_z (struct calib_Qx_obj *		rop,
			  mpz_srcptr			op);
static void	Qx_set_Zx (struct calib_Qx_obj *	rop,
			   const struct calib_Zx_obj *	op);
static void	Qx_sub (struct calib_Qx_obj *		rop,
			const struct calib_Qx_obj *	op1,
			const struct calib_Qx_obj *	op2);
static struct calib_genrep *
		Qx_to_genrep (const struct calib_Qx_obj *	op,
			      const char *			var);
static calib_bool
		Qx_zerop (const struct calib_Qx_obj *	op);

/* Standard "max" of two integers. */

#define _MAX(A,B)	(((A) > (B)) ? (A) : (B))

/*
 * Macros to allocate/zero a local polynomial of the given degree.
 *
 * Note that since the numerlying mpq_t uses memory that is dynamically
 * allocated from the heap, there is little reason to try and do any
 * fancy stack allocation here.  We put only the "struct calib_Qx_obj" guys
 * on the stack, and dynamically allocate the vectors of coefficients.
 */

#define ALLOCATE_LOCAL_POLY(PNAME, DEGREE) \
	struct calib_Qx_obj	PNAME; \
	Qx_init_degree (&PNAME, (DEGREE))

#define FREE_LOCAL_POLY(PNAME) \
	Qx_clear (&PNAME)

/*
 * Macros to allocate/zero an array of local polynomials, each having
 * the given degree.
 */

#define ALLOCATE_LOCAL_POLY_ARRAY(PNAME, NPOLY, DEGREE) \
	struct calib_Qx_obj	PNAME [NPOLY]; \
	Qx_init_array_degree (&PNAME [0], (NPOLY), (DEGREE))

#define FREE_LOCAL_POLY_ARRAY(PNAME, NPOLY) \
	Qx_clear_array (&PNAME [0], (NPOLY))

/*
 * Initialize the given polynomial.  Its value after initialization is zero.
 */

	static
	inline
	void
Qx_init (

struct calib_Qx_obj *	x		/* IN/OUT: polynomial to initialize */
)
{
	CALIB_FATAL_ERROR_IF (x EQ NULL);

	mpq_init (x -> qfact);
	mpq_set_ui (x -> qfact, 1, 1);

	x -> degree	= -1;
	x -> size	= 0;
	x -> coeff	= NULL;
}

/*
 * Initialize the given polynomial to zero, but provide it with a
 * coefficient buffer sufficient to hold a polynomial of the given degree.
 */

	static
	inline
	void
Qx_init_degree (

struct calib_Qx_obj *	x,		/* IN: polynomial to initialize */
int			degree		/* IN: maximum degree */
)
{
int		size;

	CALIB_FATAL_ERROR_IF ((x EQ NULL) OR (degree < -1));

	mpq_init (x -> qfact);
	mpq_set_ui (x -> qfact, 1, 1);

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	x -> degree	= -1;
	x -> size	= size;
	x -> coeff	= calib_new_Z_vector (size);
}

/*
 * Initialize the given array of polynomials to zero, but provide each
 * with a coefficient buffer sufficient to hold a polynomial of the
 * given degree.
 *
 * Although it is tempting to allocate all of the coefficients for all of
 * the polynomials into one big long vector, we cannot do so because the
 * coefficient vector for each polynomial must be separately free-able,
 * and re-allocatable.
 */

	static
	inline
	void
Qx_init_array_degree (

struct calib_Qx_obj *	parray,	/* IN/OUT: array of polynomials */
int			npoly,	/* IN: number of polynomials in array */
int			degree	/* IN: max degree of each polynomial */
)
{
int			i, size;
struct calib_Qx_obj *	pp;

	CALIB_FATAL_ERROR_IF (degree < 0);

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (parray EQ NULL);

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	pp = &parray [0];
	for (i = 0; i < npoly; i++, pp++) {
		mpq_init (pp -> qfact);
		mpq_set_ui (pp -> qfact, 1, 1);

		pp -> degree	= -1;
		pp -> size	= size;
		pp -> coeff	= calib_new_Z_vector (size);
	}
}

/*
 * Force the given polynomial to have buffer space sufficient to hold
 * a polynomial of at least the given degree.  This does not change
 * the value currently residing in this polynomial.
 */

	static
	void
Qx_alloc (

struct calib_Qx_obj *	rop,		/* IN/OUT: poly to alloc buffer for */
int			degree		/* IN: degree desired */
)
{
int		size, dsize;
mpz_ptr		qp, dp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (degree < -1));

	size	= degree + 1;
	dsize	= rop -> size;

	if (dsize < size) {
		dp = rop -> coeff;
		qp = calib_new_Z_vector (size);
		calib_copy_Z_vector (qp, dp, dsize);
		calib_free_Z_vector (dp, dsize);
		rop -> coeff = qp;
		rop -> size  = size;
	}
}

/*
 * Clear out the given polynomial.
 */

	static
	inline
	void
Qx_clear (

struct calib_Qx_obj *	x		/* IN/OUT: polynomial to clear */
)
{
int		size;
mpz_ptr		coeff;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	mpq_clear (x -> qfact);

	coeff = x -> coeff;
	if (coeff NE NULL) {
		size = x -> size;
		CALIB_FATAL_ERROR_IF (size <= 0);
		calib_free_Z_vector (coeff, size);
	}
	Qx_init (x);
}

/*
 * Clear out the given array of polynomials.
 */

	static
	inline
	void
Qx_clear_array (

struct calib_Qx_obj *	parray,	/* IN: array of polynomials to clear */
int			npoly	/* IN: number of polynomials in parray */
)
{
int		i;

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (parray EQ NULL);

	for (i = 0; i < npoly; i++) {
		Qx_clear (&parray [i]);
	}
}

/*
 * Set the given destination polynomial from the given coefficient
 * array.  Reallocate the destination polynomial's coefficient array,
 * if necessary.  We "canonicalize" the destination poly during this
 * operation: assuring that the Z[x] polynomial portion is primitive,
 * and that its leading coefficient (if any) is strictly positive.
 */

	static
	inline
	void
Qx_realloc_copy_poly (

struct calib_Qx_obj *	poly,		/* IN: destination polynomial */
mpq_srcptr		qfact,		/* IN: rational multiplier */
mpz_srcptr		coeff,		/* IN: source coefficient array */
int			degree		/* IN: degree to set */
)
{
int		i, cur_size, new_size, sign;
mpz_ptr		dstbuf;
mpz_t		GCD;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (degree < -1));

	poly -> degree	= degree;

	if (degree < 0) {
		/* We allow (coeff EQ NULL) here (but do not require). */
		mpq_set_ui (poly -> qfact, 1, 1);
		return;
	}

	CALIB_FATAL_ERROR_IF (coeff EQ NULL);

	new_size = degree + 1;
	cur_size = poly -> size;
	dstbuf	 = poly -> coeff;

	if (new_size > cur_size) {
		if (dstbuf NE NULL) {
			calib_free_Z_vector (dstbuf, cur_size);
		}
		dstbuf = calib_new_Z_vector (new_size);
		poly -> coeff = dstbuf;
		poly -> size  = new_size;
	}

	/* Canonicalization steps: Make Z[x] polynomial be primitive	*/
	/* during this copy, with strictly positive leading		*/
	/* coefficient.							*/
	mpz_init (GCD);

	mpz_abs (GCD, &coeff [0]);
	for (i = 1; i <= degree; i++) {
		mpz_gcd (GCD, GCD, &coeff [i]);
		if (mpz_cmp_ui (GCD, 1) EQ 0) break;
	}
	sign = mpz_sgn (&coeff [degree]);
	if (sign <= 0) {
		CALIB_FATAL_ERROR_IF (sign EQ 0);
		mpz_neg (GCD, GCD);
	}

	calib_mpq_mul_z (poly -> qfact, qfact, GCD);

	if (mpz_cmp_ui (GCD, 1) EQ 0) {
		calib_copy_Z_vector (dstbuf, coeff, new_size);
	}
	else {
		for (i = 0; i <= degree; i++) {
			mpz_divexact (&dstbuf [i], &coeff [i], GCD);
		}
	}

	mpz_clear (GCD);
}

/*
 * Copy the given polynomial.
 */

	static
	void
Qx_set (

struct calib_Qx_obj *		rop,	/* IN/OUT: destination */
const struct calib_Qx_obj *	op	/* IN: source */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {
		Qx_realloc_copy_poly (rop,
				      op -> qfact,
				      op -> coeff,
				      op -> degree);
	}
}

/*
 * Set the given polynomial to the given signed integer.
 */

	static
	void
Qx_set_si (

struct calib_Qx_obj *		rop,	/* IN/OUT: Destination polynomial */
calib_si_t			op	/* IN: Source integer */
)
{
	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	if (op EQ 0) {
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
	}
	else {
		mpq_set_si (rop -> qfact, op, 1);
		Qx_alloc (rop, 0);
		mpz_set_ui (&(rop -> coeff [0]), 1);
		rop -> degree = 0;
	}
}

/*
 * Set the given polynomial to the given GMP integer.
 */

	static
	void
Qx_set_z (

struct calib_Qx_obj *		rop,	/* IN/OUT: Destination polynomial */
mpz_srcptr			op	/* IN: Source GMP integer */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpz_sgn (op) EQ 0) {
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
	}
	else {
		mpq_set_z (rop -> qfact, op);
		Qx_alloc (rop, 0);
		mpz_set_ui (&(rop -> coeff [0]), 1);
		rop -> degree = 0;
	}
}

/*
 * Set the given polynomial to the given GMP rational.
 */

	static
	void
Qx_set_q (

struct calib_Qx_obj *		rop,	/* IN/OUT: Destination polynomial */
mpq_srcptr			op	/* IN: Source GMP rational */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpq_sgn (op) EQ 0) {
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
	}
	else {
		mpq_set (rop -> qfact, op);
		Qx_alloc (rop, 0);
		mpz_set_ui (&(rop -> coeff [0]), 1);
		rop -> degree = 0;
	}
}

/*
 * Set the Z[x] polynomial to x**power.
 */

	static
	void
Qx_set_var_power (

struct calib_Qx_obj *	rop,		/* IN/OUT: Result polynomial */
int			power		/* IN: power to set */
)
{
int		i;
mpz_ptr		coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (power < 0));

	mpq_set_ui (rop -> qfact, 1, 1);

	Qx_alloc (rop, power);
	coeff = rop -> coeff;

	for (i = 0; i < power; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
	mpz_set_ui (&coeff [power], 1);
	rop -> degree = power;
}

/*
 * Convert the given polynomial in Z[x] into Q[x] form.
 */

	static
	void
Qx_set_Zx (

struct calib_Qx_obj *		rop,	/* IN/OUT: Result polynomial */
const struct calib_Zx_obj *	op	/* IN: polynomial to convert */
)
{
int		deg, size;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	mpq_set_ui (rop -> qfact, 1, 1);

	deg = op -> degree;
	size = deg + 1;

	Qx_alloc (rop, deg);

	calib_copy_Z_vector (rop -> coeff, op -> coeff, size);

	rop -> degree	= deg;
}

/*
 * Addition of two polynomials over Q.
 */

	static
	void
Qx_add (

struct calib_Qx_obj *		rop,		/* IN/OUT: result */
const struct calib_Qx_obj *	op1,		/* IN: first operand */
const struct calib_Qx_obj *	op2		/* IN: second operand */
)
{
int			i, rdeg, deg1, deg2, rsize, size1, size2;
mpz_ptr			rp, cbuf1, cbuf2;
mpq_t			QFACT;
mpz_t			LCM, MUL1, MUL2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		CALIB_FATAL_ERROR_IF (rdeg < -1);
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	rsize	= rdeg + 1;
	size1	= deg1 + 1;
	size2	= deg2 + 1;

	/* We make both buffers be full size, so that we don't have to	*/
	/* worry about one buffer being shorter than the other.		*/
	rp	= calib_new_Z_vector (2 * rsize);
	cbuf1	= rp;		rp += rsize;
	cbuf2	= rp;

	mpq_init (QFACT);
	mpz_init (LCM);
	mpz_init (MUL1);
	mpz_init (MUL2);

	Qx_common_denominator (LCM, MUL1, MUL2, op1, op2);

	mpq_set_ui (QFACT, 1, 1);
	calib_mpq_div_z (QFACT, QFACT, LCM);

	calib_copy_Z_vector (cbuf1, op1 -> coeff, size1);
	calib_copy_Z_vector (cbuf2, op2 -> coeff, size2);

	for (i = 0; i < size1; i++) {
		mpz_mul (&cbuf1 [i], &cbuf1 [i], MUL1);
	}

	for (i = 0; i < size2; i++) {
		mpz_mul (&cbuf2 [i], &cbuf2 [i], MUL2);
	}

	for (i = 0; i < rsize; i++) {
		mpz_add (&cbuf1 [i], &cbuf1 [i], &cbuf2 [i]);
	}

	while ((rdeg >= 0) AND (mpz_sgn (&cbuf1 [rdeg]) EQ 0)) {
		--rdeg;
	}

	Qx_realloc_copy_poly (rop, QFACT, cbuf1, rdeg);

	calib_free_Z_vector (cbuf1, 2 * rsize);

	mpz_clear (MUL2);
	mpz_clear (MUL1);
	mpz_clear (LCM);
	mpq_clear (QFACT);
}

/*
 * Subtraction of two polynomials over Q.
 */

	static
	void
Qx_sub (

struct calib_Qx_obj *		rop,		/* IN/OUT: result */
const struct calib_Qx_obj *	op1,		/* IN: first operand */
const struct calib_Qx_obj *	op2		/* IN: second operand */
)
{
int			i, rdeg, deg1, deg2, rsize, size1, size2;
mpz_ptr			rp, cbuf1, cbuf2;
mpq_t			QFACT;
mpz_t			LCM, MUL1, MUL2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		CALIB_FATAL_ERROR_IF (rdeg < -1);
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	rsize	= rdeg + 1;
	size1	= deg1 + 1;
	size2	= deg2 + 1;

	/* We make both buffers be full size, so that we don't have to	*/
	/* worry about one buffer being shorter than the other.		*/
	rp	= calib_new_Z_vector (2 * rsize);
	cbuf1	= rp;		rp += rsize;
	cbuf2	= rp;

	mpq_init (QFACT);
	mpz_init (LCM);
	mpz_init (MUL1);
	mpz_init (MUL2);

	Qx_common_denominator (LCM, MUL1, MUL2, op1, op2);

	mpq_set_ui (QFACT, 1, 1);
	calib_mpq_div_z (QFACT, QFACT, LCM);

	calib_copy_Z_vector (cbuf1, op1 -> coeff, size1);
	calib_copy_Z_vector (cbuf2, op2 -> coeff, size2);

	for (i = 0; i < size1; i++) {
		mpz_mul (&cbuf1 [i], &cbuf1 [i], MUL1);
	}

	for (i = 0; i < size2; i++) {
		mpz_mul (&cbuf2 [i], &cbuf2 [i], MUL2);
	}

	for (i = 0; i < rsize; i++) {
		mpz_sub (&cbuf1 [i], &cbuf1 [i], &cbuf2 [i]);
	}

	while ((rdeg >= 0) AND (mpz_sgn (&cbuf1 [rdeg]) EQ 0)) {
		--rdeg;
	}

	Qx_realloc_copy_poly (rop, QFACT, cbuf1, rdeg);

	calib_free_Z_vector (cbuf1, 2 * rsize);

	mpz_clear (MUL2);
	mpz_clear (MUL1);
	mpz_clear (LCM);
	mpq_clear (QFACT);
}

/*
 * Compute the common denominator for Qx polynomials op1 and op2,
 * together with the multipliers mul1 and mul2 needed to conform them.
 */

	static
	void
Qx_common_denominator (

mpz_ptr				lcm,	/* IN/OUT: LCM of denominators */
mpz_ptr				mul1,	/* IN/OUT: Multiplier for op1 */
mpz_ptr				mul2,	/* IN/OUT: Multiplier for op2 */
const struct calib_Qx_obj *	op1,	/* IN: First operand polynomial */
const struct calib_Qx_obj *	op2	/* IN: Second operand polynomial */
)
{
mpz_srcptr	d1, d2;
mpz_t		G, TEMP;

	CALIB_FATAL_ERROR_IF ((lcm  EQ NULL) OR
			      (mul1 EQ NULL) OR
			      (mul2 EQ NULL) OR
			      (op1 EQ NULL) OR
			      (op2 EQ NULL));

	d1	= mpq_denref (op1 -> qfact);
	d2	= mpq_denref (op2 -> qfact);

	mpz_init (G);
	mpz_init (TEMP);

	mpz_gcd (G, d1, d2);

	mpz_set (TEMP, d1);
	mpz_divexact (TEMP, TEMP, G);
	mpz_mul (lcm, TEMP, d2);

	mpz_divexact (mul1, d2, G);
	mpz_divexact (mul2, d1, G);

	mpz_clear (TEMP);
	mpz_clear (G);
}

/*
 * Negate given polynomial over Q.
 */

	static
	void
Qx_neg (

struct calib_Qx_obj *		rop,		/* IN/OUT: result */
const struct calib_Qx_obj *	op		/* IN: operand to negate */
)
{
int		deg;
mpz_ptr		dstp;
mpz_srcptr	srcp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop EQ op) {
		/* Negation in place. */
		deg	= rop -> degree;
		if (deg < 0) {
			CALIB_FATAL_ERROR_IF (deg < -1);
			return;
		}
		mpq_neg (rop -> qfact, rop -> qfact);
	}
	else {
		deg	= op -> degree;
		srcp	= op -> coeff;
		Qx_alloc (rop, deg);
		dstp	= rop -> coeff;
		calib_copy_Z_vector (dstp, srcp, deg + 1);
		rop -> degree = deg;
		mpq_neg (rop -> qfact, op -> qfact);
	}
}

/*
 * Multiplcation of two polynomials over Q.
 */

	static
	void
Qx_mul (

struct calib_Qx_obj *		rop,		/* IN/OUT: result */
const struct calib_Qx_obj *	op1,		/* IN: first operand */
const struct calib_Qx_obj *	op2		/* IN: second operand */
)
{
int			i, j, rdeg, deg1, deg2, rsize;
mpz_ptr			res, rp1, rp2;
mpz_srcptr		cp1, cp2;
mpz_t			TEMP;
mpq_t			QFACT;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	if ((deg1 < 0) OR (deg2 < 0)) {
		/* Multipying by zero. */
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	rdeg	= deg1 + deg2;

	rsize = rdeg + 1;
	res = calib_new_Z_vector (rsize);

	mpz_init (TEMP);
	mpq_init (QFACT);

	mpq_mul (QFACT, op1 -> qfact, op2 -> qfact);

	cp1 = &(op1 -> coeff [0]);
	rp1 = &res [0];
	for (i = 0; i <= deg1; i++, rp1++, cp1++) {
		rp2 = rp1;
		cp2 = &(op2 -> coeff [0]);
		for (j = 0; j <= deg2; j++, rp2++, cp2++) {
			mpz_mul (TEMP, cp1, cp2);
			mpz_add (rp2, rp2, TEMP);
		}
	}
	while ((rdeg >= 0) AND (mpz_sgn (&res [rdeg]) EQ 0)) {
		--rdeg;
	}

	Qx_realloc_copy_poly (rop, QFACT, res, rdeg);

	mpq_clear (QFACT);
	mpz_clear (TEMP);
	calib_free_Z_vector (res, rsize);
}

/*
 * Multiply Q[x] polynomial by signed integer.
 */

	static
	void
Qx_mul_si (

struct calib_Qx_obj *		rop,		/* IN/OUT: result */
const struct calib_Qx_obj *	op1,		/* IN: first operand */
calib_si_t			op2		/* IN: second operand */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL));

	deg	= op1 -> degree;
	if ((deg < 0) OR (op2 EQ 0)) {
		/* Multipying by zero. */
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	if (rop NE op1) {
		Qx_set (rop, op1);
	}

	calib_mpq_mul_si (rop -> qfact, rop -> qfact, op2);
}

/*
 * Multiply Q[x] polynomial by GMP integer.
 */

	static
	void
Qx_mul_z (

struct calib_Qx_obj *		rop,		/* IN/OUT: result */
const struct calib_Qx_obj *	op1,		/* IN: first operand */
mpz_srcptr			op2		/* IN: second operand */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg	= op1 -> degree;
	if ((deg < 0) OR (mpz_sgn (op2) EQ 0)) {
		/* Multipying by zero. */
		mpq_set_ui (rop -> qfact, 1, 1);
		rop -> degree = -1;
		return;
	}

	if (rop NE op1) {
		Qx_set (rop, op1);
	}

	calib_mpq_mul_z (rop -> qfact, rop -> qfact, op2);
}

/*
 * Multiply Q[x] polynomial by GMP rational.
 */

	static
	void
Qx_mul_q (

struct calib_Qx_obj *		rop,		/* IN/OUT: result */
const struct calib_Qx_obj *	op1,		/* IN: first operand */
mpq_srcptr			op2		/* IN: second operand */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg	= op1 -> degree;
	if ((deg < 0) OR (mpq_sgn (op2) EQ 0)) {
		/* Multipying by zero. */
		rop -> degree = -1;
		return;
	}

	if (rop NE op1) {
		Qx_set (rop, op1);
	}

	mpq_mul (rop -> qfact, rop -> qfact, op2);
}

/*
 * Take the given polynomial to the given integer power.
 */

	static
	void
Qx_ipow (

struct calib_Qx_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_Qx_obj *	op,	/* IN: polynomial to exponentiate */
int				power	/* IN: integer power */
)
{
struct calib_Qx_obj	RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (power < 0));

	Qx_init (&RES);
	Qx_init (&PREG);

	Qx_set_si (&RES, 1);
	Qx_set (&PREG, op);

	while (power > 0) {
		if ((power & 1) NE 0) {
			Qx_mul (&RES, &RES, &PREG);
		}
		Qx_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	Qx_set (rop, &RES);
	Qx_clear (&PREG);
	Qx_clear (&RES);
}

/*
 * Create a dynamically allocated copy of the given polynomial.
 */

	static
	struct calib_Qx_obj *
Qx_dup (

const struct calib_Qx_obj *	op	/* IN: polynomial to duplicate */
)
{
int			deg, size;
struct calib_Qx_obj *	dst;
mpz_ptr			zp;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	size = deg + 1;

	dst = CALIB_NEW (struct calib_Qx_obj);
	mpq_init (dst -> qfact);
	mpq_set (dst -> qfact, op -> qfact);
	zp = NULL;
	if (size > 0) {
		zp = calib_new_Z_vector (size);
		calib_copy_Z_vector (zp, op -> coeff, size);
	}
	dst -> degree	= deg;
	dst -> size	= size;
	dst -> coeff	= zp;

	return (dst);
}

/*
 * Free the given polynomial, which must have been dynamically allocated.
 */

	static
	void
Qx_free (

struct calib_Qx_obj *	poly		/* IN: polynomial to free */
)
{
	if (poly NE NULL) {
		Qx_clear (poly);
		free (poly);
	}
}

/*
 * Evaluate the given poly(value).
 */

	static
	void
Qx_eval (

mpq_ptr				rop,	/* IN/OUT: result */
const struct calib_Qx_obj *	poly,	/* IN: polynomial to evaluate */
mpq_srcptr			value	/* IN: value to use */
)
{
int			i;
mpz_srcptr		coeff;
mpq_t			SUM, TEMP;

	CALIB_FATAL_ERROR_IF ((rop   EQ NULL) OR
			      (poly  EQ NULL) OR
			      (value EQ NULL));

	mpq_init (SUM);
	mpq_init (TEMP);

	coeff	= poly -> coeff;

	for (i = poly -> degree; i >= 0; i--) {
		mpq_mul (SUM, SUM, value);
		mpq_set_z (TEMP, &coeff [i]);
		mpq_add (SUM, SUM, TEMP);
	}
	mpq_mul (rop, poly -> qfact, SUM);

	mpq_clear (TEMP);
	mpq_clear (SUM);
}

/*
 * The division algorith (with remainder):
 * Given polynomials a, b \in Q[x], find polynomials q, r \in Q[x]
 * such that:
 *
 *	a = q*b + r	AND	degree (r) < degree (b)
 *
 * Although this is "clean" over a "pure" Q[x] implementation (because
 * the polynomial coefficients reside in a field and can be inverted),
 * that approach is quite inefficient:
 *
 *	1. Each of these rational arithmetic operations needs mpz_gcd()
 *	   to reduce to lowest terms
 *
 *	2. Most of these GCD's yield an answer of 1 (which is the slowest
 *	   case for Euclid's algorithm)
 *
 * Because the present implementation of Qx represents a polynomial as a
 * rational multiple of a primitive Z[x] polynomial, we get to use more
 * efficient pseudo-division in Z[x] here.  The solution we desire has
 * the form:
 *
 *	(A * a(x)) = (Q * q(x)) * (B * b(x)) + (R * r(x))
 *
 * while pseudo-division in Z[x] gives us:
 *
 *	d * a(x) = q(x) * b(x) + r(x)
 *
 * which means:
 *
 *	Q = R = A / (B * d)
 *
 * (but note that Q and R may diverge after moving integer contents from
 * q(x) and r(x) into Q and R, respectively).
 */

	static
	void
Qx_div (

struct calib_Qx_obj *		quotient,	/* IN/OUT: quotient */
struct calib_Qx_obj *		remainder,	/* IN/OUT: remainder */
const struct calib_Qx_obj *	a,		/* IN: dividend */
const struct calib_Qx_obj *	b		/* IN: divisor */
)
{
int				adeg, bdeg;
const struct calib_Zx_dom *	Zx;
mpz_t				DVAL;
mpq_t				QFACT;
struct calib_Zx_obj		QUOT, REM, AZX, BZX;

	CALIB_FATAL_ERROR_IF ((a EQ NULL) OR (b EQ NULL));

	adeg	= a -> degree;
	bdeg	= b -> degree;

	Zx	= calib_get_Zx_dom ();

	mpz_init (DVAL);
	mpq_init (QFACT);

	Zx -> init (&QUOT);
	Zx -> init (&REM);

	/* Assemble "fake" Zx versions of a and b.  (These are used in	*/
	/* a "read-only" fashion.)					*/
	AZX.degree	= adeg;
	AZX.size	= adeg + 1;
	AZX.coeff	= a -> coeff;

	BZX.degree	= bdeg;
	BZX.size	= bdeg + 1;
	BZX.coeff	= b -> coeff;

	Zx -> div (&QUOT, &REM, DVAL, &AZX, &BZX);

	mpq_div (QFACT, a -> qfact, b -> qfact);
	calib_mpq_div_z (QFACT, QFACT, DVAL);

	if (quotient NE NULL) {
		Qx_realloc_copy_poly (quotient,
				      QFACT,
				      QUOT.coeff,
				      QUOT.degree);
	}

	if (remainder NE NULL) {
		Qx_realloc_copy_poly (remainder,
				      QFACT,
				      REM.coeff,
				      REM.degree);
	}

	Zx -> clear (&REM);
	Zx -> clear (&QUOT);

	mpq_clear (QFACT);
	mpz_clear (DVAL);
}

/*
 * Compute the Greatest Common Divisor of two polynomials over Q.
 * The result is always monic unless a = b = 0.
 *
 * As with Qx_div() above, most of the heavy lifting here is done
 * in Zx land.  The key observation is that we want the resulting
 * GCD to be monic.
 */

	static
	void
Qx_gcd (

struct calib_Qx_obj *		gcd,	/* IN/OUT: GCD result */
const struct calib_Qx_obj *	a,	/* IN: first operand */
const struct calib_Qx_obj *	b	/* IN: second operand */
)
{
int				deg, adeg, bdeg;
mpz_srcptr			lc;
const struct calib_Zx_dom *	Zx;
mpq_t				QFACT;
struct calib_Zx_obj		GCD, AZX, BZX;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	adeg	= a -> degree;
	bdeg	= b -> degree;
	Zx	= calib_get_Zx_dom ();

	mpq_init (QFACT);
	Zx -> init (&GCD);

	/* Assemble "fake" Zx versions of a and b.  (These are used in	*/
	/* a "read-only" fashion.)					*/
	AZX.degree	= adeg;
	AZX.size	= adeg + 1;
	AZX.coeff	= a -> coeff;

	BZX.degree	= bdeg;
	BZX.size	= bdeg + 1;
	BZX.coeff	= b -> coeff;

	Zx -> gcd (&GCD, &AZX, &BZX);

	mpq_set_ui (QFACT, 1, 1);

	deg	= GCD.degree;
	if (deg < 0) {
		mpq_set_ui (gcd -> qfact, 1, 1);
		gcd -> degree = -1;
	}
	else {
		/* Lead coefficient should be strictly positive. */
		lc = &GCD.coeff [deg];
		CALIB_FATAL_ERROR_IF (mpz_sgn (lc) <= 0);

		calib_mpq_div_z (QFACT, QFACT, lc);

		Qx_realloc_copy_poly (gcd, QFACT, GCD.coeff, deg);
	}

	Zx -> clear (&GCD);
	mpq_clear (QFACT);
}

/*
 * The Extended Euclidean Algorithm for polynomials in Q[x].
 * Given polynomials a, b \in Q[x],
 * compute polynomials g, xa, xb \in Q[x] such that
 *
 * 1.	a is a multiple of g,
 * 2.	b is a multiple of g,
 * 3.	xa * a + xb * b = g.
 *
 * These three conditions imply that g = gcd(a,b).  The polynomial g
 * is monic (unless a = b = 0, in which case g = 0).
 *
 * Once again, the heavy lifting falls to Zx, and we want g to be monic.
 */

	static
	void
Qx_extgcd (

struct calib_Qx_obj *		gcd,	/* IN/OUT: GCD result */
struct calib_Qx_obj *		xa,	/* IN/OUT: a multiplier */
struct calib_Qx_obj *		xb,	/* IN/OUT: b multiplier */
const struct calib_Qx_obj *	a,	/* IN: first operand */
const struct calib_Qx_obj *	b	/* IN: second operand */
)
{
int				adeg, bdeg, gdeg;
mpz_srcptr			glc;
const struct calib_Zx_dom *	Zx;
mpq_t				QF_G, QF_XA, QF_XB;
struct calib_Zx_obj		G, XA, XB, A, B;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR
			      (xa  EQ NULL) OR
			      (xb  EQ NULL) OR
			      (a   EQ NULL) OR
			      (b   EQ NULL));

	adeg	= a -> degree;
	bdeg	= b -> degree;
	Zx	= calib_get_Zx_dom ();

	mpq_init (QF_G);
	mpq_init (QF_XA);
	mpq_init (QF_XB);

	Zx -> init (&G);
	Zx -> init (&XA);
	Zx -> init (&XB);

	mpq_set_ui (QF_G, 1, 1);
	mpq_set (QF_XA, a -> qfact);
	mpq_set (QF_XB, b -> qfact);

	/* Assemble "fake" Zx versions of a and b.  (These are used in	*/
	/* a "read-only" fashion.)					*/
	A.degree	= adeg;
	A.size		= adeg + 1;
	A.coeff		= a -> coeff;

	B.degree	= bdeg;
	B.size	= bdeg + 1;
	B.coeff	= b -> coeff;

	Zx -> extgcd (&G, &XA, &XB, &A, &B);

	gdeg	= G.degree;
	if (gdeg < 0) {
		mpq_set_ui (gcd -> qfact, 1, 1);
		gcd -> degree = -1;
	}
	else {
		glc = &G.coeff [gdeg];
		CALIB_FATAL_ERROR_IF (mpz_sgn (glc) <= 0);
		calib_mpq_div_z (QF_G, QF_G, glc);

		Qx_realloc_copy_poly (gcd, QF_G, G.coeff, gdeg);
	}

	mpq_div (QF_XA, QF_G, a -> qfact);
	mpq_div (QF_XB, QF_G, b -> qfact);

	Qx_realloc_copy_poly (xa, QF_XA, XA.coeff, XA.degree);
	Qx_realloc_copy_poly (xb, QF_XB, XB.coeff, XB.degree);

	Zx -> clear (&XB);
	Zx -> clear (&XA);
	Zx -> clear (&G);

	mpq_clear (QF_XB);
	mpq_clear (QF_XA);
	mpq_clear (QF_G);
}

/*
 * Factor the given polynomial in Q[x] completely into its monic,
 * irreducible factors over Q[x].  The correct multiplicity of each
 * factor is also determined.
 *
 * Once again, the heavy lifting falls on Zx.
 */

	static
	struct calib_Qx_factor *
Qx_factor (

const struct calib_Qx_obj *	poly		/* IN: polynomial to factor */
)
{
int				deg, deg2;
const struct calib_Zx_dom *	Zx;
struct calib_Qx_factor *	list;
struct calib_Qx_factor **	hookp;
struct calib_Qx_factor *	fp;
struct calib_Zx_factor *	zxfactors;
struct calib_Zx_factor *	zxfp;
struct calib_Zx_obj *		zxpoly;
mpz_srcptr			flc;
mpq_t				QFACT;
struct calib_Qx_obj		QXPOLY;
struct calib_Zx_obj		ZXPOLY;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	deg	= poly -> degree;
	Zx	= calib_get_Zx_dom ();

	mpq_init (QFACT);
	Qx_init (&QXPOLY);

	list = NULL;
	hookp = &list;

	if (mpq_cmp_ui (poly -> qfact, 1, 1) NE 0) {
		/* Since all of the "non-trivial" Qx factors are monic,	*/
		/* we emit a constant factor first.			*/
		Qx_set_q (&QXPOLY, poly -> qfact);

		fp = CALIB_NEW (struct calib_Qx_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= Qx_dup (&QXPOLY);
		fp -> next		= NULL;
		*hookp = fp;
		hookp = &(fp -> next);
	}

	/* Assemble a "fake" Zx version of poly. */
	ZXPOLY.degree	= deg;
	ZXPOLY.size	= deg + 1;
	ZXPOLY.coeff	= poly -> coeff;

	zxfactors = Zx -> factor (&ZXPOLY);

	for (zxfp = zxfactors; zxfp NE NULL; zxfp = zxfp -> next) {
		zxpoly	= zxfp -> factor;
		deg2	= zxpoly -> degree;

		CALIB_FATAL_ERROR_IF (deg2 <= 0);

		flc	= &(zxpoly -> coeff [deg2]);

		CALIB_FATAL_ERROR_IF (mpz_sgn (flc) <= 0);

		mpq_set_ui (QFACT, 1, 1);
		calib_mpq_div_z (QFACT, QFACT, flc);

		Qx_realloc_copy_poly (&QXPOLY, QFACT, zxpoly -> coeff, deg2);

		fp = CALIB_NEW (struct calib_Qx_factor);
		fp -> multiplicity	= zxfp -> multiplicity;
		fp -> factor		= Qx_dup (&QXPOLY);
		fp -> next		= NULL;
		*hookp = fp;
		hookp = &(fp -> next);
	}
	Zx -> free_factors (zxfactors);

	Qx_clear (&QXPOLY);
	mpq_clear (QFACT);

	return (list);
}

/*
 * Free the given factorization, as produced by Qx_factor().
 */

	static
	void
Qx_free_factors (

struct calib_Qx_factor *	factors	/* IN: factors to recursively free */
)
{
struct calib_Qx_factor *	ftemp;
struct calib_Qx_obj *		poly;

	while (factors NE NULL) {
		ftemp = factors -> next;
		factors -> next = NULL;
		poly = factors -> factor;
		factors -> factor = NULL;
		Qx_free (poly);
		free (factors);
		factors = ftemp;
	}
}

/*
 * Compute the derivative of the given polynomial in Q[x].
 */

	static
	void
Qx_derivative (

struct calib_Qx_obj *		rop,	/* IN/OUT: derivative polynomial */
const struct calib_Qx_obj *	op	/* IN: polynomial to differentiate */
)
{
int			i, deg, rsize;
mpz_ptr			result;
mpz_srcptr		pcoeff;

	deg	= op -> degree;
	pcoeff	= op -> coeff;
	rsize	= deg;
	result	= calib_new_Z_vector (rsize);

	for (i = 1; i <= deg; i++) {
		mpz_mul_si (&result [i - 1], &pcoeff [i], i);
	}

	do {
		--deg;
	} while ((deg >= 0) AND (mpz_sgn (&result [deg]) EQ 0));

	Qx_realloc_copy_poly (rop, op -> qfact, result, deg);

	calib_free_Z_vector (result, rsize);
}

/*
 * Integrate the given polynomial in Q[x].
 */

	static
	void
Qx_integral (

struct calib_Qx_obj *		rop,	/* IN/OUT: integral polynomial */
const struct calib_Qx_obj *	op	/* IN: polynomial to integrate */
)
{
int			i, deg, rsize;
mpq_ptr			result;
mpz_srcptr		pcoeff;
mpz_ptr			zcoeff;
mpz_t			LCM;
mpq_t			TEMP;

	deg	= op -> degree;
	pcoeff	= op -> coeff;
	rsize	= deg + 2;

	result	= calib_new_Q_vector (rsize);
	zcoeff	= calib_new_Z_vector (rsize);
	mpz_init (LCM);
	mpq_init (TEMP);

	for (i = 0; i <= deg; i++) {
		mpq_set_z (TEMP, &pcoeff [i]);
		calib_mpq_div_si (&result [i + 1], TEMP, i + 1);
	}
	++deg;

	while ((deg >= 0) AND (mpq_sgn (&result [deg]) EQ 0)) {
		--deg;
	}

	/* Find common denominator and convert to Z coefficients. */
	Qx_convert_Q_vector_to_Z (LCM, zcoeff, result, rsize);

	calib_mpq_div_z (TEMP, op -> qfact, LCM);

	Qx_realloc_copy_poly (rop, TEMP, zcoeff, deg);

	mpq_clear (TEMP);
	mpz_clear (LCM);
	calib_free_Z_vector (zcoeff, rsize);
	calib_free_Q_vector (result, rsize);
}

/*
 * Convert the given Q vector into the corresponding LCM (of denominators)
 * and Z vector.
 */

	static
	void
Qx_convert_Q_vector_to_Z (

mpz_ptr		lcm,		/* IN/OUT: Common denominator */
mpz_ptr		zarray,		/* IN/OUT: Vector of Z values */
mpq_srcptr	qarray,		/* IN: Vector of Q values */
int		n		/* IN: Length of vectors */
)
{
int		i;
mpz_srcptr	zp;
mpz_t		G, LCM, TEMP;

	CALIB_FATAL_ERROR_IF ((lcm    EQ NULL) OR
			      (zarray EQ NULL) OR
			      (qarray EQ NULL));

	mpz_init (G);
	mpz_init (LCM);
	mpz_init (TEMP);

	/* Compute least-common-multiple of the denominators. */
	mpz_set_ui (LCM, 1);

	for (i = 0; i < n; i++) {
		zp = mpq_denref (&qarray [i]);
		mpz_gcd (G, LCM, zp);
		mpz_divexact (TEMP, zp, G);
		mpz_mul (LCM, LCM, TEMP);
	}

	mpz_set (lcm, LCM);

	if (mpz_cmp_ui (LCM, 1) EQ 0) {
		/* All denominators are 1. */
		for (i = 0; i < n; i++) {
			mpz_set (&zarray [i], mpq_numref (&qarray [i]));
		}
	}
	else {
		for (i = 0; i < n; i++) {
			mpz_set (TEMP, LCM);
			mpz_divexact (TEMP, TEMP, mpq_denref (&qarray [i]));
			mpz_mul (&zarray [i], TEMP, mpq_numref (&qarray [i]));
		}
	}

	mpz_clear (TEMP);
	mpz_clear (LCM);
	mpz_clear (G);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically zero.
 */

	static
	calib_bool
Qx_zerop (

const struct calib_Qx_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg >= 0) {
		return (FALSE);
	}

	CALIB_FATAL_ERROR_IF (deg < -1);

	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically one.
 */

	static
	calib_bool
Qx_onep (

const struct calib_Qx_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;
calib_bool	result;
mpq_t		TEMP;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg NE 0) {
		return (FALSE);
	}

	mpq_init (TEMP);
	calib_mpq_mul_z (TEMP, op -> qfact, &(op -> coeff [0]));

	result = mpq_cmp_ui (TEMP, 1, 1);

	mpq_clear (TEMP);

	return (result);
}

/*
 * Convert the given general expression into a member of Q[x].
 * All of the variables in the given expression *MUST* match the given
 * variable name.
 */

	static
	void
Qx_set_genrep (

struct calib_Qx_obj *		rop,	/* IN/OUT: Resulting polynomial */
const struct calib_genrep *	op,	/* IN: expression to convert */
const char *			var	/* IN: main var for polynomial */
)
{
int				i;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_Qx_obj		POLY;
struct calib_Qx_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (var EQ NULL));

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Qx_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Qx_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		Qx_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		CALIB_FATAL_ERROR_IF (strcmp (op -> u.var, var) NE 0);
		Qx_set_var_power (rop, 1);
		break;

	case CALIB_GENREP_OP_ADD:
		Qx_init (&POLY);
		Qx_init (&TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Qx_set_genrep (&TEMP, lp -> operand, var);
			Qx_add (&POLY, &POLY, &TEMP);
		}
		Qx_set (rop, &POLY);

		Qx_clear (&TEMP);
		Qx_clear (&POLY);
		break;

	case CALIB_GENREP_OP_MUL:
		Qx_init (&POLY);
		Qx_init (&TEMP);
		Qx_set_si (&POLY, 1);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Qx_set_genrep (&TEMP, lp -> operand, var);
			Qx_mul (&POLY, &POLY, &TEMP);
		}
		Qx_set (rop, &POLY);

		Qx_clear (&TEMP);
		Qx_clear (&POLY);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;
		CALIB_FATAL_ERROR_IF (i < 0);

		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);

		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			CALIB_FATAL_ERROR_IF (strcmp (gp -> u.var, var) NE 0);

			/* Optimize important special case. */
			Qx_set_var_power (rop, i);
		}
		else {
			Qx_init (&TEMP);
			Qx_set_genrep (&TEMP, gp, var);
			Qx_ipow (rop, &TEMP, i);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Convert the given element of Q[x] into general representation.
 * Use the variable name given.
 */

	static
	struct calib_genrep *
Qx_to_genrep (

const struct calib_Qx_obj *	op,	/* IN: polynomial to convert */
const char *			var	/* IN: variable name to use */
)
{
int				i, degree;
mpq_ptr				coeff;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (var EQ NULL));

	degree	= op -> degree;

	/* Get array of fully rational coefficients. */
	coeff	= Qx_get_coeffs (op);

	root = NULL;
	hookp = &root;

	for (i = degree; i >= 0; i--) {
		if (mpq_sgn (&coeff [i]) EQ 0) continue;
		vgp = calib_genrep_q (&coeff [i]);
		gp1 = calib_genrep_poly_term (vgp, var, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	calib_free_Q_vector (coeff, degree + 1);

	return (gp1);
}

/*
 * Convert the given list of factors into general representation.
 * Use the variable given.
 */

	static
	struct calib_genrep *
Qx_factors_to_genrep (

const struct calib_Qx_factor *	factors,	/* IN: list of factors */
const char *			var		/* IN: variable to use */
)
{
int				pow;
const struct calib_Qx_factor *	fp;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;

	root = NULL;
	hookp = &root;

	for (fp = factors; fp NE NULL; fp = fp -> next) {
		gp1 = Qx_to_genrep (fp -> factor, var);
		pow = fp -> multiplicity;
		CALIB_FATAL_ERROR_IF (pow <= 0);
		if (pow > 1) {
			gp2 = CALIB_NEW (struct calib_genrep);
			gp2 -> op		= CALIB_GENREP_OP_IPOW;
			gp2 -> u.ipow.base	= gp1;
			gp2 -> u.ipow.power	= pow;
			gp1 = gp2;
		}
		lp = calib_genrep_new_list (gp1, NULL);
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (1);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_MUL;
		gp1 -> u.list	= root;
	}
	return (gp1);
}

/*
 * Return a dynamically allocated array of the rational coefficients
 * of this Qx polynomial.  (The caller must free it.)
 */

	static
	mpq_ptr
Qx_get_coeffs (

const struct calib_Qx_obj *	op	/* IN: Polynomial for which to	*/
					/*     get the coefficients */
)
{
int		i, deg;
mpz_srcptr	scoeff;
mpq_srcptr	qfact;
mpq_ptr		dcoeff;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	qfact	= op -> qfact;
	deg	= op -> degree;
	scoeff	= op -> coeff;

	CALIB_FATAL_ERROR_IF (deg < -1);

	dcoeff	= calib_new_Q_vector (deg + 1);

	for (i = 0; i <= deg; i++) {
		calib_mpq_mul_z (&dcoeff [i], qfact, &scoeff [i]);
	}

	return (dcoeff);
}

/*
 * Set the given Q[x] polynomial rop to have the given degree and
 * rational coefficients.
 */

	static
	void
Qx_set_coeffs (

struct calib_Qx_obj *	rop,	/* IN/OUT: Polynomial to set */
int			degree,	/* IN: Degree to set */
mpq_srcptr		coeffs	/* IN: Polynomial coefficients to set */
)
{
int		i;
mpz_ptr		zvec;
mpz_t		LCM, G, TEMP;
mpq_t		QFACT;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	if (degree < 0) {
		Qx_set_si (rop, 0);
		return;
	}

	mpz_init (LCM);
	mpz_init (G);
	mpz_init (TEMP);
	mpq_init (QFACT);

	/* Compute common demoninator. */

	mpz_set_ui (LCM, 1);
	for (i = 0; i <= degree; i++) {
		mpz_gcd (G, LCM, mpq_denref (&coeffs [i]));
		mpz_divexact (TEMP, mpq_denref (&coeffs [i]), G);
		mpz_mul (LCM, LCM, TEMP);
	}

	/* Clear denominators to get a vector of integer coefficients. */
	zvec	= calib_new_Z_vector (degree + 1);
	for (i = 0; i <= degree; i++) {
		mpz_divexact (TEMP, LCM, mpq_denref (&coeffs [i]));
		mpz_mul (&zvec [i], TEMP, mpq_numref (&coeffs [i]));
	}

	mpz_set_ui (mpq_numref (QFACT), 1);
	mpz_set    (mpq_denref (QFACT), LCM);

	/* Remove integer content and make lead-coefficient be positive. */
	Qx_realloc_copy_poly (rop, QFACT, zvec, degree);

	calib_free_Z_vector (zvec, degree + 1);

	mpq_clear (QFACT);
	mpz_clear (TEMP);
	mpz_clear (G);
	mpz_clear (LCM);
}

/*
 * The ring Q[x].
 */

static const struct calib_Qx_dom	_Qx_dom = {
	.init			= Qx_init,
	.init_degree		= Qx_init_degree,
	.alloc			= Qx_alloc,
	.clear			= Qx_clear,
	.set			= Qx_set,
	.set_si			= Qx_set_si,
	.set_z			= Qx_set_z,
	.set_q			= Qx_set_q,
	.set_var_power		= Qx_set_var_power,
	.set_Zx			= Qx_set_Zx,
	.add			= Qx_add,
	.sub			= Qx_sub,
	.neg			= Qx_neg,
	.mul			= Qx_mul,
	.mul_si			= Qx_mul_si,
	.mul_z			= Qx_mul_z,
	.mul_q			= Qx_mul_q,
	.dup			= Qx_dup,
	.free			= Qx_free,
	.ipow			= Qx_ipow,
	.eval			= Qx_eval,
	.div			= Qx_div,
	.gcd			= Qx_gcd,
	.extgcd			= Qx_extgcd,
	.factor			= Qx_factor,
	.free_factors		= Qx_free_factors,
	.derivative		= Qx_derivative,
	.integral		= Qx_integral,
	.zerop			= Qx_zerop,
	.onep			= Qx_onep,
	.set_genrep		= Qx_set_genrep,
	.to_genrep		= Qx_to_genrep,
	.factors_to_genrep	= Qx_factors_to_genrep,
	.get_coeffs		= Qx_get_coeffs,
	.set_coeffs		= Qx_set_coeffs,
};

/*
 * Return the ring Q[x] of polynomials over the rationals Q.
 */

	const struct calib_Qx_dom *
calib_get_Qx_dom (void)

{
	return (&_Qx_dom);
}
