/**********************************************************************

	$Id: Za.c,v 1.1 2024/10/20 00:36:26 warme Exp $

	File:	Za.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The ring Z(a), where a is an algebraic extension defined
	to be the root of a given monic, irreducible polynomial
	g(a) \in Z[a].  It is a quotient ring obtained as Z[a] mod g(a).

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Za.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/Zx.h"
#include <gmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Local Functions
 */

static void	Za_add (struct calib_Za_obj *		rop,
			const struct calib_Za_obj *	op1,
			const struct calib_Za_obj *	op2);
static void	Za_clear (struct calib_Za_obj *		x);
static void	Za_cvZa (struct calib_Za_obj *		rop,
			 const struct calib_Za_obj *	op);
static int	Za_degree (const struct calib_Za_obj *	op);
static void	Za_div_z_exact (
			struct calib_Za_obj *		rop,
			const struct calib_Za_obj *	op1,
			mpz_srcptr			op2);
static void	Za_init (const struct calib_Za_dom *	rp,
			 struct calib_Za_obj *		x);
static void	Za_ipow (struct calib_Za_obj *		rop,
			 const struct calib_Za_obj *	op,
			 int				power);
static void	Za_mul (struct calib_Za_obj *		rop,
			const struct calib_Za_obj *	op1,
			const struct calib_Za_obj *	op2);
static void	Za_mul_a (struct calib_Za_obj *		rop,
			  const struct calib_Za_obj *	op);
static void	Za_mul_z (struct calib_Za_obj *		rop,
			  const struct calib_Za_obj *	op1,
			  mpz_srcptr			op2);
static void	Za_neg (struct calib_Za_obj *		rop,
			const struct calib_Za_obj *	op);
static calib_bool
		Za_onep (const struct calib_Za_obj *	op);
static void	Za_pinv (struct calib_Za_obj *		rop,
			 mpz_ptr			d,
			 const struct calib_Za_obj *	op);
static void	Za_prim_part (mpz_ptr				content,
			      struct calib_Za_obj *		ppart,
			      const struct calib_Za_obj *	op);
static void	Za_set (struct calib_Za_obj *		rop,
			const struct calib_Za_obj *	op);
static void	Za_set_genrep (struct calib_Za_obj *		rop,
			       const struct calib_genrep *	op,
			       const char *			var);
static void	Za_set_q (struct calib_Za_obj *		rop,
			  mpq_srcptr			op);
static void	Za_set_si (struct calib_Za_obj *	rop,
			   calib_si_t			op);
static void	Za_set_var_power (struct calib_Za_obj *	rop,
				  int			power);
static void	Za_set_z (struct calib_Za_obj *	rop,
			  mpz_srcptr		op);
static void	Za_sub (struct calib_Za_obj *		rop,
			const struct calib_Za_obj *	op1,
			const struct calib_Za_obj *	op2);
static struct calib_genrep *
		Za_to_genrep (const struct calib_Za_obj *	op,
			      const char *			var);
static calib_bool
		Za_zerop (const struct calib_Za_obj *	op);

/*
 * Initialize the given x to hold a value from the given Z(a) ring.
 */

	static
	void
Za_init (

const struct calib_Za_dom *	rp,	/* IN: description of ring */
struct calib_Za_obj *		x	/* IN/OUT: value to initialize */
)
{
	CALIB_FATAL_ERROR_IF ((rp EQ NULL) OR (x EQ NULL));

	x -> dom	= rp;
	x -> coeff	= calib_new_Z_vector (rp -> gpoly -> degree);
}

/*
 * Clear the given Z(a) value object.
 */

	static
	void
Za_clear (

struct calib_Za_obj *		x	/* IN/OUT: value to clear */
)
{
const struct calib_Za_dom *	rp;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	rp = x -> dom;

	calib_free_Z_vector (x -> coeff, rp -> gpoly -> degree);

	x -> dom	= NULL;
	x -> coeff	= NULL;
}

/*
 * Set the destination Z(a) value from the source Z(a) value.
 */

	static
	void
Za_set (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
const struct calib_Za_obj *	op	/* IN: Za value to set */
)
{
int				k;
const struct calib_Za_dom *	rp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {

		rp = rop -> dom;
		CALIB_FATAL_ERROR_IF (op -> dom NE rp);

		k = rp -> gpoly -> degree;
		calib_copy_Z_vector (rop -> coeff, op -> coeff, k);
	}
}

/*
 * Set the destination Q(a) value from the source signed integer.
 */

	static
	void
Za_set_si (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
calib_si_t			op	/* IN: integer to convert */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				coeff;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	rp	= rop -> dom;
	coeff	= rop -> coeff;
	k	= rp -> gpoly -> degree;

	mpz_set_si (&coeff [0], op);

	for (i = 1; i < k; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
}

/*
 * Set the destination Z(a) value from the source GMP integer.
 */

	static
	void
Za_set_z (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
mpz_srcptr			op	/* IN: integer to convert */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	rp	= rop -> dom;
	coeff	= rop -> coeff;
	k	= rp -> gpoly -> degree;

	mpz_set (&coeff [0], op);

	for (i = 1; i < k; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
}

/*
 * Convert the given GMP rational into the corresponding member of Z(a).
 * (The rational must have a denominator of 1.)
 */

	static
	void
Za_set_q (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
mpq_srcptr			op	/* IN: rational to convert */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	/* The denominator *must* be 1. */
	CALIB_FATAL_ERROR_IF (mpz_cmp_ui (mpq_denref (op), 1) NE 0);

	rp	= rop -> dom;
	coeff	= rop -> coeff;
	k	= rp -> gpoly -> degree;

	mpz_set (&coeff [0], mpq_numref (op));

	for (i = 1; i < k; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
}

/*
 * Set the destination Z(a) value to be a**power.
 */

	static
	void
Za_set_var_power (

struct calib_Za_obj *	rop,		/* IN/OUT: destination */
int			power		/* IN: power to use */
)
{
int				k;
const struct calib_Za_dom *	rp;
mpz_ptr				coeff;
struct calib_Za_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (power < 0));

	rp	= rop -> dom;
	coeff	= rop -> coeff;
	k	= rp -> gpoly -> degree;

	if (power < k) {
		/* Simple case. */
		Za_set_si (rop, 0);
		mpz_set_ui (&coeff [power], 1);
	}
	else {
		/* Must reduce result modulo generator polynomial. */
		Za_init (rp, &TEMP);
		Za_set_var_power (&TEMP, 1);
		Za_ipow (rop, &TEMP, power);
		Za_clear (&TEMP);
	}
}

/*
 * Addition in Z(a).
 */

	static
	void
Za_add (

struct calib_Za_obj *		rop,		/* IN/OUT: result */
const struct calib_Za_obj *	op1,		/* IN: first operand */
const struct calib_Za_obj *	op2		/* IN: second operand */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				res, coeff;
mpz_srcptr			coeff1, coeff2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	rp	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE rp) OR (op2 -> dom NE rp));

	coeff	= rop -> coeff;
	coeff1	= op1 -> coeff;
	coeff2	= op2 -> coeff;
	k	= rp -> gpoly -> degree;

	res = calib_new_Z_vector (k);

	for (i = 0; i < k; i++) {
		mpz_add (&res [i], &coeff1 [i], &coeff2 [i]);
	}

	calib_copy_Z_vector (coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Subtraction in Z(a).
 */

	static
	void
Za_sub (

struct calib_Za_obj *		rop,		/* IN/OUT: result */
const struct calib_Za_obj *	op1,		/* IN: first operand */
const struct calib_Za_obj *	op2		/* IN: second operand */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				res, coeff;
mpz_srcptr			coeff1, coeff2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	rp	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE rp) OR (op2 -> dom NE rp));

	coeff	= rop -> coeff;
	coeff1	= op1 -> coeff;
	coeff2	= op2 -> coeff;
	k	= rp -> gpoly -> degree;

	res = calib_new_Z_vector (k);

	for (i = 0; i < k; i++) {
		mpz_sub (&res [i], &coeff1 [i], &coeff2 [i]);
	}

	calib_copy_Z_vector (coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Negation in Z(a).
 */

	static
	void
Za_neg (

struct calib_Za_obj *		rop,
const struct calib_Za_obj *	op	/* IN: operand to negate */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				coeff;
mpz_srcptr			scoeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	rp	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE rp);

	coeff	= rop -> coeff;
	k	= rp -> gpoly -> degree;

	if (rop EQ op) {
		/* Negate in-place. */
		for (i = 0; i < k; i++) {
			mpz_neg (&coeff [i], &coeff [i]);
		}
	}
	else {
		scoeff	= op -> coeff;
		for (i = 0; i < k; i++) {
			mpz_neg (&coeff [i], &scoeff [i]);
		}
	}
}

/*
 * Multiplication in Z(a).
 */

	static
	void
Za_mul (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
const struct calib_Za_obj *	op1,	/* IN: first operand */
const struct calib_Za_obj *	op2	/* IN: second operand */
)
{
int				i, j, k;
const struct calib_Za_dom *	rp;
mpz_srcptr			coeff1, coeff2, gpoly;
mpz_srcptr			bj, lc;
mpz_ptr				res, coeff;
mpz_t				FACTOR, TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	rp	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE rp) OR (op2 -> dom NE rp));

	coeff	= rop -> coeff;
	coeff1	= op1 -> coeff;
	coeff2	= op2 -> coeff;
	k	= rp -> gpoly -> degree;
	gpoly	= &(rp -> gpoly -> coeff [0]);

	res = calib_new_Z_vector (k);

	bj = &coeff2 [k - 1];
	if (mpz_sgn (bj) EQ 0) {
		/* res is currently already completely zero. */
	}
	else if (mpz_cmp_ui (bj, 1) EQ 0) {
		calib_copy_Z_vector (res, coeff1, k);
	}
	else {
		for (i = 0; i < k; i++) {
			mpz_mul (&res [i], &coeff1 [i], bj);
		}
	}

	mpz_init (FACTOR);
	mpz_init (TEMP);

	for (j = k - 2; j >= 0; j--) {
		/* Multiply by a^1. */
		lc = &res [k - 1];
		if (mpz_sgn (lc) EQ 0) {
			/* No correction necessary, just shift. */
			for (i = k - 1; i > 0; i--) {
				mpz_set (&res [i], &res [i - 1]);
			}
			mpz_set_ui (&res [0], 0);
		}
		else if (mpz_cmp_ui (lc, 1) EQ 0) {
			/* Subtract generator polynomial while shifting. */
			for (i = k - 1; i > 0; i--) {
				mpz_sub (&res [i], &res [i - 1], &gpoly [i]);
			}
			mpz_neg (&res [0], &gpoly [0]);
		}
		else {
			/* Subtract lc*gpoly while shifting. */
			mpz_neg (FACTOR, lc);
			for (i = k - 1; i > 0; i--) {
				mpz_mul (TEMP, FACTOR, &gpoly [i]);
				mpz_add (&res [i], &res [i -1], TEMP);
			}
			mpz_mul (&res [0], FACTOR, &gpoly [0]);
		}

		/* Get next bj and do res += bj*a. */
		bj = &coeff2 [j];
		if (mpz_sgn (bj) EQ 0) {
			/* Nothing to add. */
		}
		else if (mpz_cmp_ui (bj, 1) EQ 0) {
			for (i = 0; i < k; i++) {
				mpz_add (&res [i], &res [i], &coeff1 [i]);
			}
		}
		else {
			for (i = 0; i < k; i++) {
				mpz_mul (TEMP, bj, &coeff1 [i]);
				mpz_add (&res [i], &res [i], TEMP);
			}
		}
	}

	mpz_clear (TEMP);
	mpz_clear (FACTOR);

	calib_copy_Z_vector (coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Multiply the given element of Z(a) by a GMP integer.
 */

	static
	void
Za_mul_z (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
const struct calib_Za_obj *	op1,	/* IN: first operand */
mpz_srcptr			op2	/* IN: second operand */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				dcoeff;
mpz_srcptr			scoeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	rp	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE rp);

	dcoeff	= rop -> coeff;
	scoeff	= op1 -> coeff;
	k	= rp -> gpoly -> degree;

	if (mpz_sgn (op2) EQ 0) {
		/* Multiplying by zero. */
		Za_set_si (rop, 0);
		return;
	}

	if (mpz_cmp_ui (op2, 1) EQ 0) {
		/* Multiplying by one. */
		Za_set (rop, op1);
		return;
	}

	for (i = 0; i < k; i++) {
		mpz_mul (&dcoeff [i], &scoeff [i], op2);
	}
}

/*
 * Multiply the given element of Z(a) by a.  This is a "shift and remainder"
 * operation.
 */

	static
	void
Za_mul_a (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
const struct calib_Za_obj *	op	/* IN: operand to multiply by a */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
const struct calib_Zx_obj *	gpoly;
mpz_ptr				res, coeff;
mpz_srcptr			scoeff, lc, gp;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	rp	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE rp);

	coeff	= rop -> coeff;
	scoeff	= op -> coeff;
	k	= rp -> gpoly -> degree;
	gpoly	= rp -> gpoly;
	gp	= gpoly -> coeff;

	res = calib_new_Z_vector (k);

	lc = &scoeff [k - 1];
	if (mpz_sgn (lc) EQ 0) {
		/* Simple shift is sufficient. */
		calib_copy_Z_vector (&res [1], &scoeff [0], k - 1);
	}
	else {
		/* Need to reduce while copying. */
		mpz_init (TEMP);
		for (i = 1; i < k; i++) {
			mpz_mul (TEMP, lc, &gp [i]);
			mpz_sub (&res [i], &scoeff [i - 1], TEMP);
		}
		mpz_mul (TEMP, lc, &gp [0]);
		mpz_neg (&res [0], TEMP);
		mpz_clear (TEMP);
	}
	calib_copy_Z_vector (coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Take the given Z(a) value to the given integer power.
 */

	static
	void
Za_ipow (

struct calib_Za_obj *		rop,		/* IN/OUT: result */
const struct calib_Za_obj *	op,		/* IN: value to exponentiate */
int				power		/* IN: integer power */
)
{
const struct calib_Za_dom *	rp;
struct calib_Za_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (power < 0));

	rp	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE rp);

	Za_init (rp, &RES);
	Za_init (rp, &PREG);
	Za_set_si (&RES, 1);
	Za_set (&PREG, op);

	while (power > 0) {
		if ((power & 1) NE 0) {
			Za_mul (&RES, &RES, &PREG);
		}
		Za_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	Za_set (rop, &RES);

	Za_clear (&PREG);
	Za_clear (&RES);
}

/*
 * Pseudo-inverse in Z(a).
 * Given A \in Z(a), compute B \in Z(a) and d \in Z such that A*B = d.
 */

	static
	void
Za_pinv (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
mpz_ptr				d,	/* IN/OUT: denominator of result */
const struct calib_Za_obj *	op	/* IN: operand to invert */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
const struct calib_Zx_dom *	Zx;
mpz_ptr				coeff;
struct calib_Zx_obj		A, G, XA, XB;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (d EQ NULL) OR (op EQ NULL));

	rp	= rop -> dom;
	coeff	= rop -> coeff;
	k	= rp -> gpoly -> degree;

	/* Type pun from Z(a) to Z[x]. */
	A.degree	= k - 1;
	A.size		= k;
	A.coeff		= op -> coeff;

	Zx = calib_get_Zx_dom ();

	Zx -> init (&G);
	Zx -> init (&XA);
	Zx -> init (&XB);

	/* Find XA * A + XB * gpoly = G, via extended Euclidean algorithm. */

	Zx -> extgcd (&G, &XA, &XB, &A, rp -> gpoly);

	CALIB_FATAL_ERROR_IF (G.degree > 0);
	if (G.degree < 0) {
		mpz_set_ui (d, 0);
	}
	else {
		mpz_set (d, &G.coeff [0]);
	}

	for (i = 0; i < k; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
	calib_copy_Z_vector (coeff, &XA.coeff [0], XA.degree + 1);

	Zx -> clear (&XB);
	Zx -> clear (&XA);
	Zx -> clear (&G);
}

/*
 * Divide the given element op1 of Z(a) by integer op2.
 * The division must be exact.
 */

	static
	void
Za_div_z_exact (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
const struct calib_Za_obj *	op1,	/* IN: first operand */
mpz_srcptr			op2	/* IN: second operand */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				dcoeff;
mpz_srcptr			scoeff;
mpz_t				REM;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	rp	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE rp);

	k	= rp -> gpoly -> degree;
	dcoeff	= rop -> coeff;
	scoeff	= op1 -> coeff;

	mpz_init (REM);
	for (i = 0; i < k; i++) {
		mpz_tdiv_qr (&dcoeff [i], REM, &scoeff [i], op2);
		CALIB_FATAL_ERROR_IF (mpz_sgn (REM) NE 0);
	}
	mpz_clear (REM);
}

/*
 * Compute both the content and primitive-part of the given A \in Z(a).
 *
 * This computation views the operand A as simply a vector of integers,
 * and completely ignores the modulus polynomial.
 */

	static
	void
Za_prim_part (

mpz_ptr				content,	/* IN/OUT: content */
struct calib_Za_obj *		ppart,		/* IN/OUT: primitive part */
const struct calib_Za_obj *	op		/* IN: operand */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_ptr				res, coeff;
mpz_srcptr			scoeff;
mpz_t				G;

	CALIB_FATAL_ERROR_IF ((content EQ NULL) OR
			      (ppart   EQ NULL) OR
			      (op      EQ NULL));

	rp	= ppart -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE rp);

	coeff	= ppart -> coeff;
	scoeff	= op -> coeff;
	k	= rp -> gpoly -> degree;

	res = calib_new_Z_vector (k);
	calib_copy_Z_vector (res, scoeff, k);

	mpz_init (G);

	for (i = 0; i < k; i++) {
		mpz_gcd (G, G, &res [i]);
		if (mpz_cmp_ui (G, 1) EQ 0) break;
	}

	if (mpz_sgn (G) EQ 0) {
		mpz_set_ui (G, 1);
	}
	else if (mpz_cmp_ui (G, 1) NE 0) {
		for (i = 0; i < k; i++) {
			mpz_divexact (&res [i], &res [i], G);
		}
	}
	mpz_set (content, G);
	mpz_clear (G);

	calib_copy_Z_vector (coeff, res, k);
	calib_free_Z_vector (res, k);
}

/*
 * Convert the given element of Z(b) into the corresponding element of Z(a).
 * It is intended that the gpoly's defining Z(a) and Z(b) be different.
 */

	static
	void
Za_cvZa (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
const struct calib_Za_obj *	op	/* IN: element to convert */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
const struct calib_Za_dom *	src_rp;
mpz_srcptr			scoeff;
struct calib_Za_obj		RES;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	rp	= rop -> dom;
	src_rp	= op -> dom;
	scoeff	= op -> coeff;
	k	= src_rp -> gpoly -> degree;

	Za_init (rp, &RES);

	for (i = k - 1; i >= 0; i--) {
		Za_mul_a (&RES, &RES);
		mpz_add (&RES.coeff [0], &RES.coeff [0], &scoeff [i]);
	}

	Za_set (rop, &RES);
	Za_clear (&RES);
}

/*
 * Return the degree of the given element of Z(a), i.e., the highest
 * non-zero power of the corresponding polynomial in the quotient
 * ring Z[a]/g(a).  The degree of the zero element is -1.
 */

	static
	int
Za_degree (

const struct calib_Za_obj *	op	/* IN: ring element to get degree of */
)
{
int	deg;
const struct calib_Za_dom *	rp;
mpz_srcptr			scoeff;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	rp	= op -> dom;
	scoeff	= op -> coeff;
	deg	= rp -> gpoly -> degree - 1;

	while ((deg >= 0) AND (mpz_sgn (&scoeff [deg]) EQ 0)) {
		--deg;
	}
	return (deg);
}

/*
 * Is the given element of Z(a) zero?
 */

	static
	calib_bool
Za_zerop (

const struct calib_Za_obj *	op	/* IN: ring element to test */
)
{
int		deg;

	deg	= Za_degree (op);
	if (deg < 0) {
		return (TRUE);
	}
	return (FALSE);
}


/*
 * Is the given element of Z(a) one?
 */

	static
	calib_bool
Za_onep (

const struct calib_Za_obj *	op	/* IN: ring element to test */
)
{
int	deg;

	deg = Za_degree (op);
	if ((deg EQ 0) AND (mpz_cmp_ui (&(op -> coeff [0]), 1) EQ 0)) {
		return (TRUE);
	}
	return (FALSE);
}

/*
 * Convert the given general expression into a member of Z(a) = Z[a]/g(a).
 * All of the variables in the given expression *MUST* match the given
 * variable name.
 */

	static
	void
Za_set_genrep (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
const struct calib_genrep *	op,	/* IN: expression to convert */
const char *			var	/* IN:  var for polynomial */
)
{
int				i;
const struct calib_Za_dom *	rp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_Za_obj		TEMP1, TEMP2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op  EQ NULL) OR (var EQ NULL));

	rp	= rop -> dom;

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Za_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Za_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		Za_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		CALIB_FATAL_ERROR_IF (strcmp (op -> u.var, var) NE 0);
		Za_set_var_power (rop, 1);
		break;

	case CALIB_GENREP_OP_ADD:
		Za_init (rp, &TEMP1);
		Za_init (rp, &TEMP2);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Za_set_genrep (&TEMP2, lp -> operand, var);
			Za_add (&TEMP1, &TEMP1, &TEMP2);
		}
		Za_set (rop, &TEMP1);
		Za_clear (&TEMP2);
		Za_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_MUL:
		Za_init (rp, &TEMP1);
		Za_init (rp, &TEMP2);
		Za_set_si (&TEMP1, 1);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Za_set_genrep (&TEMP2, lp -> operand, var);
			Za_mul (&TEMP1, &TEMP1, &TEMP2);
		}
		Za_set (rop, &TEMP1);
		Za_clear (&TEMP2);
		Za_clear (&TEMP1);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;
		/* This is a ring, not a field -- no inverses! */
		CALIB_FATAL_ERROR_IF (i < 0);

		gp = op -> u.ipow.base;

		CALIB_FATAL_ERROR_IF (gp EQ NULL);

		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			/* Optimize important special case. */
			Za_set_var_power (rop, i);
		}
		else {
			Za_init (rp, &TEMP2);
			Za_set_genrep (&TEMP2, gp, var);
			Za_ipow (rop, &TEMP2, i);
			Za_clear (&TEMP2);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
	}
}

/*
 * Convert the given element of Z(a) = Z[a]/g(a) into
 * general representation.  Use the variable name given.
 */

	static
	struct calib_genrep *
Za_to_genrep (

const struct calib_Za_obj *	op,	/* IN: element to convert */
const char *			var	/* IN: variable name to use */
)
{
int				i, k;
const struct calib_Za_dom *	rp;
mpz_srcptr			scoeff;
mpz_srcptr			value;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (var EQ NULL));

	rp	= op -> dom;
	scoeff	= op -> coeff;
	k	= rp -> gpoly -> degree;

	root = NULL;
	hookp = &root;

	for (i = k - 1; i >= 0; i--) {
		value = &scoeff [i];
		if (mpz_sgn (value) EQ 0) continue;
		vgp = calib_genrep_z (value);
		gp1 = calib_genrep_poly_term (vgp, var, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	return (gp1);
}

/*
 * Prototype Za domain object.
 */

static const struct calib_Za_dom	Za_dom = {
	.init		= Za_init,
	.clear		= Za_clear,
	.set		= Za_set,
	.set_si		= Za_set_si,
	.set_z		= Za_set_z,
	.set_q		= Za_set_q,
	.set_var_power	= Za_set_var_power,
	.add		= Za_add,
	.sub		= Za_sub,
	.neg		= Za_neg,
	.mul		= Za_mul,
	.mul_z		= Za_mul_z,
	.mul_a		= Za_mul_a,
	.ipow		= Za_ipow,
	.pinv		= Za_pinv,
	.div_z_exact	= Za_div_z_exact,
	.prim_part	= Za_prim_part,
	.cvZa		= Za_cvZa,
	.degree		= Za_degree,
	.zerop		= Za_zerop,
	.onep		= Za_onep,
	.set_genrep	= Za_set_genrep,
	.to_genrep	= Za_to_genrep,
};

/*
 * Construct the ring Z(a) = Z[a]/g(a), where g(a) is the given
 * monic, irreducible generator polynomial in Z[a].
 */

	struct calib_Za_dom *
calib_make_Za_dom (

const struct calib_Zx_obj *	gpoly	/* IN: generator polynomial */
)
{
int				k;
struct calib_Za_dom *		rp;
const struct calib_Zx_dom *	Zx;
mpz_srcptr			gcoeff;

	/* Verify that the given generator polynomial is monic,	*/
	/* and that it has degree AT LEAST 2.  If the degree is	*/
	/* less than 2, then remainders have degree at most 1	*/
	/* and there is no representation for the var "a"!	*/
	k	= gpoly -> degree;
	gcoeff	= gpoly -> coeff;
	CALIB_FATAL_ERROR_IF ((k < 1) OR (mpz_cmp_ui (&gcoeff [k], 1) NE 0));

	Zx = calib_get_Zx_dom ();

	rp = CALIB_NEW (struct calib_Za_dom);

	/* Initialize "most" things. */
	*rp = Za_dom;

	rp -> gpoly = Zx -> dup (gpoly);

	return (rp);
}

/*
 * Free up the given Z(a) ring description.
 */

	void
calib_free_Za_dom (

struct calib_Za_dom *	rp	/* IN: ring description to free */
)
{
const struct calib_Zx_dom *	Zx;

	if (rp NE NULL) {
		Zx = calib_get_Zx_dom ();
		Zx -> free (rp -> gpoly);
		memset (rp, 0, sizeof (*rp));
		free (rp);
	}
}
