/**********************************************************************

	$Id: Zax.c,v 1.1 2024/10/20 00:36:26 warme Exp $

	File:	Zax.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring of polynomials R[x] over the ring R = Z(a).

***********************************************************************

	Modification Log:

	a-1:	07/04/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Zax.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/GFpk.h"
#include "calib/GFpkx.h"
#include "calib/gmpmisc.h"
#include "calib/new.h"
#include "calib/Qa.h"
#include "calib/Qax.h"
#include "calib/types.h"
#include "calib/Za.h"
#include "calib/Zx.h"
#include <gmp.h>
#include <stdlib.h>
#include <string.h>

/* Standard "max" of two integers. */

#define _MAX(A,B)	(((A) > (B)) ? (A) : (B))

/*
 * Macros to allocate/zero a local polynomial of the given degree.
 */

#define ALLOCATE_LOCAL_POLY(R_OF_X, PNAME, DEGREE) \
	struct calib_Zax_obj	PNAME; \
	Zax_init_degree ((R_OF_X), &PNAME, (DEGREE))

#define FREE_LOCAL_POLY(PNAME) \
	Zax_clear (&PNAME)

/*
 * Macros to allocate/zero an array of local polynomials, each having
 * the given degree.
 */

#define ALLOCATE_LOCAL_POLY_ARRAY(R_OF_X, PNAME, NPOLY, DEGREE) \
	struct calib_Zax_obj	PNAME [NPOLY]; \
	Zax_init_array_degree ((R_OF_X), &PNAME [0], (NPOLY), (DEGREE))

#define FREE_LOCAL_POLY_ARRAY(PNAME, NPOLY) \
	Zax_clear_array(&PNAME [0], (NPOLY))

/*
 * Local Functions
 */

static void	Zax_add (struct calib_Zax_obj *		rop,
			 const struct calib_Zax_obj *	op1,
			 const struct calib_Zax_obj *	op2);
static void	Zax_alloc (struct calib_Zax_obj *	rop,
			   int				degree);
static void	Zax_clear (struct calib_Zax_obj *	x);
static void	Zax_clear_array (struct calib_Zax_obj *	parray,
				 int			npoly);
static void	Zax_copy_Za_vector (struct calib_Za_obj *	dst,
				    const struct calib_Za_obj *	src,
				    int				n);
static struct calib_Zax_obj *
		Zax_cvGFpkx (const struct calib_Zax_dom *	R_of_x,
			     const struct calib_GFpkx_obj *	gfpoly);
static struct calib_Zax_obj *
		Zax_cvQax (const struct calib_Zax_dom *	R_of_x,
			   const struct calib_Qax_obj *	qpoly,
			   mpz_srcptr			d);
static struct calib_Zax_obj *
		Zax_cvZax (const struct calib_Zax_dom *	R_of_x,
			   const struct calib_Zax_obj *	op);
static void	Zax_div (struct calib_Zax_obj *		quotient,
			 struct calib_Zax_obj *		remainder,
			 mpz_ptr			d,
			 const struct calib_Zax_obj *	a,
			 const struct calib_Zax_obj *	b);
static void	Zax_div_z_exact (struct calib_Zax_obj *		rop,
				 const struct calib_Zax_obj *	op1,
				 mpz_srcptr			op2);
static struct calib_Zax_obj *
		Zax_dup (const struct calib_Zax_obj *	op);
static void	Zax_eval (struct calib_Za_obj *		rop,
			  const struct calib_Zax_obj *	op,
			  const struct calib_Za_obj *	value);
static void	Zax_extgcd (struct calib_Zax_obj *		gcd,
			    struct calib_Zax_obj *		xa,
			    struct calib_Zax_obj *		xb,
			    const struct calib_Zax_obj *	a,
			    const struct calib_Zax_obj *	b);
static void	Zax_free (struct calib_Zax_obj *	poly);
static void	Zax_free_Za_vector (struct calib_Za_obj *	array,
				    int				n);
static void	Zax_init (const struct calib_Zax_dom *	R_of_x,
			  struct calib_Zax_obj *	x);
static void	Zax_init_array_degree (
				const struct calib_Zax_dom *	R_of_x,
				struct calib_Zax_obj *		parray,
				int				npoly,
				int				degree);
static void	Zax_init_degree (const struct calib_Zax_dom *	R_of_x,
				 struct calib_Zax_obj *		x,
				 int				degree);
static void	Zax_init_si (const struct calib_Zax_dom *	R_of_x,
			     struct calib_Zax_obj *		dst,
			     calib_si_t				src);
static void	Zax_ipow (struct calib_Zax_obj *	rop,
			  const struct calib_Zax_obj *	op,
			  int				power);
static void	Zax_mul (struct calib_Zax_obj *		rop,
			 const struct calib_Zax_obj *	op1,
			 const struct calib_Zax_obj *	op2);
static void	Zax_mul_z (struct calib_Zax_obj *	rop,
			   const struct calib_Zax_obj *	op1,
			   mpz_srcptr			op2);
static void	Zax_neg (struct calib_Zax_obj *		rop,
			 const struct calib_Zax_obj *	op);
static struct calib_Za_obj *
		Zax_new_Za_vector (const struct calib_Za_dom *	rp,
				   int				n);
static calib_bool
		Zax_onep (const struct calib_Zax_obj *	op);
static void	Zax_realloc_copy_poly (
				struct calib_Zax_obj *		poly,
				const struct calib_Za_obj *	coeff,
				int				degree);
static void	Zax_set (struct calib_Zax_obj *		rop,
			 const struct calib_Zax_obj *	op);
static void	Zax_set_genrep (struct calib_Zax_obj *		rop,
				const struct calib_genrep *	op,
				const char *			xvar,
				const char *			avar);
static void	Zax_set_q (struct calib_Zax_obj *	rop,
			   mpq_srcptr			op);
static void	Zax_set_si (struct calib_Zax_obj *	rop,
			    calib_si_t			op);
static void	Zax_set_var_power (struct calib_Zax_obj *	rop,
				   int				power);
static void	Zax_set_z (struct calib_Zax_obj *	rop,
			   mpz_srcptr			op);
static void	Zax_sub (struct calib_Zax_obj *		rop,
			 const struct calib_Zax_obj *	op1,
			 const struct calib_Zax_obj *	op2);
static struct calib_genrep *
		Zax_to_genrep (const struct calib_Zax_obj *	op,
			       const char *			xvar,
			       const char *			avar);
static calib_bool
		Zax_zerop (const struct calib_Zax_obj *	op);

/*
 * Initialize the given polynomial.  Its value after initialization is zero.
 */

	static
	inline
	void
Zax_init (

const struct calib_Zax_dom *	R_of_x,	/* IN: polynomial ring */
struct calib_Zax_obj *		x	/* IN/OUT: polynomial to initialize */
)
{
	CALIB_FATAL_ERROR_IF ((R_of_x EQ NULL) OR (x EQ NULL));

	x -> degree	= -1;
	x -> size	= 0;
	x -> dom	= R_of_x;
	x -> coeff	= NULL;
}

/*
 * Initialize the given polynomial to zero, but provide it with a
 * coefficient buffer sufficient to hold a polynomial of the given degree.
 */

	static
	inline
	void
Zax_init_degree (

const struct calib_Zax_dom *	R_of_x,	/* IN: polynomial ring/domain */
struct calib_Zax_obj *		x,	/* IN: polynomial to initialize */
int				degree	/* IN: maximum degree */
)
{
int				size;
const struct calib_Za_dom *	cr;

	CALIB_FATAL_ERROR_IF ((R_of_x EQ NULL) OR
			      (x      EQ NULL) OR
			      (degree <  0));

	cr	= R_of_x -> cr;

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	x -> degree	= -1;
	x -> size	= size;
	x -> dom	= R_of_x;
	x -> coeff	= Zax_new_Za_vector (cr, size);
}

/*
 * Initialize the given polynomial.  Its value after initialization
 * is the given integer value.
 */

	static
	void
Zax_init_si (

const struct calib_Zax_dom *	R_of_x,	/* IN: polynomial ring */
struct calib_Zax_obj *		dst,	/* IN/OUT: polynomial to initialize */
calib_si_t			src	/* IN: initial value */
)
{
int				k;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		coeff;

	if (src EQ 0) {
		Zax_init (R_of_x, dst);
	}
	else {
		cr	= R_of_x -> cr;
		k	= cr -> gpoly -> degree;

		coeff = Zax_new_Za_vector (cr, k);
	
		cr -> set_si (&coeff [0], src);

		dst -> degree	= 0;
		dst -> size	= k;
		dst -> dom	= R_of_x;
		dst -> coeff	= coeff;
	}
}

/*
 * Initialize the given array of polynomials to zero, but provide each
 * with a coefficient buffer sufficient to hold a polynomial of the
 * given degree.
 *
 * Although it is tempting to allocate all of the coefficients for all of
 * the polynomials into one big long vector, we cannot do so because the
 * coefficient vector for each polynomial must be separately free-able,
 * and re-allocatable.
 */

	static
	inline
	void
Zax_init_array_degree (

const struct calib_Zax_dom *	R_of_x,	/* IN: polynomial ring/domain */
struct calib_Zax_obj *		parray,	/* IN/OUT: array of polynomials */
int				npoly,	/* IN: num polynomials in array */
int				degree	/* IN: max degree of each polynomial */
)
{
int				i, size;
const struct calib_Za_dom *	cr;
struct calib_Zax_obj *		pp;

	CALIB_FATAL_ERROR_IF ((R_of_x EQ NULL) OR
			      (parray EQ NULL) OR
			      (npoly  <= 0   ) OR
			      (degree <  0   ));

	cr	= R_of_x -> cr;

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	pp = &parray [0];
	for (i = 0; i < npoly; i++, pp++) {
		pp -> degree	= -1;
		pp -> size	= size;
		pp -> dom	= R_of_x;
		pp -> coeff	= Zax_new_Za_vector (cr, size);
	}
}

/*
 * Force the given polynomial to have buffer space sufficient to hold
 * a polynomial of at least the given degree.  This does not change
 * the value currently residing in this polynomial.
 */

	static
	void
Zax_alloc (

struct calib_Zax_obj *		rop,	/* IN/OUT: poly to alloc buffer for */
int				degree	/* IN: degree desired */
)
{
int				size, dsize;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		coeff;
struct calib_Za_obj *		scoeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (degree < -1));

	R_of_x	= rop -> dom;
	cr	= R_of_x -> cr;

	size	= degree + 1;
	dsize	= rop -> size;

	if (dsize < size) {
		scoeff	= rop -> coeff;
		coeff	= Zax_new_Za_vector (cr, size);
		Zax_copy_Za_vector (coeff, scoeff, dsize);
		Zax_free_Za_vector (scoeff, dsize);
		rop -> coeff = coeff;
		rop -> size  = size;
	}
}

/*
 * Clear out the given polynomial.
 */

	static
	inline
	void
Zax_clear (

struct calib_Zax_obj *	x	/* IN/OUT: polynomial to clear */
)
{
int			size;
struct calib_Za_obj *	coeff;

	CALIB_FATAL_ERROR_IF (x EQ NULL);
	coeff = x -> coeff;
	if (coeff NE NULL) {
		size = x -> size;
		CALIB_FATAL_ERROR_IF (size <= 0);
		Zax_free_Za_vector (coeff, size);
	}
	x -> degree	= -1;
	x -> size	= 0;
	x -> dom	= NULL;
	x -> coeff	= NULL;
}

/*
 * Clear out the given array of polynomials.
 */

	static
	inline
	void
Zax_clear_array (

struct calib_Zax_obj * parray,	/* IN: array of polynomials to clear */
int			npoly	/* IN: number of polynomials in parray */
)
{
int		i;

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (parray EQ NULL);

	for (i = 0; i < npoly; i++) {
		Zax_clear (&parray [i]);
	}
}

/*
 * Set the given destination polynomial from the given coefficient
 * array.  Reallocate the destination polynomial's coefficient array,
 * if necessary.
 */

	static
	inline
	void
Zax_realloc_copy_poly (

struct calib_Zax_obj *		poly,	/* IN: destination polynomial */
const struct calib_Za_obj *	coeff,	/* IN: source coefficient array */
int				degree	/* IN: degree to set */
)
{
int				cur_size, new_size;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (degree <  -1));

	poly -> degree	= degree;

	if (degree < 0) {
		/* We allow (coeff EQ NULL) here (but do not require). */
		return;
	}

	CALIB_FATAL_ERROR_IF (coeff EQ NULL);

	R_of_x	= poly -> dom;
	cr	= R_of_x -> cr;

	new_size = degree + 1;
	if (new_size > 0) {
		cur_size = poly -> size;
		if (new_size > cur_size) {
			if (poly -> coeff NE NULL) {
				Zax_free_Za_vector (poly -> coeff, cur_size);
			}
			poly -> coeff = Zax_new_Za_vector (cr, new_size);
			poly -> size = new_size;
		}
		Zax_copy_Za_vector (poly -> coeff, coeff, new_size);
	}
}

/*
 * Allocate and initialize a vector of Za objects.
 */

	static
	struct calib_Za_obj *
Zax_new_Za_vector (

const struct calib_Za_dom *	cr,	/* IN: Z(a) coefficient ring */
int				n	/* IN: number of elements */
)
{
int			i;
struct calib_Za_obj *	array;

	CALIB_FATAL_ERROR_IF (cr EQ NULL);

	array = NULL;

	if (n > 0) {
		array = CALIB_NEWA (n, struct calib_Za_obj);
		for (i = 0; i < n; i++) {
			cr -> init (cr, &array [i]);
		}
	}

	return (array);
}

/*
 * Free a vector of Za objects.
 */

	static
	void
Zax_free_Za_vector (

struct calib_Za_obj *		array,	/* IN/OUT: Array of Za objects */
int				n	/* IN: Number of elements */
)
{
int				i;
const struct calib_Za_dom *	cr;

	CALIB_FATAL_ERROR_IF (n < 0);

	if (array NE NULL) {
		cr	= array [0].dom;
		for (i = 0; i < n; i++) {
			cr -> clear (&array [i]);
		}
		free (array);
	}
	else {
		CALIB_FATAL_ERROR_IF (n NE 0);
	}
}

/*
 * Copy a vector of Za objects.
 */

	static
	void
Zax_copy_Za_vector (

struct calib_Za_obj *		dst,	/* IN/OUT: Destination vector */
const struct calib_Za_obj *	src,	/* IN: Source vector */
int				n	/* IN: Number of elements */
)
{
int		i;
const struct calib_Za_dom *	cr;

	if (n > 0) {
		cr = dst -> dom;
		CALIB_FATAL_ERROR_IF (src -> dom NE cr);
		for (i = 0; i < n; i++) {
			cr -> set (&dst [i], &src [i]);
		}
	}
}

/*
 * Copy the given polynomial.
 */

	static
	void
Zax_set (

struct calib_Zax_obj *		rop,		/* IN/OUT: result */
const struct calib_Zax_obj *	op		/* IN: source operand */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {
		Zax_realloc_copy_poly (rop, op -> coeff, op -> degree);
	}
}

/*
 * Set the given polynomial to the given integer value.
 */

	static
	void
Zax_set_si (

struct calib_Zax_obj *		rop,		/* IN: polynomial to set */
calib_si_t			op		/* IN: integer value to set */
)
{
const struct calib_Za_dom *	cr;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	if (op EQ 0) {
		rop -> degree = -1;
	}
	else {
		Zax_alloc (rop, 0);
		cr = rop -> dom -> cr;
		cr -> set_si (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given polynomial to the given GMP integer value.
 */

	static
	void
Zax_set_z (

struct calib_Zax_obj *	rop,		/* IN: polynomial to set */
mpz_srcptr		op		/* IN: integer value to set */
)
{
const struct calib_Za_dom *	cr;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL)OR (op EQ NULL));

	if (mpz_sgn (op) EQ 0) {
		rop -> degree = -1;
	}
	else {
		Zax_alloc (rop, 0);
		cr = rop -> dom -> cr;
		cr -> set_z (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given polynomial to the given GMP rational value.
 * (This rational must have a denominator of 1.)
 */

	static
	void
Zax_set_q (

struct calib_Zax_obj *		rop,		/* IN: polynomial to set */
mpq_srcptr			op		/* IN: integer value to set */
)
{
const struct calib_Za_dom *	cr;
mpz_srcptr			num;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	/* Denominator *must* be 1. */
	CALIB_FATAL_ERROR_IF (mpz_cmp_ui (mpq_denref (op), 1) NE 0);

	num = mpq_numref (op);

	if (mpz_sgn (num) EQ 0) {
		rop -> degree = -1;
	}
	else {
		Zax_alloc (rop, 0);
		cr = rop -> dom -> cr;
		cr -> set_z (&(rop -> coeff [0]), num);
		rop -> degree = 0;
	}
}

/*
 * Set the given Z(a)[x] polynomial to x**power.
 */

	static
	void
Zax_set_var_power (

struct calib_Zax_obj *		rop,		/* IN/OUT: result */
int				power		/* IN: power to set */
)
{
int				i;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (power < 0));

	cr	= rop -> dom -> cr;

	Zax_alloc (rop, power);
	coeff	= rop -> coeff;

	for (i = 0; i < power; i++) {
		cr -> set_si (&coeff [i], 0);
	}
	cr -> set_si (&coeff [power], 1);
	rop -> degree = power;
}

/*
 * Addition of two polynomials over Z(a).
 */

	static
	void
Zax_add (

struct calib_Zax_obj *		rop,		/* IN/OUT: result */
const struct calib_Zax_obj *	op1,		/* IN: first operand */
const struct calib_Zax_obj *	op2		/* IN: second operand */
)
{
int				i, rdeg, deg1, deg2, rsize;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		res;
struct calib_Za_obj *		rp;
const struct calib_Za_obj *	coeff1;
const struct calib_Za_obj *	coeff2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	R_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE R_of_x) OR
			      (op2 -> dom NE R_of_x));

	cr	= R_of_x -> cr;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = Zax_new_Za_vector (cr, rsize);

	coeff1 = &(op1 -> coeff [0]);
	coeff2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, coeff1++, coeff2++, rp++) {
		if (i <= deg1) {
			if (i <= deg2) {
				cr -> add (rp, coeff1, coeff2);
			}
			else {
				cr -> set (rp, coeff1);
			}
		}
		else {
			cr -> set (rp, coeff2);
		}
	}
	while ((rdeg >= 0) AND (cr -> zerop (&res [rdeg]))) {
		--rdeg;
	}

	Zax_realloc_copy_poly (rop, res, rdeg);
	Zax_free_Za_vector (res, rsize);
}

/*
 * Subtraction of two polynomials over Z(a).
 */

	static
	void
Zax_sub (

struct calib_Zax_obj *		rop,		/* IN/OUT: result */
const struct calib_Zax_obj *	op1,		/* IN: first operand */
const struct calib_Zax_obj *	op2		/* IN: second operand */
)
{
int				i, rdeg, deg1, deg2, rsize;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		res;
struct calib_Za_obj *		rp;
const struct calib_Za_obj *	coeff1;
const struct calib_Za_obj *	coeff2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	R_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE R_of_x) OR
			      (op2 -> dom NE R_of_x));

	cr	= R_of_x -> cr;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = Zax_new_Za_vector (cr, rsize);

	coeff1 = &(op1 -> coeff [0]);
	coeff2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, coeff1++, coeff2++, rp++) {
		if (i <= deg1) {
			if (i <= deg2) {
				cr -> sub (rp, coeff1, coeff2);
			}
			else {
				cr -> set (rp, coeff1);
			}
		}
		else {
			cr -> neg (rp, coeff2);
		}
	}
	while ((rdeg >= 0) AND (cr -> zerop (&res [rdeg]))) {
		--rdeg;
	}

	Zax_realloc_copy_poly (rop, res, rdeg);
	Zax_free_Za_vector (res, rsize);
}

/*
 * Negate the given Z(a)[x] polynomial.
 */

	static
	void
Zax_neg (

struct calib_Zax_obj *		rop,		/* IN/OUT: result */
const struct calib_Zax_obj *	op		/* IN: operand to negate */
)
{
int				i, deg;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		dstp;
const struct calib_Za_obj *	srcp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	R_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE R_of_x);

	cr	= R_of_x -> cr;

	if (rop EQ op) {
		deg	= rop -> degree;
		dstp	= rop -> coeff;
		for (i = 0; i <= deg; i++, dstp++) {
			cr -> neg (dstp, dstp);
		}
	}
	else {
		deg	= op -> degree;
		srcp	= op -> coeff;
		Zax_alloc (rop, deg);
		dstp	= rop -> coeff;
		for (i = 0; i <= deg; i++, dstp++, srcp++) {
			cr -> neg (dstp, srcp);
		}
		rop -> degree = deg;
	}
}

/*
 * Multiplcation of two polynomials over Z(a).
 */

	static
	void
Zax_mul (

struct calib_Zax_obj *		rop,		/* IN/OUT: result */
const struct calib_Zax_obj *	op1,		/* IN: first operand */
const struct calib_Zax_obj *	op2		/* IN: second operand */
)
{
int				i, j, rdeg, deg1, deg2, rsize;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		res;
struct calib_Za_obj *		rp1;
struct calib_Za_obj *		rp2;
struct calib_Za_obj *		coeff1;
struct calib_Za_obj *		coeff2;
struct calib_Za_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	R_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE R_of_x) OR
			      (op2 -> dom NE R_of_x));

	cr	= R_of_x -> cr;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	if ((deg1 < 0) OR (deg2 < 0)) {
		/* Multipying by zero. */
		rop -> degree = -1;
		return;
	}

	rdeg	= deg1 + deg2;

	rsize = rdeg + 1;
	res = Zax_new_Za_vector (cr, rsize);

	cr -> init (cr, &TEMP);

	coeff1 = &(op1 -> coeff [0]);
	rp1 = &res [0];
	for (i = 0; i <= deg1; i++, coeff1++) {
		rp2 = rp1;
		coeff2 = &(op2 -> coeff [0]);
		for (j = 0; j <= deg2; j++, coeff2++, rp2++) {
			cr -> mul (&TEMP, coeff1, coeff2);
			cr -> add (rp2, rp2, &TEMP);
		}
		++rp1;
	}
	while ((rdeg >= 0) AND (cr -> zerop (&res [rdeg]))) {
		--rdeg;
	}

	Zax_realloc_copy_poly (rop, res, rdeg);

	cr -> clear (&TEMP);

	Zax_free_Za_vector (res, rsize);
}

/*
 * Multiply the given polynomial by the given integer.
 */

	static
	void
Zax_mul_z (

struct calib_Zax_obj *		rop,		/* IN/OUT: result */
const struct calib_Zax_obj *	op1,		/* IN: first operand */
mpz_srcptr			op2		/* IN: second operand */
)
{
int				i, deg, size;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		coeff;
struct calib_Za_obj *		res;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	R_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE R_of_x);

	cr	= R_of_x -> cr;
	deg	= op1 -> degree;
	size	= deg + 1;
	coeff	= op1 -> coeff;

	res = Zax_new_Za_vector (cr, size);

	for (i = 0; i <= deg; i++) {
		cr -> mul_z (&res [i], &coeff [i], op2);
	}

	Zax_realloc_copy_poly (rop, res, deg);

	Zax_free_Za_vector (res, size);
}

/*
 * Take the given polynomial to the given integer power.
 */

	static
	void
Zax_ipow (

struct calib_Zax_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_Zax_obj *	op,	/* IN: polynomial to exponentiate */
int				power	/* IN: integer power */
)
{
const struct calib_Zax_dom *	R_of_x;
struct calib_Zax_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (power < 0));

	R_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE R_of_x);

	Zax_init (R_of_x, &RES);
	Zax_init (R_of_x, &PREG);

	Zax_set_si (&RES, 1);
	Zax_set (&PREG, op);

	while (power > 0) {
		if ((power & 1) NE 0) {
			Zax_mul (&RES, &RES, &PREG);
		}
		Zax_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	Zax_set (rop, &RES);

	Zax_clear (&PREG);
	Zax_clear (&RES);
}

/*
 * Create a dynamically allocated copy of the given polynomial.
 */

	static
	struct calib_Zax_obj *
Zax_dup (

const struct calib_Zax_obj *	op	/* IN: polynomial to duplicate */
)
{
int				deg, size;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Zax_obj *		dst;
struct calib_Za_obj *		cp;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	R_of_x	= op -> dom;
	cr	= R_of_x -> cr;

	deg = op -> degree;
	size = deg + 1;

	dst = CALIB_NEW (struct calib_Zax_obj);
	cp = NULL;
	if (size > 0) {
		cp = Zax_new_Za_vector (cr, size);
		Zax_copy_Za_vector (cp, op -> coeff, size);
	}
	dst -> degree	= deg;
	dst -> size	= size;
	dst -> dom	= R_of_x;
	dst -> coeff	= cp;

	return (dst);
}

/*
 * Free the given polynomial, which must have been dynamically allocated.
 */

	static
	void
Zax_free (

struct calib_Zax_obj *	poly		/* IN/OUT: ring element to free */
)
{
	if (poly NE NULL) {
		Zax_clear (poly);
		free (poly);
	}
}

/*
 * Evaluate the given poly(value).
 */

	static
	void
Zax_eval (

struct calib_Za_obj *		rop,	/* IN/OUT: result */
const struct calib_Zax_obj *	op,	/* IN: polynomial to evaulate */
const struct calib_Za_obj *	value	/* IN: variable value to use */
)
{
int				i, deg;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		coeff;
struct calib_Za_obj		SUM;

	CALIB_FATAL_ERROR_IF ((rop   EQ NULL) OR
			      (op    EQ NULL) OR
			      (value EQ NULL));

	deg	= op -> degree;
	R_of_x	= op -> dom;
	coeff	= op -> coeff;
	cr	= R_of_x -> cr;

	CALIB_FATAL_ERROR_IF ((rop -> dom NE cr) OR (value -> dom NE cr));

	cr -> init (cr, &SUM);

	for (i = deg; i >= 0; i--) {
		cr -> mul (&SUM, &SUM, value);
		cr -> add (&SUM, &SUM, &coeff [i]);
	}
	cr -> set (rop, &SUM);

	cr -> clear (&SUM);
}

/*
 * The division algorith (with remainder):
 * Given polynomials a, b \in Z(a)[x], find polynomials q, r \in Z(a)[x]
 * and constant d \in Z(a) such that:
 *
 *	d*a = q*b + r	AND	degree (r) < degree (b)
 *			AND	d is minimal
 *
 * Yes, this is messy because the coefficients are in Z(a) which is a ring,
 * not a field.
 */

	static
	void
Zax_div (

struct calib_Zax_obj *		quotient,	/* IN/OUT: quotient */
struct calib_Zax_obj *		remainder,	/* IN/OUT: remainder */
mpz_ptr				d,		/* IN/OUT: denominator */
const struct calib_Zax_obj *	a,		/* IN: first operand */
const struct calib_Zax_obj *	b		/* IN: second operand */
)
{
int				i, j, k, adeg, bdeg, qdeg, rdeg;
int				size, total_size;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Za_obj *		quot;
struct calib_Za_obj *		rem;
struct calib_Za_obj *		qp;
struct calib_Za_obj *		rp;
struct calib_Za_obj *		zp;
struct calib_Za_obj *		lcr;
struct calib_Za_obj		TMP, BLC_INV, BLC_DENOM, FACTOR;
const struct calib_Za_obj *	ap;
const struct calib_Za_obj *	bp;
mpz_t				dval, denom;

	CALIB_FATAL_ERROR_IF ((a EQ NULL) OR (b EQ NULL));

	R_of_x	= a -> dom;

	CALIB_FATAL_ERROR_IF (b -> dom NE R_of_x);
	CALIB_FATAL_ERROR_IF ((quotient NE NULL) AND
			      (quotient -> dom NE R_of_x));
	CALIB_FATAL_ERROR_IF ((remainder NE NULL) AND
			      (remainder -> dom NE R_of_x));

	cr = R_of_x -> cr;
	k  = cr -> gpoly -> degree;

	adeg	= a -> degree;
	bdeg	= b -> degree;

	/* Local coefficient buffers. */
	size = adeg + 1;
	total_size = 2 * size;
	zp = Zax_new_Za_vector (cr, total_size);
	quot	= zp;		zp += size;
	rem	= zp;		zp += size;

	mpz_init (dval);

	ap = &(a -> coeff [0]);
	bp = &(b -> coeff [0]);

	if (bdeg <= 0) {
		/* Division by a constant.  Do not allow division by zero. */
		CALIB_FATAL_ERROR_IF ((bdeg < 0) OR (cr -> zerop (bp)));

		cr -> init (cr, &BLC_INV);
		cr -> pinv (&BLC_INV, dval, bp);
		rp = &rem [0];
		for (i = 0; i <= adeg; i++, ap++, rp++) {
			cr -> mul (rp, ap, &BLC_INV);
		}
		if (quotient NE NULL) {
			Zax_realloc_copy_poly (quotient, rem, adeg);
		}
		if (remainder NE NULL) {
			remainder -> degree = -1;
		}
		if (d NE NULL) {
			mpz_set (d, dval);
		}

		mpz_clear (dval);
		Zax_free_Za_vector (quot, total_size);
		cr -> clear (&BLC_INV);

		return;
	}

	cr -> init (cr, &TMP);
	cr -> init (cr, &BLC_INV);
	cr -> init (cr, &BLC_DENOM);
	cr -> init (cr, &FACTOR);
	mpz_init (denom);

	qdeg = adeg - bdeg;
	if (qdeg < -1) {
		qdeg = -1;
	}

	rdeg = adeg;
	Zax_copy_Za_vector (&rem [0], ap, rdeg + 1);

	/* Compute multiplicative pseudo-inverse of lc(b). */
	cr -> pinv (&BLC_INV, denom, &(b -> coeff [bdeg]));
	cr -> set_z (&BLC_DENOM, denom);

	mpz_set_ui (dval, 1);

	while (rdeg >= bdeg) {
		j = rdeg - bdeg;
		lcr = &rem [rdeg];
		cr -> mul (&FACTOR, lcr, &BLC_INV);
		for (i = qdeg; i > j; i--) {
			cr -> mul (&quot [i], &quot [i], &BLC_DENOM);
		}
		for (i = rdeg - 1; i >= 0; i--) {
			cr -> mul (&rem [i], &rem [i], &BLC_DENOM);
		}
		mpz_mul (dval, dval, denom);
		qp = &quot [j];
		cr -> set (qp, &FACTOR);

		rp = &rem [rdeg - 1];
		bp = &(b -> coeff [bdeg - 1]);
		for (i = bdeg - 1; i >= 0; i--, rp -= k, bp -= k) {
			cr -> mul (&TMP, &FACTOR, bp);
			cr -> sub (rp, rp, &TMP);
		}
		--rdeg;
		rp = &rem [rdeg];
		while ((rdeg >= 0) AND (cr -> zerop (rp))) {
			--rdeg;
			--rp;
		}
	}

	if (quotient NE NULL) {
		Zax_realloc_copy_poly (quotient, quot, qdeg);
	}

	if (remainder NE NULL) {
		Zax_realloc_copy_poly (remainder, rem, rdeg);
	}

	if (d NE NULL) {
		mpz_set (d, dval);
	}

	mpz_clear (denom);
	mpz_clear (dval);
	Zax_free_Za_vector (quot, total_size);
}

/*
 * Divide the given polynomial by the given integer z.
 * The division must be exact.
 */

	static
	void
Zax_div_z_exact (

struct calib_Zax_obj *		rop,		/* IN/OUT: result */
const struct calib_Zax_obj *	op1,		/* IN: first operand */
mpz_srcptr			op2		/* IN: second operand */
)
{
int				i, deg, size;
const struct calib_Zax_dom *	R_of_x;		/* IN: polynomial ring */
const struct calib_Za_dom *	cr;
const struct calib_Za_obj *	src;
struct calib_Za_obj *		res;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	R_of_x	= rop -> dom;
	cr	= R_of_x -> cr;
	deg	= op1 -> degree;
	size	= deg + 1;
	src	= op1 -> coeff;
	res	= Zax_new_Za_vector (cr, size);
	for (i = 0; i < size; i++) {
		cr -> div_z_exact (&res [i], &src [i], op2);
	}
	Zax_realloc_copy_poly (rop, res, deg);
	Zax_free_Za_vector (res, size);
}

/*
 * The Extended Euclidean Algorithm for polynomials in R[x],
 * where R = Z(a).  Given polynomials a, b \in R[x],
 * compute polynomials g, xa, xb \in R[x] such that
 *
 * 1.	a is a multiple of g,
 * 2.	b is a multiple of g,
 * 3.	xa * a + xb * b = g.
 *
 * These three conditions imply that g = gcd(a,b).  The polynomial g
 * is monic (unless a = b = 0, in which case g = 0).
 */

	static
	void
Zax_extgcd (

struct calib_Zax_obj *		gcd,		/* IN/OUT: GCD */
struct calib_Zax_obj *		xa,		/* IN/OUT: multiple for a */
struct calib_Zax_obj *		xb,		/* IN/OUT: multiple for b */
const struct calib_Zax_obj *	a,		/* IN: first operand */
const struct calib_Zax_obj *	b		/* IN: second operand */
)
{
int				i, adeg, bdeg, maxdeg, deg;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
struct calib_Zax_obj *		ap;
struct calib_Zax_obj *		bp;
struct calib_Zax_obj *		rp;
struct calib_Zax_obj *		qp;
struct calib_Zax_obj *		m1;
struct calib_Zax_obj *		m2;
struct calib_Zax_obj *		m3;
struct calib_Zax_obj *		n1;
struct calib_Zax_obj *		n2;
struct calib_Zax_obj *		n3;
struct calib_Zax_obj *		TEMP1;
struct calib_Zax_obj *		TEMP2;
struct calib_Zax_obj *		temp_ptr;
struct calib_Za_obj *		cp;
struct calib_Zax_obj		dpoly;
struct calib_Za_obj		D, LCA_INV;
mpz_t				dval, lca_denom;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR
			      (xa  EQ NULL) OR
			      (xb  EQ NULL) OR
			      (a   EQ NULL) OR
			      (b   EQ NULL));

	R_of_x	= gcd -> dom;

	CALIB_FATAL_ERROR_IF ((xa -> dom NE R_of_x) OR
			      (xb -> dom NE R_of_x) OR
			      (a  -> dom NE R_of_x) OR
			      (b  -> dom NE R_of_x));

	cr	= R_of_x -> cr;

	adeg	= a -> degree;
	bdeg	= b -> degree;
	maxdeg	= _MAX (adeg, bdeg);

	ALLOCATE_LOCAL_POLY_ARRAY (R_of_x, TMP, 12, maxdeg);

	temp_ptr = &TMP [0];
	ap = temp_ptr++;
	bp = temp_ptr++;
	rp = temp_ptr++;
	qp = temp_ptr++;
	m1 = temp_ptr++;
	m2 = temp_ptr++;
	m3 = temp_ptr++;
	n1 = temp_ptr++;
	n2 = temp_ptr++;
	n3 = temp_ptr++;
	TEMP1 = temp_ptr++;
	TEMP2 = temp_ptr++;

	Zax_set (ap, a);
	Zax_set (bp, b);

	/* ap = m1*a + n1*b */
	Zax_set_si (m1, 1);
	Zax_set_si (n1, 0);

	/* bp = m2*a + n2*b */
	Zax_set_si (m2, 0);
	Zax_set_si (n2, 1);

	cr -> init (cr, &D);
	mpz_init (dval);

	while (bp -> degree >= 0) {
		Zax_div (qp, rp, dval, ap, bp);

		cr -> set_z (&D, dval);

		/* r = d*ap - q*bp				*/
		/*   = d*(m1*a + n1*b) - q*(m2*a + n2*b)	*/
		/*   = (d*m1 - q*m2)*a + (d*n1 - q*n2)*b	*/
		/*   = m3*a	     + m3*b,			*/
		/* so we compute:				*/
		/*	m3 = d*m1 - q*m2			*/
		/*	n3 = d*n1 - q*n2			*/

		dpoly.degree	= 0;
		dpoly.size	= 1;
		dpoly.coeff	= &D;

		Zax_mul (TEMP1, &dpoly, m1);
		Zax_mul (TEMP2, qp, m2);
		Zax_sub (m3, TEMP1, TEMP2);
		Zax_mul (TEMP1, &dpoly,  n1);
		Zax_mul (TEMP2, qp, n2);
		Zax_sub (n3, TEMP1, TEMP2);

		temp_ptr = ap;
		ap = bp;
		bp = rp;
		rp = temp_ptr;

		temp_ptr = m1;
		m1 = m2;
		m2 = m3;
		m3 = temp_ptr;

		temp_ptr = n1;
		n1 = n2;
		n2 = n3;
		n3 = temp_ptr;
	}

	/* We have ap = M1*a + N1*b. */

	adeg = ap -> degree;
	if (adeg >= 0) {
		/* Regularize the result.  This means that lc(ap) is	*/
		/* forced to be a pure integer (i.e., an element of	*/
		/* degree zero in Z(a)).				*/
		/* First compute pseudo-inverse of lc(ap).  This yields	*/
		/* LCA_INV*lc(ap) == lca_denom (mod gpoly).		*/
		cr -> init (cr, &LCA_INV);
		mpz_init (lca_denom);
		cr -> pinv (&LCA_INV, lca_denom, &(ap -> coeff [adeg]));

		if (NOT cr -> onep (&LCA_INV)) {
			cp = &(ap -> coeff [0]);
			for (i = 0; i <= adeg; i++, cp++) {
				cr -> mul (cp, cp, &LCA_INV);
			}

			cp = &(m1 -> coeff [0]);
			deg = m1 -> degree;
			for (i = 0; i <= deg; i++, cp++) {
				cr -> mul (cp, cp, &LCA_INV);
			}

			cp = &(n1 -> coeff [0]);
			deg = n1 -> degree;
			for (i = 0; i <= deg; i++, cp++) {
				cr -> mul (cp, cp, &LCA_INV);
			}
		}
		mpz_clear (lca_denom);
		cr -> clear (&LCA_INV);
	}

	Zax_set (gcd, ap);
	Zax_set (xa, m1);
	Zax_set (xb, n1);

	mpz_clear (dval);
	cr -> clear (&D);

	FREE_LOCAL_POLY_ARRAY (TMP, 12);
}

/*
 * Convert the given polynomial from Z(a)[x] to Z(b)[x].  The gpoly
 * defining Z(a) and Z(b) are intentionally different.
 */

	static
	struct calib_Zax_obj *
Zax_cvZax (

const struct calib_Zax_dom *	R_of_x,	/* IN: destination polynomial ring */
const struct calib_Zax_obj *	op	/* IN: polynomial to convert */
)
{
int				i, deg, size;
const struct calib_Za_dom *	cr1;
struct calib_Zax_obj *		new_poly;
struct calib_Za_obj *		coeff;
struct calib_Za_obj *		srcp;
struct calib_Za_obj *		dstp;

	CALIB_FATAL_ERROR_IF ((R_of_x EQ NULL) OR (op EQ NULL));

	deg	= op -> degree;

	cr1	= R_of_x -> cr;

	size = deg + 1;
	coeff = NULL;
	if (size > 0) {
		coeff = Zax_new_Za_vector (cr1, size);
	}

	srcp = op -> coeff;
	dstp = coeff;
	for (i = 0; i <= deg; i++) {
		cr1 -> cvZa (dstp, srcp);
		++dstp;
		++srcp;
	}

	new_poly = CALIB_NEW (struct calib_Zax_obj);
	new_poly -> degree	= deg;
	new_poly -> size	= size;
	new_poly -> dom		= R_of_x;
	new_poly -> coeff	= coeff;

	return (new_poly);
}

/*
 * Convert the given polynomial from GF(p^k)[x] into Z(a)[x] form.
 */

	static
	struct calib_Zax_obj *
Zax_cvGFpkx (

const struct calib_Zax_dom *	R_of_x,	/* IN: integer polynomial ring */
const struct calib_GFpkx_obj *	gfpoly	/* IN: polynomial to convert */
)
{
int				i, j, k, deg, size;
const struct calib_Zx_obj *	gpoly;
const struct calib_GFpkx_dom *	gfpkx;
const struct calib_GFpk_dom *	gfpk;
struct calib_Zax_obj *		new_poly;
struct calib_Za_obj *		coeff;
mpz_ptr				zp;
const struct calib_GFpk_obj *	gfcoeff;
mpz_srcptr			p;
mpz_t				cval, half_p;

	CALIB_FATAL_ERROR_IF ((R_of_x EQ NULL) OR (gfpoly EQ NULL));

	deg = gfpoly -> degree;

	/* Verify that the generator polynomial is identical in	*/
	/* both domains (equivalent mod p).			*/
	gpoly	= R_of_x -> cr -> gpoly;
	gfpkx	= gfpoly -> dom;
	gfpk	= gfpkx -> cf;
	k	= gfpk -> k;
	CALIB_FATAL_ERROR_IF (gpoly -> degree NE k);
	p = gfpk -> cf -> p;
	mpz_init (cval);
	for (i = 0; i <= k; i++) {
		mpz_sub (cval,
			 &(gpoly -> coeff [i]),
			 &(gfpk -> coeff [i]));
		mpz_mod (cval, cval, p);
		CALIB_FATAL_ERROR_IF (mpz_sgn (cval) NE 0);
	}
	mpz_clear (cval);

	size = deg + 1;
	coeff = NULL;
	if (size > 0) {
		coeff = Zax_new_Za_vector (R_of_x -> cr, size);
	}

	mpz_init (half_p);
	mpz_tdiv_q_2exp (half_p, p, 1);

	for (i = 0; i <= deg; i++) {
		gfcoeff = &(gfpoly -> coeff [i]);
		zp = &coeff [i].coeff [0];
		for (j = 0; j < k; j++) {
			mpz_set (zp, &(gfcoeff -> coeff [j]));
			if (mpz_cmp (zp, half_p) > 0) {
				mpz_sub (zp, zp, p);
			}
			++zp;
		}
	}

	mpz_clear (half_p);

	new_poly = CALIB_NEW (struct calib_Zax_obj);
	new_poly -> degree	= deg;
	new_poly -> size	= size;
	new_poly -> dom		= R_of_x;
	new_poly -> coeff	= coeff;

	return (new_poly);
}

/*
 * Convert the given polynomial from Q(a)[x] into Z(a)[x] form by
 * multiplying all coefficients by d to clear the denominators.
 * It is a fatal error if any non-integral coefficients remain.
 */

	static
	struct calib_Zax_obj *
Zax_cvQax (

const struct calib_Zax_dom *	R_of_x,	/* IN: integer polynomial ring */
const struct calib_Qax_obj *	qpoly,	/* IN: rational polynomial */
mpz_srcptr			d	/* IN: common denominator */
)
{
int				i, j, k, deg, deg2, size;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
const struct calib_Zx_obj *	gpoly;
const struct calib_Zx_obj *	zpoly;
struct calib_Zax_obj *		new_poly;
struct calib_Za_obj *		coeff;
struct calib_Qa_obj *		scoeff;
mpq_ptr				qcoeffs;
mpz_ptr				zp;
mpq_t				cval;

	CALIB_FATAL_ERROR_IF ((R_of_x EQ NULL) OR
			      (qpoly  EQ NULL) OR
			      (d      EQ NULL));

	deg	= qpoly -> degree;
	K_of_x	= qpoly -> dom;
	cf	= K_of_x -> cf;

	/* Verify that the generator polynomial is identical in	*/
	/* both domains.					*/
	gpoly = R_of_x -> cr -> gpoly;
	zpoly = K_of_x -> cf -> gpoly;
	k  = gpoly -> degree;
	CALIB_FATAL_ERROR_IF (k NE zpoly -> degree);
	for (i = 0; i < k; i++) {
		CALIB_FATAL_ERROR_IF (mpz_cmp (&(gpoly -> coeff [i]),
					 &(zpoly -> coeff [i])) NE 0);
	}

	size = deg + 1;
	coeff = NULL;
	if (size > 0) {
		coeff = Zax_new_Za_vector (R_of_x -> cr, size);
	}

	mpq_init (cval);
	for (i = 0; i <= deg; i++) {
		zp = &coeff [i].coeff [0];
		scoeff	= &(qpoly -> coeff [i]);
		deg2	= scoeff -> degree;
		qcoeffs	= cf -> get_coeffs (scoeff);
		for (j = 0; j <= deg2; j++) {
			mpq_set (cval, &qcoeffs [j]);
			calib_mpq_mul_z (cval, cval, d);
			/* Denominator must now be 1. */
			CALIB_FATAL_ERROR_IF (mpz_cmp_ui (mpq_denref (cval), 1) NE 0);
			mpz_set (zp, mpq_numref (cval));
			++zp;
		}
		calib_free_Q_vector (qcoeffs, deg2 + 1);
	}
	mpq_clear (cval);

	new_poly = CALIB_NEW (struct calib_Zax_obj);
	new_poly -> degree	= deg;
	new_poly -> size	= size;
	new_poly -> dom		= R_of_x;
	new_poly -> coeff	= coeff;

	return (new_poly);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically zero.
 */

	static
	calib_bool
Zax_zerop (

const struct calib_Zax_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg >= 0) {
		return (FALSE);
	}

	CALIB_FATAL_ERROR_IF (deg < -1);

	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically one.
 */

	static
	calib_bool
Zax_onep (

const struct calib_Zax_obj *	op	/* IN: Polynomial to test */
)
{
int				deg;
const struct calib_Za_dom *	cr;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg NE 0) {
		return (FALSE);
	}

	cr = op -> dom -> cr;

	if (cr -> onep (&(op -> coeff [0])) EQ 0) {
		return (TRUE);
	}

	return (FALSE);
}

/*
 * Convert the given general expression into a member of Z(a)[x].
 * All of the variables in the given expression *MUST* match either the
 * given xvar, or avar.  (xvar is the "main" variable of the polynomial,
 * whereas avar is the variable for the coefficient ring.)
 */

	static
	void
Zax_set_genrep (

struct calib_Zax_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_genrep *	op,	/* IN: expression to convert */
const char *			xvar,	/* IN: main var of polynomial */
const char *			avar	/* IN: var for coefficient ring */
)
{
int				i;
const struct calib_Zax_dom *	R_of_x;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
const struct calib_Za_dom *	cr;
struct calib_Zax_obj		POLY;
struct calib_Zax_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop    EQ NULL) OR
			      (op     EQ NULL) OR
			      (xvar   EQ NULL) OR
			      (avar   EQ NULL));

	R_of_x	= rop -> dom;
	cr	= R_of_x -> cr;

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Zax_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Zax_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		Zax_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		if (strcmp (op -> u.var, xvar) EQ 0) {
			Zax_set_var_power (rop, 1);
		}
		else if (strcmp (op -> u.var, avar) EQ 0) {
			Zax_alloc (rop, 0);
			rop -> degree	= 0;
			cr -> set_var_power (rop -> coeff, 1);
		}
		else {
			CALIB_FATAL_ERROR;
		}
		break;

	case CALIB_GENREP_OP_ADD:
		Zax_init (R_of_x, &POLY);
		Zax_init (R_of_x, &TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zax_set_genrep (&TEMP,
					lp -> operand,
					xvar,
					avar);
			Zax_add (&POLY, &POLY, &TEMP);
		}
		Zax_set (rop, &POLY);

		Zax_clear (&TEMP);
		Zax_clear (&POLY);
		break;

	case CALIB_GENREP_OP_MUL:
		Zax_init (R_of_x, &POLY);
		Zax_init (R_of_x, &TEMP);
		Zax_set_si (&POLY, 1);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zax_set_genrep (&TEMP,
					lp -> operand,
					xvar,
					avar);
			Zax_mul (&POLY, &POLY, &TEMP);
		}
		Zax_set (rop, &POLY);

		Zax_clear (&TEMP);
		Zax_clear (&POLY);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;
		CALIB_FATAL_ERROR_IF (i < 0);

		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);

		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			if (strcmp (gp -> u.var, xvar) EQ 0) {
				Zax_set_var_power (rop, i);
			}
			else if (strcmp (gp -> u.var, avar) EQ 0) {
				Zax_alloc (rop, 0);
				rop -> degree = 0;
				cr -> set_var_power (rop -> coeff, i);
			}
			else {
				CALIB_FATAL_ERROR;
			}
		}
		else {
			Zax_init (R_of_x, &TEMP);
			Zax_set_genrep (&TEMP, gp, xvar, avar);
			Zax_ipow (rop, &TEMP, i);
			Zax_clear (&TEMP);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Convert the given element of R[x] (where R = Z(a)) into general
 * representation.  The generated expression uses xvar for the
 * "main" polynomial variable, whereas avar is used for the
 * variable in the coefficient ring R.
 */

	static
	struct calib_genrep *
Zax_to_genrep (

const struct calib_Zax_obj *	op,	/* IN: polynomial to convert */
const char *			xvar,	/* IN: main var of polynomial */
const char *			avar	/* IN: var for coefficient ring */
)
{
int				i, degree;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Za_dom *	cr;
const struct calib_Za_obj *	coeff;
const struct calib_Za_obj *	valp;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (xvar EQ NULL) OR (avar EQ NULL));

	R_of_x	= op -> dom;
	cr	= R_of_x -> cr;

	degree	= op -> degree;
	coeff	= op -> coeff;

	root = NULL;
	hookp = &root;

	for (i = degree; i >= 0; i--) {
		valp = &coeff [i];
		if (cr -> zerop (valp)) continue;
		vgp = cr -> to_genrep (valp, avar);
		gp1 = calib_genrep_poly_term (vgp, xvar, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	return (gp1);
}

/*
 * Prototype Zax domain.
 */

static const struct calib_Zax_dom	Zax_dom = {
	.init			= Zax_init,
	.init_degree		= Zax_init_degree,
	.init_si		= Zax_init_si,
	.alloc			= Zax_alloc,
	.clear			= Zax_clear,
	.set			= Zax_set,
	.set_si			= Zax_set_si,
	.set_z			= Zax_set_z,
	.set_q			= Zax_set_q,
	.set_var_power		= Zax_set_var_power,
	.add			= Zax_add,
	.sub			= Zax_sub,
	.neg			= Zax_neg,
	.mul			= Zax_mul,
	.mul_z			= Zax_mul_z,
	.ipow			= Zax_ipow,
	.dup			= Zax_dup,
	.free			= Zax_free,
	.eval			= Zax_eval,
	.div			= Zax_div,
	.div_z_exact		= Zax_div_z_exact,
	.extgcd			= Zax_extgcd,
	.cvZax			= Zax_cvZax,
	.cvGFpkx		= Zax_cvGFpkx,
	.cvQax			= Zax_cvQax,
	.zerop			= Zax_zerop,
	.onep			= Zax_onep,
	.set_genrep		= Zax_set_genrep,
	.to_genrep		= Zax_to_genrep,
};

/*
 * Construct the ring of polynomials R[x], where R is the given ring Z(a).
 */

	struct calib_Zax_dom *
calib_make_Zax_dom (

const struct calib_Za_dom *	R	/* IN: coefficient ring */
)
{
struct calib_Zax_dom *	R_of_x;

	R_of_x = CALIB_NEW (struct calib_Zax_dom);

	/* Initialize "most" things. */
	*R_of_x	= Zax_dom;

	R_of_x -> cr		= R;

	return (R_of_x);
}

/*
 * Free up the given Z(a)[x] ring description.
 */

	void
calib_free_Zax_dom (

struct calib_Zax_dom *	R_of_x	/* IN: Z(a)[x] ring description */
)
{
	if (R_of_x NE NULL) {
		/* No subordinate data to free. */
		memset (R_of_x, 0, sizeof (*R_of_x));
		free (R_of_x);
	}
}
