/**********************************************************************

	$Id: Zp.c,v 1.1 2024/10/20 00:36:26 warme Exp $

	File:	Zp.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The field of integers modulo p, where p is a prime
	of at most 32 bits.  The "object" is a GMP integer.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Zp.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/random.h"
#include "calib/types.h"
#include <gmp.h>
#include <stdlib.h>
#include <string.h>

/*
 * Local Functions
 */

static void	Zp_add (const struct calib_Zp_dom *	f,
			mpz_ptr				rop,
			mpz_srcptr			op1,
			mpz_srcptr			op2);
static void	Zp_inv (const struct calib_Zp_dom *	f,
			mpz_ptr				rop,
			mpz_srcptr			op);
static void	Zp_ipow (const struct calib_Zp_dom *	f,
			 mpz_ptr			rop,
			 mpz_srcptr			op,
			 int				power);
static void	Zp_mul (const struct calib_Zp_dom *	f,
			mpz_ptr				rop,
			mpz_srcptr			op1,
			mpz_srcptr			op2);
static void	Zp_neg (const struct calib_Zp_dom *	f,
			mpz_ptr				rop,
			mpz_srcptr			op);
static void	Zp_set_genrep (const struct calib_Zp_dom *	f,
			       mpz_ptr				rop,
			       const struct calib_genrep *	op);
static void	Zp_set_q (const struct calib_Zp_dom *	f,
			  mpz_ptr			rop,
			  mpq_srcptr			op);
static void	Zp_set_random (const struct calib_Zp_dom *	f,
			       mpz_ptr				rop,
			       struct calib_Random *		randp);
static void	Zp_set_si (const struct calib_Zp_dom *	f,
			   mpz_ptr			rop,
			   calib_si_t			op);
static void	Zp_set_z (const struct calib_Zp_dom *	f,
			  mpz_ptr			rop,
			  mpz_srcptr			op);
static void	Zp_sub (const struct calib_Zp_dom *	f,
			mpz_ptr				rop,
			mpz_srcptr			op1,
			mpz_srcptr			op2);
static struct calib_genrep *
		Zp_to_genrep (const struct calib_Zp_dom *	f,
			      mpz_srcptr			op);

/*
 * Convert the given integer into the corresponding member of Zp.
 */

	static
	void
Zp_set_si (

const struct calib_Zp_dom *	f,		/* IN: description of field */
mpz_ptr				rop,		/* IN/OUT: result */
calib_si_t			op		/* IN: integer to convert */
)
{
mpz_t		TEMP;

	CALIB_FATAL_ERROR_IF ((f EQ NULL) OR (rop EQ NULL));

	mpz_init_set_si (TEMP, op);
	mpz_fdiv_r (rop, TEMP, f -> p);
	mpz_clear (TEMP);
}

/*
 * Convert the given GMP integer into the corresponding member of Zp.
 */

	static
	void
Zp_set_z (

const struct calib_Zp_dom *	f,		/* IN: description of field */
mpz_ptr				rop,		/* IN/OUT: result */
mpz_srcptr			op		/* IN: integer to convert */
)
{
	mpz_fdiv_r (rop, op, f -> p);
}

/*
 * Convert the given GMP integer into the corresponding member of Zp.
 */

	static
	void
Zp_set_q (

const struct calib_Zp_dom *	f,		/* IN: description of field */
mpz_ptr				rop,		/* IN/OUT: result */
mpq_srcptr			op		/* IN: rational to convert */
)
{
mpz_t	DEN_INV;

	CALIB_FATAL_ERROR_IF ((f EQ NULL) OR (rop EQ NULL) OR (op EQ NULL));

	mpz_init (DEN_INV);
	Zp_inv (f, DEN_INV, mpq_denref (op));
	Zp_mul (f, rop, mpq_numref (op), DEN_INV);
	mpz_clear (DEN_INV);
}

/*
 * Addition in Zp.
 */

	static
	void
Zp_add (

const struct calib_Zp_dom *	f,		/* IN: description of field */
mpz_ptr				rop,		/* IN/OUT: result */
mpz_srcptr			op1,		/* IN: first operand */
mpz_srcptr			op2		/* IN: second operand */
)
{
mpz_t		TEMP;

	mpz_init (TEMP);
	mpz_add (TEMP, op1, op2);
	mpz_fdiv_r (rop, TEMP, f -> p);
	mpz_clear (TEMP);
}

/*
 * Subtraction in Zp.
 */

	static
	void
Zp_sub (

const struct calib_Zp_dom *	f,		/* IN: description of field */
mpz_ptr				rop,		/* IN/OUT: result */
mpz_srcptr			op1,		/* IN: first operand */
mpz_srcptr			op2		/* IN: second operand */
)
{
mpz_t		TEMP;

	mpz_init (TEMP);
	mpz_sub (TEMP, op1, op2);
	mpz_fdiv_r (rop, TEMP, f -> p);
	mpz_clear (TEMP);
}

/*
 * Negation in Zp.
 */

	static
	void
Zp_neg (

const struct calib_Zp_dom *	f,		/* IN: description of field */
mpz_ptr				rop,		/* IN/OUT: result */
mpz_srcptr			op		/* IN: operand to negate */
)
{
mpz_t		TEMP;

	mpz_init (TEMP);
	mpz_neg (TEMP, op);
	mpz_fdiv_r (rop, TEMP, f -> p);
	mpz_clear (TEMP);
}

/*
 * Multiplication in Zp.
 */

	static
	void
Zp_mul (

const struct calib_Zp_dom *	f,		/* IN: description of field */
mpz_ptr				rop,		/* IN/OUT: result */
mpz_srcptr			op1,		/* IN: first operand */
mpz_srcptr			op2		/* IN: second operand */
)
{
mpz_t		TEMP;

	mpz_init (TEMP);
	mpz_mul (TEMP, op1, op2);
	mpz_fdiv_r (rop, TEMP, f -> p);
	mpz_clear (TEMP);
}

/*
 * Compute a**pow (mod p).
 */

	static
	void
Zp_ipow (

const struct calib_Zp_dom *	f,	/* IN: Description of this field */
mpz_ptr				rop,	/* IN/OUT: Result */
mpz_srcptr			op,	/* IN: Operand to exponentiate */
int				power	/* IN: Power to take */
)
{
	CALIB_FATAL_ERROR_IF ((f EQ NULL) OR (rop EQ NULL) OR (op EQ NULL));

	/* This supports negative exponents when the modulus is prime. */
	mpz_powm_ui (rop, op, power, f -> p);
}

/*
 * Multiplicative Inverse in Zp.
 */

	static
	void
Zp_inv (

const struct calib_Zp_dom *	f,		/* IN: description of field */
mpz_ptr				rop,		/* IN/OUT: result */
mpz_srcptr			op		/* IN: operand to invert */
)
{
int		success;

	success = mpz_invert (rop, op, f -> p);
	CALIB_FATAL_ERROR_IF (NOT success);
}

/*
 * Set rop to be a random Zp value (selected more or less uniformly).
 */

	static
	void
Zp_set_random (

const struct calib_Zp_dom *	f,	/* IN: description of field */
mpz_ptr				rop,	/* IN/OUT: result */
struct calib_Random *		randp	/* IN/OUT: random number generator */
					/*	   to use */
)
{
calib_int64u		word;
mpz_srcptr		p;
mpz_t			TEMP;

	CALIB_FATAL_ERROR_IF ((f EQ NULL) OR (rop EQ NULL) OR (randp EQ NULL));

	p	= f -> p;

	mpz_init (TEMP);

	do {
		word = calib_random_u64 (randp);

		mpz_mul_2exp (TEMP, TEMP, 64);
		mpz_add_ui (TEMP, TEMP, word);
	} while (p -> _mp_size > TEMP [0]._mp_size);

	mpz_fdiv_r (rop, TEMP, p);

	mpz_clear (TEMP);
}

/*
 * Convert the given genrep expression (which must be either an integer
 * or rational number) into the corresponding member of Zp.
 */

	static
	void
Zp_set_genrep (

const struct calib_Zp_dom *	f,	/* IN: description of field */
mpz_ptr				rop,	/* IN/OUT: result */
const struct calib_genrep *	op	/* IN: expression to convert */
)
{
struct calib_genrep_list *	lp;
mpz_t		A, B;

	CALIB_FATAL_ERROR_IF ((f EQ NULL) OR (rop EQ NULL) OR (op EQ NULL));

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Zp_set_si (f, rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Zp_set_z (f, rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		Zp_set_q (f, rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_ADD:
		mpz_init (A);
		mpz_init (B);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zp_set_genrep (f, B, lp -> operand);
			Zp_add (f, A, A, B);
		}

		mpz_set (rop, A);
		mpz_clear (B);
		mpz_clear (A);
		break;

	case CALIB_GENREP_OP_MUL:
		mpz_init (A);
		mpz_init (B);

		mpz_set_ui (A, 1);
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zp_set_genrep (f, B, lp -> operand);
			Zp_mul (f, A, A, B);
		}

		mpz_set (rop, A);
		mpz_clear (B);
		mpz_clear (A);
		break;

	case CALIB_GENREP_OP_IPOW:
		mpz_init (A);
		Zp_set_genrep (f, A, op -> u.ipow.base);
		Zp_ipow (f, rop, A, op -> u.ipow.power);
		mpz_clear (A);
		break;

	default:
		CALIB_FATAL_ERROR;
	}
}

/*
 * Convert the given element of Zp into general representation.
 */

	static
	struct calib_genrep *
Zp_to_genrep (

const struct calib_Zp_dom *	f,	/* IN: description of field */
mpz_srcptr			op	/* IN: element of Zp to convert */
)
{
struct calib_genrep *	gp1;
mpz_t			TEMP;

	mpz_init_set (TEMP, op);
	mpz_fdiv_r (TEMP, TEMP, f -> p);
	gp1 = calib_genrep_z (TEMP);
	mpz_clear (TEMP);

	return (gp1);
}

/*
 * Prototype Zp domain object.
 */

static const struct calib_Zp_dom	Zp_dom = {
	.set_si		= Zp_set_si,
	.set_z		= Zp_set_z,
	.set_q		= Zp_set_q,
	.add		= Zp_add,
	.sub		= Zp_sub,
	.neg		= Zp_neg,
	.mul		= Zp_mul,
	.ipow		= Zp_ipow,
	.inv		= Zp_inv,
	.set_random	= Zp_set_random,
	.set_genrep	= Zp_set_genrep,
	.to_genrep	= Zp_to_genrep,
};

/*
 * Construct the field for Zp.
 */

	struct calib_Zp_dom *
calib_make_Zp_dom (

mpz_srcptr		p	/* IN: prime modulus */
)
{
struct calib_Zp_dom *		f;

	/* FIXME -- test primality of p? */

	f = CALIB_NEW (struct calib_Zp_dom);

	/* Initialize "most" things. */
	*f = Zp_dom;

	mpz_init_set (f -> p, p);

	return (f);
}

/*
 * Construct the field for Zp.
 */

	struct calib_Zp_dom *
calib_make_Zp_dom_ui (

unsigned long		p	/* IN: prime modulus */
)
{
struct calib_Zp_dom *	f;
mpz_t			P;

	mpz_init_set_ui (P, p);

	f = calib_make_Zp_dom (P);

	mpz_clear (P);

	return (f);
}

/*
 * Free up the given Zp field description.
 */

	void
calib_free_Zp_dom (

struct calib_Zp_dom *	f	/* IN: Zp field description */
)
{
	if (f NE NULL) {
		mpz_clear (f -> p);
		memset (f, 0, sizeof (*f));
		free (f);
	}
}
