/**********************************************************************

	$Id: Zpx.c,v 1.1 2024/10/20 00:36:26 warme Exp $

	File:	Zpx.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring Zp[x] of polynomials over the field Zp, the integers
	modulo a prime p.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Zpx.h"

#include "berle.h"
#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/Qa.h"
#include "calib/random.h"
#include "calib/Zx.h"
#include <gmp.h>
#include <stdlib.h>
#include <string.h>

/*
 * Default settings object for newly-created Zpx domains.
 *
 * API permits these to be changed, which affects future creation of
 * new Zpx domains.
 */

struct calib_Zpx_settings	calib_Zpx_default_settings = {

	/* Default to Berlekamp factoring method. */
	.factor_method		= CALIB_ZPX_FACTOR_METHOD_BERLEKAMP,
};

/*
 * Local Functions
 */

static void	Zpx_add (struct calib_Zpx_obj *		rop,
			 const struct calib_Zpx_obj *	op1,
			 const struct calib_Zpx_obj *	op2);
static void	Zpx_alloc (struct calib_Zpx_obj *	rop,
			   int				degree);
static void	Zpx_clear (struct calib_Zpx_obj *	x);
static void	Zpx_clear_array (struct calib_Zpx_obj *	parray,
				 int			npoly);
static struct calib_Zpx_factor *
		Zpx_compute_square_free_factorization (
				struct calib_Zpx_obj *		poly);
static int	Zpx_count_multiplicity (
				struct calib_Zpx_obj *		poly,
				const struct calib_Zpx_obj *	factor);
static struct calib_Zpx_factor *
		Zpx_ddf (const struct calib_Zpx_obj *		sqfpoly,
			 struct calib_Random *			randp);
static void	Zpx_derivative (struct calib_Zpx_obj *		rop,
				const struct calib_Zpx_obj *	op);
static void	Zpx_div (struct calib_Zpx_obj *		quotient,
			 struct calib_Zpx_obj *		remainder,
			 const struct calib_Zpx_obj *	a,
			 const struct calib_Zpx_obj *	b);
static struct calib_Zpx_obj *
		Zpx_dup (const struct calib_Zpx_obj * op);
static void	Zpx_eval (mpz_ptr			rop,
			  const struct calib_Zpx_obj *	poly,
			  mpz_srcptr			value);
static void	Zpx_extgcd (struct calib_Zpx_obj *		gcd,
			    struct calib_Zpx_obj *		xa,
			    struct calib_Zpx_obj *		xb,
			    const struct calib_Zpx_obj *	a,
			    const struct calib_Zpx_obj *	b);
static struct calib_Zpx_factor *
		Zpx_factor (const struct calib_Zpx_obj *	poly);
static struct calib_Zpx_factor *
		Zpx_factor_one_square_free_poly (
				const struct calib_Zpx_obj *	sqfpoly,
				struct calib_Random *		randp);
static struct calib_Zpx_factor *
		Zpx_factor_square_free (const struct calib_Zpx_obj *	poly);
static struct calib_genrep *
		Zpx_factors_to_genrep (
				const struct calib_Zpx_factor *	factors,
				const char *			var);
static struct calib_Zpx_factor *
		Zpx_finish_sqf (const struct calib_Zpx_factor *	sqfactors);
static void	Zpx_free (struct calib_Zpx_obj * poly);
static void	Zpx_free_factors (struct calib_Zpx_factor * factors);
static void	Zpx_gcd (struct calib_Zpx_obj *		gcd,
			 const struct calib_Zpx_obj *	a,
			 const struct calib_Zpx_obj *	b);
static void	Zpx_init (const struct calib_Zpx_dom *	K_of_x,
			  struct calib_Zpx_obj *	x);
static void	Zpx_init_array_degree (
				const struct calib_Zpx_dom *	K_of_x,
				struct calib_Zpx_obj *		parray,
				int				npoly,
				int				degree);
static void	Zpx_init_degree (const struct calib_Zpx_dom *	K_of_x,
				 struct calib_Zpx_obj *		x,
				 int				degree);
static void	Zpx_ipow (struct calib_Zpx_obj *	rop,
			  const struct calib_Zpx_obj *	op,
			  int				power);
static void	Zpx_monicize (struct calib_Zpx_obj *	poly);
static void	Zpx_mul (struct calib_Zpx_obj *		rop,
			 const struct calib_Zpx_obj *	op1,
			 const struct calib_Zpx_obj *	op2);
static void	Zpx_mul_z (struct calib_Zpx_obj *	rop,
			   const struct calib_Zpx_obj *	op1,
			   mpz_srcptr			op2);
static void	Zpx_neg (struct calib_Zpx_obj *		rop,
			 const struct calib_Zpx_obj *	op);
static calib_bool
		Zpx_onep (const struct calib_Zpx_obj *	op);
static void	Zpx_poly_mul_mod_poly (
				struct calib_Zpx_obj *		rop,
				const struct calib_Zpx_obj *	op1,
				const struct calib_Zpx_obj *	op2,
				const struct calib_Zpx_obj *	modpoly);
static void	Zpx_poly_to_int_pow_mod_poly (
				struct calib_Zpx_obj *		rop,
				const struct calib_Zpx_obj *	op,
				mpz_srcptr			exponent,
				const struct calib_Zpx_obj *	modpoly);
static void	Zpx_print (const struct calib_Zpx_obj * op);
static int	Zpx_print_nnl (const struct calib_Zpx_obj * op, int col);
static void	Zpx_random_poly (struct calib_Zpx_obj *		poly,
				 int				d,
				 struct calib_Random *		randp);
static void	Zpx_realloc_copy_poly (struct calib_Zpx_obj *	poly,
				       mpz_srcptr		coeff,
				       int			degree);
static void	Zpx_resultant (mpz_ptr				result,
			       const struct calib_Zpx_obj *	a,
			       const struct calib_Zpx_obj *	b);
static void	Zpx_set (struct calib_Zpx_obj *		rop,
			  const struct calib_Zpx_obj *	op);
static void	Zpx_set_genrep (struct calib_Zpx_obj *		rop,
				const struct calib_genrep *	op,
				const char *			var);
static void	Zpx_set_q (struct calib_Zpx_obj *	rop,
			   mpq_srcptr			op);
static void	Zpx_set_Qa (struct calib_Zpx_obj *		rop,
			    const struct calib_Qa_obj *		op);
static void	Zpx_set_si (struct calib_Zpx_obj *		rop,
			    calib_si_t				op);
static void	Zpx_set_var_power (struct calib_Zpx_obj *	rop,
				   int				power);
static void	Zpx_set_z (struct calib_Zpx_obj *	rop,
			   mpz_srcptr			op);
static void	Zpx_set_Zx (struct calib_Zpx_obj *		rop,
			    const struct calib_Zx_obj *		op);
static struct calib_Zpx_factor **
		Zpx_split_factors (struct calib_Zpx_factor **		hookp,
				   const struct calib_Zpx_obj *	poly,
				   int				d,
				   struct calib_Random *	randp);
static void	Zpx_sub (struct calib_Zpx_obj *		rop,
			 const struct calib_Zpx_obj *	op1,
			 const struct calib_Zpx_obj *	op2);
static struct calib_genrep *
		Zpx_to_genrep (const struct calib_Zpx_obj *	op,
			       const char *			var);
static calib_bool
		Zpx_zerop (const struct calib_Zpx_obj *	op);

/* Standard "max" of two integers. */

#define _MAX(A,B)	(((A) > (B)) ? (A) : (B))

/*
 * Macros to allocate/zero a local polynomial of the given degree.
 *
 * Note that since the underlying mpz_t uses memory that is dynamically
 * allocated from the heap, there is little reason to try and do any
 * fancy stack allocation here.  We put only the "struct calib_Zpx_obj" guys
 * on the stack, and dynamically allocate the vectors of __mpz_struct's.
 */

#define ALLOCATE_LOCAL_POLY(K_OF_X, PNAME, DEGREE) \
	struct calib_Zpx_obj	PNAME; \
	Zpx_init_degree ((K_OF_X), &PNAME, (DEGREE))

#define FREE_LOCAL_POLY(PNAME) \
	Zpx_clear (&PNAME)

/*
 * Macros to allocate/zero an array of local polynomials, each having
 * the given degree.
 */

#define ALLOCATE_LOCAL_POLY_ARRAY(K_OF_X, PNAME, NPOLY, DEGREE) \
	struct calib_Zpx_obj	PNAME [NPOLY]; \
	Zpx_init_array_degree ((K_OF_X), &PNAME [0], (NPOLY), (DEGREE))

#define FREE_LOCAL_POLY_ARRAY(PNAME, NPOLY) \
	Zpx_clear_array (&PNAME [0], (NPOLY))

/*
 * Initialize the given polynomial.  Its value after initialization is zero.
 */

	static
	inline
	void
Zpx_init (

const struct calib_Zpx_dom *	K_of_x,	/* IN: polynomial domain */
struct calib_Zpx_obj *		x	/* IN/OUT: polynomial to initialize */
)
{
	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR (x EQ NULL));

	x -> degree	= -1;
	x -> size	= 0;
	x -> dom	= K_of_x;
	x -> coeff	= NULL;
}

/*
 * Initialize the given polynomial to zero, but provide it with a
 * coefficient buffer sufficient to hold a polynomial of the given degree.
 */

	static
	inline
	void
Zpx_init_degree (

const struct calib_Zpx_dom *	K_of_x,	/* IN: polynomial domain */
struct calib_Zpx_obj *		x,	/* IN: polynomial to initialize */
int				degree	/* IN: maximum degree */
)
{
int		size;

	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR
			      (x      EQ NULL) OR
			      (degree  < -1));

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	x -> degree	= -1;
	x -> size	= size;
	x -> dom	= K_of_x;
	x -> coeff	= calib_new_Z_vector (size);
}

/*
 * Initialize the given array of polynomials to zero, but provide each
 * with a coefficient buffer sufficient to hold a polynomial of the
 * given degree.
 *
 * Although it is tempting to allocate all of the coefficients for all of
 * the polynomials into one big long vector, we cannot do so because the
 * coefficient vector for each polynomial must be separately free-able,
 * and re-allocatable.
 */

	static
	inline
	void
Zpx_init_array_degree (

const struct calib_Zpx_dom *	K_of_x,	/* IN: polynomial domain */
struct calib_Zpx_obj *		parray,	/* IN/OUT: array of polynomials */
int				npoly,	/* IN: num polynomials in array */
int				degree	/* IN: max degree of each polynomial */
)
{
int			i, size;
struct calib_Zpx_obj *	pp;

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	pp = &parray [0];
	for (i = 0; i < npoly; i++, pp++) {
		pp -> degree	= -1;
		pp -> size	= size;
		pp -> dom	= K_of_x;
		pp -> coeff	= calib_new_Z_vector (size);
	}
}

/*
 * Assure that the given polynomial is big enough to hold a polynomial
 * of the given degree.  This does not change the value currently
 * residing in this polynomial.
 */

	static
	void
Zpx_alloc (

struct calib_Zpx_obj *	rop,		/* IN/OUT: Poly to alloc buffer for */
int			degree		/* IN: Degree desired */
)
{
int		size, dsize;
mpz_ptr		zp, dp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (degree < -1));

	size	= degree + 1;
	dsize	= rop -> size;

	if (dsize < size) {
		dp	= rop -> coeff;
		zp	= calib_new_Z_vector (size);
		calib_copy_Z_vector (zp, dp, dsize);
		calib_free_Z_vector (dp, dsize);
		rop -> coeff = zp;
		rop -> size  = size;
	}
}

/*
 * Clear out the given polynomial.
 */

	static
	inline
	void
Zpx_clear (

struct calib_Zpx_obj *	x	/* IN/OUT: polynomial to clear */
)
{
int		size;
mpz_ptr		coeff;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	coeff = x -> coeff;
	if (coeff NE NULL) {
		size = x -> size;
		CALIB_FATAL_ERROR_IF (size <= 0);
		calib_free_Z_vector (coeff, size);
	}
	x -> degree	= -1;
	x -> size	= 0;
	x -> dom	= NULL;
	x -> coeff	= NULL;
}

/*
 * Clear out the given array of polynomials.
 */

	static
	inline
	void
Zpx_clear_array (

struct calib_Zpx_obj * parray,	/* IN: array of polynomials to clear */
int			npoly	/* IN: number of polynomials in parray */
)
{
int		i;

	for (i = 0; i < npoly; i++) {
		Zpx_clear (&parray [i]);
	}
}

/*
 * Set the given destination polynomial from the given coefficient
 * array.  Reallocate the destination polynomial's coefficient array,
 * if necessary.
 */

	static
	inline
	void
Zpx_realloc_copy_poly (

struct calib_Zpx_obj *	poly,		/* IN: destination polynomial */
mpz_srcptr	 	coeff,		/* IN: source coefficient array */
int		 	degree		/* IN: degree to set */
)
{
int		cur_size, new_size;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (degree < -1));

	poly -> degree	= degree;

	if (degree < 0) {
		/* We allow (coeff EQ NULL) here (but do not require). */
		return;
	}

	CALIB_FATAL_ERROR_IF (coeff EQ NULL);

	new_size = degree + 1;
	if (new_size > 0) {
		cur_size = poly -> size;
		if (new_size > cur_size) {
			if (poly -> coeff NE NULL) {
				calib_free_Z_vector (poly -> coeff, cur_size);
			}
			poly -> coeff = calib_new_Z_vector (new_size);
			poly -> size = new_size;
		}
		calib_copy_Z_vector (poly -> coeff, coeff, new_size);
	}
}

/*
 * Copy the given polynomial.
 */

	static
	void
Zpx_set (

struct calib_Zpx_obj *		rop,	/* IN/OUT: destination polynomial */
const struct calib_Zpx_obj *	op	/* IN: source polynomial */
)
{
const struct calib_Zpx_dom *	K_of_x;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {
		K_of_x	= rop -> dom;
		
		CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

		Zpx_realloc_copy_poly (rop, op -> coeff, op -> degree);
	}
}

/*
 * Set the given polynomial to the given integer value.
 */

	static
	void
Zpx_set_si (

struct calib_Zpx_obj *		rop,	/* IN/OUT: destination polynomial */
calib_si_t			op	/* IN: source integer value */
)
{
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	if (op EQ 0) {
		rop -> degree = -1;
	}
	else {
		K_of_x	= rop -> dom;
		cf	= K_of_x -> cf;
		Zpx_alloc (rop, 0);
		cf -> set_si (cf, &(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given polynomial to the given GMP integer value.
 */

	static
	void
Zpx_set_z (

struct calib_Zpx_obj *		rop,	/* IN/OUT: destination polynomial */
mpz_srcptr			op	/* IN: source GMP integer */
)
{
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpz_sgn (op) EQ 0) {
		rop -> degree = -1;
	}
	else {
		K_of_x	= rop -> dom;
		cf	= K_of_x -> cf;
		Zpx_alloc (rop, 0);
		cf -> set_z (cf, &(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given polynomial to the given GMP rational value.
 */

	static
	void
Zpx_set_q (

struct calib_Zpx_obj *		rop,	/* IN/OUT: destination polynomial */
mpq_srcptr			op	/* IN: source GMP rational */
)
{
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpq_sgn (op) EQ 0) {
		rop -> degree = -1;
	}
	else {
		K_of_x	= rop -> dom;
		cf	= K_of_x -> cf;
		Zpx_alloc (rop, 0);
		cf -> set_q (cf, &(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * The the given Zp[x] polynomial to x**power.
 */

	static
	void
Zpx_set_var_power (

struct calib_Zpx_obj *		rop,	/* IN/OUT: destination polynomial */
int				power	/* IN: power to set */
)
{
int		i;
mpz_ptr		coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (power < 0));

	Zpx_alloc (rop, power);
	coeff = rop -> coeff;
	for (i = 0; i < power; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
	mpz_set_ui (&coeff [power], 1);
	rop -> degree = power;
}

/*
 * Convert the given polynomial in Z[x] into the corresponding polynomial
 * in Zp[x].
 */

	static
	void
Zpx_set_Zx (

struct calib_Zpx_obj *		rop,	/* IN/OUT: destination polynomial */
const struct calib_Zx_obj *	op	/* IN: source polynomial in Z[x] */
)
{
int				i, degree, size;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_srcptr			zcoeff;
mpz_ptr				coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;
	degree	= op -> degree;
	zcoeff	= op -> coeff;
	size	= degree + 1;

	Zpx_alloc (rop, degree);

	coeff = rop -> coeff;
	if (size > 0) {
		for (i = 0; i < size; i++) {
			cf -> set_z (cf, &coeff [i], &zcoeff [i]);
		}
	}
	rop -> degree = degree;
}

/*
 * Convert the given Q(a) value into the corresponding polynomial
 * in Zp[x].
 */

	static
	void
Zpx_set_Qa (

struct calib_Zpx_obj *		rop,	/* IN/OUT: destination polynomial */
const struct calib_Qa_obj *	op	/* IN: source Q(a) value */
)
{
int				i, degree;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_srcptr			zcoeff;
mpz_ptr				coeff;
mpz_t				QFACT;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;
	degree	= op -> degree;
	zcoeff	= op -> coeff;

	mpz_init (QFACT);

	cf -> set_q (cf, QFACT, op -> qfact);

	Zpx_alloc (rop, degree);

	coeff = rop -> coeff;
	for (i = 0; i <= degree; i++) {
		cf -> mul (cf, &coeff [i], QFACT, &zcoeff [i]);
	}
	rop -> degree = degree;

	mpz_clear (QFACT);
}

/*
 * Addition of two polynomials over Zp.
 */

	static
	void
Zpx_add (

struct calib_Zpx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zpx_obj *	op1,		/* IN: first operand */
const struct calib_Zpx_obj *	op2		/* IN: second operand */
)
{
int				i, rdeg, deg1, deg2, rsize;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_ptr				res, rp;
mpz_srcptr			cp1, cp2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE K_of_x) OR
			      (op2 -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = calib_new_Z_vector (rsize);

	cp1 = &(op1 -> coeff [0]);
	cp2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, rp++, cp1++, cp2++) {
		if (i <= deg1) {
			if (i <= deg2) {
				cf -> add (cf, rp, cp1, cp2);
			}
			else {
				mpz_set (rp, cp1);
			}
		}
		else {
			mpz_set (rp, cp2);
		}
	}
	while ((rdeg >= 0) AND (mpz_sgn (&res [rdeg]) EQ 0)) {
		--rdeg;
	}

	Zpx_realloc_copy_poly (rop, res, rdeg);
	calib_free_Z_vector (res, rsize);
}

/*
 * Subtraction of two polynomials over Zp.
 */

	static
	void
Zpx_sub (

struct calib_Zpx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zpx_obj *	op1,		/* IN: first operand */
const struct calib_Zpx_obj *	op2		/* IN: second operand */
)
{
int				i, rdeg, deg1, deg2, rsize;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_ptr				rp, res;
mpz_srcptr			cp1, cp2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE K_of_x) OR
			      (op2 -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = calib_new_Z_vector (rsize);

	cp1 = &(op1 -> coeff [0]);
	cp2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, rp++, cp1++, cp2++) {
		if (i <= deg1) {
			if (i <= deg2) {
				cf -> sub (cf, rp, cp1, cp2);
			}
			else {
				mpz_set (rp, cp1);
			}
		}
		else {
			cf -> neg (cf, rp, cp2);
		}
	}
	while ((rdeg >= 0) AND (mpz_sgn (&res [rdeg]) EQ 0)) {
		--rdeg;
	}

	Zpx_realloc_copy_poly (rop, res, rdeg);
	calib_free_Z_vector (res, rsize);
}

/*
 * Negate the given polynomial over Zp.
 */

	static
	void
Zpx_neg (


struct calib_Zpx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zpx_obj *	op		/* IN: operand to negate */
)
{
int				i, deg;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_ptr				dstp;
mpz_srcptr			srcp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	if (rop EQ op) {
		deg	= rop -> degree;
		dstp	= rop -> coeff;
		for (i = 0; i <= deg; i++) {
			cf -> neg (cf, &dstp [i], &dstp [i]);
		}
	}
	else {
		deg	= op -> degree;
		srcp	= op -> coeff;
		Zpx_alloc (rop, deg);
		dstp	= rop -> coeff;
		for (i = 0; i <= deg; i++) {
			cf -> neg (cf, &dstp [i], &srcp [i]);
		}
		rop -> degree = deg;
	}
}

/*
 * Multiplcation of two polynomials over Zp.
 */

	static
	void
Zpx_mul (

struct calib_Zpx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zpx_obj *	op1,		/* IN: first operand */
const struct calib_Zpx_obj *	op2		/* IN: second operand */
)
{
int				i, j, rdeg, deg1, deg2, rsize;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_ptr				res, rp1, rp2;
mpz_srcptr			cp1, cp2;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE K_of_x) OR
			      (op2 -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	if ((deg1 < 0) OR (deg2 < 0)) {
		/* Multipying by zero. */
		rop -> degree = -1;
		return;
	}

	rdeg	= deg1 + deg2;

	rsize = rdeg + 1;
	res = calib_new_Z_vector (rsize);

	mpz_init (TEMP);

	cp1 = &(op1 -> coeff [0]);
	rp1 = &res [0];
	for (i = 0; i <= deg1; i++) {
		rp2 = rp1;
		cp2 = &(op2 -> coeff [0]);
		for (j = 0; j <= deg2; j++) {
			cf -> mul (cf, TEMP, cp1, cp2);
			cf -> add (cf, rp2, rp2, TEMP);
			++rp2;
			++cp2;
		}
		++rp1;
		++cp1;
	}
	while ((rdeg >= 0) AND (mpz_sgn (&res [rdeg]) EQ 0)) {
		--rdeg;
	}

	mpz_clear (TEMP);

	Zpx_realloc_copy_poly (rop, res, rdeg);
	calib_free_Z_vector (res, rsize);
}

/*
 * Multiply a polynomial by the given integer.
 */

	static
	void
Zpx_mul_z (

struct calib_Zpx_obj *		rop,	/* IN/OUT: result */
const struct calib_Zpx_obj *	op1,	/* IN: polynomial to multiply */
mpz_srcptr			op2	/* IN: integer to multiply by */
)
{
int				i, rdeg, deg1, rsize;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_ptr				res, rp1;
mpz_srcptr			cp1;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	if ((deg1 < 0) OR (mpz_sgn (op2) EQ 0)) {
		/* Multipying by zero. */
		rop -> degree = -1;
		return;
	}

	rdeg = deg1;

	rsize = rdeg + 1;
	res = calib_new_Z_vector (rsize);

	cp1 = &(op1 -> coeff [0]);
	rp1 = &res [0];
	for (i = 0; i <= deg1; i++, cp1++, rp1++) {
		cf -> mul (cf, rp1, cp1, op2);
	}
	while ((rdeg >= 0) AND (mpz_sgn (&res [rdeg]) EQ 0)) {
		--rdeg;
	}

	Zpx_realloc_copy_poly (rop, res, rdeg);
	calib_free_Z_vector (res, rsize);
}

/*
 * Take the given polynomial to the given integer power.
 */

	static
	void
Zpx_ipow (

struct calib_Zpx_obj *		rop,		/* IN/OUT: result polynomial */
const struct calib_Zpx_obj *	op,		/* IN: polynomial to exponentiate */
int				power		/* IN: integer power */
)
{
const struct calib_Zpx_dom *	K_of_x;
struct calib_Zpx_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (power < 0));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	Zpx_init (K_of_x, &RES);
	Zpx_set_si (&RES, 1);

	Zpx_init (K_of_x, &PREG);
	Zpx_set (&PREG, op);

	while (power > 0) {
		if ((power & 1) NE 0) {
			Zpx_mul (&RES, &RES, &PREG);
		}
		Zpx_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	Zpx_set (rop, &RES);
	Zpx_clear (&PREG);
	Zpx_clear (&RES);
}

/*
 * Create a dynamically allocated copy of the given polynomial.
 */

	static
	struct calib_Zpx_obj *
Zpx_dup (

const struct calib_Zpx_obj *	op	/* IN: polynomial to duplicate */
)
{
int				deg, size;
const struct calib_Zpx_dom *	K_of_x;
struct calib_Zpx_obj *		dst;
mpz_ptr				zp;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	K_of_x	= op -> dom;
	deg	= op -> degree;
	size	= deg + 1;

	dst = CALIB_NEW (struct calib_Zpx_obj);
	zp = NULL;
	if (size > 0) {
		zp = calib_new_Z_vector (size);
		calib_copy_Z_vector (zp, op -> coeff, size);
	}
	dst -> degree	= deg;
	dst -> size	= size;
	dst -> dom	= K_of_x;
	dst -> coeff	= zp;

	return (dst);
}

/*
 * Free the given polynomial, which must have been dynamically allocated.
 */

	static
	void
Zpx_free (

struct calib_Zpx_obj *	poly	/* IN/OUT: polynomial to free */
)
{
	if (poly NE NULL) {
		Zpx_clear (poly);
		free (poly);
	}
}

/*
 * Modify the given polynomial to be monic.
 */

	static
	void
Zpx_monicize (

struct calib_Zpx_obj *		poly	/* IN/OUT: polynomial to monicize */
)
{
int				i, deg;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_srcptr			lc;
mpz_ptr				coeff;
mpz_t				LC_INV;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	K_of_x	= poly -> dom;
	deg	= poly -> degree;

	if (deg >= 0) {
		coeff = poly -> coeff;
		lc = &coeff [deg];
		if (mpz_cmp_ui (lc, 1) NE 0) {
			cf = K_of_x -> cf;
			mpz_init (LC_INV);
			cf -> inv (cf, LC_INV, lc);
			for (i = 0; i <= deg; i++) {
				cf -> mul (cf, &coeff [i], &coeff [i], LC_INV);
			}
			mpz_clear (LC_INV);
		}
	}
}

/*
 * Evaluate the given poly(value).
 */

	static
	void
Zpx_eval (

mpz_ptr				rop,		/* IN/OUT: result polynomial */
const struct calib_Zpx_obj *	poly,		/* IN: polynomial to evaluate */
mpz_srcptr			value		/* IN: variable value to use */
)
{
int				i;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_srcptr			coeff;
mpz_t				SUM;

	CALIB_FATAL_ERROR_IF ((rop   EQ NULL) OR
			      (poly  EQ NULL) OR
			      (value EQ NULL));

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;

	mpz_init (SUM);

	coeff = poly -> coeff;

	for (i = poly -> degree; i >= 0; i--) {
		cf -> mul (cf, SUM, SUM, value);
		cf -> add (cf, SUM, SUM, &coeff [i]);
	}

	mpz_set (rop, SUM);
	mpz_clear (SUM);
}

/*
 * The division algorithm (with remainder):
 * Given polynomials a, b \in Zp[x], find polynomials q, r \in Zp[x]
 * such that:
 *
 *	a = q*b + r	AND	degree (r) < degree (b)
 *
 * This is easy and clean because we have polynomials over a field,
 * which supports division.
 */

	static
	void
Zpx_div (

struct calib_Zpx_obj *		quotient,	/* IN/OUT: quotient */
struct calib_Zpx_obj *		remainder,	/* IN/OUT: remainder */
const struct calib_Zpx_obj *	a,		/* IN: first operand */
const struct calib_Zpx_obj *	b		/* IN: second operand */
)
{
int				i, j, adeg, bdeg, qdeg, rdeg, rsize;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_ptr				rp, quot, rem;
mpz_srcptr			ap, bp;
mpz_t				BLC_INV, FACTOR, TEMP;

	CALIB_FATAL_ERROR_IF ((a EQ NULL) OR (b EQ NULL));

	K_of_x	= a -> dom;

	CALIB_FATAL_ERROR_IF (b -> dom NE K_of_x);
	CALIB_FATAL_ERROR_IF ((quotient NE NULL) AND
			      (quotient -> dom NE K_of_x));
	CALIB_FATAL_ERROR_IF ((remainder NE NULL) AND
			      (remainder -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	adeg	= a -> degree;
	bdeg	= b -> degree;

	rsize = adeg + 1;

	/* Local coefficient buffers. */
	quot = calib_new_Z_vector (2 * rsize);
	rem  = quot + rsize;

	mpz_init (BLC_INV);

	ap = &(a -> coeff [0]);
	bp = &(b -> coeff [0]);

	if (bdeg <= 0) {
		/* Division by a constant.  Do not allow division by zero. */
		CALIB_FATAL_ERROR_IF ((bdeg < 0) OR (mpz_sgn (bp) EQ 0));

		cf -> inv (cf, BLC_INV, bp);

		rp = &rem [0];
		for (i = 0; i <= adeg; i++) {
			cf -> mul (cf, rp++, ap++, BLC_INV);
		}
		if (quotient NE NULL) {
			Zpx_realloc_copy_poly (quotient, rem, adeg);
		}
		if (remainder NE NULL) {
			remainder -> degree = -1;
		}
		mpz_clear (BLC_INV);

		calib_free_Z_vector (quot, 2 * rsize);

		return;
	}

	qdeg = adeg - bdeg;
	if (qdeg < -1) {
		qdeg = -1;
	}
	calib_copy_Z_vector (rem, &(a -> coeff [0]), adeg + 1);
	rdeg = adeg;

	mpz_init (FACTOR);
	mpz_init (TEMP);

	/* Compute multiplicative inverse of lc(b). */
	cf -> inv (cf, BLC_INV, &(b -> coeff [bdeg]));

	while (rdeg >= bdeg) {
		j = rdeg - bdeg;
		cf -> mul (cf, FACTOR, &rem [rdeg], BLC_INV);
		mpz_set (&quot [j], FACTOR);
		rp = &rem [rdeg - 1];
		bp = &(b -> coeff [bdeg - 1]);
		for (i = bdeg - 1; i >= 0; i--) {
			cf -> mul (cf, TEMP, FACTOR, bp--);
			cf -> sub (cf, rp, rp, TEMP);
			--rp;
		}
		--rdeg;
		rp = &rem [rdeg];
		while ((rdeg >= 0) AND (mpz_sgn (&rem [rdeg]) EQ 0)) {
			--rdeg;
		}
	}

	if (quotient NE NULL) {
		Zpx_realloc_copy_poly (quotient, &quot [0], qdeg);
	}

	if (remainder NE NULL) {
		Zpx_realloc_copy_poly (remainder, &rem [0], rdeg);
	}

	mpz_clear (TEMP);
	mpz_clear (FACTOR);
	mpz_clear (BLC_INV);

	calib_free_Z_vector (quot, 2 * rsize);
}

/*
 * Compute the Greatest Common Divisor of two polynomials over Zp.
 * The result is always monic unless a = b = 0.
 *
 * We use the conventional Euclidean algorithm.
 */

	static
	void
Zpx_gcd (

struct calib_Zpx_obj *		gcd,		/* IN/OUT: GCD */
const struct calib_Zpx_obj *	a,		/* IN: first operand */
const struct calib_Zpx_obj *	b		/* IN: second operand */
)
{
int				i, adeg, bdeg, maxdeg;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
struct calib_Zpx_obj *		ap;
struct calib_Zpx_obj *		bp;
struct calib_Zpx_obj *		rp;
struct calib_Zpx_obj *		qp;
struct calib_Zpx_obj *		temp_ptr;
mpz_ptr				cp;
mpz_t				LCA_INV;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	K_of_x	= gcd -> dom;

	CALIB_FATAL_ERROR_IF ((a -> dom NE K_of_x) OR (b -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	adeg	= a -> degree;
	bdeg	= b -> degree;
	maxdeg	= _MAX (adeg, bdeg);

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 4, maxdeg);

	temp_ptr = &TMP [0];
	ap = temp_ptr++;
	bp = temp_ptr++;
	rp = temp_ptr++;
	qp = temp_ptr++;

	Zpx_set (ap, a);
	Zpx_set (bp, b);

	while (bp -> degree >= 0) {
		K_of_x -> div (qp, rp, ap, bp);
		temp_ptr = ap;
		ap = bp;
		bp = rp;
		rp = temp_ptr;
	}

	adeg = ap -> degree;
	if (adeg >= 0) {
		/* Monicize result. */
		mpz_init (LCA_INV);
		cf -> inv (cf, LCA_INV, &(ap -> coeff [adeg]));
		cp = &(ap -> coeff [0]);
		for (i = 0; i <= adeg; i++) {
			cf -> mul (cf, cp, cp, LCA_INV);
			++cp;
		}
		mpz_clear (LCA_INV);
	}

	Zpx_set (gcd, ap);

	FREE_LOCAL_POLY_ARRAY (TMP, 4);
}

/*
 * The Extended Euclidean Algorithm for polynomials in Zp[x].
 * Given polynomials a, b \in Zp[x],
 * compute polynomials g, xa, xb \in Zp[x] such that
 *
 * 1.	a is a multiple of g,
 * 2.	b is a multiple of g,
 * 3.	xa * a + xb * b = g.
 *
 * These three conditions imply that g = gcd(a,b).  The polynomial g
 * is monic (unless a = b = 0, in which case g = 0).
 */

	static
	void
Zpx_extgcd (

struct calib_Zpx_obj *		gcd,		/* IN/OUT: GCD */
struct calib_Zpx_obj *		xa,		/* IN/OUT: multiplier for a */
struct calib_Zpx_obj *		xb,		/* IN/OUT: multiplier for b */
const struct calib_Zpx_obj *	a,		/* IN: first operand */
const struct calib_Zpx_obj *	b		/* IN: second operand */
)
{
int				i, adeg, bdeg, maxdeg, deg;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
struct calib_Zpx_obj *		ap;
struct calib_Zpx_obj *		bp;
struct calib_Zpx_obj *		rp;
struct calib_Zpx_obj *		qp;
struct calib_Zpx_obj *		m1;
struct calib_Zpx_obj *		m2;
struct calib_Zpx_obj *		m3;
struct calib_Zpx_obj *		n1;
struct calib_Zpx_obj *		n2;
struct calib_Zpx_obj *		n3;
struct calib_Zpx_obj *		TEMP;
struct calib_Zpx_obj *		temp_ptr;
mpz_ptr				cp;
mpz_t				LCA_INV;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR
			      (xa  EQ NULL) OR
			      (xb  EQ NULL) OR
			      (a   EQ NULL) OR
			      (b   EQ NULL));

	K_of_x	= gcd -> dom;

	CALIB_FATAL_ERROR_IF ((xa -> dom NE K_of_x) OR
			      (xb -> dom NE K_of_x) OR
			      (a  -> dom NE K_of_x) OR
			      (b  -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	adeg	= a -> degree;
	bdeg	= b -> degree;
	maxdeg	= _MAX (adeg, bdeg);

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 11, maxdeg);

	temp_ptr = &TMP [0];
	ap = temp_ptr++;
	bp = temp_ptr++;
	rp = temp_ptr++;
	qp = temp_ptr++;
	m1 = temp_ptr++;
	m2 = temp_ptr++;
	m3 = temp_ptr++;
	n1 = temp_ptr++;
	n2 = temp_ptr++;
	n3 = temp_ptr++;
	TEMP = temp_ptr++;

	Zpx_set (ap, a);
	Zpx_set (bp, b);

	/* ap = m1*a + n1*b */
	Zpx_set_si (m1, 1);
	Zpx_set_si (n1, 0);

	/* bp = m2*a + n2*b */
	Zpx_set_si (m2, 0);
	Zpx_set_si (n2, 1);

	while (bp -> degree >= 0) {
		Zpx_div (qp, rp, ap, bp);

		/* r = ap - q*bp			*/
		/*   = (m1*a + n1*b) - q*(m2*a + n2*b)	*/
		/*   = (m1 - q*m2)*a + (n1 - q*n2)*b	*/
		/*   = m3*a	     + m3*b,		*/
		/* so we compute:			*/
		/*	m3 = m1 - q*m2			*/
		/*	n3 = n1 - q*n2			*/
		Zpx_mul (TEMP, qp, m2);
		Zpx_sub (m3, m1, TEMP);
		Zpx_mul (TEMP, qp, n2);
		Zpx_sub (n3, n1, TEMP);

		temp_ptr = ap;
		ap = bp;
		bp = rp;
		rp = temp_ptr;

		temp_ptr = m1;
		m1 = m2;
		m2 = m3;
		m3 = temp_ptr;

		temp_ptr = n1;
		n1 = n2;
		n2 = n3;
		n3 = temp_ptr;
	}

	/* We have ap = M1*a + N1*b. */

	adeg = ap -> degree;
	if (adeg >= 0) {
		/* Monicize result. */
		mpz_init (LCA_INV);
		cf -> inv (cf, LCA_INV, &(ap -> coeff [adeg]));

		cp = &(ap -> coeff [0]);
		for (i = 0; i <= adeg; i++, cp++) {
			cf -> mul (cf, cp, cp, LCA_INV);
		}

		cp = &(m1 -> coeff [0]);
		deg = m1 -> degree;
		for (i = 0; i <= deg; i++, cp++) {
			cf -> mul (cf, cp, cp, LCA_INV);
		}

		cp = &(n1 -> coeff [0]);
		deg = n1 -> degree;
		for (i = 0; i <= deg; i++, cp++) {
			cf -> mul (cf, cp, cp, LCA_INV);
		}

		mpz_clear (LCA_INV);
	}

	Zpx_set (gcd, ap);
	Zpx_set (xa, m1);
	Zpx_set (xb, n1);

	FREE_LOCAL_POLY_ARRAY (TMP, 11);
}

/*
 * Factor the given polynomial in Zp[x] completely into its monic,
 * irreducible factors over Zp[x].  The correct multiplicity of each
 * factor is also determined.
 */

	static
	struct calib_Zpx_factor *
Zpx_factor (

const struct calib_Zpx_obj *		poly	/* IN: polynomial to factor */
)
{
struct calib_Zpx_factor *	sqfactors;
struct calib_Zpx_factor *	factors;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	/* First perform a square-free factorization.  This also strips	*/
	/* any leading coefficient off as a constant factor so that all	*/
	/* of the remaining factors are monic.				*/

	sqfactors = Zpx_factor_square_free (poly);

	/* Now factor each square-free factor into irreducibles. */
	factors = Zpx_finish_sqf (sqfactors);

	Zpx_free_factors (sqfactors);

	return (factors);
}

/*
 * Factor the given polynomial into square-free factors.
 */

	static
	struct calib_Zpx_factor *
Zpx_factor_square_free (

const struct calib_Zpx_obj *	poly		/* IN: Polynomial to factor */
)
{
int				deg;
const struct calib_Zpx_dom *	K_of_x;
mpz_srcptr			lc;
struct calib_Zpx_factor *	sqfactors;
struct calib_Zpx_factor *	list;
struct calib_Zpx_factor **	hookp;
struct calib_Zpx_factor *	nfp;
struct calib_Zpx_obj *		sqfpoly;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	deg	= poly -> degree;

	if (deg <= 0) {
		/* Polynomial is constant. */
		list = CALIB_NEW (struct calib_Zpx_factor);
		list -> multiplicity	= 1;
		list -> factor		= Zpx_dup (poly);
		list -> next		= NULL;
		return (list);
	}

	K_of_x	= poly -> dom;

	/* Make a local copy of the polynomial and monicize it.	*/
	ALLOCATE_LOCAL_POLY (K_of_x, lpoly, deg);
	Zpx_set (&lpoly, poly);

	list = NULL;
	hookp = &list;

	/* If the lead coefficient isn't 1, then save it off as a	*/
	/* degree 0 factor of multiplicity 1.				*/
	lc = &lpoly.coeff [deg];
	if (mpz_cmp_ui (lc, 1) NE 0) {
		sqfpoly = CALIB_NEW (struct calib_Zpx_obj);
		sqfpoly -> degree	= 0;
		sqfpoly -> size		= 1;
		sqfpoly -> dom		= K_of_x;
		sqfpoly -> coeff	= calib_new_Z_vector (1);
		mpz_set (&(sqfpoly -> coeff [0]), lc);

		nfp = CALIB_NEW (struct calib_Zpx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= sqfpoly;
		nfp -> next		= NULL;
		*hookp = nfp;
		hookp = &(nfp -> next);

		Zpx_monicize (&lpoly);
	}

	if (lpoly.degree <= 1) {
		/* Polynomial is already irreducible. */
		nfp = CALIB_NEW (struct calib_Zpx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= Zpx_dup (&lpoly);
		nfp -> next		= NULL;
		*hookp = nfp;
		hookp = &(nfp -> next);
	}
	else {
		sqfactors = Zpx_compute_square_free_factorization (&lpoly);

		while (sqfactors NE NULL) {
			nfp = sqfactors -> next;
			sqfactors -> next = NULL;

			*hookp = sqfactors;
			hookp = &(sqfactors -> next);

			sqfactors = nfp;
		}
	}

	FREE_LOCAL_POLY (lpoly);

	return (list);
}

/*
 * Factor the given polynomial into factors that are both square-free and
 * pairwise relatively prime, correctly determining the multiplicity of
 * each such factor.  This results in one non-trivial factor for each
 * distinct multiplicity present -- that factor being the product of all
 * factors having that that multiplicity.
 *
 * NOTE: the given polynomial is assumed to be monic, and may be
 *	 modified by this routine.
 */

	static
	struct calib_Zpx_factor *
Zpx_compute_square_free_factorization (

struct calib_Zpx_obj *		poly		/* IN: polynomial to factor */
)
{
int				i, deg, multiplicity, multiplicity_multiplier;
unsigned long			p;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
struct calib_Zpx_factor *	list;
struct calib_Zpx_factor *	fp;
struct calib_Zpx_factor **	hookp;
struct calib_Zpx_obj *		pp;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;

	p = mpz_get_ui (cf -> p);

	/* Given polynomial must be monic, and degree >= 1. */
	deg = poly -> degree;
	CALIB_FATAL_ERROR_IF (deg <= 0);
	CALIB_FATAL_ERROR_IF (mpz_cmp_ui (&(poly -> coeff [deg]), 1) NE 0);

	list = NULL;
	hookp = &list;

	/* Assume poly has the following factorization:			*/
	/*								*/
	/*	poly(x) = f1(x) * (f2(x))^2 * ... * (fp(x))^p * ...	*/
	/*								*/
	/* and let g = gcd(poly, poly').  Then g is like poly,		*/
	/* except that one copy of each factor "WHOSE MULTIPLICITY IS	*/
	/* *NOT* A MULTPLE OF p" is removed.				*/

	ALLOCATE_LOCAL_POLY (K_of_x, f, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, fprime, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, a, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, b, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, c, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, d, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, rem, deg - 1);

	/* Yun's algorithm with a "mod p twist", initialize. */
	multiplicity_multiplier = 1;

	Zpx_set (&f, poly);

	Zpx_derivative (&fprime, &f);

	while (f.degree > 0) {
		/* We have a new f to factor. */
		while (fprime.degree < 0) {
			/* Derivative is zero.  This implies that f has	*/
			/* the form:					*/
			/*	f(x) = g(x^p) = (g(x))^p		*/
			deg = f.degree;
			CALIB_FATAL_ERROR_IF ((deg %p) NE 0);
			deg /= p;
			for (i = 0; i <= deg; i++) {
				mpz_set (&f.coeff [i], &f.coeff [i * p]);
			}
			f.degree = deg;
			multiplicity_multiplier *= p;

			Zpx_derivative (&fprime, &f);
		}
		Zpx_gcd (&a, &f, &fprime);

		Zpx_div (&b, &rem, &f, &a);
		CALIB_FATAL_ERROR_IF (rem.degree >= 0);

		Zpx_div (&c, &rem, &fprime, &a);
		CALIB_FATAL_ERROR_IF (rem.degree >= 0);

		Zpx_derivative (&d, &b);
		Zpx_sub (&d, &c, &d);

		for (;;) {
			/* Standard Yun's algorithm, iteration. */
			Zpx_gcd (&a, &b, &d);

			if (NOT ((a.degree EQ 0) AND
				 (mpz_cmp_ui (&a.coeff [0], 1) EQ 0))) {
				/* We have a new square-free factor. */
				pp = Zpx_dup (&a);
				Zpx_monicize (pp);

				multiplicity = Zpx_count_multiplicity (&f, pp);

				fp = CALIB_NEW (struct calib_Zpx_factor);
				fp -> multiplicity	= multiplicity * multiplicity_multiplier;
				fp -> factor		= pp;
				fp -> next		= NULL;
				*hookp = fp;
				hookp = &(fp -> next);
			}
			Zpx_div (&b, &rem, &b, &a);
			CALIB_FATAL_ERROR_IF (rem.degree >= 0);

			Zpx_div (&c, &rem, &d, &a);
			CALIB_FATAL_ERROR_IF (rem.degree >= 0);

			Zpx_derivative (&d, &b);
			Zpx_sub (&d, &c, &d);

			if (b.degree <= 0) break;
		}

		/* We had better have b EQ 1. */
		CALIB_FATAL_ERROR_IF ((b.degree NE 0) OR
				      (mpz_cmp_ui (&b.coeff [0], 1) NE 0));

		if ((f.degree EQ 0) AND (mpz_cmp_ui (&f.coeff [0], 1) EQ 0)) {
			/* All factors have been divided out of f.	*/
			/* We are done.					*/
			break;
		}

		/* Must have f' = 0, demonstrating that f has the form:	*/
		/*	f(x) = g(x^p) = (h(x))^p			*/
		Zpx_derivative (&fprime, &f);
		CALIB_FATAL_ERROR_IF (fprime.degree >= 0);
	}

	FREE_LOCAL_POLY (rem);
	FREE_LOCAL_POLY (d);
	FREE_LOCAL_POLY (c);
	FREE_LOCAL_POLY (b);
	FREE_LOCAL_POLY (a);
	FREE_LOCAL_POLY (fprime);
	FREE_LOCAL_POLY (f);

	return (list);
}

/*
 * Determine the multiplicity of the given factor of the given polynomial
 * by counting how many times it divides out evenly.
 */

	static
	int
Zpx_count_multiplicity (

struct calib_Zpx_obj *		poly,	/* IN/OUT: polynomial to divide */
const struct calib_Zpx_obj *	factor	/* IN: factor to divide out */
)
{
int				pdeg, fdeg, multiplicity;
const struct calib_Zpx_dom *	K_of_x;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (factor EQ NULL));

	K_of_x	= poly -> dom;

	CALIB_FATAL_ERROR_IF (factor -> dom NE K_of_x);

	pdeg	= poly -> degree;
	fdeg	= factor -> degree;

	ALLOCATE_LOCAL_POLY (K_of_x, quot, pdeg);
	ALLOCATE_LOCAL_POLY (K_of_x, rem,  fdeg);

	multiplicity = 0;
	for (;;) {
		Zpx_div (&quot, &rem, poly, factor);
		if (rem.degree >= 0) {
			/* Remainder is not zero.  We are done. */
			break;
		}
		++multiplicity;
		Zpx_set (poly, &quot);
	}

	FREE_LOCAL_POLY (rem);
	FREE_LOCAL_POLY (quot);

	return (multiplicity);
}

/*
 * Given a list of square-free factors, "finish" the factorization by
 * factoring each into irreducible polynomials correctly labeled with
 * their respective multiplicities.
 */

	static
	struct calib_Zpx_factor *
Zpx_finish_sqf (

const struct calib_Zpx_factor *	sqfactors	/* IN: Square-free factors */
)
{
int				multiplicity;
struct calib_Zpx_factor *	list;
struct calib_Zpx_factor **	hookp;
struct calib_Zpx_factor *	nfp;
struct calib_Zpx_factor *	factors;
struct calib_Zpx_factor *	temp;
const struct calib_Zpx_obj *	sqfpoly;
struct calib_Random		random;

	list	= NULL;
	hookp	= &list;

	calib_random_init (&random);

	for (; sqfactors NE NULL; sqfactors = sqfactors -> next) {
		multiplicity	= sqfactors -> multiplicity;
		sqfpoly		= sqfactors -> factor;

		if (sqfpoly -> degree <= 1) {
			/* Factor is constant or linear, and therefore	*/
			/* already irreducible.  Copy to output list.	*/
			nfp = CALIB_NEW (struct calib_Zpx_factor);
			nfp -> multiplicity	= multiplicity;
			nfp -> factor		= Zpx_dup (sqfpoly);
			nfp -> next		= NULL;

			*hookp = nfp;
			hookp = &(nfp -> next);
			continue;
		}

		factors = Zpx_factor_one_square_free_poly (sqfpoly,
							   &random);

		/* Apply correct multiplicity and append to output list. */
		while (factors NE NULL) {
			temp = factors -> next;
			factors -> multiplicity	= multiplicity;
			factors -> next		= NULL;

			*hookp = factors;
			hookp = &(factors -> next);

			factors = temp;
		}
	}

	return (list);
}

/*
 * Given a polynomial that has degree >= 2 and is known to be
 * square-free, factor it into irreducibles over Z_p[x].
 *
 * There are several possible methods, and we choose among them here.
 * Some of these need randomness, which the caller also provides.
 */

	static
	struct calib_Zpx_factor *
Zpx_factor_one_square_free_poly (

const struct calib_Zpx_obj *	sqfpoly,	/* IN: Square-free poly to */
						/*     factor */
struct calib_Random *		randp		/* IN/OUT: Random number generator */
)
{
enum calib_Zpx_factor_method	method;
struct calib_Zpx_factor *	factors;
const struct calib_Zpx_dom *	dom;
const struct calib_Zp_dom *	cf;

	CALIB_FATAL_ERROR_IF ((sqfpoly EQ NULL) OR (randp EQ NULL));

	dom	= sqfpoly -> dom;
	cf	= dom -> cf;

	method = dom -> settings.factor_method;

	if (mpz_cmp_ui (cf -> p, 1000000) > 0) {
		/* Berlekamp cannot handle large primes, but DDF can. */
		method = CALIB_ZPX_FACTOR_METHOD_DDF;
	}

	switch (method) {
	case CALIB_ZPX_FACTOR_METHOD_BERLEKAMP:
		factors	= _calib_Zpx_berle_factor (sqfpoly, randp);
		break;

	case CALIB_ZPX_FACTOR_METHOD_DDF:
		factors	= Zpx_ddf (sqfpoly, randp);
		break;

	default:
		CALIB_FATAL_ERROR;
		factors = NULL;
		break;
	}

	return (factors);
}

/*
 * Given a square-free polynomial over Zp[x], factor it completely into
 * irreducibles using the Distinct-Degree-Factorization method.  This
 * method exploits the fact that the polynomial x^{p^d} - x is the product
 * of all irreducible polynomials whose degree divides d.
 *
 * To factor f(x), simply compute p(x) = ([x^q] mod f(x)), where q = p^d.
 * Then gcd(p(x) - x, f(x)) gives the product of all factors whose degree
 * k divides d.  If we do this for d = 1, 2, ... then we recover the
 * irreducible factors in increasing degree.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	struct calib_Zpx_factor *
Zpx_ddf (

const struct calib_Zpx_obj *	sqfpoly,	/* IN: square-free polynomial to */
						/*     factor */
struct calib_Random *		randp		/* IN/OUT: random number */
						/*	   generator */
)
{
int				d, deg, fdeg;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
struct calib_Zpx_factor *	list;
struct calib_Zpx_factor **	hookp;
struct calib_Zpx_factor *	fp;
struct calib_Zpx_obj		*lpoly, *reg1, *reg2, *f, *temp_ptr;

	CALIB_FATAL_ERROR_IF ((sqfpoly EQ NULL) OR (randp EQ NULL));

	K_of_x	= sqfpoly -> dom;
	cf	= K_of_x -> cf;

	list = NULL;
	hookp = &list;

	deg = sqfpoly -> degree;

	/* Assure that sqfpoly is monic. */
	CALIB_FATAL_ERROR_IF ((deg <= 0) OR
			      (mpz_cmp_ui (&(sqfpoly -> coeff [deg]), 1) NE 0));

	/* Allocate local polynomials. */
	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 4, deg);
	temp_ptr = &TMP [0];
	lpoly	= temp_ptr++;
	reg1	= temp_ptr++;
	reg2	= temp_ptr++;
	f	= temp_ptr++;

	/* Make a local copy of sqfpoly that we can modify. */
	Zpx_set (lpoly, sqfpoly);

	ALLOCATE_LOCAL_POLY (K_of_x, xpoly, 1);
	xpoly.degree	= 1;
	mpz_set_ui (&xpoly.coeff [1], 1);

	/* reg1 is initially x**(p**d), where d=0 (i.e., x**1). */
	Zpx_set (reg1, &xpoly);

	for (d = 1; ; d++) {
		if (2 * d > lpoly -> degree) {
			/* Residual lpoly is irreducible. */
			break;
		}

		/* reg1 = (reg1**p) mod lpoly, so that	*/
		/* reg1 = x**(p**d) mod lpoly.		*/
		Zpx_poly_to_int_pow_mod_poly (reg1,
					      reg1,
					      cf -> p,
					      lpoly);
		Zpx_sub (reg2, reg1, &xpoly);
		Zpx_gcd (f, reg2, lpoly);

		fdeg = f -> degree;
		if (fdeg > 0) {
			CALIB_FATAL_ERROR_IF (fdeg > lpoly -> degree);
			Zpx_div (lpoly, NULL, lpoly, f);
			/* Further reduce reg1 by new lpoly. */
			Zpx_div (NULL, reg1, reg1, lpoly);
			hookp = Zpx_split_factors (hookp, f, d, randp);
		}
	}

	if (lpoly -> degree > 0) {
		/* Append final irreducible factor (highest degree). */
		f = Zpx_dup (lpoly);
		Zpx_monicize (f);

		fp = CALIB_NEW (struct calib_Zpx_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= f;
		fp -> next		= NULL;
		*hookp = fp;
		hookp = &(fp -> next);
	}

	FREE_LOCAL_POLY (xpoly);
	FREE_LOCAL_POLY_ARRAY (TMP, 4);

	return (list);
}

#pragma GCC diagnostic pop

/*
 * Given a polynomial f(x) \in Zp[x] whose irreducible factors all have
 * degree d, factor f(x) into these irreducible factors.  This is trivial
 * if degree(f) is 0 or d.  Otherwise we use a randomized algorithm to
 * split f(x) into irreducible factors.
 */

	static
	struct calib_Zpx_factor **
Zpx_split_factors (

struct calib_Zpx_factor **	hookp,	/* IN: where to append next factor */
const struct calib_Zpx_obj *	poly,	/* IN: polynomial f(x) to factor */
int				d,	/* IN: degree of irreducible factors */
struct calib_Random *		randp	/* IN: random number generator */
)
{
int				deg, fdeg, rdeg;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
struct calib_Zpx_factor *	fp;
struct calib_Zpx_obj *		pp;
mpz_t				R, ONE;

	CALIB_FATAL_ERROR_IF ((hookp EQ NULL) OR
			      (poly  EQ NULL) OR
			      (d     <= 0) OR
			      (randp EQ NULL));

	deg	= poly -> degree;
	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;

	/* Degree must be a multiple of d! */
	CALIB_FATAL_ERROR_IF ((deg < 0) OR ((deg % d) NE 0));
	if (deg EQ 0) {
		/* Polynomial is constant -- nothing to split. */
		return (hookp);
	}
	if (deg EQ d) {
		/* We have a single irreducible factor! */
		pp = Zpx_dup (poly);
		Zpx_monicize (pp);

		fp = CALIB_NEW (struct calib_Zpx_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= pp;
		fp -> next		= NULL;
		*hookp = fp;
		hookp = &(fp -> next);
		return (hookp);
	}

	/* f(x) has degree k*d, where k >= 2.			*/
	/* We require that p is an odd prime.  Let:		*/
	/*							*/
	/*	q = p**d,					*/
	/*	G(x)  = x**q - x				*/
	/*	r = (q-1)/2,					*/
	/*	H1(x) = x**r - 1				*/
	/*	H2(x) = x**r + 1				*/
	/*							*/
	/* Let R(x) be a random polynomial \in Zp[x] of degree	*/
	/* d.							*/
	/*							*/
	/* Since G(x) is the product of all monic, irreducible	*/
	/* polynomials whose degree divides d, then G(R(x))	*/
	/* has these same factors (and perhaps more).  It turns	*/
	/* out that H1(R(x)) has about half of these factors,	*/
	/* so that gcd(f(x), H1(R(x))) is a non-trivial factor	*/
	/* of f(x) with probability at least 1/2.		*/

	mpz_init_set (R, cf -> p);
	mpz_pow_ui (R, R, d);
	mpz_sub_ui (R, R, 1);
	mpz_tdiv_q_2exp (R, R, 1);	/* R <<= 1 */

	mpz_init_set_ui (ONE, 1);

	ALLOCATE_LOCAL_POLY (K_of_x, xpoly, 1);
	xpoly.degree	= 1;
	mpz_set_ui (&xpoly.coeff [1], 1);

	ALLOCATE_LOCAL_POLY (K_of_x, rpoly, d);
	ALLOCATE_LOCAL_POLY (K_of_x, f, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, reg1, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, reg2, deg);

	/* Make a local, modifiable copy of f(x). */
	Zpx_set (&f, poly);

	fdeg = f.degree;
	while (fdeg > d) {
		Zpx_random_poly (&rpoly, d, randp);
		Zpx_poly_to_int_pow_mod_poly (&reg1, &rpoly, R, &f);
		cf -> sub (cf, &reg1.coeff [0], &reg1.coeff [0], ONE);
		if ((reg1.degree EQ 0) AND (mpz_sgn (&reg1.coeff [0]) EQ 0)) {
			/* H1(R(x)) == 0 (mod f(x)), so this is an	*/
			/* unlucky (and very unusual) draw.		*/
			continue;
		}
		Zpx_gcd (&reg2, &reg1, &f);
		rdeg = reg2.degree;
		if ((rdeg <= 0) OR (rdeg >= fdeg)) {
			/* We failed to split.  Try again. */
			continue;
		}

		/* We have a non-trivial split.  Recurse to either	*/
		/* save off the degree d factor, or split further.	*/
		hookp = Zpx_split_factors (hookp, &reg2, d, randp);
		Zpx_div (&f, NULL, &f, &reg2);
		fdeg = f.degree;
	}

	/* Should have one irredicible left.  Save it off. */
	CALIB_FATAL_ERROR_IF (fdeg NE d);
	hookp = Zpx_split_factors (hookp, &f, d, randp);

	FREE_LOCAL_POLY (reg2);
	FREE_LOCAL_POLY (reg1);
	FREE_LOCAL_POLY (f);
	FREE_LOCAL_POLY (rpoly);
	FREE_LOCAL_POLY (xpoly);

	mpz_clear (ONE);
	mpz_clear (R);

	return (hookp);
}

/*
 * Choose a random polynomial \in Zp[x] of degree d.
 */

	static
	void
Zpx_random_poly (

struct calib_Zpx_obj *		poly,	/* IN/OUT: random polynomial */
int				d,	/* IN: degree desired */
struct calib_Random *		randp	/* IN: random number generator */
)
{
int				i;
unsigned long			word;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_ptr				coeff;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (d <= 0) OR (randp EQ NULL));

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;
	coeff	= poly -> coeff;

	mpz_init (TEMP);

	for (i = 0; i <= d; i++) {
		word = calib_random_u64 (randp);
		mpz_set_ui (TEMP, word);
		cf -> set_z (cf, &coeff [i], TEMP);
	}

	mpz_clear (TEMP);

	/* Lead coefficient must not be zero. */
	if (mpz_sgn (&coeff [d]) EQ 0) {
		mpz_set_ui (&coeff [d], 1);
	}

	poly -> degree = d;
}

/*
 * Compute ((P(x)^K) mod M(x)), P(x), M(x) \in Zp[x], K is a possibly
 * very large non-negative integer.
 */

	static
	void
Zpx_poly_to_int_pow_mod_poly (

struct calib_Zpx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zpx_obj *	poly,		/* IN: the polynomial P(x) */
mpz_srcptr			exponent,	/* IN: the power K */
const struct calib_Zpx_obj *	modpoly		/* IN: the modulus polynomial M(x) */
)
{
int				mdeg;
const struct calib_Zpx_dom *	K_of_x;		/* IN: polynomial ring */
struct calib_Zpx_obj *		reg1;
struct calib_Zpx_obj *		reg2;
mpz_t				POWER;

	CALIB_FATAL_ERROR_IF ((rop      EQ NULL) OR
			      (poly     EQ NULL) OR
			      (exponent EQ NULL) OR
			      (modpoly  EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((poly    -> dom NE K_of_x) OR
			      (modpoly -> dom NE K_of_x));

	mdeg = modpoly -> degree;

	mpz_init_set (POWER, exponent);

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 2, mdeg - 1);
	reg1	= &TMP [0];		/* Successive squarings of poly. */
	reg2	= &TMP [1];		/* Result. */

	Zpx_div (NULL, reg1, poly, modpoly);
	reg2 -> degree	  = 0;
	mpz_set_ui (&(reg2 -> coeff [0]), 1);

	while (mpz_sgn (POWER) > 0) {
		if (mpz_odd_p (POWER)) {
			Zpx_poly_mul_mod_poly (reg2, reg2, reg1, modpoly);
		}
		Zpx_poly_mul_mod_poly (reg1, reg1, reg1, modpoly);

		/* POWER >>= 1. */
		mpz_tdiv_q_2exp (POWER, POWER, 1);
	}

	Zpx_set (rop, reg2);

	FREE_LOCAL_POLY_ARRAY (TMP, 2);

	mpz_clear (POWER);
}

/*
 * Compute ((A(x) * B(x)) mod M(x)), where A(x), B(x), M(x) \in Zp[x].
 */

	static
	void
Zpx_poly_mul_mod_poly (

struct calib_Zpx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zpx_obj *	op1,		/* IN: the polynomial A(x) */
const struct calib_Zpx_obj *	op2,		/* IN: the polynomial B(x) */
const struct calib_Zpx_obj *	modpoly		/* IN: the modulus polynomial M(x) */
)
{
int				i, j, adeg, mdeg, maxdeg;
int				r1deg, r2deg;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
struct calib_Zpx_obj		*ap, *bp, *reg1, *reg2, *temp_ptr;
mpz_ptr				apcoeff, bpcoeff, r1coeff, r2coeff, modpcoeff;
mpz_t				BI, FACTOR, TEMP;

	CALIB_FATAL_ERROR_IF ((rop     EQ NULL) OR
			      (op1     EQ NULL) OR
			      (op2     EQ NULL) OR
			      (modpoly EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1     -> dom NE K_of_x) OR
			      (op2     -> dom NE K_of_x) OR
			      (modpoly -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	mdeg	= modpoly -> degree;
	maxdeg	= mdeg - 1;		/* max degree of something mod M(X). */

	/* Verify that modpoly is monic. */
	CALIB_FATAL_ERROR_IF ((mdeg < 1) OR
			      (mpz_cmp_ui (&(modpoly -> coeff [mdeg]), 1) NE 0));

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 4, mdeg - 1);
	temp_ptr = &TMP [0];
	ap	= temp_ptr++;		/* A(x) mod M(x) */
	bp	= temp_ptr++;		/* B(x) mod M(x) */
	reg1	= temp_ptr++;		/* Result. */
	reg2	= temp_ptr++;		/* bi*A(x) */

	Zpx_div (NULL, ap, op1, modpoly);
	Zpx_div (NULL, bp, op2, modpoly);

	adeg = ap -> degree;

	reg1 -> degree	  = -1;

	apcoeff		= ap -> coeff;
	bpcoeff		= bp -> coeff;
	r1coeff		= reg1 -> coeff;
	r2coeff		= reg2 -> coeff;
	modpcoeff	= modpoly -> coeff;

	mpz_init (BI);
	mpz_init (FACTOR);
	mpz_init (TEMP);

	for (i = bp -> degree; i >= 0; i--) {
		/* Compute reg1 = ((reg1 * x**1) mod M(x)).	*/
		/* This just shifts the coefficients left,	*/
		/* mod-ing out those that fall off the end.	*/
		r1deg = reg1 -> degree;
		if (r1deg < maxdeg) {
			/* Simple shift suffices. */
			++r1deg;
			for (j = r1deg; j > 0; j--) {
				mpz_set (&r1coeff [j], &r1coeff [j-1]);
			}
			mpz_set_ui (&r1coeff [0], 0);
		}
		else {
			/* Need to mod out while shifting. */
			mpz_set (FACTOR, &r1coeff [r1deg]);
			for (j = r1deg; j > 0; j--) {
				cf -> mul (cf, TEMP, FACTOR, &modpcoeff [j]);
				cf -> sub (cf, &r1coeff [j], &r1coeff [j-1], TEMP);
			}
			cf -> mul (cf, TEMP, FACTOR, &modpcoeff [0]);
			cf -> neg (cf, &r1coeff [0], TEMP);
		}
		while ((r1deg >= 0) AND (mpz_sgn (&r1coeff [r1deg]) EQ 0)) {
			--r1deg;
		}
		reg1 -> degree = r1deg;

		mpz_set (BI, &bpcoeff [i]);
		for (j = 0; j <= adeg; j++) {
			cf -> mul (cf, &r2coeff [j], BI, &apcoeff [j]);
		}
		r2deg = adeg;
		while ((r2deg >= 0) AND (mpz_sgn (&r2coeff [r2deg]) EQ 0)) {
			--r2deg;
		}
		reg2 -> degree = r2deg;
		Zpx_add (reg1, reg1, reg2);
	}

	Zpx_set (rop, reg1);

	mpz_clear (TEMP);
	mpz_clear (FACTOR);
	mpz_clear (BI);

	FREE_LOCAL_POLY_ARRAY (TMP, 4);
}

/*
 * Free the given factorization, as produced by Zpx_factor().
 */

	static
	void
Zpx_free_factors (

struct calib_Zpx_factor *	factors		/* IN/OUT: factors to free */
)
{
struct calib_Zpx_factor *	ftemp;
struct calib_Zpx_obj *		poly;

	while (factors NE NULL) {
		ftemp = factors -> next;
		factors -> next = NULL;
		poly = factors -> factor;
		if (poly NE NULL) {
			Zpx_free (poly);
		}
		free (factors);
		factors = ftemp;
	}
}

/*
 * Compute the resultant of given polynomials a and b.
 * This is an element of Zp.
 */

	static
	void
Zpx_resultant (

mpz_ptr				result,		/* IN/OUT: Resultant value */
const struct calib_Zpx_obj *	a,		/* IN: Polynomial a */
const struct calib_Zpx_obj *	b		/* IN: Polynomial b */
)
{
int				adeg, bdeg, i;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
const struct calib_Zpx_obj *	tmp;
struct calib_Zpx_obj		poly1, poly2, quot, rem;
mpz_t				value, coeff;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR
			      (a      EQ NULL) OR
			      (b      EQ NULL));

	K_of_x	= a -> dom;

	CALIB_FATAL_ERROR_IF (b -> dom NE K_of_x);

	cf	= K_of_x -> cf;
	adeg	= a -> degree;
	bdeg	= b -> degree;

	if (adeg <= 0) {
		CALIB_FATAL_ERROR_IF (adeg < -1);

		/* Polynomial a is constant. */
		if (bdeg <= 0) {
			CALIB_FATAL_ERROR_IF (bdeg < -1);
			/* Resultant of (0,0), (1,0), (0,1), (93,-47)	*/
			/* are all 1.					*/
			mpz_set_ui (result, 1);
		}
		else if (adeg < 0) {
			/* When deg(f) >= 1, Resultant(0, f(x)) = 0. */
			mpz_set_ui (result, 0);
		}
		else {
			/* When adeg = 0 and bdeg > 0, res(a,b) = a**bdeg. */
			mpz_powm_ui (result, &(a -> coeff [0]), bdeg, cf -> p);
		}
		return;
	}
	if (bdeg <= 0) {
		CALIB_FATAL_ERROR_IF (bdeg < -1);
		/* Polynomial b is constant, deg(a) > 0. */
		if (bdeg EQ -1) {
			/* res(a(x),0) ==> 0. */
			mpz_set_ui (result, 0);
		}
		else {
			/* When adeg > 0 and bdeg = 0, res(a,b) = b**adeg. */
			mpz_powm_ui (result, &(b -> coeff [0]), adeg, cf -> p);
		}
		return;
	}

	/* The general case: (adeg > 0) AND (bdeg > 0). */
	CALIB_FATAL_ERROR_IF (NOT ((adeg > 0) AND (bdeg > 0)));

	mpz_init (value);
	mpz_init (coeff);

	mpz_set_ui (value, 1);
	if (bdeg > adeg) {
		/* Swap a and b. */
		tmp	= a;
		a	= b;
		b	= tmp;
		if (((adeg & 1) NE 0) AND ((bdeg & 1) NE 0)) {
			/* Both degrees are odd.  Negate the value. */
			cf -> neg (cf, value, value);
		}
	}

	Zpx_init (K_of_x, &poly1);
	Zpx_init (K_of_x, &poly2);
	Zpx_init (K_of_x, &quot);
	Zpx_init (K_of_x, &rem);

	Zpx_set (&poly1, a);
	Zpx_set (&poly2, b);

	for (;;) {
		/* From Wikipedia: let					*/
		/*	A = a0 x^d + a1 x^{d-1} + ... ad		*/
		/*	B = b0 x^e + b1 x^{e-1} + ... be		*/
		/* be nonzero polynomials of degree d and e,		*/
		/* respectively.  If C is another polynomial such that	*/
		/* the degree of A - C B is \delta, then		*/
		/*	res(B, A - C B) = b0^{\delta - d} res(B, A).	*/
		/*							*/
		/* We exploit this here as follows: let			*/
		/*	poly1 = quot * poly2 + rem			*/
		/* Then if we let					*/
		/*	A = rem						*/
		/*	B = poly2					*/
		/*	C = -quot					*/
		/*	A - C B = rem + quot * poly2 = poly1		*/
		/* the above yields:					*/
		/*	res(poly2, poly1)				*/
		/*	= lc(poly2)**{\delta-d} * res(poly2, rem)	*/
		/* Furthermore, since res(A,B) = (-1)**{d*e}*res(B,A),	*/
		/* we have						*/
		/*	res(poly1, poly2)				*/
		/*	= (-1)**(poly1.degree * poly2.degree)		*/
		/*	  * lc(poly2)**{\delta-d}			*/
		/*	  * res(poly2, rem).				*/

		if (((poly1.degree & 1) NE 0) AND ((poly2.degree & 1) NE 0)) {
			/* Both polys have odd degree.  Negate. */
			cf -> neg (cf, value, value);
		}

		/* Perform the division. */
		Zpx_div (&quot, &rem, &poly1, &poly2);

		i = poly1.degree - rem.degree;
		CALIB_FATAL_ERROR_IF (i <= 0);

		mpz_powm_ui (coeff, &poly2.coeff [poly2.degree], i, cf -> p);

		cf -> mul (cf, value, value, coeff);

		if (rem.degree <= 0) break;

		Zpx_set (&poly1, &poly2);
		Zpx_set (&poly2, &rem);
	}

	if (rem.degree < 0) {
		mpz_set_ui (value, 0);
	}
	else {
		mpz_powm_ui (coeff, &rem.coeff [0], poly2.degree, cf -> p);
		cf -> mul (cf, value, value, coeff);
	}

	mpz_set (result, value);

	Zpx_clear (&rem);
	Zpx_clear (&quot);
	Zpx_clear (&poly2);
	Zpx_clear (&poly1);

	mpz_clear (coeff);
	mpz_clear (value);
}

/*
 * Compute the derivative of the given polynomial in Zp[x].
 */

	static
	void
Zpx_derivative (

struct calib_Zpx_obj *		rop,	/* IN/OUT: derivative polynomial */
const struct calib_Zpx_obj *	op	/* IN: polynomial to differentiate */
)
{
int				i, deg, rsize;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_ptr				result;
mpz_srcptr			pcoeff;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	deg	= op -> degree;
	pcoeff	= op -> coeff;
	rsize	= deg;
	result	= calib_new_Z_vector (rsize);

	mpz_init (TEMP);

	for (i = 1; i <= deg; i++) {
		cf -> set_si (cf, TEMP, i);
		cf -> mul (cf, &result [i - 1], TEMP, &pcoeff [i]);
	}

	mpz_clear (TEMP);

	do {
		--deg;
	} while ((deg >= 0) AND (mpz_sgn (&result [deg]) EQ 0));

	Zpx_realloc_copy_poly (rop, result, deg);
	calib_free_Z_vector (result, rsize);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically zero.
 */

	static
	calib_bool
Zpx_zerop (

const struct calib_Zpx_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg >= 0) {
		return (FALSE);
	}

	CALIB_FATAL_ERROR_IF (deg < -1);

	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically one.
 */

	static
	calib_bool
Zpx_onep (

const struct calib_Zpx_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg NE 0) {
		return (FALSE);
	}

	if (mpz_sgn (&(op -> coeff [0])) EQ 0) {
		return (TRUE);
	}

	return (FALSE);
}

/*
 * Convert the given general expression into a member of Zp[x].
 * All of the variables in the given expression *MUST* match the given
 * variable name.
 */

	static
	void
Zpx_set_genrep (

struct calib_Zpx_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_genrep *	op,	/* IN: expression to convert */
const char *			var	/* IN: main var for polynomial */
)
{
int				i;
const struct calib_Zpx_dom *	K_of_x;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_Zpx_obj		POLY;
struct calib_Zpx_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (var EQ NULL));

	K_of_x	= rop -> dom;

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Zpx_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Zpx_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		Zpx_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		CALIB_FATAL_ERROR_IF (strcmp (op -> u.var, var) NE 0);
		Zpx_set_var_power (rop, 1);
		break;

	case CALIB_GENREP_OP_ADD:
		Zpx_init (K_of_x, &POLY);
		Zpx_init (K_of_x, &TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zpx_set_genrep (&TEMP, lp -> operand, var);
			Zpx_add (&POLY, &POLY, &TEMP);
		}
		Zpx_set (rop, &POLY);

		Zpx_clear (&TEMP);
		Zpx_clear (&POLY);
		break;

	case CALIB_GENREP_OP_MUL:
		Zpx_init (K_of_x, &POLY);
		Zpx_init (K_of_x, &TEMP);
		Zpx_set_si (&POLY, 1);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zpx_set_genrep (&TEMP, lp -> operand, var);
			Zpx_mul (&POLY, &POLY, &TEMP);
		}
		Zpx_set (rop, &POLY);

		Zpx_clear (&TEMP);
		Zpx_clear (&POLY);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;
		CALIB_FATAL_ERROR_IF (i < 0);

		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);

		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			/* Optimize important special case. */
			CALIB_FATAL_ERROR_IF (strcmp (gp -> u.var, var) NE 0);
			Zpx_set_var_power (rop, i);
		}
		else {
			Zpx_init (K_of_x, &TEMP);
			Zpx_set_genrep (&TEMP, op -> u.ipow.base, var);
			Zpx_ipow (rop, &TEMP, i);
			Zpx_clear (&TEMP);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Convert the given element of Zp[x] into general representation.
 * Use the variable name given.
 */

	static
	struct calib_genrep *
Zpx_to_genrep (

const struct calib_Zpx_obj *	op,	/* IN: polynomial to convert */
const char *			var	/* IN: variable name to use */
)
{
int				i, degree;
mpz_srcptr			value;
const struct calib_Zpx_dom *	K_of_x;
const struct calib_Zp_dom *	cf;
mpz_srcptr			coeff;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (var EQ NULL));

	K_of_x	= op -> dom;
	cf	= K_of_x -> cf;

	degree	= op -> degree;
	coeff	= op -> coeff;

	root = NULL;
	hookp = &root;

	for (i = degree; i >= 0; i--) {
		value = &coeff [i];
		if (mpz_sgn (value) EQ 0) continue;
		vgp = cf -> to_genrep (cf, value);
		gp1 = calib_genrep_poly_term (vgp, var, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	return (gp1);
}

/*
 * Convert the given list of factors into general representation.
 * Use the variable given.
 */

	static
	struct calib_genrep *
Zpx_factors_to_genrep (

const struct calib_Zpx_factor *	factors,/* IN: list of factors */
const char *			var	/* IN: variable to use */
)
{
int				pow;
const struct calib_Zpx_factor *	fp;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;

	root = NULL;
	hookp = &root;

	for (fp = factors; fp NE NULL; fp = fp -> next) {
		gp1 = Zpx_to_genrep (fp -> factor, var);
		pow = fp -> multiplicity;
		CALIB_FATAL_ERROR_IF (pow <= 0);
		if (pow > 1) {
			gp2 = CALIB_NEW (struct calib_genrep);
			gp2 -> op		= CALIB_GENREP_OP_IPOW;
			gp2 -> u.ipow.base	= gp1;
			gp2 -> u.ipow.power	= pow;
			gp1 = gp2;
		}
		lp = calib_genrep_new_list (gp1, NULL);
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (1);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_MUL;
		gp1 -> u.list	= root;
	}
	return (gp1);
}

/*
 * Print out a single polynomial, in a syntax that maxima can read. 
 */

	static
	void
Zpx_print (

const struct calib_Zpx_obj *	op	/* IN: polynomial to print */
)
{
int		col;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	col = 0;
	col = Zpx_print_nnl (op, col);
	printf ("\n");
}

/*
 * Print out a single polynomial in a syntax that maxima can read.
 */

	static
	int
Zpx_print_nnl (

const struct calib_Zpx_obj *
			op,	/* IN: polynomial to print */
int			col	/* IN: starting column number */
)
{
int		i, degree, sign, ndigits, width;
char *		string;
char *		sstr;
const char *	sep;
mpz_srcptr	zp;
char		powbuf [32];

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (col < 0));

	if (op EQ NULL) {
		printf ("\"<null ptr>\"");
		return (col + 12);
	}
	degree = op -> degree;
	if (degree < 0) {
		printf ("0");
		return (col + 1);
	}
	zp = &(op -> coeff [degree]);
	for (i = degree; i >= 0; i--, zp--) {
		sign = mpz_sgn (zp);
		if (sign EQ 0) continue;
		ndigits = calib_genrep_convert_abs_Z_to_decimal_string (zp, &string);
		if (sign < 0) {
			sstr = (i EQ degree) ? "- " : " - ";
		}
		else {
			sstr = (i EQ degree) ? "" : " + ";
		}
		width = strlen (sstr);
		if ((i > 0) AND (ndigits EQ 1) AND (string [0] EQ '1')) {
			/* 1*var^pow ==> var^pow. */
			string [0] = '\0';
			ndigits = 0;
		}
		else {
			width += ndigits;
		}
		powbuf [0] = '\0';
		if (i > 0) {
			sep = (ndigits > 0) ? "*" : "";
			if (i EQ 1) {
				sprintf (powbuf, "%sx", sep);
			}
			else {
				sprintf (powbuf, "%sx^%d", sep, i);
			}
			width += strlen (powbuf);
		}
		if (col + width >= 78) {
			printf ("\n");
			col = 0;
		}
		printf ("%s%s%s", sstr, string, powbuf);
		free (string);
	}
	return (col);
}

/*
 * Prototype Zpx domain object.
 */

static const struct calib_Zpx_dom	Zpx_dom = {
	.init			= Zpx_init,
	.init_degree		= Zpx_init_degree,
	.alloc			= Zpx_alloc,
	.clear			= Zpx_clear,
	.set			= Zpx_set,
	.set_z			= Zpx_set_z,
	.set_q			= Zpx_set_q,
	.set_Zx			= Zpx_set_Zx,
	.set_Qa			= Zpx_set_Qa,
	.add			= Zpx_add,
	.sub			= Zpx_sub,
	.neg			= Zpx_neg,
	.mul			= Zpx_mul,
	.mul_z			= Zpx_mul_z,
	.ipow			= Zpx_ipow,
	.dup			= Zpx_dup,
	.free			= Zpx_free,
	.monicize		= Zpx_monicize,
	.eval			= Zpx_eval,
	.div			= Zpx_div,
	.gcd			= Zpx_gcd,
	.extgcd			= Zpx_extgcd,
	.factor			= Zpx_factor,
	.factor_square_free	= Zpx_factor_square_free,
	.finish_sqf		= Zpx_finish_sqf,
	.free_factors		= Zpx_free_factors,
	.resultant		= Zpx_resultant,
	.derivative		= Zpx_derivative,
	.zerop			= Zpx_zerop,
	.onep			= Zpx_onep,
	.set_genrep		= Zpx_set_genrep,
	.to_genrep		= Zpx_to_genrep,
	.factors_to_genrep	= Zpx_factors_to_genrep,
	.print_maxima		= Zpx_print,
};

/*
 * Construct the ring of polynomials over field K, where K is
 * the given field GF(p^k).
 */

	struct calib_Zpx_dom *
calib_make_Zpx_dom (

const struct calib_Zp_dom *	K		/* IN: coefficient field */
)
{
struct calib_Zpx_dom *	K_of_x;

	K_of_x	= CALIB_NEW (struct calib_Zpx_dom);

	/* Initialize "most" things. */
	*K_of_x	= Zpx_dom;

	K_of_x -> cf		= K;
	K_of_x -> settings	= calib_Zpx_default_settings;

	return (K_of_x);
}

/*
 * Free up the given Zp[x] ring description.
 */

	void
calib_free_Zpx_dom (

struct calib_Zpx_dom *	K_of_x	/* IN: Zp[x] ring description */
)
{
	if (K_of_x NE NULL) {
		/* No subordinate data to free. */
		memset (K_of_x, 0, sizeof (*K_of_x));
		free (K_of_x);
	}
}
