/**********************************************************************

	$Id: berle.c,v 1.1 2024/10/20 00:36:26 warme Exp $

	File:	berle.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2007, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Berlekamp's factoring algorithm for polynomials in Z_p[x].

***********************************************************************

	Modification Log:

	a-1:	08/14/2020	warme
		: Created from earlier stand-alone code.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "berle.h"

#include "calib/cputime.h"
#include "calib/fatal.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/random.h"
#include "calib/types.h"
#include "calib/Zpx.h"
#include <gmp.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct calib_Zpx_factor *
		 _calib_Zpx_berle_factor (
				const struct calib_Zpx_obj *	poly,
				struct calib_Random *		randp);

/*
 * Local Types
 */

typedef unsigned int	COEFF;

typedef struct {
	int		degree;
	COEFF *		coeff;
} POLY;

/*
 * Internal Routines
 */

static void		compute_Q_matrix (COEFF *	Q,
					  const POLY *	poly,
					  COEFF		modulus);
static void		compute_Q_matrix_large_p (
					COEFF *		Q,
					const POLY *	poly,
					COEFF		modulus);
static void		compute_Q_matrix_small_p (
					COEFF *		Q,
					const POLY *	poly,
					COEFF		modulus);
static void		compute_null_space (
					COEFF *		Q,
					int		N,
					COEFF *		V,
					int *		num_vect,
					COEFF		p);
static struct calib_Zpx_obj *
			convert_poly_to_Zpx (
					const struct calib_Zpx_dom *	dom,
					const POLY *			src);
static void		fix_degree (POLY * poly);
static COEFF		inv (COEFF num, COEFF p);
static void		poly_clear (POLY * poly);
static void		poly_div_mod (POLY *		quotient,
				      POLY *		remainder,
				      const POLY *	dividend,
				      const POLY *	divisor,
				      COEFF		modulus);
static void		poly_expt_mod_poly (POLY *		result,
					    const POLY *	a,
					    COEFF		pow,
					    const POLY *	modpoly,
					    COEFF		p);
static void		poly_gcd_mod (POLY *		result,
				      const POLY *	poly1,
				      const POLY *	poly2,
				      COEFF		p);
static void		poly_init_degree (POLY * dst, int degree);
static void		poly_init_set (POLY * dst, const POLY * src);
static void		poly_monicize_mod (POLY * poly, COEFF p);
static void		poly_mul_mod_poly (POLY *	result,
					   const POLY *	a,
					   const POLY *	b,
					   const POLY *	modpoly,
					   COEFF	p);
static void		poly_print (const POLY * poly);
static void		poly_set (POLY * dst, const POLY * src);
static struct calib_Zpx_factor *
			recover_factors (
				const struct calib_Zpx_dom *	dom,
				const POLY *			poly,
				COEFF *				V,
				int				R,
				int				N,
				COEFF				p,
				struct calib_Random *		randp);
static void		recover_factors_large_p (
					POLY *			roots,
					COEFF *			V,
					int			R,
					int			N,
					COEFF			p,
					struct calib_Random *	randp);
static void		recover_factors_small_p (
					POLY *		roots,
					COEFF *		V,
					int		R,
					int		N,
					COEFF		p);

/*
 * Fast coefficient arithmetic.
 */

	static
	inline
	COEFF
cadd (

COEFF		a,		/* IN: First operand */
COEFF		b,		/* IN: Second operand */
COEFF		modulus		/* IN: Modulus */
)
{
calib_int64u	da, db;
calib_int32u	result;

	da	= a;
	db	= b;
	result	= (da + db) % modulus;

	return (result);
}

	static
	inline
	COEFF
csub (

COEFF		a,		/* IN: First operand */
COEFF		b,		/* IN: Second operand */
COEFF		modulus		/* IN: Modulus */
)
{
calib_int64u	da, db;
calib_int32u	result;

	da	= a;
	db	= b;
	if (da < db) {
		da += modulus;
		CALIB_FATAL_ERROR_IF (da < db);
	}
	result	= (da - db) % modulus;

	return (result);
}

	static
	inline
	COEFF
cmul (

COEFF		a,		/* IN: First operand */
COEFF		b,		/* IN: Second operand */
COEFF		modulus		/* IN: Modulus */
)
{
calib_int64u	da, db;
calib_int32u	result;

	da	= a;
	db	= b;
	result	= (da * db) % modulus;

	return (result);
}

/*
 * This routine takes a polynomial in the given modulus, and factors it.
 * The given polynomial must be monic and square-free.
 */

	struct calib_Zpx_factor *
_calib_Zpx_berle_factor (

const struct calib_Zpx_obj *	poly,	/* IN: Polynomial to factor */
struct calib_Random *		randp	/* IN/OUT: Random number generator */
)
{
int				i, N;
calib_cpu_time_t		t0, t1;
const struct calib_Zpx_dom *	dom;
COEFF				p;
const struct calib_Zp_dom *	Zp;
COEFF *				Q;
COEFF *				qp;
COEFF * 			V;
int				R;
struct calib_Zpx_factor *	nfp;
struct calib_Zpx_factor *	list;
POLY				modpoly;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (randp EQ NULL));

	dom	= poly -> dom;
	Zp	= dom -> cf;

	CALIB_FATAL_ERROR_IF (Zp EQ NULL);

	/* We want the prime modulus to fit in 32 bits so that we can	*/
	/* multiply two coefficients (staying in 64 bits) and reduce	*/
	/* the result (mod p) all in fast, hardware 64-bit unsigned	*/
	/* long arithmetic.						*/
	CALIB_FATAL_ERROR_IF (NOT mpz_fits_uint_p (Zp -> p));

	p = mpz_get_ui (Zp -> p);

	CALIB_FATAL_ERROR_IF (p <= 1);

	N = poly -> degree;

	if (N <= 1) {
		CALIB_FATAL_ERROR_IF (N <= 0);

		/* All linear factors are irreducible. */
		nfp = CALIB_NEW (struct calib_Zpx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= dom -> dup (poly);
		nfp -> next		= NULL;

		return (nfp);
	}

	t0 = calib_get_cpu_time ();

	/* Given polynomial *must* already be monic! */
	CALIB_FATAL_ERROR_IF (mpz_cmp_ui (&(poly -> coeff [N]), 1) NE 0);

	/* Get into our local polynomial form, converting to (mod p). */
	modpoly.degree	= N;
	modpoly.coeff	= CALIB_NEWA (N + 1, COEFF);
	for (i = 0; i <= N; i++) {
		modpoly.coeff [i] = mpz_fdiv_ui (&(poly -> coeff [i]), p);
	}

	Q = CALIB_NEWA (N * N, COEFF);
	compute_Q_matrix (Q, &modpoly, p);

	/* Now subtract the identity matrix I from the Q matrix. */
	qp = Q;
	for (i = 0; i < N; i++) {
		if (*qp EQ 0) {
			*qp = p - 1;
		}
		else {
			--*qp;
		}
		qp += (N + 1);
	}
	V = CALIB_NEWA (N * N, COEFF);
	compute_null_space (Q, N, V, &R, p);

	list = recover_factors (dom, &modpoly, V, R, N, p, randp);

	free (V);
	free (Q);
	free (modpoly.coeff);

	t1 = calib_get_cpu_time ();
	t1 -= t0;

	i = 0;
	for (nfp = list; nfp NE NULL; nfp = nfp -> next) {
		++i;
	}

#if 0
	printf ("Berlekamp:"
		" Degree = %d, p = %u, factors = %d, %d.%02d seconds\n",
		N, p, i, t1 / 100, t1 % 100);
#endif

	return (list);
}

/*
 * Compute the Q matrix, which is N by N, where N is the degree of the
 * given polynomial f(x).  For 0 <= i < N, the i-th row of this table
 * is the polynomial x^{i*p} (mod f(x)).
 *
 * We have two ways of computing this matrix: one for small primes p,
 * and the other for large primes p.
 */

	static
	void
compute_Q_matrix (

COEFF *		Q,		/* IN/OUT: The Q matrix */
const POLY *	poly,		/* IN: Polynomial being factored */
COEFF		p		/* IN: The prime modulus */
)
{
	CALIB_FATAL_ERROR_IF ((Q EQ NULL) OR (poly EQ NULL));

	if (p <= 13) {
		/* Primes of 4 bits or less. */
		compute_Q_matrix_small_p (Q, poly, p);
	}
	else {
		compute_Q_matrix_large_p (Q, poly, p);
	}
}

/*
 * Compute the Q matrix for small primes p.
 */

	static
	void
compute_Q_matrix_small_p (

COEFF *		Q,		/* IN/OUT: The Q matrix */
const POLY *	poly,		/* IN: Polynomial being factored */
COEFF		p		/* IN: The prime modulus */
)
{
int		N;
COEFF *		A;
COEFF * 	rowp;
COEFF		t;
int		i;
int		j;
int		row;
COEFF		tmp1, tmp2;

	N = poly -> degree;

	A = CALIB_NEWA (N, COEFF);

	rowp = Q;

	memset (A, 0, N * sizeof (A [0]));
	A [0] = 1;

	memcpy (rowp, A, N * sizeof (rowp [0]));
	rowp += N;

	for (row = 1; row < N; row++) {
		/* multiply A by X^p modulo poly. */
		for (j = 0; j < p; j++) {
			/* multiply A by X and take residue modulo poly. */
			i = N - 1;
			t = A [N - 1];
			while (i > 0) {
				tmp1 = A [i - 1];
				tmp2 = (t * poly -> coeff [i]) % p;
				if (tmp1 >= tmp2) {
					tmp1 -= tmp2;
				}
				else {
					tmp1 = (p + tmp1 - tmp2);
				}
				A [i] = tmp1;
				--i;
			}
			tmp1 = (t * poly -> coeff [0]) % p;
			if (tmp1 NE 0) {
				tmp1 = p - tmp1;
			}
			A [0] = tmp1;
		}
		/* now copy A into the next row of the Q matrix. */
		memcpy (rowp, A, N * sizeof (rowp [0]));
		rowp += N;
	}
	free (A);
}

/*
 * Compute the Q matrix for large primes p.
 */

	static
	void
compute_Q_matrix_large_p (

COEFF *		Q,		/* IN/OUT: The Q matrix */
const POLY *	poly,		/* IN: Polynomial being factored */
COEFF		p		/* IN: The prime modulus */
)
{
int		i, N, nnz, nz;
COEFF *		rowp;
POLY		ptmp, xptmp;

	N = poly -> degree;

	/* We had better be factoring degree at least 2. */
	CALIB_FATAL_ERROR_IF (N < 2);

	poly_init_degree (&ptmp,  N - 1);
	poly_init_degree (&xptmp, N - 1);

	/* xptmp = x**p mod poly. */
	xptmp.degree	= 1;
	xptmp.coeff [0]	= 0;
	xptmp.coeff [1]	= 1;
	poly_expt_mod_poly (&xptmp, &xptmp, p, poly, p);

	poly_set (&ptmp, &xptmp);

	/* Row 0 = 1. */
	rowp = Q;
	memset (rowp, 0, N * sizeof (rowp [0]));
	rowp [0] = 1;
	rowp += N;

	/* Row 1 = xptmp. */
	nnz = ptmp.degree + 1;
	CALIB_FATAL_ERROR_IF (nnz > N);
	memcpy (rowp, ptmp.coeff, nnz * sizeof (rowp [0]));
	nz	= N - nnz;
	if (nz NE 0) {
		memset (&rowp [nnz], 0, nz * sizeof (rowp [0]));
	}
	rowp += N;

	for (i = 2; i < N; i++) {
		/* Each additional row is further multiplied by xptmp. */
		poly_mul_mod_poly (&ptmp, &ptmp, &xptmp, poly, p);

		nnz = ptmp.degree + 1;
		CALIB_FATAL_ERROR_IF (nnz > N);
		memcpy (rowp, ptmp.coeff, nnz * sizeof (rowp [0]));
		nz	= N - nnz;
		if (nz NE 0) {
			memset (&rowp [nnz], 0, nz * sizeof (rowp [0]));
		}
		rowp += N;
	}

	poly_clear (&xptmp);
	poly_clear (&ptmp);
}

	static
	void
compute_null_space (

COEFF *		Q,		/* IN/OUT: The Q matrix (modified) */
int		N,		/* IN: Size of the Q matrix */
COEFF *		V,		/* IN/OUT: Null space vectors */
int *		num_vect,	/* IN/OUT: num vectors in the null space. */
COEFF		p		/* IN: The prime modulus */
)
{
int		i, j, k, col;
COEFF *		rowp;
COEFF *		colp;
COEFF *		colp2;
COEFF		z;
int *		C;
int		R;

	if (N < 1) {
		*num_vect = 0;
		return;
	}

	C = CALIB_NEWA (N, int);
	for (i = 0; i < N; i++) {
		C [i] = -1;
	}

	R = 0;

	rowp = Q;
	for (k = 0; k < N; k++) {
		for (j = 0; j < N; j++) {
			if ((rowp [j] NE 0) AND (C [j] < 0)) break;
		}
		if (j < N) {
			/* multiply column J by -1/Q[k,j] so that this */
			/* "pivot" cell becomes -1.		       */
			colp = &rowp [j];
			z = p - inv (*colp, p);
			for (i = k; i < N; i++) {
				*colp = cmul (z, *colp, p);
				colp += N;
			}
			for (col = 0; col < N; col++) {
				if (col EQ j) continue;
				colp2 = &rowp [col];
				colp  = &rowp [j];
				z = *colp2;
				for (i = k; i < N; i++) {
					*colp2 = cadd (*colp2, cmul (z, *colp, p), p);
					colp2 += N;
					colp  += N;
				}
			}
			C [j] = k;
		}
		else {
			for (j = 0; j < N; j++) {
				for (i = 0; i < N; i++) {
					if (C [i] EQ j) break;
				}
				if (i < N) {
					*V++ = rowp [i];
				}
				else if (j EQ k) {
					*V++ = 1;
				}
				else {
					*V++ = 0;
				}
			}
			++R;
		}
		rowp += N;
	}
	*num_vect = R;

	free (C);
}

/*
 * Recover the factors from the given Null Space vectors V.
 *
 * There are two ways of doing this:
 *
 *	- A deterministic procedure for small primes.
 *	- A randomized procedure for large primes.
 */

	static
	struct calib_Zpx_factor *
recover_factors (

const struct calib_Zpx_dom *
			dom,	/* IN: Polynomial domain */
const POLY *		poly,	/* IN: Polynomial being factored */
COEFF *			V,	/* IN: Null space vectors */
int			R,	/* IN: Number of null space vectors */
int			N,	/* IN: Size of null space vectors */
COEFF			p,	/* IN: Prime modulus */
struct calib_Random *	randp	/* IN/OUT: Random number generator */
)
{
int				i, ncoeff;
POLY *				roots;
COEFF *				carray;
struct calib_Zpx_factor *	list;
struct calib_Zpx_factor *	nfp;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (V EQ NULL));
	CALIB_FATAL_ERROR_IF (R <= 0);
	CALIB_FATAL_ERROR_IF (N <= 0);

	/* Create array of polynomials to hold factors we discover. */
	roots	= CALIB_NEWA (R, POLY);
	ncoeff	= poly -> degree + 1;
	carray	= CALIB_NEWA (R * ncoeff, COEFF);
	for (i = 0; i < R; i++) {
		roots [i].degree	= -1;
		roots [i].coeff		= &carray [i * ncoeff];
	}

	/* Copy the polynomial to be factored into the first element. */
	poly_set (&roots [0], poly);

	if (p <= 17) {
		recover_factors_small_p (roots, V, R, N, p);
	}
	else {
		recover_factors_large_p (roots, V, R, N, p, randp);
	}

	/* Create linked list of factors. */

	list = NULL;
	for (i = 0; i < R; i++) {
		nfp = CALIB_NEW (struct calib_Zpx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= convert_poly_to_Zpx (dom,
							       &roots [i]);
		nfp -> next		= list;
		list = nfp;
	}

	free (carray);
	free (roots);

	return (list);
}

/*
 * The deterministic procedure for recovering factors.  This is
 * really only useful for very small primes.
 */

	static
	void
recover_factors_small_p (

POLY *		roots,	/* IN/OUT: Polynomial factors (initially just one) */
COEFF *		V,	/* IN: Null space vectors */
int		R,	/* IN: Number of null space vectors */
int		N,	/* IN: Size of null space vectors */
COEFF		p	/* IN: Prime modulus */
)
{
int			i, j, s, nroots, deg, cur_nroots;
POLY *			rp;
POLY *			dstp;
POLY			tmp, tmp1;

	CALIB_FATAL_ERROR_IF ((roots EQ NULL) OR (V EQ NULL));
	CALIB_FATAL_ERROR_IF (R <= 0);
	CALIB_FATAL_ERROR_IF (N <= 0);

	deg	= roots [0].degree;

	CALIB_FATAL_ERROR_IF (N NE deg);

	poly_init_degree (&tmp, deg);
	poly_init_degree (&tmp1, deg);

	nroots = 1;

	V += N;			/* skip first vector... */
	for (j = 1; j < R; j++) {
		/* recover factors from vector V [j]. */
		tmp.degree = N - 1;
		memcpy (tmp.coeff, V, N * sizeof (COEFF));
		fix_degree (&tmp);
		cur_nroots = nroots;
		for (s = 0; s < p; s++) {
			tmp.coeff [0] = s;
			rp = roots;
			for (i = 0; i < cur_nroots; i++) {
				poly_gcd_mod (&tmp1, rp, &tmp, p);
				if ((tmp1.degree >= 1) AND
				    (tmp1.degree < rp -> degree)) {
					dstp = &roots [nroots++];
					poly_set (dstp, &tmp1);
					poly_div_mod (rp,
						      &tmp1,
						      rp,
						      &tmp1,
						      p);
					if (nroots >= R) goto done;
				}
				++rp;
			}
		}
		V += N;
	}
done:

	CALIB_FATAL_ERROR_IF (nroots NE R);

	poly_clear (&tmp1);
	poly_clear (&tmp);
}

/*
 * The randomized procedure for recovering factors.  We use this
 * version for primes that are anything but tiny.
 */

	static
	void
recover_factors_large_p (

POLY *			roots,	/* IN/OUT: Polynomial factors (initially */
				/*	   just one) */
COEFF *			V,	/* IN: Null space vectors */
int			R,	/* IN: Number of null space vectors */
int			N,	/* IN: Size of null space vectors */
COEFF			p,	/* IN: Prime modulus */
struct calib_Random *	randp	/* IN/OUT: Random number generator */
)
{
int		i, j, deg, nroots, cur_nroots;
COEFF		t, pow;
COEFF *		srcrow;
COEFF *		dstrow;
POLY *		rp;
POLY *		dstp;
POLY		tmp1, tmp2, tmp3;

	CALIB_FATAL_ERROR_IF ((roots EQ NULL) OR
			      (V EQ NULL) OR
			      (randp EQ NULL));
	CALIB_FATAL_ERROR_IF (R <= 0);
	CALIB_FATAL_ERROR_IF (N <= 0);

	/* Prime *MUST* be odd! */
	CALIB_FATAL_ERROR_IF ((p & 0x01) NE 1);

	pow = (p - 1) >> 1;

	deg	= roots [0].degree;

	CALIB_FATAL_ERROR_IF (N NE deg);

	poly_init_degree (&tmp1, deg);
	poly_init_degree (&tmp2, deg);
	poly_init_degree (&tmp3, deg);

	nroots = 1;
	while (nroots < R) {
		/* Compute random linear combination of null-space vectors. */
		dstrow = &tmp1.coeff [0];
		memset (dstrow, 0, N * sizeof (dstrow [0]));
		for (i = 0; i < R; i++) {
			t = (calib_random_u32 (randp) % p);
			if (t EQ 0) continue;
			srcrow = &V [i * N];
			if (t EQ 1) {
				for (j = 0; j < N; j++) {
					dstrow [j] = cadd (dstrow [j], srcrow [j], p);
				}
				continue;
			}
			for (j = 0; j < N; j++) {
				dstrow [j] = cadd (dstrow [j],
						   cmul (t, srcrow [j], p),
						   p);
			}
		}
		tmp1.degree = N - 1;
		fix_degree (&tmp1);

		if (tmp1.degree <= 0) {
			/* A rare case, but worth avoiding all of the	*/
			/* useless work below.				*/
			continue;
		}

		/* Use random vector to (hopefully) split existing factors. */
		rp = roots;
		cur_nroots = nroots;
		for (i = 0; i < cur_nroots; i++, rp++) {
			if (rp -> degree <= 1) {
				/* This factor is already irreducible. */
				continue;
			}
			poly_expt_mod_poly (&tmp2, &tmp1, pow, rp, p);

			/* Assure that it is valid to use tmp2.coeff [0]. */
			if (tmp2.degree <= 0) {
				/* A very rare case, but don't bother	*/
				/* computing the GCD!			*/
				continue;
			}
			tmp2.coeff [0] = csub (tmp2.coeff [0], 1, p);

			poly_gcd_mod (&tmp3, &tmp2, rp, p);

			j = tmp3.degree;
			if ((j > 0) AND (j < rp -> degree)) {
				/* We have a non-trivial factor. */
				dstp = &roots [nroots++];
				poly_set (dstp, &tmp3);
				poly_div_mod (rp, &tmp2, rp, &tmp3, p);

				if (nroots >= R) goto done;
			}
		}
	}

done:

	CALIB_FATAL_ERROR_IF (nroots NE R);

	poly_clear (&tmp3);
	poly_clear (&tmp2);
	poly_clear (&tmp1);
}

/*
 * Convert the given source polynomial into a dynamically allocated
 * Zpx_obj.
 */

	static
	struct calib_Zpx_obj *
convert_poly_to_Zpx (

const struct calib_Zpx_dom *	dom,	/* IN: Polynomial domain */
const POLY *			src	/* IN: Source polynomial */
)
{
int			i, deg;
const COEFF *		scoeff;
struct calib_Zpx_obj *	dst;
mpz_ptr			dcoeff;

	CALIB_FATAL_ERROR_IF (src EQ NULL);

	deg	= src -> degree;
	scoeff	= src -> coeff;

	dcoeff = NULL;
	if (deg >= 0) {
		dcoeff = calib_new_Z_vector (deg + 1);
	}

	for (i = 0; i <= deg; i++) {
		mpz_set_ui (&dcoeff [i], scoeff [i]);
	}

	dst = CALIB_NEW (struct calib_Zpx_obj);
	dst -> degree	= deg;
	dst -> size	= deg + 1;
	dst -> dom	= dom;
	dst -> coeff	= dcoeff;

	return (dst);
}

/*
 * Compute result = (a(x) ** pow) mod f(x), all in Z_p[x].
 * The result polynomial must have size to hold the result
 * (at most degree(f(x) - 1).
 */

	static
	void
poly_expt_mod_poly (

POLY *		result,		/* IN/OUT: Result polynomial */
const POLY *	a,		/* IN: Polynomial to exponentiate */
COEFF		pow,		/* IN: Power to take said polynomial to */
const POLY *	modpoly,	/* IN: Polynomial to mod out by */
COEFF		p		/* IN: Prime modulus */
)
{
int		i, fdeg;
POLY		ptmp, sqtmp, quot;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR
			      (a EQ NULL) OR
			      (modpoly EQ NULL));

	fdeg	= modpoly -> degree;

	i = a -> degree - fdeg;
	if (i <= 0) {
		i = 1;
	}

	poly_init_degree (&ptmp,  fdeg - 1);
	poly_init_degree (&sqtmp, fdeg - 1);
	poly_init_degree (&quot,  i);

	/* Set sqtmp = a, but reducing it mod f(x). */
	poly_div_mod (&quot, &sqtmp, a, modpoly, p);

	ptmp.degree	= 0;
	ptmp.coeff [0]	= 1;

	while (pow > 0) {
		if ((pow & 0x01) NE 0) {
			poly_mul_mod_poly (&ptmp, &ptmp, &sqtmp, modpoly, p);
		}
		poly_mul_mod_poly (&sqtmp, &sqtmp, &sqtmp, modpoly, p);
		pow >>= 1;
	}

	poly_set (result, &ptmp);

	poly_clear (&quot);
	poly_clear (&sqtmp);
	poly_clear (&ptmp);
}

/*
 * Compute result = (a(x) * b(x)) mod f(x), all in Z_p[x].
 * The result polynomial must have size to hold the result
 * (at most degree(f(x)) - 1).
 */

	static
	void
poly_mul_mod_poly (

POLY *		result,		/* IN/OUT: Result polynomial */
const POLY *	a,		/* IN: First multiplicand */
const POLY *	b,		/* IN: Second multiplicand */
const POLY *	modpoly,	/* IN: Polynomial to mod out by */
COEFF		p		/* IN: Prime modulus */
)
{
int		i, j, adeg, bdeg, fdeg;
COEFF		t;
const COEFF *	acoeff;
const COEFF *	bcoeff;
const COEFF *	fcoeff;
COEFF *		rcoeff;
POLY		la, lb, quot;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR
			      (a EQ NULL) OR
			      (b EQ NULL) OR
			      (modpoly EQ NULL));

	fdeg	= modpoly -> degree;
	fcoeff	= modpoly -> coeff;

	/* modpoly *must* be monic. */
	CALIB_FATAL_ERROR_IF (fcoeff [fdeg] NE 1);

	poly_init_degree (&la, fdeg - 1);
	poly_init_degree (&lb, fdeg - 1);

	/* Compute space needed for quotient. */
	adeg	= a -> degree;
	bdeg	= b -> degree;
	i = adeg - fdeg;
	if (i <= 0) {
		i = 1;
	}
	j = bdeg - fdeg;
	if (j <= 0) {
		j = 1;
	}
	if (j > i) {
		i = j;
	}
	poly_init_degree (&quot, i);

	/* Reduce given polys a and b mod f(x) into local polys la and lb. */
	poly_div_mod (&quot, &la, a, modpoly, p);
	poly_div_mod (&quot, &lb, b, modpoly, p);

	adeg	= la.degree;
	acoeff	= la.coeff;
	bdeg	= lb.degree;
	bcoeff	= lb.coeff;

	rcoeff = CALIB_NEWA (fdeg, COEFF);
	memset (rcoeff, 0, fdeg * sizeof (rcoeff [0]));

	for (i = adeg; i >= 0; i--) {
		/* Multiply rcoeff buffer by x (shift coefficients). */
		t = rcoeff [fdeg - 1];
		memmove (&rcoeff [1],
			 &rcoeff [0],
			 (fdeg - 1) * sizeof (rcoeff [0]));
		rcoeff [0] = 0;
		if (t EQ 0) {
			/* Do nothing. */
		}
		else if (t EQ 1) {
			for (j = 0; j < fdeg; j++) {
				rcoeff [j] = csub (rcoeff [j], fcoeff [j], p);
			}
		}
		else {
			for (j = 0; j < fdeg; j++) {
				rcoeff [j] = csub (rcoeff [j],
						   cmul (t, fcoeff [j], p),
						   p);
			}
		}
		t = acoeff [i];
		if (t EQ 0) {
			/* Do nothing. */
		}
		else if (t EQ 1) {
			for (j = 0; j <= bdeg; j++) {
				rcoeff [j] = cadd (rcoeff [j], bcoeff [j], p);
			}
		}
		else {
			for (j = 0; j <= bdeg; j++) {
				rcoeff [j] = cadd (rcoeff [j],
						   cmul (t, bcoeff [j], p),
						   p);
			}
		}
	}

	memcpy (result -> coeff, rcoeff, fdeg * sizeof (result -> coeff [0]));
	result -> degree = fdeg - 1;
	fix_degree (result);

	free (rcoeff);
	poly_clear (&quot);
	poly_clear (&lb);
	poly_clear (&la);
}

/*
 * Compute the GCD of two polynomials in Zp[x].
 */

	static
	void
poly_gcd_mod (

POLY *		result,		/* IN/OUT: Result polynomial */
const POLY *	poly1,		/* IN: First operand */
const POLY *	poly2,		/* IN: Second operand */
COEFF		p		/* IN: Prime modulus */
)
{
int		i;
COEFF		tmp1, tmp2;
POLY *		p1;
POLY *		p2;
POLY *		tmp;
COEFF *		cp1;
COEFF *		cp2;
COEFF *		endp;
POLY		a, b;

	poly_init_set (&a, poly1);
	poly_init_set (&b, poly2);

	/* Force each polynomial to be monic. */
	poly_monicize_mod (&a, p);
	poly_monicize_mod (&b, p);

	p1 = &a;
	p2 = &b;
	for (;;) {
		if (p1 -> degree > p2 -> degree) {
			tmp = p1;
			p1 = p2;
			p2 = tmp;
		}
		if (p1 -> degree < 0) break;
		cp1 = &(p1 -> coeff [p1 -> degree]);
		cp2 = &(p2 -> coeff [p2 -> degree]);
		endp = &(p1 -> coeff [0]);
		while (cp1 >= endp) {
			tmp1 = *cp1;
			tmp2 = *cp2;
			if (tmp2 < tmp1) {
				tmp2 += p;
			}
			*cp2 = tmp2 - tmp1;
			--cp1;
			--cp2;
		}
		do {
			i = --(p2 -> degree);
		} while ((i >= 0) AND (p2 -> coeff [i] EQ 0));
		poly_monicize_mod (p2, p);
	}
	poly_set (result, p2);

	poly_clear (&a);
	poly_clear (&b);
}

/*
 * Divide two polynomials in Zp[x], yielding quotient and remainder.
 * Zp is a field, so this always works unless the divisor is zero.
 */

	static
	void
poly_div_mod (

POLY *		quotient,	/* IN/OUT: Quotient polynomial */
POLY *		remainder,	/* IN/OUT: Remainder polynomial */
const POLY *	dividend,	/* IN: Dividend polynomial */
const POLY *	divisor,	/* IN: Divisor polynomial */
COEFF		modulus		/* IN: Prime modulus */
)
{
int		i, j;
int		num_deg, den_deg, quot_deg;
COEFF		factor, c, tmp1, tmp2;
POLY		num, den;

	CALIB_FATAL_ERROR_IF ((quotient EQ NULL) OR
			      (remainder EQ NULL) OR
			      (dividend EQ NULL) OR
			      (divisor EQ NULL));

	num_deg = dividend -> degree;
	den_deg = divisor -> degree;

	if (num_deg < den_deg) {
		quotient -> degree = -1;
		poly_set (remainder, dividend);
		return;
	}

	quot_deg = num_deg - den_deg;
	poly_init_set (&num, dividend);
	poly_init_set (&den, divisor);

	factor = inv (den.coeff [den_deg], modulus);
	quotient -> degree = quot_deg;
	while (quot_deg >= 0) {
		i = den_deg;
		j = num_deg;
		c = cmul (num.coeff [num_deg], factor, modulus);
		quotient -> coeff [quot_deg] = c;
		while (i >= 0) {
			tmp1 = num.coeff [j];
			tmp2 = cmul (c, den.coeff [i], modulus);
			num.coeff [j] = csub (tmp1, tmp2, modulus);
			--i;
			--j;
		}
		--quot_deg;
		--num_deg;
	}
	num.degree = num_deg;
	poly_set (remainder, &num);
	fix_degree (quotient);
	fix_degree (remainder);

	poly_clear (&den);
	poly_clear (&num);
}

/*
 * Force the given polynomial in Zp[x] to be monic.
 */

	static
	void
poly_monicize_mod (

POLY *		poly,		/* IN/OUT: Polynomial to monicize */
COEFF		p		/* IN: Prime modulus */
)
{
int	i;
int	deg;
COEFF	c;

	fix_degree (poly);
	deg = poly -> degree;
	if (deg <= 0) {
		if (deg < 0) {
			return;
		}
		CALIB_FATAL_ERROR_IF (poly -> coeff [0] EQ 0);
	}
	c = poly -> coeff [deg];
	if (c EQ 1) {
		return;
	}

	c = inv (c, p);

	for (i = 0; i <= deg; i++) {
		poly -> coeff [i] = cmul (c, poly -> coeff [i], p);
	}

	CALIB_FATAL_ERROR_IF (poly -> coeff [deg] NE 1);
}

/*
 * Reduce the degree of the given polynomial until the lead coefficient
 * is non-zero.  If the polynomial is zero, the degree becomes -1.
 */

	static
	void
fix_degree (

POLY *		poly		/* IN/OUT: Polynomial to fix degree of */
)
{
int		i;
COEFF *		coeff;

	i	= poly -> degree;
	coeff	= poly -> coeff;
	while (i >= 0) {
		if (coeff [i] NE 0) break;
		--i;
	}
	poly -> degree = i;
}

/*
 * Initialize the given polynomial to have the given maximum degree.
 */

	static
	void
poly_init_degree (

POLY *		poly,		/* IN/OUT: Polynomial to initialize */
int		degree		/* IN: Maximum degree */
)
{
COEFF *		coeff;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	coeff = NULL;
	if (degree >= 0) {
		coeff = CALIB_NEWA (degree + 1, COEFF);
	}

	poly -> degree	= -1;
	poly -> coeff	= coeff;
}

/*
 * Initialize the given destination polynomial from the given
 * source polynomial.
 */

	static
	void
poly_init_set (

POLY *		dst,		/* IN/OUT: Destination polynomial */
const POLY *	src		/* IN: Source polynomial */
)
{
int		deg;
COEFF *		coeff;

	CALIB_FATAL_ERROR_IF ((dst EQ NULL) OR (src EQ NULL));

	deg	= src -> degree;

	dst -> degree	= deg;

	if (deg < 0) {
		dst -> coeff = NULL;
	}
	else {
		coeff = CALIB_NEWA (deg + 1, COEFF);
		memcpy (coeff, src -> coeff, (deg + 1) * sizeof (COEFF));
		dst -> coeff = coeff;
	}
}

/*
 * Clear out the given polynomial.
 */

	static
	void
poly_clear (

POLY *		poly		/* IN: Polynomial to clear */
)
{
COEFF *		coeff;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	poly -> degree = -1;
	coeff = poly -> coeff;
	if (coeff NE NULL) {
		free (coeff);
		poly -> coeff = NULL;
	}
}

/*
 * Copy source polynomial into destination polynomial.  It is assumed
 * that the destination's coefficient buffer is sufficiently large to
 * hold all of the source polynomial's coefficients.
 */

	static
	void
poly_set (

POLY *		dst,	/* IN/OUT: Destination polynomial */
const POLY *	src	/* IN: Source polynomial */
)
{
int	deg;

	CALIB_FATAL_ERROR_IF ((dst EQ NULL) OR (src EQ NULL));

	if (dst NE src) {
		deg	= src -> degree;

		dst -> degree	= deg;

		memcpy (dst -> coeff,
			src -> coeff,
			(deg + 1) * sizeof (COEFF));
	}
}

/*
 * Return the multiplicative inverse of num modulo given prime p,
 * using the extended Euclidean algorithm.
 */

	static
	COEFF
inv (

COEFF		num,		/* IN: Number to take inverse of */
COEFF		p		/* IN: Prime modulus */
)
{
calib_bool	neg_flag;
COEFF		a1, a2, a3;
COEFF		b1, b2, b3;
COEFF		n, m, q, r;

	CALIB_FATAL_ERROR_IF (num EQ 0);
	CALIB_FATAL_ERROR_IF (p <= 1);
	CALIB_FATAL_ERROR_IF ((num % p) EQ 0);

	n = num;
	m = p;
	a1 = 1;
	b1 = 0;
	a2 = 0;
	b2 = 1;
	neg_flag = FALSE;
	while (n > 1) {
		q = m / n;
		r = m % n;

		a3 = a1 + q*a2;
		b3 = b1 + q*b2;
		a1 = a2;
		b1 = b2;
		a2 = a3;
		b2 = b3;
		m = n;
		n = r;
		neg_flag = NOT neg_flag;
	}
	if (n < 1) {
		/* Cannot find inverse! */
		CALIB_FATAL_ERROR;
	}
	if (neg_flag) {
		b2 = (p - b2);
	}
	return (b2);
}

/*
 * Print the given polynomial in Zp[x].
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
poly_print (

const POLY *	poly		/* IN: Polynomial to print */
)
{
int		i, N;
const COEFF *	coeff;

	N	= poly -> degree;
	coeff	= poly -> coeff;

	if (N < 0) {
		printf ("0\n");
		return;
	}
	for (i = N; i >= 0; i--) {
		if (coeff [i] EQ 0) continue;
		if (i < N) {
			printf (" + ");
		}
		if (i EQ 0) {
			printf ("%u", coeff [0]);
		}
		else {
			if (coeff [i] NE 1) {
				printf ("%u ", coeff [i]);
			}
			if (i EQ 1) {
				printf ("X");
			}
			else if (i NE 0) {
				printf ("X^%d", i);
			}
		}
	}
	printf ("\n");
}

#pragma GCC diagnostic pop
