/***********************************************************************

	$Id: cputime.c,v 1.1 2024/10/20 00:36:27 warme Exp $

	File:	cputime.c
	Rev:	a-3
	Date:	10/19/2024

	Copyright (c) 2004, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

************************************************************************

	This file contains routines for computing CPU time
	consumption for the program.

************************************************************************

	Modification Log:

	a-1:	02/24/2004	warme
		: Created.
	a-2:	08/07/2008	warme
		: Changed to use sysconf() instead of CLK_TCK.
	a-3:	10/19/2024	warme
		: Cleanup for initial release.

************************************************************************/

#define	_GNU_SOURCE	1

#include "calib/cputime.h"

#include "calib/fatal.h"
#include "calib/logic.h"

#ifdef UNIX_CPU_TIME
#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>
#include <unistd.h>
#endif

#include <time.h>


/*
 * Global Functions
 */

void			calib_convert_cpu_time (
					calib_cpu_time_t	time,
					char *			out_buf);
void			calib_convert_delta_cpu_time (
					char *			buf,
					calib_cpu_time_t *	time_in_out);
calib_cpu_time_t	calib_get_cpu_time (void);
calib_cpu_time_t	calib_get_delta_cpu_time (
					calib_cpu_time_t *	time_in_out);

/*
 * Local Types
 */

struct environment {
	clock_t		clocks_per_sec;	/* Tick rate */
	unsigned long	Q;		/* Quotient */
	unsigned long	R;		/* Remainder */
	int		method;		/* Time keeping method */
	int		initialized;	/* TRUE if we have initialized */
};

/*
 * Local Functions
 */

static void	init_cpu_time (void);

static struct environment	env;

/*
 * This is the units WE use for measuring CPU time.
 */

#define	TICKS_PER_SEC	100

/*
 * Define API_CLOCKS_PER_SECOND to be the tick rate presented to the
 * program by the chosen API.
 */

#ifdef UNIX_CPU_TIME
 #define API_CLOCKS_PER_SECOND	(sysconf(_SC_CLK_TCK)) \
						/* The Posix value */
#else
 #define API_CLOCKS_PER_SECOND	CLOCKS_PER_SEC	/* The ANSI C value */
#endif

/*
 * This routine gets the current amount of CPU time that the program has
 * used.  It is returned in a format we define that is independent of the
 * operating system (i.e. hundreth's of a second).
 *
 * Because the times we are measuring can be quite long (several days),
 * we can run into overflow problems if the operating system gives us
 * a CPU timer with too high a resolution (i.e. microseconds).  This
 * code can be compiled to use either the standard interfaces defined
 * by ANSI C, or the native UNIX interface.  Choose the one that has
 * a resolution closest to our own TICKS_PER_SEC units.
 *
 * The basic conversion formula is:
 *
 *			   sys_ticks * our_ticks_per_sec    1
 *	our_ticks = floor (----------------------------- + ---)
 *				  sys_ticks_per_sec	    2
 *
 * which can be re-written
 *
 *			   2 * sys_ticks * our_tps + sys_tps
 *	our_ticks = floor (---------------------------------)
 *				     2 * sys_tps
 *
 * We compute this using one of several methods, depending upon the
 * relationship between sys_ticks_per_sec and our_ticks_per_sec.
 * The idea of it all is to minimize the chance of arithmetic overflow.
 */

	void
init_cpu_time (void)

{
int		method;
clock_t		clocks_per_sec;
unsigned long	Q, R;

	if (env.initialized) return;

	/* On some systems this is a system call.  Do only once! */
	clocks_per_sec = API_CLOCKS_PER_SECOND;

	/* Now pick the conversion method and parameters. */
	if (TICKS_PER_SEC >= clocks_per_sec) {
		Q = TICKS_PER_SEC / clocks_per_sec;
		R = TICKS_PER_SEC % clocks_per_sec;
		method = (R EQ 0) ? 0 : 1;
	}
	else {
		R = clocks_per_sec % TICKS_PER_SEC;
		if (R EQ 0) {
			Q = clocks_per_sec / TICKS_PER_SEC;
			method = 2;
		}
		else {
			method = 3;
		}
	}

	env.clocks_per_sec = clocks_per_sec;
	env.Q = Q;
	env.R = R;
	env.method = method;
	env.initialized = TRUE;
}

/*
 * Return the total CPU time expended (system and user) by this
 * process (and all of its waited-for terminated children).
 */

	calib_cpu_time_t
calib_get_cpu_time (void)

{
clock_t			total;
clock_t			clocks_per_sec;
unsigned long		seconds;
unsigned long		ticks;
calib_cpu_time_t	cpu_time;

	init_cpu_time ();

#ifdef UNIX_CPU_TIME
	{ struct tms	t;

		times (&t);

		total	= t.tms_utime  + t.tms_stime
			+ t.tms_cutime + t.tms_cstime;
	}
#else
	/* Using the ANSI C defined interface... */
	total = clock ();
#endif

	clocks_per_sec = env.clocks_per_sec;

	seconds	= total / clocks_per_sec;
	ticks	= total % clocks_per_sec;

	/* Convert the fractions of a second ticks into .01 second units. */
	switch (env.method) {
	case 0:
		ticks *= env.Q;
		break;

	case 1:
		ticks = ticks * env.Q
			+ (2 * ticks * env.R + clocks_per_sec) /
			  (2 * clocks_per_sec);
		break;

	case 2:
		ticks = (2 * ticks + env.Q) / (2 * env.Q);
		break;

	case 3:
		ticks = ((2 * TICKS_PER_SEC) * ticks + clocks_per_sec)
			/ (2 * clocks_per_sec);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	cpu_time = seconds * TICKS_PER_SEC + ticks;

	return (cpu_time);
}

/*
 * This routine will convert the given CPU time into a printable
 * null-terminated ASCII string.  The answer contains two decimal places.
 */

	void
calib_convert_cpu_time (

calib_cpu_time_t	time,		/* IN: CPU time to convert */
char *			out_buf		/* IN/OUT: time as text string */
)
{
calib_cpu_time_t	secs;
calib_cpu_time_t	ticks;
char *			p;
char			buf [20];

#define ZERO	((calib_cpu_time_t) 0)
#define TEN	((calib_cpu_time_t) 10)

	secs	= time / TICKS_PER_SEC;
	ticks	= time % TICKS_PER_SEC;

	p = &buf [0];
	if (secs <= ZERO) {
		*p++ = '0';
	}
	else {
		while (secs > ZERO) {
			*p++ = (secs % TEN) + '0';
			secs /= TEN;
		}
	}
	while (p > &buf [0]) {
		*out_buf++ = *--p;
	}
	*out_buf++ = '.';
	ticks *= TEN;
	*out_buf++ = (ticks / TICKS_PER_SEC) + '0';
	ticks %= TICKS_PER_SEC;
	ticks *= TEN;
	*out_buf++ = (ticks / TICKS_PER_SEC) + '0';
	*out_buf++ = '\0';

#undef ZERO
#undef TEN
}

/*
 * Compute the CPU time used since the last time we called this routine.
 * Returns CPU expended since given time Tn (which is updated to "now").
 */

	calib_cpu_time_t
calib_get_delta_cpu_time (

calib_cpu_time_t *	Tn	/* IN/OUT: Prev CPU in, current CPU out */
)
{
calib_cpu_time_t	now, delta;

	now = calib_get_cpu_time ();

	delta = now - *Tn;

	*Tn = now;

	return (delta);
}

/*
 * Compute and format the CPU time used since the last time we called
 * this routine.
 */

	void
calib_convert_delta_cpu_time (

char *			buf,	/* OUT: ASCII time string, CPU seconds */
calib_cpu_time_t *	Tn	/* IN/OUT: Prev CPU in, current CPU out */
)
{
calib_cpu_time_t	delta;

	delta = calib_get_delta_cpu_time (Tn);
	calib_convert_cpu_time (delta, buf);
}
