/**********************************************************************

	$Id: genrep.c,v 1.1 2024/10/20 00:36:27 warme Exp $

	File:	genrep.c
	Rev:	a-3
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	General representation for expressions.

***********************************************************************

	Modification Log:

	a-1:	04/23/2012	warme
		: Created.
	a-2:	10/10/2019	warme
		: Split pretty print into oldprint.c.
	a-3:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/genrep.h"

#include "genrep_int.h"

#include "calib/new.h"
#include "calib/fatal.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include <gmp.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


/*
 * Global Functions.
 */

struct calib_genrep *
		calib_genrep_add2 (const struct calib_genrep *	op1,
				   const struct calib_genrep *	op2);
void		calib_genrep_clear_varlist (
				struct calib_genrep_varlist *	list);
struct calib_genrep *
		calib_genrep_div2 (const struct calib_genrep *	op1,
				   const struct calib_genrep *	op2);
struct calib_genrep *
		calib_genrep_dup (const struct calib_genrep * op);
struct calib_genrep_list *
		calib_genrep_dup_list (const struct calib_genrep_list * list);
struct calib_genrep *
		calib_genrep_fread (FILE * fp, int * status);
void		calib_genrep_free (struct calib_genrep *	op);
void		calib_genrep_free_list (struct calib_genrep_list * list);
void		calib_genrep_get_varlist (
				struct calib_genrep_varlist *	list,
				const struct calib_genrep *	op);
struct calib_genrep *
		calib_genrep_ipow (const struct calib_genrep * op1, int op2);
struct calib_genrep *
		calib_genrep_mul2 (const struct calib_genrep *	op1,
				   const struct calib_genrep *	op2);
struct calib_genrep *
		calib_genrep_neg (struct calib_genrep * op);
struct calib_genrep_list *
		calib_genrep_new_list (struct calib_genrep *		op,
				       struct calib_genrep_list *	next);
struct calib_genrep *
		calib_genrep_poly_term (struct calib_genrep *	coeff,
					const char *	var,
					int		pow);
struct calib_genrep *
		calib_genrep_q (mpq_srcptr	op);
struct calib_genrep *
		calib_genrep_read (int * status);
struct calib_genrep *
		calib_genrep_si (calib_si_t op);
struct calib_genrep *
		calib_genrep_sub2 (struct calib_genrep *	op1,
				   struct calib_genrep *	op2);
struct calib_genrep *
		calib_genrep_var (const char * var);
struct calib_genrep *
		calib_genrep_z (mpz_srcptr	op);

struct calib_genrep *
		_calib_genrep_copy_simp (const struct calib_genrep * op);
struct calib_genrep *
		_calib_genrep_finish_prod (struct _genrep_nary * root);
struct calib_genrep *
		_calib_genrep_finish_sum (struct _genrep_nary * root);
void		_calib_genrep_init_prod (struct _genrep_nary * root);
void		_calib_genrep_init_sum (struct _genrep_nary * root);
void		_calib_genrep_insert_prod (
				struct _genrep_nary *		root,
				const struct calib_genrep *	op,
				_genrep_dup_func		dup_func);
void		_calib_genrep_insert_sum (
				struct _genrep_nary *		root,
				const struct calib_genrep *	op,
				_genrep_dup_func		dup_func);

/*
 * Local Types
 */

typedef struct calib_genrep * (*grdup_func) (const struct calib_genrep * op);

struct vlist {
	const char *	vname;		/* Variable name */
	struct vlist *	next;		/* Next in list */
};

struct token {
	int		type;		/* Type of token */
	union {
		char	ident [128];	/* Identifier name */
		mpz_t	zval;		/* Integer value */
	} u;
};

#define TOK_EOF		0		/* End-of-file */
#define TOK_ICONST	1		/* Integer constant */
#define TOK_IDENT	2		/* Identifier name */
#define TOK_PLUS	3		/* Plus sign */
#define TOK_MINUS	4		/* Minus sign */
#define TOK_STAR	5		/* Asterisk */
#define TOK_SLASH	6		/* Forward slash */
#define TOK_POW		7		/* '^' or '**' */
#define TOK_LPAREN	8		/* Left parenthesis */
#define TOK_RPAREN	9		/* Right parenthesis */
#define TOK_SEMI	10		/* Semicolon */
#define TOK_COMMA	11		/* Comma */
#define TOK_ERROR	12		/* Invalid token */

/*
 * Local Functions.
 */

static struct calib_genrep *
		copy_func (const struct calib_genrep * op);
static struct calib_genrep *
		copy_simp_ipow (const struct calib_genrep *	op1,
				int				op2,
				grdup_func			dup_func);
static void	get_qval (mpq_ptr qval, const struct calib_genrep * op);
static void	_mpq_pow_si (mpq_ptr	result,
			     mpq_srcptr	op1,
			     long int	op2);
static struct calib_genrep *
		new_genrep (void);
static void	next_token (FILE * fp, struct token * token);
static int	nth_char_id (int c);
static struct calib_genrep *
		read_expr (FILE * fp, struct token * token);
static struct calib_genrep *
		read_factor (FILE * fp, struct token * token);
static struct calib_genrep *
		read_term (FILE * fp, struct token * token);
static struct calib_genrep_list *
		reverse_list (struct calib_genrep_list * lp);
static void	strip_c_comment (FILE * fp, int echo);
static void	strip_cpp_comment (FILE * fp, int echo);
static calib_bool
		strip_comment (FILE * fp, int * echo_flag);
static void	varlist_recurse (struct vlist **	root_ptr,
				 const struct calib_genrep *	op);

/*
 * Create a general expression for the given integer value.
 */

	struct calib_genrep *
calib_genrep_si (

calib_si_t	op		/* IN: integer operand */
)
{
struct calib_genrep *	gp;

	gp = new_genrep ();
	gp -> op	= CALIB_GENREP_OP_SI;
	gp -> u.siop	= op;

	return (gp);
}

/*
 * Create a general expression for the given arbirary precision integer value.
 */

	struct calib_genrep *
calib_genrep_z (

mpz_srcptr	op		/* IN: integer operand */
)
{
struct calib_genrep *	gp;

	gp = new_genrep ();
	gp -> op	= CALIB_GENREP_OP_Z;
	gp -> u.zop	= calib_new_Z_vector (1);
	mpz_set (gp -> u.zop, op);

	return (gp);
}

/*
 * Create a general expression for the given arbirary precision rational value.
 */

	struct calib_genrep *
calib_genrep_q (

mpq_srcptr	op		/* IN: integer operand */
)
{
struct calib_genrep *	gp;

	if (mpz_cmp_ui (mpq_denref (op), 1) EQ 0) {
		/* Rational value is integral.  Use integer form. */
		gp = calib_genrep_z (mpq_numref (op));
	}
	else {
		gp = new_genrep ();
		gp -> op	= CALIB_GENREP_OP_Q;
		gp -> u.qop	= calib_new_Q_vector (1);
		mpq_set (gp -> u.qop, op);
	}

	return (gp);
}

/*
 * Create a general expression for the given variable name.
 */

	struct calib_genrep *
calib_genrep_var (

const char *	var		/* IN: name of variable */
)
{
size_t			len;
char *			buf;
struct calib_genrep *	gp;

	len = strlen (var);
	CALIB_FATAL_ERROR_IF (len <= 0);

	buf = CALIB_NEWA (len + 1, char);
	memcpy (buf, var, len + 1);

	gp = new_genrep ();
	gp -> op	= CALIB_GENREP_OP_VAR;
	gp -> u.var	= buf;

	return (gp);
}

/*
 * Create a general expression that represents the sum of the two given
 * expressions.
 */

	struct calib_genrep *
calib_genrep_add2 (

const struct calib_genrep *	op1,	/* IN: first operand */
const struct calib_genrep *	op2	/* IN: second operand */
)
{
struct calib_genrep *	gp1;
struct _genrep_nary	root;

	_calib_genrep_init_sum (&root);

	_calib_genrep_insert_sum (&root, op1, calib_genrep_dup);
	_calib_genrep_insert_sum (&root, op2, calib_genrep_dup);

	gp1 = _calib_genrep_finish_sum (&root);

	return (gp1);
}

/*
 * Initialize an empty sum.
 */

	void
_calib_genrep_init_sum (

struct _genrep_nary *	root	/* IN/OUT: list of terms to init */
)
{
	/* Initial sum is 0. */
	mpq_init (root -> qval);
	root -> list = NULL;
	root -> hookp = &(root -> list);
}

/*
 * Subroutine to insert a sub-expression into a list of terms.
 * The constant term is kept separate.
 */

	void
_calib_genrep_insert_sum (

struct _genrep_nary *		root,	 /* IN/OUT: list of terms to add into */
const struct calib_genrep *	op,	 /* IN: sub-expression to insert/add */
grdup_func			dup_func /* IN: func to recursively copy subexprs */
)
{
struct calib_genrep_list *	lp1;
struct calib_genrep *		op2;
mpq_t				qval;

	if (CALIB_GENREP_NUMBERP (op)) {
		get_qval (qval, op);
		mpq_add (root -> qval, root -> qval, qval);
		mpq_clear (qval);
		return;
	}

	if (op -> op EQ CALIB_GENREP_OP_ADD) {
		for (lp1 = op -> u.list; lp1 NE NULL; lp1 = lp1 -> next) {
			_calib_genrep_insert_sum (root,
						  lp1 -> operand,
						  dup_func);
		}
		return;
	}

	/* A general term.  Simply append it to the list.		*/
	/* We don't do lots of simplification here because we expect	*/
	/* the various canonical representations to give us clean	*/
	/* sums, products, etc.						*/

	op2 = (*dup_func) (op);
	lp1 = calib_genrep_new_list (op2, NULL);

	*(root -> hookp) = lp1;
	root -> hookp = &(lp1 -> next);
}

/*
 * Finish up the given collection of summed terms.
 */

	struct calib_genrep *
_calib_genrep_finish_sum (

struct _genrep_nary *	root	/* IN: collection of terms to finish */
)
{
struct calib_genrep *		gp1;
struct calib_genrep_list *	lp1;

	if (mpz_sgn (mpq_numref (root -> qval)) NE 0) {
		/* Constant term goes last.  Use hookp to append it to	*/
		/* the end.						*/
		gp1 = calib_genrep_q (root -> qval);
		lp1 = calib_genrep_new_list (gp1, NULL);
		*(root -> hookp) = lp1;
		root -> hookp = &(lp1 -> next);
	}

	lp1 = root -> list;

	if (lp1 EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (lp1 -> next EQ NULL) {
		/* One term. */
		gp1 = lp1 -> operand;
		free (lp1);
	}
	else {
		gp1 = new_genrep ();
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= lp1;
	}

	mpq_clear (root -> qval);
	root -> list	= NULL;
	root -> hookp	= NULL;

	return (gp1);
}

/*
 * Create a general expression that represents the product of the two given
 * expressions.
 */

	struct calib_genrep *
calib_genrep_mul2 (

const struct calib_genrep *	op1,	/* IN: first operand */
const struct calib_genrep *	op2	/* IN: second operand */
)
{
struct calib_genrep *	gp1;
struct _genrep_nary	root;

	_calib_genrep_init_prod (&root);

	_calib_genrep_insert_prod (&root, op1, calib_genrep_dup);
	_calib_genrep_insert_prod (&root, op2, calib_genrep_dup);

	gp1 = _calib_genrep_finish_prod (&root);

	return (gp1);
}

/*
 * Initialize an empty product.
 */

	void
_calib_genrep_init_prod (

struct _genrep_nary *	root	/* IN/OUT: list of factors to init */
)
{
	/* Initial product is 1. */
	mpq_init (root -> qval);
	mpz_set_ui (mpq_numref (root -> qval), 1);
	root -> list = NULL;
	root -> hookp = &(root -> list);
}

/*
 * Subroutine to insert a sub-expression into a list of factors.
 * The constant factor is kept separate.
 */

	void
_calib_genrep_insert_prod (

struct _genrep_nary *		root,	 /* IN/OUT: list of factors to add into */
const struct calib_genrep *	op,	 /* IN: sub-expression to insert/add */
grdup_func			dup_func /* IN: func to recursively copy subexprs */
)
{
struct calib_genrep_list *	lp1;
struct calib_genrep *		op2;
mpq_t				qval;

	if (CALIB_GENREP_NUMBERP (op)) {
		get_qval (qval, op);
		mpq_mul (root -> qval, root -> qval, qval);
		mpq_clear (qval);
		return;
	}

	if (op -> op EQ CALIB_GENREP_OP_MUL) {
		for (lp1 = op -> u.list; lp1 NE NULL; lp1 = lp1 -> next) {
			_calib_genrep_insert_prod (root, lp1 -> operand, dup_func);
		}
		return;
	}

	/* A general factor.  Simply append it to the list.		*/
	/* We don't do lots of simplification here because we expect	*/
	/* the various canonical representations to give us clean	*/
	/* sums, products, etc.						*/

	op2 = (*dup_func) (op);
	lp1 = calib_genrep_new_list (op2, NULL);

	*(root -> hookp) = lp1;
	root -> hookp = &(lp1 -> next);
}

/*
 * Finish up the given collection of product factors.
 */

	struct calib_genrep *
_calib_genrep_finish_prod (

struct _genrep_nary *	root		/* IN: collection of factors to finish */
)
{
struct calib_genrep *		gp1;
struct calib_genrep_list *	lp1;

	if (mpq_cmp_ui (root -> qval, 0, 1) EQ 0) {
		/* Constant factor is zero! */
		calib_genrep_free_list (root -> list);
		gp1 = calib_genrep_si (0);
	}
	else {
		lp1 = root -> list;
		if (mpq_cmp_ui (root -> qval, 1, 1) NE 0) {
			/* Constant factor goes first. */
			gp1 = calib_genrep_q (root -> qval);
			lp1 = calib_genrep_new_list (gp1, lp1);
		}
		if (lp1 EQ NULL) {
			gp1 = calib_genrep_si (1);
		}
		else if (lp1 -> next EQ NULL) {
			/* One factor. */
			gp1 = lp1 -> operand;
			free (lp1);
		}
		else {
			/* Two or more factors. */
			gp1 = new_genrep ();
			gp1 -> op	= CALIB_GENREP_OP_MUL;
			gp1 -> u.list	= lp1;
		}
	}

	mpq_clear (root -> qval);
	root -> list	= NULL;
	root -> hookp	= NULL;

	return (gp1);
}

/*
 * Create a general expression that represents op1**op2.
 */

	struct calib_genrep *
calib_genrep_ipow (

const struct calib_genrep *	op1,	/* IN: first operand */
int				op2	/* IN: second operand */
)
{
struct calib_genrep *	gp1;

	gp1 = copy_simp_ipow (op1, op2, calib_genrep_dup);

	return (gp1);
}

/*
 * Take the given rational to the given integer power.
 */

	static
	void
_mpq_pow_si (

mpq_ptr		result,		/* IN/OUT: result */
mpq_srcptr	op1,		/* IN: rational operand */
long int	op2		/* IN: integer power */
)
{
unsigned long int	pow;
mpz_t			num, den;

	mpz_init (num);
	mpz_init (den);

	if (op2 >= 0) {
		pow = op2;
		mpz_pow_ui (num, mpq_numref (op1), pow);
		mpz_pow_ui (den, mpq_denref (op1), pow);
	}
	else {
		pow = - op2;
		mpz_pow_ui (den, mpq_numref (op1), pow);
		mpz_pow_ui (num, mpq_denref (op1), pow);
	}
	CALIB_FATAL_ERROR_IF (mpz_sgn (den) EQ 0);

	mpz_set (mpq_numref (result), num);
	mpz_set (mpq_denref (result), den);

	mpz_clear (den);
	mpz_clear (num);
}

/*
 * Create a general expression for op1 - op2.
 */

	struct calib_genrep *
calib_genrep_sub2 (

struct calib_genrep *	op1,		/* IN: operand 1 */
struct calib_genrep *	op2		/* IN: operand 2 */
)
{
struct calib_genrep *	gp1;
struct calib_genrep *	gp2;

	gp2 = calib_genrep_neg (op2);
	gp1 = calib_genrep_add2 (op1, gp2);

	calib_genrep_free (gp2);

	return (gp1);
}

/*
 * Create a general expression for -op.
 */

	struct calib_genrep *
calib_genrep_neg (

struct calib_genrep *	op		/* IN: operand to negate */
)
{
struct calib_genrep *	gp1;
struct calib_genrep *	gp2;

	gp1 = calib_genrep_si (-1);
	gp2 = calib_genrep_mul2 (gp1, op);
	calib_genrep_free (gp1);

	return (gp2);
}

/*
 * Create a general expression for op1 / op2.
 */

	struct calib_genrep *
calib_genrep_div2 (

const struct calib_genrep *	op1,	/* IN: operand 1 */
const struct calib_genrep *	op2	/* IN: operand 2 */
)
{
struct calib_genrep *	gp1;
struct calib_genrep *	gp2;

	gp2 = calib_genrep_ipow (op2, -1);
	gp1 = calib_genrep_mul2 (op1, gp2);

	calib_genrep_free (gp2);

	return (gp1);
}

/*
 * Create the general representation for a single polynomial term.
 * The coefficient value is given as a genrep that will either:
 *	1. Be freed by this routine, or
 *	2. Become a part of the resulting expression.
 */

	struct calib_genrep *
calib_genrep_poly_term (

struct calib_genrep *	coeff,		/* IN: coefficient value */
const char *		var,		/* IN: polynomial variable */
int			pow		/* IN: power of variable */
)
{
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;
struct calib_genrep_list *	lp1;

	CALIB_FATAL_ERROR_IF ((coeff EQ NULL) OR (var EQ NULL) OR (pow < 0));

	if (CALIB_GENREP_ZEROP (coeff) OR (pow EQ 0)) {
		return (coeff);
	}

	vgp = calib_genrep_var (var);
	if (pow > 1) {
		gp1 = new_genrep ();
		gp1 -> op		= CALIB_GENREP_OP_IPOW;
		gp1 -> u.ipow.base	= vgp;
		gp1 -> u.ipow.power	= pow;
		vgp = gp1;
	}

	if (CALIB_GENREP_ONEP (coeff)) {
		calib_genrep_free (coeff);
		gp1 = vgp;
	}
	else {
		lp1 = calib_genrep_new_list (coeff,
				       calib_genrep_new_list (vgp, NULL));
		gp1 = new_genrep ();
		gp1 -> op	= CALIB_GENREP_OP_MUL;
		gp1 -> u.list	= lp1;
	}

	return (gp1);
}

/*
 * Create general expression for f(x), where f is a builtin function (number).
 */

	struct calib_genrep *
calib_genrep_func_si_1 (

calib_si_t		fidx,	/* IN: Builtin function number */
struct calib_genrep *	op	/* IN: Argument operand */
)
{
struct calib_genrep *		gp;
struct calib_genrep *		gp1;
struct calib_genrep_list *	lp;

	CALIB_FATAL_ERROR_IF ((fidx < 0) OR (op EQ NULL));

	gp1 = calib_genrep_si (fidx);
	lp = calib_genrep_new_list (op, NULL);

	gp = calib_genrep_func (gp1, lp);

	return (gp);
}

/*
 * Create general expression for f(x,y), where f is a builtin function
 * (number).
 */

	struct calib_genrep *
calib_genrep_func_si_2 (

calib_si_t		fidx,	/* IN: Builtin function number */
struct calib_genrep *	arg1,	/* IN: First argument */
struct calib_genrep *	arg2	/* IN: Second argument */
)
{
struct calib_genrep *		gp;
struct calib_genrep *		gp1;
struct calib_genrep_list *	lp1;
struct calib_genrep_list *	lp2;

	CALIB_FATAL_ERROR_IF ((fidx < 0) OR (arg1 EQ NULL) OR (arg2 EQ NULL));

	gp1 = calib_genrep_si (fidx);
	lp1 = calib_genrep_new_list (arg2, NULL);
	lp2 = calib_genrep_new_list (arg1, lp1);

	gp = calib_genrep_func (gp1, lp2);

	return (gp);
}

/*
 * Create general expression for f(x,y,...), where f is a non-standard
 * function (given by name / textual string).
 */

	struct calib_genrep *
calib_genrep_func_str (

const char *			fname,	/* IN: Function name */
struct calib_genrep_list *	args	/* IN: Argument list */
)
{
struct calib_genrep *		gp;
struct calib_genrep *		gp1;

	CALIB_FATAL_ERROR_IF (fname EQ NULL);

	gp1 = calib_genrep_var (fname);

	gp = calib_genrep_func (gp1, args);

	return (gp);
}

/*
 * Create general expression for f(x,y,...) for given function and
 * argument list.
 */

	struct calib_genrep *
calib_genrep_func (

struct calib_genrep *		f,	/* IN: Function */
struct calib_genrep_list *	args	/* IN: Argument list */
)
{
int				op;
struct calib_genrep *		gp;

	CALIB_FATAL_ERROR_IF (f EQ NULL);

	op = f -> op;

	/* Functions must be either "builtin" (a calib_si_t), or a	*/
	/* general function name (variable).				*/
	CALIB_FATAL_ERROR_IF ((op NE CALIB_GENREP_OP_SI) AND
			      (op NE CALIB_GENREP_OP_VAR));

	gp = new_genrep ();

	gp -> op		= CALIB_GENREP_OP_FUNC;
	gp -> u.func.func	= f;
	gp -> u.func.args	= args;

	return (gp);
}

/*
 * Define functions that generate genreps for our
  * one-argument builtin functions.
 */

#define	DEF1(func, FUNC)						\
	struct calib_genrep *						\
calib_genrep_ ## func (							\
									\
struct calib_genrep *		x					\
)									\
{									\
struct calib_genrep *		gp;					\
									\
	gp = calib_genrep_func_si_1 (CALIB_GENREP_FUNC_ ## FUNC, x);	\
	return (gp);							\
}

CALIB_GENREP_BUILTIN_FUNC1_LIST (DEF1)

#undef DEF1

/*
 * Define functions that generate genreps for our
  * two-argument builtin functions.
 */

#define	DEF2(func, FUNC)						\
	struct calib_genrep *						\
calib_genrep_ ## func (							\
									\
struct calib_genrep *		x1,					\
struct calib_genrep *		x2					\
)									\
{									\
struct calib_genrep *		gp;					\
									\
	gp = calib_genrep_func_si_2 (CALIB_GENREP_FUNC_ ## FUNC, x1, x2); \
	return (gp);							\
}

CALIB_GENREP_BUILTIN_FUNC2_LIST (DEF2)

#undef DEF2

/*
 * Recursively copy the given general expression.
 */

	struct calib_genrep *
calib_genrep_dup (

const struct calib_genrep *	op	/* IN: operand to duplicate */
)
{
struct calib_genrep *	gp1;

	CALIB_FATAL_ERROR_IF (op EQ NULL);
	switch (op -> op) {
	case CALIB_GENREP_OP_ADD:
	case CALIB_GENREP_OP_MUL:
		gp1 = new_genrep ();
		gp1 -> op	= op -> op;
		gp1 -> u.list	= calib_genrep_dup_list (op -> u.list);
		break;

	case CALIB_GENREP_OP_IPOW:
		gp1 = new_genrep ();
		gp1 -> op		= CALIB_GENREP_OP_IPOW;
		gp1 -> u.ipow.base	= calib_genrep_dup (op -> u.ipow.base);
		gp1 -> u.ipow.power	= op -> u.ipow.power;
		break;

	case CALIB_GENREP_OP_VAR:
		gp1 = calib_genrep_var (op -> u.var);
		break;

	case CALIB_GENREP_OP_SI:
		gp1 = calib_genrep_si (op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		gp1 = calib_genrep_z (op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		gp1 = calib_genrep_q (op -> u.qop);
		break;

	case CALIB_GENREP_OP_FUNC:
		gp1 = new_genrep ();
		gp1 -> op	= op -> op;
		gp1 -> u.func.func	= calib_genrep_dup (op -> u.func.func);
		gp1 -> u.func.args	= calib_genrep_dup_list (op -> u.func.args);
		break;

	default:
		CALIB_FATAL_ERROR;
		gp1 = calib_genrep_si (0);
		break;
	}

	return (gp1);
}

/*
 * Recursively copy the given general expression list.
 */

	struct calib_genrep_list *
calib_genrep_dup_list (

const struct calib_genrep_list *	list	/* IN: list to duplicate */
)
{
struct calib_genrep_list *		root;
struct calib_genrep_list **		hookp;
const struct calib_genrep_list *	lp1;
struct calib_genrep_list *		lp2;

	root = NULL;
	hookp = &root;
	for (lp1 = list; lp1 NE NULL; lp1 = lp1 -> next) {
		lp2 = calib_genrep_new_list (calib_genrep_dup (lp1 -> operand), NULL);
		*hookp = lp2;
		hookp = &(lp2 -> next);
	}

	return (root);
}

/*
 * Recursively free the given general expression.
 */

	void
calib_genrep_free (

struct calib_genrep *	op		/* IN: genrep to recursively free */
)
{
	if (op EQ NULL) return;

	switch (op -> op) {
	case CALIB_GENREP_OP_ADD:
	case CALIB_GENREP_OP_MUL:
		calib_genrep_free_list (op -> u.list);
		break;

	case CALIB_GENREP_OP_IPOW:
		calib_genrep_free (op -> u.ipow.base);
		break;

	case CALIB_GENREP_OP_VAR:
		free (op -> u.var);
		break;

	case CALIB_GENREP_OP_SI:
		break;

	case CALIB_GENREP_OP_Z:
		calib_free_Z_vector (op -> u.zop, 1);
		break;

	case CALIB_GENREP_OP_Q:
		calib_free_Q_vector (op -> u.qop, 1);
		break;

	case CALIB_GENREP_OP_FUNC:
		calib_genrep_free (op -> u.func.func);
		calib_genrep_free_list (op -> u.func.args);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	memset (op, 0, sizeof (*op));
	free (op);
}

/*
 * Recursively free the given general expression list.
 */

	void
calib_genrep_free_list (

struct calib_genrep_list *	list	/* IN: list to recursively free */
)
{
struct calib_genrep_list *	tmp;

	while (list NE NULL) {
		tmp = list -> next;
		calib_genrep_free (list -> operand);
		memset (list, 0, sizeof (*list));
		free (list);
		list = tmp;
	}
}

/*
 * Compute the list of all variables occurring in the given general expression.
 * Each distinct variable occurs only once in the list, which is sorted
 * in ASCII order.
 */

	void
calib_genrep_get_varlist (

struct calib_genrep_varlist *	list,	/* IN/OUT: list of variables */
const struct calib_genrep *	op	/* IN: expression to scan */
)
{
int			i, nvars;
size_t			nbytes, len;
struct vlist *		root;
struct vlist *		vp;
const char **		vlist;
char *			cbuf;

	root = NULL;

	varlist_recurse (&root, op);

	nvars = 0;
	nbytes = 0;
	for (vp = root; vp NE NULL; vp = vp -> next) {
		nbytes += (strlen (vp -> vname) + 1);
		++nvars;
	}

	cbuf = NULL;
	if (nvars > 0) {
		cbuf	= CALIB_NEWA (nbytes, char);
	}
	vlist	= CALIB_NEWA (nvars + 1, const char *);

	i = 0;
	for (vp = root; vp NE NULL; vp = vp -> next) {
		len = strlen (vp -> vname);
		memcpy (cbuf, vp -> vname, len + 1);
		vlist [i++] = cbuf;
		cbuf += (len + 1);
	}
	vlist [i] = NULL;

	while (root NE NULL) {
		vp = root -> next;
		free (root);
		root = vp;
	}

	list -> nvars = nvars;
	list -> vlist = vlist;
}

/*
 * Clear out the given varlist.
 */

	void
calib_genrep_clear_varlist (

struct calib_genrep_varlist *	list	/* IN: varlist to free */
)
{
char **		sarray;

	sarray = (char **) list -> vlist;
	if (sarray NE NULL) {
		if (sarray [0] NE NULL) {
			free (sarray [0]);
		}
		free (sarray);
	}

	list -> nvars	= 0;
	list -> vlist	= NULL;
}

/*
 * Recursive portion of varlist extraction.
 */

	static
	void
varlist_recurse (

struct vlist **			root_ptr,	/* IN/OUT: list of var names */
const struct calib_genrep *	op		/* IN: expression to scan */
)
{
int				cmp;
size_t				len1, len2, len3;
const char *			var;
struct vlist **			hookp;
struct vlist *			vp;
struct vlist *			vp2;
struct calib_genrep_list *	lp;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	switch (op -> op) {
	case CALIB_GENREP_OP_ADD:
	case CALIB_GENREP_OP_MUL:
		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			varlist_recurse (root_ptr, lp -> operand);
		}
		break;

	case CALIB_GENREP_OP_FUNC:
		for (lp = op -> u.func.args; lp NE NULL; lp = lp -> next) {
			varlist_recurse (root_ptr, lp -> operand);
		}
		break;

	case CALIB_GENREP_OP_IPOW:
		varlist_recurse (root_ptr, op -> u.ipow.base);
		break;

	case CALIB_GENREP_OP_VAR:
		var = op -> u.var;
		len1 = strlen (var);
		for (hookp = root_ptr; ; hookp = &(vp -> next)) {
			vp = *hookp;
			if (vp EQ NULL) {
				/* Insert new entry here. */
				vp2 = CALIB_NEW (struct vlist);
				vp2 -> vname	= var;
				vp2 -> next	= vp;
				*hookp = vp2;
				return;
			}
			len2 = strlen (vp -> vname);
			len3 = (len1 < len2) ? len1 : len2;
			cmp = memcmp (var, vp -> vname, len3);
			if ((len1 EQ len2) AND (cmp EQ 0)) {
				/* Variable already present in root list. */
				return;
			}
			if ((cmp < 0) OR ((cmp EQ 0) AND (len1 < len2))) {
				/* Insert new entry here. */
				vp2 = CALIB_NEW (struct vlist);
				vp2 -> vname	= var;
				vp2 -> next	= vp;
				*hookp = vp2;
				return;
			}
		}
		break;

	case CALIB_GENREP_OP_SI:
	case CALIB_GENREP_OP_Z:
	case CALIB_GENREP_OP_Q:
		break;

	default:
		CALIB_FATAL_ERROR;
	}
}

/*
 * Read a general expression (terminated by a semicolon) from stdin.
 * Valid result combinations:
 *
 *	return value		status		Meaning
 *	------------		------		-------
 *	    NULL		  0		End of file.
 *	    NULL		  1		Syntax error in expression.
 *	non-NULL		  0		Expression read successfully.
 */

	struct calib_genrep *
calib_genrep_read (

int *		status		/* IN/OUT: error indication */
)
{
struct calib_genrep *	gp1;

	gp1 = calib_genrep_fread (stdin, status);

	return (gp1);
}

/*
 * Read a general expression (terminated by a semicolon) from the
 * given input stream.  Valid result combinations:
 *
 *	return value		status		Meaning
 *	------------		------		-------
 *	  NULL			  0		End of file.
 *	  NULL			  1		Syntax error in expression.
 *	non-NULL		  0		Expression read successfully.
 */

	struct calib_genrep *
calib_genrep_fread (

FILE *		fp,		/* IN: stream to read expression from */
int *		status		/* IN/OUT: error indication */
)
{
struct calib_genrep *	gp1;
struct token		token;

	/* Assume error until successful parse is obtained. */
	*status = 1;

	/* Read first token. */
	next_token (fp, &token);
	if (token.type EQ TOK_EOF) {
		*status = 0;
		return (NULL);
	}

	gp1 = read_expr (fp, &token);
	if ((gp1 NE NULL) AND (token.type EQ TOK_SEMI)) {
		/* We have successfully read an expression. */
		*status = 0;
		return (gp1);
	}
	else {
		calib_genrep_free (gp1);

		if (isatty (fileno (fp))) {
			/* Discard pending input. */
			__fpurge (fp);
		}
		else {
			/* Skip tokens until EOF or semicolon. */
			for (;;) {
				if (token.type EQ TOK_EOF) break;
				if (token.type EQ TOK_SEMI) break;
				next_token (fp, &token);
			}
		}

		*status = 1;
		return (NULL);
	}
}

/*
 * Recursively read an expression:
 *
 *	expr = ([-+])?term([-+]term)*
 */

	static
	struct calib_genrep *
read_expr (

FILE *		fp,		/* IN: stream to read from */
struct token *	token		/* IN/OUT: token buffer */
)
{
struct calib_genrep *	gp1;
struct calib_genrep *	gp2;
struct calib_genrep *	gp3;

	if (token -> type EQ TOK_MINUS) {
		next_token (fp, token);
		gp2 = read_term (fp, token);
		gp1 = calib_genrep_neg (gp2);
		calib_genrep_free (gp2);
	}
	else if (token -> type EQ TOK_PLUS) {
		next_token (fp, token);
		gp1 = read_term (fp, token);
	}
	else {
		gp1 = read_term (fp, token);
	}

	if (gp1 EQ NULL) {
		/* Syntax error. */
		return (NULL);
	}

	for (;;) {
		if (token -> type EQ TOK_PLUS) {
			next_token (fp, token);
			gp2 = read_term (fp, token);
			gp3 = calib_genrep_add2 (gp1, gp2);
			calib_genrep_free (gp2);
			calib_genrep_free (gp1);
			gp1 = gp3;
			continue;
		}
		if (token -> type NE TOK_MINUS) {
			break;
		}
		next_token (fp, token);
		gp3 = read_term (fp, token);
		gp2 = calib_genrep_neg (gp3);
		calib_genrep_free (gp3);
		gp3 = calib_genrep_add2 (gp1, gp2);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		gp1 = gp3;
	}

	return (gp1);
}

/*
 * Recursively read a term:
 *
 *	term = factor([ / * ]factor)*
 */

	static
	struct calib_genrep *
read_term (

FILE *		fp,		/* IN: stream to read from */
struct token *	token		/* IN/OUT: token buffer */
)
{
struct calib_genrep *	gp1;
struct calib_genrep *	gp2;
struct calib_genrep *	gp3;

	gp1 = read_factor (fp, token);

	if (gp1 EQ NULL) {
		/* Syntax error. */
		return (NULL);
	}

	for (;;) {
		if (token -> type EQ TOK_STAR) {
			next_token (fp, token);
			gp2 = read_factor (fp, token);
			gp3 = calib_genrep_mul2 (gp1, gp2);
			calib_genrep_free (gp2);
			calib_genrep_free (gp1);
			gp1 = gp3;
			continue;
		}
		if (token -> type NE TOK_SLASH) {
			break;
		}
		next_token (fp, token);
		gp2 = read_factor (fp, token);
		gp3 = calib_genrep_div2 (gp1, gp2);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		gp1 = gp3;
	}

	return (gp1);
}

/*
 * Recursively read a factor:
 *
 *	factor = ( ICONST | IDENT | '(' expr ')' | function )('^' factor)?
 *
 *	function = IDENT '(' arglist ')'
 *
 *	arglist =  {empty} | expr ( ',' expr )*
 */

	static
	struct calib_genrep *
read_factor (

FILE *		fp,		/* IN: stream to read from */
struct token *	token		/* IN/OUT: token buffer */
)
{
int				i;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep_list *	lp;

	switch (token -> type) {
	case TOK_ICONST:
		gp1 = calib_genrep_z (token -> u.zval);
		mpz_clear (token -> u.zval);
		next_token (fp, token);
		break;

	case TOK_IDENT:
		gp1 = calib_genrep_var (token -> u.ident);
		next_token (fp, token);
		if (token -> type EQ TOK_LPAREN) {
			/* A function with zero or more arguments. */
			lp = NULL;
			next_token (fp, token);
			if (token -> type NE TOK_RPAREN) {
				/* Appear to have arguments. */
				for (;;) {
					gp2 = read_expr (fp, token);
					if (gp2 EQ NULL) {
						calib_genrep_free (gp1);
						calib_genrep_free_list (lp);
						return (NULL);
					}
					lp = calib_genrep_new_list (gp2, lp);
					if (token -> type EQ TOK_RPAREN) {
						next_token (fp, token);
						break;
					}
					if (token -> type NE TOK_COMMA) {
						calib_genrep_free (gp1);
						calib_genrep_free_list (lp);
						return (NULL);
					}
					next_token (fp, token);
				}
				lp = reverse_list (lp);
			}
			/* Check for builtin functions. */
			i = calib_genrep_builtin_func_name_to_index (gp1 -> u.var);
			if (i >= 0) {
				calib_genrep_free (gp1);
				gp1 = calib_genrep_si (i);
			}
			gp1 = calib_genrep_func (gp1, lp);
		}
		break;

	case TOK_LPAREN:
		next_token (fp, token);
		gp1 = read_expr (fp, token);
		if ((gp1 EQ NULL) OR (token -> type NE TOK_RPAREN)) {
			/* Syntax error! */
			calib_genrep_free (gp1);
			return (NULL);
		}
		next_token (fp, token);
		break;

	default:
		return (NULL);
	}

	if (token -> type EQ TOK_POW) {
		next_token (fp, token);
		gp2 = read_factor (fp, token);
		if ((gp2 EQ NULL) OR
		    (NOT CALIB_GENREP_INTP (gp2)) OR
		    (mpz_cmp_si (gp2 -> u.zop, INT_MIN) < 0) OR
		    (mpz_cmp_si (gp2 -> u.zop, INT_MAX) > 0)) {
			/* Syntax error, or exponent out of range. */
			calib_genrep_free (gp2);
			calib_genrep_free (gp1);
			return (NULL);
		}
		i = mpz_get_si (gp2 -> u.zop);
		calib_genrep_free (gp2);
		gp2 = calib_genrep_ipow (gp1, i);
		calib_genrep_free (gp1);
		gp1 = gp2;
	}

	return (gp1);
}

/*
 * Reverse the given genrep_list.
 */

	static
	struct calib_genrep_list *
reverse_list (

struct calib_genrep_list *	lp	/* IN: List to destructively reverse */
)
{
struct calib_genrep_list *	root;
struct calib_genrep_list *	tmp;

	root = NULL;
	while (lp NE NULL) {
		tmp = lp -> next;
		lp -> next = root;
		root = lp;
		lp = tmp;
	}
	return (root);
}

/*
 * Check the given function name to see if it is one of the builtin
 * functions.  Return its index if so, and -1 if not.
 */

	calib_si_t
calib_genrep_builtin_func_name_to_index (

const char *	fname		/* IN: Function name to check */
)
{
#define	CHECK(func, FUNC)				\
	if (strcmp (fname, #func) EQ 0) {		\
		return (CALIB_GENREP_FUNC_ ## FUNC);	\
	}

	/* FIXME -- use a hash table here. */

	CALIB_GENREP_BUILTIN_FUNC1_LIST (CHECK)
	CALIB_GENREP_BUILTIN_FUNC2_LIST (CHECK)

	/* Not one of the builtin functions. */
	return (-1);

#undef CHECK
}

/*
 * Translate the given builtin function index into corresponding name.
 * Returns NULL if the index is invalid.
 */

	const char *
calib_genrep_builtin_func_index_to_name (

calib_si_t	fidx		/* IN: Builtin function index */
)
{
#define	CASE(func, FUNC)						\
	case CALIB_GENREP_FUNC_ ## FUNC:				\
		return (#func);

	switch (fidx) {
		CALIB_GENREP_BUILTIN_FUNC1_LIST (CASE)
		CALIB_GENREP_BUILTIN_FUNC2_LIST (CASE)
	}
	return (NULL);

#undef CASE
}

/*
 * Read the next token from the given input stream.
 */

	static
	void
next_token (

FILE *		fp,		/* IN: input stream to read from */
struct token *	token		/* IN/OUT: next token value */
)
{
int		c, echo;
char *		idbuf;
char *		idp;
char *		endp;

	echo = -1;

	for (;;) {
		c = fgetc (fp);
		if (c < 0) {
			token -> type = TOK_EOF;
			break;
		}

		switch (c) {
		case ' ': case '\t': case '\n': case '\f': case '\v':
			/* Skip blanks, newlines, and other whitespace. */
			continue;

		case '+':	token -> type = TOK_PLUS;	return;
		case ',':	token -> type = TOK_COMMA;	return;
		case '-':	token -> type = TOK_MINUS;	return;
		case '/':
			if (strip_comment (fp, &echo)) {
				continue;
			}
			token -> type = TOK_SLASH;
			return;
			token -> type = TOK_SLASH;
			return;
		case '^':	token -> type = TOK_POW;	return;
		case '(':	token -> type = TOK_LPAREN;	return;
		case ')':	token -> type = TOK_RPAREN;	return;
		case ';':	token -> type = TOK_SEMI;	return;

		case '*':
			c = fgetc (fp);
			if (c < 0) {
				token -> type = TOK_STAR;
			}
			else if (c NE '*') {
				ungetc (c, fp);
				token -> type = TOK_STAR;
			}
			else {
				token -> type = TOK_POW;
			}
			return;

		case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
		case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
		case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
		case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
		case 'Y': case 'Z':
		case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
		case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
		case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
		case 's': case 't': case 'u': case 'v': case 'w': case 'x':
		case 'y': case 'z':
		case '_': case '%':
			/* First character of an identifier. */
			idbuf	= &(token -> u.ident [0]);
			endp	= idbuf + sizeof (token -> u.ident);
			idp	= idbuf;
			*idp++ = c;
			for (;;) {
				c = fgetc (fp);
				if (c < 0) break;
				if (NOT nth_char_id (c)) {
					ungetc (c, fp);
					break;
				}
				CALIB_FATAL_ERROR_IF (idp >= endp);
				*idp++ = c;
			}
			CALIB_FATAL_ERROR_IF (idp >= endp);
			*idp = '\0';
			token -> type	= TOK_IDENT;
			return;

		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
			/* Integer constant. */
			mpz_init (token -> u.zval);
			mpz_set_ui (token -> u.zval, c - '0');
			for (;;) {
				c = fgetc (fp);
				if (c < 0) break;
				if ((c < '0') OR (c > '9')) {
					ungetc (c, fp);
					break;
				}

				mpz_mul_ui (token -> u.zval,
					    token -> u.zval,
					    10);
				mpz_add_ui (token -> u.zval,
					    token -> u.zval,
					    c - '0');
			}
			token -> type	= TOK_ICONST;
			/* Caller must do mpz_clear(token -> u.zval)! */
			return;

		default:
			token -> type = TOK_ERROR;
			return;
		}
	}
}

/*
 * Return TRUE if-and-only-if the given character is valid as the nth
 * character (not the first) of an identifier.
 */

	static
	int
nth_char_id (

int		c		/* IN: character to test */
)
{
	/* Valid nth characters of an identifier: [%0-9A-Z_a-z].	*/
	/* This code assumes ASCII!					*/

	if (c < '%') return FALSE;
	if (c EQ '%') return TRUE;
	if (c < '0') return FALSE;
	if (c <= '9') return TRUE;
	if (c < 'A') return FALSE;
	if (c <= 'Z') return TRUE;
	if (c < '_') return FALSE;
	if (c EQ '_') return TRUE;
	if (c < 'a') return FALSE;
	if (c <= 'z') return TRUE;

	return (FALSE);
}

/*
 * Having just read a '/' character, strip any C or C++ style comment.
 * Returns TRUE if a comment was stripped, or FALSE otherwise.
 *
 * Note: Comments are echoed back to stdout -- unless input is coming
 * from a (presumably interactive) TTY.
 */

	static
	calib_bool
strip_comment (

FILE *		fp,		/* IN: Input file to read from */
int *		echo_flag	/* IN/OUT: Echo output (or -1) */
)
{
int		c, echo;
calib_bool	result;

	c = fgetc (fp);
	if ((c EQ '*') OR (c EQ '/')) {
		/* Start of a C or C++ style comment. */
		echo = *echo_flag;
		if (echo < 0) {
			/* Echo comments unless input	*/
			/* is coming from a (presumably	*/
			/* interactive) TTY.		*/
			echo = NOT isatty (fileno (fp));
			*echo_flag = echo;
		}
		if (echo) {
			fputc ('/', stdout);
			fputc (c, stdout);
		}
		if (c EQ '*') {
			strip_c_comment (fp, echo);
		}
		else {
			strip_cpp_comment (fp, echo);
		}
		/* A comment was stripped. */
		result = TRUE;
	}
	else {
		ungetc (c, fp);
		result = FALSE;
	}
	return (result);
}

/*
 * Strip a C style comment, optionally echoing it to stdout.
 */

	static
	void
strip_c_comment (

FILE *		fp,		/* IN: Input file to read from */
int		echo		/* IN: Echo input to stdout? */
)
{
int		c;

	/* The slash-star has already been read. */
	for (;;) {
		c = fgetc (fp);
		if (c < 0) {
			fprintf (stderr,
				 "EOF in /*...*/ comment.\n");
			break;
		}
		if (echo) {
			fputc (c, stdout);
		}
		if (c EQ '*') {
			c = fgetc (fp);
			if (c EQ '/') {
				if (echo) {
					fputc (c, stdout);
				}
				break;
			}
			if (c >= 0) {
				ungetc (c, fp);
			}
		}
	}
}

/*
 * Strip a C++ style comment, optionally echoing it to stdout.
 */

	static
	void
strip_cpp_comment (

FILE *		fp,		/* IN: Input file to read from */
int		echo		/* IN: Echo input to stdout? */
)
{
int		c;

	for (;;) {
		c = fgetc (fp);
		if (c < 0) {
			fprintf (stderr,
				 "EOF in // comment.\n");
			break;
		}
		if (echo) {
			fputc (c, stdout);
		}
		if (c EQ '\n') break;
	}
}

/*
 * Recursively copy the given genrep tree, performing several simple but
 * important simplifications in the process.  In particular, we:
 *
 * 1. Collect and combine constant terms and factors.
 * 2. Flatten "plus-inside-of-plus" and "mul-inside-of-mul".
 */

	struct calib_genrep *
_calib_genrep_copy_simp (

const struct calib_genrep *	op	/* IN: subtree to copy/simplify */
)
{
struct calib_genrep *		gp1;
struct calib_genrep_list *	lp1;
struct _genrep_nary		root;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	switch (op -> op) {
	case CALIB_GENREP_OP_ADD:
		_calib_genrep_init_sum (&root);
		for (lp1 = op -> u.list; lp1 NE NULL; lp1 = lp1 -> next) {
			_calib_genrep_insert_sum (&root,
						  lp1 -> operand,
						   _calib_genrep_copy_simp);
		}
		gp1 = _calib_genrep_finish_sum (&root);
		break;

	case CALIB_GENREP_OP_MUL:
		_calib_genrep_init_prod (&root);
		for (lp1 = op -> u.list; lp1 NE NULL; lp1 = lp1 -> next) {
			_calib_genrep_insert_prod (&root, lp1 -> operand, _calib_genrep_copy_simp);
		}
		gp1 = _calib_genrep_finish_prod (&root);
		break;

	case CALIB_GENREP_OP_IPOW:
		gp1 = copy_simp_ipow (op -> u.ipow.base,
				      op -> u.ipow.power,
				      _calib_genrep_copy_simp);
		break;

	case CALIB_GENREP_OP_FUNC:
		gp1 = copy_func (op);
		break;

	case CALIB_GENREP_OP_VAR:
		gp1 = calib_genrep_var (op -> u.var);
		break;

	case CALIB_GENREP_OP_SI:
		gp1 = calib_genrep_si (op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		gp1 = calib_genrep_z (op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		gp1 = calib_genrep_q (op -> u.qop);
		break;

	default:
		CALIB_FATAL_ERROR;
		gp1 = calib_genrep_si (0);
		break;
	}

	return (gp1);
}

/*
 * Recursively copy the given expression op1, taking the op2-th power
 * of it in the process.
 */

	static
	struct calib_genrep *
copy_simp_ipow (

const struct calib_genrep *	op1,	 /* IN: first operand */
int				op2,	 /* IN: second operand */
grdup_func			dup_func /* IN: func to recursively */
					 /*	copy subexprs */
)
{
struct calib_genrep *		gp1;
struct calib_genrep_list *	lp;
mpq_t				qval;
struct _genrep_nary		root;

	if (op2 EQ 0) {
		if (CALIB_GENREP_ZEROP (op1)) {
			fprintf (stderr, "0**0 Generated!\n");
		}
		gp1 = calib_genrep_si (1);
	}
	else if (op2 EQ 1) {
		gp1 = (*dup_func) (op1);
	}
	else if (CALIB_GENREP_NUMBERP (op1)) {
		get_qval (qval, op1);
		_mpq_pow_si (qval, qval, op2);
		gp1 = calib_genrep_q (qval);
		mpq_clear (qval);
	}
	else if (op1 -> op EQ CALIB_GENREP_OP_MUL) {
		/* Power each factor separately. */
		_calib_genrep_init_prod (&root);
		for (lp = op1 -> u.list; lp NE NULL; lp = lp -> next) {
			_calib_genrep_insert_prod (&root, lp -> operand, dup_func);
		}
		_mpq_pow_si (root.qval, root.qval, op2);
		for (lp = root.list; lp NE NULL; lp = lp -> next) {
			gp1 = lp -> operand;
			lp -> operand = copy_simp_ipow (gp1, op2, dup_func);
			calib_genrep_free (gp1);
		}
		gp1 = _calib_genrep_finish_prod (&root);
	}
	else {
		gp1 = new_genrep ();
		gp1 -> op		= CALIB_GENREP_OP_IPOW;
		gp1 -> u.ipow.base	= (*dup_func) (op1);
		gp1 -> u.ipow.power	= op2;
	}

	return (gp1);
}

/*
 * Recursively copy the given function invocation.
 */

	static
	struct calib_genrep *
copy_func (

const struct calib_genrep *	op	/* IN: Function to copy */
)
{
struct calib_genrep *		gp;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep_list *	lp2;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR
			      (op -> op NE CALIB_GENREP_OP_FUNC));

	root = NULL;
	hookp = &root;
	for (lp = op -> u.func.args; lp NE NULL; lp = lp -> next) {
		gp = _calib_genrep_copy_simp (lp -> operand);
		lp2 = calib_genrep_new_list (gp, NULL);
		*hookp = lp2;
		hookp = &(lp2 -> next);
	}

	gp = new_genrep ();
	gp -> op		= CALIB_GENREP_OP_FUNC;
	gp -> u.func.func	= _calib_genrep_copy_simp (op -> u.func.func);
	gp -> u.func.args	= root;

	return (gp);
}

/*
 * Convert the given expression (which should satisfy CALIB_GENREP_NUMBERP())
 * into a rational, which we initialize.  The caller must clear it when done.
 */

	static
	void
get_qval (

mpq_ptr				qval,	/* IN/OUT: rational value */
const struct calib_genrep *	op	/* IN: expression to convert */
)
{
	mpq_init (qval);
	if (op -> op EQ CALIB_GENREP_OP_Z) {
		mpz_set (mpq_numref (qval), op -> u.zop);
	}
	else if (op -> op EQ CALIB_GENREP_OP_Q) {
		mpq_set (qval, op -> u.qop);
	}
	else if (op -> op EQ CALIB_GENREP_OP_SI) {
		mpq_set_si (qval, op -> u.siop, 1);
	}
	else {
		CALIB_FATAL_ERROR;
	}
}

/*
 * Allocate and zero a new genrep object.
 */

	static
	struct calib_genrep *
new_genrep (void)

{
struct calib_genrep *	gp;

	gp = CALIB_NEW (struct calib_genrep);
	memset (gp, 0, sizeof (*gp));

	return (gp);
}

/*
 * Allocate and return an empty genrep.
 */

	struct calib_genrep *
calib_genrep_new ()

{
struct calib_genrep *	gp;

	gp = new_genrep ();

	return (gp);
}

/*
 * Allocate and initializes a new genrep list object.
 */

	struct calib_genrep_list *
calib_genrep_new_list (

struct calib_genrep *		op,	/* IN: operand */
struct calib_genrep_list *	next	/* IN: other list members */
)
{
struct calib_genrep_list *	lp;

	lp = CALIB_NEW (struct calib_genrep_list);
	lp -> operand	= op;
	lp -> next	= next;

	return (lp);
}
