/**********************************************************************

	$Id: gmpmisc.c,v 1.1 2024/10/20 00:36:27 warme Exp $

	File:	gmpmisc.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Miscellaneous extensions and utilities for GMP.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/gmpmisc.h"

#include "calib/logic.h"
#include "calib/new.h"
#include <gmp.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>


/*
 * Set GMP rational rop = op1 / op2, both of which are GMP integers.
 */

	void
calib_mpq_set_z_z (

mpq_ptr		rop,		/* IN/OUT: Result */
mpz_srcptr	op1,		/* IN: First operand */
mpz_srcptr	op2		/* IN: Second operand */
)
{
mpq_t		RES;

	mpq_init (RES);

	mpz_set (mpq_numref (RES), op1);
	mpz_set (mpq_denref (RES), op2);

	mpq_canonicalize (RES);

	mpq_set (rop, RES);

	mpq_clear (RES);
}

/*
 * Multiply a GMP rational by a signed integer.
 */

	void
calib_mpq_mul_si (

mpq_ptr		rop,		/* IN/OUT: Result */
mpq_srcptr	op1,		/* IN: First operand */
calib_si_t	op2		/* IN: Second operand */
)
{
mpq_t		RES;

	mpq_init (RES);

	mpq_set (RES, op1);
	mpz_mul_si (mpq_numref (RES), mpq_numref (RES), op2);
	mpq_canonicalize (RES);

	mpq_set (rop, RES);

	mpq_clear (RES);
}

/*
 * Multiply a GMP rational by a GMP integer.
 */

	void
calib_mpq_mul_z (

mpq_ptr		rop,		/* IN/OUT: Result */
mpq_srcptr	op1,		/* IN: First operand */
mpz_srcptr	op2		/* IN: Second operand */
)
{
mpq_t		RES;

	mpq_init (RES);

	mpq_set (RES, op1);
	mpz_mul (mpq_numref (RES), mpq_numref (RES), op2);
	mpq_canonicalize (RES);

	mpq_set (rop, RES);

	mpq_clear (RES);
}

/*
 * Divide a GMP rational by a signed integer.
 */

	void
calib_mpq_div_si (

mpq_ptr		rop,		/* IN/OUT: Result */
mpq_srcptr	op1,		/* IN: First operand */
calib_si_t	op2		/* IN: Second operand */
)
{
mpq_t		RES;

	mpq_init (RES);

	mpq_set (RES, op1);
	mpz_mul_si (mpq_denref (RES), mpq_denref (RES), op2);
	mpq_canonicalize (RES);

	mpq_set (rop, RES);

	mpq_clear (RES);
}

/*
 * Divide a GMP rational by a GMP integer.
 */

	void
calib_mpq_div_z (

mpq_ptr		rop,		/* IN/OUT: Result */
mpq_srcptr	op1,		/* IN: First operand */
mpz_srcptr	op2		/* IN: Second operand */
)
{
mpq_t		RES;

	mpq_init (RES);

	mpq_set (RES, op1);
	mpz_mul (mpq_denref (RES), mpq_denref (RES), op2);
	mpq_canonicalize (RES);

	mpq_set (rop, RES);

	mpq_clear (RES);
}

/*
 * Initialize the given vector of GMP integers.
 */

	void
calib_init_Z_vector (

mpz_ptr		p,		/* IN: vector to initialize */
size_t		n		/* IN: number of elements */
)
{
size_t		i;

	for (i = 0; i < n; i++) {
		mpz_init (&p [i]);
	}
}

/*
 * Clear the given vector of GMP integers.
 */

	void
calib_clear_Z_vector (

mpz_ptr		p,		/* IN: vector to clear */
size_t		n		/* IN: number of elements */
)
{
size_t		i;

	for (i = 0; i < n; i++) {
		mpz_clear (&p [i]);
	}
}

/*
 * Allocate a vector of GMP integers.
 */

	mpz_ptr
calib_new_Z_vector (

size_t		n		/* IN: number of elements */
)
{
mpz_ptr		zp;

	if (n <= 0) {
		n = 1;
	}
	zp = CALIB_NEWA (n, __mpz_struct);
	calib_init_Z_vector (zp, n);
	return (zp);
}

/*
 * Free a vector of GMP integers.
 */

	void
calib_free_Z_vector (

mpz_ptr		array,		/* IN: vector to free */
size_t		n		/* IN: number of elements in vector */
)
{
	if (array NE NULL) {
		if (n <= 0) {
			n = 1;
		}
		calib_clear_Z_vector (array, n);
		free (array);
	}
}

/*
 * Copy a vector of integers.
 */

	void
calib_copy_Z_vector (

mpz_ptr		dst,		/* IN: destination vector */
mpz_srcptr	src,		/* IN: source vector */
size_t		n		/* IN: number of elements to copy */
)
{
size_t		i;

	for (i = 0; i < n; i++) {
		mpz_set (dst++, src++);
	}
}

/*
 * Print the given vector of integers.
 */

	void
calib_print_Z_vector (

mpz_srcptr	vector,		/* IN: vector of integers to print */
size_t		n		/* IN: size of vector */
)
{
size_t		i;
mpz_srcptr	zp;

	zp = vector;
	for (i = 0; i < n; i++, zp++) {
		printf (" ");
		mpz_out_str (stdout, 10, zp);
	}
	printf ("\n");
}

/*
 * Initialize the given vector of GMP rationals.
 */

	void
calib_init_Q_vector (

mpq_ptr		p,		/* IN: vector to initialize */
size_t		n		/* IN: number of elements */
)
{
size_t		i;

	for (i = 0; i < n; i++) {
		mpq_init (&p [i]);
	}
}

/*
 * Clear the given vector of GMP rationals.
 */

	void
calib_clear_Q_vector (

mpq_ptr		p,		/* IN: vector to clear */
size_t		n		/* IN: number of elements */
)
{
size_t		i;

	for (i = 0; i < n; i++) {
		mpq_clear (&p [i]);
	}
}

/*
 * Allocate a vector of GMP rationals.
 */

	mpq_ptr
calib_new_Q_vector (

size_t		n		/* IN: number of elements */
)
{
mpq_ptr		qp;

	if (n <= 0) {
		n = 1;
	}
	qp = CALIB_NEWA (n, __mpq_struct);
	calib_init_Q_vector (qp, n);
	return (qp);
}

/*
 * Free a vector of GMP rationals.
 */

	void
calib_free_Q_vector (

mpq_ptr		array,		/* IN: vector to free */
size_t		n		/* IN: number of elements in vector */
)
{
	if (array NE NULL) {
		if (n <= 0) {
			n = 1;
		}
		calib_clear_Q_vector (array, n);
		free (array);
	}
}

/*
 * Copy a vector of rational numbers.
 */

	void
calib_copy_Q_vector (

mpq_ptr		dst,		/* IN: destination vector */
mpq_srcptr	src,		/* IN: source vector */
size_t		n		/* IN: number of elements to copy */
)
{
size_t		i;

	for (i = 0; i < n; i++) {
		mpq_set (dst++, src++);
	}
}

/*
 * Print the given vector of rational numbers.
 */

	void
calib_print_Q_vector (

mpq_srcptr	vector,		/* IN: vector of rationals to print */
size_t		n		/* IN: size of vector */
)
{
size_t		i;
mpz_srcptr	numer, denom;

	for (i = 0; i < n; i++) {
		numer = mpq_numref (&vector [i]);
		denom = mpq_denref (&vector [i]);
		printf (" ");
		mpz_out_str (stdout, 10, numer);
		printf ("/");
		mpz_out_str (stdout, 10, denom);
	}
	printf ("\n");
}

/*
 * Return the bit-size of the given integer vector.
 */

	size_t
calib_mpz_bit_size_vector (

mpz_srcptr	op,		/* IN: mpz vector */
size_t		n		/* IN: vector length */
)
{
size_t		i, nbits, max_nbits;

	max_nbits = 0;
	for (i = 0; i < n; i++) {
		nbits = calib_mpz_bit_size (&op [i]);
		if (nbits > max_nbits) {
			max_nbits = nbits;
		}
	}
	return (max_nbits);
}


/*
 * Return the bit-size of the given rational vector.
 */

	size_t
calib_mpq_bit_size_vector (

mpq_srcptr	op,		/* IN: mpq vector */
size_t		n		/* IN: vector length */
)
{
size_t		i, nbits, max_nbits;

	max_nbits = 0;
	for (i = 0; i < n; i++) {
		nbits = calib_mpq_bit_size (&op [i]);
		if (nbits > max_nbits) {
			max_nbits = nbits;
		}
	}
	return (max_nbits);
}

/*
 * Return the bit-size of the given rational number.
 */

	size_t
calib_mpq_bit_size (

mpq_srcptr	q		/* IN: mpq to get bit size of */
)
{
size_t		nbits;
mpz_t		zval;

	mpz_init (zval);
	mpz_mul (zval, mpq_numref (q), mpq_denref (q));
	nbits = calib_mpz_bit_size (zval);
	mpz_clear (zval);

	return (nbits);
}


/*
 * Return the bit-size of the given integer.
 */

	size_t
calib_mpz_bit_size (

mpz_srcptr	z		/* IN: mpz to get bit size of */
)
{
int		size;
size_t		nbits;
mp_limb_t	hilimb;

	size = z -> _mp_size;
	if (size <= 0) {
		if (size EQ 0) {
			/* Number is zero.  This is 1 bit. */
			return (1);
		}
		size = -size;
	}

	/* Count bits in all limbs except the highest. */
	nbits = size - 1;
	nbits *= GMP_LIMB_BITS;

	/* Count bits in the highest limb. */
	hilimb = z -> _mp_d [size - 1];

	if ((hilimb & 0xFFFFFFFF00000000lu) NE 0) {
		nbits += 32;
		hilimb >>= 32;
	}
	if ((hilimb & 0xFFFF0000) NE 0) {
		nbits += 16;
		hilimb >>= 16;
	}
	if ((hilimb & 0xFF00) NE 0) {
		nbits += 8;
		hilimb >>= 8;
	}
	if ((hilimb & 0xF0) NE 0) {
		nbits += 4;
		hilimb >>= 4;
	}
	if ((hilimb & 0x0C) NE 0) {
		nbits += 2;
		hilimb >>= 2;
	}
	if ((hilimb & 0x02) NE 0) {
		nbits += 1;
		hilimb >>= 1;
	}
	if ((hilimb & 0x01) NE 0) {
		++nbits;
	}

	return (nbits);
}
