/**********************************************************************

	$Id: ivechash.c,v 1.1 2024/10/20 00:36:27 warme Exp $

	File:	ivechash.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A hash table for fixed-sized vectors of int's.

***********************************************************************

	Modification Log:

	a-1:	10/13/2015	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "ivechash.h"

#include "calib/fatal.h"
#include "calib/logic.h"
#include "calib/new.h"
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

/*
 * Macro to free up a hash table entry.
 */

#define FREE_ENTRY(ENTRY, UFUNC, UDATA)					\
	do {								\
		if ((UFUNC) NE NULL) {					\
			(*(UFUNC)) ((ENTRY), IVEOP_FREE, (UDATA));	\
		}							\
		free ((ENTRY));						\
	} while (FALSE)

/*
 * Macro to create/initialize a new hash table entry.
 */

#define CREATE_ENTRY(ENTRY, UFUNC, UDATA)				\
	do {								\
		if ((UFUNC) NE NULL) {					\
			(*(UFUNC)) ((ENTRY), IVEOP_CREATE, (UDATA));	\
		}							\
	} while (FALSE)

/*
 * Local Functions
 */

static int	compare_integer_vectors (int		vsize,
					 const int *	vec1,
					 const int *	vec2);
static int	ivec_hash_function (int vsize, const int * vec);

/*
 * Create an empty ivec_hash table for vectors of the given size.
 * The client data in each hash table entry are managed by the
 * given ufunc/udata pair.
 */

	struct ivec_hash *
_calib_ivec_hash_create (

int			vsize,	/* IN: size of int vector keys */
ivec_entry_func_t *	ufunc,	/* IN: function to handle client data */
void *			udata	/* IN: data for use by ufunc */
)
{
int				i, nbuck;
struct ivec_hash *		htp;
struct ivec_hash_entry **	buckets;
struct ivec_dlink *		dlink;

	CALIB_FATAL_ERROR_IF (vsize < 0);

	htp = CALIB_NEW (struct ivec_hash);

	memset (htp, 0, sizeof (*htp));

	nbuck = IVEC_HASH_SIZE;

	buckets = CALIB_NEWA (nbuck, struct ivec_hash_entry *);
	dlink	= CALIB_NEWA (nbuck + 1, struct ivec_dlink) + 1;

	memset (buckets, 0, nbuck * sizeof (buckets [0]));

	dlink [-1].forw = -1;
	dlink [-1].back = -1;
	for (i = 0; i < nbuck; i++) {
		dlink [i].forw = -2;
		dlink [i].back = -2;
	}

	htp -> vsize	= vsize;
	htp -> nentries	= 0;
	htp -> buckets	= buckets;
	htp -> link	= dlink;
	htp -> ufunc	= ufunc;
	htp -> udata	= udata;

	return (htp);
}

/*
 * Destroy the given ivec_hash table.
 */

	void
_calib_ivec_hash_free (

struct ivec_hash *	htp	/* IN: hash table to free up */
)
{
	if (htp EQ NULL) {
		return;
	}

	/* Empty out the hash table. */
	_calib_ivec_hash_reset (htp);

	if (htp -> link NE NULL) {
		free (htp -> link - 1);
	}
	if (htp -> buckets NE NULL) {
		free (htp -> buckets);
	}

	memset (htp, 0, sizeof (*htp));

	free (htp);
}

/*
 * Force the given hash table to be empty.
 */

	void
_calib_ivec_hash_reset (

struct ivec_hash *	htp	/* IN: hash table to reset */
)
{
int				i, j;
struct ivec_hash_entry *	ep;
struct ivec_hash_entry *	ep2;
ivec_entry_func_t *		ufunc;
void *				udata;

	if (htp EQ NULL) return;

	ufunc = htp -> ufunc;
	udata = htp -> udata;

	for (;;) {
		i = htp -> link [-1].forw;
		if (i < 0) {
			CALIB_FATAL_ERROR_IF (i < -1);
			break;
		}
		CALIB_FATAL_ERROR_IF (i >= IVEC_HASH_SIZE);
		ep = htp -> buckets [i];
		CALIB_FATAL_ERROR_IF (ep EQ NULL);
		htp -> buckets [i] = NULL;
		while (ep NE NULL) {
			ep2 = ep -> next;
			ep -> next = NULL;
			FREE_ENTRY (ep, ufunc, udata);
			--(htp -> nentries);
			ep = ep2;
		}
		CALIB_FATAL_ERROR_IF (htp -> link [i].back NE -1);
		j = htp -> link [i].forw;
		htp -> link [j].back = -1;
		htp -> link [-1].forw = j;
		htp -> link [i].forw = -2;
		htp -> link [i].back = -2;
	}

	CALIB_FATAL_ERROR_IF (htp -> nentries NE 0);
}

/*
 * Search the hash table looking for the given vector.
 * Return value is a "hook pointer" for the entry sought.
 * If (bucket NE NULL), then a hook pointer for the start of
 * the particular hash table bucket is also stored into *bucket.
 * (This is needed, e.g. for managing empty to non-empty transitions
 * of this bucket during the _add() operation, or non-empty to
 * empty transitions during the _remove() operation.)
 *
 * This function does *NOT* automatically create an entry for the
 * given vector if the entry does not exist.
 */

	struct ivec_hash_entry **
_calib_ivec_hash_find (

struct ivec_hash *	htp,	/* IN: hash table to search */
const int *		vec,	/* IN: vector to find */
struct ivec_hash_entry ***
			_bucket	/* IN/OUT: start of bucket vector belongs in */
)
{
int				h;
int				vsize, nbytes;
struct ivec_hash_entry *	ep;
struct ivec_hash_entry **	bucket;
struct ivec_hash_entry **	hookp;

	CALIB_FATAL_ERROR_IF ((htp EQ NULL) OR (vec EQ NULL));

	vsize = htp -> vsize;
	nbytes = vsize * sizeof (int);

	h = ivec_hash_function (vsize, vec);

	bucket = &(htp -> buckets [h]);

	if (_bucket NE NULL) {
		*_bucket = bucket;
	}

	hookp = bucket;
	for (;;) {
		ep = *hookp;
		if (ep EQ NULL) break;

		if (memcmp (ep -> vec, vec, nbytes) EQ 0) break;

		hookp = &(ep -> next);
	}
	return (hookp);
}

/*
 * Search the hash table looking for the given vector.
 * Return value is a "hook pointer" for the entry sought.
 * If (bucket NE NULL), then a hook pointer for the start of
 * the particular hash table bucket is also stored into *bucket.
 * (This is needed, e.g. for managing empty to non-empty transitions
 * of this bucket during the _add() operation, or non-empty to
 * empty transitions during the _remove() operation.)
 *
 * This function *DOES* automatically create an entry for the
 * given vector if the entry does not already exist.
 */

	struct ivec_hash_entry **
_calib_ivec_hash_make (

struct ivec_hash *	htp,	/* IN: hash table to search */
const int *		vec,	/* IN: vector to find */
struct ivec_hash_entry ***
			_bucket	/* IN/OUT: start of bucket vector belongs in */
)
{
int				h, i1, i2;
int				vsize, nbytes;
struct ivec_hash_entry *	ep;
struct ivec_hash_entry **	bucket;
struct ivec_hash_entry **	hookp;
struct ivec_dlink *		link;

	CALIB_FATAL_ERROR_IF ((htp EQ NULL) OR (vec EQ NULL));

	vsize = htp -> vsize;
	nbytes = vsize * sizeof (int);

	h = ivec_hash_function (vsize, vec);

	bucket = &(htp -> buckets [h]);

	if (_bucket NE NULL) {
		*_bucket = bucket;
	}

	hookp = bucket;
	for (;;) {
		ep = *hookp;
		if (ep EQ NULL) break;

		if (memcmp (ep -> vec, vec, nbytes) EQ 0) {
			/* Existing entry found! */
			return (hookp);
		}

		hookp = &(ep -> next);
	}

	/* Need to create a new entry. */
	nbytes	= IVEC_HASH_ENTRY_SIZE (vsize);
	ep	= _calib_new (nbytes);

	memset (ep, 0, IVEC_HASH_ENTRY_SIZE (0));
	memcpy (ep -> vec, vec, vsize * sizeof (int));

	CREATE_ENTRY (ep, htp -> ufunc, htp -> udata);

	if (*bucket EQ NULL) {
		/* This bucket is transitioning from empty to	*/
		/* non-empty.  Add this bucket to doubly-linked	*/
		/* list of non-empty buckets.			*/
		link = htp -> link;
		CALIB_FATAL_ERROR_IF ((link [h].forw NE -2) OR
				      (link [h].back NE -2));
		i1 = -1;
		i2 = link [i1].forw;
		link [h].back = i1;
		link [h].forw = i2;
		link [i1].forw = h;
		link [i2].back = h;
	}

	/* Insert new entries at front of bucket chain. */
	hookp = bucket;

	ep -> next	= *hookp;
	*hookp		= ep;

	++(htp -> nentries);

	return (hookp);
}

/*
 * Remove an entry from the given hash table.
 *
 * The entry is specified via two pieces of data:
 *
 *	- The "hook pointer" for the hash table entry.
 *
 *	- A "hook pointer" for the front of the chain for
 *	  the hash table bucket to which this entry belongs.
 *
 * The second entry is required in order to detect and process the
 * case where the bucket transitions from non-empty to empty.
 */

	void
_calib_ivec_hash_remove (

struct ivec_hash *	htp,	/* IN: hash table to remove entry from */
struct ivec_hash_entry **
			hookp,	/* IN/OUT: hook pointer for item to remove */
struct ivec_hash_entry **
			bucket	/* IN/OUT: start of bucket vector belongs in */
)
{
int				h, i1, i2;
struct ivec_hash_entry *	ep;
struct ivec_dlink *		link;

	CALIB_FATAL_ERROR_IF ((htp EQ NULL) OR
			      (hookp EQ NULL) OR
			      (bucket EQ NULL));

	ep = *hookp;

	CALIB_FATAL_ERROR_IF (ep EQ NULL);

	*hookp = ep -> next;

	ep -> next = NULL;

	FREE_ENTRY (ep, htp -> ufunc, htp -> udata);
	--(htp -> nentries);

	if (*bucket EQ NULL) {
		/* This bucket transitioned from non-empty to empty.	*/
		/* Remove this bucket from doubly-linked list of	*/
		/* non-empty buckets.					*/
		h = bucket - htp -> buckets;
		CALIB_FATAL_ERROR_IF ((h < 0) OR ( h >= IVEC_HASH_SIZE));
		link = htp -> link;
		i1 = link [h].back;
		i2 = link [h].forw;
		CALIB_FATAL_ERROR_IF ((i1 <= -2) OR
				      (i1 >= IVEC_HASH_SIZE) OR
				      (i2 <= -2) OR
				      (i2 >= IVEC_HASH_SIZE));
		link [i1].forw = i2;
		link [i2].back = i1;
		link [h].forw = -2;
		link [h].back = -2;
	}
}

/*
 * Hash function for a vector of integers of given size.
 */

	static
	int
ivec_hash_function (

int		vsize,		/* IN: size of vector */
const int *	vec		/* IN: vector to compute hash for */
)
{
int		i;
unsigned long	reg, word;

	/* Assure that we have at least 64 bits. */
	CALIB_FATAL_ERROR_IF (sizeof (reg) NE 8);

	/* Initial register is all one bits. */
	reg = 0;
	reg = ~reg;

	for (i = 0; i < vsize; i++) {
		/* Rotate register 9 bits to the left. */
		reg = (reg << 9) | ((reg >> 55) & 0x1FF);
		word = vec [i];
		word &= 0xFFFFFFFF;
		reg ^= word;
	}
	reg %= IVEC_HASH_SIZE;

	return (reg);
}

/*
 * Return a single array of pointers to all the entries of the given
 * hash table.  The entries are sorted according to a collating sequence
 * of the integer vector keys.
 *
 * It is the caller's responsibility to free the returned array of pointers.
 */

	struct ivec_hash_entry **
_calib_ivec_hash_sort_entries (

const struct ivec_hash *	htp	/* IN: hash table get entries from */
)
{
int				i, j, k, n, nentries, vsize, status;
struct ivec_hash_entry **	bucket;
struct ivec_dlink *		link;
struct ivec_hash_entry **	sarray;
struct ivec_hash_entry **	sp;
struct ivec_hash_entry *	ep;
struct ivec_hash_entry *	ep1;
struct ivec_hash_entry *	ep2;

	CALIB_FATAL_ERROR_IF (htp EQ NULL);

	nentries	= htp -> nentries;
	vsize		= htp -> vsize;
	bucket		= htp -> buckets;
	link		= htp -> link;

	/* Make an array of all entries in the hash table. */
	sarray = CALIB_NEWA (nentries, struct ivec_hash_entry *);
	sp = sarray;
	for (i = link [-1].forw; i >= 0; i = link [i].forw) {
		ep = bucket [i];
		CALIB_FATAL_ERROR_IF (ep EQ NULL);
		for (; ep NE NULL; ep = ep -> next) {
			*sp++ = ep;
		}
	}
	CALIB_FATAL_ERROR_IF (sp NE sarray + nentries);

	/* Construct the heap via sift-downs, in O(n) time. */
	n = nentries;
	for (k = n >> 1; k >= 0; k--) {
		j = k;
		for (;;) {
			i = (j << 1) + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the right subchild is worse. */
				ep1 = sarray [i];
				ep2 = sarray [i + 1];
				status = compare_integer_vectors (vsize,
								  ep1 -> vec,
								  ep2 -> vec);
				CALIB_FATAL_ERROR_IF (status EQ 0);
				if (status > 0) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			ep1 = sarray [j];
			ep2 = sarray [i];
			status = compare_integer_vectors (vsize,
							  ep1 -> vec,
							  ep2 -> vec);
			CALIB_FATAL_ERROR_IF (status EQ 0);
			if (status < 0) {
				/* Worst child is better.  Sift-down	*/
				/* is done. */
				break;
			}
			/* Sift down and continue. */
			sarray [j] = ep2;
			sarray [i] = ep1;
			j = i;
		}
	}

	/* Now do actual sorting.  Exchange first/last and sift down. */
	while (n > 1) {
		/* Worst is at order [0], swap with order [n-1],	*/
		/* thereby putting it into final position.		*/
		--n;
		ep1 = sarray [0];
		sarray [0] = sarray [n];
		sarray [n] = ep1;

		/* Now restore the heap by sifting order [0] down. */
		j = 0;
		for (;;) {
			i = (j << 1) + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the right subchild is worse. */
				ep1 = sarray [i];
				ep2 = sarray [i + 1];
				status = compare_integer_vectors (vsize,
								  ep1 -> vec,
								  ep2 -> vec);
				CALIB_FATAL_ERROR_IF (status EQ 0);
				if (status > 0) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			ep1 = sarray [j];
			ep2 = sarray [i];
			status = compare_integer_vectors (vsize,
							  ep1 -> vec,
							  ep2 -> vec);
			CALIB_FATAL_ERROR_IF (status EQ 0);
			if (status < 0) {
				/* Worst child is better.  Sift-down	*/
				/* is done. */
				break;
			}
			/* Sift down and continue. */
			sarray [j] = ep2;
			sarray [i] = ep1;
			j = i;
		}
	}

	return (sarray);
}

/*
 * Compare two vectors of integers lexicographically.
 */

	static
	int
compare_integer_vectors (

int		vsize,		/* IN: number of variables */
const int *	vec1,		/* IN: first vector of integers */
const int *	vec2		/* IN: second vector of integers */
)
{
int	i, val1, val2;

	for (i = 0; i < vsize; i++) {
		val1 = vec1 [i];
		val2 = vec2 [i];
		if (val1 < val2) return (-1);
		if (val1 > val2) return (+1);
	}
	return (0);
}
