/**********************************************************************

	$Id: ivechash.h,v 1.1 2024/10/20 00:36:27 warme Exp $

	File:	ivechash.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A hash table for fixed-sized vectors of int's.

***********************************************************************

	Modification Log:

	a-1:	10/13/2015	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_IVEC_HASH_H
#define CALIB_IVEC_HASH_H

struct ivec_hash;
struct ivec_hash_entry;

/*
 * Operations for ivec_entry_func_t.
 */

enum ivec_entry_op {
	IVEOP_CREATE,		/* Creat a new entry */
	IVEOP_FREE,		/* Destroy an entry */
};

/*
 * Type signature for function used to construct and/or destruct
 * user data in a single hash table entry.
 */

typedef void ivec_entry_func_t (struct ivec_hash_entry *	ep,
				enum ivec_entry_op		op,
				void *				data);

/*
 * A hash table for fixed sized arrays of int's.
 *
 * For the sake of efficiency, we want the hash table to have lots of
 * buckets (to minimize collisions), but at the same time we want to be
 * able to efficiently iterate over all items in the hash table (even if
 * the table is empty or has just one entry in it).  This means we maintain
 * a doubly-linked list of the non-empty buckets, which we update in O(1)
 * time on each empty-to-non-empty or non-empty-to-empty transition.
 * Clearing out the hash table when we are done with it also uses this
 * linked list so that we only clear the non-empty buckets.
 */

#define IVEC_HASH_SIZE	10009

struct ivec_hash {
	int		vsize;		/* Number of elements in each */
					/* integer vector */
	int		nentries;	/* Current number of table entries */
	struct ivec_hash_entry **
			buckets;	/* Array of hash table buckets */
	struct ivec_dlink *
			link;		/* Array of forward/backward links, */
					/* one per bucket, plus a "root" */
					/* entry at [-1]. */
	ivec_entry_func_t *
			ufunc;		/* User-provided function for */
					/* handling hash table entries */
	void *		udata;		/* User-provided data for ufunc */
};

/*
 * The ivec_hash table entry.
 * These are allocated with variable size so that the integer vector
 * is embedded directly in the node.
 */

struct ivec_hash_entry {
	void *		data;		/* Client data */
	struct ivec_hash_entry *
			next;		/* Next entry in linked list */
	int		vec [1];	/* Array of integers */
};

#define IVEC_HASH_ENTRY_SIZE(VSIZE) \
	(offsetof (struct ivec_hash_entry, vec [0]) + (VSIZE) * sizeof (int))

/* The doubly-linked list entry. */

struct ivec_dlink {
	short		forw;		/* Forward index */
	short		back;		/* Backward index */
};

/* Functions */

extern struct ivec_hash *
		_calib_ivec_hash_create (
				int			vsize,
				ivec_entry_func_t *	ufunc,
				void *			udata);
extern void	_calib_ivec_hash_free (struct ivec_hash * htp);
extern void	_calib_ivec_hash_reset (struct ivec_hash * htp);

extern struct ivec_hash_entry **
		_calib_ivec_hash_find (
				struct ivec_hash *		htp,
				const int *			vec,
				struct ivec_hash_entry ***	bucket);
extern struct ivec_hash_entry **
		_calib_ivec_hash_make (
				struct ivec_hash *		htp,
				const int *			vec,
				struct ivec_hash_entry ***	bucket);
extern void	_calib_ivec_hash_remove (
				struct ivec_hash *		htp,
				struct ivec_hash_entry **	hookp,
				struct ivec_hash_entry **	bucket);
extern struct ivec_hash_entry **
		_calib_ivec_hash_sort_entries (const struct ivec_hash * htp);

#endif
