/**********************************************************************

	$Id: lll.c,v 1.1 2024/10/20 00:36:27 warme Exp $

	File:	lll.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	CALIB implementation of LLL algorithm.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/lll.h"

#include "calib/fatal.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/types.h"
#include <gmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Global Functions
 */

int		_calib_LLL (mpz_ptr	res,
			    mpz_srcptr	b,
			    int		nvec,
			    int		vsize);

/* Do not warn about unused functions anywhere in this file! */
#pragma GCC diagnostic ignored "-Wunused-function"

/*
 * Global info for the LLL algorithm.
 */

struct LLLInfo {
	int		li_nvec;		/* Number of vectors */
	int		li_vsize;		/* Size of each vector */
	int		li_bs_valid;		/* Index of last valid */
						/* Gram-Schmidt vector / */
						/* mag**2 value */
	mpz_ptr		li_B;			/* Lattice basis vectors */
	int *		li_order;		/* Order of basis vectors */
	mpq_ptr		li_Bstar;		/* Gram-Schmidt vectors */
	mpq_ptr		li_Bsmag2;		/* Mag**2 of GS vectors */
};

/*
 * A matrix of GMP integers.
 */

struct ZMatrix {
	int		zm_rows;	/* Number of rows */
	int		zm_cols;	/* Number of columns */
	mpz_ptr		zm_A;		/* Matrix of integer values */
};

/*
 * A matrix of GMP rationals.
 */

struct QMatrix {
	int		qm_rows;	/* Number of rows */
	int		qm_cols;	/* Number of columns */
	mpq_ptr		qm_A;		/* Matrix of rational values */
};

/*
 * Local Functions
 */

static void	compute_mu (struct LLLInfo *	lip,
			    mpq_ptr		mu_out,
			    int			i,
			    int			j);
static void	debug_print_Gram_Schmidt (mpq_srcptr	Bstar,
					  int		nvec,
					  int		vsize);
static mpq_ptr	Gram_Schmidt (mpz_srcptr B, int nvec, int vsize);
static void	mpq_mag2 (mpq_ptr result, mpq_srcptr v, int n);
static double	mpq_mag2_double (mpq_srcptr v, int n);
static void	mpz_mag2 (mpz_ptr result, mpz_srcptr v, int n);
static int	reduce_lattice_basis_LLL (mpz_ptr	B,
					  int		nvec,
					  int		vsize,
					  mpq_srcptr	Delta,
					  int *		order);
static int	reduce_lattice_basis_LLL_old (mpz_ptr		B,
					      int		nvec,
					      int		vsize,
					      mpq_srcptr	Delta,
					      int *		order);
static calib_bool
		reduction_phase (mpz_ptr	B,
				 const int *	order,
				 mpq_srcptr	Bstar,
				 mpq_srcptr	Bsmag2,
				 int		nvec,
				 int		vsize);
static calib_bool
		swap_phase (mpz_srcptr	B,
			    int *	order,
			    mpq_srcptr	Bstar,
			    mpq_srcptr	Bsmag2,
			    mpq_srcptr	Delta,
			    int		nvec,
			    int		vsize);
static calib_bool
		test_delta_reduced (mpz_srcptr	B,
				    int		nvec,
				    int		vsize,
				    mpq_srcptr	Delta);
static void	update_bstar (struct LLLInfo * lip, int i);
static void	verify_GS_dot_B_mags (mpz_srcptr	B,
				      mpq_srcptr	Bstar,
				      mpq_srcptr	Bsmag2,
				      int		nvec,
				      int		vsize);

/*
 * The internal, CALIB implementation of the Lenstra, Lenstra, Lovasz
 * lattice-basis reduction algorithm.
 *
 * The lattice-basis arrays are structured as a sequence of basis vectors.
 * (The coefficients of each vector occupy consecutive array elements.)
 *
 * This implementation doesn't work yet.  We will remove the leading
 * underscore when it works better.
 */

	int
_calib_LLL (

mpz_ptr		res,		/* IN/OUT: Reduced lattice-basis */
mpz_srcptr	b,		/* IN: Lattic-basis to reduce */
int		nvec,		/* IN: Number of basis vectors */
int		vsize		/* IN: Size of each vector */
)
{
int		i, n, status;
int *		order;
mpq_t		Delta;
mpz_ptr		lbase;

	CALIB_FATAL_ERROR_IF ((res EQ NULL) OR (b EQ NULL));

	/* Accept only non-trivial problems. */
	CALIB_FATAL_ERROR_IF ((nvec <= 0) OR (vsize <= 0));

	n = nvec * vsize;

	if (nvec EQ 1) {
		calib_copy_Z_vector (res, b, n);
		return (0);
	}

	mpq_init (Delta);
	mpq_set_ui (Delta, 3, 4);

	/* Make local copy of the lattice-basis. */
	lbase = calib_new_Z_vector (n);
	calib_copy_Z_vector (lbase, b, n);

	order = CALIB_NEWA (nvec, int);

#if 1
	status = reduce_lattice_basis_LLL (lbase,
					   nvec,
					   vsize,
					   Delta,
					   order);
#else
	status = reduce_lattice_basis_LLL_old (lbase,
					       nvec,
					       vsize,
					       Delta,
					       order);
#endif

	/* Copy vectors back to caller in proper order. */
	for (i = 0; i < nvec; i++) {
		calib_copy_Z_vector (&res [i * vsize],
				      &lbase [order [i] * vsize],
				      vsize);
	}

	free (order);
	calib_free_Z_vector (lbase, n);

	return (status);
}

/*
 * Reduce the lattice basis BLambda using the Lenstra-Lenstra-Lovasz
 * lattice basis reduction algorithm.
 *
 * New implementation based upon pseudo-code on Wikipedia.
 */

	static
	int
reduce_lattice_basis_LLL (

mpz_ptr			B,		/* IN/OUT: Lattice-basis to reduce */
int			nvec,		/* IN: Number of vectors */
int			vsize,		/* IN: Size of each vector */
mpq_srcptr		Delta,		/* IN: The parameter Delta */
int *			order		/* IN/OUT: final order of vectors */
)
{
int			i, j, k;
mpz_srcptr		szp;
mpq_ptr			Bstar, Bsmag2;
mpz_ptr			dzp;
struct LLLInfo *	lip;
struct LLLInfo		lllInfo;
mpz_t			FACTOR, ZTEMP;
mpq_t			MU, HALF, QTEMP;

	lip = &lllInfo;
	memset (lip, 0, sizeof (*lip));

	/* Allocate buffers. */
	k	= nvec * vsize;
	Bstar	= calib_new_Q_vector (k);
	Bsmag2	= calib_new_Q_vector (vsize);

	/* Set initial ordering of B vectors. */
	for (i = 0; i < nvec; i++) {
		order [i] = i;
	}

	lip -> li_nvec		= nvec;
	lip -> li_vsize		= vsize;
	lip -> li_bs_valid	= -1;	/* No Bstar are valid. */
	lip -> li_B		= B;
	lip -> li_order		= order;
	lip -> li_Bstar		= Bstar;
	lip -> li_Bsmag2	= Bsmag2;

	mpz_init (FACTOR);
	mpz_init (ZTEMP);

	mpq_init (MU);
	mpq_init (HALF);
	mpq_init (QTEMP);

	mpq_set_ui (HALF, 1, 2);

	for (k = 1; k < nvec; ) {
		dzp = &B [order [k]];
		for (j = k - 1; j >= 0; j--) {
			compute_mu (lip, MU, k, j);
			mpq_abs (QTEMP, MU);
			if (mpq_cmp (QTEMP, HALF) <= 0) {
				continue;
			}
			mpq_add (QTEMP, MU, HALF);
			mpz_fdiv_q (FACTOR,
				    mpq_numref (QTEMP),
				    mpq_denref (QTEMP));
			CALIB_FATAL_ERROR_IF (mpz_sgn (FACTOR) EQ 0);
			szp = &B [order [j]];
			for (i = 0; i < vsize; i++) {
				mpz_mul (ZTEMP, FACTOR, &szp [i]);
				mpz_sub (&dzp [i], &dzp [i], ZTEMP);
			}
		}
		/* Mark Bstar_k and beyond invalid.  Note that the	*/
		/* previous loop only reads B_j and Bstar_j for j < k,	*/
		/* they are never modified, so we only invalidate when	*/
		/* the entire reduction update of B_k is finished.	*/
		if (lip -> li_bs_valid >= k) {
			lip -> li_bs_valid = k - 1;
		}

		/* Assure that Gram-Schmidt vectors (and mag**2) are	*/
		/* correct up to Bstar_k.				*/
		update_bstar (lip, k);

		/* Test the Lovasz condition. */
		compute_mu (lip, MU, k, k - 1);
		mpq_mul (QTEMP, MU, MU);
		mpq_sub (QTEMP, Delta, QTEMP);
		mpq_mul (QTEMP, QTEMP, &Bsmag2 [k - 1]);
		if (mpq_cmp (&Bsmag2 [k], QTEMP) >= 0) {
			++k;
		}
		else {
			/* Swap B[k] and B[k-1]. */
			i = order [k];
			order [k] = order [k - 1];
			order [k - 1] = i;
			/* Mark Bstar invalid at k-1 and beyond. */
			if (lip -> li_bs_valid >= k - 1) {
				lip -> li_bs_valid = k - 2;
			}
			if (k > 1) {
				--k;
			}
		}
	}

	mpq_clear (QTEMP);
	mpq_clear (HALF);
	mpq_clear (MU);

	mpz_clear (ZTEMP);
	mpz_clear (FACTOR);

	calib_free_Q_vector (Bsmag2, nvec);

	k = nvec * vsize;
	calib_free_Q_vector (Bstar, k);
	memset (lip, 0, sizeof (*lip));

	return (0);
}

/*
 * Compute mu[i,j] = <B[i], Bstar[j]> / <Bstar[j], Bstar[j]>.
 */

	static
	void
compute_mu (

struct LLLInfo *	lip,		/* IN/OUT: LLL algorithm info */
mpq_ptr			mu_out,		/* IN/OUT: Computed mu value */
int			i,		/* IN: Index i */
int			j		/* IN: Index j */
)
{
int		k, nvec, vsize;
mpz_srcptr	B, szp;
mpq_srcptr	Bstar, Bsmag2, sqp;
int *		order;
mpq_t		SUM, TERM;

	CALIB_FATAL_ERROR_IF ((lip EQ NULL) OR (mu_out EQ NULL));

	nvec	= lip -> li_nvec;
	vsize	= lip -> li_vsize;
	B	= lip -> li_B;
	order	= lip -> li_order;
	Bstar	= lip -> li_Bstar;
	Bsmag2	= lip -> li_Bsmag2;

	CALIB_FATAL_ERROR_IF ((i < 0) OR (i >= nvec));
	CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= nvec));

	/* Assure that Gram-Schmidt vectors and mag**2 are correct	*/
	/* up to index j.						*/
	update_bstar (lip, j);

	mpq_init (SUM);
	mpq_init (TERM);

	szp	= &B [order [i] * vsize];
	sqp	= &Bstar [j * vsize];
	for (k = 0; k < vsize; k++) {
		mpz_set (mpq_numref (TERM), &szp [k]);
		mpz_set_ui (mpq_denref (TERM), 1);
		mpq_mul (TERM, TERM, &sqp [k]);
		mpq_add (SUM, SUM, TERM);
	}

	sqp = &Bsmag2 [j];
	CALIB_FATAL_ERROR_IF (mpq_sgn (sqp) EQ 0);

	mpq_div (mu_out, SUM, sqp);

	mpq_clear (TERM);
	mpq_clear (SUM);
}

/*
 * Assure that all Bstar[j] and Bsmag2[j] are correct for all
 * 0 <= j <= i.  If not, recompute them starting from the lowest
 * invalid j on up to i.
 */

	static
	void
update_bstar (

struct LLLInfo *	lip,		/* IN/OUT: LLL algorithm info */
int			i		/* IN: Index i requested */
)
{
int		j, m, n, nvec, vsize, sign;
mpz_srcptr	B, szp;
int *		order;
mpq_ptr		Bstar, Bsmag2, dqp;
mpq_srcptr	sqp;
mpq_t		SUM, TERM, RATIO;

	CALIB_FATAL_ERROR_IF (lip EQ NULL);

	nvec	= lip -> li_nvec;

	CALIB_FATAL_ERROR_IF ((i < 0) OR (i >= nvec));

	m = lip -> li_bs_valid;
	if (i <= m) {
		/* Bstar and Bsmag2 are valid up to i! */
		return;
	}

	vsize	= lip -> li_vsize;
	B	= lip -> li_B;
	order	= lip -> li_order;
	Bstar	= lip -> li_Bstar;
	Bsmag2	= lip -> li_Bsmag2;

	mpq_init (SUM);
	mpq_init (TERM);
	mpq_init (RATIO);

	while (m < i) {
		++m;

		CALIB_FATAL_ERROR_IF ((m < 0) OR (m >= nvec));

		/* Compute Bstar[m].  Start with Bstar[m] = B[order[m]]. */
		szp = &B [order [m] * vsize];
		dqp = &Bstar [m * vsize];
		for (j = 0; j < vsize; j++) {
			mpz_set (mpq_numref (&dqp [j]), &szp [j]);
			mpz_set_ui (mpq_denref (&dqp [j]), 1);
		}

		/* Eliminate components of previous Bstar. */
		for (n = 0; n < m; n++) {
			sign = mpq_sgn (&Bsmag2 [n]);
			if (sign <= 0) {
				CALIB_FATAL_ERROR_IF (sign < 0);
				/* Bstar[n] is zero.  Subtracting any	*/
				/* multiple of this does nothing.	*/
				continue;
			}
			mpq_set_ui (SUM, 0, 1);
			sqp = &Bstar [n * vsize];
			for (j = 0; j < vsize; j++) {
				mpz_set (mpq_numref (TERM), &szp [j]);
				mpz_set_ui (mpq_denref (TERM), 1);
				mpq_mul (TERM, TERM, &sqp [j]);
				mpq_add (SUM, SUM, TERM);
			}
			mpq_div (RATIO, SUM, &Bsmag2 [n]);
			for (j = 0; j < vsize; j++) {
				mpq_mul (TERM, RATIO, &sqp [j]);
				mpq_sub (&dqp [j], &dqp [j], TERM);
			}
		}

		/* Compute Bsmag2[m]. */
		mpq_set_ui (SUM, 0, 1);
		for (j = 0; j < vsize; j++) {
			mpq_mul (TERM, &dqp [j], &dqp [j]);
			mpq_add (SUM, SUM, TERM);
		}
		mpq_set (&Bsmag2 [m], SUM);
	}

	lip -> li_bs_valid = i;

	mpq_clear (RATIO);
	mpq_clear (TERM);
	mpq_clear (SUM);
}

/*
 * Reduce the lattice basis BLambda using the Lenstra-Lenstra-Lovasz
 * lattice basis reduction algorithm.
 *
 * *THIS CODE IS NOT CORRECT!!!*
 */

	static
	int
reduce_lattice_basis_LLL_old (

mpz_ptr			B,		/* IN/OUT: Lattice-basis to reduce */
int			nvec,		/* IN: Number of vectors */
int			vsize,		/* IN: Size of each vector */
mpq_srcptr		Delta,		/* IN: The parameter Delta */
int *			order		/* IN/OUT: final order of vectors */
)
{
int			i, k;
calib_bool		any_reductions_made;
calib_bool		any_swaps_made;
mpq_ptr			Bstar, Bsmag2, sqp;
mpz_ptr			Btemp;

	Bstar = Gram_Schmidt (B, nvec, vsize);

	/* Compute Bsmag2[i] = <Bstar_i,Bstar_i>. */
	Bsmag2	= calib_new_Q_vector (nvec);
	for (i = 0; i < nvec; i++, sqp += vsize) {
		mpq_mag2 (&Bsmag2 [i], &Bstar [i * vsize], vsize);
	}

#if 1
	verify_GS_dot_B_mags (B, Bstar, Bsmag2, nvec, vsize);
#endif

	/* Set initial ordering of B vectors. */
	for (i = 0; i < nvec; i++) {
		order [i] = i;
	}

	k	= nvec * vsize;
	Btemp	= calib_new_Z_vector (k);

	for (;;) {
		any_reductions_made =
			reduction_phase (B,
					 order,
					 Bstar,
					 Bsmag2,
					 nvec,
					 vsize);

		any_swaps_made =
			swap_phase (B,
					      order,
					      Bstar,
					      Bsmag2,
					      Delta,
					      nvec,
					      vsize);

		if (any_reductions_made OR any_swaps_made) {
			continue;
		}

		for (i = 0; i < nvec; i++) {
			calib_copy_Z_vector (&Btemp [i * vsize],
					      &B [order [i] * vsize],
					      vsize);
		}
		for (i = 0; i < nvec; i++) {
			calib_copy_Z_vector (&B [i * vsize],
				       &Btemp [i * vsize],
				       vsize);
			order [i] = i;
		}

		if (test_delta_reduced (B, nvec, vsize, Delta)) {
			break;
		}

		/* Recompute Gram-Schmidt vectors. */
		k = nvec * vsize;
		calib_free_Q_vector (Bstar, k);
		Bstar = Gram_Schmidt (Btemp, nvec, vsize);

		/* Recompute Bsmag2. */
		for (i = 0; i < nvec; i++) {
			mpq_mag2 (&Bsmag2 [i], &Bstar [i * vsize], vsize);
		}

	}

	free (order);

	calib_free_Q_vector (Bsmag2, nvec);

	k = nvec * vsize;
	calib_free_Q_vector (Bstar, k);

	return (0);
}

/*
 * The reduction phase of the LLL algorithm.
 *
 * Returns TRUE if-and-only-if any reductions were peformed.
 */

	static
	calib_bool
reduction_phase (

mpz_ptr			B,		/* IN/OUT: LLL vectors to reduce */
const int *		order,		/* IN: Ordering of vectors */
mpq_srcptr		Bstar,		/* IN: Gram-Schmidt vectors */
mpq_srcptr		Bsmag2,		/* IN: Mag**2 of GS vectors */
int			nvec,		/* IN: Number of vectors */
int			vsize		/* IN: Size of each vector */
)
{
int			i, j, k;
calib_bool		any_reductions_made;
mpz_srcptr		szp;
mpq_srcptr		sqp;
mpz_ptr			dzp;
mpz_t			ZTERM, FACTOR;
mpq_t			SUM, TERM, RATIO, HALF;

	mpz_init (ZTERM);
	mpz_init (FACTOR);

	mpq_init (SUM);
	mpq_init (TERM);
	mpq_init (RATIO);
	mpq_init (HALF);

	mpq_set_ui (HALF, 1, 2);

	any_reductions_made = FALSE;
	for (i = 1; i < nvec; i++) {
		dzp = &B [order [i] * vsize];
		for (j = i - 1; j >= 0; j--) {
			/* Compute <B_i,Bstar_j> */
			mpq_set_ui (SUM, 0, 1);
			szp = &B [order [i] * vsize];
			sqp = &Bstar [order [j] * vsize];
			for (k = 0; k < vsize; k++) {
				mpz_set (mpq_numref (TERM), &szp [k]);
				mpz_set_ui (mpq_denref (TERM), 1);
				mpq_mul (TERM, TERM, &sqp [k]);
				mpq_add (SUM, SUM, TERM);
			}
			sqp = &Bsmag2 [order [j]];
			k = mpq_sgn (sqp);
			if (k <= 0) {
				/* We have bstar_j EQ 0. */
				CALIB_FATAL_ERROR_IF (k < 0);
				continue;
			}
			mpq_div (RATIO, SUM, sqp);
			mpq_add (RATIO, RATIO, HALF);
			mpz_fdiv_q (FACTOR,
				    mpq_numref (RATIO),
				    mpq_denref (RATIO));
			if (mpz_sgn (FACTOR) NE 0) {
#if 0
				mpz_t	MAG2before, MAG2after;
				mpq_t	RATIO;
				mpz_init (MAG2before);
				mpz_init (MAG2after);
				mpq_init (RATIO);
				mpz_mag2 (MAG2before,
					  &B [order [i] * vsize],
					  vsize);
#endif
				szp = &B [order [j] * vsize];
				for (k = 0; k < vsize; k++) {
					mpz_mul (ZTERM, FACTOR, &szp [k]);
					mpz_sub (&dzp [k], &dzp [k], ZTERM);
				}
#if 0
				mpz_mag2 (MAG2after,
					  &B [order [i] * vsize],
					  vsize);
				mpz_set (mpq_numref (RATIO), MAG2after);
				mpz_set (mpq_denref (RATIO), MAG2before);
				mpq_canonicalize (RATIO);
				gmp_printf ("b_%d -= %Zd * b_%d"
					    "\t(%Zd ==> %Zd, %.17g)\n",
					    i, FACTOR, j,
					    MAG2before, MAG2after,
					    mpq_get_d (RATIO));
				mpq_clear (RATIO);
				mpz_clear (MAG2after);
				mpz_clear (MAG2before);
#endif
				any_reductions_made = TRUE;
			}
		}
	}

#if 1
	/* Test invariant of reduction step:			*/
	/*							*/
	/*	|\mu_{i,j}| <= 1/2	for all i > j.		*/
	for (i = 1; i < nvec; i++) {
		for (j = 0; j < i; j++) {
			/* Compute <B_i,Bstar_j> */
			mpq_set_ui (SUM, 0, 1);
			szp = &B [order [i] * vsize];
			sqp = &Bstar [order [j] * vsize];
			for (k = 0; k < vsize; k++) {
				mpz_set (mpq_numref (TERM), &szp [k]);
				mpz_set_ui (mpq_denref (TERM), 1);
				mpq_mul (TERM, TERM, &sqp [k]);
				mpq_add (SUM, SUM, TERM);
			}
			sqp = &Bsmag2 [order [j]];
			k = mpq_sgn (sqp);
			if (k <= 0) {
				CALIB_FATAL_ERROR_IF (k < 0);
				continue;
			}
			mpq_div (RATIO, SUM, sqp);
			/* We should have |RATIO| <= 1/2. */
			mpq_abs (RATIO, RATIO);
			if (mpq_cmp (RATIO, HALF) > 0) {
				printf ("Have bad mu[%d,%d] > 1/2"
					" (%.17g)\n",
					i, j, mpq_get_d (RATIO));
			}
		}
	}
#endif

#if 1
	verify_GS_dot_B_mags (B, Bstar, Bsmag2, nvec, vsize);
#endif

	return (any_reductions_made);
}

/*
 * The swap phase of the LLL algorithm.
 *
 * Returns TRUE if-and-only-if any swaps were performed.
 */

	static
	calib_bool
swap_phase (

mpz_srcptr		B,		/* IN: LLL vectors */
int *			order,		/* IN/OUT: Ordering of vectors */
mpq_srcptr		Bstar,		/* IN: Gram-Schmidt vectors */
mpq_srcptr		Bsmag2,		/* IN: Mag**2 of GS vectors */
mpq_srcptr		Delta,		/* IN: The parameter Delta */
int			nvec,		/* IN: Number of vectors */
int			vsize		/* IN: Size of each vector */
)
{
int			i, j, k, cmp, sign;
calib_bool		any_swaps_made;
mpz_srcptr		szp;
mpq_srcptr		sqp;
mpq_t			SUM, TERM, HALF;

	mpq_init (SUM);
	mpq_init (TERM);
	mpq_init (HALF);

	mpq_set_ui (HALF, 1, 2);

	any_swaps_made = FALSE;
	for (j = 1; j < nvec; j++) {
		for (i = j; i > 0; i--) {
			/* Check for zero vector. */
			sqp = &Bsmag2 [order [i - 1]];
			sign = mpq_sgn (sqp);
			if (sign <= 0) {
				/* Stop at zero vectors, which we keep	*/
				/* at the top.				*/
				CALIB_FATAL_ERROR_IF (sign < 0);
				break;
			}
			/* Compute					     */
			/*	mu_i,{i-1} = <B_i,Bstar_{i-1}>/Bsmag2_{i-1}. */
			mpq_set_ui (SUM, 0, 1);
			szp = &B [order [i] * vsize];
			sqp = &Bstar [order [i - 1] * vsize];
			for (k = 0; k < vsize; k++) {
				mpz_set (mpq_numref (TERM), &szp [k]);
				mpz_set_ui (mpq_denref (TERM), 1);
				mpq_mul (TERM, TERM, &sqp [k]);
				mpq_add (SUM, SUM, TERM);
			}
			/* Test the Lovasz condition. */
			sqp = &Bsmag2 [order [i - 1]];
			mpq_div (SUM, SUM, sqp);
			mpq_mul (SUM, SUM, SUM);
			mpq_sub (SUM, Delta, SUM);
			mpq_mul (SUM, SUM, &Bsmag2 [order [i - 1]]);
			cmp = mpq_cmp (SUM, &Bsmag2 [order [i]]);
			if (cmp <= 0) {
				/* Lovasz condition satisfied.		*/
				/* i-th vector has been swapped up as	*/
				/* far as it needs to go.		*/
				break;
			}
#if 1
			printf ("Swap %d <--> %d\t(%d <--> %d)\n",
				i - 1,  i, order [i - 1], order [i]);
#endif
			k = order [i - 1];
			order [i - 1] = order [i];
			order [i] = k;
			any_swaps_made = TRUE;
#if 1
			goto done;
#endif
		}
	}

done:

	mpq_clear (HALF);
	mpq_clear (TERM);
	mpq_clear (SUM);

	return (any_swaps_made);
}

/*
 * Verify that <B_i,Bstar_i> EQ <Bstar_i,Bstar_i> for all i.
 */

	static
	void
verify_GS_dot_B_mags (

mpz_srcptr		B,		/* IN: LLL vectors */
mpq_srcptr		Bstar,		/* IN: Gram-Schmidt vectors */
mpq_srcptr		Bsmag2,		/* IN: Mag**2 of GS vectors */
int			nvec,		/* IN: Number of vectors */
int			vsize		/* IN: Size of each vector */
)
{
int		i, k;
mpq_srcptr	sqp;
mpz_srcptr	szp;
mpq_t		SUM, TERM;

	mpq_init (SUM);
	mpq_init (TERM);

	sqp = Bstar;
	szp = B;
	for (i = 0; i < nvec; i++, sqp += vsize, szp += vsize) {
		mpq_set_ui (SUM, 0, 1);
		for (k = 0; k < vsize; k++) {
			mpz_set (mpq_numref (TERM), &szp [k]);
			mpz_set_ui (mpq_denref (TERM), 1);
			mpq_mul (TERM, TERM, &sqp [k]);
			mpq_add (SUM, SUM, TERM);
		}
		if (mpq_cmp (SUM, &Bsmag2 [i]) NE 0) {
			printf ("We have <B_i,Bstar_i> NE Bsmag2_i for i = %d\n", i);
			CALIB_FATAL_ERROR;
		}
	}

	mpq_clear (TERM);
	mpq_clear (SUM);
}

/*
 * Test the given lattice basis to see if it is Delta-LLL-reduced.
 */

	static
	calib_bool
test_delta_reduced (

mpz_srcptr		B,	/* IN: Basis to test */
int			nvec,	/* IN: Number of vectors */
int			vsize,	/* IN: Size of each vector */
mpq_srcptr		Delta	/* IN: The parameter Delta */
)
{
int		i, j, k;
calib_bool	valid;
mpz_srcptr	szp;
mpq_srcptr	sqp;
mpq_ptr		Bstar, Bsmag2;
mpq_t		SUM, TERM, RATIO, HALF;

	Bstar = Gram_Schmidt (B, nvec, vsize);

	Bsmag2 = calib_new_Q_vector (nvec);
	for (i = 0; i < nvec; i++) {
		mpq_mag2 (&Bsmag2 [i], &Bstar [i * vsize], vsize);
	}

	mpq_init (SUM);
	mpq_init (TERM);
	mpq_init (RATIO);
	mpq_init (HALF);

	mpq_set_ui (HALF, 1, 2);

	valid = TRUE;

	/* Property 1: |\mu_{i+1,i}| <= 1/2 for all j < i. */
	for (i = 1; i < nvec; i++) {
		for (j = 0; j < i; j++) {
			/* Compute <B_i,Bstar_j> */
			mpq_set_ui (SUM, 0, 1);
			szp = &B [i * vsize];
			sqp = &Bstar [j * vsize];
			for (k = 0; k < vsize; k++) {
				mpz_set (mpq_numref (TERM), &szp [k]);
				mpz_set_ui (mpq_denref (TERM), 1);
				mpq_mul (TERM, TERM, &sqp [k]);
				mpq_add (SUM, SUM, TERM);
			}
			sqp = &Bsmag2 [j];
			k = mpq_sgn (sqp);
			if (k <= 0) {
				CALIB_FATAL_ERROR_IF (k < 0);
				printf ("Have Bsmag2[%d] EQ 0!\n", j);
				continue;
			}
			mpq_div (RATIO, SUM, sqp);
			/* We should have |RATIO| <= 1/2. */
			mpq_abs (RATIO, RATIO);
			if (mpq_cmp (RATIO, HALF) > 0) {
				printf ("Have bad mu[%d,%d] > 1/2"
					" (%.17g)\n",
					i, j, mpq_get_d (RATIO));
				valid = FALSE;
			}
		}
	}

	/* Property 2: Lovasz condition. */
	for (i = 1; i < nvec; i++) {
		/* Compute <B_i,Bstar_{i-1}>. */
		mpq_set_ui (SUM, 0, 1);
		szp = &B [i * vsize];
		sqp = &Bstar [(i - 1) * vsize];
		for (k = 0; k < vsize; k++) {
			mpz_set (mpq_numref (TERM), &szp [k]);
			mpz_set_ui (mpq_denref (TERM), 1);
			mpq_mul (TERM, TERM, &sqp [k]);
			mpq_add (SUM, SUM, TERM);
		}
		/* Test Lovasz condition. */
		mpq_mul (SUM, SUM, SUM);
		mpq_sub (SUM, Delta, SUM);
		mpq_mul (SUM, SUM, &Bsmag2 [i - 1]);
		if (mpq_cmp (&Bsmag2 [i], SUM) < 0) {
			printf ("Lovasz condition fails at mu[%d,%d]\n",
				i, i - 1);
			valid = FALSE;
		}
	}

	mpq_clear (HALF);
	mpq_clear (RATIO);
	mpq_clear (TERM);
	mpq_clear (SUM);

	calib_free_Q_vector (Bsmag2, nvec);
	calib_free_Q_vector (Bstar, nvec * vsize);

	return (valid);
}

/*
 * Compute the magnitude squared of the given rational vector.
 * Return the result as a double.
 */

	static
	double
mpq_mag2_double (

mpq_srcptr	v,		/* IN: Rational vector */
int		n		/* IN: Length of rational vector */
)
{
mpq_t		MAG2;
double		result;

	mpq_init (MAG2);

	mpq_mag2 (MAG2, v, n);

	result = mpq_get_d (MAG2);

	mpq_clear (MAG2);

	return (result);
}

/*
 * Compute the magnitude squared of the given rational vector.
 */

	static
	void
mpq_mag2 (

mpq_ptr		result,		/* IN/OUT: Resulting magnitude squared */
mpq_srcptr	v,		/* IN: Rational vector */
int		n		/* IN: Length of rational vector */
)
{
int		i;
mpq_t		SUM, TERM;

	mpq_init (SUM);
	mpq_init (TERM);

	for (i = 0; i < n; i++) {
		mpq_mul (TERM, &v [i], &v [i]);
		mpq_add (SUM, SUM, TERM);
	}

	mpq_set (result, SUM);

	mpq_clear (TERM);
	mpq_clear (SUM);
}

/*
 * Compute the magnitude squared of the given integer vector.
 */

	static
	void
mpz_mag2 (

mpz_ptr		result,		/* IN/OUT: Resulting magnitude squared */
mpz_srcptr	v,		/* IN: Integer vector */
int		n		/* IN: Length of integer vector */
)
{
int		i;
mpz_t		SUM, TERM;

	mpz_init (SUM);
	mpz_init (TERM);

	for (i = 0; i < n; i++) {
		mpz_mul (TERM, &v [i], &v [i]);
		mpz_add (SUM, SUM, TERM);
	}

	mpz_set (result, SUM);

	mpz_clear (TERM);
	mpz_clear (SUM);
}

/*
 * Compute the exact rational Gram-Schmidt orthogonalization of the
 * given integer vectors.  It is the caller's responsibility to free
 * the returned matrix.
 */

	static
	mpq_ptr
Gram_Schmidt (

mpz_srcptr		B,		/* IN: Integral vectors to	*/
					/*     orthogonalize		*/
int			nvec,		/* IN: Number of vectors	*/
int			vsize		/* IN: Size of each vector	*/
)
{
int		i, j, k, n2;
mpz_srcptr	zp;
mpq_srcptr	sqp;
mpq_ptr		Bstar, dqp;
mpq_t		TERM, SUM1, SUM2, RATIO;

	n2 = nvec * vsize;

	/* Convert integer vectors into rational form. */
	Bstar = calib_new_Q_vector (n2);
	for (i = 0; i < n2; i++) {
		mpz_set (mpq_numref (&Bstar [i]), &B [i]);
		mpz_set_ui (mpq_denref (&Bstar [i]), 1);
	}

	mpq_init (TERM);
	mpq_init (SUM1);
	mpq_init (SUM2);
	mpq_init (RATIO);

	/* Bstar_0 = B_0. */
	for (i = 1; i < nvec; i++) {
		zp  = &B [i * vsize];
		dqp = &Bstar [i * vsize];
		for (j = 0; j < i; j++) {
			/* Bstar [i] -= (<B_j,Bstar_j>/<Bstar_j,Bstar_j>)*B_j */
			mpq_set_ui (SUM1, 0, 1);
			mpq_set_ui (SUM2, 0, 1);
			sqp = &Bstar [j * vsize];
			for (k = 0; k < vsize; k++) {
				mpz_set (mpq_numref (TERM), &zp [k]);
				mpz_set_ui (mpq_denref (TERM), 1);
				mpq_mul (TERM, TERM, &sqp [k]);
				mpq_add (SUM1, SUM1, TERM);

				mpq_mul (TERM, &sqp [k], &sqp [k]);
				mpq_add (SUM2, SUM2, TERM);
			}
			k = mpq_sgn (SUM2);
			if (k <= 0) {
				CALIB_FATAL_ERROR_IF (k < 0);
				/* Bstar_j is zero.  Subtracting any	*/
				/* multiple of this does nothing.	*/
				continue;
			}
			mpq_div (RATIO, SUM1, SUM2);
			for (k = 0; k < vsize; k++) {
				mpq_mul (TERM, RATIO, &sqp [k]);
				mpq_sub (&dqp [k], &dqp [k], TERM);
			}
		}
	}

#if 0
	debug_print_Gram_Schmidt (Bstar, nvec, vsize);
#endif

	mpq_clear (RATIO);
	mpq_clear (SUM2);
	mpq_clear (SUM1);
	mpq_clear (TERM);

	return (Bstar);
}

/*
 * Print debugging information about the given Gram-Schmidt vectors.
 */

	static
	void
debug_print_Gram_Schmidt (

mpq_srcptr	Bstar,		/* IN: Gram-Schmit orthogonal vectors */
int		nvec,		/* IN: Number of vectors */
int		vsize		/* IN: Size of each vector */
)
{
int		i, j, k;
size_t		nbits;
mpq_srcptr	qip, qjp;
mpq_t		SUM, TERM;

	for (i = 0; i < nvec; i++) {
		nbits	= calib_mpq_bit_size_vector (&Bstar [i * vsize], vsize);
		printf ("%d:\t%lu\n", i, nbits);
	}

	mpq_init (SUM);
	mpq_init (TERM);

	printf ("\nDot Products:\n");
	for (i = 0; i < nvec; i++) {
		qip = &Bstar [i * vsize];
		printf ("\n");
		for (j = 0; j < nvec; j++) {
			qjp = &Bstar [j * vsize];
			mpq_set_ui (SUM, 0, 1);
			for (k = 0; k < vsize; k++) {
				mpq_mul (TERM, &qip [k], &qjp [k]);
				mpq_add (SUM, SUM, TERM);
			}
			nbits = calib_mpq_bit_size (SUM);
			printf ("%d\t%d\t%d\t%.17g\t%lu\n",
				 i,
				 j,
				 mpq_sgn (SUM),
				 mpq_get_d (SUM),
				 nbits);
		}
	}

	mpq_clear (TERM);
	mpq_clear (SUM);
}
