/**********************************************************************

	$Id: modcleanup.h,v 1.1 2024/10/20 00:36:28 warme Exp $

	File:	modcleanup.h
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Internal interfaces to cleanup the algebra library when done.

***********************************************************************

	Modification Log:

	a-1:	10/16/2015	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#ifndef CALIB_MODCLEANUP_H
#define CALIB_MODCLEANUP_H

/*
 * A mechanism for cleaning up all of the "global static" data structures
 * used by this library.
 *
 * Note that there are quite a few different modules within this library,
 * and some of these maintain data structures (e.g., freelists for
 * objects of various types) that would otherwise remain in use upon
 * termination of the program.  This can cause huge problems for tools
 * such as valgrind.
 *
 * One possible solution would be for this library to provide a single
 * "cleanup" function, and for this function to simply call the "cleanup"
 * function of each of the library's internal modules that might need
 * to perform cleanup.  This implementation suffers from the problem
 * that calling this function forces the entire library to become
 * referenced, even though only a few modules might actually be used
 * by the application.
 *
 * The solution is for individual modules to be able to dynamically
 * add "cleanup" handlers to a list.  (They carefully do this only
 * once for each handler.)  The single, application-visible function
 * to do cleanup then simply calls each cleanup function in the list.
 */

/* Include the application-visible shutdown interfaces. */
#include "calib/shutdown.h"

/*
 * Type signature of cleanup functions.
 */

typedef void calib_module_cleanup_func_t (void);

/*
 * Internal function (not visible to applications) by which a module
 * adds a cleanup handler to the list.
 */

extern void	_calib_add_cleanup_handler (
				calib_module_cleanup_func_t *	handler);

#endif
