/**********************************************************************

	$Id: oldprint.c,v 1.1 2024/10/20 00:36:28 warme Exp $

	File:	oldprint.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	General representation for expressions.

***********************************************************************

	Modification Log:

	a-1:	10/10/2019	warme
		: Split off from genrep.c
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/genrep.h"

#include "calib/fatal.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "genrep_int.h"
#include <gmp.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*
 * Global Functions.
 */

size_t	calib_genrep_convert_abs_Z_to_decimal_string (
			mpz_srcptr	zval,
			char **		str_out);
void	calib_genrep_fprint (FILE *				fp,
			     const struct calib_genrep *	op);
void	calib_genrep_fwprint (FILE *				fp,
			      const struct calib_genrep *	op,
			      int				width);
void	calib_genrep_print (const struct calib_genrep * op);
void	calib_genrep_wprint (const struct calib_genrep *	op,
			     int				width);

/*
 * Local Macros
 */

#define _MAX(A,B)	(((A) >= (B)) ? (A) : (B))

#define DEFAULT_MAX_WIDTH	79

#define	LPAREN		"("
#define	RPAREN		")"

/*
 * An object to recursive represent the physical dimensions of an
 * expression (in printed form), and all of its sub-expressions.
 */

struct grdim {
	char		op;		/* Operator */
	char		sign;		/* Sign for sub-expr */
	char		leaf;		/* TRUE if a leaf node */
	char		parens;		/* expr must be parenthesized */
	char		multiline;	/* TRUE if expression formatted */
					/* across multiple lines */
	int		width;
	int		height;
	int		ascent;		/* ascent above centerline */
	int		descent;	/* descent below centerline */
	union {
		struct grdim_list *	list;
		struct grdim *		unop;
		struct {
			struct grdim *	numer;
			struct grdim *	denom;
		} frac;
		struct {
			struct grdim *	op1;
			int		op2;
		} pow;
		struct {
			char *			fname;
			struct grdim_list *	args;
		} func;
		char *			zstr;
		const char *		var;
	} u;
};

struct grdim_list {
	struct grdim *		operand;
	struct grdim_list *	next;
};

#define	GRDIM_OP_ADD		0
#define GRDIM_OP_UPLUS		1
#define	GRDIM_OP_NEG		2
#define GRDIM_OP_MUL		3
#define GRDIM_OP_FRAC		4
#define GRDIM_OP_POW		5
#define GRDIM_OP_VAR		6
#define GRDIM_OP_Z		7
#define	GRDIM_OP_FUNC		8
#define GRDIM_OP_PARTIAL_FUNC	9

/*
 * This object represents a dynamically expandable "canvas" that we can
 * plot onto using ASCII text.
 */

struct canvas {
	int		nlines;
	struct cline *	lines;
};

struct cline {
	int		len;		/* Length of text on this line. */
	char *		text;		/* Text for this line.		*/
};

/*
 * A "rubber band" buffer for accumulating text.
 */

#define RUBBER_SIZE	512

struct rbblk {
	struct rbblk *	next;			/* Next block in linked list */
	char		buf [RUBBER_SIZE];	/* Buffer for characters */
};

typedef struct {
	char *		ptr;		/* Append next character here */
	int		cnt;		/* Num chars left in current buffer */
	size_t		nblks;		/* Number of blocks *after* first */
	struct rbblk *	last;		/* Last block in list */
	struct rbblk	first;		/* First block is embedded */
} RubberBuf;

/*
 * Local Functions.
 */

static void		clear_rubber_buf (RubberBuf * rbp);
static struct grdim *	expr_dims (const struct calib_genrep * op);
static struct grdim *	expr_dim_func (const struct calib_genrep * op);
static struct grdim *	expr_dim_gather_prod (mpz_srcptr	  zval,
					      struct grdim_list * factors);
static struct grdim *	expr_dim_pow (const struct calib_genrep * op);
static struct grdim *	expr_dim_prod (const struct calib_genrep_list * factors);
static struct grdim *	expr_dim_sum (const struct calib_genrep_list * factors);
static void		free_canvas (struct canvas * cp);
static void		free_grdim_split (struct grdim_list * dlist);
static void		grdim_free (struct grdim * dp);
static int		grdim_plot_indented (struct canvas * cp,
					     struct grdim *	 dp,
					     int		 x,
					     int		 max_x);
static int		grdim_print_indented (FILE *	 	fp,
					      struct grdim *	dp,
					      int		x,
					      int		max_x);
static struct grdim_list * grdim_split (struct grdim * dp, int width);
static struct grdim_list * grdim_split_func (struct grdim * op, int width);
static struct grdim_list * grdim_split_list (struct grdim_list * dlist,
					     int		 width,
					     int		 operator);
static struct grdim *	grdim_strip_signs (struct grdim * dp, mpz_ptr numfact);
static struct grdim *	grdim_unneg (struct grdim * op);
static struct grdim *	grdim_unplus (struct grdim * op);
static void		init_canvas (struct canvas * cp, int nlines);
static void		init_rubber_buf (RubberBuf * rbp);
static struct grdim *	make_grdim_frac (struct grdim *	numer,
					 struct grdim *	denom);
static struct grdim *	make_grdim_int (calib_si_t ival);
static struct grdim *	make_grdim_uplus (struct grdim * op);
static struct grdim *	make_grdim_neg (struct grdim * op);
static struct grdim *	make_grdim_pow (struct grdim * base, int pow);
static struct grdim *	make_grdim_prod (struct grdim_list * list);
static struct grdim *	make_grdim_q (mpq_srcptr qval);
static struct grdim *	make_grdim_var (const char * var);
static struct grdim *	make_grdim_z (mpz_srcptr zval);
static struct grdim *	new_grdim (void);
static struct grdim_list * new_grdim_list (struct grdim *	op,
					   struct grdim_list *	next);
static void		plot_canvas (struct canvas *	cp,
				     const char *	text,
				     int		x,
				     int		y);
static void		plot_canvas_z (struct canvas *	cp,
				       mpz_srcptr	zval,
				       int		x,
				       int		y);
static void		plot_grdim (struct canvas *	cp,
				    struct grdim *	dp,
				    int			x,
				    int			y);
static void		rubber_buf_putc (RubberBuf * rbp, char c);
static void		write_canvas (struct canvas * cp, FILE * fp);

/*
 * Display the given expression (expressed in general representation)
 * to stdout using the default maximum width.
 */

	void
calib_genrep_print (

const struct calib_genrep *	op	/* IN: expression to display */
)
{
	calib_genrep_fwprint (stdout, op, DEFAULT_MAX_WIDTH);
}

/*
 * Display the given expression (expressed in general representation)
 * to the given stream using the default maximum width.
 */

	void
calib_genrep_fprint (

FILE *				fp,	/* IN/OUT: stream to write to */
const struct calib_genrep *	op	/* IN: expression to display */
)
{
	calib_genrep_fwprint (fp, op, DEFAULT_MAX_WIDTH);
}

/*
 * Display the given expression (expressed in general representation)
 * to stdout using the given maximum number of columns.
 */

	void
calib_genrep_wprint (

const struct calib_genrep *	op,	/* IN: expression to display */
int				width	/* IN: maximum width of display */
)
{
	calib_genrep_fwprint (stdout, op, width);
}

/*
 * Display the given expression (expressed in general representation)
 * to the given stream using the given maximum number of columns.
 */

	void
calib_genrep_fwprint (

FILE *				fp,	/* IN/OUT: stream to write to */
const struct calib_genrep *	op,	/* IN: expression to display */
int				width	/* IN: maximum width of display */
)
{
int			x, y, dx, success, operator;
const char *		delim;
struct calib_genrep *	op2;
struct grdim *		dp;
struct grdim *		dp2;
struct grdim_list *	dlist;
struct grdim_list *	dlp;
struct canvas		canvas;

	/* Copy the given general expression, performing several simple	*/
	/* but important simplifications.				*/
	op2 = _calib_genrep_copy_simp (op);

	/* Get recursive sub-expression dimensions, assuming infinite width. */
	dp = expr_dims (op2);

	calib_genrep_free (op2);

	/* Remove leading unary plus. */
	dp = grdim_unplus (dp);

	if (dp -> width < width) {
		x = (width - (dp -> width) + 1) / 2;
		y = dp -> ascent;
		init_canvas (&canvas, dp -> height);
		plot_grdim (&canvas, dp, x, y);
		write_canvas (&canvas, fp);
		free_canvas (&canvas);
		grdim_free (dp);
		return;
	}

	/* Expression is too wide to print on one line.  Let's try	*/
	/* splitting it up across several.				*/

	x = 3;
	dlist = grdim_split (dp, width - x);
	if (dlist NE NULL) {
		dp2 = dlist -> operand;
		operator = dp2 -> op;

		/* Should be more than one line. */
		CALIB_FATAL_ERROR_IF (dlist -> next EQ NULL);

		delim = ",";

		y = dp2 -> ascent;
		init_canvas (&canvas, dp2 -> height);
		plot_grdim (&canvas, dp2, x, y);
		if (operator EQ GRDIM_OP_PARTIAL_FUNC) {
			/* Put out arg delimiter at end of line. */
			dx = dp2 -> width - 1;
			plot_canvas (&canvas, delim, x + dx, y);
		}
		write_canvas (&canvas, fp);
		free_canvas (&canvas);

		for (dlp = dlist -> next; dlp NE NULL; dlp = dlp -> next) {
			fprintf (fp, "\n");
			if (dlp -> next EQ NULL) {
				/* Last arg ends with right paren. */
				delim = RPAREN;
			}

			dp2 = dlp -> operand;
			y = dp2 -> ascent;
			init_canvas (&canvas, dp2 -> height);

			plot_grdim (&canvas, dp2, x, y);

			if (operator EQ GRDIM_OP_PARTIAL_FUNC) {
				/* Put out arg delimiter at end of line. */
				dx = dp2 -> width - 1;
				plot_canvas (&canvas, delim, x + dx, y);
			}
			write_canvas (&canvas, fp);
			free_canvas (&canvas);
		}
		free_grdim_split (dlist);
	}
	else {
		x = 3,
		success = grdim_print_indented (fp, dp, x, width - x);
		if (NOT success) {
			fprintf (fp, "Expression too wide to print.\n");
		}
	}

	grdim_free (dp);
}

/*
 * Attempt to split the given grdim sub-expression into multiple lines.
 * Returns NULL upon failure.
 */

	static
	struct grdim_list *
grdim_split (

struct grdim *	dp,		/* IN: grdim to split */
int		width		/* IN: max width of each line */
)
{
struct grdim_list *	dlp;

	dlp = NULL;

	switch (dp -> op) {
	case GRDIM_OP_ADD:
		/* Break into several sums. */
		dlp = grdim_split_list (dp -> u.list, width, GRDIM_OP_ADD);
		break;

	case GRDIM_OP_MUL:
		/* Break into several products. */
		dlp = grdim_split_list (dp -> u.list, width, GRDIM_OP_MUL);
		break;

	case GRDIM_OP_UPLUS:
	case GRDIM_OP_NEG:
	case GRDIM_OP_FRAC:
	case GRDIM_OP_POW:
	case GRDIM_OP_VAR:
	case GRDIM_OP_Z:
		break;

	case GRDIM_OP_FUNC:
		/* Break up arguments. */
		dlp = grdim_split_func (dp, width);
		break;

	default:
		CALIB_FATAL_ERROR;
	}

	return (dlp);
}

/*
 * Attempt to split the given list of terms/factors into multiple lines,
 * one sum-of-terms / product-of-factors per line.
 * Returns NULL upon failure.
 */

	static
	struct grdim_list *
grdim_split_list (

struct grdim_list *	dlist,		/* IN: list of grdim terms to split */
int			max_width,	/* IN: max width of each line */
int			operator	/* IN: operator being split */
)
{
int			width, height, ascent, descent, space;
struct grdim_list *	root;
struct grdim_list **	r_hookp;
struct grdim_list *	lterms;
struct grdim_list **	lt_hookp;
struct grdim_list *	dlp;
struct grdim_list *	dlp2;
struct grdim *		dp;
struct grdim *		dp2;

	/* Init empty list of lines. */
	root = NULL;
	r_hookp = &root;

	/* Init empty list of terms on current line. */
	lterms = NULL;
	lt_hookp = &lterms;
	width	= 0;
	height	= 0;
	ascent	= 0;
	descent	= 0;
	space	= 0;

	for (dlp = dlist; dlp NE NULL; dlp = dlp -> next) {
		dp = dlp -> operand;
		if (width + space + dp -> width > max_width) {
			/* This term/factor does not fit on the		*/
			/* current line.				*/
			if ((lterms EQ NULL) OR (dp -> width > max_width)) {
				/* This term/factor is too big all by	*/
				/* itself - fail!			*/
				goto fail;
			}

			/* Gather the current terms/factors into a sum	*/
			/* or product representing a complete line.	*/
			dp2 = new_grdim ();
			dp2 -> op	= operator;
			dp2 -> sign	= '+';
			dp2 -> width	= width;
			dp2 -> height	= height;
			dp2 -> ascent	= ascent;
			dp2 -> descent	= descent;
			dp2 -> u.list = lterms;
			dlp2 = new_grdim_list (dp2, NULL);

			/* Append this sum/product as another line. */
			*r_hookp = dlp2;
			r_hookp = &(dlp2 -> next);

			/* Start another line.  Add this term/factor as	*/
			/* the first on this new line.			*/
			lterms = NULL;
			lt_hookp = &lterms;
			width	= dp -> width;
			ascent	= dp -> ascent;
			descent	= dp -> descent;
			height	= ascent + 1 + descent;
			space	= 1;
			dlp2 = new_grdim_list (dp, NULL);
			*lt_hookp = dlp2;
			lt_hookp = &(dlp2 -> next);
		}
		else {
			/* Append this term/factor to current line. */
			dlp2 = new_grdim_list (dp, NULL);
			width	+= space + dp -> width;
			ascent	= _MAX (ascent, dp -> ascent);
			descent	= _MAX (descent, dp -> descent);
			height	= ascent + 1 + descent;
			*lt_hookp = dlp2;
			lt_hookp = &(dlp2 -> next);
			space = 1;
		}
	}

	/* All terms have been inserted into new structure.	*/
	if (lterms NE NULL) {
		/* Put remaining terms/factors out as a sum or product	*/
		/* forming the final line.				*/
		dp2 = new_grdim ();
		dp2 -> op	= operator;
		dp2 -> sign	= '+';
		dp2 -> width	= width;
		dp2 -> height	= height;
		dp2 -> ascent	= ascent;
		dp2 -> descent	= descent;
		dp2 -> u.list = lterms;
		dlp2 = new_grdim_list (dp2, NULL);

		/* Append this sum/product as another line. */
		*r_hookp = dlp2;
		r_hookp = &(dlp2 -> next);
	}

	return (root);

fail:
	free_grdim_split (root);

	return (NULL);
}

/*
 * Attempt to split the given function call into multiple lines.
 * Returns NULL upon failure.
 */

	static
	struct grdim_list *
grdim_split_func (

struct grdim *		op,		/* IN: function invocation to split */
int			max_width	/* IN: max width of each line */
)
{
int			width, height, ascent, descent, sep, delim;
char *			fname;
struct grdim_list *	root;
struct grdim_list **	r_hookp;
struct grdim_list *	lterms;
struct grdim_list **	lt_hookp;
struct grdim_list *	dlist;
struct grdim_list *	dlp;
struct grdim_list *	dlp2;
struct grdim *		dp;
struct grdim *		dp2;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR
			      (op -> op NE GRDIM_OP_FUNC));

	fname	= op -> u.func.fname;
	dlist	= op -> u.func.args;

	/* Init empty list of lines. */
	root = NULL;
	r_hookp = &root;

	/* Init empty list of terms on current line. */
	lterms = NULL;
	lt_hookp = &lterms;
	width	= strlen (fname) + 2;	/* fname space lparen */
	height	= 0;
	ascent	= 0;
	descent	= 0;
	sep	= 0;		/* separation space from prev arg */
	delim	= 1;		/* comma or rparen */

	for (dlp = dlist; dlp NE NULL; dlp = dlp -> next) {
		dp = dlp -> operand;
		if (width + sep + dp -> width + delim > max_width) {
			/* This arg does not fit on the current line.	*/
			if ((lterms EQ NULL) OR (dp -> width + delim > max_width)) {
				/* This arg is too big all by itself -	*/
				/* fail!				*/
				goto fail;
			}

			/* Gather the current args into an arg list	*/
			/* representing a complete line.		*/
			dp2 = new_grdim ();
			dp2 -> op		= GRDIM_OP_PARTIAL_FUNC;
			dp2 -> sign		= '+';
			dp2 -> width		= width;
			dp2 -> height		= height;
			dp2 -> ascent		= ascent;
			dp2 -> descent		= descent;
			dp2 -> u.func.fname	= fname;
			dp2 -> u.func.args	= lterms;

			dlp2 = new_grdim_list (dp2, NULL);

			/* Append this sum/product as another line. */
			*r_hookp = dlp2;
			r_hookp = &(dlp2 -> next);

			/* Successive lines have fname EQ NULL. */
			fname = NULL;

			/* Start another line.  Add this arg as the	*/
			/* first on this new line.			*/
			lterms = NULL;
			lt_hookp = &lterms;
			width	= dp -> width + delim;
			ascent	= dp -> ascent;
			descent	= dp -> descent;
			height	= ascent + 1 + descent;
			sep	= 1;
			dlp2 = new_grdim_list (dp, NULL);
			*lt_hookp = dlp2;
			lt_hookp = &(dlp2 -> next);
		}
		else {
			/* Append this arg to current line. */
			dlp2 = new_grdim_list (dp, NULL);
			width	+= sep + dp -> width + delim;
			ascent	= _MAX (ascent, dp -> ascent);
			descent	= _MAX (descent, dp -> descent);
			height	= ascent + 1 + descent;
			*lt_hookp = dlp2;
			lt_hookp = &(dlp2 -> next);

			/* Need 1-space separator before next arg. */
			sep = 1;
		}
	}

	/* All terms have been inserted into new structure.	*/
	if (lterms NE NULL) {
		/* Put remaining terms/factors out as a sum or product	*/
		/* forming the final line.				*/
		dp2 = new_grdim ();
		dp2 -> op	= GRDIM_OP_PARTIAL_FUNC;
		dp2 -> sign	= '+';
		dp2 -> width	= width;
		dp2 -> height	= height;
		dp2 -> ascent	= ascent;
		dp2 -> descent	= descent;
		dp2 -> u.func.fname	= fname;
		dp2 -> u.func.args	= lterms;
		dlp2 = new_grdim_list (dp2, NULL);

		/* Append this sum/product as another line. */
		*r_hookp = dlp2;
		r_hookp = &(dlp2 -> next);
	}

	return (root);

fail:
	free_grdim_split (root);

	return (NULL);
}

/*
 * Free up the nodes forming the syntactic "shell" around the terms or
 * factors that have been split across several lines.
 */

	static
	void
free_grdim_split (

struct grdim_list *	dlist		/* IN: list of lines to free */
)
{
struct grdim_list *	dlp;
struct grdim_list *	dlp2;
struct grdim *		dp;

	/* Free up newly created stuff. */
	while (dlist NE NULL) {
		dp = dlist -> operand;
		/* This should be a sum or product node containing	*/
		/* a list of terms/factors.  We need to free up the	*/
		/* sum/product node and the list nodes, but not the	*/
		/* individual sub-expressions.				*/
		switch (dp -> op) {
		case GRDIM_OP_ADD:
		case GRDIM_OP_MUL:
			dlp = dp -> u.list;
			free (dp);
			while (dlp NE NULL) {
				dlp2 = dlp -> next;
				free (dlp);
				dlp = dlp2;
			}
			break;

		case GRDIM_OP_PARTIAL_FUNC:
			dlp	= dp -> u.func.args;
			free (dp);
			while (dlp NE NULL) {
				dlp2 = dlp -> next;
				free (dlp);
				dlp = dlp2;
			}
			break;
		}

		dlp = dlist -> next;
		free (dlist);
		dlist = dlp;
	}
}

/*
 * Force printing of the given (very wide) sub-expression starting in
 * the given column, with the given maximum width.  This generally
 * results in a multi-line output.  This method assumes that variable
 * names are short -- and that it is large numeric constants that make
 * the expression be too wide to print.  This observation motivates
 * the primary technique used here to squeeze terms into the given
 * width: namely, to format large integers in a multi-line format using
 * backslash characters as a continuation indicator, as in the following
 * example:
 *
 *	      1234567890123456789\
 *	0123456789012345678901234\  2            135790246813579024\
 *	56789012345678901234567890 x    6813579024681357902468135790
 *	----------------------------- + ----------------------------
 *         2468013579246801357924680\       14703692581470369258
 *	13579246801357924680135792467
 *
 * This example is intended to stress the fact that the backslash characters
 * are used to wrap the numeric digits, not "lines" per se.  Also, the
 * present method does not place two "bignums" in a single "horizontal
 * band" as is done in this example.  At its core, the method operates
 * as follows:
 *
 * 1. Consider a product (possibly vacuous, e.g., containing only one
 *    factor).  Let t be the width of this product *excluding* any
 *    leading integer constant (t should count the space, if any, separating
 *    the leading integer from the remaining factors).
 *
 * 2. Let w = width - t.  Assuming w > k (where k is some small positive
 *    integer), format the integer coefficient as a block of \ continued
 *    digits that is w characters wide, letting it be as many lines tall
 *    as necessary.  If w <= k, then the expression is too wide for this
 *    method.
 *
 * This routine returns TRUE upon success, FALSE otherwise.
 * (Note that in the FALSE case, a partial/incomplete output will have
 * already been written to the output stream.  This is not desirable,
 * but c'es la poole.)
 */

	static
	int
grdim_print_indented (

FILE *		fp,		/* IN/OUT: stream to write to */
struct grdim *	dp,		/* IN: dimensioned sub-expr to print */
int		x,		/* IN: starting column to display in */
int		max_x		/* IN: max x coord within display box */
)
{
int		success, height;
struct canvas	canvas;

	height = 24;		/* Don't know how tall to make it! */
	init_canvas (&canvas, height);
	success = grdim_plot_indented (&canvas, dp, x, max_x);
	if (success) {
		write_canvas (&canvas, fp);
	}
	free_canvas (&canvas);

	return (success);
}

/*
 * Plot an especially wide sub-expression at the given column, forcing
 * its box to have the given maximum width.
 */


	static
	int
grdim_plot_indented (

struct canvas *	cp,		/* IN/OUT: canvas to plot into */
struct grdim *	dp,		/* IN: dimensioned sub-expr to print */
int		x,		/* IN: starting column to display in */
int		max_x		/* IN: max x coord within display box */
)
{
int		success;

	/* FIXME -- write this! */
	/* Consider doing something like "dp2 = expr_redim (dp, width);". */
	success = FALSE;

	return (success);
}

/*
 * Recursively construct an annotated form of the given expression that
 * gives the dimensions of each sub-expression when printed on an
 * infinitely tall and infinitely wide sheet of paper.
 */

	static
	struct grdim *
expr_dims (

const struct calib_genrep *	op	/* IN: expression to size up */
)
{
struct grdim *		dp;

	switch (op -> op) {
	case CALIB_GENREP_OP_ADD:
		dp = expr_dim_sum (op -> u.list);
		break;

	case CALIB_GENREP_OP_MUL:
		dp = expr_dim_prod (op -> u.list);
		break;

	case CALIB_GENREP_OP_IPOW:
		dp = expr_dim_pow (op);
		break;

	case CALIB_GENREP_OP_FUNC:
		dp = expr_dim_func (op);
		break;

	case CALIB_GENREP_OP_VAR:
		dp = make_grdim_uplus (make_grdim_var (op -> u.var));
		break;

	case CALIB_GENREP_OP_SI:
		dp = make_grdim_int (op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		dp = make_grdim_z (op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		dp = make_grdim_q (op -> u.qop);
		break;

	default:
		CALIB_FATAL_ERROR;
		dp = make_grdim_int (0);
		break;
	}

	return (dp);
}

/*
 * Recursively construct a grdim node that properly represents the sum
 * of the given list of terms.
 */

	static
	struct grdim *
expr_dim_sum (

const struct calib_genrep_list *	terms	/* IN: list of terms */
)
{
int					width, ascent, descent, first, space;
const struct calib_genrep *		gp1;
const struct calib_genrep_list *	lp;
struct grdim *				dp;
struct grdim_list *			dlp;
struct grdim_list *			root;
struct grdim_list **			hookp;

	root = NULL;
	hookp = &root;

	for (lp = terms; lp NE NULL; lp = lp -> next) {
		gp1 = lp -> operand;
		dp = expr_dims (gp1);
		dlp = new_grdim_list (dp, NULL);
		*hookp = dlp;
		hookp = &(dlp -> next);
	}

	if (root EQ NULL) {
		dp = make_grdim_int (0);
	}
	else if (root -> next EQ NULL) {
		dp = root -> operand;
		free (root);
	}
	else {
		/* Compute dimensions. */
		width = 0;
		first = TRUE;
		ascent = 0;
		descent = 0;
		space = 0;
		for (dlp = root; dlp NE NULL; dlp = dlp -> next) {
			dp = dlp -> operand;
			CALIB_FATAL_ERROR_IF ((dp -> op NE GRDIM_OP_UPLUS) AND
					      (dp -> op NE GRDIM_OP_NEG));

			/* Each term should have its own unary "+" or	*/
			/* unary "-" wrapped around it.			*/
			if (first) {
				/* Remove leading "+" from first term. */
				dp = grdim_unplus (dp);
				dlp -> operand = dp;
			}

			width	+= (space + dp -> width);
			ascent	= _MAX (ascent, dp -> ascent);
			descent	= _MAX (descent, dp -> descent);
			first = FALSE;
			space = 1;
		}
		dp = new_grdim ();
		dp -> op	= GRDIM_OP_ADD;
		dp -> u.list	= root;
		dp -> width	= width;
		dp -> height	= ascent + 1 + descent;
		dp -> ascent	= ascent;
		dp -> descent	= descent;
	}

	return (dp);
}

/*
 * Recursively construct a grdim node that properly represents the product
 * of the given list of factors.
 */

	static
	struct grdim *
expr_dim_prod (

const struct calib_genrep_list *	factors	/* IN: list of factors */
)
{
int				pow, sign;
const struct calib_genrep *	gp1;
const struct calib_genrep_list * lp;
struct grdim *			dp;
struct grdim *			dp2;
struct grdim *			dp3;
struct grdim_list *		dlp;
struct grdim_list *		numer;
struct grdim_list *		denom;
struct grdim_list **		nhookp;
struct grdim_list **		dhookp;
mpz_t				num_const;
mpz_t				den_const;

	/* Initialize numerator and denominator factor lists to empty. */
	mpz_init_set_ui (num_const, 1);
	mpz_init_set_ui (den_const, 1);
	numer = NULL;		nhookp = &numer;
	denom = NULL;		dhookp = &denom;

	/* Pass 1.  Split out numerator vs. denominator factors. */
	for (lp = factors; lp NE NULL; lp = lp -> next) {
		gp1 = lp -> operand;
		if (gp1 -> op EQ CALIB_GENREP_OP_Z) {
			/* The constant factor should be first. */
			mpz_mul (num_const, num_const, gp1 -> u.zop);
		}
		else if (gp1 -> op EQ CALIB_GENREP_OP_Q) {
			/* The constant factor (if any) should	*/
			/* be first.  Split into separate	*/
			/* numerator and denominator portions.	*/
			mpz_mul (num_const, num_const, mpq_numref (gp1 -> u.qop));
			mpz_mul (den_const, den_const, mpq_denref (gp1 -> u.qop));
		}
		else if (gp1 -> op EQ CALIB_GENREP_OP_IPOW) {
			pow = gp1 -> u.ipow.power;
			CALIB_FATAL_ERROR_IF ((pow EQ 0) OR (pow EQ 1));
			if (pow > 0) {
				dp2 = expr_dims (gp1);
				dp2 = grdim_strip_signs (dp2, num_const);
				dlp = new_grdim_list (dp2, NULL);
				*nhookp = dlp;
				nhookp = &(dlp -> next);
			}
			else {
				/* Factor belongs in denominator. */
				if (pow EQ -1) {
					/* Remove power. */
					dp2 = expr_dims (gp1 -> u.ipow.base);
				}
				else {
					struct calib_genrep	tmp;
					tmp.op = CALIB_GENREP_OP_IPOW;
					tmp.u.ipow.base  = gp1 -> u.ipow.base;
					tmp.u.ipow.power = - pow;
					dp2 = expr_dims (&tmp);
				}
				dp2 = grdim_strip_signs (dp2, num_const);
				dlp = new_grdim_list (dp2, NULL);
				*dhookp = dlp;
				dhookp = &(dlp -> next);
			}
		}
		else {
			/* Put this factor into numerator. */
			dp2 = expr_dims (gp1);
			dp2 = grdim_strip_signs (dp2, num_const);
			dlp = new_grdim_list (dp2, NULL);
			*nhookp = dlp;
			nhookp = &(dlp -> next);
		}
	}

	if (mpz_sgn (den_const) < 0) {
		/* Force negative sign into the numerator. */
		mpz_neg (num_const, num_const);
		mpz_neg (den_const, den_const);
	}
	sign = mpz_sgn (num_const);
	mpz_abs (num_const, num_const);

	/* We should not have a constant factor of zero! */
	CALIB_FATAL_ERROR_IF (sign EQ 0);

	if (denom EQ NULL) {
		if (mpz_cmp_ui (den_const, 1) EQ 0) {
			/* Denominator is 1 (empty).  No fraction needed. */
			dp = expr_dim_gather_prod (num_const, numer);
			dp = grdim_strip_signs (dp, NULL);
		}
		else {
			/* Denominator is just an integer. */
			dp2 = expr_dim_gather_prod (num_const, numer);
			dp2 = grdim_strip_signs (dp2, NULL);

			dp3 = make_grdim_z (den_const);
			dp3 = grdim_strip_signs (dp3, NULL);

			dp = make_grdim_frac (dp2, dp3);
		}
	}
	else {
		dp2 = expr_dim_gather_prod (num_const, numer);
		dp2 = grdim_strip_signs (dp2, NULL);

		dp3 = expr_dim_gather_prod (den_const, denom);
		dp3 = grdim_strip_signs (dp3, NULL);

		dp = make_grdim_frac (dp2, dp3);
	}

	if (sign < 0) {
		/* Wrap a neg(...) around the product. */
		dp = make_grdim_neg (dp);
	}
	else {
		/* Wrap a +(...) around the product. */
		dp = make_grdim_uplus (dp);
	}

	mpz_clear (den_const);
	mpz_clear (num_const);

	return (dp);
}

/*
 * Gather the given factors into a proper set of grdim nodes.
 */

	static
	struct grdim *
expr_dim_gather_prod (

mpz_srcptr		zval,	/* IN: constant factor */
struct grdim_list *	factors	/* IN: list of other factors */	
)
{
int			sign;
struct grdim *		dp;
struct grdim *		dp2;
mpz_t			VALUE;

	mpz_init (VALUE);
	mpz_abs (VALUE, zval);

	sign = mpz_sgn (zval);

	if (factors EQ NULL) {
		/* Constant factor is the only one. */
		dp = make_grdim_z (VALUE);
		dp = grdim_unplus (dp);
	}
	else if (mpz_cmp_ui (VALUE, 1) EQ 0) {
		/* Constant factor is +1 or -1.  Move into sign. */
		if (factors -> next EQ NULL) {
			/* Only one factor. */
			dp = factors -> operand;
			free (factors);
		}
		else {
			dp = make_grdim_prod (factors);
		}
	}
	else {
		/* Combine constant factor with others. */
		dp2 = make_grdim_z (VALUE);
		dp2 = grdim_unplus (dp2);
		factors = new_grdim_list (dp2, factors);
		dp = make_grdim_prod (factors);
	}

	if (sign < 0) {
		dp = make_grdim_neg (dp);
	}
	else {
		dp = make_grdim_uplus (dp);
	}

	mpz_clear (VALUE);

	return (dp);
}

/*
 * Recursively construct a grdim node that properly represents the
 * given exponent.
 */

	static
	struct grdim *
expr_dim_pow (

const struct calib_genrep *	op	/* IN: power expression to process */
)
{
int				pow;
const struct calib_genrep *	base;
struct grdim *			dp;
struct grdim *			dp2;

	base = op -> u.ipow.base;
	pow  = op -> u.ipow.power;

	CALIB_FATAL_ERROR_IF ((pow EQ 0) OR (pow EQ 1));

	dp2 = expr_dims (base);
	dp2 = grdim_unplus (dp2);

	if (pow EQ -1) {
		/* Special handling for reciprocals. */
		dp = make_grdim_frac (make_grdim_int (1), dp2);
		return (dp);
	}

	if ((dp2 -> op EQ GRDIM_OP_ADD) OR
	    (dp2 -> op EQ GRDIM_OP_NEG) OR
	    (dp2 -> op EQ GRDIM_OP_MUL)) {
		/* Must parenthesize (a+b)^2, (-a)^2 and (a*b)^2. */
		dp2 -> parens = TRUE;
		dp2 -> width += 2;
	}

	if (pow > 0) {
		dp = make_grdim_pow (dp2, pow);
	}
	else {
		dp2 = make_grdim_pow (dp2, - pow);
		dp = make_grdim_frac (make_grdim_int (1), dp2);
	}

	dp = make_grdim_uplus (dp);

	return (dp);
}

/*
 * Make a grdim node to contain a function call with args.
 */

	static
	struct grdim *
expr_dim_func (

const struct calib_genrep *	op	/* IN: function invocation */
)
{
struct calib_genrep *		fgp;
struct calib_genrep_list *	lp;
struct calib_genrep_list *	lp2;
struct grdim *			dp;
struct grdim *			dp2;
struct grdim_list *		root;
struct grdim_list **		hookp;
struct grdim_list *		dlp2;
const char *			fname;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	fgp	= op -> u.func.func;
	lp	= op -> u.func.args;

	CALIB_FATAL_ERROR_IF (fgp EQ NULL);

	if (fgp -> op EQ CALIB_GENREP_OP_VAR) {
		fname = fgp -> u.var;
	}
	else if (fgp -> op EQ CALIB_GENREP_OP_SI) {
		fname = calib_genrep_builtin_func_index_to_name (fgp -> u.siop);
		CALIB_FATAL_ERROR_IF (fname EQ NULL);
	}
	else {
		CALIB_FATAL_ERROR;
		fname = "f";
	}

	/* Process the argument list. */
	root = NULL;
	hookp = &root;
	for (lp2 = lp; lp2 NE NULL; lp2 = lp2 -> next) {
		dp2 = expr_dims (lp2 -> operand);
		/* Remove any unary plus. */
		dp2 = grdim_unplus (dp2);
		dlp2 = new_grdim_list (dp2, NULL);
		*hookp = dlp2;
		hookp = &(dlp2 -> next);
	}

	dp = new_grdim ();

	dp -> op	= GRDIM_OP_FUNC;
	dp -> sign	= '+';
	dp -> u.func.fname	= strdup (fname);
	dp -> u.func.args	= root;

	dp -> height	= 1;
	dp -> width	= strlen (fname) + 3;	/* "fname ()" */
	dp -> ascent	= 0;
	dp -> descent	= 0;

	for (dlp2 = root; dlp2 NE NULL; dlp2 = dlp2 -> next) {
		dp2 = dlp2 -> operand;
		dp -> width += dp2 -> width;
		if (dlp2 -> next NE NULL) {
			dp -> width += 2;	/* ", " between args */
		}
		if (dp2 -> ascent > dp -> ascent) {
			dp -> ascent = dp2 -> ascent;
		}
		if (dp2 -> descent > dp -> descent) {
			dp -> descent = dp2 -> descent;
		}
	}
	dp -> height = 1 + dp -> ascent + dp -> descent;

	dp = make_grdim_uplus (dp);

	return (dp);
}

/*
 * Make a grdim node for unary "+" of a sub-expression.
 */

	static
	struct grdim *
make_grdim_uplus (

struct grdim *		op	/* IN: grdim of sub-expression */
)
{
struct grdim *		dp;

	if (op -> op EQ GRDIM_OP_UPLUS) {
		return (op);
	}

	dp = new_grdim ();
	dp -> op	= GRDIM_OP_UPLUS;
	dp -> u.unop	= op;
	dp -> sign	= '-';
	dp -> width	= op -> width + 2;
	dp -> height	= op -> height;
	dp -> ascent	= op -> ascent;
	dp -> descent	= op -> descent;

	return (dp);
}

/*
 * Strip all unary '+' and '-' operations from the given grdim
 * (freeing them as you go).  Make the corresponding sign corrections
 * to the given numeric factor.
 */

	static
	struct grdim *
grdim_strip_signs (

struct grdim *		dp,	/* IN: grdim to strip signs from */
mpz_ptr			numfact	/* IN: numeric factor to correct sign of */
)
{
	for (;;) {
		if (dp -> op EQ GRDIM_OP_UPLUS) {
			/* Strip unary '+'. */
			dp = grdim_unplus (dp);
		}
		else if (dp -> op EQ GRDIM_OP_NEG) {
			/* Strip unary '-' and flip sign. */
			dp = grdim_unneg (dp);
			if (numfact NE NULL) {
				mpz_neg (numfact, numfact);
			}
		}
		else {
			break;
		}
	}

	return (dp);
}

/*
 * If the given grdim is a unary "+", strip it off (and free it),
 * leaving only the un-adorned sub-expression.  Otherwise, return
 * the given sub-expression unmodified.
 */

	static
	struct grdim *
grdim_unplus (

struct grdim *		op	/* IN: grdim to "unplus" */
)
{
struct grdim *		dp;

	if (op -> op NE GRDIM_OP_UPLUS) {
		return (op);
	}
	dp = op -> u.unop;
	memset (op, 0, sizeof (*op));
	free (op);
	return (dp);
}

/*
 * If the given grdim is a unary "-", strip it off (and free it),
 * leaving only the un-adorned sub-expression.  Otherwise, return
 * the given sub-expression unmodified.
 */

	static
	struct grdim *
grdim_unneg (

struct grdim *		op	/* IN: grdim to "unplus" */
)
{
struct grdim *		dp;

	if (op -> op NE GRDIM_OP_NEG) {
		return (op);
	}
	dp = op -> u.unop;
	memset (op, 0, sizeof (*op));
	free (op);
	return (dp);
}

/*
 * Make a grdim node for the negation of a sub-expression.
 */

	static
	struct grdim *
make_grdim_neg (

struct grdim *		op	/* IN: grdim of sub-expression */
)
{
struct grdim *		dp;

	dp = new_grdim ();
	dp -> op	= GRDIM_OP_NEG;
	dp -> u.unop	= op;
	dp -> sign	= '-';
	dp -> width	= op -> width + 2;
	dp -> height	= op -> height;
	dp -> ascent	= op -> ascent;
	dp -> descent	= op -> descent;

	return (dp);
}

/*
 * Make a grdim node for a product of 2 or more factors.
 */

	static
	struct grdim *
make_grdim_prod (

struct grdim_list *	list	/* IN: list of grdim's of factors */
)
{
int			n, width, ascent, descent;
int			space;
struct grdim *		dp;
struct grdim_list *	dlp;

	space = 0;
	width = 0;
	ascent = 0;
	descent = 0;
	n = 0;
	for (dlp = list; dlp NE NULL; dlp = dlp -> next) {
		dp = dlp -> operand;
		if (dp -> op EQ GRDIM_OP_ADD) {
			/* Must parenthesize sums inside of products. */
			dp -> parens = TRUE;
			dp -> width += 2;
		}
		/* First factor does not get a separating space, rest do. */
		width	+= (space + dp -> width);
		ascent	= _MAX (ascent, dp -> ascent);
		descent	= _MAX (descent, dp -> descent);
		space	= 1;
		++n;
	}
	CALIB_FATAL_ERROR_IF (n < 2);

	dp = new_grdim ();
	dp -> op	= GRDIM_OP_MUL;
	dp -> u.list	= list;
	dp -> sign	= '+';
	dp -> width	= width;
	dp -> height	= ascent + 1 + descent;
	dp -> ascent	= ascent;
	dp -> descent	= descent;

	return (dp);
}

/*
 * Make a grdim node for a fraction bar.
 */

	static
	struct grdim *
make_grdim_frac (

struct grdim *		numer,		/* IN: numerator of fraction */
struct grdim *		denom		/* IN: denominator of fration */
)
{
struct grdim *		dp;

	dp = new_grdim ();

	dp -> op = GRDIM_OP_FRAC;
	dp -> u.frac.numer = numer;
	dp -> u.frac.denom = denom;

	dp -> sign	= numer -> sign;
	dp -> width	= _MAX (numer -> width, denom -> width);
	dp -> ascent	= numer -> height;
	dp -> descent	= denom -> height;
	dp -> height	= 1 + dp -> ascent + dp -> descent;

	return (dp);
}

/*
 * Make a grdim node for the given integer power.
 */

	static
	struct grdim *
make_grdim_pow (

struct grdim *		base,		/* IN: item to raise to power */
int			pow		/* IN: power to raise to */
)
{
int			esize;
struct grdim *		dp;
char			buf [32];

	CALIB_FATAL_ERROR_IF (pow < 2);

	sprintf (buf, "%d", pow);
	esize = strlen (buf);

	dp = new_grdim ();
	dp -> op	= GRDIM_OP_POW;
	dp -> u.pow.op1	= base;
	dp -> u.pow.op2	= pow;

	dp -> width	= base -> width + esize;
	dp -> ascent	= _MAX (base -> ascent, 1);
	dp -> descent	= base -> descent;
	dp -> height	= dp -> ascent + 1 + dp -> descent;

	return (dp);
}

/*
 * Make a grdim node for the given variable name.
 */

	static
	struct grdim *
make_grdim_var (

const char *		var		/* IN: variable name */
)
{
struct grdim *		dp;

	dp = new_grdim ();

	dp -> op	= GRDIM_OP_VAR;
	dp -> u.var	= strdup (var);
	dp -> width	= strlen (var);

	return (dp);
}

/*
 * Make a grdim subtree to hold the given rational value.
 */

	static
	struct grdim *
make_grdim_q (

mpq_srcptr		qval	/* IN: rational value */
)
{
int			sign;
struct grdim *		dp;
struct grdim *		dp2;
struct grdim *		dp3;
mpq_t			abs_qval;

	sign = mpq_sgn (qval);

	mpq_init (abs_qval);
	mpq_abs (abs_qval, qval);

	if (mpz_cmp_ui (mpq_denref (abs_qval), 1) EQ 0) {
		/* Value is integral. */
		dp = make_grdim_z (mpq_numref (abs_qval));
		dp = grdim_strip_signs (dp, NULL);
	}
	else {
		/* Need numerator/denominator. */
		dp2 = make_grdim_z (mpq_numref (abs_qval));
		dp2 = grdim_strip_signs (dp2, NULL);

		dp3 = make_grdim_z (mpq_denref (abs_qval));
		dp3 = grdim_strip_signs (dp3, NULL);

		dp = make_grdim_frac (dp2, dp3);
	}
	if (sign < 0) {
		dp = make_grdim_neg (dp);
	}
	else {
		dp = make_grdim_uplus (dp);
	}

	mpq_clear (abs_qval);

	return (dp);
}

/*
 * Make a grdim node for the given "int" value.
 */

	static
	struct grdim *
make_grdim_int (

calib_si_t		ival		/* IN: integer value */
)
{
struct grdim *		dp;
mpz_t			zval;

	mpz_init_set_si (zval, ival);
	dp = make_grdim_z (zval);
	mpz_clear (zval);

	return (dp);
}

/*
 * Make a grdim node to contain the given integer value.
 */

	static
	struct grdim *
make_grdim_z (

mpz_srcptr		zval		/* IN: integer value to use */
)
{
int			ndigits;
struct grdim *		dp;
char *			string;

	/* Convert |zval| into string of decimal digits. */
	ndigits = calib_genrep_convert_abs_Z_to_decimal_string (zval, &string);

	dp = new_grdim ();

	dp -> op	= GRDIM_OP_Z;
	dp -> sign	= (mpz_sgn (zval) >= 0) ? '+' : '-';
	dp -> u.zstr	= string;

	dp -> height	= 1;
	dp -> width	= ndigits;
	dp -> ascent	= 0;
	dp -> descent	= 0;

	if (mpz_sgn (zval) < 0) {
		dp = make_grdim_neg (dp);
	}
	else {
		dp = make_grdim_uplus (dp);
	}

	return (dp);
}

/*
 * Create a new "struct grdim" node.
 */

	static
	struct grdim *
new_grdim (void)

{
struct grdim *		dp;

	dp = CALIB_NEW (struct grdim);
	memset (dp, 0, sizeof (*dp));

	dp -> sign	= '+';
	dp -> leaf	= FALSE;
	dp -> parens	= FALSE;
	dp -> multiline	= FALSE;
	dp -> width	= 0;
	dp -> height	= 1;
	dp -> ascent	= 0;
	dp -> descent	= 0;

	return (dp);
}

/*
 * Create a new "struct grdim_list" node.
 */

	static
	struct grdim_list *
new_grdim_list (

struct grdim *		op,		/* IN: operand */
struct grdim_list *	next		/* IN: list of successive operands */
)
{
struct grdim_list *	dlp;

	dlp = CALIB_NEW (struct grdim_list);
	dlp -> operand	= op;
	dlp -> next	= next;

	return (dlp);
}

/*
 * Recursively free the given grdim.
 */

	static
	void
grdim_free (

struct grdim *		dp	/* IN: grdim to recursively free */
)
{
struct grdim_list *	dlp;
struct grdim_list *	tmp;

	if (dp EQ NULL) return;

	switch (dp -> op) {
	case GRDIM_OP_ADD:
	case GRDIM_OP_MUL:
		dlp = dp -> u.list;
		dp -> u.list = NULL;
		while (dlp NE NULL) {
			grdim_free (dlp -> operand);
			tmp = dlp -> next;
			free (dlp);
			dlp = tmp;
		}
		break;

	case GRDIM_OP_UPLUS:
	case GRDIM_OP_NEG:
		grdim_free (dp -> u.unop);
		break;

	case GRDIM_OP_FRAC:
		grdim_free (dp -> u.frac.numer);
		grdim_free (dp -> u.frac.denom);
		break;

	case GRDIM_OP_POW:
		grdim_free (dp -> u.pow.op1);
		break;

	case GRDIM_OP_FUNC:
		free (dp -> u.func.fname);
		dp -> u.func.fname = NULL;
		dlp = dp -> u.func.args;
		dp -> u.func.args = NULL;
		while (dlp NE NULL) {
			grdim_free (dlp -> operand);
			tmp = dlp -> next;
			free (dlp);
			dlp = tmp;
		}
		break;

	case GRDIM_OP_VAR:
		free ((char *) (dp -> u.var));
		break;

	case GRDIM_OP_Z:
		free (dp -> u.zstr);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	free (dp);
}

/*
 * Plot the given grdim sub-expression into the given canvas at the given
 * x,y coordinates.
 */

	static
	void
plot_grdim (

struct canvas *		cp,		/* IN/OUT: canvas to plot into */
struct grdim *		dp,		/* IN: sub-expression to plot */
int			x,		/* IN: x coordinate */
int			y		/* IN: y coordinate */
)
{
int			w;
struct grdim_list *	dlp;
char *			buf;
const char *		fname;
struct grdim *		dp2;
struct grdim *		dp3;
char			numbuf [32];

	CALIB_FATAL_ERROR_IF ((cp EQ NULL) OR (cp EQ NULL));
	CALIB_FATAL_ERROR_IF ((x < 0) OR (y < 0));

	if (dp -> parens) {
		plot_canvas (cp, LPAREN, x, y);
		plot_canvas (cp, RPAREN, x + dp -> width - 1, y);
		++x;
	}

	switch (dp -> op) {
	case GRDIM_OP_ADD:
		dlp = dp -> u.list;
		CALIB_FATAL_ERROR_IF (dlp EQ NULL);
		for (; dlp NE NULL; dlp = dlp -> next) {
			dp2 =  dlp -> operand;
			/* Each term should have its own unary "+" or	*/
			/* unary "-" on it, except the first term won't	*/
			/* have a unary "+".				*/
			plot_grdim (cp, dp2, x, y);
			x += (dp2 -> width + 1);
		}
		break;

	case GRDIM_OP_UPLUS:
		plot_canvas (cp, "+", x, y);
		plot_grdim (cp, dp -> u.unop, x + 2, y);
		break;

	case GRDIM_OP_NEG:
		plot_canvas (cp, "-", x, y);
		plot_grdim (cp, dp -> u.unop, x + 2, y);
		break;

	case GRDIM_OP_MUL:
		dlp = dp -> u.list;
		CALIB_FATAL_ERROR_IF (dlp EQ NULL);
		dp2 = dlp -> operand;
		plot_grdim (cp, dp2, x, y);
		x += dp2 -> width + 1;
		for (dlp = dlp -> next; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			plot_grdim (cp, dp2, x, y);
			x += (dp2 -> width + 1);
		}
		break;

	case GRDIM_OP_FRAC:
		w = dp -> width;
		if (dp -> parens) {
			w -= 2;
		}
		buf = CALIB_NEWA (w + 1, char);
		memset (buf, '-', w);
		buf [w] = '\0';
		plot_canvas (cp, buf, x, y);
		free (buf);
		dp2 = dp -> u.frac.numer;
		dp3 = dp -> u.frac.denom;
		plot_grdim (cp,
			    dp2,
			    x + (w - dp2 -> width) / 2,
			    y - 1 - dp2 -> descent);
		plot_grdim (cp,
			    dp3,
			    x + (w - dp3 -> width) / 2,
			    y + 1 + dp3 -> ascent);
		break;

	case GRDIM_OP_POW:
		dp2 = dp -> u.pow.op1;
		sprintf (numbuf, "%d", dp -> u.pow.op2);
		plot_grdim (cp, dp2, x, y);
		plot_canvas (cp, numbuf, x + dp2 -> width, y - 1);
		break;

	case GRDIM_OP_FUNC:
		fname = dp -> u.func.fname;
		plot_canvas (cp, fname, x, y);
		x += strlen (fname) + 1;
		plot_canvas (cp, LPAREN, x, y);
		++x;
		for (dlp = dp -> u.func.args; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			
			plot_grdim (cp, dp2, x, y);
			x += dp2 -> width;
			if (dlp -> next NE NULL) {
				plot_canvas (cp, ",", x, y);
				x += 2;
			}
		}
		plot_canvas (cp, RPAREN, x, y);
		break;

	case GRDIM_OP_PARTIAL_FUNC:
		fname = dp -> u.func.fname;
		if (fname NE NULL) {
			plot_canvas (cp, fname, x, y);
			x += strlen (fname) + 1;
			plot_canvas (cp, LPAREN, x, y);
			++x;
		}
		for (dlp = dp -> u.func.args; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			
			plot_grdim (cp, dp2, x, y);
			x += dp2 -> width;
			if (dlp -> next NE NULL) {
				plot_canvas (cp, ",", x, y);
				x += 2;
			}
		}
		/* Closing paren provided by caller on last line of	*/
		/* arguments.						*/
		break;

	case GRDIM_OP_VAR:
		plot_canvas (cp, dp -> u.var, x, y);
		break;

	case GRDIM_OP_Z:
		plot_canvas (cp, dp -> u.zstr, x, y);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Initialize the given canvas to be the given number of lines tall.
 */

	static
	void
init_canvas (

struct canvas *		cp,		/* IN/OUT: canvas to initialize */
int			nlines		/* IN: number of lines */
)
{
int			i;
struct cline *		lines;
struct cline *		lp;

	CALIB_FATAL_ERROR_IF (nlines <= 0);

	lines = CALIB_NEWA (nlines, struct cline);
	for (i = 0; i < nlines; i++) {
		lp = &lines [i];
		lp -> len	= 0;
		lp -> text	= strdup ("");
	}

	cp -> nlines	= nlines;
	cp -> lines	= lines;
}

/*
 * Plot the given integer ***RIGHT-TO-LEFT*** starting at the given
 * x,y coordinates.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
plot_canvas_z (

struct canvas *		cp,		/* IN/OUT: canvas to plot into */
mpz_srcptr		zval,		/* IN: integer value */
int			x,		/* IN: x coordinate */
int			y		/* IN: y coordinate */
)
{
int		sign;
mpz_t		z;
char		buf [2];

	buf [1] = '\0';

	sign = mpz_sgn (zval);
	if (sign EQ 0) {
		buf [0] = '0';
		plot_canvas (cp, buf, x, y);
		return;
	}

	CALIB_FATAL_ERROR_IF (sign < 0);

	/* Make modifiable local copy of value. */
	mpz_init_set (z, zval);

	while (mpz_sgn (z) > 0) {
		buf [0] = '0' + mpz_fdiv_q_ui (z, z, 10);
		plot_canvas (cp, buf, x, y);
		--x;
	}

	mpz_clear (z);
}

#pragma GCC diagnostic pop

/*
 * Plot the given text horizontally into the given canvas, starting at
 * the given x,y coordinates.
 */

	static
	void
plot_canvas (

struct canvas *		cp,		/* IN/OUT: canvas to plot into */
const char *		text,		/* IN: text to plot */
int			x,		/* IN: x coordinate */
int			y		/* IN: y coordinate */
)
{
int			l1, l2, l3;
char *			lbuf;
struct cline *		lp;

	CALIB_FATAL_ERROR_IF ((cp EQ NULL) OR (text EQ NULL));
	CALIB_FATAL_ERROR_IF ((x < 0) OR (y < 0) OR (y >= cp -> nlines));

	CALIB_FATAL_ERROR_IF (y >= cp -> nlines);

	lp = &(cp -> lines [y]);

	l1 = strlen (text);
	l2 = x + l1;
	l3 = lp -> len;
	if (l2 > l3) {
		/* Make line longer, and pad with blanks. */
		lbuf = CALIB_NEWA (l2 + 1, char);
		memcpy (lbuf, lp -> text, l3);
		memset (&lbuf [l3], ' ', l2 - l3);
		lbuf [l2] = '\0';
		free (lp -> text);
		lp -> text = lbuf;
		lp -> len = l2;
	}
	memcpy (&(lp -> text [x]), text, l1);
}

/*
 * Write the given canvas to the given output stream.
 */

	static
	void
write_canvas (

struct canvas *		cp,		/* IN: canvas to write */
FILE *			fp		/* IN: output stream to write into */
)
{
int		i, n;
struct cline *	lines;
struct cline *	lp;

	CALIB_FATAL_ERROR_IF ((cp EQ NULL) OR (fp EQ NULL));

	n	= cp -> nlines;
	lines	= cp -> lines;
	for (i = 0; i < n; i++) {
		lp = &lines [i];
		fprintf (fp, "%s\n", lp -> text);
	}
	fflush (fp);
}

/*
 * Destroy the given canvas.
 */

	static
	void
free_canvas (

struct canvas *		cp		/* IN: canvas to clean up. */
)
{
int		i, n;
struct cline *	lines;
struct cline *	lp;

	n = cp -> nlines;
	lines = cp -> lines;
	cp -> nlines	= 0;
	cp -> lines	= NULL;
	for (i = 0; i < n; i++) {
		lp = &lines [i];
		free (lp -> text);
		lp -> len  = 0;
		lp -> text = NULL;
	}
	free (lines);
}

/*
 * Convert the given |zval| into a decimal string with no sign.
 */

	size_t
calib_genrep_convert_abs_Z_to_decimal_string (

mpz_srcptr		zval,		/* IN: convert |zval| */
char **			str_out		/* IN/OUT: decimal string */
)
{
int		i, count;
size_t		n, ndigits;
unsigned long	dig8;
char *		string;
char *		srcp;
char *		dstp;
struct rbblk *	blkp;
mpz_t		zabs, zrem;
RubberBuf	rbuf;

	if (mpz_sgn (zval) EQ 0) {
		/* Handle zero specially. */
		*str_out = strdup ("0");
		return (1);
	}

	mpz_init (zabs);
	mpz_init (zrem);
	mpz_abs (zabs, zval);
	init_rubber_buf (&rbuf);

	while (mpz_sgn (zabs) > 0) {
		/* Do groups of 8 digits.  This works for both 32-bit	*/
		/* and 64-bit limbs.					*/
		dig8 = mpz_fdiv_qr_ui (zabs, zrem, zabs, 100000000);
		if (mpz_sgn (zabs) EQ 0) {
			/* These are the last digits. */
			while (dig8 > 0) {
				rubber_buf_putc (&rbuf, '0' + dig8 % 10);
				dig8 /= 10;
			}
		}
		else {
			/* We have a full group of 8 digits. */
			for (i = 0; i < 8; i++) {
				rubber_buf_putc (&rbuf, '0' + dig8 % 10);
				dig8 /= 10;
			}
			CALIB_FATAL_ERROR_IF (dig8 NE 0);
		}
	}

	/* The full decimal string is now in the rubber buffer, but in	*/
	/* reverse order.  Allocate a correctly sized buffer for it and	*/
	/* copy in reverse order.					*/

	ndigits = rbuf.nblks * RUBBER_SIZE + (RUBBER_SIZE - rbuf.cnt);
	string = CALIB_NEWA (ndigits + 1, char);
	dstp = string + ndigits;
	*dstp = '\0';
	n = ndigits;
	for (blkp = &rbuf.first; n > 0; blkp = blkp -> next) {
		CALIB_FATAL_ERROR_IF (blkp EQ NULL);
		count = n;
		if (count > RUBBER_SIZE) {
			count = RUBBER_SIZE;
		}
		n -= count;
		srcp = &(blkp -> buf [0]);
		while (count > 0) {
			*--dstp = *srcp++;
			--count;
		}
	}
	CALIB_FATAL_ERROR_IF (dstp NE string);

	clear_rubber_buf (&rbuf);
	mpz_clear (zrem);
	mpz_clear (zabs);

	*str_out = string;

	return (ndigits);
}

/*
 * Initialize the given rubber buffer to empty.
 */

	static
	void
init_rubber_buf (

RubberBuf *	rbp		/* IN: rubber buffer to initialize */
)
{
	rbp -> ptr		= &(rbp -> first.buf [0]);
	rbp -> cnt		= RUBBER_SIZE;
	rbp -> nblks		= 0;
	rbp -> last		= &(rbp -> first);
	rbp -> first.next	= NULL;
}

/*
 * Destroy the given rubber buffer -- freeing up all of the memory
 * it currently uses.
 */

	static
	void
clear_rubber_buf (

RubberBuf *	rbp		/* IN: rubber buffer to clear */
)
{
struct rbblk *	p;
struct rbblk *	tmp;

	p = rbp -> first.next;
	rbp -> ptr	  = NULL;
	rbp -> first.next = NULL;

	while (p NE NULL) {
		tmp = p -> next;
		free (p);
		p = tmp;
	}
}

/*
 * Append the given character to the given rubber buffer.
 */

	static
	void
rubber_buf_putc (

RubberBuf *	rbp,		/* IN: rubber buffer to append to */
char		c		/* IN: character to append */
)
{
int		cnt;
char *		ptr;
struct rbblk *	new_blk;

	ptr = rbp -> ptr;
	cnt = rbp -> cnt;

	if (cnt <= 0) {
		new_blk = CALIB_NEW (struct rbblk);
		new_blk -> next = NULL;
		rbp -> last -> next = new_blk;
		rbp -> last	    = new_blk;
		++(rbp -> nblks);
		ptr = &(new_blk -> buf [0]);
		cnt = RUBBER_SIZE;
	}

	*ptr++ = c;

	rbp -> ptr = ptr;
	rbp -> cnt = cnt - 1;
}
