/**********************************************************************

	$Id: prettyprint.c,v 1.1 2024/10/20 00:36:28 warme Exp $

	File:	prettyprint.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Pretty-printing of expressions in genrep form.

***********************************************************************

	Modification Log:

	a-1:	06/11/2017	warme
		: Created.  (Split off from genrep.c.)
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/genrep.h"

#include "calib/fatal.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/types.h"
#include "genrep_int.h"
#include <gmp.h>
#include <limits.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <unistd.h>


/*
 * Global Functions.
 */

void	calib_genrep_prettyprint (const struct calib_genrep * op);
void	calib_genrep_prettyprint_file (
			FILE *				fp,
			const struct calib_genrep *	op);
void	calib_genrep_prettyprint_file_width (
			FILE *				fp,
			const struct calib_genrep *	op,
			int				width);
void	calib_genrep_prettyprint_width (
			const struct calib_genrep *	op,
			int				width);
void	calib_genrep_print_maxima (
			FILE *				fp,
			const struct calib_genrep *	op,
			int				width);

/*
 * Local Macros
 */

#define _MAX(A,B)	(((A) >= (B)) ? (A) : (B))
#define _MIN(A,B)	(((A) <= (B)) ? (A) : (B))

#define DEFAULT_MAX_WIDTH	79

#define	MIN_WRAP_DIGITS		10

#define LPAREN_STRING	"("
#define RPAREN_STRING	")"

/*
 * Global information passed throughout the Pretty Print module.
 */

struct PPInfo {
	int		flags;		/* Formatting flags */
	FILE *		fp;		/* Output stream to write to */
	int		maxwidth;	/* Max chars / line */
};

#define	FLAG_FORMAT_MAXIMA	0x001	/* Format expression to be read	*/
					/* as input by maxima		*/

/*
 * The dimensions of a given expression, when displayed as a "tile"
 * of arbitrary size.  A "tile" assumes no splitting of the expression
 * into multiple horizontal "panels", but can employ "wrapping" of very
 * wide entities (i.e., integer constants) to make the expression have
 * smaller width (at the potential expense of greater height).
 */

struct tiledim {
	int	width;		/* Width of the expression */
	int	height;		/* Height of the expression */
	int	ascent;		/* Ascent above horizontal centerline */
	int	descent;	/* Descent below horizontal centerline */
};

/*
 * An object to recursive represent the physical dimensions of an
 * expression (in printed form), and all of its sub-expressions.
 */

struct grdim {
	char		op;		/* Operator */
	signed char	sign;		/* Sign for sub-expr */
	char		parens;		/* expr must be parenthesized */
	char		type;		/* Formatting method */
	char		parent_type;	/* Parent's formatting method	*/
					/* (same as type if no parent) */
	struct tiledim	maxtile;	/* Maximum width tile (achieved */
					/* by not wrapping anything) */
	struct tiledim	mintile;	/* Minimum width tile */
	union {
		struct grdim_list *	list;
		struct grdim *		unop;
		struct {
			struct grdim *	numer;
			struct grdim *	denom;
		} frac;
		struct {
			struct grdim *	base;
			int		power;
		} ipow;
		struct {
			struct grdim *		func;
			struct grdim_list *	args;
		} func;
		struct {
			struct grdim *	op;	/* operand to prefix/suffix */
			calib_int8u	pflags;	/* Prefix flags */
			calib_int8u	sflags;	/* Suffix flags */
			char *		prefix;
			char *		suffix;
		} fix;
		char *			zstr;
		const char *		var;
	} u;
};

struct grdim_list {
	struct grdim *		operand;
	struct grdim_list *	next;
};

#define GRDIM_OP_ADD		0
#define GRDIM_OP_UPLUS		1
#define GRDIM_OP_NEG		2
#define GRDIM_OP_MUL		3
#define GRDIM_OP_FRAC		4
#define GRDIM_OP_IPOW		5
#define GRDIM_OP_VAR		6
#define GRDIM_OP_Z		7
#define GRDIM_OP_FUNC		8
#define GRDIM_OP_FIX		9	/* Prefix and/or suffix */

/* Additional "fake" operators to determine parens. */
#define GRDIM_OP_NUMERATOR	10
#define GRDIM_OP_DENOMINATOR	11

#define TYPE_TILE		0	/* Fixed width tile */
#define TYPE_VTILE		1	/* Variable width tile that can */
					/* easily fit on one line */
#define TYPE_OPT		2	/* Optimize layout */
#define TYPE_FLAT		3	/* Display as "flat" text */

/* Prefix / suffix flags.  This indicate the cases under which the	*/
/* corresponding prefix or suffix are displayed.			*/

#define	FIX_2D			0x01	/* Display if parent and self	*/
					/* are both 2d			*/
#define	FIX_TRANS		0x02	/* Transition: display if	*/
					/* parent flat and self is 2d	*/
#define	FIX_FLAT		0x04	/* Display if parent and self	*/
					/* are both flat		*/

#define	FIX_ALL			0x07	/* All cases: always display */

/*
 * This object encapsulates all of the data that results from the
 * dynamic programming layout optimization.
 */

#define STORE_DP_TABLES		1

struct DPFormat {
	int			nlines;		/* Num "expression lines" */
	int			nkids;		/* Num operands / sub-exprs */
	struct grdim **		kids;		/* Array of sub-exprs */
	int *			numops;		/* Num operands on each line */
	int *			opwidths;	/* Formatted width of */
						/* each operand */
	struct tiledim *	linedims;	/* Dimensions of rectangle */
						/* for each line */
#if STORE_DP_TABLES
	int *			min_kid_width;
	int *			max_kid_width;
	struct tiledim *	kwdim;
	struct T_entry **	opstart;
	struct T_entry *	T;
	struct T_entry *	S;
	int **			L_partition;
	calib_int64u *		L_area;
#endif
};

/*
 * This object is the data structure used to represent each element in
 * the following dynamic programming table:
 *
 *	T [I, K, W]
 *
 * which represents the best (currently known) way (i.e., least total area)
 * of formatting K consecutive operands/sub-expressions (starting with
 * sub-expression I) on a single "expression line" of total width W.
 * The table entry gives:
 *
 *	- The dimensions of the entire rectangle (of width W), and
 *	- The width w_k of the K-th operand in this sequence.
 *
 * If K > 0, this implicitly (recursively) refers to table entry
 *
 *		T [I, K-1, W-1-w_k]
 *
 * that gives specifies w_{k-1}, etc. so that the best width of each
 * operand in the "expression line" can be extracted from the table.
 */

struct T_entry {
	struct tiledim	dim;	/* Dimensions of rectangle for line */
	int		w_k;	/* Formatted width of K-th operand */
};

/*
 * This object represents a dynamically expandable "canvas" that we can
 * plot onto using ASCII text.
 */

struct canvas {
	int		nlines;
	struct cline *	lines;
};

struct cline {
	int		len;		/* Length of text on this line. */
	char *		text;		/* Text for this line.		*/
};

/*
 * Info for keeping track of the current column when printing
 * flat, "tokenized" output.
 */

struct ColInfo {
	int		column;		/* Current column */
	int		numwhite;	/* Pending whitespace, not yet */
					/* written to output stream */
};

/*
 * Local Functions.
 */

static void		append_tiles (struct tiledim *		dst,
				      const struct tiledim *	left,
				      const struct tiledim *	right);
static calib_int64u	biased_line_area (
				const struct tiledim *	tp,
				int			maxwidth);
static void		clear_format (struct DPFormat * format);
static void		determine_display_width (struct PPInfo * ppip);
static struct grdim *	expr_dims (const struct calib_genrep * op);
static struct grdim *	expr_dim_func (const struct calib_genrep * op);
static struct grdim *	expr_dim_functor (const struct calib_genrep * op);
static struct grdim *	expr_dim_gather_prod (mpz_srcptr	  zval,
					      struct grdim_list * factors);
static struct grdim *	expr_dim_pow (const struct calib_genrep * op);
static struct grdim *	expr_dim_prod (const struct calib_genrep_list * factors);
static struct grdim *	expr_dim_sum (const struct calib_genrep_list * factors);
static void		find_best_layout (struct DPFormat *	result,
					  int			nkids,
					  struct grdim **	kids,
					  int			maxwidth,
					  int			suffixlen);
static int		flat_parens (int parent_op, int op);
static void		free_canvas (struct canvas * cp);
static struct tiledim *	get_mintile (struct grdim * dp);
static void		get_operand_array (int *		nkids_out,
					   struct grdim ***	array_out,
					   struct grdim_list *	list);
static void		grdim_classify (struct PPInfo *	ppip,
					struct grdim *	dp,
					int		parent_op);
static void		grdim_free (struct grdim * dp);
static void		grdim_list_free (struct grdim_list * dlp);
static void		grdim_recompute_mintile (struct grdim * dp);
static void		grdim_set_parent_type (struct grdim *	dp,
					       int		parent_type);
static struct grdim *	grdim_strip_signs (struct grdim * dp, mpz_ptr numfact);
static struct grdim *	grdim_unneg (struct grdim * op);
static struct grdim *	grdim_unplus (struct grdim * op);
static void		init_canvas (struct canvas * cp, int nlines);
static struct grdim *	make_grdim_fix (struct grdim *	op,
					calib_int8u	pflags,
					const char *	prefix,
					calib_int8u	sflags,
					const char *	suffix);
static struct grdim *	make_grdim_frac (struct grdim *	numer,
					 struct grdim *	denom);
static struct grdim *	make_grdim_int (int ival);
static struct grdim *	make_grdim_uplus (struct grdim * op);
static struct grdim *	make_grdim_neg (struct grdim * op);
static struct grdim *	make_grdim_pow (struct grdim * base, int pow);
static struct grdim *	make_grdim_prod (struct grdim_list * list);
static struct grdim *	make_grdim_q (mpq_srcptr qval);
static struct grdim *	make_grdim_si (calib_si_t zval);
static struct grdim *	make_grdim_var (const char * var);
static struct grdim *	make_grdim_z (mpz_srcptr zval);
static calib_bool	maxima_parens (int parent_op, int op);
static struct grdim *	new_grdim (void);
static struct grdim_list * new_grdim_list (struct grdim *	op,
					   struct grdim_list *	next);
static int		opt_parens (int parent_op, int op);
static void		plot_canvas (struct canvas *	cp,
				     const char *	text,
				     int		x,
				     int		y);
static void		plot_canvas_z (struct canvas *	cp,
				       mpz_srcptr	zval,
				       int		x,
				       int		y);
static void		plot_formatted_expr_line (
					struct canvas *		cp,
					struct DPFormat *	format,
					int			line,
					int			x,
					int			y);
static void		plot_grdim (struct canvas *	cp,
				    struct grdim *	dp,
				    int			x,
				    int			y);
static void		plot_vtile_width (struct canvas *	cp,
					  struct grdim *	dp,
					  int			x,
					  int			y,
					  int			width);
static void		prettyprint_internal (
					struct PPInfo *			ppip,
					const struct calib_genrep *	op);
static char *		string_cat2 (const char * s1, const char * s2);
static calib_int64u	tile_area (const struct tiledim * tp);
static int		tile_parens (int parent_op, int op);
static void		too_wide (struct PPInfo * ppip);
static void		update_tile_heights (struct tiledim *		dst,
					     const struct tiledim *	src);
static void		vtile_dims (struct tiledim *	dims,
				    struct grdim *	dp,
				    int			width);
static void		write_canvas (struct PPInfo *	ppip,
				      struct canvas *	cp);
static void		write_formatted (struct PPInfo *	ppip,
					 struct DPFormat *	format,
					 const char *		prefix,
					 const char *		suffix);
static void		write_grdim (struct PPInfo *	ppip,
				     struct grdim *	dp,
				     const char *	prefix,
				     const char *	suffix);
static void		write_grdim_flat (
					struct PPInfo *	ppip,
					struct grdim *	dp,
					const char *	prefix,
					const char *	suffix);
static void		write_grdim_maxima (
					struct PPInfo *	 ppip,
					struct ColInfo * cip,
					struct grdim *	 dp,
					int		 parent_op);
static void		write_grdim_opt (
					struct PPInfo *	ppip,
					struct grdim *	dp,
					const char *	prefix,
					const char *	suffix);
static void		write_grdim_tile (
					struct PPInfo *	ppip,
					struct grdim *	dp,
					const char *	prefix,
					const char *	suffix);
static void		write_grdim_tile_line (
					struct PPInfo *	ppip,
					struct grdim *	dp,
					const char *	prefix,
					const char *	suffix);
static void		write_grdim_top_level (
					struct PPInfo *	ppip,
					struct grdim *	dp);
static void		write_token (struct PPInfo *	ppip,
				     struct ColInfo *	cip,
				     int		ws1,
				     const char *	token,
				     int		ws2);

/*
 * Display the given expression (expressed in general representation)
 * to stdout using the default maximum width.
 */

	void
calib_genrep_prettyprint (

const struct calib_genrep *	op	/* IN: expression to display */
)
{
struct PPInfo		ppinfo;

	ppinfo.flags	= 0;
	ppinfo.fp	= stdout;
	ppinfo.maxwidth	= 0;

	prettyprint_internal (&ppinfo, op);
}

/*
 * Display the given expression (expressed in general representation)
 * to the given stream using the default maximum width.
 */

	void
calib_genrep_prettyprint_file (

FILE *				fp,	/* IN/OUT: stream to write to */
const struct calib_genrep *	op	/* IN: expression to display */
)
{
struct PPInfo		ppinfo;

	ppinfo.flags	= 0;
	ppinfo.fp	= fp;
	ppinfo.maxwidth	= 0;

	prettyprint_internal (&ppinfo, op);
}

/*
 * Display the given expression (expressed in general representation)
 * to stdout using the given maximum number of columns.
 */

	void
calib_genrep_prettyprint_width (

const struct calib_genrep *	op,	/* IN: expression to display */
int				width	/* IN: maximum width of display */
)
{
struct PPInfo		ppinfo;

	ppinfo.flags	= 0;
	ppinfo.fp	= stdout;
	ppinfo.maxwidth	= width;

	prettyprint_internal (&ppinfo, op);
}

/*
 * Display the given expression (expressed in general representation)
 * to the given stream using the given maximum number of columns.
 */

	void
calib_genrep_prettyprint_file_width (

FILE *				fp,	/* IN/OUT: stream to write to */
const struct calib_genrep *	op,	/* IN: expression to display */
int				width	/* IN: maximum width of display */
)
{
struct PPInfo		ppinfo;

	ppinfo.flags	= 0;
	ppinfo.fp	= fp;
	ppinfo.maxwidth	= width;

	prettyprint_internal (&ppinfo, op);
}

/*
 * Write the expression in syntax allowing it to be read back in
 * by maxima (or similar packages).
 */

	void
calib_genrep_print_maxima (

FILE *				fp,	/* IN/OUT: stream to write to */
const struct calib_genrep *	op,	/* IN: expression to display */
int				width	/* IN: maximum width of display */
)
{
struct PPInfo		ppinfo;

	ppinfo.flags	= FLAG_FORMAT_MAXIMA;
	ppinfo.fp	= fp;
	ppinfo.maxwidth	= width;

	prettyprint_internal (&ppinfo, op);
}

/*
 * Display the given expression (expressed in general representation)
 * to the given stream using the given maximum number of columns.
 */

	static
	void
prettyprint_internal (

struct PPInfo *			ppip,	/* IN/OUT: Global prettyprint info */
const struct calib_genrep *	op	/* IN: expression to display */
)
{
int			parent_op;
struct calib_genrep *	op2;
struct grdim *		dp;
struct ColInfo		col_info;

	determine_display_width (ppip);

	/* Copy the given general expression, performing several simple	*/
	/* but important simplifications.				*/
	op2 = _calib_genrep_copy_simp (op);

	/* Get recursive sub-expression dimensions, assuming infinite width. */
	dp = expr_dims (op2);

	calib_genrep_free (op2);

	/* Remove leading unary plus. */
	dp = grdim_unplus (dp);

	if ((ppip -> flags & FLAG_FORMAT_MAXIMA) NE 0) {
		parent_op = -1;
		memset (&col_info, 0, sizeof (col_info));
		write_grdim_maxima (ppip, &col_info, dp, parent_op);
		write_token (ppip, &col_info, 0, ";", 0);
		if (col_info.column > 0) {
			fprintf (ppip -> fp, "\n");
		}
	}
	else {
		parent_op = -1;
		grdim_classify (ppip, dp, parent_op);

		/* Set all of the parent_type members. */
		grdim_set_parent_type (dp, dp -> parent_type);

		grdim_recompute_mintile (dp);

		write_grdim_top_level (ppip, dp);
	}

	grdim_free (dp);
}

/*
 * If the user has not specified a particular display width, see if we
 * can determine the proper display width dynamically.  Otherwise use
 * the default.
 */

	static
	void
determine_display_width (

struct PPInfo *		ppip		/* IN/OUT: Global prettyprint info */
)
{
int		maxwidth;
int		fd;

	maxwidth = ppip -> maxwidth;
	if (maxwidth > 0) {
		/* User has given a display width. */
		return;
	}

	/* See if output is to a tty. */
	fd = fileno (ppip -> fp);
	if (NOT isatty (fd)) {
		/* Not a tty.  Use default. */
		ppip -> maxwidth = DEFAULT_MAX_WIDTH;
		return;
	}

	/* This is the preferred interface to use. */
#ifdef TIOCGWINSZ
	{	struct winsize	ws;
		int i = ioctl (fd, TIOCGWINSZ, &ws);
		if (i EQ 0) {
			/* Do not attempt to use the last column. */
			int w = ws.ws_col - 1;
			if (w > 0) {
				ppip -> maxwidth = w;
				return;
			}
		}
	}
#endif

	/* The next interface to try. */
#ifdef TIOCGSIZE
	{	struct ttysize	ts;
		int i = ioctl (fd, TIOCGSIZE, &ts);
		if (i EQ 0) {
			/* Do not attempt to use the last column. */
			if w = ts.ts_cols - 1;
			if (w > 0) {
				ppip -> maxwidth = w;
				return;
			}
		}
	}
#endif

	/* Use default. */
	ppip -> maxwidth = DEFAULT_MAX_WIDTH;
}

/*
 * Recursively construct an annotated form of the given expression that
 * gives the dimensions of each sub-expression when printed on an
 * infinitely tall and infinitely wide sheet of paper.
 */

	static
	struct grdim *
expr_dims (

const struct calib_genrep *	op	/* IN: expression to size up */
)
{
struct grdim *		dp;

	switch (op -> op) {
	case CALIB_GENREP_OP_ADD:
		dp = expr_dim_sum (op -> u.list);
		break;

	case CALIB_GENREP_OP_MUL:
		dp = expr_dim_prod (op -> u.list);
		break;

	case CALIB_GENREP_OP_SI:
		dp = make_grdim_si (op -> u.siop);
		break;

	case CALIB_GENREP_OP_IPOW:
		dp = expr_dim_pow (op);
		break;

	case CALIB_GENREP_OP_VAR:
		dp = make_grdim_uplus (make_grdim_var (op -> u.var));
		break;

	case CALIB_GENREP_OP_Z:
		dp = make_grdim_z (op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		dp = make_grdim_q (op -> u.qop);
		break;

	case CALIB_GENREP_OP_FUNC:
		dp = make_grdim_uplus (expr_dim_func (op));
		break;

	default:
		CALIB_FATAL_ERROR;
		dp = make_grdim_int (0);
		break;
	}

	return (dp);
}

/*
 * Recursively construct a grdim node that properly represents the sum
 * of the given list of terms.
 */

	static
	struct grdim *
expr_dim_sum (

const struct calib_genrep_list *	terms	/* IN: list of terms */
)
{
int					first, space;
const struct calib_genrep *		gp1;
const struct calib_genrep_list *	lp;
struct grdim *				dp;
struct grdim_list *			dlp;
struct grdim_list *			root;
struct grdim_list **			hookp;
struct tiledim				maxtile, mintile;

	root = NULL;
	hookp = &root;

	for (lp = terms; lp NE NULL; lp = lp -> next) {
		gp1 = lp -> operand;
		dp = expr_dims (gp1);
		dlp = new_grdim_list (dp, NULL);
		*hookp = dlp;
		hookp = &(dlp -> next);
	}

	if (root EQ NULL) {
		dp = make_grdim_int (0);
	}
	else if (root -> next EQ NULL) {
		dp = root -> operand;
		free (root);
	}
	else {
		/* Compute dimensions. */
		memset (&maxtile, 0, sizeof (maxtile));
		memset (&mintile, 0, sizeof (mintile));
		first = TRUE;
		space = 0;
		for (dlp = root; dlp NE NULL; dlp = dlp -> next) {
			dp = dlp -> operand;
			CALIB_FATAL_ERROR_IF ((dp -> op NE GRDIM_OP_UPLUS) AND
					      (dp -> op NE GRDIM_OP_NEG));

			/* Each term should have its own unary "+" or	*/
			/* unary "-" wrapped around it.			*/
			if (first) {
				/* Remove leading "+" from first term. */
				dp = grdim_unplus (dp);
				dlp -> operand = dp;
			}

			maxtile.width	+= (space + dp -> maxtile.width);
			update_tile_heights (&maxtile, &(dp -> maxtile));
			mintile.width	+= (space + dp -> mintile.width);
			update_tile_heights (&mintile, &(dp -> mintile));
			first = FALSE;
			/* Prefix rest of terms with " " (each term has	*/
			/* an explicit unary + or - operator on it).	*/
			space = 1;
		}
		dp = new_grdim ();
		dp -> op	= GRDIM_OP_ADD;
		dp -> u.list	= root;
		dp -> maxtile	= maxtile;
		dp -> mintile	= mintile;
	}

	return (dp);
}

/*
 * Recursively construct a grdim node that properly represents the product
 * of the given list of factors.
 */

	static
	struct grdim *
expr_dim_prod (

const struct calib_genrep_list *	factors	/* IN: list of factors */
)
{
int					pow, sign;
const struct calib_genrep *		gp1;
const struct calib_genrep_list *	lp;
struct grdim *				dp;
struct grdim *				dp2;
struct grdim *				dp3;
struct grdim_list *			dlp;
struct grdim_list *			numer;
struct grdim_list *			denom;
struct grdim_list **			nhookp;
struct grdim_list **			dhookp;
mpz_t					num_const;
mpz_t					den_const;

	/* Initialize numerator and denominator factor lists to empty. */
	mpz_init_set_ui (num_const, 1);
	mpz_init_set_ui (den_const, 1);
	numer = NULL;		nhookp = &numer;
	denom = NULL;		dhookp = &denom;

	/* Pass 1.  Split out numerator vs. denominator factors. */
	for (lp = factors; lp NE NULL; lp = lp -> next) {
		gp1 = lp -> operand;
		if (gp1 -> op EQ CALIB_GENREP_OP_Z) {
			/* The constant factor should be first. */
			mpz_mul (num_const, num_const, gp1 -> u.zop);
		}
		else if (gp1 -> op EQ CALIB_GENREP_OP_Q) {
			/* The constant factor (if any) should	*/
			/* be first.  Split into separate	*/
			/* numerator and denominator portions.	*/
			mpz_mul (num_const, num_const, mpq_numref (gp1 -> u.qop));
			mpz_mul (den_const, den_const, mpq_denref (gp1 -> u.qop));
		}
		else if (gp1 -> op EQ CALIB_GENREP_OP_IPOW) {
			pow = gp1 -> u.ipow.power;
			CALIB_FATAL_ERROR_IF ((pow EQ 0) OR (pow EQ 1));
			if (pow > 0) {
				dp2 = expr_dims (gp1);
				dp2 = grdim_strip_signs (dp2, num_const);
				dlp = new_grdim_list (dp2, NULL);
				*nhookp = dlp;
				nhookp = &(dlp -> next);
			}
			else {
				/* Factor belongs in denominator. */
				if (pow EQ -1) {
					/* Remove power. */
					dp2 = expr_dims (gp1 -> u.ipow.base);
				}
				else {
					struct calib_genrep	tmp;
					tmp.op = CALIB_GENREP_OP_IPOW;
					tmp.u.ipow.base  = gp1 -> u.ipow.base;
					tmp.u.ipow.power = - pow;
					dp2 = expr_dims (&tmp);
				}
				dp2 = grdim_strip_signs (dp2, num_const);
				dlp = new_grdim_list (dp2, NULL);
				*dhookp = dlp;
				dhookp = &(dlp -> next);
			}
		}
		else {
			/* Put this factor into numerator. */
			dp2 = expr_dims (gp1);
			dp2 = grdim_strip_signs (dp2, num_const);
			dlp = new_grdim_list (dp2, NULL);
			*nhookp = dlp;
			nhookp = &(dlp -> next);
		}
	}

	if (mpz_sgn (den_const) < 0) {
		/* Force negative sign into the numerator. */
		mpz_neg (num_const, num_const);
		mpz_neg (den_const, den_const);
	}
	sign = mpz_sgn (num_const);
	mpz_abs (num_const, num_const);

	/* We should not have a constant factor of zero! */
	CALIB_FATAL_ERROR_IF (sign EQ 0);

	if (denom EQ NULL) {
		if (mpz_cmp_ui (den_const, 1) EQ 0) {
			/* Denominator is 1 (empty).  No fraction needed. */
			dp = expr_dim_gather_prod (num_const, numer);
			dp = grdim_strip_signs (dp, NULL);
		}
		else {
			/* Denominator is just an integer. */
			dp2 = expr_dim_gather_prod (num_const, numer);
			dp2 = grdim_strip_signs (dp2, NULL);

			dp3 = make_grdim_z (den_const);
			dp3 = grdim_strip_signs (dp3, NULL);

			dp = make_grdim_frac (dp2, dp3);
		}
	}
	else {
		dp2 = expr_dim_gather_prod (num_const, numer);
		dp2 = grdim_strip_signs (dp2, NULL);

		dp3 = expr_dim_gather_prod (den_const, denom);
		dp3 = grdim_strip_signs (dp3, NULL);

		dp = make_grdim_frac (dp2, dp3);
	}

	if (sign < 0) {
		/* Wrap a neg(...) around the product. */
		dp = make_grdim_neg (dp);
	}
	else {
		/* Wrap a +(...) around the product. */
		dp = make_grdim_uplus (dp);
	}

	mpz_clear (den_const);
	mpz_clear (num_const);

	return (dp);
}

/*
 * Gather the given factors into a proper set of grdim nodes.
 */

	static
	struct grdim *
expr_dim_gather_prod (

mpz_srcptr		zval,	/* IN: constant factor */
struct grdim_list *	factors	/* IN: list of other factors */	
)
{
int			sign;
struct grdim *		dp;
struct grdim *		dp2;
mpz_t			VALUE;

	mpz_init (VALUE);
	mpz_abs (VALUE, zval);

	sign = mpz_sgn (zval);

	if (factors EQ NULL) {
		/* Constant factor is the only one. */
		dp = make_grdim_z (VALUE);
		dp = grdim_unplus (dp);
	}
	else if (mpz_cmp_ui (VALUE, 1) EQ 0) {
		/* Constant factor is +1 or -1.  Move into sign. */
		if (factors -> next EQ NULL) {
			/* Only one factor. */
			dp = factors -> operand;
			free (factors);
		}
		else {
			dp = make_grdim_prod (factors);
		}
	}
	else {
		/* Combine constant factor with others. */
		dp2 = make_grdim_z (VALUE);
		dp2 = grdim_unplus (dp2);
		factors = new_grdim_list (dp2, factors);
		dp = make_grdim_prod (factors);
	}

	if (sign < 0) {
		dp = make_grdim_neg (dp);
	}
	else {
		dp = make_grdim_uplus (dp);
	}

	mpz_clear (VALUE);

	return (dp);
}

/*
 * Recursively construct a grdim node that properly represents the
 * given exponent.
 */

	static
	struct grdim *
expr_dim_pow (

const struct calib_genrep *	op	/* IN: power expression to process */
)
{
int				pow;
const struct calib_genrep *	base;
struct grdim *			dp;
struct grdim *			dp2;

	base = op -> u.ipow.base;
	pow  = op -> u.ipow.power;

	CALIB_FATAL_ERROR_IF ((pow EQ 0) OR (pow EQ 1));

	dp2 = expr_dims (base);
	dp2 = grdim_unplus (dp2);

	if (pow EQ -1) {
		/* Special handling for reciprocals. */
		dp = make_grdim_frac (make_grdim_int (1), dp2);
		dp = make_grdim_uplus (dp);
		return (dp);
	}

	if ((dp2 -> op EQ GRDIM_OP_ADD) OR
	    (dp2 -> op EQ GRDIM_OP_NEG) OR
	    (dp2 -> op EQ GRDIM_OP_MUL)) {
		/* Must parenthesize (a+b)^2, (-a)^2 and (a*b)^2. */
		dp2 -> parens = TRUE;
		dp2 -> maxtile.width += 2;
	}

	if (pow > 0) {
		dp = make_grdim_pow (dp2, pow);
	}
	else {
		dp2 = make_grdim_pow (dp2, - pow);
		dp = make_grdim_frac (make_grdim_int (1), dp2);
	}

	dp = make_grdim_uplus (dp);

	return (dp);
}

/*
 * Recursively construct a grdim node that properly represents the
 * given function invocation.
 */

	static
	struct grdim *
expr_dim_func (

const struct calib_genrep *	op	/* IN: Function to process */
)
{
int				nargs, space;
struct calib_genrep *		fgp;
struct calib_genrep_list *	lp;
struct calib_genrep_list *	lp2;
struct grdim *			dp;
struct grdim *			fdp;
struct grdim_list *		root;
struct grdim_list **		hookp;
struct grdim_list *		dlp;
const char *			prefix;
const char *			suffix;
struct tiledim			maxtile, mintile;


	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR
			      (op -> op NE CALIB_GENREP_OP_FUNC));

	fgp	= op -> u.func.func;
	lp	= op -> u.func.args;

	CALIB_FATAL_ERROR_IF (fgp EQ NULL);

	fdp = expr_dim_functor (fgp);

	if (lp EQ NULL) {
		/* No arguments.  Apply " ()" to the raw function	*/
		/* (no GRDIM_OP_FUNC "function application").  This is	*/
		/* a simpler way to do it.				*/
		fdp = make_grdim_fix (fdp, 0, NULL, FIX_ALL, " ()");

		return (fdp);
	}

	/* Process the argument list.  We prefix the first argument	*/
	/* with a left paren.  All args receive a suffix: a comma for	*/
	/* leading args; right paren for final arg.  This gets all of	*/
	/* the dimensions correct with less fuss during formatting.	*/
	nargs = 0;
	root = NULL;
	hookp = &root;
	for (lp2 = lp; lp2 NE NULL; lp2 = lp2 -> next) {
		dp = expr_dims (lp2 -> operand);
		/* Remove any unary plus. */
		dp = grdim_unplus (dp);

		prefix	= (nargs EQ 0) ? LPAREN_STRING : NULL;
		suffix	= (lp2 -> next NE NULL) ? "," : RPAREN_STRING;

		dp = make_grdim_fix (dp,
				     FIX_ALL,
				     prefix,
				     FIX_ALL,
				     suffix);

		dlp = new_grdim_list (dp, NULL);
		*hookp = dlp;
		hookp = &(dlp -> next);
		++nargs;
	}

	/* Compute dimensions. */
	maxtile	= fdp -> maxtile;
	mintile	= fdp -> mintile;
	space = 1;
	for (dlp = root; dlp NE NULL; dlp = dlp -> next) {
		dp = dlp -> operand;

		maxtile.width	+= (space + dp -> maxtile.width);
		update_tile_heights (&maxtile, &(dp -> maxtile));
		mintile.width	+= (space + dp -> mintile.width);
		update_tile_heights (&mintile, &(dp -> mintile));
	}

	dp = new_grdim ();
	dp -> op		= GRDIM_OP_FUNC;
	dp -> u.func.func	= fdp;
	dp -> u.func.args	= root;
	dp -> maxtile		= maxtile;
	dp -> mintile		= mintile;

	return (dp);
}

/*
 * Construct a grdim node representing the given functor.
 */

	static
	struct grdim *
expr_dim_functor (

const struct calib_genrep *	op	/* IN: Functor to process */
)
{
int			power, elen, vlen, idx;
struct calib_genrep *	base;
struct grdim *		dp;
struct grdim *		dp2;
const char *		fname;
char *			var;
struct tiledim		maxtile, mintile;
char			buf [32];

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	switch (op -> op) {
	case CALIB_GENREP_OP_IPOW:
		base	= op -> u.ipow.base;
		power	= op -> u.ipow.power;
		dp2 = expr_dim_functor (base);
		sprintf (buf, "%d", power);
		elen = strlen (buf);

		maxtile	= dp2 -> maxtile;
		mintile = dp2 -> mintile;

		maxtile.width	+= elen;
		maxtile.ascent	= _MAX (maxtile.ascent, 1);
		maxtile.height	= 1 + maxtile.ascent + maxtile.descent;

		mintile.width	+= elen;
		mintile.ascent	= _MAX (mintile.ascent, 1);
		mintile.height	= 1 + mintile.ascent + mintile.descent;

		dp = new_grdim ();
		dp -> op		= GRDIM_OP_IPOW;
		dp -> maxtile		= maxtile;
		dp -> mintile		= mintile;
		dp -> u.ipow.base	= dp2;
		dp -> u.ipow.power	= power;
		break;

	case CALIB_GENREP_OP_VAR:
		var	= strdup (op -> u.var);
		vlen	= strlen (var);

		dp = new_grdim ();
		dp -> op		= GRDIM_OP_VAR;
		dp -> u.var		= var;
		dp -> maxtile.width	= vlen;
		dp -> mintile.width	= vlen;
		break;

	case CALIB_GENREP_OP_SI:
		idx	= op -> u.siop;
		fname	= calib_genrep_builtin_func_index_to_name (idx);
		CALIB_FATAL_ERROR_IF (fname EQ NULL);

		var	= strdup (fname);
		vlen	= strlen (var);

		dp = new_grdim ();
		dp -> op		= GRDIM_OP_VAR;
		dp -> u.var		= var;
		dp -> maxtile.width	= vlen;
		dp -> mintile.width	= vlen;
		break;

	default:
		CALIB_FATAL_ERROR;
		dp = NULL;
	}

	return (dp);
}

/*
 * Make a grdim node for unary "+" of a sub-expression.
 */

	static
	struct grdim *
make_grdim_uplus (

struct grdim *		op	/* IN: grdim of sub-expression */
)
{
struct grdim *		dp;

	if (op -> op EQ GRDIM_OP_UPLUS) {
		return (op);
	}

	dp = new_grdim ();
	dp -> op	= GRDIM_OP_UPLUS;
	dp -> u.unop	= op;
	dp -> sign	= '-';
	dp -> maxtile	= op -> maxtile;
	dp -> mintile	= op -> mintile;
	dp -> maxtile.width += 2;
	dp -> mintile.width += 2;

	return (dp);
}

/*
 * Strip all unary '+' and '-' operations from the given grdim
 * (freeing them as you go).  Make the corresponding sign corrections
 * to the given numeric factor.
 */

	static
	struct grdim *
grdim_strip_signs (

struct grdim *		dp,		/* IN: grdim to strip signs from */
mpz_ptr			numfact		/* IN: numeric factor to correct */
					/*     sign of */
)
{
	for (;;) {
		if (dp -> op EQ GRDIM_OP_UPLUS) {
			/* Strip unary '+'. */
			dp = grdim_unplus (dp);
		}
		else if (dp -> op EQ GRDIM_OP_NEG) {
			/* Strip unary '-' and flip sign. */
			dp = grdim_unneg (dp);
			if (numfact NE NULL) {
				mpz_neg (numfact, numfact);
			}
		}
		else {
			break;
		}
	}

	return (dp);
}

/*
 * If the given grdim is a unary "+", strip it off (and free it),
 * leaving only the un-adorned sub-expression.  Otherwise, return
 * the given sub-expression unmodified.
 */

	static
	struct grdim *
grdim_unplus (

struct grdim *		op	/* IN: grdim to "unplus" */
)
{
struct grdim *		dp;

	if (op -> op NE GRDIM_OP_UPLUS) {
		return (op);
	}
	dp = op -> u.unop;
	memset (op, 0, sizeof (*op));
	free (op);
	return (dp);
}

/*
 * If the given grdim is a unary "-", strip it off (and free it),
 * leaving only the un-adorned sub-expression.  Otherwise, return
 * the given sub-expression unmodified.
 */

	static
	struct grdim *
grdim_unneg (

struct grdim *		op	/* IN: grdim to "unplus" */
)
{
struct grdim *		dp;

	if (op -> op NE GRDIM_OP_NEG) {
		return (op);
	}
	dp = op -> u.unop;
	memset (op, 0, sizeof (*op));
	free (op);
	return (dp);
}

/*
 * Make a grdim node for the negation of a sub-expression.
 */

	static
	struct grdim *
make_grdim_neg (

struct grdim *		op	/* IN: grdim of sub-expression */
)
{
struct grdim *		dp;

	dp = new_grdim ();
	dp -> op	= GRDIM_OP_NEG;
	dp -> u.unop	= op;
	dp -> sign	= '-';
	dp -> maxtile	= op -> maxtile;
	dp -> mintile	= op -> mintile;
	dp -> maxtile.width += 2;
	dp -> mintile.width += 2;

	return (dp);
}

/*
 * Make a grdim node for a product of 2 or more factors.
 */

	static
	struct grdim *
make_grdim_prod (

struct grdim_list *	list	/* IN: list of grdim's of factors */
)
{
int			n;
int			space;
struct grdim *		dp;
struct grdim_list *	dlp;
struct tiledim		maxtile, mintile;

	memset (&maxtile, 0, sizeof (maxtile));
	memset (&mintile, 0, sizeof (mintile));
	space = 0;
	n = 0;
	for (dlp = list; dlp NE NULL; dlp = dlp -> next) {
		dp = dlp -> operand;
		if (dp -> op EQ GRDIM_OP_ADD) {
			/* Must parenthesize sums inside of products. */
			dp -> parens = TRUE;
			dp -> maxtile.width += 2;
			dp -> mintile.width += 2;
		}
		/* First factor does not get a separating space, rest do. */
		maxtile.width += (space + dp -> maxtile.width);
		update_tile_heights (&maxtile, &(dp -> maxtile));
		mintile.width += (space + dp -> mintile.width);
		update_tile_heights (&mintile, &(dp -> mintile));
		space	= 1;
		++n;
	}
	CALIB_FATAL_ERROR_IF (n < 2);

	dp = new_grdim ();
	dp -> op	= GRDIM_OP_MUL;
	dp -> u.list	= list;
	dp -> sign	= '+';
	dp -> maxtile	= maxtile;
	dp -> mintile	= mintile;

	return (dp);
}

/*
 * Make a grdim node for a fraction bar.
 */

	static
	struct grdim *
make_grdim_frac (

struct grdim *		numer,		/* IN: numerator of fraction */
struct grdim *		denom		/* IN: denominator of fration */
)
{
struct grdim *		dp;
struct tiledim		maxtile, mintile;

	dp = new_grdim ();

	dp -> op = GRDIM_OP_FRAC;
	dp -> u.frac.numer = numer;
	dp -> u.frac.denom = denom;

	maxtile.width	= _MAX (numer -> maxtile.width, denom -> maxtile.width);
	maxtile.ascent	= numer -> maxtile.height;
	maxtile.descent	= denom -> maxtile.height;
	maxtile.height	= 1 + maxtile.ascent + maxtile.descent;

	mintile.width	= _MAX (numer -> mintile.width, denom -> mintile.width);
	mintile.ascent	= numer -> mintile.height;
	mintile.descent	= denom -> mintile.height;
	mintile.height	= 1 + mintile.ascent + mintile.descent;

	dp -> sign	= numer -> sign;
	dp -> maxtile	= maxtile;
	dp -> mintile	= mintile;

	return (dp);
}

/*
 * Make a grdim node for the given integer power.
 */

	static
	struct grdim *
make_grdim_pow (

struct grdim *		base,		/* IN: item to raise to power */
int			pow		/* IN: power to raise to */
)
{
int			esize;
struct grdim *		dp;
struct tiledim		maxtile, mintile;
char			buf [32];

	CALIB_FATAL_ERROR_IF (pow < 2);

	sprintf (buf, "%d", pow);
	esize = strlen (buf);

	if (base -> op EQ GRDIM_OP_FUNC) {
		/* We have f(x)^n.  Format these specially as f^n(x).	*/
		dp = base -> u.func.func;
		dp = make_grdim_pow (dp, pow);
		base -> u.func.func = dp;

		dp = base;

		/* Repair dimensions. */
		maxtile	= dp -> maxtile;
		mintile	= dp -> mintile;

		maxtile.width	+= esize;
		maxtile.ascent	= _MAX (maxtile.ascent, 1);
		maxtile.height	= 1 + maxtile.ascent + maxtile.descent;

		mintile.width	+= esize;
		mintile.ascent	= _MAX (mintile.ascent, 1);
		mintile.height	= 1 + mintile.ascent + mintile.descent;

		dp -> maxtile	= maxtile;
		dp -> mintile	= mintile;
	}
	else {
		dp = new_grdim ();
		dp -> op		= GRDIM_OP_IPOW;
		dp -> u.ipow.base	= base;
		dp -> u.ipow.power	= pow;

		maxtile.width	= base -> maxtile.width + esize;
		maxtile.ascent	= _MAX (base -> maxtile.ascent, 1);
		maxtile.descent	= base -> maxtile.descent;
		maxtile.height	= 1 + maxtile.ascent + maxtile.descent;

		mintile.width	= base -> mintile.width + esize;
		mintile.ascent	= _MAX (base -> mintile.ascent, 1);
		mintile.descent	= base -> mintile.descent;
		mintile.height	= 1 + mintile.ascent + mintile.descent;

		dp -> maxtile	= maxtile;
		dp -> mintile	= mintile;
	}

	return (dp);
}

/*
 * Make a grdim node for the given variable name.
 */

	static
	struct grdim *
make_grdim_var (

const char *		var		/* IN: variable name */
)
{
int			len;
struct grdim *		dp;

	dp = new_grdim ();

	len = strlen (var);

	dp -> op	= GRDIM_OP_VAR;
	dp -> u.var	= strdup (var);
	dp -> maxtile.width	= len;
	dp -> mintile.width	= len;

	return (dp);
}

/*
 * Make a grdim subtree to hold the given rational value.
 */

	static
	struct grdim *
make_grdim_q (

mpq_srcptr		qval	/* IN: rational value */
)
{
int			sign;
struct grdim *		dp;
struct grdim *		dp2;
struct grdim *		dp3;
mpq_t			abs_qval;

	sign = mpq_sgn (qval);

	mpq_init (abs_qval);
	mpq_abs (abs_qval, qval);

	if (mpz_cmp_ui (mpq_denref (abs_qval), 1) EQ 0) {
		/* Value is integral. */
		dp = make_grdim_z (mpq_numref (abs_qval));
		dp = grdim_strip_signs (dp, NULL);
	}
	else {
		/* Need numerator/denominator. */
		dp2 = make_grdim_z (mpq_numref (abs_qval));
		dp2 = grdim_strip_signs (dp2, NULL);

		dp3 = make_grdim_z (mpq_denref (abs_qval));
		dp3 = grdim_strip_signs (dp3, NULL);

		dp = make_grdim_frac (dp2, dp3);
	}
	if (sign < 0) {
		dp = make_grdim_neg (dp);
	}
	else {
		dp = make_grdim_uplus (dp);
	}

	mpq_clear (abs_qval);

	return (dp);
}

/*
 * Make a grdim node for the given "int" value.
 */

	static
	struct grdim *
make_grdim_int (

int		ival		/* IN: integer value */
)
{
struct grdim *		dp;
mpz_t			zval;

	mpz_init_set_si (zval, ival);
	dp = make_grdim_z (zval);
	mpz_clear (zval);

	return (dp);
}

/*
 * Make a grdim node to contain the given integer value.
 */

	static
	struct grdim *
make_grdim_si (

calib_si_t	zval		/* IN: integer value to use */
)
{
int		ndigits;
struct grdim *	dp;
char *		p;
char *		string;
char		buf [32];

	sprintf (buf, "%ld", zval);

	p = &buf [0];
	if (*p EQ '-') {
		++p;
	}
	ndigits = strlen (p);

	string	= CALIB_NEWA (ndigits + 1, char);
	strcpy (string, p);

	dp = new_grdim ();

	dp -> op	= GRDIM_OP_Z;
	dp -> sign	= (zval >= 0) ? '+' : '-';
	dp -> u.zstr	= string;

	dp -> maxtile.height	= 1;
	dp -> maxtile.width	= ndigits;
	dp -> maxtile.ascent	= 0;
	dp -> maxtile.descent	= 0;

	dp -> mintile.height	= _MAX (ndigits - 1, 1);
	dp -> mintile.width	= (ndigits <= 1) ? 1 : 2;
	dp -> mintile.ascent	= dp -> mintile.height - 1;
	dp -> mintile.descent	= 0;

	if (zval < 0) {
		dp = make_grdim_neg (dp);
	}
	else {
		dp = make_grdim_uplus (dp);
	}

	return (dp);
}

/*
 * Make a grdim node to contain the given integer value.
 */

	static
	struct grdim *
make_grdim_z (

mpz_srcptr	zval		/* IN: integer value to use */
)
{
int		ndigits;
struct grdim *	dp;
char *		string;

	/* Convert |zval| into string of decimal digits. */
	ndigits = calib_genrep_convert_abs_Z_to_decimal_string (zval, &string);

	dp = new_grdim ();

	dp -> op	= GRDIM_OP_Z;
	dp -> sign	= (mpz_sgn (zval) >= 0) ? '+' : '-';
	dp -> u.zstr	= string;

	dp -> maxtile.height	= 1;
	dp -> maxtile.width	= ndigits;
	dp -> maxtile.ascent	= 0;
	dp -> maxtile.descent	= 0;

	dp -> mintile.height	= _MAX (ndigits - 1, 1);
	dp -> mintile.width	= (ndigits <= 1) ? 1 : 2;
	dp -> mintile.ascent	= dp -> mintile.height - 1;
	dp -> mintile.descent	= 0;

	if (mpz_sgn (zval) < 0) {
		dp = make_grdim_neg (dp);
	}
	else {
		dp = make_grdim_uplus (dp);
	}

	return (dp);
}

/*
 * Make a prefix/suffix grdim node.
 */

	static
	struct grdim *
make_grdim_fix (

struct grdim *		op,		/* IN: Expression to prefix/suffix */
calib_int8u		pflags,		/* IN: When to display prefix */
const char *		prefix,		/* IN: Prefix string (or NULL) */
calib_int8u		sflags,		/* IN: When to display suffix */
const char *		suffix		/* IN: Suffix string (or NULL) */
)
{
int			plen, slen;
struct grdim *		dp;

	dp = new_grdim ();

	plen	= 0;
	slen	= 0;
	if (prefix NE NULL) {
		plen = strlen (prefix);
	}
	if (suffix NE NULL) {
		slen = strlen (suffix);
	}

	dp -> op		= GRDIM_OP_FIX;
	dp -> maxtile		= op -> maxtile;
	dp -> mintile		= op -> mintile;
	dp -> u.fix.op		= op;
	dp -> u.fix.pflags	= pflags;
	dp -> u.fix.sflags	= sflags;
	dp -> u.fix.prefix	= (prefix EQ NULL) ? NULL : strdup (prefix);
	dp -> u.fix.suffix	= (suffix EQ NULL) ? NULL : strdup (suffix);

	dp -> maxtile.width	+= (plen + slen);
	dp -> mintile.width	+= (plen + slen);

	return (dp);
}

/*
 * Create a new "struct grdim" node.
 */

	static
	struct grdim *
new_grdim (void)

{
struct grdim *		dp;

	dp = CALIB_NEW (struct grdim);
	memset (dp, 0, sizeof (*dp));

	dp -> sign	= '+';
	dp -> parens	= FALSE;

	dp -> maxtile.width	= 0;
	dp -> maxtile.height	= 1;
	dp -> maxtile.ascent	= 0;
	dp -> maxtile.descent	= 0;

	dp -> mintile.width	= 0;
	dp -> mintile.height	= 1;
	dp -> mintile.ascent	= 0;
	dp -> mintile.descent	= 0;

	return (dp);
}

/*
 * Create a new "struct grdim_list" node.
 */

	static
	struct grdim_list *
new_grdim_list (

struct grdim *		op,		/* IN: first operand */
struct grdim_list *	next		/* IN: list of successive operands */
)
{
struct grdim_list *	dlp;

	dlp = CALIB_NEW (struct grdim_list);
	dlp -> operand	= op;
	dlp -> next	= next;

	return (dlp);
}

/*
 * Recursively free the given grdim.
 */

	static
	void
grdim_free (

struct grdim *		dp	/* IN: grdim to recursively free */
)
{
	if (dp EQ NULL) return;

	switch (dp -> op) {
	case GRDIM_OP_ADD:
	case GRDIM_OP_MUL:
		grdim_list_free (dp -> u.list);
		break;

	case GRDIM_OP_UPLUS:
	case GRDIM_OP_NEG:
		grdim_free (dp -> u.unop);
		break;

	case GRDIM_OP_FRAC:
		grdim_free (dp -> u.frac.numer);
		grdim_free (dp -> u.frac.denom);
		break;

	case GRDIM_OP_IPOW:
		grdim_free (dp -> u.ipow.base);
		break;

	case GRDIM_OP_VAR:
		free ((char *) (dp -> u.var));
		break;

	case GRDIM_OP_Z:
		free (dp -> u.zstr);
		break;

	case GRDIM_OP_FUNC:
		grdim_free (dp -> u.func.func);
		grdim_list_free (dp -> u.func.args);
		break;

	case GRDIM_OP_FIX:
		grdim_free (dp -> u.fix.op);
		CALIB_FREE (dp -> u.fix.prefix);
		CALIB_FREE (dp -> u.fix.suffix);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	memset (dp, 0, sizeof (*dp));
	free (dp);
}

/*
 * Recursively free the given grdim_list.
 */

	static
	void
grdim_list_free (

struct grdim_list *	dlp	/* IN: grdim_list to free */
)
{
struct grdim_list *	tmp;

	while (dlp NE NULL) {
		grdim_free (dlp -> operand);
		tmp = dlp -> next;

		dlp -> operand	= NULL;
		dlp -> next	= NULL;
		free (dlp);

		dlp = tmp;
	}
}

/*
 * Write the given expression out in a syntax that can be read back
 * in by maxima (or similar packages).
 */

	static
	void
write_grdim_maxima (

struct PPInfo *  ppip,		/* IN/OUT: Global prettyprint info */
struct ColInfo * cip,		/* IN/OUT: Output column state */
struct grdim *	 dp,		/* IN/OUT: grdim tree to classify nodes of */
int		 parent_op	/* IN: Parent node's operator */
)
{
int			nspace;
calib_bool		parens;
struct grdim_list *	dlp;
char			numbuf [32];

	CALIB_FATAL_ERROR_IF ((ppip EQ NULL) OR (dp EQ NULL));

	parens = maxima_parens (parent_op, dp -> op);

	if (parens) {
		write_token (ppip, cip, 0, LPAREN_STRING, 0);
	}

	switch (dp -> op) {
	case GRDIM_OP_ADD:	/* (Each term has unary "+" or "-".) */
		nspace = 0;
		for (dlp = dp -> u.list; dlp NE NULL; dlp = dlp -> next) {
			cip -> numwhite += nspace;
			write_grdim_maxima (ppip, cip, dlp -> operand, dp -> op);
			nspace = 1;
		}
		break;

	case GRDIM_OP_UPLUS:
		write_token (ppip, cip, 0, "+", 1);
		write_grdim_maxima (ppip, cip, dp -> u.unop, dp -> op);
		break;

	case GRDIM_OP_NEG:
		write_token (ppip, cip, 0, "-", 1);
		write_grdim_maxima (ppip, cip, dp -> u.unop, dp -> op);
		break;

	case GRDIM_OP_MUL:
		nspace = 0;
		for (dlp = dp -> u.list; dlp NE NULL; dlp = dlp -> next) {
			if (nspace > 0) {
				write_token (ppip, cip, 0, "*", 0);
			}
			write_grdim_maxima (ppip, cip, dlp -> operand, dp -> op);
			nspace = 1;
		}
		break;

	case GRDIM_OP_FRAC:
		write_grdim_maxima (ppip, cip, dp -> u.frac.numer, GRDIM_OP_NUMERATOR);
		write_token (ppip, cip, 1, "/", 1);
		write_grdim_maxima (ppip, cip, dp -> u.frac.denom, GRDIM_OP_DENOMINATOR);
		break;

	case GRDIM_OP_IPOW:
		write_grdim_maxima (ppip, cip, dp -> u.ipow.base, dp -> op);
		write_token (ppip, cip, 0, "^", 0);
		sprintf (numbuf, "%d", dp -> u.ipow.power);
		write_token (ppip, cip, 0, numbuf, 0);
		break;

	case GRDIM_OP_VAR:
		write_token (ppip, cip, 0, dp -> u.var, 0);
		break;

	case GRDIM_OP_Z:
		write_token (ppip, cip, 0, dp -> u.zstr, 0);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	if (parens) {
		write_token (ppip, cip, 0, RPAREN_STRING, 0);
	}
}

/*
 * Return TRUE if-and-only-if the node having given parent_op and op
 * must be parenthesized when writing in "maxima" syntax.
 */

	static
	calib_bool
maxima_parens (

int		parent_op,	/* IN: parent node's operator */
int		op		/* IN: this node's operator */
)
{
	if (parent_op < 0) {
		/* Parens never needed around top-level expression. */
		return (FALSE);
	}

	switch (parent_op) {
	case GRDIM_OP_ADD:
		switch (op) {
		case GRDIM_OP_ADD:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_UPLUS:				return (FALSE);
		case GRDIM_OP_NEG:				return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:				return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_UPLUS:	/* Encapsulates terms of a sum. */
	case GRDIM_OP_NEG:	/* Encapsulates terms of a sum. */
		switch (op) {
		case GRDIM_OP_ADD:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:				return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_MUL:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_IPOW:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:				return (TRUE);
		case GRDIM_OP_IPOW:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_NUMERATOR:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_DENOMINATOR:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FRAC:
		/* Should present as NUMERATOR or DENOMINATOR. */
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_VAR:
	case GRDIM_OP_Z:
		/* Neither of these have child nodes. */
		CALIB_FATAL_ERROR;
		return (FALSE);
	}

	CALIB_FATAL_ERROR;
	return (FALSE);
}

/*
 * Write out a sequence of <whitespace1> <token> <whitespace2>.
 * Omit <whitespace1> if (col EQ 0).  Omit <whitespace2> at end of line.
 * Returns updated column.
 */

	static
	void
write_token (

struct PPInfo *	 ppip,		/* IN/OUT: Global prettyprint info */
struct ColInfo * cip,		/* IN/OUT: Output column state */
int		 ws1,		/* IN: Num spaces in whitespace 1 */
const char *	 token,		/* IN: token */
int		 ws2		/* IN: Num spaces in whitespace 2 */
)
{
int		n;

	n = strlen (token);

	if ((cip -> numwhite > 0) OR (ws1 > 0)) {
		/* Desire to prefix token with whitespace. */
		if (cip -> column + 1 + n > ppip -> maxwidth) {
			/* Doesn't fit on current line. */
			fprintf (ppip -> fp, "\n%s", token);
			cip -> column = n;
		}
		else {
			fprintf (ppip -> fp, " %s", token);
			cip -> column += (1 + n);
		}
	}
	else if (cip -> column + n > ppip -> maxwidth) {
		/* Token doesn't fit on current line. */
		fprintf (ppip -> fp, "\n%s", token);
		cip -> column = n;
	}
	else {
		fprintf (ppip -> fp, "%s", token);
		cip -> column += n;
	}
	cip -> numwhite = ws2;
}

/*
 * Recursively classify each grdim node in the tree in a bottom-up traversal.
 */

	static
	void
grdim_classify (

struct PPInfo *	ppip,		/* IN/OUT: Global prettyprint info */
struct grdim *	dp,		/* IN/OUT: grdim tree to classify nodes of */
int		parent_op	/* IN: Parent node's operator */
)
{
int			i, nkids, num_tile, num_vtile, num_opt, num_flat;
char			type;
struct grdim_list *	dlp;
struct grdim *		dp2;
struct grdim **		kids;

	CALIB_FATAL_ERROR_IF (dp EQ NULL);

	kids = NULL;

	if (dp -> op EQ GRDIM_OP_FRAC) {
		nkids = 2;
		kids = CALIB_NEWA (nkids, struct grdim *);
		kids [0] = dp -> u.frac.numer;
		kids [1] = dp -> u.frac.denom;
		grdim_classify (ppip, kids [0], GRDIM_OP_NUMERATOR);
		grdim_classify (ppip, kids [1], GRDIM_OP_DENOMINATOR);
	}
	else {
		switch (dp -> op) {
		case GRDIM_OP_ADD:
		case GRDIM_OP_MUL:
			nkids = 0;
			for (dlp = dp -> u.list; dlp NE NULL; dlp = dlp -> next) {
				++nkids;
			}
			kids = CALIB_NEWA (nkids, struct grdim *);
			for (i = 0, dlp = dp -> u.list; dlp NE NULL; dlp = dlp -> next, i++) {
				kids [i] = dlp -> operand;
			}
			break;

		case GRDIM_OP_UPLUS:
		case GRDIM_OP_NEG:
			nkids = 1;
			kids = CALIB_NEWA (nkids, struct grdim *);
			kids [0] = dp -> u.unop;
			break;

		case GRDIM_OP_FRAC:
			CALIB_FATAL_ERROR;
			break;

		case GRDIM_OP_IPOW:
			nkids = 1;
			kids = CALIB_NEWA (nkids, struct grdim *);
			kids [0] = dp -> u.ipow.base;
			break;

		case GRDIM_OP_VAR:
			dp -> type = TYPE_TILE;
			return;

		case GRDIM_OP_Z:
			if (dp -> maxtile.width >= MIN_WRAP_DIGITS) {
				dp -> type = TYPE_VTILE;
			}
			else {
				dp -> type = TYPE_TILE;
#if 0
				dp -> mintile = dp -> maxtile;
#endif
			}
			return;

		case GRDIM_OP_FUNC:
			/* The function is another "kid". */
			nkids = 1;
			for (dlp = dp -> u.func.args; dlp NE NULL; dlp = dlp -> next) {
				++nkids;
			}
			kids = CALIB_NEWA (nkids, struct grdim *);
			kids [0] = dp -> u.func.func;
			for (i = 0, dlp = dp -> u.func.args; dlp NE NULL; dlp = dlp -> next, i++) {
				kids [i + 1] = dlp -> operand;
			}
			break;

		case GRDIM_OP_FIX:
			nkids = 1;
			kids = CALIB_NEWA (nkids, struct grdim *);
			kids [0] = dp -> u.fix.op;
			break;

		default:
			CALIB_FATAL_ERROR;
		}

		for (i = 0; i < nkids; i++) {
			dp2 = kids [i];
			grdim_classify (ppip, dp2, dp -> op);
		}
	}

	/* Count various types of children. */
	num_tile   = 0;
	num_vtile  = 0;
	num_opt	   = 0;
	num_flat   = 0;
	for (i = 0; i < nkids; i++) {
		dp2 = kids [i];
		switch (dp2 -> type) {
		case TYPE_TILE:		++num_tile;		break;
		case TYPE_VTILE:	++num_vtile;		break;
		case TYPE_OPT:		++num_opt;		break;
		case TYPE_FLAT:		++num_flat;		break;
		default:		CALIB_FATAL_ERROR;	break;
		}
	}

	if ((num_flat > 0) OR (num_opt > 0)) {
		/* Formatting of this and all parents must be flat. */
		type = TYPE_FLAT;
	}
	else if (num_vtile >= 2) {
		/* Two or more vtiles. */
		if (dp -> op EQ GRDIM_OP_FRAC) {
			/* Special case: don't need to optimize a	*/
			/* fraction if numerator and denominator are	*/
			/* both vtiles.					*/
			type = TYPE_VTILE;
		}
		else {
			/* Two or more vtiles: must optimize. */
			type = TYPE_OPT;
		}
	}
	else if (num_vtile > 0) {
		type = TYPE_VTILE;
	}
	else {
		type = TYPE_TILE;
	}

	/* Consider whether to force optimize or not. */
	if (((type EQ TYPE_TILE) OR (type EQ TYPE_VTILE)) AND (nkids > 1)) {
		if (dp -> maxtile.width > ppip -> maxwidth) {
			/* Classic layout is too wide. */
			if (dp -> mintile.width > ppip -> maxwidth) {
				/* Can't crush VTILES to fit either. */
				/* Must optimize. */
				type = TYPE_OPT;
			}
			else if (num_tile EQ nkids) {
				/* All kids are TILES, so there isn't	*/
				/* any way to reformat stuff to be more	*/
				/* narrow.  Must optimize.		*/
				type = TYPE_OPT;
			}
		}
	}

#if 1	/* FIXME!  Maybe we do know how to optimize function(...). */

	/* We don't know how to optimize function(...), so make these	*/
	/* be flat if too wide.						*/
	if ((type EQ TYPE_OPT) AND
	    (dp -> op EQ GRDIM_OP_FUNC)) {
		type = TYPE_FLAT;
	}
#endif

	free (kids);

	dp -> type = type;

	/* Note: we don't need dp->maxtiles or dp->mintiles for	*/
	/* nodes that are flat or being optimized, so there is	*/
	/* no need to try and correct their widths because of	*/
	/* changes in dp->parens.				*/
	if ((type EQ TYPE_TILE) OR (type EQ TYPE_VTILE)) {
#if 0
		dp -> parens = tile_parens (parent_op, dp -> op);
#else
		(void) tile_parens;
#endif
	}
	else if (type EQ TYPE_OPT) {
		dp -> parens = opt_parens (parent_op, dp -> op);
	}
	else if (type EQ TYPE_FLAT) {
		dp -> parens = flat_parens (parent_op, dp -> op);
	}
}

/*
 * Return TRUE if-and-only-if the TYPE_TILE or TYPE_VTILE node having given
 * parent_op and op must be parenthesized.
 */

	static
	int
tile_parens (

int		parent_op,	/* IN: parent node's operator */
int		op		/* IN: this node's operator */
)
{
	if (parent_op < 0) {
		/* Parens never needed around top-level expression. */
		return (FALSE);
	}

	switch (parent_op) {
	case GRDIM_OP_ADD:
	case GRDIM_OP_UPLUS:	/* Encapsulates terms of a sum. */
	case GRDIM_OP_NEG:	/* Encapsulates terms of a sum. */
		switch (op) {
		case GRDIM_OP_ADD:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_UPLUS:				return (FALSE);
		case GRDIM_OP_NEG:				return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:				return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_MUL:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_IPOW:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:				return (TRUE);
		case GRDIM_OP_IPOW:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_NUMERATOR:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_DENOMINATOR:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FRAC:
		/* Should present as NUMERATOR or DENOMINATOR. */
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_VAR:
	case GRDIM_OP_Z:
		/* Neither of these have child nodes. */
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FUNC:
		switch (op) {
		case GRDIM_OP_ADD:				return (FALSE);
		case GRDIM_OP_UPLUS:				return (FALSE);
		case GRDIM_OP_NEG:				return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FIX:
		/* Parens never needed between prefix/suffix strings. */
		return (FALSE);
	}

	CALIB_FATAL_ERROR;
	return (FALSE);
}

/*
 * Return TRUE if-and-only-if the TYPE_OPT node having given
 * parent_op and op must be parenthesized.
 */

	static
	int
opt_parens (

int		parent_op,	/* IN: parent node's operator */
int		op		/* IN: this node's operator */
)
{
	if (parent_op < 0) {
		/* Parens never needed around top-level expression. */
		return (FALSE);
	}

	switch (parent_op) {
	case GRDIM_OP_ADD:
	case GRDIM_OP_UPLUS:	/* Encapsulates terms of a sum. */
	case GRDIM_OP_NEG:	/* Encapsulates terms of a sum. */
		switch (op) {
		case GRDIM_OP_ADD:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_UPLUS:				return (FALSE);
		case GRDIM_OP_NEG:				return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:				return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_MUL:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_IPOW:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:				return (TRUE);
		case GRDIM_OP_IPOW:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_NUMERATOR:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_DENOMINATOR:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FRAC:
		/* Should present as NUMERATOR or DENOMINATOR. */
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_VAR:
	case GRDIM_OP_Z:
		/* Neither of these have child nodes. */
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FUNC:
		switch (op) {
		case GRDIM_OP_ADD:				return (FALSE);
		case GRDIM_OP_UPLUS:				return (FALSE);
		case GRDIM_OP_NEG:				return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FUNC:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FIX:
		/* Parens never needed between prefix/suffix strings. */
		return (FALSE);
	}

	CALIB_FATAL_ERROR;
	return (FALSE);
}

/*
 * Return TRUE if-and-only-if the TYPE_FLAT node having given
 * parent_op and op must be parenthesized.
 */

	static
	int
flat_parens (

int		parent_op,	/* IN: parent node's operator */
int		op		/* IN: this node's operator */
)
{
	if (parent_op < 0) {
		/* Parens never needed around top-level expression. */
		return (FALSE);
	}

	switch (parent_op) {
	case GRDIM_OP_ADD:
		switch (op) {
		case GRDIM_OP_ADD:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_UPLUS:				return (FALSE);
		case GRDIM_OP_NEG:				return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:				return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_UPLUS:	/* Encapsulates terms of a sum. */
	case GRDIM_OP_NEG:	/* Encapsulates terms of a sum. */
		switch (op) {
		case GRDIM_OP_ADD:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:				return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_MUL:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_IPOW:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:				return (TRUE);
		case GRDIM_OP_IPOW:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_NUMERATOR:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_DENOMINATOR:
		switch (op) {
		case GRDIM_OP_ADD:				return (TRUE);
		case GRDIM_OP_UPLUS:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_NEG:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_MUL:				return (TRUE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (TRUE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FRAC:
		/* Should present as NUMERATOR or DENOMINATOR. */
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_VAR:
	case GRDIM_OP_Z:
		/* Neither of these have child nodes. */
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FUNC:
		switch (op) {
		case GRDIM_OP_ADD:				return (FALSE);
		case GRDIM_OP_UPLUS:				return (FALSE);
		case GRDIM_OP_NEG:				return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);

	case GRDIM_OP_FIX:
		switch (op) {
		case GRDIM_OP_ADD:				return (FALSE);
		case GRDIM_OP_UPLUS:				return (FALSE);
		case GRDIM_OP_NEG:				return (FALSE);
		case GRDIM_OP_MUL:				return (FALSE);
		case GRDIM_OP_FRAC:	CALIB_FATAL_ERROR;	return (FALSE);
		case GRDIM_OP_IPOW:				return (FALSE);
		case GRDIM_OP_VAR:				return (FALSE);
		case GRDIM_OP_Z:				return (FALSE);
		case GRDIM_OP_FIX:				return (FALSE);
		}
		CALIB_FATAL_ERROR;
		return (FALSE);
	}

	CALIB_FATAL_ERROR;
	return (FALSE);
}

/*
 * Recursively set the parent_type member of each grdim node.
 */

	static
	void
grdim_set_parent_type (

struct grdim *	dp,		/* IN: Node for which to set parent_type */
int		parent_type	/* IN: Value to set */
)
{
int			type;
struct grdim_list *	dlp;

	CALIB_FATAL_ERROR_IF (dp EQ NULL);

	dp -> parent_type	= parent_type;

	/* Get type to inflict upon all children parent_type. */
	type	= dp -> type;

	switch (dp -> op) {
	case GRDIM_OP_ADD:
	case GRDIM_OP_MUL:
		for (dlp = dp -> u.list; dlp NE NULL; dlp = dlp -> next) {
			grdim_set_parent_type (dlp -> operand, type);
		}
		break;

	case GRDIM_OP_UPLUS:
	case GRDIM_OP_NEG:
		grdim_set_parent_type (dp -> u.unop, type);
		break;

	case GRDIM_OP_FRAC:
		grdim_set_parent_type (dp -> u.frac.numer, type);
		grdim_set_parent_type (dp -> u.frac.denom, type);
		break;

	case GRDIM_OP_IPOW:
		grdim_set_parent_type (dp -> u.ipow.base, type);
		break;

	case GRDIM_OP_VAR:
	case GRDIM_OP_Z:
		/* No children. */
		break;

	case GRDIM_OP_FUNC:
		for (dlp = dp -> u.func.args; dlp NE NULL; dlp = dlp -> next) {
			grdim_set_parent_type (dlp -> operand, type);
		}
		break;

	case GRDIM_OP_FIX:
		grdim_set_parent_type (dp -> u.fix.op, type);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Recompute the mintile values now that the type of each grdim node
 * has been established.  This can make the mintiles wider than
 * initially estimated.
 *
 * Consider, e.g., (x^32 - 65536 x^16 + 4294967296).  Note that this
 * GRDIM_OP_ADD node is parenthesized (dp -> parens is TRUE).  This
 * expression's dimensions are initially computed to be:
 *
 *			Mintile		Maxtile
 *			-------------	-------------
 *			{19, 9, 8, 0}	{30, 2, 1, 0}
 *
 * Its three terms initially have the following maxtile / mintile values:
 *
 *	Type		Mintile		Maxtile
 *	-------		-------------	-------------
 *	TILE		{ 3, 1, 1, 0}	{ 3, 1, 1, 0}
 *	TILE		{ 8, 4, 3, 0}	{11, 1, 1, 0}
 *	VTILE		{ 4, 9, 8, 0}	{12, 1, 0, 0}
 *
 * Note that the top-level mintile of {19, 9, 8, 0} is based on the
 * assumption that integer constant of 65536 within the middle term
 * can be formatted as {2, 4, 3, 0}.  But because this integer constant
 * is shorter than MIN_WRAP_DIGITS, we forbit wrapping this integer across
 * multiple lines, making it be a TILE instead of a VTILE.  Given this
 * choice, the mintile for the middle term must be updated to
 * {11, 1, 1, 0}.  This change must be reflected up through all parent
 * nodes, making the top-level mintile of this example be {22, 9, 8, 0}.
 *
 * Failure to do this can cause us to try formatting VTILES in widths
 * that are too short, causing fatal errors.
 */

	static
	void
grdim_recompute_mintile (

struct grdim *		dp		/* IN/OUT: dims for which to */
					/*	   recompute mintile values */
)
{
int			space, esize;
struct grdim_list *	dlp;
struct grdim *		dp2;
struct grdim *		dp3;
struct tiledim *	tp2;
struct tiledim *	tp3;
struct tiledim		dims;
char			buf [32];

	switch (dp -> op) {
	case GRDIM_OP_ADD:
	case GRDIM_OP_MUL:
		memset (&dims, 0, sizeof (dims));
		space = 0;
		for (dlp = dp -> u.list; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			grdim_recompute_mintile (dp2);
			tp2 = get_mintile (dp2);
			dims.width += (space + tp2 -> width);
			update_tile_heights (&dims, tp2);
			space = 1;
		}
		dp -> mintile = dims;
		break;

	case GRDIM_OP_NEG:
	case GRDIM_OP_UPLUS:
		dp2 = dp -> u.unop;
		grdim_recompute_mintile (dp2);
		tp2 = get_mintile (dp2);
		dp -> mintile = *tp2;
		dp -> mintile.width += 2;
		break;

	case GRDIM_OP_FRAC:
		dp2 = dp -> u.frac.numer;
		dp3 = dp -> u.frac.denom;
		grdim_recompute_mintile (dp2);
		grdim_recompute_mintile (dp3);
		tp2 = get_mintile (dp2);
		tp3 = get_mintile (dp3);
		dp -> mintile.width	= (tp2 -> width > tp3 -> width) ?
						tp2 -> width : tp3 -> width;
		dp -> mintile.height	= 1 + tp2 -> height + tp3 -> height;
		dp -> mintile.ascent	= tp2 -> height;
		dp -> mintile.descent	= tp3 -> height;
		break;

	case GRDIM_OP_IPOW:
		dp2 = dp -> u.ipow.base;
		grdim_recompute_mintile (dp2);
		tp2 = get_mintile (dp2);
		sprintf (buf, "%d", dp2 -> u.ipow.power);
		esize = strlen (buf);
		dp -> mintile.width = tp2 -> width + esize;
		dp -> mintile.ascent	= _MAX (dp2 -> mintile.ascent, 1);
		dp -> mintile.descent	= dp2 -> mintile.descent;
		dp -> mintile.height	= 1 + dp -> mintile.ascent
					    + dp -> mintile.descent;
		break;

	case GRDIM_OP_VAR:
	case GRDIM_OP_Z:
		/* These should not need to change. */
		return;

	case GRDIM_OP_FUNC:
		memset (&dims, 0, sizeof (dims));
		space = 0;
		for (dlp = dp -> u.func.args; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			grdim_recompute_mintile (dp2);
			tp2 = get_mintile (dp2);
			dims.width += (space + tp2 -> width);
			update_tile_heights (&dims, tp2);
			space = 2;
		}
		dims.width += (dp -> u.func.func -> maxtile.width + 3);
		dp -> mintile = dims;
		break;

	case GRDIM_OP_FIX:
		grdim_recompute_mintile (dp -> u.fix.op);
		dp -> mintile	= dp -> u.fix.op -> mintile;
		if (dp -> u.fix.prefix NE NULL) {
			dp -> mintile.width += strlen (dp -> u.fix.prefix);
		}
		if (dp -> u.fix.suffix NE NULL) {
			dp -> mintile.width += strlen (dp -> u.fix.suffix);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	if (dp -> parens) {
		dp -> mintile.width += 2;
	}
}

/*
 * Return the proper "mintile" values for the given node, based upon
 * the node's type.
 */

	static
	struct tiledim *
get_mintile (

struct grdim *		dp	/* IN: Node to get mintile values for */
)
{
struct tiledim *	tp;

	if (dp -> type EQ TYPE_TILE) {
		/* Ignore the "wrapped" version of this node --	*/
		/* mintile is the same as maxtile.		*/
		tp = &(dp -> maxtile);
	}
	else {
		tp = &(dp -> mintile);
	}

	return (tp);
}

/*
 * Update the given destination tile heights to represent
 * the tile that results when appending the two tiles side by side.
 */

	static
	void
update_tile_heights (

struct tiledim *	dst,		/* IN/OUT: dims to update heights of */
const struct tiledim *	src		/* IN: tile being added horizontally */
)
{
int		ascent, descent, height;

	ascent	= _MAX (dst -> ascent,  src -> ascent);
	descent	= _MAX (dst -> descent, src -> descent);
	height	= ascent + 1 + descent;

	dst -> ascent	= ascent;
	dst -> descent	= descent;
	dst -> height	= height;
}

/*
 * Write the given top-level expression to the given stream.
 */

	static
	void
write_grdim_top_level (

struct PPInfo *		ppip,		/* IN/OUT: Global prettyprint info */
struct grdim *		dp		/* IN: expression to display */
)
{
int			w, x, y;
struct tiledim *	tp;
const char *		prefix;
const char *		suffix;
struct tiledim		dims;
struct canvas		canvas;

	if (dp -> type EQ TYPE_TILE) {
		tp = &(dp -> maxtile);
		w = tp -> width;
		if (w < ppip -> maxwidth) {
			/* Display centered. */
			x = (ppip -> maxwidth - w + 1) / 2;
			y = tp -> ascent;
			init_canvas (&canvas, tp -> height);
			plot_grdim (&canvas, dp, x, y);
			write_canvas (ppip, &canvas);
			free_canvas (&canvas);
		}
		else {
			too_wide (ppip);
		}
		return;
	}
	if (dp -> type EQ TYPE_VTILE) {
		tp = &(dp -> maxtile);
		w = tp -> width;
		if (w < ppip -> maxwidth) {
			/* Display centered. */
			x = (ppip -> maxwidth - w + 1) / 2;
			y = tp -> ascent;
			init_canvas (&canvas, tp -> height);
			plot_grdim (&canvas, dp, x, y);
			write_canvas (ppip, &canvas);
			free_canvas (&canvas);
			return;
		}
		tp = &(dp -> mintile);
		w = tp -> width;
		if (w > ppip -> maxwidth) {
			too_wide (ppip);
			return;
		}
		vtile_dims (&dims, dp, ppip -> maxwidth);
		CALIB_FATAL_ERROR_IF (dims.width NE ppip -> maxwidth);
		x = 0;
		y = dims.ascent;
		init_canvas (&canvas, dims.height);
		plot_vtile_width (&canvas, dp, x, y, ppip -> maxwidth);
		write_canvas (ppip, &canvas);
		free_canvas (&canvas);
		return;
	}

	prefix = "";
	suffix = "";
	write_grdim (ppip, dp, prefix, suffix);
}

/*
 * Write the given sub-expression tree out to the given stream.
 */

	static
	void
write_grdim (

struct PPInfo *		ppip,		/* IN/OUT: Global prettyprint info */
struct grdim *		dp,		/* IN: expression to display */
const char *		prefix,		/* IN: prefix string */
const char *		suffix		/* IN: suffix string */
)
{
	switch (dp -> type) {
	case TYPE_TILE:
	case TYPE_VTILE:
		write_grdim_tile_line (ppip, dp, prefix, suffix);
		break;

	case TYPE_OPT:
		write_grdim_opt (ppip, dp, prefix, suffix);
		break;

	case TYPE_FLAT:
		write_grdim_flat (ppip, dp, prefix, suffix);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Write the given TILE sub-expression out to the given stream,
 * followed by a newline.
 */

	static
	void
write_grdim_tile_line (

struct PPInfo *		ppip,		/* IN/OUT: Global prettyprint info */
struct grdim *		dp,		/* IN: expression to display */
const char *		prefix,		/* IN: prefix string */
const char *		suffix		/* IN: suffix string */
)
{
	/* Total kludge here. */
	fprintf (ppip -> fp, LPAREN_STRING);

	write_grdim_tile (ppip, dp, prefix, suffix);

	fprintf (ppip -> fp, RPAREN_STRING);

	fprintf (ppip -> fp, "\n");
}

/*
 * Write the given TILE sub-expression out to the given stream.
 */

	static
	void
write_grdim_tile (

struct PPInfo *		ppip,		/* IN/OUT: Global prettyprint info */
struct grdim *		dp,		/* IN: expression to display */
const char *		prefix,		/* IN: prefix string */
const char *		suffix		/* IN: suffix string */
)
{
FILE *			fp;
struct grdim_list *	lp;

	fp	= ppip -> fp;

	if (dp -> parens) {
		fprintf (fp, LPAREN_STRING);
	}

	switch (dp -> op) {
	case GRDIM_OP_ADD:
		for (lp = dp -> u.list; lp NE NULL; lp = lp -> next) {
			write_grdim_tile (ppip, lp -> operand, prefix, suffix);
			if (lp -> next NE NULL) {
				fprintf (fp, " ");
			}
		}
		break;

	case GRDIM_OP_UPLUS:
		fprintf (fp, "+ ");
		write_grdim_tile (ppip, dp -> u.unop, prefix, suffix);
		break;

	case GRDIM_OP_NEG:
		fprintf (fp, "- ");
		write_grdim_tile (ppip, dp -> u.unop, prefix, suffix);
		break;

	case GRDIM_OP_MUL:
		for (lp = dp -> u.list; lp NE NULL; lp = lp -> next) {
			write_grdim_tile (ppip, lp -> operand, prefix, suffix);
			if (lp -> next NE NULL) {
				fprintf (fp, " * ");
			}
		}
		break;

	case GRDIM_OP_FRAC:
		fprintf (fp, LPAREN_STRING);
		write_grdim_tile (ppip, dp -> u.frac.numer, prefix, suffix);
		fprintf (fp, RPAREN_STRING " / " LPAREN_STRING);
		write_grdim_tile (ppip, dp -> u.frac.denom, prefix, suffix);
		fprintf (fp, RPAREN_STRING);
		break;

	case GRDIM_OP_IPOW:
		write_grdim_tile (ppip, dp -> u.ipow.base, prefix, suffix);
		fprintf (fp, "^%d", dp -> u.ipow.power);
		break;

	case GRDIM_OP_VAR:
		fprintf (fp, "%s", dp -> u.var);
		break;

	case GRDIM_OP_Z:
		fprintf (fp, "%s", dp -> u.zstr);
		break;

	case GRDIM_OP_FUNC:
		write_grdim_tile (ppip, dp -> u.func.func, prefix, suffix);
		fprintf (fp, " " LPAREN_STRING);
		for (lp = dp -> u.func.args; lp NE NULL; lp = lp -> next) {
			write_grdim_tile (ppip, lp -> operand, prefix, suffix);
			if (lp -> next NE NULL) {
				fprintf (fp, ", ");
			}
		}
		fprintf (fp, RPAREN_STRING);
		break;

	case GRDIM_OP_FIX:
		if ((dp -> u.fix.pflags & FIX_FLAT) NE 0) {
			fprintf (fp, "%s", dp -> u.fix.prefix);
		}
		write_grdim_tile (ppip, dp -> u.fix.op, prefix, suffix);
		if ((dp -> u.fix.sflags & FIX_FLAT) NE 0) {
			fprintf (fp, "%s", dp -> u.fix.suffix);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	if (dp -> parens) {
		fprintf (fp, RPAREN_STRING);
	}
}

/*
 * Write the given OPT sub-expression out to the given stream.
 */

	static
	void
write_grdim_opt (

struct PPInfo *		ppip,		/* IN/OUT: Global prettyprint info */
struct grdim *		dp,		/* IN: expression to display */
const char *		prefix,		/* IN: prefix string */
const char *		suffix		/* IN: suffix string */
)
{
int			nkids, plen, slen;
struct grdim **		kids;
char *			paren_prefix;
char *			paren_suffix;
struct DPFormat		format;

	CALIB_FATAL_ERROR_IF ((dp EQ NULL) OR (dp -> type NE TYPE_OPT));

	if (dp -> parens) {
		paren_prefix	= string_cat2 (prefix, LPAREN_STRING);
		paren_suffix	= string_cat2 (RPAREN_STRING, suffix);
	}
	else {
		paren_prefix	= strdup (prefix);
		paren_suffix	= strdup (suffix);
	}

	plen = strlen (paren_prefix);
	slen = strlen (paren_suffix);

	switch (dp -> op) {
	case GRDIM_OP_ADD:
		get_operand_array (&nkids, &kids, dp -> u.list);

		find_best_layout (&format,
				  nkids,
				  kids,
				  ppip -> maxwidth - plen,
				  slen);

		write_formatted (ppip, &format, paren_prefix, paren_suffix);

		clear_format (&format);
		free (kids);
		break;

	case GRDIM_OP_MUL:
		get_operand_array (&nkids, &kids, dp -> u.list);

		find_best_layout (&format,
				  nkids,
				  kids,
				  ppip -> maxwidth - plen,
				  slen);

		write_formatted (ppip, &format, paren_prefix, paren_suffix);

		clear_format (&format);
		free (kids);
		break;

	case GRDIM_OP_FRAC:
		/* Fractions are prevented from being TYPE_OPT. */
		CALIB_FATAL_ERROR;
		break;

	case GRDIM_OP_UPLUS:
	case GRDIM_OP_NEG:
	case GRDIM_OP_IPOW:
	case GRDIM_OP_VAR:
	case GRDIM_OP_Z:
		/* These cannot be TYPE_OPT.  Having TYPE_OPT requires	*/
		/* 2 or more VTILE operands, which none of these	*/
		/* operators allow.					*/
		CALIB_FATAL_ERROR;
		break;

	case GRDIM_OP_FUNC:
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	free (paren_suffix);
	free (paren_prefix);
}

/*
 * Write the given FLAT sub-expression out to the given stream.
 */

	static
	void
write_grdim_flat (

struct PPInfo *		ppip,		/* IN/OUT: Global prettyprint info */
struct grdim *		dp,		/* IN: expression to display */
const char *		prefix,		/* IN: prefix string */
const char *		suffix		/* IN: suffix string */
)
{
int			i, j, k, nkids, plen, slen;
struct grdim *		func;
struct grdim_list *	args;
struct grdim_list *	dlp;
struct grdim **		kids;
char *			paren_prefix;
char *			paren_suffix;
char *			op_prefix;
char *			op_suffix;
char *			str;
const char *		cur_prefix;
const char *		cur_suffix;
struct grdim *		dp2;
struct DPFormat		format;
char			numbuf [32];

	CALIB_FATAL_ERROR_IF (dp EQ NULL);
#if 0
	/* We call this with non-flat sub-expressions! */
	CALIB_FATAL_ERROR_IF (dp -> type NE TYPE_FLAT);
#endif

	if (dp -> parens) {
		paren_prefix	= string_cat2 (prefix, LPAREN_STRING);
		paren_suffix	= string_cat2 (RPAREN_STRING, suffix);
	}
	else {
		paren_prefix	= strdup (prefix);
		paren_suffix	= strdup (suffix);
	}

	plen = strlen (paren_prefix);
	slen = strlen (paren_suffix);

	switch (dp -> op) {
	case GRDIM_OP_ADD:
		get_operand_array (&nkids, &kids, dp -> u.list);

		for (i = 0; i < nkids; ) {
			dp2 = kids [i];
			cur_prefix = (i EQ 0) ? paren_prefix : "";
			cur_suffix = (i EQ nkids - 1) ? paren_suffix : "";
			if ((dp2 -> type EQ TYPE_FLAT) OR
			    (dp2 -> type EQ TYPE_OPT)) {
				write_grdim (ppip,
					     dp2,
					     cur_prefix,
					     cur_suffix);
				if (i < nkids - 1) {
					fprintf (ppip -> fp, "\n");
				}
				++i;
				continue;
			}
			CALIB_FATAL_ERROR_IF ((dp2 -> type NE TYPE_TILE) AND
					      (dp2 -> type NE TYPE_VTILE));
			/* Have TILE or VTILE.  Find longest sequence	*/
			/* of consecutive TILE/VTILE sub-exprs.		*/
			for (j = i + 1; j < nkids; j++) {
				dp2 = kids [j];
				if ((dp2 -> type NE TYPE_TILE) AND
				    (dp2 -> type NE TYPE_VTILE)) break;
			}
			k = j - i;
			find_best_layout (&format,
					  k,
					  &kids [i],
					  ppip -> maxwidth - plen,
					  slen);
			write_formatted (ppip, &format, cur_prefix, cur_suffix);
			clear_format (&format);
			i = j;
			if (i < nkids - 1) {
				fprintf (ppip -> fp, "\n");
			}
		}

		free (kids);
		break;

	case GRDIM_OP_UPLUS:
		op_prefix = string_cat2 (prefix, "+ ");		
		write_grdim (ppip, dp -> u.unop, op_prefix, suffix);
		free (op_prefix);
		break;

	case GRDIM_OP_NEG:
		op_prefix = CALIB_NEWA (plen + 3, char);
		memset (op_prefix, ' ', plen + 2);
		op_prefix [plen] = '-';
		op_prefix [plen + 2] = '\0';
		write_grdim (ppip, dp -> u.unop, op_prefix, suffix);
		free (op_prefix);
		break;

	case GRDIM_OP_MUL:
		get_operand_array (&nkids, &kids, dp -> u.list);

		op_prefix = CALIB_NEWA (plen + 3, char);
		memset (op_prefix, ' ', plen + 2);
		op_prefix [plen] = '*';
		op_prefix [plen + 2] = '\0';

		for (i = 0; i < nkids; ) {
			dp2 = kids [i];
			cur_prefix = (i EQ 0) ? paren_prefix : op_prefix;
			cur_suffix = (i EQ nkids - 1) ? paren_suffix : "";
			if ((dp2 -> type EQ TYPE_FLAT) OR
			    (dp2 -> type EQ TYPE_OPT)) {
				write_grdim (ppip,
					     dp2,
					     cur_prefix,
					     cur_suffix);
				if (i < nkids - 1) {
					fprintf (ppip -> fp, "\n");
				}
				++i;
				continue;
			}
			CALIB_FATAL_ERROR_IF ((dp2 -> type NE TYPE_TILE) AND
					      (dp2 -> type NE TYPE_VTILE));
			/* Have TILE or VTILE.  Find longest sequence	*/
			/* of consecutive TILE/VTILE sub-exprs.		*/
			for (j = i + 1; j < nkids; j++) {
				dp2 = kids [j];
				if ((dp2 -> type NE TYPE_TILE) AND
				    (dp2 -> type NE TYPE_VTILE)) break;
			}
			k = j - i;
			find_best_layout (&format,
					  k,
					  &kids [i],
					  ppip -> maxwidth - plen,
					  slen);
			write_formatted (ppip, &format, op_prefix, paren_suffix);
			clear_format (&format);
			i = j;
			if (i < nkids - 1) {
				fprintf (ppip -> fp, "\n");
			}
		}

		free (op_prefix);
		free (kids);
		break;

	case GRDIM_OP_FRAC:
		write_grdim (ppip, dp -> u.frac.numer, paren_prefix, "");
		fprintf (ppip -> fp, "\n");

		op_prefix = CALIB_NEWA (plen + 3, char);
		memset (op_prefix, ' ', plen + 2);
		op_prefix [plen] = '/';
		op_prefix [plen + 2] = '\0';

		write_grdim (ppip, dp -> u.frac.denom, op_prefix, paren_suffix);
		free (op_prefix);
		break;

	case GRDIM_OP_IPOW:
		sprintf (numbuf, "^%d", dp -> u.ipow.power);
		op_suffix = string_cat2 (paren_suffix, numbuf);
		write_grdim (ppip, dp -> u.ipow.base, paren_prefix, op_suffix);
		free (op_suffix);
		break;

	case GRDIM_OP_VAR:
		fprintf (ppip -> fp, "%s%s%s", prefix, dp -> u.var, suffix);
		break;

	case GRDIM_OP_Z:
		fprintf (ppip -> fp, "%s%s%s", prefix, dp -> u.zstr, suffix);
		break;

	case GRDIM_OP_FUNC:
		func	= dp -> u.func.func;
		args	= dp -> u.func.args;

		/* Get function and args into one list. */
		dlp = new_grdim_list (func, args);

		get_operand_array (&nkids, &kids, dlp);

		dlp -> operand	= NULL;
		dlp -> next	= NULL;
		grdim_list_free (dlp);

		for (i = 0; i < nkids; ) {
			dp2 = kids [i];
			if ((dp2 -> type EQ TYPE_FLAT) OR
			    (dp2 -> type EQ TYPE_OPT)) {
				write_grdim (ppip,
					     dp2,
					     "",
					     "");
				if (i < nkids - 1) {
					fprintf (ppip -> fp, "\n");
				}
				++i;
				continue;
			}
			CALIB_FATAL_ERROR_IF ((dp2 -> type NE TYPE_TILE) AND
					      (dp2 -> type NE TYPE_VTILE));
			/* Have TILE or VTILE.  Find longest sequence	*/
			/* of consecutive TILE/VTILE sub-exprs.		*/
			for (j = i + 1; j < nkids; j++) {
				dp2 = kids [j];
				if ((dp2 -> type NE TYPE_TILE) AND
				    (dp2 -> type NE TYPE_VTILE)) break;
			}
			k = j - i;
			find_best_layout (&format,
					  k,
					  &kids [i],
					  ppip -> maxwidth - plen,
					  slen);
			write_formatted (ppip, &format, "", "");
			clear_format (&format);
			i = j;
			if (i < nkids - 1) {
				fprintf (ppip -> fp, "\n");
			}
		}

		free (kids);
		break;

	case GRDIM_OP_FIX:
		fprintf (ppip -> fp, "%s", prefix);
		str = dp -> u.fix.prefix;
		if ((str NE NULL) AND
		    ((dp -> u.fix.pflags & FIX_FLAT) NE 0)) {
			fprintf (ppip -> fp, "%s", str);
		}
		write_grdim_flat (ppip, dp -> u.fix.op, "", "");
		str = dp -> u.fix.suffix;
		if ((str NE NULL) AND
		    ((dp -> u.fix.sflags & FIX_FLAT) NE 0)) {
			fprintf (ppip -> fp, "%s", str);
		}
		fprintf (ppip -> fp, "%s", suffix);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	free (paren_suffix);
	free (paren_prefix);
}

/*
 * Write the given sequence of sub-expressions according to the given
 * format / layout.
 */

	static
	void
write_formatted (

struct PPInfo *		ppip,		/* IN/OUT: Global prettyprint info */
struct DPFormat *	format,		/* IN: Formatted expression */
const char *		prefix,		/* IN: Prefix string */
const char *		suffix		/* IN: Suffix string */
)
{
int			x, y, plen, nlines, line;
struct tiledim *	tp;
struct canvas		canvas;

	plen	= strlen (prefix);

	nlines = format -> nlines;

	CALIB_FATAL_ERROR_IF (nlines <= 0);

	/* First expression line. */
	line = 0;
	tp = &(format -> linedims [line]);
	init_canvas (&canvas, tp -> height);
	x = plen;
	y = tp -> ascent;
	plot_canvas (&canvas, prefix, 0, y);
	plot_formatted_expr_line (&canvas, format, line, x, y);

	for (line = 1; line < nlines; line++) {
		write_canvas (ppip, &canvas);
		free_canvas (&canvas);

		fprintf (ppip -> fp, "\n");

		tp = &(format -> linedims [line]);
		init_canvas (&canvas, tp -> height);
		x = plen;
		y = tp -> ascent;
		plot_formatted_expr_line (&canvas, format, line, x, y);
	}

	/* Current value of tp refers to last line's rectangle. */
	/* Apply suffix. */
	x = plen + tp -> width;
	y = tp -> ascent;
	plot_canvas (&canvas, suffix, x, y);
	write_canvas (ppip, &canvas);
	free_canvas (&canvas);
}

/*
 * Plot a given expression line from the given format to the given canvas.
 */

	static
	void
plot_formatted_expr_line (

struct canvas *		cp,		/* IN/OUT: canvas to plot into */
struct DPFormat *	format,		/* IN: Format / layout of exprs */
int			line,		/* IN: Which expr line to plot */
int			x,		/* IN: X coordinate */
int			y		/* IN: Y coordinate */
)
{
int			i, j, k, nlines, w;
struct grdim **		kids;
int *			numops;
int *			opwidths;
struct grdim *		dp;

	CALIB_FATAL_ERROR_IF ((cp EQ NULL) OR (format EQ NULL));

	nlines		= format -> nlines;
	kids		= format -> kids;
	numops		= format -> numops;
	opwidths	= format -> opwidths;

	CALIB_FATAL_ERROR_IF ((line < 0) OR (line >= nlines));

	i = 0;
	for (j = 0; j < line; j++) {
		i += numops [j];
	}
	k = format -> numops [line];
	for (j = 0; j < k; j++) {
		dp = kids [i + j];
		if (dp -> type EQ TYPE_TILE) {
			plot_grdim (cp, dp, x, y);
			x += (1 + dp -> maxtile.width);
		}
		else {
			CALIB_FATAL_ERROR_IF (dp -> type NE TYPE_VTILE);
			w = opwidths [i + j];
			plot_vtile_width (cp, dp, x, y, w);
			x += (w + 1);
		}
	}
}

/*
 * Convert a list of operands into a dynamically allocated array of
 * operand pointers.
 */

	static
	void
get_operand_array (

int *			nkids_out,	/* IN/OUT: Number of operands */
struct grdim ***	array_out,	/* IN/OUT: Array of operands */
struct grdim_list *	list		/* IN: Linked list of operands */
)
{
int			i, nkids;
struct grdim **		array;
struct grdim_list *	dlp;

	nkids = 0;
	for (dlp = list; dlp NE NULL; dlp = dlp -> next) {
		++nkids;
	}
	CALIB_FATAL_ERROR_IF (nkids < 2);

	array = CALIB_NEWA (nkids, struct grdim *);

	for (i = 0, dlp = list; dlp NE NULL; i++, dlp = dlp -> next) {
		array [i] = dlp -> operand;
	}
	CALIB_FATAL_ERROR_IF (i NE nkids);

	*nkids_out	= nkids;
	*array_out	= array;
}

/*
 * Find the best layout of the given sequence of TILE and/or VTILE
 * sub-expressions via dynamic programming.
 *
 * Results are returned in the given "results" object.
 *
 * The objective is roughly to minimize the total area of the rectangles,
 * but there is also a penalty for having lines that are partially full
 * (empty space at the right hand end).  This penalty does not apply for
 * the last line (i.e., any line containing the final child sub-expression).
 */

	static
	void
find_best_layout (

struct DPFormat *	result,		/* IN/OUT: Format info */
int			nkids,		/* IN: Number of sub-expressions */
struct grdim **		kids,		/* IN: Array of sub-expressions */
int			maxwidth,	/* IN: Max width of each line */
int			suffixlen	/* IN: Length of "suffix" needed on */
					/*     last line */
)
{
int			i, j, k, m, n, w, wk, i1, k1;
int			jlim, klim, wdim, wlim;
int			line, nlines, base, line_max_width;
calib_int64u		area, area2, start_area;
int *			min_kid_width;
int *			max_kid_width;
int *			numops;
int *			opwidths;
int *			ip1;
int *			ip2;
struct tiledim *	linedims;
struct grdim *		dp;
struct tiledim *	kwdim;
struct tiledim *	kwrow;
struct T_entry **	opstart;
struct T_entry *	T;
struct T_entry *	Trow;
struct T_entry *	prev_row;
struct T_entry *	src;
struct T_entry *	dst;
struct T_entry *	S;
struct tiledim *	tp;
struct tiledim *	tp2;
int **			L_partition;
calib_int64u *		L_area;
struct tiledim		dims;

	/* Clear out result object. */
	memset (result, 0, sizeof (*result));

	/* Verify that each sub-expression is TILE or VTILE. */
	for (i = 0; i < nkids; i++) {
		dp = kids [i];
		CALIB_FATAL_ERROR_IF ((dp -> type NE TYPE_TILE) AND
				      (dp -> type NE TYPE_VTILE));
	}

	/* Determine the minimum and maximum width for each kid. */
	min_kid_width = CALIB_NEWA (2 * nkids, int);
	max_kid_width = min_kid_width + nkids;
	for (i = 0; i < nkids; i++) {
		dp = kids [i];
		if (dp -> type EQ TYPE_TILE) {
			min_kid_width [i]	= dp -> maxtile.width;
			max_kid_width [i]	= dp -> maxtile.width;
		}
		else {
			j = dp -> maxtile.width;
			if (j > maxwidth) {
				j = maxwidth;
			}
			min_kid_width [i]	= dp -> mintile.width;
			max_kid_width [i]	= j;
		}
	}

	/* For each kid i and width j, compute the dimensions of the	*/
	/* rectangle that results when that kid is formatted to be	*/
	/* exactly that width.						*/
	wdim = maxwidth + 1;
	n = nkids * wdim;
	kwdim = CALIB_NEWA (n, struct tiledim);
	memset (kwdim, 0, n * sizeof (kwdim [0]));
	for (i = 0, kwrow = kwdim; i < nkids; i++, kwrow += wdim) {
		dp = kids [i];
		if (dp -> type EQ TYPE_TILE) {
			/* Only one valid width. */
			kwrow [dp -> maxtile.width] = dp -> maxtile;
			continue;
		}
		/* A VTILE, which can have many widths. */
		j = dp -> mintile.width;
		kwrow [j] = dp -> mintile;
		++j;
		jlim = max_kid_width [i];
		for (; j <= jlim; j++) {
			vtile_dims (&kwrow [j], dp, j);
		}
	}

	opstart = CALIB_NEWA (nkids, struct T_entry *);

	n = (nkids * (nkids + 1) / 2) * wdim;
	T = CALIB_NEWA (n, struct T_entry);
	memset (T, 0, n * sizeof (T [0]));

	Trow = T;
	for (i = 0; i < nkids; i++) {
		/* Record start of tables for operand I. */
		opstart [i] = Trow;

		line_max_width = maxwidth;
		if (i EQ nkids - 1) {
			/* Last line can be shorter because of suffix. */
			line_max_width = maxwidth - suffixlen;
		}

		/* Base case: T [I, 1, *]. */
		dp = kids [i];
		if (dp -> type EQ TYPE_TILE) {
			w = dp -> maxtile.width;
			CALIB_FATAL_ERROR_IF ((w < 0) OR (w >= wdim));
			/* It is possible that the last tile simply	*/
			/* does not fit on last line.			*/
			if (w <= line_max_width) {
				Trow [w].dim = dp -> maxtile;
				Trow [w].w_k = w;
			}
		}
		else {
			CALIB_FATAL_ERROR_IF (dp -> type NE TYPE_VTILE);

			kwrow = &kwdim [i * wdim];

			wlim = dp -> maxtile.width;
			if (wlim > line_max_width) {
				wlim = line_max_width;
			}
			for (w = dp -> mintile.width; w < wdim; w++) {
				if (kwrow [w].width > 0) {
					Trow [w].dim = kwrow [w];
					Trow [w].w_k = w;
				}
			}
		}
		/* Define T [I, J, *] in terms of T [I, J-1, *]. */
		for (i1 = i + 1; i1 < nkids; i1++) {
			prev_row = Trow;
			Trow += wdim;
			line_max_width = maxwidth;
			if (i1 EQ nkids - 1) {
				line_max_width = maxwidth - suffixlen;
			}
			dp = kids [i1];
			if (dp -> type EQ TYPE_TILE) {
				w = dp -> maxtile.width;
				CALIB_FATAL_ERROR_IF ((w <= 0) OR (w >= wdim));
				klim = line_max_width - (1 + w);
				for (k = 0; k < klim; k++) {
					tp = &(prev_row [k].dim);
					if (tp -> width <= 0) continue;
					append_tiles (&dims,
						      tp,
						      &(dp -> maxtile));
					area = tile_area (&dims);
					dst = &Trow [k + 1 + w];
					if ((dst -> dim.width <= 0) OR
					    (area < tile_area (&(dst -> dim)))) {
						dst -> dim = dims;
						dst -> w_k = w;
					}
				}
			}
			else {
				CALIB_FATAL_ERROR_IF (dp -> type NE TYPE_VTILE);
				/* This is where most of the real	*/
				/* optimization	happens.		*/
				kwrow = &kwdim [i1 * wdim];
				wlim = dp -> maxtile.width;
				if (wlim > line_max_width) {
					wlim = line_max_width;
				}

#if 1
				for (k = 0; k <= line_max_width; k++) {
					tp = &(prev_row [k].dim);
					if (tp -> width <= 0) continue;
					wlim = line_max_width - (k + 1);
					for (w = dp -> mintile.width; w <= wlim; w++) {
						tp2 = &kwrow [w];
						if (tp2 -> width <= 0) continue;
						append_tiles (&dims, tp, tp2);
						area = tile_area (&dims);
						dst = &Trow [k + 1 + w];
						if ((dst -> dim.width <= 0) OR
						    (area < tile_area (&(dst -> dim)))) {
							dst -> dim = dims;
							dst -> w_k = w;
						}
					}
				}
#else
				for (w = dp -> mintile.width; w <= wlim; w++) {
					klim = line_max_width - (1 + w);
					tp2 = &kwrow [w];
					if (tp2 -> width <= 0) continue;
					for (k = 0; k < klim; k++) {
						tp = &(prev_row [k].dim);
						if (tp -> width <= 0) continue;
						dst = &Trow [k + 1 + w];
						append_tiles (&dims, tp, tp2);
						area = tile_area (&dims);
						if ((dst -> dim.width <= 0) OR
						    (area < tile_area (&(dst -> dim)))) {
							dst -> dim = dims;
							dst -> w_k = w;
						}
					}
				}
#endif
			}
		}
		Trow += wdim;
	}

	n = (nkids * (nkids + 1)) / 2;

	/* Summarize: compute S [I, J] = min area T [I, J, *]. */
	S = CALIB_NEWA (n, struct T_entry);
	memset (S, 0, n * sizeof (S [0]));

	m = 0;
	for (i = 0; i < nkids; i++) {
		Trow = opstart [i];
		for (j = i; j < nkids; j++) {
			area = ~((calib_int64u) 0);
			k1 = -1;
			for (k = 0; k < wdim; k++) {
				if (Trow [k].dim.width <= 0) continue;
				area2 = tile_area (&(Trow [k].dim));
				if (area2 < area) {
					k1 = k;
					area = area2;
				}
			}
			if (k1 >= 0) {
				S [m] = Trow [k1];
			}
			++m;
			Trow += wdim;
		}
	}

	/* Find best partition of operands between lines.		*/
	/* L_partition [i] contains data describing the best (least	*/
	/* total area) manner for packing all kids 0, 1, 2, ..., i	*/
	/* into one or more "lines".  If non-NULL, L_partition [i] is	*/
	/* a dynamically allocated array of ints as follows:		*/
	/*								*/
	/*	Number_of_Lines						*/
	/*	Number of kids in first line				*/
	/*	Number of kids in second line				*/
	/*	etc.							*/
	/*								*/
	/* Summing up the number of kids in each line should yield	*/
	/* exactly "i".  By explicitly constructing a new such object	*/
	/* Each time we improve L_area [i], we don't interfere with	*/
	/* other partitions held elsewhere in L_partition[] -- we have	*/
	/* consistent partitions no matter what sequence we choose to	*/
	/* update/improve individual L_area[i] / L_partition[i] pairs.	*/
	
	L_partition	= CALIB_NEWA (nkids + 1, int *);
	L_area	= CALIB_NEWA (nkids + 1, calib_int64u);
	memset (L_partition, 0, (nkids + 1) * sizeof (L_partition [0]));
	for (i = 0; i <= nkids; i++) {
		L_area [i] = ~((calib_int64u) 0);
	}
	ip1 = NULL;
	m = 0;
	start_area = 0;
	for (i = 0; i < nkids; i++) {
		for (j = i; j < nkids; j++) {
			src = &S [m++];
			if (src -> dim.width <= 0) continue;
			area = start_area + biased_line_area (&(src -> dim),
							      maxwidth);
			if (area < L_area [j]) {
				nlines = (ip1 EQ NULL) ? 0 : ip1 [0];
				ip2 = CALIB_NEWA (nlines + 2, int);
				ip2 [0] = nlines + 1;
				memcpy (&ip2 [1],
					&ip1 [1],
					nlines * sizeof (ip2 [0]));
				ip2 [nlines + 1] = (j - i) + 1;

				CALIB_FREE (L_partition [j]);
				L_partition [j] = ip2;
				L_area  [j] = area;
			}
		}
		start_area = L_area [i];
		ip1 = L_partition [i];
	}

	/* Get best partition of all kids. */
	ip1 = L_partition [nkids - 1];
	CALIB_FATAL_ERROR_IF (ip1 EQ NULL);

	nlines = ip1 [0];

	/* Allocate result buffers. */
	numops		= CALIB_NEWA (nlines, int);
	opwidths	= CALIB_NEWA (nkids, int);
	linedims	= CALIB_NEWA (nlines, struct tiledim);

	memcpy (numops, &ip1 [1], nlines * sizeof (numops [0]));

	/* Left f(n) = (n * (n + 1)) / 2.  Then			*/
	/*	FIRST_ROW(i,n)	= f(n) - f(n-i)			*/
	/*			= n + n-1 + ... + n-i+1.	*/
#define	FIRST_ROW(i,n)	(((i) * (2 * (n) - (i) + 1)) / 2)

	/* Compute linedims[] and opwidths[]. */
	j = 0;
	for (line = 0; line < nlines; line++) {
		k = numops [line];
		base = FIRST_ROW (j, nkids);
		linedims [line] = S [base + k - 1].dim;
		w = S [base + k - 1].dim.width;
		for (i = k - 1; i >= 0; i--) {
			wk = T [(base + i) * wdim + w].w_k;
			opwidths [j + i] = wk;
			w -= (1 + wk);
		}
		j += k;
	}

#undef FIRST_ROW

	result -> nlines	= nlines;
	result -> nkids		= nkids;
	result -> kids		= kids;
	result -> numops	= numops;
	result -> opwidths	= opwidths;
	result -> linedims	= linedims;

#if STORE_DP_TABLES
	result -> min_kid_width	= min_kid_width;
	result -> max_kid_width	= max_kid_width;
	result -> kwdim		= kwdim;
	result -> opstart	= opstart;
	result -> T		= T;
	result -> S		= S;
	result -> L_partition	= L_partition;
	result -> L_area	= L_area;
#else

	for (i = 0; i < nkids; i++) {
		CALIB_FREE (L_partition [i]);
	}

	free (L_area);
	free (L_partition);
	free (S);
	free (T);
	free (opstart);
	free (kwdim);
	free (min_kid_width);
#endif
}

/*
 * Clear out the given DPFormat object, freeing up associated memory.
 */

	static
	void
clear_format (

struct DPFormat *	format		/* IN: Format to clear */
)
{
#if STORE_DP_TABLES
	int i, nkids;

	nkids = format -> nkids;
	for (i = 0; i < nkids; i++) {
		CALIB_FREE (format -> L_partition [i]);
	}

	CALIB_FREE (format -> L_area);
	CALIB_FREE (format -> L_partition);
	CALIB_FREE (format -> S);
	CALIB_FREE (format -> T);
	CALIB_FREE (format -> opstart);
	CALIB_FREE (format -> kwdim);
	CALIB_FREE (format -> min_kid_width);
#endif

	CALIB_FREE (format -> linedims);
	CALIB_FREE (format -> opwidths);
	CALIB_FREE (format -> numops);

	memset (format, 0, sizeof (*format));
}

/*
 * Compute the tile dimensions that result when appending the two
 * given tiledim rectangles together (side by side, with a single space
 * between them).
 */

	static
	void
append_tiles (

struct tiledim *	dst,		/* IN/OUT: Destination tiledim */
const struct tiledim *	left,		/* IN: Left tiledim to append */
const struct tiledim *	right		/* IN: Right tiledim to append */
)
{
int		width, ascent, descent, height;

	width	= left -> width + 1 + right -> width;
	ascent	= _MAX (left -> ascent,  right -> ascent);
	descent	= _MAX (left -> descent, right -> descent);
	height	= ascent + 1 + descent;

	dst -> width	= width;
	dst -> height	= height;
	dst -> ascent	= ascent;
	dst -> descent	= descent;
}

/*
 * Return the area of the rectangle represented by the given tiledim.
 */

	static
	calib_int64u
tile_area (

const struct tiledim *		tp	/* IN: Tiledim to get area of */
)
{
calib_int64u		h, w;

	h = tp -> height;
	w = tp -> width;

	return (h * w);
}

/*
 * Given the dimensions of an expression line, return a "biased" version
 * of its area that penalizes lines having lots of trailing empty space.
 * (We include a fraction of this empty space in the returned area.)
 */

	static
	calib_int64u
biased_line_area (

const struct tiledim *		tp,		/* IN: Dims to get area of */
int				maxwidth	/* IN: Max line width */
)
{
calib_int64u		h, w, w2;

	h = tp -> height;
	w = tp -> width;

	CALIB_FATAL_ERROR_IF (w > maxwidth);

	w2 = maxwidth - w;

	return (h * w + h * w2 / 4);
}

/*
 * We give up -- expression is too wide to print.
 */

	static
	void
too_wide (

struct PPInfo *		ppip		/* IN/OUT: Global prettyprint info */
)
{
	fprintf (ppip -> fp, "Expression too wide to print.\n");
}

/*
 * Plot the given grdim sub-expression into the given canvas at the given
 * x,y coordinates.
 */

	static
	void
plot_grdim (

struct canvas *		cp,		/* IN/OUT: canvas to plot into */
struct grdim *		dp,		/* IN: sub-expression to plot */
int			x,		/* IN: x coordinate */
int			y		/* IN: y coordinate */
)
{
int			w, mask;
struct grdim_list *	dlp;
char *			buf;
char *			str;
struct grdim *		dp2;
struct grdim *		dp3;
struct grdim *		func;
char			numbuf [32];

	CALIB_FATAL_ERROR_IF ((cp EQ NULL) OR (cp EQ NULL));
	CALIB_FATAL_ERROR_IF ((x < 0) OR (y < 0));

	if (dp -> parens) {
		plot_canvas (cp, "(", x, y);
		plot_canvas (cp, ")", x + dp -> maxtile.width - 1, y);
		++x;
	}

	switch (dp -> op) {
	case GRDIM_OP_ADD:
		dlp = dp -> u.list;
		CALIB_FATAL_ERROR_IF (dlp EQ NULL);
		for (; dlp NE NULL; dlp = dlp -> next) {
			dp2 =  dlp -> operand;
			/* Each term should have its own unary "+" or	*/
			/* unary "-" on it, except the first term won't	*/
			/* have a unary "+".				*/
			plot_grdim (cp, dp2, x, y);
			x += (dp2 -> maxtile.width + 1);
		}
		break;

	case GRDIM_OP_UPLUS:
		plot_canvas (cp, "+", x, y);
		plot_grdim (cp, dp -> u.unop, x + 2, y);
		break;

	case GRDIM_OP_NEG:
		plot_canvas (cp, "-", x, y);
		plot_grdim (cp, dp -> u.unop, x + 2, y);
		break;

	case GRDIM_OP_MUL:
		dlp = dp -> u.list;
		CALIB_FATAL_ERROR_IF (dlp EQ NULL);
		dp2 = dlp -> operand;
		plot_grdim (cp, dp2, x, y);
		x += dp2 -> maxtile.width + 1;
		for (dlp = dlp -> next; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			plot_grdim (cp, dp2, x, y);
			x += (dp2 -> maxtile.width + 1);
		}
		break;

	case GRDIM_OP_FRAC:
		w = dp -> maxtile.width;
		if (dp -> parens) {
			w -= 2;
		}
		buf = CALIB_NEWA (w + 1, char);
		memset (buf, '-', w);
		buf [w] = '\0';
		plot_canvas (cp, buf, x, y);
		free (buf);
		dp2 = dp -> u.frac.numer;
		dp3 = dp -> u.frac.denom;
		plot_grdim (cp,
			    dp2,
			    x + (w - dp2 -> maxtile.width) / 2,
			    y - 1 - dp2 -> maxtile.descent);
		plot_grdim (cp,
			    dp3,
			    x + (w - dp3 -> maxtile.width) / 2,
			    y + 1 + dp3 -> maxtile.ascent);
		break;

	case GRDIM_OP_IPOW:
		dp2 = dp -> u.ipow.base;
		sprintf (numbuf, "%d", dp -> u.ipow.power);
		plot_grdim (cp, dp2, x, y);
		plot_canvas (cp, numbuf, x + dp2 -> maxtile.width, y - 1);
		break;

	case GRDIM_OP_VAR:
		plot_canvas (cp, dp -> u.var, x, y);
		break;

	case GRDIM_OP_Z:
		plot_canvas (cp, dp -> u.zstr, x, y);
		break;

	case GRDIM_OP_FUNC:
		/* Note: the argument list syntax (open paren, commas	*/
		/* and closing paren) are included as GRDIM_OP_FIX	*/
		/* prefix and suffix strings in these operands, so they	*/
		/* should appear automagically here.			*/
		func	= dp -> u.func.func;
		plot_grdim (cp, func, x, y);
		x += func -> maxtile.width;
		++x;
		for (dlp = dp -> u.func.args; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			plot_grdim (cp, dp2, x, y);
			x += (dp2 -> maxtile.width + 1);
		}
		break;

	case GRDIM_OP_FIX:
		if (dp -> type EQ TYPE_FLAT) {
			/* We are flat, and so is our parent. */
			mask = FIX_FLAT;
		}
		else if (dp -> parent_type EQ TYPE_FLAT) {
			/* We are 2d, but our parent is flat. */
			mask = FIX_TRANS;
		}
		else {
			mask = FIX_2D;
		}
		str = dp -> u.fix.prefix;
		if ((str NE NULL) AND
		    ((dp -> u.fix.pflags & mask) NE 0)) {
			/* Display the prefix string. */
			plot_canvas (cp, str, x, y);
			x += strlen (str);
		}
		/* Display sub-expression being prefixed / suffixed. */
		plot_grdim (cp, dp -> u.fix.op, x, y);
		x += dp -> u.fix.op -> maxtile.width;
		str = dp -> u.fix.suffix;
		if ((str NE NULL) AND
		    ((dp -> u.fix.sflags & mask) NE 0)) {
			/* Display the suffix string. */
			plot_canvas (cp, str, x, y);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Compute the dimensions of the given VTILE when forced to have the
 * given width.
 */

	static
	void
vtile_dims (

struct tiledim *	dims,	/* IN/OUT: dimensions */
struct grdim *		dp,	/* IN: VTILE sub-expr to format */
int			width	/* IN: desired width */
)
{
int			h, w, nv, paren_w, space;
struct grdim_list *	dlp;
struct grdim *		dp2;
struct grdim *		dp3;
struct tiledim		ldims, ldims2;
char			numbuf [32];

	CALIB_FATAL_ERROR_IF ((dims EQ NULL) OR (dp EQ NULL));
	CALIB_FATAL_ERROR_IF (dp -> type NE TYPE_VTILE);
	CALIB_FATAL_ERROR_IF (width <= 0);

	paren_w = 0;
	if (dp -> parens) {
		width -= 2;
		paren_w = 2;
	}

	switch (dp -> op) {
	case GRDIM_OP_ADD:
	case GRDIM_OP_MUL:
		dlp = dp -> u.list;
		CALIB_FATAL_ERROR_IF (dlp EQ NULL);

		/* All operands must be TILE, except for one VTILE.	*/
		/* Compute aggregate dimensions of all TILEs, and find	*/
		/* the VTILE.						*/

		ldims.width	= 0;
		ldims.height	= 1;
		ldims.ascent	= 0;
		ldims.descent	= 0;

		nv = 0;
		space = 0;
		dp3 = NULL;
		for (; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			if (dp2 -> type EQ TYPE_TILE) {
				ldims.width += (space + dp2 -> maxtile.width);
				update_tile_heights (&ldims, &(dp2 -> maxtile));
				space = 1;
			}
			else {
				++nv;
				dp3 = dp2;
			}
		}
		CALIB_FATAL_ERROR_IF ((nv NE 1) OR (dp3 EQ NULL));
		w = width - ldims.width - 1;
		vtile_dims (&ldims2, dp3, w);
		ldims.width += (1 + ldims2.width);
		update_tile_heights (&ldims, &ldims2);
		ldims.width += paren_w;
		*dims = ldims;
		break;

	case GRDIM_OP_UPLUS:
	case GRDIM_OP_NEG:
		vtile_dims (&ldims, dp -> u.unop, width - 2);
		ldims.width += (2 + paren_w);
		*dims = ldims;
		break;

	case GRDIM_OP_FRAC:
		dp2 = dp -> u.frac.numer;
		dp3 = dp -> u.frac.denom;
		if (dp2 -> type EQ TYPE_TILE) {
			w = dp2 -> maxtile.width;
			CALIB_FATAL_ERROR_IF (w > width);
			ldims = dp2 -> maxtile;
		}
		else {
			CALIB_FATAL_ERROR_IF (dp2 -> type NE TYPE_VTILE);
			w = _MIN (dp2 -> maxtile.width, width);
			vtile_dims (&ldims, dp2, w);
			CALIB_FATAL_ERROR_IF (ldims.width NE w);
		}
		if (dp3 -> type EQ TYPE_TILE) {
			w = dp3 -> maxtile.width;
			CALIB_FATAL_ERROR_IF (w > width);
			ldims2 = dp3 -> maxtile;
		}
		else {
			CALIB_FATAL_ERROR_IF (dp3 -> type NE TYPE_VTILE);
			w = _MIN (dp3 -> maxtile.width, width);
			vtile_dims (&ldims2, dp3, w);
			CALIB_FATAL_ERROR_IF (ldims2.width NE w);
		}
		CALIB_FATAL_ERROR_IF ((ldims.width NE width) AND
				      (ldims2.width NE width));
		dims -> width = width + paren_w;
		dims -> height = ldims.height + 1 + ldims2.height;
		dims -> ascent = ldims.height;
		dims -> descent = ldims2.height;
		break;

	case GRDIM_OP_IPOW:
		CALIB_FATAL_ERROR_IF (dp -> u.ipow.power <= 1);
		sprintf (numbuf, "%d", dp -> u.ipow.power);
		w = width - strlen (numbuf);
		vtile_dims (&ldims, dp -> u.ipow.base, w);
		ldims.width += w + paren_w;
		if (ldims.ascent <= 0) {
			ldims.ascent = 1;
			++ldims.height;
		}
		*dims = ldims;
		break;

	case GRDIM_OP_VAR:
		w = strlen (dp -> u.var);
		CALIB_FATAL_ERROR_IF (w NE width);
		dims -> width = w + paren_w;
		dims -> height = 1;
		dims -> ascent = 0;
		dims -> descent = 0;
		break;

	case GRDIM_OP_Z:
		w = strlen (dp -> u.zstr);
		CALIB_FATAL_ERROR_IF (width < 2);
		CALIB_FATAL_ERROR_IF (w < width);
		dims -> width = width + paren_w;
		if (w EQ width) {
			h = 1;
		}
		else {
			h = (w - 1 + (width - 2)) / (width - 1);
		}
		dims -> width = width + paren_w;
		dims -> height	= h;
		dims -> ascent	= h - 1;
		dims -> descent = 0;
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Plot the given "vtile" sub-expression into the given canvas at the given
 * x,y coordinates, forcing it to have the given width.
 */

	static
	void
plot_vtile_width (

struct canvas *		cp,		/* IN/OUT: canvas to plot into */
struct grdim *		dp,		/* IN: sub-expression to plot */
int			x,		/* IN: x coordinate */
int			y,		/* IN: y coordinate */
int			width		/* IN: desired width */
)
{
int			w, nv;
struct grdim_list *	dlp;
struct grdim *		dp2;
struct grdim *		dp3;
char *			buf;
char *			beginp;
char *			endp;
char *			p;
struct tiledim		dims;
char			numbuf [32];

	CALIB_FATAL_ERROR_IF ((cp EQ NULL) OR (dp EQ NULL));
	CALIB_FATAL_ERROR_IF ((x < 0) OR (y < 0));

	CALIB_FATAL_ERROR_IF (dp -> type NE TYPE_VTILE);

	if (dp -> parens) {
		plot_canvas (cp, "(", x, y);
		plot_canvas (cp, ")", x + width - 1, y);
		++x;
		width -= 2;
	}

	switch (dp -> op) {
	case GRDIM_OP_ADD:
		dlp = dp -> u.list;
		CALIB_FATAL_ERROR_IF (dlp EQ NULL);

		/* All operands should be TILE, except for one VTILE. */
		/* Add up widths of TILEs. */
		nv = 0;
		w = -1;
		dp2 = NULL;
		for (; dlp NE NULL; dlp = dlp -> next) {
			dp3 = dlp -> operand;
			if (dp3 -> type EQ TYPE_VTILE) {
				++nv;
				dp2 = dp3;
			}
			else {
				CALIB_FATAL_ERROR_IF (dp3 -> type NE TYPE_TILE);
				w += (1 + dp3 -> maxtile.width);
			}
		}
		CALIB_FATAL_ERROR_IF (nv NE 1);
		CALIB_FATAL_ERROR_IF (dp2 EQ NULL);
		CALIB_FATAL_ERROR_IF (w <= 0);
		CALIB_FATAL_ERROR_IF (w + 1 + dp2 -> mintile.width > width);
		w = (width - (w + 1));

		for (dlp = dp -> u.list; dlp NE NULL; dlp = dlp -> next) {
			dp2 =  dlp -> operand;
			if (dp2 -> type EQ TYPE_TILE) {
				plot_grdim (cp, dp2, x, y);
				x += (dp2 -> maxtile.width + 1);
			}
			else {
				CALIB_FATAL_ERROR_IF (dp2 -> type NE TYPE_VTILE);
				plot_vtile_width (cp, dp2, x, y, w);
				x += (w + 1);
			}
		}
		break;

	case GRDIM_OP_UPLUS:
		plot_canvas (cp, "+", x, y);
		plot_vtile_width (cp, dp -> u.unop, x + 2, y, width - 2);
		break;

	case GRDIM_OP_NEG:
		plot_canvas (cp, "-", x, y);
		plot_vtile_width (cp, dp -> u.unop, x + 2, y, width - 2);
		break;

	case GRDIM_OP_MUL:
		dlp = dp -> u.list;
		CALIB_FATAL_ERROR_IF (dlp EQ NULL);

		/* All operands should be TILE, except for one VTILE. */
		/* Add up widths of TILEs. */
		nv = 0;
		w = -1;
		dp3 = NULL;
		for (; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			if (dp2 -> type EQ TYPE_VTILE) {
				++nv;
				dp3 = dp2;
			}
			else {
				CALIB_FATAL_ERROR_IF (dp2 -> type NE TYPE_TILE);
				w += (1 + dp2 -> maxtile.width);
			}
		}
		CALIB_FATAL_ERROR_IF (nv NE 1);
		CALIB_FATAL_ERROR_IF (dp3 EQ NULL);
		CALIB_FATAL_ERROR_IF (w <= 0);
		CALIB_FATAL_ERROR_IF (w + 1 + dp3 -> mintile.width > width);
		w = (width - (w + 1));

		dlp = dp -> u.list;
		dp2 = dlp -> operand;
		if (dp2 -> type EQ TYPE_TILE) {
			plot_grdim (cp, dp2, x, y);
			x += dp2 -> maxtile.width + 1;
		}
		else {
			CALIB_FATAL_ERROR_IF (dp2 -> type NE TYPE_VTILE);
			plot_vtile_width (cp, dp2, x, y, w);
			x += (w + 1);
		}
		for (dlp = dlp -> next; dlp NE NULL; dlp = dlp -> next) {
			dp2 = dlp -> operand;
			if (dp2 -> type EQ TYPE_TILE) {
				plot_grdim (cp, dp2, x, y);
				x += (dp2 -> maxtile.width + 1);
			}
			else {
				plot_vtile_width (cp, dp2, x, y, w);
				x += (w + 1);
			}
		}
		break;

	case GRDIM_OP_FRAC:
		w = width;
		buf = CALIB_NEWA (w + 1, char);
		memset (buf, '-', w);
		buf [w] = '\0';
		plot_canvas (cp, buf, x, y);
		free (buf);
		dp2 = dp -> u.frac.numer;
		dp3 = dp -> u.frac.denom;
		if ((dp2 -> type EQ TYPE_TILE) OR
		    ((dp2 -> type EQ TYPE_VTILE) AND
		     (dp2 -> maxtile.width < width))) {
			plot_grdim (cp,
				    dp2,
				    x + (w - dp2 -> maxtile.width) / 2,
				    y - 1 - dp2 -> maxtile.descent);
		}
		else {
			CALIB_FATAL_ERROR_IF (dp2 -> type NE TYPE_VTILE);
			CALIB_FATAL_ERROR_IF (dp2 -> mintile.width > width);
			vtile_dims (&dims, dp2, width);
			CALIB_FATAL_ERROR_IF (dims.width NE width);
			plot_vtile_width (cp,
					  dp2,
					  x,
					  y - 1 - dims.descent,
					  width);
		}
		if ((dp3 -> type EQ TYPE_TILE) OR
		    ((dp3 -> type EQ TYPE_VTILE) AND
		     (dp3 -> maxtile.width < width))) {
			plot_grdim (cp,
				    dp3,
				    x + (w - dp3 -> maxtile.width) / 2,
				    y + 1 + dp3 -> maxtile.ascent);
		}
		else {
			CALIB_FATAL_ERROR_IF (dp3 -> type NE TYPE_VTILE);
			CALIB_FATAL_ERROR_IF (dp3 -> mintile.width > width);
			vtile_dims (&dims, dp3, width);
			CALIB_FATAL_ERROR_IF (dims.width NE width);
			plot_vtile_width (cp,
					  dp3,
					  x,
					  y + 1 + dims.ascent,
					  width);
		}
		break;

	case GRDIM_OP_IPOW:
		dp2 = dp -> u.ipow.base;
		sprintf (numbuf, "%d", dp -> u.ipow.power);
		w = width - strlen (numbuf);
		plot_vtile_width (cp, dp2, x, y, w);
		plot_canvas (cp, numbuf, x + w, y - 1);
		break;

	case GRDIM_OP_VAR:
		CALIB_FATAL_ERROR_IF (strlen (dp -> u.var) NE width);
		plot_canvas (cp, dp -> u.var, x, y);
		break;

	case GRDIM_OP_Z:
		/* Plot number as a rectangle of the specified width. */
		w = strlen (dp -> u.zstr);
		CALIB_FATAL_ERROR_IF (width < 2);
		CALIB_FATAL_ERROR_IF (w < width);
		if (w EQ width) {
			plot_canvas (cp, dp -> u.zstr, x, y);
		}
		else {
			buf = CALIB_NEWA (width + 1, char);
			beginp = dp -> u.zstr;
			endp = beginp + w;
			CALIB_FATAL_ERROR_IF (*endp NE '\0');
			p = endp - width;
			memcpy (buf, p, width);
			buf [width] = '\0';
			plot_canvas (cp, p, x, y);
			--y;
			do {
				endp = p;
				p -= (width - 1);
				if (p < beginp) {
					p = beginp;
				}
				memcpy (buf, p, (endp - p));
				buf [endp - p] = '\\';
				buf [endp - p + 1] = '\0';
				plot_canvas (cp, buf, x + width - (endp - p) - 1, y);
				--y;
			} while (p > beginp);
			free (buf);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Initialize the given canvas to be the given number of lines tall.
 */

	static
	void
init_canvas (

struct canvas *		cp,		/* IN/OUT: canvas to initialize */
int			nlines		/* IN: number of lines */
)
{
int			i;
struct cline *		lines;
struct cline *		lp;

	CALIB_FATAL_ERROR_IF (nlines <= 0);

	lines = CALIB_NEWA (nlines, struct cline);
	for (i = 0; i < nlines; i++) {
		lp = &lines [i];
		lp -> len	= 0;
		lp -> text	= strdup ("");
	}

	cp -> nlines	= nlines;
	cp -> lines	= lines;
}

/*
 * Plot the given integer ***RIGHT-TO-LEFT*** starting at the given
 * x,y coordinates.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
plot_canvas_z (

struct canvas *		cp,		/* IN/OUT: canvas to plot into */
mpz_srcptr		zval,		/* IN: integer value */
int			x,		/* IN: x coordinate */
int			y		/* IN: y coordinate */
)
{
int		sign;
mpz_t		z;
char		buf [2];

	buf [1] = '\0';

	sign = mpz_sgn (zval);
	if (sign EQ 0) {
		buf [0] = '0';
		plot_canvas (cp, buf, x, y);
		return;
	}

	CALIB_FATAL_ERROR_IF (sign < 0);

	/* Make modifiable local copy of value. */
	mpz_init_set (z, zval);

	while (mpz_sgn (z) > 0) {
		buf [0] = '0' + mpz_fdiv_q_ui (z, z, 10);
		plot_canvas (cp, buf, x, y);
		--x;
	}

	mpz_clear (z);
}

#pragma GCC diagnostic pop

/*
 * Plot the given text horizontally into the given canvas, starting at
 * the given x,y coordinates.
 */

	static
	void
plot_canvas (

struct canvas *		cp,		/* IN/OUT: canvas to plot into */
const char *		text,		/* IN: text to plot */
int			x,		/* IN: x coordinate */
int			y		/* IN: y coordinate */
)
{
int			l1, l2, l3;
char *			lbuf;
struct cline *		lp;

	CALIB_FATAL_ERROR_IF ((cp EQ NULL) OR (text EQ NULL));
	CALIB_FATAL_ERROR_IF ((x < 0) OR (y < 0) OR (y >= cp -> nlines));

	CALIB_FATAL_ERROR_IF (y >= cp -> nlines);

	lp = &(cp -> lines [y]);

	l1 = strlen (text);
	l2 = x + l1;
	l3 = lp -> len;
	if (l2 > l3) {
		/* Make line longer, and pad with blanks. */
		lbuf = CALIB_NEWA (l2 + 1, char);
		memcpy (lbuf, lp -> text, l3);
		memset (&lbuf [l3], ' ', l2 - l3);
		lbuf [l2] = '\0';
		free (lp -> text);
		lp -> text = lbuf;
		lp -> len = l2;
	}
	memcpy (&(lp -> text [x]), text, l1);
}

/*
 * Write the given canvas to the current output stream.
 */

	static
	void
write_canvas (

struct PPInfo *		ppip,		/* IN/OUT: Global prettyprint info */
struct canvas *		cp		/* IN: canvas to write */
)
{
int		i, n;
struct cline *	lines;
struct cline *	lp;

	CALIB_FATAL_ERROR_IF ((ppip EQ NULL) OR (cp EQ NULL));

	n	= cp -> nlines;
	lines	= cp -> lines;
	for (i = 0; i < n; i++) {
		lp = &lines [i];
		fprintf (ppip -> fp, "%s\n", lp -> text);
	}
	fflush (ppip -> fp);
}

/*
 * Destroy the given canvas.
 */

	static
	void
free_canvas (

struct canvas *		cp		/* IN: canvas to clean up. */
)
{
int		i, n;
struct cline *	lines;
struct cline *	lp;

	n = cp -> nlines;
	lines = cp -> lines;
	cp -> nlines	= 0;
	cp -> lines	= NULL;
	for (i = 0; i < n; i++) {
		lp = &lines [i];
		free (lp -> text);
		lp -> len  = 0;
		lp -> text = NULL;
	}
	free (lines);
}

/*
 * Return a dynamically allocated string that is the concatenation of
 * the two given strings.
 */

	static
	char *
string_cat2 (

const char *		s1,		/* IN: First string */
const char *		s2		/* IN: Second string */
)
{
size_t		l1, l2;
char *		s;

	l1 = strlen (s1);
	l2 = strlen (s2);

	s = CALIB_NEWA (l1 + l2 + 1, char);

	memcpy (s, s1, l1);
	memcpy (s + l1, s2, l2);
	s [l1 + l2] = '\0';

	return (s);
}
