/**********************************************************************

	$Id: rat.c,v 1.1 2024/10/20 00:36:28 warme Exp $

	File:	rat.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Rational functions: the quotient field P/P,
	where P=Z[x,y,z], the ring of multi-variate polynomials
	over the integers.

***********************************************************************

	Modification Log:

	a-1:	10/01/2015	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/rat.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/types.h"
#include "calib/Zxyz.h"
#include <gmp.h>
#include "ivechash.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Some macros that avoid confusing certain editors.
 */

#define	LPAREN_STRING	"("
#define	RPAREN_STRING	")"

/*
 * An object to represent a "pool" of relatively prime polynomials (RPP's)
 * over a given ring.  We construct the pool (in a "batch" manner) from a
 * given list of polynomials.  The resulting pool of RPP's is then fixed,
 * and the pool becomes an array, with each RPP in the pool having a
 * unique index from 0 to Num_RPP-1.
 */

struct RppPool {
	int			num_rpp;	/* Number of RPPs in pool */
	const struct calib_Zxyz_dom *
				dom;		/* Common ring for all RPPs */
	struct calib_Zxyz_obj *	rpp;		/* Array of RPPs */
};

/*
 * An object to represent a polynomial in a partially factored form:
 *
 *		k * Rpp1^a1 * Rpp2^a2 * Rpp3^a3 * ... * Rppn^an
 *
 * where k is an integer constant, Rpp1 ... Rppn are relatively prime
 * polynomials from an associated RppPool object, and a1 ... an are
 * non-negative integers.  Since the RppPool object stores all of the
 * Rpp's, all this object needs to store is k and the powers a1 ... an.
 */

struct FPoly {
	mpz_t		Konst;		/* Constant factor */
	int *		pow;		/* Array of powers */
};

/*
 * A rational function with denominator in FPoly form.
 */

struct PFRat {
	struct calib_Zxyz_obj	numer;	/* Numerator */
	struct FPoly		denom;	/* Denominator */
};

/*
 * An object to store a polynomial in such a manner that
 * it may later be split into two or more factors dynamically.
 * This representation is recursive, and consists of the
 * following three cases:
 *
 *	1. A NULL pointer represents 1.
 *	2. The other base case is a simple polynomial.
 *	3. Otherwise, it PS1**k * PS2 where PS1 and
 *	   PS2 are recursive instances.
 *
 * The method is to introduce new polynomials as case 2,
 * and perform pair-wise GCD operations.  As non-trivial
 * common factors are found, they are used to convert
 * nodes from case 2 to case 3.
 */

struct PSplit {
	int			index;	/* RPP index */
	int			sign;	/* For sign flips */
	int			sfpow;	/* Power */
	struct calib_Zxyz_obj	*
				poly;	/* Polynomial this node */
					/* represents (not necessarily */
					/* an RPP) */
	struct PSplit *		sfact;	/* poly = sign*sfact**sfpow*nfact, */
					/* where sfact and nfact are */
					/* relatively prime, and */
					/* both subject to further */
					/* splitting */
	struct PSplit *		nfact;	/* Next factor (or NULL) */
	struct PSplit *		forw;	/* Forward link */
	struct PSplit *		back;	/* Backward link */
};

/*
 * A pre-computed table of powers for a single polynomial.
 */

struct poly_powers {
	int			max_power;	/* Maximum power stored. */
	struct calib_Zxyz_obj *	tab;		/* Table of powers */
						/* 1...max_power */
};

/*
 * For a certain sparse subset of variables, the following object
 * stores tables containing pre-computed powers of corresponding
 * substitution polynomials.
 */

struct subst_var_powers {
	int			nvars;	/* Number of vars */
	int			ntabs;	/* Size of powers array */
	struct poly_powers *	powers;	/* Array of power tables */
	struct poly_powers **	map;	/* Map var to appropriate */
					/* power table */
};

/*
 * This object is our client extension to the hash table entry
 * used in substitute_poly_to_pfrat().
 *
 * This is essentially the info held by an FPoly, except that the
 * array of RPP powers resides in the hash table key (a vector of integers).
 */

struct hashent {
	struct calib_Zxyz_obj
			numer;		/* Numerator */
	mpz_t		den_konst;	/* Denominator constant factor */
	struct ivec_hash_entry *
			owner;		/* Hash table entry we belont to */
};

/*
 * Local Functions
 */

static void	clear_psplit_head (struct PSplit * psp);
static void	clear_rpp_pool (struct RppPool * pool);
static struct subst_var_powers *
		compute_rpp_powers (
				struct RppPool *	pool,
				const int *		maxpow);
static struct subst_var_powers *
		compute_svar_powers (
				const struct calib_Zxyz_obj *	poly,
				struct PFRat **			subst_array,
				const struct calib_Zxyz_dom *	subring,
				const int *			varmap);
static struct PFRat *
		convert_subst_list_to_pfrat_array (
				struct RppPool *		pool,
				struct calib_subst_list *	slist,
				const struct calib_Zxyz_dom *	subring,
				const int *			varmap);
static void	eval_rpp_term (struct calib_Zxyz_obj *		result,
			       const int *			pow,
			       struct subst_var_powers *	rpp_powers,
			       struct RppPool *			pool);
static void	eval_svar_term (struct PFRat *			result,
				const int *			pow,
				struct subst_var_powers *	svar_powers,
				struct PFRat **			subst_array,
				struct RppPool *		pool);
static void	free_svar_powers (struct subst_var_powers * svp);
static int	get_sign (const struct calib_Zxyz_obj * poly);
static void	hash_entry_op (struct ivec_hash_entry *	ep,
			       enum ivec_entry_op	op,
			       void *			data);
static void	init_psplit_head (struct PSplit * psp);
static void	insert_psplit (struct PSplit *	rpp_head,
			       struct PSplit *	split_head,
			       struct PSplit *	p);
static calib_bool
		is_constant_poly (const struct calib_Zxyz_obj * poly);
static void	pfrat_clear (struct PFRat * rfp);
static void	pfrat_init (struct PFRat *			rfp,
			     const struct calib_Zxyz_dom *	ring,
			     struct RppPool *			pool);
static void	rat_add (struct calib_rat_obj *		rop,
			 const struct calib_rat_obj *	op1,
			 const struct calib_rat_obj *	op2);
static void	rat_canonicalize (struct calib_rat_obj *	op);
static void	rat_clear (struct calib_rat_obj *	x);
static void	rat_div (struct calib_rat_obj *		rop,
			 const struct calib_rat_obj *	op1,
			 const struct calib_rat_obj *	op2);
static struct calib_Zxyz_factor *
		rat_factor (const struct calib_rat_obj *	op);
static void	rat_free_slist (struct calib_subst_list *	slist);
static void	rat_init (const struct calib_rat_dom *	dom,
			  struct calib_rat_obj *	x);
static void	rat_inv (struct calib_rat_obj *		rop,
			 const struct calib_rat_obj *	op);
static void	rat_ipow (struct calib_rat_obj *	rop,
			  const struct calib_rat_obj *	op,
			  int				power);
static void	rat_mul (struct calib_rat_obj *		rop,
			 const struct calib_rat_obj *	op1,
			 const struct calib_rat_obj *	op2);
static void	rat_mul_q (struct calib_rat_obj *	rop,
			   const struct calib_rat_obj *	op1,
			   mpq_srcptr			op2);
static void	rat_mul_z (struct calib_rat_obj *	rop,
			   const struct calib_rat_obj *	op1,
			   mpz_srcptr			op2);
static void	rat_neg (struct calib_rat_obj *		rop,
			 const struct calib_rat_obj *	op);
static calib_bool
		rat_onep (const struct calib_rat_obj *	op);
static void	rat_print (const struct calib_rat_obj * op);
static void	rat_print_nnl (const struct calib_rat_obj * op);
static void	rat_set (struct calib_rat_obj *		rop,
			 const struct calib_rat_obj *	op);
static void	rat_set_q (struct calib_rat_obj *	rop,
			   mpq_srcptr			op);
static void	rat_set_si (struct calib_rat_obj *	rop,
			    calib_si_t			op);
static void	rat_set_var_power (struct calib_rat_obj *	rop,
				   int				var,
				   int				power);
static void	rat_set_z (struct calib_rat_obj *	rop,
			   mpz_srcptr			op);
static void	rat_sub (struct calib_rat_obj *		rop,
			 const struct calib_rat_obj *	op1,
			 const struct calib_rat_obj *	op2);
static void	rat_substitute_with_varmap (
				struct calib_rat_obj *		rop,
				const struct calib_rat_obj *	op,
				const int *			varmap,
				struct calib_subst_list *	slist);
static void	substitute_poly_to_pfrat (
				struct PFRat *			result,
				const struct calib_Zxyz_obj *	poly,
				const int *			varmap,
				struct PFRat **			subst_array,
				struct RppPool *		pool);
static int	traverse_psplit (int * pow, int K, struct PSplit * p);
static calib_bool
		rat_zerop (const struct calib_rat_obj *	op);

/*
 * Initialize a rational function object.
 */

	static
	void
rat_init (

const struct calib_rat_dom *	dom,	/* IN: rational function domain */
struct calib_rat_obj *		x	/* IN/OUT: rat to initialize */
)
{
const struct calib_Zxyz_dom *	ring;

	ring = dom -> ring;

	x -> dom	= dom;
	ring -> init (ring, &(x -> numer));
	ring -> init (ring, &(x -> denom));
	ring -> set_si (&(x -> denom), 1);
}

/*
 * Clear out a rational function object.
 */

	static
	void
rat_clear (

struct calib_rat_obj *	x		/* IN: rational object to clear */
)
{
const struct calib_Zxyz_dom *	ring;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	ring = x -> numer.dom;

	CALIB_FATAL_ERROR_IF (ring NE x -> denom.dom);

	ring -> clear (&(x -> denom));
	ring -> clear (&(x -> numer));

	memset (x, 0, sizeof (*x));
}

/*
 * Copy the given rational function.
 */

	static
	void
rat_set (

struct calib_rat_obj *		rop,
const struct calib_rat_obj *	op
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	if (rop NE op) {
		Zxyz	= dom -> ring;

		Zxyz -> set (&(rop -> numer), &(op -> numer));
		Zxyz -> set (&(rop -> denom), &(op -> denom));
	}
}

/*
 * Set the given rational function to have the given constant,
 * signed integer value.
 */

	static
	void
rat_set_si (

struct calib_rat_obj *	rop,		/* IN/OUT: destination rat */
calib_si_t		op		/* IN: integer value */
)
{
const struct calib_Zxyz_dom *	Zxyz;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	Zxyz	= rop -> dom -> ring;

	Zxyz -> set_si (&(rop -> numer), op);
	Zxyz -> set_si (&(rop -> denom), 1);
}

/*
 * Set the given rational function to have the given GMP integer value.
 */

	static
	void
rat_set_z (

struct calib_rat_obj *	rop,		/* IN/OUT: destination rat */
mpz_srcptr		op		/* IN: GMP integer value */
)
{
const struct calib_Zxyz_dom *	Zxyz;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	Zxyz	= rop -> dom -> ring;

	Zxyz -> set_z  (&(rop -> numer), op);
	Zxyz -> set_si (&(rop -> denom), 1);
}

/*
 * Set the given rational function to have the given GMP rational value.
 */

	static
	void
rat_set_q (

struct calib_rat_obj *	rop,		/* IN/OUT: destination rat */
mpq_srcptr		op		/* IN: GMP rational value */
)
{
const struct calib_Zxyz_dom *	Zxyz;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	Zxyz	= rop -> dom -> ring;

	Zxyz -> set_z (&(rop -> numer), mpq_numref (op));
	Zxyz -> set_z (&(rop -> denom), mpq_denref (op));
}

/*
 * Set the given rational function to be var**power.
 */

	static
	void
rat_set_var_power (

struct calib_rat_obj *	rop,		/* IN/OUT: destination rat */
int			var,		/* IN: Variable */
int			power		/* IN: Power to use */
)
{
const struct calib_Zxyz_dom *	Zxyz;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	Zxyz	= rop -> dom -> ring;

	if (power < 0) {
		power = - power;
		CALIB_FATAL_ERROR_IF (power < 0);

		Zxyz -> set_si (&(rop -> numer), 1);
		Zxyz -> set_var_power (&(rop -> denom), var, power);
	}
	else {
		Zxyz -> set_var_power (&(rop -> numer), var, power);
		Zxyz -> set_si (&(rop -> denom), 1);
	}
}

/*
 * Addition of two rational functions.
 */

	static
	void
rat_add (

struct calib_rat_obj *		rop,		/* IN/OUT: result */
const struct calib_rat_obj *	op1,		/* IN: first operand */
const struct calib_rat_obj *	op2		/* IN: second operand */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_Zxyz_obj *		cofact [2];
const struct calib_Zxyz_obj *	poly_ptrs [2];
struct calib_Zxyz_obj		G, A, B, TMP1, TMP2;
struct calib_rat_obj		RES;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));

	Zxyz	= dom -> ring;

	/* Note: it is possible to do a bit better here (e.g., cross-	*/
	/* operand cancellation).					*/

	Zxyz -> init (Zxyz, &G);
	Zxyz -> init (Zxyz, &A);
	Zxyz -> init (Zxyz, &B);
	Zxyz -> init (Zxyz, &TMP1);
	Zxyz -> init (Zxyz, &TMP2);

	dom  -> init (dom, &RES);

	/* Compute common denominator and cofactors. */

	cofact [0]	= &A;
	cofact [1]	= &B;
	poly_ptrs [0]	= &(op1 -> denom);
	poly_ptrs [1]	= &(op2 -> denom);

	Zxyz -> gcd_n (&G, cofact, 2, poly_ptrs);

	Zxyz -> mul (&TMP1, &(op1 -> numer), &B);
	Zxyz -> mul (&TMP2, &(op2 -> numer), &A);
	Zxyz -> add (&RES.numer, &TMP1, &TMP2);

	Zxyz -> mul (&TMP1, &G, &A);
	Zxyz -> mul (&RES.denom, &TMP1, &B);

	rat_canonicalize (&RES);

	rat_set (rop, &RES);

	dom -> clear (&RES);

	Zxyz -> clear (&TMP2);
	Zxyz -> clear (&TMP1);
	Zxyz -> clear (&B);
	Zxyz -> clear (&A);
	Zxyz -> clear (&G);
}

/*
 * Subtraction of two rational functions.
 */

	static
	void
rat_sub (

struct calib_rat_obj *		rop,		/* IN/OUT: result */
const struct calib_rat_obj *	op1,		/* IN: first operand */
const struct calib_rat_obj *	op2		/* IN: second operand */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_Zxyz_obj *		cofact [2];
const struct calib_Zxyz_obj *	poly_ptrs [2];
struct calib_Zxyz_obj		G, A, B, TMP1, TMP2;
struct calib_rat_obj		RES;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));

	Zxyz	= dom -> ring;

	/* Note: it is possible to do a bit better here (e.g., cross-	*/
	/* operand cancellation).					*/

	Zxyz -> init (Zxyz, &G);
	Zxyz -> init (Zxyz, &A);
	Zxyz -> init (Zxyz, &B);
	Zxyz -> init (Zxyz, &TMP1);
	Zxyz -> init (Zxyz, &TMP2);

	dom  -> init (dom, &RES);

	/* Compute common denominator and cofactors. */

	cofact [0]	= &A;
	cofact [1]	= &B;
	poly_ptrs [0]	= &(op1 -> denom);
	poly_ptrs [1]	= &(op2 -> denom);

	Zxyz -> gcd_n (&G, cofact, 2, poly_ptrs);

	Zxyz -> mul (&TMP1, &(op1 -> numer), &B);
	Zxyz -> mul (&TMP2, &(op2 -> numer), &A);
	Zxyz -> sub (&RES.numer, &TMP1, &TMP2);

	Zxyz -> mul (&TMP1, &G, &A);
	Zxyz -> mul (&RES.denom, &TMP1, &B);

	rat_canonicalize (&RES);

	rat_set (rop, &RES);

	dom -> clear (&RES);

	Zxyz -> clear (&TMP2);
	Zxyz -> clear (&TMP1);
	Zxyz -> clear (&B);
	Zxyz -> clear (&A);
	Zxyz -> clear (&G);
}

/*
 * Negation of a rational function.
 */

	static
	void
rat_neg (

struct calib_rat_obj *		rop,	/* IN/OUT: result */
const struct calib_rat_obj *	op	/* IN: rational function to negate */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_rat_obj		TMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	Zxyz	= dom -> ring;

	if (rop EQ op) {
		/* Negation in place. */
		Zxyz -> neg (&(rop -> numer), &(rop -> numer));
	}
	else {
		dom -> init (dom, &TMP);
		Zxyz -> neg (&TMP.numer, &(op -> numer));
		Zxyz -> set (&TMP.denom, &(op -> denom));
		rat_set (rop, &TMP);
		dom -> clear (&TMP);
	}
}

/*
 * Multiplication of two rational functions.
 */

	static
	void
rat_mul (

struct calib_rat_obj *		rop,		/* IN/OUT: result */
const struct calib_rat_obj *	op1,		/* IN: first operand */
const struct calib_rat_obj *	op2		/* IN: second operand */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_Zxyz_obj *		cofact [2];
const struct calib_Zxyz_obj *	poly_ptrs [2];
struct calib_Zxyz_obj		G1, G2, A1, A2, B1, B2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));

	Zxyz	= dom -> ring;

	Zxyz -> init (Zxyz, &G1);
	Zxyz -> init (Zxyz, &G2);
	Zxyz -> init (Zxyz, &A1);
	Zxyz -> init (Zxyz, &A2);
	Zxyz -> init (Zxyz, &B1);
	Zxyz -> init (Zxyz, &B2);

	/* Compute cross-GCDs and cofactors.  We don't need the GCDs,	*/
	/* only the cofactors.						*/

	cofact [0]	= &A1;
	cofact [1]	= &B1;
	poly_ptrs [0]	= &(op1 -> numer);
	poly_ptrs [1]	= &(op2 -> denom);

	Zxyz -> gcd_n (&G1, cofact, 2, poly_ptrs);

	cofact [0]	= &A2;
	cofact [1]	= &B2;
	poly_ptrs [0]	= &(op2 -> numer);
	poly_ptrs [1]	= &(op1 -> denom);

	Zxyz -> gcd_n (&G2, cofact, 2, poly_ptrs);

	Zxyz -> mul (&(rop -> numer), &A1, &A2);
	Zxyz -> mul (&(rop -> denom), &B1, &B2);

	Zxyz -> clear (&B2);
	Zxyz -> clear (&B1);
	Zxyz -> clear (&A2);
	Zxyz -> clear (&A1);
	Zxyz -> clear (&G2);
	Zxyz -> clear (&G1);
}

/*
 * Multiplication the given rational function by the given GMP integer.
 */

	static
	void
rat_mul_z (

struct calib_rat_obj *		rop,		/* IN/OUT: result */
const struct calib_rat_obj *	op1,		/* IN: first operand */
mpz_srcptr			op2		/* IN: second operand */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_rat_obj		RES;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE dom);

	Zxyz	= dom -> ring;

	dom -> init (dom, &RES);

	Zxyz -> mul_z (&RES.numer, &(op1 -> numer), op2);
	Zxyz -> set (&RES.denom, &(op1 -> denom));

	rat_canonicalize (&RES);

	rat_set (rop, &RES);

	dom -> clear (&RES);
}

/*
 * Multiplication the given rational function by the given GMP rational.
 */

	static
	void
rat_mul_q (

struct calib_rat_obj *		rop,		/* IN/OUT: result */
const struct calib_rat_obj *	op1,		/* IN: first operand */
mpq_srcptr			op2		/* IN: second operand */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_rat_obj		RES;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE dom);

	Zxyz	= dom -> ring;

	dom -> init (dom, &RES);

	Zxyz -> mul_z (&RES.numer, &(op1 -> numer), mpq_numref (op2));
	Zxyz -> mul_z (&RES.denom, &(op1 -> denom), mpq_denref (op2));

	rat_canonicalize (&RES);

	rat_set (rop, &RES);

	dom -> clear (&RES);
}

/*
 * Take the given rational function to the given integer power.
 */

	static
	void
rat_ipow (

struct calib_rat_obj *		rop,	/* IN/OUT: result rational function */
const struct calib_rat_obj *	op,	/* IN: rational func to exponentiate */
int				power	/* IN: integer power */
)
{
unsigned int			upower;
const struct calib_rat_dom *	dom;
struct calib_rat_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	dom -> init (dom, &RES);
	dom -> init (dom, &PREG);
	dom -> set_si (&RES, 1);

	if (power < 0) {
		dom -> inv (&PREG, op);
		upower = - power;
	}
	else {
		dom -> set (&PREG, op);
		upower = power;
	}

	while (upower > 0) {
		if ((upower & 1) NE 0) {
			rat_mul (&RES, &RES, &PREG);
		}
		rat_mul (&PREG, &PREG, &PREG);
		upower >>= 1;
	}

	rat_set (rop, &RES);

	rat_clear (&PREG);
	rat_clear (&RES);
}

/*
 * Take the reciprocal of the given rational function.
 */

	static
	void
rat_inv (

struct calib_rat_obj *		rop,	/* IN/OUT: result */
const struct calib_rat_obj *	op	/* IN: rational function to invert */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_Zxyz_obj		TMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	Zxyz	= dom -> ring;

	/* Must not divide by zero! */
	CALIB_FATAL_ERROR_IF (Zxyz -> zerop (&(op -> numer)));

	if (rop EQ op) {
		/* Inversion in place.  Just swap objects. */
		TMP = rop -> numer;
		rop -> numer = rop -> denom;
		rop -> denom = TMP;
	}
	else {
		Zxyz -> set (&(rop -> denom), &(op -> numer));
		Zxyz -> set (&(rop -> numer), &(op -> denom));
	}
}

/*
 * Division of two rational functions.
 */

	static
	void
rat_div (

struct calib_rat_obj *		rop,		/* IN/OUT: result */
const struct calib_rat_obj *	op1,		/* IN: first operand */
const struct calib_rat_obj *	op2		/* IN: second operand */
)
{
const struct calib_rat_dom *	dom;
struct calib_rat_obj		TMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));

	dom -> init (dom, &TMP);

	dom -> inv (&TMP, op2);
	dom -> mul (rop, op1, &TMP);

	dom -> clear (&TMP);
}

/*
 * Reduce the given rational function object to lowest terms,
 * and force a canonical choice of signs between the
 * numerator and denominator.
 */

	static
	void
rat_canonicalize (

struct calib_rat_obj *	op	/* IN/OUT: rational function to canonicalize */
)
{
int				i, j, k, m;
int				tdeg, nterms, nvars, npoly, max_tdeg;
const int *			pow;
mpz_srcptr			cp;
const struct calib_Zxyz_dom *	ring;
struct calib_Zxyz_obj *		poly_ptrs [2];
struct calib_Zxyz_obj		GCD;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	ring = op -> numer.dom;
	CALIB_FATAL_ERROR_IF (ring NE op -> denom.dom);

	ring -> init (ring, &GCD);

	npoly = 2;
	poly_ptrs [0] = &(op -> numer);
	poly_ptrs [1] = &(op -> denom);

	/* Reduce any common factors. */
	ring -> gcd_n (&GCD,
		       &poly_ptrs [0],
		       npoly,
		       (const struct calib_Zxyz_obj **) &poly_ptrs [0]);

	ring -> clear (&GCD);

	nterms = op -> denom.nterms;
	CALIB_FATAL_ERROR_IF (nterms <= 0);

	/* In order to canonicalize signs between numerator and	*/
	/* denominator, scan the denominator to find the first	*/
	/* term of maximum total degree.  Force this term to	*/
	/* have positive sign.					*/

	nvars		= ring -> nvars;
	cp		= op -> denom.coeff;
	pow		= op -> denom.pow;
	max_tdeg	= -1;
	m		= -1;
	for (i = 0; i < nterms; i++, pow += nvars) {
		tdeg = 0;
		for (j = 0; j < nvars; j++) {
			k = pow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			tdeg += k;
		}
		if (tdeg > max_tdeg) {
			max_tdeg = tdeg;
			m	 = i;
		}
	}

	CALIB_FATAL_ERROR_IF (m < 0);

	k = mpz_sgn (&cp [m]);
	if (k <= 0) {
		CALIB_FATAL_ERROR_IF (k EQ 0);

		ring -> neg (&(op -> numer), &(op -> numer));
		ring -> neg (&(op -> denom), &(op -> denom));
	}
}

/*
 * Factor the given rational function.  Returns a list of calib_Zxyz_factor
 * objects for which the denominator's factors have *negative* multiplicies!
 */

	static
	struct calib_Zxyz_factor *
rat_factor (

const struct calib_rat_obj *	op	/* IN: Rational func to factor */
)
{
const struct calib_Zxyz_dom *	Zxyz;
struct calib_Zxyz_factor *	num_factors;
struct calib_Zxyz_factor *	den_factors;
struct calib_Zxyz_factor *	fp;
struct calib_Zxyz_factor **	hookp;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	Zxyz	= op -> dom -> ring;

	num_factors	= Zxyz -> factor (&(op -> numer));
	den_factors	= Zxyz -> factor (&(op -> denom));

	/* Combine the new factor lists, negating the multiplicies of	*/
	/* the denominator's factors.					*/
	hookp = &num_factors;
	for (;;) {
		fp = *hookp;
		if (fp EQ NULL) break;
		hookp = &(fp -> next);
	}

	*hookp = den_factors;

	for (fp = den_factors; fp NE NULL; fp = fp -> next) {
		fp -> multiplicity = - (fp -> multiplicity);
	}

	return (num_factors);
}

/*
 * Efficiently perform a set of variable substitutions of the form
 *
 *		var = rat
 *
 * to the given rational expression.  This handles the numerator and
 * denominator separately, with denominators kept in "PFRat"
 * (i.e., partially factored) form, which can keep the expressions
 * smaller in some cases, and allow some cancellation to take place
 * in this more compact form.
 */

	static
	void
rat_substitute_with_varmap (

struct calib_rat_obj *		rop,	/* IN: resulting rational expression */
const struct calib_rat_obj *	op,	/* IN: rational expression to apply */
					/*     substitutions to */
const int *			varmap,	/* IN: variable mapping array */
struct calib_subst_list *	slist	/* IN: list of substitutions */
)
{
int				i, nsubst, var, nvars, npoly, num_rpp;
int				npow, dpow;
int *				maxpow;
const struct calib_Zxyz_dom *	ring;
const struct calib_Zxyz_dom *	subring;
struct PFRat *			denrfp;
struct PFRat *			rfp;
struct PFRat **			subst_array;
struct subst_var_powers *	rpp_powers;
const struct calib_Zxyz_obj *	poly_ptrs [2];
struct calib_Zxyz_obj *		cf_ptrs [2];
struct calib_Zxyz_obj		GCD, RPP_FACTORS;
mpz_t				gcd, ncont, dcont;
struct RppPool			pool;
struct PFRat			num_rf, den_rf;

	CALIB_FATAL_ERROR_IF ((rop    EQ NULL) OR
			      (op     EQ NULL) OR
			      (varmap EQ NULL) OR
			      (slist  EQ NULL));

	ring	= op -> numer.dom;
	CALIB_FATAL_ERROR_IF (op -> denom.dom NE ring);

	subring	= rop -> numer.dom;
	CALIB_FATAL_ERROR_IF (rop -> denom.dom NE subring);

	nsubst = slist -> num_subst;

	/* Convert substitution expressions into PFRat form,		*/
	/* initializing the pool of relatively prime polynomials that	*/
	/* appear in the denominators.					*/
	denrfp = convert_subst_list_to_pfrat_array (&pool,
						     slist,
						     subring,
						     varmap);

	/* Make an array that references these by variable index. */
	nvars = op -> numer.dom -> nvars;
	subst_array = CALIB_NEWA (nvars, struct PFRat *);
	memset (subst_array, 0, nvars * sizeof (subst_array [0]));
	for (i = 0; i < nsubst; i++) {
		var = slist -> subst [i].var;
		CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));
		CALIB_FATAL_ERROR_IF (subst_array [var] NE NULL);
		subst_array [var] = &denrfp [i];
	}

	/* Perform substitutions separately into both the numerator and	*/
	/* denominator polynomials, obtaining each result in the form	*/
	/* of a reduced PFRat.						*/

	pfrat_init (&num_rf, subring, &pool);
	pfrat_init (&den_rf, subring, &pool);

	substitute_poly_to_pfrat (&num_rf,
				  &(op -> numer),
				  varmap,
				  subst_array,
				  &pool);

	substitute_poly_to_pfrat (&den_rf,
				  &(op -> denom),
				  varmap,
				  subst_array,
				  &pool);

	/* Cancel common integer factors between */
	/* num_rf.numer and den_rf.numer. */
	mpz_init (gcd);
	mpz_init (ncont);
	mpz_init (dcont);
	subring -> z_content (ncont, &num_rf.numer);
	subring -> z_content (dcont, &den_rf.numer);
	mpz_gcd (gcd, ncont, dcont);
	if (mpz_cmp_ui (gcd, 1) > 0) {
		subring -> div_z_exact (&num_rf.numer, &num_rf.numer, gcd);
		subring -> div_z_exact (&den_rf.numer, &den_rf.numer, gcd);
	}

	/* Cancel common integer factors between */
	/* num_rf.denom and den_rf.denom. */
	mpz_gcd (gcd, num_rf.denom.Konst, den_rf.denom.Konst);
	if (mpz_cmp_ui (gcd, 1) > 0) {
		mpz_divexact (num_rf.denom.Konst, num_rf.denom.Konst, gcd);
		mpz_divexact (den_rf.denom.Konst, den_rf.denom.Konst, gcd);
	}
	mpz_clear (dcont);
	mpz_clear (ncont);
	mpz_clear (gcd);

	/* Cancel RPP's in common between numerator and denominator.	*/
	/* Any that do cancel can be a huge savings.			*/
	for (i = 0; i < pool.num_rpp; i++) {
		npow = num_rf.denom.pow [i];
		dpow = den_rf.denom.pow [i];
		CALIB_FATAL_ERROR_IF ((npow < 0) OR (dpow < 0));
		if ((npow EQ 0) OR (dpow EQ 0)) continue;

		if (npow <= dpow) {
			num_rf.denom.pow [i] = 0;
			den_rf.denom.pow [i] -= npow;
		}
		else {
			num_rf.denom.pow [i] -= dpow;
			den_rf.denom.pow [i] = 0;
		}
	}

	/* Our final form is going to be RAT = num_rf / den_rf.		*/
	/* num_rf and den_rf are each PFRat objects, each supposedly	*/
	/* completely reduced.  The immediately previous code cancelled	*/
	/* integer factors, and the RPP's.  Let's cancel any common	*/
	/* factors between num_rf.numer and den_rf.numer *before* we	*/
	/* apply the RPP's.						*/

	subring -> init (subring, &GCD);

	npoly = 2;
	poly_ptrs [0]	= &num_rf.numer;
	poly_ptrs [1]	= &den_rf.numer;
	cf_ptrs [0]	= &num_rf.numer;
	cf_ptrs [1]	= &den_rf.numer;
	subring -> gcd_n (&GCD, &cf_ptrs [0], npoly, &poly_ptrs [0]);

	/* Apply den_rf.denom.Konst to num_rf.numer, and vice versa. */
	if (mpz_cmp_ui (den_rf.denom.Konst, 1) NE 0) {
		subring -> mul_z (&num_rf.numer,
				  &num_rf.numer,
				  den_rf.denom.Konst);
	}
	if (mpz_cmp_ui (num_rf.denom.Konst, 1) NE 0) {
		subring -> mul_z (&den_rf.numer,
				  &den_rf.numer,
				  num_rf.denom.Konst);
	}

	subring -> clear (&GCD);

	/* Compute a power array that is the "max" of num_rf.denom.pow	*/
	/* and den_rf.denom.pow.  This gives us the largest powers that	*/
	/* either of them require.					*/
	num_rpp = pool.num_rpp;
	maxpow = CALIB_NEWA (num_rpp, int);
	memset (maxpow, 0, num_rpp * sizeof (maxpow [0]));
	for (i = 0; i < num_rpp; i++) {
		npow = num_rf.denom.pow [i];
		dpow = den_rf.denom.pow [i];
		maxpow [i] = (npow >= dpow) ? npow : dpow;
	}

	/* Create a pre-computed table of RPP powers. */
	rpp_powers = compute_rpp_powers (&pool, maxpow);
	free (maxpow);

	/* Apply proper RPP terms to the numerator. */
	subring -> init (subring, &RPP_FACTORS);

	eval_rpp_term (&RPP_FACTORS, den_rf.denom.pow, rpp_powers, &pool);

	subring -> mul (&(rop -> numer), &num_rf.numer, &RPP_FACTORS);

	/* Apply proper RPP terms to the denominator. */
	eval_rpp_term (&RPP_FACTORS, num_rf.denom.pow, rpp_powers, &pool);

	subring -> mul (&(rop -> denom), &den_rf.numer, &RPP_FACTORS);

	subring -> clear (&RPP_FACTORS);
	free_svar_powers (rpp_powers);

	/* Hopefully, this should not cancel anything.			*/
	/* But it will shift the signs properly between num and den.	*/
	rat_canonicalize (rop);

	pfrat_clear (&den_rf);
	pfrat_clear (&num_rf);

	free (subst_array);
	for (i = 0; i < nsubst; i++) {
		rfp = &denrfp [i];
		ring -> clear (&(rfp -> numer));
		mpz_clear (rfp -> denom.Konst);
		free (rfp -> denom.pow);
	}
	free (denrfp);
	clear_rpp_pool (&pool);
}

/*
 * Convert the given list of substitution "value" expressions into an array
 * of PFRat objects, where the denominators appear in partially factored
 * form.  Initialize the given RPP Pool object to contain the single, common
 * copy of each RPP (relatively prime polynomial) that is referenced by
 * all of these partially factored denominators.
 */

	static
	struct PFRat *
convert_subst_list_to_pfrat_array (

struct RppPool *	pool,		/* IN/OUT: Pool to init */
struct calib_subst_list *
			slist,		/* IN: list of substitutions */
const struct calib_Zxyz_dom *
			subring,	/* IN: ring to map RPP's into */
const int *		varmap		/* IN: variable mapping array */
)
{
int				i, nsubst, num_rpp, sign;
int *				pow;
const struct calib_Zxyz_dom *	ring;
struct calib_rsubst *		rsp;
struct PFRat *			rf_array;
struct PFRat *			rfp;
struct PSplit **		ps_array;
struct PSplit *			psp;
struct calib_Zxyz_obj *		poly;
struct calib_Zxyz_obj *		rpp_array;
struct calib_Zxyz_obj		POLY;
struct PSplit			rpp_head, split_head;

	CALIB_FATAL_ERROR_IF ((pool EQ NULL) OR (slist EQ NULL));

	nsubst	= slist -> num_subst;

	if (nsubst <= 0) {
		pool -> num_rpp		= 0;
		pool -> dom		= NULL;
		pool -> rpp		= NULL;
		return (NULL);
	}

	rsp = slist -> subst;
	ring = rsp -> val.numer.dom;
	for (i = 0; i < nsubst; i++, rsp++) {
		CALIB_FATAL_ERROR_IF (rsp -> val.numer.dom NE ring);
		CALIB_FATAL_ERROR_IF (rsp -> val.denom.dom NE ring);
	}

	/* Construct an array of substitution values in RafFac form. */
	rf_array = CALIB_NEWA (nsubst, struct PFRat);
	memset (rf_array, 0, nsubst * sizeof (rf_array [0]));
	rfp = &rf_array [0];
	for (i = 0; i < nsubst; i++, rfp++) {
		/* Note: we can't use pfrat_init() here because the	*/
		/* pool does not contain any RPP's yet -- thus it does	*/
		/* not know how many elements long to make denom.pow[].	*/
		/* We leave these pow pointers NULL for now.		*/
		subring -> init (subring, &(rfp -> numer));
		mpz_init (rfp -> denom.Konst);
	}

	/* Copy numerators into PFRat array, converting to subring	*/
	/* using the given varmap.					*/
	rsp = slist -> subst;
	rfp = &rf_array [0];
	for (i = 0; i < nsubst; i++, rfp++, rsp++) {
		ring -> convert_with_varmap (&(rfp -> numer),
					     &(rsp -> val.numer),
					     varmap);
	}

	/* Initialize two doubly-linked lists of PSplit objects: one	*/
	/* for polynomials that are RPP's (all are "case 2" form), and	*/
	/* another for polynomials that consist of products of two or	*/
	/* more RPP's.							*/
	init_psplit_head (&rpp_head);
	init_psplit_head (&split_head);

	/* Create an array to hold a PSplit object for each denominator	*/
	/* that we process.						*/
	ps_array = CALIB_NEWA (nsubst, struct PSplit *);
	memset (ps_array, 0, nsubst * sizeof (ps_array [0]));

	subring -> init (subring, &POLY);

	/* Now process each substitution denominator, creating a PSplit	*/
	/* node for each, inserting each into our data structure which	*/
	/* may cause it to get split into smaller relatively prime	*/
	/* pieces.							*/
	rsp = slist -> subst;
	rfp = &rf_array [0];
	for (i = 0; i < nsubst; i++, rsp++, rfp++) {
		ring -> convert_with_varmap (&POLY,
					     &(rsp -> val.denom),
					     varmap);

		poly = subring -> dup (&POLY);

		/* Denominator should not be zero! */
		CALIB_FATAL_ERROR_IF (poly -> nterms <= 0);

		/* Remove integer content and store in the	*/
		/* RafFac's denominator FPoly object.		*/
		ring -> strip_z_content (rfp -> denom.Konst, poly);

		psp = CALIB_NEW (struct PSplit);
		memset (psp, 0, sizeof (*psp));
		psp -> index	= -1;
		psp -> sign	= 1;
		psp -> sfpow	= 1;
		psp -> poly	= poly;
		psp -> sfact	= NULL;
		psp -> nfact	= NULL;
		psp -> forw	= NULL;
		psp -> back	= NULL;

		ps_array [i] = psp;

		insert_psplit (&rpp_head, &split_head, psp);
	}

	subring -> clear (&POLY);

	/* Count how many RPP's there are, and assign their indices. */
	i = 0;
	for (psp = rpp_head.forw; psp NE &rpp_head; psp = psp -> forw) {
		psp -> index = i;
		++i;
	}
	num_rpp = i;

	/* Create the RPP array for the pool. */
	rpp_array = CALIB_NEWA (num_rpp, struct calib_Zxyz_obj);
	i = 0;
	for (psp = rpp_head.forw; psp NE &rpp_head; psp = psp -> forw) {
		subring -> init (subring, &rpp_array [i]);
		subring -> set (&rpp_array [i], psp -> poly);
		++i;
	}

	/* Create proper RPP power vectors for each denominator. */
	rfp = &rf_array [0];
	for (i = 0; i < nsubst; i++, rfp++) {
		pow = CALIB_NEWA (num_rpp, int);
		memset (pow, 0, num_rpp * sizeof (pow [0]));

		psp = ps_array [i];
		sign = traverse_psplit (pow, 1, psp);

		if (sign < 0) {
			mpz_neg (rfp -> denom.Konst, rfp -> denom.Konst);
		}
		rfp -> denom.pow = pow;
	}

	free (ps_array);
	clear_psplit_head (&split_head);
	clear_psplit_head (&rpp_head);

	pool -> num_rpp = num_rpp;
	pool -> dom	= subring;
	pool -> rpp	= rpp_array;

	return (rf_array);
}

/*
 * Clear out the given RPP Pool, freeing all memory it holds.
 */

	static
	void
clear_rpp_pool (

struct RppPool *	pool		/* IN/OUT: Pool to clear out */
)
{
int				i, n;
const struct calib_Zxyz_dom *	ring;
struct calib_Zxyz_obj *		rpp;

	CALIB_FATAL_ERROR_IF (pool EQ NULL);

	n	= pool -> num_rpp;
	ring	= pool -> dom;
	rpp	= pool -> rpp;

	if (rpp NE NULL) {
		for (i = 0; i < n; i++, rpp++) {
			ring -> clear (rpp);
		}
		free (pool -> rpp);
	}
	memset (pool, 0, sizeof (*pool));
}

/*
 * Initialize doubly-linked list head for PSplit objects.
 */

	static
	void
init_psplit_head (

struct PSplit *		psp	/* IN: PSplit object to init */
)
{
	CALIB_FATAL_ERROR_IF (psp EQ NULL);

	memset (psp, 0, sizeof (*psp));

	psp -> forw	= psp;
	psp -> back	= psp;
}

/*
 * Clear out the given doubly-linked list of PSplit nodes.
 */

	static
	void
clear_psplit_head (

struct PSplit *		psp	/* IN/OUT: PSplit object to clear */
)
{
struct PSplit *			p;
struct PSplit *			next;
struct calib_Zxyz_obj *		poly;
const struct calib_Zxyz_dom *	ring;

	for (;;) {
		p = psp -> forw;
		if (p EQ psp) break;

		next	= p -> forw;
		next -> back = psp;
		psp  -> forw = next;

		poly = p -> poly;
		if (poly NE NULL) {
			ring = poly -> dom;
			ring -> free (poly);
		}
		free (p);
	}

	memset (psp, 0, sizeof (*psp));
}

/*
 * Insert the given PSplit node into our data structure consisting of
 * two doubly-linked lists of existing PSplit nodes: those that are
 * still RPP's, and those that are non-trivial products and/or powers
 * of other PSplit nodes.
 *
 * The basic algorithm is to GCD this polynomial with each of the
 * current RPP's.  Any non-trivial GCD causes one or both of these
 * PSplit nodes to split into additional nodes that are recursively
 * inserted.
 */

	static
	void
insert_psplit (

struct PSplit *		rpp_head,	/* IN: Head of RPP list */
struct PSplit *		split_head,	/* IN: Head of split list */
struct PSplit *		p		/* IN: PSplit node to insert */
)
{
int				npoly, sign, pow;
struct PSplit *			p2;
struct PSplit *			p3;
struct PSplit *			p4;
struct PSplit *			gsp;
struct PSplit *			next;
struct PSplit *			prev;
struct PSplit *			p2_next;
const struct calib_Zxyz_dom *	ring;
struct calib_Zxyz_obj *		poly1;
struct calib_Zxyz_obj *		poly2;
const struct calib_Zxyz_obj *	poly_ptrs [2];
struct calib_Zxyz_obj *		cf_ptrs [2];
struct calib_Zxyz_obj		GCD, CF1, CF2;

	CALIB_FATAL_ERROR_IF ((rpp_head EQ NULL) OR
			      (split_head EQ NULL) OR
			      (p EQ NULL));

	poly1	= p -> poly;
	ring	= poly1 -> dom;

	ring -> init (ring, &GCD);
	ring -> init (ring, &CF1);
	ring -> init (ring, &CF2);

	npoly = 2;
	poly_ptrs [0] = poly1;

	cf_ptrs [0] = &CF1;
	cf_ptrs [1] = &CF2;

	for (p2 = rpp_head -> forw; ; p2 = p2_next) {
		if (p2 EQ rpp_head) {
			/* p2 is relatively prime to all existing	*/
			/* entries in the RPP list.			*/
			if (is_constant_poly (p -> poly)) {
				/* Just a constant, which should be +1	*/
				/* or -1.  Convert to case 3 (by	*/
				/* threading it into the split_head	*/
				/* list), set proper sign, and make	*/
				/* sfact = nfact = NULL, which means 1.	*/
				sign = get_sign (p2 -> poly);
				p -> sign	= sign;
				p -> sfpow	= 1;
				p -> sfact	= NULL;
				p -> nfact	= NULL;

				prev = split_head;
				next = split_head -> forw;
				p -> back = prev;
				p -> forw = next;
				prev -> forw = p;
				next -> back = p;
			}
			else {
				/* A new, non-trivial RPP! */
				p -> sign	= 1;
				p -> sfpow	= 1;
				p -> sfact	= NULL;
				p -> nfact	= NULL;

				prev = rpp_head;
				next = rpp_head -> forw;
				p -> back = prev;
				p -> forw = next;
				prev -> forw = p;
				next -> back = p;
			}
			break;
		}

		/* Get next entry now, in case p2 gets removed	*/
		/* from the rpp_head list.			*/
		p2_next = p2 -> forw;

		poly2 = p2 -> poly;

		poly_ptrs [1] = poly2;
		ring -> gcd_n (&GCD, &cf_ptrs [0], npoly, &poly_ptrs [0]);

		if (is_constant_poly (&GCD)) {
			/* No common divisors.  Skip. */
			continue;
		}

		/* The GCD is non-trivial.  See how many times the GCD	*/
		/* divides poly2.					*/
		pow = 1 + ring -> div_remove (&CF2, &CF2, &GCD);

		/* poly2 = CF2 * GCD**pow, where:		*/
		/* 1. GCD is known to be non-trivial,		*/
		/* 2. GCD and CF2 are relatively prime to each	*/
		/*    other and all other members of the RPP	*/
		/*    list.					*/
		/* Make new PSplit nodes p3 for CF2 and p4 for	*/
		/* GCD, and insert at *front* of rpp_head (to	*/
		/* avoid GCD'ing them with p again).		*/
		sign = 1;
		if (NOT is_constant_poly (&CF2)) {
			/* Unthread p2 from rpp_head list. */
			prev = p2 -> back;
			next = p2 -> forw;
			prev -> forw = next;
			next -> back = prev;
			p2 -> forw = NULL;
			p2 -> back = NULL;

			/* Make new RPP p3 for CF2. */
			p3 = CALIB_NEW (struct PSplit);
			memset (p3, 0, sizeof (*p3));
			p3 -> index	= -1;
			p3 -> sign	= 1;
			p3 -> sfpow	= 1;
			p3 -> poly	= ring -> dup (&CF2);
			p3 -> sfact	= NULL;
			p3 -> nfact	= NULL;

			/* Thread p3 onto rpp_head list. */
			prev = rpp_head;
			next = rpp_head -> forw;
			p3 -> back = prev;
			p3 -> forw = next;
			prev -> forw = p3;
			next -> back = p3;

			/* Make new RPP p4 for GCD. */
			p4 = CALIB_NEW (struct PSplit);
			memset (p4, 0, sizeof (*p4));
			p4 -> index	= -1;
			p4 -> sign	= 1;
			p4 -> sfpow	= 1;
			p4 -> poly	= ring -> dup (&GCD);
			p4 -> sfact	= NULL;
			p4 -> nfact	= NULL;

			/* Thread p4 onto rpp_head list. */
			prev = rpp_head;
			next = rpp_head -> forw;
			p4 -> forw	= next;
			p4 -> back	= prev;
			next -> back	= p4;
			prev -> forw	= p4;

			/* Convert p2: poly2 = GCD**pow*CF2. */
			p2 -> sfpow	= pow;
			p2 -> sfact	= p4;
			p2 -> nfact	= p3;

			/* Thread p2 onto split_head list. */
			prev = split_head;
			next = split_head -> forw;
			p2 -> back = prev;
			p2 -> forw = next;
			prev -> forw = p2;
			next -> back = p2;

			sign = 1;
			gsp = p4;
		}
		else if (pow EQ 1) {
			/* poly2 = GCD * CF2, where CF2	is +1	*/
			/* or -1.  Furthermore, poly2 is	*/
			/* already an RPP, so			*/
			/* GCD = sign * poly2.			*/
			sign = get_sign (&CF2);
			gsp = p2;
		}
		else {
			/* poly2 = GCD**pow * CF2, where CF2	*/
			/* is +1 or -1.				*/

			/* Unthread p2 from rpp_head list. */
			prev = p2 -> back;
			next = p2 -> forw;
			prev -> forw = next;
			next -> back = prev;
			p2 -> forw = NULL;
			p2 -> back = NULL;

			/* Make new RPP p4 for GCD. */
			p4 = CALIB_NEW (struct PSplit);
			memset (p4, 0, sizeof (*p4));
			p4 -> index	= -1;
			p4 -> sign	= sign;
			p4 -> sfpow	= 1;
			p4 -> poly	= ring -> dup (&GCD);
			p4 -> sfact	= NULL;
			p4 -> nfact	= NULL;

			/* Thread p4 onto rpp_head list. */
			prev = rpp_head;
			next = rpp_head -> forw;
			p4 -> forw	= next;
			p4 -> back	= prev;
			next -> back	= p4;
			prev -> forw	= p4;

			/* Convert p2 = sign * GCD**pow. */
			p2 -> sign	= get_sign (&CF2);
			p2 -> sfpow	= pow;
			p2 -> sfact	= p4;
			p2 -> nfact	= NULL;

			/* Thread p2 onto split_head list. */
			prev = split_head;
			next = split_head -> forw;
			p2 -> back = prev;
			p2 -> forw = next;
			prev -> forw = p2;
			next -> back = p2;

			sign = 1;
			gsp = p4;
		}

		/* Find out how many powers of GCD can be	*/
		/* divided out from polynomial p.		*/
		pow = 1 + ring -> div_remove (&CF1, &CF1, &GCD);

		/* p = CF1 * GCD**pow, where GCD is represented	*/
		/* as sign*gsp.  Make PSplit node p3 for CF1.	*/
		if ((pow & 0x01) EQ 0) {
			/* Even power.  Force positive sign. */
			sign = 1;
		}
		if (is_constant_poly (&CF1)) {
			sign *= get_sign (&CF1);
			p3 = NULL;
		}
		else {
			p3 = CALIB_NEW (struct PSplit);
			memset (p3, 0, sizeof (*p3));
			p3 -> index	= -1;
			p3 -> sign	= 1;
			p3 -> sfpow	= 1;
			p3 -> poly	= ring -> dup (&CF1);
			p3 -> sfact	= NULL;
			p3 -> nfact	= NULL;
			p3 -> forw	= NULL;
			p3 -> back	= NULL;

			insert_psplit (rpp_head, split_head, p3);
		}

		/* Convert p to case 3. */
		p -> sfpow	= pow;
		p -> sign	= sign;
		p -> sfact	= gsp;
		p -> nfact	= p3;

		/* Thread p into split_head list. */
		prev = split_head;
		next = split_head -> forw;
		p -> forw = next;
		p -> back = prev;
		prev -> forw = p;
		next -> back = p;

		/* We are done inserting p! */
		break;
	}

	ring -> clear (&CF2);
	ring -> clear (&CF1);
	ring -> clear (&GCD);
}

/*
 * We are given the following:
 *
 *	1. An array "pow" of integer powers, one per RPP in the pool.
 *	2. An integer power K.
 *	3. A PSplit node representing some polynomial P.
 *
 * Let P = sign * RPP1**a1 * RPP2**a2 * ... * RPPn**an.
 *
 * Our goal is to add K*(a1, a2, ..., an) to the array "pow",
 * and return the "sign".
 *
 * In other words, multiply the given power buffer by POLY**K
 * and return the corrective sign.
 */

	static
	int
traverse_psplit (

int *		pow,		/* IN/OUT: power buffer to multiply into */
int		K,		/* IN: the power of POLY to apply */
struct PSplit *	p		/* IN: PSplit not to interpret as RPP powers */
)
{
int		idx, sign, sign2;

	if (p EQ NULL) {
		/* NULL pointer represent polynomial "1" (case 1).	*/
		/* No modification to power buffer is necessary, nor is	*/
		/* there any sign correction.				*/
		return (1);
	}

	idx = p -> index;
	if (idx >= 0) {
		/* This is case 2, a single RPP. */
		CALIB_FATAL_ERROR_IF (p -> sign NE 1);
		CALIB_FATAL_ERROR_IF (p -> sfpow NE 1);
		CALIB_FATAL_ERROR_IF (p -> sfact NE NULL);
		CALIB_FATAL_ERROR_IF (p -> nfact NE NULL);
		pow [idx] += K;
		return (1);
	}

	/* Case 3:						*/
	/*							*/
	/*	poly = sign * sfact**pow * nfact,		*/
	/*							*/
	/* so							*/
	/*							*/
	/*	poly**K = (sign * sfact**pow * nfact)**K,	*/
	/*		= sign**K * sfact**(K*pow) * nfact**K	*/

	sign = traverse_psplit (pow, K * p -> sfpow, p -> sfact);
	if ((p -> sfpow & 0x01) EQ 0) {
		/* sfpow is even, force positive sign. */
		sign = 1;
	}
	sign2 = traverse_psplit (pow, K, p -> nfact);

	sign *= sign2;

	if ((K & 0x01) EQ 0) {
		/* Power K is even.  Force positive sign. */
		sign = 1;
	}

	return (sign);
}

/*
 * Apply a set of substitutions of the form:
 *
 *		var = value
 *
 * to a given polynomial.  Each "value" is a rational expression, so
 * the result is a rational expression.  For efficiency, each of these
 * rational expressions is in RafFac form with respect to the given
 * pool of RPP's.
 *
 * The result is a fully reduced PFRat object.
 */

	static
	void
substitute_poly_to_pfrat (

struct PFRat *		result,	/* IN/OUT: reduced result */
const struct calib_Zxyz_obj *
			poly,	/* IN: polynomial to substitute into */
const int *		varmap,	/* IN: variable mapping array */
struct PFRat **		subst_array,
				/* IN: PFRat value to substitute for */
				/*     each var (if NULL, leave var as is) */
struct RppPool *	pool	/* IN: pool of RPP's being used */
)
{
int				i, j, k, nvars, nterms, num_rpp, npoly;
int *				pow;
int *				lcmpow;
int *				deficit_pow;
const struct calib_Zxyz_dom *	ring;
const struct calib_Zxyz_dom *	subring;
struct subst_var_powers *	svar_powers;
struct subst_var_powers *	rpp_powers;
struct ivec_hash *		hash;
struct calib_Zxyz_obj *		cpoly_array;
const struct calib_Zxyz_obj **	pptrs;
struct ivec_hash_entry **	hookp;
struct ivec_hash_entry **	bucket;
struct ivec_hash_entry *	ep;
struct hashent *		xp;
struct PFRat *			rf_array;
struct PFRat *			rfp;
struct ivec_hash_entry **	entries;
struct calib_Zxyz_obj *		rpp;
const struct calib_Zxyz_obj *	poly_ptrs [2];
struct calib_Zxyz_obj *		cf_ptrs [2];
mpz_t				zgcd, zlcm, zfactor;
struct calib_Zxyz_obj		SPOLY, GCD, CF1, CF2, RPP_TERM;
struct PFRat			TERM;

	CALIB_FATAL_ERROR_IF ((result	   EQ NULL) OR
			      (poly	   EQ NULL) OR
			      (varmap	   EQ NULL) OR
			      (subst_array EQ NULL) OR
			      (pool	   EQ NULL));

	ring	= poly -> dom;
	subring	= result -> numer.dom;
	num_rpp	= pool -> num_rpp;

	/* Convert the given polynomial into two parts:			*/
	/*								*/
	/*	1. A main polynomial (SPOLY) containing only variables	*/
	/*	   that are being replaced with their corresponding	*/
	/*	   values, and						*/
	/*								*/
	/*	2. The coefficients of SPOLY, which are polynomials in	*/
	/*	   all of the remaining variables.			*/
	/*								*/
	/* Note that our normal data structures allow only integer	*/
	/* coefficients.  We do this by making all of the coefficients	*/
	/* of SPOLY be simply "dummy" values of 1, and instead provide	*/
	/* a separate "cpoly_array" having one element per term of	*/
	/* SPOLY that contains the corresponding coefficient		*/
	/* polynomials.  Putting the polynomial in the form means that	*/
	/* we only have to evaluate each distinct multiplicative	*/
	/* combination of the substituted variables only once.  It also	*/
	/* gives us a chance to do some intermediate simplification.	*/

	ring -> init (ring, &SPOLY);
	cpoly_array = ring -> map_to_subring (&SPOLY, poly, subring, varmap);

	/* Pre-compute the necessary powers of (the numerator of) each	*/
	/* substituted value.						*/
	svar_powers = compute_svar_powers (poly, subst_array, subring, varmap);

	hash = _calib_ivec_hash_create (num_rpp,
					hash_entry_op,
					(void *) subring);

	pfrat_init (&TERM, subring, pool);
	mpz_init (zgcd);
	mpz_init (zlcm);
	mpz_init (zfactor);

	nterms	= SPOLY.nterms;
	nvars	= SPOLY.dom -> nvars;
	pow	= SPOLY.pow;
	for (i = 0; i < nterms; i++, pow += nvars) {
		/* Multiply out the substitution value powers for this	*/
		/* single term.  Result is a PFRat that includes	*/
		/* denominator in FPoly form (konstant, array of RPP	*/
		/* powers).						*/
		eval_svar_term (&TERM, pow, svar_powers, subst_array, pool);

		/* Now multiply in the coefficient polynomial for this	*/
		/* term, kept in separate array.			*/
		subring -> mul (&TERM.numer, &TERM.numer, &cpoly_array [i]);

		/* Find/make the hash table entry corresponding to this	*/
		/* combination of denominator RPP powers.		*/
		hookp = _calib_ivec_hash_make (hash, TERM.denom.pow, &bucket);
		CALIB_FATAL_ERROR_IF (hookp EQ NULL);
		ep = *hookp;
		CALIB_FATAL_ERROR_IF (ep EQ NULL);
		xp = (struct hashent *) (ep -> data);
		CALIB_FATAL_ERROR_IF (xp EQ NULL);

		/* We have TERM = poly1 / (Konst1 * RPP_powers), and	*/
		/* xp = poly2 / (Konst2 * RPP_powers), where the hash	*/
		/* table has assured that the RPP_powers are the same	*/
		/* for both TERM and xp.  We need to add TERM into xp,	*/
		/* but we cannot do so unless Konst1 EQ Konst2.		*/
		mpz_gcd (zgcd, TERM.denom.Konst, xp -> den_konst);
		mpz_divexact (zlcm, xp -> den_konst, zgcd);
		mpz_mul (zlcm, zlcm, TERM.denom.Konst);

		mpz_divexact (zfactor, zlcm, xp -> den_konst);
		if (mpz_cmp_ui (zfactor, 1) NE 0) {
			subring -> mul_z (&(xp -> numer),
					  &(xp -> numer),
					  zfactor);
			mpz_set (xp -> den_konst, zlcm);
		}
		mpz_divexact (zfactor, zlcm, TERM.denom.Konst);
		if (mpz_cmp_ui (zfactor, 1) NE 0) {
			subring -> mul_z (&TERM.numer, &TERM.numer, zfactor);
			mpz_set (TERM.denom.Konst, zlcm);
		}

		subring -> add (&(xp -> numer), &(xp -> numer), &TERM.numer);

		/* Check for common constant factors in xp. */
		subring -> z_content (zfactor, &(xp -> numer));
		mpz_gcd (zgcd, zfactor, xp -> den_konst);
		if (mpz_cmp_ui (zgcd, 1) > 0) {
			subring -> div_z_exact (&(xp -> numer),
						&(xp -> numer),
						zgcd);
			mpz_divexact (xp -> den_konst,
				      xp -> den_konst,
				      zgcd);
		}

		if (xp -> numer.nterms <= 0) {
			/* Result became zero!  Delete this hash table	*/
			/* entry.					*/
			_calib_ivec_hash_remove (hash, hookp, bucket);
		}
	}

	/* We don't need theses any more. */
	for (i = 0; i < SPOLY.nterms; i++) {
		ring -> clear (&cpoly_array [i]);
	}
	free (cpoly_array);
	free_svar_powers (svar_powers);

	/* Gather hash table entries into an array of PFRat objects. */
	nterms = hash -> nentries;
	entries = _calib_ivec_hash_sort_entries (hash);

	rf_array = NULL;
	if (nterms > 0) {
		rf_array = CALIB_NEWA (nterms, struct PFRat);
		memset (rf_array, 0, nterms * sizeof (rf_array [0]));
	}

	rfp = &rf_array [0];
	for (i = 0; i < nterms; i++, rfp++) {
		pfrat_init (rfp, subring, pool);
		ep = entries [i];
		xp = (struct hashent *) (ep -> data);
		subring -> set (&(rfp -> numer), &(xp -> numer));
		mpz_set (rfp -> denom.Konst, xp -> den_konst);
		memcpy (rfp -> denom.pow, ep -> vec, num_rpp * sizeof (int));
	}

	/* Don't need the hash table any more. */
	free (entries);
	_calib_ivec_hash_free (hash);

	/* The numerator z_contents, and denominator Konst factors are	*/
	/* kept in lowest terms by the above code.  However, now that	*/
	/* we are done with each term, it can be profitable to see if	*/
	/* any of the denominator RPP's cancel with any of the		*/
	/* numerator polynomials.  This is only really a win if it	*/
	/* reduces the maximum power of any RPP[j] across all terms --	*/
	/* otherwise we're just going to be multiplying these factors	*/
	/* back into the numerators below when putting all terms over	*/
	/* the least common denominator.  We go ahead and trust that	*/
	/* this will be occasionally worthwhile.			*/
	subring -> init (subring, &GCD);
	subring -> init (subring, &CF1);
	subring -> init (subring, &CF2);
	rfp = &rf_array [0];
	for (i = 0; i < nterms; i++, rfp++) {
		for (j = 0; j < num_rpp; j++) {
			k = rfp -> denom.pow [j];
			if (k <= 0) continue;

			rpp = &(pool -> rpp [j]);
			do {
				poly_ptrs [0] = &(rfp -> numer);
				poly_ptrs [1] = rpp;
				cf_ptrs [0] = &CF1;
				cf_ptrs [1] = &CF2;
				npoly = 2;
				subring -> gcd_n (&GCD,
						  &cf_ptrs [0],
						  npoly,
						  &poly_ptrs [0]);
				if (is_constant_poly (&GCD)) {
					/* This RPP does not cancel. */
					break;
				}
				if (NOT is_constant_poly (&CF2)) {
					/* Only a portion of this RPP	*/
					/* canceled.  We ignore this,	*/
					/* which is a downside of not	*/
					/* allowing new RPP's to be	*/
					/* added to the mix		*/
					/* dynamically.			*/
					continue;
				}
				/* CF2 should be just +1 or -1.	*/
				/* Fatal if not so.		*/
				get_sign (&CF2);

				subring -> set (&(rfp -> numer), &CF1);
				--k;
				rfp -> denom.pow [j] = k;
			} while (k > 0);
		}
	}

	/* Find LCM of all the denominators: both the Konst part, and	*/
	/* the RPP powers part.						*/
	mpz_set_ui (zlcm, 1);
	lcmpow = CALIB_NEWA (num_rpp, int);
	memset (lcmpow, 0, num_rpp * sizeof (lcmpow [0]));
	rfp = &rf_array [0];
	for (i = 0; i < nterms; i++, rfp++) {
		mpz_gcd (zgcd, zlcm, rfp -> denom.Konst);
		mpz_divexact (zlcm, zlcm, zgcd);
		mpz_mul (zlcm, zlcm, rfp -> denom.Konst);

		for (j = 0; j < num_rpp; j++) {
			k = rfp -> denom.pow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			if (k > lcmpow [j]) {
				lcmpow [j] = k;
			}
		}
	}

	/* Compute the largest "deficit" of each RPP across all terms.	*/
	deficit_pow = CALIB_NEWA (num_rpp, int);
	memset (deficit_pow, 0, num_rpp * sizeof (deficit_pow [0]));
	rfp = &rf_array [0];
	for (i = 0; i < nterms; i++, rfp++) {
		for (j = 0; j < num_rpp; j++) {
			k = lcmpow [j] - rfp -> denom.pow [j];
			if (k > deficit_pow [j]) {
				deficit_pow [j] = k;
			}
		}
	}

	/* Compute a power table of the RPP's that we'll need. */
	rpp_powers = compute_rpp_powers (pool, deficit_pow);

	subring -> init (subring, &RPP_TERM);

	/* Place each term into the common denominator by multiplying	*/
	/* each numerator by the proper Konst factor and the proper	*/
	/* powers of each RPP in which it is deficient.			*/
	rfp = &rf_array [0];
	for (i = 0; i < nterms; i++, rfp++) {
		mpz_divexact (zfactor, zlcm, rfp -> denom.Konst);
		mpz_set (rfp -> denom.Konst, zlcm);
		if (mpz_cmp_ui (zfactor, 1) NE 0) {
			subring -> mul_z (&(rfp -> numer),
					  &(rfp -> numer),
					  zfactor);
		}

		rpp = pool -> rpp;
		for (j = 0; j < num_rpp; j++) {
			k = lcmpow [j] - rfp -> denom.pow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			deficit_pow [j] = k;

			rfp -> denom.pow [j] = lcmpow [j];
		}
		eval_rpp_term (&RPP_TERM, deficit_pow, rpp_powers, pool);

		subring -> mul (&(rfp -> numer),
				&(rfp -> numer),
				&RPP_TERM);
	}

	/* Add up all of the terms. */
	pptrs = CALIB_NEWA (nterms, const struct calib_Zxyz_obj *);
	rfp = &rf_array [0];
	for (i = 0; i < nterms; i++, rfp++) {
		pptrs [i] = &(rfp -> numer);
	}
	mpz_set (TERM.denom.Konst, zlcm);
	memcpy (TERM.denom.pow, lcmpow, num_rpp * sizeof (int));
	subring -> add_n (&TERM.numer, nterms, &pptrs [0]);
	free (pptrs);

	rfp = &rf_array [0];
	for (i = 0; i < nterms; i++, rfp++) {
		pfrat_clear (rfp);
	}
	free (rf_array);

	/* Divide out any common integer factor. */
	subring -> z_content (zfactor, &TERM.numer);
	mpz_gcd (zgcd, zfactor, TERM.denom.Konst);
	if (mpz_cmp_ui (zgcd, 1) > 0) {
		subring -> div_z_exact (&TERM.numer, &TERM.numer, zgcd);
		mpz_divexact (TERM.denom.Konst, TERM.denom.Konst, zgcd);
	}

	/* Divide out any denominator RPP's that cancel. */
	for (i = 0; i < num_rpp; i++) {
		k = TERM.denom.pow [i];
		if (k <= 0) continue;
		rpp = &(pool -> rpp [i]);
		do {
			poly_ptrs [0]	= &TERM.numer;
			poly_ptrs [1]	= rpp;
			cf_ptrs [0]	= &CF1;
			cf_ptrs [1]	= &CF2;
			npoly = 2;
			subring -> gcd_n (&GCD,
					  &cf_ptrs [0],
					  npoly,
					  &poly_ptrs [0]);
			if (is_constant_poly (&GCD)) {
				/* This RPP does not cancel. */
				break;
			}
			if (NOT is_constant_poly (&CF2)) {
				/* Only a portion of this RPP	*/
				/* canceled.  We ignore this,	*/
				/* which is a downside of not	*/
				/* allowing new RPP's to be	*/
				/* added to the mix		*/
				/* dynamically.			*/
				break;
			}
			/* CF2 should be just +1 or -1.	*/
			/* Fatal if not so.		*/
			get_sign (&CF2);

			subring -> set (&TERM.numer, &CF1);
			--k;
			TERM.denom.pow [i] = k;
		} while (k > 0);
	}

	/* Copy result out to caller. */
	subring -> set (&(result -> numer), &TERM.numer);
	mpz_set (result -> denom.Konst, TERM.denom.Konst);
	memcpy (result -> denom.pow, TERM.denom.pow, num_rpp * sizeof (int));

	subring -> clear (&RPP_TERM);
	free_svar_powers (rpp_powers);
	free (deficit_pow);
	free (lcmpow);
	subring -> clear (&CF2);
	subring -> clear (&CF1);
	subring -> clear (&GCD);
	mpz_clear (zfactor);
	mpz_clear (zlcm);
	mpz_clear (zgcd);
	pfrat_clear (&TERM);
	ring -> clear (&SPOLY);
}

/*
 * Create a pre-computed table of powers of (the numerator of) each
 * substituted value.
 *
 * Analysis of possible alternative implementations:
 *
 * - Doing it on the fly for each term can repeat each exponentiation
 *   unnecessarily.
 *
 * - A "cache them as they are accessed" approach is more complicated,
 *   and does not actually save any space because for each substituted
 *   variable j, at the very least we need:
 *
 *		poly**K
 *
 *   where poly = subst_array [j] -> numer and K = maxdeg [j].
 *
 * So we might as well just compute them all once up front.
 */

	static
	struct subst_var_powers *
compute_svar_powers (

const struct calib_Zxyz_obj *	poly,	/* IN: polynomial, some of whose */
					/*     vars are being substituted */
struct PFRat **		subst_array,	/* IN: map of substitutions */
const struct calib_Zxyz_dom *
			subring,	/* IN: ring of substitution values */
const int *		varmap		/* IN: variable mapping array */
)
{
int				i, j, k, n, nvars, nterms;
int				nsubst;
int *				pow;
int *				maxdeg;
const struct calib_Zxyz_dom *	ring;
struct calib_Zxyz_obj *		array;
struct calib_Zxyz_obj *		npoly;
struct poly_powers *		ptabs;
struct poly_powers *		pp;
struct poly_powers **		map;
struct PFRat *			rfp;
struct subst_var_powers *	svp;
struct calib_Zxyz_obj		NPOLY;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (subst_array EQ NULL));

	ring	= poly -> dom;
	nvars	= ring -> nvars;

	/* Compute the maximum degree of each variable in SPOLY, and	*/
	/* verify that the only variables it contains are those that	*/
	/* have substitutions.						*/
	maxdeg = CALIB_NEWA (nvars, int);
	memset (maxdeg, 0, nvars * sizeof (maxdeg [0]));

	nterms	= poly -> nterms;
	pow	= poly -> pow;
	for (i = 0; i < nterms; i++, pow += nvars) {
		for (j = 0; j < nvars; j++) {
			k = pow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			rfp = subst_array [j];
			if (rfp EQ NULL) {
				/* Skip vars having no substitution. */
				continue;
			}
			if (k > maxdeg [j]) {
				maxdeg [j] = k;
			}
		}
	}

	/* Count the number of substitions that have maxdeg > 1. */
	nsubst = 0;
	for (j = 0; j < nvars; j++) {
		rfp = subst_array [j];
		if ((rfp NE NULL) AND (maxdeg [j] > 0)) {
			++nsubst;
		}
	}

	/* Create an array to map variables to their associated */
	/* power table (or NULL).				*/
	map = CALIB_NEWA (nvars, struct poly_powers *);
	memset (map, 0, nvars * sizeof (map [0]));

	subring -> init (subring, &NPOLY);

	/* Create a power table for each substitution. */
	ptabs = NULL;
	if (nsubst > 0) {
		ptabs = CALIB_NEWA (nsubst, struct poly_powers);
		memset (ptabs, 0, sizeof (ptabs [0]));
	}
	pp = ptabs;
	for (j = 0; j < nvars; j++) {
		rfp = subst_array [j];
		n   = maxdeg [j];
		CALIB_FATAL_ERROR_IF (n < 0);
		if ((rfp EQ NULL) OR (n EQ 0)) {
			continue;
		}

		array = CALIB_NEWA (n, struct calib_Zxyz_obj);
		for (i = 0; i < n; i++) {
			subring -> init (subring, &array [i]);
		}

		/* Get numerator polynomial to compute powers of.	*/
		/* We need to map from ring to subring.			*/
		/*							*/
		/* NOTE: Do this even if (ring EQ subring) because the	*/
		/* varmap might be re-ordering the variables on us!	*/

		npoly	= &(rfp -> numer);
		ring -> convert_with_varmap (&NPOLY, npoly, varmap);

		/* Compute and store NPOLY**k, for k = 1 ... n. */
		subring -> set (&array [0], &NPOLY);
		for (i = 1; i < n; i++) {
			subring -> mul (&array [i], &array [i-1], &NPOLY);
		}

		pp -> max_power = n;
		pp -> tab	= array;

		map [j] = pp;

		++pp;
	}

	subring -> clear (&NPOLY);
	free (maxdeg);

	svp = CALIB_NEW (struct subst_var_powers);
	memset (svp, 0, sizeof (*svp));
	svp -> nvars	= nvars;
	svp -> ntabs	= nsubst;
	svp -> powers	= ptabs;
	svp -> map	= map;

	return (svp);
}

/*
 * Free up the given subst_var_powers object.
 */

	static
	void
free_svar_powers (

struct subst_var_powers *	svp	/* IN: powers object to free */
)
{
int				i, j, ntabs, n;
struct poly_powers *		pp;
struct calib_Zxyz_obj *		array;
const struct calib_Zxyz_dom *	ring;

	if (svp EQ NULL) return;

	if (svp -> map NE NULL) {
		free (svp -> map);
	}

	ntabs	= svp -> ntabs;
	pp	= svp -> powers;
	if (pp NE NULL) {
		for (i = 0; i < ntabs; i++, pp++) {
			n	= pp -> max_power;
			array	= pp -> tab;
			if (array EQ NULL) continue;
			for (j = 0; j < n; j++) {
				ring = array [j].dom;
				CALIB_FATAL_ERROR_IF (ring EQ NULL);
				ring -> clear (&array [j]);
			}
			free (array);
		}
		free (svp -> powers);
	}
	memset (svp, 0, sizeof (*svp));

	free (svp);
}

/*
 * We are given a vector of integer variable powers from a single
 * term of our SPOLY, which should contain only variables for which
 * there is a substitution var_j = value_j = numer_j / (Konst_j * RPP's).
 * Suppose var_j has power aj.  We compute the product, of var_j**aj
 * over all j, using the pre-computed powers of the numer_j.
 * Note that we need to also compute the denominator of this product,
 * both the Konst factor, and the resulting RPP powers.
 *
 * The result is obtained in PFRat form.
 */

	static
	void
eval_svar_term (

struct PFRat *		result,		/* IN/OUT: result in PFRat form */
const int *		pow,		/* IN: array of var powers */
struct subst_var_powers *
			svar_powers,	/* IN: pre-computed table of */
					/*     var numerator powers */
struct PFRat **		subst_array,	/* IN: PFRat to substitute for */
					/*     each var (if NULL, leave */
					/*     var as is) */
struct RppPool *	pool		/* IN: pool of RPP's being used */
)
{
int				i, j, k, nvars, num_rpp;
int *				tpow;
const struct calib_Zxyz_dom *	ring;
struct PFRat *			rfp;
struct poly_powers *		pp;
mpz_t				ztmp;
struct PFRat			TERM;

	CALIB_FATAL_ERROR_IF ((result	   EQ NULL) OR
			      (pow	   EQ NULL) OR
			      (svar_powers EQ NULL) OR
			      (subst_array EQ NULL) OR
			      (pool	   EQ NULL));

	ring	= result -> numer.dom;
	nvars	= svar_powers -> nvars;
	num_rpp	= pool -> num_rpp;

	mpz_init (ztmp);

	/* Initialize term to 0 / 1, then set it to 1 / 1. */
	pfrat_init (&TERM, ring, pool);
	ring -> set_si (&TERM.numer, 1);
	tpow = TERM.denom.pow;

	for (i = 0; i < nvars; i++) {
		k = pow [i];
		CALIB_FATAL_ERROR_IF (k < 0);
		if (k EQ 0) continue;

		/* Term should contain *only* vars that have a substitution. */
		rfp = subst_array [i];
		CALIB_FATAL_ERROR_IF (rfp EQ NULL);

		pp = svar_powers -> map [i];
		CALIB_FATAL_ERROR_IF (pp EQ NULL);

		/* Multiply var_value.numerator**pow into numerator term. */
		ring -> mul (&TERM.numer, &TERM.numer, &(pp -> tab [k - 1]));

		/* Multiply the proper power of the denominator Konst. */
		mpz_pow_ui (ztmp, rfp -> denom.Konst, k);
		mpz_mul (TERM.denom.Konst, TERM.denom.Konst, ztmp);

		/* Update the denominator RPP's. */
		for (j = 0; j < num_rpp; j++) {
			tpow [j] += k * rfp -> denom.pow [j];
		}
	}

	/* Copy resulting TERM out to caller. */
	ring -> set (&(result -> numer), &TERM.numer);
	mpz_set (result -> denom.Konst, TERM.denom.Konst);
	memcpy (result -> denom.pow, TERM.denom.pow, num_rpp * sizeof (int));

	pfrat_clear (&TERM);
	mpz_clear (ztmp);
}

/*
 * Create a pre-computed table of the necessary powers of each RPP.
 */

	static
	struct subst_var_powers *
compute_rpp_powers (

struct RppPool *	pool,	/* IN: Pool of RPP's to gen powers for */
const int *		maxpow	/* IN: max power needed for each RPP */
)
{
int				i, j, k, ntabs, num_rpp;
struct calib_Zxyz_obj *		rpp;
struct poly_powers *		ptabs;
struct poly_powers *		pp;
struct poly_powers **		map;
struct calib_Zxyz_obj *		array;
const struct calib_Zxyz_dom *	ring;
struct subst_var_powers *	svp;

	CALIB_FATAL_ERROR_IF ((pool EQ NULL) OR (maxpow EQ NULL));

	num_rpp	= pool -> num_rpp;
	ring	= pool -> dom;

	/* Count how many of our RPP's need a power table. */
	ntabs = 0;
	for (i = 0; i < num_rpp; i++) {
		k = maxpow [i];
		if (k > 0) {
			++ntabs;
		}
	}

	/* Create an array to map RPP indices to their associated	*/
	/* power table (or NULL).					*/
	map = CALIB_NEWA (num_rpp, struct poly_powers *);
	memset (map, 0, num_rpp * sizeof (map [0]));

	/* Create power table for each RPP that needs one. */
	ptabs = NULL;
	if (ntabs > 0) {
		ptabs = CALIB_NEWA (ntabs, struct poly_powers);
		memset (ptabs, 0, ntabs * sizeof (ptabs [0]));
	}

	rpp = pool -> rpp;
	pp = ptabs;
	for (i = 0; i < num_rpp; i++, rpp++) {
		k = maxpow [i];
		if (k <= 0) continue;

		array = CALIB_NEWA (k, struct calib_Zxyz_obj);
		for (j = 0; j < k; j++) {
			ring -> init (ring, &array [j]);
		}
		ring -> set (&array [0], rpp);
		for (j = 1; j < k; j++) {
			ring -> mul (&array [j], &array [j-1], rpp);
		}

		pp -> max_power	= k;
		pp -> tab	= array;

		map [i] = pp;

		++pp;
	}

	svp = CALIB_NEW (struct subst_var_powers);
	memset (svp, 0, sizeof (*svp));
	svp -> nvars	= num_rpp;
	svp -> ntabs	= ntabs;
	svp -> powers	= ptabs;
	svp -> map	= map;

	return (svp);
}

/*
 * We are given a vector of integer RPP powers.
 * Compute the product of the RPP_j**pow_j for all j.
 * The necessary RPP powers are taken from the given rpp_powers object.
 */

	static
	void
eval_rpp_term (

struct calib_Zxyz_obj *	result,		/* IN/OUT: result */
const int *		pow,		/* IN: array of var powers */
struct subst_var_powers *
			rpp_powers,	/* IN: pre-computed table of RPP */
					/*     powers */
struct RppPool *	pool		/* IN: pool of RPP's being used */
)
{
int				i, k, num_rpp;
const struct calib_Zxyz_dom *	ring;
struct poly_powers *		pp;
struct calib_Zxyz_obj		TERM;

	CALIB_FATAL_ERROR_IF ((result	  EQ NULL) OR
			      (pow	  EQ NULL) OR
			      (rpp_powers EQ NULL) OR
			      (pool	  EQ NULL));

	ring	= pool -> dom;
	num_rpp	= pool -> num_rpp;

	CALIB_FATAL_ERROR_IF (ring NE result -> dom);

	/* Initialize term to 0 / 1, then set it to 1 / 1. */
	ring -> init (ring, &TERM);
	ring -> set_si (&TERM, 1);

	for (i = 0; i < num_rpp; i++) {
		k = pow [i];
		CALIB_FATAL_ERROR_IF (k < 0);
		if (k EQ 0) continue;

		/* Should have a power table for each RPP that needs one. */
		pp = rpp_powers -> map [i];
		CALIB_FATAL_ERROR_IF (pp EQ NULL);

		/* Multiply RPP**k into TERM. */
		ring -> mul (&TERM, &TERM, &(pp -> tab [k - 1]));
	}

	/* Copy resulting TERM out to caller. */
	ring -> set (result, &TERM);

	ring -> clear (&TERM);
}

/*
 * Function to handle creation / destruction of hash table entries
 * on behalf of substitute_poly_to_pfrat().
 */

	static
	void
hash_entry_op (

struct ivec_hash_entry *	ep,	/* IN: entry being manipulated */
enum ivec_entry_op		op,	/* IN: operation to perform */
void *				data	/* IN: client data */
)
{
const struct calib_Zxyz_dom *	subring;
struct hashent *		xp;

	CALIB_FATAL_ERROR_IF ((ep EQ NULL) OR (data EQ NULL));

	subring = (const struct calib_Zxyz_dom *) data;

	switch (op) {
	case IVEOP_CREATE:
		xp = CALIB_NEW (struct hashent);
		subring -> init (subring, &(xp -> numer));
		mpz_init_set_ui (xp -> den_konst, 1);
		xp -> owner = ep;
		ep -> data = xp;
		break;

	case IVEOP_FREE:
		xp = (struct hashent *) (ep -> data);
		CALIB_FATAL_ERROR_IF (xp EQ NULL);
		CALIB_FATAL_ERROR_IF (xp -> owner NE ep);
		mpz_clear (xp -> den_konst);
		subring -> clear (&(xp -> numer));
		memset (xp, 0, sizeof (*xp));
		free (xp);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Initialize the given PFRat object to value 0 / 1.
 */

	static
	void
pfrat_init (

struct PFRat *			rfp,	/* IN/OUT: PFRat object to init */
const struct calib_Zxyz_dom *	ring,	/* IN: ring for numerator */
struct RppPool *		pool	/* IN: pool to use RPP's from */
)
{
int		num_rpp;
int *		pow;

	CALIB_FATAL_ERROR_IF ((rfp  EQ NULL) OR
			      (ring EQ NULL) OR
			      (pool EQ NULL));

	num_rpp = pool -> num_rpp;
	CALIB_FATAL_ERROR_IF (num_rpp < 0);

	memset (rfp, 0, sizeof (*rfp));

	ring -> init (ring, &(rfp -> numer));

	mpz_init_set_ui (rfp -> denom.Konst, 1);

	pow = NULL;
	if (num_rpp > 0) {
		pow = CALIB_NEWA (num_rpp, int);
		memset (pow, 0, num_rpp * sizeof (pow [0]));
	}
	rfp -> denom.pow = pow;
}

/*
 * Clear out the given PFRat object.
 */

	static
	void
pfrat_clear (

struct PFRat *		rfp	/* IN/OUT: PFRat object to clear */
)
{
const struct calib_Zxyz_dom *	ring;

	if (rfp NE NULL) {
		if (rfp -> denom.pow NE NULL) {
			free (rfp -> denom.pow);
		}

		mpz_clear (rfp -> denom.Konst);

		ring = rfp -> numer.dom;
		CALIB_FATAL_ERROR_IF (ring EQ NULL);

		ring -> clear (&(rfp -> numer));

		memset (rfp, 0, sizeof (*rfp));
	}
}

/*
 * Return TRUE if-and-only-if the given polynomial is a constant.
 */

	static
	calib_bool
is_constant_poly (

const struct calib_Zxyz_obj *	poly		/* IN: polynomial to test */
)
{
int			i, k, nterms, nvars;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	nterms = poly -> nterms;

	if (nterms <= 0) {
		CALIB_FATAL_ERROR_IF (nterms < 0);
		/* Zero is a constant. */
		return (TRUE);
	}

	if (nterms > 1) {
		/* Cannot have multiple terms without being	*/
		/* some non-constant polynomial.		*/
		return (FALSE);
	}

	/* Exactly one term.  Need to see if constant. */

	nvars = poly -> dom -> nvars;
	for (i = 0; i < nvars; i++) {
		k = poly -> pow [i];
		if (k > 0) {
			/* At least 1 var to power >= 1. */
			return (FALSE);
		}
		CALIB_FATAL_ERROR_IF (k < 0);
	}

	/* Polynomial is a non-zero constant. */
	return (TRUE);
}

/*
 * The given polynomial must be a constant polynomial, either
 * identically +1, or -1.  Return the sign of this value.
 *
 * Will *fatal* if the polynomial isn't identically +1 or -1.
 */

	static
	int
get_sign (

const struct calib_Zxyz_obj *	poly		/* IN: polynomial to test */
)
{
int			i, nterms, nvars, sign;
const int *		pow;
mpz_srcptr		cp;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	nterms = poly -> nterms;

	CALIB_FATAL_ERROR_IF (nterms NE 1);

	/* Exactly one term.  Check coefficient value. */
	cp = poly -> coeff;
	if (mpz_cmp_ui (cp, +1) EQ 0) {
		sign = +1;
	}
	else if (mpz_cmp_si (cp, -1) EQ 0) {
		sign = -1;
	}
	else {
		CALIB_FATAL_ERROR;
	}

	/* Verify all var powers are zero. */
	nvars	= poly -> dom -> nvars;
	pow	= poly -> pow;
	for (i = 0; i < nvars; i++) {
		CALIB_FATAL_ERROR_IF (pow [i] NE 0);
	}

	return (sign);
}

/*
 * Free the given substitution list.
 */

	static
	void
rat_free_slist (

struct calib_subst_list *	slist	/* IN: substitution list to free */
)
{
int			i, n;
struct calib_rsubst *	array;
struct calib_rsubst *	rsp;

	if (slist EQ NULL) return;

	array	= slist -> subst;
	if (array NE NULL) {
		n = slist -> num_subst;
		rsp = array;
		for (i = 0; i < n; i++, rsp++) {
			rat_clear (&(rsp -> val));
			CALIB_FREE (rsp -> varname);
		}
		memset (array, 0, n * sizeof (array [0]));
		free (array);
	}

	free (slist);
}

/*
 * Is the given rational function identically zero?
 */

	static
	calib_bool
rat_zerop (

const struct calib_rat_obj *	op	/* IN: rational function to test */
)
{
const struct calib_Zxyz_dom *	ring;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	ring	= op -> dom -> ring;

	if (ring -> zerop (&(op -> numer))) {
		return (TRUE);
	}
	return (FALSE);
}


/*
 * Is the given rational function identically one?
 */

	static
	calib_bool
rat_onep (

const struct calib_rat_obj *	op	/* IN: rational function to test */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	ring;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	dom	= op -> dom;
	ring	= dom -> ring;

	if ((ring -> onep (&(op -> numer))) AND
	    (ring -> onep (&(op -> denom)))) {
		return (TRUE);
	}

	return (FALSE);
}

/*
 * Convert the given general expression into the given rational function.
 * All of the variables in the given expression *MUST* match those given
 * in the rational function domain.
 */

	static
	void
rat_set_genrep (

struct calib_rat_obj *		rop,	/* IN/OUT: resulting Zx polynomial */
const struct calib_genrep *	op	/* IN: expression to convert */
)
{
int				i, var;
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_rat_obj		POLY;
struct calib_rat_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;
	Zxyz	= dom -> ring;

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		rat_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		rat_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		rat_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		var = Zxyz -> lookup_var (Zxyz, op -> u.var);
		CALIB_FATAL_ERROR_IF (var < 0);
		rat_set_var_power (rop, var, 1);
		break;

	case CALIB_GENREP_OP_ADD:
		rat_init (dom, &POLY);
		rat_init (dom, &TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			rat_set_genrep (&TEMP, lp -> operand);
			rat_add (&POLY, &POLY, &TEMP);
		}
		rat_set (rop, &POLY);

		rat_clear (&TEMP);
		rat_clear (&POLY);
		break;

	case CALIB_GENREP_OP_MUL:
		rat_init (dom, &POLY);
		rat_init (dom, &TEMP);
		rat_set_si (&POLY, 1);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			rat_set_genrep (&TEMP, lp -> operand);
			rat_mul (&POLY, &POLY, &TEMP);
		}
		rat_set (rop, &POLY);

		rat_clear (&TEMP);
		rat_clear (&POLY);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;

		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);
		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			/* Optimize important special case. */
			var = Zxyz -> lookup_var (Zxyz, gp -> u.var);
			CALIB_FATAL_ERROR_IF (var < 0);
			rat_set_var_power (rop, var, i);
		}
		else {
			rat_init (dom, &TEMP);
			rat_set_genrep (&TEMP, gp);
			rat_ipow (rop, &TEMP, i);
			rat_clear (&TEMP);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Convert the given rational function into general representation.
 */

	static
	struct calib_genrep *
rat_to_genrep (

const struct calib_rat_obj *	op	/* IN: rat to convert */
)
{
const struct calib_rat_dom *	dom;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	dom	= op -> dom;
	Zxyz	= dom -> ring;

	gp1 = Zxyz -> to_genrep (&(op -> numer));
	if (NOT (Zxyz -> onep (&(op -> denom)))) {
		gp2 = Zxyz -> to_genrep (&(op -> denom));
		gp3 = calib_genrep_ipow (gp2, -1);
		calib_genrep_free (gp2);
		gp2 = gp3;

		gp3 = calib_genrep_mul2 (gp1, gp2);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		gp1 = gp3;
	}

	return (gp1);
}

/*
 * Print a rational expression value, terminated with a newline.
 */

	static
	void
rat_print (

const struct calib_rat_obj *	op	/* IN: rational expression to print */
)
{
	rat_print_nnl (op);
	printf ("\n");
}


/*
 * Print a rational expression value with no-newline.
 */

	static
	void
rat_print_nnl (

const struct calib_rat_obj *	op	/* IN: rational expression to print */
)
{
const struct calib_Zxyz_dom *	ring;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	ring = op -> numer.dom;
	CALIB_FATAL_ERROR_IF (ring NE op -> denom.dom);

	printf (LPAREN_STRING);
	ring -> print_maxima_nnl (&(op -> numer));
	printf (RPAREN_STRING " / " LPAREN_STRING);
	ring -> print_maxima_nnl (&(op -> denom));
	printf (RPAREN_STRING);
}

/*
 * The field of rational functions R / R, where R = Z[x,y,z,...].
 */

static const struct calib_rat_dom	_rat_dom = {
	.ring			= NULL,

	.init			= rat_init,
	.clear			= rat_clear,
	.set			= rat_set,
	.set_si			= rat_set_si,
	.set_z			= rat_set_z,
	.set_q			= rat_set_q,
	.set_var_power		= rat_set_var_power,
	.add			= rat_add,
	.sub			= rat_sub,
	.neg			= rat_neg,
	.mul			= rat_mul,
	.mul_z			= rat_mul_z,
	.mul_q			= rat_mul_q,
	.ipow			= rat_ipow,
	.inv			= rat_inv,
	.div			= rat_div,
	.canonicalize		= rat_canonicalize,
	.factor			= rat_factor,
	.substitute_with_varmap	= rat_substitute_with_varmap,
	.free_slist		= rat_free_slist,
	.zerop			= rat_zerop,
	.onep			= rat_onep,
	.set_genrep		= rat_set_genrep,
	.to_genrep		= rat_to_genrep,
	.print			= rat_print,
	.print_nnl		= rat_print_nnl,
};

/*
 * Construct the field of rational functions R / R, where R is the
 * given ring Z[x,y,z,...].
 */

	struct calib_rat_dom *
calib_make_rat_dom (

const struct calib_Zxyz_dom *	ring	/* IN: Ring for numer, denom */
)
{
struct calib_rat_dom *		dom;

	dom = CALIB_NEW (struct calib_rat_dom);
	memcpy (dom, &_rat_dom, sizeof (*dom));

	dom -> ring	= ring;

	return (dom);
}

/*
 * Free up the given rational function field description.
 */

	void
calib_free_rat_dom (

const struct calib_rat_dom *	dom	/* IN: rat domain to free */
)
{
struct calib_rat_dom *	ldom;

	if (dom NE NULL) {
		/* Cast away the constness... */
		ldom = (struct calib_rat_dom *) dom;

		memset (ldom, 0, sizeof (*ldom));

		free (ldom);
	}
}
