/**********************************************************************

	$Id: ratfac.c,v 1.1 2024/10/20 00:36:28 warme Exp $

	File:	ratfac.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Rational functions in partially-factored form.

	Note: This is currently used only in Zxyz::resultant_old()
	-- an old, slow, obsolete multi-variate polynomial resultant
	algorithm.

***********************************************************************

	Modification Log:

	a-1:	08/01/2020	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "ratfac.h"

#include "calib/logic.h"
#include "calib/new.h"
#include "calib/types.h"
#include "calib/Zxyz.h"
#include "calib/fatal.h"
#include <gmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	LPAREN_STRING	"("
#define	RPAREN_STRING	")"

/*
 * Global Functions
 */

void		_calib_RatFac_clear (struct calib_RatFac * rfp);
struct calib_Zxyz_factor *
		_calib_RatFac_denom (const struct calib_RatFac * rfp);
void		_calib_RatFac_init (const struct calib_Zxyz_dom *	dom,
				    struct calib_RatFac *		rfp);
void		_calib_RatFac_mul (struct calib_RatFac *		rfp,
				   const struct calib_Zxyz_obj *	poly,
				   int					pow);
void		_calib_RatFac_negate (struct calib_RatFac * rfp);
struct calib_Zxyz_factor *
		_calib_RatFac_numer (const struct calib_RatFac * rfp);
void		_calib_RatFac_print (const struct calib_RatFac * rfp);


/*
 * Local Types
 */

/*
 * A linked list to represent partially factorized integers
 * and/or rational numbers (Z Factor List).
 */

struct ZFList {
	mpz_t			zfl_zval;	/* Positive integer value */
	struct ZFList *		zfl_next;	/* Next in linked list */
	int			zfl_pow;	/* Exponent (signed!) */
};

/*
 * A linked list to represent partially factorized polynomials.
 */

struct PFList {
	struct calib_Zxyz_obj	pfl_poly;	/* Polynomial factor */
	struct PFList *		pfl_next;	/* Next in linked list */
	int			pfl_pow;	/* Exponent (signed!) */
};

/*
 * Small primes that we fully factorize.
 */

#define		N_SMALL_PRIMES		8

static const int	small_primes [N_SMALL_PRIMES] = {
	2, 3, 5, 7, 11, 13, 17, 19
};


/*
 * Our local "implementation" object.
 */

struct calib_RFImp {
	const struct calib_Zxyz_dom *
			rfi_dom;	/* Z[x,y,z] domain */
	int		rfi_sign;	/* Numerator sign */
	int		rfi_sp_pow [N_SMALL_PRIMES];
					/* Powers of small primes */
	struct ZFList *	rfi_zroot;	/* Integer factors */
	struct PFList *	rfi_proot;	/* Polynomial factors */
};

/*
 * Local Functions
 */

static void	RatFac_insert_integer (struct calib_RFImp *	imp,
				       mpz_ptr			zval,
				       int			pow);
static void	RatFac_insert_poly (struct calib_RFImp *	imp,
				    struct calib_Zxyz_obj *	poly,
				    int				pow);
static void	RatFac_insert_ZFList (struct calib_RFImp *	imp,
				      struct ZFList *		zflp);
static struct PFList *
		RatFac_PFList_new (const struct calib_Zxyz_obj *	poly,
				   int					pow);
static void	RatFac_PFList_free (struct PFList * pflp);
static void	RatFac_PFList_free_chain (struct PFList * pflp);
static calib_bool
		RatFac_poly_is_one (const struct calib_Zxyz_obj * poly);
static void	RatFac_ZFList_free (struct ZFList * zflp);
static void	RatFac_ZFList_free_chain (struct ZFList * zflp);
static struct ZFList *
		RatFac_ZFList_new (mpz_srcptr zval, int pow);

/*
 * Initialize the given RatFac object to be the empty product (1/1)
 * within the given polynomial domain.
 */

	void
_calib_RatFac_init (

const struct calib_Zxyz_dom *	dom,	/* IN: Polynomial domain */
struct calib_RatFac *		rfp	/* IN/OUT: RatFac to init */
)
{
struct calib_RFImp *	imp;

	CALIB_FATAL_ERROR_IF ((dom EQ NULL) OR (rfp EQ NULL));

	imp = CALIB_NEW (struct calib_RFImp);
	memset (imp, 0, sizeof (*imp));

	imp -> rfi_dom	= dom;
	imp -> rfi_sign	= +1;

	rfp -> rf_imp = imp;
}

/*
 * Clear the given RatFac object.
 */

	void
_calib_RatFac_clear (

struct calib_RatFac *	rfp	/* IN/OUT: RatFac object to clear */
)
{
struct calib_RFImp *	imp;

	CALIB_FATAL_ERROR_IF (rfp EQ NULL);

	imp = rfp -> rf_imp;
	rfp -> rf_imp = NULL;

	CALIB_FATAL_ERROR_IF (imp EQ NULL);

	RatFac_ZFList_free_chain (imp -> rfi_zroot);
	RatFac_PFList_free_chain (imp -> rfi_proot);

	memset (imp, 0, sizeof (*imp));

	free (imp);
}

/*
 * Multiply this RatFac by poly(x,y,z)**pow, where pow is a
 * signed integer.
 */

	void
_calib_RatFac_mul (

struct calib_RatFac *		rfp,	/* IN/OUT: RatFac to multiply "into" */
const struct calib_Zxyz_obj *	poly,	/* IN: Polynomial to multiply */
int				pow	/* IN: Exponent (signed!) */
)
{
int				i, nterms;
struct calib_RFImp *		imp;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_factor *	sqfactors;
struct calib_Zxyz_factor *	sqfp;
struct calib_Zxyz_obj		LPOLY;
mpz_t				ZCONT;

	CALIB_FATAL_ERROR_IF ((rfp EQ NULL) OR (poly EQ NULL));

	imp = rfp -> rf_imp;

	CALIB_FATAL_ERROR_IF (imp EQ NULL);

	dom = imp -> rfi_dom;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);
	CALIB_FATAL_ERROR_IF (poly -> dom NE dom);

	/* Should *never* multiply zero into a RatFac! */
	nterms = poly -> nterms;
	CALIB_FATAL_ERROR_IF (nterms <= 0);

	if (pow EQ 0) {
		return;
	}

	mpz_init (ZCONT);

	/* Make a local, modifiable copy of given polynomial. */
	dom -> init (dom, &LPOLY);
	dom -> set (&LPOLY, poly);

	/* Strip off any integer content. */
	dom -> strip_z_content (ZCONT, &LPOLY);

	/* Normalize sign of leading term. */
	i = mpz_sgn (&LPOLY.coeff [0]);
	if (i <= 0) {
		CALIB_FATAL_ERROR_IF (i EQ 0);
		dom -> neg (&LPOLY, &LPOLY);
		mpz_neg (ZCONT, ZCONT);
	}

	RatFac_insert_integer (imp, ZCONT, pow);

	sqfactors = dom -> sqf_factor (&LPOLY);
	for (sqfp = sqfactors; sqfp NE NULL; sqfp = sqfp -> next) {
		i = pow * sqfp -> multiplicity;
		RatFac_insert_poly (imp, sqfp -> factor, i);
	}
	dom -> free_factors (sqfactors);

	dom -> clear (&LPOLY);
	mpz_clear (ZCONT);
}

/*
 * Insert the given zval**pow into the given RatFac Implementation.
 * Note: this *modifies* the caller's zval!
 */

	static
	void
RatFac_insert_integer (

struct calib_RFImp *	imp,		/* IN/OUT: RFImp to insert into */
mpz_ptr			zval,		/* IN: Integer to insert */
int			pow		/* IN: Exponent (signed!) */
)
{
int			i, j;
unsigned long		nrem1, nrem2;
struct ZFList **	hookp;
struct ZFList *		zflp;
struct ZFList *		np;
mpz_t			P, GCD;

	CALIB_FATAL_ERROR_IF ((imp EQ NULL) OR (zval EQ NULL));

	i = mpz_sgn (zval);
	if (i <= 0) {
		CALIB_FATAL_ERROR_IF (i EQ 0);

		mpz_abs (zval, zval);
		if ((pow & 0x01) NE 0) {
			imp -> rfi_sign = - (imp -> rfi_sign);
		}
	}

	i = mpz_cmp_ui (zval, 1);
	if (i <= 0) {
		CALIB_FATAL_ERROR_IF (i < 0);
		return;
	}

	mpz_init (P);
	mpz_init (GCD);

	/* Remove small primes. */
	for (i = 0; i < N_SMALL_PRIMES; i++) {
		mpz_set_ui (P, small_primes [i]);
		nrem1 = mpz_remove (zval, zval, P);
		if (nrem1 > 0) {
			imp -> rfi_sp_pow [i] += (pow * nrem1);
			if (mpz_cmp_ui (zval, 1) EQ 0) {
				return;
			}
		}
	}

	/* Fragment with respect to existing factors.		*/
	hookp = &(imp -> rfi_zroot);
	for (;;) {
		zflp = *hookp;
		if (zflp EQ NULL) {
			CALIB_FATAL_ERROR_IF (mpz_cmp_ui (zval, 1) EQ 0);

			/* zval is now relatively prime to all	*/
			/* existing nodes.  Add a new node for	*/
			/* zval.				*/
			np = RatFac_ZFList_new (zval, pow);
			RatFac_insert_ZFList (imp, np);
			break;
		}

		mpz_gcd (GCD, zval, zflp -> zfl_zval);
		if (mpz_cmp_ui (GCD, 1) EQ 0) {
			hookp = &(zflp -> zfl_next);
			continue;
		}

		/* We have a common factor.  Remove all copies of it	*/
		/* from zval.						*/
		nrem1 = mpz_remove (zval, zval, GCD);
		CALIB_FATAL_ERROR_IF (nrem1 <= 0);

		/* Now process the existing node by seeing how many	*/
		/* copies of the GCD can be removed from it.		*/
		nrem2 = mpz_remove (P, zflp -> zfl_zval, GCD);
		CALIB_FATAL_ERROR_IF (nrem2 <= 0);

		if ((nrem2 EQ 1) AND (mpz_cmp_ui (P, 1) EQ 0)) {
			/* Should have node EQ GCD. */
			CALIB_FATAL_ERROR_IF (mpz_cmp (zflp -> zfl_zval, GCD) NE 0);
			zflp -> zfl_pow += (nrem1 * pow);

			hookp = &(zflp -> zfl_next);
		}
		else {
			/* This node must split in two: one for P and	*/
			/* another for GCD.  Remove current node from	*/
			/* the list.					*/
			*hookp = zflp -> zfl_next;
			zflp -> zfl_next = NULL;

			i = mpz_cmp_ui (P, 1);
			CALIB_FATAL_ERROR_IF (i < 0);
			if (i > 0) {
				np = RatFac_ZFList_new (P,   zflp -> zfl_pow);
				RatFac_insert_ZFList (imp, np);
			}
			/* Note: We allow nodes with zero powers to	*/
			/* remain in the list to afford the best	*/
			/* possible split of future integers.		*/
			j = (nrem1 * pow) + (nrem2 * zflp -> zfl_pow);
			np = RatFac_ZFList_new (GCD, j);
			RatFac_insert_ZFList (imp, np);

			RatFac_ZFList_free (zflp);
		}

		if (mpz_cmp_ui (zval, 1) EQ 0) {
			/* Given zval has been completely fragmented	*/
			/* into small primes and the ZFList.  We are	*/
			/* done.					*/
			break;
		}
	}

	mpz_clear (GCD);
	mpz_clear (P);
}

/*
 * Insert the given ZFList node into our list.  We keep them sorted
 * by increasing value.
 */

	static
	void
RatFac_insert_ZFList (

struct calib_RFImp *	imp,		/* IN/OUT: RFImp to insert into */
struct ZFList *		zflp		/* IN: Node to insert */
)
{
int			cmp;
struct ZFList **	hookp;
struct ZFList *		p;

	CALIB_FATAL_ERROR_IF ((imp EQ NULL) OR (zflp EQ NULL));
	CALIB_FATAL_ERROR_IF (zflp -> zfl_next NE NULL);

	hookp = &(imp -> rfi_zroot);
	for (;;) {
		p = *hookp;
		if (p EQ NULL) {
			/* New node goes last. */
			*hookp = zflp;
			break;
		}
		cmp = mpz_cmp (zflp -> zfl_zval, p -> zfl_zval);
		if (cmp < 0) {
			/* Insert new node here. */
			zflp -> zfl_next = p;
			*hookp = zflp;
			break;
		}

		/* Must not introduce any duplicates. */
		CALIB_FATAL_ERROR_IF (cmp EQ 0);

		hookp = &(p -> zfl_next);
	}
}

/*
 * Insert the given poly**pow into the given RatFac Implementation.
 * Note: this *modifies* the caller's polynomial!
 */

	static
	void
RatFac_insert_poly (

struct calib_RFImp *	imp,		/* IN/OUT: RFImp to insert into */
struct calib_Zxyz_obj *	poly,		/* IN: Polynomial to insert */
int			pow		/* IN: Exponent (signed!) */
)
{
int				j;
unsigned long			nrem1, nrem2;
const struct calib_Zxyz_dom *	dom;
struct PFList **		hookp;
struct PFList *			pflp;
struct PFList *			np;
struct calib_Zxyz_obj		P, GCD, TEMP;

	CALIB_FATAL_ERROR_IF ((imp EQ NULL) OR (poly EQ NULL));

	if (RatFac_poly_is_one (poly)) {
		return;
	}

	dom	= poly -> dom;

	dom -> init (dom, &P);
	dom -> init (dom, &GCD);
	dom -> init (dom, &TEMP);

	/* Fragment with respect to existing factors.		*/
	hookp = &(imp -> rfi_proot);
	for (;;) {
		pflp = *hookp;
		if (pflp EQ NULL) {
			CALIB_FATAL_ERROR_IF (RatFac_poly_is_one (poly));

			/* poly is now relatively prime to all	*/
			/* existing nodes.  Add a new node for	*/
			/* it.					*/
			np = RatFac_PFList_new (poly, pow);
			np -> pfl_next = imp -> rfi_proot;
			imp -> rfi_proot = np;
			break;
		}

		dom -> gcd (&GCD, poly, &(pflp -> pfl_poly));
		if (RatFac_poly_is_one (&GCD)) {
			hookp = &(pflp -> pfl_next);
			continue;
		}

		/* We have a common factor.  Remove all copies of it	*/
		/* from poly.						*/
		nrem1 = dom -> div_remove (poly, poly, &GCD);
		CALIB_FATAL_ERROR_IF (nrem1 <= 0);

		/* Now process the existing node by seeing how many	*/
		/* copies of the GCD can be removed from it.		*/
		nrem2 = dom -> div_remove (&P, &(pflp -> pfl_poly), &GCD);
		CALIB_FATAL_ERROR_IF (nrem2 <= 0);

		if ((nrem2 EQ 1) AND (RatFac_poly_is_one (&P))) {
			/* Should have node EQ GCD. */
			dom -> sub (&TEMP, &(pflp -> pfl_poly), &GCD);
			CALIB_FATAL_ERROR_IF (TEMP.nterms NE 0);

			pflp -> pfl_pow += (nrem1 * pow);

			hookp = &(pflp -> pfl_next);
		}
		else {
			/* This node must split in two: one for P and	*/
			/* another for GCD.  Remove current node from	*/
			/* the list.					*/
			*hookp = pflp -> pfl_next;
			pflp -> pfl_next = NULL;

			if (NOT RatFac_poly_is_one (&P)) {
				np = RatFac_PFList_new (&P, pflp -> pfl_pow);
				np -> pfl_next = imp -> rfi_proot;
				imp -> rfi_proot = np;
			}
			/* Note: We allow nodes with zero powers to	*/
			/* remain in the list to afford the best	*/
			/* possible split of future polynomials.	*/
			j = (nrem1 * pow) + (nrem2 * pflp -> pfl_pow);
			np = RatFac_PFList_new (&GCD, j);
			np -> pfl_next = imp -> rfi_proot;
			imp -> rfi_proot = np;

			RatFac_PFList_free (pflp);
		}

		if (RatFac_poly_is_one (poly)) {
			/* Given poly has been completely fragmented	*/
			/* into the PFList.  We are done.		*/
			break;
		}
	}

	dom -> clear (&TEMP);
	dom -> clear (&GCD);
	dom -> clear (&P);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically one.
 */

	static
	calib_bool
RatFac_poly_is_one (

const struct calib_Zxyz_obj *	poly	/* IN: Polynomial to test */
)
{
int				i , nterms, nvars;
const int *			pow;
const struct calib_Zxyz_dom *	dom;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	nterms	= poly -> nterms;

	if (nterms NE 1) {
		return (FALSE);
	}

	dom	= poly -> dom;
	pow	= poly -> pow;
	nvars	= dom -> nvars;

	for (i = 0; i < nvars; i++) {
		CALIB_FATAL_ERROR_IF (pow [i] < 0);
		if (pow [i] > 0) {
			return (FALSE);
		}
	}

	if (mpz_cmp_ui (poly -> coeff, 1) NE 0) {
		return (FALSE);
	}

	/* Polynomial is identically one! */
	return (TRUE);
}

/*
 * Negate the given RatFac object.
 */

	void
_calib_RatFac_negate (

struct calib_RatFac *	rfp	/* IN/OUT: RatFac object to access */
)
{
struct calib_RFImp *	imp;

	CALIB_FATAL_ERROR_IF (rfp EQ NULL);

	imp = rfp -> rf_imp;

	CALIB_FATAL_ERROR_IF (imp EQ NULL);

	imp -> rfi_sign = - (imp -> rfi_sign);
}

/*
 * Return partially factored numerator.
 */

	struct calib_Zxyz_factor *
_calib_RatFac_numer (

const struct calib_RatFac *	rfp	/* IN: RatFac object to access */
)
{
int				i, j;
struct calib_RFImp *		imp;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_factor *	root;
struct calib_Zxyz_factor **	hookp;
struct ZFList *			zflp;
struct PFList *			pflp;
struct calib_Zxyz_factor *	nfp;
struct calib_Zxyz_obj		TEMP;

	CALIB_FATAL_ERROR_IF (rfp EQ NULL);

	imp	= rfp -> rf_imp;

	CALIB_FATAL_ERROR_IF (imp EQ NULL);

	dom	= imp -> rfi_dom;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	dom -> init (dom, &TEMP);

	root = NULL;
	hookp = &root;

	for (i = 0; i < N_SMALL_PRIMES; i++) {
		j = imp -> rfi_sp_pow [i];
		if (j <= 0) continue;

		dom -> set_si (&TEMP, small_primes [i]);

		nfp = CALIB_NEW (struct calib_Zxyz_factor);
		nfp -> multiplicity	= j;
		nfp -> factor		= dom -> dup (&TEMP);
		nfp -> next		= NULL;

		*hookp = nfp;
		hookp = &(nfp -> next);
	}

	for (zflp = imp -> rfi_zroot; zflp NE NULL; zflp = zflp -> zfl_next) {
		j = zflp -> zfl_pow;
		if (j <= 0) continue;

		dom -> set_z (&TEMP, zflp -> zfl_zval);

		nfp = CALIB_NEW (struct calib_Zxyz_factor);
		nfp -> multiplicity	= j;
		nfp -> factor		= dom -> dup (&TEMP);
		nfp -> next		= NULL;

		*hookp = nfp;
		hookp = &(nfp -> next);
	}

	for (pflp = imp -> rfi_proot; pflp NE NULL; pflp = pflp -> pfl_next) {
		j = pflp -> pfl_pow;
		if (j <= 0) continue;

		nfp = CALIB_NEW (struct calib_Zxyz_factor);
		nfp -> multiplicity	= j;
		nfp -> factor		= dom -> dup (&(pflp -> pfl_poly));
		nfp -> next		= NULL;

		*hookp = nfp;
		hookp = &(nfp -> next);
	}

	dom -> clear (&TEMP);

	return (root);
}

/*
 * Return partially factored denominator.
 */

	struct calib_Zxyz_factor *
_calib_RatFac_denom (

const struct calib_RatFac *	rfp	/* IN: RatFac object to access */
)
{
int				i, j;
struct calib_RFImp *		imp;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_factor *	root;
struct calib_Zxyz_factor **	hookp;
struct ZFList *			zflp;
struct PFList *			pflp;
struct calib_Zxyz_factor *	nfp;
struct calib_Zxyz_obj		TEMP;

	CALIB_FATAL_ERROR_IF (rfp EQ NULL);

	imp	= rfp -> rf_imp;

	CALIB_FATAL_ERROR_IF (imp EQ NULL);

	dom = imp -> rfi_dom;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	dom -> init (dom, &TEMP);

	root = NULL;
	hookp = &root;

	for (i = 0; i < N_SMALL_PRIMES; i++) {
		j = imp -> rfi_sp_pow [i];
		if (j >= 0) continue;

		dom -> set_si (&TEMP, small_primes [i]);

		nfp = CALIB_NEW (struct calib_Zxyz_factor);
		nfp -> multiplicity	= - j;
		nfp -> factor		= dom -> dup (&TEMP);
		nfp -> next		= NULL;

		*hookp = nfp;
		hookp = &(nfp -> next);
	}

	for (zflp = imp -> rfi_zroot; zflp NE NULL; zflp = zflp -> zfl_next) {
		j = zflp -> zfl_pow;
		if (j >= 0) continue;

		dom -> set_z (&TEMP, zflp -> zfl_zval);

		nfp = CALIB_NEW (struct calib_Zxyz_factor);
		nfp -> multiplicity	= - j;
		nfp -> factor		= dom -> dup (&TEMP);
		nfp -> next		= NULL;

		*hookp = nfp;
		hookp = &(nfp -> next);
	}

	for (pflp = imp -> rfi_proot; pflp NE NULL; pflp = pflp -> pfl_next) {
		j = pflp -> pfl_pow;
		if (j >= 0) continue;

		nfp = CALIB_NEW (struct calib_Zxyz_factor);
		nfp -> multiplicity	= - j;
		nfp -> factor		= dom -> dup (&(pflp -> pfl_poly));
		nfp -> next		= NULL;

		*hookp = nfp;
		hookp = &(nfp -> next);
	}

	dom -> clear (&TEMP);

	return (root);
}

/*
 * Create a new PFList object containing the given poly**pow.
 */

	static
	struct PFList *
RatFac_PFList_new (

const struct calib_Zxyz_obj *	poly,		/* IN: Polynomial */
int				pow		/* IN: Exponent (signed!) */
)
{
int				i, k, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
const int *			ppow;
struct PFList *			pflp;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	nterms	= poly -> nterms;
	dom	= poly -> dom;
	ppow	= poly -> pow;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF (nterms <= 0);

	/* Verify that polynomial has degree >= 1 in at least	*/
	/* one variable.					*/
	k = nterms * nvars;
	for (i = 0; ; i++) {
		/* Crash if we have no variables in polynomial. */
		CALIB_FATAL_ERROR_IF (i >= k);

		if (ppow [i] > 0) break;

		CALIB_FATAL_ERROR_IF (ppow [i] < 0);
	}

	pflp = CALIB_NEW (struct PFList);
	memset (pflp, 0, sizeof (*pflp));

	dom -> init (dom, &(pflp -> pfl_poly));
	dom -> set (&(pflp -> pfl_poly), poly);

	pflp -> pfl_next	= NULL;
	pflp -> pfl_pow		= pow;

	return (pflp);
}

/*
 * Free up the given *chain* of PFList objects.
 */

	static
	void
RatFac_PFList_free_chain (

struct PFList *		pflp		/* IN: PFList chain to free */
)
{
struct PFList *		tmp;

	while (pflp NE NULL) {
		tmp = pflp -> pfl_next;
		pflp -> pfl_next = NULL;

		RatFac_PFList_free (pflp);

		pflp = tmp;
	}
}

/*
 * Free up the given PFList object.
 */

	static
	void
RatFac_PFList_free (

struct PFList *		pflp		/* IN: PFList object to free */
)
{
struct calib_Zxyz_obj *		poly;
const struct calib_Zxyz_dom *	dom;

	CALIB_FATAL_ERROR_IF (pflp EQ NULL);

	poly	= &(pflp -> pfl_poly);
	dom	= poly -> dom;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	dom -> clear (poly);

	memset (pflp, 0, sizeof (*pflp));

	free (pflp);
}

/*
 * Create a new ZFList object containing the given zval**pow.
 */

	static
	struct ZFList *
RatFac_ZFList_new (

mpz_srcptr		zval,		/* IN: Positive integer value */
int			pow		/* IN: Exponent (signed!) */
)
{
struct ZFList *		zflp;

	CALIB_FATAL_ERROR_IF (zval EQ NULL);
	CALIB_FATAL_ERROR_IF (mpz_sgn (zval) <= 0);
	CALIB_FATAL_ERROR_IF (pow EQ 0);

	zflp = CALIB_NEW (struct ZFList);
	memset (zflp, 0, sizeof (*zflp));

	mpz_init_set (zflp -> zfl_zval, zval);
	zflp -> zfl_next	= NULL;
	zflp -> zfl_pow		= pow;

	return (zflp);
}

/*
 * Free up the given *chain* of ZFList objects.
 */

	static
	void
RatFac_ZFList_free_chain (

struct ZFList *		zflp		/* IN: ZFList chain to free */
)
{
struct ZFList *		tmp;

	while (zflp NE NULL) {
		tmp = zflp -> zfl_next;
		zflp -> zfl_next = NULL;

		RatFac_ZFList_free (zflp);

		zflp = tmp;
	}
}

/*
 * Free up the given ZFList object.
 */

	static
	void
RatFac_ZFList_free (

struct ZFList *		zflp		/* IN: ZFList object to free */
)
{
	CALIB_FATAL_ERROR_IF (zflp EQ NULL);

	mpz_clear (zflp -> zfl_zval);

	memset (zflp, 0, sizeof (*zflp));

	free (zflp);
}

/*
 * Print out the given ratfac object.
 */

	void
_calib_RatFac_print (

const struct calib_RatFac *	rfp	/* IN: RatFac to print */
)
{
int				i, j, k;
const char *			sep;
const char *			sep2;
struct calib_RFImp *		imp;
struct ZFList *			zfp;
struct PFList *			pfp;
struct calib_Zxyz_obj *		poly;
const struct calib_Zxyz_dom *	ring;

	CALIB_FATAL_ERROR_IF (rfp EQ NULL);

	imp = rfp -> rf_imp;

	if (imp -> rfi_sign < 0) {
		printf ("-");
	}

	/* No factors printed yet. */
	sep = " ";
	sep2 = " * ";
	for (i = 0; i < N_SMALL_PRIMES; i++) {
		k = imp -> rfi_sp_pow [i];
		if (k EQ 0) continue;
		j = small_primes [i];
		if (k < 0) {
			printf ("%s%d^(%d)", sep, j, k);
		}
		else if (k EQ 1) {
			printf ("%s%d", sep, j);
		}
		else {
			printf ("%s%d^%d", sep, j, k);
		}
		sep = sep2;
	}
	for (zfp = imp -> rfi_zroot; zfp NE NULL; zfp = zfp -> zfl_next) {
		k = zfp -> zfl_pow;
		if (k EQ 0) continue;

		printf ("%s", sep);
		mpz_out_str (stdout, 10, zfp -> zfl_zval);
		if (k < 0) {
			printf ("^(%d)", k);
		}
		else if (k NE 1) {
			printf ("^%d", k);
		}
		sep = sep2;
	}
	for (pfp = imp -> rfi_proot; pfp NE NULL; pfp = pfp -> pfl_next) {
		k = pfp -> pfl_pow;
		if (k EQ 0) continue;

		poly	= &(pfp -> pfl_poly);
		ring	= poly -> dom;

		printf ("%s" LPAREN_STRING, sep);
		ring -> print_maxima_nnl (poly);
		printf (RPAREN_STRING);
		if (k < 0) {
			printf ("^(%d)", k);
		}
		else if (k NE 1) {
			printf ("^%d", k);
		}
		sep = sep2;
	}
	printf (";\n");
}
