/**********************************************************************

	$Id: algfac.c,v 1.1 2024/10/20 00:58:37 warme Exp $

	File:	algfac.c
	Rev:	a-3
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Factorization of polynomials over K[x], where K = Q(a).

***********************************************************************

	Modification Log:

	a-1:	07/01/2012	warme
		: Created.
	a-2:	10/11/2019	warme
		: Use new pretty printer.
	a-3:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/cputime.h"
#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/prompt.h"
#include "calib/Qa.h"
#include "calib/Qax.h"
#include "calib/Zx.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static void	usage (void);

static char *	me;

	int
main (

int		argc,
char **		argv
)
{
int				i, status;
char				c;
char *				ap;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
char *				xvar;
char *				avar;
struct calib_Zx_factor *	gfactors;
struct calib_Qax_factor *	factors;
const struct calib_Zx_dom *	Zx;
struct calib_Qa_dom *		Qa;
struct calib_Qax_dom *		Qax;
struct calib_Zx_obj		GPOLY;
struct calib_Qax_obj		FPOLY;
struct calib_genrep_varlist	varlist;
size_t				nbits;
calib_cpu_time_t		T0, T1;
char				tbuf [16];

	me = argv [0];
	for (i = 1; i < argc; i++) {
		ap = argv [i];
		if (*ap++ NE '-') usage ();
		while ((c = *ap++) NE '\0') {
			switch (c) {
			case 'c':
				calib_Qax_default_settings.WR_print_trial_factor_combinations = TRUE;
				break;

			case 'i':
				calib_Qax_default_settings.WR_print_initial_factors = TRUE;
				break;

			case 'l':
				calib_Qax_default_settings.WR_print_lifted_factors = TRUE;
				break;

			case 'n':
				calib_Qax_default_settings.factor_method = CALIB_QAX_FACTOR_METHOD_NORM;
				break;

			case 'w':
				calib_Qax_default_settings.factor_method = CALIB_QAX_FACTOR_METHOD_WEINBERGER_ROTHSCHILD;
				break;

			default:
				usage ();
				break;
			}
		}
	}

	Zx = calib_get_Zx_dom ();

	while (1) {
		T0 = calib_get_cpu_time ();

		calib_prompt ("\nCoefficient field polynomial: ");
		gp1 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			continue;
		}
		if (gp1 EQ NULL) break;

		calib_genrep_get_varlist (&varlist, gp1);
		avar = NULL;
		if (varlist.nvars EQ 1) {
			avar = strdup (varlist.vlist [0]);
		}
		calib_genrep_clear_varlist (&varlist);

		if (avar EQ NULL) {
			fprintf (stderr, "Must use exactly one variable\n");
			calib_genrep_free (gp1);
			continue;
		}
		Zx -> init (&GPOLY);
		Zx -> set_genrep (&GPOLY, gp1, avar);
		calib_genrep_free (gp1);

		/* Verify that g(x) is irreducible. */
		gfactors = Zx -> factor (&GPOLY);
		if ((gfactors EQ NULL) OR
		    (gfactors -> next NE NULL) OR
		    (gfactors -> multiplicity NE 1)) {
			gp1 = Zx -> to_genrep (&GPOLY, avar);
			gp2 = Zx -> factors_to_genrep (gfactors, avar);
			fprintf (stderr, "Polynomial must be irreducible:\n");
			calib_genrep_prettyprint_file (stderr, gp1);
			fprintf (stderr, " =\n");
			calib_genrep_prettyprint_file (stderr, gp2);
			fflush (stderr);
			calib_genrep_free (gp2);
			calib_genrep_free (gp1);
			Zx -> free_factors (gfactors);
			Zx -> clear (&GPOLY);
			free (avar);
			continue;
		}
		Zx -> free_factors (gfactors);

		calib_prompt ("\nPolynomial to Factor:\t ");
		gp1 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			Zx -> clear (&GPOLY);
			free (avar);
			continue;
		}
		if (gp1 EQ NULL) break;

		calib_genrep_get_varlist (&varlist, gp1);
		xvar = NULL;
		if (varlist.nvars EQ 2) {
			if (strcmp (varlist.vlist [0], avar) EQ 0) {
				xvar = strdup (varlist.vlist [1]);
			}
			else if (strcmp (varlist.vlist [1], avar)) {
				xvar = strdup (varlist.vlist [0]);
			}
		}
		else if (varlist.nvars EQ 1) {
			if (strcmp (varlist.vlist [0], avar) NE 0) {
				xvar = strdup (varlist.vlist [0]);
			}
		}
		calib_genrep_clear_varlist (&varlist);

		if (xvar EQ NULL) {
			fprintf (stderr, "Incorrect variables used.\n");
			calib_genrep_free (gp1);
			Zx -> clear (&GPOLY);
			free (avar);
			continue;
		}

		Qa  = calib_make_Qa_dom (&GPOLY);
		Qax = calib_make_Qax_dom (Qa);

		Qax -> init (Qax, &FPOLY);

		Qax -> set_genrep (&FPOLY, gp1, xvar, avar);
		calib_genrep_free (gp1);

		/* Factorize into irreducibles over Q(a)[x]. */

		factors = Qax -> factor (&FPOLY);

		nbits = Qax -> factors_bit_size (factors);

		gp1 = Zx -> to_genrep (&GPOLY, avar);
		gp2 = Qax -> to_genrep (&FPOLY, xvar, avar);
		gp3 = Qax -> factors_to_genrep (factors, xvar, avar);

		printf ("\n");
		calib_genrep_prettyprint (gp2);
		printf (" =\n");
		calib_genrep_prettyprint (gp3);
		printf (" mod\n");
		calib_genrep_prettyprint (gp1);
		printf ("\n");

		calib_genrep_free (gp3);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);

		Qax -> free_factors (factors);
		Qax -> clear (&FPOLY);
		calib_free_Qax_dom (Qax);
		calib_free_Qa_dom (Qa);
		free (xvar);
		Zx -> clear (&GPOLY);
		free (avar);

		T1 = calib_get_cpu_time ();
		calib_convert_cpu_time (T1 - T0, tbuf);
		printf ("Bit size:\t%lu\n", nbits);
		printf ("CPU seconds:\t%s\n", tbuf);
	}
	exit (0);
}

/*
 * Print usage information and exit.
 */

	static
	void
usage (void)

{
	fprintf (stderr,
		 "Usage: %s [-cil]\n"
		 "\n"
		 "where:\n"
		 "\t-c\tPrint trial factor combinations\n"
		 "\t-i\tPrint initial (mod p) factors\n"
		 "\t-l\tPrint lifted (mod p^k) factors\n",
		 me);
	exit (1);
}
