/**********************************************************************

	$Id: test_zpx_resultant.c,v 1.1 2024/10/20 01:06:42 warme Exp $

	File:	test_zpx_resultant.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Test polynomial resultant over Zp[x].

***********************************************************************

	Modification Log:

	a-1:	01/30/2024	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/prompt.h"
#include "calib/Zpx.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	LPAREN	"("
#define	RPAREN	")"

static struct calib_genrep *
		add_two_genreps (const struct calib_genrep *	gp1,
				 const struct calib_genrep *	gp2);

	int
main ()

{
int				status;
int				modulus;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
char *				var;
struct calib_Zp_dom *		Zp;
struct calib_Zpx_dom *		K_of_x;
struct calib_Zpx_obj		POLY1;
struct calib_Zpx_obj		POLY2;
struct calib_genrep_varlist	varlist;
mpz_t				res;

	mpz_init (res);

	for (;;) {
		calib_prompt ("Modulus: ");
		(void) scanf (" %d", &modulus);
		if (modulus < 2) break;

		/* Construct our coefficient number field. */
		Zp = calib_make_Zp_dom_ui (modulus);

		/* Construct polynomial ring. */
		K_of_x = calib_make_Zpx_dom (Zp);

		calib_prompt ("Polynomial 1: ");
		gp1 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			calib_free_Zpx_dom (K_of_x);
			calib_free_Zp_dom (Zp);
			continue;
		}
		if (gp1 EQ NULL) break;

		calib_prompt ("Polynomial 2: ");
		gp2 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			calib_genrep_free (gp1);
			calib_free_Zpx_dom (K_of_x);
			calib_free_Zp_dom (Zp);
			continue;
		}
		if (gp2 EQ NULL) break;

		/* Construct an unreduced gp1+gp2 expression (only used	*/
		/* to compute the combined varlist).			*/
		gp3 = add_two_genreps (gp1, gp2);

		/* Verify at most one variable. */
		calib_genrep_get_varlist (&varlist, gp3);
		calib_genrep_free (gp3);

		var = NULL;
		if (varlist.nvars EQ 1) {
			var = strdup (varlist.vlist [0]);
		}
		else if (varlist.nvars EQ 0) {
			var = strdup ("x");
		}
		calib_genrep_clear_varlist (&varlist);

		if (var EQ NULL) {
			fprintf (stderr,
				 "Expression must use exactly"
				 " one variable.\n");
			calib_genrep_free (gp1);
			calib_genrep_free (gp2);
			calib_free_Zpx_dom (K_of_x);
			calib_free_Zp_dom (Zp);
			continue;
		}

		K_of_x -> init (K_of_x, &POLY1);
		K_of_x -> init (K_of_x, &POLY2);

		K_of_x -> set_genrep (&POLY1, gp1, var);
		K_of_x -> set_genrep (&POLY2, gp2, var);

		/* Take canonical forms back into general representation. */
		calib_genrep_free (gp1);
		gp1 = K_of_x -> to_genrep (&POLY1, var);

		calib_genrep_free (gp2);
		gp2 = K_of_x -> to_genrep (&POLY2, var);

		/* Compute resultant. */

		K_of_x -> resultant (res, &POLY1, &POLY2);

		gmp_printf ("Mod %Zd resultant of:\n", Zp -> p);
		calib_genrep_print (gp1);
		printf ("and\n");
		calib_genrep_print (gp2);
		gmp_printf ("\t= %Zd\n", res);

		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		K_of_x -> clear (&POLY2);
		K_of_x -> clear (&POLY1);
		free (var);
		calib_free_Zpx_dom (K_of_x);
		calib_free_Zp_dom (Zp);
	}

	mpz_clear (res);

	exit (0);
}

/*
 * Return the "unreduced" sum of two given genreps.
 */

	static
	struct calib_genrep *
add_two_genreps (

const struct calib_genrep *	gp1,	/* IN: first operand */
const struct calib_genrep *	gp2	/* IN: second operand */
)
{
struct calib_genrep_list *	lp1;
struct calib_genrep_list *	lp2;
struct calib_genrep *		res;

	lp1 = CALIB_NEW (struct calib_genrep_list);
	lp2 = CALIB_NEW (struct calib_genrep_list);
	res = CALIB_NEW (struct calib_genrep);

	lp1 -> operand	= calib_genrep_dup (gp1);
	lp1 -> next	= lp2;

	lp2 -> operand	= calib_genrep_dup (gp2);
	lp2 -> next	= NULL;

	res -> op	= CALIB_GENREP_OP_ADD;
	res -> u.list	= lp1;

	return (res);
}
