/**********************************************************************

	$Id: test_zx_factor.c,v 1.1 2024/10/20 01:21:17 warme Exp $

	File:	test_zx_factor.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Test implementation of Z[x].

***********************************************************************

	Modification Log:

	a-1:	05/04/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/cputime.h"
#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/prompt.h"
#include "calib/Zpx.h"
#include "calib/Zx.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Local Functions
 */

static void	decode_args (int argc, char ** argv);
static void	usage (void);

/*
 * Local Data.
 */

static char *	me;

	int
main (

int		argc,
char **		argv
)
{
int				status;
calib_cpu_time_t		t0, t1;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
char *				var;
struct calib_Zx_factor *	factors;
const struct calib_Zx_dom *	Z_of_x;
struct calib_Zx_obj		POLY;
struct calib_genrep_varlist	varlist;
char				buf [32];

	me = argv [0];

	decode_args (argc, argv);

	/* Get the polynomial ring Z[x]. */
	Z_of_x = calib_get_Zx_dom ();

	while (1) {
		calib_prompt ("Polynomial: ");
		gp1 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			continue;
		}
		if (gp1 EQ NULL) break;

		/* Verify exactly one variable. */
		calib_genrep_get_varlist (&varlist, gp1);

		var = NULL;
		if (varlist.nvars EQ 1) {
			var = strdup (varlist.vlist [0]);
		}
		else if (varlist.nvars EQ 0) {
			var = strdup ("x");
		}
		calib_genrep_clear_varlist (&varlist);

		if (var EQ NULL) {
			fprintf (stderr,
				 "Expression must use exactly"
				 " one variable.\n");
			calib_genrep_free (gp1);
			continue;
		}

		Z_of_x -> init (&POLY);
		Z_of_x -> set_genrep (&POLY, gp1, var);
		calib_genrep_free (gp1);
		if (POLY.degree < 0) {
			Z_of_x -> clear (&POLY);
			free (var);
			break;
		}

		/* Take canonical form back into general representation. */
		gp1 = Z_of_x -> to_genrep (&POLY, var);

		    t0 = calib_get_cpu_time ();
		factors = Z_of_x -> factor (&POLY);
		    t1 = calib_get_cpu_time ();

		gp2 = Z_of_x -> factors_to_genrep (factors, var);

		printf ("\n");
		calib_genrep_prettyprint (gp1);
		printf (" =\n");
		calib_genrep_prettyprint (gp2);
		printf ("\n");
		calib_convert_cpu_time (t1 - t0, buf);
		printf ("CPU Time: %s\n", buf);

		calib_genrep_free (gp2);
		Z_of_x -> free_factors (factors);
		calib_genrep_free (gp1);
		Z_of_x -> clear (&POLY);
		free (var);
	}
	exit (0);
}

/*
 * Decode command line arguments.
 */

	static
	void
decode_args (

int		argc,
char **		argv
)
{
int		i, n;
char		c;
const char *	ap;

	for (i = 1; i < argc; i++) {
		ap = argv [i];
		if (*ap++ NE '-') {
			usage ();
		}
		for (;;) {
			c = *ap++;
			if (c EQ '\0') break;

			switch (c) {
			case 'b':
				calib_Zpx_default_settings.factor_method =
					CALIB_ZPX_FACTOR_METHOD_BERLEKAMP;
				break;

			case 'd':
				calib_Zpx_default_settings.factor_method =
					CALIB_ZPX_FACTOR_METHOD_DDF;
				break;

			case 'e':
				if (*ap EQ '\0') {
					++i;
					if (i >= argc) {
						usage ();
					}
					ap = argv [i];
				}
				n = atoi (ap);
				if (n < 3) {
					usage ();
				}
				calib_Zx_settings.VH_max_enumerate = n;
				ap = "";
				break;

			case 'l':
				calib_Zx_settings.lift_method = CALIB_ZX_LIFT_METHOD_LINEAR;
				break;

			case 'q':
				calib_Zx_settings.lift_method = CALIB_ZX_LIFT_METHOD_QUADRATIC;
				break;

			case 'r':
				calib_Zx_settings.VH_use_resultant_trace_root_bound = TRUE;
				break;

			case 'v':
				calib_Zx_settings.validate_lifted_modular_factors = TRUE;
				break;

			case 'V':
				calib_Zx_settings.factor_method = CALIB_ZX_FACTOR_METHOD_VAN_HOEIJ;
				break;

			case 'Z':
				calib_Zx_settings.factor_method = CALIB_ZX_FACTOR_METHOD_ZASSENHAUS;
				break;

			default:
				usage ();
				break;
			}
		}
	}
}

/*
 * Display command usage and exit.
 */

	static
	void
usage (void)

{
	fprintf (stderr,
		 "Usage: %s [-]\n"
		 "\t-e N\tVan Hoeij enumerates at most N>=3 trial factors.\n"
		 "\t-l\tUse linear lifting of modular factors.\n"
		 "\t-q\tUse quadratic lifting of modular factors.\n"
		 "\t-r\tUse resultant-based trace root bound in Van Hoeij.\n"
		 "\t-v\tValidate lifted modular factors.\n"
		 "\t-V\tUse Van Hoeij algorithm.\n"
		 "\t-Z\tUse Zassenhaus algorithm (exhaustive enumeration).\n",
		 me);

	exit (1);
}
