/**********************************************************************

	$Id: test_zx_resultant.c,v 1.1 2024/10/20 01:26:39 warme Exp $

	File:	test_zx_resultant.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Test Z[x] resultant().

***********************************************************************

	Modification Log:

	a-1:	05/04/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/prompt.h"
#include "calib/Zx.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

	int
main ()

{
int				status;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
char *				var;
const struct calib_Zx_dom *	Z_of_x;
struct calib_Zx_obj		POLY1;
struct calib_Zx_obj		POLY2;
struct calib_genrep_varlist	varlist;
mpz_t				RVAL;

	/* Get the polynomial ring Z[x]. */
	Z_of_x = calib_get_Zx_dom ();

	Z_of_x -> init (&POLY1);
	Z_of_x -> init (&POLY2);
	mpz_init (RVAL);

	while (1) {
		calib_prompt ("Polynomial 1: ");
		gp1 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			continue;
		}
		if (gp1 EQ NULL) break;

		/* Verify exactly one variable. */
		calib_genrep_get_varlist (&varlist, gp1);

		var = NULL;
		if (varlist.nvars EQ 1) {
			var = strdup (varlist.vlist [0]);
		}
		else if (varlist.nvars EQ 0) {
			var = strdup ("x");
		}
		calib_genrep_clear_varlist (&varlist);

		if (var EQ NULL) {
			fprintf (stderr,
				 "Expression must use exactly"
				 " one variable.\n");
			calib_genrep_free (gp1);
			continue;
		}

		Z_of_x -> set_genrep (&POLY1, gp1, var);
		calib_genrep_free (gp1);
		if ((POLY1.degree EQ 0) AND
		    (mpz_cmp_ui (&(POLY1.coeff [0]), 666) EQ 0)) {
			/* Special exit value. */
			free (var);
			break;
		}

		/* Take canonical form back into general representation. */
		gp1 = Z_of_x -> to_genrep (&POLY1, var);

		calib_prompt ("Polynomial 2: ");
		gp2 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			calib_genrep_free (gp1);
			free (var);
			continue;
		}
		if (gp2 EQ NULL) {
			calib_genrep_free (gp1);
			free (var);
			break;
		}

		/* Verify at most one variable, which must be the same. */
		calib_genrep_get_varlist (&varlist, gp2);
		if ((varlist.nvars > 1) OR
		    ((varlist.nvars EQ 1) AND
		     (strcmp (varlist.vlist [0], var) NE 0))) {
			fprintf (stderr, "Must be a polynomial in %s.\n", var);
			calib_genrep_clear_varlist (&varlist);
			calib_genrep_free (gp2);
			calib_genrep_free (gp1);
			free (var);
			continue;
		}
		calib_genrep_clear_varlist (&varlist);

		Z_of_x -> set_genrep (&POLY2, gp2, var);
		calib_genrep_free (gp2);
		gp2 = Z_of_x -> to_genrep (&POLY2, var);

		Z_of_x -> resultant (RVAL, &POLY1, &POLY2);

		gp3 = calib_genrep_z (RVAL);

		printf ("\n");
		printf ("Resultant of\n");
		calib_genrep_print (gp1);
		printf (" and\n");
		calib_genrep_print (gp2);
		printf (" is\n");
		calib_genrep_print (gp3);

		calib_genrep_free (gp3);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);
		free (var);
	}

	mpz_clear (RVAL);
	Z_of_x -> clear (&POLY2);
	Z_of_x -> clear (&POLY1);

	exit (0);
}
