/**********************************************************************

	$Id: test_zxyz_discriminant.c,v 1.1 2024/10/20 01:37:11 warme Exp $

	File:	test_zxyz_discriminant.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Test implementation of resultants in Z[x,y,z].

***********************************************************************

	Modification Log:

	a-1:	02/01/2024	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/cputime.h"
#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/prompt.h"
#include "calib/shutdown.h"
#include "calib/Zxyz.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


	int
main ()

{
int				status, var;
calib_cpu_time_t		T0, T1;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_Zxyz_factor *	factors;
struct calib_Zxyz_obj		POLY, DISC;
char				tbuf [32];
struct calib_genrep_varlist	varlist;

	while (1) {
		calib_prompt ("Polynomial: ");
		gp1 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			continue;
		}
		if (gp1 EQ NULL) break;

		calib_prompt ("Variable: ");
		gp2 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			continue;
		}
		if (gp2 EQ NULL) break;

		if (gp2 -> op NE CALIB_GENREP_OP_VAR) {
			fprintf (stderr, "Variable expected\n");
			continue;
		}

		calib_genrep_get_varlist (&varlist, gp1);

		/* Make the multi-variate domain. */
		Zxyz = calib_make_Zxyz_dom (varlist.nvars, varlist.vlist);

		calib_genrep_clear_varlist (&varlist);

		var = Zxyz -> lookup_var (Zxyz, gp2 -> u.var);
		if (var < 0) {
			printf ("Variable `%s' does not appear"
				" in given polynomial.\n",
				gp2 -> u.var);
			calib_genrep_free (gp2);
			calib_genrep_free (gp1);
			continue;
		}

		Zxyz -> init (Zxyz, &POLY);
		Zxyz -> init (Zxyz, &DISC);

		Zxyz -> set_genrep (&POLY, gp1);

		calib_genrep_free (gp1);

		/* Take canonical form back into general rep. */
		gp1 = Zxyz -> to_genrep (&POLY);

		printf ("\n");
		printf ("Discriminant of:\n");
		calib_genrep_prettyprint (gp1);
		printf ("  with respect to variable:\n");
		calib_genrep_prettyprint (gp2);
		printf ("  =\n");

		calib_genrep_free (gp2);

		T0 = calib_get_cpu_time ();
		Zxyz -> discriminant (&DISC, &POLY, var);
		T1 = calib_get_cpu_time ();

		calib_convert_cpu_time (T1 - T0, tbuf);

		gp3 = Zxyz -> to_genrep (&DISC);
		calib_genrep_prettyprint (gp3);
		printf ("  %s seconds\n", tbuf);
		calib_genrep_free (gp3);

		/* Now factor the discriminant. */
		T0 = calib_get_cpu_time ();
		factors = Zxyz -> factor (&DISC);
		T1 = calib_get_cpu_time ();

		calib_convert_cpu_time (T1 - T0, tbuf);

		gp3 = Zxyz -> factors_to_genrep (factors);
		Zxyz -> free_factors (factors);

		printf ("  =\n");
		calib_genrep_prettyprint (gp3);
		printf ("  %s seconds\n", tbuf);
		calib_genrep_free (gp3);

		Zxyz -> clear (&DISC);
		Zxyz -> clear (&POLY);

		calib_genrep_free (gp1);

		calib_free_Zxyz_dom (Zxyz);
	}
	calib_shutdown ();
	exit (0);
}
