/**********************************************************************

	$Id: test_zxyz_resultant.c,v 1.1 2024/10/20 01:37:11 warme Exp $

	File:	test_zxyz_resultant.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Test implementation of resultants in Z[x,y,z].

***********************************************************************

	Modification Log:

	a-1:	02/01/2024	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/cputime.h"
#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/prompt.h"
#include "calib/shutdown.h"
#include "calib/Zxyz.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static struct calib_genrep *
		merge_3_genreps (const struct calib_genrep * gp1,
				 const struct calib_genrep * gp2,
				 const struct calib_genrep * gp3);

	int
main ()

{
int				i, status, var;
calib_cpu_time_t		T0, T1;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
struct calib_genrep *		gp4;
const struct calib_Zxyz_dom *	Zxyz;
struct calib_Zxyz_factor *	factors;
struct calib_Zxyz_obj		POLY1, POLY2, RES;
char				tbuf [32];
struct calib_genrep_varlist	varlist;

	while (1) {
		calib_prompt ("Polynomial 1: ");
		gp1 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			continue;
		}
		if (gp1 EQ NULL) break;

		calib_prompt ("Polynomial 2: ");
		gp2 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			continue;
		}
		if (gp2 EQ NULL) break;

		calib_prompt ("Variable: ");
		gp3 = calib_genrep_read (&status);
		if (status NE 0) {
			fprintf (stderr, "Syntax error\n");
			continue;
		}
		if (gp3 EQ NULL) break;

		if (gp3 -> op NE CALIB_GENREP_OP_VAR) {
			fprintf (stderr, "Variable expected\n");
			continue;
		}

		gp4 = merge_3_genreps (gp1, gp2, gp3);
		calib_genrep_get_varlist (&varlist, gp4);
		calib_genrep_free (gp4);

		var = -1;
		for (i = 0; i < varlist.nvars; i++) {
			if (strcmp (gp3 -> u.var, varlist.vlist [i]) EQ 0) {
				var = i;
				break;
			}
		}
		if (var < 0) {
			fprintf (stderr, "Invalid variable\n");
			continue;
		}

		/* Make the multi-variate domain. */
		Zxyz = calib_make_Zxyz_dom (varlist.nvars, varlist.vlist);

		calib_genrep_clear_varlist (&varlist);

		Zxyz -> init (Zxyz, &POLY1);
		Zxyz -> init (Zxyz, &POLY2);
		Zxyz -> init (Zxyz, &RES);

		Zxyz -> set_genrep (&POLY1, gp1);
		Zxyz -> set_genrep (&POLY2, gp2);

		calib_genrep_free (gp2);
		calib_genrep_free (gp1);

		/* Take canonical forms back into general rep. */
		gp1 = Zxyz -> to_genrep (&POLY1);
		gp2 = Zxyz -> to_genrep (&POLY2);

		printf ("\n");
		printf ("Resultant of:\n");
		calib_genrep_prettyprint (gp1);
		calib_genrep_prettyprint (gp2);
		printf ("  with respect to variable:\n");
		calib_genrep_prettyprint (gp3);
		printf ("  =\n");

		T0 = calib_get_cpu_time ();
		Zxyz -> resultant (&RES, &POLY1, &POLY2, var);
		T1 = calib_get_cpu_time ();

		calib_convert_cpu_time (T1 - T0, tbuf);

		gp4 = Zxyz -> to_genrep (&RES);
		calib_genrep_prettyprint (gp4);
		printf ("  %s seconds\n", tbuf);
		calib_genrep_free (gp4);

		/* Now factor the resultant. */
		T0 = calib_get_cpu_time ();
		factors = Zxyz -> factor (&RES);
		T1 = calib_get_cpu_time ();

		calib_convert_cpu_time (T1 - T0, tbuf);

		gp4 = Zxyz -> factors_to_genrep (factors);
		Zxyz -> free_factors (factors);

		printf ("  =\n");
		calib_genrep_prettyprint (gp4);
		printf ("  %s seconds\n", tbuf);
		calib_genrep_free (gp4);

		Zxyz -> clear (&RES);
		Zxyz -> clear (&POLY2);
		Zxyz -> clear (&POLY1);

		calib_genrep_free (gp3);
		calib_genrep_free (gp2);
		calib_genrep_free (gp1);

		calib_free_Zxyz_dom (Zxyz);
	}
	calib_shutdown ();
	exit (0);
}

/*
 * Merge the three given genreps by creating an *unreduced* sum.
 * (Only used to construct common varlist.)
 */

	static
	struct calib_genrep *
merge_3_genreps (

const struct calib_genrep *	gp1,	/* IN: First genrep */
const struct calib_genrep *	gp2,	/* IN: Second genrep */
const struct calib_genrep *	gp3	/* IN: Third genrep */
)
{
struct calib_genrep_list *	lp1;
struct calib_genrep_list *	lp2;
struct calib_genrep_list *	lp3;
struct calib_genrep *		gp4;

	lp1	= CALIB_NEW (struct calib_genrep_list);
	lp2	= CALIB_NEW (struct calib_genrep_list);
	lp3	= CALIB_NEW (struct calib_genrep_list);
	gp4	= CALIB_NEW (struct calib_genrep);

	lp1 -> operand	= calib_genrep_dup (gp1);
	lp1 -> next	= lp2;

	lp2 -> operand	= calib_genrep_dup (gp2);
	lp2 -> next	= lp3;

	lp3 -> operand	= calib_genrep_dup (gp3);
	lp3 -> next	= NULL;

	gp4 -> op	= CALIB_GENREP_OP_ADD;
	gp4 -> u.list	= lp1;

	return (gp4);
}
