/**********************************************************************

	$Id: test_func_poly.c,v 1.1 2024/10/20 00:43:01 warme Exp $

	File:	test_func_poly.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	Test genrep printing of (log(x) + 1)^n expanded into a
	polynomial in log(x).

***********************************************************************

	Modification Log:

	a-1:	08/06/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/logic.h"
#include "calib/prompt.h"
#include "calib/Zx.h"
#include "calib/Zxyz.h"
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * Local Functions
 */

static struct calib_genrep *
		substitute_var (const struct calib_genrep *	op1,
				const char *			vname,
				const struct calib_genrep *	op2);

	int
main (

int		argc,
char **		argv
)
{
int				n;
calib_bool			old_printer;
char *				me;
char *				ap;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;
struct calib_genrep *		gp3;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_obj		POLY;

	me = argv [0];

	old_printer = FALSE;

	ap = NULL;
	if (argc >= 2) {
		ap = argv [1];
		if ((ap [0] EQ '-') AND (ap [1] EQ 'o')) {
			old_printer = TRUE;
			ap = NULL;
			if (argc >= 3) {
				ap = argv [2];
			}
		}
	}
	n = 24;
	if (ap NE NULL) {
		n = atoi (ap);
		if (n < 0) {
			fprintf (stderr,
				 "%s: Negative power not permitted.\n",
				 me);
			exit (1);
		}
	}

	/* Expand (x + 1)^n in Z[x]. */
	Zx = calib_get_Zx_dom ();
	Zx -> init (&POLY);
	Zx -> set_var_power (&POLY, 1);
	mpz_set_ui (&POLY.coeff [0], 1);
	Zx -> ipow (&POLY, &POLY, n);

	/* Convert to genrep. */
	gp1 = Zx -> to_genrep (&POLY, "x");

	/* Get genrep for "log(x)". */
	gp2 = calib_genrep_var ("x");
	gp2 = calib_genrep_log (gp2);

	/* Recursively replace all x with log(x). */
	gp3 = substitute_var (gp2, "x", gp1);

	printf ("\n");
	if (old_printer) {
		calib_genrep_print (gp3);
	}
	else {
		calib_genrep_prettyprint (gp3);
	}
	printf ("\n");

	calib_genrep_free (gp3);
	calib_genrep_free (gp2);
	calib_genrep_free (gp1);

	Zx -> clear (&POLY);

	exit (0);
}

/*
 * Recursively substitute op1 for variable vname within op2.
 */

	static
	struct calib_genrep *
substitute_var (

const struct calib_genrep *	op1,	/* IN: Replacement expression */
const char *			vname,	/* IN: Variable to replace */
const struct calib_genrep *	op2	/* IN: Expression in which to subst */
)
{
struct calib_genrep *		gp1;
struct calib_genrep_list *	lp1;
struct calib_genrep_list *	lp2;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;

	CALIB_FATAL_ERROR_IF ((op1   EQ NULL) OR
			      (vname EQ NULL) OR
			      (op2   EQ NULL));

	switch (op2 -> op) {
	case CALIB_GENREP_OP_ADD:
	case CALIB_GENREP_OP_MUL:
		root	= NULL;
		hookp	= &root;
		for (lp1 = op2 -> u.list; lp1 NE NULL; lp1 = lp1 -> next) {
			gp1 = substitute_var (op1, vname, lp1 -> operand);
			lp2 = calib_genrep_new_list (gp1, NULL);
			*hookp = lp2;
			hookp = &(lp2 -> next);
		}
		gp1 = calib_genrep_new ();
		gp1 -> op	= op2 -> op;
		gp1 -> u.list	= root;
		break;

	case CALIB_GENREP_OP_IPOW:
		gp1 = substitute_var (op1, vname, op2 -> u.ipow.base);
		gp1 = calib_genrep_ipow (gp1, op2 -> u.ipow.power);
		break;

	case CALIB_GENREP_OP_VAR:
		if (strcmp (op2 -> u.var, vname) EQ 0) {
			gp1 = calib_genrep_dup (op1);
		}
		else {
			gp1 = calib_genrep_dup (op2);
		}
		break;

	case CALIB_GENREP_OP_SI:
	case CALIB_GENREP_OP_Z:
	case CALIB_GENREP_OP_Q:
		gp1 = calib_genrep_dup (op2);
		break;

	case CALIB_GENREP_OP_FUNC:
		root	= NULL;
		hookp	= &root;
		for (lp1 = op2 -> u.list; lp1 NE NULL; lp1 = lp1 -> next) {
			gp1 = substitute_var (op1, vname, lp1 -> operand);
			lp2 = calib_genrep_new_list (gp1, NULL);
			*hookp = lp2;
			hookp = &(lp2 -> next);
		}
		gp1 = calib_genrep_new ();
		gp1 -> op	= op2 -> op;
		gp1 -> u.list	= root;
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	return (gp1);
}
