/**********************************************************************

	$Id: Qax.c,v 1.1 2024/10/20 00:36:25 warme Exp $

	File:	Qax.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	A ring of polynomials K[x] over the algebraic
	extension field K = Q(a).

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Qax.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/GFpk.h"
#include "calib/GFpkx.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/Qa.h"
#include "calib/Qx.h"
#include "calib/random.h"
#include "calib/types.h"
#include "calib/Za.h"
#include "calib/Zax.h"
#include "calib/Zp.h"
#include "calib/Zpx.h"
#include "calib/Zx.h"
#include "calib/Zxyz.h"
#include <gmp.h>
#include "Qax_int.h"
#include <stdlib.h>
#include <string.h>


/*
 * Default settings object for newly-created Qax domains.
 *
 * API permits these to be changed, which affects future creation of
 * new Qax domains.
 */

struct calib_Qax_settings	calib_Qax_default_settings = {

	/* Default to "norm" factoring method. */
	.factor_method		= CALIB_QAX_FACTOR_METHOD_NORM,

	/* Default is to NOT print various Weinberger-Rothschild stuff. */
	.WR_print_initial_factors		= FALSE,
	.WR_print_lifted_factors		= FALSE,
	.WR_print_trial_factor_combinations	= FALSE,
};

/*
 * Global info for the Q(a)[x] algebraic factorization routine.
 */

struct afglob {
	/* The polynomial ring that the polynomial being factored	*/
	/* is a member of.						*/
	const struct calib_Qax_dom *
				K_of_x;

	/* The original polynomial f(x) being factored.  It is already	*/
	/* known to be square free.					*/
	const struct calib_Qax_obj *
				fpoly;

	/* The original polynomial F(a) that defines the algebraic	*/
	/* number field that contains the coefficients of fpoly.	*/
	const struct calib_Zx_obj *
				apoly;

	/* A locally-modifiable copy of f(x).  This is reduced as true	*/
	/* factors are discovered.					*/
	struct calib_Qax_obj	lpoly;

	/* The list of true factors discovered so far, and the "hookp"	*/
	/* for appending onto the end of this list.			*/
	struct calib_Qax_factor *
				true_factors;
	struct calib_Qax_factor **
				true_factor_hookp;

	/* The smallest positive integer for which all the rational	*/
	/* number coefficients in b*fpoly become integral.		*/
	mpz_t			b;

	/* The |discriminant| of apoly F(a). */
	mpz_t			disc;

	/* An upper bound on the magnitude of any denominator appearing	*/
	/* in any true factor of f(x).					*/
	mpz_t			d;

	/* The domains Z(a), and Z(a)[x] corrsponding to K_of_x.	*/
	struct calib_Za_dom *	Za;
	struct calib_Zax_dom *	Zax;

	/* This is just d*f(x) stored in Z(a)[x] form. */
	struct calib_Zax_obj *	zpoly;

	/* An upper bound on the magnitude of any coefficient of any	*/
	/* true factor of d*f(x) in Z(a)[x].  Our final working modulus	*/
	/* must be at least twice this in order to properly resolve the	*/
	/* sign of each coefficient.					*/
	mpz_t			B;

	/* The prime selected to use as the initial modulus. */
	mpz_t			p;

	/* The current modulus.  This is some p^k. */
	mpz_t			modulus;

	/* This is floor(modulus/2), used for detecting negative	*/
	/* numbers in their modular form.				*/
	mpz_t			half_modulus;

	/* This is the inverse of d in our modulus, i.e.,		*/
	/* d*dinv == 1 (mod modulus).					*/
	mpz_t			dinv;

	/* Counter used to generate successive prime numbers. */
	mpz_t			prime_counter;

	/* The number of "modular" factors for apoly F(a). */
	int			nmf;

	/* The Zp and Zp[x] domains, and the apoly F(a) in Zp[x] form.	*/
	struct calib_Zp_dom *	Zp;
	struct calib_Zpx_dom *	Zpx;
	struct calib_Zpx_obj *	apoly_zpx;

	/* The apoly is factored (mod p) as:				*/
	/*								*/
	/*	F(a) == F_0(a) * F_1(a) * ... * F_{nmf-1}(a)	(mod p)	*/
	/*								*/
	/* and we insist that p be chosen so that this factorization is	*/
	/* square-free.  This item points to an array [nmf] of		*/
	/* structures containing information associated with each of	*/
	/* these factors.						*/
	struct gfinfo *		gfinfo;

	/* Degree of trial factors currently being generated. */
	int			gen_degree;

	/* A collection of vars that are used in the innermost loops of	*/
	/* Qax_CRT_one_coefficient that are pre-allocated and ready to	*/
	/* go so that we don't have to allocate/free them each time	*/
	/* this routine executes.					*/
	struct calib_Zax_obj *	crt_factor;
	struct calib_Zx_obj *	crt_SUM;
	struct calib_Zx_obj *	crt_P2;
	struct calib_Zx_obj *	crt_P3;
	struct calib_Zx_obj *	crt_Q;
	struct calib_Zx_obj *	crt_R;
	mpz_t			crt_denom;
};

/*
 * The following structure holds all of the information we need about
 * a single candidate prime p, including the complete factorization of
 * our apoly F(a) over Zp[x].
 */

struct pfact {
	int			nmf;	/* Number of apoly factors */
	mpz_t			p;	/* The prime p used */
	struct calib_Zp_dom *	Zp;	/* The Zp field */
	struct calib_Zpx_dom *	Zpx;	/* The Zp[x] polynomial ring */
	struct calib_Zpx_obj *	apoly;	/* The apoly F(a) in Zp[x] form */
	struct calib_Zpx_factor *
				factors;
					/* Factors of F(a) over Zp[x] */
	struct gfinfo *		gfinfo;	/* Factorization of f(x) in */
					/* GF(p^k)[x] for each apoly factor */
	struct pfact *		next;	/* Next candidate prime */
};

/*
 * The algebraic number field is Q(a), where a is a root of a monic,
 * irreducible polynomial F(a) \in Z[x].  We call F(a) the "apoly",
 * since it defines "a", our algebraic extension.
 *
 * We also have a polynomial f(x) \in Q(a)[x], which we call the "fpoly",
 * which we know to be square-free, and it is our goal to factor it into
 * irreducible factors over Q(a)[x].
 *
 * We proceed by factoring F(a) into irreducible factors modulo some prime p,
 * chosen so that the factorization is square-free.  For each of the apoly
 * factors F_k(a) (an irreducible in Zp[x]), we define a Galois field
 * GF(p^k), and factor f(x) in this GF(p^k)[x] domain.  We then lift all
 * of these factorizations [both the F(a) and f(x)] to a suitably high
 * modulus using the Hensel lemma.  Trial factors in Q(a)[x] are then
 * obtained by combining corresponding f(x) factors modulo each F_k(a)
 * via the Chinese remainder theorem.
 *
 * The name "gfinfo" derives from the "Galois Field" GF(p^k)[x] polynomials
 * we use to initially factor f(x), one for each apoly factor.
 *
 * The following structure records all of the info we need regarding
 * a single such apoly factor F_k(a).
 */

struct gfinfo {
    /* Info regarding apoly F_k(a): */
	/* This apoly factor F_k(a) -- the original in Zp[x] form. */
	const struct calib_Zpx_dom *
				Zpx;		/* The polynomial ring */
	struct calib_Zpx_obj *	afactor_zpx;	/* The polynomial */

	/* This apoly factor -- lifted version in Z[x] form. */
	struct calib_Zx_obj *	afactor_zx;

    /* Info regarding the factorization of f(x) modulo this F_k(a): */
	/* The original f(x), in K[x] form, where K = GF(p^k). */
	struct calib_GFpk_dom *	gfpk;		/* coefficient field K */
	struct calib_GFpkx_dom *
				gfpkx;		/* polynomial ring K[x] */
	struct calib_GFpkx_obj *
				fpoly_gfpkx;	/* f(x) in K[x] form */

	/* The original f(x) mod F_k(a), in Z(a)[x] form.  (This is the	*/
	/* original (mod p) version of F_k(a) here -- not a lifted	*/
	/* version (mod p^j, where j > 1).)				*/
	struct calib_Zax_obj *	fpoly_zax;

	/* Initial factorization of f(x) modulo this F_k(a), in GF(p^k) */
	/* form. */
	int			nfact;	 /* number of factors */
	struct calib_GFpkx_obj **
				factors; /* f(x) factors in this GF(p^k)[x] */

	/* Lifted factor of f(x) modulo this (lifted) F_k(a), and also	*/
	/* mod p^j.  We increase the power j until the modulus p^j is	*/
	/* sufficiently high to guarantee recovery of all true factors.	*/
	struct calib_Za_dom *	Za;	/* lifted Z(a) ring for lfactors */
	struct calib_Zax_dom *	Zax;	/* lifted Z(a)[x] ring for lfactors */
	struct calib_Zax_obj **	lfactors; /* lifted f(x) factors */

	/* Multiplier for  this apoly factor when using the Chinese	*/
	/* remainder theorem to combine the various "mod this apoly	*/
	/* factor" trial factors into a single trial factor that is	*/
	/* simultaneously correct modulo each of these apoly factors.	*/
	struct calib_Zx_obj *	vfactor;

	/* Product of other apoly factors, used during Chinese		*/
	/* remainder theorem to reconstruct trial factors.		*/
	struct calib_Zx_obj *	wfactor;

	/* Array of flags indicating which of our lfactors[] have been	*/
	/* included in our current trial factor.			*/
	calib_bool *		lf_try;

	/* Array of flags indicating which of our lfactors[] have been	*/
	/* used/expended already as part of some "true" factor of f(x).	*/
	calib_bool *		lf_used;

	/* Base of the stack used to recursively accumulate the product	*/
	/* of a chosen subset of the factors into a trial factor.  The	*/
	/* first element of this stack contains the polynomial "1", and	*/
	/* represents the "empty" product.  Each time we include a new	*/
	/* factor polynomial into the trial factor, we "push" the	*/
	/* new product, i.e., stack [i] = factor * stack [i-1]; ++i;	*/
	/* Done correctly, this never destroys the "empty" products,	*/
	/* and we never have to allocate or free any of the polnomials	*/
	/* during recursion.						*/
	struct calib_Zax_obj *	prod_stack_base;

	/* A pointer to the final, recursively generated trial factor	*/
	/* (for this apoly factor) is stored here.			*/
	struct calib_Zax_obj *	trial_factor;
};

/*
 * The following structure represents the Sturm sequence of a polynomial
 * in Z[x].
 */

struct sturm {
	int		n;		/* Number of entries in sequence */
	int		vsize;		/* Size of integer coeff vector */
	struct calib_Zx_obj *
			poly;		/* Polynomials in sequence */
	mpz_ptr		coeff;		/* Vector of coefficients */
};

/*
 * An object to represent a polynomial in Q(a)[x] as the product
 *
 *	s * poly(x)
 *
 * where s \in Q, poly in Z(a)[x] is primitive with respect
 * to Z (not necessarily primitive over Z(a)).
 */

struct QZaPoly {
	const struct calib_Zax_dom *
				qzap_dom;	/* Polynomial ring Z(a)[x] */
	mpq_t			qzap_s;		/* The rational constant s */
	struct calib_Zax_obj	qzap_poly;	/* Polynomial \in Z(a)[x], */
						/* primitive over Z */
};

/*
 * Local Functions
 */

static void	Qax_add (struct calib_Qax_obj *		rop,
			 const struct calib_Qax_obj *	op1,
			 const struct calib_Qax_obj *	op2);
static const struct calib_Qax_dom *
		Qax_algint_dom (const struct calib_Qax_dom * dom);
static void	Qax_alloc (struct calib_Qax_obj *	rop,
			   int				degree);
static size_t	Qax_bit_size (const struct calib_Qax_obj *	op);
static void	Qax_build_gfinfo (struct afglob * afgp, struct pfact * pfact);
static struct calib_Qax_factor *
		Qax_choose_sqf_factor_algorithm (
				const struct calib_Qax_obj *	sqfpoly);
static void	Qax_choose_prime_p (struct afglob * afgp);
static void	Qax_cleanup_afglob (struct afglob * afgp);
static void	Qax_clear (struct calib_Qax_obj *	x);
static void	Qax_clear_array (struct calib_Qax_obj * parray,
				 int			npoly);
static void	Qax_clear_gfinfo (struct gfinfo * gip);
static calib_bool
		Qax_combine_convert_and_test_trial_factors (
						struct afglob * afgp);
static void	Qax_compute_common_denominator (struct afglob * afgp);
static void	Qax_compute_factor_coefficient_bound (struct afglob * afgp);
static struct calib_Qax_factor *
		Qax_compute_square_free_factorization (
				const struct calib_Qax_obj *	poly);
static void	Qax_compute_zpoly (struct afglob * afgp);
static struct calib_GFpkx_obj *
		Qax_convert_to_GFpkx (
				const struct calib_GFpkx_dom *	gfpkx,
				const struct calib_Zax_dom *	R_of_x,
				const struct calib_Zax_obj *	zax_poly,
				mpz_srcptr			d);
static struct calib_Qax_obj *
		Qax_convert_Zax_to_Qax (struct afglob *			afgp,
					const struct calib_Zax_obj *	zpoly);
static void	Qax_copy_Qa_vector (struct calib_Qa_obj *	dst,
				    const struct calib_Qa_obj *	src,
				    int				n);
static calib_bool
		Qax_CRT_one_coefficient (struct afglob * afgp, int j);
static void	Qax_derivative (struct calib_Qax_obj *		rop,
				const struct calib_Qax_obj *	op);
static void	Qax_div (struct calib_Qax_obj *		quotient,
			 struct calib_Qax_obj *		remainder,
			 const struct calib_Qax_obj *	a,
			 const struct calib_Qax_obj *	b);
static struct calib_Qax_obj *
		Qax_dup (const struct calib_Qax_obj *	op);
static void	Qax_eval (struct calib_Qa_obj *		rop,
			  const struct calib_Qax_obj *	poly,
			  const struct calib_Qa_obj *	value);
static void	Qax_eval_poly (struct calib_Qax_obj *		rop,
			       const struct calib_Qax_obj *	poly,
			       const struct calib_Qax_obj *	value);
static void	Qax_extgcd (struct calib_Qax_obj *		gcd,
			    struct calib_Qax_obj *		xa,
			    struct calib_Qax_obj *		xb,
			    const struct calib_Qax_obj *	a,
			    const struct calib_Qax_obj *	b);
static struct calib_Qax_factor *
		Qax_factor (const struct calib_Qax_obj *	poly);
static struct calib_Qax_factor *
		Qax_factor_non_algint (
				const struct calib_Qax_obj *	poly);
static struct calib_Qax_factor *
		Qax_factor_sqf_norm (
				const struct calib_Qax_obj *	fpoly);
static struct calib_Qax_factor *
		Qax_factor_sqf_Weinberger_Rothschild (
				const struct calib_Qax_obj *	fpoly);
static struct calib_Qax_factor *
		Qax_factor_square_free (
				const struct calib_Qax_obj *	sqfpoly);
static size_t	Qax_factors_bit_size (
				const struct calib_Qax_factor *	factors);
static struct calib_genrep *
		Qax_factors_to_genrep (
				const struct calib_Qax_factor *	factors,
				const char *			xvar,
				const char *			avar);
static struct pfact *
		Qax_find_apoly_Zp_factors (struct afglob * afgp, mpz_srcptr p);
static struct calib_Qax_factor *
		Qax_finish_sqf (const struct calib_Qax_factor *	sqfactors);
static void	Qax_free (struct calib_Qax_obj *	poly);
static void	Qax_free_apoly_Zp_factors (struct pfact * list);
static void	Qax_free_CRT_buffers (struct afglob * afgp);
static void	Qax_free_factors (struct calib_Qax_factor * factors);
static void	Qax_free_gfinfo (int nmf, struct gfinfo * gfinfo);
static void	Qax_free_Qa_vector (struct calib_Qa_obj *	array,
				    int				n);
static void	Qax_free_sturm_sequence (struct sturm * ss);
static void	Qax_from_algint (struct calib_Qax_obj *		rop,
				 const struct calib_Qax_obj *	op);
static void	Qax_gcd (struct calib_Qax_obj *		gcd,
			 const struct calib_Qax_obj *	a,
			 const struct calib_Qax_obj *	b);
static const struct calib_Zax_dom *
		Qax_get_Zax_dom (const struct calib_Qax_dom *	K_of_x);
static void	Qax_init (const struct calib_Qax_dom *	K_of_x,
			  struct calib_Qax_obj *	x);
static void	Qax_init_afglob (struct afglob *		afgp,
				 const struct calib_Qax_obj *	fpoly);
static void	Qax_init_array_degree (const struct calib_Qax_dom *	K_of_x,
				       struct calib_Qax_obj *		parray,
				       int				npoly,
				       int				degree);
static void	Qax_init_CRT_buffers (struct afglob * afgp);
static void	Qax_init_degree (const struct calib_Qax_dom *	K_of_x,
				 struct calib_Qax_obj *		x,
				 int				degree);
static void	Qax_ipow (struct calib_Qax_obj *	rop,
			  const struct calib_Qax_obj *	op,
			  int				power);
static calib_bool
		Qax_is_prime (mpz_srcptr p);
static calib_bool
		Qax_is_square_free (const struct calib_Qax_obj *	poly);
static void	Qax_lift_factors (struct afglob * afgp);
static void	Qax_mark_trial_factors_as_used (struct afglob * afgp);
static void	Qax_monicize (struct calib_Qax_obj *		poly);
static void	Qax_mul (struct calib_Qax_obj *		rop,
			 const struct calib_Qax_obj *	op1,
			 const struct calib_Qax_obj *	op2);
static void	Qax_mul_non_algint (struct calib_Qax_obj *		rop,
				    const struct calib_Qax_obj *	op1,
				    const struct calib_Qax_obj *	op2);
static void	Qax_neg (struct calib_Qax_obj *		rop,
			 const struct calib_Qax_obj *	op);
static struct calib_GFpk_obj *
		Qax_new_GFpk_vector (const struct calib_GFpk_dom *	cf,
				     int   				n);
static struct calib_Qa_obj *
		Qax_new_Qa_vector (const struct calib_Qa_dom *	cf,
				   int				n);
static void	Qax_next_prime (mpz_ptr p, mpz_ptr prime_counter);
static calib_bool
		Qax_onep (const struct calib_Qax_obj *	op);
static void	Qax_print_initial_gfinfo (struct afglob * afgp,
					  const char *	  xvar,
					  const char *	  avar);
static void	Qax_print_lifted_gfinfo (struct afglob * afgp,
					 const char *	 xvar,
					 const char *	 avar);
static void	Qax_print_one_initial_gfinfo (
				const struct gfinfo *	gip,
				const char *		xvar,
				const char *		avar);
static void	Qax_print_one_lifted_gfinfo (
				const struct gfinfo *	gip,
				const char *		xvar,
				const char *		avar);
static void	Qax_print_trial_factor_combination_string (struct afglob * afgp);
static void	Qax_QZaPoly_clear (struct QZaPoly * p);
static void	Qax_QZaPoly_init (const struct calib_Zax_dom *	R_of_x,
				  struct QZaPoly *		p);
static void	Qax_QZaPoly_set_Qax (struct QZaPoly *			dst,
				     const struct calib_Qax_obj *	src);
static void	Qax_realloc_copy_poly (
				struct calib_Qax_obj *		poly,
				const struct calib_Qa_obj *	coeff,
				int				degree);
static void	Qax_set (struct calib_Qax_obj *		rop,
			 const struct calib_Qax_obj *	op);
static void	Qax_set_genrep (struct calib_Qax_obj *		rop,
				const struct calib_genrep *	op,
				const char *			xvar,
				const char *			avar);
static void	Qax_set_q (struct calib_Qax_obj *	rop,
			   mpq_srcptr			op);
static void	Qax_set_si (struct calib_Qax_obj *		rop,
			    calib_si_t				op);
static void	Qax_set_var_power (struct calib_Qax_obj *	rop,
				   int				power);
static void	Qax_set_z (struct calib_Qax_obj *	rop,
			   mpz_srcptr			op);
static void	Qax_set_Zx (struct calib_Qax_obj *		rop,
			    const struct calib_Zx_obj *		op);
static int	Qax_sturm_eval_sign (
				const struct calib_Zx_obj *	poly,
				int				inf_flag,
				mpz_srcptr			x);
static struct sturm *
		Qax_sturm_sequence (const struct calib_Zx_obj * poly);
static int	Qax_sturm_variation (const struct sturm *	ss,
				     int			inf_flag,
				     mpz_srcptr			x);
static void	Qax_sub (struct calib_Qax_obj *		rop,
			 const struct calib_Qax_obj *	op1,
			 const struct calib_Qax_obj *	op2);
static void	Qax_to_algint (struct calib_Qax_obj *		rop,
			       const struct calib_Qax_obj *	op);
static struct calib_genrep *
		Qax_to_genrep (const struct calib_Qax_obj *	op,
			       const char *			xvar,
			       const char *			avar);
static calib_bool
		Qax_try_recurse_1 (struct afglob * afgp, int i);
static calib_bool
		Qax_try_recurse_2 (struct afglob *		afgp,
				   int				i,
				   int				j,
				   struct calib_Zax_obj *	prod);
static void	Qax_try_all_trial_factors (struct afglob * afgp);
static void	Qax_upper_bound_positive_root (
				mpz_ptr				ubound,
				const struct calib_Zx_obj *	poly);
static void	Qax_validate_apoly (const struct calib_Zx_obj * apoly);
static calib_bool
		Qax_zerop (const struct calib_Qax_obj *	op);

/* Standard "max" of two integers. */

#define _MAX(A,B)	(((A) > (B)) ? (A) : (B))

/*
 * Macros to allocate/zero a local polynomial of the given degree.
 */

#define ALLOCATE_LOCAL_POLY(K_OF_X, PNAME, DEGREE) \
	struct calib_Qax_obj	PNAME; \
	Qax_init_degree (K_OF_X, &PNAME, (DEGREE))

#define FREE_LOCAL_POLY(PNAME) \
	Qax_clear (&PNAME)

/*
 * Macros to allocate/zero an array of local polynomials, each having
 * the given degree.
 */

#define ALLOCATE_LOCAL_POLY_ARRAY(K_OF_X, PNAME, NPOLY, DEGREE) \
	struct calib_Qax_obj	PNAME [NPOLY]; \
	Qax_init_array_degree (K_OF_X, &PNAME [0], (NPOLY), (DEGREE))

#define FREE_LOCAL_POLY_ARRAY(PNAME, NPOLY) \
	Qax_clear_array(&PNAME [0], (NPOLY))

/*
 * Initialize the given polynomial.  Its value after initialization is zero.
 */

	static
	inline
	void
Qax_init (

const struct calib_Qax_dom *	K_of_x,	/* IN: Polynomial ring */
struct calib_Qax_obj *		x	/* IN/OUT: polynomial to initialize */
)
{
	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR (x EQ NULL));

	x -> degree	= -1;
	x -> size	= 0;
	x -> dom	= K_of_x;
	x -> coeff	= NULL;
}

/*
 * Initialize the given polynomial to zero, but provide it with a
 * coefficient buffer sufficient to hold a polynomial of the given degree.
 */

	static
	inline
	void
Qax_init_degree (

const struct calib_Qax_dom *	K_of_x,	/* IN: Polynomial ring */
struct calib_Qax_obj *		x,	/* IN: polynomial to initialize */
int				degree	/* IN: maximum degree */
)
{
int				size;
const struct calib_Qa_dom *	cf;

	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR
			      (x      EQ NULL) OR
			      (degree  < -1));

	cf	= K_of_x -> cf;

	size = degree + 1;

	x -> degree	= -1;
	x -> size	= size;
	x -> dom	= K_of_x;
	x -> coeff	= Qax_new_Qa_vector (cf, size);
}

/*
 * Initialize the given array of polynomials to zero, but provide each
 * with a coefficient buffer sufficient to hold a polynomial of the
 * given degree.
 *
 * Although it is tempting to allocate all of the coefficients for all of
 * the polynomials into one big long vector, we cannot do so because the
 * coefficient vector for each polynomial must be separately free-able,
 * and re-allocatable.
 */

	static
	inline
	void
Qax_init_array_degree (

const struct calib_Qax_dom *	K_of_x,	/* IN: Polynomial ring */
struct calib_Qax_obj *		parray,	/* IN/OUT: array of polynomials */
int				npoly,	/* IN: number of polynomials in */
					/*     array */
int				degree	/* IN: max degree of each polynomial */
)
{
int				i;
struct calib_Qax_obj *		pp;

	CALIB_FATAL_ERROR_IF ((K_of_x EQ NULL) OR (degree < -1));

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (parray EQ NULL);

	pp = &parray [0];
	for (i = 0; i < npoly; i++, pp++) {
		Qax_init_degree (K_of_x, pp, degree);
	}
}

/*
 * Force the given polynomial to have buffer space sufficient to hold
 * a polynomial of at least the given degree.  This does not change
 * the value currently residing in this polynomial.
 */

	static
	void
Qax_alloc (

struct calib_Qax_obj *		rop,	/* IN/OUT: Polynomial to alloc */
int				degree	/* IN: Degree to assure space for */
)
{
int				size, dsize;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		scp;
struct calib_Qa_obj *		dcp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (degree < -1));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;
	dsize	= rop -> size;

	size = degree + 1;

	if (size > dsize) {
		scp = rop -> coeff;
		dcp = Qax_new_Qa_vector (cf, size);
		Qax_copy_Qa_vector (dcp, scp, dsize);
		Qax_free_Qa_vector (scp, dsize);
		rop -> coeff	= dcp;
		rop -> size	= size;
	}
}

/*
 * Clear out the given polynomial.
 */

	static
	inline
	void
Qax_clear (

struct calib_Qax_obj *	x	/* IN/OUT: polynomial to clear */
)
{
int				size;
struct calib_Qa_obj *		coeff;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	coeff	= x -> coeff;

	if (coeff NE NULL) {
		size = x -> size;
		CALIB_FATAL_ERROR_IF (size <= 0);
		Qax_free_Qa_vector (coeff, size);
	}

	x -> degree	= -1;
	x -> size	= 0;
	x -> dom	= NULL;
	x -> coeff	= NULL;
}

/*
 * Clear out the given array of polynomials.
 */

	static
	inline
	void
Qax_clear_array (

struct calib_Qax_obj * parray,	/* IN: array of polynomials to clear */
int			npoly	/* IN: number of polynomials in parray */
)
{
int		i;

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (parray EQ NULL);

	for (i = 0; i < npoly; i++) {
		Qax_clear (&parray [i]);
	}
}

/*
 * Set the given destination polynomial from the given coefficient
 * array.  Reallocate the destination polynomial's coefficient array,
 * if necessary.
 */

	static
	inline
	void
Qax_realloc_copy_poly (

struct calib_Qax_obj *		poly,	/* IN: destination polynomial */
const struct calib_Qa_obj *	coeff,	/* IN: source coefficient array */
int				degree	/* IN: degree to set */
)
{
int				cur_size, new_size;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		orig_coeff;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (degree < -1));

	poly -> degree	= degree;

	if (degree < 0) {
		/* We allow (coeff EQ NULL) here (but do not require). */
		return;
	}

	CALIB_FATAL_ERROR_IF (coeff EQ NULL);

	new_size = degree + 1;
	if (new_size > 0) {
		cur_size = poly -> size;
		K_of_x	= poly -> dom;
		cf	 = K_of_x -> cf;
		if (new_size > cur_size) {
			orig_coeff = poly -> coeff;
			if (orig_coeff NE NULL) {
				Qax_free_Qa_vector (orig_coeff, cur_size);
			}
			poly -> coeff = Qax_new_Qa_vector (cf, new_size);
			poly -> size = new_size;
		}
		Qax_copy_Qa_vector (poly -> coeff, coeff, new_size);
	}
}

/*
 * Allocate and initialize an array of Q(a) objects.
 */

	static
	struct calib_Qa_obj *
Qax_new_Qa_vector (

const struct calib_Qa_dom *	cf,	/* IN: Coefficient field */
int				n	/* IN: Number of elements */
)
{
int			i;
struct calib_Qa_obj *	array;

	array = NULL;
	if (n > 0) {
		array = CALIB_NEWA (n, struct calib_Qa_obj);
		for (i = 0; i < n; i++) {
			cf -> init (cf, &array [i]);
		}
	}

	return (array);
}

/*
 * Free up the given array of Q(a) objects.
 */

	static
	void
Qax_free_Qa_vector (

struct calib_Qa_obj *	array,		/* IN/OUT: Array to free */
int			n		/* IN: Number of elements */
)
{
int				i;
const struct calib_Qa_dom *	cf;

	if (array NE NULL) {
		CALIB_FATAL_ERROR_IF (n <= 0);
		cf = array [0].dom;
		for (i = 0; i < n; i++) {
			cf -> clear (&array [i]);
		}
		free (array);
	}
}

/*
 * Copy a vector of Q(a) objects.
 */

	static
	void
Qax_copy_Qa_vector (

struct calib_Qa_obj *		dst,	/* IN/OUT: destination vector */
const struct calib_Qa_obj *	src,	/* IN: source vector */
int				n	/* IN: number of elements */
)
{
int				i;
const struct calib_Qa_dom *	cf;

	if ((n > 0) AND (dst NE src)) {
		cf = dst -> dom;
		CALIB_FATAL_ERROR_IF (src -> dom NE cf);
		for (i = 0; i < n; i++) {
			cf -> set (&dst [i], &src [i]);
		}
	}
}

/*
 * Copy the given polynomial.
 */

	static
	void
Qax_set (

struct calib_Qax_obj *		rop,		/* IN/OUT: destination */
const struct calib_Qax_obj *	op		/* IN: source */
)
{
const struct calib_Qax_dom *	K_of_x;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	if (rop NE op) {
		Qax_realloc_copy_poly (rop, op -> coeff, op -> degree);
	}
}

/*
 * Set the given Qax polynomial to the given constant signed integer value.
 */

	static
	void
Qax_set_si (

struct calib_Qax_obj *		rop,		/* IN/OUT: destination */
calib_si_t			op		/* IN: source */
)
{
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	K_of_x	= rop -> dom;

	if (op EQ 0) {
		rop -> degree = -1;
	}
	else {
		Qax_alloc (rop, 0);
		cf = K_of_x -> cf;
		cf -> set_si (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given Qax polynomial to the given GMP integer value.
 */

	static
	void
Qax_set_z (

struct calib_Qax_obj *		rop,		/* IN/OUT: destination */
mpz_srcptr			op		/* IN: source */
)
{
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;

	if (mpz_sgn (op) EQ 0) {
		rop -> degree = -1;
	}
	else {
		Qax_alloc (rop, 0);
		cf = K_of_x -> cf;
		cf -> set_z (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given Qax polynomial to the given GMP integer value.
 */

	static
	void
Qax_set_q (

struct calib_Qax_obj *		rop,		/* IN/OUT: destination */
mpq_srcptr			op		/* IN: source */
)
{
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpq_sgn (op) EQ 0) {
		rop -> degree = -1;
	}
	else {
		Qax_alloc (rop, 0);
		K_of_x	= rop -> dom;
		cf	= K_of_x -> cf;
		cf -> set_q (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given Qax polynomial to the given Q(a) value.
 */

	static
	void
Qax_set_Qa (

struct calib_Qax_obj *		rop,		/* IN/OUT: destination */
const struct calib_Qa_obj *	op		/* IN: source */
)
{
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;

	if (cf -> zerop (op)) {
		rop -> degree = -1;
	}
	else {
		Qax_alloc (rop, 0);
		cf -> set (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the given Qax polynomial to a single term x**power,
 * where x is the Qax polynomial's variable.
 */

	static
	void
Qax_set_var_power (

struct calib_Qax_obj *		rop,		/* IN/OUT: destination */
int				power		/* IN: power to use */
)
{
int				i;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (power < 0));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;

	Qax_alloc (rop, power);

	coeff	= rop -> coeff;
	for (i = 0; i < power; i++) {
		cf -> set_si (&coeff [i], 0);
	}
	cf -> set_si (&coeff [power], 1);
	rop -> degree = power;
}

/*
 * Copy the given source polynomial (in Z[x] form) into the given
 * destination polynomial (in Q(a)[x] form).
 */

	static
	void
Qax_set_Zx (

struct calib_Qax_obj *		rop,	/* IN/OUT: Destination polynomial */
const struct calib_Zx_obj *	op	/* IN: Source polynomial */
)
{
int				i, deg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
mpz_srcptr			scoeff, scp;
struct calib_Qa_obj *		dcoeff;
struct calib_Qa_obj *		dcp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;
	deg	= op -> degree;
	scoeff	= op -> coeff;

	if (deg < 0) {
		CALIB_FATAL_ERROR_IF (deg NE -1);

		/* Source polynomial is zero.  This case is easy. */
		rop -> degree = -1;
		return;
	}

	/* A new Q vector gives us all zeros (0 / 1). */
	Qax_alloc (rop, deg);

	dcoeff	= rop -> coeff;
	scp = scoeff;
	dcp = dcoeff;
	for (i = 0; i <= deg; i++) {
		cf -> set_z (dcp, scp);
		++scp;
		++dcp;
	}
	rop -> degree = deg;
}

/*
 * Addition of two polynomials over Q(a).
 */

	static
	void
Qax_add (

struct calib_Qax_obj *		rop,		/* IN/OUT: result */
const struct calib_Qax_obj *	op1,		/* IN: first operand */
const struct calib_Qax_obj *	op2		/* IN: second operand */
)
{
int				i, rdeg, deg1, deg2, rsize;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		res;
struct calib_Qa_obj *		rp;
const struct calib_Qa_obj *	sp1;
const struct calib_Qa_obj *	sp2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE K_of_x) OR
			      (op2 -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = Qax_new_Qa_vector (cf, rsize);

	sp1 = &(op1 -> coeff [0]);
	sp2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, sp1++, sp2++, rp++) {
		if (i <= deg1) {
			if (i <= deg2) {
				cf -> add (rp, sp1, sp2);
			}
			else {
				cf -> set (rp, sp1);
			}
		}
		else {
			cf -> set (rp, sp2);
		}
	}
	while ((rdeg >= 0) AND (cf -> zerop (&res [rdeg]))) {
		--rdeg;
	}

	Qax_realloc_copy_poly (rop, res, rdeg);
	Qax_free_Qa_vector (res, rsize);
}

/*
 * Subtraction of two polynomials over Q(a).
 */

	static
	void
Qax_sub (

struct calib_Qax_obj *		rop,		/* IN/OUT: result */
const struct calib_Qax_obj *	op1,		/* IN: first operand */
const struct calib_Qax_obj *	op2		/* IN: second operand */
)
{
int				i, rdeg, deg1, deg2, rsize;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		res;
struct calib_Qa_obj *		rp;
const struct calib_Qa_obj *	sp1;
const struct calib_Qa_obj *	sp2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE K_of_x) OR
			      (op2 -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = Qax_new_Qa_vector (cf, rsize);

	sp1 = &(op1 -> coeff [0]);
	sp2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, sp1++, sp2++, rp++) {
		if (i <= deg1) {
			if (i <= deg2) {
				cf -> sub (rp, sp1, sp2);
			}
			else {
				cf -> set (rp, sp1);
			}
		}
		else {
			cf -> neg (rp, sp2);
		}
	}
	while ((rdeg >= 0) AND (cf -> zerop (&res [rdeg]))) {
		--rdeg;
	}
	Qax_realloc_copy_poly (rop, res, rdeg);
	Qax_free_Qa_vector (res, rsize);
}

/*
 * Negate the given Q(a)[x] polynomial.
 */

	static
	void
Qax_neg (

struct calib_Qax_obj *		rop,		/* IN/OUT: result */
const struct calib_Qax_obj *	op		/* IN: operand to negate */
)
{
int				i, deg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		dstp;
const struct calib_Qa_obj *	srcp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	if (rop EQ op) {
		deg	= rop -> degree;
		dstp	= rop -> coeff;
		for (i = 0; i <= deg; i++) {
			cf -> neg (dstp, dstp);
			++dstp;
		}
	}
	else {
		deg	= op -> degree;
		srcp	= op -> coeff;
		Qax_alloc (rop, deg);
		dstp	= rop -> coeff;
		for (i = 0; i <= deg; i++) {
			cf -> neg (dstp, srcp);
			++dstp;
			++srcp;
		}
		rop -> degree = deg;
	}
}

/*
 * Multiplcation of two polynomials over Q(a).
 */

	static
	void
Qax_mul (

struct calib_Qax_obj *		rop,		/* IN/OUT: result */
const struct calib_Qax_obj *	op1,		/* IN: first operand */
const struct calib_Qax_obj *	op2		/* IN: second operand */
)
{
int				i, deg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
const struct calib_Zax_dom *	Zax_dom;
const struct calib_Za_obj *	scoeff;
struct calib_Qa_obj *		dcoeff;
mpq_t				RFACTOR, TEMP;
struct calib_Zax_obj		RES;
struct QZaPoly			A, B;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE K_of_x) OR
			      (op2 -> dom NE K_of_x));

	if ((op1 -> degree < 0) OR (op2 -> degree < 0)) {
		/* Multiplying by zero. */
		rop -> degree = -1;
		return;
	}

	cf	= K_of_x -> cf;

	if (NOT (cf -> is_algint (cf))) {
		/* Generator polynomial is not monic, so we can't use	*/
		/* Za or Zax.  Translate to algebraic integer, multiply	*/
		/* and translate back.  This is likely to involve far	*/
		/* less mpz_gcd() operations than doing the multiply	*/
		/* with rational arithmetic.				*/
		Qax_mul_non_algint (rop, op1, op2);
		return;
	}

	Zax_dom	= Qax_get_Zax_dom (K_of_x);

	mpq_init (RFACTOR);
	mpq_init (TEMP);

	Qax_QZaPoly_init (Zax_dom, &A);
	Qax_QZaPoly_init (Zax_dom, &B);

	Zax_dom -> init (Zax_dom, &RES);

	Qax_QZaPoly_set_Qax (&A, op1);
	Qax_QZaPoly_set_Qax (&B, op2);

	mpq_mul (RFACTOR, A.qzap_s, B.qzap_s);

	/* Perform multiplication in Z(a)[x], which does not involve	*/
	/* any rational arithmetic (integer GCD's), and is therefore	*/
	/* reasonably fast.						*/
	Zax_dom -> mul (&RES, &A.qzap_poly, &B.qzap_poly);

	deg	= RES.degree;
	scoeff	= RES.coeff;

	/* Assure that destination buffer is sufficiently large. */
	K_of_x -> alloc (rop, deg);

	/* Convert result to rational. */
	dcoeff = rop -> coeff;
	for (i = 0; i <= deg; i++) {
		cf -> set_Za_q (&dcoeff [i], &scoeff [i], RFACTOR);
	}
	rop -> degree = deg;

	Zax_dom -> clear (&RES);

	Qax_QZaPoly_clear (&B);
	Qax_QZaPoly_clear (&A);

	mpq_clear (TEMP);
	mpq_clear (RFACTOR);
}

/*
 * Multiply two polynomials over a Q(a) field that is not an algebraic
 * integer.  Convert to algebraic integer, multiply, and convert back.
 */

	static
	void
Qax_mul_non_algint (

struct calib_Qax_obj *		rop,		/* IN/OUT: result */
const struct calib_Qax_obj *	op1,		/* IN: first operand */
const struct calib_Qax_obj *	op2		/* IN: second operand */
)
{
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qax_dom *	ai_Qax;
struct calib_Qax_obj		ROP, OP1, OP2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE K_of_x) OR
			      (op2 -> dom NE K_of_x));

	if ((op1 -> degree < 0) OR (op2 -> degree < 0)) {
		/* Multiplying by zero. */
		rop -> degree = -1;
		return;
	}

	/* Get the algebraic integer version of this Qax domain. */
	ai_Qax	= K_of_x -> algint_dom (K_of_x);

	Qax_init (ai_Qax, &ROP);
	Qax_init (ai_Qax, &OP1);
	Qax_init (ai_Qax, &OP2);

	Qax_to_algint (&OP1, op1);
	Qax_to_algint (&OP2, op2);

	Qax_mul (&ROP, &OP1, &OP2);

	Qax_from_algint (rop, &ROP);

	Qax_clear (&OP2);
	Qax_clear (&OP1);
	Qax_clear (&ROP);
}

/*
 * Take the given polynomial to the given integer power.
 */

	static
	void
Qax_ipow (

struct calib_Qax_obj *		rop,		/* IN/OUT: result polynomial */
const struct calib_Qax_obj *	op,		/* IN: polynomial to exponentiate */
int				power		/* IN: integer power */
)
{
const struct calib_Qax_dom *	K_of_x;
struct calib_Qax_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (power < 0));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	Qax_init (K_of_x, &RES);
	Qax_init (K_of_x, &PREG);

	Qax_set_si (&RES, 1);
	Qax_set (&PREG, op);

	while (power > 0) {
		if ((power & 1) NE 0) {
			Qax_mul (&RES, &RES, &PREG);
		}
		Qax_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	Qax_set (rop, &RES);

	Qax_clear (&PREG);
	Qax_clear (&RES);
}

/*
 * Create a dynamically allocated copy of the given polynomial.
 */

	static
	struct calib_Qax_obj *
Qax_dup (

const struct calib_Qax_obj *	op	/* IN: polynomial to duplicate */
)
{
int				deg, size;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qax_obj *		dst;
struct calib_Qa_obj *		coeff;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg	= op -> degree;
	K_of_x	= op -> dom;
	cf	= K_of_x -> cf;

	size = deg + 1;

	dst = CALIB_NEW (struct calib_Qax_obj);
	coeff = NULL;
	if (size > 0) {
		coeff = Qax_new_Qa_vector (cf, size);
		Qax_copy_Qa_vector (coeff, op -> coeff, size);
	}
	dst -> degree	= deg;
	dst -> size	= size;
	dst -> dom	= K_of_x;
	dst -> coeff	= coeff;

	return (dst);
}

/*
 * Free the given polynomial, which must have been dynamically allocated.
 */

	static
	void
Qax_free (

struct calib_Qax_obj *	poly		/* IN: polynomial to free */
)
{
	if (poly NE NULL) {
		Qax_clear (poly);
		free (poly);
	}
}

/*
 * Evaluate the given poly(value), for given rational value.
 */

	static
	void
Qax_eval (

struct calib_Qa_obj *		rop,		/* IN/OUT: result */
const struct calib_Qax_obj *	poly,		/* IN: polynomial to eval */
const struct calib_Qa_obj *	value		/* IN: value to use */
)
{
int				i;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj		SUM;

	CALIB_FATAL_ERROR_IF ((rop   EQ NULL) OR
			      (poly  EQ NULL) OR
			      (value EQ NULL));

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;

	CALIB_FATAL_ERROR_IF ((rop -> dom NE cf) OR (value -> dom NE cf));

	cf -> init (cf, &SUM);

	for (i = poly -> degree; i >= 0; i--) {
		cf -> mul (&SUM, &SUM, value);
		cf -> add (&SUM, &SUM, &(poly -> coeff [i]));
	}
	cf -> set (rop, &SUM);
	cf -> clear (&SUM);
}

/*
 * Evaluate the given poly(value), for given value being itself a
 * polynomial in Q(a)[x].
 */

	static
	void
Qax_eval_poly (

struct calib_Qax_obj *		rop,	/* IN/OUT: Result */
const struct calib_Qax_obj *	poly,	/* IN: Polynomial to evaluate */
const struct calib_Qax_obj *	value	/* IN: Polynomial value to plug in */
)
{
int				i, pdeg, vdeg, rdeg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qax_obj		SUM;

	CALIB_FATAL_ERROR_IF ((rop   EQ NULL) OR
			      (poly  EQ NULL) OR
			      (value EQ NULL));

	K_of_x	= poly -> dom;

	CALIB_FATAL_ERROR_IF ((rop   -> dom NE K_of_x) OR
			      (value -> dom NE K_of_x));

	cf	= K_of_x -> cf;
	pdeg	= poly -> degree;
	vdeg	= value -> degree;

	if (pdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rdeg	= pdeg * vdeg;
	if (rdeg < 0) {
		rdeg = 0;
	}

	Qax_init_degree (K_of_x, &SUM, rdeg);

	for (i = poly -> degree; i >= 0; i--) {
		Qax_mul (&SUM, &SUM, value);
		cf -> add (&SUM.coeff [0],
			   &SUM.coeff [0],
			   &(poly -> coeff [i]));
		if ((SUM.degree < 0) AND (NOT cf -> zerop (&SUM.coeff [0]))) {
			SUM.degree = 0;
		}
	}

	Qax_set (rop, &SUM);

	Qax_clear (&SUM);
}

/*
 * The division algorith (with remainder):
 * Given polynomials a, b \in Q(a)[x], find polynomials q, r \in Q(a)[x]
 * such that:
 *
 *	a = q*b + r	AND	degree (r) < degree (b)
 *
 * This is easy and clean because we have polynomials over a field,
 * which supports division.
 */

	static
	void
Qax_div (

struct calib_Qax_obj *		quotient,	/* IN/OUT: quotient */
struct calib_Qax_obj *		remainder,	/* IN/OUT: remainder */
const struct calib_Qax_obj *	a,		/* IN: dividend */
const struct calib_Qax_obj *	b		/* IN: divisor */
)
{
int				i, j, adeg, bdeg, qdeg, rdeg, size;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		quot;
struct calib_Qa_obj *		rem;
struct calib_Qa_obj *		qp;
struct calib_Qa_obj *		rp;
struct calib_Qa_obj		BLC_INV, TEMP;
const struct calib_Qa_obj *	ap;
const struct calib_Qa_obj *	bp;

	CALIB_FATAL_ERROR_IF ((a EQ NULL) OR (b EQ NULL));

	K_of_x	= a -> dom;

	CALIB_FATAL_ERROR_IF (b -> dom NE K_of_x);
	CALIB_FATAL_ERROR_IF ((quotient NE NULL) AND
			      (quotient -> dom NE K_of_x));
	CALIB_FATAL_ERROR_IF ((remainder NE NULL) AND
			      (remainder -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	adeg	= a -> degree;
	bdeg	= b -> degree;

	/* Local coefficient buffers. */
	size = adeg + 1;
	quot = Qax_new_Qa_vector (cf, 2 * size);
	rem	= quot + size;

	cf -> init (cf, &BLC_INV);
	cf -> init (cf, &TEMP);

	ap = &(a -> coeff [0]);
	bp = &(b -> coeff [0]);

	if (bdeg <= 0) {
		/* Division by a constant.  Do not allow division by zero. */
		CALIB_FATAL_ERROR_IF ((bdeg < 0) OR (cf -> zerop (bp)));

		cf -> inv (&BLC_INV, bp);
		rp = &rem [0];
		for (i = 0; i <= adeg; i++, ap++, rp++) {
			cf -> mul (rp, ap, &BLC_INV);
		}
		if (quotient NE NULL) {
			Qax_realloc_copy_poly (quotient, rem, adeg);
		}
		if (remainder NE NULL) {
			remainder -> degree = -1;
		}

		Qax_free_Qa_vector (quot, 2 * size);

		return;
	}

	qdeg = adeg - bdeg;
	if (qdeg < -1) {
		qdeg = -1;
	}

	rdeg = adeg;
	Qax_copy_Qa_vector (&rem [0], ap, rdeg + 1);

	/* Compute multiplicative inverse of lc(b). */
	cf -> inv (&BLC_INV, &(b -> coeff [bdeg]));

	while (rdeg >= bdeg) {
		j = rdeg - bdeg;
		qp = &quot [j];
		cf -> mul (qp, &rem [rdeg], &BLC_INV);
		rp = &rem [rdeg - 1];
		bp = &(b -> coeff [bdeg - 1]);
		for (i = bdeg - 1; i >= 0; i--, rp--, bp--) {
			cf -> mul (&TEMP, qp, bp);
			cf -> sub (rp, rp, &TEMP);
		}
		--rdeg;
		rp = &rem [rdeg];
		while ((rdeg >= 0) AND (cf -> zerop (rp))) {
			--rdeg;
			--rp;
		}
	}

	if (quotient NE NULL) {
		Qax_realloc_copy_poly (quotient, quot, qdeg);
	}

	if (remainder NE NULL) {
		Qax_realloc_copy_poly (remainder, rem, rdeg);
	}

	Qax_free_Qa_vector (quot, 2 * size);
}

/*
 * Compute the Greatest Common Divisor of two polynomials over Q(a).
 * The result is always monic unless a = b = 0.
 *
 * We use the conventional Euclidean algorithm.
 */

	static
	void
Qax_gcd (

struct calib_Qax_obj *		gcd,		/* IN/OUT: GCD result */
const struct calib_Qax_obj *	a,		/* IN: first operand */
const struct calib_Qax_obj *	b		/* IN: second operand */
)
{
int				i, adeg, bdeg, maxdeg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qax_obj *		ap;
struct calib_Qax_obj *		bp;
struct calib_Qax_obj *		rp;
struct calib_Qax_obj *		qp;
struct calib_Qax_obj *		temp_ptr;
struct calib_Qa_obj *		lca_inv;
struct calib_Qa_obj *		cp;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	K_of_x	= gcd -> dom;

	CALIB_FATAL_ERROR_IF ((a -> dom NE K_of_x) OR (b -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	adeg	= a -> degree;
	bdeg	= b -> degree;
	maxdeg	= _MAX (adeg, bdeg);

	if (maxdeg < 0) {
		gcd -> degree = -1;
		return;
	}

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 4, maxdeg);

	temp_ptr = &TMP [0];
	ap = temp_ptr++;
	bp = temp_ptr++;
	rp = temp_ptr++;
	qp = temp_ptr++;

	Qax_set (ap, a);
	Qax_set (bp, b);

	while (bp -> degree >= 0) {
		Qax_div (qp, rp, ap, bp);
		temp_ptr = ap;
		ap = bp;
		bp = rp;
		rp = temp_ptr;
	}

	adeg = ap -> degree;
	if (adeg >= 0) {
		/* Monicize result. */
		lca_inv = &(qp -> coeff [0]);
		cf -> inv (lca_inv, &(ap -> coeff [adeg]));
		cp = &(ap -> coeff [0]);
		for (i = 0; i <= adeg; i++, cp++) {
			cf -> mul (cp, cp, lca_inv);
		}
	}

	Qax_set (gcd, ap);

	FREE_LOCAL_POLY_ARRAY (TMP, 4);
}

/*
 * The Extended Euclidean Algorithm for polynomials in K[x],
 * where K = Q(a).  Given polynomials a, b \in K[x],
 * compute polynomials g, xa, xb \in K[x] such that
 *
 * 1.	a is a multiple of g,
 * 2.	b is a multiple of g,
 * 3.	xa * a + xb * b = g.
 *
 * These three conditions imply that g = gcd(a,b).  The polynomial g
 * is monic (unless a = b = 0, in which case g = 0).
 */

	static
	void
Qax_extgcd (

struct calib_Qax_obj *		gcd,		/* IN/OUT: GCD result */
struct calib_Qax_obj *		xa,		/* IN/OUT: a multiplier */
struct calib_Qax_obj *		xb,		/* IN/OUT: b multiplier */
const struct calib_Qax_obj *	a,		/* IN: first operand */
const struct calib_Qax_obj *	b		/* IN: second operand */
)
{
int				i, adeg, bdeg, maxdeg, deg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qax_obj *		ap;
struct calib_Qax_obj *		bp;
struct calib_Qax_obj *		rp;
struct calib_Qax_obj *		qp;
struct calib_Qax_obj *		m1;
struct calib_Qax_obj *		m2;
struct calib_Qax_obj *		m3;
struct calib_Qax_obj *		n1;
struct calib_Qax_obj *		n2;
struct calib_Qax_obj *		n3;
struct calib_Qax_obj *		TEMP;
struct calib_Qax_obj *		temp_ptr;
struct calib_Qa_obj *		lca_inv;
struct calib_Qa_obj *		cp;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR
			      (xa  EQ NULL) OR
			      (xb  EQ NULL) OR
			      (a   EQ NULL) OR
			      (b   EQ NULL));

	K_of_x	= gcd -> dom;

	CALIB_FATAL_ERROR_IF ((xa -> dom NE K_of_x) OR
			      (xb -> dom NE K_of_x) OR
			      (a  -> dom NE K_of_x) OR
			      (b  -> dom NE K_of_x));

	cf	= K_of_x -> cf;

	adeg	= a -> degree;
	bdeg	= b -> degree;
	maxdeg	= _MAX (adeg, bdeg);

	if (maxdeg < 0) {
		Qax_set_si (gcd, 0);
		Qax_set_si (xa,  1);
		Qax_set_si (xb,  0);
		return;
	}

	ALLOCATE_LOCAL_POLY_ARRAY (K_of_x, TMP, 11, maxdeg);

	temp_ptr = &TMP [0];
	ap = temp_ptr++;
	bp = temp_ptr++;
	rp = temp_ptr++;
	qp = temp_ptr++;
	m1 = temp_ptr++;
	m2 = temp_ptr++;
	m3 = temp_ptr++;
	n1 = temp_ptr++;
	n2 = temp_ptr++;
	n3 = temp_ptr++;
	TEMP = temp_ptr++;

	Qax_set (ap, a);
	Qax_set (bp, b);

	/* ap = m1*a + n1*b */
	Qax_set_si (m1, 1);
	Qax_set_si (n1, 0);

	/* bp = m2*a + n2*b */
	Qax_set_si (m2, 0);
	Qax_set_si (n2, 1);

	while (bp -> degree >= 0) {
		Qax_div (qp, rp, ap, bp);

		/* r = ap - q*bp			*/
		/*   = (m1*a + n1*b) - q*(m2*a + n2*b)	*/
		/*   = (m1 - q*m2)*a + (n1 - q*n2)*b	*/
		/*   = m3*a	     + m3*b,		*/
		/* so we compute:			*/
		/*	m3 = m1 - q*m2			*/
		/*	n3 = n1 - q*n2			*/
		Qax_mul (TEMP, qp, m2);
		Qax_sub (m3, m1, TEMP);
		Qax_mul (TEMP, qp, n2);
		Qax_sub (n3, n1, TEMP);

		temp_ptr = ap;
		ap = bp;
		bp = rp;
		rp = temp_ptr;

		temp_ptr = m1;
		m1 = m2;
		m2 = m3;
		m3 = temp_ptr;

		temp_ptr = n1;
		n1 = n2;
		n2 = n3;
		n3 = temp_ptr;
	}

	/* We have ap = M1*a + N1*b. */

	adeg = ap -> degree;
	if (adeg >= 0) {
		/* Monicize result. */
		lca_inv = &(qp -> coeff [0]);
		cf -> inv (lca_inv, &(ap -> coeff [adeg]));

		cp = &(ap -> coeff [0]);
		for (i = 0; i <= adeg; i++, cp++) {
			cf -> mul (cp, cp, lca_inv);
		}

		cp = &(m1 -> coeff [0]);
		deg = m1 -> degree;
		for (i = 0; i <= deg; i++, cp++) {
			cf -> mul (cp, cp, lca_inv);
		}

		cp = &(n1 -> coeff [0]);
		deg = n1 -> degree;
		for (i = 0; i <= deg; i++, cp++) {
			cf -> mul (cp, cp, lca_inv);
		}
	}

	Qax_set (gcd, ap);
	Qax_set (xa, m1);
	Qax_set (xb, n1);

	FREE_LOCAL_POLY_ARRAY (TMP, 11);
}

/*
 * Initialize the given QZaPoly object.
 */

	static
	void
Qax_QZaPoly_init (

const struct calib_Zax_dom *	dom,	/* IN: Z(a)[x] polynomial domain */
struct QZaPoly *		p	/* IN/OUT: QZaPoly to initialize */
)
{
	CALIB_FATAL_ERROR_IF ((dom EQ NULL) OR (p EQ NULL));

	p -> qzap_dom	= dom;
	mpq_init (p -> qzap_s);

	dom -> init (dom, &(p -> qzap_poly));
}

/*
 * Clear out the given QZaPoly object.
 */

	static
	void
Qax_QZaPoly_clear (

struct QZaPoly *	p		/* IN/OUT: QZaPoly to clear */
)
{
const struct calib_Zax_dom *	dom;

	CALIB_FATAL_ERROR_IF (p EQ NULL);

	dom	= p -> qzap_dom;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	dom -> clear (&(p -> qzap_poly));
	mpq_clear (p -> qzap_s);

	memset (p, 0, sizeof (*p));
}

/*
 * Set the given QZaPoly object from the given instance of Q(a)[x].
 */

	static
	void
Qax_QZaPoly_set_Qax (

struct QZaPoly *		dst,	/* IN/OUT: Destination QZaPoly */
					/*	   object */
const struct calib_Qax_obj *	src	/* IN: Source Q(a)[x] instance */
)
{
int				i, j, k, n, deg, deg2;
const struct calib_Zax_dom *	dstdom;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
const struct calib_Zx_dom *	Zx;
const struct calib_Zx_obj *	zpoly;
mpz_srcptr			szp;
mpq_srcptr			sqcoeff, sqp;
mpz_ptr				dcoeff, dzp;
struct QaxCoeffs		COEFFS;
struct calib_Za_obj *		za_coeff;
struct calib_Zx_obj		DIFF;
mpz_t				LCM, GCD, TEMP;

	CALIB_FATAL_ERROR_IF ((dst EQ NULL) OR (src EQ NULL));

	dstdom	= dst -> qzap_dom;
	K_of_x	= src -> dom;
	cf	= K_of_x -> cf;
	zpoly	= cf -> gpoly;
	k	= zpoly -> degree;
	deg	= src -> degree;

	/* Verify that K_of_x and dstdom are using the same polynomial.	*/
	Zx = calib_get_Zx_dom ();
	Zx -> init (&DIFF);
	Zx -> sub (&DIFF, zpoly, dstdom -> cr -> gpoly);
	CALIB_FATAL_ERROR_IF (DIFF.degree >= 0);
	Zx -> clear (&DIFF);

	CALIB_FATAL_ERROR_IF (deg < -1);

	/* Get rational coefficients from source polynomial. */
	_calib_QaxCoeffs_init (&COEFFS, src);

	/* Compute LCM of src denominators. */
	mpz_init (LCM);
	mpz_init (GCD);
	mpz_init (TEMP);

	mpz_set_ui (LCM, 1);
	for (i = 0; i <= deg; i++) {
		deg2	= COEFFS.qax_coeff [i].qa_deg;
		sqcoeff	= COEFFS.qax_coeff [i].qa_coeff;
		for (j = 0; j <= deg2; j++) {
			sqp = &sqcoeff [j];
			szp = mpq_denref (sqp);
			if (mpz_cmp_ui (szp, 1) EQ 0) continue;

			mpz_gcd (GCD, LCM, szp);
			mpz_divexact (TEMP, szp, GCD);
			mpz_mul (LCM, LCM, TEMP);
		}
	}

	/* Clear denominators into temporary buffer. */
	n = (deg + 1) * k;
	dcoeff = calib_new_Z_vector (n);
	dzp = dcoeff;
	if (mpz_cmp_ui (LCM, 1) EQ 0) {
		/* All denominators are 1. */
		for (i = 0; i <= deg; i++) {
			deg2	= COEFFS.qax_coeff [i].qa_deg;
			sqcoeff	= COEFFS.qax_coeff [i].qa_coeff;
			for (j = 0; j <= deg2; j++) {
				sqp = &sqcoeff [j];
				szp = mpq_numref (sqp);
				mpz_set (&dzp [j], szp);
			}
			dzp += k;
		}
	}
	else {
		for (i = 0; i <= deg; i++) {
			deg2	= COEFFS.qax_coeff [i].qa_deg;
			sqcoeff	= COEFFS.qax_coeff [i].qa_coeff;
			for (j = 0; j <= deg2; j++) {
				sqp = &sqcoeff [j];
				mpz_divexact (TEMP, LCM, mpq_denref (sqp));
				mpz_mul (&dzp [j], TEMP, mpq_numref (sqp));
			}
			dzp += k;
		}
	}

	/* Compute GCD of temporary buffer. */
	mpz_abs (GCD, &dcoeff [0]);
	if (mpz_cmp_ui (GCD, 1) NE 0) {
		for (i = 1; i < n; i++) {
			if (mpz_sgn (&dcoeff [i]) EQ 0) continue;

			mpz_gcd (GCD, GCD, &dcoeff [i]);
			if (mpz_cmp_ui (GCD, 1) EQ 0) break;
		}
		if (mpz_cmp_ui (GCD, 1) NE 0) {
			/* Have non-trivial GCD/content.	*/
			/* Compute primitive-part.		*/
			for (i = 0; i < n; i++) {
				mpz_divexact (&dcoeff [i], &dcoeff [i], GCD);
			}
		}
	}

	/* Assure destination buffer is sufficiently big. */
	dstdom -> alloc (&(dst -> qzap_poly), deg);

	mpz_set (mpq_numref (dst -> qzap_s), GCD);
	mpz_set (mpq_denref (dst -> qzap_s), LCM);
	mpq_canonicalize (dst -> qzap_s);

	for (i = 0; i <= deg; i++) {
		za_coeff = &(dst -> qzap_poly.coeff [i]);
		calib_copy_Z_vector (&(za_coeff -> coeff [0]),
				     &dcoeff [i * k],
				     k);
	}
	dst -> qzap_poly.degree = deg;

	calib_free_Z_vector (dcoeff, n);

	mpz_clear (TEMP);
	mpz_clear (GCD);
	mpz_clear (LCM);

	_calib_QaxCoeffs_clear (&COEFFS);
}

/*
 * Factor the given polynomial in Q(a)[x] completely into its monic,
 * irreducible factors over Q(a)[x].  The correct multiplicity of each
 * factor is also determined.
 */

	static
	struct calib_Qax_factor *
Qax_factor (

const struct calib_Qax_obj *	poly	/* IN: polynomial to factor */
)
{
int				deg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
const struct calib_Zx_obj *	apoly;
struct calib_Qax_factor *	sqfactors;
struct calib_Qax_factor *	list;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;
	apoly	= cf -> gpoly;
	deg	= poly -> degree;

	/* Verify that the apoly F(a) is valid. */
	Qax_validate_apoly (apoly);

	if (NOT (cf -> is_algint (cf))) {
		/* Coefficient field is not an algebraic integer.	*/
		/* Translate to algint, factor, and translate back.	*/
		list = Qax_factor_non_algint (poly);
		return (list);
	}

	list = NULL;

	/* Make a local copy of the polynomial. */
	ALLOCATE_LOCAL_POLY (K_of_x, lpoly, deg);
	Qax_set (&lpoly, poly);

	/* Compute square free factorization.  This deals with:		*/
	/* - Zero and constant polynomials.				*/
	/* - Making things monic by pulling out the leading coefficient	*/
	/*   as a constant leading factor.				*/

	sqfactors = Qax_compute_square_free_factorization (&lpoly);

	/* Now factor each square-free factor into irreducibles. */
	list = Qax_finish_sqf (sqfactors);

	Qax_free_factors (sqfactors);

	FREE_LOCAL_POLY (lpoly);

	return (list);
}

/*
 * Validate that the apoly F(a) is valid.  It must be:
 *
 *	- Degree at least 2,
 *	- Irreducible over Z[x].
 */

	static
	void
Qax_validate_apoly (

const struct calib_Zx_obj *	apoly	/* IN: apoly F(a) to validate */
)
{
const struct calib_Zx_dom *	Zx;
struct calib_Zx_factor *	zxfactors;

	/* Verify degree >= 2. */
	CALIB_FATAL_ERROR_IF (apoly -> degree < 2);

	/* Verify irreducible over Z[x]. */
	Zx = calib_get_Zx_dom ();
	zxfactors = Zx -> factor (apoly);
	CALIB_FATAL_ERROR_IF ((zxfactors EQ NULL) OR
			      (zxfactors -> multiplicity NE 1) OR
			      (zxfactors -> next NE NULL));
	Zx -> free_factors (zxfactors);
}

/*
 * Factor polynomial over coefficient field Q(a) that is not an algebraic
 * integer.  We do this by first factoring over the corresponding
 * algebraic integer, and the translating the coefficients back to
 * the original field.
 */

	static
	struct calib_Qax_factor *
Qax_factor_non_algint (

const struct calib_Qax_obj *	poly	/* IN: polynomial to factor */
)
{
const struct calib_Qax_dom *	dom;
const struct calib_Qax_dom *	ai_Qax;
struct calib_Qax_factor *	ai_factors;
struct calib_Qax_factor *	ai_fp;
struct calib_Qax_factor *	list;
struct calib_Qax_factor **	hookp;
struct calib_Qax_obj *		respoly;
struct calib_Qax_factor *	fp;
struct calib_Qax_obj		POLY;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	dom	= poly -> dom;

	/* Get the corresponding algebraic integer domain. */

	/* Get the corresponding algebraic integer domains. */
	ai_Qax	= Qax_algint_dom (dom);

	/* Translate given polynomial into new domain. */
	Qax_init (ai_Qax, &POLY);

	Qax_to_algint (&POLY, poly);

	/* Now factor over this algebraic integer. */
	ai_factors = Qax_factor (&POLY);

	/* Translate the factors back to the original domain. */
	list	= NULL;
	hookp	= &list;
	for (ai_fp = ai_factors; ai_fp NE NULL; ai_fp = ai_fp -> next) {

		respoly = CALIB_NEW (struct calib_Qax_obj);
		Qax_init (dom, respoly);

		Qax_from_algint (respoly, ai_fp -> factor);

		fp = CALIB_NEW (struct calib_Qax_factor);

		fp -> multiplicity	= ai_fp -> multiplicity;
		fp -> factor		= respoly;
		fp -> next		= NULL;

		*hookp = fp;
		hookp = &(fp -> next);
	}

	Qax_free_factors (ai_factors);

	ai_Qax -> clear (&POLY);

	return (list);
}

/*
 * Factor the given polynomial into factors that are both square-free and
 * pairwise relatively prime, correctly determining the multiplicity of
 * each such factor.  This results in one non-trivial factor for each
 * distinct multiplicity present -- that factor being the product of all
 * factors having that that multiplicity.
 *
 * NOTE: the given polynomial is assumed to be monic, and may be
 *	 modified by this routine.
 */

	static
	struct calib_Qax_factor *
Qax_compute_square_free_factorization (

const struct calib_Qax_obj *	poly	/* IN: polynomial to factor */
)
{
int				i, deg, multiplicity;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qax_factor *	list;
struct calib_Qax_factor *	fp;
struct calib_Qax_factor **	hookp;
struct calib_Qax_obj *		pp;
struct calib_Qax_obj *		lc_poly;
struct calib_Qa_obj *		lc;
struct calib_Qa_obj *		lc2;
struct calib_Qa_obj		LC_INV;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;

	deg = poly -> degree;

	if (deg <= 0) {
		/* Zero or constant polynomial.  Already irreducible. */
		CALIB_FATAL_ERROR_IF (deg < -1);
		list = CALIB_NEW (struct calib_Qax_factor);
		list -> multiplicity	= 1;
		list -> factor		= Qax_dup (poly);
		list -> next		= NULL;

		return (list);
	}

	list = NULL;
	hookp = &list;

	ALLOCATE_LOCAL_POLY (K_of_x, f, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, fprime, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, a, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, b, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, c, deg);
	ALLOCATE_LOCAL_POLY (K_of_x, d, deg - 1);
	ALLOCATE_LOCAL_POLY (K_of_x, rem, deg - 1);

	Qax_set (&f, poly);

	/* Force to be monic. */
	lc_poly = NULL;
	lc =  &f.coeff [deg];
	if (NOT (cf -> onep (lc))) {
		/* Lead coefficient had better not be zero! */
		CALIB_FATAL_ERROR_IF (cf -> zerop (lc));

		lc2 = Qax_new_Qa_vector (cf, 1);
		cf -> set (lc2, lc);

		lc_poly = CALIB_NEW (struct calib_Qax_obj);
		lc_poly -> degree	= 0;
		lc_poly -> size		= 1;
		lc_poly -> dom		= K_of_x;
		lc_poly -> coeff	= lc2;

		fp = CALIB_NEW (struct calib_Qax_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= lc_poly;
		fp -> next		= NULL;

		*hookp = fp;
		hookp = &(fp -> next);

		/* Compute inverse of lead coefficient. */
		cf -> init (cf, &LC_INV);
		cf -> inv (&LC_INV, lc);

		/* Force leading coefficient to be 1. */
		cf -> set_si (lc, 1);

		/* Correct non-leading coefficients. */
		for (i = 0; i < deg; i++) {
			cf -> mul (&f.coeff [i], &f.coeff [i], &LC_INV);
		}

		cf -> clear (&LC_INV);
	}

	/* Yun's algorithm, initialize. */
	Qax_derivative (&fprime, &f);

	Qax_gcd (&a, &f, &fprime);

	Qax_div (&b, &rem, &f, &a);
	CALIB_FATAL_ERROR_IF (rem.degree >= 0);

	Qax_div (&c, &rem, &fprime, &a);
	CALIB_FATAL_ERROR_IF (rem.degree >= 0);

	Qax_derivative (&d, &b);
	Qax_sub (&d, &c, &d);

	for (multiplicity = 1; ; multiplicity++) {
		/* Yun's algorithm, iteration. */
		Qax_gcd (&a, &b, &d);

		if (a.degree > 0) {
			/* We have a new square-free factor. */
			pp = Qax_dup (&a);
			Qax_monicize (pp);

			fp = CALIB_NEW (struct calib_Qax_factor);
			fp -> multiplicity	= multiplicity;
			fp -> factor		= pp;
			fp -> next		= NULL;
			*hookp = fp;
			hookp = &(fp -> next);
		}
			 
		Qax_div (&b, &rem, &b, &a);
		CALIB_FATAL_ERROR_IF (rem.degree >= 0);

		Qax_div (&c, &rem, &d, &a);
		CALIB_FATAL_ERROR_IF (rem.degree >= 0);

		Qax_derivative (&d, &b);
		Qax_sub (&d, &c, &d);

		if (b.degree <= 0) break;
	}

	/* Must have b EQ 1 here. */
	CALIB_FATAL_ERROR_IF (b.degree NE 0);
	CALIB_FATAL_ERROR_IF (NOT (cf -> onep (&b.coeff [0])));

	FREE_LOCAL_POLY (rem);
	FREE_LOCAL_POLY (d);
	FREE_LOCAL_POLY (c);
	FREE_LOCAL_POLY (b);
	FREE_LOCAL_POLY (a);
	FREE_LOCAL_POLY (fprime);
	FREE_LOCAL_POLY (f);

	return (list);
}

/*
 * Compute the derivative of the given polynomial in Q(a)[x].
 */

	static
	void
Qax_derivative (

struct calib_Qax_obj *		rop,	/* IN/OUT: derivative polynomial */
const struct calib_Qax_obj *	op	/* IN: polynomial to differentiate */
)
{
int				i, deg, size;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		result;
const struct calib_Qa_obj *	pcoeff;
struct calib_Qa_obj		CBUF;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	K_of_x	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE K_of_x);

	cf	= K_of_x -> cf;

	deg	= op -> degree;
	pcoeff	= op -> coeff;

	if (deg < 0) {
		rop -> degree = -1;
		return;
	}

	size	= deg;
	result	= Qax_new_Qa_vector (cf, size);
	cf -> init (cf, &CBUF);

	for (i = 1; i <= deg; i++) {
		cf -> set_si (&CBUF, i);
		cf -> mul (&result [i - 1], &CBUF, &pcoeff [i]);
	}

	do {
		--deg;
	} while ((deg >= 0) AND (cf -> zerop (&result [deg])));

	Qax_realloc_copy_poly (rop, result, deg);

	cf -> clear (&CBUF);
	Qax_free_Qa_vector (result, size);
}

/*
 * Modify the given polynomial to be monic.
 */

	static
	void
Qax_monicize (

struct calib_Qax_obj *		poly	/* IN/OUT: polynomial to monicize */
)
{
int				i, deg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Qa_obj *		lc;
struct calib_Qa_obj *		coeff;
struct calib_Qa_obj		LC_INV;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	deg	= poly -> degree;
	if (deg >= 0) {
		K_of_x	= poly -> dom;
		coeff	= poly -> coeff;
		cf	= K_of_x -> cf;
		lc	= &coeff [deg];
		if (NOT (cf -> onep (lc))) {
			cf -> init (cf, &LC_INV);
			cf -> inv (&LC_INV, lc);
			cf -> set_si (lc, 1);
			for (i = 0; i < deg; i++) {
				cf -> mul (&coeff [i], &coeff [i], &LC_INV);
			}
			cf -> clear (&LC_INV);
		}
	}
}

/*
 * Given a list of square-free factors, "finish" the factorization by
 * factoring each into irreducible polynomials correctly labeled with
 * their respective multiplicities.
 */

	static
	struct calib_Qax_factor *
Qax_finish_sqf (

const struct calib_Qax_factor *	sqfactors	/* IN: Square-free factors */
)
{
int				multiplicity;
const struct calib_Qax_factor *	sqfp;
const struct calib_Qax_obj *	sqfpoly;
struct calib_Qax_factor *	list;
struct calib_Qax_factor **	hookp;
struct calib_Qax_factor *	irred_factors;
struct calib_Qax_factor *	nfp;

	list = NULL;
	hookp = &list;

	for (sqfp = sqfactors; sqfp NE NULL; sqfp = sqfp -> next) {
		multiplicity	= sqfp -> multiplicity;
		sqfpoly		= sqfp -> factor;

		if (sqfpoly -> degree <= 1) {
			/* Factor is constant or linear, and therefore	*/
			/* already irreducible.  Copy to output list.	*/
			nfp = CALIB_NEW (struct calib_Qax_factor);
			nfp -> multiplicity	= multiplicity;
			nfp -> factor		= Qax_dup (sqfpoly);
			nfp -> next		= NULL;

			*hookp = nfp;
			hookp = &(nfp -> next);
			continue;
		}

		irred_factors = Qax_factor_square_free (sqfpoly);

		/* Transfer the irreducible factors onto the final	*/
		/* list of factors, adding the correct multiplicity.	*/
		while (irred_factors NE NULL) {
			nfp = irred_factors;
			irred_factors = nfp -> next;
			nfp -> multiplicity	= multiplicity;
			nfp -> next		= NULL;

			*hookp = nfp;
			hookp = &(nfp -> next);
		}
	}

	return (list);
}

/*
 * Given sqfpoly \in K = Q(a)[x], which is guaranteed to be monic,
 * square-free, and of degree at least 1, factor it into irreducible
 * factors over K.
 */

	static
	struct calib_Qax_factor *
Qax_factor_square_free (

const struct calib_Qax_obj *	sqfpoly		/* IN: polynomial to factor */
)
{
int				deg;
struct calib_Qax_factor *	fp;

	CALIB_FATAL_ERROR_IF (sqfpoly EQ NULL);

	deg	= sqfpoly -> degree;

	if (deg <= 1) {
		/* Degree must be 1 here. */
		CALIB_FATAL_ERROR_IF (deg <= 0);

		/* Degree 1 is always irreducible. */
		fp = CALIB_NEW (struct calib_Qax_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= Qax_dup (sqfpoly);
		fp -> next		= NULL;
		return (fp);
	}

	fp = Qax_choose_sqf_factor_algorithm (sqfpoly);

	return (fp);
}

/*
 * Choose which factorization algorithm to use for the given
 * square-free polynomial.
 */

	static
	struct calib_Qax_factor *
Qax_choose_sqf_factor_algorithm (

const struct calib_Qax_obj *	sqfpoly		/* IN: polynomial to factor */
)
{
const struct calib_Qax_dom *	K_of_x;
struct calib_Qax_factor *	fp;

	K_of_x	= sqfpoly -> dom;

	switch (K_of_x -> settings.factor_method) {
	case CALIB_QAX_FACTOR_METHOD_WEINBERGER_ROTHSCHILD:
		/* Use method of Weinberger and Rothschild. */
		fp = Qax_factor_sqf_Weinberger_Rothschild (sqfpoly);
		break;

	case CALIB_QAX_FACTOR_METHOD_NORM:
		/* Use the "norm" method. */
		fp = Qax_factor_sqf_norm (sqfpoly);
		break;

	default:
		printf ("Unknown Qax factor method!\n");
		fp = CALIB_NEW (struct calib_Qax_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= Qax_dup (sqfpoly);
		fp -> next		= NULL;
		break;
	}

	return (fp);
}

/*
 * Factor a monic, square-free polynomial in Q(a)[x] using the "norm" method.
 */

	static
	struct calib_Qax_factor *
Qax_factor_sqf_norm (

const struct calib_Qax_obj *	fpoly		/* IN: polynomial to factor */
)
{
int				j, deg, xvar, avar;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
struct calib_Zxyz_dom *		xadom;
const struct calib_Zx_dom *	Zx;
struct calib_Qax_factor *	list;
struct calib_Qax_factor *	nfp;
struct calib_Qax_factor **	hookp;
struct calib_Zx_factor *	zxlist;
struct calib_Zx_factor *	zxfp;
const char *			varlist [2];
mpz_t				DENOM;
struct calib_Zx_obj		ZXPOLY;
struct calib_Qax_obj		SPOLY, R, QAPOLY, SUBST;
struct calib_Zxyz_obj		MVSPOLY, MVG, Pj;

	CALIB_FATAL_ERROR_IF (fpoly EQ NULL);

	deg	= fpoly -> degree;

	CALIB_FATAL_ERROR_IF (deg < 2);

	K_of_x	= fpoly -> dom;
	cf	= K_of_x -> cf;

	Qax_init (K_of_x, &SPOLY);
	Qax_init (K_of_x, &R);
	Qax_init (K_of_x, &QAPOLY);

	/* Prepare substitution X ==> X - j*a, where a is the	*/
	/* algebraic number.					*/
	Qax_init_degree (K_of_x, &SUBST, 1);
	Qax_set_var_power (&SUBST, 1);

	for (j = 2; ; j++) {
		/* Form X ==> X - j*a substitution.	*/
		/* Set -j*a coefficient.		*/
		cf -> set_var_power (&SUBST.coeff [0], 1);
		cf -> mul_si (&SUBST.coeff [0], &SUBST.coeff [0], -j);
		Qax_eval_poly (&SPOLY, fpoly, &SUBST);

		if (Qax_is_square_free (&SPOLY)) break;
	}

	/* Modify substitution to X ==> X + j*a (undo previous subst).	*/
	/* Set +j*a coefficient.					*/
	cf -> neg (&SUBST.coeff [0], &SUBST.coeff [0]);

	Zx	= calib_get_Zx_dom ();
	Zx -> init (&ZXPOLY);

	avar	= 0;
	xvar	= 1;
	varlist [avar]	= "a";
	varlist [xvar]	= "x";

	xadom	= calib_make_Zxyz_dom (2, varlist);

	xadom -> init (xadom, &MVSPOLY);
	xadom -> init (xadom, &MVG);
	xadom -> init (xadom, &Pj);

	mpz_init (DENOM);

	/* Convert SPOLY into Zxyz form [x,a]. */
	xadom -> copy_from_Qax (&MVSPOLY, DENOM, &SPOLY, xvar, avar);

	/* Convert g(a) into Zxyz form [a]. */
	xadom -> copy_from_Zx (&MVG, cf -> gpoly, avar);

	/* Compute Pj = resultant (f(x - j*a), g(a), a). */
	xadom -> resultant (&MVSPOLY, &MVSPOLY, &MVG, avar);

	/* Factor Pj into irreducibles over Z[x]. */
	xadom -> copy_into_Zx (&ZXPOLY, &MVSPOLY, xvar);
	zxlist = Zx -> factor (&ZXPOLY);

	/* FIXME -- deal with DENOM somewhere! */

	list	= NULL;
	hookp	= &list;
	for (zxfp = zxlist; zxfp NE NULL; zxfp = zxfp -> next) {
		CALIB_FATAL_ERROR_IF (zxfp -> multiplicity NE 1);
		if (zxfp -> factor -> degree <= 0) {
			/* Skip constant factors. */
			continue;
		}

		Qax_set_Zx (&QAPOLY, zxfp -> factor);

		Qax_gcd (&R, &SPOLY, &QAPOLY);

		CALIB_FATAL_ERROR_IF (R.degree <= 0);

		/* Substitute X ==> X + j*a (reverse substitution). */
		Qax_eval_poly (&R, &R, &SUBST);

		nfp = CALIB_NEW (struct calib_Qax_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= Qax_dup (&R);
		nfp -> next		= NULL;

		*hookp = nfp;
		hookp = &(nfp -> next);
	}

	Zx -> free_factors (zxlist);

	mpz_clear (DENOM);

	xadom -> clear (&Pj);
	xadom -> clear (&MVG);
	xadom -> clear (&MVSPOLY);

	calib_free_Zxyz_dom (xadom);

	Zx -> clear (&ZXPOLY);

	Qax_clear (&SUBST);
	Qax_clear (&QAPOLY);
	Qax_clear (&R);
	Qax_clear (&SPOLY);

	return (list);
}

/*
 * Return TRUE if-and-only-if the given polynomial is square-free.
 */

	static
	calib_bool
Qax_is_square_free (

const struct calib_Qax_obj *	poly		/* IN: polynomial to test */
)
{
calib_bool			is_sqf;
struct calib_Qax_factor *	sqfactors;
struct calib_Qax_factor *	fp;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	sqfactors = Qax_compute_square_free_factorization (poly);

	CALIB_FATAL_ERROR_IF (sqfactors EQ NULL);

	is_sqf = TRUE;
	for (fp = sqfactors; fp NE NULL; fp = fp -> next) {
		if (fp -> multiplicity > 1) {
			is_sqf = FALSE;
			break;
		}
	}
	Qax_free_factors (sqfactors);

	return (is_sqf);
}

/*
 * Factor a monic, square-free polynomial in Q(a)[x] using the algorithm
 * of Weinberger and Rothschild.
 */

	static
	struct calib_Qax_factor *
Qax_factor_sqf_Weinberger_Rothschild (

const struct calib_Qax_obj *	fpoly		/* IN: polynomial to factor */
)
{
int				i, deg;
const struct calib_Qax_dom *	K_of_x;
struct calib_Qax_factor *	fp;
struct afglob *			afgp;
struct afglob			afglob;

	CALIB_FATAL_ERROR_IF (fpoly EQ NULL);

	K_of_x	= fpoly -> dom;
	deg	= fpoly -> degree;

	CALIB_FATAL_ERROR_IF (deg < 2);

	/* Initialize info used globally throughout the algorithm. */
	afgp = &afglob;
	Qax_init_afglob (afgp, fpoly);

	/* Compute the constants b, disc and d:				*/
	/* 	b	is LCM of denominators in fpoly.		*/
	/*	disc	is the discriminant of apoly.			*/
	/*	d	is a positive integer such that whenever	*/
	/*		f(x) = g(x)*h(x) for g,h \in Q(a)[x], then	*/
	/*		it is guaranteed that d*g(x) \in Z(a)[x]	*/
	/*		(d integerizes every true factor).		*/

	Qax_compute_common_denominator (afgp);

	/* Compute zpoly = d*f(x) in Z(a)[x] form (clearing denominators). */
	/* This also creates the domains Za = Z(a), and Zax = Z(a)[x].	   */
	Qax_compute_zpoly (afgp);

	/* Compute an upper bound B on the magnitudes of any single	*/
	/* coefficient of any factor.					*/
	Qax_compute_factor_coefficient_bound (afgp);

	/* Choose a "good" prime p to work with, and also factor the	*/
	/* apoly F(a) over Zp[x].					*/
	Qax_choose_prime_p (afgp);

	if (K_of_x -> settings.WR_print_initial_factors) {
		Qax_print_initial_gfinfo (afgp, "x", "a");
	}

	/* Lift all of the factors to a sufficiently high modulus. */
	Qax_lift_factors (afgp);

	if (K_of_x -> settings.WR_print_lifted_factors) {
		Qax_print_lifted_gfinfo (afgp, "x", "a");
	}

	/* Compute 1/d (mod modulus). */
	i = mpz_invert (afgp -> dinv, afgp -> d, afgp -> modulus);
	CALIB_FATAL_ERROR_IF (NOT i);

	/* Compute half_modulus = floor (modulus / 2). */
	mpz_tdiv_q_2exp (afgp -> half_modulus, afgp -> modulus, 1);

	Qax_try_all_trial_factors (afgp);
	fp = afgp -> true_factors;

	Qax_cleanup_afglob (&afglob);

	return (fp);
}

/*
 * Initialize the global information used throughout the Q(a)[x]
 * factorization algorithm.
 */

	static
	void
Qax_init_afglob (

struct afglob *			afgp,	/* IN/OUT: global info to init */
const struct calib_Qax_obj *	fpoly	/* IN: polynomial f(x) to factor */
)
{
int				deg;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
const struct calib_Zx_obj *	apoly;

	CALIB_FATAL_ERROR_IF ((afgp EQ NULL) OR (fpoly EQ NULL));

	K_of_x	= fpoly -> dom;
	cf	= K_of_x -> cf;
	apoly	= cf -> gpoly;
	deg	= fpoly -> degree;

	memset (afgp, 0, sizeof (*afgp));

	afgp -> K_of_x			= K_of_x;
	afgp -> fpoly			= fpoly;
	afgp -> apoly			= apoly;

	/* Make locally-modifiable copy of fpoly. */
	Qax_init_degree (K_of_x, &(afgp -> lpoly), deg);
	Qax_set (&(afgp -> lpoly), fpoly);

	afgp -> true_factors		= NULL;
	afgp -> true_factor_hookp	= &(afgp -> true_factors);

	mpz_init (afgp -> b);
	mpz_init (afgp -> disc);
	mpz_init (afgp -> d);
	mpz_init (afgp -> B);
	mpz_init (afgp -> p);
	mpz_init (afgp -> modulus);
	mpz_init (afgp -> half_modulus);
	mpz_init (afgp -> dinv);
	mpz_init (afgp -> prime_counter);
	mpz_init (afgp -> crt_denom);

	mpz_set_ui (afgp -> prime_counter, 1);

	afgp -> nmf			= 0;
	afgp -> gfinfo			= NULL;
}

/*
 * Clear out and shut down the global information used throughout
 * the Q(a)[x] factorization algorithm.
 */

	static
	void
Qax_cleanup_afglob (

struct afglob *		afgp	/* IN/OUT: global info to clean up */
)
{
const struct calib_Qax_dom *	K_of_x;
struct calib_Zpx_dom *		Zpx;
struct calib_Zax_dom *		Zax;

	CALIB_FATAL_ERROR_IF (afgp EQ NULL);

	K_of_x	= afgp -> K_of_x;
	Zpx	= afgp -> Zpx;
	Zax	= afgp -> Zax;

	Qax_free_gfinfo (afgp -> nmf, afgp -> gfinfo);

	if (Zpx NE NULL) {
		Zpx -> free (afgp -> apoly_zpx);
	}
	calib_free_Zpx_dom (Zpx);
	calib_free_Zp_dom (afgp -> Zp);

	mpz_clear (afgp -> crt_denom);
	mpz_clear (afgp -> dinv);
	mpz_clear (afgp -> half_modulus);
	mpz_clear (afgp -> modulus);
	mpz_clear (afgp -> p);
	mpz_clear (afgp -> B);
	mpz_clear (afgp -> d);
	mpz_clear (afgp -> disc);
	mpz_clear (afgp -> b);

	if (Zax NE NULL) {
		Zax -> free (afgp -> zpoly);
	}
	calib_free_Zax_dom (Zax);
	calib_free_Za_dom (afgp -> Za);

	/* Do not clean up true_factors, true_factor_hookp, because	*/
	/* these are returned to the caller.				*/

	K_of_x -> clear (&(afgp -> lpoly));

	memset (afgp, 0, sizeof (*afgp));
}

/*
 * Compute the constants b, disc, and d:
 *
 *	b	is the smallext positive integer such that
 *		b*f(x) \in Z(a)[x].  (i.e., b clears the denominators
 *		in fpoly.)
 *
 *	disc	is the discriminant of apoly F(a).
 *
 *	d	is a positive integer such that whenever f(x) = g(x)*h(x)
 *		for g,h \in Q(a)[x], then it is guaranteed that
 *		d*g(x) \in Z(a)[x] (i.e., d integerizes every true factor
 *		of fpoly f(x)).
 *
 * Regarding d, let K = Q(a).  The ring R of integers in K is the set of all
 * z \in K such that z is a root of *some* polynomial H(x) \in Z[x].
 * Compute an integer d >= 1 that guarantees R \subset (1/d)Z(a).
 * We use d = b*discriminant(apoly).  See the paper by Rothschild, et, al.
 *
 * Store these in the afgp.
 */

	static
	void
Qax_compute_common_denominator (

struct afglob *		afgp		/* IN/OUT: global info */
)
{
int				i, deg;
const struct calib_Qax_obj *	fpoly;
const struct calib_Zx_dom *	Zx;
const struct calib_Qa_obj *	coeff;
const struct calib_Qa_obj *	cp;
mpq_srcptr			qp;
mpz_srcptr			zp;
mpz_ptr				b, disc, d;
mpz_t				bval, g;

	fpoly	= afgp -> fpoly;
	coeff	= fpoly -> coeff;

	deg	= fpoly -> degree;

	mpz_init (bval);
	mpz_init (g);

	/* Compute bval to be the LCM of all denominators in fpoly. */
	mpz_set_ui (bval, 1);
	for (i = 0; i <= deg; i++) {
		cp = &coeff [i];
		qp = cp -> qfact;
		zp = mpq_denref (qp);
		mpz_gcd (g, bval, zp);
		mpz_mul (bval, bval, zp);
		mpz_divexact (bval, bval, g);
	}

	b	= afgp -> b;
	disc	= afgp -> disc;
	d	= afgp -> d;

	mpz_set (b, bval);

	/* Compute the |discriminant| of apoly F(a). */
	Zx = calib_get_Zx_dom ();
	Zx -> discriminant (disc, afgp -> apoly);
	mpz_abs (disc, disc);

	/* Compute d. */
	mpz_mul (d, b, disc);

	mpz_clear (g);
	mpz_clear (bval);
}

/*
 * Compute zpoly = d*f(x) in Z(a)[x] form (clearing denominators).
 *
 * This also creates the domains Za = Z(a), and Zax = Z(a)[x].
 */

	static
	void
Qax_compute_zpoly (

struct afglob *	afgp		/* IN/OUT: global info */
)
{
struct calib_Za_dom *		Za;
struct calib_Zax_dom *		Zax;
const struct calib_Zx_obj *	apoly;
struct calib_Qax_obj *		lpoly;
struct calib_Zax_obj *		zpoly;
mpz_srcptr			d;

	apoly	= afgp -> apoly;
	lpoly	= &(afgp -> lpoly);
	d	= afgp -> d;

	Za  = calib_make_Za_dom (apoly);
	Zax = calib_make_Zax_dom (Za);

	zpoly = Zax -> cvQax (Zax, lpoly, d);

	afgp -> Za	= Za;
	afgp -> Zax	= Zax;
	afgp -> zpoly	= zpoly;
}

/*
 * Compute an integer B >= 1 that is an upper bound on the magnitude of
 * any single coefficient of any factor of the polynomial.  (By magnitude,
 * we mean in units of 1/d.)  This bound is computed as:
 *
 *	B = d * B2 * B3
 *
 * where B2 is an upper bound on ||phi_j||, where phi_j \in Q(a) is
 * an algebraic number coefficient appearing in any factor:
 *
 *	B2 = binomial (m, floor(m/2)) * B1,
 *
 * where m = degree(poly), and B1 is an upper bound on the magnitude of
 * all roots of poly.  Furthermore,
 *
 *			     k(k-1)/2
 *		 k! ||alpha||
 *	B3 = -----------------------------
 *	      sqrt(|discriminant(apoly)|)
 *
 * where k = degree(apoly), and alpha is an upper bound on the magnitude
 * of every root of apoly.
 *
 * Compute an integer B1 >= 1 such that f(delta) = 0 ==> ||delta|| < B1,
 * where f(x) \in Q(a)[x] is the polynomial we are trying to factor.
 */

	static
	void
Qax_compute_factor_coefficient_bound (

struct afglob *		afgp		/* IN/OUT: global info */
)
{
int				i, j, k, deg;
const struct calib_Zx_dom *	Zx;
const struct calib_Zax_obj *	zpoly;
const struct calib_Zx_obj *	apoly;
struct calib_Zx_obj *		apoly2;
struct calib_Za_obj *		zpoly_coeff;
mpz_ptr				zp, d, disc, B;
mpz_srcptr			cp;
mpz_t				alpha, B1, B2, B3, sum, temp;
struct calib_Zx_obj		b1poly;

	zpoly	= afgp -> zpoly;
	apoly	= afgp -> apoly;
	d	= afgp -> d;
	disc	= afgp -> disc;
	B	= afgp -> B;

	k	= apoly -> degree;
	deg	= zpoly -> degree;

	/* Bound the magnitude of every root of apoly.  Specifically,	*/
	/* Find integer alpha > 0 such that:				*/
	/*	apoly(z) = 0	==>	||z|| < alpha.			*/
	/*								*/
	/* Let apoly(x) = \sum_{i=0}^n a_i x^i.  Then we compute alpha	*/
	/* by upper-bounding the unique positive real root of:		*/
	/*	|a_n| x^n - \sum^{i=0}^{n-1} |a_i| x^i.			*/

	mpz_init (alpha);
	mpz_init (B1);
	mpz_init (B2);
	mpz_init (B3);
	mpz_init (sum);
	mpz_init (temp);

	Zx = calib_get_Zx_dom ();
	apoly2 = Zx -> dup (apoly);
	zp = &(apoly2 -> coeff [0]);
	for (i = 0; i < apoly2 -> degree; i++, zp++) {
		mpz_abs (zp, zp);
		mpz_neg (zp, zp);
	}
	mpz_abs (zp, zp);
	Qax_upper_bound_positive_root (alpha, apoly2);
	Zx -> free (apoly2);

	/* Now bound the magnitude of every root of fpoly \in Q(a)[x].	*/
	/* Once again, we find the unique positive real root of a	*/
	/* polynomial:							*/
	/*	|B_n| x^n - \sum_{j=0}^{n-1} |B_j| x^j			*/
	/* where each B_j is an upper bound for ||phi_j||, and phi_j	*/
	/* is the algebraic number that is the coefficient of x^j in	*/
	/* the polynomial we are trying to factor.			*/

	zp = calib_new_Z_vector (deg + 1);
	b1poly.degree	= deg;
	b1poly.size	= deg + 1;
	b1poly.coeff	= zp;

	for (j = 0; j < deg; j++, cp += k) {
		zpoly_coeff = &(zpoly -> coeff [j]);
		mpz_set_ui (sum, 0);
		cp = &(zpoly_coeff -> coeff [0]);
		for (i = k - 1; i >= 0; i--) {
			mpz_mul (sum, sum, alpha);
			mpz_abs (temp, &cp [i]);
			mpz_add (sum, sum, temp);
		}
		mpz_abs (sum, sum);
		mpz_cdiv_q (sum, sum, d);
		mpz_neg (&zp [j], sum);
	}
	mpz_set_ui (&zp [deg], 1);
	Qax_upper_bound_positive_root (B1, &b1poly);
	calib_free_Z_vector (zp, deg + 1);

	/* Now compute B2. */
	mpz_bin_uiui (temp, deg, deg/2);
	mpz_mul (B2, B1, temp);

	/* Now compute B3. */
	mpz_fac_ui (B3, k);
	mpz_pow_ui (temp, alpha, k*(k-1)/2);
	mpz_mul (B3, B3, temp);
	mpz_abs (temp, disc);
	mpz_sqrt (temp, temp);
	mpz_cdiv_q (B3, B3, temp);

	mpz_mul (temp, d, B2);
	mpz_mul (B, temp, B3);

	mpz_clear (temp);
	mpz_clear (sum);
	mpz_clear (B3);
	mpz_clear (B2);
	mpz_clear (B1);
	mpz_clear (alpha);
}

/*
 * We are given a polynomial f(x) \in Z[x] that is guaranteed to have exactly
 * one positive real root z.  Compute the smallest integer u such that
 * z <= u.  We use Sturm sequences to isolate this root exactly.
 *
 * Note that the polynomial we are given is NOT guaranteed to be square-free.
 * However, it is guaranteed to have only one positive root.  This means
 * we will not violate the requirements of Sturms theorem as long as we
 * only compute the Sturm variation at strictly positive values.
 */

	static
	void
Qax_upper_bound_positive_root (

mpz_ptr				ubound,		/* IN/OUT: upper bound u */
const struct calib_Zx_obj *	poly		/* IN: polynomial f(x) */
)
{
int			vinf, v1, v2, vmid;
struct sturm *		ss;
mpz_t			xlo, xhi, xmid;

	/* Compute the Sturm sequence of the polynomial. */
	ss = Qax_sturm_sequence (poly);

	/* Compute vinf = var(+inf). */
	vinf = Qax_sturm_variation (ss, +1, NULL);

	/* Now compute v1 = var(1). */
	mpz_init_set_ui (xlo, 1);
	mpz_init_set_ui (xhi, 2);
	mpz_init (xmid);

	do {		/* Used only for "break". */
		v1 = Qax_sturm_variation (ss, 0, xlo);
		if (v1 EQ vinf) {
			/* The interval (1,+inf) has no roots, so we	*/
			/* must have 0 < z <= 1.			*/
			mpz_set (ubound, xlo);
			break;
		}

		/* Verify exactly one root in the interval (1,+inf). */
		CALIB_FATAL_ERROR_IF (v1 NE vinf + 1);

		/* Check successive powers of 2 until we find an upper	*/
		/* bound.						*/
		for (;;) {
			v2 = Qax_sturm_variation (ss, 0, xhi);
			if (v2 EQ vinf) {
				/* The interval (xhi,+inf) has no	*/
				/* roots, so xhi is our first, crude	*/
				/* upper bound.				*/
				break;
			}
			/* The interval (xhi,+inf) has a root in it, so	*/
			/* xhi is a lower bound on the root.		*/
			mpz_set (xlo, xhi);
			mpz_mul_ui (xhi, xhi, 2);
		}

		/* We know that there is a single root in the interval	*/
		/* (xlo,xhi].  Use binary search to narrow it down to	*/
		/* unit width.						*/
		for (;;) {
			mpz_add_ui (xmid, xlo, 1);
			if (mpz_cmp (xmid, xhi) >= 0) {
				/* Our interval is now 1 wide! */
				break;
			}

			/* Interval is at least 2 wide.  Bisect. */
			mpz_add (xmid, xlo, xhi);
			mpz_tdiv_q_2exp (xmid, xmid, 1);
			vmid = Qax_sturm_variation (ss, 0, xmid);
			if (vmid EQ vinf) {
				/* Interval (xmid,+inf) has no roots.	*/
				/* Ditto for (xmid,xhi].		*/
				mpz_set (xhi, xmid);
			}
			else {
				/* Interval (xmid,+inf) has a root, so	*/
				/* ditto for (xmid,xhi].		*/
				mpz_set (xlo, xmid);
			}
		}

		/* The positive root lies in the interval (xlo,xhi],	*/
		/* and xhi - xlo = 1.  Therefore xhi is the upper bound	*/
		/* that we seek.					*/
		mpz_set (ubound, xhi);

	} while (FALSE);

	mpz_clear (xmid);
	mpz_clear (xhi);
	mpz_clear (xlo);
	Qax_free_sturm_sequence (ss);
}

/*
 * Given a polynomial in Z[x], compute its Sturm sequence.
 */

	static
	struct sturm *
Qax_sturm_sequence (

const struct calib_Zx_obj *	poly	/* IN: polynomial */
)
{
int				i, k, n, maxp, maxc, deg;
struct calib_Zx_obj *		spoly;
struct calib_Zx_obj *		spp;
struct sturm *			ss;
const struct calib_Zx_dom *	Zx;
mpz_ptr				coeffs, zp;
struct calib_Zx_obj		QUOT, REM;
mpz_t				d;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	deg	= poly -> degree;

	CALIB_FATAL_ERROR_IF (deg < 1);

	/* Compute maximum possible number of polynomials in the Sturm	*/
	/* sequence, and the maximum possible total number of		*/
	/* coefficients in all of these polynomials.			*/

	maxp = deg + 1;
	maxc = maxp * (maxp + 1) / 2;

	spoly	= CALIB_NEWA (maxp, struct calib_Zx_obj);
	coeffs	= calib_new_Z_vector (maxc);

	spp = spoly;
	zp  = coeffs;

	n = 0;

	/* Copy given polynomial into the sequence. */
	calib_copy_Z_vector (zp, poly -> coeff, deg + 1);
	spp -> degree	= deg;
	spp -> size	= deg + 1;
	spp -> coeff	= zp;
	zp += (deg + 1);
	++spp;
	++n;

	/* Copy the derivative in. */
	for (i = 1; i <= deg; i++) {
		mpz_mul_ui (&zp [i - 1], &(poly -> coeff [i]), i);
	}
	spp -> degree	= deg - 1;
	spp -> size	= deg;
	spp -> coeff	= zp;
	zp += deg;
	++spp;
	++n;

	Zx = calib_get_Zx_dom ();

	mpz_init (d);
	Zx -> init (&QUOT);
	Zx -> init (&REM);

	while (spp [-1].degree > 0) {
		/* Compute next polynomial remainder. */
		Zx -> div (&QUOT, &REM, d, spp - 2, spp - 1);

		if (mpz_sgn (d) <= 0) {
			/* Verify that "denominator" isn't 0. */
			CALIB_FATAL_ERROR_IF (mpz_sgn (d) EQ 0);
			Zx -> neg (&QUOT, &QUOT);
			Zx -> neg (&REM,  &REM);
			mpz_neg (d, d);
		}

		/* Next entry is negative of the remainder. */
		k = REM.degree;
		for (i = 0; i <= k; i++) {
			mpz_neg (&zp [i], &REM.coeff [i]);
		}
		spp -> degree	= k;
		spp -> size	= k + 1;
		spp -> coeff	= zp;
		zp += (k + 1);
		++spp;
		++n;
	}

	Zx -> clear (&REM);
	Zx -> clear (&QUOT);
	mpz_clear (d);

	ss = CALIB_NEW (struct sturm);
	ss -> n		= n;
	ss -> vsize	= maxc;
	ss -> poly	= spoly;
	ss -> coeff	= coeffs;

	return (ss);
}

/*
 * Free up the given Sturm sequence.
 */

	static
	void
Qax_free_sturm_sequence (

struct sturm *		ss		/* IN: Sturm sequence to free */
)
{
	if (ss NE NULL) {
		if (ss -> poly NE NULL) {
			free (ss -> poly);
			ss -> poly = NULL;
		}
		calib_free_Z_vector (ss -> coeff, ss -> vsize);
		memset (ss, 0, sizeof (*ss));
		free (ss);
	}
}

/*
 * Compute the "Sturm variation" of the given Sturm sequence at x.
 * Let P_0(x), P_1(x), ..., P_{n-1}(x) be the polynomials of the
 * Sturm sequence, and let z_0, z_1, ..., z_{n-1} be their values
 * at x.  (Evaluate in the limiting sense if x = +inf or x = -inf.)
 *
 * Then the Sturm variation is the number of sign changes (zero
 * entries are ignored) in the sequence z_0, z_1, ..., z_{n-1}.
 *
 * If inf_flag < 0, then use x = -inf.
 * If inf_flag > 0, then use x = +inf.
 * Otherwise, use the x value given.
 */

	static
	int
Qax_sturm_variation (

const struct sturm *	ss,		/* IN: Sturm sequence */
int			inf_flag,	/* IN: infinity flag */
mpz_srcptr		x		/* IN: evaluation point x */
)
{
int				i, n, variation;
int				sign, prev_sign;
const struct calib_Zx_obj *	poly;

	n = ss -> n;
	variation = 0;

	poly = ss -> poly;
	for (i = 0; ; i++, poly++) {
		/* All entries of Sturm sequence eval to zero! */
		CALIB_FATAL_ERROR_IF (i >= n);

		sign = Qax_sturm_eval_sign (poly, inf_flag, x);
		if (sign NE 0) break;
	}
	prev_sign = sign;
	++i;
	++poly;

	for (; i < n; i++, poly++) {
		sign = Qax_sturm_eval_sign (poly, inf_flag, x);
		if (sign EQ 0) continue;
		if (sign NE prev_sign) {
			++variation;
			prev_sign = sign;
		}
	}

	return (variation);
}

/*
 * We are given a polynomial f(x) \in Z[x], and a value
 * x \in Z \union {-inf,+inf}.
 * Return the sign of f(x) when evaluated at the given value.
 * If that value is -inf or +inf, evaluate it in the limiting sense.
 *
 * If inf_flag < 0, then use x = -inf.
 * If inf_flag > 0, then use x = +inf.
 * Otherwise, use the x value given.
 */

	static
	int
Qax_sturm_eval_sign (

const struct calib_Zx_obj *
			poly,		/* IN: polynomial to eval sign of */
int			inf_flag,	/* IN: infinity flag */
mpz_srcptr		x		/* IN: evaluation point x */
)
{
int				sign, deg;
const struct calib_Zx_dom *	Zx;
mpz_srcptr			lc;
mpz_t				y;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	if (inf_flag EQ 0) {
		/* Evaluate at a specific integer point x. */
		CALIB_FATAL_ERROR_IF (x EQ NULL);
		Zx = calib_get_Zx_dom ();
		mpz_init (y);
		Zx -> eval (y, poly, x);
		sign = mpz_sgn (y);
		mpz_clear (y);
	}
	else {
		/* Evaluating at +inf or -inf.  Make sure no finite	*/
		/* value was passed in.					*/
		CALIB_FATAL_ERROR_IF (x NE NULL);
		deg = poly -> degree;
		lc = &(poly -> coeff [deg]);
		sign = mpz_sgn (lc);
		if ((inf_flag < 0) AND ((deg & 1) NE 0)) {
			sign = - sign;
		}
		CALIB_FATAL_ERROR_IF (sign EQ 0);
	}
	return (sign);
}

/*
 * Choose a "good" prime p to work with, and also factor the
 * apoly F(a) over Zp[x].
 *
 * The following are necessary conditions on p:
 *
 *	- It must be odd.
 *	- It must not divide d.
 *	- It must not divide degree(f(x)).
 *	- The apoly F(a) must be square-free in Zp[x].
 *
 * Along with these necessary conditions, there is also a desirable
 * condition:
 *
 *	- We want as few factors in Zp[x] as possible.
 *
 * The fewer factors we have, the faster the combinatorial trial
 * factor generation and divisibility testing phase becomes.  (Or rather,
 * having many factors makes it *monstrously* slow, which we want to avoid
 * like the plague.)  So it makes sense to expend some substantial effort
 * here to find a modulus producing as few factors as possible -- it will
 * save *much* headache later.
 *
 * We note that the apoly F(a) has already been verified to be monic, and
 * irreducible over Z[x].  There is certainly some positive integer B
 * for which F(a) is irreducible (mod p) for all p > B.  Unfortunately,
 * using an especially large p causes the factorizations of f(x) over
 * GF(p^k)[x] to become much less efficient, so we want to avoid that too.
 *
 * We use a heuristic approach here.  We count successive primes tested.
 * Every time we find a new prime p that yields fewer factors, we reset
 * this counter to zero.  After K consecutive primes with no improvement,
 * we call it quits.
 */

	static
	void
Qax_choose_prime_p (

struct afglob *		afgp	/* IN/OUT: global info */
)
{
int		deg, nmf;
int		best_nmf, attempts, bcount;
struct pfact *	pfact;
struct pfact *	best_pfact;
struct pfact **	hookp;
mpz_ptr		d;
mpz_t		p, degZ;

	deg	 = afgp -> apoly -> degree;
	d	 = afgp -> d;

	best_nmf = deg + 1;
	best_pfact = NULL;
	hookp	 = &best_pfact;

	mpz_init (p);
	mpz_init (degZ);
	mpz_set_ui (degZ, deg);

	attempts = 0;
	bcount = 0;

	do {
		/* Choose next prime that satisfies the simple checks. */
		for (;;) {
			Qax_next_prime (p, afgp -> prime_counter);
			if ((NOT mpz_divisible_p (d, p)) AND
			    (NOT mpz_divisible_p (degZ, p))) {
				/* This prime is worth trying. */
				break;
			}
		}

		pfact = Qax_find_apoly_Zp_factors (afgp, p);
		if (pfact EQ NULL) {
			/* apoly is not square-free w.r.t. this p,	*/
			/* which is therefore bad.  Try another.	*/
			continue;
		}

		if (++bcount >= 3) {
			bcount = 0;
			/* Nudge up the magnitude of future primes. */
			mpz_mul_2exp (afgp -> prime_counter,
				      afgp -> prime_counter,
				      8);
			mpz_add_ui (afgp -> prime_counter,
				    afgp -> prime_counter,
				    1);
		}

		nmf = pfact -> nmf;
		if (nmf < best_nmf) {
			/* We have a new, lower number of factors.	*/
			/* Time to do the more expensive GF(p^k)	*/
			/* factorization, to check square-freeness of	*/
			/* those factors.				*/
			Qax_build_gfinfo (afgp, pfact);

			if (pfact -> gfinfo EQ NULL) {
				/* Bad prime p -- GF(p^k)[x] factors	*/
				/* were not square-free.		*/
				Qax_free_apoly_Zp_factors (pfact);
			}
			else {
				/* We have a new, lower number of	*/
				/* factors that are completely valid.	*/
				Qax_free_apoly_Zp_factors (best_pfact);
				best_pfact = pfact;
				best_nmf = pfact -> nmf;
				hookp	 = &(pfact -> next);
				attempts = 0;
				continue;
			}
		}
		else if (pfact -> nmf EQ best_nmf) {
			/* Another "best".  Append it to the list. */
			*hookp = pfact;
			hookp = &(pfact -> next);
		}
		else {
			/* Worse than best seen so far. */
			Qax_free_apoly_Zp_factors (pfact);
		}

		/* This is a failed attempt to improve. */
		++attempts;
	} while (attempts < 10);

	mpz_clear (p);

	/* Return the first (presumably lowest p) entry, and free all	*/
	/* of the others.						*/
	CALIB_FATAL_ERROR_IF (best_pfact EQ NULL);
	Qax_free_apoly_Zp_factors (best_pfact -> next);
	best_pfact -> next = NULL;

	/* Transfer these Zp data over to the global vars. */

	afgp -> nmf		= best_pfact -> nmf;
	mpz_set (afgp -> p,	  best_pfact -> p);
	afgp -> Zp		= best_pfact -> Zp;
	afgp -> Zpx		= best_pfact -> Zpx;
	afgp -> apoly_zpx	= best_pfact -> apoly;
	afgp -> gfinfo		= best_pfact -> gfinfo;

	afgp -> Zpx -> free_factors (best_pfact -> factors);
	memset (best_pfact, 0, sizeof (*best_pfact));
	free (best_pfact);
}

/*
 * Return the next prime to try.
 */

	static
	void
Qax_next_prime (

mpz_ptr		p,			/* IN/OUT: generated prime */
mpz_ptr		prime_counter		/* IN/OUT: last prime tried */
)
{
	mpz_set (p, prime_counter);

	do {
		mpz_add_ui (p, p, 2);
	} while (NOT Qax_is_prime (p));

	mpz_set (prime_counter, p);
}

/*
 * Return TRUE if-and-only-if p is prime.
 */

	static
	calib_bool
Qax_is_prime (

mpz_srcptr	p		/* IN: number to test */
)
{
int		flag;
calib_int64u	i;
calib_int64u	p64;

	if (mpz_cmp_ui (p, 0xFFFFFFFF) <= 0) {
		/* A 32-bit number. */
		p64 = mpz_get_ui (p);
		if (p64 <= 2) {
			return (p64 EQ 2);
		}
		if ((p64 & 0x01) EQ 0) return (FALSE);

		/* Check divisibility exhaustively. */
		for (i = 3; ; i += 2) {
			if (i * i > p64) return (TRUE);
			if ((p64 % i) EQ 0) return (FALSE);
		}
	}

	flag = mpz_probab_prime_p (p, 64);

	return (flag NE 0);
}

/*
 * Factorize apoly F(a) in Zp[x].  Fail if it is not square-free.
 */

	static
	struct pfact *
Qax_find_apoly_Zp_factors (

struct afglob *		afgp,		/* IN/OUT: global info */
mpz_srcptr		p		/* IN: prime p to use */
)
{
int				nfactors, square_free;
struct calib_Zp_dom *		Zp;
struct calib_Zpx_dom *		Zpx;
struct calib_Zpx_factor *	factors;
struct calib_Zpx_factor *	fp;
struct pfact *			pfact;
struct calib_Zpx_obj		apoly_zpx;

	Zp  = calib_make_Zp_dom (p);
	Zpx = calib_make_Zpx_dom (Zp);

	Zpx -> init (Zpx, &apoly_zpx);

	Zpx -> set_Zx (&apoly_zpx, afgp -> apoly);

	factors = Zpx -> factor (&apoly_zpx);

	nfactors = 0;
	square_free = TRUE;
	for (fp = factors; fp NE NULL; fp = fp -> next) {
		if (fp -> multiplicity NE 1) {
			CALIB_FATAL_ERROR_IF (fp -> multiplicity <= 0);
			square_free = FALSE;
			break;
		}
		++nfactors;
	}

	if (NOT square_free) {
		Zpx -> free_factors (factors);
		Zpx -> clear (&apoly_zpx);
		calib_free_Zpx_dom (Zpx);
		calib_free_Zp_dom (Zp);
		return (NULL);
	}

	pfact = CALIB_NEW (struct pfact);

	pfact -> nmf		= nfactors;
	mpz_init_set (pfact -> p, p);
	pfact -> Zp		= Zp;
	pfact -> Zpx		= Zpx;
	pfact -> apoly		= Zpx -> dup (&apoly_zpx);
	pfact -> factors	= factors;
	pfact -> gfinfo		= NULL;
	pfact -> next		= NULL;

	Zpx -> clear (&apoly_zpx);

	return (pfact);
}

/*
 * Free up a list of (mod p) apoly factorizations.
 */

	static
	void
Qax_free_apoly_Zp_factors (

struct pfact *		list	/* IN: list of factorizations to free */
)
{
struct pfact *		temp;
struct calib_Zpx_dom *	Zpx;

	while (list NE NULL) {
		temp = list -> next;
		list -> next = NULL;

		Qax_free_gfinfo (list -> nmf, list -> gfinfo);
		Zpx = list -> Zpx;
		Zpx -> free_factors (list -> factors);
		Zpx -> free (list -> apoly);
		calib_free_Zpx_dom (Zpx);
		calib_free_Zp_dom (list -> Zp);
		mpz_clear (list -> p);
		list -> nmf = 0;
		free (list);

		list = temp;
	}
}

/*
 * For each Zp[x] factor of apoly F(a), create a distinct GF(p^k)
 * coefficient field (we use the Zp[x] factor directly as the
 * generator polynomial for that GF(p^k) field), and factor fpoly f(x)
 * over the resulting GF(p^k)[x].
 *
 * If all of these fpoly factorizations are square-free, save them
 * off as pfact->gfinfo.  Otherwise, leave this field NULL.
 */

	static
	void
Qax_build_gfinfo (

struct afglob *		afgp,		/* IN/OUT: global info */
struct pfact *		pfact		/* IN/OUT: apoly Zp[x] factors */
)
{
int				i, nmf, nfact;
calib_bool			square_free;
const struct calib_Zax_dom *	R_of_x;
const struct calib_Zpx_dom *	Zpx;
const struct calib_Zpx_factor *	zpx_factors;
const struct calib_Zax_obj *	zax_poly;
const struct calib_Zp_dom *	cf;
const struct calib_Zpx_factor *	zpf;
struct gfinfo *			gfinfo;
struct gfinfo *			gip;
const struct calib_Zx_dom *	Zx;
const struct calib_Zpx_obj *	afactor_zpx;
struct calib_Zx_obj *		afactor_zx;
struct calib_Za_dom *		Za;
struct calib_Zax_dom *		Zax;
struct calib_GFpk_dom *		gfpk;
struct calib_GFpkx_dom *	gfpkx;
struct calib_GFpkx_obj *	fpoly;
struct calib_GFpkx_factor *	factors;
struct calib_GFpkx_factor *	fp;
struct calib_GFpkx_obj *	kfactor;
mpz_srcptr			d;

	R_of_x		= afgp -> Zax;
	nmf		= pfact -> nmf;
	Zpx		= pfact -> Zpx;
	zpx_factors	= pfact -> factors;
	zax_poly	= afgp -> zpoly;
	d		= afgp -> d;
	cf		= Zpx -> cf;
	Zx		= calib_get_Zx_dom ();

	/* Make an array of info for each of these factors. */
	gfinfo = CALIB_NEWA (nmf, struct gfinfo);

	memset (gfinfo, 0, nmf * sizeof (gfinfo [0]));

	square_free = TRUE;

	/* Now fill in one element of this array for each factor. */
	gip = gfinfo;
	for (zpf = zpx_factors; zpf NE NULL; zpf = zpf -> next, gip++) {
		afactor_zpx = zpf -> factor;

		gip -> Zpx	= Zpx;
		gip -> afactor_zpx = Zpx -> dup (afactor_zpx);

		/* Convert this factor from Zp[x] form into Z[x] form.	*/
		afactor_zx = Zx -> cvZpx (afactor_zpx);
		gip -> afactor_zx = afactor_zx;

		/* Create the Z(a) and Z(a)[x] domains for this factor.	*/
		Za = calib_make_Za_dom (afactor_zx);
		Zax = calib_make_Zax_dom (Za);

		gip -> Za  = Za;
		gip -> Zax = Zax;

		/* Create the coefficient field, K = GF(p^k). */
		gfpk = calib_make_GFpk_dom (cf, afactor_zpx);

		/* Create the polynomial ring GFp^k[x]. */
		gfpkx = calib_make_GFpkx_dom (gfpk);

		gip -> gfpk  = gfpk;
		gip -> gfpkx = gfpkx;

		/* Convert the polynomial into GFp^k[x] form. */

		fpoly = Qax_convert_to_GFpkx (gfpkx, R_of_x, zax_poly, d);

		gip -> fpoly_gfpkx = fpoly;

		/* Reduce the original Z(a)[x] polynomial modulo this	*/
		/* apoly factor and store it.				*/
		gip -> fpoly_zax = Zax -> cvZax (Zax, zax_poly);

		/* Now factor fpoly over this GFp^k[x]. */

		factors = gfpkx -> factor (fpoly);

		/* Count the factors and verify they are square-free. */
		nfact = 0;
		for (fp = factors; fp NE NULL; fp = fp -> next) {
			if (fp -> multiplicity > 1) {
				square_free = FALSE;
			}
			++nfact;
		}
		if (NOT square_free) {
			gfpkx -> free_factors (factors);
			break;
		}

		/* Store the factors in array form, in both	*/
		/* GF(p^k)[x] form, and in Z(a)[x] form.	*/
		gip -> nfact	= nfact;
		gip -> factors	= CALIB_NEWA (nfact, struct calib_GFpkx_obj *);
		gip -> lfactors	= CALIB_NEWA (nfact, struct calib_Zax_obj *);
		i = 0;
		for (fp = factors; fp NE NULL; fp = fp -> next) {
			kfactor = gfpkx -> dup (fp -> factor);
			gip -> factors [i] = kfactor;
			gip -> lfactors [i] = Zax -> cvGFpkx (Zax, kfactor);
			++i;
		}
		gfpkx -> free_factors (factors);
	}

	if (NOT square_free) {
		/* Modulus was bad.  Return an error code. */
		Qax_free_gfinfo (nmf, gfinfo);
		gfinfo = NULL;
	}

	pfact -> gfinfo = gfinfo;
}

/*
 * Convert the given polynomial zpoly = d*f(x) \in Z(a)[x] into the
 * corresponding member of GFp^k[x].  This requires modding out both
 * by p and by the generator polynomial for GFp^k.  Note that this
 * generator polynomial typically has lower degree than apoly which
 * defines Q(a).
 */

	static
	struct calib_GFpkx_obj *
Qax_convert_to_GFpkx (

const struct calib_GFpkx_dom *	gfpkx,	/* IN: poly ring to convert to */
const struct calib_Zax_dom *	R_of_x,	/* IN: poly ring to convert from */
const struct calib_Zax_obj *	zpoly,	/* IN: polynomial to convert */
mpz_srcptr			d	/* IN: common denominator */
)
{
int				i, k1, k2, deg;
const struct calib_Za_dom *	Za;
const struct calib_GFpk_dom *	gfpk;
const struct calib_Zp_dom *	Zp;
const struct calib_Za_obj *	srcp;
struct calib_GFpkx_obj *	poly;
struct calib_GFpk_obj *		coeffs;
struct calib_GFpk_obj *		dstp;
mpz_t				temp, dinv;

	CALIB_FATAL_ERROR_IF ((gfpkx  EQ NULL) OR
			      (R_of_x EQ NULL) OR
			      (zpoly  EQ NULL) OR
			      (d      EQ NULL));

	deg	= zpoly -> degree;
	Za	= R_of_x -> cr;
	k1	= Za -> gpoly -> degree;

	gfpk	= gfpkx -> cf;
	Zp	= gfpk -> cf;

	k2	= gfpk -> k;

	/* Degree of coefficients must not increase! */
	CALIB_FATAL_ERROR_IF (k2 > k1);

	/* Compute 1/d (mod p). */
	mpz_init (dinv);
	mpz_init (temp);
	Zp -> set_z (Zp, temp, d);
	Zp -> inv (Zp, dinv, temp);
	mpz_clear (temp);

	coeffs = Qax_new_GFpk_vector (gfpk, deg + 1);

	dstp = &coeffs [0];
	for (i = 0; i <= deg; i++) {
		srcp = &(zpoly -> coeff [i]);

		/* Convert this coefficient from Z(a) to GF(p^k). */
		gfpk -> cvZa (dstp, srcp);

		/* Multiply by (1/d) (mod p). */
		gfpk -> mul_z (dstp, dstp, dinv);

		++dstp;
	}

	mpz_clear (dinv);

	poly = CALIB_NEW (struct calib_GFpkx_obj);
	poly -> degree	= deg;
	poly -> size	= deg + 1;
	poly -> dom	= gfpkx;
	poly -> coeff	= coeffs;

	return (poly);
}

/*
 * Allocate and initialize an array of GF(p^k) objects.
 */

	static
	struct calib_GFpk_obj *
Qax_new_GFpk_vector (

const struct calib_GFpk_dom *	cf,	/* IN: Coefficient field */
int				n	/* IN: Number of elements */
)
{
int				i;
struct calib_GFpk_obj *		array;

	array = NULL;
	if (n > 0) {
		array = CALIB_NEWA (n, struct calib_GFpk_obj);
		for (i = 0; i < n; i++) {
			cf -> init (cf, &array [i]);
		}
	}

	return (array);
}

/*
 * Lift all of the following to a sufficiently high modulus > 2*B:
 *
 * 1. The factors of apoly, F_k(a).
 * 2. The factors of fpoly, f_{jk}(x).
 * 3. The wfactor: wfactor_i(a) = \prod_{k \ne i} F_k(a).
 * 4. The vfactor: \sum_i vfactor_i(a) * wfactor_i(a) == 1 (mod modulus).
 *
 * Let for all 0 <= k < nmf, let  F_k(a) be the final lifted factor of F(a).
 * Each F_k(a) should be irreducible over Z[x] because they are each
 * irreducible in Zp[x], and lifting to higher powers of p does not change
 * that fact.  The F_k(a) are therefore pairwise relatively prime members
 * of Z[x], and we can therefore apply the Chinese remainder theorem to
 * them.
 *
 * Let wfactor[k] = \product_{l \ne k} F_l(x).
 * Compute an array vfactor[k] of polynomials in Z[x] having the following
 * properties:
 *
 *	(vfactor_k * w_k) + (wfactor_k * F_k(x)) == 1 (mod modulus)
 *
 * These vfactor and wfactor polynomials are used to construct trial factors
 * for f(x) that are equivalent to each chosen trial factor modulo F_k(a).
 * These can only be correctly computed by starting with (mod p) versions
 * and then lifting them to higher powers.  There appear to be uniqueness
 * issues that spoil the broth if you try to compute them directly at
 * full modulus using extended Euclidean algorithm.
 *
 * FIXME - This would run much faster using modulus-squaring techniques.
 * Better yet would be to use a hybrid technique that uses both
 * incrementation and doubling of the modulus exponent intelligently to
 * precisely hit the smallest prime power we need without unnecessary
 * overshoot.
 */

	static
	void
Qax_lift_factors (

struct afglob *		afgp		/* IN/OUT: global info */
)
{
int				i, j, k, nmf, nf;
const struct calib_Zx_obj *	apoly;
const struct calib_Zpx_dom *	Zpx;
const struct calib_Zpx_obj *	apoly_zpx;
struct gfinfo *			gfinfo;
const struct calib_Zx_dom *	Zx;
struct calib_Zpx_obj *		afp;
struct calib_Zpx_obj **		afu;
struct gfinfo *			gip;
struct calib_GFpk_dom *		gfpk;
struct calib_GFpkx_dom *	gfpkx;
struct calib_GFpkx_obj ***	uarray;
struct calib_GFpkx_obj **	ulist;
struct calib_GFpkx_obj *	gfdelta;
struct calib_Zax_obj *		fdiff;
struct calib_Za_dom *		Za;
struct calib_Zax_dom *		Zax;
struct calib_Zx_obj *		zxpoly;
struct calib_Zx_obj *		adiff;
struct calib_Zax_obj **		lfactors;
struct calib_Zax_obj *		lfp;
struct calib_Za_obj *		lfcoeff;
struct calib_Zpx_obj		zpquot, zprem, zpdelta, G, XA, XB;
struct calib_Zax_obj		zax_prod, fdelta, zax_quot, zax_rem;
struct calib_Zx_obj		prod, adelta, Qzx, Rzx, sum;
mpz_srcptr			d, p, B;
mpz_ptr				modulus;
mpz_t				BOUND, dinv_modp, denom;

	apoly		= afgp -> apoly;
	Zpx		= afgp -> Zpx;
	apoly_zpx	= afgp -> apoly_zpx;
	nmf		= afgp -> nmf;
	gfinfo		= afgp -> gfinfo;
	d		= afgp -> d;
	p		= afgp -> p;
	B		= afgp -> B;
	modulus		= afgp -> modulus;

	Zx = calib_get_Zx_dom ();

	/* BOUND = K * (2*B).  The 2*B is to allow for both positive	*/
	/* and negative coefficients.  The K part is to provide some	*/
	/* extra "headroom" so that with high probability, coefficients	*/
	/* that are not part of a true factor will exceed 2*B in	*/
	/* magnitude, and can therefore be easily detected.  We use	*/
	/* K = 1024 = 2^10, recognizing that this gives us more than a	*/
	/* 1 in 1024 chance of detecting bogus coefficients because	*/
	/* there are many coefficients in each trial factor, and all of	*/
	/* these coefficients must pass the magnitude test before we do	*/
	/* the expensive division check in Q(a)[x].			*/

	mpz_init (BOUND);
	mpz_mul_2exp (BOUND, B, 1 + 10);

	mpz_set (modulus, p);

	mpz_init (dinv_modp);
	i = mpz_invert (dinv_modp, d, modulus);
	CALIB_FATAL_ERROR_IF (NOT i);

	/* Prepare list afu[] of polynomials for lifting factors of apoly. */
	Zpx -> init (Zpx, &zpquot);
	Zpx -> init (Zpx, &zprem);
	Zpx -> init (Zpx, &G);
	Zpx -> init (Zpx, &XA);
	Zpx -> init (Zpx, &XB);
	afu	 = CALIB_NEWA (nmf, struct calib_Zpx_obj *);
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		afp = gip -> afactor_zpx;
		Zpx -> div (&zpquot, &zprem, apoly_zpx, afp);
		/* Division must be exact. */
		CALIB_FATAL_ERROR_IF (zprem.degree >= 0);
		Zpx -> extgcd (&G, &XA, &XB, &zpquot, afp);
		CALIB_FATAL_ERROR_IF ((G.degree NE 0) OR
				      (mpz_cmp_ui (&G.coeff [0], 1) NE 0));
		afu [i] = Zpx -> dup (&XA);
	}
	Zpx -> clear (&XB);
	Zpx -> clear (&XA);
	Zpx -> clear (&G);

	/* Compute the h_{jk}(x) and u_{jk}(x), use to lift the f(x)	*/
	/* factors.							*/

	uarray = CALIB_NEWA (nmf, struct calib_GFpkx_obj **);
	for (i = 0; i < nmf; i++) {
		struct calib_GFpkx_obj		QUOT, REM, GCD, A, B;
		struct calib_GFpkx_obj *	kfactor;

		gip = &gfinfo [i];

		nf    = gip -> nfact;
		gfpk  = gip -> gfpk;
		gfpkx = gip -> gfpkx;

		gfpkx -> init (gfpkx, &QUOT);
		gfpkx -> init (gfpkx, &REM);
		gfpkx -> init (gfpkx, &GCD);
		gfpkx -> init (gfpkx, &A);
		gfpkx -> init (gfpkx, &B);

		ulist = CALIB_NEWA (nf, struct calib_GFpkx_obj *);
		for (j = 0; j < nf; j++) {
			kfactor = gip -> factors [j];

			/* Compute h_{jk}(x). */
			gfpkx -> div (&QUOT,
				      &REM,
				      gip -> fpoly_gfpkx,
				      kfactor);
			CALIB_FATAL_ERROR_IF (REM.degree >= 0);

			/* Compute u_{jk}(x). */
			gfpkx -> extgcd (&GCD, &A, &B, &QUOT, kfactor);
			CALIB_FATAL_ERROR_IF ((GCD.degree NE 0) OR
					      (NOT (gfpk -> onep (&GCD.coeff [0]))));

			ulist [j] = gfpkx -> dup (&A);
		}
		uarray [i] = ulist;

		gfpkx -> clear (&B);
		gfpkx -> clear (&A);
		gfpkx -> clear (&GCD);
		gfpkx -> clear (&REM);
		gfpkx -> clear (&QUOT);
	}

	/* Initialize the vfactor and wfactor for (modulus EQ p^1) state. */
	Zx -> init_si (&prod, 1);
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		/* Convert apoly factor from Zp[x] to Z[x]. */
		zxpoly = Zx -> cvZpx (gip -> afactor_zpx);
		Zx -> mul (&prod, &prod, zxpoly);
		Zx -> free (zxpoly);
	}
	mpz_init (denom);
	Zx -> init (&Qzx);
	Zx -> init (&Rzx);
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		/* Convert apoly factor from Zp[x] to Z[x]. */
		zxpoly = Zx -> cvZpx (gip -> afactor_zpx);

		Zx -> div (&Qzx, &Rzx, denom, &prod, zxpoly);
		/* Division must be exact. */
		CALIB_FATAL_ERROR_IF (Rzx.degree >= 0);

		gip -> vfactor = Zx -> cvZpx (afu [i]);
		gip -> wfactor = Zx -> dup (&Qzx);
	}
	Zx -> clear (&Rzx);
	Zx -> clear (&Qzx);
	mpz_clear (denom);
	Zx -> clear (&prod);

	/* The main lifting loop. */

	Zx -> init (&adelta);
	while (mpz_cmp (modulus, BOUND) <= 0) {
		/* First lift the factors of apoly. */
		Zx -> init_si (&prod, 1);
		for (i = 0; i < nmf; i++) {
			gip = &gfinfo [i];
			Zx -> mul (&prod, &prod, gip -> afactor_zx);
		}
		CALIB_FATAL_ERROR_IF (prod.degree NE apoly -> degree);
		Zx -> sub (&adelta, apoly, &prod);
		Zx -> div_z_exact (&adelta, &adelta, modulus);
		Zpx -> init (Zpx, &zpdelta);
		Zpx -> set_Zx (&zpdelta, &adelta);
		for (i = 0; i < nmf; i++) {
			gip = &gfinfo [i];
			afp = gip -> afactor_zpx;
			Zpx -> mul (&zpquot, &zpdelta, afu [i]);
			Zpx -> div (&zpquot, &zprem, &zpquot, afp);
			adiff = Zx -> cvZpx (&zprem);
			Zx -> mul_z (adiff, adiff, modulus);
			Zx -> add (gip -> afactor_zx, gip -> afactor_zx, adiff);
			Zx -> free (adiff);
		}
		Zpx -> clear (&zpdelta);

		/* Recompute the wfactor from the lifted apoly factors. */
		Zx -> init_si (&prod, 1);
		for (i = 0; i < nmf; i++) {
			gip = &gfinfo [i];
			Zx -> mul (&prod, &prod, gip -> afactor_zx);
		}
		mpz_init (denom);
		Zx -> init (&Rzx);
		for (i = 0; i < nmf; i++) {
			gip = &gfinfo [i];
			Zx -> div (gip -> wfactor, &Rzx, denom,
				   &prod, gip -> afactor_zx);
			/* Division must be exact. */
			CALIB_FATAL_ERROR_IF ((Rzx.degree >= 0) OR
					      (mpz_cmp_ui (denom, 1) NE 0));
		}
		mpz_clear (denom);
		Zx -> clear (&Rzx);
		/* Now lift the vfactor using the lifted apoly factors,	*/
		/* and the updated wfactor.				*/
		Zx -> init_si (&sum, 1);
		for (i = 0; i < nmf; i++) {
			gip = &gfinfo [i];
			Zx -> mul (&prod, gip -> vfactor, gip -> wfactor);
			Zx -> sub (&sum, &sum, &prod);
		}
		Zx -> clear (&prod);
		Zx -> div_z_exact (&sum, &sum, modulus);
		Zpx -> init (Zpx, &zpdelta);
		Zpx -> set_Zx (&zpdelta, &sum);
		Zx -> clear (&sum);
		for (i = 0; i < nmf; i++) {
			gip = &gfinfo [i];
			Zpx -> mul (&zprem, &zpdelta, afu [i]);
			Zpx -> div (NULL, &zprem, &zprem, gip -> afactor_zpx);
			zxpoly = Zx -> cvZpx (&zprem);
			Zx -> mul_z (zxpoly, zxpoly, modulus);
			Zx -> add (gip -> vfactor, gip -> vfactor, zxpoly);
			Zx -> free (zxpoly);
		}
		Zpx -> clear (&zpdelta);

		/* Now lift all the factors of fpoly. */
		for (i = 0; i < nmf; i++) {
			gip = &gfinfo [i];
			/* Use *CALIB_NEWLY LIFTED* apoly factor as domain. */
			Za  = calib_make_Za_dom (gip -> afactor_zx);
			Zax = calib_make_Zax_dom (Za);

			/* Replace old domains with new lifted ones. */
			calib_free_Zax_dom (gip -> Zax);
			calib_free_Za_dom (gip -> Za);
			gip -> Zax = Zax;
			gip -> Za  = Za;

			/* Force new domains into lfactors[] and all of	*/
			/* their coefficients.				*/
			nf = gip -> nfact;
			lfactors = gip -> lfactors;
			for (j = 0; j < nf; j++) {
				lfp = lfactors [j];
				lfp -> dom = Zax;
				lfcoeff = lfp -> coeff;
				for (k = 0; k <= lfp -> degree; k++) {
					lfcoeff [k].dom = Za;
				}
			}
			/* Force new domains into gip->fpoly_zax. */
			lfp = gip -> fpoly_zax;
			lfp -> dom = Zax;
			lfcoeff = lfp -> coeff;
			for (k = 0; k <= lfp -> degree; k++) {
				lfcoeff [k].dom = Za;
			}

			Zax -> init (Zax, &fdelta);
			Zax -> init (Zax, &zax_quot);
			Zax -> init (Zax, &zax_rem);

			Zax -> init_si (Zax, &zax_prod, 1);
			for (j = 0; j < nf; j++) {
				Zax -> mul (&zax_prod, &zax_prod, lfactors [j]);
			}
			Zax -> mul_z (&zax_prod, &zax_prod, d);
			Zax -> sub (&fdelta, gip -> fpoly_zax, &zax_prod);
			Zax -> div_z_exact (&fdelta, &fdelta, modulus);
			gfpkx = gip -> gfpkx;
			gfdelta = gfpkx -> cvZax (gfpkx, &fdelta);
			gfpkx -> mul_z (gfdelta, gfdelta, dinv_modp);
			struct calib_GFpkx_obj	TEMP, Q, R;
			gfpkx -> init (gfpkx, &TEMP);
			gfpkx -> init (gfpkx, &Q);
			gfpkx -> init (gfpkx, &R);
			ulist = uarray [i];
			for (j = 0; j < nf; j++) {
				gfpkx -> mul (&TEMP, gfdelta, ulist [j]);
				gfpkx -> div (&Q,
					      &R,
					      &TEMP,
					      gip -> factors [j]);
				fdiff = Zax -> cvGFpkx (Zax, &R);
				Zax -> mul_z (fdiff, fdiff, modulus);
				Zax -> add (lfactors [j],
					    lfactors [j],
					    fdiff);
				Zax -> free (fdiff);
			}
			gfpkx -> clear (&R);
			gfpkx -> clear (&Q);
			gfpkx -> clear (&TEMP);
			gfpkx -> free (gfdelta);
			Zax -> clear (&zax_prod);

			Zax -> clear (&zax_rem);
			Zax -> clear (&zax_quot);
			Zax -> clear (&fdelta);
		}

		mpz_mul (modulus, modulus, p);
	}

	Zx -> clear (&adelta);
	Zpx -> clear (&zprem);
	Zpx -> clear (&zpquot);

	/* Initialize our lf_try / lf_used flags. */
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		nf = gip -> nfact;
		gip -> lf_try = CALIB_NEWA (nf, calib_bool);
		gip -> lf_used = CALIB_NEWA (nf, calib_bool);
		memset (gip -> lf_try,  FALSE, nf * sizeof (calib_bool));
		memset (gip -> lf_used, FALSE, nf * sizeof (calib_bool));
	}

	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		gfpkx = gip -> gfpkx;
		ulist = uarray [i];
		uarray [i] = NULL;
		for (j = 0; j < gip -> nfact; j++) {
			gfpkx -> free (ulist [j]);
			ulist [j] = NULL;
		}
		free (ulist);
	}
	free (uarray);

	for (i = 0; i < nmf; i++) {
		Zpx -> free (afu [i]);
	}
	free (afu);

	mpz_clear (dinv_modp);
	mpz_clear (BOUND);
}

/*
 * Top level for the trial division portion of the Q(a)[x] factorization
 * algorith, which is multiply recursive in the following respects:
 *
 *	1. We iterate for each degree h from 1, 2, ... until 2*h > deg(f(x)).
 *
 *	2. Recursively process each ginfo [i], for 0 <= i < nmf.
 *
 *	3. For each ginfo [i], recursively generate all all combinations
 *	   of ginfo[i].lfactors that have total degree of h.  (There may be
 *	   no such combinations, or many of them).  Compute the degree h
 *	   product of each such combination, and recurse to the next
 *	   gfinfo [i+1].
 *
 *	4. Apply the Chinese remainder theorem to combine the nmf
 *	   products (each of degree h) into a single degree h polynomial
 *	   that is correct modulo F(A).  Convert this polynomial from
 *	   (1/d) Z(a)[x] form into Q(a)[x] form.  Then perform trial
 *	   division to see if it is a true factor.  If so, divide this
 *	   factor out of the polynomial being factored, and append it
 *	   onto the list of true Q(a)[x] factors.
 */

	static
	void
Qax_try_all_trial_factors (

struct afglob *		afgp		/* IN/OUT: global info */
)
{
int				i, j, nmf, nf, nstack, gen_degree;
struct gfinfo *			gfinfo;
struct gfinfo *			gip;
struct calib_Zax_dom *		Zax;
struct calib_Qax_factor *	fp;
struct calib_Zax_obj *		product_stack;
struct calib_Zax_obj *		psp;
struct calib_Zax_obj *		f1;
struct calib_Zax_obj *		f2;

	Qax_init_CRT_buffers (afgp);

	/* As we recursively include/exclude individual factors into	*/
	/* our trial factor combination, we need to compute the product	*/
	/* of the included factors.  At the base-case of the recursion	*/
	/* we will have one product polynomial for each apoly factor.	*/
	/* Rather than multiplying these out in full at each base case,	*/
	/* we incrementally compute these factors each time we choose	*/
	/* to "include" a factor in the combination.  To make this	*/
	/* more efficient, we create some data structures:		*/
	/*	- A stack of product polynomials, one for each apoly	*/
	/*	  factor.						*/
	/*	- An array of pointers (one per apoly factor) that	*/
	/*	  gives the full trial factor polynomial (product of	*/
	/*	  all chosen trial factors).				*/
	/* An apoly factor F_k(a) having j fpoly factors will need at	*/
	/* most j+1 elements in its stack (one for the "empty" product	*/
	/* whose value is 1), plus one element for each of the j	*/
	/* possible fpoly factors.  We allocate all of these stacks as	*/
	/* a single array and parcel the correct numer of them out to	*/
	/* each apoly factor.						*/

	nmf	= afgp -> nmf;
	gfinfo	= afgp -> gfinfo;
	Zax	= afgp -> Zax;

	nstack	= 0;
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		j = gip -> nfact;
		nstack += (1 + j);
	}
	product_stack = CALIB_NEWA (nstack, struct calib_Zax_obj);
	psp = product_stack;
	for (i = 0; i < nstack; i++, psp++) {
		Zax -> init (Zax, psp);
	}
	psp = product_stack;
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		j = gip -> nfact;

		/* Initialize "base" product to 1. */
		Zax -> set_si (psp, 1);

		gip -> prod_stack_base	= psp;
		gip -> trial_factor	= NULL;

		psp += (1 + j);
	}

	/* Transform all of the lifted factors into top-level Zax domain. */
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		nf = gip -> nfact;
		for (j = 0; j < nf; j++) {
			f1 = gip -> lfactors [j];
			f2 = Zax -> cvZax (Zax, f1);
			gip -> Zax -> free (f1);
			gip -> lfactors [j] = f2;
		}
	}
	/* All of the trial factors now have the same Zax domain as	*/
	/* the stack into which we will multiply them together.		*/

	for (gen_degree = 1; ; gen_degree++) {
		if (2 * gen_degree > afgp -> lpoly.degree) {
			/* No need to try any larger degree.  If we did	*/
			/* try factors this big, the "other" factor	*/
			/* would be of a smaller degree that we have	*/
			/* already fully enumerated.			*/
			break;
		}
		afgp -> gen_degree = gen_degree;


		if (afgp -> K_of_x -> settings.WR_print_trial_factor_combinations) {
			printf ("Degree %d:\n", gen_degree);
		}

		/* Start with gfinfo [i], where i = 0. */
		i = 0;
		Qax_try_recurse_1 (afgp, i);
	}

	if (afgp -> lpoly.degree > 0) {
		/* Record final, residual factor. */
		fp = CALIB_NEW (struct calib_Qax_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= Qax_dup (&(afgp ->  lpoly));
		fp -> next		= NULL;
		*(afgp -> true_factor_hookp) = fp;
		afgp -> true_factor_hookp = &(fp -> next);
	}

	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		gip -> prod_stack_base	= NULL;
		gip -> trial_factor	= NULL;
	}
	psp = product_stack;
	for (i = 0; i < nstack; i++, psp++) {
		Zax -> clear (psp);
	}
	free (product_stack);

	Qax_free_CRT_buffers (afgp);
}

/*
 * Recursively process gfinfo[i].  Returns TRUE if-and-only-if a true
 * factor was discovered during the invocation.
 */

	static
	calib_bool
Qax_try_recurse_1 (

struct afglob *		afgp,		/* IN/OUT: global info */
int			i		/* IN: index into gfinfo */
)
{
int			j;
calib_bool		success;
struct gfinfo *		gip;
struct calib_Zax_obj *	empty_prod;

	if (i >= afgp -> nmf) {
		/* Base case!						*/
		/* We have assembled one trial factor (of degree	*/
		/* gen_degree) for each gfinfo[] entry (i.e., for each	*/
		/* factor of "apoly" or F(a)).  Combine these together	*/
		/* using the Chinese remainder theorem, convert them	*/
		/* into Q(a)[x] form, and perform trial division.	*/
		success = Qax_combine_convert_and_test_trial_factors (afgp);
	}
	else {
		/* Process this gfinfo[i] entry. */
		gip = &(afgp -> gfinfo [i]);

		/* Recursively generate all combinations of the		*/
		/* remaining gip -> lfactors[] having total degree EQ	*/
		/* gen_degree.						*/

		empty_prod = gip -> prod_stack_base;
		j = 0;
		success = Qax_try_recurse_2 (afgp, i, j, empty_prod);
	}

	return (success);
}

/*
 * Recurse across all combinations of remaining gip -> lfactors[] having
 * total degree of gen_degree.  Returns TRUE if-and-only-if a true
 * factor was discovered during the invocation.
 */

	static
	calib_bool
Qax_try_recurse_2 (

struct afglob *		afgp,		/* IN/OUT: global info */
int			i,		/* IN: index into gfinfo */
int			j,		/* IN: lfactor[j] to try next */
struct calib_Zax_obj *	prod		/* IN: partial product */
)
{
calib_bool		success;
struct gfinfo *		gip;
struct calib_Zax_dom *	Zax;
struct calib_Zax_obj *	f;

	gip = &(afgp -> gfinfo [i]);

	if (prod -> degree >= afgp -> gen_degree) {
		/* Base case 1: we have a trial factor of proper degree. */
		/* Record this trial factor and advance to next		 */
		/* gfinfo[] entry.					 */

		CALIB_FATAL_ERROR_IF (prod -> degree > afgp -> gen_degree);
		gip -> trial_factor = prod;

		success = Qax_try_recurse_1 (afgp, i + 1);
		return (success);
	}

	/* Current trial factor is not of sufficiently high degree yet. */

	if (j >= gip -> nfact) {
		/* Base case 2.  No additional lfactors[] to try. */
		return (FALSE);
	}

	f = gip -> lfactors [j];

	if ((NOT (gip -> lf_used [j])) AND
	    (prod -> degree + f -> degree <= afgp -> gen_degree)) {
		/* This factor can be included.  Do so now. */
		Zax = gip -> Zax;

		Zax -> mul (prod + 1, prod, f);

		gip -> lf_try [j] = TRUE;

		success = Qax_try_recurse_2 (afgp, i, j + 1, prod + 1);

		gip -> lf_try [j] = FALSE;

		if (success AND (i > 0)) {
			return (TRUE);
		}
	}

	/* Also try NOT including this factor in the product. */

	success = Qax_try_recurse_2 (afgp, i, j + 1, prod);

	return (success);
}

/*
 * We have assembled each gfinfo[i].trial_factor (of degree gen_degree).
 * We now:
 *
 * 1. Combine these together into a single Z(a)[x] trial factor
 *    using the Chinese remainder theorem.
 *
 * 2. Convert it from (1/d) Z(a)[x] form into Q(a)[x] form.
 *
 * 3. Perform trial division (lpoly divided by this trial factor).
 *
 * If the trial division is exact, then we:
 *	- store the Q(a)[x] trial factor as a true f(x) factor,
 *	- remove this factor from lpoly,
 *	- mark the various components comprising this trial factor
 *	  as "used" so that they are not included within any future
 *	  trial factors generated.
 *
 * Returns TRUE if-and-only-if a true factor was discovered.
 */

	static
	calib_bool
Qax_combine_convert_and_test_trial_factors (

struct afglob *		afgp	/* IN/OUT: global info */
)
{
int				i, j, nmf, gen_degree;
calib_bool			success;
calib_bool			const_term_division_exact;
const struct calib_Qax_dom *	K_of_x;
struct calib_Qax_obj *		lpoly;
struct gfinfo *			gfinfo;
struct gfinfo *			gip;
struct calib_Zax_obj *		crt_factor;
struct calib_Qax_obj *		qfactor;
struct calib_Qax_factor *	fp;
struct calib_Qax_obj		Q_qax, R_qax;

	K_of_x		= afgp -> K_of_x;
	lpoly		= &(afgp -> lpoly);
	nmf		= afgp -> nmf;
	gfinfo		= afgp -> gfinfo;
	gen_degree	= afgp -> gen_degree;
	crt_factor	= afgp -> crt_factor;

	/* Make sure all trial factors have the correct degree. */
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		CALIB_FATAL_ERROR_IF (gip -> trial_factor -> degree NE gen_degree);
	}

	if (K_of_x -> settings.WR_print_trial_factor_combinations) {
		Qax_print_trial_factor_combination_string (afgp);
	}

	/* For the sake of efficiency, we construct the coefficients	*/
	/* of the trial factor (via the Chinese Remainder Theorem) one	*/
	/* by one.  If any of these exceed the magnitude that any	*/
	/* coefficient is supposed to be, then we can know without any	*/
	/* further effort that this trial factor cannot be a true	*/
	/* factor.  This saves significant computational effort.  Also,	*/
	/* we construct the coefficients from degree 0 first, on up to	*/
	/* gen_degree.  The constant coefficient is the one most	*/
	/* likely to violate the bound (e.g., the lead coefficient is	*/
	/* always going to be 1).					*/

	for (j = 0; j <= gen_degree; j++) {
		success = Qax_CRT_one_coefficient (afgp, j);
		if (NOT success) {
			/* Magnitude too big -- this cannot be a true	*/
			/* factor.					*/
			return (FALSE);
		}
	}
	crt_factor -> degree = gen_degree;

	/* Convert this (1/d)Z(a)[x] into Q(a)[x] form. */

	qfactor = Qax_convert_Zax_to_Qax (afgp, crt_factor);

	success = FALSE;

	/* Perform trial division.  Test the constant term first. */

#if 0
	/* This doesn't work in Q(a) because it has an inv() operation,	*/
	/* not a div() operation -- division is always exact in a	*/
	/* field!  The solution is to figure out how to test this in	*/
	/* (1/d)Z(a)[x] before we convert it.  This is all just an	*/
	/* efficiency hack -- disable it for now.			*/
	/* FIXME -- figure out how to do this correctly.  One approach	*/
	/* would be to examine the constant term of crt_factor (which	*/
	/* is an element of Z(a)[x]): if the sum of the magnitude of	*/
	/* its coefficients exceeds d*B2, then it cannot be a true	*/
	/* factor of f(x).						*/
    {	int				k;
	const struct calib_Qa_dom *	Qa;
	mpq_ptr				Q_qa, R_qa;
	  
		Qa = K_of_x -> cf;
		k = Qa -> gpoly -> degree;
		Q_qa = calib_new_Q_vector (2 * k);
		R_qa  = Q_qa + k;
		/* This div() function doesn't exist, only inv()!	*/
		/* Q(a) is a field, and division is always exact!	*/
		Qa -> div (Qa, Q_qa, R_qa, lpoly -> coeff, qfactor.coeff);
		const_term_division_exact = Qa -> zerop (Qa, R_qa);
		calib_free_Q_vector (Q_qa, 2 * k);
    }
#else
	/* Temporary workaround. */
	const_term_division_exact = TRUE;
#endif

	if (const_term_division_exact) {
		/* Constant factor divides properly!  Try the	*/
		/* entire polynomial.				*/
		Qax_init (K_of_x, &Q_qax);
		Qax_init (K_of_x, &R_qax);
		Qax_div (&Q_qax, &R_qax, lpoly, qfactor);
		if (R_qax.degree < 0) {
			/* Division is exact!  We have a true	*/
			/* factor!  Record it.			*/
			fp = CALIB_NEW (struct calib_Qax_factor);
			fp -> multiplicity	= 1;
			fp -> factor		= Qax_dup (qfactor);
			fp -> next		= NULL;
			*(afgp -> true_factor_hookp) = fp;
			afgp -> true_factor_hookp = &(fp -> next);

			/* Update remaining un-factored poly. */
			Qax_set (lpoly, &Q_qax);

			/* Mark all of the pieces that comprise	*/
			/* this true factor as being "used", so	*/
			/* that we don't try to re-use them	*/
			/* within future trial factors.		*/
			Qax_mark_trial_factors_as_used (afgp);
			if (K_of_x -> settings.WR_print_trial_factor_combinations) {
				printf ("\t** TRUE FACTOR **\n");
			}
			success = TRUE;
		}
		Qax_clear (&R_qax);
		Qax_clear (&Q_qax);
	}
	Qax_free (qfactor);

	return (success);
}

/*
 * Construct a single coefficient of the current trial factor using
 * the Chinese Remainder Theorem -- specifically, the coefficient
 * for the x^j term.
 *
 * The constructed coefficient is simultaneously correct modulo *all*
 * of the lifted apoly factors F_K(a).
 *
 * The coefficient is constructed into afgp->crt_factor.
 *
 * Returns TRUE if-and-only-if the constructed coefficients all have
 * magnitude at most B.  (If they exceed this bound, it means that
 * this cannot be a coefficient of any true factor.)
 */

	static
	calib_bool
Qax_CRT_one_coefficient (

struct afglob *		afgp,	/* IN/OUT: global info */
int			j	/* IN: which x^j coefficient to construct */
)
{
int				i, k1, nmf, sdeg;
struct gfinfo *			gip;
struct gfinfo *			gfinfo;
const struct calib_Zx_dom *	Zx;
struct calib_Zax_dom *		Zax;
const struct calib_Za_dom *	Za;
struct calib_Zax_obj *		fp;
mpz_srcptr			d, modulus, half_modulus;
struct calib_Za_obj *		dstp;
struct calib_Za_obj *		srcp;
mpz_ptr				zp;
mpz_ptr				denom, B;
struct calib_Zx_obj		*SUM, *P2, *P3, *Q, *R;
struct calib_Zx_obj		P1;

	CALIB_FATAL_ERROR_IF ((afgp EQ NULL) OR
			      (j < 0) OR
			      (j > afgp -> gen_degree));

	nmf	 	= afgp -> nmf;
	gfinfo	 	= afgp -> gfinfo;
	k1	 	= afgp -> apoly -> degree;
	d		= afgp -> d;
	modulus	 	= afgp -> modulus;
	half_modulus	= afgp -> half_modulus;
	B		= afgp -> B;

	denom		= afgp -> crt_denom;
	SUM		= afgp -> crt_SUM;
	P2		= afgp -> crt_P2;
	P3		= afgp -> crt_P3;
	Q		= afgp -> crt_Q;
	R		= afgp -> crt_R;

	Zx = calib_get_Zx_dom ();

	/* Reconstruct the specified coefficient of the polynomial. */
	dstp = &(afgp -> crt_factor -> coeff [j]);

	SUM -> degree = -1;	/* Force SUM to zero. */
	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		Zax = gip -> Zax;
		Za  = Zax -> cr;
		fp = gip -> trial_factor;

		/* Get the coefficient of the x^j term of the	*/
		/* trial factor of the i-th apoly factor in	*/
		/* Z[x] form in P1.  Note that P1 shares part	*/
		/* of fp's coeff vector!  This is OK because	*/
		/* Zx->mul() only uses its input operands in a	*/
		/* read-only manner.				*/

		srcp = &(fp -> coeff [j]);

		P1.degree = Za -> degree (srcp);
		P1.size	  = P1.degree + 1;
		P1.coeff  = &(srcp -> coeff [0]);
		Zx -> mul (P2, &P1, gip -> vfactor);
		memset (&P1, 0, sizeof (P1));

		Zx -> div (Q, R, denom, P2, gip -> afactor_zx);
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (denom, 1) NE 0);

		Zx -> mul (P3, R, gip -> wfactor);
		Zx -> add (SUM, SUM, P3);
	}
	sdeg = SUM -> degree;
	CALIB_FATAL_ERROR_IF (sdeg >= k1);

	/* Multiply by d and reduce by modulus. */
	zp = &(SUM -> coeff [0]);
	for (i = 0; i <= sdeg; i++, zp++) {
		mpz_mul (zp, d, zp);
		mpz_mod (zp, zp, modulus);
		if (mpz_cmp (zp, half_modulus) > 0) {
			mpz_sub (zp, zp, modulus);
		}
		mpz_abs (denom, zp);
		if (mpz_cmp (denom, B) > 0) {
			/* Magnitude is too big to be a true factor. */
			return (FALSE);
		}
	}
	calib_copy_Z_vector (&(dstp -> coeff [0]), SUM -> coeff, sdeg + 1);

	/* Force all other higher-order coefficients to zero. */
	for (i = sdeg + 1; i < k1; i++) {
		mpz_set_ui (&(dstp -> coeff [i]), 0);
	}

	/* Success! */
	return (TRUE);
}

/*
 * Convert the given Z(a)[x] factor into Q(a)[x] form.
 */

	static
	struct calib_Qax_obj *
Qax_convert_Zax_to_Qax (

struct afglob *		afgp,	/* IN: global info */
const struct calib_Zax_obj *
			zpoly	/* IN: polynomial in Z(a)[x] to convert */
)
{
int				i, j, k1, k2, deg;
const struct calib_Qax_dom *	Qax;
const struct calib_Qa_dom *	Qa;
const struct calib_Zax_dom *	Zax;
const struct calib_Za_dom *	Za;
struct calib_Qax_obj *		qpoly;
struct calib_Za_obj *		srcp;
mpz_srcptr			d, modulus, half_modulus;
struct calib_Qa_obj *		coeff;
mpz_srcptr			szp;
mpz_t				TEMP;
mpq_t				QFACT;
struct calib_Za_obj		ZAVAL;

	Qax		= afgp -> K_of_x;
	Qa		= Qax  -> cf;
	Zax		= afgp -> Zax;
	Za		= Zax -> cr;
	d		= afgp -> d;
	modulus		= afgp -> modulus;
	half_modulus	= afgp -> half_modulus;

	k1	= Qa -> gpoly -> degree;
	k2	= Za -> gpoly -> degree;
	deg	= zpoly -> degree;

	CALIB_FATAL_ERROR_IF (k1 NE k2);

	mpz_init (TEMP);
	mpq_init (QFACT);

	Za -> init (Za, &ZAVAL);

	mpq_set_ui (QFACT, 1, 1);
	calib_mpq_div_z (QFACT, QFACT, d);

	coeff	= Qax_new_Qa_vector (Qa, deg + 1);
	for (i = 0; i <= deg; i++) {
		srcp = &(zpoly -> coeff [i]);
		szp = &(srcp -> coeff [0]);
		for (j = 0; j < k1; j++, szp++) {
			/* crt_factor has already been multiplied by d	*/
			/* and reduced (mod modulus).			*/
			mpz_set (TEMP, szp);
			if (mpz_cmp (TEMP, half_modulus) > 0) {
				mpz_sub (TEMP, TEMP, modulus);
			}
			mpz_set (&ZAVAL.coeff [j], TEMP);
		}
		Qa -> set_Za_q (&coeff [i], srcp, QFACT);
	}

	mpz_clear (TEMP);

	qpoly = CALIB_NEW (struct calib_Qax_obj);
	qpoly -> degree	= deg;
	qpoly -> size	= deg + 1;
	qpoly -> dom	= Qax;
	qpoly -> coeff	= coeff;

	return (qpoly);
}

/*
 * Allocate and initialize all of the buffers used and reused over and over
 * again by the inner-loops of the Chinese Remainder Theorem stuff.
 */

	static
	void
Qax_init_CRT_buffers (

struct afglob *	afgp		/* IN/OUT: global info */
)
{
int				i, deg;
struct calib_Zax_dom *		Zax;
const struct calib_Zx_dom *	Zx;
struct calib_Zax_obj *		zpoly;
struct calib_Zx_obj *		pp;

	Zax	= afgp -> Zax;
	zpoly	= afgp -> zpoly;
	deg	= afgp -> apoly -> degree;
	Zx	= calib_get_Zx_dom ();

	/* By using dup() on zpoly, we are forcing these polynomials	*/
	/* to all have internal buffers that are already fully		*/
	/* allocated up to the degree we need.  We can access each	*/
	/* coefficient and be assured that it is already there.		*/

	afgp -> crt_factor = Zax -> dup (zpoly);

	pp = CALIB_NEWA (5, struct calib_Zx_obj);
	for (i = 0; i < 5; i++) {
		Zx -> init (&pp [i]);
		Zx -> alloc (&pp [i], deg);
	}
	afgp -> crt_SUM	= pp++;
	afgp -> crt_P2	= pp++;
	afgp -> crt_P3	= pp++;
	afgp -> crt_Q	= pp++;
	afgp -> crt_R	= pp++;
	/* crt_denom is allocated already. */
}

/*
 * Free up all of the buffers used by the Chinese Remainder Theorem stuff.
 */

	static
	void
Qax_free_CRT_buffers (

struct afglob *	afgp		/* IN/OUT: global info */
)
{
int				i;
struct calib_Zax_dom *		Zax;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_obj *		pp;

	Zax	= afgp -> Zax;
	Zx	= calib_get_Zx_dom ();

	/* crt_denom is freed later. */
	pp = afgp -> crt_SUM;

	afgp -> crt_SUM	= NULL;
	afgp -> crt_P2	= NULL;
	afgp -> crt_P3	= NULL;
	afgp -> crt_Q	= NULL;
	afgp -> crt_R	= NULL;

	for (i = 0; i < 5; i++) {
		Zx -> clear (&pp [i]);
	}
	free (pp);

	Zax -> free (afgp -> crt_factor);
	afgp -> crt_factor = NULL;
}

/*
 * Mark all of the pieces that comprise the current "true" factor as
 * being "used", so that we don't try to re-use them within future
 * trial factors that we generate.
 */

	static
	void
Qax_mark_trial_factors_as_used (

struct afglob *		afgp	/* IN/OUT: global info */
)
{
int			i, j, nmf, nfact;
struct gfinfo *		gfinfo;
struct gfinfo *		gip;

	nmf	= afgp -> nmf;
	gfinfo	= afgp -> gfinfo;

	for (i = 0; i < nmf; i++) {
		gip = &gfinfo [i];
		nfact = gip -> nfact;
		for (j = 0; j < nfact; j++) {
			if (gip -> lf_try [j]) {
				/* This factor is part of the trial	*/
				/* factor that we tried, and that	*/
				/* turned out to be a "true" factor.	*/

				/* Better not already be marked as "used"! */
				CALIB_FATAL_ERROR_IF (gip -> lf_used [j]);

				gip -> lf_used [j] = TRUE;
			}
		}
	}
}

/*
 * Print out the initial gfinfo -- the original F(a) (mod p) and
 * f(x) (GF(p^k)) factorizations.
 */

	static
	void
Qax_print_initial_gfinfo (

struct afglob *	afgp,		/* IN: global info */
const char *	xvar,		/* IN: main polynomial var */
const char *	avar		/* IN: algebraic extension var */
)
{
int			i, nmf;
struct gfinfo *		gfinfo;

	nmf	= afgp -> nmf;
	gfinfo	= afgp -> gfinfo;

	for (i = 0; i < nmf; i++) {
		printf ("Modulular factor %d:\n", i);
		Qax_print_one_initial_gfinfo (&gfinfo [i], xvar, avar);
	}
}

/*
 * Print out the "initial" portion of a single gfinfo object --
 * i.e., just the original F(a) (mod p) and f(x) (GF(p^k)) factorizations.
 */

	static
	void
Qax_print_one_initial_gfinfo (

const struct gfinfo *	gip,	/* IN: gfinfo object to print */
const char *		xvar,	/* IN: main polynomial var */
const char *		avar	/* IN: algebraic extension var */
)
{
int				j;
struct calib_genrep *		gp1;
const struct calib_Zpx_dom *	Zpx;
const struct calib_GFpkx_dom *	gfpkx;
const struct calib_GFpkx_obj *	factor;

	Zpx	= gip -> Zpx;
	gfpkx	= gip -> gfpkx;

	gp1 = Zpx -> to_genrep (gip -> afactor_zpx, avar);
	calib_genrep_fprint (stdout, gp1);
	calib_genrep_free (gp1);

	printf ("  K[x] Poly:\n");
	gp1 = gfpkx -> to_genrep (gip -> fpoly_gfpkx, xvar, avar);
	calib_genrep_fprint (stdout, gp1);
	calib_genrep_free (gp1);

	printf ("  K[x] Factors:\n");
	for (j = 0; j < gip -> nfact; j++) {
		factor = gip -> factors [j];
		gp1 = gfpkx -> to_genrep (factor, xvar, avar);
		calib_genrep_fprint (stdout, gp1);
		calib_genrep_free (gp1);
	}
}

/*
 * Print out the "lifted" gfinfo -- i.e., the lifted F(a) (mod p^j) and
 * f(x) (mod F_k(a),p^j) factorizations.
 */

	static
	void
Qax_print_lifted_gfinfo (

struct afglob *	afgp,		/* IN: global info */
const char *	xvar,		/* IN: main polynomial var */
const char *	avar		/* IN: algebraic extension var */
)
{
int		i, nmf;
struct gfinfo *	gfinfo;

	printf ("\nLifted Modulus = ");
	mpz_out_str (stdout, 10, afgp -> modulus);
	printf ("\n");

	nmf	= afgp -> nmf;
	gfinfo	= afgp -> gfinfo;

	for (i = 0; i < nmf; i++) {
		printf ("Modulular factor %d:\n", i);
		Qax_print_one_lifted_gfinfo (&gfinfo [i], xvar, avar);
	}
}

/*
 * Print out the "lifted" portion of a single gfinfo object --
 * i.e., the lifted F(a) (mod p^j) and f(x) (mod F_k(a), mod p^j)
 * factorizations.
 */

	static
	void
Qax_print_one_lifted_gfinfo (

const struct gfinfo *	gip,	/* IN: gfinfo object to print */
const char *		xvar,	/* IN: main polynomial var */
const char *		avar	/* IN: algebraic extension var */
)
{
int				j;
const struct calib_Zx_dom *	Zx;
const struct calib_Zax_dom *	Zax;
struct calib_genrep *		gp1;
const struct calib_Zax_obj *	lfactor;

	printf ("  Lifted F(a) factor:\n");
	Zx = calib_get_Zx_dom ();
	gp1 = Zx -> to_genrep (gip -> afactor_zx, "a");
	calib_genrep_fprint (stdout, gp1);
	calib_genrep_free (gp1);

	printf ("  Lifted f(x) factors:\n");
	Zax = gip -> Zax;
	for (j = 0; j < gip -> nfact; j++) {
		lfactor = gip -> lfactors [j];
		gp1 = Zax -> to_genrep (lfactor, xvar, avar);
		calib_genrep_fprint (stdout, gp1);
		calib_genrep_free (gp1);
	}
}

/*
 * Print out a rather cryptic string that indicates the particular
 * combination of trial factors that is being tried.
 */

	static
	void
Qax_print_trial_factor_combination_string (

struct afglob *		afgp	/* IN: global info */
)
{
int			i, j, nmf, nfact;
struct gfinfo *		gfinfo;
struct gfinfo *		gip;

	nmf	= afgp -> nmf;
	gfinfo	= afgp -> gfinfo;

	for (i = 0; i < nmf; i++) {
		if (i > 0) {
			printf (" ");
		}
		gip = &gfinfo [i];
		nfact = gip -> nfact;
		for (j = 0; j < nfact; j++) {
			if (gip -> lf_try [j]) {
				printf ("X");
			}
			else {
				printf (".");
			}
		}
	}
	printf ("\n");
}

/*
 * Free up the array of gfinfo objects.
 */

	static
	void
Qax_free_gfinfo (

int			nmf,	/* IN: number of modular factors */
struct gfinfo *		gfinfo	/* IN: array of info for each modular */
				/*     factor */
)
{
int		i;

	if (gfinfo NE NULL) {
		for (i = 0; i < nmf; i++) {
			Qax_clear_gfinfo (&gfinfo [i]);
		}
		free (gfinfo);
	}
}

/*
 * Free up all of the memory used by a single gfinfo object,
 * but not the object itself.
 */

	static
	void
Qax_clear_gfinfo (

struct gfinfo *		gip	/* IN: gfinfo object to clear */
)
{
int				i, nfact;
const struct calib_Zx_dom *	Zx;
const struct calib_Zpx_dom *	Zpx;
struct calib_GFpkx_dom *	gfpkx;
struct calib_Zax_dom *		Zax;

	if (gip EQ NULL) return;

	nfact = gip -> nfact;
	Zx	= calib_get_Zx_dom ();
	Zpx	= gip -> Zpx;
	gfpkx	= gip -> gfpkx;
	Zax	= gip -> Zax;

	if (gip -> lf_used NE NULL) {
		free (gip -> lf_used);
	}
	if (gip -> lf_try NE NULL) {
		free (gip -> lf_try);
	}
	if (gip -> wfactor NE NULL) {
		Zx -> free (gip -> wfactor);
	}
	if (gip -> vfactor NE NULL) {
		Zx -> free (gip -> vfactor);
	}
	if (gip -> lfactors NE NULL) {
		for (i = 0; i < nfact; i++) {
			Zax -> free (gip -> lfactors [i]);
		}
	}
	if (gip -> factors NE NULL) {
		for (i = 0; i < nfact; i++) {
			CALIB_FATAL_ERROR_IF (gfpkx EQ NULL);
			gfpkx -> free (gip -> factors [i]);
		}
		free (gip -> factors);
	}
	if (Zax NE NULL) {
		Zax -> free (gip -> fpoly_zax);
	}
	if (gfpkx NE NULL) {
		gfpkx -> free (gip -> fpoly_gfpkx);
	}

	Zx -> free (gip -> afactor_zx);
	if (Zpx NE NULL) {
		Zpx -> free (gip -> afactor_zpx);
	}

	calib_free_Zax_dom (gip -> Zax);
	calib_free_Za_dom (gip -> Za);
	calib_free_GFpkx_dom (gip -> gfpkx);
	calib_free_GFpk_dom (gip -> gfpk);

	memset (gip, 0, sizeof (*gip));
}

/*
 * Free the given factorization, as produced by Qax_factor().
 */

	static
	void
Qax_free_factors (

struct calib_Qax_factor *	factors	/* IN: factors to recursively free */
)
{
struct calib_Qax_factor *	ftemp;

	while (factors NE NULL) {
		ftemp = factors -> next;
		Qax_free (factors -> factor);
		memset (factors, 0, sizeof (*factors));
		free (factors);
		factors = ftemp;
	}
}

/*
 * For given Qax domain dom, return a Qax domain defining the corresponding
 * algebraic integer.  If dom's generator polynomial is monic, this will
 * always return dom.  Otherwise, we return a new Qa domain representing
 * the corresponding algebraic integer.
 * DO NOT FREE THIS DOMAIN, SINCE IT IS OWNED BY dom!
 */

	static
	const struct calib_Qax_dom *
Qax_algint_dom (

const struct calib_Qax_dom *	dom	/* IN: Qax domain for which to get */
					/*     algebraic integer domain */
)
{
const struct calib_Qa_dom *	cf;
const struct calib_Qa_dom *	ai_cf;
struct calib_Qax_dom *		ai_dom;
struct calib_Qax_dom *		writable_dom;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	ai_dom	= dom -> ai_dom;
	if (ai_dom NE NULL) {
		/* Previously constructed the algint domain. */
		return (ai_dom);
	}

	cf = dom -> cf;

	ai_cf	= cf -> algint_dom (cf);
	if (ai_cf EQ cf) {
		/* This domain is already over an algebraic integer. */
		return (dom);
	}

	ai_dom	= calib_make_Qax_dom (ai_cf);

	/* Cache this polynomial for later use. */
	writable_dom	= (struct calib_Qax_dom *) dom;

	writable_dom -> ai_dom	= ai_dom;

	return (ai_dom);
}

/*
 * Translates a Qax polynomial from coefficient domain B to domain A.
 * Two cases are permitted:
 *
 *	1. A = B, or
 *	2. A is the algebraic integer corresponding to B.
 *
 * The result (interpreted with respect to domain A) is a polynomial giving
 * exactly the same values as the source polynomial (interpreted with
 * respect to domain B).
 */

	static
	void
Qax_to_algint (

struct calib_Qax_obj *		rop,	/* IN/OUT: Resulting Qax polynomial */
const struct calib_Qax_obj *	op	/* IN: Source Qax polynomial */
)
{
int				i, deg;
const struct calib_Qa_dom *	A;
const struct calib_Qa_dom *	B;
const struct calib_Qa_obj *	coeff;
struct calib_Qa_obj *		ai_coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	A	= rop -> dom -> cf;
	B	= op  -> dom -> cf;

	if (A EQ B) {
		/* No change of domain. */
		Qax_set (rop, op);
	}
	else {
		/* Verify that A is the algebraic integer of B. */
		CALIB_FATAL_ERROR_IF (A NE B -> ai_dom);

		/* Do the conversion. */

		deg	= op -> degree;
		coeff	= op -> coeff;

		Qax_alloc (rop, deg);

		ai_coeff = rop -> coeff;

		for (i = 0; i <= deg; i++) {
			B -> to_algint (&ai_coeff [i], &coeff [i]);
		}
		rop -> degree = deg;
	}
}

/*
 * Translates a Qax polynomial from coefficient domain B to domain A.
 * Two cases are permitted:
 *
 *	1. A = B, or
 *	2. B is the algebraic integer corresponding to A.
 *
 * The result (interpreted with respect to domain A) is a polynomial giving
 * exactly the same values as the source polynomial (interpreted with
 * respect to domain B).
 */

	static
	void
Qax_from_algint (

struct calib_Qax_obj *		rop,	/* IN/OUT: Resulting Qax polynomial */
const struct calib_Qax_obj *	op	/* IN: Source Qax polynomial */
)
{
int				i, deg;
const struct calib_Qa_dom *	A;
const struct calib_Qa_dom *	B;
const struct calib_Qa_obj *	ai_coeff;
struct calib_Qa_obj *		coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	A	= rop -> dom -> cf;
	B	= op  -> dom -> cf;

	if (A EQ B) {
		/* No change of domain. */
		Qax_set (rop, op);
	}
	else {
		/* Verify that B is the algebraic integer of A. */
		CALIB_FATAL_ERROR_IF (B NE A -> ai_dom);

		/* Do the conversion. */

		deg		= op -> degree;
		ai_coeff	= op -> coeff;

		Qax_alloc (rop, deg);

		coeff = rop -> coeff;

		for (i = 0; i <= deg; i++) {
			B -> from_algint (&coeff [i], &ai_coeff [i]);
		}
		rop -> degree = deg;
	}
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically zero.
 */

	static
	calib_bool
Qax_zerop (

const struct calib_Qax_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg >= 0) {
		return (FALSE);
	}

	CALIB_FATAL_ERROR_IF (deg < -1);

	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically one.
 */

	static
	calib_bool
Qax_onep (

const struct calib_Qax_obj *	op	/* IN: Polynomial to test */
)
{
int				deg;
const struct calib_Qa_dom *	cf;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg NE 0) {
		return (FALSE);
	}

	cf = op -> dom -> cf;

	if (cf -> onep (&(op -> coeff [0])) EQ 0) {
		return (TRUE);
	}

	return (FALSE);
}

/*
 * Convert the given general expression into a member of Q(a)[x].
 * All of the variables in the given expression *MUST* match either the
 * given xvar, or avar.  (xvar is the "main" variable of the polynomial,
 * whereas avar is the variable for the coefficient field.)
 */

	static
	void
Qax_set_genrep (

struct calib_Qax_obj *		rop,	/* IN/OUT: resulting polynomial */
const struct calib_genrep *	op,	/* IN: expression to convert */
const char *			xvar,	/* IN: main var of polynomial */
const char *			avar	/* IN: var for coefficient field */
)
{
int				i;
const struct calib_Qax_dom *	K_of_x;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
const struct calib_Qa_dom *	cf;
struct calib_Qax_obj		POLY;
struct calib_Qax_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop    EQ NULL) OR
			      (op     EQ NULL) OR
			      (xvar   EQ NULL) OR
			      (avar   EQ NULL));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Qax_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Qax_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		Qax_set_q (rop, op -> u.qop);
		break;

	case CALIB_GENREP_OP_VAR:
		if (strcmp (op -> u.var, xvar) EQ 0) {
			Qax_set_var_power (rop, 1);
		}
		else if (strcmp (op -> u.var, avar) EQ 0) {
			Qax_alloc (rop, 0);
			rop -> degree	= 0;
			cf -> set_var_power (&(rop -> coeff [0]), 1);
		}
		else {
			CALIB_FATAL_ERROR;
		}
		break;

	case CALIB_GENREP_OP_ADD:
		Qax_init (K_of_x, &POLY);
		Qax_init (K_of_x, &TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Qax_set_genrep (&TEMP,
					lp -> operand,
					xvar,
					avar);
			Qax_add (&POLY, &POLY, &TEMP);
		}
		Qax_set (rop, &POLY);

		Qax_clear (&TEMP);
		Qax_clear (&POLY);
		break;

	case CALIB_GENREP_OP_MUL:
		Qax_init (K_of_x, &POLY);
		Qax_init (K_of_x, &TEMP);
		Qax_set_si (&POLY, 1);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Qax_set_genrep (&TEMP,
					lp -> operand,
					xvar,
					avar);
			Qax_mul (&POLY, &POLY, &TEMP);
		}
		Qax_set (rop, &POLY);

		Qax_clear (&TEMP);
		Qax_clear (&POLY);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;
		CALIB_FATAL_ERROR_IF (i < 0);

		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);

		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			/* Optimize important special case. */
			if (strcmp (gp -> u.var, xvar) EQ 0) {
				Qax_set_var_power (rop, i);
			}
			else if (strcmp (gp -> u.var, avar) EQ 0) {
				Qax_alloc (rop, 0);
				rop -> degree = 0;
				cf -> set_var_power (&(rop -> coeff [0]), i);
			}
			else {
				CALIB_FATAL_ERROR;
			}
		}
		else {
			Qax_init (K_of_x, &TEMP);
			Qax_set_genrep (&TEMP, gp, xvar, avar);
			Qax_ipow (rop, &TEMP, i);
			Qax_clear (&TEMP);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Convert the given element of K[x] (where K = Q(a)) into
 * general representation.  The generated expression uses xvar for
 * the "main" polynomial variable, whereas avar is used for the
 * variable in the coefficient field K.
 */

	static
	struct calib_genrep *
Qax_to_genrep (

const struct calib_Qax_obj *	op,	/* IN: polynomial to convert */
const char *			xvar,	/* IN: main var of polynomial */
const char *			avar	/* IN: var for coefficient field */
)
{
int				i, degree;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
const struct calib_Qa_obj *	coeff;
const struct calib_Qa_obj *	valp;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op   EQ NULL) OR
			      (xvar EQ NULL) OR
			      (avar EQ NULL));

	K_of_x	= op -> dom;
	cf	= K_of_x -> cf;

	degree	= op -> degree;
	coeff	= op -> coeff;

	root = NULL;
	hookp = &root;

	for (i = degree; i >= 0; i--) {
		valp = &coeff [i];
		if (cf -> zerop (valp)) continue;
		vgp = cf -> to_genrep (valp, avar);
		gp1 = calib_genrep_poly_term (vgp, xvar, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	return (gp1);
}

/*
 * Convert the given list of factors into general representation.
 * Use the variables given: xvar for the main variable, avar for
 * coefficient field.
 */

	static
	struct calib_genrep *
Qax_factors_to_genrep (

const struct calib_Qax_factor *	factors,/* IN: list of factors */
const char *			xvar,	/* IN: polynomial main var */
const char *			avar	/* IN: var for coefficient field */
)
{
int				pow;
const struct calib_Qax_factor *	fp;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;

	CALIB_FATAL_ERROR_IF ((xvar EQ NULL) OR (avar EQ NULL));

	root = NULL;
	hookp = &root;

	for (fp = factors; fp NE NULL; fp = fp -> next) {
		gp1 = Qax_to_genrep (fp -> factor, xvar, avar);
		pow = fp -> multiplicity;
		CALIB_FATAL_ERROR_IF (pow <= 0);
		if (pow > 1) {
			gp2 = CALIB_NEW (struct calib_genrep);
			gp2 -> op		= CALIB_GENREP_OP_IPOW;
			gp2 -> u.ipow.base	= gp1;
			gp2 -> u.ipow.power	= pow;
			gp1 = gp2;
		}
		lp = calib_genrep_new_list (gp1, NULL);
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (1);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_MUL;
		gp1 -> u.list	= root;
	}
	return (gp1);
}

/*
 * Compute the bit-size of the given polynomial.
 */

	static
	size_t
Qax_bit_size (

const struct calib_Qax_obj *	poly		/* IN: polynomial to get */
						/*     size of */
)
{
int				i, deg;
size_t				nbits, max_nbits;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
const struct calib_Qa_obj *	coeff;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	deg	= poly -> degree;
	K_of_x	= poly -> dom;
	cf	= K_of_x -> cf;
	coeff	= poly -> coeff;

	max_nbits = 0;
	for (i = 0; i <= deg; i++) {
		nbits = cf -> bit_size (&coeff [i]);
		if (nbits > max_nbits) {
			max_nbits = nbits;
		}
	}

	return (max_nbits);
}

/*
 * Compute the bit-size of the given factorization.
 */

	static
	size_t
Qax_factors_bit_size (

const struct calib_Qax_factor *	factors	/* IN: factors to get size of */
)
{
size_t				nbits, max_nbits;
const struct calib_Qax_factor *	fp;

	max_nbits = 0;
	for (fp = factors; fp NE NULL; fp = fp -> next) {
		nbits = Qax_bit_size (fp -> factor);
		if (nbits > max_nbits) {
			max_nbits = nbits;
		}
	}
	return (max_nbits);
}

/*
 * Return the Z(a)[x] version of this Q(a)[x] domain.
 *
 * NOTE: Do *NOT* free this domain, because it is owned by the Q(a)[x] domain!
 */

	static
	const struct calib_Zax_dom *
Qax_get_Zax_dom (

const struct calib_Qax_dom *	K_of_x	/* IN: Given Q(a)[x] domain */
)
{
struct calib_Zax_dom *		dom;
const struct calib_Qa_dom *	cf;
const struct calib_Za_dom *	Za;
struct calib_Qax_dom *		writable_K_of_x;

	CALIB_FATAL_ERROR_IF (K_of_x EQ NULL);

	dom = K_of_x -> Zax_dom;
	if (dom EQ NULL) {
		/* Get corresponding Z(a) domain. */
		cf = K_of_x -> cf;
		Za = cf -> get_Za_dom (cf);

		dom = calib_make_Zax_dom (Za);

		/* We must force write access to K_of_x -> Zax_dom here! */
		writable_K_of_x = (struct calib_Qax_dom *) K_of_x;
		writable_K_of_x -> Zax_dom = dom;
	}

	return (dom);
}

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
Qax_print (

const struct calib_Qax_obj *	op	/* IN: Polynomial to print */
)
{
int				i, j, deg, deg2;
const struct calib_Qa_obj *	qap;
const char *			lbrack;
const char *			rbrack;

	if (op EQ NULL) {
		printf ("<null ptr>\n");
		return;
	}

	lbrack	= "[";
	rbrack	= "]";

	deg	= op -> degree;
	for (i = deg; i >= 0; i--) {
		qap	= &(op -> coeff [i]);
		gmp_printf (" + %s (%Zd/%Zd)",
			    lbrack,
			    mpq_numref (qap -> qfact),
			    mpq_denref (qap -> qfact));
		deg2	= qap -> degree;
		for (j = deg2; j >= 0; j--) {
			gmp_printf (" %Zd", &(qap -> coeff [j]));
		}
		gmp_printf (" %s x^%d", rbrack, i);
	}
	printf ("\n");
}

#pragma GCC diagnostic pop

/*
 * Prototype Qax domain object.
 */

static const struct calib_Qax_dom	Qax_dom = {
	.init			= Qax_init,
	.init_degree		= Qax_init_degree,
	.alloc			= Qax_alloc,
	.clear			= Qax_clear,
	.set			= Qax_set,
	.set_si			= Qax_set_si,
	.set_z			= Qax_set_z,
	.set_q			= Qax_set_q,
	.set_Qa			= Qax_set_Qa,
	.set_var_power		= Qax_set_var_power,
	.set_Zx			= Qax_set_Zx,
	.add			= Qax_add,
	.sub			= Qax_sub,
	.neg			= Qax_neg,
	.mul			= Qax_mul,
	.ipow			= Qax_ipow,
	.dup			= Qax_dup,
	.free			= Qax_free,
	.eval			= Qax_eval,
	.eval_poly		= Qax_eval_poly,
	.div			= Qax_div,
	.gcd			= Qax_gcd,
	.extgcd			= Qax_extgcd,
	.factor			= Qax_factor,
	.factor_square_free	= Qax_compute_square_free_factorization,
	.finish_sqf		= Qax_finish_sqf,
	.free_factors		= Qax_free_factors,
	.algint_dom		= Qax_algint_dom,
	.to_algint		= Qax_to_algint,
	.from_algint		= Qax_from_algint,
	.zerop			= Qax_zerop,
	.onep			= Qax_onep,
	.set_genrep		= Qax_set_genrep,
	.to_genrep		= Qax_to_genrep,
	.factors_to_genrep	= Qax_factors_to_genrep,
	.bit_size		= Qax_bit_size,
	.factors_bit_size	= Qax_factors_bit_size,
	.get_Zax_dom		= Qax_get_Zax_dom,
};

/*
 * Construct the ring of polynomials over field K, where K is
 * the given field Q(a).
 */

	struct calib_Qax_dom *
calib_make_Qax_dom (

const struct calib_Qa_dom *	K	/* IN: coefficient field */
)
{
struct calib_Qax_dom *		K_of_x;

	K_of_x = CALIB_NEW (struct calib_Qax_dom);

	/* Initialize "most" things. */
	*K_of_x = Qax_dom;

	K_of_x -> cf		= K;
	K_of_x -> Zax_dom	= NULL;
	K_of_x -> ai_dom	= NULL;
	K_of_x -> settings	= calib_Qax_default_settings;

	return (K_of_x);
}

/*
 * Free up the given Q(a)[x] ring description.
 */

	void
calib_free_Qax_dom (

struct calib_Qax_dom *	K_of_x	/* IN: Q(a)[x] ring description */
)
{
	if (K_of_x NE NULL) {
		calib_free_Zax_dom (K_of_x -> Zax_dom);

		calib_free_Qax_dom (K_of_x -> ai_dom);

		memset (K_of_x, 0, sizeof (*K_of_x));

		free (K_of_x);
	}
}
