/**********************************************************************

	$Id: Zx.c,v 1.1 2024/10/20 00:36:26 warme Exp $

	File:	Zx.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The ring Z[x] of polynomials over the integers Z.

***********************************************************************

	Modification Log:

	a-1:	04/10/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Zx.h"

#include "bitmask.h"
#include "calib/config.h"
#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/random.h"
#include "calib/types.h"
#include "calib/Zp.h"
#include "calib/Zpx.h"
#include "calib/Zxyz.h"
#include <float.h>
#include <gmp.h>
#include <limits.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

/* Consider using Van Hoeij if we have FPLLL available.  (Our own	*/
/* implementation of LLL is not yet competitive, so only attempt to use	*/
/* Van Hoeij when FPLLL is present for now.)				*/

#ifdef CALIB_HAVE_FPLLL
	#define ENABLE_VAN_HOEIJ
#endif

#ifdef ENABLE_VAN_HOEIJ
	#define	VAN_HOEIJ_ENABLED_BY_DEFAULT	TRUE
	#define	DEFAULT_FACTOR_METHOD	CALIB_ZX_FACTOR_METHOD_VAN_HOEIJ
#else
	#define VAN_HOEIJ_ENABLED_BY_DEFAULT	FALSE
	#define	DEFAULT_FACTOR_METHOD	CALIB_ZX_FACTOR_METHOD_ZASSENHAUS
#endif

/*
 * The global "settings" object for the Zx domain.
 */

struct calib_Zx_settings	calib_Zx_settings = {
	/* Default factor method depends upon what is available:	*/
	/*	Van Hoeij	(if available), otherwise		*/
	/*	Zassenhaus						*/
	.factor_method		= DEFAULT_FACTOR_METHOD,

	/* Default to quadratic lifting. */
	.lift_method		= CALIB_ZX_LIFT_METHOD_QUADRATIC,

	/* Default to modular gcd. */
	.gcd_method		= CALIB_ZX_GCD_METHOD_MODULAR,

	/* If number of remaining modular factors does not exceed this,	*/
	/* use exhaustive enumeration instead of Van Hoeij.		*/
	.VH_max_enumerate	= 10,

	/* Enable trace output from Z[x] factorization algorithms.	*/
	/* Zero is none.  Higher values give more output.		*/
	.factor_trace_level		= 0,

	/* Turn this flag on to use resultants to calculate the i-th	*/
	/* trace root bounds within the Van Hoeij algorithm.  (This is	*/
	/* not usually a good idea with the present implementation,	*/
	/* which uses the general Zxyz resultant algorithm to compute	*/
	/* these resultants.  Given polynomial f(x) \in Z[x] having	*/
	/* roots \alpha, there is probably a very nice, fast trick for	*/
	/* computing polynomial h_i(x) \in Z[x] (of same degree) having	*/
	/* roots \alpha**i.  Leaving this off until I can generalize	*/
	/* this trick and make the code use it.  The present version is	*/
	/* way too slow for general use.				*/
	.VH_use_resultant_trace_root_bound	= FALSE,

	/* Validate lifted modular factors? */
	.validate_lifted_modular_factors	= FALSE,
};

/* Prime number generator for modular GCD. */

struct gcdpgen {
	struct calib_Zp_dom *
			Zp;	/* Put prime p into this Zp domain */
	mpz_srcptr	avoid;	/* Avoid primes that divide this */
	int		tries;	/* Number of tries at this nwords */
	int		nwords;	/* Number of 32-bit words in desired primes */
	struct calib_Random
			rndgen;	/* Random number generator */
};

/*
 * A matrix of GMP integers.
 */

struct ZMatrix {
	int		zm_rows;	/* Number of rows */
	int		zm_cols;	/* Number of columns */
	mpz_ptr		zm_A;		/* Matrix of integer values */
};

/*
 * A matrix of GMP rationals.
 */

struct QMatrix {
	int		qm_rows;	/* Number of rows */
	int		qm_cols;	/* Number of columns */
	mpq_ptr		qm_A;		/* Matrix of rational values */
};

/*
 * Global information used by the Z[x] factorization algorithm,
 * including the Van Hoeij lattice-reduction algorithm.
 */

struct FInfo {
	const struct calib_Zx_obj *
				fi_poly;	/* Z[x] poly being factored */
	struct calib_Zx_obj	fi_lpoly;	/* Polynomial "left" to be */
						/* factored */
	struct calib_Zpx_obj *	fi_Zpx_poly;	/* fi_poly (mod p) */
	struct calib_Zpx_obj *	fi_monic_Zpx_poly;
						/* Monic version of */
						/* fi_Zpx_poly */
	struct calib_Zpx_factor *
				fi_zpfactor_list;
						/* List of Zp[x] factors */
	int			fi_num_fact;	/* Original # of factors, */
						/* size of various factor */
						/* arrays */
	int			fi_nleft;	/* Number of untaken modular */
						/* factors remaining */
	int			fi_lm_a;	/* p**a for Landau/Mignotte */
	int			fi_a;		/* modulus = p**fi_a */
	struct calib_Zx_factor *
				fi_list;	/* Linked list of factors */
						/* irreducible over Z[x] */
	struct calib_Zx_factor **
				fi_hookp;	/* Where to append Z[x] factors */
	struct calib_Zp_dom *	fi_Zp;		/* Zp domain */
	struct calib_Zpx_dom *	fi_Zpx;		/* Zp[x] domain */
	struct calib_Zx_obj **	fi_zfactors;	/* Factors (mod p**fi_a) */
	struct calib_Zpx_obj **	fi_zpfactors;	/* Factors in Zp[x] */
	struct calib_Zpx_obj **	fi_vlist;	/* Zp[x] polys for lifting */
	struct calib_Zx_obj **	fi_zvlist;	/* Z[x] polys for lifting */
	mpz_t			fi_lcinv;	/* Multiplicative inverse of */
						/* lc(poly) mod p**fi_a. */
						/* Used only with quadratic */
						/* lifting (and we need to */
						/* lift this value too!) */
	int			fi_num_osp;	/* Num omit sub-products */
	int *			fi_osp_struct;	/* Structure for computing */
						/* sub-products */
	int *			fi_omit_struct;	/* Sub-product structure of */
						/* each omit[i] */
	calib_bool *		fi_taken;	/* TRUE for modular factors */
						/* already used to construct */
						/* a Z[x] factor */
	calib_bool *		fi_chosen;	/* TRUE for modular factors */
						/* chosen to be part of the */
						/* current trial divisor */
	int *			fi_findex;	/* Index of remaining factors */
	int *			fi_comb;	/* Stack of factor indices */
						/* being combined */
	mpz_t			fi_cbound;	/* Upper bound on magnitude */
						/* of coefficient in any */
						/* Z[x] factor */
	mpz_t			fi_modulus;	/* Current modulus (p**fi_a) */
	mpz_t			fi_lc_tc;	/* lc(lpoly) * tc(lpoly) */
	mpz_ptr			fi_tc_stack;	/* Stack containing products */
						/* of lc(lpoly) * product of */
						/* tc values of all chosen */
						/* trial factors */
	int			fi_desired_I;	/* The target value of */
						/* "Information" that we */
						/* are trying to hit */
	mpz_ptr			fi_Tr_bounds;	/* Upper bounds on magnitude */
						/* of the i-th trace of any */
						/* Z[x] factor of degree d */
	mpz_ptr			fi_Tr_fj_i;	/* For each factor j, the */
						/* i-th traces, i \in [0,d], */
						/* signed, mod p**a */
	int			fi_a_Tr_fj;	/* Power of p for which	the */
						/* factor j i-th traces have */
						/* been computed */
	int			fi_strategy;	/* Strategy being used */
	int			fi_s;		/* Num rows in matrix A */
	int			fi_d;		/* Num cols in matrix A */
	int			fi_a_b_gap;	/* Desired gap = a_i - b_i */
	int			fi_rfails;	/* Number of consecutive */
						/* failed attempts to reduce */
						/* the size of the basis */
	int			fi_zfail;	/* Basis reduced to no	*/
						/* M-short vectors */
	int *			fi_ai;		/* The a_i values */
	int *			fi_bi;		/* The b_i values */
	int *			fi_A;		/* The A matrix (NULL means */
						/* the identity matrix) */
#ifdef ENABLE_VAN_HOEIJ
	struct ZMatrix		fi_B_L;		/* Basis for lattice L */
	int			fi_C;		/* The chosen value of C */
	struct ZMatrix		fi_matrix_C;	/* Matrix of rounded modular */
						/* trace values */
	mpq_t			fi_M2;		/* Value of M**2 */
	struct ZMatrix		fi_BLambda;	/* Lattice basis matrix */
	struct ZMatrix		fi_Blll;	/* LLL-Reduced lattice basis */
	int *			fi_short;	/* List of M-short vectors */
	struct calib_Random	fi_rand;	/* Random number generator */
#endif
};

/*
 * Objects to record square-free factors, including the special ability
 * to further split each such factor into two smaller factors.
 * These nodes form a binary tree, where leaf nodes are factors that
 * (while they may or may not be irreducible) have not yet been split into
 * two smaller factors.  Interior nodes represent factors that *have*
 * been split into smaller factors.
 * The SqfTree object represents the entire tree.
 * The SqfNode object represents a single node (factor) in the tree.
 */

struct SqfList;
struct SqfNode;
struct SqfTree;

struct SqfTree {
	int			st_node_size;	/* Size of st_nodes array */
	int			st_node_alloc;	/* Next node to allocate */
	int			st_list_alloc;	/* Next SqfList to allocate */
	int			st_nleaves;	/* Number of leaf nodes */
	int *			st_leaves;	/* Indices of leaf nodes */
	struct SqfNode *	st_nodes;	/* Array of nodes */
	struct SqfList **	st_poly_list;	/* Linked list of factors */
						/* for each polynomial */
	struct SqfList *	st_lists;	/* Array of SqfList nodes */
	int **			st_poly_node_mult;
						/* For each polynomial, an */
						/* array giving the */
						/* multiplicity of each */
						/* node. (Zero is a very */
						/* popular multiplicity: */
						/* only leaf nodes have */
						/* positive counts herein.) */
	int *			st_fm_array;	/* Actual storage pointed */
						/* to by st_poly_node_mult */
	int *			st_min_mult;	/* Minimum multiplicity of */
						/* each leaf node across all */
						/* polynomials */
	int *			st_temp_leaves;	/* Temp copy of st_leaves[] */
};


struct SqfNode {
	struct calib_Zx_obj *	sn_factor;	/* Factor for this node */
	int			sn_left;	/* Left  node index (or -1) */
	int			sn_right;	/* Right node index (or -1) */
	int			sn_index;	/* Index of this node */
						/* within st_leaves (or -1) */
};

/*
 * A linked list to represent the factors (not necessarily irreducible)
 * for each polynomial we are taking the GCD of.
 */

struct SqfList {
	int			sl_fact_idx;	/* Index of this factor */
	int			sl_mult;	/* Multiplicity */
	struct SqfList *	sl_next;	/* Next node in linked list */
};

/*
 * Let r be the number of modular factors f_i in Zq[x], q = p**k.  During
 * lifting, we need to compute:
 *
 *		omit_i(x) = \prod{1<=j<=r, j NE i} f_j
 *
 * Two major approaches: (1) form product P of all f_i and then compute
 * omit_i = P / f_i by division; or (2) compute each omit_i entirely by
 * multiplication, which can be done with O(r*log(r)) multiplies if we
 * are careful to first form various sub-products, organized into rows:
 * Row 0 contains the r original f_i.
 * Row 1 contains products of consecutive pairs of f_i.
 * Row 2 contains products of consecutive groups of 4 f_i.
 * In general, row j contains m = floor(r / 2**k) "full" products of
 * 2**k consecutive factors, plus perhaps an additional "residual"
 * factor that is the product of the last r - m*2**k factors f_i.
 *
 * Given such a set of previously-computed sub-products, each omit_i can
 * then be computed as the product of O(log(r)) of these sub-products.
 *
 * This object describes a single "row" of these sub-products, helping us
 * to organize the construction of these sub-products.
 */

struct sprod_row {
	int	start;		/* Index of first sub-product in this row */
	int	end;		/* Index of first sub-product in next row */
	int	fspsize;	/* Size 2**k of "full" products in this row */
	int	nfsp;		/* Number of "full" sub-products in this row */
	int	nesp;		/* Number of "extra" sub-products in this row*/
				/* These are for residual sub-products that */
				/* must be combined from previous */
				/* sub-products */
	int	nresid;		/* Number of "residual" (non-full) factors */
				/* at end of this row */
	int	residx;		/* Index of final, residual sub-product for */
				/* this row.  This can be: */
				/*   *	-1 (when nresid EQ 0 and there is no */
				/*	 residual sub-product); */
				/*   *	The index of a (residual or full) */
				/*	sub-product from a previous row */
				/*   *	The index of one additional sub- */
				/*	product in this row, formed as the */
				/*	product of two sub-products from */
				/*	previous rows. */
};

/* Old tricks... */
#define	KEEP_LOWEST_ONE_BIT(n)	((n) & -(n))

#define IS_POWER_OF_2(n) (((n) > 0) AND (KEEP_LOWEST_ONE_BIT (n) EQ (n)))

/*
 * An object to represent distinct "residual" values.
 */

struct sprod_resid {
	int	nresid;		/* Number of trailing residual factors */
	int	row;		/* First row that needs this */
	int	pow;		/* Power of 2 we split off (or -1) */
	int	r2;		/* Residual after split */
	int	residx;		/* Index of this sub-product */
	int	idx1;		/* Index of most-signif sub-prod (or -1) */
	int	idx2;		/* Index of least-signif sub-prod (or -1) */
};

/*
 * Local Functions.
 */

static void	Zx_add (struct calib_Zx_obj *		rop,
			const struct calib_Zx_obj *	op1,
			const struct calib_Zx_obj *	op2);
static void	Zx_adjust_gcd_n_result_signs (struct calib_Zx_obj *	gcd,
					      struct calib_Zx_obj **	cofact,
					      int			npoly);
static void	Zx_alloc (struct calib_Zx_obj *	rop,
			  int			degree);
static void	Zx_choose_gcd_prime (struct gcdpgen * gp);
static void	Zx_choose_prime_modulus (struct FInfo * fip);
static void	Zx_cleanup_FInfo (struct FInfo * fip);
static void	Zx_clear (struct calib_Zx_obj *	x);
static void	Zx_clear_array (struct calib_Zx_obj *	parray,
				int			npoly);
static int	Zx_compare_factors (const struct calib_Zx_factor *	fp1,
				    const struct calib_Zx_factor *	fp2);
static void	Zx_compute_lifting_target_power_of_p (struct FInfo * fip);
static void	Zx_compute_mod_p_gcd (
				struct calib_Zpx_obj *		zpgcd,
				int				npoly,
				const int *			order,
				const struct calib_Zx_obj *	poly_array,
				struct calib_Zpx_obj *		zpx_array,
				struct calib_Zpx_obj *		zpx_cofactor);
static struct calib_Zx_obj *
		Zx_compute_omit (struct FInfo * fip, mpz_srcptr q);
static void	Zx_compute_omit_sub_product_structure (struct FInfo * fip);
static struct calib_Zx_factor *
		Zx_compute_square_free_factorization (
					const struct calib_Zx_obj *	poly);
static void	Zx_content_poly3 (mpz_ptr			zcont,
				  const struct calib_Zx_obj *	poly1,
				  const struct calib_Zx_obj *	poly2,
				  const struct calib_Zx_obj *	poly3);
static void	Zx_convert_Zpx_to_Zx (struct FInfo * fip);
static void	Zx_cumulative_content (mpz_ptr				zcont,
				       const struct calib_Zx_obj *	poly);
static struct calib_Zx_obj *
		Zx_cvZpx (const struct calib_Zpx_obj *	op);
static void	Zx_derivative (struct calib_Zx_obj *		rop,
			       const struct calib_Zx_obj *	op);
static void	Zx_discriminant (mpz_ptr			rop,
				 const struct calib_Zx_obj *	op);
static void	Zx_div (struct calib_Zx_obj *		quotient,
			struct calib_Zx_obj *		remainder,
			mpz_ptr				d,
			const struct calib_Zx_obj *	a,
			const struct calib_Zx_obj *	b);
static void	Zx_div_z_exact (struct calib_Zx_obj *		rop,
				const struct calib_Zx_obj *	op1,
				mpz_srcptr			op2);
static struct calib_Zx_obj *
		Zx_dup (const struct calib_Zx_obj *	op);
static void	Zx_eval (mpz_ptr			rop,
			 const struct calib_Zx_obj *	poly,
			 mpz_srcptr			value);
static calib_bool
		Zx_exactly_divides (struct calib_Zx_obj *	quotient,
				    const struct calib_Zx_obj *	a,
				    const struct calib_Zx_obj *	b);
static void	Zx_exhaustive_enumeration (struct FInfo * fip);
static void	Zx_extgcd (struct calib_Zx_obj *	gcd,
			   struct calib_Zx_obj *	xa,
			   struct calib_Zx_obj *	xb,
			   const struct calib_Zx_obj *	a,
			   const struct calib_Zx_obj *	b);
static struct calib_Zx_factor *
		Zx_factor (const struct calib_Zx_obj *	poly);
static void	Zx_factor_coefficient_upper_bound (struct FInfo * fip);
static struct calib_Zx_factor *
		Zx_factor_square_free (const struct calib_Zx_obj *	poly);
static struct calib_genrep *
		Zx_factors_to_genrep (
				const struct calib_Zx_factor *	factors,
				const char *			var);
static struct calib_Zx_factor *
		Zx_finish_sqf (const struct calib_Zx_factor * sqfactors);
static void	Zx_free (struct calib_Zx_obj *	poly);
static void	Zx_free_factors (struct calib_Zx_factor * factors);
static void	Zx_gcd (struct calib_Zx_obj *		gcd,
			const struct calib_Zx_obj *	a,
			const struct calib_Zx_obj *	b);
static void	Zx_gcd_modular (struct calib_Zx_obj *		gcd,
				const struct calib_Zx_obj *	a,
				const struct calib_Zx_obj *	b);
static void	Zx_gcd_n (struct calib_Zx_obj *		gcd,
			  struct calib_Zx_obj **	cofact,
			  int				npoly,
			  const struct calib_Zx_obj * const *
							poly_ptrs);
static void	Zx_gcd_prs (struct calib_Zx_obj *		gcd,
			    const struct calib_Zx_obj *		a,
			    const struct calib_Zx_obj *		b);
static void	Zx_hard_case_gcd_n (struct calib_Zx_obj *	pgcd,
				    int				npoly,
				    const struct calib_Zx_obj *	poly_array,
				    struct calib_Zx_obj *	cf_array);
static void	Zx_init (struct calib_Zx_obj *	x);
static void	Zx_init_array_degree (struct calib_Zx_obj *	parray,
				      int			npoly,
				      int			degree);
static void	Zx_init_degree (struct calib_Zx_obj *	poly,
				int			degree);
static void	Zx_init_FInfo (struct FInfo *			fip,
			       const struct calib_Zx_obj *	poly);
static void	Zx_init_si (struct calib_Zx_obj *	x,
			    calib_si_t			op);
static void	Zx_ipow (struct calib_Zx_obj *		result,
			 const struct calib_Zx_obj *	poly,
			 int				power);
static int	Zx_is_prime (unsigned long n);
static void	Zx_lift_factors (struct FInfo * fip, int target_pow);
static void	Zx_lift_factors_linear (struct FInfo * fip, int target_pow);
static void	Zx_lift_factors_quadratic (struct FInfo * fip, int target_pow);
static calib_bool
		Zx_lift_mod_p_gcd_to_z (
				const struct calib_Zpx_obj *	zpgcd,
				struct calib_Zx_obj *		pgcd,
				mpz_srcptr			cubound,
				int				npoly,
				const int *			order,
				const struct calib_Zx_obj *	poly_array,
				struct calib_Zx_obj *		cf_array,
				const struct calib_Zpx_obj *	zpx_array,
				const struct calib_Zpx_obj *	zpx_cofactor);
static void	Zx_make_zpfactors_array (struct FInfo * fip);
static void	Zx_mod_out_signed (struct calib_Zx_obj *	poly,
				   mpz_srcptr			modulus);
static void	Zx_mul (struct calib_Zx_obj *		rop,
			const struct calib_Zx_obj *	op1,
			const struct calib_Zx_obj *	op2);
static void	Zx_mul_z (struct calib_Zx_obj *		rop,
			  const struct calib_Zx_obj *	op1,
			  mpz_srcptr			op2);
static void	Zx_neg (struct calib_Zx_obj *		rop,
			const struct calib_Zx_obj *	op);
static calib_bool
		Zx_onep (const struct calib_Zx_obj *	op);
static void	Zx_pfrac (struct FInfo * fip);
static calib_int64u
		Zx_previous_prime (calib_int64u p);
static void	Zx_prim_part (mpz_ptr				content,
			      struct calib_Zx_obj *		ppart,
			      const struct calib_Zx_obj *	op);
static void	Zx_print (const struct calib_Zx_obj * op);
static void	Zx_print_array (const struct calib_Zx_obj * array, int npoly);
static int	Zx_print_nnl (const struct calib_Zx_obj * op, int col);
static void	Zx_realloc_copy_poly (struct calib_Zx_obj *	poly,
				      mpz_srcptr		coeff,
				      int			degree);
static void	Zx_remainder_monic_poly_mod_q (
				struct calib_Zx_obj *		remainder,
				const struct calib_Zx_obj *	fpoly,
				const struct calib_Zx_obj *	gpoly,
				mpz_srcptr			q);
static int	Zx_resid_recurse (int			i,
				  struct sprod_resid *	array,
				  int			j,
				  int			k);
static void	Zx_resultant (mpz_ptr				result,
			      const struct calib_Zx_obj *	a,
			      const struct calib_Zx_obj *	b);
static void	Zx_set (struct calib_Zx_obj *		rop,
			const struct calib_Zx_obj *	op);
static void	Zx_set_genrep (struct calib_Zx_obj *		rop,
			       const struct calib_genrep *	op,
			       const char *			var);
static void	Zx_set_si (struct calib_Zx_obj *	rop,
			   calib_si_t			op);
static void	Zx_set_var_power (struct calib_Zx_obj *	rop,
				  int			power);
static void	Zx_set_z (struct calib_Zx_obj *	rop,
			  mpz_srcptr		op);
static struct calib_Zx_factor *
		Zx_sort_factors (struct calib_Zx_factor * factors);
static int *	Zx_sort_gcd_polys (int				npoly,
				   const struct calib_Zx_obj *	poly_array);
static struct SqfList **
		Zx_SqfTree_append (struct SqfTree *		tp,
				   const struct calib_Zx_obj *	factor,
				   int				mult,
				   struct SqfList **		hookp);
static void	Zx_SqfTree_clear (struct SqfTree * tp);
static calib_bool
		Zx_SqfTree_compute_min_mult (struct SqfTree *	tp,
					     int		last_poly);
static struct SqfList **
		Zx_SqfTree_insert (struct SqfTree *		tp,
				   const struct calib_Zx_obj *	factor,
				   int				multiplicity,
				   struct SqfList **		hookp);
static void	Zx_SqfTree_init (struct SqfTree *		tp,
				 int				npoly,
				 const struct calib_Zx_obj *	poly_array);
static void	Zx_SqfTree_mul_factors (struct calib_Zx_obj *	prod,
					struct SqfTree *	tp,
					const int *		node_mult,
					const int *		min_mult);
static int	Zx_SqfTree_new_leaf_node (
				struct SqfTree *		tp,
				const struct calib_Zx_obj *	factor);
static struct SqfList **
		Zx_SqfTree_new_list (struct SqfTree *		tp,
				     struct SqfList **		hookp,
				     int			node_idx,
				     int			multiplicity);
static void	Zx_SqfTree_split_node (struct SqfTree *		tp,
				       int			parent,
				       int			left,
				       int			right);
static void	Zx_SqfTree_traverse_node_leaves (
					int *			node_mult,
					struct SqfTree *	tp,
					int			node_idx,
					int			multiplicity);
static void	Zx_strip_integer_content (mpz_ptr		content,
					  struct calib_Zx_obj *	poly);
static int	Zx_strip_varpow_factor (struct calib_Zx_obj * poly);
static void	Zx_sub (struct calib_Zx_obj *		rop,
			const struct calib_Zx_obj *	op1,
			const struct calib_Zx_obj *	op2);
static int	Zx_sub_product_row (int n);
static struct calib_genrep *
		Zx_to_genrep (const struct calib_Zx_obj *	op,
			      const char *			var);
static int	Zx_trial_check (struct FInfo *		fip,
				int			cardinality,
				mpz_srcptr		tc_prod);
static void	Zx_trial_division (struct FInfo * fip, int card_limit);
static int	Zx_trial_recurse (
				struct FInfo *		fip,
				int			idx,
				int			nchosen,
				int			cardinality);
static void	Zx_upper_bound_gcd_coeff (
				mpz_ptr				B,
				mpz_ptr				Avoid,
				int				npoly,
				const struct calib_Zx_obj *	poly_array);
static void	Zx_validate_lifted_modular_factors (struct FInfo * fip);
static calib_bool
		Zx_zerop (const struct calib_Zx_obj *	op);
static void	Zx_Zm_to_Z (mpz_ptr value, mpz_srcptr modulus);

#ifdef ENABLE_VAN_HOEIJ

static void	Zx_find_factors_Van_Hoeij (struct FInfo * fip);
static void	Zx_Van_Hoeij_main (struct FInfo * fip);
static calib_bool
		Zx_VH_all_columns_are_units (const struct QMatrix * rref);
static calib_bool
		Zx_VH_basis_has_factored_poly (struct FInfo * fip);
static void	Zx_VH_build_Lambda_lattice_basis (struct FInfo * fip);
static void	Zx_VH_choose_initial_matrix_and_parameters (struct FInfo * fip);
static void	Zx_VH_choose_new_matrix_and_parameters (struct FInfo * fip);
static void	Zx_VH_clear_BLambda (struct FInfo * fip);
static void	Zx_VH_clear_QMatrix (struct QMatrix * qmp);
static void	Zx_VH_clear_ZMatrix (struct ZMatrix * zmp);
static void	Zx_VH_compute_ai_bi (struct FInfo * fip);
static void	Zx_VH_compute_C_M2 (struct FInfo * fip);
static void	Zx_VH_compute_factor_traces (struct FInfo * fip);
static void	Zx_VH_compute_reduced_row_echelon_form (
					struct QMatrix *	rref,
					const struct ZMatrix *	src);
static void	Zx_VH_compute_trace_bound (struct FInfo *	fip,
					   mpz_ptr		result,
					   int			i,
					   int			d);
static void	Zx_VH_compute_trace_matrix_C (struct FInfo * fip);
static void	Zx_VH_debug_print_Gram_Schmidt (mpq_srcptr	Bstar,
						int		nvec,
						int		vsize);
static int	Zx_VH_find_M_short_vectors (struct FInfo * fip);
static calib_bool
		Zx_VH_full_divisibility_test (struct FInfo *	fip,
					      int		npart,
					      int **		partition);
static mpq_ptr	Zx_VH_Gram_Schmidt (mpz_srcptr B, int nvec, int vsize);
static void	Zx_VH_init_B_L (struct FInfo * fip);
static void	Zx_VH_init_QMatrix (struct QMatrix * qmp, int rows, int cols);
static void	Zx_VH_init_ZMatrix (struct ZMatrix * zmp, int rows, int cols);
static void	Zx_VH_init_trace_bounds (struct FInfo * fip);
static calib_bool
		Zx_VH_lc_tc_are_divisible (struct FInfo *	fip,
					   int			npart,
					   int **		partition);
static void	Zx_VH_mpq_mag2 (mpq_ptr result, mpq_srcptr v, int n);
static double	Zx_VH_mpz_log2 (mpz_srcptr x);
static void	Zx_VH_print_B_L (struct FInfo * fip);
static void	Zx_VH_print_Information (struct FInfo * fip);
static void	Zx_VH_print_ZMatrix (const struct ZMatrix * zmp);
static void	Zx_VH_reorder_remaining_factors_first (struct FInfo * fip);
static void	Zx_VH_replace_lattice (struct FInfo * fip, int r);
static void	Zx_VH_root_bound_Z (mpz_ptr			UBOUND,
				 const struct calib_Zx_obj *	poly);
static calib_bool
		Zx_VH_short_vectors_do_not_span_factors (struct FInfo * fip,
							 int		r);
static struct calib_Zx_factor *
		Zx_VH_trace_bounds_compute_h_i (
					const struct calib_Zx_obj *	f,
					int				i);
static void	Zx_VH_trace_root_bound_Z (mpz_ptr			UBOUND,
					  const struct calib_Zx_obj *	poly,
					  int				power);
static calib_bool
		Zx_VH_true_factorization_found (struct FInfo *		fip,
						const struct QMatrix *	rref);

#endif	/* ENABLE_VAN_HOEIJ */

/* Standard "max" of two integers. */

#define _MAX(A,B)	(((A) > (B)) ? (A) : (B))

/*
 * Macros to allocate/zero a local polynomial of the given degree.
 *
 * Note that since the underlying mpz_t uses memory that is dynamically
 * allocated from the heap, there is little reason to try and do any
 * fancy stack allocation here.  We put only the "struct calib_Zx_obj" guys
 * on the stack, and dynamically allocate the vectors of __mpz_struct's.
 */

#define ALLOCATE_LOCAL_POLY(PNAME, DEGREE) \
	struct calib_Zx_obj	PNAME; \
	Zx_init_degree (&PNAME, (DEGREE))

#define FREE_LOCAL_POLY(PNAME) \
	Zx_clear (&PNAME)

/*
 * Macros to allocate/zero an array of local polynomials, each having
 * the given degree.
 */

#define ALLOCATE_LOCAL_POLY_ARRAY(PNAME, NPOLY, DEGREE) \
	struct calib_Zx_obj	PNAME [NPOLY]; \
	Zx_init_array_degree (&PNAME [0], (NPOLY), (DEGREE))

#define FREE_LOCAL_POLY_ARRAY(PNAME, NPOLY) \
	Zx_clear_array (&PNAME [0], (NPOLY))

/*
 * Initialize the given polynomial.  Its value after initialization is zero.
 */

	static
	inline
	void
Zx_init (

struct calib_Zx_obj *	x	/* IN/OUT: polynomial to initialize */
)
{
	CALIB_FATAL_ERROR_IF (x EQ NULL);

	x -> degree	= -1;
	x -> size	= 0;
	x -> coeff	= NULL;
}

/*
 * Initialize the given polynomial to have the given constant value.
 */

	static
	void
Zx_init_si (

struct calib_Zx_obj *	x,	/* IN/OUT: polynomial to initialize */
calib_si_t		op	/* IN: initial value */
)
{
mpz_ptr			zp;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	if (op EQ 0) {
		Zx_init (x);
	}
	else {
		zp = calib_new_Z_vector (1);
		mpz_set_si (zp, op);

		x -> degree	= 0;
		x -> size	= 1;
		x -> coeff	= zp;
	}
}

/*
 * Initialize the given polynomial to zero, but provide it with a
 * coefficient buffer sufficient to hold a polynomial of the given degree.
 */

	static
	inline
	void
Zx_init_degree (

struct calib_Zx_obj *	poly,	/* IN: polynomial to initialize */
int			degree	/* IN: maximum degree */
)
{
int		size;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (degree < -1));

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	poly -> degree	= -1;
	poly -> size	= size;
	poly -> coeff	= calib_new_Z_vector (size);
}

/*
 * Initialize the given array of polynomials to zero, but provide each
 * with a coefficient buffer sufficient to hold a polynomial of the
 * given degree.
 *
 * Although it is tempting to allocate all of the coefficients for all of
 * the polynomials into one big long vector, we cannot do so because the
 * coefficient vector for each polynomial must be separately free-able,
 * and re-allocatable.
 */

	static
	inline
	void
Zx_init_array_degree (

struct calib_Zx_obj *	parray,	/* IN/OUT: array of polynomials */
int			npoly,	/* IN: number of polynomials in array */
int			degree	/* IN: max degree of each polynomial */
)
{
int			i, size;
struct calib_Zx_obj *	pp;

	CALIB_FATAL_ERROR_IF ((parray EQ NULL) OR
			      (npoly   < 0) OR
			      (degree  < -1));

	size = degree + 1;
	if (size <= 0) {
		size = 1;
	}

	pp = &parray [0];
	for (i = 0; i < npoly; i++, pp++) {
		pp -> degree	= -1;
		pp -> size	= size;
		pp -> coeff	= calib_new_Z_vector (size);
	}
}

/*
 * Force the given polynomial to have buffer space sufficient to hold
 * a polynomial of at least the given degree.  This does not change
 * the value currently residing in this polynomial.
 */

	static
	void
Zx_alloc (

struct calib_Zx_obj *	rop,		/* IN/OUT: poly to alloc buffer for */
int			degree		/* IN: degree desired */
)
{
int		size, dsize;
mpz_ptr		zp, dp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (degree < -1));

	size	= degree + 1;
	dsize	= rop -> size;

	if (dsize < size) {
		dp = rop -> coeff;
		zp = calib_new_Z_vector (size);
		calib_copy_Z_vector (zp, dp, dsize);
		calib_free_Z_vector (dp, dsize);
		rop -> coeff = zp;
		rop -> size  = size;
	}
}

/*
 * Clear out the given polynomial.
 */

	static
	inline
	void
Zx_clear (

struct calib_Zx_obj *	x	/* IN/OUT: polynomial to clear */
)
{
int		size;
mpz_ptr		coeff;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	coeff = x -> coeff;
	if (coeff NE NULL) {
		size = x -> size;
		CALIB_FATAL_ERROR_IF (size <= 0);
		calib_free_Z_vector (coeff, size);
	}
	Zx_init (x);
}

/*
 * Clear out the given array of polynomials.
 */

	static
	inline
	void
Zx_clear_array (

struct calib_Zx_obj *	parray,	/* IN: array of polynomials to clear */
int			npoly	/* IN: number of polynomials in parray */
)
{
int		i;

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (parray EQ NULL);

	for (i = 0; i < npoly; i++) {
		Zx_clear (&parray [i]);
	}
}

/*
 * Set the given destination polynomial from the given coefficient
 * array.  Reallocate the destination polynomial's coefficient array,
 * if necessary.
 */

	static
	inline
	void
Zx_realloc_copy_poly (

struct calib_Zx_obj *	poly,		/* IN: destination polynomial */
mpz_srcptr		coeff,		/* IN: source coefficient array */
int			degree		/* IN: degree to set */
)
{
int		cur_size, new_size;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	poly -> degree	= degree;

	if (degree < 0) {
		CALIB_FATAL_ERROR_IF (degree < -1);
		/* We allow (coeff EQ NULL) here (but do not require). */
		return;
	}

	CALIB_FATAL_ERROR_IF (coeff EQ NULL);

	new_size = degree + 1;
	if (new_size > 0) {
		cur_size = poly -> size;
		if (new_size > cur_size) {
			if (poly -> coeff NE NULL) {
				calib_free_Z_vector (poly -> coeff, cur_size);
			}
			poly -> coeff = calib_new_Z_vector (new_size);
			poly -> size = new_size;
		}
		calib_copy_Z_vector (poly -> coeff, coeff, new_size);
	}
}

/*
 * Copy the given polynomial.
 */

	static
	void
Zx_set (

struct calib_Zx_obj *		rop,	/* IN/OUT: destination */
const struct calib_Zx_obj *	op	/* IN: source */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (rop NE op) {
		Zx_realloc_copy_poly (rop, op -> coeff, op -> degree);
	}
}

/*
 * Set the polynomial to be the given integer constant.
 */

	static
	void
Zx_set_si (

struct calib_Zx_obj *	rop,		/* IN/OUT: destination */
calib_si_t		op		/* IN: source */
)
{
	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	if (op EQ 0) {
		rop -> degree = -1;
	}
	else {
		Zx_alloc (rop, 0);
		mpz_set_si (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the polynomial to be the given GMP integer constant.
 */

	static
	void
Zx_set_z (

struct calib_Zx_obj *	rop,		/* IN/OUT: destination */
mpz_srcptr		op		/* IN: source */
)
{
	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpz_sgn (op) EQ 0) {
		rop -> degree = -1;
	}
	else {
		Zx_alloc (rop, 0);
		mpz_set (&(rop -> coeff [0]), op);
		rop -> degree = 0;
	}
}

/*
 * Set the polynomial to be x**power (where x is the Zx polynomial variable).
 */

	static
	void
Zx_set_var_power (

struct calib_Zx_obj *	rop,		/* IN/OUT: destination */
int			power		/* IN: power to use */
)
{
int		i;
mpz_ptr		coeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (power < 0));

	Zx_alloc (rop, power);

	coeff = rop -> coeff;
	for (i = 0; i < power; i++) {
		mpz_set_ui (&coeff [i], 0);
	}
	mpz_set_ui (&coeff [power], 1);

	rop -> degree = power;
}

/*
 * Addition of two polynomials over Z.
 */

	static
	void
Zx_add (

struct calib_Zx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zx_obj *	op1,		/* IN: first operand */
const struct calib_Zx_obj *	op2		/* IN: second operand */
)
{
int			i, rdeg, deg1, deg2, rsize;
mpz_ptr			rp, res;
mpz_srcptr		cp1, cp2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = calib_new_Z_vector (rsize);

	cp1 = &(op1 -> coeff [0]);
	cp2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, rp++, cp1++, cp2++) {
		if (i <= deg1) {
			if (i <= deg2) {
				mpz_add (rp, cp1, cp2);
			}
			else {
				mpz_set (rp, cp1);
			}
		}
		else {
			mpz_set (rp, cp2);
		}
	}
	while ((rdeg >= 0) AND (mpz_sgn (&res [rdeg]) EQ 0)) {
		--rdeg;
	}

	Zx_realloc_copy_poly (rop, res, rdeg);
	calib_free_Z_vector (res, rsize);
}

/*
 * Subtraction of two polynomials over Z.
 */

	static
	void
Zx_sub (

struct calib_Zx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zx_obj *	op1,		/* IN: first operand */
const struct calib_Zx_obj *	op2		/* IN: second operand */
)
{
int			i, rdeg, deg1, deg2, rsize;
mpz_ptr			rp, res;
mpz_srcptr		cp1, cp2;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	rdeg	= _MAX (deg1, deg2);

	if (rdeg < 0) {
		rop -> degree = -1;
		return;
	}

	rsize = rdeg + 1;
	res = calib_new_Z_vector (rsize);

	cp1 = &(op1 -> coeff [0]);
	cp2 = &(op2 -> coeff [0]);
	rp = &res [0];
	for (i = 0; i <= rdeg; i++, rp++, cp1++, cp2++) {
		if (i <= deg1) {
			if (i <= deg2) {
				mpz_sub (rp, cp1, cp2);
			}
			else {
				mpz_set (rp, cp1);
			}
		}
		else {
			mpz_neg (rp, cp2);
		}
	}
	while ((rdeg >= 0) AND (mpz_sgn (&res [rdeg]) EQ 0)) {
		--rdeg;
	}

	Zx_realloc_copy_poly (rop, res, rdeg);
	calib_free_Z_vector (res, rsize);
}

/*
 * Negation of a polynomial over Z.
 */

	static
	void
Zx_neg (

struct calib_Zx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zx_obj *	op		/* IN: polynomial to negate */
)
{
int			i, deg, size;
mpz_ptr			rp, res;
mpz_srcptr		cp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	deg	= op -> degree;

	if (rop EQ op) {
		/* Negation in place. */
		rp = &(rop -> coeff [0]);
		for (i = 0; i <= deg; i++, rp++) {
			mpz_neg (rp, rp);
		}
	}
	else {
		size = deg + 1;
		res = calib_new_Z_vector (size);

		cp = &(op -> coeff [0]);
		rp   = &res [0];
		for (i = 0; i <= deg; i++, rp++, cp++) {
			mpz_neg (rp, cp);
		}

		Zx_realloc_copy_poly (rop, res, deg);
		calib_free_Z_vector (res, size);
	}
}

/*
 * Multiplcation of two polynomials over Z.
 */

	static
	void
Zx_mul (

struct calib_Zx_obj *		rop,		/* IN/OUT: result */
const struct calib_Zx_obj *	op1,		/* IN: first operand */
const struct calib_Zx_obj *	op2		/* IN: second operand */
)
{
int			i, j, rdeg, deg1, deg2, rsize;
mpz_ptr			res, rp1, rp2;
mpz_srcptr		cp1, cp2;
mpz_t			TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg1	= op1 -> degree;
	deg2	= op2 -> degree;
	if ((deg1 < 0) OR (deg2 < 0)) {
		/* Multipying by zero. */
		rop -> degree = -1;
		return;
	}

	rdeg	= deg1 + deg2;

	rsize = rdeg + 1;
	res = calib_new_Z_vector (rsize);

	mpz_init (TEMP);

	cp1 = &(op1 -> coeff [0]);
	rp1 = &res [0];
	for (i = 0; i <= deg1; i++, rp1++, cp1++) {
		rp2 = rp1;
		cp2 = &(op2 -> coeff [0]);
		for (j = 0; j <= deg2; j++, rp2++, cp2++) {
			mpz_mul (TEMP, cp1, cp2);
			mpz_add (rp2, rp2, TEMP);
		}
	}
	while ((rdeg >= 0) AND (mpz_sgn (&res [rdeg]) EQ 0)) {
		--rdeg;
	}

	mpz_clear (TEMP);

	Zx_realloc_copy_poly (rop, res, rdeg);
	calib_free_Z_vector (res, rsize);
}

/*
 * Multiply the given polynomial by the given integer.
 */

	static
	void
Zx_mul_z (

struct calib_Zx_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_Zx_obj *	op1,	/* IN: polynomial to multiply */
mpz_srcptr			op2	/* IN: integer to multiply by */
)
{
int		i, deg;
mpz_srcptr	coeff;
mpz_ptr		res;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg	= op1 -> degree;
	coeff	= op1 -> coeff;
	res	= calib_new_Z_vector (deg + 1);
	for (i = 0; i <= deg; i++) {
		mpz_mul (&res [i], &coeff [i], op2);
	}
	Zx_realloc_copy_poly (rop, res, deg);
	calib_free_Z_vector (res, deg + 1);
}

/*
 * Take the given polynomial to the given integer power.
 */

	static
	void
Zx_ipow (

struct calib_Zx_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_Zx_obj *	op,	/* IN: polynomial to exponentiate */
int				power	/* IN: integer power */
)
{
struct calib_Zx_obj	RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (power < 0));

	Zx_init_si (&RES, 1);
	Zx_init (&PREG);
	Zx_set (&PREG, op);

	while (power > 0) {
		if ((power & 1) NE 0) {
			Zx_mul (&RES, &RES, &PREG);
		}
		Zx_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	Zx_set (rop, &RES);
	Zx_clear (&PREG);
	Zx_clear (&RES);
}

/*
 * Create a dynamically allocated copy of the given polynomial.
 */

	static
	struct calib_Zx_obj *
Zx_dup (

const struct calib_Zx_obj *	op	/* IN: polynomial to duplicate */
)
{
int			deg, size;
struct calib_Zx_obj *	dst;
mpz_ptr			zp;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	size = deg + 1;

	dst = CALIB_NEW (struct calib_Zx_obj);
	zp = NULL;
	if (size > 0) {
		zp = calib_new_Z_vector (size);
		calib_copy_Z_vector (zp, op -> coeff, size);
	}
	dst -> degree	= deg;
	dst -> size	= size;
	dst -> coeff	= zp;

	return (dst);
}

/*
 * Free the given polynomial, which must have been dynamically allocated.
 */

	static
	void
Zx_free (

struct calib_Zx_obj *	poly		/* IN: polynomial to free */
)
{
	if (poly NE NULL) {
		Zx_clear (poly);
		free (poly);
	}
}

/*
 * Evaluate the given poly(value).
 */

	static
	void
Zx_eval (

mpz_ptr				rop,		/* IN/OUT: result */
const struct calib_Zx_obj *	poly,		/* IN: polynomial to eval*/
mpz_srcptr			value		/* IN: value to use */
)
{
int			i;
mpz_srcptr		coeff;
mpz_t			SUM;

	CALIB_FATAL_ERROR_IF ((rop   EQ NULL) OR
			      (poly  EQ NULL) OR
			      (value EQ NULL));

	mpz_init (SUM);

	coeff = poly -> coeff;

	for (i = poly -> degree; i >= 0; i--) {
		mpz_mul (SUM, SUM, value);
		mpz_add (SUM, SUM, &coeff [i]);
	}

	mpz_set (rop, SUM);
	mpz_clear (SUM);
}

/*
 * The division algorithm (with remainder):
 * Given polynomials a, b \in Z[x], find polynomials q, r \in Z[x]
 * and constant d \in Z such that:
 *
 *	d*a = q*b + r	AND	degree (r) < degree (b)
 *			AND	d >= 1
 *			AND	d is minimal
 *
 * Yes, this is messy because the coefficients are in Z which is a ring,
 * not a field.
 */

	static
	void
Zx_div (

struct calib_Zx_obj *		quotient,	/* IN/OUT: quotient */
struct calib_Zx_obj *		remainder,	/* IN/OUT: remainder */
mpz_ptr				d,		/* IN/OUT: denominator */
const struct calib_Zx_obj *	a,		/* IN: dividend */
const struct calib_Zx_obj *	b		/* IN: divisor */
)
{
int			i, j, adeg, bdeg, qdeg, rdeg, rsize;
mpz_ptr			rp, quot, rem;
mpz_srcptr		ap, bp, lcb, lcr;
struct calib_Zx_obj	ppart;
mpz_t			DVAL, CONTENT, G, M, N, TEMP;

	CALIB_FATAL_ERROR_IF ((a EQ NULL) OR (b EQ NULL));

	adeg	= a -> degree;
	bdeg	= b -> degree;

	rsize = adeg + 1;

	/* Local coefficient buffers. */
	quot = calib_new_Z_vector (2 * rsize);
	rem  = quot + rsize;

	bp = &(b -> coeff [0]);

	mpz_init (DVAL);
	mpz_init (G);
	mpz_init (M);

	if (bdeg <= 0) {
		/* Division by a constant.  Do not allow division by zero. */
		CALIB_FATAL_ERROR_IF ((bdeg < 0) OR (mpz_sgn (bp) EQ 0));

		/* Compute content and primitive part of dividend. */
		mpz_init (CONTENT);
		Zx_init (&ppart);
		Zx_prim_part (CONTENT, &ppart, a);

		mpz_gcd (G, CONTENT, bp);
		mpz_divexact (M, CONTENT, G);
		mpz_divexact (DVAL, bp, G);

		rp = &quot [0];
		ap = &ppart.coeff [0];
		for (i = 0; i <= adeg; i++) {
			mpz_mul (rp++, ap++, M);
		}
		qdeg = adeg;
		Zx_clear (&ppart);
		mpz_clear (CONTENT);
		rdeg = -1;
	}
	else {
		mpz_init (N);
		mpz_init (TEMP);

		mpz_set_ui (DVAL, 1);

		qdeg = adeg - bdeg;
		if (qdeg < -1) {
			qdeg = -1;
		}
		calib_copy_Z_vector (rem, &(a -> coeff [0]), adeg + 1);
		rdeg = adeg;

		lcb = &bp [bdeg];

		while (rdeg >= bdeg) {
			j = rdeg - bdeg;
			lcr = &rem [rdeg];
			mpz_gcd (G, lcr, lcb);
			mpz_divexact (M, lcr, G);
			mpz_divexact (N, lcb, G);

			for (i = qdeg; i > j; i--) {
				mpz_mul (&quot [i], &quot [i], N);
			}
			for (i = rdeg - 1; i >= 0; i--) {
				mpz_mul (&rem [i], &rem [i], N);
			}

			mpz_mul (DVAL, DVAL, N);
			mpz_set (&quot [j], M);

			rp = &rem [rdeg - 1];
			bp = &(b -> coeff [bdeg - 1]);
			for (i = bdeg - 1; i >= 0; i--) {
				mpz_mul (TEMP, M, bp);
				mpz_sub (rp, rp, TEMP);
				--rp;
				--bp;
			}
			--rdeg;
			rp = &rem [rdeg];
			while ((rdeg >= 0) AND (mpz_sgn (rp) EQ 0)) {
				--rdeg;
				--rp;
			}
		}

		mpz_clear (TEMP);
		mpz_clear (N);
	}

	if (quotient NE NULL) {
		Zx_realloc_copy_poly (quotient, &quot [0], qdeg);
	}

	if (remainder NE NULL) {
		Zx_realloc_copy_poly (remainder, &rem [0], rdeg);
	}

	if (d NE NULL) {
		mpz_set (d, DVAL);
	}

	mpz_clear (M);
	mpz_clear (G);
	mpz_clear (DVAL);

	calib_free_Z_vector (quot, 2 * rsize);
}

/*
 * Return TRUE if-and-only-if a is exactly divisible by b.
 * If so, then a = b * quotient.  Returns FALSE otherwise (without
 * modifying the quotient).
 */

	static
	calib_bool
Zx_exactly_divides (

struct calib_Zx_obj *		quotient,	/* IN/OUT: quotient */
const struct calib_Zx_obj *	a,		/* IN: dividend */
const struct calib_Zx_obj *	b		/* IN: divisor */
)
{
int			i, j, adeg, bdeg, qdeg, rdeg, rsize;
calib_bool		success;
mpz_ptr			rp, quot, rem;
mpz_srcptr		ap, bp, lcb, lcr;
mpz_t			M, N, TEMP;

	CALIB_FATAL_ERROR_IF ((a EQ NULL) OR (b EQ NULL));

	adeg	= a -> degree;
	bdeg	= b -> degree;

	rsize = adeg + 1;

	/* Local coefficient buffers. */
	quot = calib_new_Z_vector (2 * rsize);
	rem  = quot + rsize;

	bp = &(b -> coeff [0]);

	mpz_init (M);

	success = TRUE;

	if (bdeg <= 0) {
		/* Division by a constant.  Do not allow division by zero. */
		CALIB_FATAL_ERROR_IF ((bdeg < 0) OR (mpz_sgn (bp) EQ 0));

		ap = &(a -> coeff [0]);
		qdeg = adeg;
		rp = &quot [0];
		for (i = 0; i <= adeg; i++) {
			mpz_tdiv_qr (rp, M, ap, bp);
			if (mpz_sgn (M) NE 0) {
				success = FALSE;
				break;
			}
			++rp;
			++ap;
			++bp;
		}
	}
	else {
		mpz_init (N);
		mpz_init (TEMP);

		qdeg = adeg - bdeg;
		if (qdeg < -1) {
			qdeg = -1;
		}
		calib_copy_Z_vector (rem, &(a -> coeff [0]), adeg + 1);
		rdeg = adeg;

		lcb = &bp [bdeg];

		for (;;) {
			if (rdeg < bdeg) {
				if (rdeg > 0) {
					/* We have a remainder, so the	*/
					/* division is not exact.	*/
					success = FALSE;
				}
				break;
			}

			j = rdeg - bdeg;
			lcr = &rem [rdeg];
			
			mpz_tdiv_qr (M, N, lcr, lcb);
			if (mpz_sgn (N) NE 0) {
				/* Division is not exact! */
				success = FALSE;
				break;
			}

			mpz_set (&quot [j], M);

			rp = &rem [rdeg - 1];
			bp = &(b -> coeff [bdeg - 1]);
			for (i = bdeg - 1; i >= 0; i--) {
				mpz_mul (TEMP, M, bp);
				mpz_sub (rp, rp, TEMP);
				--rp;
				--bp;
			}
			--rdeg;
			rp = &rem [rdeg];
			while ((rdeg >= 0) AND (mpz_sgn (rp) EQ 0)) {
				--rdeg;
				--rp;
			}
		}

		mpz_clear (TEMP);
		mpz_clear (N);
	}

	if (success AND (quotient NE NULL)) {
		Zx_realloc_copy_poly (quotient, &quot [0], qdeg);
	}

	mpz_clear (M);

	calib_free_Z_vector (quot, 2 * rsize);

	return (success);
}

/*
 * Divide the given polynomial in Z[x] by the given integer.
 * It is a fatal error if the division is not exact.
 */

	static
	void
Zx_div_z_exact (

struct calib_Zx_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_Zx_obj *	op1,	/* IN: polynomial to divide */
mpz_srcptr			op2	/* IN: integer to divide by */
)
{
int		i, deg;
mpz_srcptr	src;
mpz_ptr		res;
mpz_t		REM;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	deg = op1 -> degree;
	src = op1 -> coeff;
	res = calib_new_Z_vector (deg + 1);
	mpz_init (REM);
	for (i = 0; i <= deg; i++) {
		mpz_tdiv_qr (&res [i], REM, &src [i], op2);
		CALIB_FATAL_ERROR_IF (mpz_sgn (REM) NE 0);
	}
	mpz_clear (REM);
	Zx_realloc_copy_poly (rop, res, deg);
	calib_free_Z_vector (res, deg + 1);
}

/*
 * Compute the Greatest Common Divisor of two polynomials in Z[x].
 *
 * This is just a stub that chooses between algorithms.
 */

	static
	void
Zx_gcd (

struct calib_Zx_obj *		gcd,	/* IN/OUT: GCD result */
const struct calib_Zx_obj *	a,	/* IN: first operand */
const struct calib_Zx_obj *	b	/* IN: second operand */
)
{
	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	switch (calib_Zx_settings.gcd_method) {
	case CALIB_ZX_GCD_METHOD_MODULAR:
		Zx_gcd_modular (gcd, a, b);
		break;

	case CALIB_ZX_GCD_METHOD_PRS:
		Zx_gcd_prs (gcd, a, b);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * A better implementation of gcd() that just invokes gcd_n() with n=2.
 * This gains us the efficiency of modular methods.
 */

	static
	void
Zx_gcd_modular (

struct calib_Zx_obj *		gcd,	/* IN/OUT: GCD result */
const struct calib_Zx_obj *	a,	/* IN: first operand */
const struct calib_Zx_obj *	b	/* IN: second operand */
)
{
int				n;
const struct calib_Zx_obj *	poly_ptrs [2];

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	n = 2;

	poly_ptrs [0]	= a;
	poly_ptrs [1]	= b;

	Zx_gcd_n (gcd, NULL, n, poly_ptrs);
}

/*
 * Compute the Greatest Common Divisor of two polynomials in Z[x].
 * There are lots of ways of defining this operation, and they are
 * not all equivalent.  First, we reduce the problem to polynomials
 * that are primitive, as follows:
 *
 *	gcd(a,b) = gcd(content(a),content(b)) * gcd(pp(a),pp(b))
 *
 * When a and b are both primitive, then we define:
 *
 *	gcd(a(x),b(x)) = g(x)
 *
 * such that there exist c(x), d(x) \in Z[x] and p,q,r,s \in Z such that
 *
 *	p*a(x) = q*g(x)*c(x),
 *	r*b(x) = s*g(x)*d(x),
 *	g(x), c(x) and d(x) are primitive,
 *	g(x) is of maximal degree,
 *	p,q,r,s >= 1,
 *	lc(g(x)) >= 1,
 *	gcd(p,q) = 1,
 *	gcd(r,s) = 1.
 *
 * These criteria allow us to to immediately write down z(x) = gcd(a,b)
 * over Q[x], where z(x) is the unique, monic greatest common divisor:
 *
 *		z(x) = g(x)/lc(g(x)),
 *		a(x) = z(x) * (q/p)*lc(g(x))*c(x),
 *		b(x) = z(x) * (r/s)*lc(g(x))*d(x).
 *
 * We use a fully reduced "primitive" Polynomial Remainder Sequence iteration.
 * Other options:
 *
 * 1. A modular GCD should be much faster when the result is 1 or of low
 *    degree, especially for inputs of high degree.
 *
 * 2. A true "subresultant" algorithm would be faster at the expense of
 *    slightly higher coefficient growth.
 */

	static
	void
Zx_gcd_prs (

struct calib_Zx_obj *		gcd,	/* IN/OUT: GCD result */
const struct calib_Zx_obj *	a,	/* IN: first operand */
const struct calib_Zx_obj *	b	/* IN: second operand */
)
{
int			i, adeg, bdeg;
mpz_ptr			ap, bp, lca, lcb;
struct calib_Zx_obj *	res;
struct calib_Zx_obj	a_ppart, b_ppart;
mpz_t			ACONT, BCONT, GCONT;
mpz_t			G, M, N, TEMP;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	adeg = a -> degree;
	bdeg = b -> degree;

	/* Compute primitive-parts and contents into local copies. */
	mpz_init (ACONT);
	mpz_init (BCONT);
	mpz_init (GCONT);
	Zx_init (&a_ppart);
	Zx_init (&b_ppart);
	Zx_prim_part (ACONT, &a_ppart, a);
	Zx_prim_part (BCONT, &b_ppart, b);
	mpz_gcd (GCONT, ACONT, BCONT);

	mpz_init (G);
	mpz_init (M);
	mpz_init (N);
	mpz_init (TEMP);

	/* FIXME: We could use modular GCD here.  A really simple way	*/
	/* to get there would be to just use gcd_n() with n=2.		*/

	for (;;) {
		if (adeg < 0) {
			b_ppart.degree = bdeg;
			res = &b_ppart;
			break;
		}
		if (bdeg < 0) {
			a_ppart.degree = adeg;
			res = &a_ppart;
			break;
		}
		lca = &a_ppart.coeff [adeg];
		lcb = &b_ppart.coeff [bdeg];
		mpz_gcd (G, lca, lcb);
		mpz_divexact (M, lca, G);
		mpz_divexact (N, lcb, G);
		if (adeg >= bdeg) {
			/* a(x) <== N*a(x) - M*b(x)*x^j */
			ap = lca - 1;
			for (i = adeg - 1; i >= 0; i--, ap--) {
				mpz_mul (ap, ap, N);
			}
			ap = lca - 1;
			bp = lcb - 1;
			for (i = bdeg - 1; i >= 0; i--, ap--, bp--) {
				mpz_mul (TEMP, M, bp);
				mpz_sub (ap, ap, TEMP);
			}
			ap = lca;
			do {
				--adeg;
				--ap;
			} while ((adeg >= 0) AND (mpz_sgn (ap) EQ 0));

			/* Make a(x) primitive again, if necessary. */
			a_ppart.degree = adeg;
			Zx_prim_part (NULL, &a_ppart, &a_ppart);
		}
		else {
			/* b(x) <== M*b(x) - N*a(x)*x^j */
			bp = lcb - 1;
			for (i = bdeg - 1; i >= 0; i--, bp--) {
				mpz_mul (bp, bp, M);
			}
			ap = lca - 1;
			bp = lcb - 1;
			for (i = adeg - 1; i >= 0; i--, ap--, bp--) {
				mpz_mul (TEMP, N, ap);
				mpz_sub (bp, bp, TEMP);
			}
			bp = lcb;
			do {
				--bdeg;
				--bp;
			} while ((bdeg >= 0) AND (mpz_sgn (bp) EQ 0));

			/* Make b(x) primitive again, if necessary. */
			b_ppart.degree = bdeg;
			Zx_prim_part (NULL, &b_ppart, &b_ppart);
		}
	}

	if (mpz_cmp_ui (GCONT, 1) > 0) {
		/* Apply constant factor to gcd. */
		ap = &(res -> coeff [0]);
		for (i = 0; i <= res -> degree; i++, ap++) {
			mpz_mul (ap, ap, GCONT);
		}
	}

	if ((res -> degree >= 0) AND
	    (mpz_sgn (&(res -> coeff [res -> degree])) < 0)) {
		/* Make lead-coefficient be positive. */
		ap = &(res -> coeff [0]);
		for (i = 0; i <= res -> degree; i++, ap++) {
			mpz_neg (ap, ap);
		}
	}

	Zx_set (gcd, res);

	mpz_clear (TEMP);
	mpz_clear (N);
	mpz_clear (M);
	mpz_clear (G);

	Zx_clear (&b_ppart);
	Zx_clear (&a_ppart);
	mpz_clear (GCONT);
	mpz_clear (BCONT);
	mpz_clear (ACONT);
}

/*
 * Compute the GCD of the given array of n polynomials, as well as
 * the array of "cofactor" polynomials (the original polynomials with
 * the GCD removed).
 *
 * One application for this routine is in computing contents of
 * multi-variate polynomials via the EEZGCD algorithm, which substitutes
 * "random" integers for all but one variable.  We get a distinct
 * univariate polynomial for each coefficient of the original multi-variate
 * polynomial.  In most cases the GCD is going to be 1, so it pays to use
 * a modular GCD method, and to compute the GCD of all n polynomials over
 * Zp[x] before trying to lift the Zp[x] GCD into Z[x] (e.g., using p-adic
 * lifting).  In fact, we should only convert the given Z[x] polynomials
 * into Zp[x] form as they are needed, in case we get a GCD (in Zp[x]) of
 * 1 before all of the input polynomials have been considered.
 */

	static
	void
Zx_gcd_n (

struct calib_Zx_obj *		gcd,		/* IN/OUT: GCD result */
struct calib_Zx_obj **		cofact,		/* IN/OUT: cofactor polys */
int				npoly,		/* IN: # of polynomials */
const struct calib_Zx_obj * const *
				poly_ptrs	/* IN: array of pointers */
						/*     to polynomials to */
						/*     take GCD of */
)
{
int				i, k;
int				pow, min_pow, max_gcd_deg;
int *				order;
const struct calib_Zx_obj *	cp;
struct calib_Zx_obj *		p;
const struct calib_Zx_obj **	pptrs;
struct calib_Zx_obj **		cfptrs;
struct calib_Zx_obj *		poly_array;
struct calib_Zx_obj *		cf_array;
struct calib_Zx_obj		pgcd, REM;
struct calib_Zp_dom *		Zp;
struct calib_Zpx_dom *		Zpx;
struct calib_Zpx_obj *		zpx_array;
struct calib_Zpx_obj *		zpx_cofactor;
struct gcdpgen			pgen;
struct calib_Zpx_obj		zpgcd;
mpz_t				z_cont, g, B, D, Avoid;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (npoly < 0));

	if (npoly EQ 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		gcd -> degree = -1;		/* Result is zero. */
		return;
	}

	CALIB_FATAL_ERROR_IF (poly_ptrs EQ NULL);

	/* Validate inputs.  The cofactors are optional. */
	for (i = 0; i < npoly; i++) {
		CALIB_FATAL_ERROR_IF (poly_ptrs [i] EQ NULL);
	}

	/* Check for special case where all polynomials are zero. */
	for (i = 0; i < npoly; i++) {
		if (poly_ptrs [i] -> degree >= 0) break;
	}
	if (i >= npoly) {
		/* All input polynomials are zero.  Result is zero,	*/
		/* as are all cofactors.				*/
		gcd -> degree = -1;
		if (cofact NE NULL) {
			for (i = 0; i < npoly; i++) {
				if (cofact [i] NE NULL) {
					cofact [i] -> degree = -1;
				}
			}
		}
		return;
	}

	/* Make locally modifiable copies of the two pointer arrays.	*/
	pptrs	= CALIB_NEWA (npoly, const struct calib_Zx_obj *);
	cfptrs	= CALIB_NEWA (npoly, struct calib_Zx_obj *);
	memcpy (pptrs, poly_ptrs, npoly * sizeof (pptrs [0]));
	if (cofact NE NULL) {
		memcpy (cfptrs, cofact, npoly * sizeof (cfptrs [0]));
	}
	else {
		memset (cfptrs, 0, npoly * sizeof (cfptrs [0]));
	}

	/* Remove any polynomials that are zero. */
	k = 0;
	for (i = 0; i < npoly; i++) {
		cp = pptrs [i];
		p  = cfptrs [i];
		if (cp -> degree >= 0) {
			pptrs [k]	= cp;
			cfptrs [k]	= p;
			++k;
		}
		else {
			/* This polynomial is zero -- remove it. */
			if (p NE NULL) {
				p -> degree = -1;
			}
		}
	}
	npoly = k;

	if (npoly EQ 0) {
		gcd -> degree = -1;	/* Result is zero. */
		free (cfptrs);
		free (pptrs);
		return;
	}
	if (npoly EQ 1) {
		/* One polynomial -- a very simple case. */
		Zx_set (gcd, pptrs [0]);
		if (cfptrs [0] NE NULL) {
			Zx_set_si (cfptrs [0], 1);
		}
		Zx_adjust_gcd_n_result_signs (gcd, cfptrs, npoly);
		free (cfptrs);
		free (pptrs);
		return;
	}

	/* Make locally modifiable copies of each polynomial, and also	*/
	/* make an array for the cofactors.  Give each cofactor enough	*/
	/* space to hold the corresponding original polynomial.		*/
	poly_array	= CALIB_NEWA (2 * npoly, struct calib_Zx_obj);
	cf_array	= poly_array + npoly;
	for (i = 0; i < npoly; i++) {
		cp = pptrs [i];
		Zx_init (&poly_array [i]);
		Zx_set (&poly_array [i], cp);

		Zx_init_degree (&cf_array [i], cp -> degree);
	}

	mpz_init (z_cont);
	mpz_init (g);
	mpz_init (B);
	mpz_init (Avoid);
	order = NULL;

	/* Strip off x^k factors from each polynomial. */
	min_pow = Zx_strip_varpow_factor (&poly_array [0]);
	for (i = 1; i < npoly; i++) {
		pow = Zx_strip_varpow_factor (&poly_array [i]);
		if (pow < min_pow) {
			min_pow = pow;
		}
	}

	/* Strip integer content from each polynomial. */
	Zx_strip_integer_content (z_cont, &poly_array [0]);
	for (i = 1; i < npoly; i++) {
		Zx_strip_integer_content (g, &poly_array [i]);
		mpz_gcd (z_cont, z_cont, g);
	}

	/* Compute an upper bound on the degree of the GCD, namely the	*/
	/* smallest degree among the non-zero polynomials.		*/
	max_gcd_deg = INT_MAX;
	for (i = 0; i < npoly; i++) {
		k = poly_array [i].degree;
		if (k < 0) continue;
		if (k < max_gcd_deg) {
			max_gcd_deg = k;
		}
	}
	CALIB_FATAL_ERROR_IF ((max_gcd_deg >= INT_MAX) OR (max_gcd_deg < 0));

	/* Init polynomial to hold the GCD of the primitive polys.	*/
	Zx_init_degree (&pgcd, max_gcd_deg);

	/* If any of the polynomials has degree 0 (i.e., poly is	*/
	/* equal to some non-zero constant), then we are done -- the	*/
	/* result is just z_cont.					*/
	if (max_gcd_deg EQ 0) {
		mpz_set_si (&pgcd.coeff [0], 1);
		pgcd.degree = 0;
		goto finish;
	}

	/* Compute an upper bound B on the magnitude of the	*/
	/* coefficients in the GCD of the given polynomials.	*/
	/* Also compute the LCM of the leading coefficients	*/
	/* of all of the polynomials -- this lets us avoid	*/
	/* picking primes that decrease the degree of any of	*/
	/* of the polynomials.  Having all of the polynomails	*/
	/* be full-degree (mod p) means that all of them can	*/
	/* potentially be used for lifting from Zp[x] to Z[x].	*/
	Zx_upper_bound_gcd_coeff (B, Avoid, npoly, poly_array);

	/* Sort the polynomials from lowest degree to highest. */
	order = Zx_sort_gcd_polys (npoly, poly_array);

	/* Force first GCD to be accepted as a "new best degree".	*/
	++max_gcd_deg;

	/* Construct the Zp and Zp[x] domains, and scratch polynomials	*/
	/* for doing our arithmetic in.  Use p=3 just to get started --	*/
	/* we actually modify this choice each time through the loop	*/
	/* below.							*/

	Zp = calib_make_Zp_dom_ui (3);
	Zpx = calib_make_Zpx_dom (Zp);
	Zpx -> init (Zpx, &zpgcd);
	zpx_array = CALIB_NEWA (2 * npoly, struct calib_Zpx_obj);
	zpx_cofactor = zpx_array + npoly;
	for (i = 0; i < npoly; i++) {
		k = poly_array [i].degree;
		Zpx -> init_degree (Zpx, &zpx_array [i], k);
		Zpx -> init_degree (Zpx, &zpx_cofactor [i], k);
	}

	/* Initialize the prime number generator. */
	pgen.Zp		= Zp;
	pgen.avoid	= Avoid;
	pgen.tries	= 0;
	pgen.nwords	= 0;
	calib_random_init (&pgen.rndgen);

	/* The main modular GCD loop.  Each iteration of this loop	*/
	/* tries a different prime modulus.  Our prime generator	*/
	/* produces a stream of primes that ratchets up rapidly in	*/
	/* magnitude, so it shouldn't take too long before we find a	*/
	/* prime that does the trick.					*/

	for (;;) {
		Zx_choose_gcd_prime (&pgen);

		Zx_compute_mod_p_gcd (&zpgcd,
				      npoly,
				      order,
				      poly_array,
				      zpx_array,
				      zpx_cofactor);

		if (zpgcd.degree <= 0) {
			CALIB_FATAL_ERROR_IF (zpgcd.degree < 0);
			/* GCD of primitive polys is 1! */
			mpz_set_ui (&pgcd.coeff [0], 1);
			pgcd.degree = 0;
			break;
		}

		if (zpgcd.degree > max_gcd_deg) {
			/* Degree too high.  Try next prime. */
			continue;
		}
		if (zpgcd.degree < max_gcd_deg) {
			/* We have a new, lower degree GCD over Zp[x]. */
			max_gcd_deg = zpgcd.degree;
		}

		if (Zx_lift_mod_p_gcd_to_z (&zpgcd,
					    &pgcd,
					    B,
					    npoly,
					    order,
					    poly_array,
					    cf_array,
					    zpx_array,
					    zpx_cofactor)) {
			/* This routine only returns successfully when	*/
			/* pgcd evenly divides all of the primitive	*/
			/* polynomials in the poly_array.  This means	*/
			/* we have our GCD over Z[x]! */
			break;
		}

		/* Lifting to Z failed.  Degree of true GCD must be lower. */
		--max_gcd_deg;
		if (max_gcd_deg <= 0) {
			/* True GCD (of primitive polys) must be 1. */
			mpz_set_ui (&pgcd.coeff [0], 1);
			pgcd.degree = 0;
			break;
		}
		/* Try next prime. */
	}
	k = 2 * npoly;
	for (i = 0; i < k; i++) {
		Zpx -> clear (&zpx_array [i]);
	}
	free (zpx_array);
	Zpx -> clear (&zpgcd);
	calib_free_Zpx_dom (Zpx);
	calib_free_Zp_dom (Zp);

finish:
	/* Return the solution to the caller. */
	if (mpz_cmp_ui (z_cont, 1) NE 0) {
		Zx_mul_z (&pgcd, &pgcd, z_cont);
	}
	if (min_pow > 0) {
		/* Multiply by x**min_pow (shift coefficients). */
		Zx_alloc (&pgcd, pgcd.degree + min_pow);
		for (i = pgcd.degree; i >= 0; i--) {
			mpz_set (&pgcd.coeff [i + min_pow], &pgcd.coeff [i]);
		}
		for (i = 0; i < min_pow; i++) {
			mpz_set_ui (&pgcd.coeff [i], 0);
		}
		pgcd.degree += min_pow;
	}
	Zx_set (gcd, &pgcd);
	if (cfptrs NE NULL) {
		if ((pgcd.degree EQ 0) AND
		    (mpz_cmp_ui (&pgcd.coeff [0], 1) EQ 0)) {
			for (i = 0; i < npoly; i++) {
				if (cfptrs [i] EQ NULL) continue;
				Zx_set (cfptrs [i], pptrs [i]);
			}
		}
		else {
			Zx_init (&REM);
			mpz_init (D);
			for (i = 0; i < npoly; i++) {
				if (cfptrs [i] EQ NULL) continue;
				/* Division must be exact. */
				Zx_div (cfptrs [i], &REM, D, pptrs [i], &pgcd);
				CALIB_FATAL_ERROR_IF ((REM.degree >= 0) OR
						      (mpz_cmp_ui (D, 1) NE 0));
			}
			mpz_clear (D);
			Zx_clear (&REM);
		}
	}

	/* Clean up. */
	if (order NE NULL) {
		free (order);
	}
	Zx_clear (&pgcd);
	mpz_clear (Avoid);
	mpz_clear (B);
	mpz_clear (g);
	mpz_clear (z_cont);
	Zx_clear_array (&poly_array [0], 2 * npoly);
	free (poly_array);
	free (cfptrs);
	free (pptrs);
}

/*
 * We choose to enforce the following convention about the Zx GCD
 * routines: The leading coefficient of the returned GCD shall have a
 * strictly *positive* coefficient.  If this is not the case, this
 * routine flips all the signs in both the GCD and each cofactor.
 */

	static
	void
Zx_adjust_gcd_n_result_signs (

struct calib_Zx_obj *	gcd,		/* IN/OUT: GCD result */
struct calib_Zx_obj **	cofact,		/* IN/OUT: Cofactor results */
int			npoly		/* IN: Number of cofactors */
)
{
int			j, deg, sign;
struct calib_Zx_obj *	cfp;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (npoly < 0));

	deg = gcd -> degree;
	if (deg < 0) {
		/* GCD is zero.  No sign needs to be flipped. */
		return;
	}

	/* Check sign of leading coefficient. */
	sign = mpz_sgn (&(gcd -> coeff [deg]));
	if (sign >= 0) {
		CALIB_FATAL_ERROR_IF (sign EQ 0);

		/* Leading coefficient is positive.  Nothing to do. */
		return;
	}

	/* The leading coefficient is negative.  Negate all results. */
	Zx_neg (gcd, gcd);
	if (cofact NE NULL) {
		for (j = 0; j < npoly; j++) {
			cfp = cofact [j];
			if (cfp NE NULL) {
				Zx_neg (cfp, cfp);
			}
		}
	}
}

/*
 * Strip off the largest x^k factor from the given polynomial, returning
 * the power k actually stripped.  Do this by counting trailing zero
 * coefficients.  The input polynomial is modified to remove this factor.
 */

	static
	int
Zx_strip_varpow_factor (

struct calib_Zx_obj *	poly	/* IN/OUT: polynomial to strip */
)
{
int		i, k, degree;
mpz_ptr		coeff;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	degree	= poly -> degree;
	coeff	= poly -> coeff;

	if (degree < 0) {
		return (0);
	}

	k = 0;
	for (i = 0; i <= degree; i++) {
		if (mpz_sgn (&coeff [i]) NE 0) break;
		++k;
	}

	/* Should not have all coefficients (even leading one!) be zero. */
	CALIB_FATAL_ERROR_IF (i > degree);

	if (k > 0) {
		/* Shift coefficients to divide out this x^k factor. */
		for (i = k; i <= degree; i++) {
			mpz_set (&coeff [i - k], &coeff [i]);
		}
		poly -> degree = degree - k;
	}
	return (k);
}

/*
 * Reduce the given polynomial "in place" to be primitive by computing
 * the GCD of all the coefficients and then dividing by it.
 */

	static
	void
Zx_strip_integer_content (

mpz_ptr			content,	/* IN/OUT: content of polynomial */
struct calib_Zx_obj *	poly		/* IN: polynomial to make primitive */
)
{
int		i, degree;
mpz_ptr		coeff;
mpz_t		g;

	CALIB_FATAL_ERROR_IF ((content EQ NULL) OR (poly EQ NULL));

	degree	= poly -> degree;
	coeff	= poly -> coeff;

	if (degree < 0) {
		mpz_set_ui (content, 0);
		return;
	}

	mpz_init_set (g, &coeff [0]);
	for (i = 1; i <= degree; i++) {
		if (mpz_cmp_ui (g, 1) EQ 0) break;
		mpz_gcd (g, g, &coeff [i]);
	}
	if (mpz_cmp_ui (g, 1) NE 0) {
		for (i = 0; i <= degree; i++) {
			mpz_divexact (&coeff [i], &coeff [i], g);
		}
	}
	mpz_set (content, g);
	mpz_clear (g);
}

/*
 * Compute an upper bound on the magnitude of any coefficient appearing
 * in the GCD of the given polynomials.  Also compute the LCM of the
 * leading coefficients of all the polynomials.  (Any prime that divides
 * this value will reduce the degree of one or more of the polynomials.
 * Although it is only essential that at least one of the polynomials
 * remain full degree (mod p), keeping them all full degree (mod p)
 * means that any of the polynomials may be used for lifting from Zp[x]
 * to Z[x] -- one may still have to address polynomials that are not
 * square-free, but at least all of the polynomials are candidates for
 * the lifting process.)
 *
 * The bound we use here is the classic bound from the Landau-Mignotte
 * inequality, adjusted for use with an arbitrary number of polynomials.
 */

	static
	void
Zx_upper_bound_gcd_coeff (

mpz_ptr			B,		/* IN/OUT: upper bound */
mpz_ptr			Avoid,		/* IN/OUT: LCM of all lead coeffs */
int			npoly,		/* IN: number of polynomials */
const struct calib_Zx_obj *
			poly_array	/* IN: polynomials to upper bound */
)
{
int				i, j, deg, k1;
const struct calib_Zx_obj *	p;
mpz_srcptr			coeff;
mpz_t				K2, K3, B1, B2, SUM, TMP1, CMAG, LC_LCM;

	/* Let k1 be the minimum degree of the polynomials	*/
	/* (ignore polynomials that are zero).			*/
	k1 = INT_MAX;
	for (i = 0; i < npoly; i++) {
		p = &poly_array [i];
		deg = p -> degree;
		if (deg < 0) continue;
		if (deg < k1) {
			k1 = deg;
		}
	}
	CALIB_FATAL_ERROR_IF (k1 >= INT_MAX);

	mpz_init (K2);
	mpz_init (K3);
	mpz_init (B1);
	mpz_init (B2);
	mpz_init (SUM);
	mpz_init (TMP1);
	mpz_init (CMAG);
	mpz_init (LC_LCM);

	/* Compute GCD of leading and trailing coefficients.  We divide	*/
	/* by these down below, so make sure they are non-zero.  (Note	*/
	/* that our caller should have removed trailing zero		*/
	/* coefficients.)  Also compute the LCM of the leading		*/
	/* coefficients.						*/

	mpz_set_ui (LC_LCM, 1);
	for (i = 0; i < npoly; i++) {
		p = &poly_array [i];
		deg = p -> degree;
		if (deg < 0) continue;
		coeff = p -> coeff;
		mpz_gcd (K2, K2, &coeff [deg]);
		mpz_gcd (K3, K3, &coeff [0]);

		mpz_gcd (TMP1, LC_LCM, &coeff [deg]);
		mpz_divexact (LC_LCM, LC_LCM, TMP1);
		mpz_mul (LC_LCM, LC_LCM, &coeff [deg]);
	}
	mpz_set (Avoid, LC_LCM);

	mpz_set_si (B1, -1);
	mpz_set_si (B2, -1);
	for (i = 0; i < npoly; i++) {
		p = &poly_array [i];
		deg = p -> degree;
		if (deg < 0) continue;
		coeff = p -> coeff;
		mpz_set_ui (SUM, 0);
		for (j = 0; j <= deg; j++) {
			mpz_mul (TMP1, &coeff [j], &coeff [j]);
			mpz_add (SUM, SUM, TMP1);
		}
		mpz_sqrtrem (SUM, TMP1, SUM);
		if (mpz_sgn (TMP1) NE 0) {
			mpz_add_ui (SUM, SUM, 1);
		}
		mpz_abs (CMAG, &coeff [deg]);
		mpz_mul (TMP1, K2, SUM);
		mpz_cdiv_q (TMP1, TMP1, CMAG);
		if ((mpz_sgn (B1) < 0) OR (mpz_cmp (TMP1, B1) < 0)) {
			mpz_set (B1, TMP1);
		}

		mpz_abs (CMAG, &coeff [0]);
		mpz_mul (TMP1, K3, SUM);
		mpz_cdiv_q (TMP1, TMP1, CMAG);
		if ((mpz_sgn (B2) < 0) OR (mpz_cmp (TMP1, B2) < 0)) {
			mpz_set (B2, TMP1);
		}
	}

	mpz_mul_2exp (B,
		      (mpz_cmp (B1, B2) <= 0) ? B1 : B2,
		      k1 + 1);

	mpz_clear (LC_LCM);
	mpz_clear (CMAG);
	mpz_clear (TMP1);
	mpz_clear (SUM);
	mpz_clear (B2);
	mpz_clear (B1);
	mpz_clear (K3);
	mpz_clear (K2);
}

/*
 * Sort the given polynomials in order from lowest degree to highest,
 * except that we put all of the "identically zero" polynomials last.
 *
 * We want to process the lowest-degree polynomials first, because
 * (a) lower degree polynomials require the least amount of work to
 * compute the GCD of, and (b) lower degree polynomials are likely to
 * result in a GCD of low degree early on, thereby reducing the cost
 * of GCD'ing with the higher degree polynomials later on.
 *
 * We use original index as the tie-breaker to keep the sort stable.
 */

	static
	int *
Zx_sort_gcd_polys (

int			npoly,		/* IN: number of polynomials */
const struct calib_Zx_obj *
			poly_array	/* IN: array of polynomials */
)
{
int				i, i1, i2, j, k, n;
const struct calib_Zx_obj *	p1;
const struct calib_Zx_obj *	p2;
int *				index;
int *				zlist;

	n = npoly;

	/* Make a list of the polynomial indices, but put all of the	*/
	/* degree zero polynomials last.				*/
	index = CALIB_NEWA (n, int);
	zlist = CALIB_NEWA (n, int);
	j = 0;
	k = 0;
	for (i = 0; i < n; i++) {
		p1 = &poly_array [i];
		if (p1 -> degree < 0) {
			zlist [k++] = i;
		}
		else {
			index [j++] = i;
		}
	}
	n = j;
	for (i = 0; i < k; i++) {
		index [n + i] = zlist [i];
	}
	free (zlist);

	/* Construct the heap via sift-downs, in O(n) time. */
	for (k = n >> 1; k >= 0; k--) {
		j = k;
		for (;;) {
			i = (j << 1) + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the right subchild is greater. */
				i1 = index [i];
				i2 = index [i + 1];
				p1 = &poly_array [i1];
				p2 = &poly_array [i2];
				if ((p2 -> degree > p1 -> degree) OR
				    ((p2 -> degree EQ p1 -> degree) AND
				     (i2 > i1))) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			i1 = index [j];
			i2 = index [i];
			p1 = &poly_array [i1];
			p2 = &poly_array [i2];
			if ((p1 -> degree > p2 -> degree) OR
			    ((p1 -> degree EQ p2 -> degree) AND
			     (i1 > i2))) {
				/* Greatest child is smaller.  Sift-	*/
				/* down is done. */
				break;
			}
			/* Sift down and continue. */
			index [j] = i2;
			index [i] = i1;
			j = i;
		}
	}

	/* Now do actual sorting.  Exchange first/last and sift down. */
	while (n > 1) {
		/* Largest is at index [0], swap with index [n-1],	*/
		/* thereby putting it into final position.		*/
		--n;
		i = index [0];
		index [0] = index [n];
		index [n] = i;

		/* Now restore the heap by sifting index [0] down. */
		j = 0;
		for (;;) {
			i = (j << 1) + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the right subchild is greater. */
				i1 = index [i];
				i2 = index [i + 1];
				p1 = &poly_array [i1];
				p2 = &poly_array [i2];
				if ((p2 -> degree > p1 -> degree) OR
				    ((p2 -> degree EQ p1 -> degree) AND
				     (i2 > i1))) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			i1 = index [j];
			i2 = index [i];
			p1 = &poly_array [i1];
			p2 = &poly_array [i2];
			if ((p1 -> degree > p2 -> degree) OR
			    ((p1 -> degree EQ p2 -> degree) AND
			     (i1 > i2))) {
				/* Greatest child is smaller.  Sift-	*/
				/* down is done. */
				break;
			}
			/* Sift down and continue. */
			index [j] = i2;
			index [i] = i1;
			j = i;
		}
	}

	return (index);
}

/*
 * Choose the next prime modulus to try.
 */

	static
	void
Zx_choose_gcd_prime (

struct gcdpgen *	gp		/* IN/OUT: prime generator info */
)
{
int		i;
unsigned int	word;
mpz_ptr		p;
mpz_t		reg;

 static int smallprimes [] = {
	46337, 46327, 46309, 46307, 46301, 46279, 46273, 46271, -1,
 };

	CALIB_FATAL_ERROR_IF (gp EQ NULL);

	p = gp -> Zp -> p;

	for (;;) {
		do {		/* Used only for "break". */
			/* Initially, use the smallprimes list. */
			if (gp -> nwords EQ 0) {
				i = smallprimes [(gp -> tries)++];
				if (i > 0) {
					mpz_set_ui (p, i);
					break;
				}
				gp -> nwords = 1;
				gp -> tries = 0;
			}

			/* Now using larger multi-word random primes. */
			++(gp -> tries);
			if (gp -> tries > 4) {
				++(gp -> nwords);
				gp -> tries = 0;
			}

			mpz_init (reg);
			for (i = 0; i < gp -> nwords; i++) {
				word = calib_random_u32 (&(gp -> rndgen));
				mpz_mul_2exp (reg, reg, 32);
				mpz_add_ui (reg, reg, word);
			}
			mpz_nextprime (p, reg);
			mpz_clear (reg);
		} while (FALSE);

		/* We have chosen a prime.  Make sure it is not one	*/
		/* that we are supposed to avoid.			*/
		if (NOT mpz_divisible_p (gp -> avoid, p)) {
			/* We have an acceptable prime! */
			break;
		}
		/* Need to try another. */
	}
}

/*
 * Compute the GCD of all of the polynomials over Zp[x].
 *
 * Note that the polys in zpx_array[] and zpx_cofactor have been
 * pre-allocated to hold polynomials of the necessary degree, so there
 * is no need to do allocation checks.
 */

	static
	void
Zx_compute_mod_p_gcd (

struct calib_Zpx_obj *	zpgcd,		/* IN/OUT: GCD over Zp[x] */
int			npoly,		/* IN: number of polynomials */
const int *		order,		/* IN: helpful ordering of polys */
const struct calib_Zx_obj *
			poly_array,	/* IN: polynomials to GCD */
struct calib_Zpx_obj *	zpx_array,	/* IN/OUT: Zp[x] version of polys */
struct calib_Zpx_obj *	zpx_cofactor	/* IN/OUT: Zp[x] cofactors */
)
{
int				i, j, k, deg;
const struct calib_Zpx_dom *	Zpx;
const struct calib_Zx_obj *	p;
struct calib_Zpx_obj *		p2;
const struct calib_Zp_dom *	Zp;
struct calib_Zpx_obj		REM;

	CALIB_FATAL_ERROR_IF (zpgcd EQ NULL);

	Zpx	= zpgcd -> dom;
	Zp	= Zpx -> cf;

	zpgcd -> degree = -1;
	for (i = 0; i < npoly; i++) {
		/* Get next polynomial in recommended ordering. */
		j = order [i];
		p = &poly_array [j];
		deg = p -> degree;
		if (deg < 0) continue;

		p2 = &zpx_array [j];
		for (k = 0; k <= deg; k++) {
			Zp -> set_z (Zp, &(p2 -> coeff [k]), &(p -> coeff [k]));
		}
		p2 -> degree = deg;
		if (zpgcd -> degree < 0) {
			Zpx -> set (zpgcd, p2);
		}
		else {
			Zpx -> gcd (zpgcd, zpgcd, p2);
		}
		if (zpgcd -> degree EQ 0) {
			/* We have a GCD of 1.  We are done. */
			break;
		}
	}

	if (zpgcd -> degree > 0) {
		/* We only need the cofactors if degree > 0. */
		Zpx -> init (Zpx, &REM);
		for (i = 0; i < npoly; i++) {
			Zpx -> div (&zpx_cofactor [i],
				    &REM,
				    &zpx_array [i],
				    zpgcd);
			CALIB_FATAL_ERROR_IF (REM.degree NE -1);
		}
		Zpx -> clear (&REM);
	}
}

/*
 * We have a GCD (mod p) for the given set of polynomials.  In other words:
 *
 *	zpx_cofactor [i] * zpgcd === poly_array [i] (mod p),	for all i.
 *
 * Lift zpgcd to be correct (mod p^k), where p^k > 2*B.  Once this is done,
 * return TRUE if-and-only-if the lifted gcd evenly divides all of the
 * poly_array [i].
 *
 * We use p-adic lifting of the factorization on the left-hand-side of the
 * equation above.  Note that in order to apply such p-adic lifting, we
 * need to be able to find polynomials A and B in Zp[x] such that:
 *
 *	A * zpx_cofactor [i] + B * zpgcd = 1
 *
 * (We only need one particular i for this to work, and we use that particular
 * factorization to lift zpgcd.)  As long as zpx_cofactor [i] and zpgcd are
 * relatively prime, we can easily compute A and B using the extended
 * Euclid algorithm.  Unfortunately, it is possible that zpgcd is NOT
 * relatively prime to zpx_cofactor [i] for EVERY i.  Consider, e.g., the
 * following set of N polynomials:
 *
 *		(x+1)^3 * (x+2)^2 * ... * (x+N)^2
 *		(x+1)^2 * (x+2)^3 * ... * (x+N)^2
 *		...
 *		(x+1)^2 * (x+2)^2 * ... * (x+N)^3
 *
 * The GCD of these polynomials is ((x+1)*(x+2)*...*(x+N))^2, which is
 * not relatively prime to each of the N given polynomials.
 *
 * The algorithm simply tries each i in turn, using the extended Euclid
 * algorithm over Zp[x] to solve A*zpx_cofactor[i] + B*zpgcd = G, (where
 * G = GCD (zpx_cofactor[i], zpgcd)), until we find an i where G=1.
 * If there is no such i, then we use a more complex method that amounts
 * to performing a square-free factorization of each of the given
 * polynomials (over Z[x]), and using the resulting relatively prime
 * polynomials to perform the lifting.  This case is relegated to another
 * routine.
 */

	static
	calib_bool
Zx_lift_mod_p_gcd_to_z (

const struct calib_Zpx_obj *
			zpgcd,		/* IN: GCD over Zp[x] to lift */
struct calib_Zx_obj *	pgcd,		/* IN/OUT: GCD over Z[x] */
mpz_srcptr		cubound,	/* IN: coefficient upper bound */
int			npoly,		/* IN: number of polynomials */
const int *		order,		/* IN: helpful ordering of polys */
const struct calib_Zx_obj *
			poly_array,	/* IN: polynomials in Z[x] */
struct calib_Zx_obj *	cf_array,	/* IN: cofactors in Z[x] */
const struct calib_Zpx_obj *
			zpx_array,	/* IN: Zp[x] versions of poly_array */
const struct calib_Zpx_obj *
			zpx_cofactor	/* IN: cofactors over Zp[x] */
)
{
int				i, j, k, deg;
calib_bool			gcd_is_correct;
const struct calib_Zpx_dom *	Zpx;
mpz_srcptr			p;
mpz_srcptr			lcpp;
const struct calib_Zp_dom *	Zp;
const struct calib_Zx_obj *	pp;
const struct calib_Zpx_obj *	cfp;
struct calib_Zpx_obj		A, B, G, zpxtmp1, zpxtmp2, monic_cofactor;
struct calib_Zx_obj		ZxREM, Residue, Cofactor;
struct calib_Zx_obj		zxtmp1, zxtmp2, zxtmp3;
mpz_t				D, TMP1, TMP2, Modulus, NextModulus, MLB;
mpz_t				LC_INV;

	CALIB_FATAL_ERROR_IF ((zpgcd	    EQ NULL) OR
			      (pgcd	    EQ NULL) OR
			      (cubound	    EQ NULL) OR
			      (npoly	    < 2) OR
			      (poly_array   EQ NULL) OR
			      (cf_array	    EQ NULL) OR
			      (zpx_array    EQ NULL) OR
			      (zpx_cofactor EQ NULL));

	Zpx	= zpgcd -> dom;
	deg	= zpgcd -> degree;

	CALIB_FATAL_ERROR_IF (deg <= 0);

	j = order [0];
	if (deg EQ poly_array [j].degree) {
		/* The GCD has the same degree as one of the given	*/
		/* polynomials (i,e, one of lowest degree).  Since they	*/
		/* are all in Z[x] and primitive, then the given	*/
		/* polynomial must be the actual GCD over Z[x].  No	*/
		/* "heavy lifting" required!				*/
		Zx_set (pgcd, &poly_array [j]);
		goto do_trial_divisions;
	}

	/* Look for an A*zpx_cofactor[i] + B*zpgcd = 1. */
	Zpx -> init (Zpx, &A);
	Zpx -> init (Zpx, &B);
	Zpx -> init (Zpx, &G);
	pp = NULL;
	cfp = NULL;
	for (i = 0; i < npoly; i++) {
		pp  = &poly_array [i];
		cfp = &zpx_cofactor [i];

		if (pp -> degree < 0) continue;
		Zpx -> extgcd (&G, &A, &B, cfp, zpgcd);
		if (G.degree EQ 0) {
			CALIB_FATAL_ERROR_IF (mpz_cmp_ui (&G.coeff [0], 1) NE 0);
			/* Found one! */
			break;
		}
	}
	if (i >= npoly) {
		/* There is no suitable cofactor.  This is the hard	*/
		/* case.  We perform a square-free factorization of	*/
		/* each polynomial, compute pairwise GCD's on the	*/
		/* square-free factors, and combine the results into	*/
		/* the correct final answer.				*/

		Zx_hard_case_gcd_n (pgcd, npoly, poly_array, cf_array);

		Zpx -> clear (&G);
		Zpx -> clear (&B);
		Zpx -> clear (&A);

		gcd_is_correct = TRUE;

		return (gcd_is_correct);
	}

	/* Convert the Zp[x] GCD and selected cofactor into Z[x] form.	*/
	/* The Zp[x] GCD is already monic, but we need to split off the	*/
	/* lead coefficient and force the cofactor to be monic (in	*/
	/* Zp[x]) also.							*/
	/* Note that our caller has pre-allocated pgcd, so it should be	*/
	/* big enough to do this without size checks.			*/
	Zp = Zpx -> cf;
	p  = Zp -> p;
	mpz_init (TMP1);
	mpz_init (TMP2);
	calib_copy_Z_vector (pgcd -> coeff, zpgcd -> coeff, deg + 1);
	pgcd -> degree = deg;
	Zx_mod_out_signed (pgcd, p);
	lcpp = &(pp -> coeff [pp -> degree]);
	mpz_init (LC_INV);
	i = mpz_invert (LC_INV, lcpp, p);
	CALIB_FATAL_ERROR_IF (i EQ 0);
	Zpx -> init (Zpx, &monic_cofactor);
	Zpx -> mul_z (&monic_cofactor, cfp, LC_INV);
#if 0
	/* The first A = A*lc is because cfp has leading coefficient of	*/
	/* (lc mod p), which corrects for making cfp be monic.  We then	*/
	/* set A = A/lc and B = B/lc so that we don't have to multiply	*/
	/* by LC_INV inside the loop below.  The two ops on A cancel.	*/
	Zpx -> mul_z (&A, &A, lcpp);
	Zpx -> mul_z (&A, &A, LC_INV);
#endif
	Zpx -> mul_z (&B, &B, LC_INV);
	k = pp -> degree - zpgcd -> degree;
	CALIB_FATAL_ERROR_IF (cfp -> degree > k);
	Zx_init_degree (&Cofactor, k);
	calib_copy_Z_vector (Cofactor.coeff, monic_cofactor.coeff, k + 1);
	Cofactor.degree = k;
	Zx_mod_out_signed (&Cofactor, p);

	/* Compute lower bound for the modulus, making sure that we	*/
	/* also account for the leading coefficient problem.		*/
	mpz_init_set (MLB, cubound);
	mpz_mul_ui (MLB, MLB, 2);
	mpz_abs (TMP1, lcpp);
	mpz_mul (MLB, MLB, TMP1);

	/* Do lifting. */
	mpz_init_set (Modulus, p);
	mpz_init (NextModulus);
	Zx_init (&Residue);
	Zpx -> init (Zpx, &zpxtmp2);
	Zx_init (&zxtmp2);
	Zx_init (&zxtmp3);
	for (;;) {
		/* Convert lc*pgcd to integer, remove Z content, and	*/
		/* see if it divides pp over Z[x].			*/
		Zx_mul_z (&zxtmp3, pgcd, lcpp);
		Zx_mod_out_signed (&zxtmp3, Modulus);
		Zx_strip_integer_content (TMP1, &zxtmp3);
		Zx_div (NULL, &zxtmp2, TMP1, pp, &zxtmp3);
		mpz_abs (TMP1, TMP1);
		if ((zxtmp2.degree < 0) AND (mpz_cmp_ui (TMP1, 1) EQ 0)) {
			/* We have found the GCD! */
			Zx_set (pgcd, &zxtmp3);
			break;
		}

		if (mpz_cmp (Modulus, MLB) > 0) {
			/* We have made the modulus sufficiently large. */
			Zx_set (pgcd, &zxtmp3);
			break;
		}

		mpz_mul (NextModulus, Modulus, p);

		/* Do linear lifting: p^k ==> p^{k+1}. */
		Zx_mul (&Residue, pgcd, &Cofactor);
		Zx_mul_z (&Residue, &Residue, lcpp);
		Zx_mod_out_signed (&Residue, NextModulus);
		Zx_sub (&Residue, pp, &Residue);
		k = Residue.degree;
		CALIB_FATAL_ERROR_IF (k < 0);

		/* Should be an exact multiple of modulus. */
		for (i = 0; i <= k; i++) {
			mpz_tdiv_qr (&Residue.coeff [i],
				     TMP1,
				     &Residue.coeff [i],
				     Modulus);
			CALIB_FATAL_ERROR_IF (mpz_sgn (TMP1) NE 0);
		}
		/* Don't need to multiply Residue by LC_INV here	*/
		/* because we multiplied LC_INV into A and B.		*/
		Zx_mod_out_signed (&Residue, p);

		/* Lift pgcd.					*/
		/* Type pun to Zp[x] by faking a header.	*/
		zpxtmp1.degree	= Residue.degree;
		zpxtmp1.size	= Residue.degree + 1;
		zpxtmp1.dom	= Zpx;
		zpxtmp1.coeff	= Residue.coeff;
		Zpx -> mul (&zpxtmp2, &zpxtmp1, &A);
		Zpx -> div (NULL, &zpxtmp2, &zpxtmp2, zpgcd);
		/* Type pun back to Z[x]. */
		zxtmp1.degree	= zpxtmp2.degree;
		zxtmp1.size	= zpxtmp2.degree + 1;
		zxtmp1.coeff	= zpxtmp2.coeff;
		Zx_set (&zxtmp2, &zxtmp1);
		Zx_mod_out_signed (&zxtmp2, p);
		Zx_mul_z (&zxtmp2, &zxtmp2, Modulus);
		Zx_add (pgcd, pgcd, &zxtmp2);

		/* Lift Cofactor.				*/
		Zpx -> mul (&zpxtmp2, &zpxtmp1, &B);
		Zpx -> div (NULL, &zpxtmp2, &zpxtmp2, &monic_cofactor);
		zxtmp1.degree	= zpxtmp2.degree;
		zxtmp1.size	= zpxtmp2.degree + 1;
		zxtmp1.coeff	= zpxtmp2.coeff;
		Zx_set (&zxtmp2, &zxtmp1);
		Zx_mod_out_signed (&zxtmp2, p);
		Zx_mul_z (&zxtmp2, &zxtmp2, Modulus);
		Zx_add (&Cofactor, &Cofactor, &zxtmp2);

		mpz_set (Modulus, NextModulus);
	}

	Zx_clear (&zxtmp3);
	Zx_clear (&zxtmp2);
	Zpx -> clear (&zpxtmp2);
	Zx_clear (&Residue);
	mpz_clear (NextModulus);
	mpz_clear (Modulus);
	mpz_clear (MLB);
	Zx_clear (&Cofactor);
	Zpx -> clear (&monic_cofactor);
	mpz_clear (LC_INV);
	mpz_clear (TMP2);
	mpz_clear (TMP1);
	Zpx -> clear (&G);
	Zpx -> clear (&B);
	Zpx -> clear (&A);

do_trial_divisions:

	/* First flip sign of pgcd if its lc is negative. */
	if (mpz_sgn (&(pgcd -> coeff [pgcd -> degree])) < 0) {
		for (i = 0; i <= pgcd -> degree; i++) {
			mpz_neg (&(pgcd -> coeff [i]), &(pgcd -> coeff [i]));
		}
	}

	mpz_init (D);
	Zx_init (&ZxREM);
	gcd_is_correct = TRUE;
	for (i = 0; i < npoly; i++) {
		j = order [i];
		Zx_div (&cf_array [j],
			&ZxREM,
			D,
			&poly_array [j],
			pgcd);
		if ((ZxREM.degree NE -1) OR (mpz_cmp_ui (D, 1) NE 0)) {
			/* Division was not exact. */
			gcd_is_correct = FALSE;
			break;
		}
	}
	Zx_clear (&ZxREM);
	mpz_clear (D);

	return (gcd_is_correct);
}

/*
 * Handle the case where there is no cofactor CF for which
 * P and Q \in Zp[x] exist such that
 *
 *	P * GCD + Q * CF = 1 (mod p)
 *
 * This happens, e.g., for two polynomials of the form A*C^2 and C^3.
 *
 * We perform a square-free factorization of each polynomial, compute
 * pairwise GCD's on the square-free factors, and combine the results
 * into the correct final answer.
 */

	static
	void
Zx_hard_case_gcd_n (

struct calib_Zx_obj *	pgcd,		/* IN/OUT: GCD over Z[x] */
int			npoly,		/* IN: Number of polynomials */
const struct calib_Zx_obj *
			poly_array,	/* IN: Polynomials in Z[x] */
struct calib_Zx_obj *	cf_array	/* IN/OUT: Cofactors in Z[x] */
)
{
int				i, j, deg;
calib_bool			gcd_is_one;
const struct calib_Zx_obj *	poly;
const struct calib_Zx_obj *	src;
struct calib_Zx_obj *		dst;
mpz_ptr				lc;
calib_bool *			plc_is_negative;
struct calib_Zx_obj *		lpoly_array;
struct calib_Zx_factor *	sqf_list;
struct calib_Zx_factor *	fp;
struct SqfList **		hookp;
struct SqfTree			tree;

	CALIB_FATAL_ERROR_IF ((pgcd EQ NULL) OR
			      (npoly < 2) OR
			      (poly_array EQ NULL) OR
			      (cf_array EQ NULL));

	plc_is_negative	= CALIB_NEWA (npoly, calib_bool);
	lpoly_array		= CALIB_NEWA (npoly, struct calib_Zx_obj);
	memset (plc_is_negative, FALSE, npoly * sizeof (plc_is_negative [0]));
	for (i = 0; i < npoly; i++) {
		dst = &lpoly_array [i];
		src = & poly_array [i];
		deg = src -> degree;
		Zx_init_degree (dst, deg);
		Zx_set (dst, src);
		lc = &(dst -> coeff [deg]);
		j = mpz_sgn (lc);
		if (j <= 0) {
			CALIB_FATAL_ERROR_IF (j EQ 0);
			Zx_neg (dst, dst);
			plc_is_negative [i] = TRUE;
		}
		CALIB_FATAL_ERROR_IF (mpz_sgn (lc) <= 0);
	}

	Zx_SqfTree_init (&tree, npoly, lpoly_array);

	/* Compute square-free factorization of first polynomial. */
	poly = &lpoly_array [0];
	sqf_list = Zx_compute_square_free_factorization (poly);

	CALIB_FATAL_ERROR_IF ((sqf_list EQ NULL) OR
			      ((sqf_list -> next EQ NULL) AND
			      (sqf_list -> multiplicity < 2)));

	/* The square-free factors are guaranteed to be pairwise	*/
	/* relatively prime.  Each becomes a single node.		*/
	hookp = &(tree.st_poly_list [0]);
	for (fp = sqf_list; fp NE NULL; fp = fp -> next) {
		hookp = Zx_SqfTree_append (&tree,
					   fp -> factor,
					   fp -> multiplicity,
					   hookp);
	}
	Zx_free_factors (sqf_list);

	gcd_is_one = FALSE;
	for (i = 1; i < npoly; i++) {
		poly = &lpoly_array [i];
		sqf_list = Zx_compute_square_free_factorization (poly);

		hookp = &(tree.st_poly_list [i]);
		for (fp = sqf_list; fp NE NULL; fp = fp -> next) {
			/* Each factor we insert can be split into	*/
			/* several factors in its own linked list of	*/
			/* factors, and this can also split existing	*/
			/* factors already in the tree.			*/
			hookp = Zx_SqfTree_insert (&tree,
						   fp -> factor,
						   fp -> multiplicity,
						   hookp);
		}
		Zx_free_factors (sqf_list);

		/* A quick check for GCD = 1.  If so, we don't have to	*/
		/* process any more of the polynomials.			*/
		gcd_is_one = Zx_SqfTree_compute_min_mult (&tree, i);

		if (gcd_is_one) break;
	}

	if (gcd_is_one) {
		for (i = 0; i < npoly; i++) {
			Zx_set (&cf_array [i], &poly_array [i]);
		}
		Zx_set_si (pgcd, 1);
	}
	else {
		for (i = 0; i < npoly; i++) {
			Zx_SqfTree_mul_factors (&cf_array [i],
						&tree,
						tree.st_poly_node_mult [i],
						tree.st_min_mult);
			if (plc_is_negative [i]) {
				Zx_neg (&cf_array [i], &cf_array [i]);
			}
		}
		Zx_SqfTree_mul_factors (pgcd,
					&tree,
					tree.st_min_mult,
					NULL);
	}

	Zx_SqfTree_clear (&tree);

	for (i = 0; i < npoly; i++) {
		Zx_clear (&lpoly_array [i]);
	}
	free (lpoly_array);
	free (plc_is_negative);
}

/*
 * Initialize the given SqfTree for a problem of the given sizes.
 */

	static
	void
Zx_SqfTree_init (

struct SqfTree *	tp,		/* IN/OUT: Tree to initialize */
int			npoly,		/* IN: Number of polynomials */
const struct calib_Zx_obj *
			poly_array	/* IN: Array of polynomials */
)
{
int		i, k, tdeg, max_nodes;

	memset (tp, 0, sizeof (*tp));

	/* Consider one poly p(x) of degree d.  At worst it splits into	*/
	/* d linear factors.  At worst, there is a node for p(x), then	*/
	/* a node for one linear factor plus another for the rest.  If	*/
	/* we repeat this until it has split completely into d linear	*/
	/* factors, we have consumed 1 + 2*d nodes.  The bound we	*/
	/* compute next follows.					*/
	tdeg = 0;
	for (i = 0; i < npoly; i++) {
		k = poly_array [i].degree;

		/* Should never get here if any of the polynomials are	*/
		/* constant.						*/
		CALIB_FATAL_ERROR_IF (k <= 0);

		tdeg += poly_array [i].degree;
	}
	max_nodes = npoly + tdeg;

	tp -> st_node_size	= max_nodes;
	tp -> st_node_alloc	= 0;
	tp -> st_list_alloc	= 0;
	tp -> st_nleaves	= 0;
	tp -> st_leaves		= CALIB_NEWA (max_nodes, int);
	tp -> st_nodes		= CALIB_NEWA (max_nodes, struct SqfNode);
	tp -> st_poly_list	= CALIB_NEWA (npoly, struct SqfList *);
	tp -> st_lists		= CALIB_NEWA (max_nodes, struct SqfList);
	tp -> st_poly_node_mult		= CALIB_NEWA (npoly, int *);
	tp -> st_fm_array	= CALIB_NEWA (npoly * max_nodes, int);
	tp -> st_min_mult	= CALIB_NEWA (max_nodes, int);
	tp -> st_temp_leaves	= CALIB_NEWA (max_nodes, int);

	memset (tp -> st_poly_list, 0, npoly * sizeof (tp -> st_poly_list [0]));
	for (i = 0; i < npoly; i++) {
		tp -> st_poly_node_mult [i] = &(tp -> st_fm_array [i * max_nodes]);
	}
}

/*
 * Clear out the given SqfTree.
 */

	static
	void
Zx_SqfTree_clear (

struct SqfTree *	tp		/* IN/OUT: SqfTree to clear */
)
{
int				i, size, nnodes;
struct SqfNode *		nodes;
struct SqfNode *		np;
const struct calib_Zx_dom *	dom;

	dom = calib_get_Zx_dom ();

	CALIB_FREE (tp -> st_temp_leaves);
	CALIB_FREE (tp -> st_min_mult);
	CALIB_FREE (tp -> st_fm_array);
	CALIB_FREE (tp -> st_poly_node_mult);
	CALIB_FREE (tp -> st_lists);
	CALIB_FREE (tp -> st_poly_list);

	size	= tp -> st_node_size;
	nnodes	= tp -> st_node_alloc;
	nodes	= tp -> st_nodes;
	CALIB_FATAL_ERROR_IF (nnodes > size);
	if (nodes NE NULL) {
		for (i = 0; i < nnodes; i++) {
			np = &nodes [i];
			dom -> free (np -> sn_factor);
			np -> sn_factor = NULL;
		}
		free (nodes);
	}

	CALIB_FREE (tp -> st_leaves);

	memset (tp, 0, sizeof (*tp));
}

/*
 * Append a new node for the given factor.  It is guaranteed to be
 * relatively prime to all previous nodes.
 * Construct a (list of one) factor(s) that we append at the given
 * hookp location.  This factor has the given multiplicity.
 */

	static
	struct SqfList **
Zx_SqfTree_append (

struct SqfTree *	tp,		/* IN/OUT: SqfTree to append node to */
const struct calib_Zx_obj *
			factor,		/* IN: Factor to make new node for */
int			multiplicity,	/* IN: Multiplicity of this factor */
struct SqfList **	hookp		/* IN: Where to append list nodes */
)
{
int		node_idx;

	CALIB_FATAL_ERROR_IF ((tp EQ NULL) OR
			      (factor EQ NULL) OR
			      (multiplicity <= 0) OR
			      (hookp EQ NULL));

	node_idx = Zx_SqfTree_new_leaf_node (tp, factor);

	hookp = Zx_SqfTree_new_list (tp, hookp, node_idx, multiplicity);

	return (hookp);
}

/*
 * Insert the given factor (with the given multiplicity) into the
 * given SqfTree.
 *
 * An initial scan is made for an existing node that identically matches
 * the given factor (whether the node is a leaf node or an interior node
 * that has already been split).  If an exact match exists, then we
 * can generate only one SqfList referencing that node.
 *
 * When this fails, we proceed to the computationally more intense phase
 * where we compute the GCD of the given factor with *every* current
 * leaf node in the tree.  Any of these that have non-trivial GCD's
 * cause either the existing leaf node and/or this factor to get split
 * into smaller pieces.
 */

	static
	struct SqfList **
Zx_SqfTree_insert (

struct SqfTree *	tp,		/* IN/OUT: SqfTree to insert into */
const struct calib_Zx_obj *
			factor,		/* IN: Factor to make new node for */
int			multiplicity,	/* IN: Multiplicity of this factor */
struct SqfList **	hookp		/* IN: Where to append list nodes */
)
{
int				i, j, k, nnodes, nleaves, deg;
int				idx1, idx2;
calib_bool			match;
struct SqfNode *		nodes;
struct SqfNode *		np;
int *				leaves;
int *				temp_leaves;
const struct calib_Zx_obj *	poly;
mpz_srcptr			coeff1, coeff2;
const struct calib_Zx_obj *	poly_ptrs [2];
struct calib_Zx_obj *		cf_ptrs [2];
struct calib_Zx_obj		FACTOR, GCD, CF1, CF2;

	CALIB_FATAL_ERROR_IF ((tp EQ NULL) OR
			      (factor EQ NULL) OR
			      (multiplicity <= 0) OR
			      (hookp EQ NULL));

	nnodes	= tp -> st_node_alloc;
	nodes	= tp -> st_nodes;

	deg	= factor -> degree;
	coeff1	= factor -> coeff;

	/* First scan all nodes looking for an exact match. */
	np = nodes;
	match = FALSE;
	for (i = 0; i < nnodes; i++, np++) {
		poly = np -> sn_factor;
		if (poly -> degree NE deg) continue;
		match = TRUE;
		coeff2 = poly -> coeff;
		for (j = 0; j <= deg; j++) {
			if (mpz_cmp (&coeff1 [j], &coeff2 [j]) NE 0) {
				match = FALSE;
				break;
			}
		}
		if (match) {
			/* Exact match already exists. */
			hookp = Zx_SqfTree_new_list (tp,
						     hookp,
						     i,
						     multiplicity);
			return (hookp);	
		}
	}

	Zx_init (&FACTOR);
	Zx_init (&GCD);
	Zx_init (&CF1);
	Zx_init (&CF2);

	/* Make a modifiable local copy of this factor. */
	Zx_set (&FACTOR, factor);

	/* For each leaf node, compute GCD (and cofactors) of remaining	*/
	/* FACTOR, and the leaf node's factor.  A non-trivial GCD will	*/
	/* split FACTOR, the leaf node's factor, or both.		*/

	nleaves		= tp -> st_nleaves;
	leaves		= tp -> st_leaves;
	temp_leaves	= tp -> st_temp_leaves;

	/* NOTE: leaves[0:nleaves-1] changes, both when new leaf nodes	*/
	/* are created, and (especially) when existing leaf nodes are	*/
	/* split.  Splitting causes the last leaf node to displace the	*/
	/* current leaves[] entry, meaning the following loop would	*/
	/* never test GCD with said factor.  We remedy this by making a	*/
	/* *copy* of the leaves[] array (in temp_leaves[]), and loop	*/
	/* over that array instead.  This assures that *every* leaf	*/
	/* node factor that exists *right now* gets GCD'd against the	*/
	/* factor we are now inserting.					*/
	memcpy (temp_leaves, leaves, nleaves * sizeof (temp_leaves [0]));

	for (i = 0; i < nleaves; i++) {
		j = temp_leaves [i];
		CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= nnodes));
		np = &nodes [j];

		/* Compute GCD (and cofactors) with this leaf. */
		poly_ptrs [0]	= &FACTOR;
		poly_ptrs [1]	= np -> sn_factor;	
		cf_ptrs [0]	= &CF1;
		cf_ptrs [1]	= &CF2;
		Zx_gcd_n (&GCD, cf_ptrs, 2, poly_ptrs);

		k = GCD.degree;
		if (k <= 0) {
			/* No common factor.  Skip to next leaf node. */
			CALIB_FATAL_ERROR_IF (k < 0);
			continue;
		}
		/* We have a non-trivial GCD.  Recall that all of the	*/
		/* factors entering this algorithm are square-free, so	*/
		/* when we get a non-trivial GCD, it will be relatively	*/
		/* prime to both cofactors.  Furthermore, all leaf	*/
		/* nodes are pairwise relatively prime.			*/
		if (CF2.degree > 0) {
			/* Leaf node factor has split. */
			idx1 = Zx_SqfTree_new_leaf_node (tp, &GCD);
			idx2 = Zx_SqfTree_new_leaf_node (tp, &CF2);
			Zx_SqfTree_split_node (tp, j, idx1, idx2);
		}
		else {
			/* Leaf node has not split.  GCD is node j. */
			idx1 = j;
		}
		hookp = Zx_SqfTree_new_list (tp, hookp, idx1, multiplicity);
		Zx_set (&FACTOR, &CF1);

		k = FACTOR.degree;
		if (k <= 0) {
			/* FACTOR has been completely distributed into	*/
			/* one or more nodes.				*/
			CALIB_FATAL_ERROR_IF (k < 0);
			break;
		}
	}

	if (FACTOR.degree > 0) {
		/* There is a residual FACTOR that is relatively prime	*/
		/* to all existing leaf nodes.  It becomes one more new	*/
		/* leaf node.						*/
		idx1 = Zx_SqfTree_new_leaf_node (tp, &FACTOR);
		hookp = Zx_SqfTree_new_list (tp, hookp, idx1, multiplicity);
	}


	Zx_clear (&CF2);
	Zx_clear (&CF1);
	Zx_clear (&GCD);
	Zx_clear (&FACTOR);

	return (hookp);
}

/*
 * Create a new "leaf" SqfNode for the given factor.
 * Return the index of this new node.
 */

	static
	int
Zx_SqfTree_new_leaf_node (

struct SqfTree *	tp,	/* IN: Tree to make new node in */
const struct calib_Zx_obj *
			factor	/* IN: Factor to put into new node */
)
{
int			j, size, nnodes, nleaves, node_idx;
struct SqfNode *	nodes;
int *			leaves;
struct SqfNode *	np;

	CALIB_FATAL_ERROR_IF ((tp EQ NULL) OR (factor EQ NULL));
	CALIB_FATAL_ERROR_IF (factor -> degree <= 0);

	size	= tp -> st_node_size;
	nnodes	= tp -> st_node_alloc;
	nleaves	= tp -> st_nleaves;
	leaves	= tp -> st_leaves;
	nodes	= tp -> st_nodes;

	CALIB_FATAL_ERROR_IF (nnodes >= size);

	node_idx = nnodes;

	/* Record this node_idx in the leaves[j]. */
	j = nleaves;
	CALIB_FATAL_ERROR_IF (j >= size);
	leaves [j] = node_idx;
	tp -> st_nleaves = j + 1;

	/* Allocate new SqfNode. */
	np = &nodes [node_idx];
	tp -> st_node_alloc = nnodes + 1;

	/* Fill in new node. */
	np -> sn_factor	= Zx_dup (factor);
	np -> sn_left	= -1;
	np -> sn_right	= -1;
	np -> sn_index	= j;

	return (node_idx);
}

/*
 * Split the given node, giving them the two specified child nodes.
 */

	static
	void
Zx_SqfTree_split_node (

struct SqfTree *	tp,	/* IN: SqfTree to manipulate */
int			parent,	/* IN: Index of parent node to split */
int			left,	/* IN: Index of left child */
int			right	/* IN: Index of right child */
)
{
int			i, k, size, nnodes, nleaves;
struct SqfNode *	nodes;
int *			leaves;
struct SqfNode *	p1;
struct SqfNode *	p2;
struct SqfNode *	p3;
struct SqfNode *	p4;

	CALIB_FATAL_ERROR_IF (tp EQ NULL);
	CALIB_FATAL_ERROR_IF (parent EQ left);
	CALIB_FATAL_ERROR_IF (parent EQ right);
	CALIB_FATAL_ERROR_IF (left EQ right);

	size		= tp -> st_node_size;
	nnodes		= tp -> st_node_alloc;
	nodes		= tp -> st_nodes;
	nleaves		= tp -> st_nleaves;
	leaves		= tp -> st_leaves;

	CALIB_FATAL_ERROR_IF ((parent < 0) OR (parent >= size));
	CALIB_FATAL_ERROR_IF ((left   < 0) OR (left   >= size));
	CALIB_FATAL_ERROR_IF ((right  < 0) OR (right  >= size));

	p1	= &nodes [parent];
	p2	= &nodes [left];
	p3	= &nodes [right];

	/* Verify that parent, left and right are all leaf nodes. */
	CALIB_FATAL_ERROR_IF ((p1 -> sn_left NE -1) OR (p1 -> sn_right NE -1));
	CALIB_FATAL_ERROR_IF ((p2 -> sn_left NE -1) OR (p2 -> sn_right NE -1));
	CALIB_FATAL_ERROR_IF ((p3 -> sn_left NE -1) OR (p3 -> sn_right NE -1));

	/* Remove parent node from leaves[] list. */
	i = p1 -> sn_index;
	CALIB_FATAL_ERROR_IF ((i < 0) OR (i >= nleaves));
	CALIB_FATAL_ERROR_IF (leaves [i] NE parent);
	--nleaves;

	k = leaves [nleaves];
	CALIB_FATAL_ERROR_IF ((k < 0) OR (k >= nnodes));

	leaves [i] = k;
	tp -> st_nleaves = nleaves;

	/* Update node k's position within leaves[]. */
	p4 = &nodes [k];
	CALIB_FATAL_ERROR_IF (p4 -> sn_index NE nleaves);
	p4 -> sn_index = i;

	/* Update parent node. */
	p1 -> sn_left	= left;
	p1 -> sn_right	= right;
	p1 -> sn_index	= -1;
}

/*
 * Create a new SqfList node, fill it in with the given node_idx and
 * multiplicity, and append it to a linked list using the given hookp.
 */

	static
	struct SqfList **
Zx_SqfTree_new_list (

struct SqfTree *	tp,		/* IN/OUT: Tree to use */
struct SqfList **	hookp,		/* IN: Where to append new entry */
int			node_idx,	/* IN: Node index */
int			multiplicity	/* IN: Multiplicity */
)
{
int			size, nlist;
struct SqfList *	lists;
struct SqfList *	lp;

	CALIB_FATAL_ERROR_IF ((tp EQ NULL) OR (hookp EQ NULL));

	size	= tp -> st_node_size;
	nlist	= tp -> st_list_alloc;
	lists	= tp -> st_lists;

	CALIB_FATAL_ERROR_IF (nlist >= size);

	/* Allocate new list node. */
	lp = &lists [nlist];
	tp -> st_list_alloc = nlist + 1;

	/* Fill in new list node. */
	lp -> sl_fact_idx	= node_idx;
	lp -> sl_mult		= multiplicity;
	lp -> sl_next		= NULL;

	/* Append it onto end of linked list. */
	*hookp = lp;
	hookp = &(lp -> sl_next);

	return (hookp);
}

/*
 * Compute the minimum multiplicity of each leaf node factor across
 * all of the polynomials (up to and including last_poly).
 *
 * Return TRUE if-and-only-if the minimum leaf node multiplicity is
 * zero for every leaf node.  This indicates that the GCD of polynomials
 * 0 through last_poly is 1.
 */

	static
	calib_bool
Zx_SqfTree_compute_min_mult (

struct SqfTree *	tp,		/* IN/OUT: SqfTree being used */
int			last_poly	/* IN: Last polynomial to test */
)
{
int			i, j, k, mult, nleaves;
calib_bool		gcd_is_one;
struct SqfList **	poly_list;
int *			leaves;
int **			poly_node_mult;
int *			min_mult;
int *			node_mult;
struct SqfList *	lp;

	nleaves		= tp -> st_nleaves;
	leaves		= tp -> st_leaves;
	poly_list	= tp -> st_poly_list;
	poly_node_mult	= tp -> st_poly_node_mult;
	min_mult	= tp -> st_min_mult;

	for (i = 0; i <= last_poly; i++) {
		/* Fill in all leaf-node multiplicities for	*/
		/* this polynomial.				*/
		node_mult = poly_node_mult [i];
		for (j = 0; j < nleaves; j++) {
			k = leaves [j];
			node_mult [k] = 0;
		}
		/* Store the multiplicity (within this i-th polynomial)	*/
		/* of each leaf node/factor into node_mult[].		*/
		for (lp = poly_list [i]; lp NE NULL; lp = lp -> sl_next) {
			Zx_SqfTree_traverse_node_leaves (
						node_mult,
						tp,
						lp -> sl_fact_idx,
						lp -> sl_mult);
		}
	}

	/* For each leaf node, compute its minimum multiplicity across	*/
	/* all of the polynomials.					*/
	for (j = 0; j < nleaves; j++) {
		k = leaves [j];
		min_mult [k] = INT_MAX;
	}
	for (i = 0; i <= last_poly; i++) {
		node_mult = poly_node_mult [i];
		for (j = 0; j < nleaves; j++) {
			k = leaves [j];
			mult = node_mult [k];
			if (mult < min_mult [k]) {
				min_mult [k] = mult;
			}
		}
	}

	/* See if the GCD is already one (among polynomials 0 throut	*/
	/* last_poly).							*/
	gcd_is_one = TRUE;
	for (j = 0; j < nleaves; j++) {
		k = leaves [j];
		mult = min_mult [k];
		if (mult > 0) {
			/* Leaf node (factor) k occurs in every		*/
			/* polynomial so far, so the GCD cannot (yet)	*/
			/* be 1.	    				*/
			gcd_is_one = FALSE;
			break;
		}
	}

	return (gcd_is_one);
}

/*
 * Recursively traverse all nodes starting at the given node_idx.
 * At each leaf node, store the given multiplicity into the given
 * node_mult array (at index of said leaf node).
 */

	static
	void
Zx_SqfTree_traverse_node_leaves (

int *			node_mult,	/* IN/OUT: Multiplicity of each node */
struct SqfTree *	tp,		/* IN: Tree to traverse */
int			node_idx,	/* IN: Index of node to traverse */
int			multiplicity	/* IN: Multiplicity to store */
)
{
int			n, left, right, lidx;
int *			ip;
struct SqfNode *	np;

	n	= tp -> st_node_alloc;

	/* Verify that given node_idx is valid. */
	CALIB_FATAL_ERROR_IF ((node_idx < 0) OR (node_idx >= n));
	CALIB_FATAL_ERROR_IF (multiplicity <= 0);

	np = &(tp -> st_nodes [node_idx]);

	left	= np -> sn_left;
	right	= np -> sn_right;
	lidx	= np -> sn_index;

	if (left < 0) {
		CALIB_FATAL_ERROR_IF (right >= 0);
		CALIB_FATAL_ERROR_IF ((lidx < 0) OR
				      (lidx >= tp -> st_nleaves));
		CALIB_FATAL_ERROR_IF (tp -> st_leaves [lidx] NE node_idx);

		/* This is a leaf node. */
		ip = &node_mult [node_idx];

		/* Double stores are *bad*! */
		CALIB_FATAL_ERROR_IF (*ip NE 0);

		*ip = multiplicity;
	}
	else {
		CALIB_FATAL_ERROR_IF (right < 0);
		CALIB_FATAL_ERROR_IF (lidx NE -1);

		/* This is an interior node with two children. */

		Zx_SqfTree_traverse_node_leaves (node_mult,
						 tp,
						 left,
						 multiplicity);

		Zx_SqfTree_traverse_node_leaves (node_mult,
						 tp,
						 right,
						 multiplicity);
	}
}

/*
 * Compute the product of the factors associated with leaf nodes
 * having the given multiplicities.  If min_mult is non-NULL, then
 * those multiplicities (corresponding to the GCD) are subtracted
 * out first before powering up the leaf node's factor.
 */

	static
	void
Zx_SqfTree_mul_factors (

struct calib_Zx_obj *	prod,		/* IN/OUT: Product of factors */
struct SqfTree *	tp,		/* IN: Tree with factors */
const int *		node_mult,	/* IN: Multiplicity of each leaf node */
const int *		min_mult	/* IN: Minimum multiplicities to */
					/*     subtract out (if non-NULL) */
)
{
int			i, j, k, m, nnodes, nleaves;
const int *		leaves;
const struct SqfNode *	nodes;
const struct SqfNode *	np;
struct calib_Zx_obj	PROD, POW;

	CALIB_FATAL_ERROR_IF ((prod	 EQ NULL) OR
			      (tp	 EQ NULL) OR
			      (node_mult EQ NULL));

	nnodes	= tp -> st_node_alloc;
	nleaves	= tp -> st_nleaves;
	nodes	= tp -> st_nodes;
	leaves	= tp -> st_leaves;

	Zx_init (&PROD);
	Zx_init (&POW);

	Zx_set_si (&PROD, 1);
	for (i = 0; i < nleaves; i++) {
		j = leaves [i];
		CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= nnodes));
		np = &nodes [j];

		m = node_mult [j];
		if (min_mult NE NULL) {
			/* Subtract off multiplicities that are part of GCD. */
			CALIB_FATAL_ERROR_IF (m < 0);
			k = min_mult [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			m -= k;
		}
		if (m <= 0) {
			CALIB_FATAL_ERROR_IF (m < 0);
			/* Ignore this leaf factor. */
		}
		else if (m EQ 1) {
			Zx_mul (&PROD, &PROD, np -> sn_factor);
		}
		else {
			Zx_ipow (&POW, np -> sn_factor, m);
			Zx_mul (&PROD, &PROD, &POW);
		}
	}

	Zx_set (prod, &PROD);

	Zx_clear (&POW);
	Zx_clear (&PROD);
}

/*
 * Mod out the coefficients of the given polynomial by the given modulus,
 * producing signed integers.  Signed integers are better in the context
 * of lifting, because negative numbers stop changing when they become
 * correct.  These same numbers, however, if kept in unsigned form, must
 * keep getting bigger and bigger as the modulus is increased.
 */

	static
	void
Zx_mod_out_signed (

struct calib_Zx_obj *	poly,		/* IN/OUT: polynomial to mod out */
mpz_srcptr		modulus		/* IN: modulus to use */
)
{
int		i, degree;
mpz_ptr		cp;
mpz_t		TMP;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (modulus EQ NULL));

	mpz_init (TMP);

	degree	= poly -> degree;
	cp	= poly -> coeff;
	for (i = 0; i <= degree; i++, cp++) {
		mpz_fdiv_r (cp, cp, modulus);
		mpz_mul_ui (TMP, cp, 2);
		if (mpz_cmp (TMP, modulus) >= 0) {
			mpz_sub (cp, cp, modulus);
		}
	}
	mpz_clear (TMP);
	/* Adjust degree of polynomial, if necessary. */
	cp = poly -> coeff;
	if ((degree >= 0) AND (mpz_sgn (&cp [degree]) EQ 0)) {
		for (i = degree - 1; i >= 0; i--) {
			if (mpz_sgn (&cp [i]) NE 0) break;
		}
		poly -> degree = i;
	}
}

/*
 * The Extended Euclidean Algorithm for polynomials in Z[x].
 * Given polynomials a, b \in Z[x],
 * compute polynomials g, xa, xb \in Z[x] such that
 *
 * 1.	xa * a + xb * b = g.
 * 2.	xa and xb are of minimum degree.
 */

	static
	void
Zx_extgcd (

struct calib_Zx_obj *		gcd,	/* IN/OUT: GCD result */
struct calib_Zx_obj *		xa,	/* IN/OUT: multiplier for a */
struct calib_Zx_obj *		xb,	/* IN/OUT: multiplier for b */
const struct calib_Zx_obj *	a,	/* IN: first operand */
const struct calib_Zx_obj *	b	/* IN: second operand */
)
{
int			i, adeg, bdeg, maxdeg;
struct calib_Zx_obj	*ap, *bp, *rp, *qp;
struct calib_Zx_obj	*m1, *m2, *m3, *n1, *n2, *n3;
struct calib_Zx_obj	*TEMP1, *TEMP2, *temp_ptr;
struct calib_Zx_obj	dval_poly;
mpz_t			DVAL, ZCONT;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR
			      (xa  EQ NULL) OR
			      (xb  EQ NULL) OR
			      (a   EQ NULL) OR
			      (b   EQ NULL));

	adeg = a -> degree;
	bdeg = b -> degree;
	maxdeg = _MAX (adeg, bdeg);

	mpz_init (DVAL);
	mpz_init (ZCONT);
	dval_poly.degree = 0;
	dval_poly.size	 = 1;
	dval_poly.coeff	 = DVAL;

	ALLOCATE_LOCAL_POLY_ARRAY (tmp, 12, maxdeg);

	temp_ptr = &tmp [0];
	ap = temp_ptr++;
	bp = temp_ptr++;
	rp = temp_ptr++;
	qp = temp_ptr++;
	m1 = temp_ptr++;
	m2 = temp_ptr++;
	m3 = temp_ptr++;
	n1 = temp_ptr++;
	n2 = temp_ptr++;
	n3 = temp_ptr++;
	TEMP1 = temp_ptr++;
	TEMP2 = temp_ptr++;

	Zx_set (ap, a);
	Zx_set (bp, b);

	/* ap = m1*a + n1*b */
	Zx_set_si (m1, 1);
	Zx_set_si (n1, 0);

	/* bp = m2*a + n2*b */
	Zx_set_si (m2, 0);
	Zx_set_si (n2, 1);

	while (bp -> degree >= 0) {
		Zx_div (qp, rp, DVAL, ap, bp);

		/* r = DVAL*ap - q*bp				*/
		/*   = DVAL*(m1*a + n1*b) - q*(m2*a + n2*b)	*/
		/*   = (DVAL*m1 - q*m2)*a + (DVAL*n1 - q*n2)*b	*/
		/*   = m3*a	     + m3*b,			*/
		/* so we compute:				*/
		/*	m3 = DVAL*m1 - q*m2			*/
		/*	n3 = DVAL*n1 - q*n2			*/
		Zx_mul (TEMP1, &dval_poly, m1);
		Zx_mul (TEMP2, qp,	   m2);
		Zx_sub (m3, TEMP1, TEMP2);
		Zx_mul (TEMP1, &dval_poly, n1);
		Zx_mul (TEMP2, qp,	   n2);
		Zx_sub (n3, TEMP1, TEMP2);

#if 1
		/* This can be *very* worthwhile, e.g., while	*/
		/* computing Q(a) inverses while factoring	*/
		/* quintic.dat, during which tt cancels		*/
		/* hundreds of thousands of bits per iteration.	*/

		Zx_content_poly3 (ZCONT, m3, n3, rp);
		if (mpz_cmp_ui (ZCONT, 1) NE 0) {
			Zx_div_z_exact (rp, rp, ZCONT);
			Zx_div_z_exact (m3, m3, ZCONT);
			Zx_div_z_exact (n3, n3, ZCONT);
#if 0
			printf ("Zx_extgcd: Cancelled %lu bits!\n",
				calib_mpz_bit_size (ZCONT));
			fflush (stdout);
#endif
		}
#endif

		temp_ptr = ap;
		ap = bp;
		bp = rp;
		rp = temp_ptr;

		temp_ptr = m1;
		m1 = m2;
		m2 = m3;
		m3 = temp_ptr;

		temp_ptr = n1;
		n1 = n2;
		n2 = n3;
		n3 = temp_ptr;
	}

	/* We have ap = M1*a + N1*b. */
	adeg = ap -> degree;
	if ((adeg >= 0) AND (mpz_sgn (&ap -> coeff [adeg]) < 0)) {
		/* Flip sign of lead-coefficient to make it positive. */
		for (i = 0; i <= adeg; i++) {
			mpz_neg (&(ap -> coeff [i]), &(ap -> coeff [i]));
		}
		for (i = 0; i <= m1 -> degree; i++) {
			mpz_neg (&(m1 -> coeff [i]), &(m1 -> coeff [i]));
		}
		for (i = 0; i <= n1 -> degree; i++) {
			mpz_neg (&(n1 -> coeff [i]), &(n1 -> coeff [i]));
		}
	}

	Zx_set (gcd, ap);
	Zx_set (xa, m1);
	Zx_set (xb, n1);

	FREE_LOCAL_POLY_ARRAY (tmp, 12);

	mpz_clear (ZCONT);
	mpz_clear (DVAL);
}

/*
 * Special routine to compute the mutual content of 3 Z[x] polynomials.
 * The caller ensures that the polynomials are given in the order of
 * likely increasing content.  (We want to stop early with a mutual
 * content of 1 with the least amount of effort.)
 */

	static
	void
Zx_content_poly3 (

mpz_ptr				zcont,	/* IN/OUT: Mutual content */
const struct calib_Zx_obj *	poly1,	/* IN/OUT: Polynomial 1 */
const struct calib_Zx_obj *	poly2,	/* IN/OUT: Polynomial 2 */
const struct calib_Zx_obj *	poly3	/* IN/OUT: Polynomial 3 */
)
{
mpz_t		ZCONT;

	CALIB_FATAL_ERROR_IF ((zcont EQ NULL) OR
			      (poly1 EQ NULL) OR
			      (poly2 EQ NULL) OR
			      (poly3 EQ NULL));

	mpz_init (ZCONT);

	do {		/* Used only for "break". */

		/* Polynomial most likely to have content of 1. */
		Zx_cumulative_content (ZCONT, poly1);
		if (mpz_cmp_ui (ZCONT, 1) EQ 0) break;

		/* Polynomial next-most-likely to have content of 1. */
		Zx_cumulative_content (ZCONT, poly2);
		if (mpz_cmp_ui (ZCONT, 1) EQ 0) break;

		/* Polynomial least likely to have content of 1. */
		Zx_cumulative_content (ZCONT, poly3);

	} while (FALSE);

	mpz_set (zcont, ZCONT);

	mpz_clear (ZCONT);
}

/*
 * Compute the cumulative content of the given content and that of the
 * given polynomial.  We scan the coefficients in the order most likely
 * to give us a GCD of 1 as soon as possible.  For many polynomials,
 * the coefficients of highest degree have the smallest magnitude.
 * However, the reciprocal polynomial (whose coefficients appear in the
 * reverse order) behave exactly the opposite.  We start at the end
 * having the coefficient of smallest magnitude.
 */

	static
	void
Zx_cumulative_content (

mpz_ptr				zcont,	/* IN/OUT: Comulative content */
const struct calib_Zx_obj *	poly	/* IN: Polynomial */
)
{
int		i, deg;
mpz_srcptr	coeff;
mpz_t		C1, C2, ZCONT;

	CALIB_FATAL_ERROR_IF ((zcont EQ NULL) OR (poly EQ NULL));

	deg	= poly -> degree;
	if (deg < 0) {
		return;
	}

	coeff	= poly -> coeff;

	mpz_init (C1);
	mpz_init (C2);
	mpz_init (ZCONT);

	mpz_abs (C1, &coeff [0]);
	mpz_abs (C2, &coeff [deg]);
	mpz_set (ZCONT, zcont);

	/* Scan in order likely to give us smallest GCDs soonest. */
	if (mpz_cmp (C1, C2) <= 0) {
		for (i = 0; i <= deg; i++) {
			mpz_gcd (ZCONT, ZCONT, &coeff [i]);
			if (mpz_cmp_ui (ZCONT, 1) EQ 0) break;
		}
	}
	else {
		for (i = deg; i >= 0; i--) {
			mpz_gcd (ZCONT, ZCONT, &coeff [i]);
			if (mpz_cmp_ui (ZCONT, 1) EQ 0) break;
		}
	}

	mpz_set (zcont, ZCONT);

	mpz_clear (ZCONT);
	mpz_clear (C2);
	mpz_clear (C1);
}

/*
 * Given f \in Z[x], compute both the "content" and "primitive part" of f.
 */

	static
	void
Zx_prim_part (

mpz_ptr			content,	/* IN/OUT: content of poly */
struct calib_Zx_obj *	ppart,		/* IN/OUT: primitive part */
const struct calib_Zx_obj *
			op		/* IN: polynomial to get ppart of */
)
{
int		i, deg, size, sign;
mpz_ptr		result, cp;
mpz_t		CONT;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg	= op -> degree;

	sign = +1;

	mpz_init (CONT);

	size = deg + 1;
	result = NULL;

	if (size <= 0) {
		/* Polynomial is zero.  Content is one. */
		mpz_set_ui (CONT, 1);
		goto all_done;
	}

	result = calib_new_Z_vector (size);
	calib_copy_Z_vector (result, op -> coeff, size);
	if (mpz_sgn (&result [deg]) < 0) {
		sign = -1;
		for (i = 0; i <= deg; i++) {
			mpz_neg (&result [i], &result [i]);
		}
	}

	for (i = 0, cp = result; i <= deg; i++, cp++) {
		mpz_gcd (CONT, CONT, cp);
		if (mpz_cmp_ui (CONT, 1) EQ 0) goto all_done;
	}

	/* We only get here if there is a non-trivial content to remove. */
	if (ppart NE NULL) {
		for (i = 0, cp = result; i <= deg; i++, cp++) {
			mpz_divexact (cp, cp, CONT);
		}
	}

all_done:
	if (content NE NULL) {
		mpz_mul_si (content, CONT, sign);
	}
	if (ppart NE NULL) {
		Zx_realloc_copy_poly (ppart, result, deg);
	}

	mpz_clear (CONT);
	calib_free_Z_vector (result, size);
}

/*
 * Given polynomials a,b \in Z[x], compute resultant(a,b), which
 * is an integer.
 *
 * This implementation uses resultants over Zp[x] (for various primes p),
 * combining them using the Chinese Remainder Theorem.
 */

	static
	void
Zx_resultant (

mpz_ptr				result,	/* IN/OUT: the resultant value */
const struct calib_Zx_obj *	a,	/* IN: first polynomial */
const struct calib_Zx_obj *	b	/* IN: second polynomial */
)
{
int			i, n, adeg, bdeg;
calib_int64u		p;
struct calib_Zp_dom *	Zp;
struct calib_Zpx_dom *	Zpx;
struct calib_Zpx_obj	zpoly1;
struct calib_Zpx_obj	zpoly2;
mpz_srcptr		lca, lcb;
mpz_t			MAX, MAG, LB, UB, RES, MODULUS, VALUE, G, S, T;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	adeg	= a -> degree;
	bdeg	= b -> degree;

	if ((adeg <= 0) AND (bdeg <= 0)) {
		/* Resultant (0,0), Resultant (1,0), Resultant (0,1),	*/
		/* Resultant (1,1), and Resultant (15, 18) are all 1.	*/
		mpz_set_ui (result, 1);
		return;
	}

	mpz_init (MAX);
	mpz_init (MAG);
	mpz_init (LB);
	mpz_init (UB);
	mpz_init (RES);
	mpz_init (MODULUS);
	mpz_init (VALUE);
	mpz_init (G);
	mpz_init (S);
	mpz_init (T);

	/* Compute an upper bound on the magnitude of the result.  The	*/
	/* Sylvester matrix is n x n (where n = adeg + bdeg).  Its	*/
	/* determinant has n! terms, each a product of n coefficients	*/
	/* of maximum magnitude MAX:					*/
	/*								*/
	/*			UB = MAX^n * n!				*/

	mpz_set_ui (MAX, 1);	/* Don't let this be zero! */
	for (i = 0; i <= adeg; i++) {
		mpz_abs (MAG, &(a -> coeff [i]));
		if (mpz_cmp (MAG, MAX) > 0) {
			mpz_set (MAX, MAG);
		}
	}
	for (i = 0; i <= bdeg; i++) {
		mpz_abs (MAG, &(b -> coeff [i]));
		if (mpz_cmp (MAG, MAX) > 0) {
			mpz_set (MAX, MAG);
		}
	}
	n = adeg + bdeg;
	mpz_pow_ui (MAG, MAX, n);
	mpz_fac_ui (UB, n);
	mpz_mul (UB, UB, MAG);

	/* We must have Modulus >= LB = 2*UB + 1 so that the signed	*/
	/* result can be properly recovered.				*/
	mpz_mul_ui (LB, UB, 2);
	mpz_add_ui (LB, LB, 1);

	Zp	= calib_make_Zp_dom_ui (3);
	Zpx	= calib_make_Zpx_dom (Zp);

	lca = &(a -> coeff [adeg]);
	lcb = &(b -> coeff [bdeg]);

	/* Start with largest 64-bit prime. */
	p = 18446744073709551557lu;
	for (;;) {
		if (mpz_divisible_ui_p (lca, p) OR
		    mpz_divisible_ui_p (lcb, p)) {
			/* This prime changes degrees!  Skip it. */
		}
		else {
			break;
		}
		p = Zx_previous_prime (p);
	}

	/* Use modulus p. */
	mpz_set_ui (Zp -> p, p);

	/* Compute first resultant modulo p. */
	Zpx -> init (Zpx, &zpoly1);
	Zpx -> init (Zpx, &zpoly2);
	Zpx -> set_Zx (&zpoly1, a);
	Zpx -> set_Zx (&zpoly2, b);
	Zpx -> resultant (RES, &zpoly1, &zpoly2);
	Zpx -> clear (&zpoly2);
	Zpx -> clear (&zpoly1);

	/* First modular solution. */
	mpz_set (VALUE, RES);
	mpz_set_ui (MODULUS, p);

	/* Use additional primes (if necessary), combining the results	*/
	/* using th Chinese Remainder Theorem, until the modulus is	*/
	/* sufficiently large.						*/
	while (mpz_cmp (MODULUS, LB) < 0) {
		/* Select next prime p. */
		for (;;) {
			p = Zx_previous_prime (p);
			if (mpz_divisible_ui_p (lca, p) OR
			    mpz_divisible_ui_p (lcb, p)) {
				/* This prime changes degrees!  Skip it. */
			}
			else {
				break;
			}
		}
		mpz_set_ui (Zp -> p, p);

		/* Compute resultant modulus this p. */
		Zpx -> init (Zpx, &zpoly1);
		Zpx -> init (Zpx, &zpoly2);
		Zpx -> set_Zx (&zpoly1, a);
		Zpx -> set_Zx (&zpoly2, b);
		Zpx -> resultant (RES, &zpoly1, &zpoly2);
		Zpx -> clear (&zpoly2);
		Zpx -> clear (&zpoly1);

		/* Apply Chinese Remainder Theorem. */
		mpz_gcdext (G, S, T, MODULUS, Zp -> p);

		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (G, 1) NE 0);

		/* VALUE = VALUE*T*p + RES*S*MODULUS		*/

		mpz_mul (T, T, VALUE);
		mpz_mul (T, T, Zp -> p);
		mpz_mul (S, S, RES);
		mpz_mul (S, S, MODULUS);
		mpz_add (VALUE, T, S);
		mpz_mul (MODULUS, MODULUS, Zp -> p);
		mpz_fdiv_r (VALUE, VALUE, MODULUS);
	}

	mpz_mul_ui (T, VALUE, 2);
	if (mpz_cmp (T, MODULUS) > 0) {
		/* Result is negative. */
		mpz_sub (result, VALUE, MODULUS);
	}
	else {
		/* Result is non-negative. */
		mpz_set (result, VALUE);
	}

	mpz_clear (T);
	mpz_clear (S);
	mpz_clear (G);
	mpz_clear (VALUE);
	mpz_clear (MODULUS);
	mpz_clear (RES);
	mpz_clear (UB);
	mpz_clear (LB);
	mpz_clear (MAG);
	mpz_clear (MAX);
}

/*
 * Given an odd integer, return the next-lowest prime number.
 */

	static
	calib_int64u
Zx_previous_prime (

calib_int64u	p		/* IN: Given integer p */
)
{
int		i, reps;
mpz_t		NUM;

	/* Must have an odd number. */
	CALIB_FATAL_ERROR_IF ((p & 1) EQ 0);

	mpz_init (NUM);

	reps = 30;

	for (;;) {
		CALIB_FATAL_ERROR_IF (p <= 3);

		p -= 2;

		mpz_set_ui (NUM, p);

		i = mpz_probab_prime_p (NUM, reps);
		if (i NE 0) break;
	}

	mpz_clear (NUM);

	return (p);
}

/*
 * Given polynomial a \in Z[x], compute the discriminant of a,
 * which is an element of Z.  The standard definition of the
 * discriminant in terms of the resultant is:
 *
 *	disc(f) = (-1)^k * res(f,f')/ lc(f)
 *
 * where k = deg(f)*(deg(f)-1)/2.
 *
 * In terms of the roots r_i, it is defined:
 *
 *	disc(f) = lc(f)^{2n-2} \prod_{i<j} (r_i - r_j)^2
 */

	static
	void
Zx_discriminant (

mpz_ptr				rop,	/* IN/OUT: the discriminant value */
const struct calib_Zx_obj *	op	/* IN: polynomial to get discriminant of */
)
{
int			n;
mpz_t			res, rem;
struct calib_Zx_obj	APRIME;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	n = op -> degree;

	if (n <= 0) {
		/* Discriminant of zero polynomial is not very well	*/
		/* defined.  (Is the degree 0 or -1?  All complex	*/
		/* numbers are roots, what is resultant(0,0), etc.)	*/
		CALIB_FATAL_ERROR_IF (n < 0);

		/* Discriminant of a non-zero constant.  This is also	*/
		/* not very well defined.				*/
		mpz_set_ui (rop, 1);
		return;
	}

	Zx_init_degree (&APRIME, n - 1);
	Zx_derivative (&APRIME, op);

	mpz_init (res);
	mpz_init (rem);

	Zx_resultant (res, op, &APRIME);
	if ((n & 0x02) NE 0) {
		/* We have n*(n-1)/2 odd.  Negate result. */
		mpz_neg (res, res);
	}
	/* Divide by lead coefficient.  Verify remainder is zero. */
	mpz_tdiv_qr (rop, rem, res, &(op -> coeff [n]));
	CALIB_FATAL_ERROR_IF (mpz_sgn (rem) NE 0);

	mpz_clear (rem);
	mpz_clear (res);
	Zx_clear (&APRIME);
}

/*
 * Compute the derivative of the given polynomial in Q[x].
 */

	static
	void
Zx_derivative (

struct calib_Zx_obj *		rop,		/* IN/OUT: derivative polynomial */
const struct calib_Zx_obj *	op		/* IN: polynomial to differentiate */
)
{
int			i, deg, rsize;
mpz_ptr			result;
mpz_srcptr		pcoeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	deg	= op -> degree;
	if (deg <= 0) {
		rop -> degree = -1;
		return;
	}

	pcoeff	= op -> coeff;
	rsize	= deg;
	result	= calib_new_Z_vector (rsize);

	for (i = 1; i <= deg; i++) {
		mpz_mul_ui (&result [i - 1], &pcoeff [i], i);
	}

	do {
		--deg;
	} while ((deg >= 0) AND (mpz_sgn (&result [deg]) EQ 0));

	Zx_realloc_copy_poly (rop, result, deg);
	calib_free_Z_vector (result, rsize);
}

/*
 * Create a polynomial in Z[x] that is the correpsonding representative
 * of the given polynomial in Zp[x].  The Z coefficients chosen
 * have smallest absolute value that are equivalent to the corresponding
 * Zp coefficient.
 */

	static
	struct calib_Zx_obj *
Zx_cvZpx (

const struct calib_Zpx_obj *	op	/* IN: poly in Zp[x] to convert */
)
{
int				i, deg, size;
const struct calib_Zpx_dom *	Zpx;
struct calib_Zx_obj *		new_poly;
mpz_ptr				coeff;
mpz_srcptr			p;
mpz_t				half_p;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	Zpx	= op -> dom;
	p	= Zpx -> cf -> p;
	deg	= op -> degree;

	coeff	= NULL;
	size	= deg + 1;
	if (size > 0) {
		coeff = calib_new_Z_vector (size);
		calib_copy_Z_vector (coeff, op -> coeff, size);

		/* Compute half_p = floor(p/2). */
		mpz_init (half_p);
		mpz_tdiv_q_2exp (half_p, p, 1);

		for (i = 0; i < size; i++) {
			mpz_mod (&coeff [i], &coeff [i], p);
			if (mpz_cmp (&coeff [i], half_p) > 0) {
				/* Make it negative. */
				mpz_sub (&coeff [i], &coeff [i], p);
			}
		}
		mpz_clear (half_p);
	}

	new_poly = CALIB_NEW (struct calib_Zx_obj);
	new_poly -> degree	= deg;
	new_poly -> size	= size;
	new_poly -> coeff	= coeff;

	return (new_poly);
}

/*
 * Factor the given polynomial in Z[x] completely into its irreducible
 * factors over Z[x].  The correct multiplicity of each
 * factor is also determined.
 */

	static
	struct calib_Zx_factor *
Zx_factor (

const struct calib_Zx_obj *	poly	/* IN: polynomial to factor */
)
{
struct calib_Zx_factor *	sqfactors;
struct calib_Zx_factor *	list;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	/* This handles constant polynomials, and stripping off any	*/
	/* integer content as a leading constant factor.		*/
	sqfactors = Zx_compute_square_free_factorization (poly);

	/* Now factor each square-free factor into irreducibles. */
	list = Zx_finish_sqf (sqfactors);

	Zx_free_factors (sqfactors);

	list = Zx_sort_factors (list);

	return (list);
}

/*
 * Factor the given polynomial into factors that are both square-free and
 * pairwise relatively prime, correctly determining the multiplicity of
 * each such factor.  This results in one non-trivial factor for each
 * distinct multiplicity present -- that factor being the product of all
 * factors having that that multiplicity.
 *
 * NOTE: This produces a constant leading factor (of multiplicity 1)
 * if the given polynomial is not primitive.
 */

	static
	struct calib_Zx_factor *
Zx_compute_square_free_factorization (

const struct calib_Zx_obj *	poly	/* IN: polynomial to factor */
)
{
int				deg, multiplicity;
struct calib_Zx_factor *	list;
struct calib_Zx_factor *	fp;
struct calib_Zx_factor **	hookp;
struct calib_Zx_obj *		pp;
const struct calib_Zx_obj *	polys [2];
struct calib_Zx_obj *		cofactors [2];
mpz_t				CONTENT;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	deg = poly -> degree;

	if (deg <= 0) {
		/* Zero or constant polynomial. */
		CALIB_FATAL_ERROR_IF (deg < -1);
		list = CALIB_NEW (struct calib_Zx_factor);
		list -> multiplicity	= 1;
		list -> factor		= Zx_dup (poly);
		list -> next		= NULL;

		return (list);
	}

	list = NULL;
	hookp = &list;

	ALLOCATE_LOCAL_POLY (f, deg);
	ALLOCATE_LOCAL_POLY (fprime, deg - 1);
	ALLOCATE_LOCAL_POLY (a, deg - 1);
	ALLOCATE_LOCAL_POLY (b, deg);
	ALLOCATE_LOCAL_POLY (c, deg);
	ALLOCATE_LOCAL_POLY (d, deg - 1);

	Zx_set (&f, poly);

	/* Remove any integer content. */
	mpz_init (CONTENT);
	Zx_prim_part (CONTENT, &f, &f);
	if (mpz_cmp_ui (CONTENT, 1) NE 0) {
		pp = CALIB_NEW (struct calib_Zx_obj);
		pp -> degree	= 0;
		pp -> size	= 1;
		pp -> coeff	= calib_new_Z_vector (1);
		mpz_set (pp -> coeff, CONTENT);

		fp = CALIB_NEW (struct calib_Zx_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= pp;
		fp -> next		= NULL;

		*hookp = fp;
		hookp = &(fp -> next);
	}
	mpz_clear (CONTENT);

	/* Yun's algorithm, initialize. */
	Zx_derivative (&fprime, &f);
	polys [0]	= &f;
	polys [1]	= &fprime;
	cofactors [0]	= &b;
	cofactors [1]	= &c;
	Zx_gcd_n (&a, &cofactors [0], 2, &polys [0]);
	Zx_derivative (&d, &b);
	Zx_sub (&d, &c, &d);

	for (multiplicity = 1;; multiplicity++) {
		/* Yun's algorithm, iteration. */
		polys [0]	= &b;
		polys [1]	= &d;
		cofactors [0]	= &b;
		cofactors [1]	= &c;
		Zx_gcd_n (&a, &cofactors [0], 2, &polys [0]);

		if (NOT ((a.degree EQ 0) AND
			 (mpz_cmp_ui (&a.coeff [0], 1) EQ 0))) {
			/* We have a new square-free factor. */
			pp = Zx_dup (&a);

			fp = CALIB_NEW (struct calib_Zx_factor);
			fp -> multiplicity	= multiplicity;
			fp -> factor		= pp;
			fp -> next		= NULL;
			*hookp = fp;
			hookp = &(fp -> next);
		}

		Zx_derivative (&d, &b);
		Zx_sub (&d, &c, &d);

		if (b.degree <= 0) break;
	}

	if ((b.degree EQ 0) AND (mpz_cmp_ui (&b.coeff [0], 1) NE 0)) {
		/* Remaining factor had better be -1. */
		CALIB_FATAL_ERROR_IF (mpz_cmp_si (&b.coeff [0], -1) NE 0);

		/* Add this factor to the *front* of the list. */
		pp = Zx_dup (&b);

		fp = CALIB_NEW (struct calib_Zx_factor);
		fp -> multiplicity = multiplicity;
		fp -> factor	   = pp;
		fp -> next	   = list;
		list = fp;
	}

	FREE_LOCAL_POLY (d);
	FREE_LOCAL_POLY (c);
	FREE_LOCAL_POLY (b);
	FREE_LOCAL_POLY (a);
	FREE_LOCAL_POLY (fprime);
	FREE_LOCAL_POLY (f);

	return (list);
}

/*
 * Given a list of square-free factors, "finish" the factorization by
 * factoring each into irreducible polynomials correctly labeled with
 * their respective multiplicities.
 */

	static
	struct calib_Zx_factor *
Zx_finish_sqf (

const struct calib_Zx_factor *	sqfactors	/* IN: Square-free factors */
)
{
int				multiplicity;
const struct calib_Zx_factor *	sqfp;
const struct calib_Zx_obj *	sqfpoly;
struct calib_Zx_factor *	list;
struct calib_Zx_factor **	hookp;
struct calib_Zx_factor *	irred_factors;
struct calib_Zx_factor *	nfp;

	list = NULL;
	hookp = &list;

	for (sqfp = sqfactors; sqfp NE NULL; sqfp = sqfp -> next) {
		multiplicity	= sqfp -> multiplicity;
		sqfpoly		= sqfp -> factor;

		if (sqfpoly -> degree <= 1) {
			/* Factor is constant or linear, and therefore	*/
			/* already irreducible.  Copy to output list.	*/
			nfp = CALIB_NEW (struct calib_Zx_factor);
			nfp -> multiplicity	= multiplicity;
			nfp -> factor		= Zx_dup (sqfpoly);
			nfp -> next		= NULL;

			*hookp = nfp;
			hookp = &(nfp -> next);
			continue;
		}

		irred_factors = Zx_factor_square_free (sqfpoly);

		/* Transfer the irreducible factors onto the final	*/
		/* list of factors, adding the correct multiplicity.	*/
		while (irred_factors NE NULL) {
			nfp = irred_factors;
			irred_factors = nfp -> next;
			nfp -> multiplicity	= multiplicity;
			nfp -> next		= NULL;

			*hookp = nfp;
			hookp = &(nfp -> next);
		}
	}

	return (list);
}

/*
 * Factor the given polynomial f \in Z[x} (which must be both primitive and
 * square-free) into its irreducible factors.  Algorithm:
 *
 * 1. Compute an upper-bound B on the magnitude of coefficients that
 *    appear in any factor of f.
 * 2. Choose a relatively small prime p, and compute g \in Zp[x] such that
 *    g \equiv f (mod p), and g remains square-free in Zp[x].
 * 3. Factorize g = g1*g2*...*gk in Zp[x], which implies that
 *    f \equiv g1*g2*...*gk (mod p).
 * 4. P-adic lifting: refine g1...gk so that f \equiv g1*g2*...*gk (mod p^j)
 *    where p^j > 2*B.
 * 5. Trial division: for each subset S of {g1,g2,...,gk} in increasing
 *    total degree, compute h = product{S} \in Z[x].  Check if h divides f.
 *    If so, then h is an irreducible factor of f.
 */

	static
	struct calib_Zx_factor *
Zx_factor_square_free (

const struct calib_Zx_obj *	poly	/* IN: primitive, square-free */
					/*     polynomial to factor */
)
{
struct calib_Zx_factor *	list;
struct calib_Zx_factor *	nfp;
struct calib_Zp_dom *		Zp;
struct calib_Zpx_dom *		Zpx;
struct FInfo *			fip;
struct FInfo			finfo;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	fip = &finfo;

	/* Initialize our global factoring data. */
	Zx_init_FInfo (fip, poly);

	/* Compute upper bound on the magnitude of coefficients in any	*/
	/* Z[x] factor of this polynomial.				*/
	Zx_factor_coefficient_upper_bound (fip);

	/* Create a modular coefficient field Zp, and the polynomial	*/
	/* ring Zp[x].  We use a dummy value of p right now, but this	*/
	/* is soon changed to the correct modulus.			*/
	Zp = calib_make_Zp_dom_ui (3);
	Zpx = calib_make_Zpx_dom (Zp);

	fip -> fi_Zp	= Zp;
	fip -> fi_Zpx	= Zpx;

	/* Now choose the prime modulus p so that f(x) factors		*/
	/* square-free (mod p).  Compute this mod-p factorization.	*/
	/* (This simultaneously chooses/sets the following fip-> data:	*/
	/*								*/
	/*	fi_Zp -> p		the prime modulus		*/
	/*	fi_Zpx_poly		f(x) (mod p)			*/
	/*	fi_monic_Zpx_poly					*/
	/*	fi_zpfactor_list	linked list of Zp[x] factors	*/
	Zx_choose_prime_modulus (fip);

	/* Make an array of the Zp[x] factors, omitting the degree 0	*/
	/* (constant) factor, if one exists.  We do *NOT* free up the	*/
	/* fi_zpfactor_list at this time, because it provides a		*/
	/* convenient way to free all of the Zp[x] factors when we are	*/
	/* done.							*/

	Zx_make_zpfactors_array (fip);

	if (fip -> fi_nleft <= 1) {
		CALIB_FATAL_ERROR_IF (fip -> fi_nleft <= 0);

		/* Polynomial is irreducible over Zp[x], so it is also	*/
		/* irreducible over Z[x]!  We are done.			*/
		nfp = CALIB_NEW (struct calib_Zx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= Zx_dup (poly);
		nfp -> next		= NULL;
		*(fip -> fi_hookp) = nfp;
		fip -> fi_hookp = &(nfp -> next);
		goto clean_up;
	}

	/* Convert the Zp[x] factors into Z[x] form.  The coefficients	*/
	/* become signed during this conversion because this makes the	*/
	/* correction coefficients potentially more sparse.  (If we use	*/
	/* coefficients in [0,p**i-1], then coefficients that are	*/
	/* ultimately negative (> p**i/2 for all i) will need to be	*/
	/* corrected for every i, whereas if we allow such coefficients	*/
	/* to be negative, their corrections eventually become zero.)	*/
	/* We store these in Z[x] form so that the coefficients can	*/
	/* grow arbitrarily during p-adic construction / lifting.	*/

	Zx_convert_Zpx_to_Zx (fip);

	/* Let r be the number of modular factors f_i(x), for 1<=i<=r.	*/
	/* Compute computational structure/plan for computing each of	*/
	/* the omit_i(x):						*/
	/*								*/
	/*	omit_i(x) = \prod_{1<=j<=r, j NE i} f_j(x)	(mod q)	*/
	/*								*/
	/* in O(r * log(r)) polynomial multiplies (mod q).		*/

	Zx_compute_omit_sub_product_structure (fip);

	/* Compute v_i(x) \in Zp[x] such that:				*/
	/*								*/
	/*	\sum_{i} v_i(x) * omit_i(x) \equiv 1 (mod p),		*/
	/*								*/
	/* where							*/
	/*								*/
	/*	omit_i(x) = \prod_{j NE i} f_j(x).			*/
	/*								*/
	/* These are used in the p-adic lifting step.			*/
	Zx_pfrac (fip);

	/* Modular factors are correct (mod p**1). */
	fip -> fi_a	= 1;
	mpz_set (fip -> fi_modulus, Zp -> p);

	/* Let CBOUND be the current upper bound on the magnitude of	*/
	/* coefficients in any remaining factor (fip->fi_cbound).	*/
	/* Compute power of p that we need to lift factors to, i.e.,	*/
	/* the smallest integer "a" such that p**a > 2*|lc|*CBOUND.	*/

	Zx_compute_lifting_target_power_of_p (fip);

	if (calib_Zx_settings.factor_trace_level >= 1) {
		printf ("We have %d modular factors\n", fip -> fi_nleft);
	}

	/* We don't use Van Hoeij unless the number of modular factors	*/
	/* exceeds the threshold at which VH uses exhaustive		*/
	/* enumeration instead of lattice reduction.			*/

	switch (calib_Zx_settings.factor_method) {
	case CALIB_ZX_FACTOR_METHOD_ZASSENHAUS:
		Zx_exhaustive_enumeration (fip);
		break;

	case CALIB_ZX_FACTOR_METHOD_VAN_HOEIJ:
#ifdef ENABLE_VAN_HOEIJ
		{ int num_fact = fip -> fi_num_fact;
		  if (num_fact <= calib_Zx_settings.VH_max_enumerate) {
			Zx_exhaustive_enumeration (fip);
		  }
		  else {
			Zx_find_factors_Van_Hoeij (fip);
		  }
		}
#else
		printf ("Warning: Van Hoeij Z[x] factoring method requested"
			" but not available.\n");

		Zx_exhaustive_enumeration (fip);
#endif	/* ENABLE_VAN_HOEIJ */
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

clean_up:

	list = fip -> fi_list;

	Zx_cleanup_FInfo (fip);

	return (list);
}

/*
 * Compute an upper bound on the magnitude of any single coefficient
 * in any factor of the given polynomial over Z[x].
 */

	static
	void
Zx_factor_coefficient_upper_bound (

struct FInfo *		fip	/* IN/OUT: Factoring algorithm info */
)
{
int				i, deg;
const struct calib_Zx_obj *	lpoly;
mpz_srcptr			coeff;
mpz_t				A, B, SUM, TEMP, CBOUND;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	lpoly	= &(fip -> fi_lpoly);

	deg	= lpoly -> degree;
	coeff	= lpoly -> coeff;

	mpz_init (A);
	mpz_init (B);
	mpz_init (SUM);
	mpz_init (TEMP);
	mpz_init (CBOUND);

	mpz_set_ui (A, 1);
	if (deg > 0) {
		mpz_mul_2exp (A, A, deg);
	}

	for (i = 0; i <= deg; i++) {
		mpz_mul (TEMP, &coeff [i], &coeff [i]);
		mpz_add (SUM, SUM, TEMP);
	}
	mpz_sqrt (B, SUM);

	mpz_mul (CBOUND, A, B);

	/* If fip->fi_cbound is zero, then this is the first time we	*/
	/* are computing this bound -- on the original polynomial.	*/
	/* Otherwise, we are recomputing (a hopefully smaller bound)	*/
	/* on the polynomial after one or more factors have been	*/
	/* discovered and removed.					*/

	if ((mpz_sgn (fip -> fi_cbound) > 0) AND
	    (mpz_cmp (fip -> fi_cbound, CBOUND) < 0)) {
		/* Coefficient upper bound actually *increased* after	*/
		/* removing one or more factors!			*/
		CALIB_FATAL_ERROR;
	}

	mpz_set (fip -> fi_cbound, CBOUND);

	mpz_clear (CBOUND);
	mpz_clear (TEMP);
	mpz_clear (SUM);
	mpz_clear (B);
	mpz_clear (A);
}

/*
 * Find a suitable prime p to use as a modulus for factorization of
 * f(x) \in Z[x].  Let p be an odd prime, and let g(x) \in Zp[x] such that
 * f(x) \equiv g(x) (mod p).  Then p is acceptable if g(x) is
 * square-free within Zp[x].
 *
 * Note that the simple test gcd(g,g') = 1 can fail in Zp[x] because this
 * construct removes one copy of each factor of g(x)
 * WHOSE MULTIPLICITY IS NOT A MULTIPLE OF p!  The more efficient
 * strategy would be to start with a "square-free factorization" in Zp[x]
 * and verify all of the multiplicities are 1 before proceeding to a
 * full factorization into irreducibles over Zp[x].  Our implementation
 * of Zp[x] does not yet expose all of this machinery, so we take the
 * slower (but safe) approach of factoring g(x) completely into
 * irreducibles over Zp[x] and then checking the multiplicities.
 *
 * Sets the following members of fip->:
 *
 *	fi_Zp -> p		the prime modulus
 *	fi_Zpx_poly		f(x) (mod p)
 *	fi_monic_Zpx_poly
 *	fi_zpfactor_list	linked list of Zp[x] factors
 */

	static
	void
Zx_choose_prime_modulus (

struct FInfo *		fip	/* IN/OUT: Factoring algorithm info */
)
{
int				i, degree, square_free;
int				nmf, best_nmf, attempts;
unsigned long			p, best_p;
const struct calib_Zx_obj *	poly;
struct calib_Zpx_obj *		irred_poly;
struct calib_Zp_dom *		Zp;
struct calib_Zpx_dom *		Zpx;
mpz_srcptr			coeff, lc;
mpz_ptr				gcoeff;
struct calib_Zpx_factor *	zpx_sqfactors;
struct calib_Zpx_factor *	zpx_factors;
struct calib_Zpx_factor *	fp;
struct calib_Zpx_factor *	best_factors;
struct calib_Zpx_obj *		best_zpoly;
struct calib_Zpx_obj		G;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	poly	= fip -> fi_poly;
	Zp	= fip -> fi_Zp;
	Zpx	= fip -> fi_Zpx;

	degree	= poly -> degree;
	coeff	= poly -> coeff;

	/* Make buffer to hold Zp[x] version of poly (for various p). */
	Zpx -> init_degree (Zpx, &G, degree);
	gcoeff = G.coeff;
	G.degree = degree;

	/* Make another Zp[x] polynomial to hold the "best" polynomial. */
	best_zpoly = CALIB_NEW (struct calib_Zpx_obj);
	Zpx -> init_degree (Zpx, best_zpoly, degree);

	/* Get lead coefficient of polynomial. */
	lc = &coeff [degree];

	best_nmf	= degree + 1;
	best_factors	= NULL;
	best_p		= 0;

	attempts = 0;

#if 1
	p = 17;
#else
	/* Start with a reasonably large prime. */
	p = 1009;
#endif

	/* Choose next prime that satisfies the simple checks. */
	for (; attempts < 3; p += 2) {
		if (NOT Zx_is_prime (p)) continue;

		if ((degree % p) EQ 0) {
			/* Don't use primes that divide the	*/
			/* degree. These will not be square-	*/
			/* free.				*/
			continue;
		}
		if (mpz_divisible_ui_p (lc, p)) {
			/* Don't use primes that evenly divide	*/
			/* the lead coefficient of our		*/
			/* polynomial, because this changes	*/
			/* (i.e., decreases) the degree of the	*/
			/* polynomial.				*/
			continue;
		}

		/* Set modulus for Zp and for Zp[x]. */
		mpz_set_ui (Zp -> p, p);

		/* Compute g(x) \equiv f(x) (mod p). */
		for (i = 0; i <= degree; i++) {
			Zp -> set_z (Zp, &gcoeff [i], &coeff [i]);
		}

		/* Do quick square-free factorization (mod p).  (We	*/
		/* don't want to waste time factoring all the way down	*/
		/* to irreducibles if the polynomial isn't		*/
		/* square-free.)					*/
		zpx_sqfactors = Zpx -> factor_square_free (&G);

		square_free = TRUE;
		nmf = 0;
		for (fp = zpx_sqfactors; fp NE NULL; fp = fp -> next) {
			if (fp -> factor -> degree <= 0) continue;
			if (fp -> multiplicity NE 1) {
				CALIB_FATAL_ERROR_IF (fp -> multiplicity < 1);
				square_free = FALSE;
				break;
			}
			++nmf;
		}
		if (NOT square_free) {
			/* Free up this factorization and try	*/
			/* another prime.			*/
			Zpx -> free_factors (zpx_sqfactors);
			continue;
		}

		/* Polynomial is square free.  Finish factorization all	*/
		/* the way down to irreducible factors.			*/

		zpx_factors = Zpx -> finish_sqf (zpx_sqfactors);

		Zpx -> free_factors (zpx_sqfactors);

		/* We have a "valid" modulus.  See if it is "good",	*/
		/* count irreducible factors.				*/
		nmf = 0;
		for (fp = zpx_factors; fp NE NULL; fp = fp -> next) {
			if (fp -> factor -> degree <= 0) continue;
			CALIB_FATAL_ERROR_IF (fp -> multiplicity NE 1);
			irred_poly = fp -> factor;
			if (irred_poly -> degree > 0) {
				++nmf;
			}
		}

		if (nmf < best_nmf) {
			/* We have a new, lower number of	*/
			/* factors.				*/
			if (best_factors NE NULL) {
				Zpx -> free_factors (best_factors);
			}
			best_nmf	= nmf;
			best_factors	= zpx_factors;
			best_p		= p;
			Zpx -> set (best_zpoly, &G);
		}
		else if (nmf EQ best_nmf) {
			/* Same number of factors, but a larger prime	*/
			/* modulus.  This is preferable, since it will	*/
			/* take fewer liftings to reach the same	*/
			/* coefficient bound.  Chalk this up as another	*/
			/* attempt.					*/
			++attempts;
			if (best_factors NE NULL) {
				Zpx -> free_factors (best_factors);
			}
			best_nmf	= nmf;
			best_factors	= zpx_factors;
			best_p		= p;
			Zpx -> set (best_zpoly, &G);
		}
		else {
			/* This is a failed attempt to improve. */
			Zpx -> free_factors (zpx_factors);
			++attempts;
		}
	}

	/* Store outputs from this routine. */

	mpz_set_ui (Zp -> p, best_p);

	fip -> fi_Zpx_poly	 = best_zpoly;

	fip -> fi_monic_Zpx_poly = Zpx -> dup (best_zpoly);
	Zpx -> monicize (fip -> fi_monic_Zpx_poly);

	fip -> fi_zpfactor_list	 = best_factors;

	/* Clean up. */

	Zpx -> clear (&G);
}

/*
 * Return true if the given number is prime.
 */

	static
	int
Zx_is_prime (

unsigned long	n		/* IN: number to test for primeness */
)
{
unsigned long	i;

	if (n <= 1) return (FALSE);
	if (n EQ 2) return (TRUE);
	if ((n & 0x01) EQ 0) return (FALSE);

	for (i = 3; i * i <= n; i += 2) {
		if ((n % i) EQ 0) return (FALSE);
	}
	return (TRUE);
}

/*
 * Create an array of the Zp[x] factors, omitting any constant (degree 0)
 * factors.  Sets the following members of fip->:
 *
 *	fi_num_fact		original number of Zp[x] factors
 *	fi_nleft		current number of Zp[x] factors
 *	fi_zpfactors		monic irreducible Zp[x] factors
 *	fi_taken
 *	fi_chosen
 *	fi_findex
 *	fi_comb
 *	fi_tc_stack
 */

	static
	void
Zx_make_zpfactors_array (

struct FInfo *		fip	/* IN/OUT: Factoring algorithm info */
)
{
int				i, n;
const struct calib_Zpx_factor *	zpfactor_list;
const struct calib_Zpx_factor *	fp;
struct calib_Zpx_obj *		poly;
struct calib_Zpx_obj **		array;
calib_bool *			taken;
calib_bool *			chosen;
int *				findex;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	zpfactor_list	= fip -> fi_zpfactor_list;

	n = 0;
	for (fp = zpfactor_list; fp NE NULL; fp = fp -> next) {
		poly = fp -> factor;
		if (poly -> degree <= 0) continue;
		++n;
	}

	array = CALIB_NEWA (n, struct calib_Zpx_obj *);
	for (i = 0, fp = zpfactor_list; fp NE NULL; fp = fp -> next) {
		poly = fp -> factor;
		if (poly -> degree <= 0) continue;
		array [i++] = poly;
	}
	CALIB_FATAL_ERROR_IF (i NE n);

	taken	= CALIB_NEWA (n, calib_bool);
	chosen	= CALIB_NEWA (n, calib_bool);
	findex	= CALIB_NEWA (n, int);
	for (i = 0; i < n; i++) {
		taken [i]  = FALSE;
		chosen [i] = FALSE;
		findex [i] = i;
	}

	fip -> fi_num_fact	= n;
	fip -> fi_nleft		= n;
	fip -> fi_zpfactors	= array;
	fip -> fi_taken		= taken;
	fip -> fi_chosen	= chosen;
	fip -> fi_findex	= findex;
	fip -> fi_comb		= CALIB_NEWA (n, int);
	fip -> fi_tc_stack	= calib_new_Z_vector (n + 1);
}

/*
 * Convert the given array of Zp[x] factors into Z[x] form.  The
 * coefficients become signed during this conversion because this makes
 * the correction coefficients potentially more sparse during lifing.
 * (If we use coefficients in [0,p**i-1], then coefficients that are
 * ultimately negative (> p**i/2 for all sufficiently large i) will
 * need to be corrected every iteration in order to maintain this fixed
 * negative value.  If we let these coefficients be negative, however,
 * then their correction coefficients eventually become zero and remain
 * zero for all higher powers of p.)
 *
 * We store these in Z[x] form so that the coefficients can grow
 * arbitrarily during p-adic construction / lifting.
 *
 *  Sets fip -> fi_zfactors.
 */

	static
	void
Zx_convert_Zpx_to_Zx (

struct FInfo *		fip	/* IN/OUT: Factoring algorithm info */
)
{
int				i, j, num_fact, deg;
const struct calib_Zpx_dom *	Zpx;
struct calib_Zpx_obj **		zpfactors;
struct calib_Zx_obj **		zfactors;
struct calib_Zpx_obj *		zp_factor;
struct calib_genrep *		gp1;
struct calib_Zx_obj		z_factor;
mpz_ptr				coeff, cj;
mpz_srcptr			p;
mpz_t				p_over_2;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	Zpx		= fip -> fi_Zpx;
	num_fact	= fip -> fi_num_fact;
	zpfactors	= fip -> fi_zpfactors;

	p = Zpx -> cf -> p;

	Zx_init (&z_factor);
	mpz_init (p_over_2);
	mpz_tdiv_q_2exp (p_over_2, p, 1);

	zfactors = CALIB_NEWA (num_fact, struct calib_Zx_obj *);

	for (i = 0; i < num_fact; i++) {
		/* This is crude, but effective.  To genrep and back! */
		zp_factor = zpfactors [i];
		gp1 = Zpx -> to_genrep (zp_factor, "x");
		Zx_set_genrep (&z_factor, gp1, "x");
		calib_genrep_free (gp1);
		/* Make coefficients > p/2 be negative. */
		deg	= z_factor.degree;
		coeff	= z_factor.coeff;
		for (j = 0, cj = coeff; j <= deg; j++, cj++) {
			if (mpz_cmp (cj, p_over_2) > 0) {
				mpz_sub (cj, cj, p);
				CALIB_FATAL_ERROR_IF (mpz_sgn (cj) >= 0);
			}
		}
		zfactors [i] = Zx_dup (&z_factor);
	}

	mpz_clear (p_over_2);
	Zx_clear (&z_factor);

	fip -> fi_zfactors = zfactors;
}

/*
 * Let r be the number of modular factors f_i(x), for 1<=i<=r.
 * Compute computational structure/plan for computing each of
 * the omit_i(x):
 *
 *	omit_i(x) = \prod_{1<=j<=r, j NE i} f_j(x)	(mod q)
 *
 * in O(r * log(r)) polynomial multiplies (mod q).
 *
 * For r=15, the sub-products have the following structure:
 *
 *	0:	1--------------	(Individual factors)
 *	1:	-1-------------
 *	2:	--1------------
 *	3:	---1-----------
 *	4:	----1----------
 *	...	...
 *	12:	------------1--
 *	13:	-------------1-
 *	14:	--------------1
 *	15:	11-------------	(Products of adjacent factor pairs)
 *	16:	--11-----------
 *	17:	----11---------
 *	18:	------11-------
 *	19:	--------11-----
 *	20:	----------11---
 *	21:	------------11-
 *	22:	1111-----------	(Products of 4 adjacent factors)
 *	23:	----1111-------
 *	24:	--------1111---
 *	25:	------------111	(Truncated 4-product, but > 2 factors)
 *	26:	11111111-------	(Products of 8 adjacent factors)
 *	27:	--------1111111	(Truncated 8-product, but > 4 factors)
 *
 * The computational structure/plan for computing, e.g., omit_5(x)
 * is then:
 *
 *	27, 22, 18, 4, -1
 *
 * In other words, omit_5(x) is the product of sub-products 27, 22, 18 and 4.
 * We similarly decompose the computation of each of each sub-product j
 * (for r <= j < num_osp) as the product of two previous sub-products.
 *
 * This routine computes the following data:
 *
 *	fip -> fi_num_osp	Number of sub-products (14 in this example)
 *	fip -> fi_osp_struct	Structure for computing sub-products
 *	fip -> fi_omit_struct	Sequence of r (-1 terminated) "plans", one
 *				for each omit_i(x)
 *
 * We do this using r-bit masks (for clarity and correctness).  There may
 * be more efficient bit-twiddling that can get the right structure directly
 * from the i and j without resorting to r-bit masks, but we only do this
 * once -- it is the resulting "plans" of total size O(r * log(r)) that are
 * used computationally.
 */

	static
	void
Zx_compute_omit_sub_product_structure (

struct FInfo *		fip	/* IN/OUT: Factoring algorithm info */
)
{
int			i, i1, i2, j, k, m, n, r;
int			nrows, num_osp, nmasks, total, resid_size;
int			num_resid, pow, r2;
struct sprod_row *	rows;
struct sprod_row *	rowp;
struct sprod_resid *	resid;
struct sprod_resid *	rp;
struct sprod_resid *	rp2;
struct sprod_resid *	rp_end;
calib_bitmask_t		word;
calib_bitmask_t *	bp;
calib_bitmask_t *	bp1;
calib_bitmask_t *	bp2;
calib_bitmask_t *	all_ones;
calib_bitmask_t *	cur_mask;
calib_bitmask_t *	sub_prod_masks;
int **			plan_ptrs;
int *			buffer;
int *			ip;
int *			ip2;
int *			osp_struct;
int *			ospsp;
int *			omit_struct;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	r	= fip -> fi_num_fact;
	nmasks	= BMASK_ELTS (r);

	/* Must have 2 or more modular factors. */
	CALIB_FATAL_ERROR_IF (r <= 1);

	/* Compute number of rows we will need. */
	nrows	= Zx_sub_product_row (r);
	if (NOT IS_POWER_OF_2 (r)) {
		/* Need one more row to represent 2 sub-products:	*/
		/* a max-sized power-of-2 block, and the residual.	*/
		++nrows;
	}
	rows	= CALIB_NEWA (nrows, struct sprod_row);
	memset (rows, 0, nrows * sizeof (rows [0]));

	/* Allocate our resid array conservatively. */
	resid_size = 2 * nrows;
	resid = CALIB_NEWA (resid_size, struct sprod_resid);
	memset (resid, 0, resid_size * sizeof (resid [0]));

	/* Compute fspsize, nfsp and nresid for each row. */
	i = 1;
	for (k = 0; k < nrows; k++) {
		j = r / i;
		rows [k].fspsize = i;
		rows [k].nfsp	 = j;
		rows [k].nresid	 = r - i*j;
		i <<= 1;
	}

	/* Compute all of the residual values we will need, placing	*/
	/* them in the order in which they should be computed.		*/
	i = 0;
	for (k = 0; k < nrows; k++) {
		i = Zx_resid_recurse (i, resid, rows [k].nresid, k);
	}
	CALIB_FATAL_ERROR_IF (i > resid_size);
	num_resid = i;

	/* Now fill in start, end, and compute additional residual	*/
	/* sub-products each row needs.					*/
	i = 0;
	rp	= &resid [0];
	rp_end	= &resid [num_resid];
	for (k = 0; k < nrows; k++) {
		rows [k].start		= i;
		i += rows [k].nfsp;

		/* Process any additional residual sub-products this	*/
		/* row might need.					*/
		for (; (rp < rp_end) AND (rp -> row EQ k); rp++) {
			n	= rp -> nresid;
			pow	= rp -> pow;
			r2	= rp -> r2;

			/* This power of 2 should be in an earlier row,	*/
			/* which has now been completed.		*/
			CALIB_FATAL_ERROR_IF (pow >= k);

			/* Get index of this last "full" sub-product	*/
			/* in that row.					*/
			rowp = &rows [pow];
			i2 = rowp -> start + rowp -> nfsp - 1;

			if (r2 EQ 0) {
				/* Residual is an exact power of 2.  No	*/
				/* new sub-product needed.  Just refer	*/
				/* to existing sub-product.		*/
				rp -> residx	= i2;
				rp -> idx1	= -1;
				rp -> idx2	= -1;
			}
			else {
				/* Find index of residual r2. */
				for (rp2 = rp;; rp2--) {
					CALIB_FATAL_ERROR_IF (rp2 < resid);
					if (rp2 -> nresid EQ r2) {
						i1 = rp2 -> residx;
						break;
					}
				}
				/* Add an extra sub-product for this row. */
				rp -> residx	= i;
				rp -> idx1	= i1;
				rp -> idx2	= i2;
				++i;
			}
			++m;
		}

		rows [k].end = i;
	}
	num_osp = i;

	/* Allocate and zero all of the bit masks we will need:		*/
	/*								*/
	/*	- An "all ones" masks (first r bits set, rest cleared)	*/
	/*	- A working mask for the "current" omit_i(x)		*/
	/*	- One mask for each sub-product (num_osp)		*/

	n = (2 + num_osp) * nmasks;
	bp	= CALIB_NEWA (n, calib_bitmask_t);
	memset (bp, 0, n * sizeof (bp [0]));
	all_ones	= bp;			bp += nmasks;
	cur_mask	= bp;			bp += nmasks;
	sub_prod_masks	= bp;

	/* Initialize the all_ones mask. */
	n	= nmasks - 1;
	word	= ~((calib_bitmask_t) 0);
	for (i = 0; i < n; i++) {
		all_ones [i] = word;
	}
	/* Compute number of '1' bits in final word. */
	j = r - n * BPW;
	if (j < BPW) {
		word	= 1;
		word	<<= j;
		word	-= 1;
	}
	else {
		/* Shifting left by j >= word size is undefined. */
		/* Just keep our all-ones word. */
	}
	all_ones [i] = word;

	/* Initialize the first r sub_prod_masks. */
	bp = sub_prod_masks;
	for (i = 0; i < r; i++, bp += nmasks) {
		SETBIT (bp, i);
	}

	/* Create buffer to hold computational structure / plan for	*/
	/* computing sub-products.					*/
	osp_struct = CALIB_NEWA (2 * (num_osp - r), int);
	ospsp = osp_struct;

	/* Compute sub_prod_masks for each successive row. */
	rp	= resid;
	rp_end	= &resid [num_resid];
	for (k = 1; k < nrows; k++) {
		i = rows [k - 1].start;
		j = rows [k].start;
		bp1 = &sub_prod_masks [i * nmasks];
		bp  = &sub_prod_masks [j * nmasks];
		bp2 = bp1 + nmasks;
		m = rows [k].nfsp;
		for (n = 0; n < m; n++) {
			/* Sub-product is two consecutive sub-products	*/
			/* from previous row k-1.			*/
			*ospsp++	= i++;
			*ospsp++	= i++;
			for (j = 0; j < nmasks; j++) {
				CALIB_FATAL_ERROR_IF ((*bp1 & *bp2) NE 0);
				*bp++ = (*bp1++ | *bp2++);
			}
			bp1 = bp2;
			bp2 = bp1 + nmasks;
		}

		j = rows [k].start + m;
		for (; (rp < rp_end) AND (rp -> row EQ k); rp++) {
			i1	= rp -> idx1;
			i2	= rp -> idx2;
			if (i1 < 0) {
				CALIB_FATAL_ERROR_IF ((i1 NE -1) OR
						      (i2 NE -1));
				/* Exact power of 2.  No new	*/
				/* sub-product needed.		*/
				continue;
			}
			CALIB_FATAL_ERROR_IF ((i1 <= 0) OR (i2 <= 0));
			CALIB_FATAL_ERROR_IF ((i1 >= j) OR (i2 >= j));
			*ospsp++	= i1;
			*ospsp++	= i2;
			bp1 = &sub_prod_masks [i1 * nmasks];
			bp2 = &sub_prod_masks [i2 * nmasks];
			for (n = 0; n < nmasks; n++) {
				CALIB_FATAL_ERROR_IF ((*bp1 & *bp2) NE 0);
				*bp++ = (*bp1++ | *bp2++);
			}
			rows [k].residx = j;
			++j;
		}
		CALIB_FATAL_ERROR_IF (j NE rows [k].end);
	}

	CALIB_FATAL_ERROR_IF (bp NE &sub_prod_masks [num_osp * nmasks]);

#if 1
	/* Validate osp_struct while masks are still present. */
	ospsp = osp_struct;
	for (m = r; m < num_osp; m++) {
		i = *ospsp++;
		j = *ospsp++;
		CALIB_FATAL_ERROR_IF ((i < 0) OR (i >= m));
		CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= m));
		CALIB_FATAL_ERROR_IF (i EQ j);
		bp1 = &sub_prod_masks [i * nmasks];
		bp2 = &sub_prod_masks [j * nmasks];
		bp = &cur_mask [0];
		for (n = 0; n < nmasks; n++) {
			CALIB_FATAL_ERROR_IF ((*bp1 & *bp2) NE 0);
			*bp++ = (*bp1++ | *bp2++);
		}
		n = memcmp (&sub_prod_masks [m * nmasks],
			    cur_mask,
			    nmasks * sizeof (cur_mask [0]));
		CALIB_FATAL_ERROR_IF (n NE 0);
	}
#endif

	/* Computation of sub-product masks is now complete.		*/
	/* Now compute the structure/plan for each omit_i(x).		*/
	plan_ptrs	= CALIB_NEWA (r, int *);
	buffer		= CALIB_NEWA (r, int);
	total = 0;
	for (i = 0; i < r; i++) {
		memcpy (cur_mask, all_ones, nmasks * sizeof (cur_mask [0]));
		CLRBIT (cur_mask, i);

		k = 0;
		for (j = num_osp - 1; j >= 0; j--) {
			bp = &sub_prod_masks [j * nmasks];
			bp1 = bp;
			bp2 = cur_mask;
			for (m = 0; m < nmasks; m++) {
				if ((*bp1 & *bp2) NE *bp1) break;
				++bp1;
				++bp2;
			}
			if (m < nmasks) {
				/* bp is *not* a subset of cur_mask. */
				continue;
			}
			/* bp is a subset of cur_mask.  Include sub-	*/
			/* product j in product for omit_i(x).		*/
			buffer [k++] = j;
			bp1 = bp;
			bp2 = cur_mask;
			/* Remove these factors from cur_mask. */
			for (m = 0; m < nmasks; m++) {
				*bp2++ ^= *bp1++;
			}
		}
		buffer [k++] = -1;
		CALIB_FATAL_ERROR_IF (k > r);

		/* cur_mask should now be entirely zero. */
		for (m = 0; m < nmasks; m++) {
			CALIB_FATAL_ERROR_IF (cur_mask [m] NE 0);
		}

		ip = CALIB_NEWA (k, int);
		memcpy (ip, buffer, k * sizeof (ip [0]));
		plan_ptrs [i] = ip;

		total += k;
	}

	/* Combine all plans into a single array. */
	omit_struct = CALIB_NEWA (total, int);
	ip = omit_struct;
	for (i = 0; i < r; i++) {
		ip2 = plan_ptrs [i];
		do {
			j = *ip2++;
			*ip++ = j;
		} while (j >= 0);
		free (plan_ptrs [i]);
		plan_ptrs [i] = NULL;
	}
	CALIB_FATAL_ERROR_IF (ip NE &omit_struct [total]);

	free (buffer);
	free (plan_ptrs);
	free (all_ones);
	free (resid);
	free (rows);

	fip -> fi_num_osp	= num_osp;
	fip -> fi_osp_struct	= osp_struct;
	fip -> fi_omit_struct	= omit_struct;
}

/*
 * Recursively compute the structure of the given residual, placing them
 * in the order in which they should be computed.
 */

	static
	int
Zx_resid_recurse (

int			i,	/* IN: Index at which to place next residual */
struct sprod_resid *	array,	/* IN/OUT: Array of residual descriptions */
int			j,	/* IN: Residual value to represent */
int			k	/* IN: Row requesting this residual */
)
{
int		m, pow, r2;

	CALIB_FATAL_ERROR_IF ((i < 0) OR
			      (array EQ NULL) OR
			      (j < 0) OR
			      (k < 0));

	if (j EQ 0) {
		/* No residual.  Nothing needed. */
		return (i);
	}

	/* Residual is positive. */

	if (i > 0) {
		/* Not our first residual.  See if it matches previous. */
		m = array [i - 1].nresid;
		if (j <= m) {
			CALIB_FATAL_ERROR_IF (j < m);
			/* This residual matches last one.  Nothing needed. */
			return (i);
		}
	}

	/* We have a larger residual. */
	if (IS_POWER_OF_2 (j)) {
		/* This residual is available as the last "full"	*/
		/* sub-product of the appropriate row.  Determine that	*/
		/* power / row.  Setting r2 = 0 indicates this is an	*/
		/* exact power of 2.					*/
		pow	= Zx_sub_product_row (j);
		r2	= 0;
	}
	else {
		/* Determine the largest power of 2 that we can split	*/
		/* off the left end of this residual.			*/
		pow	= Zx_sub_product_row (j) - 1;
		r2	= j - (1 << pow);
		CALIB_FATAL_ERROR_IF (r2 <= 0);

		/* Recursively process the r2 residual. */
		i = Zx_resid_recurse (i, array, r2, k);

		/* Now store this larger residual j. */
	}

	array [i].nresid	= j;
	array [i].row		= k;
	array [i].pow		= pow;
	array [i].r2		= r2;
	/* Determined later. */
	array [i].residx	= -1;
	array [i].idx1		= -1;
	array [i].idx2		= -1;
	++i;
	return (i);
}

/*
 * For all n >= 1, compute the following:
 *
 *	f(n) = k such that 2**(k-1) < n <= 2**k.
 *
 * This has various uses.  For example, for r modular factors, the largest
 * row index we need is m = f(r) - 1.  [m + 1 rows indexed 0 through m.]
 *  For given nresid, k = f(nresid) - 1 is the last row one need search for
 * the existing nresid sub-product, or sub-products needed to form the new
 * nresid sub-product.
 */

	static
	int
Zx_sub_product_row (

int		n		/* IN: Given value n >= 2 */
)
{
int		k, bit;

	CALIB_FATAL_ERROR_IF (n < 1);

	k = 0;
	bit = 1;
	while (bit < n) {
		bit <<= 1;
		++k;
	}
	return (k);
}

/*
 * Let r be the number of modular factors, r >= 2.
 * Compute and return an array of polynomials omit_i(x):
 *
 *	omit_i(x) = \prod_{1<=j<=r, j NE i} f_j(x)	(mod q)
 *
 * where f_j(x) is the j-th modular factor.  We do this in O(r * log(r))
 * polynomial multiplies due to the computational plans established by
 * Zx_compute_omit_sub_product_structure().
 */

	static
	struct calib_Zx_obj *
Zx_compute_omit (

struct FInfo *		fip,	/* IN/OUT: Factoring algorithm info */
mpz_srcptr		q	/* IN: Modulus to use */
)
{
int			i, j, k, m, r, num_osp;
struct calib_Zx_obj **	zfactors;
struct calib_Zx_obj *	sub_prod;
struct calib_Zx_obj *	omit;
const int *		osp_struct;
const int *		omit_struct;

	CALIB_FATAL_ERROR_IF ((fip EQ NULL) OR (q EQ NULL));

	r		= fip -> fi_num_fact;
	zfactors	= fip -> fi_zfactors;
	num_osp		= fip -> fi_num_osp;
	osp_struct	= fip -> fi_osp_struct;
	omit_struct	= fip -> fi_omit_struct;

	/* Compute sub-products. */
	sub_prod = CALIB_NEWA (num_osp, struct calib_Zx_obj);
	for (i = 0; i < num_osp; i++) {
		Zx_init (&sub_prod [i]);
	}
	/* First row: single factors. */
	k = 0;
	for (i = 0; i < r; i++) {
		Zx_set (&sub_prod [k], zfactors [i]);
		Zx_mod_out_signed (&sub_prod [k], q);
		++k;
	}

	/* Compute remaining sub-products. */
	for (m = r; m < num_osp; m++) {
		i = *osp_struct++;
		j = *osp_struct++;
		CALIB_FATAL_ERROR_IF ((i < 0) OR (i >= k));
		CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= k));
		CALIB_FATAL_ERROR_IF (i EQ j);
		Zx_mul (&sub_prod [k], &sub_prod [i], &sub_prod [j]);
		Zx_mod_out_signed (&sub_prod [k], q);
		++k;
	}
	CALIB_FATAL_ERROR_IF (k NE num_osp);

	omit = CALIB_NEWA (r, struct calib_Zx_obj);
	for (i = 0; i < r; i++) {
		Zx_init (&omit [i]);
		/* Just copy first sub-product. */
		j = *omit_struct++;
		CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= num_osp));
		Zx_set (&omit [i], &sub_prod [j]);
		/* Multiply remaining sub-products into omit[i]. */
		for (;;) {
			j = *omit_struct++;
			if (j < 0) {
				CALIB_FATAL_ERROR_IF (j NE -1);
				break;
			}
			CALIB_FATAL_ERROR_IF (j >= num_osp);
			Zx_mul (&omit [i], &omit [i], &sub_prod [j]);
			Zx_mod_out_signed (&omit [i], q);
		}
	}

	Zx_clear_array (sub_prod, num_osp);
	free (sub_prod);

	return (omit);
}

/*
 * We are given the factorization g(x) = \prod_i f_i(x), over Zp[x],
 * where g(x) and f_i(x) are all monic.  Compute v_i(x) \in Zp[x] such that
 *
 *	\sum_i v_i(x) * omit_i(x) \equiv 1	(mod p)
 *
 * where
 *
 *	omit_i(x) = \prod_{j NE i} f_j(x).
 *
 * Sets fip -> fi_vlist.
 */

	static
	void
Zx_pfrac (

struct FInfo *		fip	/* IN/OUT: Factoring algorithm info */
)
{
int				i, num_fact;
const struct calib_Zpx_dom *	Zpx;
struct calib_Zpx_obj **		farray;
struct calib_Zpx_obj **		varray;
const struct calib_Zpx_obj *	fi;
struct calib_Zx_obj *		omit;
mpz_srcptr			p;
struct calib_Zpx_obj		G, A, B, R, OMIT;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	Zpx		= fip -> fi_Zpx;
	num_fact	= fip -> fi_num_fact;
	farray		= fip -> fi_zpfactors;

	p	= Zpx -> cf -> p;

	omit = Zx_compute_omit (fip, p);

	varray = CALIB_NEWA (num_fact, struct calib_Zpx_obj *);

	Zpx -> init (Zpx, &G);
	Zpx -> init (Zpx, &A);
	Zpx -> init (Zpx, &B);
	Zpx -> init (Zpx, &R);
	Zpx -> init (Zpx, &OMIT);

	for (i = 0; i < num_fact; i++) {
		/* Get omit[i] in Zp[x] form. */
		Zpx -> set_Zx (&OMIT, &omit [i]);

		fi = farray [i];
		Zpx -> extgcd (&G, &A, &B, &OMIT, fi);
		CALIB_FATAL_ERROR_IF ((G.degree NE 0) OR
				      (mpz_cmp_ui (&G.coeff [0], 1) NE 0));
		CALIB_FATAL_ERROR_IF (A.degree >= fi -> degree);

		varray [i] = Zpx -> dup (&A);
	}

	Zx_clear_array (omit, num_fact);
	free (omit);

	Zpx -> clear (&OMIT);
	Zpx -> clear (&R);
	Zpx -> clear (&B);
	Zpx -> clear (&A);
	Zpx -> clear (&G);

	fip -> fi_vlist	= varray;
}

/*
 * Let CBOUND be the current upper bound on the magnitude of coefficients
 * in any remaining factor (fip->fi_cboune).  Compute power of p that we
 * need to lift modular factors to, i.e., the smallest integer "a" such
 * that:
 *
 *		p**a > 2 * |lc| * CBOUND
 *
 * Sets fip -> fi_lm_a.
 */

	static
	void
Zx_compute_lifting_target_power_of_p (

struct FInfo *		fip	/* IN/OUT: Factoring algorithm info */
)
{
int				a, a_lm, degree;
const struct calib_Zx_obj *	lpoly;
const struct calib_Zp_dom *	Zp;
mpz_srcptr			p, lc, cbound;
mpz_t				ppow, min_modulus;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	lpoly	= &(fip -> fi_lpoly);
	Zp	= fip -> fi_Zp;
	cbound	= fip -> fi_cbound;

	degree	= lpoly -> degree;
	lc	= &(lpoly -> coeff [degree]);

	p	= Zp -> p;

	mpz_init (ppow);
	mpz_init (min_modulus);

	mpz_abs (min_modulus, lc);
	mpz_mul_ui (min_modulus, min_modulus, 2);
	mpz_mul (min_modulus, min_modulus, cbound);

	mpz_set (ppow, p);
	for (a = 1; ; a++) {
		if (mpz_cmp (ppow, min_modulus) > 0) break;
		mpz_mul (ppow, ppow, p);
	}

	/* If fip->fi_lm_a is zero, then this is the first time we	*/
	/* are computing this power -- on the original polynomial.	*/
	/* Otherwise, we are recomputing a (hopefully smaller) power	*/
	/* on the polynomial after one or more factors have been	*/
	/* discovered and removed.					*/

	a_lm	= fip -> fi_lm_a;
	if ((a_lm > 0) AND (a > a_lm)) {
		/* Power actually *increased* after removing one or	*/
		/* more factors!					*/
		CALIB_FATAL_ERROR;
	}

	mpz_clear (min_modulus);
	mpz_clear (ppow);

	fip -> fi_lm_a	= a;
}

#ifdef ENABLE_VAN_HOEIJ

/*
 * Find the irreducible Z[x] factors using Van Hoeij's lattice reduction
 * algorithm.  Van Hoeij recommends first enumerating all subsets of
 * cardinality 1, 2 and 3.
 */

	static
	void
Zx_find_factors_Van_Hoeij (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int				a, a_lm, nleft, nleft2, card_limit;
struct calib_Zx_obj *		lpoly;
struct calib_Zx_factor *	nfp;
struct calib_Zx_factor **	hookp;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	lpoly	= &(fip -> fi_lpoly);
	nleft	= fip -> fi_nleft;

	/* Compute initial power "a" to lift to.  We choose 4 unless	*/
	/* this exceeds the power implied by Landau-Mignotte.		*/
	a	= 4;
	a_lm	= fip -> fi_lm_a;
	if (a > a_lm) {
		a = a_lm;
	}

	/* Lift modular factors (in Z[x] form) from current (mod p**1)	*/
	/* to target (mod p**a).					*/

#if 1
	/* Trial division (even partial enumeration isn't *guaranteed*	*/
	/* to work unless we lift all the way up to the Landau-Mignotte	*/
	/* bound.							*/
	a = a_lm;
#endif

	Zx_lift_factors (fip, a);

	nleft		= fip -> fi_nleft;

	if (calib_Zx_settings.factor_trace_level >= 2) {
		int i;
		for (i = 0; i < nleft; i++) {
			printf ("%2d = ", i);
			Zx_print (fip -> fi_zfactors [i]);
		}
	}

	/* Exhaustively enumerate all subsets S of factors up to	*/
	/* cardinality 3.						*/

	if (calib_Zx_settings.factor_trace_level >= 1) {
		printf ("Enumerating subsets 3/%d\n", nleft);
	}

#if 0
	card_limit	= 3;
#else
	card_limit	= 0;
#endif
	Zx_trial_division (fip, card_limit);

	nleft2	= fip -> fi_nleft;

	if (calib_Zx_settings.factor_trace_level >= 1) {
		printf ("We have %d factors left\n", nleft2);
	}

	if (nleft2 < nleft) {
		/* We found some factors, so lpoly is now smaller,	*/
		/* which means we can probably get a smaller Landau-	*/
		/* Mignotte bound and lower lifting power.  Recompute	*/
		/* these now.						*/
		Zx_factor_coefficient_upper_bound (fip);

		Zx_compute_lifting_target_power_of_p (fip);

		/* Put remaining factors first in their respective arrays. */
		Zx_VH_reorder_remaining_factors_first (fip);
	}

	/* When we have 7 or fewer modular factors left, then the	*/
	/* remaining polynomial is irreducible, because we have tried	*/
	/* all combinations of 3 or fewer modular factors.		*/

	nleft = fip -> fi_nleft;
	if ((nleft <= 7) AND (lpoly -> degree > 0)) {
		/* Append final irreducible factor. */
		nfp = CALIB_NEW (struct calib_Zx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= Zx_dup (lpoly);
		nfp -> next		= NULL;
		hookp	= fip -> fi_hookp;
		*hookp = nfp;
		hookp = &(nfp -> next);
		fip -> fi_hookp = hookp;
	}
	else if (nleft <= calib_Zx_settings.VH_max_enumerate) {
		/* The number of modular factors remaining is	*/
		/* sufficiently small that we finish using	*/
		/* exhaustive enumeration instead of using	*/
		/* Van Hoeij lattice-reduction.			*/
		Zx_trial_division (fip, nleft);

		if (lpoly -> degree > 0) {
			/* Append final irreducible factor. */
			nfp = CALIB_NEW (struct calib_Zx_factor);
			nfp -> multiplicity	= 1;
			nfp -> factor		= Zx_dup (lpoly);
			nfp -> next		= NULL;
			hookp	= fip -> fi_hookp;
			*hookp = nfp;
			hookp = &(nfp -> next);
			fip -> fi_hookp = hookp;
		}
	}
	else {
		/* Do Van Hoeij lattice-reduction. */

		Zx_Van_Hoeij_main (fip);
	}
}

/*
 * This is the part of the Van Hoeij algorithm at which we have committed
 * to solving the problen using lattice reduction.
 */

	static
	void
Zx_Van_Hoeij_main (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int				nleft, l, r, a_at_last_trial_div;
struct calib_Zx_factor *	fp;
mpq_t				Delta;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	/* Compute the desired "Information" target I_desired that we	*/
	/* are trying to hit: I_target = 0.12*n**2.			*/

	nleft = fip -> fi_nleft;
	fip -> fi_desired_I = (nleft * nleft) >> 3;

	Zx_VH_init_trace_bounds (fip);

	Zx_VH_init_B_L (fip);

	Zx_VH_choose_initial_matrix_and_parameters (fip);

	mpq_init (Delta);
	mpq_set_ui (Delta, 3, 4);

	a_at_last_trial_div = 0;

	for (;;) {

		Zx_VH_compute_trace_matrix_C (fip);

		Zx_VH_compute_C_M2 (fip);

		Zx_VH_build_Lambda_lattice_basis (fip);

		{ int			status, nvec, vsize;
		  struct ZMatrix *	smp;
		  struct ZMatrix *	dmp;

		  smp = &(fip -> fi_BLambda);
		  dmp = &(fip -> fi_Blll);
		  nvec  = smp -> zm_rows;
		  vsize = smp -> zm_cols;
		  Zx_VH_init_ZMatrix (dmp, nvec, vsize);

#ifdef CALIB_HAVE_FPLLL
		  extern int calib_FPLLL (mpz_ptr, mpz_srcptr, int, int);

		  status = calib_FPLLL (dmp -> zm_A, smp -> zm_A, nvec, vsize);
		  if (calib_Zx_settings.factor_trace_level >= 1) {
			printf ("calib_FPLLL: status = %d\n", status);
		  }
		  CALIB_FATAL_ERROR_IF (status NE 0);
#else
		  _calib_LLL (dmp -> zm_A, smp -> zm_A, nvec, vsize);
#endif
		}	

		l = fip -> fi_B_L.zm_rows;

		r = Zx_VH_find_M_short_vectors (fip);

		if (calib_Zx_settings.factor_trace_level >= 2) {
			printf ("l = %d\tr = %d\n", l, r);
		}

		fip -> fi_zfail = FALSE;

		CALIB_FATAL_ERROR_IF (r < 0);

		if (r EQ 0) {
			/* Reduced too far.  Give BLambda info an extra	*/
			/* bump.					*/
			fip -> fi_zfail = TRUE;
			Zx_VH_clear_BLambda (fip);
			Zx_VH_choose_new_matrix_and_parameters (fip);
			continue;
		}

		if (Zx_VH_short_vectors_do_not_span_factors (fip, r)) {
			/* Not enough information in BLambda for the	*/
			/* reduced lattice to span all factors!		*/
			/* We must increase information I.		*/
			Zx_VH_clear_BLambda (fip);
			Zx_VH_choose_new_matrix_and_parameters (fip);
			continue;
		}

		if (r EQ 1) {
			/* Remaining polynomial is irreducible. */
			fp = CALIB_NEW (struct calib_Zx_factor);
			fp -> multiplicity	= 1;
			fp -> factor		= Zx_dup (&(fip -> fi_lpoly));
			fp -> next		= NULL;
			*(fip -> fi_hookp) = fp;
			fip -> fi_hookp = &(fp -> next);
			break;
		}

		if (r >= l) {
			/* Dimension did not decrease. */
			if (fip -> fi_a > a_at_last_trial_div) {
				/* The basis has not decreased in size,	*/
				/* but the modulus has increased, so it	*/
				/* is worth re-testing the current	*/
				/* basis.				*/
				if (Zx_VH_basis_has_factored_poly (fip)) {
					break;
				}
			}

			/* Tally consecutive reduction failures. */
			++(fip -> fi_rfails);
			Zx_VH_clear_BLambda (fip);
			Zx_VH_choose_new_matrix_and_parameters (fip);
			continue;
		}

		fip -> fi_rfails = 0;

		Zx_VH_replace_lattice (fip, r);

		Zx_VH_clear_BLambda (fip);

#if 0		/* Initial trial division isn't guaranteed to work	*/
		/* unless you lift all the way to the LM bound before	*/
		/* hand.  Disabling this check allows us to disable	*/
		/* initial trial division with no ill effects.		*/
		if (4 * r > nleft) {
			/* New lattice L' cannot be W. */
			Zx_VH_choose_new_matrix_and_parameters (fip);
			continue;
		}
#endif

		if (Zx_VH_basis_has_factored_poly (fip)) {
			break;
		}
		Zx_VH_choose_new_matrix_and_parameters (fip);
	}

	mpq_clear (Delta);
}

/*
 * Let f(x) \in Z[x] be the poly we are factoring, and g(x) be any
 * (complex or p-adic) factor of f(x), n = degree(f), and d = degree(g).
 *
 * Computing an upper bound on the magnitude of the i-th trace of g(x)
 * is potentially fairly expensive, so we compute these only as they
 * are needed, and store them in an array so that they are computed at
 * most once each, and only if needed.
 *
 * We store only the d=1 entry for each i (i.e., an upper bound on the
 * magnitude of \alpha**i, where \alpha is a root of f(x).
 * The best bound that the current methods can give is
 *
 *		d * Root_Power_Upper_Bound[i],
 *
 * so our array is indexed only by i and we multiply the entry by d
 * during access.
 *
 * This routine initializes this array to those values that are
 * trivially computable, and -1 (the "not computed yet value) for all
 * of the other elements.
 */

	static
	void
Zx_VH_init_trace_bounds (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int				i, N;
const struct calib_Zx_obj *	f;
mpz_ptr				UB;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	f	= &(fip -> fi_lpoly);
	N	= f -> degree;

	UB = calib_new_Z_vector (N + 1);

	/* UB[0] = 1, because \alpha**0 = 1 for all \alpha. */
	mpz_set_ui (&UB [0], 1);

	/* The polynomial we are factoring is square free, so don't	*/
	/* waste time trying to do a square-free factorization of it.	*/
	/* Just compute a bound on its roots.				*/
	Zx_VH_root_bound_Z (&UB [1], f);

	/* All of the other elements are computed at most once, only if	*/
	/* they are needed.						*/
	for (i = 2; i <= N; i++) {
		mpz_set_si (&UB [i], -1);
	}

	fip -> fi_Tr_bounds	= UB;
}

/*
 * Initialize the starting basis for lattice L.
 */

	static
	void
Zx_VH_init_B_L (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, n;
struct ZMatrix *	zmp;
mpz_ptr			rowp;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	n	= fip -> fi_nleft;
	zmp	= &(fip -> fi_B_L);

	Zx_VH_init_ZMatrix (zmp, n, n);

	rowp	= zmp -> zm_A;

	/* Initialize n x n identity matrix.  Lattice L = Z^n. */
	for (i = 0; i < n; i++, rowp += n) {
		mpz_set_ui (&rowp [i], 1);
	}
}

/*
 * Choose the following:
 *
 * - Positive integers s, d.
 * - An s by d matrix A of small integers.
 * - Upper bounds B_1, B_2, ..., B_s on TA_i(g) for any rational factor
 *   g(x) of f(x).
 * - Non-negative integers a_1 > b_1, a_2 > b_2, ... a_s > b_s, such
 *   that b_i > log(2 B_i) / log(p).
 *
 * We want to choose these so that the "Information" of our lattice matrix:
 *
 *	I = (\sum_{i=1}^s a_i - b_i) / log(p)
 *
 * is approximately I_target.
 *
 * This version always uses s = d with A an identity matrix.
 */

	static
	void
Zx_VH_choose_initial_matrix_and_parameters (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, j, k, n, N, min_d, deg;
struct calib_Zx_obj **	zfactors;
struct calib_Zx_obj *	fp;
mpz_srcptr		coeff;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	n		= fip -> fi_nleft;
	N		= fip -> fi_lpoly.degree;
	zfactors	= fip -> fi_zfactors;

	/* Modular factors having many zero coefficients just below	*/
	/* the leading (monic) coefficient wind up having corresponding	*/
	/* traces that are zero.  This leads to BLambda matrices having	*/
	/* rows where the last s columns are all zero, which certainly	*/
	/* qualifies as "small" with respect to M. To avoid this (and	*/
	/* the incremental "ramp up" of d value that would otherwise	*/
	/* result), find the modular factor having the longest sequence	*/
	/* of such zero coefficients.  This establishes a lower bound	*/
	/* on d.							*/
	min_d = 2;
	for (j = 0; j < n; j++) {
		fp	= zfactors [j];
		deg	= fp -> degree;
		coeff	= fp -> coeff;
		for (i = deg - 1; ; i--) {
			/* x^deg is not a valid modular factor because	*/
			/* it is not square-free!			*/
			CALIB_FATAL_ERROR_IF (i < 0);
			if (mpz_sgn (&coeff [i]) NE 0) break;
		}
		k = deg - i;
		if (k > min_d) {
			min_d = k;
		}
	}

	if (calib_Zx_settings.factor_trace_level >= 2) {
		printf ("Have min_d = %d\n", min_d);
	}

	/* We always start out with strategy 1, where s=d and A is	*/
	/* (implicityly) the identity matrix. */

	fip -> fi_strategy	= 1;
	fip -> fi_s		= min_d;
	fip -> fi_d		= min_d;
	fip -> fi_A		= NULL;
	fip -> fi_ai		= CALIB_NEWA (N + 1, int);
	fip -> fi_bi		= CALIB_NEWA (N + 1, int);
	fip -> fi_a_b_gap	= 2;

	Zx_VH_compute_ai_bi (fip);
}

/*
 * Compute the matrix C[i,j] (n by s) of rounded modular trace values.
 */

	static
	void
Zx_VH_compute_trace_matrix_C (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, j, k, n, N, s, d, ai, bi;
int *			A;
const int *		srcirow;
struct ZMatrix *	zmp;
mpz_srcptr		Tr_fj_i, p, lc, srczrow;
mpz_ptr			dstrow, lcpow, lc_Tr_fj, lc_TA_fj;
mpz_t			P_TO_B, P_TO_A_MINUS_B, TEMP, TEMP2, HMOD1, HMOD2;
mpz_t			SUM, TERM;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	/* Get factor j i-th trace values up to date. */
	Zx_VH_compute_factor_traces (fip);

	n	= fip -> fi_nleft;
	s	= fip -> fi_s;
	d	= fip -> fi_d;
	A	= fip -> fi_A;
	N	= fip -> fi_lpoly.degree;
	Tr_fj_i	= fip -> fi_Tr_fj_i;
	p	= fip -> fi_Zp -> p;
	zmp	= &(fip -> fi_matrix_C);

	/* Compute needed powers of lead coefficient. */
	lc	= &(fip -> fi_lpoly.coeff [N]);
	lcpow = calib_new_Z_vector (1 + d);
	mpz_abs (&lcpow [1], lc);
	lc = &lcpow [1];
	for (i = 2; i <= d; i++) {
		mpz_mul (&lcpow [i], lc, &lcpow [i - 1]);
	}

	/* Compute lc_Tr_fj[i,j] = lcpow[i] * Tr_i(f_j).  This is where	*/
	/* the leading coefficient gets applied to the modular factor	*/
	/* traces for non-monic polynomials.				*/
	lc_Tr_fj = calib_new_Z_vector (n * (1 + d));
	for (j = 0; j < n; j++) {
		for (i = 1; i <= d; i++) {
			mpz_mul (&lc_Tr_fj [j * (d + 1) + i],
				 &lcpow [i],
				 &Tr_fj_i [j * (N + 1) + i]);
		}
	}

	/* Compute lc_TA_fj = T_A(f_j). */
	lc_TA_fj = calib_new_Z_vector (n * (1 + s));
	if (fip -> fi_strategy EQ 1) {
		CALIB_FATAL_ERROR_IF (A NE NULL);
		CALIB_FATAL_ERROR_IF (s NE d);

		/* Strategy 1: s=d and A is (implicitly) an identity matrix. */
		calib_copy_Z_vector(lc_TA_fj, lc_Tr_fj, n * (1 + s));
	}
	else {
		CALIB_FATAL_ERROR_IF (fip -> fi_strategy NE 2);
		CALIB_FATAL_ERROR_IF (A EQ NULL);

		/* Strategy 2: Must multiply by A matrix. */
		mpz_init (SUM);
		mpz_init (TERM);
		for (j = 0; j < n; j++) {
			dstrow = &lc_TA_fj [j * (1 + s)];
			srczrow = &lc_Tr_fj [j * (1 + d)];
			for (i = 1; i <= s; i++) {
				mpz_set_ui (SUM, 0);
				srcirow = &A [i * d];
				for (k = 0; k < d; k++) {
					mpz_set_ui (TERM, srcirow [k]);
					mpz_mul (TERM, TERM, &srczrow [1 + k]);
					mpz_add (SUM, SUM, TERM);
				}
				mpz_fdiv_r (SUM, SUM, fip -> fi_modulus);
				mpz_set (&dstrow [i], SUM);
			}
		}
		mpz_clear (TERM);
		mpz_clear (SUM);
	}

	mpz_init (P_TO_B);
	mpz_init (P_TO_A_MINUS_B);
	mpz_init (TEMP);
	mpz_init (TEMP2);
	mpz_init (HMOD1);
	mpz_init (HMOD2);

	Zx_VH_init_ZMatrix (zmp, n, s);

	dstrow = zmp -> zm_A;
	for (j = 0; j < n; j++, dstrow += s) {
		srczrow = &lc_TA_fj [j * (1 + s)];
		for (i = 1; i <= s; i++) {
			ai	= fip -> fi_ai [i];
			bi	= fip -> fi_bi [i];

			mpz_pow_ui (P_TO_B, p, bi);
			mpz_fdiv_q_2exp (HMOD1, P_TO_B, 1);
			mpz_pow_ui (P_TO_A_MINUS_B, p, ai - bi);
			mpz_fdiv_q_2exp (HMOD2, P_TO_A_MINUS_B, 1);

			mpz_set (TEMP, &srczrow [i]);
			mpz_fdiv_r (TEMP2, TEMP, P_TO_B);
			if (mpz_cmp (TEMP2, HMOD1) > 0) {
				mpz_sub (TEMP2, TEMP2, P_TO_B);
			}
			mpz_sub (TEMP2, TEMP, TEMP2);
			mpz_divexact (TEMP2, TEMP2, P_TO_B);
			mpz_fdiv_r (TEMP2, TEMP2, P_TO_A_MINUS_B);
			if (mpz_cmp (TEMP2, HMOD2) > 0) {
				mpz_sub (TEMP2, TEMP2, P_TO_A_MINUS_B);
			}
			mpz_set (&dstrow [i - 1], TEMP2);
		}
	}

	mpz_clear (HMOD2);
	mpz_clear (HMOD1);
	mpz_clear (TEMP2);
	mpz_clear (TEMP);
	mpz_clear (P_TO_A_MINUS_B);
	mpz_clear (P_TO_B);

	calib_free_Z_vector (lc_TA_fj,	n * (1 + s));
	calib_free_Z_vector (lc_Tr_fj,	n * (1 + d));
	calib_free_Z_vector (lcpow,		1 + d);
}

/*
 * For each modular factor f_j, compute (and store) the i-th trace,
 * for 0 <= i <= d, where d is the degree of the polynomial being factored.
 * These traces are computed as signed integers, (mod p**a).
 *
 * Let f(x) be the original polynomial being factored, and d = deg(f),
 *
 *	f(x) = \sum_{j=0}^d a_j*x^j.
 *
 * We note that a_j = 0 for j < 0 and for j > d.
 *
 * The recurrence relations for computing P_i (the i-th trace) of f(x)
 * as follows:
 *
 *	P_i = - i * a_{d-i} - \sum_{k=1}^{i-1} P_k * ~E_{i-k}
 *
 * where ~E_q = a_{d-q}.  This means:
 *
 *	P_i = - i * a_{d-i} - \sum_{k=1}^{i-1} P_k * a_{d-i+k}
 *
 * Now consider modular factor f_j(x) of degree m < d:
 *
 *	f_j(x) = \sum_{j=0}^m b_j*x^j,
 *
 * and we note that b_j = 0 for (j < 0) and for (j > m).  For 1 <= i <= d,
 * the i-th trace of f_j(x) is then
 *
 *	P_i = - i * b_{m-i} - \sum_{k=1}^{i-1} P_k * b_{m-i+k}
 *
 * We note that the first term is zero for i > m.  In the summation,
 * the coefficient b_{m-i+k} = 0 for {m-i+k} < 0 and for {m-i+k} > m.
 * The valid, non-zero b coeficients therefore satisfy:
 *
 *		{m-i+k} >= 0	AND	{m-i+k} <= m
 *
 * ==>		k >= i - m	AND	k <= i.
 *
 * So the summation bounds should be:
 *
 *		lo = max(1, i - m)
 *		hi = min(i - 1, i) = i - 1
 */

	static
	void
Zx_VH_compute_factor_traces (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, j, k, m, N, n, ncols, lo;
mpz_ptr			array;
mpz_ptr			rowp;
struct calib_Zx_obj **	zfactors;
struct calib_Zx_obj *	fp;
mpz_srcptr		coeff, modulus;
mpz_t			SUM, TERM, HMOD;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	N		= fip -> fi_lpoly.degree;
	n		= fip -> fi_nleft;
	zfactors	= fip -> fi_zfactors;
	modulus		= fip -> fi_modulus;
	array		= fip -> fi_Tr_fj_i;

	if (fip -> fi_a_Tr_fj >= fip -> fi_a) {
		CALIB_FATAL_ERROR_IF (fip -> fi_a_Tr_fj > fip -> fi_a);
		CALIB_FATAL_ERROR_IF (array EQ NULL);

		/* Traces are up to date with respect to the modulus. */
		return;
	}

	ncols = N + 1;

	if (array EQ NULL) {
		/* Allocate array the first time we call this. */
		array = calib_new_Z_vector (n * ncols);
		fip -> fi_Tr_fj_i = array;
	}

	mpz_init (SUM);
	mpz_init (TERM);
	mpz_init (HMOD);

	mpz_tdiv_q_2exp (HMOD, modulus, 1);

	rowp = array;
	for (j = 0; j < n; j++, rowp += ncols) {
		fp = zfactors [j];

		m	= fp -> degree;
		coeff	= fp -> coeff;

		mpz_set_ui (&rowp [0], m);
		for (i = 1; i <= N; i++) {
			/* Compute i-th trace of fp.  Start with the	*/
			/* first time, which can be missing.		*/
			if (i > m) {
				mpz_set_ui (SUM, 0);
			}
			else {
				mpz_mul_si (SUM, &coeff [m - i], - i);
			}
			lo = i - m;
			if (lo < 1) {
				lo = 1;
			}
			for (k = lo; k < i; k++) {
				mpz_mul (TERM, &rowp [k], &coeff [m - i + k]);
				mpz_sub (SUM, SUM, TERM);
			}
			/* Convert to (mod p**a) signed. */
			mpz_fdiv_r (SUM, SUM, modulus);
			if (mpz_cmp (SUM, HMOD) > 0) {
				mpz_sub (SUM, SUM, modulus);
			}
			mpz_set (&rowp [i], SUM);
		}
	}

	mpz_clear (HMOD);
	mpz_clear (TERM);
	mpz_clear (SUM);

	/* i-th trace values for j-th factors are now up to date with	*/
	/* respect to modulus p**a.					*/
	fip -> fi_a_Tr_fj = fip -> fi_a;
}

/*
 * Compute the parameters C and M (M2 = M**2).
 */

	static
	void
Zx_VH_compute_C_M2 (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			n, s, C, c1, c2;
double			foo, sqrt_half;
mpq_t			TERM1, TERM2;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	n	= fip -> fi_nleft;
	s	= fip -> fi_s;

	/* Let								*/
	/*								*/
	/*	M = sqrt (C**2 * n + s * (n / 2)**2)			*/
	/*								*/
	/* where C is a positive integer chosen in such a way that	*/
	/* neither one of the two terms under the square root is much	*/
	/* larger than the other one.					*/

	/* This means that C**2 ~ foo, where foo = (s * n / 4).		*/
	/* So we consider						*/
	/*								*/
	/*	C1 = floor (sqrt (foo))		and			*/
	/*	C2 = ceil  (sqrt (foo)),				*/
	/*								*/
	/* and pick the one that is closer in a quadratic sense.	*/

	foo	= s * n;
	foo	*= 0.25;
	sqrt_half = sqrt (0.5);
	c1 = (int) floor (sqrt (foo));
	c2 = (int) ceil  (sqrt (foo));
	if (c1 EQ c2) {
		C = c1;
	}
	else if ((((double) (c1 * c2)) / foo) >= sqrt_half) {
		C = c1;
	}
	else {
		C = c2;
	}
	fip -> fi_C	= C;

	mpq_init (TERM1);
	mpq_init (TERM2);

	mpq_set_ui (TERM1, C * C * n, 1);
	mpq_set_ui (TERM2, s * n * n, 4);
	mpq_canonicalize (TERM2);

	mpq_add (fip -> fi_M2, TERM1, TERM2);

	mpq_clear (TERM2);
	mpq_clear (TERM1);
}

/*
 * Build the basis matrix describing the integer lattice Lambda.
 */

	static
	void
Zx_VH_build_Lambda_lattice_basis (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, j, k, n, s, C, r, rows, cols, ai, bi;
const struct ZMatrix *	vmp;
const struct ZMatrix *	cmp;
struct ZMatrix *	bmp;
mpz_srcptr		p, vp, szp, coeff;
mpz_ptr			drowp;
mpz_t			TERM;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	n	= fip -> fi_nleft;
	s	= fip -> fi_s;
	p	= fip -> fi_Zp -> p;
	C	= fip -> fi_C;

	/* The basis matrix for our lattice L (B_L) consists of r	*/
	/* vectors of length n.						*/
	vmp	= &(fip -> fi_B_L);
	r	= vmp -> zm_rows;
	CALIB_FATAL_ERROR_IF (vmp -> zm_cols NE n);

	cmp = &(fip -> fi_matrix_C);
	CALIB_FATAL_ERROR_IF ((cmp -> zm_rows NE n) OR (cmp -> zm_cols NE s));

	rows	= r + s;
	cols	= n + s;

	bmp = &(fip -> fi_BLambda);

	Zx_VH_init_ZMatrix (bmp, rows, cols);

	mpz_init (TERM);

	vp	= vmp -> zm_A;
	drowp	= bmp -> zm_A;
	for (j = 0; j < r; j++) {
		/* First n cols are C*v_j. */
		for (i = 0; i < n; i++) {
			mpz_mul_ui (&drowp [i], &vp [i], C);
		}
		drowp += n;
		/* Last s colums are v_j*Cmatrix. */
		for (k = 0; k < s; k++) {
			mpz_set_ui (&drowp [k], 0);
		}
		for (i = 0; i < n; i++) {
			coeff = &vp [i];
			if (mpz_sgn (coeff) EQ 0) continue;
			szp = &(fip -> fi_matrix_C.zm_A [i * s]);
			for (k = 0; k < s; k++) {
				mpz_mul (TERM, coeff, &szp [k]);
				mpz_add (&drowp [k], &drowp [k], TERM);
			}
		}
		drowp += s;
		vp += n;
	}
	for (i = 1; i <= s; i++) {
		ai	= fip -> fi_ai [i];
		bi	= fip -> fi_bi [i];
		CALIB_FATAL_ERROR_IF (ai <= bi);

		drowp += n;

		mpz_pow_ui (&drowp [i - 1], p, ai - bi);

		drowp += s;
	}

	mpz_clear (TERM);
}

/*
 * Compute the Gram-Schmidt orthogonalization of the reduced lattice basis
 * and determine the set of all such vectors whose magnitude is at most M.
 * (We compare magnitude squared -- M^2 is always purely an integer.)
 *
 * Van Hoeij's paper recommends doing this orthgonalization using
 * floating-point arithmetic, and comparing against M using a tolerance
 * bounded by the maximum floating-point rounding error.  We may get to
 * this implementation eventually, but for now we have an exact rational
 * Gram-Schmidt code above, and we "cheat" (in reverse, because the fast
 * floating-point approximation is the *real* efficiency cheat) for
 * simplicity by using it for now.
 */

	static
	int
Zx_VH_find_M_short_vectors (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, j, nvec, vsize, r;
const struct ZMatrix *	bmp;
int *			list;
mpz_srcptr		Blll;
mpq_ptr			Bstar;
mpq_srcptr		sqp, M2;
mpq_t			MAG2;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	bmp	= &(fip -> fi_Blll);
	M2	= fip -> fi_M2;

	nvec	= bmp -> zm_rows;
	vsize	= bmp -> zm_cols;
	Blll	= bmp -> zm_A;

	Bstar = Zx_VH_Gram_Schmidt (Blll, nvec, vsize);

	mpq_init (MAG2);

	list = CALIB_NEWA (nvec, int);

	r = 0;
	for (i = 0; i < nvec; i++) {
		sqp = &Bstar [i * vsize];
		Zx_VH_mpq_mag2 (MAG2, sqp, vsize);
		j = mpq_sgn (MAG2);
		if (j <= 0) {
			CALIB_FATAL_ERROR_IF (j < 0);
			/* Vector j is exactly a linear combination	*/
			/* of the earlier vectors, and is therefore	*/
			/* unsuitable to include in any new basis.	*/
			if (calib_Zx_settings.factor_trace_level >= 2) {
				printf ("Vector %d is *too* short (0)\n", i);
			}
		}
		else if (mpq_cmp (MAG2, M2) <= 0) {
			if (calib_Zx_settings.factor_trace_level >= 2) {
				printf ("Vector %d is M-short (%.17g)\n",
					i, mpq_get_d (MAG2));
				}
			list [r] = i;
			++r;
		}
		else {
			if (calib_Zx_settings.factor_trace_level >= 2) {
				printf ("Vector %d is NOT M-short (%.17g)\n",
					i, mpq_get_d (MAG2));
			}
		}
	}

	mpq_clear (MAG2);

	calib_free_Q_vector (Bstar, nvec * vsize);

	fip -> fi_short	= list;

	return (r);
}

/*
 * Compute the magnitude squared of the given rational vector.
 */

	static
	void
Zx_VH_mpq_mag2 (

mpq_ptr		result,		/* IN/OUT: Resulting magnitude squared */
mpq_srcptr	v,		/* IN: Rational vector */
int		n		/* IN: Length of rational vector */
)
{
int		i;
mpq_t		SUM, TERM;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (n < 0));

	mpq_init (SUM);
	mpq_init (TERM);

	if (n > 0) {
		CALIB_FATAL_ERROR_IF (v EQ NULL);

		for (i = 0; i < n; i++) {
			mpq_mul (TERM, &v [i], &v [i]);
			mpq_add (SUM, SUM, TERM);
		}
	}

	mpq_set (result, SUM);

	mpq_clear (TERM);
	mpq_clear (SUM);
}

/*
 * Compute the exact rational Gram-Schmidt orthogonalization of the
 * given integer vectors.  It is the caller's responsibility to free
 * the returned matrix.
 */

	static
	mpq_ptr
Zx_VH_Gram_Schmidt (

mpz_srcptr		B,		/* IN: Integral vectors to	*/
					/*     orthogonalize		*/
int			nvec,		/* IN: Number of vectors	*/
int			vsize		/* IN: Size of each vector	*/
)
{
int		i, j, k, n2;
mpz_srcptr	zp;
mpq_srcptr	sqp;
mpq_ptr		Bstar, dqp;
mpq_t		TERM, SUM1, SUM2, RATIO;

	n2 = nvec * vsize;

	/* Convert integer vectors into rational form. */
	Bstar = calib_new_Q_vector (n2);
	for (i = 0; i < n2; i++) {
		mpz_set (mpq_numref (&Bstar [i]), &B [i]);
		mpz_set_ui (mpq_denref (&Bstar [i]), 1);
	}

	mpq_init (TERM);
	mpq_init (SUM1);
	mpq_init (SUM2);
	mpq_init (RATIO);

	/* Bstar_0 = B_0. */
	for (i = 1; i < nvec; i++) {
		zp  = &B [i * vsize];
		dqp = &Bstar [i * vsize];
		for (j = 0; j < i; j++) {
			/* Bstar [i] -= (<B_j,Bstar_j>/<Bstar_j,Bstar_j>)*B_j */
			mpq_set_ui (SUM1, 0, 1);
			mpq_set_ui (SUM2, 0, 1);
			sqp = &Bstar [j * vsize];
			for (k = 0; k < vsize; k++) {
				mpz_set (mpq_numref (TERM), &zp [k]);
				mpz_set_ui (mpq_denref (TERM), 1);
				mpq_mul (TERM, TERM, &sqp [k]);
				mpq_add (SUM1, SUM1, TERM);

				mpq_mul (TERM, &sqp [k], &sqp [k]);
				mpq_add (SUM2, SUM2, TERM);
			}
			k = mpq_sgn (SUM2);
			if (k <= 0) {
				CALIB_FATAL_ERROR_IF (k < 0);
				/* Bstar_j is zero.  Subtracting any	*/
				/* multiple of this does nothing.	*/
				continue;
			}
			mpq_div (RATIO, SUM1, SUM2);
			for (k = 0; k < vsize; k++) {
				mpq_mul (TERM, RATIO, &sqp [k]);
				mpq_sub (&dqp [k], &dqp [k], TERM);
			}
		}
	}

	if (calib_Zx_settings.factor_trace_level >= 3) {
		Zx_VH_debug_print_Gram_Schmidt (Bstar, nvec, vsize);
	}

	mpq_clear (RATIO);
	mpq_clear (SUM2);
	mpq_clear (SUM1);
	mpq_clear (TERM);

	return (Bstar);
}

/*
 * Print debugging information about the given Gram-Schmidt vectors.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
Zx_VH_debug_print_Gram_Schmidt (

mpq_srcptr	Bstar,		/* IN: Gram-Schmit orthogonal vectors */
int		nvec,		/* IN: Number of vectors */
int		vsize		/* IN: Size of each vector */
)
{
int		i, j, k;
size_t		nbits;
mpq_srcptr	qip, qjp;
mpq_t		SUM, TERM;

	for (i = 0; i < nvec; i++) {
		nbits	= calib_mpq_bit_size_vector (&Bstar [i * vsize], vsize);
		printf ("%d:\t%lu\n", i, nbits);
	}

	mpq_init (SUM);
	mpq_init (TERM);

	printf ("\nDot Products:\n");
	for (i = 0; i < nvec; i++) {
		qip = &Bstar [i * vsize];
		printf ("\n");
		for (j = 0; j < nvec; j++) {
			qjp = &Bstar [j * vsize];
			mpq_set_ui (SUM, 0, 1);
			for (k = 0; k < vsize; k++) {
				mpq_mul (TERM, &qip [k], &qjp [k]);
				mpq_add (SUM, SUM, TERM);
			}
			nbits = calib_mpq_bit_size (SUM);
			printf ("%d\t%d\t%d\t%.17g\t%lu\n",
				 i,
				 j,
				 mpq_sgn (SUM),
				 mpq_get_d (SUM),
				 nbits);
		}
	}

	mpq_clear (TERM);
	mpq_clear (SUM);
}

#pragma GCC diagnostic pop

/*
 * Return TRUE if-and-only-if the M-short vectors of the reduced
 * lattice do not span the space of all factors.  This is a clear
 * indication that the information I present in BLambda is not
 * sufficiently high and must be increased.
 */

	static
	calib_bool
Zx_VH_short_vectors_do_not_span_factors (

struct FInfo *		fip,		/* IN/OUT: Factoring algorithm info */
int			r		/* IN: Number of M-short vectors */
)
{
int			i, j, k, n, nvec, vsize;
calib_bool		span_failure, all_zero;
const struct ZMatrix *	rmp;
int *			list;
mpz_srcptr		szp;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	n	= fip -> fi_nleft;
	rmp	= &(fip -> fi_Blll);
	list	= fip -> fi_short;

	nvec	= rmp -> zm_rows;
	vsize	= rmp -> zm_cols;
	szp	= rmp -> zm_A;

	span_failure = FALSE;

	/* Test each column to see if it is all zero. */
	for (j = 0; j < n; j++) {
		all_zero = TRUE;
		for (k = 0; k < r; k++) {
			i = list [k];
			CALIB_FATAL_ERROR_IF ((i < 0) OR (i >= nvec));
			if (mpz_sgn (&szp [i * vsize + j]) NE 0) {
				all_zero = FALSE;
				break;
			}
		}
		if (all_zero) {
			span_failure = TRUE;
			break;
		}
	}

	return (span_failure);
}

/*
 * Clear out all of the BLambda stuff that gets recomputed
 * during each iteration.
 */

	static
	void
Zx_VH_clear_BLambda (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	Zx_VH_clear_ZMatrix (&(fip -> fi_Blll));
	Zx_VH_clear_ZMatrix (&(fip -> fi_BLambda));
	Zx_VH_clear_ZMatrix (&(fip -> fi_matrix_C));

	if (fip -> fi_A NE NULL) {
		free (fip -> fi_A);
		fip -> fi_A = NULL;
	}

	free (fip -> fi_short);
	fip -> fi_short = NULL;
}

/*
 * Replace the lattice L with the r M-short vectors contained in
 * fip -> fi_short.
 */

	static
	void
Zx_VH_replace_lattice (

struct FInfo *		fip,		/* IN/OUT: Factoring algorithm info */
int			r		/* IN: Number of M-short vectors */
)
{
int			i, j, k, n, snvecs, svsize, C;
const int *		list;
const struct ZMatrix *	rmp;
struct ZMatrix *	lmp;
mpz_srcptr		szp;
mpz_ptr			dzp;
mpz_t			REM;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	n	= fip -> fi_nleft;
	C	= fip -> fi_C;
	rmp	= &(fip -> fi_Blll);
	lmp	= &(fip -> fi_B_L);
	list	= fip -> fi_short;

	CALIB_FATAL_ERROR_IF (lmp -> zm_cols NE n);
	CALIB_FATAL_ERROR_IF (r >= lmp -> zm_rows);

	/* Free up existing lattice basis. */
	Zx_VH_clear_ZMatrix (lmp);

	/* Allocate new lattice basis. */
	Zx_VH_init_ZMatrix (lmp, r, n);

	snvecs	= rmp -> zm_rows;
	svsize	= rmp -> zm_cols;

	mpz_init (REM);

	dzp = lmp -> zm_A;

	for (j = 0; j < r; j++) {
		i = list [j];
		CALIB_FATAL_ERROR_IF ((i < 0) OR (i >= snvecs));

		szp = &(rmp -> zm_A [i * svsize]);
		for (k = 0; k < n; k++) {
			mpz_fdiv_qr_ui (&dzp [k], REM, &szp [k], C);
			CALIB_FATAL_ERROR_IF (mpz_sgn (REM) NE 0);
		}

		dzp += n;
	}

	mpz_clear (REM);
}

/*
 * Select a new s, d, matrix A and parameters a_i, b_i.
 */

	static
	void
Zx_VH_choose_new_matrix_and_parameters (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, k, n, N, s, d, gap, gapinc;
calib_bool		switched;
int *			A;
mpz_srcptr		p;
double			fgap, log2p, fbits;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	n		= fip -> fi_nleft;
	N		= fip -> fi_lpoly.degree;
	s		= fip -> fi_s;
	d		= fip -> fi_d;
	A		= fip -> fi_A;
	p		= fip -> fi_Zp -> p;
	gap		= fip -> fi_a_b_gap;

	CALIB_FATAL_ERROR_IF (A NE NULL);

	log2p = Zx_VH_mpz_log2 (p);

	switched = FALSE;

	if ((fip -> fi_strategy EQ 1) AND (fip -> fi_rfails >= 3)) {
		/* Three or more consecutive failures to reduce the	*/
		/* number of basis vectors.  Switch to strategy 2.	*/

		switched = TRUE;
		fip -> fi_strategy = 2;
		d = N / 2;
#if 1
		s = 2;
#else
		/* Let s ~ log2 (d) (count bits in d). */
		s = 1;
		k = d;
		while (k > 1) {
			k >>= 1;
			++s;
		}
		/* Force s >= 4. */
		if (s < 4) {
			s = 4;
		}
#endif

		fgap = 0.125 * (n * n) / (log2p * s);
		gap = (int) ceil (fgap);
		if (gap < 4) {
			gap = 4;
		}

		if (calib_Zx_settings.factor_trace_level >= 2) {
			printf ("Switching to strategy 2: s = %d, d = %d, gap = %d\n",
				s, d, gap);
		}

		/* Initialize the random number generator. */
		calib_random_init (&(fip -> fi_rand));
	}

	if (fip -> fi_strategy EQ 1) {
		CALIB_FATAL_ERROR_IF (s NE d);

#if 1
		/* FIXME -- Temporary kludge to let me write and	*/
		/* develop the rest of the algorithm.  Refine this	*/
		/* later.						*/

		++s;
		++d;
		++gap;

		CALIB_FATAL_ERROR_IF ((s > N) OR (d > N));
#endif
	}
	else if (fip -> fi_strategy EQ 2) {
		/* Choose a new random matrix A.  We use integers in	*/
		/* the range of [0,31].					*/
		if (NOT switched) {
			/* Increase info. */
			++s;
			/* Bump precision of each trace value by at	*/
			/* least this many bits.			*/
			fbits = 16.0;
			if (log2p >= fbits) {
				gapinc = 1;
			}
			else {
				gapinc = (int) ceil (fbits / log2p);
			}
			gap += gapinc;
		}
		k = s * d;
		A = CALIB_NEWA (k, int);
		for (i = 0; i < k; i++) {
			A [i] = calib_random_u32 (&(fip -> fi_rand)) & 0x1f;
		}
		fip -> fi_A = A;
	}
	else {
		CALIB_FATAL_ERROR;
	}

	fip -> fi_s		= s;
	fip -> fi_d		= d;
	fip -> fi_a_b_gap	= gap;

	Zx_VH_compute_ai_bi (fip);
}

/*
 * Compute the a_i and b_i values for the current s, d, and matrix A.
 */

	static
	void
Zx_VH_compute_ai_bi (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, j, N, s, d, gap, aval, bval, max_a;
int			Aij;
int *			A;
mpz_srcptr		p, lc;
mpz_ptr			UB_Tr, UB_lc_Tr, B;
double			log2_p, log2Bi;
mpz_t			LCPOW, P_POW_B, TWO_P_POW_B, SUM, TERM;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	N		= fip -> fi_lpoly.degree;
	s		= fip -> fi_s;
	d		= fip -> fi_d;
	A		= fip -> fi_A;
	p		= fip -> fi_Zp -> p;
	gap		= fip -> fi_a_b_gap;

	/* Let N = degree(f(x)), n be the number of modular factors,	*/
	/* and lc be the leading coefficient of f(x).			*/
	/* For all 1 <= d, compute UB_Tr[i] to be an upper bound on	*/
	/* Tr_i(g), the i-th trace of any rational factor g(x) of f(x).	*/

	UB_Tr = calib_new_Z_vector (1 + d);
	for (i = 1; i <= d; i++) {
		Zx_VH_compute_trace_bound (fip, &UB_Tr [i], i, N);
	}

	/* Let lc be the absolute value of the leading coefficient of	*/
	/* f(x).  Compute						*/
	/*								*/
	/*		UB_lc_Tr[i] = lc^i * UB_Tr [i]			*/
	/*								*/
	/* This is where the leading coefficient is incorporated into	*/
	/* the computation of the bounds b_i.				*/

	lc = &(fip -> fi_lpoly.coeff [N]);
	UB_lc_Tr = calib_new_Z_vector (1 + d);
	mpz_init (LCPOW);
	mpz_abs (LCPOW, lc);
	for (i = 1; i <= d; i++) {
		mpz_mul (&UB_lc_Tr [i], LCPOW, &UB_Tr [i]);
		mpz_mul (LCPOW, LCPOW, lc);
	}
	mpz_clear (LCPOW);

	/* Now compute B[i] to be an upper bound on the i-th entry	*/
	/* of A.UB_lc_Tr (called T_A(g) in the paper -- this is version	*/
	/* that has been corrected for non-monic polynomials).		*/
	/* (Note: we use zero-origin indexing here becasue matrix A is	*/
	/* stored with zero-origin indices.)				*/
	B = calib_new_Z_vector (1 + s);
	if (fip -> fi_strategy EQ 1) {
		CALIB_FATAL_ERROR_IF (A NE NULL);
		CALIB_FATAL_ERROR_IF (s NE d);

		/* Strategy 1: Matrix A is (implicitly) an identity matrix. */
		calib_copy_Z_vector (B, UB_lc_Tr, 1 + s);
	}
	else {
		CALIB_FATAL_ERROR_IF (fip -> fi_strategy NE 2);
		CALIB_FATAL_ERROR_IF (A EQ NULL);

		/* Strategy 2: Must explicitly multiply by matrix A	*/
		/* (using absolute values of its entrys) to obtain the	*/
		/* upper bound on T_A(g) (corrected for non-monic	*/
		/* leading coefficients).				*/

		mpz_init (SUM);
		mpz_init (TERM);
		for (i = 0; i < s; i++) {
			mpz_set_ui (SUM, 0);
			for (j = 0; j < d; j++) {
				Aij = A [i * d + j];
				if (Aij < 0) {
					Aij = - Aij;
				}
				mpz_set_ui (TERM, Aij);
				mpz_mul (TERM, TERM, &UB_lc_Tr [1 + j]);
				mpz_add (SUM, SUM, TERM);
			}
			mpz_set (&B [1 + i], SUM);
		}
	}

	/* FIXME -- make this next loop choose a_i and b_i, and	stop	*/
	/* when \sum_{k=1}^i (a_k - b_k) is sufficiently large.  This	*/
	/* chooses d to achieve a pre-specified information value.	*/

	log2_p = Zx_VH_mpz_log2 (p);

	mpz_init (P_POW_B);
	mpz_init (TWO_P_POW_B);

	max_a = 0;
	for (i = 1; i <= s; i++) {
		log2Bi = 1.0 + Zx_VH_mpz_log2 (&B [i]);
		bval = (int) ceil (log2Bi / log2_p);

		CALIB_FATAL_ERROR_IF (bval <= 0);
		mpz_pow_ui (P_POW_B, p, bval);
		mpz_mul_2exp (TWO_P_POW_B, P_POW_B, 1);

		for (j = 0; ; j++) {
			CALIB_FATAL_ERROR_IF (j > 2);

			if (mpz_cmp (P_POW_B, &B [i]) > 0) {
				/* bval is now sufficiently large. */
				break;
			}
			/* Floating-point approximation was too crude. */
			++bval;
			mpz_mul (P_POW_B, P_POW_B, p);
			mpz_mul_2exp (TWO_P_POW_B, P_POW_B, 1);
		}

		aval = bval + gap;

		fip -> fi_ai [i] = aval;
		fip -> fi_bi [i] = bval;

		if (aval > max_a) {
			max_a = aval;
		}
	}

	mpz_clear (TWO_P_POW_B);
	mpz_clear (P_POW_B);

	if (max_a > fip -> fi_a) {
		/* Need to do more lifting. */
		Zx_lift_factors (fip, max_a);
	}

	calib_free_Z_vector (B,	 1 + s);
	calib_free_Z_vector (UB_lc_Tr, 1 + d);
	calib_free_Z_vector (UB_Tr,	 1 + d);

	if (calib_Zx_settings.factor_trace_level >= 1) {
		Zx_VH_print_Information (fip);
	}
}

/*
 * Print out the "information content" I of basis Lambda.
 */

	static
	void
Zx_VH_print_Information (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int		i, n, s, sum, n2;
int *		ai;
int *		bi;
mpz_srcptr	p;
double		Info, log2_p, frac;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	n	= fip -> fi_nleft;
	s	= fip -> fi_s;
	ai	= fip -> fi_ai;
	bi	= fip -> fi_bi;
	p	= fip -> fi_Zp -> p;

	n2 = n * n;

	sum = 0;
	for (i = 1; i <= s; i++) {
		sum += (ai [i] - bi [i]);
	}

	log2_p = Zx_VH_mpz_log2 (p);

	Info = sum * log2_p;

	CALIB_FATAL_ERROR_IF (Info <= 0.0);

	frac = Info / n2;

	printf ("I = %.17g ~ %.17g * n^2\ts = %d\tsum = %d\n",
		Info, frac, s, sum);
}

/*
 * Compute y = log_2 (|x|), where x is an arbitrary GMP integer.
 * 
 * For convenience, we define log_2 (0) = 0.
 */

	static
	double
Zx_VH_mpz_log2 (

mpz_srcptr		x	/* IN: Value to take base-2 logarithm of */
)
{
size_t		bits;
mpz_t		absx;
double		mantissa, xval, y;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	mpz_init_set (absx, x);

	mpz_abs (absx, absx);

	bits = calib_mpz_bit_size (x);
	if (bits <= 1) {
		return (0);
	}

	if (bits > DBL_MANT_DIG) {
		/* Number is longer than the mantissa of an IEEE	*/
		/* double precision floating point number.  Round to	*/
		mpz_tdiv_q_2exp (absx, absx, bits - (1 + DBL_MANT_DIG));
		mpz_add_ui (absx, absx, 1);
		mpz_tdiv_q_2exp (absx, absx, 1);
	}
	else if (bits < DBL_MANT_DIG) {
		/* Number is shorter than mantissa of IEEE double.  Pad	*/
		/* with least-significant zeros.			*/
		mpz_mul_2exp (absx, absx, DBL_MANT_DIG - bits);
	}
	mantissa = mpz_get_ui (absx);
	/* Scale value into range [1.0, 2.0). */
	xval = ldexp (mantissa, 1 - DBL_MANT_DIG);

	mpz_clear (absx);

	y = (bits - 1) + log2 (xval);

	return (y);
}

/*
 * Let f(x) \in Z[x] be the poly we are factoring, and g(x) be any
 * (complex or p-adic) factor of f(x), n = degree(f), and d = degree(g).
 *
 * Return an upper bound on the magnitude of the i-th trace of g(x).
 * (This is a function of i and d, where 0 <= i <= n and 1 <= d <= n.)
 *
 * Let Brt > 0 be an upper bound on the magnitude of any complex root
 * of f(x).  Then bound[i,d] = d*|Brt|**i.  Our array contains
 * memoized values of |Brt|**i.
 *
 * Let h_i(x) = resultant (x - t^i, f(t), t).  Then degree(h_i) = n, and
 * the roots of h_i(x) have the form \alpha^i for each root \alpha of
 * f(x).  We explicitly compute h_i(x) \in Z[x], and compute an upper
 * bound on the magnitude of the roots of h_i(x).  This gives us
 * bound[i,1] = |Brt|**i, which we store in fip->fi_Tr_bounds[i].
 *
 * Note that one could also compute
 *	h_id(x) = resultant (x - d*t, h_i(t), t)
 * and then compute an upper bound on the roots of h_id(x).  But the
 * various methods we use to compute these bounds wind up with bounds
 * proportional to d.  So when asking for root_upper_bound(h_id), we
 * just use d*root_upper_bound(h_i).
 */

	static
	void
Zx_VH_compute_trace_bound (

struct FInfo *		fip,		/* IN/OUT: Factoring algorithm info */
mpz_ptr			result,		/* IN/OUT: Upper bound */
int			i,		/* IN: The i-th trace */
int			d		/* IN: The degree d of factor g(x) */
)
{
int				N;
const struct calib_Zx_obj *	f;
mpz_ptr				UB, entry;
mpz_t				MAX, TEMP;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	f	= &(fip -> fi_lpoly);
	N	= f -> degree;

	CALIB_FATAL_ERROR_IF ((i < 0) OR (i > N));
	CALIB_FATAL_ERROR_IF ((d < 1) OR (d > N));

	UB	= fip -> fi_Tr_bounds;

	entry = &UB [i];
	if (mpz_sgn (entry) < 0) {
		/* We must compute this array entry now. */
		mpz_init (MAX);
		mpz_init (TEMP);

		if (calib_Zx_settings.VH_use_resultant_trace_root_bound) {
			struct calib_Zx_factor *	res_factors;
			struct calib_Zx_factor *	rfp;
			res_factors = Zx_VH_trace_bounds_compute_h_i (f, i);

			mpz_set_si (MAX, -1);
			for (rfp = res_factors; rfp NE NULL; rfp = rfp -> next) {
				if (rfp -> factor -> degree <= 0) {
					/* Skip integer content. */
					continue;
				}
				Zx_VH_trace_root_bound_Z (TEMP, rfp -> factor, i);
				CALIB_FATAL_ERROR_IF (mpz_sgn (TEMP) <= 0);

				if (mpz_cmp (TEMP, MAX) > 0) {
					mpz_set (MAX, TEMP);
				}
			}
			Zx_free_factors (res_factors);
		}
		else {
			/* Computationally cheap, but loose bound. */
			mpz_pow_ui (MAX, &UB [1], i);
		}

		CALIB_FATAL_ERROR_IF (mpz_sgn (MAX) <= 0);

		mpz_set (entry, MAX);

		mpz_clear (TEMP);
		mpz_clear (MAX);
	}

	mpz_mul_ui (result, entry, d);
}

/*
 * Compute h_i(x) = resultant (x - t^i, f(t), t).
 *
 * This is a polynomial whose roots have the form \alpha^i for all
 * roots \alpha of the given f(x).
 *
 * It is returned as a dynamically allocated list of Z[x] factors.
 */

	static
	struct calib_Zx_factor *
Zx_VH_trace_bounds_compute_h_i (

const struct calib_Zx_obj *	f,	/* IN: The polynomial f(x) */
int				i	/* IN: The root power i */
)
{
int				var;
struct calib_Zxyz_dom *		dom;
struct calib_Zx_factor *	sqfactors;
struct calib_Zx_factor *	nfp;
struct calib_Zxyz_obj		RES, F, G, T_pow_i;
struct calib_Zx_obj		ZX_RES;
const char *			varlist [2];

	CALIB_FATAL_ERROR_IF ((f EQ NULL) OR (i <= 0));

	if (i EQ 1) {
		/* Asking for roots of the form alpha^1 EQ alpha.	*/
		/* The original polynomial f(x) does this.		*/
		nfp = CALIB_NEW (struct calib_Zx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= Zx_dup (f);
		nfp -> next		= NULL;

		return (nfp);
	}

	/* Construct domain Z[x,t]. */
	varlist [0] = "x";
	varlist [1] = "t";
	dom = calib_make_Zxyz_dom (2, varlist);

	Zx_init (&ZX_RES);

	dom -> init (dom, &RES);
	dom -> init (dom, &F);
	dom -> init (dom, &G);
	dom -> init (dom, &T_pow_i);

	/* Copy f(t) (in Z[x] form) into F (in Z[x,t] form). */
	var = 1;			/* var = "t" */
	dom -> copy_from_Zx (&F, f,  var);

	/* Construct g(x,t) = x - t^i \in Z[x,t]. */
	var = 0;			/* var = "x" */
	dom -> set_var_power (&G, var, 1);
	var = 1;			/* var = "t" */
	dom -> set_var_power (&T_pow_i, var, i);
	dom -> sub (&G, &G, &T_pow_i);

	var = 1;	/* Eliminate "t". */
	dom -> resultant (&RES, &F, &G, var);

	/* Extract "x" polynomial to Z[x] form. */
	var = 0;
	dom -> copy_into_Zx (&ZX_RES, &RES, var);

	sqfactors = Zx_compute_square_free_factorization (&ZX_RES);

	dom -> clear (&T_pow_i);
	dom -> clear (&G);
	dom -> clear (&F);
	dom -> clear (&RES);

	Zx_clear (&ZX_RES);

	calib_free_Zxyz_dom (dom);

	return (sqfactors);
}

/*
 * Compute an *integer* upper bound on the magnitude of any (complex)
 * root of the given polynomial.  This routine wraps some extra effort
 * around Zx_VH_root_bound_Z(), computing said bound:
 *
 * 1. On the given polynomial.
 *
 * 2. On the individual factors of a square-free factorization of the
 *    given polynomial.  (This uses the maximum root bound among all
 *    such square-free factors.)
 *
 * The smaller of doors #1 and #2 is used.
 *
 * Consider factoring x^96 - 1.  Trial division by all modular factor
 * subsets of cardinality <= 3 yields a degree 56 polynomial.
 * Computing the polynomial whose roots are \alpha^32 yields another
 * degree 56 polynomial that is actually (x^2 + x + 1)^28.  Our best
 * integer root bound on the expanded polynomial is 812, but the
 * root bound on x^2 + x + 1 is 1, which is completely tight.
 */

	static
	void
Zx_VH_trace_root_bound_Z (

mpz_ptr				UBOUND,	/* IN/OUT: Root upper bound */
const struct calib_Zx_obj *	poly,	/* IN: Polynomial to bound roots of */
int				power	/* IN: Power roots taken to */
)
{
int				deg, k, cmp;
struct calib_Zx_factor *	sqf_list;
struct calib_Zx_factor *	fp;
mpz_t				UB1, UB2, UB3;
struct calib_Zx_obj		LPOLY;

	CALIB_FATAL_ERROR_IF ((UBOUND EQ NULL) OR (poly EQ NULL));

	deg	= poly -> degree;

	CALIB_FATAL_ERROR_IF (deg <= 0);

	mpz_init (UB1);
	mpz_init (UB2);
	mpz_init (UB3);

	/* Make local copy of polynomial. */
	Zx_init_degree (&LPOLY, deg);
	Zx_prim_part (UB1, &LPOLY, poly);

	/* Lead coefficient should now be strictly positive. */
	CALIB_FATAL_ERROR_IF (mpz_sgn (&(LPOLY.coeff [LPOLY.degree])) <= 0);

	/* Compute root bound of original polynomial. */
	Zx_VH_root_bound_Z (UB1, &LPOLY);

	/* Compute square-free factorization. */
	sqf_list = Zx_compute_square_free_factorization (&LPOLY);

	/* Compute UB2 = max root bound among square-free factors. */
	mpz_set_ui (UB2, 0);

	/* Get total multiplicity of non-trivial square-free factors. */
	k = 0;
	for (fp = sqf_list; fp NE NULL; fp = fp -> next) {
		if (fp -> factor -> degree <= 0) continue;
		k += fp -> multiplicity;
	}

	if (k <= 1) {
		/* No non-trivial split.  Use only first bound. */
		if (calib_Zx_settings.factor_trace_level >= 1) {
			gmp_printf ("Root^%d bound of orignial poly:\t%Zd\n",
				    power, UB1);
		}
		mpz_set (UBOUND, UB1);
	}
	else {
		if (calib_Zx_settings.factor_trace_level >= 1) {
			printf ("Have %d square-free factors.\n", k);
		}
		for (fp = sqf_list; fp NE NULL; fp = fp -> next) {
			if (fp -> factor -> degree <= 0) continue;

			Zx_VH_root_bound_Z (UB3, fp -> factor);

			if (mpz_cmp (UB3, UB2) > 0) {
				mpz_set (UB2, UB3);
			}
		}

		if (calib_Zx_settings.factor_trace_level >= 1) {
			gmp_printf ("Root^%d bound of original poly:\t%Zd\n",
				    power, UB1);
			gmp_printf ("Root^%d bound of sqfactors:\t%Zd\n",
				    power, UB2);
		}
		cmp = mpz_cmp (UB1, UB2);
		if (cmp < 0) {
			if (calib_Zx_settings.factor_trace_level >= 1) {
				printf ("Original poly has best root bound\n");
			}
			mpz_set (UBOUND, UB1);
		}
		else if (cmp EQ 0) {
			if (calib_Zx_settings.factor_trace_level >= 1) {
				printf ("Same root bound from both\n");
			}
			mpz_set (UBOUND, UB1);
		}
		else {
			if (calib_Zx_settings.factor_trace_level >= 1) {
				printf ("Square-free factors have best root bound\n");
			}
			mpz_set (UBOUND, UB2);
		}
	}
	Zx_free_factors (sqf_list);

	Zx_clear (&LPOLY);

	mpz_clear (UB3);
	mpz_clear (UB2);
	mpz_clear (UB1);
}

/*
 * Compute an *integer* upper bound on the magnitude of any (complex)
 * root of the given polynomial.  We use the best of several well-known
 * root bounds.
 */

	static
	void
Zx_VH_root_bound_Z (

mpz_ptr				UBOUND,	/* IN/OUT: Root upper bound */
const struct calib_Zx_obj *	poly	/* IN: Polynomial to bound roots of */
)
{
int		i, n;
mpz_srcptr	a, ai, an;
mpz_t		BEST_UB, UB, K, REM;
mpq_t		RATIO, QSUM;

	CALIB_FATAL_ERROR_IF ((UBOUND EQ NULL) OR (poly EQ NULL));

	n	= poly -> degree;
	a	= poly -> coeff;

	an = &a [n];

	CALIB_FATAL_ERROR_IF (mpz_sgn (an) EQ 0);

	mpz_init (BEST_UB);
	mpz_init (UB);
	mpz_init (K);
	mpz_init (REM);

	mpq_init (RATIO);
	mpq_init (QSUM);

	/* A bound from Lagrange: max{1, \sum_{i=0}^{n-1} |a_i/a_n|}. */
	for (i = 0; i < n; i++) {
		ai = &a [i];
		if (mpz_sgn (ai) EQ 0) continue;
		mpz_abs (mpq_numref (RATIO), ai);
		mpz_abs (mpq_denref (RATIO), an);
		mpq_canonicalize (RATIO);
		mpq_add (QSUM, QSUM, RATIO);
	}
	mpz_cdiv_q (UB, mpq_numref (QSUM), mpq_denref (QSUM));
	if (mpz_cmp_ui (UB, 1) < 0) {
		mpz_set_ui (UB, 1);
	}
	mpz_set (BEST_UB, UB);

	/* A bound from Cauchy:						*/
	/*	1 + max{|a_n-1/a_n|, |a_n-2/a_n|, ..., |a_0/a_n|}	*/
	mpz_set_ui (UB, 0);
	for (i = 0; i < n; i++) {
		ai = &a [i];
		if (mpz_sgn (ai) EQ 0) continue;
		mpz_abs (mpq_numref (RATIO), ai);
		mpz_abs (mpq_denref (RATIO), an);
		mpz_cdiv_q (K, mpq_numref (RATIO), mpq_denref (RATIO));
		if (mpz_cmp (K, UB) > 0) {
			mpz_set (UB, K);
		}
	}
	mpz_add_ui (UB, UB, 1);
	if (mpz_cmp (UB, BEST_UB) < 0) {
		mpz_set (BEST_UB, UB);
	}

	/* Another bound from Lagrange:					*/
	/*	2 * max {|a_n-1/a_n|, |a_n-2/a_n|^(1/2), ...		*/
	/*		 |a_0/a_n|^(1/n)}				*/
	mpz_set_ui (UB, 0);
	if (n >= 1) {
		ai = &a [n - 1];
		if (mpz_sgn (ai) NE 0) {
			mpz_abs (mpq_numref (RATIO), ai);
			mpz_abs (mpq_denref (RATIO), an);
			mpz_cdiv_q (UB, mpq_numref (RATIO), mpq_denref (RATIO));
		}
	}
	if (n >= 2) {
		ai = &a [n - 2];
		if (mpz_sgn (ai) NE 0) {
			mpz_abs (mpq_numref (RATIO), ai);
			mpz_abs (mpq_denref (RATIO), an);
			mpz_cdiv_q (UB, mpq_numref (RATIO), mpq_denref (RATIO));
			mpz_sqrtrem (K, REM, K);
			if (mpz_sgn (REM) > 0) {
				mpz_add_ui (K, K, 1);
			}
			if (mpz_cmp (K, UB) > 0) {
				mpz_set (UB, K);
			}
		}
	}
	for (i = 3; i <= n; i++) {
		ai = &a [n - i];
		if (mpz_sgn (ai) EQ 0) continue;
		mpz_abs (mpq_numref (RATIO), ai);
		mpz_abs (mpq_denref (RATIO), an);
		mpz_cdiv_q (K, mpq_numref (RATIO), mpq_denref (RATIO));
		mpz_rootrem (K, REM, K, i);
		if (mpz_sgn (REM) > 0) {
			mpz_add_ui (K, K, 1);
		}
		if (mpz_cmp (K, UB) > 0) {
			mpz_set (UB, K);
		}
	}
	mpz_mul_ui (UB, UB, 2);
	if (mpz_cmp (UB, BEST_UB) < 0) {
		mpz_set (BEST_UB, UB);
	}

	/* FIXME -- insert addtional root bounds here. */

	mpz_set (UBOUND, BEST_UB);

	mpq_clear (QSUM);
	mpq_clear (RATIO);

	mpz_clear (REM);
	mpz_clear (K);
	mpz_clear (UB);
	mpz_clear (BEST_UB);
}

/*
 * We have a reduced lattice basis that "could" represent a correct
 * grouping of modular factors into true integer factors.  Do the
 * more expensive test now to see if it really is our final basis --
 * and if so, generate these true integer factors.
 *
 * Returns TRUE if-and-only-if the true factors have been discovered.
 */

	static
	calib_bool
Zx_VH_basis_has_factored_poly (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int			nvec, vsize;
calib_bool		factored;
const struct ZMatrix *	lmp;
struct QMatrix		rref_mat;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	/* Get current lattice. */
	lmp	= &(fip -> fi_B_L);
	nvec	= lmp -> zm_rows;
	vsize	= lmp -> zm_cols;

	CALIB_FATAL_ERROR_IF (vsize NE fip -> fi_nleft);

	factored = FALSE;

	Zx_VH_init_QMatrix (&rref_mat, nvec, vsize);

	Zx_VH_compute_reduced_row_echelon_form (&rref_mat, lmp);

	if (Zx_VH_all_columns_are_units (&rref_mat)) {
		/* Every column is all zeros, except for a single 1.	*/
		/* See if the rows represent true factors over Z[x].	*/
		if (Zx_VH_true_factorization_found (fip, &rref_mat)) {
			factored = TRUE;
		}
	}

	Zx_VH_clear_QMatrix (&rref_mat);

	return (factored);
}

/*
 * Compute the Reduced Row Echolon Form of the given integer matrix.
 * The result can have rational entries.
 */

	static
	void
Zx_VH_compute_reduced_row_echelon_form (

struct QMatrix *	rref,		/* IN/OUT: Reduced Row Echelon Form */
const struct ZMatrix *	src		/* IN: Source integer matrix */
)
{
int		i, j, k, m, rows, cols, best_row, best_col, limit;
int *		order;
mpz_srcptr	szp;
mpq_srcptr	srcrowp;
mpq_ptr		M, dstrowp, piv, pivrowp, konst;
mpq_t		TEMP;

	CALIB_FATAL_ERROR_IF ((rref EQ NULL) OR (src EQ NULL));

	rows	= src -> zm_rows;
	cols	= src -> zm_cols;

	CALIB_FATAL_ERROR_IF (rref -> qm_rows NE rows);
	CALIB_FATAL_ERROR_IF (rref -> qm_cols NE cols);

	M = calib_new_Q_vector (rows * cols);

	/* Copy source integer matrix into rational RREF matrix. */
	k = rows * cols;
	szp = src -> zm_A;
	for (i = 0; i < k; i++) {
		mpz_set (mpq_numref (&M [i]), &szp [i]);
		mpz_set_ui (mpq_denref (&M [i]), 1);
	}

	/* Allow efficient re-ordering of the rows. */
	order = CALIB_NEWA (rows, int);
	for (i = 0; i < rows; i++) {
		order [i] = i;
	}

	mpq_init (TEMP);

	for (m = 0; m < rows; m++) {
		/* Identify remaining row with left-most non-zero. */
		best_col	= cols + 1;
		best_row	= -1;
		for (i = m; i < rows; i++) {
			srcrowp = &M [order [i] * cols];
			limit = (best_col < cols) ? best_col : cols;
			for (j = 0; j < limit; j++) {
				if ((mpq_sgn (&srcrowp [j]) NE 0) AND
				    (j < best_col)) {
					best_col	= j;
					best_row	= i;
					break;
				}
			}
		}
		if (best_row < 0) {
			/* All remaining rows are zero. */
			break;
		}

		/* Swap best_row with row m. */
		k = order [best_row];
		order [best_row] = order [m];
		order [m] = k;

		pivrowp = &M [order [m] * cols];
		piv = &pivrowp [best_col];
		if (mpq_cmp_ui (piv, 1, 1) NE 0) {
			/* Divide row by this pivot element. */
			for (j = best_col + 1; j < cols; j++) {
				mpq_div (&pivrowp [j], &pivrowp [j], piv);
			}
			mpq_set_ui (piv, 1, 1);
		}

		/* Eliminate all rows below. */
		for (i = m + 1; i < rows; i++) {
			dstrowp = &M [order [i] * cols];
			konst = &dstrowp [best_col];
			if (mpq_sgn (konst) EQ 0) continue;
			for (j = best_col + 1; j < cols; j++) {
				mpq_mul (TEMP, konst, &pivrowp [j]);
				mpq_sub (&dstrowp [j], &dstrowp [j], TEMP);
			}
			mpq_set_ui (konst, 0, 1);
		}
	}

	/* Copy rows (in proper order) back to caller. */
	for (i = 0; i < rows; i++) {
		calib_copy_Q_vector (&(rref -> qm_A [i * cols]),
			       &M [order [i] * cols],
			       cols);
	}

	mpq_clear (TEMP);

	free (order);

	calib_free_Q_vector (M, rows * cols);
}

/*
 * Return TRUE if-and-only-if every column of the given matrix
 * (in reduced row echelon form) is all zero except for a single one.
 */

	static
	calib_bool
Zx_VH_all_columns_are_units (

const struct QMatrix *	rref	/* IN: Reduced row echelon form to test */
)
{
int			i, j, k, rows, cols;
mpq_srcptr		M, p;

	CALIB_FATAL_ERROR_IF (rref EQ NULL);

	rows	= rref -> qm_rows;
	cols	= rref -> qm_cols;
	M	= rref -> qm_A;

	for (j = 0; j < cols; j++) {
		p = &M [j];
		k = 0;
		for (i = 0; i < rows; i++, p += cols) {
			if (mpq_sgn (p) EQ 0) continue;
			if (mpq_cmp_ui (p, 1, 1) NE 0) {
				/* Not zero and not one. */
				return (FALSE);
			}
			/* Count number of ones. */
			++k;
			if (k > 1) return (FALSE);
		}
		if (k <= 0) return (FALSE);

		/* Column j is a unit! */
	}

	/* All columns are units! */

	return (TRUE);
}

/*
 * Each row of the given RREF matrix identifies a subset of the modular
 * factors, and together these rows partition all of the modular factors.
 * Return TRUE if-and-only-if *all* such non-empty subsets represents
 * a true factor over Z[x].  If TRUE is returned, we also append said
 * Z[x] factors to the output list of factors.
 */

	static
	calib_bool
Zx_VH_true_factorization_found (

struct FInfo *		fip,		/* IN/OUT: Factoring algorithm info */
const struct QMatrix *	rref		/* IN: RREF matrix giving partition */
					/*     of the modular factors */
)
{
int		i, j, nvec, vsize;
calib_bool	factored;
mpq_srcptr	A, rowp;
int *		array;
int *		ip;
int **		partition;


	CALIB_FATAL_ERROR_IF (fip EQ NULL);
	CALIB_FATAL_ERROR_IF (rref EQ NULL);

	nvec	= rref -> qm_rows;
	vsize	= rref -> qm_cols;
	A	= rref -> qm_A;

	CALIB_FATAL_ERROR_IF (vsize NE fip -> fi_nleft);

	factored = FALSE;

	/* Construct partition indicated by given RREF matrix. */
	array		= CALIB_NEWA (vsize, int);
	partition	= CALIB_NEWA (nvec + 1, int *);
	ip = array;
	rowp = A;
	for (i = 0; i < nvec; i++, rowp += vsize) {
		partition [i] = ip;
		for (j = 0; j < vsize; j++) {
			if (mpq_sgn (&rowp [j]) NE 0) {
				*ip++ = j;
			}
		}
	}
	partition [i] = ip;
	CALIB_FATAL_ERROR_IF (ip NE &array [vsize]);

	/* For simplicity, lift all factors up to the Landau-Mignotte	*/
	/* bound right here.  If the modulus is too small, divisibility	*/
	/* can otherwise fail even though the proper partition of	*/
	/* modular factors has been identified.				*/
	Zx_lift_factors (fip, fip -> fi_lm_a);

	/* Efficiency: test leading/trailing coefficients first. */
	if (Zx_VH_lc_tc_are_divisible (fip, nvec, partition)) {
		/* Now do full divisibility test. */
		if (Zx_VH_full_divisibility_test (fip, nvec, partition)) {
			factored = TRUE;
		}
	}

	free (partition);
	free (array);

	return (factored);
}

/*
 * Let F be the set of remaining modular factors.  (|F| = nleft.)
 * We are given a partition F = \union_i S_i of these factors (where the
 * S_i are mutually disjoint).  Let P_i = \prod_{f \in S_i} f (mod modulus)
 * where we take the coefficients of P_i to be *signed* integers.
 *
 * Return TRUE if-and-only-if the following divisibility property holds
 * for all i:
 *
 *	lc(lpoly) * tc(P_i) divides lc(lpoly) * tc(lpoly)
 *
 * Of course we do not need to compute the full polynomial P_i at this
 * point, only its trailing coeffcient.
 */

	static
	calib_bool
Zx_VH_lc_tc_are_divisible (

struct FInfo *	fip,		/* IN/OUT: Factoring algorithm info */
int		npart,		/* IN: Num subsets in partition */
int **		partition	/* IN: Partition of modular factors */
)
{
int			i, j, nleft, deg, result;
calib_bool		all_tc_divisible, monic;
struct calib_Zx_obj *	lpoly;
int *			ip1;
int *			ip2;
mpz_srcptr		lclpoly, tclpoly, modulus;
struct calib_Zx_obj **	zfactors;
struct calib_Zx_obj *	fp;
mpz_t			LCTC, PROD, HMOD, TEMP;

	CALIB_FATAL_ERROR_IF ((fip EQ NULL) OR (partition EQ NULL));

	nleft		= fip -> fi_nleft;
	lpoly		= &(fip -> fi_lpoly);
	zfactors	= fip -> fi_zfactors;
	modulus		= fip -> fi_modulus;

	CALIB_FATAL_ERROR_IF (npart NE fip -> fi_B_L.zm_rows);
	CALIB_FATAL_ERROR_IF (nleft NE fip -> fi_B_L.zm_cols);

	deg = lpoly -> degree;
	lclpoly = &(lpoly -> coeff [deg]);
	tclpoly = &(lpoly -> coeff [0]);

	mpz_init (LCTC);
	mpz_init (PROD);
	mpz_init (HMOD);
	mpz_init (TEMP);

	/* See if we are factoring a monic polynomial. */
	mpz_abs (LCTC, lclpoly);
	monic = (mpz_cmp_ui (LCTC, 1) EQ 0);

	mpz_fdiv_q_2exp (HMOD, modulus, 1);
	mpz_mul (LCTC, lclpoly, tclpoly);

	all_tc_divisible = TRUE;

	for (i = 0; i < npart; i++) {
		ip1 = partition [i];
		ip2 = partition [i + 1];
		if (ip1 >= ip2) {
			/* This subset contains no factors.  Skip. */
			continue;
		}
		mpz_set (PROD, lclpoly);
		while (ip1 < ip2) {
			j = *ip1++;
			CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= nleft));
			fp = zfactors [j];
			mpz_mul (PROD, PROD, &(fp -> coeff [0]));
			mpz_fdiv_r (PROD, PROD, modulus);
		}
		if (mpz_cmp (PROD, HMOD) > 0) {
			mpz_sub (PROD, PROD, modulus);
		}

		if (monic) {
			result = mpz_divisible_p (LCTC, PROD);
		}
		else {
			mpz_gcd (TEMP, lclpoly, PROD);
			mpz_divexact (TEMP, PROD, TEMP);
			result = mpz_divisible_p (TEMP, tclpoly);
		}
		if (NOT result) {
			/* The divisibility property is *NOT* satisfied! */
			all_tc_divisible = FALSE;
			break;
		}
	}

	mpz_clear (TEMP);
	mpz_clear (HMOD);
	mpz_clear (PROD);
	mpz_clear (LCTC);

	return (all_tc_divisible);
}

/*
 * Do the full polynomial divisibility check.
 */

	static
	calib_bool
Zx_VH_full_divisibility_test (

struct FInfo *	fip,		/* IN/OUT: Factoring algorithm info */
int		npart,		/* IN: Num subsets in partition */
int **		partition	/* IN: Partition of modular factors */
)
{
int				i, j, k, nleft, deg;
calib_bool			factored;
struct calib_Zx_obj *		factors;
struct calib_Zx_obj *		lpoly;
int *				ip1;
int *				ip2;
mpz_srcptr			lclpoly, modulus;
struct calib_Zx_obj *		prod;
struct calib_Zx_obj *		fp;
struct calib_Zx_obj **		zfactors;
struct calib_Zx_factor **	hookp;
struct calib_Zx_factor *	nfp;
struct calib_Zx_obj		QUOT, REM;
mpz_t				D;

	CALIB_FATAL_ERROR_IF ((fip EQ NULL) OR (partition EQ NULL));

	nleft		= fip -> fi_nleft;
	lpoly		= &(fip -> fi_lpoly);
	zfactors	= fip -> fi_zfactors;
	modulus		= fip -> fi_modulus;

	CALIB_FATAL_ERROR_IF (npart NE fip -> fi_B_L.zm_rows);
	CALIB_FATAL_ERROR_IF (nleft NE fip -> fi_B_L.zm_cols);

	deg = lpoly -> degree;
	lclpoly = &(lpoly -> coeff [deg]);

	factors = CALIB_NEWA (npart, struct calib_Zx_obj);
	for (i = 0; i < npart; i++) {
		Zx_init_degree (&factors [i], 0);
	}

	Zx_init (&QUOT);
	Zx_init (&REM);

	mpz_init (D);

	factored = TRUE;

	for (i = 0; i < npart; i++) {
		ip1 = partition [i];
		ip2 = partition [i + 1];
		if (ip1 >= ip2) {
			/* This subset contains no factors.  Skip. */
			continue;
		}
		prod = &factors [i];
		prod -> degree = 0;
		mpz_set (&(prod -> coeff [0]), lclpoly);
		while (ip1 < ip2) {
			j = *ip1++;
			CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= nleft));
			fp = zfactors [j];
			Zx_mul (prod, prod, fp);
		}
		k = prod -> degree;
		for (j = 0; j <= k; j++) {
			Zx_Zm_to_Z (&(prod -> coeff [j]), modulus);
		}

		/* Take the primitive part. */
		Zx_prim_part (NULL, prod, prod);

		/* See if it divides evenly. */
		if (NOT Zx_exactly_divides (&QUOT, lpoly, prod)) {
			/* This factor does not divide evenly. */
			factored = FALSE;
			break;
		}
	}

	mpz_clear (D);

	Zx_clear (&REM);
	Zx_clear (&QUOT);

	if (factored) {
		/* Append true Z[x] factors to output list. */
		hookp = fip -> fi_hookp;
		for (i = 0; i < npart; i++) {
			fp = &factors [i];
			if (fp -> degree <= 0) continue;
			nfp = CALIB_NEW (struct calib_Zx_factor);
			nfp -> multiplicity	= 1;
			nfp -> factor		= Zx_dup (fp);
			nfp -> next		= NULL;
			*hookp = nfp;
			hookp = &(nfp -> next);
		}
		fip -> fi_hookp = hookp;
	}

	Zx_clear_array (factors, npart);
	free (factors);

	return (factored);
}

/*
 * Print the current lattice basis L.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
Zx_VH_print_B_L (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
	Zx_VH_print_ZMatrix (&(fip -> fi_B_L));
}

#pragma GCC diagnostic pop

/*
 * Print the given ZMatrix.
 */

	static
	void
Zx_VH_print_ZMatrix (

const struct ZMatrix *	zmp	/* IN: ZMatrix to print */
)
{
int			i, j, k, nvec, vsize, w, pad;
mpz_srcptr		B_L;
char **			smat;
int *			width;
char *			s;

	CALIB_FATAL_ERROR_IF (zmp EQ NULL);

	nvec	= zmp -> zm_rows;
	vsize	= zmp -> zm_cols;
	B_L	= zmp -> zm_A;

	k = nvec * vsize;
	smat = CALIB_NEWA (k, char *);
	memset (smat, 0, k * sizeof (smat [0]));
	for (i = 0; i < k; i++) {
		smat [i] = mpz_get_str (NULL, 10, &B_L [i]);
	}

	width = CALIB_NEWA (vsize, int);
	memset (width, 0, vsize * sizeof (width [0]));
	
	for (i = 0; i < nvec; i++) {
		for (j = 0; j < vsize; j++) {
			k = i * vsize + j;
			w = strlen (smat [k]);
			if (w > width [j]) {
				width [j] = w;
			}
		}
	}

	for (i = 0; i < nvec; i++) {
		for (j = 0; j < vsize; j++) {
			k = i * vsize + j;
			s = smat [k];
			w = strlen (s);
			for (pad = width [j] - w; pad > 0; pad--) {
				putchar (' ');
			}
			printf (" %s", s);
		}
		printf ("\n");
	}

	free (width);
	k = nvec * vsize;
	for (i = 0; i < k; i++) {
		free (smat [i]);
		smat [i] = NULL;
	}
	free (smat);
}

/*
 * Re-oder all of the factors so that:
 *
 *	1. All "non-taken" factors come first.
 *	2. All "taken" factors come last.
 *	3. The sort is otherwise stable.
 *
 * The number of "non-taken" factors is verified to be fip->fi_nleft.
 *
 * We do this so that most of the software does not have to deal with
 * indirection through fi_findex.
 *
 * The following fip-> arrays are reordered here:
 *
 *	fi_zfactors
 *	fi_zpfactors
 *	fi_vlist
 *	fi_zvlist
 *	fi_taken
 *	fi_findex
 */

	static
	void
Zx_VH_reorder_remaining_factors_first (

struct FInfo *	fip		/* IN/OUT: Factoring algorithm info */
)
{
int			i, j, num_fact, nleft;
calib_bool *		taken;
int *			findex;
struct calib_Zx_obj **	zfactors;
struct calib_Zpx_obj **	zpfactors;
struct calib_Zpx_obj **	vlist;
struct calib_Zx_obj **	zvlist;
struct calib_Zx_obj **	zx_temp;
struct calib_Zpx_obj **	zpx_temp;
int *			order;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	num_fact	= fip -> fi_num_fact;
	zfactors	= fip -> fi_zfactors;
	zpfactors	= fip -> fi_zpfactors;
	vlist		= fip -> fi_vlist;
	zvlist		= fip -> fi_zvlist;
	taken		= fip -> fi_taken;
	findex		= fip -> fi_findex;

	/* Construct order[] giving the order ordering of existing factors. */
	order = CALIB_NEWA (num_fact, int);
	j = 0;
	for (i = 0; i < num_fact; i++) {
		if (NOT taken [i]) {
			order [j++] = i;
		}
	}
	nleft = j;

	CALIB_FATAL_ERROR_IF (fip -> fi_nleft NE nleft);

	for (i = 0; i < num_fact; i++) {
		if (taken [i]) {
			order [j++] = i;
		}
	}
	CALIB_FATAL_ERROR_IF (j NE num_fact);

	zx_temp		= CALIB_NEWA (num_fact, struct calib_Zx_obj *);
	zpx_temp	= CALIB_NEWA (num_fact, struct calib_Zpx_obj *);

	memset (zx_temp, 0, num_fact * sizeof (zx_temp));
	for (i = 0; i < num_fact; i++) {
		j = order [i];
		zx_temp [i] = zfactors [j];
	}
	for (i = 0; i < num_fact; i++) {
		CALIB_FATAL_ERROR_IF (zx_temp [i] EQ NULL);
	}
	memcpy (zfactors, zx_temp, num_fact * sizeof (zfactors [0]));

	memset (zpx_temp, 0, num_fact * sizeof (zpx_temp));
	for (i = 0; i < num_fact; i++) {
		j = order [i];
		zpx_temp [i] = zpfactors [j];
	}
	for (i = 0; i < num_fact; i++) {
		CALIB_FATAL_ERROR_IF (zpx_temp [i] EQ NULL);
	}
	memcpy (zpfactors, zpx_temp, num_fact * sizeof (zpfactors [0]));

	memset (zpx_temp, 0, num_fact * sizeof (zpx_temp));
	for (i = 0; i < num_fact; i++) {
		j = order [i];
		zpx_temp [i] = vlist [j];
	}
	for (i = 0; i < num_fact; i++) {
		CALIB_FATAL_ERROR_IF (zpx_temp [i] EQ NULL);
	}
	memcpy (vlist, zpx_temp, num_fact * sizeof (zpfactors [0]));

	if (zvlist NE NULL) {
		/* Only present when doing quadratic lifting. */
		memset (zx_temp, 0, num_fact * sizeof (zvlist [0]));
		for (i = 0; i < num_fact; i++) {
			j = order [i];
			zx_temp [i] = zvlist [j];
		}
		for (i = 0; i < num_fact; i++) {
			CALIB_FATAL_ERROR_IF (zx_temp [i] EQ NULL);
		}
		memcpy (zvlist, zx_temp, num_fact * sizeof (zvlist [0]));
	}

	for (i = 0; i < nleft; i++) {
		taken [i] = FALSE;
	}
	for (; i < num_fact; i++) {
		taken [i] = TRUE;
	}

	for (i = 0; i < num_fact; i++) {
		findex [i] = i;
	}

	free (zpx_temp);
	free (zx_temp);
	free (order);
}

#endif	/* ENABLE_VAN_HOEIJ */

/*
 * Initialize the given FInfo.
 */

	static
	void
Zx_init_FInfo (

struct FInfo *			fip,	/* IN/OUT: Factoring algorithm info */
const struct calib_Zx_obj *	poly	/* IN: Polynomial to factor */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF ((fip EQ NULL) OR (poly EQ NULL));

	deg = poly -> degree;

	memset (fip, 0, sizeof (*fip));

	fip -> fi_poly	= poly;

	/* Make a copy of poly from which we can divide out Z[x]	*/
	/* factors, as they are discovered.				*/
	Zx_init_degree (&(fip -> fi_lpoly), deg);
	Zx_set (&(fip -> fi_lpoly), poly);

	/* Initialize list of computed Z[x] factors. */
	fip -> fi_hookp = &(fip -> fi_list);

	mpz_init (fip -> fi_lcinv);
	mpz_init (fip -> fi_modulus);
	mpz_init (fip -> fi_cbound);
	mpz_init (fip -> fi_lc_tc);

#ifdef ENABLE_VAN_HOEIJ
	mpq_init (fip -> fi_M2);
#endif
}

/*
 * Clean up the given FInfo.
 */

	static
	void
Zx_cleanup_FInfo (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int				i, deg, num_fact, nleft;
struct calib_Zp_dom *		Zp;
struct calib_Zpx_dom *		Zpx;
struct calib_Zx_obj **		zfactors;
struct calib_Zpx_obj **		zpfactors;
struct calib_Zpx_obj **		vlist;
struct calib_Zx_obj **		zvlist;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	Zp		= fip -> fi_Zp;
	Zpx		= fip -> fi_Zpx;
	num_fact	= fip -> fi_num_fact;
	nleft		= fip -> fi_nleft;

	deg	= fip -> fi_lpoly.degree;

	fip -> fi_poly		= NULL;		/* We do not own this. */

	Zx_clear (&(fip -> fi_lpoly));

	if (fip -> fi_Zpx_poly NE NULL) {
		Zpx -> free (fip -> fi_Zpx_poly);
		fip -> fi_Zpx_poly	= NULL;
	}
	if (fip -> fi_monic_Zpx_poly NE NULL) {
		Zpx -> free (fip -> fi_monic_Zpx_poly);
		fip -> fi_monic_Zpx_poly = NULL;
	}

	Zpx -> free_factors (fip -> fi_zpfactor_list);
	fip -> fi_zpfactor_list = NULL;

	fip -> fi_num_fact	= 0;
	fip -> fi_nleft		= 0;
	fip -> fi_lm_a		= 0;
	fip -> fi_a		= 0;

	fip -> fi_list		= NULL;		/* We don't own this. */
	fip -> fi_hookp		= NULL;		/* We don't own this. */

	fip -> fi_Zp		= NULL;		/* freeing Zp below. */
	fip -> fi_Zpx		= NULL;		/* freeing Zpx below. */

	zfactors = fip -> fi_zfactors;
	fip -> fi_zfactors = NULL;
	if (zfactors NE NULL) {
		for (i = 0; i < num_fact; i++) {
			Zx_free (zfactors [i]);
			zfactors [i] = NULL;
		}
		free (zfactors);
	}

	zpfactors = fip -> fi_zpfactors;
	fip -> fi_zpfactors = NULL;
	if (zpfactors NE NULL) {
		/* The factors are freed in linked-list form above. */
		memset (zpfactors, 0, num_fact * sizeof (zpfactors [0]));
		free (zpfactors);
	}

	vlist = fip -> fi_vlist;
	fip -> fi_vlist = NULL;
	if (vlist NE NULL) {
		for (i = 0; i < num_fact; i++) {
			Zpx -> free (vlist [i]);
			vlist [i] = NULL;
		}
		free (vlist);
	}
	zvlist = fip -> fi_zvlist;
	fip -> fi_zvlist = NULL;
	if (zvlist NE NULL) {
		for (i = 0; i < num_fact; i++) {
			Zx_free (zvlist [i]);
			zvlist [i] = NULL;
		}
		free (zvlist);
	}
	fip -> fi_num_osp	= 0;
	if (fip -> fi_osp_struct NE NULL) {
		free (fip -> fi_osp_struct);
	}
	if (fip -> fi_omit_struct NE NULL) {
		free (fip -> fi_omit_struct);
	}
	if (fip -> fi_taken NE NULL) {
		free (fip -> fi_taken);
	}
	if (fip -> fi_chosen NE NULL) {
		free (fip -> fi_chosen);
	}
	if (fip -> fi_findex NE NULL) {
		free (fip -> fi_findex);
	}
	if (fip -> fi_comb NE NULL) {
		free (fip -> fi_comb);
	}

	mpz_clear (fip -> fi_lcinv);
	mpz_clear (fip -> fi_cbound);
	mpz_clear (fip -> fi_modulus);
	mpz_clear (fip -> fi_lc_tc);

	calib_free_Z_vector (fip -> fi_tc_stack, num_fact + 1);

	calib_free_Z_vector (fip -> fi_Tr_bounds, deg + 1);

	calib_free_Z_vector (fip -> fi_Tr_fj_i, nleft * (deg + 1));

	if (fip -> fi_ai NE NULL) {
		free (fip -> fi_ai);
	}
	if (fip -> fi_bi NE NULL) {
		free (fip -> fi_bi);
	}
	if (fip -> fi_A NE NULL) {
		free (fip -> fi_A);
	}

#ifdef ENABLE_VAN_HOEIJ
	if (fip -> fi_B_L.zm_A NE NULL) {
		Zx_VH_clear_ZMatrix (&(fip -> fi_B_L));
	}
	if (fip -> fi_matrix_C.zm_A NE NULL) {
		Zx_VH_clear_ZMatrix (&(fip -> fi_matrix_C));
	}
	mpq_clear (fip -> fi_M2);
	if (fip -> fi_BLambda.zm_A NE NULL) {
		Zx_VH_clear_ZMatrix (&(fip -> fi_BLambda));
	}
	if (fip -> fi_Blll.zm_A NE NULL) {
		Zx_VH_clear_ZMatrix (&(fip -> fi_Blll));
	}
	if (fip -> fi_short NE NULL) {
		free (fip -> fi_short);
	}
#endif

	calib_free_Zpx_dom (Zpx);
	calib_free_Zp_dom (Zp);

	memset (fip, 0, sizeof (*fip));
}

#ifdef ENABLE_VAN_HOEIJ

/*
 * Initialize the given ZMatrix to all zeros.
 */

	static
	void
Zx_VH_init_ZMatrix (

struct ZMatrix *	zmp,	/* IN/OUT: ZMatrix to initialize */
int			rows,	/* IN: Number of rows */
int			cols	/* IN: Number of columns */
)
{
	CALIB_FATAL_ERROR_IF ((zmp EQ NULL) OR (rows <= 0) OR (cols <= 0));

	zmp -> zm_rows	= rows;
	zmp -> zm_cols	= cols;
	zmp -> zm_A	= calib_new_Z_vector (rows * cols);
}

/*
 * Clear out the given ZMatrix.
 */

	static
	void
Zx_VH_clear_ZMatrix (

struct ZMatrix *	zmp	/* IN/OUT: ZMatrix to clear out */
)
{
	CALIB_FATAL_ERROR_IF (zmp EQ NULL);

	calib_free_Z_vector (zmp -> zm_A, zmp -> zm_rows * zmp -> zm_cols);

	memset (zmp, 0, sizeof (*zmp));
}

/*
 * Initialize the given QMatrix to all zeros.
 */

	static
	void
Zx_VH_init_QMatrix (

struct QMatrix *	qmp,	/* IN/OUT: QMatrix to initialize */
int			rows,	/* IN: Number of rows */
int			cols	/* IN: Number of columns */
)
{
	CALIB_FATAL_ERROR_IF ((qmp EQ NULL) OR (rows <= 0) OR (cols <= 0));

	qmp -> qm_rows	= rows;
	qmp -> qm_cols	= cols;
	qmp -> qm_A	= calib_new_Q_vector (rows * cols);
}

/*
 * Clear out the given QMatrix.
 */

	static
	void
Zx_VH_clear_QMatrix (

struct QMatrix *	qmp	/* IN/OUT: QMatrix to clear out */
)
{
	CALIB_FATAL_ERROR_IF (qmp EQ NULL);

	calib_free_Q_vector (qmp -> qm_A, qmp -> qm_rows * qmp -> qm_cols);

	memset (qmp, 0, sizeof (*qmp));
}

#endif	/* ENABLE_VAN_HOEIJ */

/*
 * Determine irreducible Z[x] factors via exhaustive enumeration
 * (the so-called Berlekamp-Zassenhaus method).
 *
 * For each subset S of the factors (ordered by increasing cardinality
 * from 1 to num_fact/2), compute
 *
 *	A = lc(poly) * prod_{f \in S} f		(mod MODULUS),
 *
 *	B = convert_to_Zx (A),
 *
 *	C = prim_part (B),
 *
 * and then test if C(x) evenly divides poly(x).  If so, then
 * C is an irreducible factor of poly, which should be removed,
 * and S should be removed from the set of modular factors.
 */

	static
	void
Zx_exhaustive_enumeration (

struct FInfo *		fip		/* IN/OUT: Factoring algorithm info */
)
{
int				num_fact, nleft, a_lm;
struct calib_Zx_factor **	hookp;
struct calib_Zx_factor *	nfp;
struct calib_Zx_obj *		lpoly;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	lpoly		= &(fip -> fi_lpoly);
	a_lm		= fip -> fi_lm_a;
	num_fact	= fip -> fi_num_fact;

	/* Lift modular factors (in Z[x] form) from current (mod p**1)	*/
	/* to target (mod p**a).  We always lift all of the way to the	*/
	/* power implied by the Landau-Mignotte bound.			*/

	Zx_lift_factors (fip, a_lm);

	/* Exhaustively enumerate all subsets S of factors in order	*/
	/* by increasing cardinality (but avoid enumerating both S and	*/
	/* its complement).						*/

	nleft = num_fact;
	Zx_trial_division (fip, nleft);

	if (lpoly -> degree > 0) {
		/* Append final irreducible factor. */
		nfp = CALIB_NEW (struct calib_Zx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= Zx_dup (lpoly);
		nfp -> next		= NULL;

		hookp		= fip -> fi_hookp;
		*hookp = nfp;
		hookp = &(nfp -> next);
		fip -> fi_hookp = hookp;
	}
}

/*
 * Use the Hensel Lemma (p-adic construction) to lift the given
 * factorization to a sufficiently large modulus = p^k > 2*cbound.
 *
 * Choose which method to use: linear or quadratic.
 */

	static
	void
Zx_lift_factors (

struct FInfo *		fip,		/* IN/OUT: Factoring algorithm info */
int			desired_pow	/* IN/OUT: Desired power of p */
)
{
	switch (calib_Zx_settings.lift_method) {
	case CALIB_ZX_LIFT_METHOD_LINEAR:
		Zx_lift_factors_linear (fip, desired_pow);
		break;

	case CALIB_ZX_LIFT_METHOD_QUADRATIC:
		Zx_lift_factors_quadratic (fip, desired_pow);
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}

	if (calib_Zx_settings.validate_lifted_modular_factors) {
		Zx_validate_lifted_modular_factors (fip);
	}
}

/*
 * Use the Hensel Lemma (p-adic construction) to lift the given
 * factorization to a sufficiently large modulus = p^k > 2*cbound.
 *
 * We use linear lifting here, instead of quadratic (i.e.,
 * modulus goes from p^k to p^{k+1}, not from p^{k} to p^{2*k}).
 * This is simpler, and avoids gross overshoot on the modulus.
 */

	static
	void
Zx_lift_factors_linear (

struct FInfo *		fip,		/* IN/OUT: Factoring algorithm info */
int			desired_pow	/* IN/OUT: Desired power of p */
)
{
int				i, j, k, num_fact, deg, fdeg, cur_pow;
const struct calib_Zx_obj *	poly;
const struct calib_Zp_dom *	Zp;
struct calib_Zpx_dom *		Zpx;
struct calib_Zx_obj **		zfactors;
struct calib_Zpx_obj **		zpfactors;
struct calib_Zpx_obj **		vlist;
mpz_srcptr			lc;
struct calib_Zx_obj *		zfact;
struct calib_Zpx_obj		HZp;
struct calib_Zpx_obj		FHAT;
struct calib_Zx_obj		PROD, H;
mpz_ptr				coeff, cj, modulus;
mpz_t				MOD_OVER_2, TEMP, LC_INV;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	cur_pow		= fip -> fi_a;
	if (cur_pow >= desired_pow) {
		/* Already at (or above) the requested power. */
		return;
	}

	poly		= fip -> fi_poly;
	Zp		= fip -> fi_Zp;
	Zpx		= fip -> fi_Zpx;
	num_fact	= fip -> fi_num_fact;
	zfactors	= fip -> fi_zfactors;
	zpfactors	= fip -> fi_zpfactors;
	vlist		= fip -> fi_vlist;
	modulus		= fip -> fi_modulus;

	deg	= poly -> degree;

	mpz_init (MOD_OVER_2);
	mpz_init (TEMP);
	mpz_init (LC_INV);

	/* Get the lead coefficient.  We have to treat this specially	*/
	/* because all of our modular factors are monic.		*/
	lc = &(poly -> coeff [deg]);

	/* Compute 1/lc (mod p). */
	Zp -> set_z (Zp, LC_INV, lc);
	Zp -> inv (Zp, LC_INV, LC_INV);

	Zx_init_degree (&PROD, deg);
	Zx_init (&H);

	/* Verify that modulus is what we think it is. */
	mpz_pow_ui (TEMP, Zp -> p, cur_pow);
	CALIB_FATAL_ERROR_IF (mpz_cmp (modulus, TEMP) NE 0);

	/* Make a local Zp[x] buffers for computing the lifing polys.	*/
	Zpx -> init (Zpx, &FHAT);

	for (k = cur_pow; k < desired_pow; k++) {
		/* Compute:						*/
		/*							*/
		/*	     poly - lc * \prod_i zfactors [i]		*/
		/*	H = ----------------------------------		*/
		/*			modulus				*/
		/*							*/
		/* Using arithmetic in Z[x].  The division should	*/
		/* always be exact.					*/

		if (calib_Zx_settings.factor_trace_level >= 1) {
			printf ("Zx_lift_factors_linear: pow = %d\n", k);
		}
		PROD.degree = 0;
		mpz_set (&PROD.coeff [0], lc);
		for (i = 0; i < num_fact; i++) {
			Zx_mul (&PROD, &PROD, zfactors [i]);
		}
		Zx_sub (&H, poly, &PROD);
		for (i = 0; i <= H.degree; i++) {
			mpz_fdiv_qr (&H.coeff [i],
				     TEMP,
				     &H.coeff [i],
				     modulus);
			/* Verify that division is exact. */
			CALIB_FATAL_ERROR_IF (mpz_sgn (TEMP) NE 0);
		}

		/* Now convert Hzp = (H mod p), and then Hzp = Hzp/lc,	*/
		/* using arithmetic in Zp[x].				*/
		Zpx -> init (Zpx, &HZp);
		Zpx -> set_Zx (&HZp, &H);
		if (mpz_cmp_ui (LC_INV, 1) NE 0) {
			for (i = 0; i <= HZp.degree; i++) {
				Zp -> mul (Zp,
					   &(HZp.coeff [i]),
					   &(HZp.coeff [i]),
					   LC_INV);
			}
		}
		for (i = 0; i < num_fact; i++) {
			/* Compute FHAT_i = (v_i * HZp) mod f_i,	*/
			/* with  all arithmetic in Zp[x].		*/
			Zpx -> mul (&FHAT, &HZp, vlist [i]);
			Zpx -> div (NULL, &FHAT, &FHAT, zpfactors [i]);

			zfact = zfactors [i];
			CALIB_FATAL_ERROR_IF (FHAT.degree >= zfact -> degree);

			/* Lift zfactor[i]:	f_i = f_i + FHAT*modulus. */
			for (j = 0; j <= FHAT.degree; j++) {
				mpz_mul (TEMP, modulus, &FHAT.coeff [j]);
				mpz_add (&(zfact -> coeff [j]),
					 &(zfact -> coeff [j]),
					 TEMP);
			}
		}
		Zpx -> clear (&HZp);

		mpz_mul (modulus, modulus, Zp -> p);

		/* Convert coefficients > modulus/2 to be negative. */
		mpz_fdiv_q_2exp (MOD_OVER_2, modulus, 1);
		for (i = 0; i < num_fact; i++) {
			zfact = zfactors [i];
			fdeg	= zfact -> degree;
			coeff	= zfact -> coeff;
			for (j = 0, cj = coeff; j <= fdeg; j++, cj++) {
				if (mpz_cmp (cj, MOD_OVER_2) > 0) {
					mpz_sub (cj, cj, modulus);
					CALIB_FATAL_ERROR_IF (mpz_sgn (cj) >= 0);
				}
			}
		}
	}

	fip -> fi_a = desired_pow;

	Zpx -> clear (&FHAT);

	Zx_clear (&H);
	Zx_clear (&PROD);

	mpz_clear (LC_INV);
	mpz_clear (TEMP);
	mpz_clear (MOD_OVER_2);
}

/*
 * Use the Hensel Lemma (p-adic construction) to lift the given
 * factorization to a sufficiently large modulus = p^k > 2*cbound.
 *
 * We use quadratic lifting here -- modulus q = p^k increases to q^2,
 * doubling k each iteration.  For simplicity, We make no attempt to
 * avoid overshoot.
 */

	static
	void
Zx_lift_factors_quadratic (

struct FInfo *		fip,		/* IN/OUT: Factoring algorithm info */
int			desired_pow	/* IN/OUT: Desired power of p */
)
{
int				i, j, k, num_fact, deg, cur_pow;
const struct calib_Zx_obj *	poly;
const struct calib_Zp_dom *	Zp;
struct calib_Zx_obj **		zfactors;
struct calib_Zpx_obj **		vlist;
struct calib_Zx_obj **		zvlist;
mpz_srcptr			lc;
struct calib_Zx_obj *		zfact;
struct calib_Zx_obj *		omit;
struct calib_Zx_obj		PROD, H, FHAT, REM;
mpz_ptr				q, lcinv;
mpz_t				MOD_OVER_2, TEMP, L1, L2, Q2, RESID, A, B;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	cur_pow		= fip -> fi_a;
	if (cur_pow >= desired_pow) {
		/* Already at (or above) the requested power. */
		return;
	}

	poly		= fip -> fi_poly;
	Zp		= fip -> fi_Zp;
	num_fact	= fip -> fi_num_fact;
	zfactors	= fip -> fi_zfactors;
	vlist		= fip -> fi_vlist;
	zvlist		= fip -> fi_zvlist;
	lcinv		= fip -> fi_lcinv;
	q		= fip -> fi_modulus;

	deg	= poly -> degree;

	mpz_init (MOD_OVER_2);
	mpz_init (TEMP);
	mpz_init (L1);
	mpz_init (L2);
	mpz_init (Q2);
	mpz_init (RESID);
	mpz_init (A);
	mpz_init (B);

	if (zvlist EQ NULL) {
		/* Make Z[x] versions of the Vi polys. */
		zvlist = CALIB_NEWA (num_fact, struct calib_Zx_obj *);
		for (i = 0; i < num_fact; i++) {
			zvlist [i] = Zx_cvZpx (vlist [i]);
		}
		fip -> fi_zvlist = zvlist;
	}

	/* Get the lead coefficient.  We have to treat this specially	*/
	/* because all of our modular factors are monic.		*/
	lc = &(poly -> coeff [deg]);

	if (cur_pow EQ 1) {
		/* Initialize lcinv = 1/lc (mod p). */
		Zp -> set_z (Zp, TEMP, lc);
		Zp -> inv (Zp, lcinv, TEMP);
	}

	Zx_init_degree (&PROD, deg);
	Zx_init (&H);
	Zx_init (&FHAT);
	Zx_init (&REM);

	/* Verify that q is what we think it is. */
	mpz_pow_ui (TEMP, Zp -> p, cur_pow);
	CALIB_FATAL_ERROR_IF (mpz_cmp (q, TEMP) NE 0);

	if (calib_Zx_settings.factor_trace_level >= 1) {
		printf ("Zx_lift_factors_quadratic: desired_pow = %d\n",
			desired_pow);
	}

	for (k = cur_pow; k < desired_pow; k <<= 1) {
		/* Compute:						*/
		/*							*/
		/*	     poly - lc * \prod_i zfactors [i]		*/
		/*	H = ----------------------------------		*/
		/*			   q				*/
		/*							*/
		/* Using arithmetic in Z[x].  The division should	*/
		/* always be exact.					*/

		mpz_mul (Q2, q, q);
		mpz_fdiv_q_2exp (MOD_OVER_2, q, 1);

		if (calib_Zx_settings.factor_trace_level >= 1) {
			printf ("Zx_lift_factors_quadratic: lifting from"
				" q = p**%d to Q2 = p**%d\n",
				k, 2*k);
		}

		mpz_fdiv_r (L2, lc, Q2);
		Zx_set_z (&PROD, L2);
		for (i = 0; i < num_fact; i++) {
			Zx_mul (&PROD, &PROD, zfactors [i]);
			Zx_mod_out_signed (&PROD, Q2);
		}
		Zx_sub (&H, poly, &PROD);
		Zx_mod_out_signed (&H, Q2);
		Zx_div_z_exact (&H, &H, q);

		/* Now multiply H by 1/lc (mod q) so that it		*/
		/* corresponds with the monic modular factors.  Make	*/
		/* coefficients be signed.				*/
		if (mpz_cmp_ui (lcinv, 1) NE 0) {
			Zx_mul_z (&H, &H, lcinv);
			Zx_mod_out_signed (&H, q);
		}
		for (i = 0; i < num_fact; i++) {
			/* Compute FHAT_i = (v_i * H) mod f_i,	*/
			/* with  all arithmetic in Zp[x].		*/
			Zx_mul (&FHAT, &H, zvlist [i]);
			Zx_mod_out_signed (&FHAT, q);
			zfact = zfactors [i];
#if 1
			Zx_remainder_monic_poly_mod_q (&FHAT, &FHAT, zfact, q);
#else
			Zx_div (NULL, &FHAT, D, &FHAT, zfact);

			/* Since divisor is monic, should have D=1. */
			CALIB_FATAL_ERROR_IF (mpz_cmp_ui (D, 1) NE 0);

			Zx_mod_out_signed (&FHAT, q);
#endif

			CALIB_FATAL_ERROR_IF (FHAT.degree >= zfact -> degree);

			/* Lift zfactor[i]:	f_i = f_i + FHAT*q. */
			Zx_mul_z (&FHAT, &FHAT, q);
			Zx_add (zfact, zfact, &FHAT);
			Zx_mod_out_signed (zfact, Q2);
		}

		/* Now lift the zvlist[i].  First compute the "omit"	*/
		/* polynomials mod q**2.				*/
		omit = Zx_compute_omit (fip, Q2);

		/* Compute the residual H. */
		Zx_set_si (&H, 1);
		for (i = 0; i < num_fact; i++) {
			Zx_mul (&FHAT, zvlist [i], &omit [i]);
			Zx_mod_out_signed (&FHAT, Q2);
			Zx_sub (&H, &H, &FHAT);
		}
		Zx_clear_array (omit, num_fact);
		free (omit);

		Zx_div_z_exact (&H, &H, q);
		Zx_mod_out_signed (&H, q);
		for (i = 0; i < num_fact; i++) {
			Zx_mul (&FHAT, &H, zvlist [i]);
			Zx_mod_out_signed (&FHAT, q);
#if 1
			zfact = zfactors [i];
			Zx_remainder_monic_poly_mod_q (&FHAT, &FHAT, zfact, q);
#else
			Zx_div (NULL, &FHAT, D, &FHAT, zfactors [i]);
			CALIB_FATAL_ERROR_IF (mpz_cmp_ui (D, 1) NE 0);
			Zx_mod_out_signed (&FHAT, q);
#endif
			Zx_mul_z (&FHAT, &FHAT, q);
			Zx_add (zvlist [i], zvlist [i], &FHAT);
			Zx_mod_out_signed (zvlist [i], Q2);
		}

		/* Now lift lcinv.  Let V = lcinv.  We have		*/
		/* V*L1 = 1 (mod q), and we want			*/
		/*	(V + A*q)*(L1 + B*q) = 1 (moq q2)		*/
		/* ==> V*L1 + (A*L1 + B*V)*q + A*B*q2 = 1 (mod q2)	*/
		/* ==> A*L1 + b*V = (1 - V*L1)/q (mod q).		*/
		/*  Compute the residual (1 - V*L1)/q (mod q).		*/
		mpz_fdiv_r (L1, L2, q);
		mpz_mul (RESID, L1, lcinv);
		mpz_ui_sub (RESID, 1, RESID);
		mpz_divexact (RESID, RESID, q);
		mpz_fdiv_r (RESID, RESID, q);
		/* Need to solve A*L1 + B*lcinv = RESID (mod q).	*/
		/* Fortunately, we can compute B directly:		*/
		/*	B = (L2 - L1) / q (mod q).			*/
		mpz_sub (B, L2, L1);
		mpz_divexact (B, B, q);
		/* Now							*/
		/*		A*L1 + B*lcinv = RESID			*/
		/* ==>		A = inv(L1)*(RESID - B*lcinv).		*/
		mpz_mul (A, B, lcinv);
		mpz_sub (A, RESID, A);
		j = mpz_invert (TEMP, L1, q);
		CALIB_FATAL_ERROR_IF (j EQ 0);
		mpz_mul (A, A, TEMP);
		mpz_fdiv_r (A, A, q);
		mpz_mul (A, A, q);
		mpz_add (lcinv, lcinv, A);
		mpz_fdiv_r (lcinv, lcinv, Q2);
		mpz_add (TEMP, lcinv, lcinv);
		if (mpz_cmp (TEMP, Q2) >= 0) {
			mpz_sub (lcinv, lcinv, Q2);
		}
		/* Verify lifted lcinv. */
		mpz_mul (TEMP, lc, lcinv);
		mpz_fdiv_r (TEMP, TEMP, Q2);
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (TEMP, 1) NE 0);

		mpz_set (q, Q2);
	}

	fip -> fi_a = k;

	Zx_clear (&REM);
	Zx_clear (&FHAT);
	Zx_clear (&H);
	Zx_clear (&PROD);

	mpz_clear (B);
	mpz_clear (A);
	mpz_clear (RESID);
	mpz_clear (Q2);
	mpz_clear (L2);
	mpz_clear (L1);
	mpz_clear (TEMP);
	mpz_clear (MOD_OVER_2);
}

/*
 * Compute the remainder of f(x) / g(x) (in Zq[x]) where g(x) is monic.
 * It is very important to keep successive remainders reduced modulo q here.
 */

	static
	void
Zx_remainder_monic_poly_mod_q (

struct calib_Zx_obj *		remainder,	/* IN/OUT: Remainder */
const struct calib_Zx_obj *	fpoly,		/* IN: dividend poly f(x) */
const struct calib_Zx_obj *	gpoly,		/* IN: divisor  poly g(x) */
mpz_srcptr			q		/* IN: coefficient modulus */
)
{
int				i, fdeg, gdeg, rdeg;
mpz_ptr				zp1, zp2, rlc;
mpz_t				HALF_Q, TEMP;
struct calib_Zx_obj		REM;

	CALIB_FATAL_ERROR_IF ((remainder EQ NULL) OR
			      (fpoly	 EQ NULL) OR
			      (gpoly	 EQ NULL) OR
			      (q	 EQ NULL));

	fdeg	= fpoly -> degree;
	gdeg	= gpoly -> degree;

	/* g(x) must be monic. */
	CALIB_FATAL_ERROR_IF (mpz_cmp_ui (&(gpoly -> coeff [gdeg]), 1) NE 0);

	mpz_init (HALF_Q);
	mpz_init (TEMP);

	mpz_fdiv_q_2exp (HALF_Q, q, 1);

	Zx_init_degree (&REM, fdeg);
	Zx_set (&REM, fpoly);

	rdeg = REM.degree;
	for (;;) {
		if (rdeg < gdeg) {
			/* We are done. */
			break;
		}
		zp1 = &REM.coeff [rdeg];
		zp2 = &(gpoly -> coeff [gdeg]);

		rlc = zp1;
		--zp1;
		--zp2;
		for (i = 0; i < gdeg; i++) {
			mpz_mul (TEMP, rlc, zp2);
			mpz_sub (TEMP, zp1, TEMP);
			mpz_fdiv_r (TEMP, TEMP, q);
			if (mpz_cmp (TEMP, HALF_Q) >= 0) {
				mpz_sub (zp1, TEMP, q);
			}
			else {
				mpz_set (zp1, TEMP);
			}
			--zp1;
			--zp2;
		}
		mpz_set_ui (rlc, 0);
		--rdeg;
		while ((rdeg >= 0) AND (mpz_sgn (&REM.coeff [rdeg]) EQ 0)) {
			--rdeg;
		}
	}
	REM.degree = rdeg;

	Zx_set (remainder, &REM);

	Zx_clear (&REM);
	mpz_clear (TEMP);
	mpz_clear (HALF_Q);
}

/*
 * Validate the lifted modular factors.
 */

	static
	void
Zx_validate_lifted_modular_factors (

struct FInfo *		fip		/* IN: Factoring algorithm info */
)
{
int				i, num_fact, deg;
const struct calib_Zx_obj *	poly;
struct calib_Zx_obj **		zfactors;
mpz_ptr				coeff;
mpz_srcptr			modulus, lc;
struct calib_Zx_obj		PROD, H;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	poly		= fip -> fi_poly;
	num_fact	= fip -> fi_num_fact;
	zfactors	= fip -> fi_zfactors;
	modulus		= fip -> fi_modulus;

	deg		= poly -> degree;

	CALIB_FATAL_ERROR_IF (deg <= 1);

	lc		= &(poly -> coeff [deg]);

	Zx_init_degree (&PROD, deg);
	Zx_init (&H);

	Zx_set_z (&PROD, lc);
	for (i = 0; i < num_fact; i++) {
		Zx_mul (&PROD, &PROD, zfactors [i]);
	}

	Zx_sub (&H, poly, &PROD);

	for (i = H.degree; i >= 0; i--) {
		coeff = &H.coeff [i];
		mpz_fdiv_r (coeff, coeff, modulus);
		CALIB_FATAL_ERROR_IF (mpz_sgn (coeff) NE 0);
		H.degree = i - 1;
	}

	Zx_clear (&H);
	Zx_clear (&PROD);
}

/*
 * Trial division: for each subset S of the factors (ordered by
 * increasing cardinality from 1 to num_fact/2), compute
 *
 *	A = lc(poly) * prod_{f \in S} f		(mod MODULUS),
 *
 *	B = convert_to_Zx (A),
 *
 *	C = prim_part (B),
 *
 * and then test if C(x) evenly divides poly(x).  If so, then
 * C is an irreducible factor of poly, which should be removed,
 * and S should be removed from the set of modular factors.
 */

	static
	void
Zx_trial_division (

struct FInfo *		fip,		/* IN/OUT: Factoring algorithm info */
int			card_limit	/* IN: maximum cardinality to test */
)
{
int				i, j, k;
int				nleft, cardinality;
struct calib_Zx_obj *		lpoly;
calib_bool *			taken;
int *				findex;
mpz_ptr				tc_stack, lc_tc;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	lpoly		= &(fip -> fi_lpoly);
	nleft		= fip -> fi_nleft;
	taken		= fip -> fi_taken;
	findex		= fip -> fi_findex;
	tc_stack	= fip -> fi_tc_stack;
	lc_tc		= fip -> fi_lc_tc;

	/* Put lead coefficient in first element of stack.		*/
	mpz_set (&tc_stack [0], &(lpoly -> coeff [lpoly -> degree]));

	for (cardinality = 1; ; cardinality++) {
		if (2 * cardinality > nleft) {
			/* For subsets this big, the complementary	*/
			/* subset of factors will have already been	*/
			/* tested (unsuccessfully) as divisors of	*/
			/* lpoly, so we are done.  The remaining lpoly	*/
			/* is the remaining irreducible factor.		*/
			break;
		}
		if (cardinality > card_limit) {
			/* Caller wants us to stop enumerating subsets.	*/
			/* We do *not* know if the remaining lpoly is	*/
			/* irreducible or not.  But we do know that all	*/
			/* irreducible Z[x] factors consisting of	*/
			/* card_limit or fewer modular factors have	*/
			/* been found.					*/
			break;
		}

		/* Compute lead_coeff * trail_coeff for fast		*/
		/* divisibility check.					*/
		mpz_mul (lc_tc,
			 &(lpoly -> coeff [lpoly -> degree]),
			 &(lpoly -> coeff [0]));

		/* Test all subsets of this cardinality. */
		Zx_trial_recurse (fip,
				  0,		/* pos in findex to consider */
				  0,		/* num factors selected */
				  cardinality);

		/* Remove all factors from findex that have been taken. */
		j = 0;
		for (i = 0; i < nleft; i++) {
			k = findex [i];
			if (taken [k]) continue;
			findex [j++] = k;
		}
		nleft = j;
		fip -> fi_nleft = nleft;
	}
}

/*
 * Recursive portion of trial division.
 */

	static
	int
Zx_trial_recurse (

struct FInfo *	fip,		/* IN/OUT: Factoring algorithm info */
int		idx,		/* IN: pos in findex to consider */
int		nchosen,	/* IN: number of factors chosen */
int		cardinality	/* IN: cardinality desired */
)
{
int			i, nleft, result, deg;
calib_bool *		taken;
calib_bool *		chosen;
int *			findex;
int *			comb;
struct calib_Zx_obj *	tfp;
struct calib_Zx_obj **	zfactors;
struct calib_Zx_obj *	lpoly;
mpz_ptr			tc_stack, modulus, lc;

	tc_stack = fip -> fi_tc_stack;

	if (nchosen >= cardinality) {
		CALIB_FATAL_ERROR_IF (nchosen > cardinality);

		/* Base case of recursion -- we have a subset of the	*/
		/* desired cardinality.					*/
		result = Zx_trial_check (fip,
					 cardinality,
					 &tc_stack [nchosen]);
		return (result);
	}
	/* nchosen < cardinality. */

	nleft	= fip -> fi_nleft;

	if (idx >= nleft) return (FALSE);
	if (nchosen + (nleft - idx) < cardinality) {
		/* There are not enough remaining factors to obtain	*/
		/* the desired cardinality, so terminate the recursion.	*/
		return (FALSE);
	}

	taken	= fip -> fi_taken;
	chosen	= fip -> fi_chosen;
	findex	= fip -> fi_findex;

	/* Consider including the next factor. */
	i = findex [idx];
	if (taken [i]) {
		/* This factor has already been taken.  Can't use it again. */
	}
	else {
		comb		= fip -> fi_comb;
		zfactors	= fip -> fi_zfactors;
		tc_stack	= fip -> fi_tc_stack;
		modulus		= fip -> fi_modulus;

		comb [nchosen]	= i;
		chosen [i]	= TRUE;

		/* Update tc_stack to reflect trailing coefficient of	*/
		/* this trial factor.					*/
		tfp = zfactors [i];
		mpz_mul (&tc_stack [nchosen + 1],
			 &tc_stack [nchosen],
			 &(tfp -> coeff [0]));
		mpz_fdiv_r (&tc_stack [nchosen + 1],
			    &tc_stack [nchosen + 1],
			    modulus);

		result = Zx_trial_recurse (fip,
					   idx + 1,
					   nchosen + 1,
					   cardinality);
		chosen [i] = FALSE;

		if (result) {
			/* A factor was removed, meaning that all	*/
			/* currently chosen factors are now out of	*/
			/* commision.  Back out of the recursion until	*/
			/* no factors are "chosen".			*/
			if (nchosen > 0) return (result);
		}
	}

	/* Update base of trailing-coefficient product stack to contain	*/
	/* the latest lc(lpoly).					*/
	lpoly	= &(fip -> fi_lpoly);
	deg	= lpoly -> degree;
	lc	= &(lpoly -> coeff [deg]);
	mpz_set (&tc_stack [0], lc);

	/* Now consider excluding the next factor. */
	result = Zx_trial_recurse (fip,
				   idx + 1,
				   nchosen,
				   cardinality);

	return (result);
}

/*
 * We have chosen a subset of the factors having the desired cardinality.
 * Multiply them out and see if they result in a true factor.
 */

	static
	int
Zx_trial_check (

struct FInfo *	fip,		/* IN/OUT: Factoring algorithm info */
int		cardinality,	/* IN: cardinality desired */
mpz_srcptr	tc_prod		/* IN: lead_coeff * product of all trial */
				/*     factor trail_coeff, mod modulus */
)
{
int				i, j, deg, result;
calib_bool			poly_is_signed_monic;
int *				comb;
calib_bool *			taken;
calib_bool *			chosen;
struct calib_Zx_obj *		lpoly;
struct calib_Zx_obj **		zfactors;
struct calib_Zx_obj *		fp;
struct calib_Zx_factor **	hookp;
struct calib_Zx_factor *	nfp;
struct calib_Zx_obj		PROD, QUOT;
mpz_srcptr			lc, lc_tc, modulus;
mpz_t				PROD0, TEMP;

	CALIB_FATAL_ERROR_IF (fip EQ NULL);

	lpoly		= &(fip -> fi_lpoly);
	modulus		= fip -> fi_modulus;
	lc_tc		= fip -> fi_lc_tc;

	/* Efficiency hack.  Test constant term first. */
	mpz_init (PROD0);
	mpz_init (TEMP);
	mpz_set (PROD0, tc_prod);
	mpz_mul_2exp (TEMP, PROD0, 1);
	if (mpz_cmp (TEMP, modulus) >= 0) {
		/* Negate constant coefficient. */
		mpz_sub (PROD0, PROD0, modulus);
	}

	deg		= lpoly -> degree;
	lc		= &(lpoly -> coeff [deg]);

	mpz_abs (TEMP, lc);
	poly_is_signed_monic = (mpz_cmp_ui (TEMP, 1) EQ 0);

	if (NOT poly_is_signed_monic) {
		/* Polynomial being factored is not monic.  We mimic	*/
		/* the full divisibility check below, which computes	*/
		/* the actual trial factor TF as follows:		*/
		/*							*/
		/*	TF = prim_part (signed_int (lc*\prod factors))	*/
		/*							*/
		/* and then testing if TF divides lpoly.  In this	*/
		/* context we do not have all coefficients of TF, only	*/
		/* the leading and trailing coefficients -- the GCD of	*/
		/* which will be >= to the content removed when		*/
		/* computing TF.  The adjusted trailing coefficient	*/
		/* is therefore <= tc(TF).				*/
		mpz_gcd (TEMP, lc, PROD0);
		mpz_divexact (TEMP, PROD0, TEMP);
		mpz_mul (PROD0, lc, &(lpoly -> coeff [0]));
		result = mpz_divisible_p (PROD0, TEMP);
	}
	else {
		result = mpz_divisible_p (lc_tc, PROD0);
	}
	if (result AND poly_is_signed_monic) {
		/* Now check the degree N-1 coefficient of this product. */
		mpz_t		TEMP2;
		int		nleft;
		nleft	 = fip -> fi_nleft;
		taken	 = fip -> fi_taken;
		chosen	 = fip -> fi_chosen;
		zfactors = fip -> fi_zfactors;
		mpz_init (TEMP2);
		mpz_set_ui (TEMP, 0);
		for (i = 0 ; i < nleft; i++) {
			if (taken [i]) continue;
			fp = zfactors [i];
			if (chosen [i]) {
				mpz_add (TEMP,
					 TEMP,
					 &(fp -> coeff [fp -> degree - 1]));
			}
			else {
				mpz_add (TEMP2,
					 TEMP2,
					 &(fp -> coeff [fp -> degree - 1]));
			}
		}
		mpz_mul (TEMP,  lc, TEMP);
		mpz_mul (TEMP2, lc, TEMP2);
		Zx_Zm_to_Z (TEMP,  modulus);
		Zx_Zm_to_Z (TEMP2, modulus);
		mpz_add (TEMP, TEMP, TEMP2);
		/* Adjust for when lpoly is - monic. */
		mpz_mul (PROD0, lc, &(lpoly -> coeff [deg - 1]));
		result = (mpz_cmp (PROD0, TEMP) EQ 0);
		mpz_clear (TEMP2);
	}

	mpz_clear (TEMP);
	mpz_clear (PROD0);

	if (NOT result) {
		return (result);
	}

	zfactors	= fip -> fi_zfactors;
	taken		= fip -> fi_taken;
	comb		= fip -> fi_comb;

	/* This combination looks promising.  Do the full polynomial.	*/
	Zx_init_degree (&PROD, deg - 1);
	Zx_init_degree (&QUOT, deg - 1);

	PROD.degree = 0;
	mpz_set (&PROD.coeff [0], lc);

	for (i = 0; i < cardinality; i++) {
		j = comb [i];
		fp = zfactors [j];
		Zx_mul (&PROD, &PROD, fp);
	}
	/* Mod out result by modulus, and convert to signed integer. */
	for (i = 0; i <= PROD.degree; i++) {
		Zx_Zm_to_Z (&PROD.coeff [i], modulus);
	}

	/* Take the primitive part. */
	Zx_prim_part (NULL, &PROD, &PROD);

	/* See if it divides evenly. */
	result = Zx_exactly_divides (&QUOT, lpoly, &PROD);

	if (result) {
		/* It divides out evenly!  Append a new factor to the list. */
		hookp = fip -> fi_hookp;
		nfp = CALIB_NEW (struct calib_Zx_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= Zx_dup (&PROD);
		nfp -> next		= NULL;
		*hookp = nfp;
		hookp = &(nfp -> next);
		fip -> fi_hookp = hookp;

		/* Remove this factor from the polynomial being factored. */
		Zx_set (lpoly, &QUOT);

		/* Mark all of the chosen factors as taken. */
		for (i = 0; i < cardinality; i++) {
			j = comb [i];
			taken [j] = TRUE;
		}
	}

	Zx_clear (&QUOT);
	Zx_clear (&PROD);

	return (result);
}

/*
 * Sort the final list of factors.
 *
 * It sorts the list destructively by re-rethreading the linked list
 * entries into the proper order.  The return value is the only
 * valid result from this function.
 */

	static
	struct calib_Zx_factor *
Zx_sort_factors (

struct calib_Zx_factor *	factors		/* IN: List of factors to sort */
)
{
int				i, j, k, n, nfact, cmp;
struct calib_Zx_factor *	fp;
struct calib_Zx_factor *	fp2;
struct calib_Zx_factor **	array;

	nfact = 0;
	for (fp = factors; fp NE NULL; fp = fp -> next) {
		++nfact;
	}

	array = CALIB_NEWA (nfact, struct calib_Zx_factor *);

	for (i = 0, fp = factors; fp NE NULL; i++, fp = fp -> next) {
		array [i] = fp;
	}
	CALIB_FATAL_ERROR_IF (i NE nfact);

	n = nfact;

	/* Construct the heap via sift-downs, in O(n) time. */
	for (k = n >> 1; k >= 0; k--) {
		j = k;
		for (;;) {
			i = (j << 1) + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the right subchild is greater. */
				fp  = array [i];
				fp2 = array [i + 1];
				cmp = Zx_compare_factors (fp, fp2);
				if (cmp < 0) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			fp  = array [j];
			fp2 = array [i];
			cmp = Zx_compare_factors (fp, fp2);
			if (cmp >= 0) {
				/* Greatest child is smaller.  Sift-	*/
				/* down is done. */
				break;
			}
			/* Sift down and continue. */
			array [j] = fp2;
			array [i] = fp;
			j = i;
		}
	}

	/* Now do actual sorting.  Exchange first/last and sift down. */
	while (n > 1) {
		/* Largest is at index [0], swap with index [n-1],	*/
		/* thereby putting it into final position.		*/
		--n;
		fp = array [0];
		array [0] = array [n];
		array [n] = fp;

		/* Now restore the heap by sifting array [0] down. */
		j = 0;
		for (;;) {
			i = (j << 1) + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the right subchild is greater. */
				fp  = array [i];
				fp2 = array [i + 1];
				cmp = Zx_compare_factors (fp, fp2);
				if (cmp < 0) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			fp  = array [j];
			fp2 = array [i];
			cmp = Zx_compare_factors (fp, fp2);
			if (cmp >= 0) {
				/* Greatest child is smaller.  Sift-	*/
				/* down is done. */
				break;
			}
			/* Sift down and continue. */
			array [j] = fp2;
			array [i] = fp;
			j = i;
		}
	}

	/* Re-thread the entries into a new list of factors. */
	factors = NULL;
	for (i = nfact - 1; i >= 0; i--) {
		fp = array [i];
		fp -> next = factors;
		factors = fp;
	}

	free (array);

	return (factors);
}

/*
 * Compare two factors lexicographically.
 */

	static
	int
Zx_compare_factors (

const struct calib_Zx_factor *	fp1,	/* IN: First factor */
const struct calib_Zx_factor *	fp2	/* IN: Second factor */
)
{
int				i, m1, m2, d1, d2, s1, s2, cmp;
const struct calib_Zx_obj *	poly1;
const struct calib_Zx_obj *	poly2;
mpz_srcptr			cp1, cp2;
mpz_t				fake1, fake2;

	CALIB_FATAL_ERROR_IF ((fp1 EQ NULL) OR (fp2 EQ NULL));
	CALIB_FATAL_ERROR_IF (fp1 EQ fp2);

	m1 = fp1 -> multiplicity;
	m2 = fp2 -> multiplicity;

	if (m1 < m2) return (-1);
	if (m1 > m2) return (+1);

	poly1	= fp1 -> factor;
	poly2	= fp2 -> factor;

	d1 = poly1 -> degree;
	d2 = poly2 -> degree;

	if (d1 < d2) return (-1);
	if (d1 > d2) return (+1);

	/* Same multiplicity, same degree.  Compare coefficients	*/
	/* lexicographically.						*/
	cp1 = poly1 -> coeff;
	cp2 = poly2 -> coeff;
	for (i = d1; i >= 0; i--) {
		/* Prepare fake1 and fake2 to be mpz_t's that are the	*/
		/* absolute values of cp1 [i] and cp2 [i], but sharing	*/
		/* the latter's storage.				*/
		fake1 [0] = cp1 [i];
		fake2 [0] = cp2 [i];
		if (fake1 [0]._mp_size < 0) {
			fake1 [0]._mp_size = - fake1 [0]._mp_size;
		}
		if (fake2 [0]._mp_size < 0) {
			fake2 [0]._mp_size = - fake2 [0]._mp_size;
		}
		cmp = mpz_cmp (fake1, fake2);

		/* Put smaller magnitude coefficients first. */
		if (cmp < 0) return (-1);
		if (cmp > 0) return (+1);

		s1 = mpz_sgn (&cp1 [i]);
		s2 = mpz_sgn (&cp2 [i]);

		/* Put positive coefficients before negative. */
		if (s1 > s2) return (-1);
		if (s1 < s2) return (+1);
	}

	/* Should never have two identical factors. */
	CALIB_FATAL_ERROR;

	return (0);
}

/*
 * Convert the given value (modulo the given modulus) to a signed integer.
 */

	static
	void
Zx_Zm_to_Z (

mpz_ptr		value,		/* IN/OUT: value to convert */
mpz_srcptr	modulus		/* IN: modulus */
)
{
mpz_t		TEMP;

	/* First mod out by the modulus. */
	mpz_fdiv_r (value, value, modulus);

	mpz_init (TEMP);
	mpz_mul_ui (TEMP, value, 2);

	if (mpz_cmp (TEMP, modulus) >= 0) {
		/* Value must be negated. */
		mpz_sub (value, value, modulus);
	}
	mpz_clear (TEMP);
}

/*
 * Free the given factorization, as produced by Zx_factor().
 */

	static
	void
Zx_free_factors (

struct calib_Zx_factor *	factors	/* IN: factors to recursively free */
)
{
struct calib_Zx_factor *	ftemp;
struct calib_Zx_obj *		poly;

	while (factors NE NULL) {
		ftemp = factors -> next;
		factors -> next = NULL;
		poly = factors -> factor;
		factors -> factor = NULL;
		Zx_free (poly);
		free (factors);
		factors = ftemp;
	}
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically zero.
 */

	static
	calib_bool
Zx_zerop (

const struct calib_Zx_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg >= 0) {
		return (FALSE);
	}

	CALIB_FATAL_ERROR_IF (deg < -1);

	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is identically one.
 */

	static
	calib_bool
Zx_onep (

const struct calib_Zx_obj *	op	/* IN: Polynomial to test */
)
{
int		deg;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	deg = op -> degree;
	if (deg NE 0) {
		return (FALSE);
	}

	if (mpz_sgn (&(op -> coeff [0])) EQ 0) {
		return (TRUE);
	}

	return (FALSE);
}

/*
 * Convert the given general expression into a member of Q[x].
 * All of the variables in the given expression *MUST* match the given
 * variable name.
 */

	static
	void
Zx_set_genrep (

struct calib_Zx_obj *		rop,	/* IN/OUT: resulting Zx polynomial */
const struct calib_genrep *	op,	/* IN: expression to convert */
const char *			var	/* IN: main var for polynomial */
)
{
int				i;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_Zx_obj		POLY;
struct calib_Zx_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (var EQ NULL));

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Zx_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Zx_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		/* Denominator *MUST* be 1!!! */
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (mpq_denref (op -> u.qop), 1) NE 0);
		Zx_set_z (rop, mpq_numref (op -> u.qop));
		break;

	case CALIB_GENREP_OP_VAR:
		CALIB_FATAL_ERROR_IF (strcmp (op -> u.var, var) NE 0);
		Zx_set_var_power (rop, 1);
		break;

	case CALIB_GENREP_OP_ADD:
		Zx_init (&POLY);
		Zx_init (&TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zx_set_genrep (&TEMP, lp -> operand, var);
			Zx_add (&POLY, &POLY, &TEMP);
		}
		Zx_set (rop, &POLY);

		Zx_clear (&TEMP);
		Zx_clear (&POLY);
		break;

	case CALIB_GENREP_OP_MUL:
		Zx_init_si (&POLY, 1);
		Zx_init (&TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zx_set_genrep (&TEMP, lp -> operand, var);
			Zx_mul (&POLY, &POLY, &TEMP);
		}
		Zx_set (rop, &POLY);

		Zx_clear (&TEMP);
		Zx_clear (&POLY);
		break;

	case CALIB_GENREP_OP_IPOW:
		i = op -> u.ipow.power;
		CALIB_FATAL_ERROR_IF (i < 0);

		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);
		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			/* Optimize important special case. */
			Zx_set_var_power (rop, i);
		}
		else {
			Zx_init (&TEMP);
			Zx_set_genrep (&TEMP, gp, var);
			Zx_ipow (rop, &TEMP, i);
			Zx_clear (&TEMP);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Convert the given element of Q[x] into general representation.
 * Use the variable name given.
 */

	static
	struct calib_genrep *
Zx_to_genrep (

const struct calib_Zx_obj *
			op,		/* IN: polynomial to convert */
const char *		var		/* IN: variable name to use */
)
{
int				i, degree;
mpz_srcptr			coeff;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		vgp;

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (var EQ NULL));

	degree	= op -> degree;
	coeff	= op -> coeff;

	root = NULL;
	hookp = &root;

	for (i = degree; i >= 0; i--) {
		if (mpz_sgn (&coeff [i]) EQ 0) continue;
		vgp = calib_genrep_z (&coeff [i]);
		gp1 = calib_genrep_poly_term (vgp, var, i);
		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= root;
	}

	return (gp1);
}

/*
 * Convert the given list of factors into general representation.
 * Use the variable given.
 */

	static
	struct calib_genrep *
Zx_factors_to_genrep (

const struct calib_Zx_factor *	factors,	/* IN: list of factors */
const char *			var		/* IN: variable to use */
)
{
int				pow;
const struct calib_Zx_factor *	fp;
struct calib_genrep_list *	root;
struct calib_genrep_list **	hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;

	CALIB_FATAL_ERROR_IF (var EQ NULL);

	root = NULL;
	hookp = &root;

	for (fp = factors; fp NE NULL; fp = fp -> next) {
		gp1 = Zx_to_genrep (fp -> factor, var);
		pow = fp -> multiplicity;
		CALIB_FATAL_ERROR_IF (pow <= 0);
		if (pow > 1) {
			gp2 = CALIB_NEW (struct calib_genrep);
			gp2 -> op		= CALIB_GENREP_OP_IPOW;
			gp2 -> u.ipow.base	= gp1;
			gp2 -> u.ipow.power	= pow;
			gp1 = gp2;
		}
		lp = calib_genrep_new_list (gp1, NULL);
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (1);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_MUL;
		gp1 -> u.list	= root;
	}
	return (gp1);
}

/*
 * Print out a single polynomial, in a syntax that maxima can read. 
 */

	static
	void
Zx_print (

const struct calib_Zx_obj *	op	/* IN: polynomial to print */
)
{
int		col;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	col = 0;
	col = Zx_print_nnl (op, col);
	printf ("\n");
}

/*
 * Print out an array of polynomials, in a syntax that maxima can read.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
Zx_print_array (

const struct calib_Zx_obj *
			array,		/* IN: array of polynomials to print */
int			npoly		/* IN: number of polynomials */
)
{
int		i, col;
char		buf [32];

	if (npoly <= 0) {
		CALIB_FATAL_ERROR_IF (npoly < 0);
		return;
	}

	CALIB_FATAL_ERROR_IF (array EQ NULL);

	for (i = 0; i < npoly; i++) {
		printf ("p%02d : ", i);
		col = 6;
		Zx_print_nnl (&array [i], col);
		printf ("$\n");
	}
	printf ("plist : [p00");
	col = 12;
	for (i = 1; i < npoly; i++) {
		if (col + 4 >= 78) {
			printf ("\n");
			col = 0;
		}
		sprintf (buf, ",p%02d", i);
		printf ("%s", buf);
		col += strlen (buf);
	}
	printf ("]$\n");
}

#pragma GCC diagnostic pop

/*
 * Print out a single polynomial in a syntax that maxima can read.
 */

	static
	int
Zx_print_nnl (

const struct calib_Zx_obj *
			op,	/* IN: polynomial to print */
int			col	/* IN: starting column number */
)
{
int		i, degree, sign, ndigits, width;
char *		string;
char *		sstr;
const char *	sep;
mpz_srcptr	zp;
char		powbuf [32];

	CALIB_FATAL_ERROR_IF ((op EQ NULL) OR (col < 0));

	if (op EQ NULL) {
		printf ("\"<null ptr>\"");
		return (col + 12);
	}
	degree = op -> degree;
	if (degree < 0) {
		printf ("0");
		return (col + 1);
	}
	zp = &(op -> coeff [degree]);
	for (i = degree; i >= 0; i--, zp--) {
		sign = mpz_sgn (zp);
		if (sign EQ 0) continue;
		ndigits = calib_genrep_convert_abs_Z_to_decimal_string (zp, &string);
		if (sign < 0) {
			sstr = (i EQ degree) ? "- " : " - ";
		}
		else {
			sstr = (i EQ degree) ? "" : " + ";
		}
		width = strlen (sstr);
		if ((i > 0) AND (ndigits EQ 1) AND (string [0] EQ '1')) {
			/* 1*var^pow ==> var^pow. */
			string [0] = '\0';
			ndigits = 0;
		}
		else {
			width += ndigits;
		}
		powbuf [0] = '\0';
		if (i > 0) {
			sep = (ndigits > 0) ? "*" : "";
			if (i EQ 1) {
				sprintf (powbuf, "%sx", sep);
			}
			else {
				sprintf (powbuf, "%sx^%d", sep, i);
			}
			width += strlen (powbuf);
		}
		if (col + width >= 78) {
			printf ("\n");
			col = 0;
		}
		printf ("%s%s%s", sstr, string, powbuf);
		free (string);
	}
	return (col);
}

/*
 * The ring Q[x].
 */

static const struct calib_Zx_dom	_Zx_dom = {
	.init			= Zx_init,
	.init_si		= Zx_init_si,
	.alloc			= Zx_alloc,
	.clear			= Zx_clear,
	.set			= Zx_set,
	.set_si			= Zx_set_si,
	.set_z			= Zx_set_z,
	.set_var_power		= Zx_set_var_power,
	.add			= Zx_add,
	.sub			= Zx_sub,
	.neg			= Zx_neg,
	.mul			= Zx_mul,
	.mul_z			= Zx_mul_z,
	.ipow			= Zx_ipow,
	.dup			= Zx_dup,
	.free			= Zx_free,
	.eval			= Zx_eval,
	.div			= Zx_div,
	.exactly_divides	= Zx_exactly_divides,
	.div_z_exact		= Zx_div_z_exact,
	.gcd			= Zx_gcd,
	.gcd_n			= Zx_gcd_n,
	.extgcd			= Zx_extgcd,
	.prim_part		= Zx_prim_part,
	.resultant		= Zx_resultant,
	.discriminant		= Zx_discriminant,
	.derivative		= Zx_derivative,
	.cvZpx			= Zx_cvZpx,
	.factor			= Zx_factor,
	.factor_square_free	= Zx_compute_square_free_factorization,
	.finish_sqf		= Zx_finish_sqf,
	.free_factors		= Zx_free_factors,
	.zerop			= Zx_zerop,
	.onep			= Zx_onep,
	.set_genrep		= Zx_set_genrep,
	.to_genrep		= Zx_to_genrep,
	.factors_to_genrep	= Zx_factors_to_genrep,
	.print_maxima		= Zx_print,
};

/*
 * Return the ring Z[x] of polynomials over the integers Z.
 */

	const struct calib_Zx_dom *
calib_get_Zx_dom (void)

{
	return (&_Zx_dom);
}
