/**********************************************************************

	$Id: Zxyz.c,v 1.1 2024/10/20 00:36:26 warme Exp $

	File:	Zxyz.c
	Rev:	a-2
	Date:	10/19/2024

	Copyright (c) 2012, 2024 by David M. Warme.  All rights
	reserved.  This work is licensed under a Creative Commons
	Attribution-NonCommercial-ShareAlike 4.0 International License.

***********************************************************************

	The ring Z[x,y,z] of multi-variate polynomials over the
	ring Z, of integers.

***********************************************************************

	Modification Log:

	a-1:	08/05/2012	warme
		: Created.
	a-2:	10/19/2024	warme
		: Cleanup for initial release.

***********************************************************************/

#include "calib/Zxyz.h"

#include "calib/fatal.h"
#include "calib/genrep.h"
#include "calib/gmpmisc.h"
#include "calib/logic.h"
#include "calib/new.h"
#include "calib/Qa.h"
#include "calib/Qax.h"
#include "calib/random.h"
#include "calib/types.h"
#include "calib/Zp.h"
#include "calib/Zpx.h"
#include "calib/Zx.h"
#include <gmp.h>
#include <limits.h>
#include "modcleanup.h"
#include "Qax_int.h"
#include "ratfac.h"
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

#define		PRINT_LIFTING_INFO		0
#define		TRACE_CFRECURSE			0

/* Some exceptional conditions to alert and/or check for. */
#define		SUCCESS			0
#define		BAD_EVAL_POINT		-1
#define		EXTRANEOUS_FACTORS	-2

/* Which application is using the EEZ algorithm. */
#define		MODE_EEZGCD		0
#define		MODE_FACTOR		1

/* Do we attempt to pre-etermine trailing coefficients also? */
#define		PREDETERMINE_TRAILING_COEFFICIENTS	1

/* Factorization mode. */

enum FactorMode {
	FMODE_FULL,		/* Factor down to irreducible polynomials */
	FMODE_SQUARE_FREE	/* Factor only down to square-free */
};

/*
 * This structure represents a block of memory that we can carve up into
 * fixed size nodes.
 */

#define	MBLK_TOTAL_SIZE		4096
#define MBLK_SIZE		(MBLK_TOTAL_SIZE - sizeof (struct mblk *))

struct mblk {
	struct mblk *	next;			/* Next block in chain. */
	char		bytes [MBLK_SIZE];	/* Usable block of memory */
};

/*
 * The following structure is used to represent a linked list of
 * polynomial terms.  In actual use, we allocate these with variable size,
 * to hold the correct exponent for all vars.
 */

struct pterm {
	mpz_t		coeff;		/* Constant coefficient for term */
	struct pterm *	next;		/* Next term in linked list */
	int		pow [1];	/* Array of variable exponents */
};

#define PTERM_SIZE(NVARS) \
	(offsetof (struct pterm, pow [0]) + (NVARS) * sizeof (int))

/*
 * A hash table that we use as an "accumulator" for polynomial terms
 * (represented via "struct pterm" nodes).  The hash table "key" that we
 * search on is the "pow" array of variable exponents in these nodes.
 *
 * We need to be able to dynamically:
 *	- Find a given term (exponent list), and update the coefficient,
 *	- Add a new term if missing,
 *	- Remove an existing term if its coefficient becomes zero
 *	  (via arithmetic cancellation).
 *
 * For the sake of efficiency, we want the hash table to have lots of
 * buckets (to minimize collisions), but at the same time we want to be
 * able to efficiently iterate over all items in the hash table (even if
 * the table is empty or has just one term in it).  This means we maintain
 * a doubly-linked list of the non-empty buckets, which we update in O(1)
 * time on each empty-to-non-empty or non-empty-to-empty transition.
 * Clearing out the hash table when we are done with it also uses this
 * linked list so that we only clear the non-empty buckets.
 *
 * Finally, we maintain a free-list of these accumulator objects, each
 * of which is completely initialized (to the empty) state, and ready to
 * go.  We clear each accumulator object before putting it back onto this
 * freelist.
 *
 * The individual "pterm" nodes that reside in the hash table are allocated
 * and freed from a freelist that is local to each hash table.  When empty,
 * the freelist is populated from nodes allocated from "struct mblk" objects.
 * These mblk's are allocated/freed using a global freelist -- but once a
 * hash table grabs an mblk, that entire mblk is completely reserved for
 * use by that hash table.  (There is no sharing of one mblk by several
 * hash tables.)  This means that a hash table can clear itself quickly
 * by (1) NULL-ing out all non-empty buckets, and (2) returning all of its
 * mblks to the global freelist.
 */

#define PACCUM_HASH_SIZE	10009

struct paccum {
	const struct calib_Zxyz_dom *
			dom;		/* Domain of terms we contain */
	int		nterms;		/* Current number of terms */
	struct pterm **	bucket;		/* Array buckets, each containing */
					/* a linked-list of pterm nodes */
	struct dlink *	link;		/* Array of forward/backward links, */
					/* one per bucket, plus a "root" */
					/* entry at [-1]. */
	struct pterm *	freelist;	/* Free list of term nodes */
	char *		allocp;		/* Allocation pointer for new terms */
	int		nleft;		/* Number of bytes of allocp left */
	struct mblk *	blocks;		/* Linked list of memory blocks */
	struct paccum *	next;		/* Next paccum in freelist */
};

/* The doubly-linked list entry. */

struct dlink {
	short		forw;		/* Forward index */
	short		back;		/* Backward index */
};

/*
 * Recursive form of multi-variate polynomials.  We lexically sort the
 * variables so that if var1 > var2, then var1 is "higher up" in the
 * recursive representation -- i.e., polynomials in var1 have coefficients
 * that are polynomials that may contain var2, but NOT vice-versa.
 *
 * Note that the "var" fields here may contain a *DIFFERENT* numbering of
 * the variables than the order given in the Zxyz_dom object!!!  We use
 * separate arrays to map between these distinct variable orderings
 * (recursive to flat, and flat to recursive).
 */

struct RPoly {
	int			var;	/* Var (or -1, ==> coefficient) */
	union {
		mpz_ptr		zval;	/* Coefficient value (var EQ -1) */
		struct RTerm *	terms;	/* Terms for this var (var >= 0) */
		struct RPoly *	next;	/* For forming freelists of these */
	} u;
};

struct RTerm {
	struct RPoly *		coeff;	/* Coefficient polynomial */
	struct RTerm *		next;	/* Next term */
	int			pow;	/* Exponent for this term */
};

/*
 * A special pointer value that means "division was not exact."
 */

#define	DIVISION_NOT_EXACT	((struct RPoly *) 1)

/*
 * A memory pool for allocating recursive polynomial structures
 * (both RPoly and RTerm).
 */

struct RPool {
	struct RPoly *	pfree;		/* Freelist of RPoly nodes */
	struct RTerm *	tfree;		/* Freelist of RTerm nodes */
	char *		allocp;		/* Allocation pointer for new nodes */
	int		nleft;		/* Number of bytes of allocp left */
	struct mblk *	blocks;		/* Linked list of memory blocks */
};

/*
 * An object to encapsulate generation of evaluation points.
 */

struct EvalPtGen {
	int		tries;	/* Number of tries at this nbytes */
	int		nbytes;	/* Number of bytes long to make numbers */
	int		nvars;	/* Total number of vars */
	int		nv;	/* Num vars being evaluated */
	const int *	r2fmap;	/* Which vars being evaluated */
	struct calib_Random *
			rndgen;	/* Random number generator */
};

/*
 * A structure to cache the complete factorization of the leading and/or
 * trailing coefficients of a polynomial (with respect to some main variable).
 *
 * Note that the "valid" flags are needed since the correct lcf and/or tcf
 * values can be NULL if the corresponding symbolic coefficient is 1.
 */

struct CFCache {			/* Coefficient Factor Cache */
	calib_bool		lvalid;	/* If true, lcf is valid */
	calib_bool		tvalid;	/* If true, tcf is valid */
	struct calib_Zxyz_factor *
				lcf;	/* Factors of leading coefficient */
	struct calib_Zxyz_factor *
				tcf;	/* Factors of trailing coefficient */
};

/*
 * Information that we cache about a single variable during the EEZ algorithm.
 * We maintain an array of these that is simply indexed by variable number
 * (i.e., independent of lifting order).
 */

struct EEZvar {
	int		maxdeg;		/* Max degree of this var, globally */
	mpz_srcptr	bval;		/* Value substituted for this var */
	mpz_ptr		pvalpows;	/* bval^k for k \in [0,maxdeg] */
	mpz_ptr		mvalpows;	/* (-bval)^k, for k \in [0,maxdeg] */
	struct calib_Zxyz_obj *
			mon_pows;	/* (var-bval)^k, for k \in [0,maxdeg] */
					/* These start off at 0, and are */
					/* computed on demand */
};

/*
 * Global EEZ algorithm info.
 */

struct EEZglob {
	struct GInfo *	gip;		/* Global info */
	const struct calib_Zxyz_dom *
			dom;		/* Polynomial domain */
	int		mode;		/* MODE_EEZGCD or MODE_FACTOR */
	int		nvars;		/* Total number of vars in domain */
					/* (whether actually present or not) */
	int		x1;		/* The main variable */
	int		nv;		/* Number of vars evaluated away */
	int *		varlist;	/* Vars in lifting order */
	const int *	r2fmap;		/* Recursive to flat var mapping. */
	const calib_bool *
			evflags;	/* Which vars evaluated away */
	mpz_srcptr	bvals;		/* Values substituted for each var */
	int		pg_tries;	/* Num evaluation pts tried so far */
	int		pg_nbytes;	/* Size of numbers to pick */
	const struct calib_Zxyz_obj *
			poly;		/* The polynomial being factored */
	struct calib_Zxyz_obj *
			poly2;		/* Modifiable local copy of poly */
	int		nf;		/* Number of factors being lifted */
	int		total_degree;	/* Total degree of poly */
	struct calib_Zxyz_obj *
			factors;	/* The factors to be lifted (each */
					/* of degree >= 1 in variable x1) */
	struct calib_Zxyz_obj *
			pdfactors;	/* The factors[], but with leading, */
					/* trailing, and perhaps other */
					/* coefficients predetermine, and */
					/* perhaps constants shifted between */
					/* the factors to make this happen */
	struct calib_Zxyz_obj *
			lfactors;	/* These are just pdfactors[], but */
					/* with all the variables we haven't */
					/* lifted yet (e.g., from pre- */
					/* determined coefficients) */
					/* evaluated away */
	struct calib_Zxyz_obj *
			l2factors;	/* These are just lfactors[], but */
					/* with the current variable we  are */
					/* lifting *also* evaluated away */
	struct calib_Zxyz_obj *
			ypolys;		/* ypolys[i]=prod_{j<>i}l2factors[j] */
	struct calib_Zxyz_obj *
			evpoly;		/* Successively evaluated versions */
					/* of poly.  evpoly[0] is univariate */
					/* evpoly[nv] is original multi- */
					/* variate poly */
	struct calib_Zxyz_obj *
			f0fact;		/* Univariate verion of factors[] */
	struct calib_Zx_obj *
			zx_f0fact;	/* f0fact[] in Z[x] form */
	struct calib_Zx_obj *
			zx_y0;		/* y0[i] = prod_{j<>i} zx_f0fact[j] */
	struct calib_Zx_obj *
			zx_cfact;	/* Z[x] correction factors */
	struct calib_Zx_obj *
			zx_f0bar;	/* Monic versions of zx_f0fact */
	const struct calib_Zxyz_obj *
			upoly;		/* Univariate version of poly. */
	struct calib_Zxyz_obj *
			cfact;		/* Correction factors */
	struct calib_Zxyz_obj *
			c2fact;		/* Corrections (of some total degree */
					/* m) for each cfact[] */
	struct calib_Zxyz_obj *
			Dm_stack;	/* Stack of Dm residual values */
	struct calib_Zxyz_obj *
			deriv_stack;	/* Stack of derivatives */
	struct calib_Zxyz_obj *
			prodstack;	/* Stack of monomial products, one */
					/* for each lifted var, plus base */
					/* case of 1 */
	struct calib_Zx_obj *
			cf_zx_temp;	/* Temp Z[x] polys for use by */
					/* cfact_recurse() */
	struct calib_Zxyz_obj
			cf_zxyz_temp;	/* Temp Z[x,y,z] poly for use by */
					/* cfact_recurse() */
	mpz_t		cf_z_temp;	/* Temp for use by cfact_recurse() */
	mpz_ptr		invfactorial;	/* Precomputed values of */
					/* 1/n! (mod modulus) */
	int		max_fact;	/* invfactorial[max_fact] is the */
					/* last invfactorial[] entry */
					/* currently available */
	mpz_t		fact_mod;	/* Modulus used to compute the */
					/* current invfactorial[] table */
	struct EEZvar *	var_info;	/* Info for each variable */
	struct calib_Zp_dom *
			Zp;		/* The domain of integers modulo p */
	struct calib_Zpx_dom *
			Zpx;		/* The domain Zp[x] */
	mpz_t		B;		/* The coefficient bound */
	mpz_t		p;		/* Prime modulus base */
	int		p_pow_k;	/* This is k, where modulus = p^k */
	mpz_t		modulus;	/* modulus being used (p^k) */
	unsigned long	num_cf_terms;	/* Number of Taylor series expansion */
					/* terms in correction factor */
};

/*
 * Info for use by algorithms that choose random integer lattice points
 * (random evaluation points).
 */

struct evstate {
	int	tries;
	int	nbytes;
};

/*
 * Global information used in many places.
 */

struct GInfo {
	int			nbinom;	/* Num binomial rows computed */
	mpz_ptr *		binom;	/* Array of binomial coeff rows */
	struct RPool		pool;	/* Pool to alloc/free from */
	struct calib_Random	rng;	/* Random number generator */
};

/*
 * Local Functions
 */

static void	Zxyz_add (struct calib_Zxyz_obj *	rop,
			  const struct calib_Zxyz_obj *	op1,
			  const struct calib_Zxyz_obj *	op2);
static void	Zxyz_add_n (struct calib_Zxyz_obj *		rop,
			    int					npoly,
			    const struct calib_Zxyz_obj **	poly_ptrs);
static struct calib_Zxyz_dom *
		Zxyz_add_vars (const struct calib_Zxyz_dom *	dom,
			       int				nvars,
			       const char * const *		newvars);
static void	Zxyz_adjust_gcd_n_result_signs (
				struct calib_Zxyz_obj *		gcd,
				struct calib_Zxyz_obj **	cofact,
				int				npoly);
static void	Zxyz_alloc (struct calib_Zxyz_obj *	rop,
			    int				nterms);
static mpz_srcptr
		Zxyz_binomial_row (int n, struct GInfo * gip);
static void	Zxyz_choose_evaluation_point (mpz_ptr		bvals,
					      struct EEZglob *	eezgip);
static void	Zxyz_choose_dense_evaluation_point (
					mpz_ptr			bvals,
					int			n,
					struct evstate *	evs,
					struct GInfo *		gip);
static int	Zxyz_choose_gcd_cofactor_to_lift (
				struct calib_Zxyz_obj *		gcd,
				const struct calib_Zxyz_obj *	ugcd,
				const struct calib_Zx_obj *	zx_gcd,
				int				npoly,
				struct calib_Zxyz_obj **	poly_ptrs,
				const struct calib_Zxyz_obj *	upoly_array,
				const struct calib_Zxyz_obj *	ucf_array,
				const struct calib_Zx_obj *	zx_upoly_array,
				const struct calib_Zx_obj *	zx_ucf_array,
				struct CFCache *		cfcache,
				struct EEZglob *		eezgip);
static int *	Zxyz_choose_var_order (
				int *				nvars_out,
				int				npoly,
				int				nvars,
				const struct calib_Zxyz_obj **	array,
				const int *			mindeg,
				const int *			maxdeg);
static void	Zxyz_cleanup_handler (void);
static void	Zxyz_clear (struct calib_Zxyz_obj *	x);
static void	Zxyz_clear_array (struct calib_Zxyz_obj *	parray,
				  int				npoly);
static int	Zxyz_compare_pow (int		nvars,
				  const int *	pow1,
				  const int *	pow2);
static struct calib_Zxyz_factor *
		Zxyz_compute_square_free_factorization (
				const struct calib_Zxyz_obj *	poly,
				struct GInfo *			gip);
static void	Zxyz_convert_with_varmap (
				struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op,
				const int *			varmap);
static void	Zxyz_copy_from_Qax (struct calib_Zxyz_obj *		rop,
				    mpz_ptr				denom,
				    const struct calib_Qax_obj *	op,
				    int					xvar,
				    int					avar);
static void	Zxyz_copy_from_Zpx (struct calib_Zxyz_obj *		rop,
				    const struct calib_Zpx_obj *	op,
				    int					var);
static void	Zxyz_copy_from_Zx (struct calib_Zxyz_obj *	rop,
				   const struct calib_Zx_obj *	op,
				   int				var);
static void	Zxyz_copy_into_Qax (struct calib_Qax_obj *		rop,
				    const struct calib_Zxyz_obj *	op,
				    int					xvar,
				    int					avar);
static void	Zxyz_copy_into_superring (
				struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op);
static void	Zxyz_copy_into_Zpx (struct calib_Zpx_obj *		rop,
				    const struct calib_Zxyz_obj *	op,
				    int					var);
static void	Zxyz_copy_into_Zx (struct calib_Zx_obj *		rop,
				   const struct calib_Zxyz_obj *	op,
				   int					var);
static struct calib_Zx_obj *
		Zxyz_cv_array_to_Zx (int			npoly,
				     struct calib_Zxyz_obj **	array,
				     int			x1);
static void	Zxyz_cv_from_Zx (struct calib_Zxyz_obj *	result,
				 const struct calib_Zx_obj *	src,
				 int				x1);
static void	Zxyz_cv_poly_to_Zx_form (struct calib_Zx_obj *	 	result,
					 const struct calib_Zxyz_obj *	poly,
					 int			 	x1);
static void	Zxyz_cvZpxyz (struct calib_Zxyz_obj *		rop,
			      const struct calib_Zp_dom *	Zp,
			      const struct calib_Zxyz_obj *	op);
static int	Zxyz_degree_in_var (const struct calib_Zxyz_obj *	poly,
				    int					var);
static void	Zxyz_derivative (struct calib_Zxyz_obj *	result,
				 const struct calib_Zxyz_obj *	poly,
				 int				var);
static calib_bool
		Zxyz_determine_coeffs (
			struct calib_Zxyz_obj *			C,
			const struct calib_Zxyz_factor *	flist,
			mpz_srcptr				delta,
			mpz_srcptr				ucoeff,
			mpz_ptr					dpart,
			struct EEZglob *			eezgip);
static void	Zxyz_discriminant (struct calib_Zxyz_obj *		rop,
				   const struct calib_Zxyz_obj *	op,
				   int					var);
static void	Zxyz_div (struct calib_Zxyz_obj *	quotient,
			  struct calib_Zxyz_obj *	remainder,
			  struct calib_Zxyz_obj *	d,
			  const struct calib_Zxyz_obj *	a,
			  const struct calib_Zxyz_obj *	b,
			  int				var);
static void	Zxyz_div_exact (struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	op1,
				const struct calib_Zxyz_obj *	op2,
				struct GInfo *			gip);
static void	Zxyz_div_z_exact (struct calib_Zxyz_obj *	result,
				  const struct calib_Zxyz_obj *	poly,
				  mpz_srcptr			z);
static int	Zxyz_div_remove (struct calib_Zxyz_obj *	rop,
				 const struct calib_Zxyz_obj *	op1,
				 const struct calib_Zxyz_obj *	op2);
static void	Zxyz_drop_coeff (struct calib_Zxyz_obj * poly, int var, int k);
static struct calib_Zxyz_obj *
		Zxyz_dup (const struct calib_Zxyz_obj *	op);
static void	Zxyz_eez_cfact_recurse (struct EEZglob *	eezgip,
					int			nlift,
					const int *		maxdeg,
					int			varidx,
					struct calib_Zxyz_obj *	Dm_ptr,
					struct calib_Zxyz_obj *	deriv,
					struct calib_Zxyz_obj *	prod,
					int			deg,
					int			pow);
static void	Zxyz_eez_choose_random_prime (struct EEZglob * eezgip);
static void	Zxyz_eez_clear_global_info (struct EEZglob * eezgip);
static void	Zxyz_eez_clear_var_info (struct EEZvar * vip);
static calib_bool
		Zxyz_eez_compute_correction (
					struct EEZglob *		eezgip,
					int				nlift,
					const int *			maxdeg,
					const struct calib_Zxyz_obj *	C);
static void	Zxyz_eez_compute_evpoly (struct EEZglob * eezgip);
static void	Zxyz_eez_compute_f0fact (struct EEZglob * eezgip);
static void	Zxyz_eez_compute_inverse_factorials (struct EEZglob * eezgip);
static calib_bool
		Zxyz_eez_compute_modp_cfact (
				struct EEZglob *	eezgip,
				struct calib_Zpx_obj *	modp_cfact);
static void	Zxyz_eez_compute_modulus_info (struct EEZglob * eezgip);
static void	Zxyz_eez_compute_upper_bound (struct EEZglob * eezgip);
static void	Zxyz_eez_compute_var_info (struct EEZglob * eezgip);
static const struct calib_Zxyz_obj *
		Zxyz_eez_get_var_val_pow (struct EEZglob *	eezgip,
					  int			var,
					  int			m);
static void	Zxyz_eez_init_global_info (
				struct EEZglob *		eezgip,
				int				mode,
				struct GInfo *			gip,
				const struct calib_Zxyz_dom *	dom,
				int				nv,
				const int *			r2fmap,
				const calib_bool *		evflags,
				mpz_srcptr			bvals,
				int				nf);
static void	Zxyz_eez_lift_cfact (
				struct EEZglob *	eezgip,
				struct calib_Zpx_obj *	modp_cfact);
static int	Zxyz_eez_lift_factors (struct EEZglob * eezgip);
static int	Zxyz_eez_main (struct CFCache *		cfcache,
			       struct EEZglob *		eezgip);
static void	Zxyz_eezgcd (struct calib_Zxyz_obj *		gcd,
			     struct calib_Zxyz_obj *		af,
			     struct calib_Zxyz_obj *		bf,
			     const struct calib_Zxyz_obj *	a,
			     const struct calib_Zxyz_obj *	b);
static void	Zxyz_establish_cleanup_handler (void);
static void	Zxyz_eval (mpz_ptr				rop,
			   const struct calib_Zxyz_obj *	poly,
			   mpz_srcptr				values);
static void	Zxyz_eval_one_var (
				struct calib_Zxyz_obj *		result,
				const struct calib_Zxyz_obj *	poly,
				int				var,
				mpz_srcptr			value);
static void	Zxyz_eval_var_subset (
				struct calib_Zxyz_obj *		rop,
				const struct calib_Zxyz_obj *	poly,
				const calib_bool *		evflags,
				mpz_srcptr			values);
static void	Zxyz_extgcd (struct calib_Zxyz_obj *		gcd,
			     struct calib_Zxyz_obj *		xa,
			     struct calib_Zxyz_obj *		xb,
			     mpz_ptr				d,
			     const struct calib_Zxyz_obj *	a,
			     const struct calib_Zxyz_obj *	b);
static struct calib_Zxyz_factor *
		Zxyz_factor (const struct calib_Zxyz_obj * poly);
static void	Zxyz_factor_coefficient_upper_bound (
				mpz_ptr				B,
				const struct calib_Zxyz_obj *	poly);
static struct calib_Zxyz_factor *
		Zxyz_factor_poly_coeff (const struct calib_Zxyz_obj *	poly,
					int				x1,
					int				k,
					struct GInfo *			gip);
static struct calib_Zxyz_factor *
		Zxyz_factor_square_free (const struct calib_Zxyz_obj *	poly,
					 struct GInfo *			gip);
static struct calib_genrep *
		Zxyz_factors_to_genrep (
			const struct calib_Zxyz_factor *	factors);
static void	Zxyz_free (struct calib_Zxyz_obj *	poly);
static void	Zxyz_free_array (struct calib_Zxyz_obj * array, int npoly);
static void	Zxyz_free_factors (struct calib_Zxyz_factor * factors);
static void	Zxyz_free_Zpx_array (struct calib_Zpx_dom *	Zpx,
				     struct calib_Zpx_obj *	array,
				     int			n);
static void	Zxyz_free_Zx_array (struct calib_Zx_obj * array, int n);
static void	Zxyz_gcd (struct calib_Zxyz_obj *	gcd,
			  const struct calib_Zxyz_obj *	a,
			  const struct calib_Zxyz_obj *	b);
static void	Zxyz_gcd_n (struct calib_Zxyz_obj *		gcd,
			    struct calib_Zxyz_obj **		cofact,
			    int					npoly,
			    const struct calib_Zxyz_obj **	array);
static void	Zxyz_get_coeff (struct calib_Zxyz_obj *		result,
				const struct calib_Zxyz_obj *	poly,
				int				var,
				int				power);
static struct mblk *
		Zxyz_get_mblk (void);
static struct paccum *
		Zxyz_get_paccum (const struct calib_Zxyz_dom * dom);
static void	Zxyz_get_poly_array_degrees (int **			mindeg,
					     int **			maxdeg,
					     int			npolys,
					     int			nvars,
					     struct calib_Zxyz_obj **	array);
static void	Zxyz_get_poly_degrees (int **				mindeg,
				       int **				maxdeg,
				       const struct calib_Zxyz_obj *	poly);
static struct pterm *
		Zxyz_get_pterm (struct paccum * pap);
static struct pterm **
		Zxyz_get_sorted_terms (const struct paccum * pap);
static int	Zxyz_get_var_coeffs (struct calib_Zxyz_obj **		coeffs,
				     int				var,
				     const struct calib_Zxyz_obj *	p);
static void	Zxyz_get_var_pow_coeff (struct calib_Zxyz_obj *		rop,
					const struct calib_Zxyz_obj *	op,
					int				var,
					int				pow);
static void	Zxyz_ginfo_clear (struct GInfo * gip);
static void	Zxyz_ginfo_init (struct GInfo * gip);
static void	Zxyz_init (const struct calib_Zxyz_dom *	dom,
			   struct calib_Zxyz_obj *		x);
static void	Zxyz_init_array (const struct calib_Zxyz_dom *	dom,
				 struct calib_Zxyz_obj *	parray,
				 int				npoly);
static void	Zxyz_init_si (const struct calib_Zxyz_dom *	dom,
			      struct calib_Zxyz_obj *		x,
			      calib_si_t			op);
static void	Zxyz_int_content_ppart (
				struct calib_Zxyz_obj *		content,
				struct calib_Zxyz_obj *		ppart,
				const struct calib_Zxyz_obj *	poly,
				int				var,
				struct GInfo *			gip);
static void	Zxyz_int_eezgcd_prim_polys (
				struct calib_Zxyz_obj *		gcd,
				int				npoly,
				struct calib_Zxyz_obj **	array,
				int				nv,
				const int *			r2fmap,
				struct GInfo *			gip);
static void	Zxyz_internal_eezgcd_n (
				struct calib_Zxyz_obj *		gcd,
				struct calib_Zxyz_obj **	cofact,
				int				npoly,
				const struct calib_Zxyz_obj **	a,
				struct GInfo *		 	gip);
static struct calib_Zxyz_factor *
		Zxyz_internal_factor (
				const struct calib_Zxyz_obj *	poly,
				struct GInfo *			gip,
				enum FactorMode			factor_mode);
static void	Zxyz_ipow (struct calib_Zxyz_obj *		rop,
			   const struct calib_Zxyz_obj *	op,
			   int					power);
static calib_bool
		Zxyz_is_constant_poly (const struct calib_Zxyz_obj * poly);
static calib_bool
		Zxyz_is_one_poly (const struct calib_Zxyz_obj * poly);
static calib_bool
		Zxyz_is_non_zero_constant_poly (
				const struct calib_Zxyz_obj * poly);
static int	Zxyz_lookup_var (const struct calib_Zxyz_dom *	dom,
				 const char *			var);
static void	Zxyz_make_poly_of_var (
				int *				deg_out,
				struct calib_Zxyz_obj **	array_out,
				const struct calib_Zxyz_obj *	poly,
				int				var);
static struct calib_Zxyz_obj *
		Zxyz_map_to_subring (
				struct calib_Zxyz_obj *		result,
				const struct calib_Zxyz_obj *	poly,
				const struct calib_Zxyz_dom *	subring,
				const int *			varmap);
static void	Zxyz_mod_out (struct calib_Zxyz_obj *	poly,
			      mpz_srcptr		modulus);
static void	Zxyz_mod_out_Zx (struct calib_Zx_obj *	poly,
				 mpz_srcptr		modulus);
static void	Zxyz_mul (struct calib_Zxyz_obj *	rop,
			  const struct calib_Zxyz_obj *	op1,
			  const struct calib_Zxyz_obj *	op2);
static void	Zxyz_mul_var_pow_k (
				struct calib_Zxyz_obj *	poly,
				int				var,
				int				k);
static void	Zxyz_mul_z (struct calib_Zxyz_obj *		rop,
			    const struct calib_Zxyz_obj *	op1,
			    mpz_srcptr				op2);
static void	Zxyz_neg (struct calib_Zxyz_obj *	rop,
			  const struct calib_Zxyz_obj *	op);
static struct calib_Zxyz_obj *
		Zxyz_new_array (const struct calib_Zxyz_dom * dom, int npoly);
static struct calib_Zpx_obj *
		Zxyz_new_Zpx_array (struct calib_Zpx_dom * Zpx, int n);
static struct calib_Zx_obj *
		Zxyz_new_Zx_array (int n);
static void	Zxyz_nth_derivative (struct calib_Zxyz_obj *		result,
				     const struct calib_Zxyz_obj *	poly,
				     int				var,
				     int				n);
static int	Zxyz_num_vars_present (const struct calib_Zxyz_obj * poly);
static calib_bool
		Zxyz_onep (const struct calib_Zxyz_obj *	op);
static void	Zxyz_paccum_add_term (struct paccum *		pap,
				      mpz_srcptr		coeff,
				      const int *		pow);
static void	Zxyz_poly_from_coeffs (
				struct calib_Zxyz_obj *		result,
				int				degree,
				const struct calib_Zxyz_obj *	coeffs,
				int				var);
static void	Zxyz_prim_part (struct calib_Zxyz_obj *		content,
				struct calib_Zxyz_obj *		ppart,
				const struct calib_Zxyz_obj *	op,
				int				var);
static void	Zxyz_print (const struct calib_Zxyz_obj * op);
static void	Zxyz_print_nnl (const struct calib_Zxyz_obj * op);
static int	Zxyz_pterm_hash (int nvars, const int * pow);
static void	Zxyz_put_mblk (struct mblk * mp);
static void	Zxyz_put_paccum (struct paccum * pap);
static void	Zxyz_put_pterm (struct pterm * tp, struct paccum * pap);
static void	Zxyz_realloc_copy_poly (struct calib_Zxyz_obj *	poly,
					int			nterms,
					const int *		pow,
					mpz_srcptr		coeff);
static void	Zxyz_remove_v_contents (
				struct calib_Zxyz_obj *		vcont,
				int				npoly,
				int				nvars,
				struct calib_Zxyz_obj **	array,
				int *				mindeg,
				int *				maxdeg);
static void	Zxyz_remove_z_contents (mpz_ptr				zcont,
					int				npoly,
					struct calib_Zxyz_obj **	array);
static void	Zxyz_res_CRT_poly (
				struct calib_Zxyz_obj *		C,
				const struct calib_Zxyz_obj *	R,
				mpz_srcptr			MODULUS,
				calib_int64u			p);
static void	Zxyz_res_coeff_norm (mpz_ptr				cnorm,
				     const struct calib_Zxyz_obj *	poly,
				     int				var);
static void	Zxyz_res_eval_var_Zp (struct calib_Zxyz_obj *		result,
				      const struct calib_Zxyz_obj *	poly,
				      mpz_srcptr			value,
				      int				var,
				      const struct calib_Zp_dom *	Zp);
static void	Zxyz_res_mod_Zp (
				struct calib_Zxyz_obj *		result,
				const struct calib_Zxyz_obj *	poly,
				const struct calib_Zp_dom *	Zp);
static void	Zxyz_res_norm (mpz_ptr				norm,
			       const struct calib_Zxyz_obj *	poly);
static calib_int64u
		Zxyz_res_prev_prime (calib_int64u p);
static void	Zxyz_res_pseudo_div (
				struct calib_Zxyz_obj *		quotient,
				int *				qdeg_out,
				struct calib_Zxyz_obj *		remainder,
				int *				rdeg_out,
				struct calib_Zxyz_obj *		D_out,
				const struct calib_Zxyz_obj *	a_in,
				int				adeg,
				const struct calib_Zxyz_obj *	b_in,
				int				bdeg);
static void	Zxyz_res_recurse (int				i,
				  struct calib_Zxyz_obj *	C,
				  const struct calib_Zxyz_obj *	A,
				  const struct calib_Zxyz_obj *	B,
				  const struct calib_Zpx_dom *	Zpx,
				  calib_int64u			p,
				  int				var);
static void	Zxyz_reset_paccum (struct paccum * pap);
static void	Zxyz_resultant (struct calib_Zxyz_obj *		result,
				const struct calib_Zxyz_obj *	a,
				const struct calib_Zxyz_obj *	b,
				int				var);
static struct calib_Zxyz_factor *
		Zxyz_resultant_old (const struct calib_Zxyz_obj *	a,
				    const struct calib_Zxyz_obj *	b,
				    int					var);
static struct RPoly *
		Zxyz_rpoly_add (const struct RPoly *	a,
				const struct RPoly *	b,
				struct GInfo *		gip);
static struct RPoly *
		Zxyz_rpoly_cv_f2r (
				const struct calib_Zxyz_obj *	poly,
				int				nv,
				const int *			r2f_map,
				struct GInfo *			gip);
static void	Zxyz_rpoly_cv_r2f (struct calib_Zxyz_obj *	result,
				   const struct RPoly *		pp,
				   const int *			r2f_map);
static struct RPoly *
		Zxyz_rpoly_div_exact (const struct RPoly *	a,
				      const struct RPoly *	b,
				      struct GInfo *		gip);
static struct RPoly *
		Zxyz_rpoly_dup (const struct RPoly * pp, struct GInfo * gip);
static struct RPoly *
		Zxyz_rpoly_finish_terms (int		var,
					 struct RTerm *	terms,
					 struct GInfo *	gip);
static void	Zxyz_rpoly_free (struct RPoly * pp, struct GInfo * gip);
static struct RPoly *
		Zxyz_rpoly_mul (const struct RPoly *	a,
				const struct RPoly *	b,
				struct GInfo *		gip);
static struct RPoly *
		Zxyz_rpoly_mul_si (const struct RPoly *	a,
				   long			zval,
				   struct GInfo *	gip);
static struct RPoly *
		Zxyz_rpoly_mul_z (const struct RPoly *	a,
				  mpz_srcptr		zval,
				  struct GInfo *	gip);
static struct RPoly *
		Zxyz_rpoly_new_si (long zval, struct GInfo * gip);
static struct RPoly *
		Zxyz_rpoly_new_varpow (int var, int pow, struct GInfo * gip);
static struct RPoly *
		Zxyz_rpoly_new_z (mpz_srcptr zval, struct GInfo * gip);
static void	Zxyz_rpoly_print (const struct RPoly * pp);
static void	Zxyz_rpoly_print_recurse (const struct RPoly * pp);
static void	Zxyz_rpoly_r2f_recurse (struct paccum *		pap,
					int *			pow,
					calib_bool *		varseen,
					const int *		r2f_map,
					int			parent_var,
					const struct RPoly *	pp);
static struct RPoly *
		Zxyz_rpoly_sub (const struct RPoly *	a,
				const struct RPoly *	b,
				struct GInfo *		gip);
static void	Zxyz_rpool_clear (struct RPool * pool);
static void	Zxyz_rpool_free_rpoly (struct RPoly * pp, struct GInfo * gip);
static void	Zxyz_rpool_free_rterm (struct RTerm * tp, struct GInfo * gip);
static void	Zxyz_rpool_init (struct RPool * pool);
static struct RPoly *
		Zxyz_rpool_new_rpoly (struct GInfo * gip);
static struct RTerm *
		Zxyz_rpool_new_rterm (struct GInfo * gip);
static void	Zxyz_rterms_free (struct RTerm * tp, struct GInfo * gip);
static void	Zxyz_set (struct calib_Zxyz_obj *	rop,
			  const struct calib_Zxyz_obj *	op);
static void	Zxyz_set_genrep (struct calib_Zxyz_obj *	rop,
				 const struct calib_genrep *	op);
static void	Zxyz_set_poly_from_paccum (struct calib_Zxyz_obj *	 dst,
					   const struct paccum *	pap);
static void	Zxyz_set_si (struct calib_Zxyz_obj *	rop,
			     calib_si_t			op);
static void	Zxyz_set_var_power (struct calib_Zxyz_obj *	rop,
				    int				var,
				    int				power);
static void	Zxyz_set_z (struct calib_Zxyz_obj *	rop,
			    mpz_srcptr			op);
static struct calib_Zxyz_factor *
		Zxyz_sqf_factor (const struct calib_Zxyz_obj * poly);
static void	Zxyz_strip_z_content (mpz_ptr			zcont,
				      struct calib_Zxyz_obj *	poly);
static void	Zxyz_sub (struct calib_Zxyz_obj *	rop,
			  const struct calib_Zxyz_obj *	op1,
			  const struct calib_Zxyz_obj *	op2);
static struct calib_genrep *
		Zxyz_to_genrep (const struct calib_Zxyz_obj *	op);
static int	Zxyz_total_degree (const struct calib_Zxyz_obj * poly);
static void	Zxyz_z_content (mpz_ptr				zcont,
				const struct calib_Zxyz_obj *	op);
static calib_bool
		Zxyz_zerop (const struct calib_Zxyz_obj *	op);


/*
 * Local Variables
 */

static calib_bool	cleanup_handler_established;
static struct mblk *	mblk_freelist;
static struct paccum *	paccum_freelist;

/*
 * Initialize the given polynomial.  Its value after initialization is zero.
 */

	static
	inline
	void
Zxyz_init (

const struct calib_Zxyz_dom *	dom,	/* IN: polynomial domain */
struct calib_Zxyz_obj *		x	/* IN/OUT: polynomial to initialize */
)
{
	x -> nterms	= 0;
	x -> size	= 0;
	x -> dom	= dom;
	x -> pow	= NULL;
	x -> coeff	= NULL;
}

/*
 * Initialize the given polynomial to have the given constant value.
 */

	static
	void
Zxyz_init_si (

const struct calib_Zxyz_dom *	dom,	/* IN: polynomial domain */
struct calib_Zxyz_obj *		x,	/* IN/OUT: polynomial to initialize */
calib_si_t			op	/* IN: initial value */
)
{
mpz_ptr			zp;
int *			pow;

	if (op EQ 0) {
		Zxyz_init (dom, x);
	}
	else {
		pow = CALIB_NEWA (dom -> nvars, int);
		memset (pow, 0, dom -> nvars * sizeof (*pow));

		zp = calib_new_Z_vector (1);
		mpz_set_si (zp, op);

		x -> nterms	= 1;
		x -> size	= 1;
		x -> dom	= dom;
		x -> pow	= pow;
		x -> coeff	= zp;
	}
}

/*
 * Initialize the given array of polynomials to zero.
 */

	static
	inline
	void
Zxyz_init_array (

const struct calib_Zxyz_dom *	dom,	/* IN: polynomial domain */
struct calib_Zxyz_obj *		parray,	/* IN/OUT: array of polynomials */
int				npoly	/* IN: num polynomials in array */
)
{
int			i;
struct calib_Zxyz_obj *	pp;

	pp = &parray [0];
	for (i = 0; i < npoly; i++, pp++) {
		Zxyz_init (dom, pp);
	}
}

/*
 * Force the given polynomial to have buffer space sufficient to hold
 * a polynomial having at least the given number of terms.  This does not
 * change the value currently residing in this polynomial.
 */

	static
	void
Zxyz_alloc (

struct calib_Zxyz_obj *	rop,		/* IN/OUT: poly to alloc buffer for */
int			nterms		/* IN: number of terms desired */
)
{
int	size, dsize, dterms, nvars;
int *	spow;
int *	dpow;
mpz_ptr	zp, sp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (nterms < 0));

	size	= nterms;
	dterms	= rop -> nterms;
	dsize	= rop -> size;

	if (dsize < size) {
		nvars = rop -> dom -> nvars;
		spow = rop -> pow;
		dpow = CALIB_NEWA (size * nvars, int);
		if (dterms > 0) {
			memcpy (dpow, spow, dterms * nvars * sizeof (int));
		}

		sp = rop -> coeff;
		zp = calib_new_Z_vector (size);
		calib_copy_Z_vector (zp, sp, dsize);

		rop -> pow	= dpow;
		rop -> coeff	= zp;
		rop -> size	= size;

		calib_free_Z_vector (sp, dsize);
		free (spow);
	}
}

/*
 * Clear out the given polynomial.
 */

	static
	inline
	void
Zxyz_clear (

struct calib_Zxyz_obj *	x		/* IN/OUT: polynomial to clear */
)
{
int		size;
int *		pow;
mpz_ptr		coeff;

	CALIB_FATAL_ERROR_IF (x EQ NULL);

	size	= x -> size;
	pow	= x -> pow;
	coeff	= x -> coeff;

	x -> nterms	= 0;
	x -> size	= 0;
	x -> dom	= NULL;
	x -> pow	= NULL;
	x -> coeff	= NULL;

	if (pow NE NULL) {
		free (pow);
	}

	if (coeff NE NULL) {
		CALIB_FATAL_ERROR_IF (size <= 0);
		calib_free_Z_vector (coeff, size);
	}
}

/*
 * Clear out the given array of polynomials.
 */

	static
	inline
	void
Zxyz_clear_array (

struct calib_Zxyz_obj *	parray,	/* IN: array of polynomials to clear */
int			npoly	/* IN: number of polynomials in parray */
)
{
int		i;

	for (i = 0; i < npoly; i++) {
		Zxyz_clear (&parray [i]);
	}
}

/*
 * Dynamically allocate and initialize an array of polynomials.
 */

	static
	struct calib_Zxyz_obj *
Zxyz_new_array (

const struct calib_Zxyz_dom *	dom,	/* IN: polynomial domain */
int				npoly	/* IN: number of polynomials */
)
{
struct calib_Zxyz_obj *	array;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	array = NULL;
	if (npoly > 0) {
		array = CALIB_NEWA (npoly, struct calib_Zxyz_obj);
		Zxyz_init_array (dom, array, npoly);
	}

	return (array);
}

/*
 * Free up the given dynamically allocated array of polynomials.
 */

	static
	void
Zxyz_free_array (

struct calib_Zxyz_obj *	array,		/* IN: polynomial array */
int			npoly		/* IN: number of polynomials */
)
{
	if (array NE NULL) {
		CALIB_FATAL_ERROR_IF (npoly <= 0);
		Zxyz_clear_array (array, npoly);
		free (array);
	}
	else {
		CALIB_FATAL_ERROR_IF (npoly NE 0);
	}
}

/*
 * Set the given destination polynomial from the given power/coefficient
 * arrays.  Reallocate the destination polynomial's coefficient array,
 * if necessary.
 */

	static
	inline
	void
Zxyz_realloc_copy_poly (

struct calib_Zxyz_obj *	poly,		/* IN: destination polynomial */
int			nterms,		/* IN: number of terms to set */
const int *		pow,		/* IN: source power array */
mpz_srcptr		coeff		/* IN: source coefficient array */
)
{
int			cur_size, new_size, nvars;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (nterms < 0));

	poly -> nterms	= nterms;

	if (nterms <= 0) {
		/* We allow (coeff EQ NULL) AND/OR (pow EQ NULL) here. */
		return;
	}

	CALIB_FATAL_ERROR_IF ((pow EQ NULL) OR (coeff EQ NULL));

	new_size = nterms;
	if (new_size > 0) {
		nvars = poly -> dom -> nvars;
		cur_size = poly -> size;
		if (new_size > cur_size) {
			if (poly -> pow NE NULL) {
				free (poly -> pow);
			}
			if (poly -> coeff NE NULL) {
				calib_free_Z_vector (poly -> coeff, cur_size);
			}
			poly -> pow	= CALIB_NEWA (new_size * nvars, int);
			poly -> coeff	= calib_new_Z_vector (new_size);
			poly -> size	= new_size;
		}
		if (pow NE poly -> pow) {
			memcpy (poly -> pow, pow, new_size * nvars * sizeof (int));
		}
		calib_copy_Z_vector (poly -> coeff, coeff, new_size);
	}
}

/*
 * Copy the given polynomial.
 */

	static
	void
Zxyz_set (

struct calib_Zxyz_obj *		rop,
const struct calib_Zxyz_obj *	op
)
{
	CALIB_FATAL_ERROR_IF ((rop	  EQ NULL) OR
			      (op	  EQ NULL) OR
			      (rop -> dom NE op -> dom));

	if (rop NE op) {
		Zxyz_realloc_copy_poly (rop, op -> nterms, op -> pow, op -> coeff);
	}
}

/*
 * Set the given polynomial to be a constant polynomial having the
 * given signed integer value.
 */

	static
	void
Zxyz_set_si (

struct calib_Zxyz_obj *	rop,		/* IN/OUT: destination polynomial */
calib_si_t		op		/* IN: integer value */
)
{
mpz_t		z;

	mpz_init_set_si (z, op);
	Zxyz_set_z (rop, z);
	mpz_clear (z);
}

/*
 * Set the given polynomial to be a constant polynomial having the
 * given integer value.
 */

	static
	void
Zxyz_set_z (

struct calib_Zxyz_obj *	rop,		/* IN/OUT: destination polynomial */
mpz_srcptr		op		/* IN: integer value */
)
{
int		nvars;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	if (mpz_sgn (op) EQ 0) {
		rop -> nterms = 0;
	}
	else {
		Zxyz_alloc (rop, 1);

		mpz_set (rop -> coeff, op);

		nvars = rop -> dom -> nvars;
		memset (rop -> pow, 0, nvars * sizeof (int));
		rop -> nterms = 1;
	}
}

/*
 * Set the given polynomial to be a single term var**power.
 */

	static
	void
Zxyz_set_var_power (

struct calib_Zxyz_obj *	rop,		/* IN/OUT: destination polynomial */
int			var,		/* IN: Variable for term */
int			power		/* IN: Power to use */
)
{
int		nvars;

	CALIB_FATAL_ERROR_IF (rop EQ NULL);

	nvars = rop -> dom -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));
	CALIB_FATAL_ERROR_IF (power < 0);

	Zxyz_alloc (rop, 1);

	mpz_set_ui (rop -> coeff, 1);
	memset (rop -> pow, 0, nvars * sizeof (int));
	rop -> pow [var] = power;
	rop -> nterms = 1;
}

/*
 * Addition of two polynomials over Z.
 */

	static
	void
Zxyz_add (

struct calib_Zxyz_obj *		rop,		/* IN/OUT: result */
const struct calib_Zxyz_obj *	op1,		/* IN: first operand */
const struct calib_Zxyz_obj *	op2		/* IN: second operand */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
mpz_srcptr			cp;
const int *			pp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom = rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));

	pap = Zxyz_get_paccum (dom);

	nvars	= dom -> nvars;

	nterms	= op1 -> nterms;
	pp	= op1 -> pow;
	cp	= op1 -> coeff;
	for (i = 0; i < nterms; i++, cp++, pp += nvars) {
		Zxyz_paccum_add_term (pap, cp, pp);
	}

	nterms	= op2 -> nterms;
	pp	= op2 -> pow;
	cp	= op2 -> coeff;
	for (i = 0; i < nterms; i++, cp++, pp += nvars) {
		Zxyz_paccum_add_term (pap, cp, pp);
	}

	Zxyz_set_poly_from_paccum (rop, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Addition of N polynomials over Z.
 */

	static
	void
Zxyz_add_n (

struct calib_Zxyz_obj *	rop,		/* IN/OUT: result */
int			npoly,		/* IN: number of polys to add */
const struct calib_Zxyz_obj **
			poly_ptrs	/* IN: array of ptrs to polynomials */
)
{
int				i, j, nterms, nvars;
const struct calib_Zxyz_obj *	poly;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
mpz_srcptr			cp;
const int *			pp;

	CALIB_FATAL_ERROR_IF ((rop	 EQ NULL) OR
			      (npoly	 <  0) OR
			      (poly_ptrs EQ NULL));

	dom = rop -> dom;
	for (i = 0; i < npoly; i++) {
		poly = poly_ptrs [i];
		CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR
				      (poly -> dom NE dom));
	}

	pap = Zxyz_get_paccum (dom);

	nvars	= dom -> nvars;

	for (i = 0; i < npoly; i++) {
		poly = poly_ptrs [i];
		nterms	= poly -> nterms;
		pp	= poly -> pow;
		cp	= poly -> coeff;
		for (j = 0; j < nterms; j++, cp++, pp += nvars) {
			Zxyz_paccum_add_term (pap, cp, pp);
		}
	}

	Zxyz_set_poly_from_paccum (rop, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Subtraction of two polynomials over Z.
 */

	static
	void
Zxyz_sub (

struct calib_Zxyz_obj *		rop,		/* IN/OUT: result */
const struct calib_Zxyz_obj *	op1,		/* IN: first operand */
const struct calib_Zxyz_obj *	op2		/* IN: second operand */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
mpz_srcptr			cp;
const int *			pp;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom = rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));

	pap = Zxyz_get_paccum (dom);

	nvars	= dom -> nvars;

	nterms	= op1 -> nterms;
	pp	= op1 -> pow;
	cp	= op1 -> coeff;
	for (i = 0; i < nterms; i++, cp++, pp += nvars) {
		Zxyz_paccum_add_term (pap, cp, pp);
	}

	mpz_init (TEMP);
	nterms	= op2 -> nterms;
	pp	= op2 -> pow;
	cp	= op2 -> coeff;
	for (i = 0; i < nterms; i++, cp++, pp += nvars) {
		mpz_neg (TEMP, cp);
		Zxyz_paccum_add_term (pap, TEMP, pp);
	}
	mpz_clear (TEMP);

	Zxyz_set_poly_from_paccum (rop, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Negation of a polynomial over Z.
 */

	static
	void
Zxyz_neg (

struct calib_Zxyz_obj *		rop,		/* IN/OUT: result */
const struct calib_Zxyz_obj *	op		/* IN: polynomial to negate */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
mpz_ptr				rcp;
mpz_srcptr			scp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	nterms	= op -> nterms;

	if (rop EQ op) {
		/* Negation in place. */
		rcp = rop -> coeff;
		for (i = 0; i < nterms; i++, rcp++) {
			mpz_neg (rcp, rcp);
		}
		return;
	}

	Zxyz_alloc (rop, nterms);

	nvars	= dom -> nvars;

	memcpy (rop -> pow, op -> pow, nterms * nvars * sizeof (int));

	rcp = rop -> coeff;
	scp = op -> coeff;

	for (i = 0; i < nterms; i++, rcp++, scp++) {
		mpz_neg (rcp, scp);
	}
	rop -> nterms = nterms;
}

/*
 * Multiplcation of two polynomials over Z.
 */

	static
	void
Zxyz_mul (

struct calib_Zxyz_obj *		rop,		/* IN/OUT: result */
const struct calib_Zxyz_obj *	op1,		/* IN: first operand */
const struct calib_Zxyz_obj *	op2		/* IN: second operand */
)
{
int				i, j, k, nterms1, nterms2, nvars;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
const int *			pp1;
const int *			pp2;
mpz_srcptr			cp1, cp2;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom = rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));

	pap = Zxyz_get_paccum (dom);

	nvars	= dom -> nvars;
	nterms1	= op1 -> nterms;
	nterms2	= op2 -> nterms;

	int	ppow [nvars];

	mpz_init (TEMP);

	cp1 = op1 -> coeff;
	pp1 = op1 -> pow;
	for (i = 0; i < nterms1; i++, cp1++, pp1 += nvars) {
		cp2 = op2 -> coeff;
		pp2 = op2 -> pow;
		for (j = 0; j < nterms2; j++, cp2++, pp2 += nvars) {
			mpz_mul (TEMP, cp1, cp2);
			for (k = 0; k < nvars; k++) {
				ppow [k] = pp1 [k] + pp2 [k];
			}
			Zxyz_paccum_add_term (pap, TEMP, ppow);
		}
	}

	mpz_clear (TEMP);

	Zxyz_set_poly_from_paccum (rop, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Multiply the given polynomial by the given integer.
 */

	static
	void
Zxyz_mul_z (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_Zxyz_obj *	op1,	/* IN: polynomial to multiply */
mpz_srcptr			op2	/* IN: integer to multiply by */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
mpz_ptr				rcp;
mpz_srcptr			acp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op1 -> dom NE dom);

	if (mpz_sgn (op2) EQ 0) {
		/* Multiplying by zero. */
		rop -> nterms = 0;
	}
	if (mpz_cmp_ui (op2, 1) EQ 0) {
		/* Multiplying by one. */
		if (rop NE op1) {
			Zxyz_set (rop, op1);
		}
		return;
	}

	nterms	= op1 -> nterms;

	if (rop EQ op1) {
		/* Multiplication in place. */
		rcp = rop -> coeff;
		for (i = 0; i < nterms; i++, rcp++) {
			mpz_mul (rcp, rcp, op2);
		}
		return;
	}

	Zxyz_alloc (rop, nterms);

	nvars	= dom -> nvars;
	memcpy (rop -> pow, op1 -> pow, nterms * nvars * sizeof (int));

	rcp = rop -> coeff;
	acp = op1 -> coeff;

	for (i = 0; i < nterms; i++, rcp++, acp++) {
		mpz_mul (rcp, acp, op2);
	}

	rop -> nterms = nterms;
}

/*
 * Take the given polynomial to the given integer power.
 */

	static
	void
Zxyz_ipow (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_Zxyz_obj *	op,	/* IN: polynomial to exponentiate */
int				power	/* IN: integer power */
)
{
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj		RES, PREG;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR
			      (op   EQ NULL) OR
			      (power  < 0));

	dom = rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	Zxyz_init_si (dom, &RES, 1);
	Zxyz_init (dom, &PREG);
	Zxyz_set (&PREG, op);

	while (power > 0) {
		if ((power & 1) NE 0) {
			Zxyz_mul (&RES, &RES, &PREG);
		}
		Zxyz_mul (&PREG, &PREG, &PREG);
		power >>= 1;
	}
	Zxyz_set (rop, &RES);
	Zxyz_clear (&PREG);
	Zxyz_clear (&RES);
}

/*
 * Create a dynamically allocated copy of the given polynomial.
 */

	static
	struct calib_Zxyz_obj *
Zxyz_dup (

const struct calib_Zxyz_obj *	op	/* IN: polynomial to duplicate */
)
{
int				nterms, size, nvars;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		dst;
int *				pow;
mpz_ptr				zp;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	nterms	= op -> nterms;
	dom	= op -> dom;

	size = nterms;

	dst = CALIB_NEW (struct calib_Zxyz_obj);
	pow = NULL;
	zp = NULL;
	if (size > 0) {
		nvars = dom -> nvars;
		pow = CALIB_NEWA (size * nvars, int);
		memcpy (pow, op -> pow, size * nvars * sizeof (int));

		zp = calib_new_Z_vector (size);
		calib_copy_Z_vector (zp, op -> coeff, size);
	}

	dst -> nterms	= nterms;
	dst -> size	= size;
	dst -> dom	= dom;
	dst -> pow	= pow;
	dst -> coeff	= zp;

	return (dst);
}

/*
 * Free the given polynomial, which must have been dynamically allocated.
 */

	static
	void
Zxyz_free (

struct calib_Zxyz_obj *	poly	/* IN: polynomial to free */
)
{
	if (poly NE NULL) {
		Zxyz_clear (poly);
		free (poly);
	}
}

/*
 * Evaluate the given poly(val1, val2, val3, ...).
 */

	static
	void
Zxyz_eval (

mpz_ptr				rop,		/* IN/OUT: result */
const struct calib_Zxyz_obj *	poly,		/* IN: polynomial to eval */
mpz_srcptr			values		/* IN: values to use */
)
{
int				i, j, k, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
const int *			pow;
const int *			pp;
mpz_ptr				zp;
mpz_srcptr			coeff;
mpz_t				SUM, PROD, POW;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR
			      (poly   EQ NULL) OR
			      (values EQ NULL));

	nterms	= poly -> nterms;
	dom	= poly -> dom;
	pow	= poly -> pow;
	nvars	= dom -> nvars;

	/* Compute the highest power for each var. */
	int maxpow [nvars];
	memset (maxpow, 0, sizeof (maxpow));
	pp = pow;
	for (i = 0; i < nterms; i++) {
		for (j = 0; j < nvars; j++) {
			k = *pp++;
			CALIB_FATAL_ERROR_IF (k < 0);
			if (k < maxpow [j]) {
				maxpow [j] = k;
			}
		}
	}

	mpz_init (SUM);
	mpz_init (PROD);
	mpz_init (POW);

	/* Compute table powers for each var's value. */
	mpz_ptr	vpow [nvars];
	for (j = 0; j < nvars; j++) {
		k = maxpow [j];
		zp = calib_new_Z_vector (1 + k);
		mpz_set_ui (POW, 1);
		mpz_set (&zp [0], POW);
		for (i = 1; i <= k; i++) {
			mpz_mul (POW, POW, &values [j]);
			mpz_set (&zp [i], POW);
		}
		vpow [j] = zp;
	}

	coeff = poly -> coeff;

	pp = pow;
	for (i = 0; i < nterms; i++) {
		mpz_set (PROD, &coeff [i]);
		for (j = 0; j < nvars; j++) {
			k = *pp++;
			mpz_mul (PROD, PROD, &vpow [j] [k]);
		}
		mpz_add (SUM, SUM, PROD);
	}

	mpz_set (rop, SUM);

	for (j = 0; j < nvars; j++) {
		calib_free_Z_vector (vpow [j], 1 + maxpow [j]);
	}

	mpz_clear (POW);
	mpz_clear (PROD);
	mpz_clear (SUM);
}

/*
 * Evaluate the given polynomial at specified values for a given
 * subset of the variables.  The result is a polynomial that does
 * not contain any of the variables that were evaluated away.
 *
 * When (evflags[i] EQ TRUE), then values[i] is substituted in for
 * variable i.  The other elements of values[] are ignored.
 */

	static
	void
Zxyz_eval_var_subset (

struct calib_Zxyz_obj *	rop,		/* IN/OUT: resulting polynomial */
const struct calib_Zxyz_obj *
			poly,		/* IN: polynomial to evaluate */
const calib_bool *	evflags,	/* IN: which vars to evaluate */
mpz_srcptr		values		/* IN: variable values */
)
{
int				i, j, k, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
mpz_srcptr			cp;
const int *			pp;
mpz_t				cval, tmp;

	CALIB_FATAL_ERROR_IF ((rop     EQ NULL) OR
			      (poly    EQ NULL) OR
			      (evflags EQ NULL) OR
			      (values  EQ NULL));

	dom = rop -> dom;

	CALIB_FATAL_ERROR_IF (poly -> dom NE dom);

	pap = Zxyz_get_paccum (dom);

	nvars	= dom -> nvars;

	int	pow [nvars];

	mpz_init (cval);
	mpz_init (tmp);

	nterms	= poly -> nterms;
	pp	= poly -> pow;
	cp	= poly -> coeff;

	for (i = 0; i < nterms; i++, cp++, pp += nvars) {
		mpz_set (cval, cp);
		memcpy (&pow [0], pp, nvars * sizeof (int));

		for (j = 0; j < nvars; j++) {
			if (NOT (evflags [j])) continue;
			k = pow [j];
			if (k > 0) {
				pow [j] = 0;	/* Drop var from this term */
				mpz_pow_ui (tmp, &values [j], k);
				mpz_mul (cval, cval, tmp);
			}
			else {
				CALIB_FATAL_ERROR_IF (k < 0);
			}
		}

		Zxyz_paccum_add_term (pap, cval, &pow [0]);
	}

	mpz_clear (tmp);
	mpz_clear (cval);

	Zxyz_set_poly_from_paccum (rop, pap);

	Zxyz_put_paccum (pap);
}

/*
 * The pseudo-division algorithm (with remainder):
 * Given polynomials a, b \in Z[x,y,z] and a choice x of main variable,
 * find polynomials q, r, d \in Z[x,y,z] such that:
 *
 *	d*a = q*b + r	AND	degree_x (r) < degree_x (b)
 *			AND	d is free of x
 *
 * Yes, this is messy because the coefficients are in Z which is a ring,
 * not a field.
 */

	static
	void
Zxyz_div (

struct calib_Zxyz_obj *		quotient,	/* IN/OUT: quotient */
struct calib_Zxyz_obj *		remainder,	/* IN/OUT: remainder */
struct calib_Zxyz_obj *		d,		/* IN/OUT: denominator */
const struct calib_Zxyz_obj *	a,		/* IN: dividend */
const struct calib_Zxyz_obj *	b,		/* IN: divisor */
int				var		/* IN: main variable */
)
{
int				i, j, adeg, bdeg;
const struct calib_Zxyz_dom *	dom;
int *				pp;
struct calib_Zxyz_obj *		acoeff;
struct calib_Zxyz_obj *		bcoeff;
struct calib_Zxyz_obj *		alc;
struct calib_Zxyz_obj *		blc;
struct calib_Zxyz_obj		qpoly, rpoly, dpoly, GCD, af, bf, tmp1, tmp2;

	CALIB_FATAL_ERROR_IF ((a EQ NULL) OR (b EQ NULL));

	dom = a -> dom;

	CALIB_FATAL_ERROR_IF (b -> dom NE dom);
	CALIB_FATAL_ERROR_IF ((quotient NE NULL) AND
			      (quotient -> dom NE dom));
	CALIB_FATAL_ERROR_IF ((remainder NE NULL) AND
			      (remainder -> dom NE dom));
	CALIB_FATAL_ERROR_IF ((d NE NULL) AND (d -> dom NE dom));
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= dom -> nvars));

	/* Reorganize a and b as polynomials in the given var, getting	*/
	/* an array of coefficient polynomials (in the remaining vars)	*/
	/* for each power of var from 0 to its degree in x.		*/

	adeg = Zxyz_get_var_coeffs (&acoeff, var, a);
	bdeg = Zxyz_get_var_coeffs (&bcoeff, var, b);

	Zxyz_init (dom, &qpoly);
	Zxyz_init (dom, &rpoly);
	Zxyz_init_si (dom, &dpoly, 1);

	if (bdeg <= 0) {
		/* Catch attempts to divide by zero. */
		CALIB_FATAL_ERROR_IF (bdeg < 0);

		/* Division by the "constant" b.			*/
		/* Result is: q = a, r = 0, d = b.			*/
		Zxyz_set (&qpoly, a);
		/* rpoly is zero already. */
		Zxyz_set (&dpoly, b);
	}
	else {
		/* Have degree(b) >= 1. */
		Zxyz_init (dom, &GCD);
		Zxyz_init (dom, &af);
		Zxyz_init (dom, &bf);
		Zxyz_init (dom, &tmp1);
		Zxyz_init (dom, &tmp2);

		blc = &bcoeff [bdeg];
		while (adeg >= bdeg) {
			j = adeg - bdeg;
			alc = &acoeff [adeg];
			/* Compute GCD and cofactors of alc and blc. */
			Zxyz_eezgcd (&GCD, &af, &bf, alc, blc);
			Zxyz_mul (&dpoly, &dpoly, &bf);
			for (i = 0; i <= adeg; i++) {
				Zxyz_mul (&acoeff [i], &acoeff [i], &bf);
			}
			for (i = bdeg; i >= 0; i--) {
				Zxyz_mul (&tmp1, &acoeff [i + j], &bf);
				Zxyz_mul (&tmp2, &bcoeff [i],     &af);
				Zxyz_sub (&acoeff [i + j], &tmp1, &tmp2);
			}

			/* Add af*var^j to quotient. */
			pp = &af.pow [var];
			for (i = 0; i < af.nterms; i++) {
				CALIB_FATAL_ERROR_IF (*pp NE 0);
				*pp = j;
			}
			Zxyz_add (&qpoly, &qpoly, &af);

			/* Verify that acoeff [adeg] is now zero.	*/
			CALIB_FATAL_ERROR_IF (acoeff [adeg].nterms NE 0);

			/* Adjust degree of a downward, clearing as we go. */
			do {
				Zxyz_clear (&acoeff [adeg]);
				--adeg;
				if (acoeff [adeg].nterms > 0) break;
			} while (adeg >= 0);
		}
		Zxyz_clear (&tmp2);
		Zxyz_clear (&tmp1);
		Zxyz_clear (&bf);
		Zxyz_clear (&af);
		Zxyz_clear (&GCD);

		Zxyz_poly_from_coeffs (&rpoly, adeg, &acoeff [0], var);
	}

	Zxyz_free_array (bcoeff, bdeg + 1);
	Zxyz_free_array (acoeff, adeg + 1);

	if (quotient NE NULL) {
		CALIB_FATAL_ERROR_IF (quotient -> dom NE dom);
		Zxyz_set (quotient, &qpoly);
	}
	if (remainder NE NULL) {
		CALIB_FATAL_ERROR_IF (remainder -> dom NE dom);
		Zxyz_set (remainder, &rpoly);
	}
	if (d NE NULL) {
		CALIB_FATAL_ERROR_IF (d -> dom NE dom);
		Zxyz_set (d, &dpoly);
	}
	Zxyz_clear (&dpoly);
	Zxyz_clear (&rpoly);
	Zxyz_clear (&qpoly);
}

/*
 * Divide the given polynomials a by b.  It is a fatal error if
 * the division is not exact.
 */

	static
	void
Zxyz_div_exact (

struct calib_Zxyz_obj *		rop,		/* IN/OUT: quotient polynomial */
const struct calib_Zxyz_obj *	op1,		/* IN: dividend polynomial */
const struct calib_Zxyz_obj *	op2,		/* IN: divisor polynomial */
struct GInfo *			gip		/* IN/OUT: global info */
)
{
int				i, nvars;
const struct calib_Zxyz_dom *	dom;
int *				r2fmap;
struct RPoly *			rp1;
struct RPoly *			rp2;
struct RPoly *			qp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR
			      (op1 EQ NULL) OR
			      (op2 EQ NULL) OR
			      (gip EQ NULL));

	dom	= rop -> dom;
	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));
	nvars	= dom -> nvars;

	r2fmap = CALIB_NEWA (nvars, int);
	for (i = 0; i < nvars; i++) {
		r2fmap [i] = i;
	}

	rp1 = Zxyz_rpoly_cv_f2r (op1, nvars, r2fmap, gip);
	rp2 = Zxyz_rpoly_cv_f2r (op2, nvars, r2fmap, gip);
	qp  = Zxyz_rpoly_div_exact (rp1, rp2, gip);

	Zxyz_rpoly_free (rp2, gip);
	Zxyz_rpoly_free (rp1, gip);

	CALIB_FATAL_ERROR_IF (qp EQ DIVISION_NOT_EXACT);

	Zxyz_rpoly_cv_r2f (rop, qp, r2fmap);

	Zxyz_rpoly_free (qp, gip);

	free (r2fmap);
}

/*
 * Divide the given polynomial in Z[x,y,z] by the given integer.
 * It is a fatal error if the division is not exact.
 */

	static
	void
Zxyz_div_z_exact (

struct calib_Zxyz_obj *		result,		/* IN/OUT: result polynomial */
const struct calib_Zxyz_obj *	poly,		/* IN: polynomial to divide */
mpz_srcptr			z		/* IN: integer to divide by */
)
{
int		i, nterms;
mpz_srcptr	src;
mpz_ptr		res;
mpz_t		REM;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR
			      (poly   EQ NULL) OR
			      (z      EQ NULL));
	CALIB_FATAL_ERROR_IF (result -> dom NE poly -> dom);

	nterms = poly -> nterms;
	src = poly -> coeff;
	res = calib_new_Z_vector (nterms);

	mpz_init (REM);
	for (i = 0; i < nterms; i++) {
		mpz_tdiv_qr (&res [i], REM, &src [i], z);
		CALIB_FATAL_ERROR_IF (mpz_sgn (REM) NE 0);
	}
	mpz_clear (REM);

	Zxyz_realloc_copy_poly (result, nterms, poly -> pow, res);
	calib_free_Z_vector (res, nterms);
}

/*
 * Repeatedly divide a by b until no more factors of b can be
 * removed.  Returns the power k such that result = a / b**k.
 */

	static
	int
Zxyz_div_remove (

struct calib_Zxyz_obj *		rop,		/* IN/OUT: result polynomial */
const struct calib_Zxyz_obj *	op1,		/* IN: polynomial to divide */
const struct calib_Zxyz_obj *	op2		/* IN: polynomial to divide by */
)
{
int				i, nvars, pow;
const struct calib_Zxyz_dom *	dom;
int *				r2fmap;
struct RPoly *			rp1;
struct RPoly *			rp2;
struct RPoly *			qp;
struct GInfo *			gip;
struct GInfo			ginfo;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op1 EQ NULL) OR (op2 EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op1 -> dom NE dom) OR (op2 -> dom NE dom));

	Zxyz_ginfo_init (&ginfo);
	gip = &ginfo;

	nvars	= dom -> nvars;

	r2fmap = CALIB_NEWA (nvars, int);
	for (i = 0; i < nvars; i++) {
		r2fmap [i] = i;
	}

	rp1 = Zxyz_rpoly_cv_f2r (op1, nvars, r2fmap, gip);
	rp2 = Zxyz_rpoly_cv_f2r (op2, nvars, r2fmap, gip);

	pow = 0;
	for (;;) {
		qp  = Zxyz_rpoly_div_exact (rp1, rp2, gip);

		if (qp EQ DIVISION_NOT_EXACT) {
			break;
		}
		Zxyz_rpoly_free (rp1, gip);
		rp1 = qp;
		++pow;
	}

	Zxyz_rpoly_cv_r2f (rop, rp1, r2fmap);

	Zxyz_rpoly_free (rp2, gip);
	Zxyz_rpoly_free (rp1, gip);
	free (r2fmap);

	Zxyz_ginfo_clear (gip);

	return (pow);
}

/*
 * Compute the Greatest Common Divisor of two polynomials in Z[x,y,z].
 */

	static
	void
Zxyz_gcd (

struct calib_Zxyz_obj *		gcd,	/* IN/OUT: GCD result */
const struct calib_Zxyz_obj *	a,	/* IN: first operand */
const struct calib_Zxyz_obj *	b	/* IN: second operand */
)
{
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj		AF, BF;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	dom = gcd -> dom;

	CALIB_FATAL_ERROR_IF ((a -> dom NE dom) OR (b -> dom NE dom));

	Zxyz_init (dom, &AF);
	Zxyz_init (dom, &BF);

	Zxyz_eezgcd (gcd, &AF, &BF, a, b);

	Zxyz_clear (&BF);
	Zxyz_clear (&AF);
}

/*
 * Compute the GCD of N different polynomials, which can be done more
 * efficiently than doing it two polynomials at a time.
 *
 * This interface also allows the cofactors to be returned.
 */

	static
	void
Zxyz_gcd_n (

struct calib_Zxyz_obj *		gcd,	/* IN/OUT: Greatest Common Divisor */
struct calib_Zxyz_obj **	cofact,	/* IN/OUT: Array of cofactors (or NULL) */
int				npoly,	/* IN: number of polynomials to GCD */
const struct calib_Zxyz_obj **	array	/* IN: array of polynomials to GCD */
)
{
struct GInfo		ginfo;

	Zxyz_ginfo_init (&ginfo);

	Zxyz_internal_eezgcd_n (gcd, cofact, npoly, array, &ginfo);

	Zxyz_ginfo_clear (&ginfo);
}

/*
 * The Extended Euclidean Algorithm for polynomials in Z[x,y,z].
 * Given polynomials a, b \in Z[x,y,z],
 * compute polynomials g, xa, xb \in Z[x,y,z] such that
 *
 * 1.	xa * a + xb * b = g.
 * 2.	xa and xb are of minimum degree.
 */

	static
	void
Zxyz_extgcd (

struct calib_Zxyz_obj *		gcd,	/* IN/OUT: GCD result */
struct calib_Zxyz_obj *		xa,	/* IN/OUT: multiplier for a */
struct calib_Zxyz_obj *		xb,	/* IN/OUT: multiplier for b */
mpz_ptr				d,	/* IN/OUT: denominator for xa, xb */
const struct calib_Zxyz_obj *	a,	/* IN: first operand */
const struct calib_Zxyz_obj *	b	/* IN: second operand */
)
{
	/* FIXME -- write this! */
	CALIB_FATAL_ERROR;
}

/*
 * Compute gcd, af, and bf such that:
 *
 *	gcd is the greatest-common-divisor of a and b,
 *	af = a / gcd,
 *	bf = b / gcd.
 *
 * The quantities af and bf are sometimes called the "cofactors".
 * We use the "Enhanced, Extended Zassenhaus" method - EEZ.
 */

	static
	void
Zxyz_eezgcd (

struct calib_Zxyz_obj *		gcd,	/* IN/OUT: GCD result */
struct calib_Zxyz_obj *		af,	/* IN/OUT: multiplier for a */
struct calib_Zxyz_obj *		bf,	/* IN/OUT: multiplier for b */
const struct calib_Zxyz_obj *	a,	/* IN: first operand */
const struct calib_Zxyz_obj *	b	/* IN: second operand */
)
{
const struct calib_Zxyz_obj *	array [2];
struct calib_Zxyz_obj *		cofact [2];
struct GInfo			ginfo;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (a EQ NULL) OR (b EQ NULL));

	array [0]	= a;
	array [1]	= b;
	cofact [0]	= af;
	cofact [1]	= bf;

	Zxyz_ginfo_init (&ginfo);

	Zxyz_internal_eezgcd_n (gcd, cofact, 2, array, &ginfo);

	Zxyz_ginfo_clear (&ginfo);
}

/*
 * Compute the integer content of the given polynomial.
 */

	static
	void
Zxyz_z_content (

mpz_ptr				zcont,	/* IN/OUT: integer content */
const struct calib_Zxyz_obj *	op	/* IN: polynomial to get integer content of */
)
{
int			i, nterms;
mpz_srcptr		zp;
mpz_t			g;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	nterms	= op -> nterms;
	zp	= op -> coeff;

	if (nterms <= 0) {
		CALIB_FATAL_ERROR_IF (nterms < 0);
		mpz_set_ui (zcont, 0);
		return;
	}

	mpz_init (g);

	mpz_abs (g, &zp [0]);
	if (mpz_cmp_ui (g, 1) > 0) {
		for (i = 1; i < nterms; i++) {
			mpz_gcd (g, g, &zp [i]);
			if (mpz_cmp_ui (g, 1) EQ 0) break;
		}
	}

	mpz_set (zcont, g);

	mpz_clear (g);
}

/*
 * Compute and remove the integer content from the given polynomial.
 */

	static
	void
Zxyz_strip_z_content (

mpz_ptr			zcont,	/* IN/OUT: integer content */
struct calib_Zxyz_obj *	poly	/* IN/OUT: polynomial to strip content from */
)
{
int		i, nterms;
mpz_ptr		zp;

	Zxyz_z_content (zcont, poly);
	if (mpz_cmp_ui (zcont, 1) NE 0) {
		nterms	= poly -> nterms;
		zp	= poly -> coeff;
		for (i = 0; i < nterms; i++, zp++) {
			mpz_divexact (zp, zp, zcont);
		}
	}
}

/*
 * Compute the content and primitive part of the given polynomial
 * with respect to the given variable.
 */

	static
	void
Zxyz_prim_part (

struct calib_Zxyz_obj *		content,	/* IN/OUT: the content */
struct calib_Zxyz_obj *		ppart,		/* IN/OUT: the primitive part */
const struct calib_Zxyz_obj *	op,		/* IN: input polynomial */
int				var		/* IN: main variable to use */
)
{
struct GInfo		ginfo;

	Zxyz_ginfo_init (&ginfo);

	Zxyz_int_content_ppart (content, ppart, op, var, &ginfo);

	Zxyz_ginfo_clear (&ginfo);
}

/*
 * Internal routine to compute the content and primitive part of the
 * given polynomial with respect to the given variable.
 */

	static
	void
Zxyz_int_content_ppart (

struct calib_Zxyz_obj *		content,	/* IN/OUT: the content */
struct calib_Zxyz_obj *		ppart,		/* IN/OUT: the primitive part */
const struct calib_Zxyz_obj *	poly,		/* IN: input polynomial */
int				var,		/* IN: main variable to use */
struct GInfo *			gip		/* IN/OUT: global info */
)
{
int				i, nvars, nv, degree;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		coeff_polys;
struct calib_Zxyz_obj **	poly_ptrs;
struct calib_Zxyz_obj *		zpp_array [1];
struct calib_Zxyz_obj		g, zpp;
mpz_t				zcont;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (var < 0));
	dom	= poly -> dom;
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF (((content NE NULL) AND
			      (content -> dom NE dom)) OR
			      ((ppart NE NULL) AND
			       (ppart -> dom NE dom)) OR
			      (var >= nvars));

	/* Break the polynomial up into separate polynomials for each	*/
	/* coefficient of the given variable.				*/
	degree = Zxyz_get_var_coeffs (&coeff_polys, var, poly);

	if (degree <= 0) {
		/* Polynomial does not contain requested variable. */
		Zxyz_free_array (coeff_polys, degree + 1);
		if (content NE NULL) {
			Zxyz_set (content, poly);
		}
		if (ppart NE NULL) {
			Zxyz_set (ppart, poly);
		}
		return;
	}

	/* Count the number of variables occurring in poly. */
	nv = Zxyz_num_vars_present (poly);
	if (nv <= 1) {
		CALIB_FATAL_ERROR_IF (nv <= 0);
		/* Polynomial is univariate.  Take integer content. */
		Zxyz_free_array (coeff_polys, degree + 1);
		mpz_init (zcont);
		Zxyz_init (dom, &zpp);
		Zxyz_set (&zpp, poly);
		zpp_array [0] = &zpp;

		Zxyz_remove_z_contents (zcont, 1, zpp_array);

		if (content NE NULL) {
			Zxyz_set_z (content, zcont);
		}
		if (ppart NE NULL) {
			Zxyz_set (ppart, &zpp);
		}
		Zxyz_clear (&zpp);
		mpz_clear (zcont);
		return;
	}

	/* There are two or more variables. */

	poly_ptrs = CALIB_NEWA (degree + 1, struct calib_Zxyz_obj *);
	for (i = 0; i <= degree; i++) {
		poly_ptrs [i]	= &coeff_polys [i];
	}

	/* Compute gcd and cofactors for all these polynomials. */
	Zxyz_init (dom, &g);
	Zxyz_internal_eezgcd_n (&g, &poly_ptrs [0], degree + 1,
		       (const struct calib_Zxyz_obj **) (&poly_ptrs [0]),
		       gip);

	if (content NE NULL) {
		Zxyz_set (content, &g);
	}
	if (ppart NE NULL) {
		Zxyz_poly_from_coeffs (ppart, degree, &coeff_polys [0], var);
	}

	Zxyz_clear (&g);
	free (poly_ptrs);
	Zxyz_free_array (coeff_polys, degree + 1);
}

/*
 * Internal routine that we recurse with.
 *
 * Compute the gcd of the given array of n polynomials, as well as
 * the array of "cofactor" polynomials (the original polynomials with
 * the GCD removed).
 *
 * We use the "Enhanced, Extended Zassenhaus" GCD method - EEZGCD.
 */

	static
	void
Zxyz_internal_eezgcd_n (

struct calib_Zxyz_obj *		gcd,	/* IN/OUT: GCD result */
struct calib_Zxyz_obj **	cofact,	/* IN/OUT: cofactors */
int				npoly,	/* IN: number of polynomials */
const struct calib_Zxyz_obj **	a,	/* IN: array of polys to GCD */
struct GInfo *			gip	/* IN/OUT: global info */
)
{
int				i, k, nvars, nv, x1;
calib_bool			constant_poly_seen;
const struct calib_Zxyz_dom *	dom;
const struct calib_Zxyz_obj **	poly_ptrs;
struct calib_Zxyz_obj **	cf_ptrs;
struct calib_Zxyz_obj **	array;
const struct calib_Zxyz_obj *	cp;
struct calib_Zxyz_obj *		p;
struct calib_Zxyz_obj *		pcont_polys;
const struct calib_Zxyz_obj **	pcont_ptrs;
int *				mindeg;
int *				maxdeg;
int *				r2fmap;
struct calib_Zx_obj *		zx_array;
const struct calib_Zx_obj **	zx_ptrs;
const struct calib_Zx_dom *	Zx;
struct RPoly *			rgcd;
struct RPoly *			rp;
struct RPoly *			qp;
struct calib_Zxyz_obj		vcont, pcont;
struct calib_Zx_obj		zx_gcd;
mpz_t				zcont;

	CALIB_FATAL_ERROR_IF ((gcd EQ NULL) OR (npoly < 0));

	if (npoly EQ 0) {
		gcd -> nterms = 0;	/* Result is zero. */
		return;
	}

	/* Validate inputs.  The cofactors are all optional. */
	dom	= gcd -> dom;
	nvars	= dom -> nvars;
	for (i = 0; i < npoly; i++) {
		cp = a [i];
		CALIB_FATAL_ERROR_IF ((cp EQ NULL) OR (cp -> dom NE dom));
	}
	if (cofact NE NULL) {
		for (i = 0; i < npoly; i++) {
			p = cofact [i];
			CALIB_FATAL_ERROR_IF ((p NE NULL) AND
					      (p -> dom NE dom));
		}
	}

	/* Make locally modifiable copies of the two pointer arrays. */
	poly_ptrs = CALIB_NEWA (npoly, const struct calib_Zxyz_obj *);
	cf_ptrs = CALIB_NEWA (npoly, struct calib_Zxyz_obj *);
	memcpy (poly_ptrs, a, npoly * sizeof (poly_ptrs [0]));
	if (cofact NE NULL) {
		memcpy (cf_ptrs, cofact, npoly * sizeof (cf_ptrs [0]));
	}
	else {
		memset (cf_ptrs, 0, npoly * sizeof (cf_ptrs [0]));
	}

	/* Let's remove any polynomials that are zero. */
	k = 0;
	for (i = 0; i < npoly; i++) {
		cp = poly_ptrs [i];
		p = cf_ptrs [i];
		if (cp -> nterms NE 0) {
			poly_ptrs [k] = cp;
			cf_ptrs [k] = p;
			++k;
		}
		else {
			/* This polynomial is zero -- remove it. */
			if (p NE NULL) {
				p -> nterms = 0;
			}
		}
	}
	npoly = k;

	if (npoly EQ 0) {
		gcd -> nterms = 0;	/* Result is zero. */
		free (cf_ptrs);
		free (poly_ptrs);
		return;
	}
	if (npoly EQ 1) {
		/* One polynomial -- a very simple case. */
		Zxyz_set (gcd, poly_ptrs [0]);
		if (cf_ptrs [0] NE NULL) {
			Zxyz_set_si (cf_ptrs [0], 1);
		}
		Zxyz_adjust_gcd_n_result_signs (gcd, cf_ptrs, npoly);
		free (cf_ptrs);
		free (poly_ptrs);
		return;
	}

	/* Two or more polynomials.  Make locally modifiable copies. */
	array = CALIB_NEWA (npoly, struct calib_Zxyz_obj *);
	for (i = 0; i < npoly; i++) {
		array [i] = Zxyz_dup (poly_ptrs [i]);
	}

	/* Remove integer contents. */
	mpz_init (zcont);
	Zxyz_remove_z_contents (zcont, npoly, array);

	/* Check to see if any of the polynomials is a non-zero		*/
	/* constant.  If so, then the GCD is just zcont.		*/
	constant_poly_seen = FALSE;
	for (i = 0; i < npoly; i++) {
		p = array [i];
		if (Zxyz_is_non_zero_constant_poly (p)) {
			constant_poly_seen = TRUE;
			break;
		}
	}
	if (constant_poly_seen) {
		Zxyz_set_z (gcd, zcont);
		if (cofact NE NULL) {
			for (i = 0; i < npoly; i++) {
				p = cf_ptrs [i];
				if (p NE NULL) {
					Zxyz_div_z_exact (p,
							  poly_ptrs [i],
							  zcont);
				}
			}
		}
		Zxyz_adjust_gcd_n_result_signs (gcd, cf_ptrs, npoly);
		mpz_clear (zcont);
		for (i = 0; i < npoly; i++) {
			Zxyz_free (array [i]);
		}
		free (array);
		free (cf_ptrs);
		free (poly_ptrs);
		return;
	}

	/* Get the min/max degree of each var in each polynomial. */
	Zxyz_get_poly_array_degrees (&mindeg, &maxdeg, npoly, nvars, array);

	/* Determine which variables are actually present in the	*/
	/* problem, and pick a suitable ordering of these variables.	*/
	/* (Do this *before* removing vconts, below.)			*/
	r2fmap = Zxyz_choose_var_order (&nv,
					npoly,
					nvars,
					(const struct calib_Zxyz_obj **) array,
					mindeg,
					maxdeg);

	/* Remove contents that consist of variables raised to some power. */
	Zxyz_init (dom, &vcont);
	Zxyz_remove_v_contents (&vcont, npoly, nvars, array, mindeg, maxdeg);

	/* Combine zcont into vcont. */
	Zxyz_mul_z (&vcont, &vcont, zcont);
	mpz_clear (zcont);

	if (nv <= 0) {
		/* None of the polynomials have any variables. */
		free (r2fmap);
		Zxyz_set (gcd, &vcont);
		if (cofact NE NULL) {
			for (i = 0; i < npoly; i++) {
				p = cf_ptrs [i];
				if (p NE NULL) {
					Zxyz_div_exact (p,
							poly_ptrs [i],
							&vcont,
							gip);
				}
			}
		}
		Zxyz_adjust_gcd_n_result_signs (gcd, cf_ptrs, npoly);
		Zxyz_clear (&vcont);
		for (i = 0; i < npoly; i++) {
			Zxyz_free (array [i]);
		}
		free (array);
		free (cf_ptrs);
		free (poly_ptrs);
		return;
	}

	/* Get the "main" variable. */
	x1 = r2fmap [nv - 1];

	/* Compute the content and primitive part of each polynomial	*/
	/* with respect to x1.  These are recursive calls, but the	*/
	/* polynomials each have one fewer variable.			*/
	pcont_polys = Zxyz_new_array (dom, npoly);
	for (i = 0; i < npoly; i++) {
		p = array [i];
		Zxyz_int_content_ppart (&pcont_polys [i], p, p, x1, gip);
	}

	/* Compute the GCD of all the contents.  We don't need the	*/
	/* cofactors of these, however, just the GCD.			*/
	pcont_ptrs = CALIB_NEWA (npoly, const struct calib_Zxyz_obj *);
	for (i = 0; i < npoly; i++) {
		pcont_ptrs [i] = &pcont_polys [i];
	}
	Zxyz_init (dom, &pcont);
	Zxyz_internal_eezgcd_n (&pcont, NULL, npoly, pcont_ptrs, gip);
	free (pcont_ptrs);
	Zxyz_free_array (pcont_polys, npoly);

	/* Combine vcont into pcont. */
	Zxyz_mul (&pcont, &pcont, &vcont);
	Zxyz_clear (&vcont);

	/* We now have primitive polynomials. */

	if (nv <= 1) {
		CALIB_FATAL_ERROR_IF (nv <= 0);
		/* Polynomials are univariate, x1 is the sole var remaining. */
		Zx = calib_get_Zx_dom ();
		zx_array = Zxyz_cv_array_to_Zx (npoly, array, x1);
		zx_ptrs = CALIB_NEWA (npoly, const struct calib_Zx_obj *);
		for (i = 0; i < npoly; i++) {
			zx_ptrs [i] = &zx_array [i];
		}
		Zx -> init (&zx_gcd);
		Zx -> gcd_n (&zx_gcd, NULL, npoly, zx_ptrs);
		Zxyz_cv_from_Zx (gcd, &zx_gcd, x1);
		Zx -> clear (&zx_gcd);
		free (zx_ptrs);
		Zxyz_free_Zx_array (zx_array, npoly);
	}
	else {
		/* Polynomials have at least 2 variables. */
		Zxyz_int_eezgcd_prim_polys (gcd,
					    npoly,
					    array,
					    nv,
					    r2fmap,
					    gip);
	}

	/* Form final gcd. */
	Zxyz_mul (gcd, gcd, &pcont);
	Zxyz_clear (&pcont);

	if (cofact NE NULL) {
		/* Do final division within Zxyz to get cofactors. */
		rgcd = Zxyz_rpoly_cv_f2r (gcd, nv, r2fmap, gip);
		for (i = 0; i < npoly; i++) {
			if (cf_ptrs [i] EQ NULL) continue;
			rp = Zxyz_rpoly_cv_f2r (poly_ptrs [i], nv, r2fmap, gip);
			qp = Zxyz_rpoly_div_exact (rp, rgcd, gip);
			CALIB_FATAL_ERROR_IF (qp EQ DIVISION_NOT_EXACT);
			Zxyz_rpoly_cv_r2f (cf_ptrs [i], qp, r2fmap);
			Zxyz_rpoly_free (qp, gip);
			Zxyz_rpoly_free (rp, gip);
		}
		Zxyz_rpoly_free (rgcd, gip);
	}

	Zxyz_adjust_gcd_n_result_signs (gcd, cf_ptrs, npoly);

	/* Clean up. */
	free (r2fmap);
	free (mindeg);
	for (i = 0; i < npoly; i++) {
		Zxyz_free (array [i]);
	}
	free (array);
	free (cf_ptrs);
	free (poly_ptrs);
}

/*
 * We choose to enforce the following convention about the Zxyz GCD
 * routines: The "leading coefficient" (lexicographically largest power
 * vector) shall have a strictly *positive* coefficient in the returned
 * GCD.  If this is not the case, this routine flips all the signs in
 * both the GCD and in each cofactor.
 */

	static
	void
Zxyz_adjust_gcd_n_result_signs (

struct calib_Zxyz_obj *		gcd,		/* IN/OUT: GCD result */
struct calib_Zxyz_obj **	cofact,		/* IN/OUT: Cofactor results */
int				npoly		/* IN: Number of cofactors */
)
{
int			j, n, sign;
struct calib_Zxyz_obj *	cfp;

	/* Note: the term having lexicographically largest var powers	*/
	/* is always the first term in the sequence.			*/
	n = gcd -> nterms;
	if (n <= 0) {
		CALIB_FATAL_ERROR_IF (n < 0);
		/* GCD is zero.  No sign needs to be flipped. */
		return;
	}

	/* GCD has at least 1 term.  Check sign of first term. */
	sign = mpz_sgn (&(gcd -> coeff [0]));
	if (sign >= 0) {
		CALIB_FATAL_ERROR_IF (sign EQ 0);

		/* Leading coefficient is positive.  Nothing to do. */
		return;
	}

	/* The leading coefficient is negative.  Negate all results. */
	Zxyz_neg (gcd, gcd);
	if (cofact NE NULL) {
		for (j = 0; j < npoly; j++) {
			cfp = cofact [j];
			if (cfp NE NULL) {
				Zxyz_neg (cfp, cfp);
			}
		}
	}
}

/*
 * Remove all integer contents from all of the given polynomials.
 * After this is done, compute the GCD of these contents.
 * All polynomials get individually reduced, even if they have no
 * common integer divisor > 1.
 */

	static
	void
Zxyz_remove_z_contents (

mpz_ptr			zcont,	/* IN/OUT: integer GCD of all given polys */
int			npoly,	/* IN: number of polynomials */
struct calib_Zxyz_obj **
			array	/* IN - array of polynomials to reduce */
)
{
int				i, j, nterms;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		p;
mpz_ptr				zp;
mpz_ptr				contents;
mpz_t				g;

	CALIB_FATAL_ERROR_IF ((zcont	 EQ NULL) OR
			      (npoly	 <= 0) OR
			      (array	 EQ NULL) OR
			      (array [0] EQ NULL));
	dom = array [0] -> dom;
	for (i = 1; i < npoly; i++) {
		p = array [i];
		CALIB_FATAL_ERROR_IF ((p EQ NULL) OR (p -> dom NE dom));
	}

	mpz_init (g);
	contents = calib_new_Z_vector (npoly);

	for (i = 0; i < npoly; i++) {
		p = array [i];
		CALIB_FATAL_ERROR_IF (p EQ NULL);
		nterms	= p -> nterms;
		zp	= p -> coeff;
		CALIB_FATAL_ERROR_IF (nterms <= 0);

		Zxyz_z_content (g, p);

		mpz_set (&contents [i], g);
		if (mpz_cmp_ui (g, 1) EQ 0) continue;
		/* Reduce this polynomial. */
		for (j = 0; j < nterms; j++) {
			mpz_divexact (&zp [j], &zp [j], g);
		}
	}

	/* Compute integer GCD across all polys. */
	mpz_set (g, &contents [0]);
	if (mpz_cmp_ui (g, 1) NE 0) {
		for (i = 1; i < npoly; i++) {
			mpz_gcd (g, g, &contents [i]);
			if (mpz_cmp_ui (g, 1) EQ 0) break;
		}
	}
	mpz_set (zcont, g);

	calib_free_Z_vector (contents, npoly);
	mpz_clear (g);
}

/*
 * Return TRUE if-and-only-if the given polynomial is constant.
 */

	static
	calib_bool
Zxyz_is_constant_poly (

const struct calib_Zxyz_obj *	poly	/* IN: polynomial to test */
)
{
int		i, k, nvars, nterms;
const int *	pow;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	nterms = poly -> nterms;

	if (nterms <= 0) {
		CALIB_FATAL_ERROR_IF (nterms NE 0);

		/* The zero polynomial is certainly constant. */
		return (TRUE);
	}

	if (nterms NE 1) return (FALSE);

	CALIB_FATAL_ERROR_IF (mpz_sgn (poly -> coeff) EQ 0);

	nvars	= poly -> dom -> nvars;
	pow	= poly -> pow;
	for (i = 0; i < nvars; i++) {
		k = pow [i];
		if (k > 0) return (FALSE);
		CALIB_FATAL_ERROR_IF (k < 0);
	}

	/* Only one term, all exponents are zero, and coefficient is	*/
	/* non-zero.  This is a "non-zero constant" polynomial.		*/

	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is exactly one.
 */

	static
	calib_bool
Zxyz_is_one_poly (

const struct calib_Zxyz_obj *	poly	/* IN: polynomial to test */
)
{
int		i, k, nvars;
const int *	pow;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	if (poly -> nterms NE 1) return (FALSE);

	if (mpz_cmp_ui (poly -> coeff, 1) NE 0) return (FALSE);

	nvars	= poly -> dom -> nvars;
	pow	= poly -> pow;
	for (i = 0; i < nvars; i++) {
		k = pow [i];
		if (k > 0) return (FALSE);
		CALIB_FATAL_ERROR_IF (k < 0);
	}

	/* Only one term, all exponents are zero, and coefficient is	*/
	/* one.  This is the "constant one" polynomial.			*/

	return (TRUE);
}

/*
 * Return TRUE if-and-only-if the given polynomial is exactly some
 * non-zero constant.
 */

	static
	calib_bool
Zxyz_is_non_zero_constant_poly (

const struct calib_Zxyz_obj *	poly	/* IN: polynomial to test */
)
{
int		i, k, nvars;
const int *	pow;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	if (poly -> nterms NE 1) return (FALSE);

	CALIB_FATAL_ERROR_IF (mpz_sgn (poly -> coeff) EQ 0);

	nvars	= poly -> dom -> nvars;
	pow	= poly -> pow;
	for (i = 0; i < nvars; i++) {
		k = pow [i];
		if (k > 0) return (FALSE);
		CALIB_FATAL_ERROR_IF (k < 0);
	}

	/* Only one term, all exponents are zero, and coefficient is	*/
	/* non-zero.  This is a non-zero, constant polynomial.		*/

	return (TRUE);
}

/*
 * Get the minimum and maximum degree of each variable in each polynomial.
 * The caller must free the returned "mindeg" array (which also frees maxdeg).
 */

	static
	void
Zxyz_get_poly_array_degrees (

int **				mindeg,	/* IN/OUT: min degrees */
int **				maxdeg,	/* IN/OUT: max degrees */
int				npoly,	/* IN: number of polynomials */
int				nvars,	/* IN: number of vars in each poly */
struct calib_Zxyz_obj **	array	/* IN: array of polynomials */
)
{
int			i, j, k, m, nterms;
int			*minarray, *maxarray, *minp, *maxp, *ip;
struct calib_Zxyz_obj *	p;

	CALIB_FATAL_ERROR_IF ((mindeg EQ NULL) OR
			      (maxdeg EQ NULL) OR
			      (npoly  <= 0) OR
			      (nvars  <= 0) OR
			      (array  EQ NULL));

	minarray = CALIB_NEWA (2 * npoly * nvars, int);
	maxarray = minarray + npoly * nvars;

	minp = minarray;
	maxp = maxarray;
	for (i = 0; i < npoly; i++) {
		p = array [i];
		CALIB_FATAL_ERROR_IF (p EQ NULL);
		nterms	= p -> nterms;
		ip	= p -> pow;
		CALIB_FATAL_ERROR_IF (nterms <= 0);
		/* Min degrees = Max degrees = degrees from first term. */
		memcpy (minp, ip, nvars * sizeof (int));
		memcpy (maxp, ip, nvars * sizeof (int));
		ip += nvars;
		for (j = 1; j < nterms; j++) {
			for (k = 0; k < nvars; k++) {
				m = *ip++;
				CALIB_FATAL_ERROR_IF (m < 0);
				if (m < minp [k]) {
					minp [k] = m;
				}
				if (m > maxp [k]) {
					maxp [k] = m;
				}
			}
		}
		minp += nvars;
		maxp += nvars;
	}

	*mindeg	= minarray;
	*maxdeg	= maxarray;
}

/*
 * Get the minimum and maximum degree of each variable in the given polynomial.
 * The caller must free the returned "mindeg" array (which also frees maxdeg).
 */

	static
	void
Zxyz_get_poly_degrees (

int **				mindeg,	/* IN/OUT: min degrees */
int **				maxdeg,	/* IN/OUT: max degrees */
const struct calib_Zxyz_obj *	poly	/* IN: polynomial to get degrees of */
)
{
int				j, k, m, nterms, nvars;
int				*minarray, *maxarray, *minp, *maxp, *ip;
const struct calib_Zxyz_dom *	dom;

	CALIB_FATAL_ERROR_IF ((mindeg EQ NULL) OR
			      (maxdeg EQ NULL) OR
			      (poly   EQ NULL));

	nterms	= poly -> nterms;
	dom	= poly -> dom;
	nvars	= dom -> nvars;

	minarray = CALIB_NEWA (2 * nvars, int);
	maxarray = minarray + nvars;

	minp = minarray;
	maxp = maxarray;

	if (nterms <= 0) {
		for (k = 0; k < nvars; k++) {
			minp [k] = 0;
			maxp [k] = 0;
		}
	}
	else {
		ip = poly -> pow;
		/* Min degrees = Max degrees = degrees from first term. */
		memcpy (minp, ip, nvars * sizeof (int));
		memcpy (maxp, ip, nvars * sizeof (int));
		ip += nvars;
		for (j = 1; j < nterms; j++) {
			for (k = 0; k < nvars; k++) {
				m = *ip++;
				CALIB_FATAL_ERROR_IF (m < 0);
				if (m < minp [k]) {
					minp [k] = m;
				}
				if (m > maxp [k]) {
					maxp [k] = m;
				}
			}
		}
	}

	*mindeg	= minarray;
	*maxdeg	= maxarray;
}

/*
 * Remove contents from each polynomial that consist of variables raised
 * to some power.  Once each polynomial is reduced in this fashion,
 * then also compute the GCD of these "variable contents" across all
 * polynomials.
 *
 * Let k = mindeg[var].  If k > 0, then var^k is a non-trivial factor
 * of the polynomial, which we remove.  We also subtract k from mindeg[var]
 * and from maxdeg[var] to keep these tables up-to-date with respect to
 * the reduced polynomials.
 */

	static
	void
Zxyz_remove_v_contents (

struct calib_Zxyz_obj *	vcont,		/* IN/OUT: variable content */
int			npoly,		/* IN: number of polynomials */
int			nvars,		/* IN: number of variables */
struct calib_Zxyz_obj **
			array,		/* IN/OUT: array of polynomials */
int *			mindeg,		/* IN/OUT: min degrees [poly,var] */
int *			maxdeg		/* IN/OUT: max degrees [poly,var] */
)
{
int			i, j, k, m, nterms;
calib_bool		reducible;
int			*vcarray, *minp, *maxp, *ip;
struct calib_Zxyz_obj *	p;

	CALIB_FATAL_ERROR_IF ((vcont  EQ NULL) OR
			      (npoly  <= 0) OR
			      (nvars  <= 0) OR
			      (array  EQ NULL) OR
			      (mindeg EQ NULL) OR
			      (maxdeg EQ NULL));

	/* Initialize vcarray to be the minimum degrees from the first	*/
	/* polynomial.							*/
	vcarray = CALIB_NEWA (nvars, int);
	memcpy (vcarray, mindeg, nvars * sizeof (int));

	minp = mindeg;
	maxp = maxdeg;
	for (i = 0; i < npoly; i++, minp += nvars, maxp += nvars) {
		p = array [i];
		CALIB_FATAL_ERROR_IF (p EQ NULL);

		/* See if this polynomial can be reduced. */
		reducible = FALSE;
		for (j = 0; j < nvars; j++) {
			k = minp [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			if (k < vcarray [j]) {
				vcarray [j] = k;
			}
			if (k > 0) {
				reducible = TRUE;
			}
		}
		if (NOT reducible) {
			/* This polynomial cannot be reduced. */
			continue;
		}

		/* Reduce each term of this polynomial. */
		nterms	= p -> nterms;
		ip	= p -> pow;
		for (j = 0; j < nterms; j++) {
			for (k = 0; k < nvars; k++) {
				m = *ip;
				m -= minp [k];
				CALIB_FATAL_ERROR_IF (m < 0);
				*ip++ = m;
			}
		}

		/* Now reduce the mindeg and maxdeg tables. */
		for (j = 0; j < nvars; j++) {
			m = minp [j];
			CALIB_FATAL_ERROR_IF (m < 0);
			minp [j] = 0;
			m = maxp [j] - m;
			CALIB_FATAL_ERROR_IF (m < 0);
			maxp [j] = m;
		}
	}

	/* Now construct (in vcont) a one-term polynomial having the	*/
	/* variable degrees listed in vcarray.				*/
	Zxyz_alloc (vcont, 1);
	vcont -> nterms = 1;
	memcpy (vcont -> pow, vcarray, nvars * sizeof (int));
	mpz_set_ui (&(vcont -> coeff [0]), 1);

	free (vcarray);
}

/*
 * Determine the set of variables that are actually present in the given
 * problem instance, and choose an ordering of those variables.
 *
 * Currently, this is simply by degree -- the variable of highest degree
 * (within any of the polynomials) becomes the "main" variable.
 * Subsequent ordering is by descending degree, and ties are resolved
 * arbitrarily using variable index as a second key.
 *
 * Note: we do not currently use the "array" argument.  It is being
 * passed in so that future versions can take into account the term
 * distribution of the polynomials if necessary.
 *
 * Further note: some authors recommend picking a main variable of
 * *MINIMUM* degree.  They argue that fewer integer evaluation points
 * will be needed before discoving a GCD of minimum degree.  While
 * this may be true, one can also argue that the later cost of p-adic
 * construction of the eliminated variables becomes vastly greater.
 * In my experience, it is better to take a few punches up front in
 * order to reduce the exponential behavior of p-adic construction of
 * the eliminated variables.  The variables constructed last are
 * particularly expensive, and should therefore be of lowest degree.
 */

	static
	int *
Zxyz_choose_var_order (

int *			nvars_out,	/* OUT: # vars actually present */
int			npoly,		/* IN: number of polynomials */
int			nvars,		/* IN: number of variables */
const struct calib_Zxyz_obj **
			array,		/* IN: array of polynomials */
const int *		mindeg,		/* IN: min degrees [poly,var] */
const int *		maxdeg		/* IN: max degrees [poly,var] */
)
{
int			i, j, k, nv, limit, next_limit;
int *			darray;
const int *		maxp;
int *			order;

	CALIB_FATAL_ERROR_IF ((npoly  <= 0) OR
			      (nvars  <= 0) OR
			      (array  EQ NULL) OR
			      (mindeg EQ NULL) OR
			      (maxdeg EQ NULL));

	order = CALIB_NEWA (nvars, int);

	/* Compute darray to be the maximum degree of each variable,	*/
	/* over all of the polynomials.					*/
	darray = CALIB_NEWA (nvars, int);
	memcpy (darray, maxdeg, nvars * sizeof (int));
	maxp = maxdeg;
	maxp += nvars;
	for (i = 1; i < npoly; i++, maxp += nvars) {
		for (j = 0; j < nvars; j++) {
			k = maxp [j];
			if (k > darray [j]) {
				darray [j] = k;
			}
		}
	}

	nv = 0;
	for (i = 0; i < nvars; i++) {
		if (darray [i] > 0) {
			order [nv++] = i;
		}
	}

	/* Don't puke -- we are using bubble sort here. */
	limit = nv;
	do {
		next_limit = 0;
		for (i = 1; i < limit; i++) {
			j = order [i - 1];
			k = order [i];
			if ((darray [j] > darray [k]) OR
			    ((darray [j] EQ darray [k]) AND (j > k))) {
				order [i - 1]	= k;
				order [i]	= j;
				next_limit = i;
			}
		}
		limit = next_limit;
	} while (limit > 0);

	free (darray);

	*nvars_out = nv;

	return (order);
}

/*
 * Determine the number of variables present in the given polynomial.
 */

	static
	int
Zxyz_num_vars_present (

const struct calib_Zxyz_obj *	poly		/* IN: polynomial to test */
)
{
int				i, j, nterms, nvars, nvp;
const struct calib_Zxyz_dom *	dom;
int *				ip;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	dom	= poly -> dom;
	nterms	= poly -> nterms;
	ip	= poly -> pow;

	nvars	= dom -> nvars;

	calib_bool varseen [nvars];
	memset (&varseen [0], FALSE, nvars * sizeof (calib_bool));

	nvp = 0;
	for (i = 0; i < nterms; i++) {
		for (j = 0; j < nvars; j++) {
			if ((ip [j] > 0) AND (NOT varseen [j])) {
				varseen [j] = TRUE;
				++nvp;
			}
		}
		ip += nvars;
	}
	return (nvp);
}

/*
 * Convert an array of Zxyz polynomials (which have only variable x1
 * remaining) into an array of Z[x] polynomials.
 */

	static
	struct calib_Zx_obj *
Zxyz_cv_array_to_Zx (

int			npoly,		/* IN: number of polynomials */
struct calib_Zxyz_obj **
			array,		/* IN: array of polys to convert */
int			x1		/* IN: sole var remaining */
)
{
int				i, nvars;
struct calib_Zx_obj *		zx_array;
struct calib_Zxyz_obj *		src;
struct calib_Zx_obj *		dst;
const struct calib_Zxyz_dom *	dom;

	zx_array = Zxyz_new_Zx_array (npoly);
	dst = zx_array;
	dom = NULL;
	nvars = 0;
	for (i = 0; i < npoly; i++, dst++) {
		src = array [i];
		if (dom EQ NULL) {
			dom = src -> dom;
			nvars = dom -> nvars;
			CALIB_FATAL_ERROR_IF ((x1 < 0) OR (x1 >= nvars));
		}
		else {
			CALIB_FATAL_ERROR_IF (src -> dom NE dom);
		}

		Zxyz_cv_poly_to_Zx_form (dst, src, x1);
	}
	return (zx_array);
}

/*
 * Convert a Zxyz polynomial (which has only variable x1 remaining)
 * into Z[x] form.
 */

	static
	void
Zxyz_cv_poly_to_Zx_form (

struct calib_Zx_obj *	result,		/* IN/OUT: resulting Z[x] poly */
const struct calib_Zxyz_obj *
			poly,		/* IN: Z[x,y,z] poly to convert */
int			x1		/* IN: sole var remaining */
)
{
int				i, j, k, nterms, nvars, deg;
const struct calib_Zxyz_dom *	dom;
const int *			pow;
mpz_srcptr			coeff;
int *				vardeg;
const struct calib_Zx_dom *	Zx;
mpz_ptr				zp;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (poly EQ NULL));
	dom	= poly -> dom;
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((x1 < 0) OR (x1 >= nvars));

	nterms	= poly -> nterms;
	pow	= poly -> pow;
	coeff	= poly -> coeff;

	/* Compute degrees of all vars. */
	vardeg = CALIB_NEWA (nvars, int);
	memset (vardeg, 0, nvars * sizeof (int));
	for (i = 0; i < nterms; i++) {
		for (j = 0; j < nvars; j++) {
			k = *pow++;
			if (k > vardeg [j]) {
				vardeg [j] = k;
			}
		}
	}

	/* Get degree of var x1, and verify that all other vars	*/
	/* have degree zero (i.e., are not present).		*/
	deg = vardeg [x1];
	vardeg [x1] = 0;
	for (k = 0; k < nvars; k++) {
		CALIB_FATAL_ERROR_IF (vardeg [k] NE 0);
	}
	free (vardeg);

	/* Make result big enough to hold a polynomial of this degree. */
	Zx = calib_get_Zx_dom ();
	Zx -> alloc (result, deg);

	zp  = result -> coeff;
	for (i = 0; i <= deg; i++) {
		mpz_set_ui (&zp [i], 0);
	}

	pow = poly -> pow;
	for (i = 0; i < nterms; i++, pow += nvars) {
		k = pow [x1];
		CALIB_FATAL_ERROR_IF ((k < 0) OR (k > deg));
		mpz_add (&zp [k], &zp [k], &coeff [i]);
	}
	result -> degree = deg;
}

/*
 * Convert the given Z[x] polynomial into Zxyz form, using the given
 * variable number.
 */

	static
	void
Zxyz_cv_from_Zx (

struct calib_Zxyz_obj *		result,	/* IN/OUT: resulting Zxyz poly */
const struct calib_Zx_obj *	src,	/* IN: source Z[x] poly */
int				x1	/* IN: variable to use */
)
{
int				i, nvars, nterms, deg;
const struct calib_Zxyz_dom *	dom;
mpz_srcptr			coeff;
int *				pow;
mpz_ptr				zp;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (src EQ NULL));
	dom = result -> dom;
	nvars = dom -> nvars;
	CALIB_FATAL_ERROR_IF ((x1 < 0) OR (x1 >= nvars));

	/* Count the number of non-zero terms in source poly. */
	deg	= src -> degree;
	coeff	= src -> coeff;
	nterms	= 0;
	for (i = 0; i <= deg; i++) {
		if (mpz_sgn (&coeff [i]) NE 0) {
			++nterms;
		}
	}
	/* Force destination arrays to be sufficiently large. */
	Zxyz_alloc (result, nterms);

	memset (result -> pow, 0, nterms * nvars * sizeof (int));
	pow = result -> pow;
	zp  = result -> coeff;
	for (i = deg; i >= 0; i--) {
		if (mpz_sgn (&coeff [i]) NE 0) {
			pow [x1] = i;
			mpz_set (zp, &coeff [i]);
			pow += nvars;
			++zp;
		}
	}
	result -> nterms = nterms;
}

/*
 * Compute the GCD of N *primitive* multi-variate polynomials using the
 * EEZ GCD algorithm.  The main variable x1 has been chosen, and the
 * coefficients are polynomials in at least 1 variable (although many
 * of these coefficients may be constant, they are not all constant).
 *
 * This routine does not return cofactors, because our caller does not
 * need them.
 */

	static
	void
Zxyz_int_eezgcd_prim_polys (

struct calib_Zxyz_obj *	gcd,	/* IN/OUT: the gcd */
int			npoly,	/* IN/OUT: number of polynomials */
struct calib_Zxyz_obj **
			array,	/* IN: array of polynomials to GCD */
int			nv,	/* IN: num vars in polys (includes x1) */
const int *		r2fmap,	/* IN: recursive to flat variable mapping */
struct GInfo *		gip	/* IN/OUT: global info */
)
{
int				i, j, nvars, nf, x1, max_gcd_deg;
int				iter;
calib_bool			bad, success;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		p;
mpz_ptr				bvals;
calib_bool *			evflags;
int *				x1_degree;
struct calib_Zxyz_obj		*upoly_array, **upoly_ptrs;
struct calib_Zxyz_obj		*ucf_array, **ucf_ptrs;
struct calib_Zx_obj		*zx_upoly_array, *zx_ucf_array, **zx_ucf_ptrs;
const struct calib_Zx_obj **	zx_upoly_ptrs;
const struct calib_Zx_dom *	Zx;
struct CFCache *		cfcache;
struct RPoly			*rdp, *rpp, *rqp;
struct calib_Zx_obj		zx_gcd;
struct calib_Zxyz_obj		ugcd, TEMP;
struct EEZglob			eez_global;

	CALIB_FATAL_ERROR_IF ((gcd    EQ NULL) OR
			      (npoly  < 0) OR
			      (array  EQ NULL) OR
			      (nv     <= 1) OR
			      (r2fmap EQ NULL) OR
			      (gip    EQ NULL) OR
			      (rand   EQ NULL));

	dom	= gcd -> dom;
	nvars	= dom -> nvars;
	for (i = 0; i < npoly; i++) {
		p = array [i];
		CALIB_FATAL_ERROR_IF ((p	   EQ NULL) OR
				      (p -> dom    NE dom) OR
				      (p -> nterms <= 0));
	}

	if (npoly EQ 0) {
		/* No polynomials.  GCD is always zero. */
		gcd -> nterms = 0;
		return;
	}
	if (npoly EQ 1) {
		/* One polynomial, which is the GCD. */
		Zxyz_set (gcd, array [0]);
		return;
	}

	/* Get the "main" variable, x1. */
	x1 = r2fmap [nv - 1];

	/* Determine the degree of each polynomial in x1.  We need	*/
	/* these to verify that the evaluation points do not decrease	*/
	/* the degrees of any of the polynomials.			*/
	x1_degree = CALIB_NEWA (npoly, int);
	for (i = 0; i < npoly; i++) {
		j = Zxyz_degree_in_var (array [i], x1);
		x1_degree [i] = j;
	}

	/* Compute an upper bound on the degree of the GCD, namely the	*/
	/* smallest degree among the polynomials (all of which have at	*/
	/* least one term, so they are not identically zero).		*/
	max_gcd_deg = INT_MAX;
	for (i = 0; i < npoly; i++) {
		j = x1_degree [i];
		CALIB_FATAL_ERROR_IF (j < 0);
		if (j < max_gcd_deg) {
			max_gcd_deg = j;
		}
	}

	/* Create buffer for the evaluation point and flags indicating	*/
	/* which variables we are evaluating away (all but x1).		*/
	bvals = calib_new_Z_vector (nvars);
	evflags = CALIB_NEWA (nvars, calib_bool);
	memset (evflags, FALSE, nvars * sizeof (evflags [0]));
	for (i = 0; i < nv - 1; i++) {
		j = r2fmap [i];
		evflags [j] = TRUE;
	}

	/* Initialize the global EEZ algorithm info. */
	nf = 2;				/* 2 factors: GCD and cofactor. */
	Zxyz_eez_init_global_info (&eez_global,
				   MODE_EEZGCD,
				   gip,
				   dom,
				   nv,
				   r2fmap,
				   evflags,
				   bvals,
				   nf);

	eez_global.poly		= NULL;		/* Selected below */

	/* Create buffers for the reduced (univariate) polynomials,	*/
	/* their GCD, and their cofactors (in both Zxyz and Zx form).	*/
	upoly_array	= Zxyz_new_array (dom, 2 * npoly);
	ucf_array	= upoly_array + npoly;
	upoly_ptrs	= CALIB_NEWA (2 * npoly, struct calib_Zxyz_obj *);
	ucf_ptrs	= upoly_ptrs + npoly;
	zx_upoly_array	= Zxyz_new_Zx_array (2 * npoly);
	zx_ucf_array	= zx_upoly_array + npoly;
	zx_ucf_ptrs	= CALIB_NEWA (2 * npoly, struct calib_Zx_obj *);
	zx_upoly_ptrs	= (const struct calib_Zx_obj **) (zx_ucf_ptrs + npoly);
	for (i = 0; i < npoly; i++) {
		upoly_ptrs [i]		= &upoly_array [i];
		ucf_ptrs [i]		= &ucf_array [i];
		zx_upoly_ptrs [i]	= &zx_upoly_array [i];
		zx_ucf_ptrs [i]		= &zx_ucf_array [i];
	}
	Zx = calib_get_Zx_dom ();
	Zx -> init (&zx_gcd);
	Zxyz_init (dom, &ugcd);

	/* Prepare an empty cache of info (computed on demand) about	*/
	/* each polynomial we are taking the GCD of.			*/
	cfcache = CALIB_NEWA (npoly, struct CFCache);
	memset (cfcache, 0, npoly * sizeof (cfcache [0]));

	/* Force first GCD to be accepted as a "new best degree".	*/
	++max_gcd_deg;

	iter = 0;
	for (;;) {
		++iter;
#if 0
		if ((iter & 0x07) EQ 0) {
			printf ("Zxyz_int_eezgcd_prim_polys: iter = %d\n", iter);
		}
#endif

		/* Choose a new evaluation point. */
		Zxyz_choose_evaluation_point (bvals, &eez_global);

		/* Each time we pick a new evaluation point, we need to	*/
		/* recompute the coefficient upper bound B, prime p,	*/
		/* modulus = p^k > 2*B, and all that depend on these.	*/
		mpz_set_ui (eez_global.modulus, 0);

		/* Evaluate each poly down to univariate, and verify	*/
		/* that their degrees do not decrease.  Also verify	*/
		/* that their trailing coefficients do not become	*/
		/* zero, since this messes up pre-determination of	*/
		/* trailing coefficients.  We should have gotten rid of	*/
		/* all the "v contents" by now, which are factors of	*/
		/* form var^k with k>=1, which means that the trailing	*/
		/* coefficient evaluated to zero from something that	*/
		/* was not zero.					*/
		Zxyz_init (dom, &TEMP);
		bad = FALSE;
		for (i = 0; i < npoly; i++) {
			Zxyz_eval_var_subset (&upoly_array [i],
					      array [i],
					      evflags,
					      bvals);
			j = Zxyz_degree_in_var (&upoly_array [i], x1);
			if (j NE x1_degree [i]) {
				bad = TRUE;
				break;
			}
			Zxyz_get_coeff (&TEMP, &upoly_array [i], x1, 0);
			if (TEMP.nterms <= 0) {
				CALIB_FATAL_ERROR_IF (TEMP.nterms < 0);
				/* Trailing coefficient became zero. */
				bad = TRUE;
				break;
			}
		}
		Zxyz_clear (&TEMP);
		if (bad) {
			/* This evaluation point was bad.  Try another. */
			continue;
		}

		/* Convert univariate polynomials into Z[x] form. */
		for (i = 0; i < npoly; i++) {
			Zxyz_cv_poly_to_Zx_form (&zx_upoly_array [i],
						 &upoly_array [i],
						 x1);
		}

		/* Compute GCD/cofactors of all the univariate polys. */
		Zx -> gcd_n (&zx_gcd, zx_ucf_ptrs, npoly, zx_upoly_ptrs);

		if (zx_gcd.degree <= 0) {
			CALIB_FATAL_ERROR_IF (zx_gcd.degree < 0);
			/* GCD of primitive polys is 1! */
			Zxyz_set_si (gcd, 1);
			break;
		}

		if (zx_gcd.degree > max_gcd_deg) {
			/* Degree too high.  Try another evaluation. */
			continue;
		}
		if (zx_gcd.degree < max_gcd_deg) {
			/* We have a new, lower degree GCD over Z[x]. */
			max_gcd_deg = zx_gcd.degree;
		}

		/* Suppose ucf_array [j] EQ 1.  With high probability,	*/
		/* array [j] is our GCD.  Do a division check to see	*/
		/* if this is so.  If the division check fails it means	*/
		/* that we have extraneous factors and we need to pick	*/
		/* a new evaluation point.				*/
		for (j = 0; j < npoly; j++) {
			if (zx_ucf_array [j].degree <= 0) {
				mpz_ptr		zp;
				CALIB_FATAL_ERROR_IF (zx_ucf_array [j].degree < 0);
				zp = &zx_ucf_array [j].coeff [0];
				if ((mpz_cmp_ui (zp, 1) EQ 0) OR
				    (mpz_cmp_si (zp, -1) EQ 0)) {
					break;
				}
			}
		}
		if (j < npoly) {
			bad = FALSE;
			rdp = Zxyz_rpoly_cv_f2r (array [j], nv, r2fmap, gip);
			for (i = 0; i < npoly; i++) {
				if (i EQ j) continue;
				rpp = Zxyz_rpoly_cv_f2r (array [i], nv, r2fmap, gip);
				rqp = Zxyz_rpoly_div_exact (rpp, rdp, gip);
				success = (rqp NE DIVISION_NOT_EXACT);
				Zxyz_rpoly_free (rqp, gip);
				Zxyz_rpoly_free (rpp, gip);
				if (NOT success) {
					bad = TRUE;
					break;
				}
			}
			Zxyz_rpoly_free (rdp, gip);
			if (bad) {
				/* Need a smaller GCD. */
				--max_gcd_deg;
				continue;
			}
			/* upoly_array [j] is our GCD! */
			Zxyz_set (gcd, array [j]);
			break;
		}

		/* Convert univariate GCD back into Z[x,y,z] form. */
		Zxyz_cv_from_Zx (&ugcd, &zx_gcd, x1);
		for (i = 0; i < npoly; i++) {
			Zxyz_cv_from_Zx (&ucf_array [i], &zx_ucf_array [i], x1);
		}

		j = Zxyz_choose_gcd_cofactor_to_lift (gcd,
						      &ugcd,
						      &zx_gcd,
						      npoly,
						      array,
						      upoly_array,
						      ucf_array,
						      zx_upoly_array,
						      zx_ucf_array,
						      cfcache,
						      &eez_global);
		if (j < 0) {
			if (j EQ BAD_EVAL_POINT) {
				/* Evaluation point does not pass the	*/
				/* tests needed to determine the	*/
				/* leading and/or trailing coefficient	*/
				/* of the GCD polynomial.  Try another	*/
				/* evaluation point.			*/
				continue;
			}
			/* We have detected extraneous factors.  We	*/
			/* need a GCD of lower degree.			*/
			--max_gcd_deg;
			if (max_gcd_deg <= 0) {
				/* The GCD must be 1. */
				Zxyz_set_si (gcd, 1);
				break;
			}
			/* Try another evaluation point. */
			continue;
		}
		CALIB_FATAL_ERROR_IF (j >= npoly);

		/* We lifted a factorization of polynomial j, meaning	*/
		/* that gcd*cofactor = array [j] over Z[x,y,z].  Now do	*/
		/* the trial division of all the other polynomials by	*/
		/* this gcd.						*/

		rdp = Zxyz_rpoly_cv_f2r (gcd, nv, r2fmap, gip);

		success = TRUE;
		for (i = 0; i < npoly; i++) {
			if (i EQ j) continue;

			rpp = Zxyz_rpoly_cv_f2r (array [i], nv, r2fmap, gip);
			rqp = Zxyz_rpoly_div_exact (rpp, rdp, gip);

			success = (rqp NE DIVISION_NOT_EXACT);

			Zxyz_rpoly_free (rqp, gip);
			Zxyz_rpoly_free (rpp, gip);

			if (NOT success) break;
		}
		Zxyz_rpoly_free (rdp, gip);

		if (success) {
			/* We have found the true GCD! */
			break;
		}

		/* One or more of the polynomials did not	*/
		/* divide evenly by the current GCD.  This	*/
		/* means that the degree of the true GCD must	*/
		/* be lower.					*/
		--max_gcd_deg;
		if (max_gcd_deg <= 0) {
			/* True GCD must be 1. */
			Zxyz_set_si (gcd, 1);
			break;
		}

		/* Try another evaluation point, looking for a	*/
		/* smaller GCD.					*/
	}

	/* Clean up. */
	for (i = 0; i < npoly; i++) {
		Zxyz_free_factors (cfcache [i].lcf);
		Zxyz_free_factors (cfcache [i].tcf);
	}
	free (cfcache);
	Zxyz_clear (&ugcd);
	Zx -> clear (&zx_gcd);
	free (zx_ucf_ptrs);
	Zxyz_free_Zx_array (zx_upoly_array, 2 * npoly);
	free (upoly_ptrs);
	Zxyz_free_array (upoly_array, 2 * npoly);
	Zxyz_eez_clear_global_info (&eez_global);
	free (evflags);
	calib_free_Z_vector (bvals, nvars);
	free (x1_degree);
}

/*
 * Return the degree of the given polynomial in the given variable.
 */

	static
	int
Zxyz_degree_in_var (

const struct calib_Zxyz_obj *
			poly,	/* IN: polynomial to get degree of */
int			var	/* IN: variable to get degree of */
)
{
int				i, k, nvars, nterms, degree;
const struct calib_Zxyz_dom *	dom;
const int *			ip;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);
	dom	= poly -> dom;
	CALIB_FATAL_ERROR_IF (dom EQ NULL);
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));
	nterms	= poly -> nterms;
	ip	= poly -> pow;
	degree = -1;
	for (i = 0; i < nterms; i++) {
		k = ip [var];
		CALIB_FATAL_ERROR_IF (k < 0);
		if (k > degree) {
			degree = k;
		}
		ip += nvars;
	}
	return (degree);
}

/*
 * Return the total degree of the given polynomial.
 */

	static
	int
Zxyz_total_degree (

const struct calib_Zxyz_obj *	poly	/* IN: polynomial to get total degree of */
)
{
int				i, j, k, nvars, nterms, degree, sum;
const struct calib_Zxyz_dom *	dom;
const int *			ip;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);
	dom	= poly -> dom;
	nvars	= dom -> nvars;
	nterms	= poly -> nterms;
	ip	= poly -> pow;
	degree = -1;
	for (i = 0; i < nterms; i++) {
		sum = 0;
		for (j = 0; j < nvars; j++) {
			k = *ip++;
			CALIB_FATAL_ERROR_IF (k < 0);
			sum += k;
		}
		if (sum > degree) {
			degree = sum;
		}
	}
	return (degree);
}

/*
 * Choose the next evaluation point.  We generate random integers
 * which slowly increase in magnitude as more attempts are needed
 * to find a suitable evaluation point.
 */

	static
	void
Zxyz_choose_evaluation_point (

mpz_ptr			bvals,		/* IN/OUT: evaluation point */
struct EEZglob *	eezgip		/* IN/OUT: EEZ global info */
)
{
int			i, j, var;
int			tries, nbytes, nwords, nv;
unsigned int		word, mask;
struct GInfo *		gip;
const int *		r2fmap;
struct calib_Random *	rand;
mpz_t 			REG;

	CALIB_FATAL_ERROR_IF ((bvals EQ NULL) OR (eezgip EQ NULL));

	gip	= eezgip -> gip;
	tries	= eezgip -> pg_tries;
	nbytes	= eezgip -> pg_nbytes;
	nv	= eezgip -> nv;
	r2fmap	= eezgip -> r2fmap;
	rand	= &(gip -> rng);

	++tries;
	if (tries > 10) {
		/* Time to increase magnitude of the integers we choose. */
		++nbytes;
		eezgip -> pg_nbytes = nbytes;
	}
	eezgip -> pg_tries = tries;

	nwords = (nbytes >> 2);
	nbytes &= 0x03;
	mask = 1;
	mask <<= (8 * nbytes);
	--mask;

	mpz_init (REG);

	for (i = 0; i < nv; i++) {
		var = r2fmap [i];
		word = 0;
		if (nbytes > 0) {
			word = calib_random_u32 (rand);
			word &= mask;
		}
		mpz_set_ui (REG, word);
		for (j = 0; j < nwords; j++) {
			word = calib_random_u32 (rand);
			mpz_mul_2exp (REG, REG, 32);
			mpz_add_ui (REG, REG, word);
		}
		mpz_set (&bvals [var], REG);
	}

	mpz_clear (REG);
}

/*
 * Choose an appropriate polynomial j -- we need the following:
 *
 *		gcd (ucf_array [j], ugcd) = 1
 *
 * If none of the polynomials satisfy this condition, then we use an
 * alternate technique (implemented in another function).
 *
 * Assuming we have found a suitable univariate (gcd,cofactor) pair,
 * we apply Wang's EEZ algorithm (which can fail at the point either
 * because an unlucky evaluation point does not permit leading or
 * trailing coefficients to be found, or because of extraneous factors
 * (i.e., GCD is of too high a degree)).
 *
 * Returns BAD_EVAL_POINT, EXTRANEOUS_FACTORS (both < 0), or the index
 * of a cofactor that was successfully lifted (>= 0).
 */

	static
	int
Zxyz_choose_gcd_cofactor_to_lift (

struct calib_Zxyz_obj *	gcd,		/* IN/OUT: GCD over Z[x,y,z] */
const struct calib_Zxyz_obj *
			ugcd,		/* IN: univariate GCD */
const struct calib_Zx_obj *
			zx_gcd,		/* IN: univariate GCD in Z[x] form */
int			npoly,		/* IN: number of polynomials */
struct calib_Zxyz_obj **
			poly_ptrs,	/* IN: original polynomials */
const struct calib_Zxyz_obj *
			upoly_array,	/* IN: array of univariate polys */
const struct calib_Zxyz_obj *
			ucf_array,	/* IN: array of univariate cofactors */
const struct calib_Zx_obj *
			zx_upoly_array,	/* IN: univariate polys in Z[x] form */
const struct calib_Zx_obj *
			zx_ucf_array,	/* IN: cofactors in Z[x] form */
struct CFCache *	cfcache,	/* IN: coefficient factorization cache */
struct EEZglob *	eezgip		/* IN/OUT: EEZ algorithm global info */
)
{
int				j, status;
calib_bool			found;
const struct calib_Zxyz_dom *	dom;
const struct calib_Zx_dom *	Zx;
struct GInfo *			gip;
struct calib_Zxyz_factor	*sqfp, *fp;
struct calib_Zxyz_obj **	ptrs;
struct calib_Zx_obj		G, A, B;
struct calib_Zxyz_obj		TEMP2, TEMP3, TEMP4;

	CALIB_FATAL_ERROR_IF ((gcd	   EQ NULL) OR
			      (ugcd	   EQ NULL) OR
			      (npoly	   < 2) OR
			      (upoly_array EQ NULL) OR
			      (ucf_array   EQ NULL) OR
			      (cfcache	   EQ NULL) OR
			      (eezgip	   EQ NULL));

	dom = gcd -> dom;

	gip	= eezgip -> gip;

	Zx = calib_get_Zx_dom ();
	Zx -> init (&G);
	Zx -> init (&A);
	Zx -> init (&B);

	found = FALSE;
	for (j = 0; j < npoly; j++) {
		Zx -> extgcd (&G, &A, &B, zx_gcd, &zx_ucf_array [j]);
		if (G.degree EQ 0) {
			found = TRUE;
			break;
		}
	}

	if (found) {
		/* We have A*zx_gcd + B*zx_ucf_array [j] = G, where	*/
		/* G \in Z.						*/

		/* Set up the two univariate factors we want to lift,	*/
		/* together with the multi-variate polynomial that they	*/
		/* should equal when multiplied out.			*/

		eezgip -> poly	= poly_ptrs [j];
		Zxyz_set (&(eezgip -> factors [0]), ugcd);
		Zxyz_set (&(eezgip -> factors [1]), &ucf_array [j]);

		/* Do the main EEZ algorithm. */
		status = Zxyz_eez_main (&cfcache [j], eezgip);

		eezgip -> poly = NULL;

		if (status EQ SUCCESS) {
			/* Copy lifted GCD factor back to caller. */
			Zxyz_set (gcd, &(eezgip -> pdfactors [0]));
			status = j;
		}
		else {
			/* Status is BAD_EVAL_POINT or EXTRANEOUS_FACTORS. */
		}
	}
	else {

		/* None of the cofactors works with this GCD, due to	*/
		/* the presence of numerous factors of multiplicity	*/
		/* > 1.  The following code works around this special	*/
		/* case by performing a square-free factorization of	*/
		/* one of the input polynomials (we choose the first,	*/
		/* since it is of minimum degree), and performing a	*/
		/* separate GCD over each such factor.  There are more	*/
		/* efficient ways to do this, but in our application	*/
		/* this case will occur rarely (if ever) -- we choose	*/
		/* simplicity over efficiency.				*/

		sqfp = Zxyz_compute_square_free_factorization (poly_ptrs [0],
							       gip);

		ptrs = CALIB_NEWA (npoly, struct calib_Zxyz_obj *);
		memcpy (ptrs, poly_ptrs, npoly * sizeof (ptrs [0]));
		Zxyz_init (dom, &TEMP2);
		Zxyz_init (dom, &TEMP3);
		Zxyz_init_si (dom, &TEMP4, 1);
		for (fp = sqfp; fp NE NULL; fp = fp -> next) {
			/* Polynomial should be primitive, so if there	*/
			/* is a constant factor, it should be -1.	*/
			CALIB_FATAL_ERROR_IF (fp -> factor -> nterms <= 0);
			if (Zxyz_is_non_zero_constant_poly (fp -> factor)) {
				CALIB_FATAL_ERROR_IF (mpz_cmp_si (&(fp -> factor -> coeff [0]), -1) NE 0);
				/* Skip -1 units. */
				continue;
			}
			Zxyz_ipow (&TEMP2, fp -> factor, fp -> multiplicity);
			ptrs [0] = &TEMP2;
			Zxyz_internal_eezgcd_n (&TEMP3,
						NULL,
						npoly,
						(const struct calib_Zxyz_obj **) ptrs,
						gip);

			Zxyz_mul (&TEMP4, &TEMP4, &TEMP3);
		}
		Zxyz_set (gcd, &TEMP4);

		Zxyz_clear (&TEMP4);
		Zxyz_clear (&TEMP3);
		Zxyz_clear (&TEMP2);
		free (ptrs);
		Zxyz_free_factors (sqfp);
		status = SUCCESS;
	}

	Zx -> clear (&B);
	Zx -> clear (&A);
	Zx -> clear (&G);

	return (status);
}

/*
 * The main portion of the EEZ algorithm, whose objective is to lift the
 * given univariate vactors to be full multi-variate factors over Z, or
  report that this is impossible.  The process consists of three major
 * steps:
 *
 * 1. Pre-determine the correct leading (and trailing) coefficient of
 *    each factor.
 *
 * 2. [Optional]  Pre-determine additional factor coefficients, if possible.
 *
 * 3. Use the EEZ construction to lift each factor to its full multi-variate
 *    form.
 *
 * There are two ways this process might fail:
 *
 * a. The current evaluation point might not have the proprties necessary
 *    for pre-determining the leading (or trailing) coefficients of each
 *    factor.
 *
 * b. Two or more of the given univariate factors might be extraneous,
 *    i.e., there is no "true factor" of the given original multi-variate
 *    polynomial that (when evaluated at the current evaluation point)
 *    corresponds exactly to the given univariate factors.
 *
 * Returns SUCCESS, BAD_EVAL_POINT or EXTRANEOUS_FACTORS.
 */

	static
	int
Zxyz_eez_main (

struct CFCache *	cfcache,	/* IN/OUT: cached coefficient factors */
struct EEZglob *	eezgip		/* IN/OUT: global EEZ info */
)
{
int				i, k, nf, x1, degree, total_degree, status;
calib_bool			success;
const struct calib_Zxyz_obj *	poly;
struct calib_Zxyz_obj *		poly2;
const struct calib_Zxyz_dom *	dom;
struct GInfo *			gip;
struct calib_Zxyz_obj *		factors;
struct calib_Zxyz_obj *		pdfactors;
struct calib_Zxyz_obj *		factor;
struct calib_Zxyz_factor *	lcf;
struct calib_Zxyz_factor *	tcf;
mpz_ptr				factor_lc_values;
struct calib_Zxyz_obj *		factor_lc_array;
mpz_ptr				factor_tc_values;
struct calib_Zxyz_obj *		factor_tc_array;
mpz_ptr				zp;
mpz_ptr				zcont, dpart1, dpart2, dpart3;
struct calib_Zxyz_obj		TEMP;
mpz_t				delta, distributed_delta;
mpz_t				C1, GCD, LCM;

	x1		= eezgip -> x1;
	poly		= eezgip -> poly;
	nf		= eezgip -> nf;
	factors		= eezgip -> factors;
	pdfactors	= eezgip -> pdfactors;
	gip		= eezgip -> gip;

	/* Get the total degree of the polynomial we are factoring. */
	total_degree = Zxyz_total_degree (eezgip -> poly);
	eezgip -> total_degree = total_degree;

	/* Make a locally modifiable copy of it too. */
	poly2 = Zxyz_dup (poly);
	eezgip -> poly2 = poly2;

	dom = poly -> dom;

	factor_lc_values = NULL;
	factor_lc_array	 = NULL;
	factor_tc_values = NULL;
	factor_tc_array	 = NULL;

	zp = calib_new_Z_vector (3 * nf);
	dpart1	= zp;	zp += nf;
	dpart2	= zp;	zp += nf;
	dpart3	= zp;

	mpz_init (delta);
	mpz_init (distributed_delta);
	mpz_init (GCD);
	mpz_init (LCM);
	mpz_init (C1);
	zcont = calib_new_Z_vector (nf);
	Zxyz_init (dom, &TEMP);

	/* Compute pdfactors[i] = z_pp(factors[i]), and			*/
	/* zcont[i] = z_content(factors[i]).  In the MODE_EEZGCD case	*/
	/* our GCD and cofactor may both have integer contents that	*/
	/* need to be removed here.					*/
	/* Let delta be the product of the zcont[i].			*/
	mpz_set_ui (delta, 1);
	for (i = 0; i < nf; i++) {
		Zxyz_set (&pdfactors [i], &factors [i]);
		Zxyz_strip_z_content (&zcont [i], &pdfactors [i]);
		mpz_mul (delta, delta, &zcont [i]);
	}

	degree = Zxyz_degree_in_var (poly, x1);

	do {		/* Used only for "break". */

		/* Get complete factorization of the lead coefficient	*/
		/* (might have computed this already and cached it).	*/
		if (cfcache NE NULL) {
			lcf = cfcache -> lcf;
			if (NOT (cfcache -> lvalid)) {
				/* Factor the leading coefficient. */
				lcf = Zxyz_factor_poly_coeff (poly, x1, degree, gip);
				cfcache -> lcf	  = lcf;
				cfcache -> lvalid = TRUE;
			}
		}
		else {		/* No cache being used.  Must compute it. */
			lcf = Zxyz_factor_poly_coeff (poly, x1, degree, gip);
		}

		factor_lc_values = calib_new_Z_vector (nf);
		for (i = 0; i < nf; i++) {
			factor = &pdfactors [i];
			k = Zxyz_degree_in_var (factor, x1);
			Zxyz_get_coeff (&TEMP, factor, x1, k);
			CALIB_FATAL_ERROR_IF (NOT Zxyz_is_non_zero_constant_poly (&TEMP));
			mpz_set (&factor_lc_values [i], &TEMP.coeff [0]);
		}
		factor_lc_array = Zxyz_new_array (dom, nf);

		/* Partition the symbolic factors of the lead		*/
		/* coefficient correctly across the lead coefficients	*/
		/* of the factors.					*/
		success = Zxyz_determine_coeffs (factor_lc_array,
						 lcf,
						 delta,
						 factor_lc_values,
						 dpart1,
						 eezgip);
		if (NOT success) {
			/* Could not determine lead coefficients. */
			status = BAD_EVAL_POINT;
			break;
		}

		/* Get complete factorization of the trailing		*/
		/* coefficient (might have computed this already and	*/
		/* cached it).						*/
		if (cfcache NE NULL) {
			tcf = cfcache -> tcf;
			if (NOT (cfcache -> tvalid)) {
				/* Factor the trailing coefficient. */
				tcf = Zxyz_factor_poly_coeff (poly, x1, 0, gip);
				cfcache -> tcf	  = tcf;
				cfcache -> tvalid = TRUE;
			}
		}
		else {		/* No cache being used.  Must compute it. */
			tcf = Zxyz_factor_poly_coeff (poly, x1, 0, gip);
		}

		factor_tc_values = calib_new_Z_vector (nf);
		for (i = 0; i < nf; i++) {
			factor = &pdfactors [i];
			Zxyz_get_coeff (&TEMP, factor, x1, 0);
			CALIB_FATAL_ERROR_IF (NOT Zxyz_is_non_zero_constant_poly (&TEMP));
			mpz_set (&factor_tc_values [i], &TEMP.coeff [0]);
		}
		factor_tc_array = Zxyz_new_array (dom, nf);

		/* Partition the symbolic factors of the trailing	*/
		/* coefficient correctly across the trailing		*/
		/* coefficients	of the factors.				*/
		success = Zxyz_determine_coeffs (factor_tc_array,
						 tcf,
						 delta,
						 factor_tc_values,
						 dpart2,
						 eezgip);

		if (NOT success) {
			/* Could not determine trailing coefficients. */
			status = BAD_EVAL_POINT;
			break;
		}

		/* Let our given polynomial be:				*/
		/*							*/
		/*	U(x, x2, ..., xt) = Vn*x^n + ... + V1*x + V0	*/
		/*							*/
		/* which is primitive and square-free, and factors as:	*/
		/*							*/
		/*	U = prod (G_i (x, x2, ..., xt), i, 1, r)	*/
		/*							*/
		/* Let U0(x) = U(x, b2, b3, ..., bt) be the univariate	*/
		/* reduction (with b2, ..., bt being the chosen integer	*/
		/* bvals to substitute for the eliminated vars).  Now	*/
		/* factor						*/
		/*							*/
		/*	U0(x) = delta * u1(x) * ... * ur(x)		*/
		/*							*/
		/* over Z[x], with delta \in Z.  Note that for all	*/
		/* 1 <= i <= r,						*/
		/*							*/
		/*	G_i (x, b2, b3, ..., bt) = delta_i * ui(x)	*/
		/*							*/
		/* and that						*/
		/*							*/
		/*	delta = delta_1 * delta_2 * ... * delta_r	*/
		/*							*/
		/* We don't necessarily know the individual delta_i	*/
		/* values for certain at this point, but clearly each	*/
		/* delta_i *belongs* to a unique factor G_i!  delta_i	*/
		/* cannot belong to factor 1 in the leading		*/
		/* coefficient, and switch to factor 3 in the trailing	*/
		/* coefficient!  In particular dpart1[i] is a LOWER	*/
		/* BOUND on delta_i imposed by the factorization of the	*/
		/* leading coefficient Vn.  Similarly, dpart2[i] is	*/
		/* another LOWER BOUND on delta_i, this time imposed by	*/
		/* the factorization of the trailing coefficient V0.	*/
		/* We combine lower bounds dpart1[i] and dpart2[i] into	*/
		/* dpart3[i] by computing their LCM.  Also compute the	*/
		/* total portion of delta that has been unambiguously	*/
		/* distributed among the factors.			*/

		mpz_set_ui (distributed_delta, 1);
		for (i = 0; i < nf; i++) {
			mpz_gcd (GCD, &dpart1 [i], &dpart2 [i]);
			mpz_mul (LCM, &dpart1 [i], &dpart2 [i]);
			mpz_divexact (LCM, LCM, GCD);
			mpz_set (&dpart3 [i], LCM);
			mpz_mul (distributed_delta, distributed_delta, LCM);
		}

		/* The dpart3[i] should now all divide the delta_i	*/
		/* (whose exact values may still not be clear).		*/

		if (mpz_cmp (distributed_delta, delta) > 0) {
			/* The distribution of the various delta_i	*/
			/* values between the factors of Vn (lc(U))	*/
			/* and V0 (tc(U)) is not compatible!  This	*/
			/* should not happen unless we have extraneous	*/
			/* factors.					*/
#if 1
			/* Crash for now.  We want to know if this can	*/
			/* really happen.				*/
			CALIB_FATAL_ERROR;
#endif
			/* (FIXME? downgrade to BAD_EVAL_POINT?)	*/
			status = EXTRANEOUS_FACTORS;
			break;
		}

		/* Impose dpart3[i] on each pdfactor[i]. */
		for (i = 0; i < nf; i++) {
			factor = &pdfactors [i];
			Zxyz_mul_z (factor, factor, &dpart3 [i]);
		}

		/* The dpart1[i] and dpart2[i] have already been	*/
		/* imposed upon their respective coefficients of	*/
		/* factor i.  Let C1 = dpart3[i] / dpart1[i].  C1 is a	*/
		/* factor that we impose in the leading coefficients	*/
		/* now.  Ditto for C1 = dpart3[i] / dpart2[i] in the	*/
		/* trailing coefficients.				*/

		for (i = 0; i < nf; i++) {
			mpz_divexact (C1, &dpart3 [i], &dpart1 [i]);
			factor = &factor_lc_array [i];
			Zxyz_mul_z (factor, factor, C1);

			mpz_divexact (C1, &dpart3 [i], &dpart2 [i]);
			factor = &factor_tc_array [i];
			Zxyz_mul_z (factor, factor, C1);
		}

		/* Now determine the remaining constant factor that has	*/
		/* yet to be distributed.				*/

		mpz_divexact (C1, delta, distributed_delta);

		/* Impose one copy of C1 upon:				*/
		/*							*/
		/*	- Each pdfactors[i]				*/
		/*	- Each leading coefficient i,			*/
		/*	- Each trailing coefficient i.			*/
		for (i = 0; i < nf; i++) {
			factor = &factor_lc_array [i];
			Zxyz_mul_z (factor, factor, C1);

			factor = &factor_tc_array [i];
			Zxyz_mul_z (factor, factor, C1);

			factor = &pdfactors [i];
			Zxyz_mul_z (factor, factor, C1);
		}

		/* Set poly2 = C1**(nf-1) * poly.  This is where poly2	*/
		/* may diverge from poly (if C1 NE 1).			*/
		mpz_pow_ui (C1, C1, nf - 1);
		Zxyz_mul_z (poly2, poly2, C1);

#if 0
		/* FIXME -- there are no end to the list of things that	*/
		/* are clearly wrong with this switch statement:	*/
		/*							*/
		/* 1. The FACTOR (for which nf >= 2) case currently	*/
		/*    falls into the EEZGCD case, which immediately	*/
		/*    fatals if (nf NE 2).				*/
		/* 2. Provide an example of what we are correcting for	*/
		/*    in the EEZGCD case.				*/

		mpz_t		C2;
		switch (eezgip -> mode) {
		case MODE_FACTOR:
 #if 1
			/* Constant factor has been imposed upon the	*/
			/* first factor -- treat just like EEZGCD.	*/
 #else
			/* The constants should not shift back and	*/
			/* forth between the factors in this case.	*/
			for (i = 0; i < nf; i++) {
				if (mpz_cmp (&zcont [i], &dpart1 [i]) NE 0) {
					success = FALSE;
					break;
				}
			}
			break;
 #endif

		case MODE_EEZGCD:
			/* In this case, the GCD (which was computed in	*/
			/* the univariate space) can "steal" integer	*/
			/* contents that do not belong to the multi-	*/
			/* variate GCD.  Shift them back if necessary	*/
			CALIB_FATAL_ERROR_IF (nf NE 2);
			if ((mpz_cmp (&zcont [0], &dpart1 [0]) EQ 0) AND
			    (mpz_cmp (&zcont [1], &dpart1 [0]) EQ 0)) {
				/* No shifting occurred. */
				break;
			}
			/* Verify that a constant factor moved from	*/
			/* factor 0 (the GCD) to factor 1 (cofactor).	*/
			mpz_init (C2);
			mpz_divexact (C1, &zcont [0], &dpart1 [0]);
			mpz_divexact (C2, &dpart1 [1], &zcont [1]);
			success = (mpz_cmp (C1, C2) EQ 0);
			mpz_clear (C2);
			break;

		default:
			CALIB_FATAL_ERROR;
		}
		if (NOT success) {
			status = BAD_EVAL_POINT;
			break;
		}
#endif

		/* Apply the correct leading and trailing coefficient	*/
		/* to each factor.					*/
		for (i = 0; i < nf; i++) {
			factor = &pdfactors [i];

			k = Zxyz_degree_in_var (factor, x1);
			Zxyz_drop_coeff (factor, x1, k);
			Zxyz_set (&TEMP, &factor_lc_array [i]);
			Zxyz_mul_var_pow_k (&TEMP, x1, k);
			Zxyz_add (factor, factor, &TEMP);

			Zxyz_drop_coeff (factor, x1, 0);
			Zxyz_add (factor, factor, &factor_tc_array [i]);
		}

		/* Compute the array of successively evaluated versions	*/
		/* of poly2.						*/
		Zxyz_eez_compute_evpoly (eezgip);

		/* We have leading and trailing coefficients determined	*/
		/* (up to constant factors) for all the factors.	*/
		/* Do the EEZ lifting now.				*/

		status = Zxyz_eez_lift_factors (eezgip);

		if (status NE SUCCESS) break;

		/* Remove integer contents from our factors.  Any that	*/
		/* are present were introduced to handle the "leading	*/
		/* coefficient problem."				*/
		for (i = 0; i < nf; i++) {
			Zxyz_strip_z_content (delta, &pdfactors [i]);
		}

	} while (FALSE);

	Zxyz_clear (&TEMP);
	calib_free_Z_vector (zcont, nf);
	mpz_clear (C1);
	mpz_clear (LCM);
	mpz_clear (GCD);
	mpz_clear (distributed_delta);
	mpz_clear (delta);

	calib_free_Z_vector (dpart1, 3 * nf);

	calib_free_Z_vector (factor_lc_values, nf);
	if (factor_lc_array NE NULL) {
		Zxyz_free_array (factor_lc_array, nf);
	}
	calib_free_Z_vector (factor_tc_values, nf);
	if (factor_tc_array NE NULL) {
		Zxyz_free_array (factor_tc_array, nf);
	}

	Zxyz_free (eezgip -> poly2);
	eezgip -> poly2 = NULL;

	return (status);
}

/*
 * Compute the array evpoly, such that evpoly [0] is univariate,
 * evpoly [nv] is the original multi-variate polynomial, and the
 * intervening entries each have one variable successively eliminated.
 */

	static
	void
Zxyz_eez_compute_evpoly (

struct EEZglob *	eezgip		/* IN: EEZ algorithm global info */
)
{
int				i, j, nv, nvars;
const int *			r2fmap;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		evpoly;
mpz_srcptr			bvals;
struct calib_Zxyz_obj		POLY;

	CALIB_FATAL_ERROR_IF (eezgip EQ NULL);

	dom	= eezgip -> dom;
	nv	= eezgip -> nv;
	r2fmap	= eezgip -> r2fmap;
	bvals	= eezgip -> bvals;
	nvars	= dom -> nvars;

	/* Create the evpoly array if it does not already exist.	*/
	evpoly = eezgip -> evpoly;
	if (evpoly EQ NULL) {
		evpoly = Zxyz_new_array (dom, nv + 1);
		eezgip -> evpoly = evpoly;

		/* The univariate polynomial winds up being the first one. */
		eezgip -> upoly	 = &evpoly [0];
	}

	calib_bool varseen [nvars];
	memset (varseen, FALSE, nvars * sizeof (calib_bool));

	/* Make a local copy of the polynomial. */
	Zxyz_init (dom, &POLY);
	Zxyz_set (&POLY, eezgip -> poly2);

	/* Set final (unevaluated) polynomial. */
	Zxyz_set (&evpoly [nv], &POLY);

	for (i = 0; i < nv; i++) {
		/* Get next var to eliminate. */
		j = r2fmap [i];

		CALIB_FATAL_ERROR_IF ((j < 0) OR
				      (j >= nvars) OR
				      (varseen [j]));
		varseen [j] = TRUE;

		Zxyz_eval_one_var (&POLY, &POLY, j, &bvals [j]);
		Zxyz_set (&evpoly [nv - i - 1], &POLY);
	}
	Zxyz_clear (&POLY);
}

/*
 * Let Y be the coefficient of x1^k within the given polynomial.
 * Return the complete factorization of Y.
 */

	static
	struct calib_Zxyz_factor *
Zxyz_factor_poly_coeff (

const struct calib_Zxyz_obj *	poly,	/* IN: polynomial */
int				x1,	/* IN: variable */
int				k,	/* IN: power of desired coefficient */
struct GInfo *			gip	/* IN/OUT: global info */
)
{
int				nvars;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_factor *	factors;
struct calib_Zxyz_obj		COEFF;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (x1 < 0));
	dom	= poly -> dom;
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF (x1 >= nvars);

	/* Get the desired coefficient. */
	Zxyz_init (dom, &COEFF);
	Zxyz_get_coeff (&COEFF, poly, x1, k);

	factors = Zxyz_internal_factor (&COEFF, gip, FMODE_FULL);

	Zxyz_clear (&COEFF);

	return (factors);
}

/*
 * Let U(x,x2,x3,...,xt) be the primitive, (not necessarily square-free)
 * polynomial to be factored, and G(x2,x3,...,xt) be the leading coefficient
 * of U with respect to x.  Let
 *
 *	G = omega * F1^e1 * F2^e2 * F3^e3 * ... * Fk^ek		(**)
 *
 * be the complete factorization of G into irreducible polynomials F1...Fk
 * (omega is an integer), and let b2,b3,...,bt be given integers (the current
 * evaluation point).  Then one can easily compute U0(x) = U(x,b2,b3,...,bt),
 * delta = content (U0(x)), u(x) = prim_part (U0(x)).
 * Suppose we have a factorization
 *
 *	u(x) = u1(x) * u2(x) * ... * ur(x),			(***)
 *
 * Then U factors into at most r corresponding factors U = U1 * U2 * ... * Ur.
 * (This factorization is unknown, but we want to determine it.)
 * Let C1 = lc_x(U1), C2 = lc_x(U2), ...,
 * Cr = lc_x(Ur) be the leading coefficients of each factor, and also let
 *
 *	Ftilde_1 = F1(b2,b3,...,bt),
 *	Ftilde_2 = F2(b2,b3,...,bt),
 *	...,
 *	Ftilde_k = Fk(b2,b3,...,bt),
 * and
 *	v1 = lc(u1(x)),
 *	v2 = lc(u2(x)),
 *	...
 *	vr = lc(ur(x)).
 *
 * Then if the Ftilde_i, delta and omega have certain properties, it is
 * possible to determine C1...Cr precisely from the values of
 * Ftilde_1...Ftilde_k and v1...vr (although in some cases, omega and delta
 * cannot be partitioned precisely among the Ci, and it is necessary to
 * multiply each factor U1...Ur and u1...ur by integers c1...cr (whose product
 * divides omega).  (Note that U is assumed to be primitive, so when its
 * factors U1...Ur are finally determined, simply removing any integer
 * contents from them will discard these additional factors c1...cr.)
 *
 * This routine first determines whether Ftilde_i, delta and omega have the
 * necessary properties.  If not, it returns FALSE.  Otherwise it computes
 * C1...Cr (and when necessary, c1...cr) and returns TRUE.
 *
 * Finally, note that this routine works just as well for the trailing
 * coefficients of U, U1...Ur, and u(x), u1(x)...ur(x) as it does for
 * their leading coefficients.
 */

	static
	calib_bool
Zxyz_determine_coeffs (

struct calib_Zxyz_obj *		C,	/* IN/OUT: determined coefficient */
					/*	    for each factor */
const struct calib_Zxyz_factor *
				flist,	/* IN: fully factored coefficient */
mpz_srcptr			delta,	/* IN: content of univariate poly */
mpz_srcptr			ucoeff,	/* IN: coeff values from each ui(x) */
mpz_ptr				dpart,	/* IN/OUT: partition of delta */
					/*	    among the factors */
struct EEZglob *		eezgip	/* IN/OUT: global EEZ algorithm info */
)
{
int					i, j, k, r;
calib_bool				success;
const struct calib_Zxyz_dom *		dom;
const struct calib_Zxyz_factor *	fp;
const struct calib_Zxyz_obj *		factor;
const calib_bool *			evflags;
mpz_srcptr				bvals;
struct calib_Zxyz_obj *			D;
const struct calib_Zxyz_obj **		farray;
int *					expon;
unsigned long				nf;
mpz_ptr					Ftilde;
mpz_ptr					d, u2;
struct calib_Zxyz_obj			pvalue;
mpz_t					omega, TEMP1, _q, _r, Delta;

	/* Note -- flist is allowed to be NULL! */
	CALIB_FATAL_ERROR_IF ((C      EQ NULL) OR
			      (delta  EQ NULL) OR
			      (ucoeff EQ NULL) OR
			      (eezgip EQ NULL));

	success = FALSE;

	r	= eezgip -> nf;
	evflags	= eezgip -> evflags;
	bvals	= eezgip -> bvals;

	CALIB_FATAL_ERROR_IF (r < 2);

	dom = C [0].dom;
	for (i = 1; i < r; i++) {
		CALIB_FATAL_ERROR_IF (C [i].dom NE dom);
	}

	/* Count the non-constant multi-variate factors.  Factors that	*/
	/* are non-zero constants get multiplied into omega.		*/
	mpz_init_set_ui (omega, 1);
	mpz_init (TEMP1);
	k = 0;
	for (fp = flist; fp NE NULL; fp = fp -> next) {
		CALIB_FATAL_ERROR_IF (fp -> multiplicity <= 0);
		factor = fp -> factor;
		/* Factor of zero should not occur! */
		CALIB_FATAL_ERROR_IF (factor -> nterms <= 0);
		CALIB_FATAL_ERROR_IF (factor -> dom NE dom);
		if (Zxyz_is_non_zero_constant_poly (factor)) {
			mpz_pow_ui (TEMP1,
				    &(factor -> coeff [0]),
				    fp -> multiplicity);
			mpz_mul (omega, omega, TEMP1);
			continue;
		}
		++k;
	}

	/* Create local array of determined coefficient polynomials,	*/
	/* one per factor.						*/
	D = Zxyz_new_array (dom, r);
	for (i = 0; i < r; i++) {
		Zxyz_set_si (&D [i], 1);
		mpz_set_ui (&dpart [i], 1);
	}

#if 0
	/* We have to be very careful with signs (e.g., the units +1	*/
	/* and -1).  Our Z[x] factoring algorithm's policy is to make	*/
	/* the leading coefficient of each factor be strictly positive,	*/
	/* and to impose a -1 upon the (or as a) constant factor, if	*/
	/* necessary.  If omega < 0 here, then we need to impose this	*/
	/* sign somehow.  Note that it is *NOT* acceptable to return	*/
	/* it as a constant to "impose" upon each factor, and remove	*/
	/* the resulting integer contents from each factor when we are	*/
	/* done -- because integer contents are sign agnostic, and we	*/
	/* will wind up producing the factorization for -U instead of	*/
	/* for +U, as required.  Instead, we simply impose this sign	*/
	/* upon the determined coefficient for the *LAST* factor -- the	*/
	/* choice is pretty arbitrary when factoring, but in the case	*/
	/* of EEZ GCD, this alters the cofactor, and not the GCD.	*/
	if (mpz_sgn (omega) <= 0) {
		CALIB_FATAL_ERROR_IF (mpz_sgn (omega) EQ 0);
		mpz_abs (omega, omega);
		Zxyz_neg (&D [r - 1], &D [r - 1]);
	}
#endif

	mpz_init_set (Delta, delta);
	mpz_init (_q);
	mpz_init (_r);
	Zxyz_init (dom, &pvalue);

	/* Create a local array u2 that contains the evaluated integer	*/
	/* leading/trailing coefficient from each factor.		*/
	u2 = calib_new_Z_vector (r);
	calib_copy_Z_vector (u2, ucoeff, r);

	farray	= NULL;
	expon	= NULL;
	Ftilde	= NULL;
	d	= NULL;

	if (k > 0) {
		/* Allocate arrays for exponents and evaluated factors. */
		farray	= CALIB_NEWA (k, const struct calib_Zxyz_obj *);
		expon	= CALIB_NEWA (k, int);
		Ftilde	= calib_new_Z_vector (k);
		d	= calib_new_Z_vector (k + 1);

		/* Second pass through factors, evaluating them as we go. */
		i = 0;
		for (fp = flist; fp NE NULL; fp = fp -> next) {
			factor = fp -> factor;
			if (Zxyz_is_non_zero_constant_poly (factor)) continue;

			farray [i] = factor;
			expon [i]  = fp -> multiplicity;

			Zxyz_eval_var_subset (&pvalue, factor, evflags, bvals);

			/* The evaluated result MUST be a non-zero	*/
			/* integer constant.				*/
			CALIB_FATAL_ERROR_IF ((pvalue.nterms NE 1) OR
					      (NOT Zxyz_is_non_zero_constant_poly (&pvalue)));
			mpz_set (&Ftilde [i], &pvalue.coeff [0]);
			++i;
		}
		CALIB_FATAL_ERROR_IF (i NE k);

		/* We now have all of the inputs needed for Algorithm N	*/
		/* (Nondivisors) in Wang's paper, "An Improved		*/
		/* Multivariate	Factoring Algorithm," Mathematics of	*/
		/* Computation, Vol 32,	No. 144, October 1978, pages	*/
		/* 1215-1231.						*/

		mpz_mul (&d [0], Delta, omega);

		for (i = 1; i <= k; i++) {
			mpz_abs (_q, &Ftilde [i - 1]);
			for (j = i - 1; j >= 0; j--) {
				mpz_abs (_r, &d [j]);
				do {
					mpz_gcd (_r, _r, _q);
					mpz_divexact (_q, _q, _r);
				} while (mpz_cmp_ui (_r, 1) NE 0);
				if (mpz_cmp_ui (_q, 1) EQ 0) {
					/* Conditions for pre-determining */
					/* this	coefficient are NOT met.  */
					success = FALSE;
					goto clean_up;
				}
			}
			mpz_set (&d [i], _q);
		}

		/* Now distribute the multivariate factors F1...Fk	*/
		/* among the Dj.					*/
		for (i = k - 1; i >= 0; i--) {
			factor = farray [i];

			/* Distribute the correct number of copies of	*/
			/* "factor" to each D[j].			*/
			for (j = 0; j < r; j++) {
				mpz_mul (_q, Delta, &u2 [j]);
				mpz_abs (_q, _q);
				mpz_abs (_r, &Ftilde [i]);
				nf = mpz_remove (_r, _q, _r);
				CALIB_FATAL_ERROR_IF (nf > expon [i]);
				expon [i] -= nf;
				if (nf > 0) {
					Zxyz_ipow (&pvalue, factor, nf);
					Zxyz_mul (&D [j], &D [j], &pvalue);
					/* Remove Ftilde[i]^nf, first	*/
					/* from u2[j], and rest from	*/
					/* Delta.			*/
					mpz_pow_ui (_r, &Ftilde [i], nf);
					if (mpz_sgn (_r) < 0) {
						mpz_neg (&u2 [j], &u2 [j]);
						mpz_neg (_r, _r);
					}
					mpz_gcd (_q, _r, &u2 [j]);
					mpz_divexact (&u2 [j], &u2 [j], _q);
					mpz_divexact (_r, _r, _q);
					mpz_divexact (Delta, Delta, _r);
					mpz_mul (&dpart [j], &dpart [j], _r);
				}
			}
			/* All copies of this factor should be		*/
			/* distributed among the D[j].  If not, then we	*/
			/* need to pick a new evaluation point.		*/
			if (expon [i] > 0) {
				/* We must pick a new evaluation point. */
				success = FALSE;
				goto clean_up;
			}
		}
	}

	/* The non-constant factors F1...Fk have all been distributed.	*/
	/* The product of Delta and all of the remaining u2[i] values	*/
	/* should be exactly "omega".  Validate this now.  This		*/
	/* demonstrates that omega has been completely accounted for,	*/
	/* once we take care of Delta.					*/
	mpz_set (_q, Delta);
	for (i = 0; i < r; i++) {
		mpz_mul (_q, _q, &u2 [i]);
	}
	CALIB_FATAL_ERROR_IF (mpz_cmp (_q, omega) NE 0);

	/* Multiply the remaining u2[i] back into each D[i].  This	*/
	/* correctly distributes all of the factors of omega (except	*/
	/* those occurring in Delta), and corrects the signs of this	*/
	/* coeffcient within each factor.				*/
	for (i = 0; i < r; i++) {
		Zxyz_mul_z (&D [i], &D [i], &u2 [i]);
	}

	/* Copy determined coefficients out to our caller. */
	for (i = 0; i < r; i++) {
		Zxyz_set (&C [i], &D [i]);
	}

	success = TRUE;

clean_up:
	calib_free_Z_vector (d, k + 1);
	calib_free_Z_vector (Ftilde, k);
	if (expon NE NULL) {
		free (expon);
	}
	if (farray NE NULL) {
		free (farray);
	}
	calib_free_Z_vector (u2, r);
	Zxyz_clear (&pvalue);
	mpz_clear (_r);
	mpz_clear (_q);
	mpz_clear (Delta);
	Zxyz_free_array (D, r);
	mpz_clear (TEMP1);
	mpz_clear (omega);

	return (success);
}

/*
 * Modify the given polynomial by dropping all terms comprising the
 * coefficient of var^k.
 */

	static
	void
Zxyz_drop_coeff (

struct calib_Zxyz_obj *	poly,		/* IN/OUT: polynomial to modify */
int			var,		/* IN: variable */
int			k		/* IN: power to drop */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
int				*sip, *dip;
mpz_ptr				szp, dzp;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (k < 0));
	dom	= poly -> dom;
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	nterms = poly -> nterms;
	sip = dip = poly -> pow;
	szp = dzp = poly -> coeff;
	for (i = 0; i < nterms; i++) {
		if (sip [var] NE k) {
			/* Retain this term. */
			if (sip NE dip) {
				memcpy (dip, sip, nvars * sizeof (int));
			}
			mpz_set (dzp, szp);
			dip += nvars;
			++dzp;
		}
		sip += nvars;
		++szp;
	}
	poly -> nterms = (dzp - poly -> coeff);
}

/*
 * Multiply the given polynomial (in place) by var^k.
 */

	static
	void
Zxyz_mul_var_pow_k (

struct calib_Zxyz_obj *		poly,	/* IN/OUT: polynomial to multiply */
int				var,	/* IN: variable */
int				k	/* IN: power to multiply by */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
int *				ip;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (k < 0));
	dom	= poly -> dom;
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	nterms	= poly -> nterms;
	ip	= poly -> pow;
	for (i = 0; i < nterms; i++) {
		ip [var] += k;
		ip += nvars;
	}
}

/*
 * Do the actual EEZ multi-variate lifting.
 *
 * Returns SUCCESS, BAD_EVAL_POINT, or EXTRANEOUS_FACTORS.
 */

	static
	int
Zxyz_eez_lift_factors (

struct EEZglob *	eezgip		/* IN/OUT: global EEZ info */
)
{
int				i, j, m, nf, nv, nvars, var, nlift, status;
calib_bool			success;
const struct calib_Zxyz_dom *	dom;
const struct calib_Zxyz_obj *	poly;
int *				varlist;
mpz_srcptr			bvals, modulus;
struct calib_Zxyz_obj *		pdfactors;
struct calib_Zxyz_obj *		lfactors;
struct calib_Zxyz_obj *		l2factors;
struct calib_Zxyz_obj *		evpoly;
struct calib_Zxyz_obj *		redfactors;
struct calib_Zxyz_obj *		cfact;
calib_bool *			lift_evflags;
calib_bool *			lift2_evflags;
int *				mindeg;
int *				maxdeg;
const struct calib_Zxyz_obj *	vvp;
struct calib_Zxyz_obj		PROD, RESID, C;
mpz_t				mfact;

	dom		= eezgip -> dom;
	nf		= eezgip -> nf;
	nv		= eezgip -> nv;
	nvars		= eezgip -> nvars;
	pdfactors	= eezgip -> pdfactors;
	lfactors	= eezgip -> lfactors;
	l2factors	= eezgip -> l2factors;
	evpoly		= eezgip -> evpoly;
	cfact		= eezgip -> cfact;
	varlist		= eezgip -> varlist;
	bvals		= eezgip -> bvals;
	modulus		= eezgip -> modulus;

	/* Compute all of the factorials we will ever need. */

#if PRINT_LIFTING_INFO
printf ("Enter eez_lift_factors with"
	" nf=%d, nvars=%d, nv=%d, nterms=%d, totaldeg=%d\n",
	nf,
	nvars,
	nv,
	eezgip -> poly -> nterms,
	eezgip -> total_degree);
#endif

	/* Compute info about each variable. */
	Zxyz_eez_compute_var_info (eezgip);

	/* Compute the f0fact and zx_f0fact -- fully evaluated		*/
	/* univariate versions of each factor.				*/
	Zxyz_eez_compute_f0fact (eezgip);

	/* Create a local lift_evflags that only evaluates away		*/
	/* variables that are not yet being lifted, and also another	*/
	/* lift2_evflags that also evaluates away the one we are now	*/
	/* lifting.							*/
	lift_evflags = CALIB_NEWA (2 * nvars, calib_bool);
	lift2_evflags = lift_evflags + nvars;
	memcpy (lift_evflags,  eezgip -> evflags, nvars * sizeof (calib_bool));
	memcpy (lift2_evflags, eezgip -> evflags, nvars * sizeof (calib_bool));

	mpz_init (mfact);
	Zxyz_init (dom, &PROD);
	Zxyz_init (dom, &RESID);
	Zxyz_init (dom, &C);
	redfactors = Zxyz_new_array (dom, nf);

	mindeg = maxdeg = NULL;

	status = SUCCESS;

	/* Lift each variable successively. */
	for (i = 0; i < nv; i++) {
		/* Get next variable to lift back in. */
		var = varlist [i];

#if PRINT_LIFTING_INFO
		printf ("  Lifing %s:\n", dom -> vars [var]);
#endif

		/* This routine is lifting "var".  Our correction	*/
		/* factors involve only the previously lifted vars,	*/
		/* e.g., varlist [0], ..., varlist [i-1].		*/
		nlift = i;

		/* Stop evaluating this variable away. */
		lift_evflags [var] = FALSE;

		/* Get the "objective" for this lifting cycle: the	*/
		/* product polynomial, having just this one additional	*/
		/* variable in it.  The goal is to make our factors	*/
		/* multiply out to this (over Z[x,y,z]).		*/
		poly = &evpoly [i + 1];

		/* Get the maximum degree of each var in this evaluated	*/
		/* polynomial (may be less than in original).		*/
		Zxyz_get_poly_degrees (&mindeg, &maxdeg, poly);

		/* Our factors have pre-determined coefficients in	*/
		/* them, containing variables that we have not yet	*/
		/* lifted, as well as the one we are now lifting.	*/
		/* Compute l2factors[] to have all of these vars	*/
		/* evaluated away.					*/
		for (j = 0; j < nf; j++) {
			Zxyz_eval_var_subset (&l2factors [j],
					      &pdfactors [j],
					      lift2_evflags,
					      bvals);
		}

		/* Lift successively higher powers m of this var. */
		mpz_set_ui (mfact, 1);
		for (m = 1; ; m++) {

			/* Our factors have pre-determined coefficients	*/
			/* in them, containing variables that we have	*/
			/* not lifted yet.  Evaluate these away,	*/
			/* storing them in lfactors[].			*/
			for (j = 0; j < nf; j++) {
				Zxyz_eval_var_subset (&lfactors [j],
						      &pdfactors [j],
						      lift_evflags,
						      bvals);
			}

			/* Compute the current residual. */
			Zxyz_set (&PROD, &lfactors [0]);
			for (j = 1; j < nf; j++) {
				Zxyz_mul (&PROD, &PROD, &lfactors [j]);
			}
			Zxyz_sub (&RESID, poly, &PROD);

			if (RESID.nterms <= 0) {
#if PRINT_LIFTING_INFO
				printf ("    done\n");
#endif
				status = SUCCESS;
				break;
			}
			if (m > maxdeg [var]) {
				/* We have lifted this var all the way	*/
				/* up to its maximum degree, and we	*/
				/* still do not have equality over Z.	*/
				/* Must have extraneous factors.	*/
#if PRINT_LIFTING_INFO
			printf ("    degree failure!\n");
#endif
				status = EXTRANEOUS_FACTORS;
				break;
			}

			/* RESID *must* be an exact multiple of		*/
			/* (var - value)^m.  Use EEZ differentiation	*/
			/* to divide out this factor.			*/
			Zxyz_nth_derivative (&C, &RESID, var, m);
			Zxyz_div_z_exact (&C, &C, mfact);
			Zxyz_eval_one_var (&C, &C, var, &bvals [var]);

			/* Compute the correction coefficients for C.	*/
			eezgip -> num_cf_terms = 0;
			success = Zxyz_eez_compute_correction (eezgip,
							       nlift,
							       maxdeg,
							       &C);
			if (NOT success) {
				/* Failed to compute correction factor. */
				/* We must have extraneous factors.	*/
#if PRINT_LIFTING_INFO
				printf ("    Correction factor failure!\n");
#endif
				status = EXTRANEOUS_FACTORS;
				break;
			}

#if PRINT_LIFTING_INFO
			printf ("    Degree %d correction: %lu Taylor terms\n",
				m, eezgip -> num_cf_terms);
#endif

			/* Get (var-val)^m polynomial. */
			vvp = Zxyz_eez_get_var_val_pow (eezgip, var, m);

			/* Apply correction to each factor. */
			for (j = 0; j < nf; j++) {
				Zxyz_mul (&C, &cfact [j], vvp);
				Zxyz_mod_out (&C, modulus);
				Zxyz_add (&pdfactors [j], &pdfactors [j], &C);
				Zxyz_mod_out (&pdfactors [j], modulus);
			}

			mpz_mul_ui (mfact, mfact, m + 1);
		}
		if (status NE SUCCESS) break;

		/* This var is lifted.  Stop evaluating it away. */
		lift2_evflags [var] = FALSE;

		free (mindeg);
		mindeg = maxdeg = NULL;
	}

	if (mindeg NE NULL) {
		free (mindeg);
	}

	Zxyz_free_array (redfactors, nf);
	Zxyz_clear (&C);
	Zxyz_clear (&RESID);
	Zxyz_clear (&PROD);
	mpz_clear (mfact);

	free (lift_evflags);

#if PRINT_LIFTING_INFO
printf ("Leave eez_lift_factors with status=%d\n", status);
#endif

	return (status);
}

/*
 * Compute various information about each variable that is needed by
 * the EEZ algorithm.
 */

	static
	void
Zxyz_eez_compute_var_info (

struct EEZglob *	eezgip	/* IN/OUT: EEZ algorithm global info */
)
{
int				i, j, k, nvars, nv, x1, var;
const struct calib_Zxyz_dom *	dom;
const int *			r2fmap;
mpz_srcptr			bvals, bval;
struct EEZvar *			var_info;
struct EEZvar *			vip;
struct calib_Zxyz_obj *		poly2;
int *				mindeg;
int *				maxdeg;
mpz_ptr				pvalpows;
mpz_ptr				mvalpows;

	dom		= eezgip -> dom;
	nvars		= eezgip -> nvars;
	x1		= eezgip -> x1;
	nv		= eezgip -> nv;
	r2fmap		= eezgip -> r2fmap;
	bvals		= eezgip -> bvals;
	poly2		= eezgip -> poly2;
	var_info	= eezgip -> var_info;

	/* Get the degrees of each variable. */
	Zxyz_get_poly_degrees (&mindeg, &maxdeg, poly2);

	/* We only need info about vars being lifted. */
	for (i = 0; i < nv; i++) {
		var = r2fmap [i];
		CALIB_FATAL_ERROR_IF ((var < 0) OR
				      (var >= nvars) OR
				      (var EQ x1));

		vip = &var_info [var];

		/* First, free up arrays for this var. */
		Zxyz_eez_clear_var_info (vip);

		k = maxdeg [var];
		bval = &bvals [var];

		/* Precompute powers of +bval and -bval. */
		pvalpows = calib_new_Z_vector (2 * (k + 1));
		mvalpows = pvalpows + (k + 1);
		mpz_set_ui (&pvalpows [0], 1);
		mpz_set    (&pvalpows [1], bval);
		for (j = 2; j <= k; j++) {
			mpz_mul (&pvalpows [j], &pvalpows [j - 1], bval);
		}
		calib_copy_Z_vector (mvalpows, pvalpows, k + 1);
		for (j = 1; j <= k; j += 2) {
			mpz_neg (&mvalpows [j], &mvalpows [j]);
		}

		vip -> maxdeg	= k;
		vip -> bval	= &bvals [var];
		vip -> pvalpows	= pvalpows;
		vip -> mvalpows	= mvalpows;
		vip -> mon_pows	= Zxyz_new_array (dom, k + 1);
	}
	free (mindeg);
}

/*
 * Compute the f0fact and zx_f0fact -- fully evaluated univariate versions
 * of each factor.  Note that these only change if one of the following
 * changes:
 *
 *	1. The polynomial (being factored), or
 *	2. The evaluation point.
 */

	static
	void
Zxyz_eez_compute_f0fact (

struct EEZglob *	eezgip	/* IN/OUT: global EEZ algorithm info */
)
{
int				i, j, nf, x1;
const calib_bool *		evflags;
mpz_srcptr			bvals;
struct calib_Zxyz_obj *		pdfactors;
struct calib_Zxyz_obj *		f0fact;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_obj *		zx_f0fact;
struct calib_Zx_obj *		zx_y0;
struct calib_Zx_obj *		yp1;

	x1		= eezgip -> x1;
	nf		= eezgip -> nf;
	evflags		= eezgip -> evflags;
	bvals		= eezgip -> bvals;
	pdfactors	= eezgip -> pdfactors;
	f0fact		= eezgip -> f0fact;
	zx_f0fact	= eezgip -> zx_f0fact;
	zx_y0		= eezgip -> zx_y0;

	for (i = 0; i < nf; i++) {
		Zxyz_eval_var_subset (&f0fact [i], &pdfactors [i], evflags, bvals);
		Zxyz_cv_poly_to_Zx_form (&zx_f0fact [i], &f0fact [i], x1);
	}

	/* Compute the y0 [i] (in Z[x] form). */
	Zx = calib_get_Zx_dom ();
	for (i = 0; i < nf; i++) {
		yp1 = &zx_y0 [i];
		/* Need Zx -> set_si (yp1, 1) here. */
		Zx -> alloc (yp1, 0);
		mpz_set_ui (&(yp1 -> coeff [0]), 1);
		yp1 -> degree = 0;

		for (j = 0; j < nf; j++) {
			if (j EQ i) continue;
			Zx -> mul (yp1, yp1, &zx_f0fact [j]);
		}
	}
}

/*
 * Compute the EEZ correction coefficients for the given C.
 */

	static
	calib_bool
Zxyz_eez_compute_correction (

struct EEZglob *	eezgip,	/* IN/OUT: global EEZ algorithm info */
int			nlift,	/* IN: num varlist[] vars being lifted */
const int *		maxdeg,	/* IN: degree of each var in current evpoly */
const struct calib_Zxyz_obj *
			C	/* IN: Polynomial C to correct for */
)
{
int				i, j, m, nf, x1, total_degree;
calib_bool			success;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		l2factors;
struct calib_Zxyz_obj *		ypolys;
struct calib_Zxyz_obj *		cfact;
struct calib_Zxyz_obj *		c2fact;
struct calib_Zxyz_obj *		Dm_stack;
struct calib_Zxyz_obj *		deriv_stack;
struct calib_Zxyz_obj *		prodstack;
struct calib_Zx_obj *		zx_cfact;
struct calib_Zx_obj *		zx_f0bar;
const struct calib_Zx_dom *	Zx;
mpz_srcptr			modulus;
struct calib_Zxyz_obj		C0, Dm, PROD;
struct calib_Zx_obj		zx_C0, zx_temp1, zx_temp2;
mpz_t				Dval;

	dom		= eezgip -> dom;
	nf		= eezgip -> nf;
	x1		= eezgip -> x1;
	total_degree	= eezgip -> total_degree;
	l2factors	= eezgip -> l2factors;
	ypolys		= eezgip -> ypolys;
	cfact		= eezgip -> cfact;
	c2fact		= eezgip -> c2fact;
	Dm_stack	= eezgip -> Dm_stack;
	deriv_stack	= eezgip -> deriv_stack;
	prodstack	= eezgip -> prodstack;
	zx_cfact	= eezgip -> zx_cfact;
	zx_f0bar	= eezgip -> zx_f0bar;
	modulus		= eezgip -> modulus;

	if (mpz_sgn (modulus) EQ 0) {
		/* Choose a new modulus and recompute all info	*/
		/* pertaining to it.				*/
		Zxyz_eez_compute_modulus_info (eezgip);
	}

	Zx = calib_get_Zx_dom ();

	Zxyz_init (dom, &C0);
	Zxyz_init (dom, &Dm);
	Zxyz_init (dom, &PROD);

	Zx -> init (&zx_C0);
	Zx -> init (&zx_temp1);
	Zx -> init (&zx_temp2);

	/* Compute the ypolys[i] = prod_{j<>i} l2factors[j].  They do	*/
	/* NOT contain the one variable our caller is now lifting, but	*/
	/* only the previously lifted variables.			*/
	for (i = 0; i < nf; i++) {
		Zxyz_set_si (&PROD, 1);
		for (j = 0; j < nf; j++) {
			if (j NE i) {
				Zxyz_mul (&PROD, &PROD, &l2factors [j]);
			}
		}
		Zxyz_set (&ypolys [i], &PROD);
	}

	/* Compute zx_C0, which is C reduced to univariate, in Z[x] form. */
	Zxyz_eval_var_subset (&C0, C, eezgip -> evflags, eezgip -> bvals);
	Zxyz_cv_poly_to_Zx_form (&zx_C0, &C0, x1);

	/* Compute initial univariate correction factors.  This is one	*/
	/* place where we are very glad that the zx_f0bar[] are monic.	*/
	mpz_init (Dval);
	for (i = 0; i < nf; i++) {
		Zx -> mul (&zx_temp1, &zx_cfact [i], &zx_C0);
		Zxyz_mod_out_Zx (&zx_temp1, modulus);
		Zx -> div (&zx_temp2, &zx_temp1, Dval,
			   &zx_temp1, &zx_f0bar [i]);
		/* Division must not be fractional! */
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (Dval, 1) NE 0);

		Zxyz_mod_out_Zx (&zx_temp1, modulus);
		Zxyz_cv_from_Zx (&cfact [i], &zx_temp1, x1);
	}
	mpz_clear (Dval);
	++(eezgip -> num_cf_terms);

	/* Solution should be correct over x1.  This is correct modulo	*/
	/* total degree m = 0 in the other (xk - ak)^ek.  Each		*/
	/* iteration of this loop lifts this total degree in these	*/
	/* monomials from m to m+1.					*/

	for (m = 0; ; m++) {

		Zxyz_mul (&Dm, &cfact [0], &ypolys [0]);
		for (i = 1; i < nf; i++) {
			Zxyz_mul (&PROD, &cfact [i], &ypolys [i]);
			Zxyz_add (&Dm, &Dm, &PROD);
		}
		Zxyz_sub (&Dm, C, &Dm);
		Zxyz_mod_out (&Dm, modulus);

		if (Dm.nterms <= 0) {
			/* We have equality over Z.  Lifting of this	*/
			/* correction factor is complete.		*/
			success = TRUE;
			break;
		}

		if (m >= total_degree) {
			/* We have lifted up to order total_degree	*/
			/* without obtaining equality over Z.  This	*/
			/* means that we have extraneous factors.	*/
			success = FALSE;
			break;
		}

		/* Zero out the buckets into which we accumulate	*/
		/* corrections (of total degree m) to each correction	*/
		/* factor.						*/
		for (i = 0; i < nf; i++) {
			c2fact [i].nterms = 0;
		}

		/* Set first residual in stack. */
		Zxyz_set (&Dm_stack [0], &Dm);

		/* Set the initial product to 1. */
		Zxyz_set_si (&prodstack [0], 1);

		/* Recursively enumerate (and sum into c2fact[]) all	*/
		/* corrections of total degree m + 1 to each of the	*/
		/* correction factors cfact[].				*/

		Zxyz_eez_cfact_recurse (eezgip,
					nlift,
					maxdeg,
					nlift - 1, /* cur varlist[] index */
					Dm_stack,
					deriv_stack,
					prodstack,
					0,	/* degree(*prodstack) */
					m + 1);

		/* Apply these corrections: cfact[] += c2fact[]. */
		for (i = 0; i < nf; i++) {
			Zxyz_add (&cfact [i], &cfact [i], &c2fact [i]);
			Zxyz_mod_out (&cfact [i], modulus);
		}
	}

	Zx -> clear (&zx_temp2);
	Zx -> clear (&zx_temp1);
	Zx -> clear (&zx_C0);

	Zxyz_clear (&PROD);
	Zxyz_clear (&Dm);
	Zxyz_clear (&C0);

	return (success);
}

/*
 * Compute (or recompute) all of the following:
 *
 * 1. The coefficient upper bound B.
 * 2. A suitable prime p.
 * 3. A new modulus = p**k such that modulus > 2*B.
 * 4. The "raw" univariate correction cofficients.
 */

	static
	void
Zxyz_eez_compute_modulus_info (

struct EEZglob *	eezgip	/* IN/OUT: global EEZ algorithm info */
)
{
int			i, nf;
calib_bool		bad, success;
struct calib_Zx_obj *	zx_f0fact;
struct calib_Zx_obj *	fp;
mpz_srcptr		p, lc;
struct calib_Zpx_obj *	modp_cfact;
struct calib_Zpx_dom *	Zpx;

	/* Compute the upper bound B on the magnitude of the	*/
	/* coefficients of any factor we are trying to lift.	*/
	Zxyz_eez_compute_upper_bound (eezgip);

	nf		= eezgip -> nf;
	p		= eezgip -> p;
	zx_f0fact	= eezgip -> zx_f0fact;
	Zpx		= eezgip -> Zpx;

	modp_cfact = Zxyz_new_Zpx_array (Zpx, nf);

	/* Try random primes p until we find one that works. */
	for (;;) {
		Zxyz_eez_choose_random_prime (eezgip);

		/* The prime must not divide the leading coefficient of	*/
		/* any of the (univariate) factors we are trying to	*/
		/* lift.  Using such primes would cause the degree of	*/
		/* these factors to decrease.				*/
		bad = FALSE;
		for (i = 0; i < nf; i++) {
			/* Get leading coefficient of univariate factor. */
			fp = &zx_f0fact [i];
			lc = &(fp -> coeff [fp -> degree]);
			if (mpz_divisible_p (lc, p)) {
				bad = TRUE;
				break;
			}
		}
		if (bad) {
			/* Must choose another prime. */
			continue;
		}

		success = Zxyz_eez_compute_modp_cfact (eezgip, modp_cfact);
		if (success) break;

		/* Try another prime. */
	}

	/* Need to lift the modp_cfact to a sufficiently large		*/
	/*  modulus = p^k, and also handle the leading coefficient	*/
	/* problem.							*/

	Zxyz_eez_lift_cfact (eezgip, modp_cfact);

	/* Now that we have a new modulus, we will need to compute (or	*/
	/* recompute) our table of (1/n! mod modulus) values.		*/
	Zxyz_eez_compute_inverse_factorials (eezgip);

	Zxyz_free_Zpx_array (Zpx, modp_cfact, nf);
}

/*
 * Compute an upper bound B on the magnitude of the coefficients of any
 * factor we are trying to lift.  A couple of notes:
 *
 * 1. Since we like to assure the correct factorization (over Z) for
 *    each variable successively lifted, we compute this bound for
 *    each evpoly[], and take the maximum.
 *
 * 2. This bound does not change for the original multi-variate polynomial,
 *    but it can change for the other evpoly[] as we choose different
 *    evaluation points.  However, in the case of EEZ GCD, the original
 *    multi-variate polynomial can even change each time.
 *
 * The easiest option is to recompute this each time a new evaluation
 * point is chosen (i.e., when Zxyz_eez_compute_modulus_info() is called).
 */

	static
	void
Zxyz_eez_compute_upper_bound (

struct EEZglob *	eezgip	/* IN/OUT: global EEZ algorithm info */
)
{
int			i, nv;
struct calib_Zxyz_obj *	evpoly;
mpz_ptr			B;
mpz_t			Bi;

	nv	= eezgip -> nv;
	evpoly	= eezgip -> evpoly;
	B	= eezgip -> B;

	mpz_set_ui (B, 3);	/* Minimum reasonable bound: 1st odd prime. */

	mpz_init (Bi);

	for (i = nv; i >= 0; i--) {
		Zxyz_factor_coefficient_upper_bound (Bi, &evpoly [i]);
		if (mpz_cmp (Bi, B) > 0) {
			mpz_set (B, Bi);
		}
	}

	mpz_clear (Bi);
}

/*
 * Compute an upper bound on the magnitude of the coefficients of any
 * factor of the given polynomial.  This bound is due to:
 *
 * Gel'Fond, A. O., "Transcendental and Algebraic Numbers," GITTL,
 * Moscow, 1952; English translation, Dover, New York, 1960, MR 15,
 * 292; 22 #5598, p. 135.
 *
 *	B = Umax * e^(m1 + m2 + ... + mn),
 *
 * where
 *
 *	Umax is the maximum coefficient magnitude, and
 *	m1, m2, ..., mn are the maximum magnitudes of each variable.
 *
 * We use (11/4) = 2.75 > e = 2.7182818 for simplicity.
 */

	static
	void
Zxyz_factor_coefficient_upper_bound (

mpz_ptr				B,		/* IN/OUT: upper bound */
const struct calib_Zxyz_obj *	poly		/* IN: polynomial to bound */
)
{
int				i, j, k, nterms, nvars, sum;
const struct calib_Zxyz_dom *	dom;
const int *			ip;
mpz_srcptr			zp;
mpz_t				Umax, mag, tmp1, tmp2, tmp3;

	CALIB_FATAL_ERROR_IF ((B EQ NULL) OR (poly EQ NULL));

	nterms	= poly -> nterms;
	dom	= poly -> dom;
	nvars	= dom -> nvars;

	mpz_init (Umax);
	mpz_init (mag);
	mpz_init (tmp1);
	mpz_init (tmp2);
	mpz_init (tmp3);

	int	maxdeg [nvars];
	memset (&maxdeg [0], 0, nvars * sizeof (int));

	ip = poly -> pow;
	zp = poly -> coeff;
	for (i = 0; i < nterms; i++) {
		for (j = 0; j < nvars; j++) {
			k = *ip++;
			if (k > maxdeg [j]) {
				maxdeg [j] = k;
			}
		}
		mpz_abs (mag, zp);
		++zp;
		if (mpz_cmp (mag, Umax) > 0) {
			mpz_set (Umax, mag);
		}
	}
	sum = 0;
	for (j = 0; j < nvars; j++) {
		sum += maxdeg [j];
	}
	mpz_set_ui (tmp1, 11);
	mpz_pow_ui (tmp1, tmp1, sum);
	/* tmp2 = ceil (sqrt (tmp1)). */
	mpz_sqrt (tmp2, tmp1);
	mpz_mul (tmp3, tmp2, tmp2);
	if (mpz_cmp (tmp3, tmp1) < 0) {
		mpz_add_ui (tmp2, tmp2, 1);
	}
	/* tmp1 = ceil (sqrt (tmp2)). */
	mpz_sqrt (tmp1, tmp2);
	mpz_mul (tmp3, tmp1, tmp1);
	if (mpz_cmp (tmp3, tmp2) < 0) {
		mpz_add_ui (tmp1, tmp1, 1);
	}
	mpz_mul (B, Umax, tmp1);

	mpz_clear (tmp3);
	mpz_clear (tmp2);
	mpz_clear (tmp1);
	mpz_clear (mag);
	mpz_clear (Umax);
}

/*
 * Choose a random prime number p.
 */

	static
	void
Zxyz_eez_choose_random_prime (

struct EEZglob *	eezgip	/* IN/OUT: global EEZ algorithm info */
)
{
unsigned int		word;
struct calib_Random *	rand;
mpz_ptr			p;

	rand	= &(eezgip -> gip -> rng);
	p	= eezgip -> p;

	word = calib_random_u32 (rand);
	mpz_set_ui (p, word);
	mpz_nextprime (p, p);

	/* Set it in our Zp and Zp[x] domains also. */
	mpz_set (eezgip -> Zp -> p, p);
}

/*
 * Compute the (mod p) version of the correction coefficients, which
 * satisfy the following:
 *
 *
 *	sum_{i=1}^nf  modp_cfact [i] * my0 [i] == 1 (mod p)
 *
 * where my0 [i] is the image (in Zp[x]) of zx_y0 [i].
 *
 * This will fail if any pair of the factors fails to be relatively
 * prime modulo p.  This indicates a "bad" choice of prime p.  Our
 * caller simply picks another prime and re-tries.
 *
 * Returns TRUE upon success, FALSE upon failure.
 */

	static
	calib_bool
Zxyz_eez_compute_modp_cfact (

struct EEZglob *	eezgip,		/* IN/OUT: global EEZ info */
struct calib_Zpx_obj *	modp_cfact	/* IN/OUT: mod p correction factors */
)
{
int			i, nf;
calib_bool		success;
struct calib_Zpx_dom *	Zpx;
struct calib_Zx_obj *	zx_f0fact;
struct calib_Zx_obj *	zx_y0;
struct calib_Zpx_obj	A, G, F, Y;

	nf		= eezgip -> nf;
	Zpx		= eezgip -> Zpx;
	zx_f0fact	= eezgip -> zx_f0fact;
	zx_y0		= eezgip -> zx_y0;

	Zpx -> init (Zpx, &A);
	Zpx -> init (Zpx, &G);
	Zpx -> init (Zpx, &F);
	Zpx -> init (Zpx, &Y);

	success = TRUE;

	for (i = 0; i < nf; i++) {
		Zpx -> set_Zx (&F, &zx_f0fact [i]);
		Zpx -> set_Zx (&Y, &zx_y0 [i]);
		Zpx -> extgcd (&G, &A, &modp_cfact [i], &F, &Y);
		if ((G.degree NE 0) OR (mpz_cmp_ui (&G.coeff [0], 1) NE 0)) {
			/* Two factors are not relatively prime! */
			success = FALSE;
			break;
		}
	}

	Zpx -> clear (&Y);
	Zpx -> clear (&F);
	Zpx -> clear (&G);
	Zpx -> clear (&A);

	return (success);
}

/*
 * Lift the given modulo-p correction factors to a sufficiently high
 * modulus p^k > 2*B.
 *
 * Note that we have to take some very special care when computing
 * these correction factors, to get them into a form that is
 * computationally useful.  The first problem is that we will be working
 * in K[x], where K is the integers modulo p^k -- which is not a field!
 * The computation we need to perform (given some c(x)) is this:
 *
 *	a_i(x) <== [[ c(x) * cfact_i(x) ] mod f0_i(x) ]	(mod p^k)	(**)
 *
 * where the polynomial arithmetic is done in Z[x], and the final result
 * reduced modulo p^k.
 *
 * But note that f0_i(x) is not necessarily monic!  So we need to compute
 * f0bar_i(x) in \Z[x] that is monic, and otherwise a scalar multiple of
 * f0_i(x) (mod p^k).  Fortunately, the *remainder* of a polynomial
 * division does not change when the divisor polynomial is scaled by any
 * non-zero constant (only the quotient changes, which we do not care about).
 * There is therefore no need to correct the cfact_i(x) values.
 *
 * The solution is to compute integer constants d_i such that:
 *
 *	lc (d_i * f0_i(x)) == 1	(mod p^k), and
 *
 * define f0bar_i(x) = ((d_i * f0_i(x)) mod (p^k)).  We need to lift
 * the d_i from modulus p^1 up to p^k.
 *
 * We lift the following equation to get the cfact_i(x):
 *
 *	\sum_i cfact_i(x) * y0_i(x) == 1	(mod p^k)
 *
 * When we are done, the computation (**) becomes:
 *
 *	a_i(x) <== [[ c(x) * cfact_i(x) ] mod f0bar_i(x) ] (mod p^k)
 */

	static
	void
Zxyz_eez_lift_cfact (

struct EEZglob *	eezgip,		/* IN/OUT: global EEZ info */
struct calib_Zpx_obj *	modp_cfact	/* IN: mod p correction factors */
)
{
int				i, j, k, nf;
mpz_srcptr			p, lc;
mpz_ptr				modulus;
mpz_ptr				d, lcf_inv;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_obj *		zx_f0fact;
struct calib_Zx_obj *		zx_y0;
struct calib_Zx_obj *		zx_cfact;
struct calib_Zx_obj *		zx_ptr;
struct calib_Zx_obj *		zx_f0bar;
struct calib_Zx_obj *		init_zx_cfact;
struct calib_Zx_obj *		init_f0bar;
struct calib_Zpx_obj *		zpx_ptr;
struct calib_Zx_obj		RESID, TEMP1, TEMP2;
mpz_t				BOUND, HALF_MOD, NEXT_MODULUS, temp1, temp2;

	nf		= eezgip -> nf;
	p		= eezgip -> p;
	zx_f0fact	= eezgip -> zx_f0fact;
	zx_y0		= eezgip -> zx_y0;
	zx_cfact	= eezgip -> zx_cfact;
	zx_f0bar	= eezgip -> zx_f0bar;
	modulus		= eezgip -> modulus;

	Zx = calib_get_Zx_dom ();

	/* Compute inverses (mod p) of the leading coefficients of the	*/
	/* f0_i(x).							*/
	lcf_inv = calib_new_Z_vector (nf);
	for (i = 0; i < nf; i++) {
		zx_ptr = &zx_f0fact [i];
		lc = &(zx_ptr -> coeff [zx_ptr -> degree]);
		j = mpz_invert (&lcf_inv [i], lc, p);
		CALIB_FATAL_ERROR_IF (NOT j);
	}

	/* Initialize d[i] = lcf_inv[i].  The d[i] change as we lift,	*/
	/* but the lcf_inv[i] do not.					*/
	d = calib_new_Z_vector (nf);
	calib_copy_Z_vector (d, lcf_inv, nf);

	/* Initialize the Z[x] correction factors from the	*/
	/* (mod p) versions.					*/
	for (i = 0; i < nf; i++) {
		zpx_ptr = &modp_cfact [i];
		k = zpx_ptr -> degree;
		zx_ptr = &zx_cfact [i];
		Zx -> alloc (zx_ptr, k);
		calib_copy_Z_vector (zx_ptr -> coeff, zpx_ptr -> coeff, k + 1);
		zx_ptr -> degree = k;
	}

	/* Now initialize the zx_f0bar. */
	for (i = 0; i < nf; i++) {
		zx_ptr = &zx_f0bar [i];
		Zx -> mul_z (zx_ptr, &zx_f0fact [i], &d [i]);
		Zxyz_mod_out_Zx (zx_ptr, p);
	}

	Zx -> init (&RESID);
	Zx -> init (&TEMP1);
	Zx -> init (&TEMP2);

#if 1
	/* Verify that the zx_f0bar[i] are all monic. */
	for (i = 0; i < nf; i++) {
		zx_ptr = &zx_f0bar [i];
		k = zx_ptr -> degree;
		CALIB_FATAL_ERROR_IF (k < 1);
		lc = &(zx_ptr -> coeff [k]);
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (lc, 1) NE 0);
	}

	/* Verify that \sum_i zx_cfact [i] * zx_y0 [i] EQ 1 (mod p).	*/
	Zx -> alloc (&TEMP1, 1);
	mpz_set_si (&TEMP1.coeff [0], -1);
	TEMP1.degree = 0;
	for (i = 0; i < nf; i++) {
		Zx -> mul (&TEMP2, &zx_cfact [i], &zx_y0 [i]);
		Zx -> add (&TEMP1, &TEMP1, &TEMP2);
	}
	Zxyz_mod_out_Zx (&TEMP1, p);
	CALIB_FATAL_ERROR_IF (TEMP1.degree NE -1);
#endif

	/* Save off the initial (mod p) versions of zx_cfact and zx_f0bar. */
	init_zx_cfact = Zxyz_new_Zx_array (2 * nf);
	init_f0bar = init_zx_cfact + nf;
	for (i = 0; i < nf; i++) {
		Zx -> set (&init_zx_cfact [i], &zx_cfact [i]);
		Zx -> set (&init_f0bar [i],    &zx_f0bar [i]);
	}

	mpz_init (BOUND);
	mpz_init (HALF_MOD);
	mpz_init (NEXT_MODULUS);
	mpz_init (temp1);
	mpz_init (temp2);

	mpz_mul_2exp (BOUND, eezgip -> B, 1);

	mpz_set (modulus, p);

	for (k = 1; mpz_cmp (modulus, BOUND) <= 0; k++) {

		mpz_fdiv_q_2exp (HALF_MOD, modulus, 1);
		mpz_mul (NEXT_MODULUS, modulus, p);

		/* First lift the d[i]. */
		for (i = 0; i < nf; i++) {
			/* Lift d[i]. */
			zx_ptr = &zx_f0fact [i];
			lc = &(zx_ptr -> coeff [zx_ptr -> degree]);
			mpz_mul (temp1, lc, &d [i]);
			mpz_neg (temp1, temp1);
			mpz_add_ui (temp1, temp1, 1);
			mpz_fdiv_qr (temp1, temp2, temp1, modulus);
			CALIB_FATAL_ERROR_IF (mpz_sgn (temp2) NE 0);
			mpz_mul (temp1, temp1, &lcf_inv [i]);
			mpz_mod (temp1, temp1, modulus);
			if (mpz_cmp (temp1, HALF_MOD) >= 0) {
				mpz_sub (temp1, temp1, modulus);
			}
			mpz_mul (temp1, temp1, modulus);
			mpz_add (&d [i], &d [i], temp1);
		}

		/* Update the zx_f0bar [i] using the new d[i], and	*/
		/* verify that they are monic.				*/
		for (i = 0; i < nf; i++) {
			zx_ptr = &zx_f0bar [i];
			Zx -> mul_z (zx_ptr, &zx_f0fact [i], &d [i]);
			Zxyz_mod_out_Zx (zx_ptr, NEXT_MODULUS);
			lc = &(zx_ptr -> coeff [zx_ptr -> degree]);
			CALIB_FATAL_ERROR_IF (mpz_cmp_ui (lc, 1) NE 0);
		}

		/* Now lift the zx_cfact.  First compute the residual. */
		Zx -> alloc (&RESID, 0);   /* Zx -> set_si (&RESID, -1); */
		mpz_set_si (&RESID.coeff [0], -1);
		RESID.degree = 0;
		for (i = 0; i < nf; i++) {
			Zx -> mul (&TEMP1, &zx_cfact [i], &zx_y0 [i]);
			Zx -> add (&RESID, &RESID, &TEMP1);
		}
		Zx -> neg (&RESID, &RESID);
		Zx -> div_z_exact (&RESID, &RESID, modulus);
		Zxyz_mod_out_Zx (&RESID, p);

		/* Now correct each zx_cfact. */
		for (i = 0; i < nf; i++) {
			Zx -> mul (&TEMP1, &RESID, &init_zx_cfact [i]);
			Zx -> div (&TEMP2, &TEMP1, temp1, &TEMP1, &init_f0bar [i]);
			/* Remainder must not be fractional. */
			CALIB_FATAL_ERROR_IF (mpz_cmp_ui (temp1, 1) NE 0);
			Zxyz_mod_out_Zx (&TEMP1, p);
			Zx -> mul_z (&TEMP1, &TEMP1, modulus);
			Zx -> add (&zx_cfact [i], &zx_cfact [i], &TEMP1);
			Zxyz_mod_out_Zx (&zx_cfact [i], NEXT_MODULUS);
		}

		mpz_set (modulus, NEXT_MODULUS);

#if 1
		/* Verify that the lifted cfact are correct. */
		Zx -> alloc (&RESID, 0);   /* Zx -> set_si (&RESID, -1); */
		mpz_set_si (&RESID.coeff [0], -1);
		RESID.degree = 0;
		for (i = 0; i < nf; i++) {
			Zx -> mul (&TEMP1, &zx_cfact [i], &zx_y0 [i]);
			Zx -> add (&RESID, &RESID, &TEMP1);
		}
		Zxyz_mod_out_Zx (&RESID, modulus);
		CALIB_FATAL_ERROR_IF (RESID.degree NE -1);
#endif
	}

	/* Remember the power k for which modulus = p^k. */
	eezgip -> p_pow_k = k;

	mpz_clear (temp2);
	mpz_clear (temp1);
	mpz_clear (NEXT_MODULUS);
	mpz_clear (HALF_MOD);
	mpz_clear (BOUND);

	Zxyz_free_Zx_array (init_zx_cfact, 2 * nf);

	Zx -> clear (&TEMP2);
	Zx -> clear (&TEMP1);
	Zx -> clear (&RESID);

	calib_free_Z_vector (d, nf);
	calib_free_Z_vector (lcf_inv, nf);
}

/*
 * Precompute a table of 1/n! (mod modulus) values for all
 * n \in [0, total_degree].
 */

	static
	void
Zxyz_eez_compute_inverse_factorials (

struct EEZglob *	eezgip	/* IN/OUT: global EEZ algorithm info */
)
{
int		k, n, p_pow_k, max_fact, total_degree;
mpz_ptr		array, recip;
mpz_srcptr	p, modulus, fact_mod;
mpz_t		TEMP, CURMOD, NEXTMOD, HALFMOD;

	CALIB_FATAL_ERROR_IF (eezgip EQ NULL);

	total_degree	= eezgip -> total_degree;
	array		= eezgip -> invfactorial;
	max_fact	= eezgip -> max_fact;
	fact_mod	= eezgip -> fact_mod;
	modulus		= eezgip -> modulus;

	CALIB_FATAL_ERROR_IF ((total_degree < 1) OR (max_fact < 0));

	if ((array NE NULL) AND
	    (max_fact >= total_degree) AND
	    (mpz_cmp (fact_mod, modulus) EQ 0)) {
		/* The existing table works. */
		return;
	}

	if (array NE NULL) {
		calib_free_Z_vector (array, max_fact + 1);
	}

	max_fact = total_degree;
	if (max_fact < 2) {
		max_fact = 2;
	}
	array = calib_new_Z_vector (max_fact + 1);
	recip = calib_new_Z_vector (max_fact + 1);

	p		= eezgip -> p;
	p_pow_k		= eezgip -> p_pow_k;
	mpz_init (TEMP);
	mpz_init (CURMOD);
	mpz_init (NEXTMOD);
	mpz_init (HALFMOD);

	/* The first two entries (1/0! and 1/1!) are both 1, no matter	*/
	/* what the modulus is.						*/
	mpz_set_ui (&array [0], 1);
	mpz_set_ui (&array [1], 1);

	/* Now initialize both the array[] and recip[] tables with	*/
	/* (1/n mod p) values.						*/
	for (n = 2; n <= max_fact; n++) {
		mpz_set_ui (TEMP, n);
		mpz_invert (TEMP, TEMP, p);
		mpz_set (&array [n], TEMP);
		mpz_set (&recip [n], TEMP);
	}

	/* Now iteratively lift array[] from p^1 to p^2, p^3, ..., p^k	*/
	/* so that we get to the current modulus = p^p_pow_k.		*/
	mpz_set (CURMOD, p);
	for (k = 1; k < p_pow_k; k++) {
		/* Lift all entries from CURMOD = p^k to NEXTMOD = p^{k+1}. */
		mpz_mul (NEXTMOD, CURMOD, p);
		for (n = 2; n <= max_fact; n++) {
			mpz_mul_ui (TEMP, &array [n], n);
			mpz_sub_ui (TEMP, TEMP, 1);
			mpz_divexact (TEMP, TEMP, CURMOD);
			mpz_mod (TEMP, TEMP, p);
			mpz_mul (TEMP, TEMP, &recip [n]);
			mpz_mod (TEMP, TEMP, p);
			mpz_mul (TEMP, TEMP, CURMOD);
			mpz_sub (TEMP, &array [n], TEMP);
			mpz_mod (TEMP, TEMP, NEXTMOD);
			mpz_set (&array [n], TEMP);
		}
		mpz_set (CURMOD, NEXTMOD);
	}

	CALIB_FATAL_ERROR_IF (mpz_cmp (CURMOD, modulus) NE 0);

	/* Now multiply successive reciprocals to obtain inverse factorials. */
	for (n = 3; n <= max_fact; n++) {
		mpz_mul (&array [n], &array [n], &array [n - 1]);
		mpz_mod (&array [n], &array [n], modulus);
	}

	/* Now correct the signs of the results. */
	mpz_tdiv_q_2exp (HALFMOD, modulus, 1);
	for (n = 2; n <= max_fact; n++) {
		if (mpz_cmp (&array [n], HALFMOD) >= 0) {
			/* Make this entry be negative. */
			mpz_sub (&array [n], &array [n], modulus);
		}
	}

	mpz_clear (HALFMOD);
	mpz_clear (NEXTMOD);
	mpz_clear (CURMOD);
	mpz_clear (TEMP);

	calib_free_Z_vector (recip, max_fact + 1);

	/* Store new inverse factorial table. */
	eezgip -> invfactorial	= array;
	eezgip -> max_fact	= total_degree;
	mpz_set (eezgip -> fact_mod, modulus);
}

/*
 * Recursively enumerate all products of the form:
 *
 *	PROD = (x_1 - b_1)^e_1 * (x_2 - b_2)^e_2 * ... * (x_t - b_t)^e_t
 *
 * such that:
 *
 *	0 <= e_i <= n_i			\forall i \in [1,t], and
 *	\sum_{i=1}^t e_i EQ pow
 *
 * where n_i is the degree of the current evpoly[] in variable i.
 *
 * For each such product (of total degree "pow"), compute the (univariate)
 * coefficient of the current residual Dm, compute univariate correction
 * factors a[] , and adjust each c2fact[] accordingly:
 *
 *		c2fact [j] += a[j] * PROD
 */

	static
	void
Zxyz_eez_cfact_recurse (

struct EEZglob *	eezgip,	/* IN/OUT: global EEZ algorithm info */
int			nlift,	/* IN: num vars currently being lifted */
const int *		maxdeg,	/* IN: max degree of each var */
int			varidx,	/* IN: current varlist[] index */
struct calib_Zxyz_obj *	Dm_ptr,	/* IN: current residual Dm */
struct calib_Zxyz_obj *	deriv,	/* IN: derivative temporary stack */
struct calib_Zxyz_obj *	prod,	/* IN: current product */
int			deg,	/* IN: degree of current product */
int			pow	/* IN: total degree desired */
)
{
int				i, e, nf, x1, xk, nk;
struct calib_Zxyz_obj *		Dm;
struct calib_Zxyz_obj *		next_Dm;
struct calib_Zxyz_obj *		next_prod;
struct calib_Zxyz_obj *		c2fact;
const struct calib_Zxyz_obj *	vvp;
struct calib_Zx_obj *		zx_cfact;
struct calib_Zx_obj *		zx_f0bar;
struct calib_Zx_obj *		zxp;
struct calib_Zx_obj *		cftemp1;
struct calib_Zx_obj *		cftemp2;
struct calib_Zx_obj *		cftemp3;
struct calib_Zxyz_obj *		cftemp4;
mpz_ptr				cftemp5;
mpz_srcptr			modulus, bk, invfactorial;
const struct calib_Zx_dom *	Zx;

#if TRACE_CFRECURSE
	printf ("Entering cfrecurse: varidx=%d, deg=%d, pow=%d\n",
		varidx, deg, pow);
#endif

	Dm = Dm_ptr;

	if (Dm -> nterms <= 0) {
		/* Current residual is zero.  Nothing of this form to	*/
		/* correct.						*/
#if TRACE_CFRECURSE
		printf ("Leaving cfrecurse - zero\n");
#endif
		return;
	}

	if (varidx < 0) {
		/* All variables have been reduced.  Our current Dm	*/
		/* should be a polynomial in x1.			*/
		if (deg NE pow) {
			CALIB_FATAL_ERROR_IF (deg > pow);
			/* Product is too low in degree. */
#if TRACE_CFRECURSE
			printf ("Leaving cfrecurse - deg < pow\n");
#endif
			return;
		}
		/* This term has the proper degree.  Decompose our	*/
		/* current Dm (which should be univariate in x1), using	*/
		/* the univariate correction factors, and apply the	*/
		/* proper multiple of this product to each c2fact[].	*/
		/* This is the other place where we are glad we have	*/
		/* gone to the trouble of making the zx_f0bar be monic.	*/

		nf		= eezgip -> nf;
		x1		= eezgip -> x1;
		c2fact		= eezgip -> c2fact;
		zx_cfact	= eezgip -> zx_cfact;
		zx_f0bar	= eezgip -> zx_f0bar;
		zxp		= eezgip -> cf_zx_temp;

		cftemp1 	= zxp++;
		cftemp2		= zxp++;
		cftemp3		= zxp++;

		cftemp4		= &(eezgip -> cf_zxyz_temp);
		cftemp5		= eezgip -> cf_z_temp;
		modulus		= eezgip -> modulus;

		Zx = calib_get_Zx_dom ();

		Zxyz_cv_poly_to_Zx_form (cftemp1, Dm, x1);
		for (i = 0; i < nf; i++) {
			Zx -> mul (cftemp2, &zx_cfact [i], cftemp1);
			Zxyz_mod_out_Zx (cftemp2, modulus);
			Zx -> div (cftemp3, cftemp2, cftemp5,
				   cftemp2, &zx_f0bar [i]);
			/* Division must not be fractional! */
			CALIB_FATAL_ERROR_IF (mpz_cmp_ui (cftemp5, 1) NE 0);
			Zxyz_mod_out_Zx (cftemp2, modulus);

			/* Convert from Z[x] to Z[x,y,z] form. */
			Zxyz_cv_from_Zx (cftemp4, cftemp2, x1);
			Zxyz_mul (cftemp4, cftemp4, prod);
			Zxyz_mod_out (cftemp4, modulus);
			Zxyz_add (&c2fact [i], &c2fact [i], cftemp4);
		}
		++(eezgip -> num_cf_terms);
#if TRACE_CFRECURSE
		printf ("Leaving cfrecurse - base case\n");
#endif
		return;
	}

	/* Handle next variable. */
	xk = eezgip -> varlist [varidx--];
	bk = &(eezgip -> bvals [xk]);
	nk = maxdeg [xk];

	/* First try this term with zero-th derivative. */
	next_Dm = Dm + 1;
	modulus	= eezgip -> modulus;
	Zxyz_eval_one_var (next_Dm, Dm, xk, bk);
	Zxyz_mod_out (next_Dm, modulus);
	Zxyz_eez_cfact_recurse (eezgip,
				nlift,
				maxdeg,
				varidx,
				next_Dm,
				deriv,
				prod,
				deg,
				pow);

	/* Now try successively higher derivatives. */
	Zxyz_set (deriv, Dm);
	next_prod = prod + 1;
	invfactorial = eezgip -> invfactorial;
	for (e = 1; ; e++) {
		if (e > nk) break;
		if (deg + e > pow) break;

		Zxyz_derivative (deriv, deriv, xk);
		Zxyz_mod_out (deriv, modulus);
		Zxyz_eval_one_var (next_Dm, deriv, xk, bk);
		Zxyz_mod_out (next_Dm, modulus);
		if (e > 1) {
			Zxyz_mul_z (next_Dm, next_Dm, &invfactorial [e]);
			Zxyz_mod_out (next_Dm, modulus);
		}

		vvp = Zxyz_eez_get_var_val_pow (eezgip, xk, e);
		Zxyz_mul (next_prod, prod, vvp);

		Zxyz_eez_cfact_recurse (eezgip,
					nlift,
					maxdeg,
					varidx,
					next_Dm,
					deriv + 1,
					next_prod,
					deg + e,
					pow);
	}

#if TRACE_CFRECURSE
	printf ("Leaving cfrecurse - end\n");
#endif
}

/*
 * Return a pointer to the cached (var-val))^m polynomial.
 * Compute this cached entry if not already available.
 */

	static
	const struct calib_Zxyz_obj *
Zxyz_eez_get_var_val_pow (

struct EEZglob *	eezgip,	/* IN: global EEZ algorithm info */
int			var,	/* IN: variable to use */
int			m	/* IN: power to get */
)
{
int			i, nvars, nterms;
struct EEZvar *		vip;
struct calib_Zxyz_obj *	mon_pows;
struct calib_Zxyz_obj *	p;
int *			ip;
mpz_ptr			cp;
mpz_srcptr		brow, mvalpows;

	CALIB_FATAL_ERROR_IF (eezgip EQ NULL);

	nvars = eezgip -> nvars;
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars) OR (var EQ eezgip -> x1));

	vip = &(eezgip -> var_info [var]);

	CALIB_FATAL_ERROR_IF ((m < 0) OR (m > vip -> maxdeg));
	mon_pows = vip -> mon_pows;
	CALIB_FATAL_ERROR_IF (mon_pows EQ NULL);

	p = &mon_pows [m];

	if (p -> nterms > 0) {
		/* This entry has already been computed.  Just return it. */
		return (p);
	}

	if (m EQ 0) {
		/* Power is zero, answer is 1. */
		Zxyz_set_si (p, 1);
		return (p);
	}

	/* Cached entry is not yet initialized.  Need to compute	*/
	/* (var - val)^m.  Use the binomial expansion.			*/

	nterms = m + 1;
	Zxyz_alloc (p, nterms);

	ip = p -> pow;
	cp = p -> coeff;

	memset (ip, 0, nvars * nterms * sizeof (int));

	brow = Zxyz_binomial_row (m, eezgip -> gip);
	mvalpows = vip -> mvalpows;

	ip [var] = m;
	mpz_set_ui (cp, 1);
	ip += nvars;
	++cp;
	for (i = 1; i < m; i++) {
		ip [var] = m - i;
		mpz_mul (cp, &brow [i], &mvalpows [i]);
		ip += nvars;
		++cp;
	}
	mpz_set (cp, &mvalpows [m]);
	p -> nterms = nterms;

	return (p);
}

/*
 * Mod out the coefficients of the given polynomial.
 */

	static
	void
Zxyz_mod_out (

struct calib_Zxyz_obj *	poly,		/* IN/OUT: polynomial to mod out */
mpz_srcptr		modulus		/* IN: modulus to use */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
int *				sip;
int *				dip;
mpz_ptr				zp, szp, dzp;
mpz_t				TEMP, HALF_MOD;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (modulus EQ NULL));

	dom = poly -> dom;
	nvars = dom -> nvars;

	mpz_init (TEMP);
	mpz_init (HALF_MOD);
	mpz_tdiv_q_2exp (HALF_MOD, modulus, 1);

	nterms	= poly -> nterms;
	zp	= poly -> coeff;
	for (i = 0; i < nterms; i++, zp++) {
		mpz_fdiv_r (TEMP, zp, modulus);
		if (mpz_cmp (TEMP, HALF_MOD) > 0) {
			/* Negate result. */
			mpz_sub (TEMP, TEMP, modulus);
		}
		mpz_set (zp, TEMP);
	}

	mpz_clear (HALF_MOD);
	mpz_clear (TEMP);

	/* Check for any coefficients that became zero. */
	sip = poly -> pow;
	szp = poly -> coeff;
	for (i = 0; ; i++) {
		if (i >= nterms) {
			/* No deletion of zero terms necessary. */
			return;
		}
		if (mpz_sgn (szp) EQ 0) break;
		sip += nvars;
		++szp;
	}

	/* We need to delete zeros starting right here. */
	dip = sip;
	dzp = szp;
	sip += nvars;
	++szp;
	++i;
	for (; i < nterms; i++) {
		if (mpz_sgn (szp) NE 0) {
			memcpy (dip, sip, nvars * sizeof (int));
			mpz_set (dzp, szp);
			dip += nvars;
			++dzp;
		}
		sip += nvars;
		++szp;
	}
	poly -> nterms = (dzp - poly -> coeff);
}

/*
 * Mod out the coefficients of the given Z[x] polynomial.
 */

	static
	void
Zxyz_mod_out_Zx (

struct calib_Zx_obj *	poly,		/* IN/OUT: polynomial to mod out */
mpz_srcptr		modulus		/* IN: modulus to use */
)
{
int			i, degree;
mpz_ptr			zp;
mpz_t			TEMP, HALF_MOD;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (modulus EQ NULL));

	mpz_init (TEMP);
	mpz_init (HALF_MOD);
	mpz_tdiv_q_2exp (HALF_MOD, modulus, 1);

	degree	= poly -> degree;
	zp	= poly -> coeff;
	for (i = 0; i <= degree; i++, zp++) {
		mpz_fdiv_r (TEMP, zp, modulus);
		if (mpz_cmp (TEMP, HALF_MOD) > 0) {
			/* Negate result. */
			mpz_sub (TEMP, TEMP, modulus);
		}
		mpz_set (zp, TEMP);
	}

	mpz_clear (HALF_MOD);
	mpz_clear (TEMP);

	/* Strip zero coefficients to obtain correct degree. */
	zp = poly -> coeff;
	while ((degree >= 0) AND (mpz_sgn (&zp [degree]) EQ 0)) {
		--degree;
	}
	poly -> degree = degree;
}

/*
 * Initialize the given EEZ algorithm global information structure.
 */

	static
	void
Zxyz_eez_init_global_info (

struct EEZglob *	eezgip,	/* IN/OUT: EEZ algorithm global info */
int			mode,	/* IN: mode: MODE_EEZGCD or MODE_FACTOR */
struct GInfo *		gip,	/* IN: global info */
const struct calib_Zxyz_dom *
			dom,	/* IN: polynomial domain */
int			nv,	/* IN: nuber of variables in r2fmap */
const int *		r2fmap,	/* IN: var ordering, least to greatest */
const calib_bool *	evflags,/* IN: which vars evaluated away */
mpz_srcptr		bvals,	/* IN: values substituted for evals vars */
int			nf	/* IN: number of factors */
)
{
int			i, nvars, x1;
int *			varlist;
struct EEZvar *		var_info;
struct calib_Zx_obj *	zx_f0fact;
struct calib_Zx_obj *	zx_y0;
struct calib_Zx_obj *	zx_cfact;
struct calib_Zx_obj *	zx_f0bar;

	CALIB_FATAL_ERROR_IF (NOT ((mode EQ MODE_EEZGCD) OR (mode EQ MODE_FACTOR)));

	CALIB_FATAL_ERROR_IF ((eezgip EQ NULL) OR
			      (gip    EQ NULL) OR
			      (dom    EQ NULL) OR
			      (nv     < 2) OR
			      (r2fmap EQ NULL) OR
			      (nf     < 2));

	nvars = dom -> nvars;

	/* Get/remove the main variable x1 from our list. */
	x1 = r2fmap [--nv];

	/* Create a list of variables we are working with, listed in	*/
	/* the order that we will be lifting them.			*/
	varlist = CALIB_NEWA (nv, int);
	for (i = 0; i < nv; i++) {
		varlist [i] = r2fmap [nv - 1 - i];
	}

	memset (eezgip, 0, sizeof (*eezgip));

	eezgip -> gip		= gip;
	eezgip -> mode		= mode;
	eezgip -> dom		= dom;
	eezgip -> nvars		= nvars;
	eezgip -> x1		= x1;
	eezgip -> nv		= nv;
	eezgip -> varlist	= varlist;
	eezgip -> r2fmap	= r2fmap;
	eezgip -> evflags	= evflags;
	eezgip -> bvals		= bvals;

	eezgip -> pg_tries	= 0;
	eezgip -> pg_nbytes	= 1;

	eezgip -> nf		= nf;

	eezgip -> factors	= Zxyz_new_array (dom, nf);
	eezgip -> pdfactors	= Zxyz_new_array (dom, nf);
	eezgip -> lfactors	= Zxyz_new_array (dom, nf);
	eezgip -> l2factors	= Zxyz_new_array (dom, nf);
	eezgip -> ypolys	= Zxyz_new_array (dom, nf);
	eezgip -> f0fact	= Zxyz_new_array (dom, nf);

	zx_f0fact	= Zxyz_new_Zx_array (4 * nf);
	zx_y0		= zx_f0fact + nf;
	zx_cfact	= zx_y0 + nf;
	zx_f0bar	= zx_cfact + nf;

	eezgip -> zx_f0fact	= zx_f0fact;
	eezgip -> zx_y0		= zx_y0;
	eezgip -> zx_cfact	= zx_cfact;
	eezgip -> zx_f0bar	= zx_f0bar;

	eezgip -> cfact		= Zxyz_new_array (dom, nf);
	eezgip -> c2fact	= Zxyz_new_array (dom, nf);
	eezgip -> Dm_stack	= Zxyz_new_array (dom, nv + 1);
	eezgip -> deriv_stack	= Zxyz_new_array (dom, nv);
	eezgip -> prodstack	= Zxyz_new_array (dom, nv + 1);

	eezgip -> cf_zx_temp	= Zxyz_new_Zx_array (3);
	Zxyz_init (dom, &(eezgip -> cf_zxyz_temp));
	mpz_init (eezgip -> cf_z_temp);

	mpz_init (eezgip -> fact_mod);

	var_info = CALIB_NEWA (nvars, struct EEZvar);
	memset (var_info, 0, nvars * sizeof (var_info [0]));
	eezgip -> var_info	= var_info;

	/* Use p = 3 as a placeholder.  We adjust as needed. */
	eezgip -> Zp		= calib_make_Zp_dom_ui (3);
	eezgip -> Zpx		= calib_make_Zpx_dom (eezgip -> Zp);

	mpz_init (eezgip -> B);
	mpz_init (eezgip -> p);
	mpz_init (eezgip -> modulus);
}

/*
 * Clean up the given EEZ algorithm global information structure.
 */

	static
	void
Zxyz_eez_clear_global_info (

struct EEZglob *	eezgip	/* IN/OUT: EEZ algorithm global info */
)
{
int			i, nf, nv, nvars;
struct EEZvar *		var_info;

	CALIB_FATAL_ERROR_IF (eezgip EQ NULL);

	nf	= eezgip -> nf;
	nvars	= eezgip -> nvars;
	nv	= eezgip -> nv;

	var_info = eezgip -> var_info;
	if (var_info NE NULL) {
		for (i = 0; i < nvars; i++) {
			Zxyz_eez_clear_var_info (&var_info [i]);
		}
		free (var_info);
		eezgip -> var_info = NULL;
	}

	if (eezgip -> evpoly NE NULL) {
		Zxyz_free_array (eezgip -> evpoly, nv + 1);
	}

	if (eezgip -> invfactorial NE NULL) {
		calib_free_Z_vector (eezgip -> invfactorial, eezgip -> max_fact + 1);
	}

	mpz_clear (eezgip -> fact_mod);

	mpz_clear (eezgip -> cf_z_temp);
	Zxyz_clear (&(eezgip -> cf_zxyz_temp));
	Zxyz_free_Zx_array (eezgip -> cf_zx_temp, 3);

	Zxyz_free_array (eezgip -> prodstack, nv + 1);
	Zxyz_free_array (eezgip -> deriv_stack, nv);
	Zxyz_free_array (eezgip -> Dm_stack, nv + 1);
	Zxyz_free_array (eezgip -> c2fact, nf);
	Zxyz_free_array (eezgip -> cfact, nf);

	Zxyz_free_Zx_array (eezgip -> zx_f0fact, 4 * nf);

	Zxyz_free_array (eezgip -> f0fact, nf);
	Zxyz_free_array (eezgip -> ypolys, nf);
	Zxyz_free_array (eezgip -> l2factors, nf);
	Zxyz_free_array (eezgip -> lfactors, nf);
	Zxyz_free_array (eezgip -> pdfactors, nf);
	Zxyz_free_array (eezgip -> factors, nf);

	calib_free_Zpx_dom (eezgip -> Zpx);
	calib_free_Zp_dom (eezgip -> Zp);

	if (eezgip -> varlist NE NULL) {
		free (eezgip -> varlist);
	}

	mpz_clear (eezgip -> modulus);
	mpz_clear (eezgip -> p);
	mpz_clear (eezgip -> B);
}

/*
 * Clear out a single EEZvar object.
 */

	static
	void
Zxyz_eez_clear_var_info (

struct EEZvar *		vip		/* IN: var info to clear */
)
{
int		nelts;

	nelts = vip -> maxdeg + 1;

	if (vip -> pvalpows NE NULL) {
		/* These two are allocated as one. */
		calib_free_Z_vector (vip -> pvalpows, 2 * nelts);
		vip -> pvalpows = NULL;
		vip -> mvalpows = NULL;
	}

	if (vip -> mon_pows NE NULL) {
		Zxyz_free_array (vip -> mon_pows, nelts);
		vip -> mon_pows = NULL;
	}
}

/*
 * Compute res = resultant (a, b, var), where a and b are polynomials
 * in the same domain.
 *
 * This implementation is quite efficient, but still has execution
 * time that is exponential in the number of variables.  (The size of
 * the solution can be exponential.)
 */

	static
	void
Zxyz_resultant (

struct calib_Zxyz_obj *		result,		/* IN/OUT: Result */
const struct calib_Zxyz_obj *	a,		/* IN: Polynomial a */
const struct calib_Zxyz_obj *	b,		/* IN: Polynomial b */
int				var		/* IN: Which var to eliminate */
)
{
int				i, nvars;
int				adeg, bdeg, a_mod_deg, b_mod_deg, start_var;
calib_int64u			p;
int *				a_mindeg;
int *				a_maxdeg;
int *				b_mindeg;
int *				b_maxdeg;
mpz_ptr				cp;
const struct calib_Zxyz_dom *	dom;
struct calib_Zp_dom *		Zp;
struct calib_Zpx_dom *		Zpx;
struct calib_Zxyz_obj		C, R, A_MOD_POLY, B_MOD_POLY;
mpz_t				MODULUS, ACNORM, BCNORM, UB, LB, TEMP;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (a EQ NULL) OR (b EQ NULL));
	dom	= result -> dom;
	CALIB_FATAL_ERROR_IF ((dom	EQ NULL) OR
			      (a -> dom NE dom) OR
			      (b -> dom NE dom));
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	/* Get degrees of each variable of each polynomial. */
	Zxyz_get_poly_degrees (&a_mindeg, &a_maxdeg, a);
	Zxyz_get_poly_degrees (&b_mindeg, &b_maxdeg, b);

	adeg	= a_maxdeg [var];
	bdeg	= b_maxdeg [var];

	CALIB_FATAL_ERROR_IF ((adeg < 0) OR (bdeg < 0));

	/* Use degree EQ -1 to represent identically zero polynomials. */
	if (a -> nterms <= 0) {
		adeg = -1;
	}
	if (b -> nterms <= 0) {
		bdeg = -1;
	}

	/* Result C is initially zero. */
	Zxyz_init (dom, &C);

	if (adeg <= 0) {
		/* Polynomial a is free of var. */
		if (adeg < 0) {
			/* Polynomial a is identically zero. */
			if (bdeg <= 0) {
				/* res(0, const) ==> 1. */
				Zxyz_set_si (&C, 1);
			}
			else {
				/* res(0, poly) ==> 0. */
			}
		}
		else {
			/* Poly a is a non-zero constant. */
			if (bdeg <= 0) {
				/* res(const1, const2) ==> 1. */
				Zxyz_set_si (&C, 1);
			}
			else {
				/* res(const, poly) ==> a**bdeg. */
				Zxyz_ipow (&C, a, bdeg);
			}
		}
	}
	else if (bdeg <= 0) {
		/* Polynomial b is free of var, a is not. */
		if (bdeg < 0) {
			/* res(poly, 0) ==> 0. */
		}
		else {
			/* res(poly, const) ==> b**adeg. */
			Zxyz_ipow (&C, b, adeg);
		}
	}
	else {
		/* Both a and b have positive degree. */

		mpz_init (MODULUS);
		mpz_init (ACNORM);
		mpz_init (BCNORM);
		mpz_init (UB);
		mpz_init (LB);
		mpz_init (TEMP);

		Zxyz_init (dom, &R);
		Zxyz_init (dom, &A_MOD_POLY);
		Zxyz_init (dom, &B_MOD_POLY);

		/* Compute upper bound on magnitude of any coefficient	*/
		/* in the result.					*/

		/* Compute maximum norm of a and b's coefficients. */
		Zxyz_res_coeff_norm (ACNORM, a, var);
		Zxyz_res_coeff_norm (BCNORM, b, var);

		mpz_pow_ui (UB,   ACNORM, bdeg);
		mpz_pow_ui (TEMP, BCNORM, adeg);
		mpz_mul    (UB, UB, TEMP);
		mpz_fac_ui (TEMP, adeg + bdeg);
		mpz_mul    (UB, UB, TEMP);

		/* Compute lower bound on modulus that we need in order	*/
		/* to represent the signed result.			*/
		mpz_mul_ui (LB, UB, 2);
		mpz_add_ui (LB, LB, 1);

		/* Get highest numbered variable to eliminate via	*/
		/* evaluation.						*/
		start_var = nvars - 1;
		if (start_var EQ var) {
			--start_var;
		}

		Zp = calib_make_Zp_dom_ui (3);
		Zpx = calib_make_Zpx_dom (Zp);

		/* Start with largest 64-bit prime. */
		p = 18446744073709551557lu;

		for (;;) {
			/* Reduce polynomials a and b modulo p. */
			mpz_set_ui (Zp -> p, p);
			Zxyz_res_mod_Zp (&A_MOD_POLY, a, Zp);
			Zxyz_res_mod_Zp (&B_MOD_POLY, b, Zp);

			a_mod_deg = Zxyz_degree_in_var (&A_MOD_POLY, var);
			b_mod_deg = Zxyz_degree_in_var (&B_MOD_POLY, var);

			if ((a_mod_deg EQ adeg) AND (b_mod_deg EQ bdeg)) {
				/* Degrees are preserved.  This is an	*/
				/* acceptable modulus.			*/
				break;
			}
			p = Zxyz_res_prev_prime (p);
		}

		/* Compute C = res(a,b,var), modulo p. */
		Zxyz_res_recurse (start_var,
				  &C,
				  &A_MOD_POLY,
				  &B_MOD_POLY,
				  Zpx,
				  p,
				  var);

		mpz_set_ui (MODULUS, p);

		while (mpz_cmp (MODULUS, LB) < 0) {
			/* Get next smallest acceptable prime p. */
			for (;;) {
				p = Zxyz_res_prev_prime (p);

				/* Reduce polynomials a and b modulo p. */
				mpz_set_ui (Zp -> p, p);
				Zxyz_res_mod_Zp (&A_MOD_POLY, a, Zp);
				Zxyz_res_mod_Zp (&B_MOD_POLY, b, Zp);

				a_mod_deg = Zxyz_degree_in_var (&A_MOD_POLY, var);
				b_mod_deg = Zxyz_degree_in_var (&B_MOD_POLY, var);

				if ((a_mod_deg EQ adeg) AND
				    (b_mod_deg EQ bdeg)) {
					/* Degrees are preserved.	*/
					/* p is an acceptable modulus.	*/
					break;
				}
			}

			/* Compute R = res(a,b,var), modulo p. */
			Zxyz_res_recurse (start_var,
					  &R,
					  &A_MOD_POLY,
					  &B_MOD_POLY,
					  Zpx,
					  p,
					  var);

			/* Use Chinese Remainder Theorem to combine	*/
			/* R with C. */
			Zxyz_res_CRT_poly (&C, &R, MODULUS, p);

			mpz_mul_ui (MODULUS, MODULUS, p);
		}

		/* Adjust signs to negative. */
		for (i = 0; i < C.nterms; i++, cp++) {
			cp = &(C.coeff [i]);
			mpz_mul_ui (TEMP, cp, 2);
			if (mpz_cmp (TEMP, MODULUS) > 0) {
				mpz_sub (cp, cp, MODULUS);
			}
		}

		calib_free_Zpx_dom (Zpx);
		calib_free_Zp_dom  (Zp);

		Zxyz_clear (&B_MOD_POLY);
		Zxyz_clear (&A_MOD_POLY);
		Zxyz_clear (&R);

		mpz_clear (TEMP);
		mpz_clear (LB);
		mpz_clear (UB);
		mpz_clear (BCNORM);
		mpz_clear (ACNORM);
		mpz_clear (MODULUS);
	}

	Zxyz_set (result, &C);

	Zxyz_clear (&C);

	free (b_mindeg);
	free (a_mindeg);
}

/*
 * Compute the maximum norm of the coefficients of the given polynomial,
 * viewing it as a polynomial in the given variable.
 */

	static
	void
Zxyz_res_coeff_norm (

mpz_ptr				cnorm,	/* IN/OUT: Max coefficient norm */
const struct calib_Zxyz_obj *	poly,	/* IN: Polynomial */
int				var	/* IN: Main variable */
)
{
int			i, deg;
struct calib_Zxyz_obj *	cpolys;
mpz_t			MAX, NORM;

	Zxyz_make_poly_of_var (&deg, &cpolys, poly, var);

	mpz_init (MAX);
	mpz_init (NORM);
	for (i = 0; i <= deg; i++) {
		Zxyz_res_norm (NORM, &cpolys [i]);
		if (mpz_cmp (NORM, MAX) > 0) {
			mpz_set (MAX, NORM);
		}
	}

	mpz_set (cnorm, MAX);

	mpz_clear (NORM);
	mpz_clear (MAX);

	Zxyz_clear_array (cpolys, deg + 1);
	free (cpolys);
}

/*
 * Compute the "norm" of the given polynomial, as used by the resultant
 * algorithm.  This is just the L1 norm of the coefficients.
 */

	static
	void
Zxyz_res_norm (

mpz_ptr				norm,	/* IN/OUT: Value of the norm */
const struct calib_Zxyz_obj *	poly	/* IN: Polynomial for which to compute norm */
)
{
int				i, nterms;
const struct calib_Zxyz_dom *	dom;
mpz_t				NORM, MAG;

	CALIB_FATAL_ERROR_IF ((norm EQ NULL) OR (poly EQ NULL));

	dom = poly -> dom;
	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	nterms = poly -> nterms;

	mpz_init (NORM);
	mpz_init (MAG);

	for (i = 0; i < nterms; i++) {
		mpz_abs (MAG, &(poly -> coeff [i]));
		mpz_add (NORM, NORM, MAG);
	}

	mpz_set (norm, NORM);

	mpz_clear (MAG);
	mpz_clear (NORM);
}

/*
 * Reduce the given polynomial to coefficients in Zp.
 */

	static
	void
Zxyz_res_mod_Zp (

struct calib_Zxyz_obj *		result,	/* IN/OUT: Result polynomial */
const struct calib_Zxyz_obj *	poly,	/* IN: Polynomial to reduce */
const struct calib_Zp_dom *	Zp	/* IN: Prime domain to reduce to */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
mpz_srcptr			coeff, srcp;
const int *			pow;
mpz_t				COEFF;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR
			      (poly   EQ NULL) OR
			      (Zp     EQ NULL));

	nterms	= poly -> nterms;
	dom	= poly -> dom;
	pow	= poly -> pow;
	coeff	= poly -> coeff;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);
	CALIB_FATAL_ERROR_IF (result -> dom NE dom);

	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF (nvars <= 0);

	pap = Zxyz_get_paccum (dom);

	srcp	= coeff;

	mpz_init (COEFF);
	for (i = 0; i < nterms; i++, srcp++, pow += nvars) {
		Zp -> set_z (Zp, COEFF, srcp);

		if (mpz_sgn (COEFF) NE 0) {
			Zxyz_paccum_add_term (pap, COEFF, pow);
		}
	}
	mpz_clear (COEFF);

	Zxyz_set_poly_from_paccum (result, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Return the next lowest odd prime number.
 */

	static
	calib_int64u
Zxyz_res_prev_prime (

calib_int64u		p		/* IN: Current prime number */
)
{
int		reps;
mpz_t		NUM;

	CALIB_FATAL_ERROR_IF ((p & 0x01) EQ 0);

	mpz_init (NUM);

	reps = 30;

	for (;;) {
		CALIB_FATAL_ERROR_IF (p <= 3);
		p -= 2;

		mpz_set_ui (NUM, p);
		if (mpz_probab_prime_p (NUM, reps)) {
			/* We have p is prime, with very high probability. */
			break;
		}
	}

	mpz_clear (NUM);

	return (p);
}

/*
 * The recursive portion of the resultant routine.  We process non-main
 * variable "i" at this level.  (All non-main variables > i have been
 * evaluated away.)
 */

	static
	void
Zxyz_res_recurse (

int				i,	/* IN: Non-main var to process */
struct calib_Zxyz_obj *		C,	/* IN/OUT: Result */
const struct calib_Zxyz_obj *	A,	/* IN: Polynomial A (mod p) */
const struct calib_Zxyz_obj *	B,	/* IN: Polynomial B (mod p) */
const struct calib_Zpx_dom *	Zpx,	/* IN: The Zp[x] domain */
calib_int64u			p,	/* IN: Current prime modulus p */
int				var	/* IN: Main var of resultant */
)
{
int				nvars, k, next_var;
int				avdeg, bvdeg, aideg, bideg, adeg2, bdeg2;
calib_int64s			b;
int *				a_mindeg;
int *				a_maxdeg;
int *				b_mindeg;
int *				b_maxdeg;
mpz_ptr				dst;
const struct calib_Zxyz_dom *	dom;
const struct calib_Zp_dom *	Zp;
struct calib_Zxyz_obj		G, R, CPOLY, A_EVPOLY, B_EVPOLY, VARPOLY, TEMP;
struct calib_Zpx_obj		APOLY, BPOLY;
struct calib_Zpx_obj		D, DFACTOR;
mpz_t				BVAL, DVAL, ZVAL;

	CALIB_FATAL_ERROR_IF ((C EQ NULL) OR (A EQ NULL) OR (B EQ NULL));
	CALIB_FATAL_ERROR_IF (i EQ var);

	dom	= C -> dom;
	CALIB_FATAL_ERROR_IF ((dom	EQ NULL) OR
			      (A -> dom NE dom) OR
			      (B -> dom NE dom));

	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((i >= nvars) OR (var < 0) OR (var >= nvars));

	if (i < 0) {
		/* The base case.  The only un-evaluated variable	*/
		/* remaining is var.  Use Zp[x] module to compute the	*/
		/* resultant over Zp[x].				*/
		mpz_init (ZVAL);

		Zpx -> init (Zpx, &APOLY);
		Zpx -> init (Zpx, &BPOLY);

		Zxyz_copy_into_Zpx (&APOLY, A, var);
		Zxyz_copy_into_Zpx (&BPOLY, B, var);

		Zpx -> resultant (ZVAL, &APOLY, &BPOLY);

		Zpx -> clear (&BPOLY);
		Zpx -> clear (&APOLY);

		dom -> set_z (C, ZVAL);

		mpz_clear (ZVAL);

		return;
	}

	/* Recursive case.  determine next variable to evaluate. */
	next_var = i - 1;
	if (next_var EQ var) {
		--next_var;
	}

	/* Compute degree bound k. */
	Zxyz_get_poly_degrees (&a_mindeg, &a_maxdeg, A);
	Zxyz_get_poly_degrees (&b_mindeg, &b_maxdeg, B);

	avdeg	= a_maxdeg [var];
	bvdeg	= b_maxdeg [var];
	aideg	= a_maxdeg [i];
	bideg	= b_maxdeg [i];

	k = avdeg * bideg + bvdeg * aideg + 1;

	free (b_mindeg);
	free (a_mindeg);

	mpz_init (BVAL);

	Zxyz_init (dom, &G);
	Zxyz_init (dom, &R);
	Zxyz_init (dom, &CPOLY);
	Zxyz_init (dom, &A_EVPOLY);
	Zxyz_init (dom, &B_EVPOLY);
	Zxyz_init (dom, &VARPOLY);
	Zxyz_init (dom, &TEMP);

	Zpx -> init_degree (Zpx, &D, 0);
	Zpx -> init_degree (Zpx, &DFACTOR, 1);
	mpz_init (DVAL);

	/* D = 1. */
	mpz_set_ui (&D.coeff [0], 1);
	D.degree = 0;

	/* DFACTOR = (x + 0). */
	mpz_set_ui (&DFACTOR.coeff [1], 1);
	DFACTOR.degree = 1;

	Zp = Zpx -> cf;

	b = 0;

	for (;;) {
		/* Evaluate variable i. */
		mpz_set_ui (BVAL, b);
		Zxyz_res_eval_var_Zp (&A_EVPOLY, A, BVAL, i, Zp);
		Zxyz_res_eval_var_Zp (&B_EVPOLY, B, BVAL, i, Zp);

		/* Check degrees. */
		adeg2 = Zxyz_degree_in_var (&A_EVPOLY, var);
		bdeg2 = Zxyz_degree_in_var (&B_EVPOLY, var);
		if ((adeg2 NE avdeg) OR (bdeg2 NE bvdeg)) {
			/* Must choose a new evaluation point. */
			++b;
			CALIB_FATAL_ERROR_IF (b EQ p);
			continue;
		}

		Zxyz_res_recurse (next_var,
				  &R,
				  &A_EVPOLY,
				  &B_EVPOLY,
				  Zpx,
				  p,
				  var);

		/* Interpolate polynomial G. */
		Zpx -> eval (DVAL, &D, BVAL);
		Zp  -> inv (Zp, DVAL, DVAL);

		/* Get D(var) in Z[xzy] form. */
		Zxyz_copy_from_Zpx (&VARPOLY, &D, i);
		Zxyz_mul_z (&VARPOLY, &VARPOLY, DVAL);
		Zxyz_res_mod_Zp (&VARPOLY, &VARPOLY, Zp);

		Zxyz_res_eval_var_Zp (&TEMP, &CPOLY, BVAL, i, Zp);
		Zxyz_sub (&TEMP, &R, &TEMP);
		Zxyz_res_mod_Zp (&TEMP, &TEMP, Zp);
		Zxyz_mul (&TEMP, &TEMP, &VARPOLY);
		Zxyz_res_mod_Zp (&TEMP, &TEMP, Zp);
		Zxyz_add (&G, &TEMP, &CPOLY);
		Zxyz_res_mod_Zp (&G, &G, Zp);

		Zxyz_set (&CPOLY, &G);

		/* Update D. */
		dst = &DFACTOR.coeff [0];
		mpz_set_ui (dst, b);
		Zp -> neg (Zp, dst, dst);

		Zpx -> mul (&D, &D, &DFACTOR);

		if (D.degree > k) {
			/* We have interpolated to a sufficiently hih	*/
			/* degree.					*/
			break;
		}

		++b;
		CALIB_FATAL_ERROR_IF (b EQ p);
	}

	Zxyz_set (C, &CPOLY);

	mpz_clear (DVAL);
	Zpx -> clear (&DFACTOR);
	Zpx -> clear (&D);

	Zxyz_clear (&TEMP);
	Zxyz_clear (&VARPOLY);
	Zxyz_clear (&B_EVPOLY);
	Zxyz_clear (&A_EVPOLY);
	Zxyz_clear (&CPOLY);
	Zxyz_clear (&R);
	Zxyz_clear (&G);

	mpz_clear (BVAL);
}

/*
 * Evaluate a single variable of given polynomial using Zp arithmetic.
 */

	static
	void
Zxyz_res_eval_var_Zp (

struct calib_Zxyz_obj *		result,	/* IN/OUT: Result */
const struct calib_Zxyz_obj *	poly,	/* IN: Polynomial to evaluate */
mpz_srcptr			value,	/* IN: Value to plug in */
int				var,	/* IN: Variable to evaluate away */
const struct calib_Zp_dom *	Zp	/* IN: Zp coefficient domain */
)
{
int				i, k, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
const int *			pow;
mpz_srcptr			srcp;
mpz_ptr				cp1, cp2;
int *				ip1;
int *				ip2;
struct paccum *			pap;
struct calib_Zxyz_obj		POLY;
mpz_t				TEMP;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR
			      (poly   EQ NULL) OR
			      (value  EQ NULL));

	nterms	= poly -> nterms;
	dom	= poly -> dom;
	pow	= poly -> pow;
	srcp	= poly -> coeff;

	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	int	new_pow [nvars];

	pap = Zxyz_get_paccum (dom);

	mpz_init (TEMP);

	for (i = 0; i < nterms; i++, srcp++, pow += nvars) {
		k = pow [var];

		memcpy (new_pow, pow, nvars * sizeof (new_pow [0]));
		new_pow [var] = 0;

		Zp -> ipow (Zp, TEMP, value, k);		
		Zp -> mul (Zp, TEMP, TEMP, srcp);

		Zxyz_paccum_add_term (pap, TEMP, new_pow);
	}

	mpz_clear (TEMP);

	Zxyz_init (dom, &POLY);

	Zxyz_set_poly_from_paccum (&POLY, pap);

	Zxyz_put_paccum (pap);

	/* The Zxyz_paccum_add_term() routine uses arithmetic in Z (not	*/
	/* Zp) to combine equivalent terms.  Mod out all of the		*/
	/* coefficients, and get rid of those that become zero.		*/

	nterms = POLY.nterms;

	cp1	= POLY.coeff;
	cp2	= POLY.coeff;
	ip1	= POLY.pow;
	ip2	= POLY.pow;
	k	= 0;
	for (i = 0; i < nterms; i++, cp1++, ip1 += nvars) {
		/* Mod out this coefficient. */
		Zp -> set_z (Zp, cp1, cp1);
		if (mpz_sgn (cp1) NE 0) {
			/* Retain this term. */
			if (cp2 < cp1) {
				mpz_set (cp2, cp1);
				memcpy (ip2, ip1, nvars * sizeof (ip2 [0]));
			}
			++cp2;
			ip2 += nvars;
			++k;
		}
	}
	POLY.nterms = k;

	Zxyz_set (result, &POLY);

	dom -> clear (&POLY);
}

/*
 * Use the Chinese Remainder Theorem to combine:
 *
 *		polynomial R (mod p)
 *		polynomial C (mod MODULUS)
 *
 * to produce an *updated* polynomial C that is correct modulus (MODULUS * p).
 */

	static
	void
Zxyz_res_CRT_poly (

struct calib_Zxyz_obj *	C,		/* IN/OUT: Polynomial C to update */
const struct calib_Zxyz_obj *
			R,		/* IN: Polynomial R (mod p) */
mpz_srcptr		MODULUS,	/* IN: Modulus for polynomial C */
calib_int64u		p		/* IN: Modulus for polynomial R */
)
{
int				nterms1, nterms2, nvars, cmp;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
mpz_srcptr			src1, src2, end1, end2;
const int *			pow1;
const int *			pow2;
mpz_t				G, S, T, PVAL, PTIMESMOD, TEMP1, TEMP2;

	CALIB_FATAL_ERROR_IF ((C EQ NULL) OR (R EQ NULL) OR (MODULUS EQ NULL));

	dom = C -> dom;
	CALIB_FATAL_ERROR_IF ((dom EQ NULL) OR (R -> dom NE dom));

	nvars = dom -> nvars;

	nterms1	= C -> nterms;
	pow1	= C -> pow;
	src1	= C -> coeff;
	end1	= src1 + nterms1;

	nterms2	= R -> nterms;
	pow2	= R -> pow;
	src2	= R -> coeff;
	end2	= src2 + nterms2;

	mpz_init (G);
	mpz_init (S);
	mpz_init (T);
	mpz_init (PVAL);
	mpz_init (PTIMESMOD);
	mpz_init (TEMP1);
	mpz_init (TEMP2);

	mpz_set_ui (PVAL, p);
	mpz_mul_ui (PTIMESMOD, MODULUS, p);

	pap = Zxyz_get_paccum (dom);

	mpz_gcdext (G, S, T, MODULUS, PVAL);
	CALIB_FATAL_ERROR_IF (mpz_cmp_ui (G, 1) NE 0);

	/* First loop runs so long as both polynomials have terms left. */
	for (;;) {
		if (src1 >= end1) break;
		if (src2 >= end2) break;

		/* Chinese Remainder Theorem update:		*/
		/*	x = src1 * T * P + src2 * S * MODULUS	*/

		cmp = Zxyz_compare_pow (nvars, pow1, pow2);
		if (cmp EQ 0) {
			/* Both polynomials have this term.  Combine	*/
			/* them using CRT.				*/
			mpz_mul (TEMP1, src1, T);
			mpz_mul_ui (TEMP1, TEMP1, p);
			mpz_mul (TEMP2, src2, S);
			mpz_mul (TEMP2, TEMP2, MODULUS);
			mpz_add (TEMP1, TEMP1, TEMP2);
			mpz_fdiv_r (TEMP1, TEMP1, PTIMESMOD);

			Zxyz_paccum_add_term (pap, TEMP1, pow1);

			++src1;
			++src2;
			pow1 += nvars;
			pow2 += nvars;
		}
		else if (cmp < 0) {
			/* Only first polynomial has this term. */
			mpz_mul (TEMP1, src1, T);
			mpz_mul_ui (TEMP1, TEMP1, p);
			mpz_fdiv_r (TEMP1, TEMP1, PTIMESMOD);

			Zxyz_paccum_add_term (pap, TEMP1, pow1);

			++src1;
			pow1 += nvars;
		}
		else {
			/* Only second polynomial has this term. */
			mpz_mul (TEMP2, src2, S);
			mpz_mul (TEMP2, TEMP2, MODULUS);
			mpz_fdiv_r (TEMP2, TEMP2, PTIMESMOD);

			Zxyz_paccum_add_term (pap, TEMP2, pow2);

			++src2;
			pow2 += nvars;
		}
	}

	/* Process any remaining terms of first polynomial. */
	while (src1 < end1) {
		mpz_mul (TEMP1, src1, T);
		mpz_mul_ui (TEMP1, TEMP1, p);
		mpz_fdiv_r (TEMP1, TEMP1, PTIMESMOD);

		Zxyz_paccum_add_term (pap, TEMP1, pow1);

		++src1;
		pow1 += nvars;
	}

	/* Process any remaining terms of second polynomial. */
	while (src2 < end2) {
		mpz_mul (TEMP2, src2, S);
		mpz_mul (TEMP2, TEMP2, MODULUS);
		mpz_fdiv_r (TEMP2, TEMP2, PTIMESMOD);

		Zxyz_paccum_add_term (pap, TEMP2, pow2);

		++src2;
		pow2 += nvars;
	}

	mpz_clear (TEMP2);
	mpz_clear (TEMP1);
	mpz_clear (PTIMESMOD);
	mpz_clear (PVAL);
	mpz_clear (T);
	mpz_clear (S);
	mpz_clear (G);

	Zxyz_set_poly_from_paccum (C, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Compute res = resultant (a, b, var), where a and b are polynomials
 * in the same domain.  Returns result as a dynamically allocated
 * list of factors, some of which may be integer constants.  Caller
 * must free this list.
 *
 * WARNING: This is a very old and naive implementation!!!  Use only
 * on small polynomials of fairly low degree!
 */

	static
	struct calib_Zxyz_factor *
Zxyz_resultant_old (

const struct calib_Zxyz_obj *	a,	/* IN: Polynomial a */
const struct calib_Zxyz_obj *	b,	/* IN: Polynomial b */
int				var	/* IN: Which var to eliminate */
)
{
int				i, nvars;
int				adeg, bdeg, itemp, maxdeg, qdeg, rdeg;
int				fdeg, gdeg;
calib_bool			result_is_zero;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		apoly;
struct calib_Zxyz_obj *		bpoly;
struct calib_Zxyz_obj *		f;
struct calib_Zxyz_obj *		g;
struct calib_Zxyz_obj *		ptemp;
struct calib_Zxyz_obj *		quot;
struct calib_Zxyz_obj *		rem;
struct calib_Zxyz_obj *		lc;
struct calib_Zxyz_factor *	result;
struct calib_Zxyz_factor *	nfp;
const struct calib_Zxyz_obj *	polys [2];
struct calib_Zxyz_obj *		cofact [2];
struct calib_Zxyz_obj		LC_GCD, A_COFACT, B_COFACT, D;
struct calib_RatFac		COEFF;

	CALIB_FATAL_ERROR_IF ((a EQ NULL) OR (b EQ NULL));
	dom	= a -> dom;
	CALIB_FATAL_ERROR_IF ((dom EQ NULL) OR (b -> dom NE dom));
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	Zxyz_init (dom, &LC_GCD);
	Zxyz_init (dom, &A_COFACT);
	Zxyz_init (dom, &B_COFACT);
	Zxyz_init (dom, &D);

	/* We keep this in partially factored form.  Otherwise it tends	*/
	/* to explode into a huge, dense mess.				*/
	_calib_RatFac_init (dom, &COEFF);

	/* Let adeg be the degree of "a" in given variable "var".	*/
	/* Convert it into an array of "adeg + 1" polynomials, the i-th	*/
	/* element of which is the coeficient of "var**i" expressed as	*/
	/* a polynomial in all other variables.				*/
	Zxyz_make_poly_of_var (&adeg, &apoly, a, var);

	/* Ditto for "b". */
	Zxyz_make_poly_of_var (&bdeg, &bpoly, b, var);

	maxdeg = adeg;
	if (maxdeg < bdeg) {
		maxdeg = bdeg;
	}

	/* Remove any GCD from lc(a), lc(b).  We can multiply the final	*/
	/* answer by LC_GCD**(adeg + bdeg).				*/
	polys [0]	= &apoly [adeg];
	polys [1]	= &bpoly [bdeg];
	cofact [0]	= &A_COFACT;
	cofact [1]	= &B_COFACT;
	dom -> gcd_n (&LC_GCD, &cofact [0], 2, &polys [0]);

	if (NOT (Zxyz_is_constant_poly (&LC_GCD))) {
		/* Have a non-constant LC_GCD.  Remove it from both	*/
		/* lc(apoly) and lc(bpoly).				*/
		Zxyz_set (&apoly [adeg], &A_COFACT);
		Zxyz_set (&bpoly [bdeg], &B_COFACT);

		_calib_RatFac_mul (&COEFF, &LC_GCD, adeg + bdeg);
	}

	quot	= CALIB_NEWA (maxdeg + 1, struct calib_Zxyz_obj);
	rem	= CALIB_NEWA (maxdeg + 1, struct calib_Zxyz_obj);
	Zxyz_init_array (dom, quot, maxdeg + 1);
	Zxyz_init_array (dom, rem,  maxdeg + 1);

	f	= apoly;
	fdeg	= adeg;
	g	= bpoly;
	gdeg	= bdeg;

	result_is_zero = FALSE;
	for (;;) {
		if (fdeg > gdeg) {
			ptemp = f;	f = g;		g = ptemp;
			itemp = fdeg;	fdeg = gdeg;	gdeg = itemp;
			if ((fdeg & gdeg & 0x01) NE 0) {
				/* Both of odd degree.  Negate. */
				_calib_RatFac_negate (&COEFF);
			}
			continue;
		}
		lc = &f [fdeg];
		if (fdeg <= 0) {
			_calib_RatFac_mul (&COEFF, lc, gdeg);
			break;
		}
		/* Compute quot, rem, d, such that D*g = quot*f + rem,	*/
		/* and degree(rem) < degree(f) (in var).		*/
		Zxyz_res_pseudo_div (quot, &qdeg,
				     rem,  &rdeg,
				     &D,
				     g,    gdeg,
				     f,    fdeg);
		if (rdeg < 0) {
			result_is_zero = TRUE;
			break;
		}
		/* We have res(f,d*g) = d^n*res(f,g), so this means	*/
		/* res(f,g) = res(f, d*g)/d^n.  This denominator must	*/
		/* cancel before we are all done.			*/

		_calib_RatFac_mul (&COEFF, &D, - fdeg);

		_calib_RatFac_mul (&COEFF, lc, gdeg - rdeg);

		for (i = 0; i <= rdeg; i++) {
			Zxyz_set (&g [i], &rem [i]);
		}
		gdeg = rdeg;
	}

	if (result_is_zero) {
		struct calib_Zxyz_obj		TEMP;
		Zxyz_init (dom, &TEMP);
		TEMP.nterms = 0;
		nfp = CALIB_NEW (struct calib_Zxyz_factor);
		nfp -> multiplicity	= 1;
		nfp -> factor		= Zxyz_dup (&TEMP);
		nfp -> next		= NULL;

		Zxyz_clear (&TEMP);

		result = nfp;
	}
	else {
		/* Denominator must now be 1. */
		nfp = _calib_RatFac_denom (&COEFF);
		CALIB_FATAL_ERROR_IF (nfp NE NULL);

		result = _calib_RatFac_numer (&COEFF);
	}

	Zxyz_clear_array (rem, maxdeg + 1);
	Zxyz_clear_array (quot, maxdeg + 1);
	free (rem);
	free (quot);

	Zxyz_clear_array (apoly, adeg + 1);
	Zxyz_clear_array (bpoly, bdeg + 1);
	free (apoly);
	free (bpoly);

	_calib_RatFac_clear (&COEFF);

	Zxyz_clear (&D);
	Zxyz_clear (&A_COFACT);
	Zxyz_clear (&B_COFACT);
	Zxyz_clear (&LC_GCD);

	return (result);
}

/*
 * Let d be the degree of the given polynomial in variable "var".
 * Create an array of d+1 polynomials, the i-th of which contains
 * the coefficient of "var**i" in all of the other variables.
 *
 * Returns the degree d and the array.
 */

	static
	void
Zxyz_make_poly_of_var (

int *			deg_out,	/* IN/OUT: Degree in given var */
struct calib_Zxyz_obj **
			array_out,	/* IN/OUT: Array of coeff polys */
const struct calib_Zxyz_obj *
			poly,		/* IN: Polynomial to decompose */
int			var		/* IN: Variable to get coeffs of */
)
{
int				i, j, k, d, nvars, nterms;
int *				pow;
int *				ip;
int *				count;
mpz_srcptr			coeff, cp;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		array;
struct calib_Zxyz_obj *		dst_poly;

	CALIB_FATAL_ERROR_IF ((deg_out EQ NULL) OR
			      (array_out EQ NULL) OR
			      (poly EQ NULL));

	nterms	= poly -> nterms;
	dom	= poly -> dom;
	pow	= poly -> pow;
	coeff	= poly -> coeff;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);

	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));
	CALIB_FATAL_ERROR_IF (nterms < 0);

	/* Pass 1: Compute degree in var. */
	d = -1;
	ip = pow;
	for (i = 0; i < nterms; i++, ip += nvars) {
		j = ip [var];
		CALIB_FATAL_ERROR_IF (j < 0);
		if (j > d) {
			d = j;
		}
	}

	/* Pass 2: Compute number of var**i terms for 0 <= i <= d. */
	count = CALIB_NEWA (d + 1, int);
	memset (count, 0, (d + 1) * sizeof (count [0]));
	ip = pow;
	for (i = 0; i < nterms; i++, ip += nvars) {
		j = ip [var];
		CALIB_FATAL_ERROR_IF ((j < 0) OR (j > d));
		++(count [j]);
	}

	/* Pass3: Initialize and pre-allocate array of polynomials. */
	array = CALIB_NEWA (d + 1, struct calib_Zxyz_obj);
	Zxyz_init_array (dom, array, d + 1);
	for (i = 0; i <= d; i++) {
		Zxyz_alloc (&array [i], count [i]);
	}

	/* Pass 4: Copy each term into proper element of array. */
	ip = pow;
	cp = coeff;
	for (i = 0; i < nterms; i++, ip += nvars, cp++) {
		j = ip [var];
		CALIB_FATAL_ERROR_IF ((j < 0) OR (j > d));
		dst_poly = &array [j];
		k = dst_poly -> nterms;
		mpz_set (&(dst_poly -> coeff [k]), cp);
		memcpy (&(dst_poly -> pow [k * nvars]),
			ip,
			nvars * sizeof (dst_poly -> pow [0]));
		/* Eliminate var**j factor from this term. */
		dst_poly -> pow [k * nvars + var] = 0;
		++(dst_poly -> nterms);
		CALIB_FATAL_ERROR_IF (dst_poly -> nterms > dst_poly -> size);
	}

	/* Pass 5: Verify each array[i] has proper number of terms. */
	for (i = 0; i <= d; i++) {
		j = count [i];
		CALIB_FATAL_ERROR_IF (array [i].nterms NE j);
		CALIB_FATAL_ERROR_IF (array [i].size < j);
	}

	free (count);

	*deg_out	= d;
	*array_out	= array;
}

/*
 * Polynomial pseudo-division for the resultant calculation.
 */

	static
	void
Zxyz_res_pseudo_div (

struct calib_Zxyz_obj *	quotient,	/* IN/OUT: Quotient coefficients */
int *			qdeg_out,	/* OUT: Degree of quotient */
struct calib_Zxyz_obj *	remainder,	/* IN/OUT: Remainder coefficients */
int *			rdeg_out,	/* OUT: Degree of remainder */
struct calib_Zxyz_obj *	D_out,		/* IN/OUT: denominator (of quotient) */
const struct calib_Zxyz_obj *
			a_in,		/* IN: Dividend coefficients */
int			adeg,		/* IN: Degree of dividend */
const struct calib_Zxyz_obj *
			b_in,		/* IN: Divisor coefficients */
int			bdeg		/* IN: Degree of divisor */
)
{
int				i, j, qdeg, rdeg, rsize;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		quot;
struct calib_Zxyz_obj *		rem;
const struct calib_Zxyz_obj *	lcb;
struct calib_Zxyz_obj *		lcr;
struct calib_Zxyz_obj *		rp;
const struct calib_Zxyz_obj *	bp;
const struct calib_Zxyz_obj *	polys [2];
struct calib_Zxyz_obj *		cofact [2];
struct calib_Zxyz_obj		DVAL, G, M, N, TEMP;

	dom = D_out -> dom;

	/* This routine is only used with polynomials of strictly	*/
	/* positive degree.						*/
	CALIB_FATAL_ERROR_IF (bdeg <= 0);

	Zxyz_init (dom, &DVAL);
	Zxyz_init (dom, &G);
	Zxyz_init (dom, &M);
	Zxyz_init (dom, &N);
	Zxyz_init (dom, &TEMP);

	rsize = adeg + 1;

	/* Local coefficient buffers. */
	quot = CALIB_NEWA (2 * rsize, struct calib_Zxyz_obj);
	Zxyz_init_array (dom, quot, 2 * rsize);
	rem = quot + rsize;

	Zxyz_set_si (&DVAL, 1);

	qdeg = adeg - bdeg;
	if (qdeg < -1) {
		qdeg = -1;
	}

	/* Copy dividend into internal remainder buffer. */
	for (i = 0; i <= adeg; i++) {
		Zxyz_set (&rem [i], &a_in [i]);
	}
	rdeg	= adeg;

	lcb = &b_in [bdeg];

	while (rdeg >= bdeg) {
		j = rdeg - bdeg;
		lcr = &rem [rdeg];
		cofact [0]	= &M;
		cofact [1]	= &N;
		polys [0]	= lcr;
		polys [1]	= lcb;
		Zxyz_gcd_n (&G, &cofact [0], 2, &polys [0]);

		for (i = qdeg; i > j; i--) {
			Zxyz_mul (&quot [i], &quot [i], &N);
		}

		for (i = rdeg - 1; i >= 0; i--) {
			Zxyz_mul (&rem [i], &rem [i], &N);
		}

		Zxyz_mul (&DVAL, &DVAL, &N);
		Zxyz_set (&quot [j], &M);

		rp = &rem [rdeg - 1];
		bp = &b_in [bdeg - 1];
		for (i = bdeg - 1; i >= 0; i--) {
			Zxyz_mul (&TEMP, &M, bp);
			Zxyz_sub (rp, rp, &TEMP);
			--rp;
			--bp;
		}
		--rdeg;
		rp = &rem [rdeg];
		while (rdeg >= 0) {
			if (rp -> nterms > 0) {
				CALIB_FATAL_ERROR_IF (rp -> nterms < 0);
				break;
			}
			--rdeg;
			--rp;
		}
	}

	for (i = 0; i <= qdeg; i++) {
		Zxyz_set (&quotient [i], &quot [i]);
	}
	*qdeg_out = qdeg;
	for (i = 0; i <= rdeg; i++) {
		Zxyz_set (&remainder [i], &rem [i]);
	}
	*rdeg_out = rdeg;
	Zxyz_set (D_out, &DVAL);

	Zxyz_clear (&TEMP);
	Zxyz_clear (&N);
	Zxyz_clear (&M);
	Zxyz_clear (&G);
	Zxyz_clear (&DVAL);

	Zxyz_clear_array (quot, 2 * rsize);
	free (quot);
}

/*
 * Compute the discriminant of the given polynomial with respect to
 * the given variable.
 */

	static
	void
Zxyz_discriminant (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: discriminant */
const struct calib_Zxyz_obj *	op,	/* IN: polynomial */
int				var	/* IN: variable to eliminate */
)
{
int				i, n, nvars;
const struct calib_Zxyz_dom *	dom;
int *				r2fmap;
struct GInfo *			gip;
struct RPoly *			rp1;
struct RPoly *			rp2;
struct RPoly *			qp;
struct calib_Zxyz_obj		DERIV, RES, LC;
struct GInfo			ginfo;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (var < 0));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF ((op -> dom NE dom) OR (var >= dom -> nvars));

	nvars	= dom -> nvars;
	n	= Zxyz_degree_in_var (op, var);

	if (n <= 0) {
		/* Discriminant of zero polynomial is not very well	*/
		/* defined.  (Is the degree 0 or -1?  All complex	*/
		/* numbers are roots, what is resultant(0,0), etc.)	*/
		CALIB_FATAL_ERROR_IF (n < 0);

		/* Discriminant of a non-zero constant.  This is also	*/
		/* not very well defined.				*/
		Zxyz_set_si (rop, 1);
		return;
	}

	gip = &ginfo;
	Zxyz_ginfo_init (gip);

	Zxyz_init (dom, &DERIV);
	Zxyz_init (dom, &RES);
	Zxyz_init (dom, &LC);

	Zxyz_derivative (&DERIV, op, var);

	Zxyz_resultant (&RES, op, &DERIV, var);

	if ((n & 0x02) NE 0) {
		/* We have n*(n-1)/2 odd.  Negate result. */
		Zxyz_neg (&RES, &RES);
	}

	/* Divide by lead coefficient.  Verify remainder is zero. */
	Zxyz_get_var_pow_coeff (&LC, op, var, n);

	r2fmap = CALIB_NEWA (nvars, int);
	for (i = 0; i < nvars; i++) {
		r2fmap [i] = i;
	}
	rp1 = Zxyz_rpoly_cv_f2r (&RES, nvars, r2fmap, gip);
	rp2 = Zxyz_rpoly_cv_f2r (&LC, nvars, r2fmap, gip);
	qp = Zxyz_rpoly_div_exact (rp1, rp2, gip);

	CALIB_FATAL_ERROR_IF (qp EQ DIVISION_NOT_EXACT);

	Zxyz_rpoly_cv_r2f (rop, qp, r2fmap);

	Zxyz_rpoly_free (qp, gip);
	Zxyz_rpoly_free (rp2, gip);
	Zxyz_rpoly_free (rp1, gip);
	free (r2fmap);

	Zxyz_clear (&LC);
	Zxyz_clear (&RES);
	Zxyz_clear (&DERIV);

	Zxyz_ginfo_clear (gip);
}

/*
 * Create a polynomial in Z[x,y,z] that is the correpsonding representative
 * of the given polynomial in Zp[x,y,z].  The Z coefficients chosen
 * have smallest absolute value that are equivalent to the corresponding
 * Zp coefficient.
 */

	static
	void
Zxyz_cvZpxyz (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: Result polynomial */
const struct calib_Zp_dom *	Zp,	/* IN: Zp description */
const struct calib_Zxyz_obj *	op	/* IN: poly in Zp[x,y,z] to convert */
)
{
int				i, nterms, size, nvars;
const struct calib_Zxyz_dom *	dom;
int *				pow;
mpz_ptr				coeff;
mpz_srcptr			p;
mpz_t				half_p;
struct calib_Zxyz_obj		TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (Zp EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	p	= Zp -> p;
	nterms	= op -> nterms;

	Zxyz_init (dom, &TEMP);
	Zxyz_alloc (&TEMP, nterms);

	pow	= TEMP.pow;
	coeff	= TEMP.coeff;
	size	= nterms;
	if (size > 0) {
		nvars = dom -> nvars;
		memcpy (pow, op -> pow, size * nvars * sizeof (int));

		calib_copy_Z_vector (coeff, op -> coeff, size);

		/* Compute half_p = floor(p/2). */
		mpz_init (half_p);
		mpz_tdiv_q_2exp (half_p, p, 1);

		for (i = 0; i < size; i++) {
			mpz_mod (&coeff [i], &coeff [i], p);
			if (mpz_cmp (&coeff [i], half_p) > 0) {
				/* Make it negative. */
				mpz_sub (&coeff [i], &coeff [i], p);
			}
		}
		mpz_clear (half_p);
	}

	Zxyz_set (rop, &TEMP);

	Zxyz_clear (&TEMP);
}

/*
 * Factor the given polynomial in Z[x,y,z] completely into its irreducible
 * factors over Z[x,y,z].  The correct multiplicity of each
 * factor is also determined.
 */

	static
	struct calib_Zxyz_factor *
Zxyz_factor (

const struct calib_Zxyz_obj *	poly		/* IN: polynomial to factor */
)
{
struct calib_Zxyz_factor *	factors;
struct GInfo			ginfo;

	Zxyz_ginfo_init (&ginfo);

	factors = Zxyz_internal_factor (poly, &ginfo, FMODE_FULL);

	Zxyz_ginfo_clear (&ginfo);

	return (factors);
}

/*
 * Perform Square-Free Factorization of the given polynomial in Z[x,y,z],
 * which is less expensive, and does not go all the way down to
 * irreducible polynomials.
 */

	static
	struct calib_Zxyz_factor *
Zxyz_sqf_factor (

const struct calib_Zxyz_obj *	poly		/* IN: polynomial to factor */
)
{
struct calib_Zxyz_factor *	factors;
struct GInfo			ginfo;

	Zxyz_ginfo_init (&ginfo);

	factors = Zxyz_internal_factor (poly, &ginfo, FMODE_SQUARE_FREE);

	Zxyz_ginfo_clear (&ginfo);

	return (factors);
}

/*
 * Factor the given polynomial in Z[x,y,z] completely into its irreducible
 * factors over Z[x,y,z].  The correct multiplicity of each
 * factor is also determined.
 */

	static
	struct calib_Zxyz_factor *
Zxyz_internal_factor (

const struct calib_Zxyz_obj *	poly,		/* IN: polynomial to factor */
struct GInfo *			gip,		/* IN/OUT: global info */
enum FactorMode			factor_mode	/* IN: factorization mode */
)
{
int				i, j, k, nvars, nterms, multiplicity, next_var;
int *				ip;
int *				mindeg;
int *				maxdeg;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		pp;
struct calib_Zxyz_obj	*	lpoly;
struct calib_Zxyz_obj	*	lpoly2;
struct calib_Zxyz_obj *		sqfpoly;
struct calib_Zxyz_factor *	flp;
struct calib_Zxyz_factor *	flp2;
struct calib_Zxyz_factor *	flist_in;
struct calib_Zxyz_factor *	flist_out;
struct calib_Zxyz_factor **	flist_hookp;
struct calib_Zxyz_factor *	sqfactors;
struct calib_Zxyz_factor *	olist;
struct calib_Zxyz_factor **	ohookp;
struct calib_Zxyz_factor *	fp;
struct calib_Zxyz_factor *	next_list;
struct calib_Zxyz_factor *	irred_factors;
struct calib_Zxyz_obj		pcont;
mpz_t				CONTENT;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	dom	= poly -> dom;
	nvars	= dom -> nvars;
	nterms	= poly -> nterms;

	olist = NULL;
	ohookp = &olist;

	if (nterms <= 0) {
		CALIB_FATAL_ERROR_IF (nterms < 0);

		/* Polynomial is zero, which factors easily. */
		fp = CALIB_NEW (struct calib_Zxyz_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= Zxyz_dup (poly);
		fp -> next		= olist;
		olist = fp;
		return (olist);
	}
	if (nterms EQ 1) {
		/* Another easy case -- a single term. */
		if (mpz_cmp_ui (&(poly -> coeff [0]), 1) NE 0) {
			pp = CALIB_NEW (struct calib_Zxyz_obj);
			Zxyz_init (dom, pp);
			Zxyz_set_z (pp, &(poly -> coeff [0]));
			fp = CALIB_NEW (struct calib_Zxyz_factor);
			fp -> multiplicity	= 1;
			fp -> factor		= pp;
			fp -> next		= NULL;
			*ohookp = fp;
			ohookp = &(fp -> next);
		}
		for (i = 0; i < nvars; i++) {
			k = poly -> pow [i];
			if (k <= 0) continue;
			pp = CALIB_NEW (struct calib_Zxyz_obj);
			Zxyz_init (dom, pp);
			Zxyz_set_si (pp, 1);
			pp -> pow [i] = 1;
			fp = CALIB_NEW (struct calib_Zxyz_factor);
			fp -> multiplicity	= k;
			fp -> factor		= pp;
			fp -> next		= NULL;
			*ohookp = fp;
			ohookp = &(fp -> next);
		}
		return (olist);
	}

	/* At least 2 terms.  We need to do some real work. */

	/* Compute the content and primitive part into a local copy. */
	lpoly = Zxyz_dup (poly);
	mpz_init (CONTENT);
	Zxyz_remove_z_contents (CONTENT, 1, &lpoly);

	/* If the content isn't 1, then save it off as a degree 0	*/
	/* factor of multiplicity 1.					*/
	if (mpz_cmp_ui (CONTENT, 1) NE 0) {
		sqfpoly = CALIB_NEW (struct calib_Zxyz_obj);
		sqfpoly -> nterms	= 1;
		sqfpoly -> size		= 1;
		sqfpoly -> dom		= dom;
		sqfpoly -> pow		= CALIB_NEWA (nvars, int);
		sqfpoly -> coeff	= calib_new_Z_vector (1);
		memset (sqfpoly -> pow, 0, nvars * sizeof (int));
		mpz_set (&(sqfpoly -> coeff [0]), CONTENT);

		fp = CALIB_NEW (struct calib_Zxyz_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= sqfpoly;
		fp -> next		= NULL;
		*ohookp = fp;
		ohookp = &(fp -> next);
	}

	/* Now handle factors of the form var^power. */
	Zxyz_get_poly_degrees (&mindeg, &maxdeg, lpoly);
	for (i = 0; i < nvars; i++) {
		k = mindeg [i];
		if (k > 0) {
			sqfpoly = CALIB_NEW (struct calib_Zxyz_obj);
			sqfpoly -> nterms	= 1;
			sqfpoly -> size		= 1;
			sqfpoly -> dom		= dom;
			sqfpoly -> pow		= CALIB_NEWA (nvars, int);
			sqfpoly -> coeff	= calib_new_Z_vector (1);
			memset (sqfpoly -> pow, 0, nvars * sizeof (int));
			sqfpoly -> pow [i] = 1;
			mpz_set_ui (&(sqfpoly -> coeff [0]), 1);

			fp = CALIB_NEW (struct calib_Zxyz_factor);
			fp -> multiplicity	= k;
			fp -> factor		= sqfpoly;
			fp -> next		= NULL;
			*ohookp = fp;
			ohookp = &(fp -> next);

			ip = &(lpoly -> pow [i]);
			for (j = 0; j < nterms; j++, ip += nvars) {
				*ip -= k;
				CALIB_FATAL_ERROR_IF (*ip < 0);
			}
			mindeg [i] = 0;
			maxdeg [i] -= k;
		}
	}

	/* Now identify factors that are missing one or more of the	*/
	/* variables by taking contents w.r.t. each variable.  Note	*/
	/* that this can split our one unfactored piece (lpoly) into	*/
	/* several separate factors, each of which is potentially still	*/
	/* further reducible.  So we form two lists of factors, an	*/
	/* "input list" (initially just lpoly) and an "output list"	*/
	/* (initially empty).  Each entry in the input list needs to	*/
	/* have an integer telling us the next variable to try on that	*/
	/* factor (so that we don't repeat prior content computations	*/
	/* on that factor).  We hijack the "multiplicity" field for	*/
	/* that purpose here, since each of these factors still has	*/
	/* multiplicity 1 at this point.  Factors from the input list	*/
	/* have contents removed w.r.t. to each variable (the contents	*/
	/* removed get put back into the input list), and the final	*/
	/* primitive factor is then moved to the output list.		*/

	flp = CALIB_NEW (struct calib_Zxyz_factor);
	flp -> multiplicity	= 0;	/* Try variable 0 first. */
	flp -> factor		= Zxyz_dup (lpoly);
	flp -> next		= NULL;

	flist_in = flp;
	flist_out = NULL;
	flist_hookp = &flist_out;

	Zxyz_init (dom, &pcont);
	while (flist_in NE NULL) {
		/* Remove first entry from list. */
		flp = flist_in;
		flist_in = flp -> next;

		/* Process this entry. */
		lpoly2 = flp -> factor;
		next_var = flp -> multiplicity;
		for (i = next_var; i < nvars; i++) {
			if (Zxyz_degree_in_var (lpoly2, i) <= 0) {
				/* Poly does not contain this var, so	*/
				/* skip it.				*/
				continue;
			}

			/* Compute the content and primitive-part of	*/
			/* this factor w.r.t. variable i.		*/
			Zxyz_int_content_ppart (&pcont, lpoly2, lpoly2, i, gip);

			/* Note that every non-trivial factor must now	*/
			/* have at least 2 terms.			*/
			if (pcont.nterms > 1) {
				/* Add this non-trivial content to our	*/
				/* input list.				*/
				flp2 = CALIB_NEW (struct calib_Zxyz_factor);
				flp2 -> multiplicity	= i + 1;
				flp2 -> factor		= Zxyz_dup (&pcont);
				flp2 -> next		= flist_in;
				flist_in = flp2;
			}
		}

		/* lpoly2 is now completely primitive.  Move it to the	*/
		/* output list, with multiplicity 1.   			*/
		flp -> multiplicity	= 1;
		flp -> factor		= lpoly2;
		flp -> next		= NULL;
		*flist_hookp = flp;
		flist_hookp = &(flp -> next);
	}
	Zxyz_clear (&pcont);
	free (mindeg);

	/* Now that everything is primitive, do square-free		*/
	/* factorization on each piece.					*/

	while (flist_out NE NULL) {
		flp2 = flist_out;
		flist_out = flp2 -> next;
		flp2 -> next = NULL;
		lpoly2 = flp2 -> factor;
		free (flp2);

		sqfactors = Zxyz_compute_square_free_factorization (lpoly2, gip);

		Zxyz_free (lpoly2);

		if (factor_mode EQ FMODE_SQUARE_FREE) {
			/* Stop at square-free factors.  Do not go all	*/
			/* the way down to irreducible factors.  Splice	*/
			/* these factors onto output list, and update	*/
			/* ohookp to point to the new end of the list.	*/
			*ohookp = sqfactors;
			for (;;) {
				flp = *ohookp;
				if (flp EQ NULL) break;
				ohookp = &(flp -> next);
			}
			continue;
		}

		while (sqfactors NE NULL) {
			/* Remove first square-free factor from the list. */
			multiplicity	= sqfactors -> multiplicity;
			sqfpoly		= sqfactors -> factor;
			next_list	= sqfactors -> next;
			sqfactors -> next = NULL;
			free (sqfactors);
			sqfactors = next_list;

			CALIB_FATAL_ERROR_IF (sqfpoly EQ NULL);

			irred_factors = Zxyz_factor_square_free (sqfpoly, gip);

			/* Transfer the irreducible factors onto the	*/
			/* final list of factors, adding the correct	*/
			/* multiplicity.				*/
			while (irred_factors NE NULL) {
				next_list = irred_factors -> next;
				irred_factors -> next = NULL;
				irred_factors -> multiplicity = multiplicity;
				*ohookp = irred_factors;
				ohookp = &(irred_factors -> next);
				irred_factors = next_list;
			}
			Zxyz_free (sqfpoly);
		}
	}

	mpz_clear (CONTENT);
	Zxyz_free (lpoly);

	return (olist);
}

/*
 * Factor the given polynomial into factors that are both square-free and
 * pairwise relatively prime, correctly determining the multiplicity of
 * each such factor.  This results in one non-trivial factor for each
 * distinct multiplicity present -- that factor being the product of all
 * factors having that that multiplicity.
 *
 * NOTE: the given polynomial is assumed to be primitive.
 */

	static
	struct calib_Zxyz_factor *
Zxyz_compute_square_free_factorization (

const struct calib_Zxyz_obj *	poly,	/* IN: polynomial to factor */
struct GInfo *			gip	/* IN/OUT: global info */
)
{
int				i, k, var, vardeg;
int				multiplicity, nvars;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_factor *	list;
struct calib_Zxyz_factor *	fp;
struct calib_Zxyz_factor **	hookp;
struct calib_Zxyz_obj *		pp;
int *				mindeg;
int *				maxdeg;
const struct calib_Zxyz_obj *	polys [2];
struct calib_Zxyz_obj *		cofactors [2];
struct calib_Zxyz_obj		f, fprime, a, b, c, d;

	CALIB_FATAL_ERROR_IF (poly EQ NULL);

	dom	= poly -> dom;
	nvars	= dom -> nvars;

	list = NULL;
	hookp = &list;

	/* Choose a "main" variable.  Pick highest degree for now.	*/
	Zxyz_get_poly_degrees (&mindeg, &maxdeg, poly);
	var = 0;
	vardeg = maxdeg [var];
	for (i = 1; i < nvars; i++) {
		k = maxdeg [i];
		if (k > vardeg) {
			var = i;
			vardeg = k;
		}
	}
	free (mindeg);

	Zxyz_init (dom, &f);
	Zxyz_init (dom, &fprime);
	Zxyz_init (dom, &a);
	Zxyz_init (dom, &b);
	Zxyz_init (dom, &c);
	Zxyz_init (dom, &d);

	/* Yun's algorithm, initialize. */
	Zxyz_set (&f, poly);
	Zxyz_derivative (&fprime, &f, var);
	polys [0] = &f;
	polys [1] = &fprime;
	cofactors [0] = &b;
	cofactors [1] = &c;
	dom -> gcd_n (&a, &cofactors [0], 2, &polys [0]);
	Zxyz_derivative (&d, &b, var);
	dom -> sub (&d, &c, &d);

	for (multiplicity = 1;; multiplicity++) {
		/* Yun's algorithm, iteration. */
		polys [0]	= &b;
		polys [1]	= &d;
		cofactors [0]	= &b;
		cofactors [1]	= &c;
		dom -> gcd_n (&a, &cofactors [0], 2, &polys [0]);
		if (NOT Zxyz_is_one_poly (&a)) {
			pp = Zxyz_dup (&a);

			fp = CALIB_NEW (struct calib_Zxyz_factor);
			fp -> multiplicity	= multiplicity;
			fp -> factor		= pp;
			fp -> next		= NULL;
			*hookp = fp;
			hookp = &(fp -> next);
		}
		Zxyz_derivative (&d, &b, var);
		Zxyz_sub (&d, &c, &d);

		if (Zxyz_is_constant_poly (&b)) break;
	}

	if (NOT Zxyz_is_one_poly (&b)) {
		/* Constant, but not 1 -- it had better be -1. */
		CALIB_FATAL_ERROR_IF (b.nterms NE 1);
		CALIB_FATAL_ERROR_IF (mpz_cmp_si (&b.coeff [0], -1) NE 0);

		/* Add this factor to the *front* of the list. */
		pp = Zxyz_dup (&b);

		fp = CALIB_NEW (struct calib_Zxyz_factor);
		fp -> multiplicity = 1;
		fp -> factor	   = pp;
		fp -> next	   = list;
		list = fp;
		if (list -> next EQ NULL) {
			hookp = &(list -> next);
		}
	}

	Zxyz_clear (&d);
	Zxyz_clear (&c);
	Zxyz_clear (&b);
	Zxyz_clear (&a);
	Zxyz_clear (&fprime);
	Zxyz_clear (&f);

	return (list);
}

/*
 * Choose a random dense evaluation point.  We generate random integers
 * which slowly increase in magnitude as more attempts are needed
 * to find a suitable evaluation point.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
Zxyz_choose_dense_evaluation_point (

mpz_ptr			bvals,		/* IN/OUT: evaluation point */
int			n,		/* IN: length of vector to generate */
struct evstate *	evs,		/* IN/OUT: state info */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
int			i, j;
int			tries, nbytes, nwords;
unsigned int		word, mask;
struct calib_Random *	rand;
mpz_t 			REG;

	CALIB_FATAL_ERROR_IF ((bvals EQ NULL) OR (n <= 0));

	tries	= evs -> tries;
	nbytes	= evs -> nbytes;
	rand	= &(gip -> rng);

	++tries;
	if (tries > 10) {
		/* Time to increase magnitude of the integers we choose. */
		++nbytes;
		evs -> nbytes = nbytes;
	}
	evs -> tries = tries;

	nwords = (nbytes >> 2);
	nbytes &= 0x03;
	mask = 1;
	mask <<= (8 * nbytes);
	--mask;

	mpz_init (REG);

	for (i = 0; i < n; i++) {
		word = 0;
		if (nbytes > 0) {
			word = calib_random_u32 (rand);
			word &= mask;
		}
		mpz_set_ui (REG, word);
		for (j = 0; j < nwords; j++) {
			word = calib_random_u32 (rand);
			mpz_mul_2exp (REG, REG, 32);
			mpz_add_ui (REG, REG, word);
		}
		mpz_set (&bvals [i], REG);
	}

	mpz_clear (REG);
}

#pragma GCC diagnostic pop

/*
 * Factor the given polynomial (which is both primitive and square-free)
 * into irreducible factors.
 */

	static
	struct calib_Zxyz_factor *
Zxyz_factor_square_free (

const struct calib_Zxyz_obj *
			poly,	/* IN: primitive, square-free polynomial */
				/*     to factor */
struct GInfo *		gip	/* IN/OUT: global info */
)
{
int				i, j, k, nvars, nv, x1, x1_degree, status;
int				nf, max_nfact, nf_ubound;
calib_bool			bad;
int *				mindeg;
int *				maxdeg;
int *				r2fmap;
const struct calib_Zxyz_dom *	dom;
struct calib_Zxyz_obj *		p;
struct calib_Zxyz_factor *	fp;
struct calib_Zxyz_factor *	olist;
struct calib_Zxyz_factor **	ohookp;
mpz_ptr				bvals;
calib_bool *			evflags;
const struct calib_Zx_dom *	Zx;
struct calib_Zx_factor *	zx_factors;
struct calib_Zx_factor *	zxfp;
struct CFCache			cfcache;
struct calib_Zx_obj		zx_poly;
struct calib_Zxyz_obj		UPOLY, TEMP;
mpz_t				omega;
struct EEZglob			eez_global;

	CALIB_FATAL_ERROR_IF ((poly EQ NULL) OR (gip EQ NULL));

	dom	= poly -> dom;
	nvars	= dom -> nvars;

	olist = NULL;
	ohookp = &olist;

	Zxyz_get_poly_degrees (&mindeg, &maxdeg, poly);

	/* Compute a bound on the number of factors -- we use the max	*/
	/* degree of any one factor.					*/
	nf_ubound = 0;
	for (i = 0; i < nvars; i++) {
		k = maxdeg [i];
		if (k > nf_ubound) {
			nf_ubound = k;
		}
	}

	if (nf_ubound <= 1) {
		/* The polynomial is degree at most 1 in every variable. */
		/* It is therefore already irreducible.			 */
		free (mindeg);
		fp = CALIB_NEW (struct calib_Zxyz_factor);
		fp -> multiplicity	= 1;
		fp -> factor		= Zxyz_dup (poly);
		fp -> next		= NULL;
		*ohookp = fp;
		ohookp = &(fp -> next);
		return (olist);
	}

	/* Determine which variabless are actually present in the	*/
	/* problem, and pick a suitable ordering of these variables.	*/
	r2fmap = Zxyz_choose_var_order (&nv, 1, nvars, &poly, mindeg, maxdeg);
	free (mindeg);

	/* Had better have at least 1 variable at this point. */
	CALIB_FATAL_ERROR_IF (nv <= 0);

	/* Get the "main" variable. */
	x1 = r2fmap [nv - 1];

	Zx = calib_get_Zx_dom ();
	Zx -> init (&zx_poly);

	if (nv EQ 1) {
		/* Univariate. */
		free (r2fmap);
		Zxyz_cv_poly_to_Zx_form (&zx_poly, poly, x1);
		zx_factors = Zx -> factor (&zx_poly);
		Zx -> clear (&zx_poly);
		Zxyz_init (dom, &TEMP);
		for (zxfp = zx_factors; zxfp NE NULL; zxfp = zxfp -> next) {
			Zxyz_cv_from_Zx (&TEMP, zxfp -> factor, x1);
			fp = CALIB_NEW (struct calib_Zxyz_factor);
			fp -> multiplicity	= zxfp -> multiplicity;
			fp -> factor		= Zxyz_dup (&TEMP);
			fp -> next		= NULL;
			*ohookp = fp;
			ohookp = &(fp -> next);
		}
		Zxyz_clear (&TEMP);
		Zx -> free_factors (zx_factors);
		return (olist);
	}

	/* More than one variable.  Set upper bound on number of factors. */
	max_nfact = nf_ubound;

	/* Create buffer for the evaluation point and flags indicating	*/
	/* which variables we are evaluating away (all but x1).		*/
	bvals = calib_new_Z_vector (nvars);
	evflags = CALIB_NEWA (nvars, calib_bool);
	memset (evflags, FALSE, nvars * sizeof (evflags [0]));
	for (i = 0; i < nv - 1; i++) {
		j = r2fmap [i];
		evflags [j] = TRUE;
	}

	/* Create empty cache for factorization of the leading and	*/
	/* trailing coefficients of our multivariate polynomial.	*/
	memset (&cfcache, 0, sizeof (cfcache));

	/* Initialize the global EEZ algorithm info.  Use nf_ubound to	*/
	/* dimension all of the arrays that are one element per factor,	*/
	/* since we don't really know how many there are going to be	*/
	/* yet.								*/
	nf = nf_ubound;
	Zxyz_eez_init_global_info (&eez_global,
				   MODE_FACTOR,
				   gip,
				   dom,
				   nv,
				   r2fmap,
				   evflags,
				   bvals,
				   nf);

	eez_global.poly = poly;

	x1_degree = Zxyz_degree_in_var (poly, x1);

	Zxyz_init (dom, &UPOLY);
	mpz_init (omega);

	for (;;) {
		/* Choose a new evaluation point. */
		Zxyz_choose_evaluation_point (bvals, &eez_global);

		/* Each time we pick a new evaluation point, we need to	*/
		/* recompute the coefficient upper bound B, prime p,	*/
		/* modulus = p^k > 2*B, and all that depend on these.	*/
		mpz_set_ui (eez_global.modulus, 0);

		/* Evaluate the poly down to univariate (x1 only), and	*/
		/* verify that the degree in x1 does not decrease.	*/
		Zxyz_eval_var_subset (&UPOLY, poly, evflags, bvals);
		j = Zxyz_degree_in_var (&UPOLY, x1);
		if (j NE x1_degree) {
			/* Need to pick another evaluation point. */
			continue;
		}

		/* Convert univariate polynomial into Z[x] form. */
		Zxyz_cv_poly_to_Zx_form (&zx_poly, &UPOLY, x1);

		if (mpz_sgn (&zx_poly.coeff [0]) EQ 0) {
			/* Trailing coefficient is zero.  Pick another */
			/* evaluation point. */
			continue;
		}

		/* Factorize over Z[x]. */
		zx_factors = Zx -> factor (&zx_poly);

		/* Count the non-trivial factors.  Collect the constant	*/
		/* factors into omega.					*/
		mpz_set_ui (omega, 1);
		bad = FALSE;
		nf = 0;
		for (zxfp = zx_factors; zxfp NE NULL; zxfp = zxfp -> next) {
			if (zxfp -> multiplicity NE 1) {
				bad = TRUE;
				break;
			}
			if (zxfp -> factor -> degree > 0) {
				++nf;
			}
			else {
				mpz_mul (omega,
					 omega,
					 &(zxfp -> factor -> coeff [0]));
			}
		}
		if (bad) {
			/* Need to choose another evaluation point. */
			Zx -> free_factors (zx_factors);
			continue;
		}
		if (nf < 2) {
			CALIB_FATAL_ERROR_IF (nf <= 0);
			/* Multivariate polynomial is irreducible! */
			fp = CALIB_NEW (struct calib_Zxyz_factor);
			fp -> multiplicity	= 1;
			fp -> factor		= Zxyz_dup (poly);
			fp -> next		= NULL;
			*ohookp = fp;
			ohookp = &(fp -> next);
			Zx -> free_factors (zx_factors);
			break;
		}

		if (nf > max_nfact) {
			/* Too many factors.  Pick another evaluation point. */
			Zx -> free_factors (zx_factors);
			continue;
		}
		else if (nf < max_nfact) {
			/* Insist on at most this many factors from now on. */
			max_nfact = nf;
		}

		/* Convert univariate factors back into Z[x,y,z] form	*/
		/* into our array of factors to lift via EEZ.		*/
		CALIB_FATAL_ERROR_IF (nf > nf_ubound);
		i = 0;
		for (zxfp = zx_factors; zxfp NE NULL; zxfp = zxfp -> next) {
			if (zxfp -> factor -> degree <= 0) continue;
			Zxyz_cv_from_Zx (&eez_global.factors [i],
					 zxfp -> factor,
					 x1);
			++i;
		}
		CALIB_FATAL_ERROR_IF (i NE nf);
		Zx -> free_factors (zx_factors);

		/* Impose omega upon the first factor. */
		Zxyz_mul_z (&eez_global.factors [0],
			    &eez_global.factors [0],
			    omega);

		/* Do the main EEZ algorithm.  */
		eez_global.nf = nf;
		status = Zxyz_eez_main (&cfcache, &eez_global);

		if (status EQ SUCCESS) {
			/* We have factored into irreducibles! */
			for (i = 0; i < nf; i++) {
				p = Zxyz_dup (&eez_global.pdfactors [i]);
				fp = CALIB_NEW (struct calib_Zxyz_factor);
				fp -> multiplicity	= 1;
				fp -> factor		= p;
				fp -> next		= NULL;
				*ohookp = fp;
				ohookp = &(fp -> next);
			}
			break;
		}
		if (status EQ BAD_EVAL_POINT) {
			/* Just pick a new evaluation point. */
			continue;
		}

		CALIB_FATAL_ERROR_IF (status NE EXTRANEOUS_FACTORS);

		/* We have extraneous factors.  Insist on fewer. */
		--max_nfact;
		if (max_nfact <= 1) {
			/* Polynomial must be irreducible. */
			fp = CALIB_NEW (struct calib_Zxyz_factor);
			fp -> multiplicity	= 1;
			fp -> factor		= Zxyz_dup (poly);
			fp -> next		= NULL;
			*ohookp = fp;
			ohookp = &(fp -> next);
			Zx -> free_factors (zx_factors);
			break;
		}
		/* Pick a new evaluation point. */
	}

	mpz_clear (omega);
	Zxyz_clear (&UPOLY);

	eez_global.poly = NULL;

	eez_global.nf = nf_ubound;
	Zxyz_eez_clear_global_info (&eez_global);

	Zxyz_free_factors (cfcache.tcf);
	Zxyz_free_factors (cfcache.lcf);

	free (evflags);
	calib_free_Z_vector (bvals, nvars);
	Zx -> clear (&zx_poly);
	free (r2fmap);

	return (olist);
}

/*
 * Compute the n-th partial derivative of the given polynomial with respect
 * to the given variable.  It is assumed that there are no functional
 * dependencies between this variable and any of the others.
 */

	static
	void
Zxyz_nth_derivative (

struct calib_Zxyz_obj *	result,		/* IN/OUT: derivative polynomial */
const struct calib_Zxyz_obj *
			poly,		/* IN: polynomial to differentiate */
int			var,		/* IN: var to differentiate w.r.t. */
int			n		/* IN: order of derivative to take */
)
{
int				i, j, k, m, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
mpz_srcptr			szp;
mpz_ptr				dzp;
const int *			sip;
int *				dip;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (poly EQ NULL) OR (n < 0));
	dom	= poly -> dom;
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((result -> dom NE dom) OR
			(var < 0) OR (var >= nvars));

	if (n <= 1) {
		if (n EQ 0) {
			if (result NE poly) {
				Zxyz_set (result, poly);
			}
		}
		else {
			Zxyz_derivative (result, poly, var);
		}
		return;
	}

	nterms	= poly -> nterms;
	sip	= poly -> pow;
	szp	= poly -> coeff;

	if (result EQ poly) {
		/* Taking derivative in place.  The number of terms	*/
		/* cannot increase, so there is no need to do any	*/
		/* allocation check.					*/
	}
	else {
		/* Count number of terms needed. */
		k = 0;
		for (i = 0; i < nterms; i++, sip += nvars) {
			m = sip [var];
			CALIB_FATAL_ERROR_IF (m < 0);
			if (m >= n) {
				++k;
			}
		}
		/* Make sure result can hold this many terms. */
		Zxyz_alloc (result, k);
		sip = poly -> pow;
	}
	dip = result -> pow;
	dzp = result -> coeff;
	k = 0;
	for (i = 0; i < nterms; i++) {
		m = sip [var];
		if (m >= n) {
			mpz_mul_ui (dzp, szp, m);
			for (j = 1; j < n; j++) {
				mpz_mul_ui (dzp, dzp, m - j);
			}
			memcpy (dip, sip, nvars * sizeof (int));
			dip [var] = m - n;
			dip += nvars;
			++dzp;
			++k;
		}
		sip += nvars;
		++szp;
	}
	result -> nterms = k;
}

/*
 * Compute the partial derivative of the given polynomial with respect
 * to the given variable.  It is assumed that there are no functional
 * dependencies between this variable and any of the others.
 */

	static
	void
Zxyz_derivative (

struct calib_Zxyz_obj *	result,		/* IN/OUT: derivative polynomial */
const struct calib_Zxyz_obj *
			poly,		/* IN: polynomial to differentiate */
int			var		/* IN: var to differentiate w.r.t. */
)
{
int				i, k, m, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
mpz_srcptr			szp;
mpz_ptr				dzp;
const int *			sip;
int *				dip;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (poly EQ NULL));
	dom	= poly -> dom;
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((result -> dom NE dom) OR
			      (var < 0) OR
			      (var >= nvars));

	nterms	= poly -> nterms;
	sip	= poly -> pow;
	szp	= poly -> coeff;

	if (result EQ poly) {
		/* Taking derivative in place.  The number of terms	*/
		/* cannot increase, so there is no need to do any	*/
		/* allocation check.					*/
	}
	else {
		/* Count number of terms needed. */
		k = 0;
		for (i = 0; i < nterms; i++, sip += nvars) {
			m = sip [var];
			CALIB_FATAL_ERROR_IF (m < 0);
			if (m > 0) {
				++k;
			}
		}
		/* Make sure result can hold this many terms. */
		Zxyz_alloc (result, k);
		sip = poly -> pow;
	}
	dip = result -> pow;
	dzp = result -> coeff;
	k = 0;
	for (i = 0; i < nterms; i++) {
		m = sip [var];
		if (m > 0) {
			mpz_mul_ui (dzp, szp, m);
			if (sip NE dip) {
				memcpy (dip, sip, nvars * sizeof (int));
			}
			dip [var] = m - 1;
			dip += nvars;
			++dzp;
			++k;
		}
		sip += nvars;
		++szp;
	}
	result -> nterms = k;
}

/*
 * Evaluate the given polynomial with a single variable bound to a
 * given integer value.  The result is a polynomial that does not
 * contain the variable that was evaluated away.
 */

	static
	void
Zxyz_eval_one_var (

struct calib_Zxyz_obj *	result,		/* IN/OUT: resulting polynomial */
const struct calib_Zxyz_obj *
			poly,		/* IN: polynomial to evaluate */
int			var,		/* IN: variable to evaluate */
mpz_srcptr		value		/* IN: value to substitute */
)
{
int				i, k, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
mpz_srcptr			cp;
const int *			pp;
mpz_t				cval, tmp;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR
			      (poly   EQ NULL) OR
			      (value  EQ NULL));
	dom = result -> dom;
	CALIB_FATAL_ERROR_IF (poly -> dom NE dom);

	pap = Zxyz_get_paccum (dom);

	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	int	pow [nvars];

	mpz_init (cval);
	mpz_init (tmp);

	nterms	= poly -> nterms;
	pp	= poly -> pow;
	cp	= poly -> coeff;

	for (i = 0; i < nterms; i++, cp++, pp += nvars) {
		mpz_set (cval, cp);
		memcpy (&pow [0], pp, nvars * sizeof (int));

		k = pow [var];
		if (k > 0) {
			pow [var] = 0;	/* Drop var from this term */
			mpz_pow_ui (tmp, value, k);
			mpz_mul (cval, cval, tmp);
		}

		Zxyz_paccum_add_term (pap, cval, &pow [0]);
	}

	mpz_clear (tmp);
	mpz_clear (cval);

	Zxyz_set_poly_from_paccum (result, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Get the coefficient of var^power from the given polynomial.
 * The returned polynomial is a multivariate polynomial that is free of var.
 */

	static
	void
Zxyz_get_coeff (

struct calib_Zxyz_obj *	result,		/* IN/OUT: resulting coefficient */
const struct calib_Zxyz_obj *
			poly,		/* IN: poly to get coefficient from */
int			var,		/* IN: var to get coefficient of */
int			power		/* IN: which power to get */
)
{
int				i, j, k, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
const int *			sip;
int *				dip;
mpz_ptr				dzp;
mpz_srcptr			szp;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR
			      (poly   EQ NULL) OR
			      (power  < 0));
	dom	= result -> dom;
	nvars	= dom -> nvars;
	CALIB_FATAL_ERROR_IF ((poly -> dom NE dom) OR
			      (var < 0) OR
			      (var >= nvars));

	/* Count the number of terms in the desired coefficient. */
	nterms	= poly -> nterms;
	sip	= poly -> pow;
	j = 0;
	for (i = 0; i < nterms; i++) {
		if (sip [var] EQ power) {
			++j;
		}
		sip += nvars;
	}
	Zxyz_alloc (result, j);

	sip	= poly -> pow;
	szp	= poly -> coeff;
	dip	= result -> pow;
	dzp	= result -> coeff;
	k = 0;
	for (i = 0; i < nterms; i++) {
		if (sip [var] EQ power) {
			memcpy (dip, sip, nvars * sizeof (int));
			dip [var] = 0;
			mpz_set (dzp, szp);
			dip += nvars;
			++dzp;
			++k;
		}
		sip += nvars;
		++szp;
	}
	CALIB_FATAL_ERROR_IF (j NE k);
	result -> nterms = j;
}

/*
 * Free the given factorization, as produced by Zx_factor().
 */

	static
	void
Zxyz_free_factors (

struct calib_Zxyz_factor *	factors	/* IN: factors to recursively free */
)
{
struct calib_Zxyz_factor *	ftemp;
struct calib_Zxyz_obj *		poly;

	while (factors NE NULL) {
		ftemp = factors -> next;
		factors -> next = NULL;
		poly = factors -> factor;
		factors -> factor = NULL;
		Zxyz_free (poly);
		free (factors);
		factors = ftemp;
	}
}

/*
 * Map the given polynomial from its original ring into a new polynomial
 * whose coefficients reside in the given subring.  The caller gives us
 * a "varmap" array that controls this mapping on a variable-by-variable
 * basis.  Let i be a variable index within the original polynomial ring,
 * and let j = varmap [i].  Then:
 *
 *	j = -1	==> variable i remains in the source ring.
 *	j >= 0	==> variable i (from the original ring) maps to variable j
 *		    of the subring (j must satisfy 0 <= j < subring.nvars).
 *
 * Suppose the original ring has varlist {"a", "b", "c", "x", "y", "z"},
 * and the subring has varlist {"a", "b", "c"}.  This means that "result"
 * will be a polynomial in x, y and z whose coefficients are polynomials
 * in the subring in variables a, b and c.  Typically the varmap would
 * be {0, 1, 2, -1, -1, -1} in this example, although other variable
 * mappings (e.g., those that do not respect the symbolic variable names
 * in each ring's varlist) are of course possible.  This routine does
 * very little checking of the varmap beyond what is necessary for safety.
 *
 * Since the Zxyz_obj representation only handles "integer" coefficients,
 * the "result" object contains only dummy "1" coefficient values.  The
 * "actual" coefficients are found in the return value of this function,
 * which is an array of Zxyz_obj objects, each of whose "dom" member
 * is the given subring.  The returned array has the same number of
 * elements as result -> nterms.  It is the caller's responsibility to
 * free up the coefficient array when done with it.
 *
 * This routine essentially rearranges the various terms of the original
 * polynomial, putting them into a different logical structure.  The new
 * structure is likely to be of similar size as the old, although it
 * could be substantially smaller if the subring has a much smaller
 * "nvars" value.  Algorithmically, this operation is best thought of as
 * partitioning the terms of the original polynomial into distinct
 * "class" buckets, with the set of "class" buckets determined dynamically.
 */

	static
	struct calib_Zxyz_obj *
Zxyz_map_to_subring (

struct calib_Zxyz_obj *	result,		/* IN/OUT: result in original ring */
const struct calib_Zxyz_obj *
			poly,		/* IN: polynomial to map */
const struct calib_Zxyz_dom *
			subring,	/* IN: subring to map coefficients */
					/*     into */
const int *		varmap		/* IN: array describing variable */
					/*     mapping from source ring */
					/*     to subring */
)
{
int				i, i1, i2, j, k;
int				h, res_nterms;
int				nvars1, nvars2, nterms;
size_t				nbytes;
const int *			spow;
int *				tpow;
int *				dpow;
mpz_srcptr			scoeff;
const struct calib_Zxyz_dom *	ring;
struct paccum *			pap;
struct pterm **			bucket;
struct pterm **			hookp;
struct pterm *			tp;
struct dlink *			link;
struct pterm **			tarray;
struct calib_Zxyz_obj *		carray;
struct calib_Zxyz_obj *		cfp;

	CALIB_FATAL_ERROR_IF ((result  EQ NULL) OR
			      (poly    EQ NULL) OR
			      (subring EQ NULL) OR
			      (varmap  EQ NULL));

	nterms	= poly -> nterms;
	ring	= poly -> dom;

	nvars1	= ring -> nvars;
	nvars2	= subring -> nvars;

	/* Perform minimal validation of the varmap[]. */
	for (i = 0; i < nvars1; i++) {
		j = varmap [i];
		CALIB_FATAL_ERROR_IF ((j < -1) OR (j >= nvars2));
	}

	/* Note that we use the "paccum" and "pterm" objects to help us	*/
	/* determine the "class" buckets -- but we don't acually use	*/
	/* the pterm.coeff field as a GMP integer, but in fact we	*/
	/* steal its _mp_alloc and _mp_size fields as follows:		*/
	/*								*/
	/*	_mp_alloc	- holds the index of this term within	*/
	/*			  our "result" polynomial.		*/
	/*								*/
	/*	_mp_size	- holds the number of subring terms	*/
	/*			  appearing (so far) in this term of	*/
	/*			  the "result"				*/
	/*								*/
	/* On the first path through the terms of the original poly,	*/
	/* dynamically create all of the needed pterms, and count how	*/
	/* many subring terms will reside within each.			*/

	pap = Zxyz_get_paccum (ring);
	link = pap -> link;

	tpow = CALIB_NEWA (nvars1, int);
	nbytes = nvars1 * sizeof (int);

	spow	= poly -> pow;
	for (i = 0; i < nterms; i++, spow += nvars1) {
		for (j = 0; j < nvars1; j++) {
			k = varmap [j];
			if (k < 0) {
				/* Un-mapped variable, which remains in	*/
				/* the top-level result poly.  Its	*/
				/* power goes into the bucket key.	*/
				tpow [j] = spow [j];
			}
			else {
				/* A mapped variable, which disappears	*/
				/* from the top-level poly.  It goes	*/
				/* into the bucket key as a zero power.	*/
				tpow [j] = 0;
			}
		}
		h = Zxyz_pterm_hash (nvars1, tpow);
		bucket = &(pap -> bucket [h]);
		hookp = bucket;
		for (;;) {
			tp = *hookp;
			if (tp EQ NULL) break;
			if (memcmp (tp -> pow, tpow, nbytes) EQ 0) goto found1;
			hookp = &(tp -> next);
		}

		tp = Zxyz_get_pterm (pap);

		/* Clear out the mpz_t coefficient, so we can steal its */
		/* fields for other purposes.				*/
		mpz_clear (tp -> coeff);
		memset (tp -> coeff, 0, sizeof (tp -> coeff));

		memcpy (tp -> pow, tpow, nbytes);

		if (*bucket EQ NULL) {
			/* This hash table bucket is transitioning from	*/
			/* empty to non-empty.  Add it to the doubly-	*/
			/* linked list of non-empty buckets.		*/
			CALIB_FATAL_ERROR_IF ((link [h].forw NE -2) OR
					      (link [h].back NE -2));
			i1 = -1;
			i2 = link [i1].forw;
			link [h].back = i1;
			link [h].forw = i2;
			link [i1].forw = h;
			link [i2].back = h;
		}
		tp -> next = *bucket;
		*bucket = tp;
		++(pap -> nterms);
found1:
		/* Increment the number of terms residing in this	*/
		/* bucket (i.e., term of the "result" polynomial, as	*/
		/* opposed to hash table bucket).			*/
		++(tp -> coeff [0]._mp_size);
	}

	res_nterms = pap -> nterms;

	/* Sort the pterms in our paccum object into the order in which	*/
	/* they must appear within "result"				*/
	tarray = Zxyz_get_sorted_terms (pap);

	/* Mark each pterm with its index. */
	for (i = 0; i < res_nterms; i++) {
		tp = tarray [i];
		tp -> coeff [0]._mp_alloc = i;
	}

	/* Allocate the correct number of terms for "result". */
	ring -> alloc (result, res_nterms);

	/* Copy the proper powers into each term of "result", and force	*/
	/* each coefficient to be the dummy value 1.			*/
	for (i = 0; i < res_nterms; i++) {
		tp = tarray [i];
		memcpy (&(result -> pow [i * nvars1]),
			&(tp -> pow [0]),
			nvars1 * sizeof (int));
		mpz_set_ui (&(result -> coeff [i]), 1);
	}
	result -> nterms = res_nterms;

	/* Allocate the array of coefficient polynomials, and pre-	*/
	/* allocate the correct number of terms into each.		*/
	carray = Zxyz_new_array (subring, res_nterms);
	for (i = 0; i < res_nterms; i++) {
		tp = tarray [i];
		k = tp -> coeff [0]._mp_size;
		CALIB_FATAL_ERROR_IF (k <= 0);
		subring -> alloc (&carray [i], k);
	}

	/* Make a second pass through the terms of the original		*/
	/* polynomial, this time moving the coefficients into their	*/
	/* final position within carray.  Note: we simply append each	*/
	/* coefficient as a new term onto the end of the appropriate	*/
	/* carray[i] -- these terms don't wind up in the correct order	*/
	/* when doing this, but we re-sort the terms of each carray[i]	*/
	/* down below when we are done.					*/

	dpow = CALIB_NEWA (nvars2, int);

	spow	= poly -> pow;
	scoeff	= poly -> coeff;
	for (i = 0; i < nterms; i++, spow += nvars1, scoeff++) {
		memset (dpow, 0, nvars2 * sizeof (int));
		for (j = 0; j < nvars1; j++) {
			k = varmap [j];
			if (k < 0) {
				/* Un-mapped variable, which remains in	*/
				/* the top-level result poly.  Its	*/
				/* power goes into the bucket key.	*/
				tpow [j] = spow [j];
			}
			else {
				/* A mapped variable, which disappears	*/
				/* from the top-level poly.  It goes	*/
				/* into the bucket key as a zero power.	*/
				tpow [j] = 0;

				/* It becomes var k within subring. */
				dpow [k] = spow [j];
			}
		}
		h = Zxyz_pterm_hash (nvars1, tpow);
		bucket = &(pap -> bucket [h]);
		hookp = bucket;
		for (;;) {
			tp = *hookp;
			/* We *MUST* find it during pass 2! */
			CALIB_FATAL_ERROR_IF (tp EQ NULL);
			if (memcmp (tp -> pow, tpow, nbytes) EQ 0) break;
			hookp = &(tp -> next);
		}

		j = tp -> coeff [0]._mp_alloc;
		CALIB_FATAL_ERROR_IF ((j < 0) OR (j >= res_nterms));
		cfp = &carray [j];
		k = cfp -> nterms;
		mpz_set (&(cfp -> coeff [k]), scoeff);
		memcpy (&(cfp -> pow [k * nvars2]),
			dpow,
			nvars2 * sizeof (int));
		cfp -> nterms = k + 1;
	}

	/* Verify that we got the expected number of terms within each	*/
	/* carray [i].							*/
	for (i = 0; i < res_nterms; i++) {
		cfp = &carray [i];
		CALIB_FATAL_ERROR_IF (cfp -> nterms NE cfp -> size);
	}

	/* Re-initialize each pterm.coeff back to being a GMP integer. */
	for (i = 0; i < res_nterms; i++) {
		tp = tarray [i];
		mpz_init (tp -> coeff);
	}

	free (dpow);
	free (tarray);
	free (tpow);

	/* Free up the paccum object. */
	Zxyz_put_paccum (pap);

	/* Re-sort the individual terms within each carray[i] to be	*/
	/* in the proper order.  Just use normal paccum machinery to	*/
	/* do this.							*/

	pap = Zxyz_get_paccum (subring);

	for (i = 0; i < res_nterms; i++) {
		cfp = &carray [i];
		for (j = 0; j < cfp -> nterms; j++) {
			Zxyz_paccum_add_term (pap,
					      &(cfp -> coeff [j]),
					      &(cfp -> pow [j * nvars2]));
		}
		Zxyz_set_poly_from_paccum (cfp, pap);

		Zxyz_reset_paccum (pap);
	}

	Zxyz_put_paccum (pap);

	return (carray);
}

/*
 * Copy the given source polynomial into the destination polynomial.
 * The domain of the destination polynomial need not be the same as that
 * of the source polynomial, but the destination polynomial must have
 * at least as many variables.  These extra variables receive a power
 * of zero in every term copied.
 *
 * No checking of variable names is done.  This just blindly copies a
 * polynomial into a new domain with more variables.
 */

	static
	void
Zxyz_copy_into_superring (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: Destination polynomial */
const struct calib_Zxyz_obj *	op	/* IN: Source polynomial */
)
{
int				i, snvars, dnvars, nterms;
const struct calib_Zxyz_dom *	sdom;
const struct calib_Zxyz_dom *	ddom;
const int *			spow;
int *				dpow;
mpz_srcptr			scoeff;
mpz_ptr				dcoeff;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	ddom	= rop -> dom;
	sdom	= op -> dom;

	CALIB_FATAL_ERROR_IF ((sdom EQ NULL) OR (ddom EQ NULL));

	if (sdom EQ ddom) {
		Zxyz_set (rop, op);
		return;
	}

	snvars	= sdom -> nvars;
	dnvars	= ddom -> nvars;

	CALIB_FATAL_ERROR_IF (dnvars < snvars);

	nterms	= op -> nterms;
	spow	= op -> pow;
	scoeff	= op -> coeff;

	dpow	= rop -> pow;
	dcoeff	= rop -> coeff;

	/* Zero all destination powers. */
	memset (dpow, 0, nterms * dnvars * sizeof (dpow [0]));

	ddom -> alloc (rop, nterms);

	for (i = 0; i < nterms; i++) {
		memcpy (dpow, spow, snvars * sizeof (dpow [0]));
		mpz_set (&dcoeff [i], &scoeff [i]);
		spow += snvars;
		dpow += dnvars;
	}
	rop -> nterms = nterms;
}

/*
 * Convert a polynomial from one ring to another, mapping variables
 * according to the given varmap.  Vars mapping to -1 in the varmap
 * must not appear in the source polynomial.
 */

	static
	void
Zxyz_convert_with_varmap (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: result polynomial */
const struct calib_Zxyz_obj *	op,	/* IN: polynomial to convert */
const int *			varmap	/* IN: variable mapping array */
)
{
int				i, j, k, nterms, src_nvars, dst_nvars, var;
const int *			spow;
int *				dpow;
const struct calib_Zxyz_dom *	sdom;
const struct calib_Zxyz_dom *	ddom;

	CALIB_FATAL_ERROR_IF ((rop    EQ NULL) OR
			      (op     EQ NULL) OR
			      (varmap EQ NULL));

	nterms	= op -> nterms;
	sdom	= op -> dom;
	ddom	= rop -> dom;

	src_nvars = sdom -> nvars;
	dst_nvars = ddom -> nvars;

	ddom -> alloc (rop, nterms);

	spow	= op -> pow;
	dpow	= rop -> pow;
	memset (dpow, 0, nterms * dst_nvars * sizeof (dpow [0]));

	for (i = 0; i < nterms; i++, spow += src_nvars, dpow += dst_nvars) {
		for (j = 0; j < src_nvars; j++) {
			var = varmap [j];
			k = spow [j];
			CALIB_FATAL_ERROR_IF (k < 0);
			if (var < 0) {
				CALIB_FATAL_ERROR_IF (k NE 0);
				continue;
			}
			/* Doing += here permits multiple source vars	*/
			/* to map to a single destination var.		*/
			dpow [var] += k;
		}
		mpz_set (&(rop -> coeff [i]), &(op -> coeff [i]));
	}
	rop -> nterms = nterms;
}

/*
 * Copy the given source polynomial (in Z[x] form) into the given
 * destination polynomial (in multi-variable Zxyz form).  The source
 * polynomial becomes a polynomial in the given variable of the
 * destination polynomial's domain.
 */

	static
	void
Zxyz_copy_from_Zx (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: Destination polynomial */
const struct calib_Zx_obj *	op,	/* IN: Source polynomial in Z[x] */
int				var	/* IN: Destination variable to use */
)
{
int				i, k, nvars, deg, nterms;
const struct calib_Zxyz_dom *	dom;
int *				dpow;
int *				pp;
mpz_ptr				dcoeff, dcp;
mpz_srcptr			scoeff, scp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	deg	= op -> degree;
	scoeff	= op -> coeff;

	/* Count the number of terms in source polynomial. */
	nterms = 0;
	for (i = 0; i <= deg; i++) {
		if (mpz_sgn (&scoeff [i]) NE 0) {
			++nterms;
		}
	}

	Zxyz_alloc (rop, nterms);

	dpow	= rop -> pow;
	dcoeff	= rop -> coeff;

	/* Zero all of the powers. */
	k = nterms * nvars;
	memset (dpow, 0, k * sizeof (dpow [0]));

	pp = dpow;
	dcp = dcoeff;
	scp = &scoeff [deg];
	for (i = deg; i >= 0; i--, scp--) {
		if (mpz_sgn (scp) NE 0) {
			pp [var] = i;
			mpz_set (dcp, scp);
			pp += nvars;
			++dcp;
		}
	}
	rop -> nterms = nterms;
}

/*
 * Copy the given source polynomial (in multi-variable Zxyz form) into
 * the given destination polynomial (in Z[x] form).  The source polynomial
 * is in the given variable, and must contain no other variables.
 */

	static
	void
Zxyz_copy_into_Zx (

struct calib_Zx_obj *		rop,	/* IN/OUT: Destination Z[x] polynomial */
const struct calib_Zxyz_obj *	op,	/* IN: Source polynomial */
int				var	/* IN: Variable of source polynomial */
)
{
int				i, j, nvars, deg, nterms;
const struct calib_Zxyz_dom *	dom;
const int *			spow;
const int *			spp;
mpz_ptr				dcoeff, dcp;
mpz_srcptr			scoeff, scp;
const struct calib_Zx_dom *	Zx;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= op -> dom;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	nterms	= op -> nterms;
	spow	= op -> pow;
	scoeff	= op -> coeff;

	/* Verify that no other vars exist in source polynomial.	*/
	/* Also determine the degree of source poly in var.		*/
	deg = -1;
	spp = spow;
	for (i = 0; i < nterms; i++) {
		for (j = 0; j < var; j++) {
			CALIB_FATAL_ERROR_IF (spp [j] NE 0);
		}
		j = spp [j];
		if (j > deg) {
			deg = j;
		}
		for (j = var + 1; j < nvars; j++) {
			CALIB_FATAL_ERROR_IF (spp [j] NE 0);
		}
	}

	Zx = calib_get_Zx_dom ();

	Zx -> alloc (rop, deg);

	scp	= scoeff;
	dcoeff	= rop -> coeff;
	dcp = dcoeff;
	for (i = 0; i <= deg; i++, dcp++) {
		mpz_set_ui (dcp, 0);		
	}

	/* Copy coefficients into destination poly. */
	spp = spow;
	for (i = 0; i < nterms; i++, scp++) {
		CALIB_FATAL_ERROR_IF (mpz_sgn (scp) EQ 0);
		j = spp [var];
		mpz_set (&dcoeff [j], scp);
		spp += nvars;
	}
	rop -> degree = deg;
}

/*
 * Copy the given source polynomial (in Zp[x] form) into the given
 * destination polynomial (in multi-variable Zxyz form).  The source
 * polynomial becomes a polynomial in the given variable of the
 * destination polynomial's domain.
 */

	static
	void
Zxyz_copy_from_Zpx (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: Destination polynomial */
const struct calib_Zpx_obj *	op,	/* IN: Source polynomial in Zp[x] */
int				var	/* IN: Destination variable to use */
)
{
int				i, k, nvars, deg, nterms;
const struct calib_Zxyz_dom *	dom;
int *				dpow;
int *				pp;
mpz_ptr				dcoeff, dcp;
mpz_srcptr			scoeff, scp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	deg	= op -> degree;
	scoeff	= op -> coeff;

	/* Count the number of terms in source polynomial. */
	nterms = 0;
	for (i = 0; i <= deg; i++) {
		if (mpz_sgn (&scoeff [i]) NE 0) {
			++nterms;
		}
	}

	Zxyz_alloc (rop, nterms);

	dpow	= rop -> pow;
	dcoeff	= rop -> coeff;

	/* Zero all of the powers. */
	k = nterms * nvars;
	memset (dpow, 0, k * sizeof (dpow [0]));

	pp = dpow;
	dcp = dcoeff;
	scp = &scoeff [deg];
	for (i = deg; i >= 0; i--, scp--) {
		if (mpz_sgn (scp) NE 0) {
			pp [var] = i;
			mpz_set (dcp, scp);
			pp += nvars;
			++dcp;
		}
	}
	rop -> nterms = nterms;
}

/*
 * Copy the given source polynomial (in multi-variable Zxyz form) into
 * the given destination polynomial (in Zp[x] form).  The source polynomial
 * is in the given variable, and must contain no other variables.
 */

	static
	void
Zxyz_copy_into_Zpx (

struct calib_Zpx_obj *		rop,	/* IN/OUT: Destination Zp[x] polynomial */
const struct calib_Zxyz_obj *	op,	/* IN: Source polynomial */
int				var	/* IN: Variable of source polynomial */
)
{
int				i, j, nvars, deg, nterms;
const struct calib_Zpx_dom *	Zpx;
const struct calib_Zxyz_dom *	dom;
const int *			spow;
const int *			spp;
mpz_ptr				dcoeff, dcp;
mpz_srcptr			scoeff, scp;
const struct calib_Zp_dom *	Zp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	Zpx	= rop -> dom;
	dom	= op -> dom;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	Zp = Zpx -> cf;
	CALIB_FATAL_ERROR_IF (Zp EQ NULL);

	nterms	= op -> nterms;
	spow	= op -> pow;
	scoeff	= op -> coeff;

	/* Verify that no other vars exist in source polynomial.	*/
	/* Also determine the degree of source poly in var.		*/
	deg = -1;
	spp = spow;
	for (i = 0; i < nterms; i++) {
		for (j = 0; j < var; j++) {
			CALIB_FATAL_ERROR_IF (spp [j] NE 0);
		}
		j = spp [var];
		if (j > deg) {
			deg = j;
		}
		for (j = var + 1; j < nvars; j++) {
			CALIB_FATAL_ERROR_IF (spp [j] NE 0);
		}
	}

	/* Force destination sufficiently large for this degree. */
	Zpx -> alloc (rop, deg);

	scp	= scoeff;
	dcoeff	= rop -> coeff;
	dcp = dcoeff;
	for (i = 0; i <= deg; i++, dcp++) {
		mpz_set_ui (dcp, 0);		
	}

	/* Copy coefficients into destination poly. */
	spp = spow;
	for (i = 0; i < nterms; i++, scp++) {
		CALIB_FATAL_ERROR_IF (mpz_sgn (scp) EQ 0);
		j = spp [var];
		Zp -> set_z (Zp, &dcoeff [j], scp);
		spp += nvars;
	}
	rop -> degree = deg;
}

/*
 * Copy the given source polynomial (in Q(a)[x] form) into the given
 * destination polynomial (in multi-variable Zxyz form).  The source
 * polynomial's main variable becomes the given xvar of the destination
 * polynomial, while the source polynomial's algebraic variable becomes
 * the given avar of the destination polynomial.
 */

	static
	void
Zxyz_copy_from_Qax (

struct calib_Zxyz_obj *	rop,	/* IN/OUT: Destination polynomial */
mpz_ptr			denom,	/* IN/OUT: Common denominator of coeffs */
const struct calib_Qax_obj *
			op,	/* IN: Source polynomial in Q(a)[x] */
int			xvar,	/* IN: Dest var to use for "main" src var */
int			avar	/* IN: Dest var to use for algebraic src var */
)
{
int				i, j, nvars, deg, deg2;
const struct calib_Zxyz_dom *	dom;
int *				dpow;
struct paccum *			pap;
mpq_ptr				qcoeffs;
mpq_srcptr			sqp;
mpz_srcptr			szp;
struct QaxCoeffs		COEFFS;
mpz_t				G, L, TEMP;

	CALIB_FATAL_ERROR_IF ((rop    EQ NULL) OR
			      (denom  EQ NULL) OR
			      (op     EQ NULL));

	dom	= rop -> dom;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF (nvars < 2);

	CALIB_FATAL_ERROR_IF ((xvar < 0) OR (xvar >= nvars));
	CALIB_FATAL_ERROR_IF ((avar < 0) OR (avar >= nvars));
	CALIB_FATAL_ERROR_IF (xvar EQ avar);

	deg	= op -> degree;

	/* Get all of the coefficients in rational form. */
	_calib_QaxCoeffs_init (&COEFFS, op);

	/* Compute LCM of coefficient denominators. */
	mpz_init (G);
	mpz_init (L);
	mpz_init (TEMP);

	mpz_set_ui (L, 1);
	for (i = deg; i >= 0; i--) {
		deg2	= COEFFS.qax_coeff [i].qa_deg;
		qcoeffs	= COEFFS.qax_coeff [i].qa_coeff;
		for (j = 0; j <= deg2; j++) {
			sqp = &qcoeffs [j];
			szp = mpq_denref (sqp);
			if (mpz_cmp_ui (szp, 1) EQ 0) continue;
			mpz_gcd (G, L, szp);
			mpz_divexact (TEMP, szp, G);
			mpz_mul (L, L, TEMP);
		}
	}
	mpz_set (denom, L);

	dpow = CALIB_NEWA (nvars, int);
	memset (dpow, 0, nvars * (sizeof (dpow [0])));

	pap = Zxyz_get_paccum (dom);

	for (i = 0; i <= deg; i++) {
		dpow [xvar] = i;
		deg2	= COEFFS.qax_coeff [i].qa_deg;
		qcoeffs	= COEFFS.qax_coeff [i].qa_coeff;
		for (j = 0; j <= deg2; j++) {
			sqp = &qcoeffs [j];
			if (mpq_sgn (sqp) EQ 0) continue;
			szp = mpq_denref (sqp);
			mpz_divexact (TEMP, L, szp);
			mpz_mul (TEMP, TEMP, mpq_numref (sqp));
			dpow [avar] = j;
			Zxyz_paccum_add_term (pap, TEMP, dpow);
		}
	}

	Zxyz_set_poly_from_paccum (rop, pap);

	Zxyz_put_paccum (pap);

	mpz_clear (TEMP);
	mpz_clear (L);
	mpz_clear (G);

	_calib_QaxCoeffs_clear (&COEFFS);
}

/*
 * Copy the given source polynomial (in multi-variable Zxyz form) into
 * the given destination polynomial (in Q(a)[x] form).  The destination's
 * main variable is taken from xvar in the source polynomial, whereas the
 * destination's algebraic variable is taken from avar in the source
 * polynomial.
 */

	static
	void
Zxyz_copy_into_Qax (

struct calib_Qax_obj *	rop,	/* IN/OUT: Destination Q(a)[x] polynomial */
const struct calib_Zxyz_obj *
			op,	/* IN: Source polynomial */
int			xvar,	/* IN: Main var of source polynomial */
int			avar	/* IN: Algebraic var of source polynomial */
)
{
int				i, j, k, m, nvars, nterms;
int				xdeg, adeg, xpow, apow;
const struct calib_Zxyz_dom *	dom;
const struct calib_Qax_dom *	K_of_x;
const struct calib_Qa_dom *	cf;
const int *			spow;
const int *			spp;
int *				sdeg;
struct calib_Qa_obj *		dcp;
mpz_srcptr			scoeff, scp;
struct calib_Qax_obj		DST;
struct calib_Qa_obj		QAVAL;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= op -> dom;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF (nvars < 2);

	CALIB_FATAL_ERROR_IF ((xvar < 0) OR (xvar >= nvars));
	CALIB_FATAL_ERROR_IF ((avar < 0) OR (avar >= nvars));

	K_of_x	= rop -> dom;
	cf	= K_of_x -> cf;
	k	= cf -> gpoly -> degree;
	nterms	= op -> nterms;
	spow	= op -> pow;
	scoeff	= op -> coeff;

	/* Determine degree of source poly in each var.			*/
	sdeg	= CALIB_NEWA (nvars, int);
	memset (sdeg, 0, nvars * sizeof (sdeg [0]));
	spp = spow;
	for (i = 0; i < nterms; i++, spp += nvars) {
		for (j = 0; j < nvars; j++) {
			m = spp [j];
			if (m > sdeg [j]) {
				sdeg [j] = m;
			}
		}
	}
	/* Verify that no other vars exist in source polynomial.	*/
	for (i = 0; i < nvars; i++) {
		if (sdeg [i] <= 0) continue;
		CALIB_FATAL_ERROR_IF ((i NE xvar) AND (i NE avar));
	}

	xdeg = sdeg [xvar];
	adeg = sdeg [avar];
	CALIB_FATAL_ERROR_IF (adeg >= k);

	free (sdeg);

	K_of_x -> init_degree (K_of_x, &DST,  xdeg);
	cf -> init (cf, &QAVAL);

	scp	= scoeff;
	spp	= spow;
	for (i = 0; i < nterms; i++) {
		xpow	= spp [xvar];
		apow	= spp [avar];

		cf -> set_var_power (&QAVAL, apow);
		cf -> mul_z (&QAVAL, &QAVAL, scp);

		dcp	= &DST.coeff [xpow];
		cf -> add (dcp, dcp, &QAVAL);

		++scp;
		spp += nvars;
	}
	DST.degree = xdeg;

	K_of_x -> set (rop, &DST);

	cf -> clear (&QAVAL);
	K_of_x -> clear (&DST);
}

/*
 * Is the given polynomial identically zero?
 */

	static
	calib_bool
Zxyz_zerop (

const struct calib_Zxyz_obj *	op	/* IN: polynomial to test */
)
{
	CALIB_FATAL_ERROR_IF (op EQ NULL);

	if (op -> nterms <= 0) {
		return (TRUE);
	}
	return (FALSE);
}


/*
 * Is the given polynomial identically one?
 */

	static
	calib_bool
Zxyz_onep (

const struct calib_Zxyz_obj *	op	/* IN: polynomial to test */
)
{
int				i, nvars;
const struct calib_Zxyz_dom *	dom;
const int *			pow;
mpz_srcptr			coeff;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	if (op -> nterms NE 1) {
		return (FALSE);
	}

	coeff	= op -> coeff;

	if (mpz_cmp_ui (&coeff [0], 1) NE 0) {
		return (FALSE);
	}

	dom	= op -> dom;
	pow	= op -> pow;
	nvars	= dom -> nvars;

	for (i = 0; i < nvars; i++) {
		if (pow [i] NE 0) {
			return (FALSE);
		}
	}

	return (TRUE);
}

/*
 * Convert the given general expression into a member of Z[x,y,z].
 * All of the variables in the given expression *MUST* match exactly
 * one of the variables of the given domain.
 */

	static
	void
Zxyz_set_genrep (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: Result polynomial */
const struct calib_genrep *	op	/* IN: expression to convert */
)
{
int				nvars, var;
const struct calib_Zxyz_dom *	dom;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp;
struct calib_Zxyz_obj		POLY, TEMP;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL));

	dom	= rop -> dom;
	nvars	= dom -> nvars;

	switch (op -> op) {
	case CALIB_GENREP_OP_SI:
		Zxyz_set_si (rop, op -> u.siop);
		break;

	case CALIB_GENREP_OP_Z:
		Zxyz_set_z (rop, op -> u.zop);
		break;

	case CALIB_GENREP_OP_Q:
		/* Denominator *MUST* be 1!!! */
		CALIB_FATAL_ERROR_IF (mpz_cmp_ui (mpq_denref (op -> u.qop), 1) NE 0);
		Zxyz_set_z (rop, mpq_numref (op -> u.qop));
		break;

	case CALIB_GENREP_OP_VAR:
		var = Zxyz_lookup_var (dom, op -> u.var);
		CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));
		Zxyz_set_var_power (rop, var, 1);
		break;

	case CALIB_GENREP_OP_ADD:
		Zxyz_init (dom, &POLY);
		Zxyz_init (dom, &TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zxyz_set_genrep (&TEMP, lp -> operand);
			Zxyz_add (&POLY, &POLY, &TEMP);
		}
		Zxyz_set (rop, &POLY);

		Zxyz_clear (&TEMP);
		Zxyz_clear (&POLY);
		break;

	case CALIB_GENREP_OP_MUL:
		Zxyz_init_si (dom, &POLY, 1);
		Zxyz_init (dom, &TEMP);

		for (lp = op -> u.list; lp NE NULL; lp = lp -> next) {
			Zxyz_set_genrep (&TEMP, lp -> operand);
			Zxyz_mul (&POLY, &POLY, &TEMP);
		}
		Zxyz_set (rop, &POLY);

		Zxyz_clear (&TEMP);
		Zxyz_clear (&POLY);
		break;

	case CALIB_GENREP_OP_IPOW:
		gp = op -> u.ipow.base;
		CALIB_FATAL_ERROR_IF (gp EQ NULL);
		if (gp -> op EQ CALIB_GENREP_OP_VAR) {
			/* Special case to handle this more efficiently	*/
			/* that doing Zxyz_ipow() on a single var**1	*/
			/* term.					*/
			var = Zxyz_lookup_var (dom, gp -> u.var);
			CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));
			Zxyz_set_var_power (rop, var, op -> u.ipow.power);
		}
		else {
			/* Exponentiating something else. */
			Zxyz_init (dom, &TEMP);
			Zxyz_set_genrep (&TEMP, op -> u.ipow.base);
			Zxyz_ipow (rop, &TEMP, op -> u.ipow.power);
			Zxyz_clear (&TEMP);
		}
		break;

	default:
		CALIB_FATAL_ERROR;
		break;
	}
}

/*
 * Return the variable index for the given variable name, or -1 if
 * the given variable is not one of the variables in the domain.
 */

	static
	int
Zxyz_lookup_var (

const struct calib_Zxyz_dom *	dom,	/* IN: domain to look variable up within */
const char *			var	/* IN: variable name */
)
{
int	i, nvars;
char **	vars;

	CALIB_FATAL_ERROR_IF ((dom EQ NULL) OR (var EQ NULL));

	nvars	= dom -> nvars;
	vars	= dom -> vars;
	for (i = 0; i < nvars; i++) {
		if (strcmp (var, vars [i]) EQ 0) {
			return (i);
		}
	}
	return (-1);
}

/*
 * Convert the given element of Z[x,y,z] into general representation.
 */

	static
	struct calib_genrep *
Zxyz_to_genrep (

const struct calib_Zxyz_obj *	op	/* IN: polynomial to convert */
)
{
int				i, j, k, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
const int *			pow;
const int *			pp;
mpz_srcptr			coeff;
mpz_srcptr			cp;
struct calib_genrep_list *	terms;
struct calib_genrep_list **	term_hookp;
struct calib_genrep_list *	factors;
struct calib_genrep_list **	factor_hookp;
struct calib_genrep_list *	lp;
struct calib_genrep *		gp1;
struct calib_genrep *		gp2;

	CALIB_FATAL_ERROR_IF (op EQ NULL);

	nterms	= op -> nterms;
	dom	= op -> dom;
	pow	= op -> pow;
	coeff	= op -> coeff;
	nvars	= dom -> nvars;

	terms = NULL;
	term_hookp = &terms;

	pp = pow;
	for (i = 0; i < nterms; i++) {
		factors = NULL;
		factor_hookp = &factors;
		cp = &coeff [i];
		CALIB_FATAL_ERROR_IF (mpz_sgn (cp) EQ 0);
		if (mpz_cmp_ui (cp, 1) NE 0) {
			lp = CALIB_NEW (struct calib_genrep_list);
			lp -> operand	= calib_genrep_z (cp);
			lp -> next	= NULL;
			*factor_hookp = lp;
			factor_hookp = &(lp -> next);
		}
		for (j = 0; j < nvars; j++) {
			k = *pp++;
			if (k EQ 0) continue;
			gp1 = calib_genrep_var (dom -> vars [j]);
			if (k NE 1) {
				gp2 = calib_genrep_ipow (gp1, k);
				calib_genrep_free (gp1);
				gp1 = gp2;
			}
			lp = CALIB_NEW (struct calib_genrep_list);
			lp -> operand	= gp1;
			lp -> next	= NULL;
			*factor_hookp = lp;
			factor_hookp = &(lp -> next);
		}

		if (factors EQ NULL) {
			/* Coefficient was 1 and all powers were zero. */
			gp1 = calib_genrep_z (cp);
		}
		else if (factors -> next EQ NULL) {
			gp1 = factors -> operand;
			free (factors);
		}
		else {
			gp1 = CALIB_NEW (struct calib_genrep);
			gp1 -> op = CALIB_GENREP_OP_MUL;
			gp1 -> u.list = factors;
		}

		lp = CALIB_NEW (struct calib_genrep_list);
		lp -> operand = gp1;
		lp -> next = NULL;

		*term_hookp = lp;
		term_hookp = &(lp -> next);
	}

	if (terms EQ NULL) {
		gp1 = calib_genrep_si (0);
	}
	else if (terms -> next EQ NULL) {
		gp1 = terms -> operand;
		free (terms);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_ADD;
		gp1 -> u.list	= terms;
	}

	return (gp1);
}

/*
 * Convert the given list of factors into general representation.
 */

	static
	struct calib_genrep *
Zxyz_factors_to_genrep (

const struct calib_Zxyz_factor *	factors	/* IN: list of factors */
)
{
int					pow;
const struct calib_Zxyz_factor *	fp;
struct calib_genrep_list *		root;
struct calib_genrep_list **		hookp;
struct calib_genrep_list *		lp;
struct calib_genrep *			gp1;
struct calib_genrep *			gp2;

	root = NULL;
	hookp = &root;

	for (fp = factors; fp NE NULL; fp = fp -> next) {
		gp1 = Zxyz_to_genrep (fp -> factor);
		pow = fp -> multiplicity;

		/* NOTE: rat_factor() produces a list of	*/
		/* calib_Zxyz_factor objects for which the	*/
		/* denominator's factors have *negative*	*/
		/* multiplicities!  Only zero multiplicity is	*/
		/* bad here.					*/
		CALIB_FATAL_ERROR_IF (pow EQ 0);
		if ((pow > 1) OR (pow < 0)) {
			gp2 = CALIB_NEW (struct calib_genrep);
			gp2 -> op		= CALIB_GENREP_OP_IPOW;
			gp2 -> u.ipow.base	= gp1;
			gp2 -> u.ipow.power	= pow;
			gp1 = gp2;
		}
		lp = calib_genrep_new_list (gp1, NULL);
		*hookp = lp;
		hookp = &(lp -> next);
	}

	if (root EQ NULL) {
		gp1 = calib_genrep_si (1);
	}
	else if (root -> next EQ NULL) {
		gp1 = root -> operand;
		free (root);
	}
	else {
		gp1 = CALIB_NEW (struct calib_genrep);
		gp1 -> op	= CALIB_GENREP_OP_MUL;
		gp1 -> u.list	= root;
	}
	return (gp1);
}

/*
 * Given a polynomial P \in Z[x,y,z], and a main variable x,
 * let n = degree_x(P), and split P into an array A of n+1 polynomials
 * such that P = \sum_{i=0}^n A_i*x^i.  In other words, reorganize P
 * so that the "main" variable is x.  The degree n is returned, and
 * the dynamically allocated array of polynomials A is passed back
 * via a parameter.
 */

	static
	int
Zxyz_get_var_coeffs (

struct calib_Zxyz_obj **	coeffs,	/* IN/OUT: coefficient polynomials */
int				var,	/* IN: main variable to use */
const struct calib_Zxyz_obj *	p	/* IN: polynomial to reorganize */
)
{
int				i, j, n;
int				nterms, nvars;
const struct calib_Zxyz_dom *	dom;
const int *			pp;
mpz_srcptr			zp;
struct paccum **		accum;
struct calib_Zxyz_obj *		parray;

	CALIB_FATAL_ERROR_IF ((p EQ NULL) OR (coeffs EQ NULL));

	dom	= p -> dom;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	nterms	= p -> nterms;

	if (nterms <= 0) {
		/* Polynomial is zero.  No coefficients, degree is -1. */
		*coeffs = NULL;
		return (-1);
	}

	n = Zxyz_degree_in_var (p, var);

	/* Allocate accumulators for each coefficient.		*/
	accum = CALIB_NEWA (n + 1, struct paccum *);
	for (i = 0; i <= n; i++) {
		accum [i] = Zxyz_get_paccum (dom);
	}

	int pow [nvars];
	pp = &(p -> pow [0]);
	zp = &(p -> coeff [0]);
	for (i = 0; i < nterms; i++) {
		j = pp [var];
		/* Make copy of power vector for term that is freeof	*/
		/* var.							*/
		memcpy (&pow [0], &pp [0], nvars * sizeof (int));
		pow [var] = 0;
		/* Add this term to proper coeficient of var. */
		Zxyz_paccum_add_term (accum [j], zp, &pow [0]);
		pp += nvars;
		zp += 1;
	}

	/* Allocate/init array of polynomials. */
	parray = Zxyz_new_array (dom, n + 1);

	/* Transfer coefficients from paccum's into polynomials. */
	for (i = 0; i <= n; i++) {
		Zxyz_set_poly_from_paccum (&parray [i], accum [i]);
		Zxyz_put_paccum (accum [i]);
		accum [i] = NULL;
	}
	free (accum);

	*coeffs = parray;

	return (n);
}

/*
 * Get the coefficient of var^pow of the given polynomial.
 */

	static
	void
Zxyz_get_var_pow_coeff (

struct calib_Zxyz_obj *		rop,	/* IN/OUT: Resulting coefficient */
const struct calib_Zxyz_obj *	op,	/* IN: polynomial for which to */
					/*     get var^pow coefficient */
int				var,	/* IN: variable to select */
int				pow	/* IN: power to select */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
const int *			pp;
mpz_srcptr			cp;

	CALIB_FATAL_ERROR_IF ((rop EQ NULL) OR (op EQ NULL) OR (pow < 0));

	dom	= rop -> dom;

	CALIB_FATAL_ERROR_IF (op -> dom NE dom);

	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (var >= nvars));

	int	lpow [nvars];

	pap = Zxyz_get_paccum (dom);

	nterms	= op -> nterms;
	pp	= op -> pow;
	cp	= op -> coeff;
	for (i = 0; i < nterms; i++, cp++, pp += nvars) {
		if (pp [var] EQ pow) {
			memcpy (lpow, pp, nvars * sizeof (int));
			lpow [var] = 0;
			Zxyz_paccum_add_term (pap, cp, lpow);
		}
	}

	Zxyz_set_poly_from_paccum (rop, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Given an array of n+1 coefficient polynomials (each freeof the given var),
 * construct a polynomial from these coefficients.
 */

	static
	void
Zxyz_poly_from_coeffs (

struct calib_Zxyz_obj *	result,		/* IN/OUT: resulting polynomial */
int			degree,		/* IN: degree of coefficient array */
const struct calib_Zxyz_obj *
			coeffs,		/* IN: array of coefficient polys */
int			var		/* IN: new polynomial's main var */
)
{
int				i, j, nvars, nterms;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;
const int *			pp;
mpz_srcptr			zp;
const struct calib_Zxyz_obj *	poly;

	CALIB_FATAL_ERROR_IF (result EQ NULL);
	CALIB_FATAL_ERROR_IF ((degree < -1) OR
			      ((degree >= 0) AND (coeffs EQ NULL)));

	dom = result -> dom;
	for (i = 0; i >= degree; i++) {
		CALIB_FATAL_ERROR_IF (coeffs [i].dom NE dom);
	}

	nvars = dom -> nvars;
	int pow [nvars];
	pap = Zxyz_get_paccum (dom);
	for (i = 0; i <= degree; i++) {
		poly = &coeffs [i];
		nterms = poly -> nterms;
		zp = poly -> coeff;
		pp = poly -> pow;
		for (j = 0; j < nterms; j++) {
			memcpy (&pow [0], &pp [0], nvars * sizeof (int));
			CALIB_FATAL_ERROR_IF (pow [var] NE 0);
			pow [var] = i;
			Zxyz_paccum_add_term (pap, zp, &pow [0]);
			zp += 1;
			pp += nvars;
		}
	}

	Zxyz_set_poly_from_paccum (result, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Return a new, empty paccum object.  This paccum object will hold
 * terms that belong to the given domain.
 */

	static
	struct paccum *
Zxyz_get_paccum (

const struct calib_Zxyz_dom *	dom	/* IN: domain of terms we contain */
)
{
int			i;
struct paccum *		pap;
struct pterm **		bucket;
struct dlink *		dlink;

	pap = paccum_freelist;
	if (pap NE NULL) {
		paccum_freelist = pap -> next;
		pap -> next = NULL;

		CALIB_FATAL_ERROR_IF (pap -> dom NE NULL);
		pap -> dom = dom;

		return (pap);
	}

	/* Must create a new one from scratch. */

	bucket	= CALIB_NEWA (PACCUM_HASH_SIZE, struct pterm *);
	dlink	= CALIB_NEWA (PACCUM_HASH_SIZE + 1, struct dlink) + 1;

	memset (bucket, 0, PACCUM_HASH_SIZE * sizeof (bucket [0]));
	dlink [-1].forw = -1;
	dlink [-1].back = -1;
	for (i = 0; i < PACCUM_HASH_SIZE; i++) {
		dlink [i].forw	= -2;
		dlink [i].back	= -2;
	}

	pap = CALIB_NEW (struct paccum);	/* FIXME -- need to free these at end! */
	memset (pap, 0, sizeof (*pap));

	pap -> dom	= dom;
	pap -> nterms	= 0;
	pap -> bucket	= bucket;
	pap -> link	= dlink;
	pap -> blocks	= NULL;
	pap -> next	= NULL;

	Zxyz_establish_cleanup_handler ();

	return (pap);
}

/*
 * Free up the given paccum object.
 */

	static
	void
Zxyz_put_paccum (

struct paccum *		pap	/* IN: paccum object to free back to pool */
)
{
	if (pap EQ NULL) return;

	CALIB_FATAL_ERROR_IF (pap -> next NE NULL);

	/* Empty out this paccum. */
	Zxyz_reset_paccum (pap);

	pap -> dom = NULL;
	pap -> next = paccum_freelist;
	paccum_freelist = pap;
}

/*
 * Add the given term to the given paccum object.
 */

	static
	void
Zxyz_paccum_add_term (

struct paccum *		pap,	/* IN/OUT: paccum to add term to */
mpz_srcptr		coeff,	/* IN: numeric coefficient */
const int *		pow	/* IN: array of variable powers */
)
{
int		i1, i2, h;
int		nvars, nbytes;
struct pterm *	tp;
struct pterm **	bucket;
struct pterm **	hookp;
struct dlink *	link;

	CALIB_FATAL_ERROR_IF ((pap   EQ NULL) OR
			      (coeff EQ NULL) OR
			      (pow   EQ NULL));

	if (mpz_sgn (coeff) EQ 0) {
		/* Adding zero times anything is a no-op. */
		return;
	}

	nvars = pap -> dom -> nvars;
	nbytes = nvars * sizeof (int);

	h = Zxyz_pterm_hash (nvars, pow);

	bucket = &(pap -> bucket [h]);
	hookp = bucket;
	for (;;) {
		tp = *hookp;
		if (tp EQ NULL) break;

		if (memcmp (tp -> pow, pow, nbytes) EQ 0) goto found;

		hookp = &(tp -> next);
	}

	/* No existing term.  Need to add one. */
	tp = Zxyz_get_pterm (pap);
	mpz_set (tp -> coeff, coeff);
	memcpy (tp -> pow, pow, nbytes);

	/* This bucket transitioning from empty to non-empty? */
	if (*bucket EQ NULL) {
		/* Add this bucket to doubly-linked list of non-empty	*/
		/* buckets.						*/
		link = pap -> link;
		CALIB_FATAL_ERROR_IF ((link [h].forw NE -2) OR
				      (link [h].back NE -2));
		i1 = -1;
		i2 = link [i1].forw;
		link [h].back = i1;
		link [h].forw = i2;
		link [i1].forw = h;
		link [i2].back = h;
	}
	tp -> next = *bucket;
	*bucket = tp;
	++(pap -> nterms);
	return;

found:
	mpz_add (tp -> coeff, tp -> coeff, coeff);
	if (mpz_sgn (tp -> coeff) NE 0) {
		return;
	}

	/* This term cancelled to zero.  Remove it from hash table. */
	--(pap -> nterms);
	*hookp = tp -> next;
	Zxyz_put_pterm (tp, pap);

	/* Did this bucket become empty? */
	if (*bucket EQ NULL) {
		/* Yes.  Remove this bucket from doubly-linked list. */
		link = pap -> link;
		i1 = link [h].back;
		i2 = link [h].forw;
		CALIB_FATAL_ERROR_IF ((i1 <= -2) OR (i1 >= PACCUM_HASH_SIZE) OR
				      (i2 <= -2) OR (i2 >= PACCUM_HASH_SIZE));
		link [i1].forw = i2;
		link [i2].back = i1;
		link [h].forw = -2;
		link [h].back = -2;
	}
}

/*
 * Compute hash of given array of exponents.
 */

	static
	int
Zxyz_pterm_hash (

int		nvars,		/* IN: number of vars */
const int *	pow		/* IN: exponent for each var */
)
{
int		i;
unsigned long	reg;

	/* Assure that we have at least 64 bits. */
	CALIB_FATAL_ERROR_IF (sizeof (reg) NE 8);

	/* Initial register is all one bits. */
	reg = 0;
	reg = ~reg;

	for (i = 0; i < nvars; i++) {
		/* Rotate register 9 bits to the left. */
		reg = (reg << 9) | ((reg >> 55) & 0x1FF);
		reg ^= (pow [i] & 0xFFFFFFFF);
	}
	reg %= PACCUM_HASH_SIZE;

	return (reg);
}

/*
 * Copy all of the terms from the given paccum object into the
 * given polynomial object.
 */

	static
	void
Zxyz_set_poly_from_paccum (

struct calib_Zxyz_obj *	dst,		/* IN/OUT: polynomial to set */
const struct paccum *	pap		/* IN: paccum to copy terms from */
)
{
int				i, nterms, nvars;
const struct calib_Zxyz_dom *	dom;
struct pterm **			terms;
mpz_ptr				dcp;
int *				dpp;
const struct pterm *		tp;


	CALIB_FATAL_ERROR_IF ((dst EQ NULL) OR (pap EQ NULL));

	dom	= pap -> dom;
	nterms	= pap -> nterms;
	nvars	= dom -> nvars;

	CALIB_FATAL_ERROR_IF (dom NE dst -> dom);

	terms = Zxyz_get_sorted_terms (pap);

	Zxyz_alloc (dst, nterms);

	dcp = dst -> coeff;
	dpp = dst -> pow;

	for (i = 0; i < nterms; i++, dcp++, dpp += nvars) {
		tp = terms [i];
		mpz_set (dcp, tp -> coeff);
		memcpy (dpp, tp -> pow, nvars * sizeof (int));
	}
	dst -> nterms = nterms;

	free (terms);
}

/*
 * Return a single array of pointers to all the terms in the given
 * paccum object.  To be nice, we sort the array order by their
 * power tuple.
 *
 * It is the caller's responsibility to free the returned array of pointers.
 */

	static
	struct pterm **
Zxyz_get_sorted_terms (

const struct paccum *	pap		/* IN: paccum to get terms from */
)
{
int			i, j, k, n, nterms, nvars, status;
struct pterm **		bucket;
struct dlink *		link;
struct pterm **		sarray;
struct pterm **		sp;
struct pterm *		tp;
struct pterm *		tp1;
struct pterm *		tp2;

	CALIB_FATAL_ERROR_IF (pap EQ NULL);

	nterms	= pap -> nterms;
	nvars	= pap -> dom -> nvars;
	bucket	= pap -> bucket;
	link	= pap -> link;

	/* Make a link-list of all terms in paccum object. */
	sarray = CALIB_NEWA (nterms, struct pterm *);
	sp = sarray;
	for (i = link [-1].forw; i >= 0; i = link [i].forw) {
		tp = bucket [i];
		CALIB_FATAL_ERROR_IF (tp EQ NULL);
		for (; tp NE NULL; tp = tp -> next) {
			*sp++ = tp;
		}
	}
	CALIB_FATAL_ERROR_IF (sp NE sarray + nterms);

	/* Construct the heap via sift-downs, in O(n) time. */
	n = nterms;
	for (k = n >> 1; k >= 0; k--) {
		j = k;
		for (;;) {
			i = (j << 1) + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the right subchild is worse. */
				tp1 = sarray [i];
				tp2 = sarray [i + 1];
				status = Zxyz_compare_pow (nvars,
							   tp1 -> pow,
							   tp2 -> pow);
				CALIB_FATAL_ERROR_IF (status EQ 0);
				if (status > 0) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			tp1 = sarray [j];
			tp2 = sarray [i];
			status = Zxyz_compare_pow (nvars,
						   tp1 -> pow,
						   tp2 -> pow);
			CALIB_FATAL_ERROR_IF (status EQ 0);
			if (status < 0) {
				/* Worst child is better.  Sift-down	*/
				/* is done. */
				break;
			}
			/* Sift down and continue. */
			sarray [j] = tp2;
			sarray [i] = tp1;
			j = i;
		}
	}

	/* Now do actual sorting.  Exchange first/last and sift down. */
	while (n > 1) {
		/* Worst is at order [0], swap with order [n-1],	*/
		/* thereby putting it into final position.		*/
		--n;
		tp1 = sarray [0];
		sarray [0] = sarray [n];
		sarray [n] = tp1;

		/* Now restore the heap by sifting order [0] down. */
		j = 0;
		for (;;) {
			i = (j << 1) + 1;
			if (i + 1 < n) {
				/* Increment i (to right subchild of j) */
				/* if the right subchild is worse. */
				tp1 = sarray [i];
				tp2 = sarray [i + 1];
				status = Zxyz_compare_pow (nvars,
							   tp1 -> pow,
							   tp2 -> pow);
				CALIB_FATAL_ERROR_IF (status EQ 0);
				if (status > 0) {
					++i;
				}
			}
			if (i >= n) {
				/* Hit bottom of heap, sift-down is done. */
				break;
			}
			tp1 = sarray [j];
			tp2 = sarray [i];
			status = Zxyz_compare_pow (nvars,
						   tp1 -> pow,
						   tp2 -> pow);
			CALIB_FATAL_ERROR_IF (status EQ 0);
			if (status < 0) {
				/* Worst child is better.  Sift-down	*/
				/* is done. */
				break;
			}
			/* Sift down and continue. */
			sarray [j] = tp2;
			sarray [i] = tp1;
			j = i;
		}
	}

	return (sarray);
}

/*
 * Compare two arrays of integer powers.
 */

	static
	int
Zxyz_compare_pow (

int		nvars,		/* IN: number of variables */
const int *	pow1,		/* IN: first array of powers */
const int *	pow2		/* IN: second array of powers */
)
{
int	i, p1, p2;

	for (i = 0; i < nvars; i++) {
		p1 = pow1 [i];
		p2 = pow2 [i];
		if (p1 < p2) return (-1);
		if (p1 > p2) return (+1);
	}
	return (0);
}

/*
 * Allocate a new pterm object for the given paccum.
 */

	static
	struct pterm *
Zxyz_get_pterm (

struct paccum *		pap	/* IN: paccum to get term for/from */
)
{
int			nbytes;
struct pterm *		tp;
struct mblk *		mp;

	CALIB_FATAL_ERROR_IF (pap EQ NULL);

	tp = pap -> freelist;
	if (tp NE NULL) {
		pap -> freelist = tp -> next;
		tp -> next = NULL;
		return (tp);
	}

	/* Compute size of pterm object (these are variable size). */
	nbytes = PTERM_SIZE (pap -> dom -> nvars);

	if (pap -> nleft < nbytes) {
		mp = Zxyz_get_mblk ();
		mp -> next = pap -> blocks;
		pap -> blocks	= mp;
		pap -> allocp	= &(mp -> bytes [0]);
		pap -> nleft	= sizeof (mp -> bytes);

		CALIB_FATAL_ERROR_IF (nbytes > pap -> nleft);
	}

	tp = (struct pterm *) (pap -> allocp);
	mpz_init (tp -> coeff);
	pap -> allocp += nbytes;
	pap -> nleft  -= nbytes;
	tp -> next = NULL;

	return (tp);
}

/*
 * Free the given pterm object back to the freelist of the given
 * paccum object.
 */

	static
	void
Zxyz_put_pterm (

struct pterm *		tp,	/* IN: term to free */
struct paccum *		pap	/* IN: paccum to put term back into */
)
{
	CALIB_FATAL_ERROR_IF ((tp EQ NULL) OR (pap EQ NULL));

	tp -> next = pap -> freelist;
	pap -> freelist = tp;
}

/*
 * Force the given paccum object to be empty.
 */

	static
	void
Zxyz_reset_paccum (

struct paccum *		pap	/* IN: paccum object to empty out */
)
{
int		i, j;
struct pterm *	tp;
struct mblk *	mp;
struct mblk *	mp2;

	if (pap EQ NULL) return;

	for (;;) {
		i = pap -> link [-1].forw;
		if (i < 0) {
			CALIB_FATAL_ERROR_IF (i < -1);
			break;
		}
		CALIB_FATAL_ERROR_IF (i >= PACCUM_HASH_SIZE);
		tp = pap -> bucket [i];
		CALIB_FATAL_ERROR_IF (tp EQ NULL);
		pap -> bucket [i] = NULL;
		for (; tp NE NULL; tp = tp -> next) {
			mpz_clear (tp -> coeff);
			--(pap -> nterms);
		}
		CALIB_FATAL_ERROR_IF (pap -> link [i].back NE -1);
		j = pap -> link [i].forw;
		pap -> link [j].back = -1;
		pap -> link [-1].forw = j;
		pap -> link [i].forw = -2;
		pap -> link [i].back = -2;
	}

	CALIB_FATAL_ERROR_IF (pap -> nterms NE 0);

	for (tp = pap -> freelist; tp NE NULL; tp = tp -> next) {
		mpz_clear (tp -> coeff);
	}
	pap -> freelist	= NULL;

	pap -> allocp	= NULL;
	pap -> nleft	= 0;

	mp = pap -> blocks;
	pap -> blocks = NULL;

	while (mp NE NULL) {
		mp2 = mp -> next;
		mp -> next = NULL;
		Zxyz_put_mblk (mp);
		mp = mp2;
	}
}

/*
 * Allocate and initialize an array of Z[x] polynomials.
 */

	static
	struct calib_Zx_obj *
Zxyz_new_Zx_array (

int		n	/* IN: number of elements in created array */
)
{
int				i;
struct calib_Zx_obj *		array;
const struct calib_Zx_dom *	Zx;

	array = NULL;
	if (n > 0) {
		Zx = calib_get_Zx_dom ();
		array = CALIB_NEWA (n, struct calib_Zx_obj);
		for (i = 0; i < n; i++) {
			Zx -> init (&array [i]);
		}
	}
	return (array);
}

/*
 * Free up the given array of Z[x] polynomials.
 */

	static
	void
Zxyz_free_Zx_array (

struct calib_Zx_obj *	array,		/* IN: array to free */
int			n		/* IN: number of elements */
)
{
int				i;
const struct calib_Zx_dom *	Zx;

	if (array EQ NULL) return;

	Zx = calib_get_Zx_dom ();
	for (i = 0; i < n; i++) {
		Zx -> clear (&array [i]);
	}
	free (array);
}

/*
 * Allocate and initialize an array of Zp[x] polynomials.
 */

	static
	struct calib_Zpx_obj *
Zxyz_new_Zpx_array (

struct calib_Zpx_dom *	Zpx,	/* IN: Zp[x] domain to use */
int			n	/* IN: number of elements in created array */
)
{
int			i;
struct calib_Zpx_obj *	array;

	CALIB_FATAL_ERROR_IF (Zpx EQ NULL);

	array = NULL;
	if (n > 0) {
		array = CALIB_NEWA (n, struct calib_Zpx_obj);
		for (i = 0; i < n; i++) {
			Zpx -> init (Zpx, &array [i]);
		}
	}
	return (array);
}

/*
 * Free up the given array of Zp[x] polynomials.
 */

	static
	void
Zxyz_free_Zpx_array (

struct calib_Zpx_dom *	Zpx,	/* IN: Zp[x] domain to use */
struct calib_Zpx_obj *	array,	/* IN: array to free */
int			n	/* IN: number of elements */
)
{
int			i;

	if (array EQ NULL) return;

	for (i = 0; i < n; i++) {
		Zpx -> clear (&array [i]);
	}
	free (array);
}

/*
 * Convert the given polynomial into recursive form, using the given
 * "recursive-to-flat" variable mapping.
 */

	static
	struct RPoly *
Zxyz_rpoly_cv_f2r (

const struct calib_Zxyz_obj *
			poly,		/* IN: flat polynomial to convert */
int			nv,		/* IN: num vars in r2f_map */
const int *		r2f_map,	/* IN: recursive-to-flat var map */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
int				i, k, nterms, nvars, sp;
int				rvar, fvar;
const struct calib_Zxyz_dom *	dom;
struct RPoly *			result;
const int *			ip;
mpz_srcptr			zp;
mpz_ptr				zp2;
struct RPoly **			phookp;
struct RPoly *			pp;
struct RPoly *			pp2;
struct RTerm **			thookp;
struct RTerm *			tp;
struct RTerm *			tp2;

	CALIB_FATAL_ERROR_IF ((poly    EQ NULL) OR
			      (r2f_map EQ NULL) OR
			      (gip     EQ NULL));

	nterms	= poly -> nterms;
	dom	= poly -> dom;
	ip	= poly -> pow;
	zp	= poly -> coeff;
	nvars	= dom -> nvars;

	struct RPoly ** phook_stack [nvars + 1];
	struct RTerm ** thook_stack [nvars + 1];

	result = NULL;

	for (i = 0; i < nterms; i++, ip += nvars, zp++) {
		CALIB_FATAL_ERROR_IF (mpz_sgn (zp) EQ 0);
		phookp = &result;
		sp = 0;
		for (rvar = nv-1; rvar >= 0; rvar--) {
			fvar = r2f_map [rvar];
			k = ip [fvar];
			if (k EQ 0) continue;
			CALIB_FATAL_ERROR_IF (k < 0);

			/* Need rvar^k entry. */
			for (;;) {
				pp = *phookp;
				if ((pp EQ NULL) OR (pp -> var < rvar)) {
					tp2 = NULL;
					if (pp NE NULL) {
						tp2 = Zxyz_rpool_new_rterm (gip);
						tp2 -> coeff	= pp;
						tp2 -> next	= NULL;
						tp2 -> pow	= 0;
					}
					pp2 = Zxyz_rpool_new_rpoly (gip);
					pp2 -> var	= rvar;
					pp2 -> u.terms	= tp2;
					*phookp = pp2;
					pp = pp2;
					break;
				}
				if (pp -> var EQ rvar) break;
				/* Have pp -> var > rvar. */
				thookp = &(pp -> u.terms);
				for (;;) {
					tp = *thookp;
					if (tp EQ NULL) {
						tp = Zxyz_rpool_new_rterm (gip);
						tp -> coeff	= NULL;
						tp -> next	= NULL;
						tp -> pow	= 0;
						*thookp = tp;
						break;
					}
					if (tp -> pow EQ 0) break;
					thookp = &(tp -> next);
				}
				phook_stack [sp] = phookp;
				thook_stack [sp] = thookp;
				++sp;
				phookp = &(tp -> coeff);
			}
			/* Find (or make) k-th power term. */
			thookp = &(pp -> u.terms);
			for (;;) {
				tp = *thookp;
				if ((tp EQ NULL) OR (tp -> pow < k)) {
					tp2 = Zxyz_rpool_new_rterm (gip);
					tp2 -> coeff	= NULL;
					tp2 -> next	= tp;
					tp2 -> pow	= k;
					*thookp = tp2;
					tp = tp2;
					break;
				}
				if (tp -> pow EQ k) break;
				/* tp -> pow > k.  Search farther down. */
				thookp = &(tp -> next);
			}
			phook_stack [sp] = phookp;
			thook_stack [sp] = thookp;
			++sp;
			phookp = &(tp -> coeff);
		}
		for (;;) {
			pp = *phookp;
			if (pp EQ NULL) {
				pp = Zxyz_rpoly_new_si (0, gip);
				*phookp = pp;
				break;
			}
			if (pp -> var < 0) break;
			thookp = &(pp -> u.terms);
			for (;;) {
				tp = *thookp;
				if (tp EQ NULL) {
					tp = Zxyz_rpool_new_rterm (gip);
					tp -> coeff	= NULL;
					tp -> next	= NULL;
					tp -> pow	= 0;
					*thookp = tp;
					break;
				}
				if (tp -> pow EQ 0) break;
				thookp = &(tp -> next);
			}
			phook_stack [sp] = phookp;
			thook_stack [sp] = thookp;
			++sp;
			phookp = &(tp -> coeff);
		}
		/* We were hitting this error.  The immediately		*/
		/* preceeding loop should allviate this problem.	*/
		CALIB_FATAL_ERROR_IF (pp -> var >= 0);

		zp2 = pp -> u.zval;
		mpz_add (zp2, zp2, zp);
		if (mpz_sgn (zp2) NE 0) continue;

		/* This term cancelled.  Delete stuff back up the tree. */
		pp -> u.zval = NULL;
		calib_free_Z_vector (zp2, 1);
		Zxyz_rpool_free_rpoly (pp, gip);
		*phookp = NULL;
		while (sp > 0) {
			--sp;
			thookp = thook_stack [sp];
			tp = *thookp;
			CALIB_FATAL_ERROR_IF ((tp EQ NULL) OR
					      (phookp NE &(tp -> coeff)));
			*thookp = tp -> next;
			Zxyz_rpool_free_rterm (tp, gip);

			phookp = phook_stack [sp];
			pp = *phookp;
			CALIB_FATAL_ERROR_IF (pp EQ NULL);
			if (pp -> u.terms NE NULL) break;
			CALIB_FATAL_ERROR_IF (thookp NE &(pp -> u.terms));
			*phookp = NULL;
			Zxyz_rpool_free_rpoly (pp, gip);
		}
	}

	return (result);
}

/*
 * Convert the given recursive polynomial into flat form, using the given
 * "recursive-to-flat" variable mapping.
 */

	static
	void
Zxyz_rpoly_cv_r2f (

struct calib_Zxyz_obj *	result,		/* IN: flat poly to convert into */
const struct RPoly *	pp,		/* IN: recursive poly to convert */
const int *		r2f_map		/* IN: recursive-to-flat var map */
)
{
int				nvars;
const struct calib_Zxyz_dom *	dom;
struct paccum *			pap;

	CALIB_FATAL_ERROR_IF ((result EQ NULL) OR (r2f_map EQ NULL));

	dom	= result -> dom;
	nvars	= dom -> nvars;

	pap = Zxyz_get_paccum (dom);

	int		pow [nvars];
	calib_bool	varseen [nvars];

	memset (&pow [0], 0, nvars * sizeof (int));
	memset (&varseen [0], FALSE, nvars * sizeof (calib_bool));

	/* Recursively copy terms into pap. */
	Zxyz_rpoly_r2f_recurse (pap, pow, varseen, r2f_map, nvars, pp);

	Zxyz_set_poly_from_paccum (result, pap);

	Zxyz_put_paccum (pap);
}

/*
 * Recursive portion of Zxyz_rpoly_cv_r2f().
 */

	static
	void
Zxyz_rpoly_r2f_recurse (

struct paccum *		pap,		/* IN/OUT: paccum to add terms to */
int *			pow,		/* IN: array of flat var powers */
calib_bool *		varseen,	/* IN: which flat vars seen? */
const int *		r2f_map,	/* IN: recursive to flat var map */
int			parent_var,	/* IN: parent var */
const struct RPoly *	pp		/* IN: recursive poly to convert */
)
{
int		rvar, fvar;
struct RTerm *	tp;

	if (pp EQ NULL) {
		/* NULL is the zero polynomial. */
		return;
	}

	rvar = pp -> var;
	if (rvar < 0) {
		/* A constant coefficient. */
		Zxyz_paccum_add_term (pap, pp -> u.zval, pow);
		return;
	}

	/* Check for recursive vars out of order. */
	CALIB_FATAL_ERROR_IF (rvar >= parent_var);
	fvar = r2f_map [rvar];
	CALIB_FATAL_ERROR_IF (varseen [fvar]);

	varseen [fvar] = TRUE;

	for (tp = pp -> u.terms; tp NE NULL; tp = tp -> next) {
		pow [fvar] = tp -> pow;
		Zxyz_rpoly_r2f_recurse (pap, pow, varseen, r2f_map, rvar, tp -> coeff);
	}

	pow [fvar] = 0;
	varseen [fvar] = FALSE;
}

/*
 * Given recursive polynomials a and b, compute the quotient a/b.
 * If a is not evenly divisible by b, then return DIVISION_NOT_EXACT,
 * a special pointer value designed to indicate this fact.
 *
 * This code assumes that both polynomials are fully simplified (i.e.,
 * all vars are in proper lexicographic order, no RPoly nodes having only
 * degree-zero terms, no RTerms with NULL coeff ptrs, etc.).
 */

	static
	struct RPoly *
Zxyz_rpoly_div_exact (

const struct RPoly *	a,		/* IN: Dividend polynomial */
const struct RPoly *	b,		/* IN: Divisor polynomial */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
int		av, bv, exact, apow, bpow;
struct RTerm **	thookp;
struct RTerm	*terms, *tp, *tp2, *atp, *btp;
struct RPoly	*qp, *qp2, *qp3, *alc, *blc, *varpow, *prod, *newrem;
mpz_srcptr	azp, bzp;
mpz_t		QUOT, REM;

	/* Must not divide by zero! */
	CALIB_FATAL_ERROR_IF (b EQ NULL);

	if (a EQ NULL) {
		/* Zero is evenly divisible by any non-zero polynomial. */
		return (NULL);
	}

	av = a -> var;
	bv = b -> var;

	if (bv < 0) {
		/* Dividing by an integer coefficient. */
		bzp = b -> u.zval;

		if (av >= 0) goto divide_by_lesser_poly;

		/* Dividing an integer by an integer. */
		azp = a -> u.zval;
		mpz_init (QUOT);
		mpz_init (REM);
		mpz_tdiv_qr (QUOT, REM, azp, bzp);
		if (mpz_sgn (REM) EQ 0) {
			qp = Zxyz_rpoly_new_z (QUOT, gip);
		}
		else {
			qp = DIVISION_NOT_EXACT;
		}
		mpz_clear (REM);
		mpz_clear (QUOT);

		return (qp);
	}

	/* Dividing by a polynomial. */

	if (av < 0) {
		/* Dividing an integer (non-zero) by a polynomial (of	*/
		/* degree at least 1).					*/
		return (DIVISION_NOT_EXACT);
	}

	/* Dividing a polynomial by a polynomial. */

	if (bv > av) {
		/* Divisor has bigger main variable. */
		return (DIVISION_NOT_EXACT);
	}
	if (bv < av) goto divide_by_lesser_poly;

	/* We have a and b both polynomials in the same var.  Time to	*/
	/* do actual "long division."					*/

	atp = a -> u.terms;
	btp = b -> u.terms;
	CALIB_FATAL_ERROR_IF ((atp EQ NULL) OR (btp EQ NULL));
	alc	= atp -> coeff;
	apow	= atp -> pow;
	blc	= btp -> coeff;
	bpow	= btp -> pow;

	if (apow < bpow) {
		return (DIVISION_NOT_EXACT);
	}
	qp = Zxyz_rpoly_div_exact (alc, blc, gip);
	if (qp EQ DIVISION_NOT_EXACT) {
		return (DIVISION_NOT_EXACT);
	}
	varpow = Zxyz_rpoly_new_varpow (av, apow - bpow, gip);
	qp2 = Zxyz_rpoly_mul (varpow, qp, gip);
	Zxyz_rpoly_free (varpow, gip);
	Zxyz_rpoly_free (qp, gip);
	qp = qp2;
	prod = Zxyz_rpoly_mul (qp, b, gip);
	newrem = Zxyz_rpoly_sub (a, prod, gip);
	Zxyz_rpoly_free (prod, gip);
	qp2 = Zxyz_rpoly_div_exact (newrem, b, gip);
	Zxyz_rpoly_free (newrem, gip);
	if (qp2 EQ DIVISION_NOT_EXACT) {
		Zxyz_rpoly_free (qp, gip);
		return (DIVISION_NOT_EXACT);
	}
	qp3 = Zxyz_rpoly_add (qp, qp2, gip);
	Zxyz_rpoly_free (qp, gip);
	Zxyz_rpoly_free (qp2, gip);
	return (qp3);

	/* Come here whenever av > bv, regardless of whether divisor b	*/
	/* is an integer (bv < 0) or a polynomial (bv >= 0).  Simply	*/
	/* compute the quotient term by term.				*/

divide_by_lesser_poly:

	terms = NULL;
	thookp = &terms;
	exact = TRUE;
	for (tp = a -> u.terms; tp NE NULL; tp = tp -> next) {
		qp = Zxyz_rpoly_div_exact (tp -> coeff, b, gip);
		if (qp EQ DIVISION_NOT_EXACT) {
			exact = FALSE;
			break;
		}
		tp2 = Zxyz_rpool_new_rterm (gip);
		tp2 -> coeff	= qp;
		tp2 -> next	= NULL;
		tp2 -> pow	= tp -> pow;
		*thookp = tp2;
		thookp = &(tp2 -> next);
	}
	if (exact) {
		qp = Zxyz_rpoly_finish_terms (av, terms, gip);
	}
	else {
		Zxyz_rterms_free (terms, gip);
		qp = DIVISION_NOT_EXACT;
	}
	return (qp);
}

/*
 * Finish off the given list of terms (a polynomial in the given var),
 * by wrapping an RPoly around it.  Performs several simplifications
 * along the way.
 */

	static
	struct RPoly *
Zxyz_rpoly_finish_terms (

int		var,		/* IN: main variable for terms */
struct RTerm *	terms,		/* IN: linked list of terms */
struct GInfo *	gip		/* IN/OUT: global info */
)
{
struct RPoly *	pp;

	CALIB_FATAL_ERROR_IF (var < 0);

	if (terms EQ NULL) return (NULL);

	if (terms -> pow EQ 0) {
		CALIB_FATAL_ERROR_IF (terms -> next NE NULL);
		/* The highest (and only) power for this var is zero.	*/
		/* Drop all mention of this variable.			*/
		pp = terms -> coeff;
		Zxyz_rpool_free_rterm (terms, gip);
	}
	else {
		pp = Zxyz_rpool_new_rpoly (gip);
		pp -> var	= var;
		pp -> u.terms	= terms;
	}
	return (pp);
}

/*
 * Construct a polynomial that consists of the single term var^pow.
 */

	static
	struct RPoly *
Zxyz_rpoly_new_varpow (

int		var,		/* IN: variable to use */
int		pow,		/* IN: power to use */
struct GInfo *	gip		/* IN/OUT: global info */
)
{
struct RPoly *	pp;
struct RTerm *	tp;

	CALIB_FATAL_ERROR_IF ((var < 0) OR (pow < 0));

	pp = Zxyz_rpoly_new_si (1, gip);

	if (pow > 0) {
		tp = Zxyz_rpool_new_rterm (gip);
		tp -> coeff	= pp;
		tp -> next	= NULL;
		tp -> pow	= pow;

		pp = Zxyz_rpool_new_rpoly (gip);
		pp -> var	= var;
		pp -> u.terms	= tp;
	}
	return (pp);
}

/*
 * Add two recursive polynomials.
 */

	static
	struct RPoly *
Zxyz_rpoly_add (

const struct RPoly *	a,		/* IN: first polynomial */
const struct RPoly *	b,		/* IN: second polynomial */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
int			av, bv, tmpv, k, pow;
struct RTerm *		terms;
struct RTerm **		thookp;
struct RTerm *		atp;
struct RTerm *		btp;
struct RTerm *		tp2;
struct RPoly *		pp;
const struct RPoly *	tmpp;
mpz_t			TEMP;

	if (a EQ NULL) {
		return (Zxyz_rpoly_dup (b, gip));
	}
	if (b EQ NULL) {
		return (Zxyz_rpoly_dup (a, gip));
	}

	av = a -> var;
	bv = b -> var;

	if (av EQ bv) {
		if (av < 0) {
			/* Adding two integers. */
			mpz_init (TEMP);
			mpz_add (TEMP, a -> u.zval, b -> u.zval);
			pp = Zxyz_rpoly_new_z (TEMP, gip);
			mpz_clear (TEMP);
			return (pp);
		}

		terms = NULL;
		thookp = &terms;
		atp = a -> u.terms;
		btp = b -> u.terms;
		for (;;) {
			if (atp EQ NULL) {
				if (btp EQ NULL) break;
				pp = Zxyz_rpoly_dup (btp -> coeff, gip);
				pow = btp -> pow;
				btp = btp -> next;
			}
			else if ((btp EQ NULL) OR (atp -> pow > btp -> pow)) {
				pp = Zxyz_rpoly_dup (atp -> coeff, gip);
				pow = atp -> pow;
				atp = atp -> next;
			}
			else if (atp -> pow < btp -> pow) {
				pp = Zxyz_rpoly_dup (btp -> coeff, gip);
				pow = btp -> pow;
				btp = btp -> next;
			}
			else {
				pp = Zxyz_rpoly_add (atp -> coeff, btp -> coeff, gip);
				pow = atp -> pow;
				atp = atp -> next;
				btp = btp -> next;
			}
			if (pp NE NULL) {
				tp2 = Zxyz_rpool_new_rterm (gip);
				tp2 -> coeff	= pp;
				tp2 -> next	= NULL;
				tp2 -> pow	= pow;
				*thookp = tp2;
				thookp = &(tp2 -> next);
			}
		}
		pp = Zxyz_rpoly_finish_terms (av, terms, gip);
		return (pp);
	}

	if (av > bv) {
		tmpp = a;	a = b;		b = tmpp;
		tmpv = av;	av = bv;	bv = tmpv;
	}

	CALIB_FATAL_ERROR_IF ((bv < 0) OR (bv <= av));
	terms = NULL;
	thookp = &terms;
	for (btp = b -> u.terms; btp NE NULL; btp = btp -> next) {
		k = btp -> pow;
		if (k <= 0) {
			CALIB_FATAL_ERROR_IF ((k < 0) OR
					      (btp -> next NE NULL));
			break;
		}
		tp2 = Zxyz_rpool_new_rterm (gip);
		tp2 -> coeff = Zxyz_rpoly_dup (btp -> coeff, gip);
		tp2 -> next  = NULL;
		tp2 -> pow   = btp -> pow;
		*thookp = tp2;
		thookp = &(tp2 -> next);
	}
	if (btp NE NULL) {
		pp = Zxyz_rpoly_add (a, btp -> coeff, gip);
	}
	else {
		pp = Zxyz_rpoly_dup (a, gip);
	}
	if (pp NE NULL) {
		tp2 = Zxyz_rpool_new_rterm (gip);
		tp2 -> coeff	= pp;
		tp2 -> next	= NULL;
		tp2 -> pow	= 0;
		*thookp = tp2;
		thookp = &(tp2 -> next);
	}
	pp = Zxyz_rpoly_finish_terms (bv, terms, gip);
	return (pp);
}

/*
 * Subtract two recursive polynomials.
 */

	static
	struct RPoly *
Zxyz_rpoly_sub (

const struct RPoly *	a,		/* IN: first polynomial */
const struct RPoly *	b,		/* IN: second polynomial */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
struct RPoly *		minus_b;
struct RPoly *		res;

	/* Simple, but not the most efficient method. */
	minus_b = Zxyz_rpoly_mul_si (b, -1, gip);

	res = Zxyz_rpoly_add (a, minus_b, gip);

	Zxyz_rpoly_free (minus_b, gip);

	return (res);
}

/*
 * Multiply two recursive polynomials.
 */

	static
	struct RPoly *
Zxyz_rpoly_mul (

const struct RPoly *	a,	/* IN: first polynomial */
const struct RPoly *	b,	/* IN: second polynomial */
struct GInfo *		gip	/* IN/OUT: global info */
)
{
int			av, bv, tmpv;
int			apow, pow;
const struct RPoly *	tmpp;
struct RPoly *		pp;
struct RPoly *		pp2;
struct RPoly *		acoeff;
struct RTerm *		terms;
struct RTerm **		thookp;
struct RTerm *		tp;
struct RTerm *		tp2;
struct RTerm *		atp;
struct RTerm *		btp;

	if ((a EQ NULL) OR (b EQ NULL)) return (NULL);

	av = a -> var;
	bv = b -> var;

	if (av < 0) {
		pp = Zxyz_rpoly_mul_z (b, a -> u.zval, gip);
		return (pp);
	}
	if (bv < 0) {
		pp = Zxyz_rpoly_mul_z (a, b -> u.zval, gip);
		return (pp);
	}

	if (av EQ bv) goto same_var;

	if (av < bv) {
		tmpp = a;	a = b;		b = tmpp;
		tmpv = av;	av = bv;	bv = tmpv;
	}
	terms = NULL;
	thookp = &terms;
	for (atp = a -> u.terms; atp NE NULL; atp = atp -> next) {
		pp = Zxyz_rpoly_mul (atp -> coeff, b, gip);
		if (pp EQ NULL) continue;

		tp2 = Zxyz_rpool_new_rterm (gip);
		tp2 -> coeff	= pp;
		tp2 -> next	= NULL;
		tp2 -> pow	= atp -> pow;
		*thookp = tp2;
		thookp = &(tp2 -> next);
	}
	pp = Zxyz_rpoly_finish_terms (av, terms, gip);
	return (pp);

same_var:

	terms = NULL;
	for (atp = a -> u.terms; atp NE NULL; atp = atp -> next) {
		acoeff	= atp -> coeff;
		apow	= atp -> pow;
		thookp = &terms;
		for (btp = b -> u.terms; btp NE NULL; btp = btp -> next) {
			pp = Zxyz_rpoly_mul (acoeff, btp -> coeff, gip);
			if (pp EQ NULL) continue;

			pow = apow + btp -> pow;
			for (;;) {
				tp = *thookp;
				if ((tp EQ NULL) OR (tp -> pow < pow)) {
					tp2 = Zxyz_rpool_new_rterm (gip);
					tp2 -> coeff	= pp;
					tp2 -> next	= tp;
					tp2 -> pow	= pow;
					*thookp = tp2;
					thookp = &(tp2 -> next);
					break;
				}
				if (tp -> pow EQ pow) {
					pp2 = Zxyz_rpoly_add (tp -> coeff,
							      pp,
							      gip);
					Zxyz_rpoly_free (pp, gip);
					Zxyz_rpoly_free (tp -> coeff, gip);
					tp -> coeff = pp2;
					if (pp2 EQ NULL) {
						*thookp = tp -> next;
						Zxyz_rpool_free_rterm (tp, gip);
					}
					break;
				}
				thookp = &(tp -> next);
			}
		}
	}
	pp = Zxyz_rpoly_finish_terms (av, terms, gip);
	return (pp);
}

/*
 * Multiply the given recursive polynomial by the given integer.
 */

	static
	struct RPoly *
Zxyz_rpoly_mul_z (

const struct RPoly *	a,	/* IN: polynomial to multiply */
mpz_srcptr		zval,	/* IN: integer value to multiply by */
struct GInfo *		gip	/* IN/OUT: global info */
)
{
struct RPoly *	result;
struct RTerm *	terms;
struct RTerm **	thookp;
struct RTerm *	tp;
struct RTerm *	tp2;
mpz_t		TEMP;

	if ((a EQ NULL) OR (mpz_sgn (zval) EQ 0)) return (NULL);
	if (a -> var < 0) {
		mpz_init (TEMP);
		mpz_mul (TEMP, a -> u.zval, zval);
		result = Zxyz_rpoly_new_z (TEMP, gip);
		mpz_clear (TEMP);
	}
	else {
		terms = NULL;
		thookp = &terms;
		for (tp = a -> u.terms; tp NE NULL; tp = tp -> next) {
			result = Zxyz_rpoly_mul_z (tp -> coeff, zval, gip);
			if (result NE NULL) {
				tp2 = Zxyz_rpool_new_rterm (gip);
				tp2 -> coeff	= result;
				tp2 -> next	= NULL;
				tp2 -> pow	= tp -> pow;
				*thookp = tp2;
				thookp = &(tp2 -> next);
			}
		}
		result = Zxyz_rpoly_finish_terms (a -> var, terms, gip);
	}
	return (result);
}

/*
 * Multiply the given recursive polynomial by the given signed integer value.
 */

	static
	struct RPoly *
Zxyz_rpoly_mul_si (

const struct RPoly *	a,	/* IN: polynomial to multiply */
long			zval,	/* IN: integer value to multiply by */
struct GInfo *		gip	/* IN/OUT: global info */
)
{
struct RPoly *	result;
struct RTerm *	terms;
struct RTerm **	thookp;
struct RTerm *	tp;
struct RTerm *	tp2;
mpz_t		TEMP;

	if ((a EQ NULL) OR (zval EQ 0)) return (NULL);
	if (a -> var < 0) {
		mpz_init (TEMP);
		mpz_mul_si (TEMP, a -> u.zval, zval);
		result = Zxyz_rpoly_new_z (TEMP, gip);
		mpz_clear (TEMP);
	}
	else {
		terms = NULL;
		thookp = &terms;
		for (tp = a -> u.terms; tp NE NULL; tp = tp -> next) {
			result = Zxyz_rpoly_mul_si (tp -> coeff, zval, gip);
			if (result NE NULL) {
				tp2 = Zxyz_rpool_new_rterm (gip);
				tp2 -> coeff	= result;
				tp2 -> next	= NULL;
				tp2 -> pow	= tp -> pow;
				*thookp = tp2;
				thookp = &(tp2 -> next);
			}
		}
		result = Zxyz_rpoly_finish_terms (a -> var, terms, gip);
	}
	return (result);
}

/*
 * Create a recursive copy of the given recursive polynomial.
 * The new copy is allocated from the global pool.
 */

	static
	struct RPoly *
Zxyz_rpoly_dup (

const struct RPoly *	pp,		/* IN: polynomial to copy */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
struct RTerm *		terms;
struct RTerm **		thookp;
struct RTerm *		tp;
struct RTerm *		tp2;
struct RPoly *		pp2;

	if (pp EQ NULL) return (NULL);

	if (pp -> var < 0) {
		return (Zxyz_rpoly_new_z (pp -> u.zval, gip));
	}

	terms = NULL;
	thookp = &terms;
	for (tp = pp -> u.terms; tp NE NULL; tp = tp -> next) {
		tp2 = Zxyz_rpool_new_rterm (gip);
		tp2 -> coeff	= Zxyz_rpoly_dup (tp -> coeff, gip);
		tp2 -> next	= NULL;
		tp2 -> pow	= tp -> pow;
		*thookp = tp2;
		thookp = &(tp2 -> next);
	}

	pp2 = Zxyz_rpoly_finish_terms (pp -> var, terms, gip);

	return (pp2);
}

/*
 * Free the given recursive polynomial to the global pool.
 */

	static
	void
Zxyz_rpoly_free (

struct RPoly *		pp,		/* IN: polynomial to free */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
	if (pp EQ NULL) return;
	if (pp EQ DIVISION_NOT_EXACT) return;

	if (pp -> var < 0) {
		calib_free_Z_vector (pp -> u.zval, 1);
		pp -> u.zval = NULL;
	}
	else {
		Zxyz_rterms_free (pp -> u.terms, gip);
		pp -> u.terms = NULL;
	}
	Zxyz_rpool_free_rpoly (pp, gip);
}

/*
 * Free the given recursive term list to the global pool.
 */

	static
	void
Zxyz_rterms_free (

struct RTerm *	tp,		/* IN: list of terms to free */
struct GInfo *	gip		/* IN/OUT: global info */
)
{
struct RTerm *	tp2;

	while (tp NE NULL) {
		tp2 = tp -> next;
		Zxyz_rpoly_free (tp -> coeff, gip);
		tp -> coeff	= NULL;
		tp -> next	= NULL;
		Zxyz_rpool_free_rterm (tp, gip);
		tp = tp2;
	}
}

/*
 * Allocate a new RPoly that represents the given integer value.
 */

	static
	struct RPoly *
Zxyz_rpoly_new_z (

mpz_srcptr		zval,		/* IN: integer value */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
mpz_ptr			zp;
struct RPoly *		pp;

	CALIB_FATAL_ERROR_IF ((zval EQ NULL) OR (gip EQ NULL));

	pp = NULL;
	if (mpz_sgn (zval) NE 0) {
		zp = calib_new_Z_vector (1);
		mpz_set (zp, zval);

		pp = Zxyz_rpool_new_rpoly (gip);
		pp -> var	= -1;
		pp -> u.zval	= zp;
	}

	return (pp);
}

/*
 * Allocate a new RPoly that represents the given signed integer value.
 *
 * NOTE: This routine NEVER returns a NULL pointer when given the
 *	 value of zero!  This behavior is intentional in that it
 *	 permits the creation of "place holder" elements, e.g.,
 *	 that can be added into.
 */

	static
	struct RPoly *
Zxyz_rpoly_new_si (

long		zval,		/* IN: integer value */
struct GInfo *	gip		/* IN/OUT: global info */
)
{
mpz_ptr			zp;
struct RPoly *		pp;

	CALIB_FATAL_ERROR_IF (gip EQ NULL);

	zp = calib_new_Z_vector (1);
	mpz_set_si (zp, zval);

	pp = Zxyz_rpool_new_rpoly (gip);
	pp -> var	= -1;
	pp -> u.zval	= zp;

	return (pp);
}

/*
 * Print the given "flat" polynomial, terminated by a newline.
 */

	static
	void
Zxyz_print (

const struct calib_Zxyz_obj *	op	/* IN: polynomial to print */
)
{
	Zxyz_print_nnl (op);
	printf ("\n");
}

/*
 * Print the given "flat" polynomial, with No NewLine.
 */

	static
	void
Zxyz_print_nnl (

const struct calib_Zxyz_obj *	op	/* IN: polynomial to print */
)
{
int				i, j, k, nvars, nterms;
const struct calib_Zxyz_dom *	dom;
const int *			ip;
mpz_srcptr			zp;
char **				varlist;
const char *			sep;
mpz_t				cmag;

	if (op EQ NULL) {
		printf ("{null pointer}\n");
		return;
	}

	nterms	= op -> nterms;

	if (nterms <= 0) {
		if (nterms < 0) {
			printf ("{nterms=%d}", nterms);
		}
		else {
			printf ("0");
		}
		return;
	}

	dom	= op -> dom;
	ip	= op -> pow;
	zp	= op -> coeff;
	nvars	= dom -> nvars;
	varlist	= dom -> vars;

	mpz_init (cmag);

	/* Display first term. */
	if (mpz_sgn (zp) < 0) {
		printf (" -");
	}
	/* Compute total degree of this term. */
	for (k = 0, j = 0; j < nvars; j++) {
		k += ip [j];
	}
	mpz_abs (cmag, zp);
	/* Need to print coefficient if no vars, or if not 1. */
	sep = " ";
	if ((k EQ 0) OR (mpz_cmp_ui (cmag, 1) NE 0)) {
		printf (" ");
		mpz_out_str (stdout, 10, cmag);
		sep = "*";
	}
	for (j = 0; j < nvars; j++) {
		k = ip [j];
		if (k <= 0) continue;
		if (k EQ 1) {
			printf ("%s%s", sep, varlist [j]);
		}
		else {
			printf ("%s%s^%d", sep, varlist [j], k);
		}
		sep = "*";
	}
	ip += nvars;
	++zp;
	/* Display other terms, if any. */
	for (i = 1; i < nterms; i++) {
		if (mpz_sgn (zp) < 0) {
			printf (" -");
		}
		else {
			printf (" +");
		}
		/* Compute total degree of this term. */
		for (k = 0, j = 0; j < nvars; j++) {
			k += ip [j];
		}
		mpz_abs (cmag, zp);
		/* Need to print coefficient if no vars, or if not 1. */
		sep = " ";
		if ((k EQ 0) OR (mpz_cmp_ui (cmag, 1) NE 0)) {
			printf (" ");
			mpz_out_str (stdout, 10, cmag);
			sep = "*";
		}
		for (j = 0; j < nvars; j++) {
			k = ip [j];
			if (k <= 0) continue;
			if (k EQ 1) {
				printf ("%s%s", sep, varlist [j]);
			}
			else {
				printf ("%s%s^%d", sep, varlist [j], k);
			}
			sep = "*";
		}
		ip += nvars;
		++zp;
	}
	mpz_clear (cmag);
}

/*
 * Print the given recursive polynomial, terminated by a newline.
 * This is just for debugging.  The variables are named x0, x1, etc.
 */

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wunused-function"

	static
	void
Zxyz_rpoly_print (

const struct RPoly *	pp		/* IN: polynomial to print */
)
{
	if (pp EQ NULL) {
		printf ("0\n");
	}
	else {
		Zxyz_rpoly_print_recurse (pp);
		printf ("\n");
	}
}

#pragma GCC diagnostic pop

/*
 * Recursively print the given recursive polynomial.
 * The format is useful for debugging, but definitely not nice.
 */

	static
	void
Zxyz_rpoly_print_recurse (

const struct RPoly *	pp		/* IN: polynomial to print */
)
{
int		j, var;
struct RTerm *	tp;

	if (pp EQ NULL) {
		printf ("NULL");
		return;
	}

	var = pp -> var;
	if (var < 0) {
		mpz_out_str (stdout, 10, pp -> u.zval);
		return;
	}
	tp = pp -> u.terms;
	if (tp EQ NULL) {
		printf ("{var x%d, but no terms}", var);
	}
	else {
		printf ("(");
		Zxyz_rpoly_print_recurse (tp -> coeff);
		printf (")");
		j = tp -> pow;
		if (j EQ 0) {
		}
		else if (j EQ 1) {
			printf (" x%d", var);
		}
		else {
			printf (" x%d^%d", var, j);
		}
		tp = tp -> next;
		for (; tp NE NULL; tp = tp -> next) {
			printf (" + (");
			Zxyz_rpoly_print_recurse (tp -> coeff);
			printf (")");
			j = tp -> pow;
			if (j EQ 0) {
			}
			else if (j EQ 1) {
				printf (" x%d", var);
			}
			else {
				printf (" x%d^%d", var, j);
			}
		}
	}
}

/*
 * Return the nth row of the binomial coefficients, i.e., the returned
 * array is such that array [j] = (n choose j).
 */

	static
	mpz_srcptr
Zxyz_binomial_row (

int			n,		/* IN: which row to get */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
int		i, k;
mpz_ptr *	binom;
mpz_ptr *	new_ptrs;
mpz_ptr		prev_row;
mpz_ptr		new_row;

	CALIB_FATAL_ERROR_IF (gip EQ NULL);

	k	= gip -> nbinom;
	binom	= gip -> binom;

	if (n >= k) {
		new_ptrs = CALIB_NEWA (n + 1, mpz_ptr);
		memcpy (new_ptrs, binom, k * sizeof (new_ptrs [0]));
		free (binom);
		binom = new_ptrs;
		gip -> nbinom = n + 1;
		gip -> binom = binom;

		for (; k <= n; k++) {
			new_row = calib_new_Z_vector (k + 1);
			if (k <= 1) {
				for (i = 0; i <= k; i++) {
					mpz_set_ui (&new_row [i], 1);
				}
			}
			else {
				prev_row = binom [k - 1];
				mpz_set_ui (&new_row [0], 1);
				for (i = 1; i < k; i++) {
					mpz_add (&new_row [i],
						 &prev_row [i - 1],
						 &prev_row [i]);
				}
				mpz_set_ui (&new_row [k], 1);
			}
			binom [k] = new_row;
		}
	}

	return (binom [n]);
}

/*
 * Initialize the given global info structure.
 */

	static
	void
Zxyz_ginfo_init (

struct GInfo *		gip		/* IN/OUT: global info to init */
)
{
	CALIB_FATAL_ERROR_IF (gip EQ NULL);

	memset (gip, 0, sizeof (*gip));

	Zxyz_rpool_init (&(gip -> pool));
	calib_random_init (&(gip -> rng));
}

/*
 * Clear out the given global info structure.
 */

	static
	void
Zxyz_ginfo_clear (

struct GInfo *		gip		/* IN/OUT: global info to clear */
)
{
int		i, n;
mpz_ptr *	binom;

	CALIB_FATAL_ERROR_IF (gip EQ NULL);

	/* Clear any binomial rows we computed. */
	n	= gip -> nbinom;
	binom	= gip -> binom;
	if (binom NE NULL) {
		CALIB_FATAL_ERROR_IF (n <= 0);
		for (i = 0; i < n; i++) {
			calib_free_Z_vector (binom [i], i + 1);
		}
		free (binom);
	}
	else {
		CALIB_FATAL_ERROR_IF (n NE 0);
	}

	Zxyz_rpool_clear (&(gip -> pool));
	/* No need to clear out the random number generator. */
}

/*
 * Initialize the given RPool.
 */

	static
	void
Zxyz_rpool_init (

struct RPool *		pool		/* IN: pool to initialize */
)
{
	memset (pool, 0, sizeof (*pool));
}

/*
 * Clear out the given pool.
 */

	static
	void
Zxyz_rpool_clear (

struct RPool *		pool		/* IN: pool to clear */
)
{
struct mblk *		bp;
struct mblk *		bp2;

	bp = pool -> blocks;
	memset (pool, 0, sizeof (*pool));
	while (bp NE NULL) {
		bp2 = bp -> next;
		bp -> next = NULL;
		Zxyz_put_mblk (bp);
		bp = bp2;
	}
}

/*
 * Allocate a new RPoly node from the global pool.
 */

	static
	struct RPoly *
Zxyz_rpool_new_rpoly (

struct GInfo *		gip		/* IN/OUT: global info */
)
{
int			nbytes;
struct RPool *		pool;
struct RPoly *		pp;
struct mblk *		mp;

	CALIB_FATAL_ERROR_IF (gip EQ NULL);

	pool = &(gip -> pool);
	pp = pool -> pfree;
	if (pp NE NULL) {
		pool -> pfree = pp -> u.next;
		pp -> u.next = NULL;
		return (pp);
	}

	nbytes = sizeof (*pp);
	if (pool -> nleft >= nbytes) {
		pp = (struct RPoly *) (pool -> allocp);
		pool -> allocp	+= nbytes;
		pool -> nleft	-= nbytes;
		pp -> u.next	= NULL;
		return (pp);
	}

	mp = Zxyz_get_mblk ();
	mp -> next = pool -> blocks;
	pool -> blocks	= mp;
	pool -> allocp	= &(mp -> bytes [0]);
	pool -> nleft	= sizeof (mp -> bytes);

	CALIB_FATAL_ERROR_IF (nbytes > pool -> nleft);

	pp = (struct RPoly *) (pool -> allocp);
	pool -> allocp	+= nbytes;
	pool -> nleft	-= nbytes;
	pp -> u.next = NULL;

	return (pp);
}

/*
 * Free an RPoly node.
 */

	static
	void
Zxyz_rpool_free_rpoly (

struct RPoly *		pp,		/* IN: RPoly node to free */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
	CALIB_FATAL_ERROR_IF ((pp EQ NULL) OR (gip EQ NULL));

	pp -> u.next = gip -> pool.pfree;
	gip -> pool.pfree = pp;
}

/*
 * Allocate a new RTerm node from the global pool.
 */

	static
	struct RTerm *
Zxyz_rpool_new_rterm (

struct GInfo *		gip		/* IN/OUT: global info */
)
{
int			nbytes;
struct RPool *		pool;
struct RTerm *		tp;
struct mblk *		mp;

	CALIB_FATAL_ERROR_IF (gip EQ NULL);

	pool = &(gip -> pool);
	tp = pool -> tfree;
	if (tp NE NULL) {
		pool -> tfree = tp -> next;
		tp -> next = NULL;
		return (tp);
	}

	nbytes = sizeof (*tp);
	if (pool -> nleft >= nbytes) {
		tp = (struct RTerm *) (pool -> allocp);
		pool -> allocp	+= nbytes;
		pool -> nleft	-= nbytes;
		tp -> next	= NULL;
		return (tp);
	}

	mp = Zxyz_get_mblk ();
	mp -> next = pool -> blocks;
	pool -> blocks	= mp;
	pool -> allocp	= &(mp -> bytes [0]);
	pool -> nleft	= sizeof (mp -> bytes);

	CALIB_FATAL_ERROR_IF (nbytes > pool -> nleft);

	tp = (struct RTerm *) (pool -> allocp);
	pool -> allocp	+= nbytes;
	pool -> nleft	-= nbytes;
	tp -> next = NULL;

	return (tp);
}

/*
 * Free an RTerm node.
 */

	static
	void
Zxyz_rpool_free_rterm (

struct RTerm *		tp,		/* IN: RTerm node to free */
struct GInfo *		gip		/* IN/OUT: global info */
)
{
	CALIB_FATAL_ERROR_IF ((tp EQ NULL) OR (gip EQ NULL));

	tp -> next = gip -> pool.tfree;
	gip -> pool.tfree = tp;
}

/*
 * Return a new mblk node.
 */

	static
	struct mblk *
Zxyz_get_mblk (void)

{
struct mblk *	mp;

	mp = mblk_freelist;
	if (mp EQ NULL) {
		mp = CALIB_NEW (struct mblk);

		Zxyz_establish_cleanup_handler ();
	}
	else {
		mblk_freelist = mp -> next;
		mp -> next = NULL;
	}
	return (mp);
}

/*
 * Free the given mblk node.
 */

	static
	void
Zxyz_put_mblk (

struct mblk *	mp		/* IN: mblk to put back on freelist */
)
{
	CALIB_FATAL_ERROR_IF ((mp EQ NULL) OR (mp -> next NE NULL));

	mp -> next = mblk_freelist;
	mblk_freelist = mp;
}

/*
 * Establish our cleanup handler, if we haven't done so already.
 */

	static
	void
Zxyz_establish_cleanup_handler (void)

{
	if (NOT cleanup_handler_established) {
		/* Add a handler allowing these to be cleaned up at end. */
		_calib_add_cleanup_handler (Zxyz_cleanup_handler);

		cleanup_handler_established = TRUE;
	}
}

/*
 * Free up the all of the paccum and mblk objects.
 */

	static
	void
Zxyz_cleanup_handler (void)

{
struct paccum *		pap;
struct mblk *		mp;

	/* Free up all of the paccum objects. */
	for (;;) {
		pap = paccum_freelist;
		if (pap EQ NULL) break;

		paccum_freelist = pap -> next;
		pap -> next = NULL;

		/* Free up all pterm's (and their associated GMP	*/
		/* coefficients), and put all its mblk's back onto	*/
		/* the global mblk freelist.				*/
		Zxyz_reset_paccum (pap);

		if (pap -> link NE NULL) {
			free (pap -> link - 1);
		}
		if (pap -> bucket NE NULL) {
			free (pap -> bucket);
		}

		memset (pap, 0, sizeof (*pap));
		free (pap);
	}

	/* Free up all of the mblk objects. */
	for (;;) {
		mp = mblk_freelist;
		if (mp EQ NULL) break;

		mblk_freelist = mp -> next;
		mp -> next = NULL;

		memset (mp, 0, sizeof (*mp));
		free (mp);
	}

	cleanup_handler_established = FALSE;

	/* We should now be able to start all over again. */
}

/*
 * Prototype Zxyz domain.
 */

static const struct calib_Zxyz_dom	_Zxyz_dom = {
	.nvars			= 0,
	.vars			= NULL,

	.init			= Zxyz_init,
	.init_si		= Zxyz_init_si,
	.alloc			= Zxyz_alloc,
	.clear			= Zxyz_clear,
	.set			= Zxyz_set,
	.set_si			= Zxyz_set_si,
	.set_z			= Zxyz_set_z,
	.set_var_power		= Zxyz_set_var_power,
	.add			= Zxyz_add,
	.add_n			= Zxyz_add_n,
	.sub			= Zxyz_sub,
	.neg			= Zxyz_neg,
	.mul			= Zxyz_mul,
	.mul_z			= Zxyz_mul_z,
	.ipow			= Zxyz_ipow,
	.dup			= Zxyz_dup,
	.free			= Zxyz_free,
	.eval			= Zxyz_eval,
	.eval_var_subset	= Zxyz_eval_var_subset,
	.div			= Zxyz_div,
	.div_z_exact		= Zxyz_div_z_exact,
	.div_remove		= Zxyz_div_remove,
	.gcd			= Zxyz_gcd,
	.gcd_n			= Zxyz_gcd_n,
	.extgcd			= Zxyz_extgcd,
	.z_content		= Zxyz_z_content,
	.strip_z_content	= Zxyz_strip_z_content,
	.prim_part		= Zxyz_prim_part,
	.resultant		= Zxyz_resultant,
	.resultant_old		= Zxyz_resultant_old,
	.discriminant		= Zxyz_discriminant,
	.cvZpxyz		= Zxyz_cvZpxyz,
	.factor			= Zxyz_factor,
	.sqf_factor		= Zxyz_sqf_factor,
	.free_factors		= Zxyz_free_factors,
	.map_to_subring		= Zxyz_map_to_subring,
	.copy_into_superring	= Zxyz_copy_into_superring,
	.convert_with_varmap	= Zxyz_convert_with_varmap,
	.copy_from_Zx		= Zxyz_copy_from_Zx,
	.copy_into_Zx		= Zxyz_copy_into_Zx,
	.copy_from_Zpx		= Zxyz_copy_from_Zpx,
	.copy_into_Zpx		= Zxyz_copy_into_Zpx,
	.copy_from_Qax		= Zxyz_copy_from_Qax,
	.copy_into_Qax		= Zxyz_copy_into_Qax,
	.add_vars		= Zxyz_add_vars,
	.zerop			= Zxyz_zerop,
	.onep			= Zxyz_onep,
	.set_genrep		= Zxyz_set_genrep,
	.to_genrep		= Zxyz_to_genrep,
	.factors_to_genrep	= Zxyz_factors_to_genrep,
	.print_maxima		= Zxyz_print,
	.print_maxima_nnl	= Zxyz_print_nnl,
	.lookup_var		= Zxyz_lookup_var,
};

/*
 * Construct the ring of polynomials Z[x,y,z], for a given list of vars.
 */

	struct calib_Zxyz_dom *
calib_make_Zxyz_dom (

int			nvars,	/* IN: Number of variables */
const char * const *	vars	/* IN: Name of each variable */
)
{
int			i, j;
const char *		vi;
const char *		vj;
char **			vlist;
struct calib_Zxyz_dom *	dom;

	CALIB_FATAL_ERROR_IF ((nvars <= 0) OR (vars EQ NULL));

	/* Validate the variable names. */
	for (i = 0; i < nvars; i++) {
		vi = vars [i];
		CALIB_FATAL_ERROR_IF (vi EQ NULL);
		CALIB_FATAL_ERROR_IF (*vi EQ '\0');
		/* Verify variable names are distinct. */
		for (j = i + 1; j < nvars; j++) {
			vj = vars [j];
			CALIB_FATAL_ERROR_IF (vj EQ NULL);
			CALIB_FATAL_ERROR_IF (strcmp (vi, vj) EQ 0);
		}
	}

	vlist = CALIB_NEWA (nvars + 1, char *);
	for (i = 0; i < nvars; i++) {
		vlist [i] = strdup (vars [i]);
	}
	vlist [i] = NULL;

	dom = CALIB_NEW (struct calib_Zxyz_dom);
	memcpy (dom, &_Zxyz_dom, sizeof (*dom));

	dom -> nvars	= nvars;
	dom -> vars	= vlist;

	return (dom);
}

/*
 * Free up the given Z[x,y,z] ring description.
 */

	void
calib_free_Zxyz_dom (

const struct calib_Zxyz_dom *	dom	/* IN: Z[x,y,z] ring description */
)
{
int			i, nvars;
struct calib_Zxyz_dom *	ldom;
char *			vi;
char **			vars;

	if (dom NE NULL) {
		nvars	= dom -> nvars;
		vars	= dom -> vars;
		if (vars NE NULL) {
			for (i = 0; i < nvars; i++) {
				vi = vars [i];
				if (vi NE NULL) {
					free (vi);
					vars [i] = NULL;
				}
			}
			free (vars);
		}
		ldom = (struct calib_Zxyz_dom *) dom;
		memset (ldom, 0, sizeof (*ldom));
		free (ldom);
	}
}

/*
 * Create a new domain Y having the given additional variables over
 * those in the given domain X.
 */

	static
	struct calib_Zxyz_dom *
Zxyz_add_vars (

const struct calib_Zxyz_dom *	dom,	/* IN: Source domain */
int				nvars,	/* IN: Number of new vars */
const char * const *		newvars	/* IN: Array of new variable names */
)
{
int			i, j, orig_nvars, new_nvars;
char **			orig_vars;
char **			new_vars;
struct calib_Zxyz_dom *	dst;

	CALIB_FATAL_ERROR_IF (dom EQ NULL);
	CALIB_FATAL_ERROR_IF (nvars <= 0);
	CALIB_FATAL_ERROR_IF (newvars EQ NULL);

	orig_nvars	= dom -> nvars;
	orig_vars	= dom -> vars;

	new_nvars	= orig_nvars + nvars;
	new_vars	= CALIB_NEWA (new_nvars + 1, char *);

	for (i = 0; i < orig_nvars; i++) {
		new_vars [i]	= strdup (orig_vars [i]);
	}
	for (j = 0; j < nvars; j++) {
		new_vars [i++] = strdup (newvars [j]);
	}
	new_vars [i] = NULL;
	CALIB_FATAL_ERROR_IF (i NE new_nvars);

	dst = CALIB_NEW (struct calib_Zxyz_dom);
	memcpy (dst, &_Zxyz_dom, sizeof (*dst));

	dst -> nvars	= nvars;
	dst -> vars	= new_vars;

	return (dst);
}
